//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: wadcheck.c
// Revision 2.1.0 build 62 03/Oct/2006 gmj
// Revision 2.1.0 build 63 03/Oct/2006 gmj (b211-0)
//
// DESCRIPTION: file routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include	"..\Risen3D\include\lzss.h"
#include	<stdio.h>
#include	<io.h>
#include	<stdlib.h>
#include	<string.h>
#include	<direct.h>

#include "wadcheck.h"

// MACROS ------------------------------------------------------------------

//#define		true	1
//#define		false	0

// Max entries in directory
#define MAX_WADNAMES		4096	// And if that's not enough bad luck!

// bits 17-20 backward compatibility flags
#define	VERSION_MASK			0xFFFFFFF0

// The version must not include any compatibility flags
#define	SAVEGAME_VER			0xF3D35300	// v220-35 sub v3

// TYPES -------------------------------------------------------------------

typedef struct
{
	int		magic;
	int		gamemode;		
	int		leveltime;
	uint	gameid;
	char	description[SAVESTRINGSIZE];
	byte	players[MAXPLAYERS];
	byte	skill;
	byte	episode;
	byte	map;
	byte	deathmatch;
	byte	nomonsters;
	byte	respawn;
	byte	cfx;
	byte	nmodels;
} saveheader_t;

typedef struct
{
	char pwadname[MAXPWADNAME];
} savepwadname_t;

char *level_lumps[] =
{
	"LINEDEFS", "SIDEDEFS", "VERTEXES", NULL
};

char *gl_lumps[] =
{
	"GL_VERT", "GL_SEGS", "GL_SSECT", "GL_NODES", /*"GL_PVS",*/ NULL
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static LZFILE *savefile;

static saveheader_t hdr;

static int file_version;
static int  num_wadnames;

//  LISTWADS =================================================

void CloseSavefile()
{
	if (savefile)
	{
		lzClose(savefile);
		savefile = NULL;
	}
}

long SV_ReadLong()
{
	return lzGetL(savefile);
}

void SV_GetHdrDescription(char *filename, char *str)
{
	if ((savefile = lzOpen(filename, "rp")) != NULL)
	{
		file_version = SV_ReadLong();	// Version

		if ((file_version & VERSION_MASK) == SAVEGAME_VER)
		{
			// Read the header.
			lzRead(&hdr, sizeof(hdr), savefile);
			strncpy(str, hdr.description, SAVESTRINGSIZE);
		}
		CloseSavefile();
	}
}

void AddFile(char *name, char *wadnames[], int type)
{
	int		len = strlen(name);

	if (num_wadnames >= MAX_WADNAMES) return;

	if (type)	// .wad
	{
		char	header[5];
		FILE	*temp = fopen(name, "rb" );
		int		hlen = temp ? fread(header, sizeof(char), 4, temp) : 0;

		if (temp) fclose(temp);

		if (hlen != 4 || strnicmp(header, "PWAD", 4))
			return;
	}

	wadnames[num_wadnames] = calloc(len + 1, 1);

	strcpy(wadnames[num_wadnames++], name);
}

int __cdecl compWadname( const void *p1, const void *p2 )
{
	return _stricmp( * ( char** ) p1, * ( char** ) p2 );
}

int Dir_ChExt(char *fname, char *ext)
{
	char file_ext[256];

	_splitpath(fname, NULL, NULL, NULL, file_ext);

	return !stricmp(file_ext, ext);
}

int ListWads(char *wadspath)
{
	char	*wadnames[MAX_WADNAMES];
	struct	_finddata_t g_file;
	long	hFile;
	int		i;

	if (_chdir(wadspath)) return 1;

	// 1) Find .wad files in current directory
	// b211-0 revised
	if( (hFile = _findfirst( "*.wad", &g_file )) != -1L )
	{
		do
		{
			if (Dir_ChExt(g_file.name, ".WAD"))
				AddFile(g_file.name, wadnames, 1);
		}
		while( _findnext( hFile, &g_file ) == 0 );
		
		_findclose( hFile );
	}
	
	// 2) Find .deh files in current directory
	// b211-0 revised
	if( (hFile = _findfirst( "*.deh", &g_file )) != -1L )
	{
		do
		{
			if (Dir_ChExt(g_file.name, ".DEH"))
				AddFile(g_file.name, wadnames, 0);
		}
		while( _findnext( hFile, &g_file ) == 0 );
		
		_findclose( hFile );
	}

	_chdir(callpath);

	fprintf(stdout, "%i\n", num_wadnames);

	if (num_wadnames)
	{
		if (num_wadnames > 1)
			qsort((void*)wadnames, (size_t)num_wadnames, sizeof(char*), compWadname );

		for (i = 0; i < num_wadnames; i++)
			fprintf(stdout, "%s\n", wadnames[i]);

		for (i = 0; i < num_wadnames; i++)
			free(wadnames[i]);
	}
	return 0;
}

//  CHECKWADS =================================================

byte	levelnum[100];

static int W_CheckWAD(FILE *f, int  *count, int	 *glcount)
{
	struct wad_header_t
	{
		char		type[4];
		int			num_entries;
		unsigned int dir_start;

	} header;

	struct lump_header_t
	{
		unsigned int start, length;
		char		name[8];

	} lumpinf;

	char	buff[10];	
	char	levelname[10];
	char	priorname[10];
	char	gl_mapname[10];

	int		i, p, num, num_entries;
	int		glmapcount = 0, mapcount = 0;
	int		wadtype;

	//-------------------------

	memset(priorname, 0, sizeof(priorname));
	memset(levelname, 0, sizeof(levelname));
	memset(buff, 0, sizeof(buff));

	*count   = 0;
	*glcount = 0;

	memset(levelnum, 0, 100);
	strcpy(gl_mapname, "GL_");

	fseek(f, 0, SEEK_SET);

	if ((i = fread(&header, sizeof(header), 1, f)) != 1) return 0;

	if (!strncmp(header.type, "IWAD", 4))
		wadtype = 1;
	else
	if (!strncmp(header.type, "PWAD", 4))
		wadtype = 2;
	else
		return 0;
	
	if  (fseek(f, header.dir_start, SEEK_SET) != 0)
		return 0;

	num_entries = header.num_entries;
	num = -1;

	for(p = 0; p < num_entries; p++)
	{
		if ((i = fread(&lumpinf, sizeof(lumpinf), 1, f)) != 1)
		{
			// Values returned as wad is usable even if faulty
			*count   = mapcount;
			*glcount = glmapcount;
			return wadtype;
		}

		memcpy(buff, lumpinf.name, 8);

		if (!strnicmp(buff,"THINGS", 7))
		{   
			//The level name always immediately precedes THINGS
			memcpy(levelname, priorname, 8);

            if ( (!levelname[4]
				&& levelname[0] == 'E'
				&& levelname[2] == 'M' 
                && CharIsNum(levelname[1])
				&& CharIsNum(levelname[3])) 
                ||
				( !levelname[5]
				&& !strncmp(levelname, "MAP", 3) 
                && CharIsNum(levelname[3])
				&& CharIsNum(levelname[4])) )
			{
				int c = 0, nametype = 0;
                int found = true;

			    if (levelname[0] == 'M')  // MAPxx
				{
					num = (levelname[3] - '0')*10 + (levelname[4] - '0');
					nametype = 1;
					strncpy(&gl_mapname[3], levelname, 5);
					gl_mapname[8] = 0;
                }
				else  // ExMx
				{
					num = (levelname[1] - '0')*10 + (levelname[3] - '0');
					nametype = 2;
					strncpy(&gl_mapname[3], levelname, 4);
					gl_mapname[7] = 0;
                }

     			if (num < 1 || num > 99) continue;	// bad wad or entry

				for(;p < num_entries && level_lumps[c]; p++)
				{
					if ( (i = fread(&lumpinf, sizeof(lumpinf), 1, f)) != 1)
					{
						found = false;
						break;		// bad entry
					}

					memcpy(buff, lumpinf.name, 8);

					if (strcmp(buff, level_lumps[c++]))
					{
						found = false;
						break;		// required match not found
					}
				}

				if (found && !(levelnum[num] & nametype))
				{
					mapcount++;
					levelnum[num] |= nametype;
				}

				continue;
			}
		}

		if (!strcmp(buff, gl_mapname))
		{
			int			c = 0;
			int			found = true;
			unsigned	start = 0;
			char		bf[4];
			fpos_t		pos;

			gl_mapname[4] = 0;	// Clear for next set of entries
					
			for(;p < num_entries && gl_lumps[c]; p++)
			{
				if ((i = fread(&lumpinf, sizeof(lumpinf), 1, f)) != 1)
					return 0;

				memcpy(buff, lumpinf.name, 8);

				if (strcmp(buff, gl_lumps[c]))
				{
					found = false;
					break;
				}

				if (!(c++)) start = lumpinf.start; // "GL_VERT"
			}

			if (!found) continue;

			if (fgetpos(f, &pos)			||	// Store current file position
				fseek(f, start, SEEK_SET)	||	// Set it to start of lump
				!fread(&bf, 4, 1, f)		||	// Read first 4 bytes
				fsetpos(f, &pos))				// Restore position
				return 0;

			if (!strncmp(bf, "gNd2", 4) ||
				!strncmp(bf, "gNd3", 4) ||
				!strncmp(bf, "gNd4", 4) ||
				!strncmp(bf, "gNd5", 4))
			{
				if (!(levelnum[num] & 4))
				{
					glmapcount++;
					levelnum[num] |= 4;
				}
			}
		}
		memcpy(priorname,lumpinf.name,8);
	}

	*count   = mapcount;
	*glcount = glmapcount;

	return wadtype;
}


int CheckWad(char *fname, char *wadspath)
{
	int		i, count, res;
	int		glnodes;
	int		wadtype;

	FILE	*fhandle = fopen(wadspath, "rb");

	if (!fhandle) return 1;

	// wadtype  = 0 - neither an IWAD or a PWAD
	//			= 1 - IWAD
	//			= 2 - PWAD
	wadtype = W_CheckWAD (fhandle, &count, &glnodes);

	fclose(fhandle);
	fhandle = NULL;

	if (!wadtype) return 1;	// Neither an IWAD or a PWAD or faulty
	if (!fname  ) return 2;	// No valid ouput path

	if (strcmp(fname, "stdout")) // I.e. if not stdout
	{
		char buff[MAXPATHNAMESIZE];

		DefFullPathBin(fname, buff);

		fhandle = fopen(buff, "w");

		if (!fhandle)
		{
			fprintf(stdout, "%s could not be opened\n", buff);
			return 0;
		}
	}

	res = count > 0 && glnodes == count;

	if (fhandle)
	{
		fprintf(fhandle, "%s\n", wadtype == 1 ? "IWAD" : "PWAD");
		fprintf(fhandle, "NUMLEVELS %i\nGLNODES %i\n", count, res);
	}
	else
	{
		fprintf(stdout, "%s\n", wadtype == 1 ? "IWAD" : "PWAD");
		fprintf(stdout, "NUMLEVELS %i\nGLNODES %i\n", count, res);
	}

	if (count)
	{
		for (i = 0; i < 100; i++)
		{
			if (levelnum[i] & 2)
			{
				if (fhandle)
					fprintf(fhandle, "E%iM%i\n", i/10, i % 10);
				else
					fprintf(stdout , "E%iM%i\n", i/10, i % 10);
			}
		}

		for (i = 0; i < 100; i++)
		{
			if (levelnum[i] & 1)
			{
				if (fhandle)
					fprintf(fhandle, "MAP%.2i\n", i);
				else
					fprintf(stdout , "MAP%.2i\n", i);
			}
		}
	}

	if (fhandle)
	{
		fflush(fhandle);
		fclose(fhandle);
	}
	return 0;	// Everything's OK
}

// GETSAVEHDRS  =================================================

int GetSaveHeaders(BOOL usefile, char *savefilepath)
{
	char fname[MAXPATHNAMESIZE];
	char hdrnames[NUMSAVESLOTS][SAVESTRINGSIZE];
	int i, ret = 1;

	FILE	*fhandle = NULL;

	if (usefile)
	{
		if (callpath[strlen(callpath) - 1] != '\\')
			sprintf(fname, "%s%s", callpath, "\\r3dshrs.out");
		else
			sprintf(fname, "%s%s", callpath, "r3dshrs.out");

		fhandle = fopen(fname, "w");
		if (!fhandle) return 1;
	}

	if ( (_chdir(savefilepath)) == 0 )
	{
		for (i = 0; i < 10; i++)
		{
			char filename[32];

			hdrnames[i][0] = 0;

			sprintf(filename, "Risen3DSav%.1i.r3dsg", i);
			if (_access(filename, 4) != -1)
				SV_GetHdrDescription(filename, hdrnames[i]);
		}

		_chdir(callpath);

		for (i = 0; i < NUMSAVESLOTS; i++)
			fprintf(fhandle == NULL ? stdout : fhandle, "%i %s\n", i, hdrnames[i]);

		ret = 0;
	}
	if (fhandle) fclose(fhandle);
	return ret;
}
