//-----------------------------------------------------------------------------
//
// ACFNLINK.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __ACFNLINK_H__
#define __ACFNLINK_H__

#include "dd_share.h"
#include "doom/mbf_player.h"

// NOTE: the following defines must match the functions
//       set in the actionlinks array,

typedef enum
{
	nmNullHandler,		// 0 R3D - must be 0
// Player sprite references
	nmLight0,
	nmWeaponReady,
	nmLower,
	nmRaise,
	nmPunch,
	nmReFire,
	nmFirePistol,
	nmLight1,
	nmFireShotgun,
	nmLight2,			// 10
	nmFireShotgun2,
	nmCheckReload,
	nmOpenShotgun2,
	nmLoadShotgun2,
	nmCloseShotgun2,
	nmFireCGun,
	nmGunFlash,
	nmFireMissile,
	nmSaw,
	nmFirePlasma,		// 20
	nmBFGsound,
	nmFireBFG,
	nmWeapSound,

// Enemy references
	nmBFGSpray,
	nmExplode,
	nmPain,
	nmPlayerScream,
	nmFall,
	nmXScream,
	nmLook,				// 30
	nmChase,
	nmFaceTarget,
	nmPosAttack,
	nmScream,
	nmSPosAttack,
	nmVileChase,
	nmVileStart,
	nmVileTarget,
	nmVileAttack,
	nmStartFire,		// 40
	nmFire,
	nmFireCrackle,
	nmTracer,
	nmSkelWhoosh,
	nmSkelFist,
	nmSkelMissile,
	nmFatRaise,
	nmFatAttack1,
	nmFatAttack2,
	nmFatAttack3,		// 50
	nmBossDeath,
	nmCPosAttack,
	nmCPosRefire,
	nmTroopAttack,
	nmSargAttack,
	nmHeadAttack,
	nmBruisAttack,
	nmSkullAttack,
	nmMetal,
	nmSpidRefire,		// 60
	nmBabyMetal,
	nmBspiAttack,
	nmHoof,
	nmCyberAttack,
	nmPainAttack,
	nmPainDie,
	nmKeenDie,
	nmBrainPain,
	nmBrainScream,
	nmBrainDie,			// 70
	nmBrainAwake,
	nmBrainSpit,
	nmSpawnSound,
	nmSpawnFly,
	nmBrainExplode,

// killough
	nmDetonate,
	nmMushroom,
	nmDie,
	nmSpawn,
	nmTurn,				// 80
	nmFace,
	nmScratch,
	nmPlaySound,
	nmRandomJump,
	nmLineEffect,

// R3D specific
	nmdbAttack,
	nmEnd,

} nmActions_e;

typedef struct
{
	nmActions_e	ref;	// Numeric position in list
	char		*name;	// Name of the routine.
	void		(__cdecl *func)(struct mobj_s*);	// Pointer to mobj functions.
	void		(__cdecl *plfn)(mbf_player_t*, pspdef_t*);	// Pointer to the player function.
} actionlink_t;

extern actionlink_t actionlinks[];

void __cdecl A_NullHandler(mobj_t*);
// Player specific actions
void __cdecl A_Light0(mbf_player_t*, pspdef_t*);
void __cdecl A_WeaponReady(mbf_player_t*, pspdef_t*);
void __cdecl A_Lower(mbf_player_t*, pspdef_t*);
void __cdecl A_Raise(mbf_player_t*, pspdef_t*);
void __cdecl A_Punch(mbf_player_t*, pspdef_t*);
void __cdecl A_ReFire(mbf_player_t*, pspdef_t*);
void __cdecl A_FirePistol(mbf_player_t*, pspdef_t*);
void __cdecl A_Light1(mbf_player_t*, pspdef_t*);
void __cdecl A_FireShotgun(mbf_player_t*, pspdef_t*);
void __cdecl A_Light2(mbf_player_t*, pspdef_t*);
void __cdecl A_FireShotgun2(mbf_player_t*, pspdef_t*);
void __cdecl A_CheckReload(mbf_player_t*, pspdef_t*);
void __cdecl A_OpenShotgun2(mbf_player_t *player, pspdef_t *psp);
void __cdecl A_LoadShotgun2(mbf_player_t *player, pspdef_t *psp);
void __cdecl A_CloseShotgun2(mbf_player_t*, pspdef_t*);
void __cdecl A_FireCGun(mbf_player_t*, pspdef_t*);
void __cdecl A_GunFlash(mbf_player_t*, pspdef_t*);
void __cdecl A_FireMissile(mbf_player_t*, pspdef_t*);
void __cdecl A_Saw(mbf_player_t*, pspdef_t*);
void __cdecl A_FirePlasma(mbf_player_t*, pspdef_t*);
void __cdecl A_BFGsound(mbf_player_t*, pspdef_t*);
void __cdecl A_FireBFG(mbf_player_t*, pspdef_t*);
void __cdecl A_WeapSound(mbf_player_t*, pspdef_t*);

// Enemy actions
void __cdecl A_BFGSpray(mobj_t*);
void __cdecl A_Explode(mobj_t*);
void __cdecl A_Pain(mobj_t*);
void __cdecl A_PlayerScream(mobj_t*);
void __cdecl A_Fall(mobj_t*);
void __cdecl A_XScream(mobj_t*);
void __cdecl A_Look(mobj_t*);
void __cdecl A_Chase(mobj_t*);
void __cdecl A_FaceTarget(mobj_t*);
void __cdecl A_PosAttack(mobj_t*);
void __cdecl A_Scream(mobj_t*);
void __cdecl A_SPosAttack(mobj_t*);
void __cdecl A_VileChase(mobj_t*);
void __cdecl A_VileStart(mobj_t*);
void __cdecl A_VileTarget(mobj_t*);
void __cdecl A_VileAttack(mobj_t*);
void __cdecl A_StartFire(mobj_t*);
void __cdecl A_Fire(mobj_t*);
void __cdecl A_FireCrackle(mobj_t*);
void __cdecl A_Tracer(mobj_t*);
void __cdecl A_SkelWhoosh(mobj_t*);
void __cdecl A_SkelFist(mobj_t*);
void __cdecl A_SkelMissile(mobj_t*);
void __cdecl A_FatRaise(mobj_t*);
void __cdecl A_FatAttack1(mobj_t*);
void __cdecl A_FatAttack2(mobj_t*);
void __cdecl A_FatAttack3(mobj_t*);
void __cdecl A_BossDeath(mobj_t*);
void __cdecl A_CPosAttack(mobj_t*);
void __cdecl A_CPosRefire(mobj_t*);
void __cdecl A_TroopAttack(mobj_t*);
void __cdecl A_SargAttack(mobj_t*);
void __cdecl A_HeadAttack(mobj_t*);
void __cdecl A_BruisAttack(mobj_t*);
void __cdecl A_SkullAttack(mobj_t*);
void __cdecl A_Metal(mobj_t*);
void __cdecl A_SpidRefire(mobj_t*);
void __cdecl A_BabyMetal(mobj_t*);
void __cdecl A_BspiAttack(mobj_t*);
void __cdecl A_Hoof(mobj_t*);
void __cdecl A_CyberAttack(mobj_t*);
void __cdecl A_PainAttack(mobj_t*);
void __cdecl A_PainDie(mobj_t*);
void __cdecl A_KeenDie(mobj_t*);
void __cdecl A_BrainPain(mobj_t*);
void __cdecl A_BrainScream(mobj_t*);
void __cdecl A_BrainDie(mobj_t*);
void __cdecl A_BrainAwake(mobj_t*);
void __cdecl A_BrainSpit(mobj_t*);
void __cdecl A_SpawnSound(mobj_t*);
void __cdecl A_SpawnFly(mobj_t*);
void __cdecl A_BrainExplode(mobj_t*);


// killough code pointers
void __cdecl A_Detonate(mobj_t*);
void __cdecl A_Mushroom(mobj_t*);
void __cdecl A_Die(mobj_t*);
void __cdecl A_Spawn(mobj_t*);
void __cdecl A_Turn(mobj_t*);
void __cdecl A_Face(mobj_t*);
void __cdecl A_Scratch(mobj_t*);
void __cdecl A_PlaySound(mobj_t*);
void __cdecl A_RandomJump(mobj_t*);
void __cdecl A_LineEffect(mobj_t*);

// R3D specific
void __cdecl A_dbAttack(mobj_t*);

#endif	// __ACFNLINK_H__
