//-----------------------------------------------------------------------------
//
// MBF_DEFS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __MBF_DEFS_H__
#define __MBF_DEFS_H__

// Level 4: 
// C4152 nonstandard extension, function/data pointer conversion in expression
// C4155 named type definition in parentheses
// C4201 onstandard extension used : nameless struct/union
// C4204 nonstandard extension used : non-constant aggregate initializer
// C4706 assignment within conditional expression
#pragma warning (disable:4115 4152 4201 4204 4706)

#include "dd_share.h"
#include "dd_defns.h"
#include "m_args.h"
#include "doom/info.h"
#include "common_vars.h"

//
// Key cards.
//
typedef enum
{
    it_bluecard,
    it_yellowcard,
    it_redcard,
    it_blueskull,
    it_yellowskull,
    it_redskull,
    
    NUMCARDS
    
} card_t;

// The defined weapons,
//  including a marker indicating
//  user has not changed weapon.
typedef enum
{
    wp_fist,
    wp_pistol,
    wp_shotgun,
    wp_chaingun,
    wp_missile,
    wp_plasma,
    wp_bfg,
    wp_chainsaw,
    wp_supershotgun,

    NUMWEAPONS,
    
    // No pending weapon change.
    wp_nochange

} weapontype_t;


// Ammunition types defined.
typedef enum
{
    am_clip,	// Pistol / chaingun ammo.
    am_shell,	// Shotgun / double barreled shotgun.
    am_cell,	// Plasma rifle, BFG.
    am_misl,	// Missile launcher.
    NUMAMMO,
    am_noammo	// Unlimited for chainsaw / fist.	

} ammotype_t;

extern int ammo_sprite[NUMAMMO];


// Power up artifacts.
typedef enum
{
    pw_invulnerability,
    pw_strength,
    pw_invisibility,
    pw_ironfeet,
    pw_allmap,
    pw_infrared,
    NUMPOWERS
    
} powertype_t;

//
// Power up durations,
//  how many seconds till expiration,
//  assuming TICRATE is 35 ticks/second.
//
typedef enum
{
    INVULNTICS	= (30*TICRATE),
    INVISTICS	= (60*TICRATE),
    INFRATICS	= (120*TICRATE),
    IRONTICS	= (60*TICRATE)
    
} powerduration_t;

// A combination of patch data and its lump number.
typedef struct
{
	int	width, height;
	int xoffset, yoffset;
	int	lump;
} dpatch_t;

// #define snd_xx commented out b220-1 14/Jun/2008

#define ORIG_FRICTION			0xE800	// default value
#define ORIG_FRICTION_FACTOR	0x0800	//      "

#define NUM_XHAIRS	6

#define	NUMSAVESLOTS	10
#define SAVESTRINGSIZE	26

enum
{ 
	JOYAXIS_NONE,
	JOYAXIS_MOVE,
	JOYAXIS_TURN,
	JOYAXIS_STRAFE,
	JOYAXIS_LOOK
};

//
// Button/action code definitions.
//
typedef enum
{
    // Press "Fire".
    BT_ATTACK		= 1,
    // Use button, to open doors, activate switches.
    BT_USE			= 2,

    // Flag: game events, not really buttons.
    BT_SPECIAL		= 128,
    BT_SPECIALMASK	= 3,

	// Center player look angle (pitch back to zero).
	//BT_LOOKCENTER	= 64,
    
    // Flag, weapon change pending.
    // If true, the next 3 bits hold weapon num.
    BT_CHANGE		= 4,
    // The 3bit weapon mask and shift, convenience.
    BT_WEAPONMASK	= (8+16+32+64),
    BT_WEAPONSHIFT	= 3,

	BT_JUMP			= 8,

    // Pause the game.
    BTS_PAUSE		= 1,
    // Save the game at each console.
    //BTS_SAVEGAME	= 2,

    // Savegame slot numbers
    //  occupy the second byte of buttons.    
    //BTS_SAVEMASK	= (4+8+16),
    //BTS_SAVESHIFT 	= 2,

	// Special weapon change flags.
	BTS_NEXTWEAPON	= 4,
	BTS_PREVWEAPON	= 8,
  
} buttoncode_t;

// Game states
typedef enum
{
    GS_LEVEL,
    GS_INTERMISSION,
    GS_FINALE,
    GS_DEMOSCREEN,
	GS_WAITING,
	GS_INFINE,
	GS_LOADING
} gamestate_t;

typedef enum
{
    ga_nothing,
    ga_loadlevel,
    ga_newgame,
    ga_loadgame,
    ga_savegame,
    ga_playdemo,
    ga_completed,
    ga_victory,
    ga_worlddone,
    ga_screenshot

} gameaction_t;

// IWAD versions (US)
typedef enum
{
  shareware,	// DOOM 1 shareware, E1, M9
  registered,	// DOOM 1 registered, E3, M27
  commercial,	// DOOM 2 retail, E1 M34
  retail,		// DOOM 1 retail, E4, M36
  undetermined	// Well, no IWAD found.
} GameMode_t;

// Mission packs
typedef enum
{
  doom,			// DOOM 1
  doom2,		// DOOM 2
  pack_tnt,		// TNT mission pack
  pack_plut,	// Plutonia pack
  none_match

} GameMission_t;

#define sk_baby			0
#define sk_easy			1
#define sk_medium		2
#define sk_hard			3
#define sk_nightmare	4

int		GetDefInt(char *def, int *returned_value);

extern gameaction_t		gameaction;
extern GameMission_t	gamemission;
extern GameMode_t		gamemode;
extern int				gameskill;
extern gamestate_t		gamestate;

extern int	gameepisode;
extern int	gamemap;

// -------------------------
// Status flags for refresh.
//

// Depending on view size - no status bar?
// Note that there is no way to disable the
//  status bar explicitely.
//extern RDBOOL	statusbaractive;
//extern RDBOOL	nodrawers;
//extern RDBOOL	noblit;

// -------------------------------------
// Scores, rating.
// Statistics on a given map, for intermission.
//
extern  int	totalkills;
extern  int	totalfriends;
extern	int	totalitems;
extern	int	totalsecret;

// Timer, for scores.
extern  int	levelstarttic;		// gametic at level start
extern  int	actual_leveltime;	// still runs if in menu
extern  int	leveltime;			// tics in game play for par

// Other variables.
extern int		bodyqueslot;

// if true, load all graphics at level load
extern RDBOOL	precache;
#endif