//-----------------------------------------------------------------------------
//
// P_LOCAL.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// Revision 2.1.1 build 0 07/Dec/2006 gmj (b211-0)
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_LOCAL__
#define __P_LOCAL__

#include "doom/mbf_defs.h"
#include "doom/mbf_player.h"

#include "p_think.h"

RDBOOL	P_IsCamera(mobj_t *mo);
//void	P_CameraThink(mbf_player_t *player);
//int		P_CameraXYMovement(mobj_t *mo);
//int		P_CameraZMovement(mobj_t *mo);
void	P_Thrust3D(mbf_player_t *player, angle_t angle, float lookdir, 
				int forwardmove, int sidemove);
// Palette indices.
// For damage/bonus red-/gold-shifts
#define STARTREDPALS		1
#define STARTBONUSPALS		9
#define NUMREDPALS			8
#define NUMBONUSPALS		4

#define FLOATSPEED		(FRACUNIT*4)

#define DELTAMUL		6.324555320	// Used when calculating ticcmd_t.lookdirdelta
#define MAXHEALTH		maxhealth //100

// mapblocks are used to check movement
// against lines and things
#define MAPBLOCKUNITS	128
#define MAPBLOCKSIZE	(MAPBLOCKUNITS*FRACUNIT)
#define MAPBLOCKSHIFT	(FRACBITS+7)
#define MAPBMASK		(MAPBLOCKSIZE-1)
#define MAPBTOFRAC		(MAPBLOCKSHIFT-FRACBITS)

#define TOCENTER		-8

// player radius for movement checking
#define PLAYERRADIUS	16*FRACUNIT

// MAXRADIUS is for precalculated sector block boxes
#define MAXRADIUS		128*FRACUNIT

// Max step up before blocking
#define	MAXSTEP			24*FRACUNIT

// follow a player exlusively for 3 seconds
#define	BASETHRESHOLD	 	100

// Note: this excludes the player since the player has no seestate.
//       Thus, where used, it should be the case that the player
//       also requires to be excluded
#define sentient(mobj) ((mobj)->health > 0 && (mobj)->info->seestate)

#define STOPSPEED		(FRACUNIT/16)

//
// P_PSPR
//
void P_SetupPsprites (mbf_player_t* curplayer);
void P_MovePsprites (mbf_player_t* curplayer);
void P_DropWeapon (mbf_player_t* player);
void P_SetPsprite (mbf_player_t* player, int position, statenum_t stnum);

//
// P_USER
//
void	P_PlayerThink (mbf_player_t* player);
void	P_SetMessage (mbf_player_t *pl, char *msg);
void	P_playSetMessage(mbf_player_t *pl, char *msg); // R3D

extern int maxhealth, healthlimit;
extern int armorpoints[2];	// Green and blue points.

//
// P_MOBJ
//
#define ONFLOORZ		((int)0x80000000)
#define ONCEILINGZ		((int)0x7fffffff)

// For resetting in p_setup
extern int		plrviewheight;
extern int		plrfiringheight;

mobj_t* P_SpawnMobj(fixed_t x, fixed_t y, fixed_t z, mobjtype_t type, int iden_num);
void 	P_RemoveMobj (mobj_t* th, int from);
RDBOOL	P_SetMobjState (mobj_t* mobj, int state);
void	P_SpawnPuff (fixed_t x, fixed_t y, fixed_t z, RDBOOL ricoh);
void 	P_SpawnBlood (fixed_t x, fixed_t y, fixed_t z, int damage, int type);
mobj_t* P_SpawnMissile (mobj_t* source, mobj_t* dest, mobjtype_t type);
void	P_SpawnPlayerMissile (mobj_t* source, mobjtype_t type);
mobj_t* P_SpawnPlayer (mapthing_t* mthing, int pnum, int iden_num);
mobj_t*	P_SpawnTeleFog(int x, int y);
int		P_SpawnSplash(int x, int y, int floorh, float vol, int flags, RDBOOL chkmo);

void	P_SetDDFlags(mobj_t *mo);

//
// P_ENEMY
//
void P_NoiseAlert (mobj_t* target);

#define			MAXBRAINTARGETS 64	// Note 0x7f max if wanting to change this
									// else the save code will have to be changed (see braineasy)
extern mobj_t*	braintargets[MAXBRAINTARGETS];
extern int		numbraintargets;
extern int		braintargeton;
extern int		braineasy;

//
// P_MAPUTL
//

void P_SetThingPosition (mobj_t* thing);
int  P_Massacre(RDBOOL notfriends);

// P_CEILNG
void P_NullActiveCeilings(void);

// P_PLATS
void P_NullActivePlats(void);

// P_MAP
RDBOOL checkStepupClear(mobj_t *chkmo, fixed_t z);

//	If "floatok" true, move would be ok
//	if within "tmfloorz - tmceilingz".

typedef struct
{
	mobj_t		*tmthing;

	int			tmflags;
	int			plrblocked;			// b211-0

	// These are all cleared using memset so don't change the order
	//----------------------------------------
	RDBOOL		blocked3D;
	RDBOOL		blockmove;			// v220-5
	RDBOOL		exitgame;			// v220-6
	RDBOOL		explodemissile;		// v220-5
	RDBOOL		floatok;
	RDBOOL		mlblocking;			// v220-6
	RDBOOL		thingblocked;		// v220-4

	line_t		*ceilingline;		// For checking missiles with skywalls
	line_t		*floorline;			// $unstuck: Highest touched floor
	line_t		*blockline;			// $unstuck: blocking linedef
	// ----------------------------------------

	line_t		*plrlineblocked;

	subsector_t	*ssec;				// v220-5
	subsector_t	*newsubsec;

	short		fpic;				// b211-0

	RDBOOL		player;
	RDBOOL		plrceilblocked;		// Used in fake water sector
	RDBOOL		plrisjumping;
	RDBOOL		solid;				// v220-4
	RDBOOL		adjustz;			// v220-7
	RDBOOL		compHeights;		// v220-31
	RDBOOL		loslope;			// v220-31

	fixed_t		tmbbox[4];
	fixed_t		tmx, tmy, tmz, tmheight;

	fixed_t		z;					// v220-5
	fixed_t		fz;
	fixed_t		tmfloorz;
	fixed_t		tmdropoffz;
	fixed_t		tmceilingz;
	fixed_t		tmstptr;
	fixed_t		slopefht;			// b220-1 14/Jun/2008

	fixed_t		oldx;
	fixed_t		oldy;
	fixed_t		oldslopez;

	fixed_t		pt_oldx;			// pre-teleport
	fixed_t		pt_oldy;
	fixed_t		pt_oldz;

	// 3D ladders
	RDBOOL		onladder;
	fixed_t		laddertop, ladderbase;

	// v220-31
	fixed_t		moheight;			// z + height
	fixed_t		nextlowest;			// next lowest floor height line contacted
	fixed_t		nexthighest;		// next highest floor height line contacted
	fixed_t		dropoffz3D;
	fixed_t		same3D;

} gamedata_t;

extern gamedata_t gdt;

// v220-35 used with R3D_WADanalyse
typedef struct {
	int	mapnummoremove;
	int	mapnummorespawn;
	int triglinenum[3];
	int	trigset[3];
	mobj_t	*mobj;
} wadani_t;

extern wadani_t wadani;


RDBOOL	P_CheckPosition(mobj_t *thing, fixed_t x, fixed_t y, fixed_t z, int pcptest);
RDBOOL	P_TryMove(mobj_t *thing, fixed_t x, fixed_t y, int from);
RDBOOL	P_TeleportMove (mobj_t* thing, fixed_t x, fixed_t y, fixed_t z, RDBOOL is3d, RDBOOL boss);
RDBOOL	Check_Sides(mobj_t *actor, int x, int y);

void	P_Thrust(mbf_player_t*	player, angle_t angle, fixed_t move);
void	P_SlideMove (mobj_t* mo);
void 	P_UseLines (mbf_player_t* player);

int P_GetMoveFactor(mobj_t *mo, int *frictionp);

RDBOOL P_ChangeSector (sector_t *sector, sector_t *sec3D, RDBOOL crunch);

extern mobj_t*	mobjtarget;	// who got hit (or NULL)

fixed_t
P_AimLineAttack
( mobj_t*	t1,
  angle_t	angle,
  fixed_t	distance,
  int		mask);

void
P_LineAttack
( mobj_t*	t1,
  angle_t	angle,
  fixed_t	distance,
  fixed_t	slope,
  int		damage);

void
P_RadiusAttack
( mobj_t*	spot,
  mobj_t*	source,
  int		damage );


//
// P_INTER
//
extern int		maxammo[NUMAMMO];
extern int		clipammo[NUMAMMO];

void P_TouchSpecialThing(mobj_t *special);	// v220-16
RDBOOL P_CheckUnblocked(mobj_t *t1, mobj_t *t2);	// v220-16


void
P_DamageMobj
( mobj_t*	target,
  mobj_t*	inflictor,
  mobj_t*	source,
  int		damage,
  int		from);

//
// P_SETUP
//
//
int MakeFilterARGB(int lump, int *lim);


// P_SPEC
//
#include "p_spec.h"

#endif	// __P_LOCAL__
//-----------------------------------------------------------------------------
//
// $Log:$
//
//-----------------------------------------------------------------------------


