//-----------------------------------------------------------------------------
//
// P_SPEC.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_SPEC__
#define __P_SPEC__

#include "mbf_data.h"
#include "doom/mbf_player.h"

#define TriggerType			0x0007
#define TriggerTypeShift		 0

//jff 3/14/98 add bits and shifts for generalized sector types
#define DAMAGE_MASK			0x0060
#define DAMAGE_SHIFT			 5
#define SECRET_MASK			0x0080
#define SECRET_SHIFT			 7
#define FRICTION_MASK		0x0100
#define FRICTION_SHIFT			 8
#define PUSH_MASK			0x0200
#define PUSH_SHIFT				 9

//jff 02/04/98 Define masks, shifts, for fields in generalized linedef types
#define GenLimit			0x8000
#define GenFloorBase		0x6000
#define GenCeilingBase		0x4000
#define GenDoorBase			0x3c00
#define GenLockedBase		0x3800
#define GenLiftBase			0x3400
#define GenStairsBase		0x3000
#define GenCrusherBase		0x2F80

// define masks and shifts for the floor type fields
#define FloorCrush			0x1000
#define FloorChange			0x0c00
#define FloorTarget			0x0380
#define FloorDirection		0x0040
#define FloorModel			0x0020
#define FloorSpeed			0x0018

#define FloorCrushShift			12
#define FloorChangeShift		10
#define FloorTargetShift		 7
#define FloorDirectionShift		 6
#define FloorModelShift			 5
#define FloorSpeedShift			 3
                               
// define masks and shifts for the ceiling type fields
#define CeilingCrush		0x1000
#define CeilingChange		0x0c00
#define CeilingTarget		0x0380
#define CeilingDirection	0x0040
#define CeilingModel		0x0020
#define CeilingSpeed		0x0018

#define CeilingCrushShift		12
#define CeilingChangeShift		10
#define CeilingTargetShift		 7
#define CeilingDirectionShift	 6
#define CeilingModelShift		 5
#define CeilingSpeedShift		 3

// define masks and shifts for the lift type fields
#define LiftTarget			0x0300
#define LiftDelay			0x00c0
#define LiftMonster			0x0020
#define LiftSpeed			0x0018

#define LiftTargetShift			 8
#define LiftDelayShift			 6
#define LiftMonsterShift		 5
#define LiftSpeedShift			 3

// define masks and shifts for the stairs type fields
#define StairIgnore			0x0200
#define StairDirection		0x0100
#define StairStep			0x00c0
#define StairMonster		0x0020
#define StairSpeed			0x0018

#define StairIgnoreShift		 9
#define StairDirectionShift		 8
#define StairStepShift			 6
#define StairMonsterShift		 5
#define StairSpeedShift			 3

// define masks and shifts for the crusher type fields
#define CrusherSilent		0x0040
#define CrusherMonster		0x0020
#define CrusherSpeed		0x0018

#define CrusherSilentShift		 6
#define CrusherMonsterShift		 5
#define CrusherSpeedShift		 3

// define masks and shifts for the door type fields
#define DoorDelay			0x0300
#define DoorMonster			0x0080
#define DoorKind			0x0060
#define DoorSpeed			0x0018

#define DoorDelayShift			 8
#define DoorMonsterShift		 7
#define DoorKindShift			 5
#define DoorSpeedShift			 3

// define masks and shifts for the locked door type fields
#define LockedNKeys			0x0200
#define LockedKey			0x01c0
#define LockedKind			0x0020
#define LockedSpeed			0x0018

#define LockedNKeysShift		 9
#define LockedKeyShift			 6
#define LockedKindShift			 5
#define LockedSpeedShift		 3

#define GLOWSPEED				 8
#define STROBEBRIGHT			 5
#define FASTDARK				15
#define SLOWDARK				35

#define MAXPLATS				30
#define PLATWAITTAGGED			 4
#define PLATWAIT				 3
#define PLATSPEED		  FRACUNIT

#define VDOORSPEED		FRACUNIT*2
#define VDOORWAIT			   150

#define CEILSPEED		  FRACUNIT
#define CEILWAIT			   150

#define ELEVATORSPEED	FRACUNIT*4
#define FLOORSPEED		  FRACUNIT

#define MAXBUTTONS	  4*MAXPLAYERS // 4/player max or 48 for single player!
#define BUTTONTIME				35 // 1 second, in ticks.              

// define names for the TriggerType field of the general linedefs
typedef enum
{
 
    lowerFloor,			// lower floor to highest surrounding floor
    lowerFloorToLowest,	// lower floor to lowest surrounding floor
    turboLower,			// lower floor to highest surrounding floor VERY FAST
    raiseFloor,			// raise floor to lowest surrounding CEILING
    raiseFloorToNearest,// raise floor to next highest surrounding floor
	lowerFloorToNearest,// jff 02/03/98 lower floor to next lowest neighbor
	lowerFloor24,		// jff 02/03/98 lower floor 24 absolute
	lowerFloor32Turbo,	// jff 02/03/98 lower floor 32 absolute
    raiseToTexture,		// raise floor to shortest height texture around it
    lowerAndChange,		// as lowerFloorToLowest but also change floorpic
    raiseFloor24,
	raiseFloor32Turbo,	// jff 02/03/98 raise floor 32 absolute (seems it was never implemented gmj)
    raiseFloor24AndChange,
    raiseFloorCrush,
    raiseFloorTurbo,	// raise to next highest floor, turbo-speed
    donutRaise,
    raiseFloor512,
    
	//jff 02/04/98  add types for generalized floor mover
	genFloor,
	genFloorChg,
	genFloorChg0,
	genFloorChgT,

	//new types for stair builders
	buildStair,
	genBuildStair,

} floor_e;

typedef enum
{
    build8,	// slowly build by 8
    turbo16	// quickly build by 16
    
} stair_e;


typedef enum
{
	elevateUp,
	elevateDown,
	elevateCurrent,
	elevateR3DDown,
} elevator_e;

typedef enum
{
    lowerToFloor,
    raiseToHighest,
	lowerToLowest,
	lowerToMaxFloor,
    lowerAndCrush,
    crushAndRaise,
    fastCrushAndRaise,
    silentCrushAndRaise,

	//jff 02/04/98 add types for generalized ceiling mover
	genCeiling,
	genCeilingChg,
	genCeilingChg0,
	genCeilingChgT,

	//jff 02/05/98 add types for generalized ceiling mover
	genCrusher,
	genSilentCrusher,

} ceiling_e;

typedef enum
{
    normal,
    close30ThenOpen,
    closeNormal,
    openNormal,
    raiseIn5Mins,
    blazeRaise,
    blazeOpen,
    blazeClose,

	//jff 02/05/98 add generalize door types
	genRaise,
	genBlazeRaise,
	genOpen,
	genBlazeOpen,
	genClose,
	genBlazeClose,
	genCdO,
	genBlazeCdO,
} vldoor_e;

typedef enum
{
	floor_special,
	ceiling_special,
	lighting_special,
} special_e;

//  texture/type change
typedef enum
{
	trigChangeOnly,
	numChangeOnly,
} change_e;

typedef enum
{
    button_top,
    button_middle,
    button_bottom

} bwhere_e;

typedef enum
{
    up,
    down,
    waiting,
    in_stasis

} plat_e;

typedef enum
{
    perpetualRaise,
    downWaitUpStay,
    raiseAndChange,
    raiseToNearestAndChange,
    blazeDWUS,

	genLift,      //jff added to support generalized Plat types
	genPerpetual, 
	toggleUpDn,   //jff 3/14/98 added to support instant toggle type

} plattype_e;

typedef enum
{
	WalkOnce,
	WalkMany,
	SwitchOnce,
	SwitchMany,
	GunOnce,
	GunMany,
	PushOnce,
	PushMany,
} triggertype_e;

// define names for the Target field of the general floor
typedef enum
{
	FtoHnF,
	FtoLnF,
	FtoNnF,
	FtoLnC,
	FtoC,
	FbyST,
	Fby24,
	Fby32,
} floortarget_e;

// define names for the Changer Type field of the general floor
typedef enum
{
	FNoChg,
	FChgZero,
	FChgTxt,
	FChgTyp,
} floorchange_e;

// define names for the Speed field of the general linedefs
typedef enum
{
	SpeedSlow,
	SpeedNormal,
	SpeedFast,
	SpeedTurbo,
} motionspeed_e;

// define names for the Target field of the general ceiling
typedef enum
{
	CtoHnC,
	CtoLnC,
	CtoNnC,
	CtoHnF,
	CtoF,
	CbyST,
	Cby24,
	Cby32,
} ceilingtarget_e;

// define names for the Changer Type field of the general ceiling
typedef enum
{
	CNoChg,
	CChgZero,
	CChgTxt,
	CChgTyp,
} ceilingchange_e;


/*
// define names for the Change Model field of the general floor
typedef enum
{
  FTriggerModel,
  FNumericModel,
} floormodel_e;


// define names for the Change Model field of the general ceiling
typedef enum
{
	CTriggerModel,
	CNumericModel,
} ceilingmodel_e;
*/

// define names for the Target field of the general lift
typedef enum
{
	F2LnF,
	F2NnF,
	F2LnC,
	LnF2HnF,
} lifttarget_e;

// define names for the door Kind field of the general ceiling

typedef enum
{
	OdCDoor,
	ODoor,
	CdODoor,
	CDoor,
} doorkind_e;

// define names for the locked door Kind field of the general ceiling

typedef enum
{
	AnyKey,
	RCard,
	BCard,
	YCard,
	RSkull,
	BSkull,
	YSkull,
	AllKeys,
} keykind_e;

// Lump order in a map WAD: each map needs a couple of lumps
// to provide a complete scene geometry description.
#define ML_LABEL		0	// A separator, name, ExMx or MAPxx
#define ML_THINGS		1	// Monsters, items..
#define ML_LINEDEFS		2	// LineDefs, from editing
#define ML_SIDEDEFS		3	// SideDefs, from editing
#define ML_VERTEXES		4	// Vertices, edited and BSP splits generated
#define ML_SEGS			5	// LineSegs, from LineDefs split by BSP
#define ML_SSECTORS		6	// SubSectors, list of LineSegs
#define ML_NODES		7	// BSP nodes
#define ML_SECTORS		8	// Sectors, from editing
#define ML_REJECT		9	// LUT, sector-sector visibility	
#define ML_BLOCKMAP		10	// LUT, motion clipping, walls/grid element
#define ML_BEHAVIOR		11	// BEHAVIOR lump identifies HEXEN format levels


// structures

#pragma pack(4)		// Align to int

typedef struct
{
	thinker_t	thinker;
	int			*slist;	// sectors to light, in order
	int			ptr;	// current ptr in list
	int			count;  // number of sectors
	int			phase_tics;

} phased_t;

typedef struct
{
	thinker_t	thinker;
	sector_t*	sector;
	int		count;
	int		maxlight;
	int		minlight;
	int		maxtime;
	int		mintime;

} lightflash_t;

typedef struct
{
	thinker_t	thinker;
	sector_t*	sector;
	int		count;
	int		minlight;
	int		maxlight;
	int		darktime;
	int		brighttime;

} strobe_t;

typedef struct
{
	thinker_t	thinker;
	sector_t*	sector;
	int		minlight;
	int		maxlight;
	int		direction;

} glow_t;

typedef struct
{
	thinker_t	thinker;
	sector_t*	sector;
	int		count;
	int		maxlight;
	int		minlight;

} fireflicker_t;

typedef struct
{
	char	name1[9];
	char	name2[9];
	short	episode;

} switchlist_t;


typedef struct
{
	line_t		*line;
	bwhere_e	where;
	int			btexture;
	int			btimer;
	degenmobj_t	*soundorg;

} button_t;

typedef struct
{
	thinker_t	thinker;
	sector_t*	sector;
	fixed_t		speed;
	fixed_t		low;
	fixed_t		high;
	int			wait;
	int			count;
	plat_e		status;
	plat_e		oldstatus;
	int			crush;
	int			tag;
	plattype_e	type;

	struct platlist *list;   // killough

} plat_t;

// New limit-free plat structure -- killough

typedef struct platlist {
	plat_t	*plat; 
	struct	platlist *next, **prev;

} platlist_t;

typedef struct
{
	thinker_t	thinker;
	vldoor_e	type;
	sector_t*	sector;
	fixed_t	topheight;
	fixed_t	speed;

	// 1 = up, 0 = waiting at top, -1 = down
	int             direction;

	// tics to wait at the top
	int             topwait;
	// (keep in case a door going down is reset)
	// when it reaches 0, start going down
	int             topcountdown;

	//jff 1/31/98 keep track of line door is triggered by
	line_t *line;

	int lighttag; //killough 10/98: sector tag for gradual lighting effects

} vldoor_t;

typedef struct
{
	thinker_t	thinker;
	ceiling_e	type;
	sector_t	*sector;
	fixed_t		bottomheight;
	fixed_t		topheight;
	fixed_t		speed;
	fixed_t		oldspeed;
	int			crush;

	//jff 02/04/98 add these to support ceiling changers
	int		newspecial;
	int		oldspecial; //jff 3/14/98 add to fix bug in change transfers
	short	texture;

	// 1 = up, 0 = waiting, -1 = down
	int		direction;

	// ID
	int		tag;                   
	int		olddirection;

	struct ceilinglist *list;   // jff 2/22/98 copied from killough's plats

} ceiling_t;

typedef struct ceilinglist {
	ceiling_t	*ceiling; 
	struct ceilinglist *next, **prev;

} ceilinglist_t;

typedef struct
{
	thinker_t	thinker;
	floor_e		type;
	int			crush;
	sector_t	*sector;
	int			direction;
	int			newspecial;
	int			oldspecial;   //jff 3/14/98 add to fix bug in change transfers
	short		texture;
	fixed_t		floordestheight;
	fixed_t		speed;

} floormove_t;


typedef struct
{
	thinker_t	thinker;
	elevator_e	type;
	sector_t	*sector;
	int			direction;
	fixed_t		floordestheight;
	fixed_t		ceilingdestheight;
	fixed_t		speed;
} elevator_t;

// killough 3/7/98: Add generalized scroll effects

#define	sc_side				0
#define	sc_floor			1
#define	sc_ceiling			2
#define	sc_carry			3
#define	sc_carry_ceiling	4	// killough 4/11/98: carry objects hanging on ceilings

typedef struct {
	thinker_t	thinker;		// Thinker structure for scrolling
	fixed_t		dx, dy;			// (dx,dy) scroll speeds
	int			affectee;		// Number of affected sidedef, sector, tag, or whatever
	int			control;		// Control sector (-1 if none) used to control scrolling
	fixed_t		last_height;	// Last known height of control sector
	fixed_t		vdx, vdy;		// Accumulated velocity if accelerative
	int			accel;			// Whether it's accelerative
	int			type;			// Type of scroll effect

} scroll_t;

typedef struct {
	thinker_t thinker;   // Thinker structure for friction
	int friction;        // friction value (E800 = normal)
	int movefactor;      // inertia factor when adding to momentum
	int affectee;        // Number of affected sector
} friction_t;

// phares 3/20/98: added new model of Pushers for push/pull effects
#define p_push		0
#define p_pull		1
#define p_wind		2
#define p_current	3

typedef struct {
	thinker_t thinker;   // Thinker structure for Pusher
	mobj_t* source;	// Point source if point pusher
	int x_mag;			// X Strength
	int y_mag;			// Y Strength
	int magnitude;		// Vector strength for point pusher
	int radius;			// Effective radius for point pusher
	int x;				// X of point source if point pusher
	int y;				// Y of point source if point pusher
	int affectee;		// Number of affected sector
	int	type;			// type of effect
} pusher_t;

typedef enum
{
	ok,
	crushed,
	pastdest
} result_e;

result_e T_MovePlane
(
	sector_t*	sector,
	fixed_t		speed,
	fixed_t		dest,
	int			crush,
	int			floorOrCeiling,
	int			direction
);

typedef struct {
	int		secnum[2];
	int		x_pos[2];
	int		y_pos[2];
} tpfake_t;

#pragma pack()

extern	tpfake_t tpfake[100];

// general
RDBOOL		subfn_UseSpecialLine(mobj_t* thing, line_t* line, int from);
RDBOOL		P_ShootSpecialLine(mobj_t *thing, line_t* line, int from);
RDBOOL		P_UseSpecialLine(mobj_t *thing, line_t*	line, int side, int from);
void		P_CrossSpecialLine(line_t *line, mobj_t *thing);
void		P_PlayerInSpecialSector (mbf_player_t* player);

// utilities
sector_t	*getSector(int secnum, int lnum, int side);
sector_t	*getNextSector(line_t* line, sector_t* sec);
side_t		*getSide(int secnum, int lnum, int side);
int			twoSided(int secnum, int lnum);

fixed_t		P_FindLowestFloorSurrounding(sector_t* sec);
fixed_t		P_FindHighestFloorSurrounding(sector_t* sec);
fixed_t		P_FindNextHighestFloor(sector_t* sec, fixed_t currentheight);
fixed_t		P_FindLowestCeilingSurrounding(sector_t* sec);
fixed_t		P_FindHighestCeilingSurrounding(sector_t* sec);

// v220-5 was P_FindSectorFromLineTag
//        name changed to reflect that the tag is now sent directly
int			P_FindSectorFromTag(int tag, int start);

int			P_FindLineFromLineTag(const line_t* line, int start);
int			P_FindMinSurroundingLight(sector_t* sector, int	max );

// v220-5 was P_FindSectorFromLineTagWithLowerBound
int			P_FindSectorFromTagWithLowerBound(int tag, int start, int min);

void		P_InitPicAnims (void);		// one time load at start
void		P_SpawnSpecials (void);		// at map load
void		P_UpdateSpecials (void);	// every tic

int			EV_DoGenFloor(line_t *line);
int			EV_DoGenCeiling(line_t *line);
int			EV_DoGenLift(line_t *line);
int			EV_DoGenStairs(line_t *line);
int			EV_DoGenCrusher(line_t *line);
int			EV_DoGenDoor(line_t *line);
int			EV_DoGenLockedDoor(line_t *line);

// v220-31
int			P_CheckCompositeCeilings(int tag, fixed_t amount);
fixed_t		P_GetLowestCompositeFloor(int tag);
fixed_t		P_GetHighestCompositeFloor(int tag);
RDBOOL		P_CheckCompositeLower(int tag, fixed_t amount, fixed_t lolimit);
RDBOOL		P_CheckCompositeRaise(int tag, fixed_t amount, fixed_t hilimit);

//
// P_SPEC
//
void		T_Scroll(scroll_t *s);
void		T_Pusher(pusher_t *p);
void		T_Phased(phased_t *d);
void		T_SinglePhased(phased_t *d);
mobj_t		*P_GetPushThing(int s);
RDBOOL		checkTP_special(int special);

// P_LIGHTS
void		P_SpawnFireFlicker (sector_t* sector);
void		P_SpawnLightFlash (sector_t* sector);
void		P_SpawnStrobeFlash(sector_t*sector, int fastOrSlow, int inSync );
void		P_SpawnGlowingLight(sector_t* sector);

void		T_LightFlash (lightflash_t* flash);
void		T_StrobeFlash (strobe_t* flash);
void		T_Glow(glow_t* g);
void		T_FireFlicker (fireflicker_t* flick);

void		EV_StartLightStrobing(line_t* line);
void		EV_TurnTagLightsOff(line_t* line);
void		EV_LightTurnOn(line_t* line, int bright);
void		EV_LightTurnOnPartway(line_t *line, fixed_t level);

// v220-33 COD walkover line support
void		EV_LightSector(line_t* line, int value);

// P_GENLIN
int			P_FindShortestTextureAround(int secnum);
fixed_t		targetMAX(fixed_t startval, int texval);

// P_SWITCH
void		P_ChangeSwitchTexture(line_t* line, int useAgain);

// P_PLATS
void		T_PlatRaise(plat_t*	plat);
void		T_PlatInStasis(void *data);
void		P_AddActivePlat(plat_t* plat);
void		P_RemoveActivePlat(plat_t* plat);
void		EV_StopPlat(line_t* line);
void		P_ActivateInStasis(int tag);
int			EV_DoPlat(line_t* line, plattype_e type, int amount);

// P_DOORS
void		EV_VerticalDoor(line_t* line, mobj_t* thing);
void		T_VerticalDoor (vldoor_t* door);
void		P_SpawnDoorCloseIn30 (sector_t* sec);
void		P_SpawnDoorRaiseIn5Mins(sector_t* sec);
int			EV_DoDoor(line_t* line, vldoor_e type);
int			EV_DoLockedDoor(line_t* line, vldoor_e type, mobj_t* thing);

// P_CEILNG
void		T_MoveCeiling (ceiling_t* ceiling);
void		T_CeilInStasis(void *data);
void		P_AddActiveCeiling(ceiling_t* c);
void		P_RemoveActiveCeiling(ceiling_t* c);
int			EV_DoCeiling(line_t* line, ceiling_e type);
int			EV_CeilingCrushStop(line_t* line);
int			P_ActivateInStasisCeiling(line_t* line);

// miscellaneous
fixed_t		P_FindNextLowestFloor(sector_t *sec, int currentheight);
int			P_SectorActive(special_e t, sector_t *s);
int			P_CheckTag(line_t *line); // jff 2/27/98
void		P_SpawnFriction(void);
RDBOOL		P_CanUnlockGenDoor(line_t *line, mbf_player_t *player, RDBOOL type);

// P_FLOOR
int			EV_DoChange(line_t *line, change_e changetype);
int			EV_DoDonut(line_t* line);
int			EV_DoElevator(line_t* line, elevator_e elevtype);
int			EV_BuildStairs(line_t* line, stair_e type);
int			EV_DoFloor(line_t*line, floor_e floortype);

void		T_MoveFloor(floormove_t* floor);
void		T_MoveElevator(elevator_t* elevator);

// P_TELEPT
int			EV_Teleport(line_t *line, mobj_t *thing);			// killough 2/14/98
int			EV_SilentTeleport(line_t *line, mobj_t *thing);	// killough 1/31/98
int			EV_SilentLineTeleport(line_t *line, mobj_t *thing, RDBOOL reverse);

// EXTERN data
extern button_t		buttonlist[MAXBUTTONS]; 
extern platlist_t	*activeplats;		// killough 2/14/98

extern ceilinglist_t *activeceilings;	// jff 2/22/98

#endif
