//-----------------------------------------------------------------------------
//
// ST_LIB.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __ST_LIB__
#define __ST_LIB__


// We are referring to patches.
//#include "r_defs.h"


//
// Background and foreground screen numbers
//
#define BG 4
#define FG 0



//
// Typedefs of widgets
//

// Number widget

typedef struct
{
    // upper right-hand corner
    //  of the number (right-justified)
    int		x;
    int		y;

    // max # of digits in number
    int width;    

    // last number value
    int		oldnum;
    
    // pointer to current value
    int*	num;

    // pointer to RDBOOL stating
    //  whether to update number
    RDBOOL*	on;

    // list of patches for 0-9
    dpatch_t*	p;

    // user data
    int data;
    
} st_number_t;



// Percent widget ("child" of number widget,
//  or, more precisely, contains a number widget.)
typedef struct
{
    // number information
    st_number_t		n;

    // percent sign graphic
    dpatch_t*		p;
	    
} st_percent_t;



// Multiple Icon widget
typedef struct
{
     // center-justified location of icons
    int			x;
    int			y;

    // last icon number
    int			oldinum;

    // pointer to current icon
    int*		inum;

    // pointer to RDBOOL stating
    //  whether to update icon
    RDBOOL*		on;

    // list of icons
    dpatch_t*		p;
    
    // user data
    int			data;
    
} st_multicon_t;




// Binary Icon widget

typedef struct
{
    // center-justified location of icon
    int			x;
    int			y;

    // last icon value
    int			oldval;

    // pointer to current icon status
    RDBOOL*	val;

    // pointer to RDBOOL
    //  stating whether to update icon
    RDBOOL*	on;  

    dpatch_t*	p;	// icon
    int			data;   // user data
    
} st_binicon_t;



//
// Widget creation, access, and update routines
//

// Initializes widget library.
// More precisely, initialize STMINUS,
//  everything else is done somewhere else.
//
void STlib_init(void);



// Number widget routines
void
STlib_initNum
( st_number_t*		n,
  int			x,
  int			y,
  dpatch_t*		pl,
  int*			num,
  RDBOOL*		on,
  int			width );

void
STlib_updateNum
( st_number_t*		n,
  RDBOOL		refresh );


// Percent widget routines
void
STlib_initPercent
( st_percent_t*		p,
  int			x,
  int			y,
  dpatch_t*		pl, 
  int*			num,
  RDBOOL*		on,
  dpatch_t*		percent );


void
STlib_updatePercent(st_percent_t* per, RDBOOL refresh);


// Multiple Icon widget routines
void
STlib_initMultIcon
( st_multicon_t*	mi,
  int			x,
  int			y,
  dpatch_t*		il, 
  int*			inum,
  RDBOOL*		on );


void
STlib_updateMultIcon
( st_multicon_t*	mi,
  RDBOOL		refresh );

// Binary Icon widget routines

void
STlib_initBinIcon
( st_binicon_t*		b,
  int			x,
  int			y,
  dpatch_t*		i,
  RDBOOL*		val,
  RDBOOL*		on );

void
STlib_updateBinIcon
( st_binicon_t*		bi,
  RDBOOL		refresh );

#endif
//-----------------------------------------------------------------------------
//
// $Log:$
//
//-----------------------------------------------------------------------------
