//-----------------------------------------------------------------------------
//
// CVARS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __CVARS_H__
#define __CVARS_H__

#include "dd_share.h"

#define MAX_LINELEN		140

// Console variable flags.
#define CVF_NO_ARCHIVE		0x0001	// Not written in/read from the defaults file.
#define CVF_PROTECTED		0x0002	// Can't be changed unless forced.
#define CVF_NO_MIN			0x0004	// Don't use the minimum.
#define CVF_NO_MAX			0x0008	// Don't use the maximum.
#define CVF_CAN_FREE		0x0010	// The string can be freed.
#define CVF_HIDE			0x0020
#define CVF_SCP_FIXED		0x0040	// v220-3 can't be changed if a scripted map
#define CVF_CPBUTTON		0x0080	// v220-4 can only be changed using a CP button
#define CVF_DOPPLER			0x0100	// v220-5 call the D3D doppler function
#define CVF_INT_PANEL		0x0200	// v220-13 integer value can only be changed in control panel or from .cfg

enum 
{
	HUD_HEALTH,
	HUD_ARMOR,
	HUD_AMMO,
	HUD_KEYS,
	HUD_FRAGS
};

// Console variable types.
typedef enum
{
	CVT_NULL,
	CVT_BYTE,
	CVT_INT,
	CVT_FLOAT,
	CVT_CHARPTR					// ptr points to a char*, which points to the string.
} 
cvartype_t;

// Console variable.
typedef struct
{
	char		*name;
	int			flags;
	cvartype_t	type;
	void		*ptr;			// Pointer to the data.
	float		min, max;		// Minimum and maximum values (for ints and floats).
	char		*help;			// A short help text.
} cvar_t;

// List them all
extern cvar_t gameCVars[];
extern cvar_t inputCVars[];
extern cvar_t engineCVars[];

typedef struct config_s
{
	int		dclickuse;
	int		usemlook;		// Mouse look (mouse Y => viewpitch)
	int		mlookInverseY;	// Inverse mlook Y axis.
	int		mouseSensiX;
	int		mouseSensiY;
	int		joyEnable;
	int		joyaxis[8];
	int		usejlook;		// Joy look (joy Y => viewpitch)
	int		jlookInverseY;	// Inverse jlook Y axis.
	int		povLookAround;
	int		joySensitivity;
	int		joyDeadzone;
	int		jlookDeltaMode;
	int		lookSpeed;
	int		echoMsg;
	int		alwaysRun;		// Always run.
	int		lookSpring;
	int		sbarscale;
	float	flashcolor[3];
	int		flashspeed;
	int		turningSkull;
	int		hudShown[5];	// HUD data visibility.
	float	hudScale;		// How to scale HUD data?
	float	hudColor[3];

	// automap.
	float	automapAlpha;
	int		automapRotate;
	int		counterCheat;
	float	linesz;			// v220-22
	int		show_dead;		//   "

	// Crosshair.
	byte	xhairColor[3];
	byte	align2;
	int		xhairType;
	int		xhairSize;

	// Player bobbing
	float	bobView;
	float	bobWeapon;

	// Audio
	int		ricochet;
	int		weaponsounds;

	// Shadows
	int		useShadows;

	// Miscellaneous
	int		jumpEnabled;
	float	jumpPower;
	int		weaponAutoSwitch;
	int		cameraNoClip;
	int		corpseTime;
	int		secretMsg;
	float	menuScale;
	int		menuEffects;
	int		levelTitle;
	float	menuColor[3];
	int		menuFog;
	int		msgCount;
	float	msgScale;
	int		msgUptime;
	int		msgBlink;

	int		lightCompress;
	int		fakeglow;
	int		inFight;
	// v220-3
	int		berserkBug;	
	int		fallingDamage;
	int		underwaterDamage;
	int		underwaterAirbar;
	// v220-4
	int		footsteps;
	float	footstepvol;	// v220-35
	float	splashvol;		// v220-35
	int		usemapmusic;
	float	text_delay;
	// v220-5
	float	flashdamage;
	float	flashpickup;
	float	doppler;
	int		pitch;
	int		use43;
	// v220-6
	int		noberserkBug;	
	int		PlayerColor[MAXPLAYERS];
	// v220-13
	int		vid_cap;		// Set the video cap fps
	// v220-16
#if 0		// $bulletrange
	int		bulletRange;
#endif
	// v220-31
	int		useTorchLight;
	int		sineCompress;
	int		setVtxLight;
	int		glflush;
	// v220-33
	int		jumpkill;
	int		oldbarscale;
	float	zbias;
	// v220-35
	int		amDevboxWidth;
	int		amEditboxWidth;
	int		olddoommessage;
	int		splashsprite;
	int		weapsmoke;

} config_t;

extern config_t	cfg;

extern RDBOOL	hu_showallfrags;
extern int		showMessages;
extern char		*chat_macros[];

// video vars
extern int	vidShowFPS;
extern int	vidVsync;

void Con_AddVariableList(cvar_t *varlist);
#endif