//-----------------------------------------------------------------------------
//
// DD_DEFNS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __DD_DEFNS_H__
#define __DD_DEFNS_H__

#include "dedfile.h"
#include "dd_dfdat.h"

typedef struct
{
	int wall_texture;
	int flat_lump;
	int detail_lump;
} detailtex_t;

typedef struct sfxinfo_s
{
	void	*data;				// Pointer to sound data.
	int		lumpnum;
	char	lumpname[9];		// Lump name.
	int		group;
	int		flags;
	int		sfxmaxchan;			// Max. channels for the sound to occupy.

/*	// The following are not used - R3D
	struct	sfxinfo_s *link;	// Link to another sound.
	int		link_pitch;
	int		link_volume;
	int		usefulness;			// Used to determine when to cache out.

	// Although read in from the defs this is not used.  A more sophisticated method for
	// determining priority was introduced by JK using the MS sound code (R3D comment)
	int		priority;
	char	external[256];		// Path to external file (R3D not supported).
*/
} sfxinfo_t;

extern ded_t		defs;		// The main definitions database.
extern sprname_t	*sprnames;	// Sprite name list.
extern state_t		*states;	// State list.
extern mobjinfo_t	*mobjinfo;	// Map object info database.
extern sfxinfo_t	*sounds;	// Sound effect list.
extern ddtext_t		*texts;		// Definition text list (these remain unchanged)
extern detailtex_t	*dettex;	// Detail texture mappings.
extern mobjinfo_t	**stateowners;
extern ded_count_t	count_sprnames;
extern ded_count_t	count_states;

// From DD_Setup.c.
extern ded_mapinfo_t *mapinfo;	// Current mapinfo, can be NULL.

char	*DefCheckBase(int *val);
void	Def_Init(void);

// Destroy databases.
void Def_Destroy(void);

// Reads the specified definition file, and creates the sprite name, 
// state, mobjinfo, sound, music and text databases accordingly.
void Def_Read(RDBOOL autoload);
RDBOOL BaseModelDefsExist(void);

void Def_PostInit(void);

int				Def_GetStateNum(char *id);
int				Def_GetSpriteNum(char *name);
int				Def_GetMusicNum(char *id);
int				Def_EvalFlags(char *ptr);
ded_mapinfo_t	*Def_GetMapInfo(char *map_id);
ded_light_t		*Def_GetLightDef(int spr, int frame);
int				Def_Get(int type, char *id, void *out);
int				Def_SameStateSequence(state_t *snew, state_t *sold);

// R3D audio
int		R3D_ChangeSoundUsingAlias(int sound_id, char *lumpname);
int		R3D_AddNewSound(char *lumpname);
char	*R3D_GetNameForSoundID(int sound_id);

#endif