//-----------------------------------------------------------------------------
//
// DD_INPUT.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __DD_INPUT_H__
#define __DD_INPUT_H__

//------------------------------------------------------------------------
//
// Events
//
//------------------------------------------------------------------------

typedef enum
{
	ev_keydown,
	ev_keyup,
	ev_keyrepeat,
	ev_mouse,
	ev_mousebdown,
	ev_mousebup,
	ev_joystick,			// Joystick main axes (xyz + Rxyz)
	ev_joyslider,			// Joystick sliders
	ev_joybdown,
	ev_joybup,
	ev_povdown,
	ev_povup
} evtype_t;

typedef struct
{
	evtype_t	type;
	int			data1;		// keys/mouse/joystick buttons
	int			data2;		// mouse/joystick x move
	int			data3;		// mouse/joystick y move
	int			data4;
	int			data5;
	int			data6;
} event_t;

#define MAXEVENTS	64

#define KEYBUFSIZE	32

// The mouse wheel is considered two extra mouse buttons.
#define DDMB_MWHEELUP		0x1000
#define DDMB_MWHEELDOWN		0x2000

extern event_t	events[MAXEVENTS];

extern int		repWait1, repWait2;
extern int		mouseDisableX, mouseDisableY;
extern int		mouseInverseY;
extern int		mouseWheelSensi;
extern int		showScanCodes;
extern RDBOOL	shiftDown, altDown;

void DD_PostEvent(event_t *ev);
void DD_ProcessEvents(void);
void DD_ClearKeyRepeaters(void);
void DD_InitInput(void);
byte DD_ScanToKey(byte scan);
byte DD_ModKey(byte key);

#endif 