//-----------------------------------------------------------------------------
//
// DD_WAD.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __DD_WAD_H__
#define __DD_WAD_H__

#include <io.h>

// File record flags.
#define	FRF_RUNTIME		0x1		// Loaded at runtime (for reset).

#define MAX_PWADS		32		// For save game checks

// lumpinfo liflags
#define	LINF_MAPNAME	0x0001	//	set but ignored
#define	LINF_PWAD		0x0008	// v220-3	no longer used
#define	LINF_PISTOL		0x0010	// v220-3	set but ignored
#define	LINF_TRUNCATED	0x0020	// v220-5
#define	LINF_LMPISWAD	0x0040	// v220-6

// Lump Grouping Tags identifies data between markers shown
// See marker_name_t groups in dd_wad.c for correspondance
// lump types (NOTE; see also grpdesc)
#define 	LGT_NONE		0
#define 	LGT_FLATS		1	// F_START and F_END
#define 	LGT_SPRITES		2	// S_START and S_END
#define 	LGT_PNAMES		3	// P_START and P_END
#define 	LGT_TX			4	// TX_START and TX_END  DeepSea
#define 	LGT_COLORMAPS	5	// C_START and C_END
#define 	LGT_HI			6	// HI_START and HI_END Replacement textures
#define 	NUM_LGTAGS		7	// search size for above
#define 	LGT_DETAIL		8	// tag for detail textures
#define 	LGT_MARKER		9	// zero sized group marker

// lump types (NOTE; see also typedesc)
#define LMP_UNDEF       0   // not determined yet
#define LMP_MARKER      1   // empty lump used as a marker
#define LMP_IGNORE      2   // cannot be determined or ignore duplicate
#define LMP_TXT         3   // lumps expected to be txt
#define LMP_MAP         4	// stuff to do with the map

// textures (see Add_NewTexture()) before changing anything
#define LMP_PNG         5
#define LMP_PCX         6
#define LMP_JPG         7	// first supported in v220-6
#define LMP_BMP         8
#define LMP_FLAT        9   // flat format (includes hires > 64x64 )
#define LMP_FLAT320    10   // DOOM flat 320x200
#define LMP_PATCH      11   // patch format 

// sound types
#define LMP_MIDI       12
#define LMP_RMI	       13
#define LMP_MUS        14
#define LMP_WAV        15
#define LMP_SND        16	// DOOM snd

extern const char		*grpdesc[], *typedesc[];

typedef struct
{
	char	filename[512];	// Full filename (every '\' -> '/').
	int		numlumps;		// Number of lumps.
	int		handle;			// File handle.
	int     numlevels;		// May only be set true (i.e. cannot be relied on for actual number of levels)
	int		pwad;
	uint	checksum;
	long	time_create;
	long	time_write;
} filerecord_t;

typedef struct
{
	char		name[9];		// End in \0
	char		groupMarker;	// Lump grouping tag (LGT_*).
	short		liflags;
	int			handle;
	size_t		position, size;
	int			sent;
	int			width, height;
	int			lumptype;
	short		modidx;
} lumpinfo_t;

typedef struct
{
	char     type[4];
	int      num_entries;
	unsigned dir_start;
} wad_header_t;

typedef struct 
{
	unsigned start, length;
	char	 name[8];
} lump_header_t;

#pragma pack(2)

typedef struct
{
	short	textureoffset;
	short	rowoffset;
	char	toptexture[8];
	char	bottomtexture[8];
	char	midtexture[8];
	ushort	sector;
} mapsidedef_t;

#pragma pack()

extern lumpinfo_t *lumpinfo;
extern int numlumps;

extern int		IWADhandle;
extern int		RISEN3DWADhandle;

char	*W_LumpName(int lump);
char	*W_LumpSourceFile(int lump);
char	*W_GetLumpNamePtr(int lumpnum);

void	*M_CacheLumpNum(int lump, size_t *length, RDBOOL error);
void	*M_CacheLumpName(char *name, size_t *length, RDBOOL error);

int		W_CheckTextureNumForName(char *name);
int		W_CheckTextureMissingNumForName(char *name);	// v220-5
int		W_CheckFlatNumForName(char *name);
int		W_CheckNumForName(char *name);
int		W_CheckTXNumForName(char *name);
int		W_GetNumForName(char *name);
int		W_IWADtype(void);
int		W_LumpLength(int lump);

// R3D
void	CheckLumpRange(int lump, int num);
int		W_CheckIWADNumForName(char *name);
int		W_CheckPWADNumForName(char *name);
int		W_GetPWADsound(char *name);
int		W_CheckPWADPatchForName(char *name, int *type);	// v220-13
int		W_CheckPWADNumForLoResName(char *name);
int		W_CheckColormapForName(char *name);
int		W_CheckRisen3DWADNumForName(char *name, RDBOOL eror);
int		W_CheckPwadSwitch(int *type);
int		W_CheckTextureNONTXNumForName(char *name);
int		W_CheckOriginalPatch(char *name);
int		W_FindScriptLump (char *name, int *lastlump);
RDBOOL	pwadHandle(int handle);
RDBOOL	W_CheckPWADtexture(char *name);	// v220-3


RDBOOL	W_AddFile(char *fpath, char *fname, RDBOOL pwadHasLevels);	// v220-6 gwa parm removed

int		W_CheckDetailNumForName(char *name);
int		W_CheckSpriteNumForName(char *name);
RDBOOL	W_OpenLatestScriptLump (char *name, char *mapid);
int		W_GetMapIndex(char *mapid);

// R3D file time/date handling - jav
void	FREEDOOMcheck(void);
long	RetrieveFileDateByName(char *filename);
long	RetrieveFileDateByNum(int fnum);
RDBOOL	MapFileDateGreaterThan(long time_write);
int     RetrieveFileNumberByName(char *filename);
int		RetrieveFileDateByModidx(int ref, uint *crc);	//v220-6 crc added v220-16

extern  filerecord_t	*filenameRecords;	// jav
extern int				numfilenameRecords;	// jav

extern uint	pwadchecksum[MAX_PWADS + 1];//iwad crc stored in 0
#endif