//-----------------------------------------------------------------------------
//
// GL_TGA.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __GL_TGA_H_
#define __GL_TGA_H_

#include "sys_file.h"

#ifdef __cplusplus
extern "C" {
#endif

// Screen origins.
#define	TGA_SCREEN_ORIGIN_LOWER	0	// Lower left-hand corner.
#define	TGA_SCREEN_ORIGIN_UPPER	1	// Upper left-hand corner.

// Data interleaving.
#define	TGA_INTERLEAVE_NONE		0	// Non-interleaved.
#define	TGA_INTERLEAVE_TWOWAY	1	// Two-way (even/odd) interleaving.
#define	TGA_INTERLEAVE_FOURWAY	2	// Four-way interleaving.

enum
{
	TGA_FALSE,
	TGA_TRUE,
	TGA_TARGA16,
	TGA_TARGA24,		// rgb888
	TGA_TARGA32			// rgba8888
};

// Level 4: nonstandard extension used : bit field types other than int
#pragma warning (disable:4214)

#pragma pack(1)	// Confirm that the structures are formed correctly.
typedef struct 		// Targa image descriptor byte; a bit field
{
	byte	attributeBits	: 4;	// Attribute bits associated with each pixel.
    byte	reserved		: 1;	// A reserved bit; must be 0.
    byte	screenOrigin	: 1;	// Location of screen origin; must be 0.
    byte	dataInterleave	: 2;	// TGA_INTERLEAVE_*
} TARGA_IMAGE_DESCRIPTOR;

typedef struct
{
	byte	idFieldSize;			// Identification field size in bytes.
	byte	colorMapType;			// Type of the color map.
    byte	imageType;				// Image type code.
    // Color map specification.
	ushort  colorMapOrigin;			// Index of first color map entry.
    ushort	colorMapLength;			// Number of color map entries.
	byte	colorMapEntrySize;		// Number of bits in a color map entry (16/24/32).
    // Image specification.
    ushort	xOrigin;				// X coordinate of lower left corner.
    ushort	yOrigin;				// Y coordinate of lower left corner.
    ushort	imageWidth;				// Width of the image in pixels.
    ushort	imageHeight;			// Height of the image in pixels.
    byte	imagePixelSize;			// Number of bits in a pixel (16/24/32).
	TARGA_IMAGE_DESCRIPTOR imageDescriptor;	// A bit field.
} TARGA_HEADER;
#pragma pack()		// Back to the default value.

// Save the rgb888 buffer as Targa 24.
int TGA_Save24_rgb888(char *filename, int w, int h, byte *buffer);

// Load an rgb 888 (24 bits per pixel) or rgba 8888 image (32 bits per pixel)
byte *TGA_LoadFile888(char *filename, int *w, int *h, int *pixSize);

// Get the dimensions of a Targa image.
int TGA_GetSize(char *filename, int *w, int *h);

#ifdef __cplusplus
}
#endif

#endif
