//-----------------------------------------------------------------------------
//
// M_NODEPILE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __M_NODEPILE_H__
#define __M_NODEPILE_H__

#define NP_ROOT_NODE	((void*) -1 )

//------------------------------------------------------------------------
//
// Linknodes
//
//  Linknodes are used when linking mobjs to lines. Each mobj has a ring
//  of linknodes, each node pointing to a line the mobj has been linked to.
//  Correspondingly each line has a ring of nodes, with pointers to the 
//  mobjs that are linked to that particular line. This way it is possible
//  that a single mobj is linked simultaneously to multiple lines (which 
//  is common).
//
//  All these rings are maintained by P_(Un)LinkThing().
//
//------------------------------------------------------------------------

#pragma pack(4)

typedef struct linknode_s
{
	nodeindex_t		prev, next;
	int				data;
	void			*object_ptr;
} linknode_t;

#pragma pack()

typedef struct nodepile_s
{
	int			count;
	int			pos;
	linknode_t	*nodes;
} nodepile_t;

void NP_Init(nodepile_t *pile, int initial);
nodeindex_t NP_New(nodepile_t *pile, void *ptr);
void NP_Link(nodepile_t *pile, nodeindex_t node, nodeindex_t root);
void NP_Unlink(nodepile_t *pile, nodeindex_t node);

#endif 