//-----------------------------------------------------------------------------
//
// R3D_MAINSPEC.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __R3D_MAINSPEC_H__
#define __R3D_MAINSPEC_H__

#include "r3d_compat.h"
#include "p_data.h"
#include "r_main.h"
#include "con_console.h"
#include "dd_main.h"
#include "m_args.h"
#include "r_data.h"
#include "doom/p_spec.h"

#include "r3d_intersect.h"
#include "r3d_lscheck.h"
#include "r3d_sl3D.h"
#include "r3d_movsec.h"
#include "r3d_ssline.h"
#include "r3d_utils.h"
#include "r3d_modelbox.h"


// v220-18
typedef struct {
	int	x, y;
}dmarray_t;

#define DM_MAXCOUNT	16
extern dmarray_t dmarray[DM_MAXCOUNT];
extern int dmcount;

typedef struct
{
	// For 3D stuff
	char	hiname[9];
	char	midname[9];
	char	loname[9];
} texname3D_t;

extern	texname3D_t *texname3D;

typedef struct 
{
	int		bounds[4];		// Bounding box for the sector.
	short	type;			// Set true if bounded by four outer lines
} bounds_t;

extern bounds_t *secbbox;

#define	MAXSORTBOXES	1024

typedef struct {
	__int64	area;
	int		box[4];
	int		*llist;
	int		listcnt;
	int		lnum;
	RDBOOL	done;

} sortbox_t;

// texerror
extern RDBOOL texerror, texwarn;
extern char *texerrfix;

// r3d_closed.c
void	LS_CheckMissingHiClosedLoops(void);
void	LS_CheckMissingLoClosedLoops(void);
void	LS_CheckOpenUntexturedContained(void);

// r3d_chained.c
void	LS_SetChainedLines(void);
void	LS_CheckLinesR3D_01(void);

// r3d_contained.c
void	LS_CheckLinkedAnimSecs(void);
void	LS_CheckMovCeilInAnim(void);
void	LS_CheckMovAnimContained(void);
void	LS_FindLinkedAnimSectors(void);
void	LS_ChkContained808Csectors(void);
void	LS_ChkBoundedSector(void);
void	LS_CheckForRingLinePairs(void);
void	LS_SetMissLowtexMovSectorsContained(void);
void	LS_MarkLoopsWithContainedLoops(void);
void	LS_SameSectorContained(void);	// v220-10

// r3d_consolidate
void subfn_PRESET();
void LS_Consolidate(void);

// r3d_ctrllines.c
void	LS_PresetContainedSSLines(void);	// b211-0

// r3d_fogwalls.c
void	FW_SetFogwallGroups(void);			// v220-3

// r3d_hack.c
void	LS_ChkNestedUnconnectedLinesInDifferentSectors(void);
void	LS_RTP_Hacks(void);
void	LS_XlineHacks(void);				// v220-10
void	LS_FalseFloorRingSpan(void);		// v220-10
void	LS_FalseCeilRingSpan(void);			// v220-13
void	LS_SupportSecInside(void);			// v220-17

// r3d_inherit.c
void	LS_chkInherit(void);
void	LS_InheritanceMovFloorLoop(void);	// v220-10

// r3d_iterate.c
void	LS_CheckOpenStaticContained(void);
void	LS_CheckSupportLineRings(void);

// r3d_lcross.c
void	LS_SetTypeR3D_2ECrossing(void);
void	LS_MiscLineCrossAndLineCoincident(void);

// r3d_linedebug.c
void	Setspec(line_t *ld, int val, int num);
void	Setloop(line_t *ld, int val, int num);
void	SetLFGSecSet(line_t *ld, int num);
void	Settype(line_t *ld, byte val, int num);
void	SettypePA(line_t *ld, byte val, int num);
void	Forcetype(line_t *ld, byte val, int num);
void	Absolutetype(line_t *ld, byte val, int num);
void	SetLFG(line_t *ld, short lfgflag, int num);
void	Setlfsec(sector_t *sec, int secnum, int num);
void	Setr3dch(sector_t *sec, sector_t *s, int num);
void	Setr3dch_byH(sector_t *sec, int height, int num);
void	Setr3dchangech(sector_t *sec, int ht, int num);
void	Setr3dfh(sector_t *sec, sector_t *s, int num);
void	Setr3dfh_byH(sector_t *sec, int height, RDBOOL test, int num);
void	Sethir3dssfs(subsector_t *ss, int newssfs, int num);
void	Setlor3dsscs(subsector_t *ss, int newsscs, int num);
void	Setr3dfs(sector_t *s, int newsec, int num);
void	Setr3dcs(sector_t *s, int newsec, int num);
void	Setr3dssfs(subsector_t *ss, int newssfs, int num);
void	Setr3dsscs(subsector_t *ss, int newsscs, int num);
void	Setr3dsslcsec(subsector_t *ss, int newsscs, int num);
void	Setr3dsslfsec(subsector_t *ss, int newsslfs, int num);
void	SetNoskywall(int lnum, int num);
void	Setsecdone(sector_t *sec, int num);
void	SetSkywall(line_t *ld, int num);
void	Setsecdone2(sector_t *sec, int num);
void	Setsecdone3(sector_t *sec, int num);
void	SetSecidxDone(int secidx, int num);
void	SetMissinglo(line_t *ld, int num);
void	ResetMissinglo(line_t *ld, int num);
void	Clr_allmtflags(line_t *ld, int num);
void	Clr_mthiflags(line_t *ld, int num);
void	Clr_mtloflags(line_t *ld, int num);
void	inhibitSSEC(subsector_t *ss, int num);
void	setSSECflag(subsector_t *ss, int flag, int num);
void	ChangeSSECsec(subsector_t *ss, sector_t *s, int num);
void	SetSSF_BASESSECFIX(subsector_t *ss, int num);
void	SetNOocclude(sector_t *sec, int num);
void	SetOverride(subsector_t *ss, int ovrnum, int num);
void	SetControlline(line_t *ld, int num);
void	SetFloorlightsec(subsector_t *ss, int newsec, int num);

// r3d_linegroups.c
void	LG_CheckMovTexUnset(void);
void	LG_AnalyseMovTexUnset(void);
void	LG_RemoveAashitty(void);
void	LG_UnclosedMovSectors(void);
void	LG_CheckFixedInMovSector(void);
void	LG_CheckUnsetStepLinePlats(void);

// r3d_MAINspec.c
void	R3D_MAINAnalyseMap(int thinglumpnum);
void	R3D_MainSetupInit(int thinglumpnum, RDBOOL analyse);

// r3d_maperror.c
void	LS_CheckBacksector(void);
void	LS_CheckFloorCeilHeightError(void);
void	LS_FixSubsecErrorsWithShortSegs(void);
void	LS_AddTexDefinitions(void);

// r3d_mapGeneral.c
void	LS_SetSameSecInGroupFlag(void);
void	LS_SetLinesSpanning1sLines(void);
void	LS_CheckMapErrors(void);
int		LS_ClearEmptySectorSpecials(void);
void	LS_InitTagLists(void);
void	LS_Process3DLines(int analyse);
void	LS_SetLineTaggedSectors(void);
int		LS_CheckTaggedErrors(void);
void	LS_CheckDoorType(void);
void	LS_SetHASTEXTUREflags(void);
void	LS_Setlinetypes(void);

// r3d miscEffects.c
void	LS_CheckFakeSectorInAnimated(void);
void	LS_CheckR3D_1ALines(void);
void	LS_CheckReversedLineMovCeilSecs(void);
void	LS_CheckUnclosedHighTexSectors(void);

// r3d_misc.c
void	LS_DelSpecSecWithUnconnectedLinesInAnimSec(void);
void	LS_CheckR3D_1AforR3D_3F(void);
void	LS_SetTexInSector(void);
void	LS_CheckRingSectors(void);
void	LS_SetMiscStaticType(void);
void	LS_SetMiscMoveType(void);
void	LS_SetMiscLineType(void);
void	LS_CheckDualRingSectors(void);
void	LS_CheckDualRingMissingSectors(void);
void	LS_CheckDonut(void);
void	LS_PresetLCSEC(void);
void	LS_CheckSameSecOwner(void);
void	LS_SetVisSplit(void);
void	LS_chkOrphanedMov(void);

// r3d_misContained.c
void	MC_CheckDoubleLowMove(void);
void	MC_CheckFlatRings(void);
void	MC_Nested(void);
void	MC_DoubleLoop(void);
void	LS_SameSidedLineSectors(void);
void	MC_FalseFloorInBox(void);
void	MC_ContiguousMissingHi(void);
void	MC_MissingLoMove(void);
void	MC_CheckNonContiguous(sector_t *sc);
void	MC_CheckIfHeightsecInvalid(void);
void	MC_FixedMLoContained(void);
void	MC_CheckSSLinesContained(void);
RDBOOL	mc_MovsecMLoContained(sector_t *sc);

// r3d_missingtex.c
void	LS_SetLineMTFlags(void);
void	LS_SetMissingTexLineType(void);
void	LS_CheckLinesInFakedSecs(void);
void	LS_SetHiLoTexFlags(void);
void	LS_CheckLineLoops(void);
RDBOOL	LS_CheckMissing(line_t *ld);
void	LS_DeleteUnwantedHiTextures(void);
void	LS_CheckInvalidLoTexFinal(void);

// r3d_mtex.c v220-9/v220-10
void	MX_RemoveUnwantedTexures(void);
void	MX_CheckInvalidPlats(void);
void	MX_CheckInvalidDoors(void);
void	MX_Prelim(void);
void	MX_MultiHiLineRing(void);
void	MX_Main(void);
void	MX_HeightsecMissing(void);
void	MX_CheckNestedTexInMovRings(void);
void	MX_CheckNoTexMovRings(void);
void	MX_MissingTexLoopSSLines(void);
void	MX_UnsetLines(void);
void	MX_Contained(void);
void	MX_CheckMovFloor(void);
void	MX_ContainedBySSLineRing(void);
void	MX_CheckLineRings(void);
void	MX_Unset(void);
void	MX_Postlim(void);
void	MX_MovMixed(void);
void	MX_Multi(void);
void	MX_MissingContained(void);
void	MX_R3D_16Nested(void);
void	MX_PartialMissing(void);
void	MX_CheckMultiFloorMove(void);
void	MX_ConnectedLo(void);
void	MX_MovInside(void);
void	MX_InnerMovLoopContained(void);
void	MX_CheckLoMissingLoops(void);
void	MX_CheckLoMissingFloorMovLoops(void);
void	MX_CheckMovCeilInside(void);
void	MX_CheckMovFloorInside(void);
void	MX_CheckMissingHiLoSkySectors(void);
void	MX_CheckNestedNotexSectors(void);
void	MX_CheckMoveNoUpperTexSectors(void);
void	MX_UnsetR3D_16(void);
void	MX_CheckNoTexDualRings(void);
void	MX_CheckNoTexRings(void);
void	MX_CheckLowMoveInFalseCeiling(void);
void	MX_CheckCojoinedMissing(void);
void	MX_CheckContained(void);
void	MX_CheckUnclosedMissingContained(void);
void	MX_ReversedMissingContained(void);
void	MX_CheckHighMissingContained(void);
void	MX_CheckPartialHiMissing(void);
void	MX_CheckAllConnectedLowMissing(void);
void	MX_CheckNoHiAndSameHeightLine(void);

// r3d_NewFunc.c
// names starting nf_ are globally available sub functions
RDBOOL	nf_CheckBoxRing(sector_t *s, int *segbox, line_t *ld);
int		nf_lineList(line_t *ld, sector_t *sec, int *bbox, int *lnarray, int condition);
int		nf_chkLineListRing(int count, int *bbox, int lptr, byte *linemem);
RDBOOL	nf_boxAreaSameSec(sector_t *sec, int *cs, int *box, int count);

void	NF_ContainRingsFINAL(void);
void	NF_ContainRingsCTRL(void);
void	NF_EnclosedFalseFloors(void);
void	NF_UnclosedSecGroupsContained(void);
void	NF_UnclosedSecGroupsContained2(void);	// v220-13
void	NF_CheckLinesForInheritance(void);
void	NF_ContainerRings(void);
void	NF_CheckDualSectorTypeLineRings(void);	// v220-10
void	NF_CheckRingInheritance(void);
void	NF_CheckSharedVertexRing(void);	// v220-10
void	NF_CheckInheritance(void);
void	NF_CheckInheritance2(void);	// v220-4
RDBOOL	NF_CheckLoop(sector_t *sec, int fnnum);
void	NF_Check1sLinesBoundedLiquidFloors(void);
void	NS_CheckSSLineFlag(void);
void	NF_SpecialFX_MidtexRing(void);
void	NF_CheckForSubset(void);	// v220-16

// r3d_nodesfix.c
void	LS_CheckDoorInCorrectSubsector(void);
void	subfn_CheckNodesWithLine(sector_t *sec, sector_t *s, line_t *ldf);
int		LS_CheckDualLineSectorChoice(void);
int		LS_CheckOrphanedSectors(void);
void	LS_CheckSingleSameSidedLinesSector(void);
void	LS_CheckSpecSecSettings(void);

// r3d_nodesfixspec.c
void	LS_SNF1(void);
void	LS_NestedSectorRings(void);
void	LS_Find4RingSectorsCrossing(void);
void	LS_ChkSupportAssign(void);
void	LS_ChkShortSegAssignment(void);
void	LS_ChkLinesInSkySec(void);

// r3d_occlude.c
void	LS_CheckSkyfix(void);
void	LS_SetSSECNo_occlude(void);
void	PS_CheckSubsecOcclude(void);			// v220-16

// r3d_postcheck.c
void	LS_CheckLinesInWrongSectors(void);
void	LS_CheckMovSecWithAASHITTYTexLines(void);
void	LS_CheckOVF_MISLOTEXLINEReset(void);
void	LS_PostSetupRoutines(void);

// r3d_precheck.c
void	LS_CheckMissingTexInherited(void);
void	LS_CheckR3D_2ASectors(void);
void	LS_CheckForR3D_04Doors(void);
void	LS_CheckForR3D_04InFakedSectors(void);
void	LS_CheckForR3D_30Lines(void);
void	LS_CheckR3D_2F(void);
void	LS_CheckTex7WithR3D_03(void);
void	LS_ResetClosedFlagForRenderer(void);
void	LS_CheckZeroHeightSector(void);
void	LS_CheckR3D_3FContained(void);
void	LS_CheckUnsetR3D_1D(void);
void	LS_CheckR3D_29Loops(void);
void	LS_CheckR3D_04SpecialCase(void);
void	LS_CheckMovSecWithSingleLine(void);
void	LS_MovLogicalSecInRenderSec(void);
void	LS_PreCheck_Types(void);

// r3d_PSroutines.c
void	PS_UnwantedTextures(void);
void	PS_SetTexInhibitFsecEqualBsec(void);
void	PS_CheckUnsetLoTexSectors(void);
void	PS_CheckUnsetSupportLines(void);
void	PS_FunctionsVarious(void);
void	PS_FunctionsVarious(void);
void	PS_CheckUnsetOwnerSectors(void);
void	PS_CheckLiquidRings(void);
void	PS_CheckLowerTexRings(void);
void	PS_BoundedLiquidSectors(void);
void	PS_MiscellaneousRoutines(void);
void	PS_CheckFakeSecStructures(void);
void	PS_CheckUnsetR3D_16Lines(void);
void	PS_MidtexDoorsSpanning1SLines(void);
void	PS_UnsetSkysecFromMidtexLine(void);
void	PS_SetTranslucentWaterFlags(void);
void	PS_SectorHeightChecks(void);
void	PS_CheckUnclosedNoHILOtexSectors(void);
void	PS_CheckCTRLSpecial(void);
void	PS_CheckLiqLines(void);
void	PS_CheckGenInheritance(void);
void	PS_CheckNestedFakeMovSectors(void);
void	PS_CheckUnsetCTRLsectors(void);
void	PS_FinaliseSSECsettings(void);
void	PS_FinaliseAASHITTYLines(void);
void	PS_FinaliseMissingTexLoops(void);
void	PS_FinaliseCrossingLines(void);
void	PS_CheckEndLineTypes(void);
void	PS_CheckEndSectorTypes(void);
void	PS_CheckR3D_16Type(void);
void	PS_CheckInvalidHiWalls(void);
void	PS_CheckSSLinesAccessingMidTEX(void);
void	PS_CheckMultiSecMissing(void);
void	PS_ReplaceMissingHi(void);				// v220-7
void	PS_CheckNoHiMovSectors(void);			// v220-7
void	PS_CheckSameSidedLineRings(void);		// v220-7
void	PS_Check3LineMissingLoTexSecs(void);	// v220-9
void	PS_CheckSingleLineSsecs(void);			// v220-9
void	PS_CheckSameSecMovSectors(void);		// v220-10
void	PS_CheckPartialMissingHi(void);			// v220-13
void	PS_CheckPartialMissingLo(void);			// v220-13
void	PS_CheckLotexPreference(void);			// v220-13
void	PS_CheckMissingHeightsec(void);			// v220-16
void	PS_SelfreferencingMovCeil(void);		// v220-18
void	PS_UnsetPartialMissingTEXLoops(void);	// v220-22

// r3d_secarbiter.c
void	LS_SectorArbitrator(void);

// r3d_secbbox.c
int		LS_GetContainingSecnum(sector_t *sec, RDBOOL type);
void	LS_CheckFlatsFromOrigTEX(void);
void	LS_SetBOUNDS(void);
void	LS_CheckRaisedFloors(void);
RDBOOL subfn_CheckContainingSectorLine(line_t *ld, sector_t *csec);
sector_t*LS_GetContainingSectorOf(sector_t *sec, RDBOOL type);

// r3d_SinSceil
void	LS_CheckSecInSecCeiling(void);

// r3d_skyloops.c
void	LS_ResetUnwantedSkyfix(void);
void	LS_CheckAdjacentSkyFromSubsector(void);
void	LS_MovSkyLoops(void);
void	LS_SetSkyLoops(void);

// r3d_special.c
RDBOOL	P_CheckSpecial(int type);
RDBOOL	P_CheckLightSpecial(int type);


// r3d_ssecchk.c
void	LS_CheckBaseSSECsectors(void);

// r3d_traceLim.c
void	TL_TraceBoundary(void);

// r3d_xssec.c
void		XS_setLineTable(void);
sector_t	*XS_GetBoundary(sector_t *sec, int *ref);
int			XS_GetBoundarySecnum(sector_t *sec);
void		XS_SetLineSegs(line_t *ld);
void		XS_GetBoundsFromLine(line_t *ld, int *losec, int *hisec, int *rf);

RDBOOL	LS_SetSectorArea(sector_t *s, int fillnum, int callnum);
RDBOOL	LS_FillGenArea(sector_t *s, int fillnum, int type, int callnum);

// r3d_traceReversed.c
RDBOOL	LS_MarkSubsecsInAreaGeneral(sector_t *s, byte *array, int loopptr, int *looplist, int *bbox, int rsec, int flags);
RDBOOL	LS_MarkSubsecsInLoop(sector_t *s, int loopcnt, int *looplist, int *bbox, int flags, int *seccnt);
RDBOOL	LS_MarkSubsecsInMultiSecLoop(int chkptr, int *chksec, int loopcnt, int *bbox, int *other, RDBOOL usesidelist, int flags);
RDBOOL	LS_MarkSsecsUsingSidelist(int loopcnt, int *bbox, int flags);	// v220-10

// r3d_traceInherited.c
void	LS_TraceContainedLinesInRing(void);

// r3d_traceMissing.c b211-0
void	TM_FindFlatRingSectors(void);
void	TM_TraceMissing(void);

// r3d_traceMissingNew.cpp v220-22
void	TM2_TraceHiLoops(void);
void	TM2_CheckLoLoopIn0xCsectors(void);
void	TM2_CheckMissLoNestedInMissHi(void);
void	TM2_CheckMissHiInDoubleMovCeil(void);
void	TM2_CheckMisHiAsymetricMovCeil(void);
void	TM2_PartialMissingLoMov(void);
void	TM2_SingleMissingLowLoop(void);	// v220-33

// r3d_traceNormal.c
void	LS_TraceNormal(void);
void	LS_CheckLinkedFloors(void);

// r3d_tricks.c
void	TR_TrickChainedFloorMissing(void);
void	TR_MissingBoundTo1sLine(void);
void	TR_MovFloorInside(void);
void	TR_MovFloorContained_X(void);

// r3d_xcheck.c (extra checks)
void	LS_XcheckMissingClosedInFaked(void);
void	LS_XcheckR3D_04defs(void);
void	LS_XcheckR3D_1Bdefs(void);
void	LS_XcheckInheritanceGeneralLines(void);

extern int	vtxisonline;

extern line_t *sampleline;
extern line_t *multicrossline;

// arrays for R3D_MAINAnalyseMap

typedef struct {
	int num;		// num of subsectors the line references
	int maxnum;		// max slots allocated
	int *list;		// a list of the subsectors
} ssline_t;

extern ssline_t *sslines;

extern int		*countsec;
extern byte		*conarray;
extern byte		*excludeline;
extern byte		*hitline;
extern int		*hitssec;
extern int		*lfsecarray;
extern int		*lgrpnum;
extern int		*linearray;
extern int		*lnsec;
extern int		*sidelist;
extern short	*lineflags;
extern byte		*linetable;
extern byte		*lxarray;
extern byte		*mtarray;
extern char		*secdone;
extern char		*secmarker;
extern short	*secspflags;
extern int		*ssinitial;
extern byte		*ssecarray;
extern byte		*vtxcheck;
extern char		*vtxarray;
extern char		*vtxflags;
extern char		*vtxflags2;
extern char		*origTEX;
extern int		csbox[4];
extern byte		*asyflags;

// Shared trace vars
extern int	 hitssecptr;
extern int	 ssec_midx, ssec_midy;

#endif