//-----------------------------------------------------------------------------
//
// R3D_ARRAYS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __R3D_ARRAYS_H__
#define __R3D_ARRAYS_H__

#include "m_winmem.h"

#define MAXTDNUM		16

// v220-31 SectorHook sound support
#define SH_STOPMUSIC	 1
#define SH_MAXSOUNDS	16

#pragma pack(1)

typedef struct {
	byte cp;	// container ptr
	byte fp;	// flags
} SH_t;

#pragma pack()

extern SH_t *SHarray;
extern int	SHcontainer[SH_MAXSOUNDS];
extern int	last_player_sector;


// Entries here are set once on startup or
// must not be changed when loading a new map
typedef struct
{
	int suppress_duplicate;

	// b211-0 following added;
	int isfriendly;
	int grenade_id;	
	int grenade_tics;
	int switch_flags;
	int hasscp_defn;

	// b211-0 used as a def check for savegames
	int states_count;
	int models_count;
	int spritemodels;	// Used to override duplicate sprites
	int skincount;
	int skill;

	// v220-3
	int num_modelLight;
	int flags;

	int	use_sky_box;
	int was_water_opacity;	// v220-4
	int nojoystick;
	int nomouse;
	int devmap;
	int scp_devpu;
	int conprogress_disabled;
	float maxLightDistance;

	// v220-4 vars moved here
	int	UseList;
	int	noobjects;

	// v220-4 new vars
	int	missingpatch;
	int customhits;
	int customdamage;
	int loadlevel;
	int isDoom;
	int usettfont;
	int maxcheat;			// For map stress testing ammo can be increased 10x
	int allowCorpse;		// Load corpse models
	int amwds;
	float farclip;

	// v220-5 new vars
	int	invalidpatch;
	int noskyreset;			// If unset a tga or jpg sky can be reset
	int	clear_screen;

	// v220-6
	int	mdefflags;
	int first_map;

	// v220-7
	short	addon_flags;
	int		addon_handle;

	// v220-8
	int		xstat;			// Set with -xstat		If set extended stats info is displayed

	// v220-9
	int		nomapinfo;		// Set with -nmapinfo	If set then MAPINFO lumps will be ignored
	int		noxhair;		// Used to suppress all xhair drawing (development mode only)
	int		useSitters;		// Use Sitters models

	// v220-13
	int		rolandResetLmp;	// roland midi reset lump in Risen3D.wad
	int		usepwadmusic;	// Set with -upm pwad music overrides list
	int		useFMODsoftware;// Set with -useFMODs Less featured soundcards may need this set
	int		noHudSpriteScale;	// Hud weapon sprites are normally scaled to the window unless they need to go to the edge (e.g. HACX)
	int		replaceall;		// Set from PWADmodify with sprites ALL. Models will substitute all sprites
	int		infightsametype;// Set with -noif1 (no bullet damage with same type things)
							// and/or   -noif2 (no skull damage between skulls)
	float	jumpPower;		// Used to reset jumpPower back to original value in case changed in scripted map

	// v220-14
	int		modifier;		// Used for node selection/testing

	// v220-20
	int		usefmodmidi;	// Set with -FMODmidi Forces fmod to play midis in place of the sequencer

	// v220-21
	int		nodup;			// Set with -nodup if pwad duplicates of iwad patches should not prevent hi-res replacements

	// v220-22
	int		customdtex;		// Used to indicate a custom detail.wad has been loaded
	int		oldgamenum;		// Used to determine whether music needs to be restarted
	int		showconsistency;	// Used to show results in checkIntegrity as long as error_c0 is set

	// v220-22
	int		batman;

	// v220-27
	int		iwadid;

	// v220-29
	int		tdcount;
	int		tdnum[MAXTDNUM];

	// v220-31
	char	nextname[6];
	int		l3dreport;
	int		l3dreportc;
	int		infflags;
	int		mapset;
	int		showskytagerrors;
	int		showtptagerrors;

	// v220-33
	int		realtekCompat;
	int		arctlev;
	int		reportdamage;
	int		reduceNearClip;

	// v220-35
	int		hasdd_defns;
	int		flashintensity;
	int		flashoffset;
	int		codlev;
	int		holycrosswad;
	int		pwadplaypal;
	int		hadpwadplaypal;
	int		textbrighten;
	int		doomwad;
	int		EXAPfound;
	int		MBFfriendallow;

} geninfo_t;

// General info (values that are only set once on startup)
extern geninfo_t ge_inf;

typedef struct
{
	uint	textureidx[4096];	// MAXTEXINDEX
	byte	textype[4096];		// MAXTEXINDEX
	int		downscaleskintex;
	int		downscaledskintexcount;
	uint	graphicstexmemoryused;
	uint	skintexmemoryused;
	int		skintexcount;
	int		graphicstexcount;
} skininfo_t;

extern skininfo_t skin_inf;

// Entries here are set on a per map basis
typedef struct
{
	int scriptthings77;
	int scriptthings78;
	int scriptthings79;
	int activethinkers;
	int activemodels;		// $stresschange	number of which are to be rendered as models
	int activedropped;		// limit on number dropped
	int	isdropped;
	int architmodels;
	int projecterr1;
	int projecterr2;
	int thingheight;
	int sseclogical;
	int duallinesec;
	int orphaned;
	int emptysec;
	int taggederr;
	int setsscount;
	int lfseccount;
	int maperror;
	int seclinefixed;
	int misstexfixed;
	int yofferror;
	int called_arbitrator;
	int unrequiredtex;
	int invalidtex;
	int nogroup;
	int ovrcount;
	int dbcount[6];		// b211-0 replace rdcount, sscount
	int sscount2;
	int sscount3;
	int sslinecheck;
	int slopeerror;
	int slopecount;
	int numthinkers;
	int thinkersremoved;
	int mempointers[NUM_MCLASS];
	uint memsize;
	int	num3Dlines;
	int	reverbthings;

	// b211-0 new vars
	int badsslines;
	int boundssegfail;
	int doombuilder;
	int mapEXM8;
	int mapsides;
	int max_polynum;
	int max_ownersec[2];
	int maxseclinecount;
	int minisegs;
	int world_starttime;
	float sound_distance;

	int	noleveltitle;

	// v220-3
	int	fogwalls_set;
	int	iwadmap;
	float global_yaw;

	// v220-4
	int massacre_count;
	int collateraldamage_count;
	int modifier;
	int friendscarriedover;

	// v220-5
	int voodoodolls;
	int degenerate;

	// v220-6
	int exitlevel;		// timeout when hitting switch (case 11:)
	int nodesFormat;
	int iwadedit;
	int	tagerrors;

	// v220-8
	int	unspawned;
	int	midtexerror;
	int	plr_melee_ht;

	// v220-9
	int	hasboss;

	// v220-10
	int	hasfakewater;

	// v220-13
	int	holdstate;
	int	playerdead;

	// v220-14
	int	zdnormal;
	int	loadgame;

	// v220-16
	int	mapchange;
	int	map_lumpnum;
	int	playerdamage;
	int	playerteleport;
	int	eliminate_sectors;

	// v220-18
	int	checkPlayer;

	// v220-22
	int	changeconflicting;
	int	startlevelmusic;

	// v220-31
	int		no_light[4];		// If player is killed in a heightless sector a red filter is used
	int		noItemDrop;			// prevent items being spawned when using the massacre cheat
	int		suspendsound;		// to keep track of whether repeating sounds have been suspended or not
	uint	mobjsResurrected;	// stress info
	uint	mobjsCreated;		// stress info
	uint	mobjsRemoved;		// stress info
	uint	noscriptset;		// used to set flags from 15xxx Things or SCP_DEFN if lighting set in SCP_ZOFF
	int		useSectorLight;		// if set then uses sector light to light models instead of floorlightsec
	int		keyoverride;		// for backward script compatibility
	int		r3dskyoffsets;		// ditto

	// v220-33
	int	annumsides;

	// v220-34
	int		NOINFIGHTALL;

	// v220-35
	float	radtime;
	float	horizonOffset;
	int		checkSkyOffy;
	int		noskychange;
	int		noMovUnset;
	int		numsidecheck;
	int		ambtrigger;
	int		report18010;
	int		dogsunspawned;
	int		addlolight;
	int		forcedyn;
	int		checkforcesky;
	int		skykeep[16][2];
	int		skykeepcount;
	int		sd_err_check;
	int		friendsfound;

} maininfo_t;

extern maininfo_t ma_inf;	// map specific

#endif