//-----------------------------------------------------------------------------
//
// R3D_DEFS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
// Revision 2.1.1 build 0 07/Dec/2006 gmj (b211-0)
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __R3D_DEFS_H__
#define __R3D_DEFS_H__

//------------------------------------------------------------------------
//
// Risen3D definitions
//
//------------------------------------------------------------------------

#include "dd_types.h"

// Identifier for use with R_FindModelFile()
#define FSEARCH_MODEL	1
#define FSEARCH_SKIN	2
#define FPATH_SKIN		3

// slopetype (floor/ceiling)
#define		LOSLOPE		0
#define		HISLOPE		1

#define FIX2DBL(x)		( (x) / (double) FRACUNIT )

#define MAX_SECHEIGHT	1024 * FRACUNIT

// Used for cacheing
#define CACHE_GRAPHICS	0x01
#define CACHE_SPRITES	0x02

// Used with lineflags rev43 sep 2004
#define LFG_MTRING		0x0001	// Set if found in a ring of missing tex lines
#define LFG_NOCHANGE	0x0002	// Set if line is to be ignored by ChangeLineTypes
#define LFG_NHCONNOBSLN	0x0004	// Set up line with upper missing tex is connected to a line with no backsec
#define LFG_SEC2SET		0x0008	// Used by analyser to indicate whether secSpec2 should be used
#define LFG_LINESSECSET	0x0010	// Used when detecting lines in set subsectors
#define LFG_EXCLUDE		0x0020	// MarkLines to exclude all lines in secloop tests
#define LFG_FIXED		0x0040	// Set where type is R3D_01 and no change can be tolerated
#define LFG_DUAL		0x0080	// Set where identified as a dual loop sec

// ring analysis
#define LFG_NEWSET		0x0100	// b211-0 Set in r3d_traceMissing, r3d_traceNormal etc.
#define LFG_MARKED		0x0200	// b211-0 Set in r3d_MTex.c
#define LFG_XDIFFH		0x0400	// b211-0 Set in r3d_MAINspec
#define LFG_NOARBITER	0x0800	// b211-0 Set if to be ignored by sector arbitrator
#define LFG_BADLINE		0x1000	// b211-0 Set if line has side/sector errors
#define LFG_FIXABS		0x2000	// v220-4 Set if specType must never be changed
#define LFG_OVERRIDE	0x4000	// v220-33 Set if R3D_WADanalyse has specific nomisshi instruction
#define LFG_3DTYPE		0x8000	// v220-4 Set if a slope or 3D line

// Used to check line loop tests
#define MLM_SECLOOPS	0x01
#define MLM_TRACEGEN	0x02

// Used with trace test
#define MSIA_NONE		0x0000
#define MSIA_SKY		0x0001
#define MSIA_REVERSED	0x0002
#define MSIA_TESTINNER	0x0004
#define MSIA_TESTAREA	0x0008
#define MSIA_MSREVERSED	0x0010

// Used with sector specFlags (test with MSFC())

// N.B. These must not be renumbered (unless numerical 0x masks are changed to use the definitions)

#define	SF_HASSECLIGHT	0x00000001 // Used by floor/ceil lighting algos
#define	SF_USESECLIGHT	0x00000002 // Overrides any other settings in renderer
#define	SF_NOLOTEX		0x00000004
#define	SF_NOHITEX		0x00000008

#define	SF_SECCLOSED	0x00000010 // All lines in sector are closed
#define	SF_03_CLOSED	0x00000020 // All R3D_03 lines in sector are closed
#define	SF_04_CLOSED	0x00000040 // All R3D_04 lines in sector are closed

#define	SF_SIDETEXTRUE	0x00000080

#define	SF_HILOTEXTRUE	0x00000100
#define	SF_HITEXTRUE	0x00000200
#define	SF_SSLINETRUE	0x00000400 // Set if single sided lines found in sector
#define SF_HASMIDTEXL	0x00000800

#define SF_HASORIGLOTEX	0x00001000
#define SF_HASORIGHITEX	0x00002000
#define SF_SPARE		0x00004000
#define SF_HASTEXTURE	0x00008000 // Set if any line in the sector has a texture

#define	SF_MOVFLOOR		0x00010000 // Set if a sector can change floor height
#define	SF_MOVCEIL		0x00020000 // Set if a sector can change ceil height
#define	SF_MOVSECTOR	0x00030000 // Set when building list - used as mask

#define	SF_MOVLOWER		0x00080000 // Set if a floor only moves down - used for skyfix sprite height checking.

#define	SF_MOVINVALID	0x00100000 // Set if a sector's direction is invalid
#define	SF_MOVTYPE_X	0x00200000 // Set if line->special == 83 or 91 in a tagged sector
#define	SF_MOVFKR3D_11	0x00400000 // Set if faked and contains line->type == R3D_11
#define	SF_LINESCLOSED	0x00800000 // Set if all lines in a sector form a ring, head to tail, and where
								   // the lines other sector is always the same.

#define	SF_MASKEDDOOR	0x01000000 // Set if sector has a masked door
#define	SF_NOOCCLUDE	0x02000000 // Set if sector should not occlude with a skywall
#define	SF_MOVFIXED		0x04000000 // v220-10 set if a plat never moves
#define	SF_HEIGHSKY		0x08000000 // Set if a sky's sector ceiling height has been adjusted

#define	SF_MULTILOOP	0x10000000 // Possibly set if the sector has control lines

#define	SF_IGNORE		0x20000000 // Set if an invalid/unset bounds box
#define	SF_DUALFMOVSEC	0x40000000 // Set if a line has a different mov floor each side
#define	SF_SKYWALL		0x80000000 // Set to indicate sky walls should be drawn rev94
#define SF_TESTMASK		0x09B3BF9C
#define SF_TESTMASK2	0x0933BF9C


// R3D - Used with sector specFlags2
// v220-24 resized from short to long
#define SF2_FAKEMOVSECTOR	0x00000001		// Special fake moving floor
#define SF2_STNMOVREP		0x00000002		// Set if a plat requires the sound sfx_stnmov to be started repeatedly
#define SF2_SETFOG			0x00000004		// v220-3 Used with outside fog to determine whether sector should have fog set
#define SF2_HASWATERFLATS	0x00000008		// v220-3 Sector has a ceiling and/or floor flat water texture
#define SF2_ADJOINSWATER	0x00000010		// v220-3 Sector has a low, middle or upper watermap 
#define SF2_MOVSKYCEIL		0x00000020		// v220-3 Sector is a mov ceil tex change sector with a skyflat 
#define SF2_MOV8ABOVEFLR	0x00000040		// v220-3 Set if mov invalid and the dest is 8 above the surrounding floor
#define SF2_FAKETPSEC		0x00000080		// v220-3 Set if the sector is a fake TP type
#define SF2_NOMOVSOUND		0x00000100		// v220-3 Set if the sector should not play a mov sound
#define SF2_DUMMYFLOOR		0x00000200		// v220-5 Set floor with sector even if it has heightsec >= 0
#define SF2_NOFAKESMOOTH	0x00000400		// v220-5 Set with a floor with water mod
#define SF2_SKYSCROLLY		0x00000800		// v220-5 Flag scrolling sky
#define SF2_HASTYPE272		0x00001000		// v220-5 Special identifier
#define SF2_STNMOVCEIL		0x00002000		// v220-22 As SF2_STNMOVREP but for ceilings.
#define SF2_LOSLOPE			0x00004000		// v220-5
#define SF2_HISLOPE			0x00008000		// v220-5
#define SF2_NONEMISSMOVLO	0x00010000		// v220-24 set if a mov floor has no missing lo textures
#define SF2_ELEVATOR3D		0x00020000		// v220-31 set if a mov floor is a valid 3D elevator type activated by specials 227 to 234
#define SF2_ELEVATOR		0x00040000		// v220-31 set if a mov floor is an invalid  3D elevator type activated by specials 235 to 238
#define SF2_PSEUDOFAKE		0x00080000		// v220-31 set if using an SCP_OFF SETPSEUDOFAKE command
#define SF2_PSFAKEDAMAGE	0x00100000		// v220-31 set if damage should be applied if under a pseudofake floor
#define SF2_FAKEDHIGHMISS	0x00200000		// v220-31 set if a hightex is missing where a sector's ceiling is lower than it's heightsec ceiling
#define SF2_FAKEDLOWMISS	0x00400000		// v220-31 set if a lowtex is missing where a sector's floor is higher than it's fakesec floor
#define SF2_NEGOFFSET		0x00800000		// v220-31 set if a lowtex needs to be offset by -1
#define SF2_CEILNOSLOW		0x01000000		// v220-35 It would appear that generalised crushers of type 0x4000 - 0x5FFF do not slow whilst crushing

// Used with lsiflags (renumbered v220-33)
#define LS_LOWTEX		0x0001
#define LS_MIDTEX		0x0002	
#define LS_HITEX		0x0004
#define LS_MIDTEXH		0x0008	// v220-33 Used for high triangle
#define LS_MIDTEXL		0x0010	// v220-33 Used for low triangle
#define LS_TEXPRESENT	0x001F

#define LS_LOWALPHA		0x0020
#define LS_MIDALPHA		0x0040
#define LS_HIALPHA		0x0080
#define LS_LOWTRANS		0x0100
#define LS_MIDTRANS		0x0200
#define LS_HITRANS		0x0400

#define LS_SIDESCROLL	0x0800	// v220-33
#define LS_ORIGTEXOFF	0x1000	// v220-33
#define LS_LOSLOPEB		0x2000
#define LS_HISLOPEB		0x4000
#define LS_NOGLOW		0x8000

#define LS_FAKESOUND	0x8000	// used on setup then cleared	

// Used with asyflags
#define ASY_L			0x01
#define ASY_M			0x02
#define ASY_H			0x04

// Used with texFlags
#define	T_LOW		LS_LOWTEX
#define	T_MID		LS_MIDTEX
#define	T_HI		LS_HITEX
#define T_ALL		LS_TEXPRESENT

// Used with LS_TestLineTex
#define	LT_LO		0x0001
#define	LT_HI		0x0004

#define	LT_MISLO	0x0008
#define	LT_MISHI	0x0010

#define	LT_DIFFLO	0x0020
#define	LT_DIFFHI	0x0040


// Used with line mtflags
#define	R3D_INVLTEX1			0x01	// Invalid texture identifiers
#define	R3D_INVHTEX1			0x02
#define	R3D_INVLTEX2			0x04
#define	R3D_INVHTEX2			0x08

#define	R3D_INVLOTEX			0x05	// Masks
#define	R3D_INVHITEX			0x0A
#define	R3D_INVTEXANY			0x0F

#define	R3D_MISLTEX1			0x10	// Missing texture identifiers
#define	R3D_MISHTEX1			0x20
#define	R3D_MISLTEX2			0x40
#define	R3D_MISHTEX2			0x80

#define	R3D_MISLOTEX			0x50	// Masks
#define	R3D_MISHITEX			0xA0
#define	R3D_MISTEXANY			0xF0

// Used with line spflags
#define SPF_MIDTRANSLUCENT		0x01	// Translucent flag mid tex
#define SPF_LOWTRANSLUCENT		0x02	// Translucent flag low tex
#define SPF_NOULTEX				0x04	// No upper/lower texture flag
#define SPF_NOTEX				0x08	// No texture flag

#define SPF_LINEONLINE			0x10	// Project forwards (type 1)
#define SPF_LINEONLINE2			0x20	// Project forwards (type 2)
#define SPF_LOL_MASK			0x30	// Line on Line mask
#define SPF_DRAWTOP				0x40	// Draw a top texture 1 fracunit deep
#define SPF_CONTROLLINE			0x80	// Control line identifier

// Used with line tyflags
#define	TYF_NOMIDTEX			0x01
#define	TYF_MIDTEXLINE			0x02
#define	TYF_MIDTEXNEGY			0x04
#define	TYF_MIDTEXPOSY			0x08
#define	TYF_MIDMASK				0x0F

#define	TYF_SKYWALL				0x10	// Set after analysis completed
#define	TYF_SKYWALLCANDIDATE	0x20	// Set after analysis completed
#define	TYF_MISHILOOP			0x40	// Set if missing hi tex line is in a closed loop (used to determine rendering sector)
#define	TYF_MISLOLOOP			0x80	// Set if missing lo tex line is in a closed loop (only used to set light level)

// Used with line ovflags
#define	OVF_3DSWITCH			0x01	// Flags a 3D switch (height checked relative to the activator)
#define OVF_3DLADDERLINE		0x02
#define OVF_MAPPED				0x04	// v220-13 $rotate removed was OVF_3DROTATESTRUCT v220-15 set if ML_MAPPED set in map editor and same sector sided
#define OVF_NOSKYWALL			0x08

#define OVF_SWITCH				0x10	// Indicates line could have a switch texture
#define OVF_NOOCCLUDE			0x20	// Used for analysis
#define OVF_INVALIDS2			0x40	// Set if side2 is invalid but is later cleared so could be used for other purposes further on
#define OVF_MISHITEXLINE		0x80	// Used in the renderer

// v220-31 - used with line xtraflags
// sound types
#define XT_PLAYLSONCE			0x0001
#define XT_PLAYLSREPEAT			0x0002
#define XT_TURNLSOFF			0x0004
// lighting types
#define XT_SETLIGHT				0x0010
#define XT_TURNTORCHON			0x0020
#define XT_TURNVTXON			0x0040
#define XT_TURNSINEON			0x0080
#define XT_LIGHTFLAGS			0x00F0
// walkover trigger type
#define XT_PLRTRIGONLY			0x0100
#define XT_MBLINE				0x0200
#define XT_OCCLUDE				0x0400	// v220-33 can be used to force occlusion
#define XT_TRIGGER				0x0800	// v220-35 used to identify as the main trigger line in the ltarray
#define XT_TRIGGER1				0x1000	// v220-35 used to identify as the main trigger line in the ltarray
#define XT_FRACPROJECT			0x2000	// v220-35 used to project vertices by a FRACUNIT
#define XT_FRACPROJECTBACK		0x4000	// v220-35 used to project vertices by a FRACUNIT

// Used with lnflags2
#define	LNF2_DOORTYPE		0x00000001	// Used to identify untagged doors
#define LNF2_NOCHANGE		0x00000002	// Set if a line's type must not be changed
#define LNF2_DRAWSKY		0x00000004	// v220-6 Set if a missing uppertex and sky backceil and midtex && spans 1's lines
#define	LNF2_NORENDLOW		0x00000008	// Set if the renderer should not render a lower texture
#define	LNF2_NORENDHI		0x00000010	// Set if the renderer should not render a upper texture
#define	LNF2_RENDVIEWH		0x00000020	// Set if the renderer should use a seg's backsector in place of the front
#define LNF2_MISSINGSPEC	0x00000040	// b221-0 Set if a line's original tex is 0x55
#define	LNF2_LOWTRIWALL		0x00000080	// mid or bottom texture - used with floor slopes
#define	LNF2_HITRIWALL		0x00000100	// mid or top texture - used with ceiling slopes
#define LNF2_BLOCKALL		0x00000200	// modelswitch blocking line
#define LNF2_ALLOWSIGHT		0x00000400	// Used with a modelswitch blocking line to indicate sight is allowed
#define LNF2_LOWTEXMASKED	0x00000800	// Used by P_BulletLineOpening() to allow 'bullets' to pass 3D masked textures
#define LNF2_MIDTEXMASKED	0x00001000	// Used by P_BulletLineOpening() to allow 'bullets' to pass 3D masked textures
#define LNF2_3DNOBLOCK		0x00002000	// Special triggered R3D blocking line v220-16 renamed from LNF2_NOBLOCK
#define LNF2_AMBIENTTRIG	0x00004000	// Ambient sound trigger line
#define LNF2_SPAN1DOORS		0x00008000	// b211-0 Used in renderer to identify a masked door spanning 1's lines
#define LNF2_NOALPHA		0x00010000	// b211-0 Used in renderer to signal not build a non-alpha texture
#define LNF2_3DTELEPLINE	0x00020000	// Set if an R3D_40 textureless line coincident with an MT_TELEPORTMAN
#define LNF2_LINESOUND		0x00040000	// Set if not to be played unless a sector's moving
#define LNF2_OCCLUDE		0x00080000	// Set if line should occlude
#define LNF2_TRIGCOINCIDENT	0x00100000	// Set if line with no special is coincident with a line with a special
#define LNF2_SPARE			0x00200000	// Unused

#define LNF2_SLOPESECLN		0x00400000	// Used to stop analyser assigning an override sec
#define LNF2_BADSSECLN		0x00800000	// b211-0 Used in analyser to identify a line in a malformed ssec caused by a map error

#define	LNF2_OWNS_SWITCH	0x01000000	// b211-0 Set if the line has a line special and an upper or lower switch texture
#define	LNF2_OWNEQUALTOONE	0x02000000	// b211-0 Set if a line is two sided and both vertexes each own a single sector
#define	LNF2_OWNBOTHGTRONE	0x04000000	// b211-0 Set if both vertexes own more than one sectors
#define	LNF2_OWNEQUALTOTWO	0x08000000	// b211-0 Set if both a lines vertexes each own two sectors Note: this does not guarantee that the line is two sided
#define	LNF2_OWNGTRTHANTWO	0x10000000	// b211-0 Set if either vertex owns more than two sectors
#define	LNF2_OWNTWOANDPLUS	0x20000000	// b211-0 Set if one vertex owns two sectors and the other > 2
#define LNF2_OWNBOTHGTRTWO	0x40000000	// b211-0 Set if a line is two sided and both vertexes each own more than two sectors
#define LNF2_VTXUNMERGED	0x80000000	// b211-0 Set if a line has a single unmerged vertex

// flags re-assigned following analysis
#define	LNF2_SKY2FOGWALL	LNF2_NOCHANGE		// v220-3 Set if a line with sky sectors and no mid tex spans 1's lines
// These flags are always recalculated and do not change in game so need not be saved allowing for backward savegame compatibility
#define LNF2_SKYOCCLUDE		LNF2_MISSINGSPEC	// v220-16 set if the renderer should just draw a skywall with occlusion
#define LNF2_UNBLOCKED		LNF2_VTXUNMERGED	// v220-16 set to quickly identify a line that should not block bullets/sight
#define LNF2_REASSIGN		(LNF2_MISSINGSPEC|LNF2_VTXUNMERGED)
#define LNF2_NOSAVE 		(LNF2_MISSINGSPEC|LNF2_VTXUNMERGED|LNF2_NORENDLOW|LNF2_NORENDHI|LNF2_DOORTYPE|LNF2_SKY2FOGWALL)

// scpFlags definitions
	// stored with savegame
#define SCPF_BLOCKFRIEND	0x01
#define SCPF_PLAYERONLY		0x02
#define SCPF_LOWMASK		0x03

	// v220-9 used for analysis
#define SCPF_REPERROR		0x04	// v220-8
#define SCPF_DRAWBSEC		0x08
#define SCPF_IGNOREBS		0x10	// v220-9	Map error id - ignore backsec
#define SCPF_DRAWTOLIMITS	0x20	// v220-10	Used with horrible software renderer hack
#define SCPF_HIMASK			0xFC	// stored for analysis

// Used in r3d_traceGen.c
// Note must start at 1 NOT 0 (see ELIMINATE_SINGLELINE)
#define	FGA_SECTOR			0x0001
#define	FGA_LFSEC			0x0002
#define	FGA_SSFLOORSEC		0x0003
#define	FGA_SSCEILSEC		0x0004
#define	FGA_SSOVRSEC		0x0005

// Line crossing conditions
#define	LC_NOCONDITION		0x00000000
#define	LC_NORMAL			0x00000001
#define	LC_NOVTXONLINE		0x00000002
#define	LC_IFVTXONLINE		0x00000004
#define	LC_VTXISONLINE		0x00000008	
#define	LC_LI_CLOSED		0x00000010
#define	LC_NOT_SHARED		0x00000020
#define	LC_SAMESECLINE		0x00000040
#define	LC_DIFFSECLINE		0x00000080
#define	LC_NOTEX			0x00000100
#define	LC_SAMEWITHMID		0x00000200
#define	LC_DIFF_FLR_HT		0x00000400
#define	LC_SINGLESIDED		0x00000800
#define	LC_DIFFSECMOV		0x00001000
#define	LC_DIFFSECSTATIC	0x00002000
#define	LC_OWNSECLINE		0x00004000	// b211-0
#define	LC_VTXNOTSHARED		0x00008000	// b211-0
#define	LC_NOVTXONLINEGEN	0x00010000	// b211-0
#define	LC_VTXPR0JECTED		0x00020000	// b211-0
#define	LC_SUPERIMPOSED		0x00040000	// b211-0
#define	LC_NOTEXMOV			0x00080000	// v220-6
#define	LC_MIDTEXONLY		0x00100000	// v220-10
#define	LC_SAMESECFMOV		0x00200000	// v220-22

// Function internal only
#define	LC_TESTLINE			0x80000000

// secspflags definitions
#define SECSPF_RAISEFLR		 0x0001
#define SECSPF_TELEPTAG		 0x0002
#define SECSPF_HASACTION	 0x0004
#define SECSPF_CTRLLINES	 0x0008
#define SECSPF_LINONESEC	 0x0010
#define SECSPF_MIDTEXCLOSED	 0x0020
#define SECSPF_LNLOOPTEX	 0x0040
#define SECSPF_FLOORTEXHT	 0x0080

#define SECSPF_NOSKYFIX		 0x0100
#define SECSPF_SAMESECLINES	 0x0200
#define SECSPF_DOORSECTOR	 0x0800

#define SECSPF_MASK1		 (SECSPF_FLOORTEXHT|SECSPF_LNLOOPTEX|SECSPF_MIDTEXCLOSED|SECSPF_LINONESEC|SECSPF_CTRLLINES|SECSPF_HASACTION)
#define SECSPF_MASK2		 (SECSPF_MASK1|SECSPF_TELEPTAG|SECSPF_RAISEFLR)

// For sky from linedef support
#define PL_SKYFLAT 0x80000000


// Model switch activate type flags
#define MS_USE			0
#define MS_GUN			1
#define MS_WALKOVER		2

// Used with DoomLumpIdentify
#define DLI_ANY			0
#define DLI_GRAPHIC		1
#define DLI_SOUND		3

#define R3D_00	0x00 // Standard lines
#define R3D_01	0x01 // Textureless lines that may require specSec
#define R3D_02	0x02 // Set if a midtex's lower unpegged top has to be fixed at its line's backsector's floorheight. Weird? Damn right.
#define R3D_03	0x03 // Step lines STEPLAD1, STEPTOP etc.
#define R3D_04	0x04 // General midtex lines. Note: this line type alone also suppresses the rendering of the ceilingflat.
#define R3D_05	0x05 // Tagged sector R3D_04 lines.
#define R3D_06	0x06 // Unconnected R3D_04 lines that, if enclosed by R3D_03 lines, inherit specSec from R3D_03.
#define R3D_07	0x07 // Used to flag a map hack line that is an error to which the author is oblivious.
#define R3D_08	0x08 // b211-0 changed to mark a specific type for debug purposes
#define R3D_09	0x09 // Line in sky sector with no height one side needing a missing lotex to be drawn (rare);
#define R3D_0A	0x0a // (obsolete: set - as may inhibit something else - but otherwise not used)
		// v220-7 pressed into use for hitex that needs the ss->floorsec to set the lower height
#define R3D_0B	0x0b // Identifies possible need for use of backsec for flats (Structures overhanging R3D_03 step lines)
#define R3D_0C	0x0c // R3D_04 type with one end unconnected crossing a line - only used in map analyser.
#define R3D_0D	0x0d // Used to identify flat to be rendered where line is facing a same sided line in a different sector.
#define R3D_0E	0x0e // R3D_04 lines tagged GenFloorBase-GenCeilingBase.
#define R3D_0F	0x0f // Marks line in a sector identified as invalid.
#define R3D_10	0x10 // Trick line with midtex side 2 but no height
#define R3D_11	0x11 // As R3D_10 but with midtex side 1 and hightex side1 (also need to use ldef->sidenum[1]->sector in renderer)
#define R3D_12	0x12 // Lines with missing hitex1 and only midtex1
#define R3D_13	0x13 // Similar (but not same) to 12, and could be tied to door type operation
#define R3D_14	0x14 // Similar to R3D_16 but only affects floors
#define R3D_15	0x15 // Unused
#define R3D_16	0x16 // Identifies line where side flat selected is view height determined.

// Missing tex line types
#define R3D_17	0x17 // Similar to R3D_1F but missing walls not drawn.
#define R3D_18	0x18 // As R3D_1F but flats must be changed.
#define R3D_19	0x19 // Line where side 2 always used.
#define R3D_1A	0x1a // Line with no valid textures and single missing texture used to set flats.
#define R3D_1B	0x1b // Line with no valid textures and two missing textures on one side used to set flats.
#define R3D_1C	0x1c // Line with no valid textures and a missing texture on each side used to set flats.
#define R3D_1D	0x1d // Variant on R3D_1C where midtex is present front and back.
#define R3D_1E	0x1e // Variant on R3D_1A where a top or bottom tex on the other side from the missing tex is present
#define R3D_1F	0x1f // Variant on R3D_1A where mis tex combined with invalid tex

// Intersected and related line types
#define R3D_20	0x20 // Used first by analyser.  Hihacked at end as an R3D_16 type with floor light level change.
#define R3D_21	0x21 // Used to tile a midtexture under certain (rare) conditions	v220-13
#define R3D_22	0x22 // Unconnected line (previously R3D_01) that intersects another line.
#define R3D_23	0x23 // Unconnected line (previously R3D_01) that does not intersect any other line.
#define R3D_24	0x24 // Unused
#define R3D_25	0x25 // Used to identify lines where specSec floorheight can be > line sec's floorheight

// Special types
#define R3D_26	0x26 // v220-26 Force backfloor height of rendered basetex even if > frontceil. Only set using the linemark command in r3d_wadanalyse
#define R3D_27	0x27 // Similar to R3D_16 but unconditional
#define R3D_28	0x28 // Identifies a movsec line that can inherit one line side sector.
#define R3D_29	0x29 // Specific missing tex line type with each vertex joined to an R3D_1A line.
#define R3D_2A	0x2a // As R3D_1A but missing walls not drawn.
#define R3D_2B	0x2b // As R3D_1B but missing walls not drawn.
#define R3D_2C	0x2c // Special type that adds a sky wall poly.
#define R3D_2D	0x2d // Special doors type 2 or texture position defined by having odd numbering for heights.
#define R3D_2E	0x2e // Used to identify lines crossing for analysis.

// Temp types used in analyser
#define R3D_2F	0x2f // Used in analyser to flag that it may need further testing.
#define R3D_30	0x30 // Temporary id set in ld_missingtex.c for later analysis.
#define R3D_31	0x31 // Used to suppress further analysis in map analyser.
#define R3D_32	0x32 // As R3D_31 but drawing of walls also suppressed in renderer.
#define R3D_33	0x33 // Identifies line type in floor sectors with animated floor flats that force the subsec override value.

#define R3D_34	0x34 // Unused
#define R3D_35	0x35 // Unused v220-10
#define R3D_36	0x36 // Used as a "do nothing" version of R3D_00
#define R3D_37	0x37 // No walls drawn but skyfix run (only used in analyser).
#define R3D_38	0x38 // Don't draw missing upper walls.
#define R3D_39	0x39 // Used to indicate special missing hi tex in loop.	// rev170s
#define R3D_3A	0x3a // b211-0 Used for drawing wall tex in special cases.
#define R3D_3B	0x3b // Identifies line type used with nodes correction and/or rend specSec override.
#define R3D_3C	0x3c // Related to R3D_1A but search for containing sector with missing tex's.
#define R3D_3D	0x3d // marked to stop processing other than in LS_XcheckInheritanceR3D_3DLines.
#define R3D_3E	0x3e // R3D_03 type marked to stop all further processing.
#define R3D_3F	0x3f // Textureless line.

// 3D stuff - Note >= R3D_40 used for comparisons
#define R3D_40	0x40 // Contiguous step lines  with no flats
#define R3D_41	0x41 // Step line with unseen bounding fence with flats
#define R3D_42	0x42 // Contiguous step lines  with flats
#define R3D_43	0x43 // Step line with unseen bounding fence with flats
// R3D_44, R3D_45, R3D_46, R3D_47, R3D_48, R3D_49,	// Unused.
#define R3D_4A	0x4a // Scrolling translucent tex (set using line->tag -3).
#define R3D_4B	0x4b // Used to set lightlevels where a higher floorheight is always put on the line's backsec

// R3D_4C, R3D_4D, R3D_4E, R3D_4F	// Unused.

typedef struct
{
	int		bardrop_id;
	int		breath_id;
	int		drown_id;
	int		fakewt_id;
	int		jumpdn_id;
	int		r3dp1_id;
	int		r3dp2_id;
	int		ricoh1_id;
	int		ricoh2_id;
	int		stnmov_id;
	int		pstop_id;
	int		uwater_id;
	int		wade_id;
	int		wprdy1_id;
	int		wprdy2_id;
	int		wprdy3_id;
	int		xtrabfg_id;
	int		zoomin_id;
	int		zoomout_id;

} r3dsound_t;

extern r3dsound_t r3ds;

#endif