//-----------------------------------------------------------------------------
//
// R_DATA.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __R_DATA_H__
#define __R_DATA_H__

#include "p_data.h"
#include "m_nodepile.h"

// Types.
#ifndef DGLubyte
typedef unsigned char	DGLubyte;
#endif
#ifndef DGLunit
typedef unsigned int	DGLuint;
#endif

// siflags flags
#define SIF_VISIBLE			0x0001		// Sector is visible on this frame
#define SIF_RENDERED		0x0002		// b211-0
#define SIF_CLEARMASK		0x0003		// b211-0

// scroll flags
#define	SIF_FLOOR			0x0004		// v220-4
#define	SIF_CEILING			0x0008		// v220-4
#define	SIF_CARRY			0x0010		// v220-5
#define	SIF_WASSECRET		0x0020		// v220-15
#define	SIF_WASSEEN			0x0040		// v220-16

#define	SIF_MOVHIGH1		0x0100		// v220-22 set if a floor is adjusted using P_FindHighestFloorSurrounding
#define	SIF_MOVHIGH2		0x0200		// v220-22 set if a floor is adjusted using P_FindNextHighestFloor
#define	SIF_SOUNDBLOCK		0x0400		// v220-33 set if a sector with deaf Things is obscured with a plane

// Texture/flats flags - used with txf_flags and flflags.
#define TXF_MASKED			0x00000001
#define TXF_ANIMPIC			0x00000002

// Texture flags - used only with txf_flags.
#define TXF_HIRES			0x00000004
#define TXF_NOHIRES			0x00000008	// Signals not to be replaced with a high-res version
#define TXF_3DTEX			0x00000010
#define TXF_NOALPHACUT		0x00000020	// Signals that blend must be used
#define TXF_FLAT			0x00000040	// Signals a flat that's been added as a texture
#define TXF_NOCLAMP			0x00000080	// Signals that a texture should not be clamped
#define TXF_CLAMPS			0x00000100	// Signals that a texture can be clamped horizontally
#define TXF_SWITCH			0x00000200	// Identifies a group tex as switch
#define TXF_DELETE			0x00000400	// Identifies that tex should be deleted prior to map being loaded
#define TXF_DELMIRROR		0x00000800	// Identifies tex that should only have the top half used

// Flats flags - used only with flflags.
#define FLT_WATER			0x00000004		
#define FLT_LAVA			0x00000008		
#define FLT_SLIME			0x00000010		
#define FLT_BLOOD			0x00000020
#define FLT_NUKE			0x00000040
#define FLT_BLKWATER		0x00000080		
#define FLT_SPLASH			(FLT_WATER|FLT_LAVA|FLT_SLIME|FLT_BLOOD|FLT_NUKE|FLT_BLKWATER)

#define FLT_MASKED			0x00000100
#define FLT_FLATFROMTEX		0x00000200
#define FLT_3DMASKED		0x00000400	// Note: this flag is only used to load a separate 3D flats version
#define FLT_GLOW			0x00000800	// For lava etc, textures that glow.


// flats3D[].flags
// Note FLT_WATER to FLT_NUKE are transferred to these
// As such 0x0004 to 0x0080 must be kept available.
#define	F3D_MOVFLAGS		0x00FFF000	// move flags for each floor - only set in base flatslist[4]
#define	F3D_MOVFLAG			0x00001000	// (shifted right and or'ed for each flat stack)
#define F3D_MARKER			0x04000000	// Used to check if a list has been flagged
#define F3D_3DFLAT			0x08000000


// bounding box references
#define BOXTOP		0
#define BOXBASE		1
#define BOXLEFT		2
#define BOXRIGHT	3

#define DI_EAST			0
#define DI_NORTHEAST	1
#define DI_NORTH		2
#define DI_NORTHWEST	3
#define DI_WEST			4
#define DI_SOUTHWEST	5
#define DI_SOUTH		6
#define DI_SOUTHEAST	7
#define DI_NODIR		8

// This is the dummy mobj_t used for blockring roots.
// It has some excess information since it has to be compatible with 
// regular mobjs (otherwise the rings won't work).
typedef struct linkmobj_s
{
	thinker_t		thinker;
	fixed_t			x,y,z;	
	struct mobj_s	*next, *prev;
} linkmobj_t;

typedef struct vertexowner_s
{
	ushort	num;			// Number of owners.
	int		*list;			// Sector indices.
} vertexowner_t;

// The sector divisions list is similar to vertexowners.
// typedef struct vertexowner_s sector_divisions_t;

// Detail texture information.
typedef struct detailinfo_s {
	DGLuint	tex;
	int		width, height;
	float	strength;
	float	scale;
	float	maxdist;
} detailinfo_t;

typedef struct
{
	int		originx;	// block origin (always UL), which has already
	int		originy;	// accounted  for the patch's internal origin
	int		patch;
} texpatch_t;

// b211-0 newheight and offset removed
// v220-5 transll removed
typedef struct
{
	char	name[10];		// for switch changing, etc
	char	oldname[10];	// for loading hi-res 3D tex's
	short	width;
	short	height;
	short	patchcount;
	short	groupMarker;
	ushort	anim_group;		// Animation group (or zero).

	int		txf_flags;		// TXF_* flags.

	DGLuint	gltex;			// The texname
	DGLuint	noalphatex;		// Index of an alternative no alpha tex.
	detailinfo_t	detail;	// Detail texture information.

	texpatch_t	patches[1];	// [patchcount] drawn back to front
							//  into the cached texture
} texture_t;


typedef struct
{
	float	texscalex, texscaley;	// For DeeP method of texture scaling
	float	rendw, rendh;			// b211-0 scaled width and height used by renderer
	float	high, high2, offset;	// v220-31 for doom midtex offset bug v220-33 add high2, offset
	fixed_t loalpha, hialpha;		// v220-33 used when editing midtex alpha lines;

} texscale_t;

typedef struct
{
	byte		rgb[4];
} rgbcol_t;

typedef struct flat_s
{
	int			lump;
	int			flat_translation;
	int			flflags;
	rgbcol_t	color;
	detailinfo_t detail;		// Detail texture information.
	int			group;			// Animation group (or zero).
	int			width;
	int			height;
	DGLuint		gltex;
} flat_t;

// v220-13 wm, wh removed
typedef struct lumptexinfo_s
{
	DGLuint		ltitex;
	short		width, height;
	short		offx, offy;
	ushort		origw, origh;	// For storing original doom patch sizes
} lumptexinfo_t;

// v220-35
typedef struct {
	int		line0;	// 8096 line
	int		line1;	// 1st trigger line to start height controlling sector
	int		line2;	// 2nd trigger line to stop  height controlling sector
	int		line3;	// 2nd trigger line, if set, of linked stop height controlling sector 
//	mobj_t	*self;	// remember self as used to reset model switch of linked Thing
} linetrig_t;

#define LT_MAXCOUNT 100
extern linetrig_t ltarray[LT_MAXCOUNT];

extern vertexowner_t	*vertexowners;
extern nodeindex_t		*linelinks;
extern long				*blockmaplump;	// offsets in blockmap are from here
extern long				*blockmap;
extern int				bmapwidth, bmapheight;	// in mapblocks
extern fixed_t			bmaporgx, bmaporgy;		// origin of block map
extern linkmobj_t		*blockrings;		
extern byte				*rejectmatrix;			// NOT USED (was for fast sight rejection)
extern nodepile_t		thingnodes, linenodes;

extern	lumptexinfo_t	*lumptexinfo;
extern	int				numlumptexinfo;
extern	int				numtextures;		
extern	texture_t		**texArray;
extern	texscale_t		*texscales;
extern  int				*texturetranslation;	// for global animation
extern	int				LevelFullBright;
extern	flat_t			**flatsarray;

extern  byte	*translationtables;

void	addPnameErrorToList(char *name);	// v220-5
void	R_PrecacheLevel(int flags);
void	R_PrecacheNoAlphaTex(void);
short	R_TextureFlags(int texture);
flat_t	*R_GetFlat(int lumpnum);
RDBOOL	R_CheckFlatFlag(int lumpnum, int value);
RDBOOL	R_CheckFlatGlowFlag(int lumpnum);
int		R_GetFlatFlags(int lumpnum);
int		R_FlatNumForName(char *name);
int		R_CheckTextureNumForName(char *name);
int		R_TextureNumForName(char *name, RDBOOL skytex);
char	*R_TextureNameForNum(int num);
char	*R_LumpNameForNum(int num);
void	R_InitFlatsArray(void);
void	R_InitTextures(void);
void	R_InitLumpTexInfo(void);
void    R_InitTranslationTables(void);
void	R_SetNewTexMemory(int extrapatches, float scale, int width, int height);

#endif