//-----------------------------------------------------------------------------
//
// REND_DYN.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __REND_DYN_H__
#define __REND_DYN_H__

#include "rend_list.h"

#define DYN_ASPECT		1.1f			// 1.2f is just too round for Doom.
#define DYN_ASPECTRCP	0.90909091f		// 1/DYN_ASPECT

// Lumobj Flags.
#define LUMF_MISSILE		0x0001			// For missile lighting
#define LUMF_MISSILE1		0x0002			// For missile lighting
#define LUMF_TORCH			0x0004			// For torch light checks
#define LUMF_PLAYER			0x0008			// For player light checks

#define LUMF_NOHALO			0x0100			// Not used but left for backward compatibility reasons $halo removed

// v220-3 Common dl light factors
//#define DL_FACTOR		0.6f			// was dlFactor
#define DL_FOGFACTOR	0.36f
#define DL_FOGDIST		100*FRACUNIT

extern float	dlFactor;				// dlFactor - reinstated v220-4

typedef struct lumobj_s					// For dynamic lighting.
{
	struct lumobj_s *next;				// Next in the same DL block, or NULL.
	struct lumobj_s *ssnext;			// Next in the same subsector, or NULL.

	mobj_t	*th;
	int		flags;
	int		radius;						// Radius: lights are spheres.
	//	int		distance; $halo removed

	// v220-31 mobj z position (stored separately to allow two heights for the same Thing)
	fixed_t	z;							
	float	center; 					// Offset to center from mobj Z.
	float	xoff;
	byte	rgb[3];						// The color.
	byte	align;

} lumobj_t;

extern int		numLuminous;
//extern int		maxDynLights; $halo removed
extern int		dlMaxRad;
extern fixed_t	dlMaxRadft, dlMaxMissileRadft;
extern float	dlRadFactor;
extern int		useWallGlow, glowHeight;

// Setup.
void DL_InitBlockMap();

// Action.
void DL_InitForNewFrame(void);
void DL_ProcessSubsector(rendpoly_t *poly, subsector_t *ssec, int segcount, seg_t **segptr);

// Helpers.
RDBOOL DL_RadiusIterator(fixed_t ix, fixed_t iy, fixed_t radius, int pspr, RDBOOL (*func)(lumobj_t*,fixed_t));

#endif