//-----------------------------------------------------------------------------
//
// REND_SKY.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __REND_SKY_H__
#define __REND_SKY_H__

// Sky hemispheres.
#define SKYHEMI_UPPER		0x1
#define SKYHEMI_LOWER		0x2
#define SKYHEMI_JUST_CAP	0x4	// Just draw the top or bottom cap.
#define SKYHEMI_FADEOUT_BG	0x8	// Draw the fadeout bg when drawing the cap.

#define MAXSKYLAYERS		 2		

typedef struct
{
	float	rgb[3];				// The RGB values.
	short	set, use;			// Is this set? Should be used?
	float	limit;				// .3 by default.

} fadeout_t;

// Sky layer flags.
#define SLF_ENABLED		0x01		// Layer enabled.
#define SLF_MASKED		0x02		// Mask the layer texture.
#define SLF_SKYBOX		0x04		// v220-4 Use skybox
#define SLF_SKYFLIPY	0x08		// v220-5 Flip sky

typedef struct
{
	int			flags;	
	int			texture;		// Not the OpenGL texture.
	fixed_t		sky_offy;		// v220-5
	float		height;
	fadeout_t	fadeout;

} skylayer_t;

// v220-5 scrolling/rotating sky stuff
typedef struct {

	float	rotate_inc;
	float   rotate_angle;

} skyvar_t;

extern skyvar_t *skyvars;
extern int		numskyvar;
extern int		current_skyvarnum;
// End sky stuff

extern skylayer_t	skyLayers[MAXSKYLAYERS];
extern float		skyAngleGlobal;	// applied globally
extern float		skyAngle;

extern int		skyhemispheres;
extern int		r_skybox;
extern int		has_skybox;

// Functions:
void	Rend_RenderSky(void);
void	Init_SkyParams(void);

// v220-3 skyboxes
void	Rend_DeleteSkyAddOnTexures(RDBOOL skycaps);

#endif
