//-----------------------------------------------------------------------------
//
// S_MAIN.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __S_MAIN_H__
#define __S_MAIN_H__

#include "s_sfx.h"
#include "s_mus.h"
#include "s_sound.h"

#define SFX_RANDOM_SHIFT	0x0001	// Random frequency shift
#define SFX_RANDOM_SHIFT2	0x0002	// 2x bigger random frequency shift
#define SFX_R3DSOUND		0x0004	// Identifies lump as a Risen3D.wad entry
#define SFX_AMBIENT			0x0010	// Ambient sound
#define SFX_DONT_STOP		0x0020	// Sound can't be stopped while playing
#define SFX_LOADFLAGS		0x003F

// The following values are for internal use only
#define SFX_ISSPECIAL		0x0080	// If stopped should be restarted whena channel becomes free
#define SFX_SPARE1			0x0100	// Unused
#define SFX_SPARE2			0x0200	// Unused
#define SFX_SPARE3			0x0400	// Unused

#define SFX_MENUONLY		0x0800	// Identifies sound as a menu sound
#define SFX_REPEAT			0x1000	// Repeats until stopped
#define SFX_HALFATTN		0x2000	// Set to half attenuation
#define SFX_BADSOUND		0x4000	// Identifies invalid sound lump

#define DEF_ATTN			1.0f

#define MAX_MUS_LEN			128
#define MAX_MUS_ENTRIES		100

typedef struct			// DOOM length = 8
{
	short	three;		//0x0300    (3)
	ushort	smplrate;	//0x112B    (11025) varies in user stuff
	int		datasize;

} SOUNDHdr_t;

// v220-13 R3D sound list vars moved here from r3d_compat.h
#define MLIST_FILE	1
#define MAX_LIST_ENTRIES	109	// 99 maps + 1 (0 always clear) + 8 possible inters (see interMus);

typedef struct {
	ushort	mlist_type;
	ushort	track;
	char	*fname;
} musiclist_t;

extern musiclist_t	musicLIST[MAX_LIST_ENTRIES];

extern int	ambientfound;
extern int	ambientcount[64];

extern char	*playStatus;
extern byte	sound_info;
extern int	music_array[MAX_MUS_ENTRIES];
extern int	currentMapMusLumpnum;

RDBOOL		S_Init(void);
void		S_Shutdown(void);
void		S_Reset(void);

void		S_LevelChange(void);
void		S_StartFrame(void);

int			S_StartMusic(char *musicid);
int			S_StartMusicNum(int id);
void		S_StopMusic(void);

int			S_StartSound(int sound_id, mobj_t *origin);
int			S_StartSoundAtVolume(int sound_id, mobj_t *origin, float volume, float attenuation);
int			S_StartSoundAtVolumeNoStop(int sound_id, mobj_t *origin, float volume, float attenuation);
int			S_SoundRepeat(int sound_id, mobj_t *origin, float volume, float attenuation);
int			S_LocalSound(int sound_id, mobj_t *origin);
int			S_LocalSoundAtVolume(int sound_id, mobj_t *origin, float volume);
int			S_LocalSoundAtPos(int sound_id, float *fixedPos, float volume);
int			S_StopSound(int sound_id, mobj_t *origin);
int			S_StopAllSounds(mobj_t *origin);

int			S_LocalMenuSound(int sound_id);
int			S_LocalMenuSoundAtVolume(int sound_id, float vol);
int			S_ConsoleSound(int sound_id, mobj_t *origin, int target_console);

RDBOOL		S_CheckSoundDistance(mobj_t *origin, float *fixedPos, float attenuation);
RDBOOL		S_IsPlaying(int sound_id, mobj_t *origin);

#endif 