//-----------------------------------------------------------------------------
//
// SYS_FILE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __SYS_FILE_H__
#define __SYS_FILE_H__

#include <stdio.h>

#include "dd_types.h"

#define deof(file) ((file)->flags.eof != 0)

// Level 4: nonstandard extension used : bit field types other than int
#pragma warning (disable:4214)


typedef struct
{
	struct DFILE_flags_s
	{
		byte open  : 1;
		byte file  : 1;
		byte write : 1;
		byte eof   : 1;
	} flags; 
	int		size;
	int		handle;
	void	*data;
	char	*pos;

} DFILE;

typedef int (*f_forall_func_t)(char *fn, int parm);

void	F_InitDirec(void);
void	F_ShutdownDFILES(void);
int		F_Access(char *path);
DFILE  *F_OpenData(byte *data, uint size);
DFILE  *F_OpenFile(const char *path, const char *mymode);
DFILE  *F_Open(char *path, const char *mode);
DFILE  *F_OpenLump(int lump);
void	F_Close(DFILE *file);
int		F_Read(void *dest, int length, DFILE *file);
int		F_GetC(DFILE *file);
int		F_Tell(DFILE *file);
int		F_Seek(DFILE *file, int offset, int origin);
void	F_Rewind(DFILE *file);
int		F_ForAll(char *filespec, int parm, f_forall_func_t func);

// Functions added v220-6
DFILE  *F_OpenFileForWrite(const char *path, const char *mymode);
int		F_FileWrite(void *data, uint length, DFILE *file, int from);

#endif	// __SYS_FILE_H__