//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: hu_msg.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:  Common HUD psprite handling.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_compat.h"

#include "doom/mbf_player.h"
#include "doom/st_stuff.h"
#include "net_main.h"
#include "con_console.h"

#include "r_main.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern float	lookOffset;
extern int		customPunchRestState;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int	xydelay;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
// HU_PSpriteYOffset
//	Calculates the Y offset for the player's psprite. The offset depends
//	on the size of the game window.
//==========================================================================
int HU_PSpriteYOffset()
{
	int offy = 0;

	// If the status bar is visible, the sprite is moved up a bit.
	if(cfg.sbarscale == 5)
	{
		offy -= FRACUNIT * (ST_HTDIV2 - 1);
	}

	return offy;
}

//==========================================================================
// HU_UpdatePlayerSprite
//	Calculates the Y offset for the player's psprite.
//==========================================================================
void HU_UpdatePlayerSprite(int pnum)
{
	mbf_player_t	*pl = mbf_players + pnum;
	ddpsprite_t		*ddpsp;
	pspdef_t		*psp;
	int				i;

	//--------------------------------------

	if (!pl->plr->mo->thinker.id) return;

	for(i = 0; i < NUMPSPRITES; i++)
	{
		psp = pl->psprites + i;
		ddpsp = pl->plr->psprites + i;

		if (!psp->state) // A NULL state?
		{
			//ddpsp->sprite = -1;
			ddpsp->stateptr = 0;
			continue;
		}

		ddpsp->stateptr = psp->state;
		ddpsp->tics = psp->tics;

		// ALPHA.---------------------

		if (psp->state - states == customPunchRestState)
			ddpsp->alpha = customPunch > 10 ? 1.0f : customPunch/10.0f;		// COD support
		else if (pl->powers[pw_invisibility] > 4*32
			|| pl->powers[pw_invisibility] & 8)
		{
			ddpsp->alpha = .25f;	// partial invisibility
		}
		else
			ddpsp->alpha = 1.0f;	// fully opaque

		// INTENSITY.-----------------

		if (   psp->state->frame & FF_FULLBRIGHT
			|| pl->powers[pw_infrared] > 4*32
			|| pl->powers[pw_infrared] & 8
			|| pl->powers[pw_invulnerability] > 30)
		{
			ddpsp->plight = 1.0f;	// fullbright
		}
		else // default local light
			ddpsp->plight = (float)(ppt.viewer_lightlevel) / 255.0f;

		// OFFSETS.-------------------

		// v220-5 Use different offsets depending on whether the weapon is changing or static
		if (   mbf_players[0].pendingweapon == wp_nochange
			&& mbf_players[0].playerstate != PST_DEAD
			&& (!xydelay || !xydelay--))	// if the first test fails then so will the second. This is just to ensure xydelay is decremented
		{
			int			a, b;

			a = FixedMul((int)(FRACUNIT*cfg.bobWeapon), mbf_players[0].bob);
			b = (128 * leveltime) & FINEMASK;
			
			// Offset from center
			ddpsp->x = FIX2FLT(FRACUNIT + FixedMul(a, finecosine[b])) - lookOffset*1300;
			ddpsp->y = FIX2FLT(32*FRACUNIT + FixedMul(a, finesine[b & (FINEANGLES/2-1)]));
		}
		else
		{
			// Changeover offset
			ddpsp->x = FIX2FLT(psp->sx) - lookOffset*1300.0f;
			ddpsp->y = FIX2FLT(psp->sy);
		}
	}
}

//==========================================================================
// HU_UpdatePsprites
//	Updates the state of the player sprites (gives their data to the
//	engine so it can render them). Servers handle psprites of all players.
//==========================================================================
void HU_UpdatePsprites(void)
{
	pspOffY = HU_PSpriteYOffset() >> (FRACBITS-4);

//	for(int i = 0; i < MAXPLAYERS; i++)	// R3D Single player only
		if(mbf_players[0].plr->ingame) 
			HU_UpdatePlayerSprite(0);
}


