//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 id Software, Inc., G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: hu_stuff.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:  Heads-up displays
//
//-----------------------------------------------------------------------------


#include <string.h>
#include <ctype.h>

#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/hu_stuff.h"
#include "doom/hu_lib.h"
#include "doom/hu_msg.h"
#include "doom/Mn_def.h"
#include "doom/d_englsh.h"

#include "con_console.h"
#include "m_misc.h"
#include "s_main.h"
#include "net_main.h"
#include "con_functions.h"
#include "r3d_compat.h"
#include "r3d_mapinfo.h"

//
// Locally used constants, shortcuts.
//
#define HU_TITLEHEIGHT	1

#define HU_INPUTTOGGLE	't'
#define HU_INPUTX	HU_MSGX
#define HU_INPUTY	(HU_MSGY + HU_MSGHEIGHT*(hu_font[0].height)+1)
#define HU_INPUTWIDTH	64
#define HU_INPUTHEIGHT	1

extern int			showMessages;
extern RDBOOL		flagfault;

int subfn_chkNameStart(char *name);

RDBOOL				hu_showallfrags = false;
byte				chatchar = 0; 
static mbf_player_t	*plr;
dpatch_t			hu_font[HU_FONTSIZE];
dpatch_t			hu_font_a[HU_FONTSIZE], hu_font_b[HU_FONTSIZE];
static hu_textline_t w_title;
RDBOOL				chat_on;
static hu_itext_t	w_chat;
static RDBOOL		always_off = false;
//static char			chat_dest[MAXPLAYERS];
static hu_itext_t	w_inputbuffer[MAXPLAYERS];

static RDBOOL		message_on;
RDBOOL				message_dontmesswithme;
static RDBOOL		message_nottobemessedwith;
RDBOOL				message_noecho;

static hu_stext_t	w_message;
static int			message_counter;

static RDBOOL		headsupactive = false;

const char*	shiftxform;
/*
const char french_shiftxform[] =
{
    0,
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
    11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
    31,
    ' ', '!', '"', '#', '$', '%', '&',
    '"', // shift-'
    '(', ')', '*', '+',
    '?', // shift-,
    '_', // shift--
    '>', // shift-.
    '?', // shift-/
    '0', // shift-0
    '1', // shift-1
    '2', // shift-2
    '3', // shift-3
    '4', // shift-4
    '5', // shift-5
    '6', // shift-6
    '7', // shift-7
    '8', // shift-8
    '9', // shift-9
    '/',
    '.', // shift-;
    '<',
    '+', // shift-=
    '>', '?', '@',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '[', // shift-[
    '!', // shift-backslash - OH MY GOD DOES WATCOM SUCK
    ']', // shift-]
    '"', '_',
    '\'', // shift-`
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '{', '|', '}', '~', 127

};
*/

const char english_shiftxform[] =
{

    0,
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
    11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
    31,
    ' ', '!', '"', '#', '$', '%', '&',
    '"', // shift-'
    '(', ')', '*', '+',
    '<', // shift-,
    '_', // shift--
    '>', // shift-.
    '?', // shift-/
    ')', // shift-0
    '!', // shift-1
    '@', // shift-2
    '#', // shift-3
    '$', // shift-4
    '%', // shift-5
    '^', // shift-6
    '&', // shift-7
    '*', // shift-8
    '(', // shift-9
    ':',
    ':', // shift-;
    '<',
    '+', // shift-=
    '>', '?', '@',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '[', // shift-[
    '!', // shift-backslash - OH MY GOD DOES WATCOM SUCK
    ']', // shift-]
    '"', '_',
    '\'', // shift-`
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '{', '|', '}', '~', 127
};

char frenchKeyMap[128]=
{
    0,
    1,2,3,4,5,6,7,8,9,10,
    11,12,13,14,15,16,17,18,19,20,
    21,22,23,24,25,26,27,28,29,30,
    31,
    ' ','!','"','#','$','%','&','%','(',')','*','+',';','-',':','!',
    '0','1','2','3','4','5','6','7','8','9',':','M','<','=','>','?',
    '@','Q','B','C','D','E','F','G','H','I','J','K','L',',','N','O',
    'P','A','R','S','T','U','V','Z','X','Y','W','^','\\','$','^','_',
    '@','Q','B','C','D','E','F','G','H','I','J','K','L',',','N','O',
    'P','A','R','S','T','U','V','Z','X','Y','W','^','\\','$','^',127
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void R_SetPatch(dpatch_t *dp, char *name, int nearest);

// CODE --------------------------------------------------------------------

char ForeignTranslation(unsigned char ch)
{
    return (char)(ch < 128 ? frenchKeyMap[ch] : ch);
}

static void initStrings(char **x, int *x_idx, uint size_idx)
{
	uint i;

	for (i = 0; i < size_idx/sizeof(int); i++)
		x[i] = (x_idx[i] == -1 ? "NEWLEVEL" : GET_TXT(x_idx[i]));
}

void HU_Init(void)
{
	extern byte		*iwadpalbuffer;
    int		i;
    int		j;
    char	buffer[9];

	// Mapnames are now loaded in dd_defns

/*	if (french)
		shiftxform = french_shiftxform;
    else*/
	
	shiftxform = english_shiftxform;

    // load the heads-up fonts
    j = HU_FONTSTART;

    for (i = 0; i < HU_FONTSIZE; i++, j++)
    {
		// The original small red font.
		sprintf(buffer, "STCFN%.3d", j);
		R_SetPatch(&hu_font[i], buffer, 0);

			// Small white font.
		if (hu_font[i].lump >= 0 && ge_inf.hadpwadplaypal)
		{
			// Use pwad font
			hu_font_a[i] = hu_font[i];
		}
		else
		{
			sprintf(buffer, "FONTA%.3d", j);
			R_SetPatch(&hu_font_a[i], buffer, 0);
		}
			
		// Large (12) white font.
		sprintf(buffer, "FONTB%.3d", j);
		R_SetPatch(&hu_font_b[i], buffer, 0);

		if(hu_font_b[i].lump < 0)
		{
			if (!i) Con_ErrorD("hu_font_b[0] is missing");

			// else this character is missing.  Substitute the
			// first character as this is supposedly always found.
			memcpy(hu_font_b+i, hu_font_b, sizeof(dpatch_t));
		}
    }
}

void HU_Stop(void)
{
    headsupactive = false;
}

void HU_Start(void)
{
    int		i;
    char*	s = NULL;

    if (headsupactive)
	HU_Stop();

    plr = &mbf_players[0];
    message_on = false;
    message_dontmesswithme = false;
    message_nottobemessedwith = false;
    chat_on = false;

    // create the message widget
    HUlib_initSText(&w_message,
		    HU_MSGX, HU_MSGY, HU_MSGHEIGHT,
		    hu_font_a, 
		    HU_FONTSTART, &message_on);

    // create the map title widget
    HUlib_initTextLine(&w_title,
		       HU_TITLEX, HU_TITLEY,
		       hu_font_a, 
		       HU_FONTSTART);

	s = mapinfo->mapname;

	while(*s)
		HUlib_addCharToTextLine(&w_title, *(s++));

    // create the chat widget
    HUlib_initIText(&w_chat,
		    HU_INPUTX, HU_INPUTY,
		    hu_font_a,
		    HU_FONTSTART, &chat_on);

    // create the inputbuffer widgets
    for (i=0 ; i<MAXPLAYERS ; i++)
		HUlib_initIText(&w_inputbuffer[i], 0, 0, 0, 0, &always_off);

    headsupactive = true;
}

void HU_Drawer(void)
{
	int i, k, x, y;
	char buf[80];

	HUMsg_Drawer();
    HUlib_drawIText(&w_chat);

	if(hu_showallfrags)
	{
		for(y=8, i=0; i<MAXPLAYERS; i++, y+= 10)
		{
			sprintf(buf, "%i%s", i, i==0 ? "=" : ":");
			M_WriteText(0, y, buf);
			x = 20;
			for(k=0; k<MAXPLAYERS; k++, x+= 18)
			{
				sprintf(buf, "%i", mbf_players[i].frags[k]);			
				M_WriteText(x, y, buf);
			}
		}
	}

    if (automap_active)
	{
		int scalar = cfg.sbarscale == 5 ? 0 : cfg.sbarscale == 1 ?  32 : 19;
		// Position the map title according to sbarscale.
		w_title.y = HU_TITLEY + scalar;
		strcpy(w_title.l, mapinfo->mapname);
		w_title.len = strlen(mapinfo->mapname);
		HUlib_drawTextLine(&w_title, false);
	}
}

void HU_Erase(void)
{

    HUlib_eraseSText(&w_message);
    HUlib_eraseIText(&w_chat);
    HUlib_eraseTextLine(&w_title);
}

void HU_Ticker(void)
{
	HUMsg_Ticker();

    // tick down message counter if message is up
    if (message_counter > 0 && !--message_counter)
    {
		message_on = false;
		message_nottobemessedwith = false;
    }
	
    if(showMessages || message_dontmesswithme)
    {
		// display message if necessary
		if (   (plr->hint_msg && !message_nottobemessedwith)
			|| (plr->hint_msg &&  message_dontmesswithme))
		{
			HUMsg_Message(plr->hint_msg);
			plr->hint_msg = NULL;
			message_on = true;
			message_counter = HU_MSGTIMEOUT;
			message_nottobemessedwith = message_dontmesswithme;
			message_dontmesswithme = false;
		}
		
    } // else message_on = false;

	message_noecho = false;
}

#define QUEUESIZE		128

static char	chatchars[QUEUESIZE];
static int	head = 0;
static int	tail = 0;

RDBOOL	shiftdown = false;
int chat_to = 0;	// 0=all, 1=player 0, etc.

void HU_queueChatChar(char c)
{
}

char HU_dequeueChatChar(void)
{
    return 0;
}

int CCmdMsgRefresh(int argc, char **argv)
{
	if(chat_on) return false;
	message_on = true;
	message_counter = HU_MSGTIMEOUT;
	return true;
}

static char lastmessage[HU_MAXLINELENGTH+1];

void HU_sendMessage(char *msg)
{
	strncpy(lastmessage, msg, HU_MAXLINELENGTH);
	lastmessage[HU_MAXLINELENGTH] = 0;	// in case truncated

	if(commercial_game)
		S_LocalSound(sfx_radio, 0);
	else
		S_LocalSound(sfx_tink, 0);
}

RDBOOL HU_Responder(event_t *ev)
{	
	static RDBOOL	altdown = false;
	static int		num_nobrainers = 0;

	RDBOOL			eatkey = false;
	byte		 	c;

	//------------------------------------

	// v220-4 players in game check removed
	
    if (ev->data1 == DDKEY_RSHIFT)
    {
		shiftdown = (ev->type == ev_keydown || ev->type == ev_keyrepeat);
		return false;
    }
    
	if (ev->data1 == DDKEY_RALT)	// DDKEY_LALT == DDKEY_RALT
    {
		altdown = (ev->type == ev_keydown || ev->type == ev_keyrepeat);
		return false;
    }

	if (ev->type != ev_keydown && ev->type != ev_keyrepeat)
		return false;
	
    if (chat_on)
    {
		c = (byte)ev->data1;
	
		if (altdown)	// send a chat macro
		{
			return false;	// not supported in r3d
		}
		else
		{
			/*if (french) c = ForeignTranslation(c);*/

			if (shiftdown || (c >= 'a' && c <= 'z'))
				c = shiftxform[c];

			eatkey = HUlib_keyInIText(&w_chat, c);

			if (eatkey)
			{
				// static unsigned char buf[20]; // DEBUG
				HU_queueChatChar(c);
				
				// sprintf(buf, "KEY: %d => %d", ev->data1, c);
				//      plr->message = buf;
			}
			if (c == DDKEY_ENTER)
			{
				chat_on = false;
				if(w_chat.l.len)
				{
					HU_sendMessage(w_chat.l.l);
				}
			}
			else if (c == DDKEY_ESCAPE)
				chat_on = false;
		}
    }
	
    return eatkey;
}
