//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_ctrl.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Moved to main engine, Multiplayer menu commented out
//  Moved 'actions' array here
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:  Controls Menu
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "r3d_compat.h"

#include "doom/Mn_def.h"
#include "doom/hu_stuff.h"
#include "doom/m_menu.h"
#include "doom/SndIdx.h"
#include "doom/g_game.h"

#include "con_action.h"
#include "con_bind.h"
#include "con_console.h"
#include "s_main.h"
#include "sys_input.h"


// MACROS ------------------------------------------------------------------

// Control flags.
#define CLF_ACTION		0x1		// The control is an action (+/- in front).
#define CLF_REPEAT		0x2		// Bind down + repeat.

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL	GL_CheckWindow(int *width, int *height);
void	UI_ChangeResolution(int w, int h, RDBOOL c);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void M_DrawControlsMenu(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// These must correspond with the first entries in controls[]

action_t actions[NUM_ACTIONS+1] =
{
	{"left"		, 0},
	{"right"	, 0},
	{"forward"	, 0},
	{"backward"	, 0},
	{"strafel"	, 0},
	{"strafer"	, 0},
	{"fire"		, 0},
	{"use"		, 0},
	{"strafe"	, 0},
	{"speed"	, 0},
	{"weap1"	, 0},	// Fist.
	{"weapon2"	, 0},
	{"weap3"	, 0},	// Shotgun.
	{"weapon4"	, 0},
	{"weapon5"	, 0},
	{"weapon6"	, 0},
	{"weapon7"	, 0},
	{"weapon8"	, 0},	// Chainsaw
	{"weapon9"	, 0},	// Super sg.
	{"nextwpn"	, 0},
	{"prevwpn"	, 0},
	{"mlook"	, 0},
	{"jlook"	, 0},
	{"lookup"	, 0},
	{"lookdown"	, 0},
	{"lookcntr"	, 0},
	{"jump"		, 0},
	{"weapon1"	, 0},	// Weapon cycle 1: fist/chain saw
	{"weapon3"	, 0},	// Weapon cycle 2: shotgun/super shotgun
	{"", 0}				// Terminator.
};

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static RDBOOL SCControlConfig(int option);

typedef struct
{
	char		*command;		// The command to execute.
	int			flags;
	int			defKey;			// 
	int			defMouse;		// Zero means there is no default.
	int			defJoy;			//
} Control_t;

static Control_t controls[] =
{
	// Actions (must be first so the A_* constants can be used).
	// These must correspond with actions above and the
	// enumerated types in con_action.h

	"left",			CLF_ACTION,		DDKEY_LEFTARROW , 0, 0,
	"right",		CLF_ACTION,		DDKEY_RIGHTARROW, 0, 0,
	"forward",		CLF_ACTION,		DDKEY_UPARROW   , 0, 0,
	"backward",		CLF_ACTION,		DDKEY_DOWNARROW , 0, 0,
	"strafel",		CLF_ACTION,		','			, 0, 0,
	"strafer",		CLF_ACTION,		'.'			, 0, 0,
	"fire",			CLF_ACTION,		DDKEY_RCTRL , 1, 1,
	"use",			CLF_ACTION,		' '			, 0, 4,
	"strafe",		CLF_ACTION,		DDKEY_RALT  , 3, 2,
	"speed",		CLF_ACTION,		DDKEY_RSHIFT, 0, 3,

	"weap1",		CLF_ACTION,		0			, 0, 0,
	"weapon2",		CLF_ACTION,		'2'			, 0, 0,
	"weap3",		CLF_ACTION,		0			, 0, 0,
	"weapon4",		CLF_ACTION,		'4'			, 0, 0,
	"weapon5",		CLF_ACTION,		'5'			, 0, 0,
	"weapon6",		CLF_ACTION,		'6'			, 0, 0,
	"weapon7",		CLF_ACTION,		'7'			, 0, 0,
	"weapon8",		CLF_ACTION,		'8'			, 0, 0,
	"weapon9",		CLF_ACTION,		'9'			, 0, 0,
	"nextwpn",		CLF_ACTION,		']'			, 0, 0,
	"prevwpn",		CLF_ACTION,		'['			, 0, 0,

	"mlook",		CLF_ACTION,		'm'			, 0, 0,
	"jlook",		CLF_ACTION,		'j'			, 0, 0,
	"lookup",		CLF_ACTION,		0			, 0, 6,
	"lookdown",		CLF_ACTION,		0			, 0, 7,
	"lookcntr",		CLF_ACTION,		DDKEY_DEL	, 0, 0,
	"jump",			CLF_ACTION,		'd'			, 0, 0,

	// Two weapons per key
	"weapon1",		CLF_ACTION,		'1'			, 0, 0,	
	"weapon3",		CLF_ACTION,		'3'			, 0, 0,

	// end of actions correspondance

	"zoom",			0,				DDKEY_PGUP	, 0, 0,	/*29*/

	// Menu hotkeys (default: F1 - F12).
	"HelpScreen",	0,				DDKEY_F1	, 0, 0,

	"SaveGame",		0,				DDKEY_F2	, 0, 0,
	"LoadGame",		0,				DDKEY_F3	, 0, 0,
	"SoundMenu",	0,				DDKEY_F4	, 0, 0,
	"QuickSave",	0,				DDKEY_F6	, 0, 0,
	"EndGame",		0,				DDKEY_F7	, 0, 0,
	"ToggleMsgs",	0,				DDKEY_F8	, 0, 0,
	"QuickLoad",	0,				DDKEY_F9	, 0, 0,
	"Quit",			0,				DDKEY_F10	, 0, 0,
	"ToggleGamma",	0,				DDKEY_F11	, 0, 0,

	// Screen controls.
	"sbcycle -",	CLF_REPEAT,		'-'			, 0, 0,
	"sbcycle +",	CLF_REPEAT,		'='			, 0, 0,

	// Misc.
	"screenshot",	0,				DDKEY_F5	, 0, 0,
	"automap",		0,				DDKEY_TAB	, 0, 0,
	"torch",		0,				't'			, 0, 0,
	"walkrun",		0,				DDKEY_CAPSLOCK, 0, 0,
	"",				0,				0, 0, 0		// terminator
};

static Control_t *grabbing = NULL;


static MenuItem_t ControlsItems[] =
{
	// Page 1 (modulus 16)
	{ ITT_EMPTY, "PLAYER ACTIONS", NULL, 0 },
	{ ITT_EFUNC, "LEFT :"			, (void (__cdecl *)(int))SCControlConfig, A_TURNLEFT		},
	{ ITT_EFUNC, "RIGHT :"			, (void (__cdecl *)(int))SCControlConfig, A_TURNRIGHT		},
	{ ITT_EFUNC, "FORWARD :"		, (void (__cdecl *)(int))SCControlConfig, A_FORWARD		},
	{ ITT_EFUNC, "BACKWARD :"		, (void (__cdecl *)(int))SCControlConfig, A_BACKWARD		},
	{ ITT_EFUNC, "STRAFE LEFT :"	, (void (__cdecl *)(int))SCControlConfig, A_STRAFELEFT		},
	{ ITT_EFUNC, "STRAFE RIGHT :"	, (void (__cdecl *)(int))SCControlConfig, A_STRAFERIGHT	},
	{ ITT_EFUNC, "FIRE :"			, (void (__cdecl *)(int))SCControlConfig, A_FIRE			},
	{ ITT_EFUNC, "USE :"			, (void (__cdecl *)(int))SCControlConfig, A_USE			},	
	{ ITT_EFUNC, "JUMP : "			, (void (__cdecl *)(int))SCControlConfig, A_JUMP			},
	{ ITT_EFUNC, "STRAFE :"			, (void (__cdecl *)(int))SCControlConfig, A_STRAFE			},
	{ ITT_EFUNC, "SPEED :"			, (void (__cdecl *)(int))SCControlConfig, A_SPEED			},
	{ ITT_EFUNC, "LOOK UP :"		, (void (__cdecl *)(int))SCControlConfig, A_LOOKUP			},
	{ ITT_EFUNC, "LOOK DOWN :"		, (void (__cdecl *)(int))SCControlConfig, A_LOOKDOWN		},
	{ ITT_EFUNC, "LOOK CENTER :"	, (void (__cdecl *)(int))SCControlConfig, A_LOOKCENTER		},
	{ ITT_EFUNC, "MOUSE LOOK :"		, (void (__cdecl *)(int))SCControlConfig, A_MLOOK			},

	// Page 2
	{ ITT_EMPTY, "MORE ACTIONS", NULL, 0 },
	{ ITT_EFUNC, "JOYSTICK LOOK :"	, (void (__cdecl *)(int))SCControlConfig, A_JLOOK			},
	{ ITT_EFUNC, "NEXT WEAPON :"	, (void (__cdecl *)(int))SCControlConfig, A_NEXTWEAPON		},
	{ ITT_EFUNC, "PREV WEAPON :"	, (void (__cdecl *)(int))SCControlConfig, A_PREVIOUSWEAPON	},
	{ ITT_EFUNC, "FIST/CHAINSAW :"	, (void (__cdecl *)(int))SCControlConfig, A_WEAPONCYCLE1	},
	{ ITT_EFUNC, "FIST :"			, (void (__cdecl *)(int))SCControlConfig, A_WEAPON1		},
	{ ITT_EFUNC, "CHAINSAW :"		, (void (__cdecl *)(int))SCControlConfig, A_WEAPON8		},
	{ ITT_EFUNC, "PISTOL :"			, (void (__cdecl *)(int))SCControlConfig, A_WEAPON2		},
	{ ITT_EFUNC, "SHOTGUN/SUPER SG :",(void (__cdecl *)(int))SCControlConfig, A_WEAPONCYCLE2	},
	{ ITT_EFUNC, "SHOTGUN :"		, (void (__cdecl *)(int))SCControlConfig, A_WEAPON3		},
	{ ITT_EFUNC, "SUPER SHOTGUN :"	, (void (__cdecl *)(int))SCControlConfig, A_WEAPON9		},
	{ ITT_EFUNC, "CHAINGUN :"		, (void (__cdecl *)(int))SCControlConfig, A_WEAPON4		},
	{ ITT_EFUNC, "ROCKET LAUNCHER :", (void (__cdecl *)(int))SCControlConfig, A_WEAPON5		},
	{ ITT_EFUNC, "PLASMA RIFLE :"	, (void (__cdecl *)(int))SCControlConfig, A_WEAPON6		},
	{ ITT_EFUNC, "BFG 9000 :"		, (void (__cdecl *)(int))SCControlConfig, A_WEAPON7		},

	// The remainder point to Control_t using NUM_ACTIONS + x
	// but are not used to index action_t 
	{ ITT_EFUNC, "TOGGLE ZOOM :"	, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS		},

	// Page 3
	{ ITT_EMPTY, "MENU HOTKEYS", NULL, 0 },
	{ ITT_EFUNC, "HELP :"			, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  1	},
	{ ITT_EFUNC, "SAVE GAME :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  2	},
	{ ITT_EFUNC, "LOAD GAME :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  3	},
	{ ITT_INERT, "SOUND MENU :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  4	},
	{ ITT_EFUNC, "QUICK SAVE :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  5	},
	{ ITT_EFUNC, "END GAME :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  6	},
	{ ITT_EFUNC, "QUICK LOAD :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  7	},
	{ ITT_EFUNC, "MESSAGES ON/OFF:"	, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  8	},
	{ ITT_EFUNC, "QUIT :"			, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS +  9	},
	{ ITT_EFUNC, "GAMMA CORRECTION :",(void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 10	},
	{ ITT_EMPTY,  NULL, NULL, 0 },
	{ ITT_EMPTY, "SCREEN", NULL, 0 },
	{ ITT_EFUNC, "CYCLE STATBAR+ :"	, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 11	},
	{ ITT_EFUNC, "CYCLE STATBAR- :"	, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 12	},
	{ ITT_EMPTY,  NULL, NULL, 0 },

	// Page 4
	{ ITT_EMPTY, "MISCELLANEOUS", NULL, 0 },
	{ ITT_EFUNC, "SCREENSHOT :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 13	},
	{ ITT_EFUNC, "AUTOMAP :"		, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 14	},
	{ ITT_EFUNC, "TORCH TOGGLE :"	, (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 15	},
	{ ITT_EFUNC, "WALK-RUN TOGGLE :", (void (__cdecl *)(int))SCControlConfig, NUM_ACTIONS + 16	}
};

// This is public
Menu_t ControlsDef =
{
	32, 40,
	M_DrawControlsMenu,
	53, ControlsItems,
	1, MENU_OPTIONS,
	hu_font_a, //1, 0, 0, 
	LINEHEIGHT_A,
	0, 16
};

// CODE --------------------------------------------------------------------

static RDBOOL SCControlConfig(int option)
{
	if(grabbing != NULL) 
		Con_ErrorD("SCControlConfig: grabbing is not NULL");
	
	grabbing = controls + option;
	return true;
}

static void spacecat(char *str, const char *catstr)
{
	// If str is not empty append a space
	if (str[0]) strcat(str, " ");

	strcat(str, catstr);	// append catstr
}

static void M_DrawControlsMenu(void)
{
	int			i;
	char		controlCmd[80];
	char		buff[80], prbuff[80], *token;
	Menu_t		*menu = &ControlsDef;
	MenuItem_t	*item = menu->items + menu->firstItem;
	Control_t	*ctrl;

	M_DrawTitle("CONTROLS", menu->y-28);
	sprintf(buff, "PAGE %i/%i", menu->firstItem/menu->numVisItems+1,
		menu->itemCount/menu->numVisItems+1);
	M_WriteText2(160-M_StringWidth(buff, hu_font_a)/2, 
		menu->y-14, buff, hu_font_a, 1, 0, 0);

	for(i=0; i<menu->numVisItems && menu->firstItem+i < menu->itemCount; i++, item++)
	{
		if(item->type == ITT_EMPTY) continue;
		
		ctrl = controls + item->option;
		strcpy(buff, "");
		if(ctrl->flags & CLF_ACTION)
			sprintf(controlCmd, "+%s", ctrl->command);
		else
			strcpy(controlCmd, ctrl->command);
		// Let's gather all the bindings for this command.
		if(!B_BindingsForCommand(controlCmd, buff))
			strcpy(buff, "NONE");
	
		// Now we must interpret what the bindings string says.
		// It may contain characters we can't print.
		strcpy(prbuff, "");
		token = strtok(buff, " ");

		while(token)
		{
			if(token[0] == '+')
				spacecat(prbuff, token+1);
			if(token[0] == '*' && !(ctrl->flags & CLF_REPEAT) || token[0] == '-')
				spacecat(prbuff, token);
			token = strtok(NULL, " ");
		}

		_strupr(prbuff);

		if(grabbing == ctrl)
		{
			// We're grabbing for this control.
			spacecat(prbuff, "...");
		}

		M_WriteText2(menu->x+134, menu->y + i*menu->itemHeight, 
			prbuff, hu_font_a, 1, 1, 1);
	}
}

// Set default bindings for unbound Controls.
void D_DefaultBindings()
{
	int			i;
	Control_t	*ctr;
	char		evname[80], cmd[256], buff[256];
	event_t		event;
	
	// Check all Controls.
	for(i=0; controls[i].command[0]; i++)
	{
		ctr = controls + i;
		// If this command is bound to something, skip it.
		sprintf(cmd, "%s%s", ctr->flags & CLF_ACTION? "+" : "",
			ctr->command);
		if(B_BindingsForCommand(cmd, buff)) continue;

		// This Control has no bindings, set it to the default.
		sprintf(buff, "\"%s\"", ctr->command);
		if(ctr->defKey)
		{
			event.type = ev_keydown;
			event.data1 = ctr->defKey;
			B_EventBuilder(evname, &event, false);
			sprintf(cmd, "%s %s %s", ctr->flags & CLF_REPEAT? "safebindr" : "safebind",
				evname+1, buff);
			Con_Execute(cmd, true);
		}
		if(ctr->defMouse)
		{
			event.type = ev_mousebdown;
			event.data1 = 1 << (ctr->defMouse-1);
			B_EventBuilder(evname, &event, false);
			sprintf(cmd, "%s %s %s", ctr->flags & CLF_REPEAT? "safebindr" : "safebind",
				evname+1, buff);
			Con_Execute(cmd, true);
		}
		if(ctr->defJoy)
		{
			event.type = ev_joybdown;
			event.data1 = 1 << (ctr->defJoy-1);
			B_EventBuilder(evname, &event, false);
			sprintf(cmd, "%s %s %s", ctr->flags & CLF_REPEAT? "safebindr" : "safebind",
				evname+1, buff);
			Con_Execute(cmd, true);
		}
	}
}

//---------------------------------------------------------------------------
//
// FUNC D_PrivilegedResponder
//
//---------------------------------------------------------------------------

static int findtoken(char *string, char *token, char *delim)
{
	char *ptr = strtok(string, delim);
	while(ptr)
	{
		if(!_stricmp(ptr, token)) return true;
		ptr = strtok(NULL, delim);
	}
	return false;
}

int D_PrivilegedResponder(event_t *event)
{
	// We're interested in key or button down events.
	if (   grabbing
		&& (event->type == ev_keydown
		||  (mouseInited && event->type == ev_mousebdown) 
		||  (joyInited   && (event->type == ev_joybdown || event->type == ev_povdown))) )
	{
		// We'll grab this event.
		char	cmd[256], buff[256], evname[80];
		RDBOOL	del = false;

		// Check for a cancel.
		if (event->type == ev_keydown)
		{
			if (event->data1 == DDKEY_ESCAPE || event->data1 == DDKEY_F4)
			{
				grabbing = NULL;
				return true;
			}
		}

		// We shall issue a silent console command, but first we need
		// a textual representation of the event.
		B_EventBuilder(evname, event, false); // "Deconstruct" into a name.

		// If this binding already exists, remove it.
		sprintf(cmd, "%s%s", grabbing->flags & CLF_ACTION? "+" : "",
													grabbing->command);

		if(B_BindingsForCommand(cmd, buff))
		{
			if(findtoken(buff, evname, " "))		// Get rid of it?
			{
				del = true;
				strcpy(buff, "");
			}
		}

		if(!del) sprintf(buff, "\"%s\"", grabbing->command);

		sprintf(cmd, "%s %s %s", grabbing->flags & CLF_REPEAT? "bindr" : "bind",
			evname+1, buff);

		Con_Execute(cmd, false);

		// We've finished the grab.
		grabbing = NULL;
		S_LocalSound(sfx_pistol, NULL);

		return true;
	}

	// v220-3 check for a change in window size
	if (   cMode.w_checkwin < 0
		&& !paused
		&& !grabbing
		&& event->type != ev_mousebdown)
	{
		int	w, h;

		if (GL_CheckWindow(&w, &h))
			UI_ChangeResolution(w, h, false);
		cMode.w_checkwin = 1;
		return true;
	}

	return false;
}
