//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_actor.c
// Revision v2.1.0 build 62 04/Oct/2006 - gmj
//  moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code
//
// Common code relating to actors, or "monsters".
// Actor movement smoothing; the "Servo".
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "doom/p_local.h"
#include "doom/p_start.h"

// MACROS ------------------------------------------------------------------

#define MIN_STEP	((10 * (ANG45/45)) >> 16) // degrees per tic
#define MAX_STEP	(ANG90 >> 16)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------
//==========================================================================
// P_SRVOAngleTicker
//	Turn visual angle towards real angle. An engine cvar controls whether
//	the visangle or the real angle is used in rendering.
//	Real-life analogy: angular momentum (you can't suddenly just take a
//	90 degree turn in zero time).
//==========================================================================
void P_SRVOAngleTicker(mobj_t *mo)
{
	short target, step, diff;
	int lstep, hgt;

	// Check requirements.
	if (mo->flags & MF_MISSILE || !(mo->flags & MF_COUNTKILL))
	{
		mo->visangle = mo->angle >> 16;
		return;	// This is not for us.
	}

	target = mo->angle >> 16;
	diff = target - mo->visangle;

	if(mo->turntime)
	{
		if(mo->tics)
			step = abs(diff) / mo->tics;
		else
			step = abs(diff);
		if(!step) step = 1;
	}
	else
	{
		// Calculate a good step size. Take thing height and diff into account.
		hgt = mo->height >> 16;
		if(hgt < 30) hgt = 30;
		if(hgt > 60) hgt = 60;
		lstep = abs(diff) * 8 / hgt;
		if(lstep < MIN_STEP) lstep = MIN_STEP;
		if(lstep > MAX_STEP) lstep = MAX_STEP;
		step = lstep;
	}
	
	// Do the step.
	if(abs(diff) <= step)
		mo->visangle = target;
	else if(diff > 0)
		mo->visangle += step;
	else if(diff < 0)
		mo->visangle -= step;
}

//===========================================================================
// P_UpdateAlpha
//	Sets the mobj's selector's alpha bits according to the alpha level.
//  Note: never called with dynlight mobjs
//===========================================================================
void P_UpdateAlpha(mobj_t *mo)
{
	// Clear the highest three bits (the alpha bits).
	mo->selector &= 0x1fffffff;
	mo->selector |= (mo->inverseAlpha >> 5) << 29;
}

//===========================================================================
// P_UpdateHealthBits
//	The first three bits of the selector special byte contain a
//	relative health level.
//===========================================================================
void P_UpdateHealthBits(mobj_t *mobj)
{
	int i;

	if (mobj->info && mobj->info->spawnhealth > 0)
	{
		mobj->selector &= DDMOBJ_SELECTOR_MASK; // Clear high byte.
		i = (mobj->health << 3) / mobj->info->spawnhealth;
		if(i > 7) i = 7;
		if(i < 0) i = 0;
		mobj->selector |= i << DDMOBJ_SELECTOR_SHIFT;
	}
}
