//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 id Software, Inc., G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: wi_stuff.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Reformatted for sanity's sake, Moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 6 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:   Intermission screens.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#include "r3d_compat.h"

#include "doom/mbf_player.h"
#include "doom/wi_stuff.h"
#include "doom/g_game.h"
#include "doom/Mn_def.h"
#include "doom/hu_stuff.h"

#include "con_console.h"
#include "gl_draw.h"
#include "dd_main.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "s_main.h"
#include "net_main.h"
#include "sys_system.h"

#include "r3d_mapinfo.h"

// MACROS ------------------------------------------------------------------

#define NUM_TEAMS		4	// Color = team.

//
// Data needed to add patches to full screen intermission pics.
// Patches are statistics messages, and animations.
// Loads of by-pixel layout and placement, offsets etc.
//

//
// Different vetween registered DOOM (1994) and
//  Ultimate DOOM - Final edition (retail, 1995?).
// This is supposedly ignored for commercial
//  release (aka DOOM II), which had 34 maps
//  in one episode. So there.
#define NUMEPISODES		9
#define NUMMAPS			9

// GLOBAL LOCATIONS
#define WI_TITLEY		2
#define WI_SPACINGY    	33

// SINGPLE-PLAYER STUFF
#define SP_STATSX		50
#define SP_STATSY		50
#define SP_TIMEX		16
#define SP_TIMEY		(SCREENHEIGHT-32)

// NET GAME STUFF
#define NG_STATSY		50
#define NG_SPACINGX    	64
#define NG_STATSX		(NG_SPACINGX + star.width/2)

// DEATHMATCH STUFF
#define DM_MATRIXX		42
#define DM_MATRIXY		68
#define DM_SPACINGX		40
#define DM_TOTALSX		269
#define DM_KILLERSX		10
#define DM_KILLERSY		100
#define DM_VICTIMSX    	5
#define DM_VICTIMSY		50

#define FB 0

// States for single-player
#define SP_KILLS		0
#define SP_ITEMS		2
#define SP_SECRET		4
#define SP_FRAGS		6 
#define SP_TIME			8 
#define SP_PAR			ST_TIME
#define SP_PAUSE		1

// in seconds
#define SHOWNEXTLOCDELAY	4

// TYPES -------------------------------------------------------------------

typedef enum
{
    ANIM_ALWAYS,
    ANIM_RANDOM,
    ANIM_LEVEL
} animenum_t;

typedef struct
{
    int		x;
    int		y;
} point_t;

//
// Animation.
//
typedef struct
{
    animenum_t	type;

    int		period;		// period in tics between animations
    int		nanims;		// number of animation frames

    point_t	loc;		// location of animation

    int		data1;	// ALWAYS: n/a, RANDOM: period deviation (<256), LEVEL: level
    int		data2;	// ALWAYS: n/a, RANDOM: random base period,      LEVEL: n/a

    dpatch_t	p[3];	// actual graphics for frames of animations

    // following must be initialized to zero before use!
    int		nexttic;	// next value of bcnt (used in conjunction with period)
    int		lastdrawn;	// last drawn animation frame
    int		ctr;		// next frame number to animate
    int		state;		// used by RANDOM and LEVEL when animating

} wianim_t;

typedef struct
{
	int members;			// 0 if team not present.
	int frags[NUM_TEAMS];
	int totalfrags;			// Kills minus suicides.
	int items;
	int kills;
	int secret;
} teaminfo_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	R_DrawMapTitle(char *lname, char *msg);
void	R_DrawScriptedMapInf(int lh, float alpha, RDBOOL completed);
void	R_SetPatch(dpatch_t *dp, char *name, int nearest);
void	ST_SetPatch(dpatch_t *dp, int nearest);
void	GL_SetFlat(int idx);
void	subfn_DrawBackground(void);
int		FI_Debriefing(char *levelid, RDBOOL runCast);


// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

int		wimap_pic = -1;
RDBOOL	wimap_el;

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	epihaspwad[10];	// v220-6

// PUBLIC DATA DEFINITIONS -------------------------------------------------

draw_lumpnum_t	dr_lnum;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static teaminfo_t teaminfo[NUM_TEAMS];	

static point_t lnodes[NUMEPISODES][NUMMAPS] =
{
    // Episode 0 World Map
    {
	{ 185, 164 },	// location of level 0 (CJ)
	{ 148, 143 },	// location of level 1 (CJ)
	{  69, 122 },	// location of level 2 (CJ)
	{ 209, 102 },	// location of level 3 (CJ)
	{ 116,  89 },	// location of level 4 (CJ)
	{ 166,  55 },	// location of level 5 (CJ)
	{  71,  56 },	// location of level 6 (CJ)
	{ 135,  29 },	// location of level 7 (CJ)
	{  71,  24 }	// location of level 8 (CJ)
    },

    // Episode 1 World Map should go here
    {
	{ 254,  25 },	// location of level 0 (CJ)
	{  97,  50 },	// location of level 1 (CJ)
	{ 188,  64 },	// location of level 2 (CJ)
	{ 128,  78 },	// location of level 3 (CJ)
	{ 214,  92 },	// location of level 4 (CJ)
	{ 133, 130 },	// location of level 5 (CJ)
	{ 208, 136 },	// location of level 6 (CJ)
	{ 148, 140 },	// location of level 7 (CJ)
	{ 235, 158 }	// location of level 8 (CJ)
    },

    // Episode 2 World Map should go here
    {
	{ 156, 168 },	// location of level 0 (CJ)
	{  48, 154 },	// location of level 1 (CJ)
	{ 174,  95 },	// location of level 2 (CJ)
	{ 265,  75 },	// location of level 3 (CJ)
	{ 130,  48 },	// location of level 4 (CJ)
	{ 279,  23 },	// location of level 5 (CJ)
	{ 198,  48 },	// location of level 6 (CJ)
	{ 140,  25 },	// location of level 7 (CJ)
	{ 281, 136 }	// location of level 8 (CJ)
    }
};

static short epsdSet[] = { 1, 1, 1, 0, 0, 0, 0, 0, 0 };
static short animSet[] = { 1, 1, 1, 0, 0, 0, 0, 0, 0 };
static int   wpicSet[] = {-1,-1,-1,-1,-1,-1,-1,-1,-1 };

//
// Animation locations for episode 0 (1).
// Using patches saves a lot of space,
//  as they replace 320x200 full screen frames.
//
static wianim_t epsd0animinfo[] =
{
    { ANIM_ALWAYS, TICRATE/3, 3, { 224, 104 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 184, 160 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 112, 136 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  72, 112 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  88,  96 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  64,  48 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 192,  40 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 136,  16 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  80,  16 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  64,  24 } }
};

static wianim_t epsd1animinfo[] =
{
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 1 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 2 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 3 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 4 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 5 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 6 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 7 },
    { ANIM_LEVEL, TICRATE/3, 3, { 192, 144 }, 8 },
    { ANIM_LEVEL, TICRATE/3, 1, { 128, 136 }, 8 }
};

static wianim_t epsd2animinfo[] =
{
    { ANIM_ALWAYS, TICRATE/3, 3, { 104, 168 } },
    { ANIM_ALWAYS, TICRATE/3, 3, {  40, 136 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 160,  96 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 104,  80 } },
    { ANIM_ALWAYS, TICRATE/3, 3, { 120,  32 } },
    { ANIM_ALWAYS, TICRATE/4, 3, {  40,   0 } }
};

static int numAnims[NUMEPISODES] =
{
    sizeof(epsd0animinfo)/sizeof(wianim_t),
    sizeof(epsd1animinfo)/sizeof(wianim_t),
    sizeof(epsd2animinfo)/sizeof(wianim_t)
};

static wianim_t *anims[NUMEPISODES] =
{
    epsd0animinfo,
    epsd1animinfo,
    epsd2animinfo
};

// used to accelerate or skip a stage
static int		acceleratestage;

// wbs->pnum
static int		me;
static int		myteam;

 // specifies current state
static stateenum_t	state;

// contains information passed into intermission
static wbstartstruct_t	*wbs;

static wbplayerstruct_t *plrs;  // wbs->plyr[]

// used for general timing
static int 		cnt;  

// used for timing of background animation
static int 		bcnt;

// signals to refresh everything for one frame
static int 		firstrefresh;

// Set true if titlepic drawn as background
static int		isCompleted;

static int		cnt_kills[NUM_TEAMS];
static int		cnt_items[NUM_TEAMS];
static int		cnt_secret[NUM_TEAMS];
static int		cnt_time;
static int		cnt_par;
static int		cnt_pause;

//	GRAPHICS patches

//static dpatch_t		bg;		// background (map of levels).
static dpatch_t		yah[2];		// You Are Here graphic
static dpatch_t		splat;		// splat
static dpatch_t		percent;	// %, : graphics
static dpatch_t		colon;
static dpatch_t		grnum[10];	// 0-9 graphic
static dpatch_t		wiminus;	// minus sign
static dpatch_t		finished;	// "Finished!" graphics
static dpatch_t		entering;	// "Entering" graphic
static dpatch_t		sp_secret;	// "secret"

 // "Kills", "Scrt", "Items", "Frags"
static dpatch_t		kills;
static dpatch_t		secret;
static dpatch_t		items;
static dpatch_t		frags;

// Time sucks.
static dpatch_t		time;
static dpatch_t		par;
static dpatch_t		sucks;

// "killers", "victims"
static dpatch_t		killers;
static dpatch_t		victims; 

// "Total", your face, your dead face
static dpatch_t		total;
static dpatch_t		star;
static dpatch_t		bstar;

// These need only be dimensioned to size NUM_TEAMS - but left as is
static dpatch_t		 p[MAXPLAYERS];	// "red  P[1..MAXPLAYERS]"
static dpatch_t		bp[MAXPLAYERS];	// "gray P[1..MAXPLAYERS]"

// CODE --------------------------------------------------------------------

// Draws "<Levelname> FINISHED" or "<Levelname> COMPLETED"
void WI_drawLF(void)
{
	if (isCompleted)
		R_DrawMapTitle(analysename, "COMPLETED");
	else
		R_DrawMapTitle(mapinfo->mapname, "FINISHED");
}

void WI_drawEL()
{
	if (wminfo.nextmapid[0])
		R_DrawMapTitle(wminfo.nextmapid, "ENTERING");
}

void WI_drawOnLnode(int n, dpatch_t *c)
{
	int		i = 0, left, top, right, bottom;
	RDBOOL	fits = false;
	
	do
	{
		left   = lnodes[wbs->epsd][n].x - c[i].xoffset;
		top    = lnodes[wbs->epsd][n].y - c[i].yoffset;
		right  = left + c[i].width;
		bottom = top  + c[i].height;
		
		if (left >= 0 && right < SCREENWIDTH && top >= 0 && bottom < SCREENHEIGHT)
			fits = true;
		else
			i++;
		
	} while (!fits && i!=2);
	
	if (fits && i<2)
	{
		GL_DrawPatchNr(lnodes[wbs->epsd][n].x, lnodes[wbs->epsd][n].y,
			c[i].lump);
	}
}

void WI_initAnimatedBack(void)
{
    int			i;
    wianim_t*	a;

    if (commercial_game || !animSet[wbs->epsd] || R3D_ScriptedMap) return;

    for (i = 0; i < numAnims[wbs->epsd]; i++)
    {
		a = &anims[wbs->epsd][i];

		// init variables
		a->ctr = -1;

		// specify the next time to draw it
		if (a->type == ANIM_ALWAYS)
			a->nexttic = bcnt + 1 + (rand()%a->period);
		else if (a->type == ANIM_RANDOM)
			a->nexttic = bcnt + 1 + a->data2+(rand()%a->data1);
		else if (a->type == ANIM_LEVEL)
			a->nexttic = bcnt + 1;
    }
}

void WI_updateAnimatedBack(void)
{
    int		i;
    wianim_t*	a;

    if (commercial_game || !animSet[wbs->epsd] || R3D_ScriptedMap) return;

    for (i = 0; i < numAnims[wbs->epsd]; i++)
    {
		a = &anims[wbs->epsd][i];

		if(bcnt == a->nexttic)
		{
			switch (a->type)
			{
			case ANIM_ALWAYS:
				if (++a->ctr >= a->nanims) a->ctr = 0;
				a->nexttic = bcnt + a->period;
				break;

			case ANIM_RANDOM:
				a->ctr++;
				if (a->ctr == a->nanims)
				{
					a->ctr = -1;
					a->nexttic = bcnt+a->data2+(rand()%a->data1);
				}
				else a->nexttic = bcnt + a->period;
				break;
		
			case ANIM_LEVEL:
				// hack for level anims
				if(!(state == StatCount && i == 7)
					&& wbs->next == a->data1)
				{
					a->ctr++;
					if (a->ctr == a->nanims) a->ctr--;
					a->nexttic = bcnt + a->period;
				}
				break;
			}
		}
    }
}

void WI_drawAnimatedBack(void)
{
    wianim_t	*a;
    int			i;

    if (commercial_game || !animSet[wbs->epsd] || R3D_ScriptedMap) return;

    for (i = 0; i < numAnims[wbs->epsd] ; i++)
    {
		a = &anims[wbs->epsd][i];
		if(a->ctr >= 0) GL_DrawPatch_CS((float)a->loc.x, (float)a->loc.y, a->p[a->ctr].lump, 1);
    }
}

//
// Draws a number.
// If digits > 0, then use that many digits minimum,
//  otherwise only use as many as necessary.
// Returns new x position.
//

int WI_drawNum(int x, int y, int n, int	digits)
{
    int		fontwidth = grnum[0].width;
    int		neg = n < 0;

    if(digits < 0)
    {
		if(!n)
			digits = 1;	// make variable-length zeros 1 digit long
		else
		{
			int temp = n;
			digits = 0;	// figure out # of digits in #
			while (temp) { temp /= 10; digits++; }
		}
    }

    if(neg) n = -n;

    // if non-number, do not draw it (Yep. Clear as mud)
    if(n == 1994) return 0;

    // draw the new number
    while(digits--)
    {
		x -= fontwidth;
		GL_DrawPatchNr(x, y, grnum[ n % 10 ].lump);
		n /= 10;
    }

    // draw a minus sign if necessary
    if(neg) GL_DrawPatchNr(x -= 8, y, wiminus.lump);

    return x;
}

void WI_drawPercent(int x, int y, int p)
{
    if (p < 0) return;

    GL_DrawPatchNr(x, y, percent.lump);
    WI_drawNum(x, y, p, -1);
}

// Display level completion time and par,
//  or "sucks" message if overflow.
void WI_drawTime(int x, int y, int t)
{
    int		div, n;

    if(t<0) return;

    if(t <= 61*59)
    {
		div = 1;
		do
		{
			n = (t / div) % 60;
			x = WI_drawNum(x, y, n, 2) - colon.width;
			div *= 60;

		    if (div==60 || t / div)
				GL_DrawPatchNr(x, y, colon.lump);	// draw
		} while (t / div);
    }
    else
    {
		GL_DrawPatchNr(x - sucks.width, y, sucks.lump); // "sucks"
    }
}

void WI_initNoState(void)
{
    state = NoState;
    acceleratestage = 0;
    cnt = 10;
}

void WI_updateNoState(void)
{
	WI_updateAnimatedBack();
	
	if (!--cnt)
	{
		// Is there a debriefing for this map?
		if (   wminfo.inflid[0]
			&& FI_Debriefing(wminfo.inflid, wminfo.runCast))
		{
			return;
		}


		if (isCompleted)
		{
			if (compat[c_nextmap])
			{
				int i;

				for (i = 0; i < 100; i++)
				{
					if (hasIWADlevel[i])
					{
						if (commercial_game)
						{
							haslevelMapxx[i] = true;
						}
						else if (i > 9)
						{
							haslevelExMx[i] = true;
							epihaspwad[ (i/10) ] = true;
						}
					}
				}
			}
			else
			{
				state = StatCount;
				return;
			}
		}

		G_WorldDone();
	}
}

static RDBOOL		snl_pointeron = false;

void WI_initShowNextLoc(void)
{
    state = ShowNextLoc;
    acceleratestage = 0;
    cnt = SHOWNEXTLOCDELAY * TICRATE;

    WI_initAnimatedBack();
}

void WI_updateShowNextLoc(void)
{
    WI_updateAnimatedBack();

    if (!--cnt || acceleratestage)
		WI_initNoState();
    else
		snl_pointeron = (cnt & 31) < 20;
}

void WI_drawShowNextLoc(void)
{
    int		i, last;

	if (R3D_ScriptedMap || isCompleted) return;

    if (commercial_game || !epsdSet[wbs->epsd])
	{
		wimap_el = true;
		GL_DrawBackground();
		WI_drawEL();
	}
	else
    {
		GL_DrawBackground();
		WI_drawAnimatedBack(); // draw animated background (if applicable)

		last = (wbs->last == 8) ? wbs->next - 1 : wbs->last;

		// draw a splat on taken cities.
		for (i = 0; i <= last; i++)
			WI_drawOnLnode(i, &splat);

		// splat the secret level?
		if (wbs->didsecret)
			WI_drawOnLnode(8, &splat);

		// draw flashing ptr
		if (snl_pointeron)
			WI_drawOnLnode(wbs->next, yah); 
    }
}

void WI_drawNoState(void)
{
    snl_pointeron = true;
    WI_drawShowNextLoc();
}

static int		sp_state;
static float	alpha;

void WI_initStats(void)
{
    state = StatCount;
    acceleratestage = 0;
    sp_state = 1;
    cnt_kills[0] = cnt_items[0] = cnt_secret[0] = -1;
    cnt_time = cnt_par = -1;
    cnt_pause = TICRATE;
    WI_initAnimatedBack();
}

void WI_updateStats(void)
{
    WI_updateAnimatedBack();

	if (R3D_ScriptedMap)
	{
		sp_state = 10;

		if (!(bcnt&3) && alpha < 1.0f)
		{
			S_LocalSound(sfx_pistol, 0);
			alpha += 0.05f;
		}

		if (alpha < 1.0f || isCompleted) return;
	}

    if (acceleratestage && sp_state != 10)
    {
		acceleratestage = 0;
		cnt_kills[0] = (plrs[me].skills * 100) / wbs->maxkills;
		cnt_items[0] = (plrs[me].sitems * 100) / wbs->maxitems;
		cnt_secret[0] = (plrs[me].ssecret * 100) / wbs->maxsecret;
		cnt_time = plrs[me].stime;
		cnt_par = wbs->partime;
		S_LocalSound(sfx_barexp, 0);
		sp_state = 10;
    }

	if (sp_state == 2)
	{
		cnt_kills[0] += 2;

		if (!(bcnt&3))
			S_LocalSound(sfx_pistol, 0);

		if (cnt_kills[0] >= (plrs[me].skills * 100) / wbs->maxkills)
		{
			cnt_kills[0] = (plrs[me].skills * 100) / wbs->maxkills;
			S_LocalSound(sfx_barexp, 0);
			sp_state++;
		}
    }
    else if (sp_state == 4)
    {
		cnt_items[0] += 2;

		if (!(bcnt&3))
			S_LocalSound(sfx_pistol, 0);

		if (cnt_items[0] >= (plrs[me].sitems * 100) / wbs->maxitems)
		{
			cnt_items[0] = (plrs[me].sitems * 100) / wbs->maxitems;
			S_LocalSound(sfx_barexp, 0);
			sp_state++;
		}
    }
    else if (sp_state == 6)
    {
		cnt_secret[0] += 2;

		if (!(bcnt&3))
			S_LocalSound(sfx_pistol, 0);

		if (cnt_secret[0] >= (plrs[me].ssecret * 100) / wbs->maxsecret)
		{
			cnt_secret[0] = (plrs[me].ssecret * 100) / wbs->maxsecret;
			S_LocalSound(sfx_barexp, 0);
			sp_state++;
		}
    }

    else if (sp_state == 8)
    {
		if (!(bcnt&3)) S_LocalSound(sfx_pistol, 0);

		cnt_time += 3;

		if (cnt_time >= plrs[me].stime)
			cnt_time = plrs[me].stime;

		cnt_par += 3;

		if (cnt_par >= wbs->partime)
		{
			cnt_par = wbs->partime;

		    if (cnt_time == plrs[me].stime)
		    {
				S_LocalSound(sfx_barexp, 0);
				sp_state++;
			}
		}
    }
    else if (sp_state == 10)
    {
		if (acceleratestage)
		{
			if (commercial_game)
			{
				WI_initNoState();
			}
			else
			{
				WI_initShowNextLoc();
				S_LocalSound(sfx_sgcock, 0);
			}
		}
    }
    else if (sp_state & 1)
    {
		if (!--cnt_pause)
		{
			sp_state++;
			cnt_pause = TICRATE;
		}
    }
}

void WI_drawStats(void)
{
    int lh = (3*grnum[0].height)/2;    // line height

    GL_DrawBackground();
	if (wimap_el) WI_drawEL();

	if (R3D_ScriptedMap)
	{
		R_DrawScriptedMapInf(lh, alpha, isCompleted);
	}
	else
	{
		WI_drawAnimatedBack();		// draw animated background
    
		WI_drawLF();
		GL_DrawPatchNr(SP_STATSX, SP_STATSY, kills.lump);
		WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY, cnt_kills[0]);

		GL_DrawPatchNr(SP_STATSX, SP_STATSY+lh, items.lump);
		WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY+lh, cnt_items[0]);

		GL_DrawPatchNr(SP_STATSX, SP_STATSY+2*lh, sp_secret.lump);
		WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY+2*lh, cnt_secret[0]);

		GL_DrawPatchNr(SP_TIMEX, SP_TIMEY, time.lump);
		WI_drawTime(SCREENWIDTH/2 - SP_TIMEX, SP_TIMEY, cnt_time);

		GL_DrawPatchNr(SCREENWIDTH/2 + SP_TIMEX, SP_TIMEY, par.lump);
		WI_drawTime(SCREENWIDTH - SP_TIMEX, SP_TIMEY, cnt_par);
	}
}

void WI_checkForAccelerate(void)
{
    int   i;
    mbf_player_t  *player;

    // check for button presses to skip delays
    for(i=0, player = mbf_players ; i<1/*MAXPLAYERS*/ ; i++, player++)
    {
		if(mbf_players[i].plr->ingame)
		{
			if(player->cmd.buttons & BT_ATTACK)
			{
				if (!player->attackdown) acceleratestage = 1;
				player->attackdown = true;
			}
			else
				player->attackdown = false;

			if(player->cmd.buttons & BT_USE)
			{
				if (!player->usedown) acceleratestage = 1;
				player->usedown = true;
			}
			else
				player->usedown = false;
		}
    }
}

RDBOOL WI_Drawer (void)
{
	int ret = false;

    if (state == StatCount)
	{
		WI_drawStats();
		ret = true;
	}
    else if (state == ShowNextLoc) { WI_drawShowNextLoc(); ret = true; }
    else if (state == NoState	 ) { WI_drawNoState(); ret = true; }
	return ret;
}

// Updates stuff each tick
void WI_Ticker(void)
{
    if (++bcnt == 1)   // counter for general background animation
    {
		if (!compat[c_nomusic])
		{
			global_looped = 1;

			// intermission music
  			if (commercial_game)
				S_StartMusicNum(mus_dm2int);
			else
				S_StartMusicNum(mus_inter);
		}
	}

    WI_checkForAccelerate();

    if      (state == StatCount  ) WI_updateStats();
    else if (state == ShowNextLoc) WI_updateShowNextLoc();
    else if (state == NoState	 ) WI_updateNoState();
}

void WI_loadData()
{
    wianim_t	*a;
	char		name[9];
    int			i, j;
	int			pic;
	int			pWIURH0  = -1;
	int			pWIURH1  = -1;
	int			pWISPLAT = -1;

	wimap_el = false;

	if (!commercial_game)
	{
		if (wbs->epsd < 3)
		{
			if (wpicSet[wbs->epsd] < 0)
			{
				sprintf(name, "WIMAP%d", wbs->epsd);

				// v220-16 check for pwad replacement
				if ( (pic = W_CheckPWADNumForName(name)) >= 0)
					wpicSet[wbs->epsd] = pic;
				else
					wpicSet[wbs->epsd] = hiresPatches ? W_CheckIWADNumForName(name) : W_CheckNumForName(name);	// v220-8
			}

			wimap_pic = wpicSet[wbs->epsd];

			R_SetPatch(&yah[0], "WIURH0", 1);	// you are here
			R_SetPatch(&yah[1], "WIURH1", 1);	// you are here (alt.)
			R_SetPatch(&splat, "WISPLAT", 1);	// splat
		
			for (j = 0; j < numAnims[wbs->epsd]; j++)
			{
				a = &anims[wbs->epsd][j];

				for (i = 0;i < a->nanims; i++)
				{
					// MONDO HACK!
					if (wbs->epsd != 1 || j != 8) 
					{
						// animations
						sprintf(name, "WIA%d%.2d%.2d", wbs->epsd, j, i);  
						R_SetPatch(&a->p[i], name, 1);
					}
					else
					{
						// HACK ALERT!
						memcpy(&a->p[i], &anims[1][4].p[i], sizeof(dpatch_t));
					}
				}
			}
		}
		// v220-4 limited support for Eternity lump as accessed via a ZDoom mapinfo
		else if (wbs->epsd > 3
			&& mapinfo->epitext[0]
			&& W_CheckNumForName(mapinfo->epitext) >= 0)
		{
			static int firsttime = true;
			int		l1v[4], l2v[12];
			char	str[6];

			animSet[wbs->epsd] = false; // for possible later support

			if (epsdSet[wbs->epsd])
			{
				wimap_pic = wpicSet[wbs->epsd];
			}
			else
			{
				if (SR_GetInfoFromScript(mapinfo->epitext, "BACKGROUND", NULL, false, 1, l1v, false,0, NULL))
				{
					wimap_pic = wpicSet[wbs->epsd] = l1v[0];
				}
				else
					goto epsd_error;

				if (SR_GetInfoFromScript(mapinfo->epitext, "SPLAT", NULL, false, 1, l1v, false, 0, NULL))
				{
					splat.lump = l1v[0]; ST_SetPatch(&splat, 1);
				}
				else
					goto epsd_error;

				if (SR_GetInfoFromScript(mapinfo->epitext, "POINTER", NULL, false, 2, l1v, false, 0, NULL))
				{
					yah[0].lump = l1v[0]; ST_SetPatch(&splat, 1);
					yah[1].lump = l1v[1]; ST_SetPatch(&splat, 1);
				}
				else
					goto epsd_error;

				for (i = 0; i < 9; i++)
				{
					sprintf(str, "E%iM%i", wbs->epsd + 1, i + 1);

					if (SR_GetInfoFromScript(mapinfo->epitext, "SPOTS", str, false, 0, NULL, true, 2, l2v))
					{
						lnodes[wbs->epsd][i].x = l2v[0];
						lnodes[wbs->epsd][i].y = l2v[1];
					}
					else
						goto epsd_error;
				}

				epsdSet[wbs->epsd] = true;
			}
epsd_error:;

		}
	}

    for (i = 0; i < 10; i++)
    {
		sprintf(name, "WINUM%d", i);	// numbers 0-9    
		R_SetPatch(&grnum[i], name, 0);
    }

	R_SetPatch(&wiminus	, "WIMINUS"	, 0);	// minus sign hack
    R_SetPatch(&percent	, "WIPCNT"	, 0);	// percent sign
    R_SetPatch(&finished, "WIF"		, 0);	// "finished"
    R_SetPatch(&entering, "WIENTER"	, 0);	// "entering"
    R_SetPatch(&kills	, "WIOSTK"	, 0);	// "kills"
    R_SetPatch(&secret	, "WIOSTS"	, 0);	// "scrt"
    R_SetPatch(&sp_secret, "WISCRT2", 0);	// "secret"
	R_SetPatch(&items	, "WIOSTI"	, 0);	// "items"
    R_SetPatch(&frags	, "WIFRGS"	, 0);	// "frgs"
    R_SetPatch(&colon	, "WICOLON"	, 0);	// ":"
    R_SetPatch(&time	, "WITIME"	, 0);	// "time"
    R_SetPatch(&sucks	, "WISUCKS"	, 0);	// "sucks"
    R_SetPatch(&par		, "WIPAR"	, 0);	// "par"
    R_SetPatch(&killers	, "WIKILRS"	, 0);	// "killers" (vertical)
    R_SetPatch(&victims	, "WIVCTMS"	, 0);	// "victims" (horiz)
    R_SetPatch(&total	, "WIMSTT"	, 0);	// "total" 
    R_SetPatch(&star	, "STFST01"	, 0);	// your face
    R_SetPatch(&bstar	, "STFDEAD0", 0);	// dead face

    for (i = 0; i < NUM_TEAMS ; i++)
    {
		// "1,2,3,4"
		sprintf(name, "STPB%d", i);      
		R_SetPatch(&p[i], name, 0);

		// "1,2,3,4"
		sprintf(name, "WIBP%d", i+1);     
		R_SetPatch(&bp[i], name, 0);
    }

}

void WI_initVariables(wbstartstruct_t* wbstartstruct)
{
	wbs = wbstartstruct;
 	acceleratestage = 0;
	cnt = bcnt = 0;
	firstrefresh = 1;
	me = wbs->pnum;
	myteam = cfg.PlayerColor[wbs->pnum];
	plrs = wbs->plyr;

	if (!wbs->maxkills ) wbs->maxkills = 1;
	if (!wbs->maxitems ) wbs->maxitems = 1;
	if (!wbs->maxsecret) wbs->maxsecret = 1;

	alpha = 0;	// b211-0 added for scripted maps
}

void WI_Start(wbstartstruct_t* wbstartstruct)
{
	teaminfo_t *tin;
	int i, j, k;

	isCompleted = wminfo.show_completed;

    WI_initVariables(wbstartstruct);
    WI_loadData();

	// Calculate team stats.
	memset(teaminfo, 0, sizeof(teaminfo));

	for (i=0, tin=teaminfo; i<NUM_TEAMS; i++, tin++)
	{
		for (j=0; j<MAXPLAYERS; j++)
		{
			// Is the player in this team?
			if(!plrs[j].in || cfg.PlayerColor[j] != i) continue;
			tin->members++;
			// Check the frags.
			for(k=0; k<MAXPLAYERS; k++)
				tin->frags[cfg.PlayerColor[k]] += plrs[j].frags[k];
			// Counters.
			if(plrs[j].sitems > tin->items) tin->items = plrs[j].sitems;
			if(plrs[j].skills > tin->kills) tin->kills = plrs[j].skills;
			if(plrs[j].ssecret > tin->secret) tin->secret = plrs[j].ssecret;
		}

		// Calculate team's total frags.
		for(j=0; j<NUM_TEAMS; j++)
		{
			if(j == i) // Suicides are negative frags.
				tin->totalfrags -= tin->frags[j];
			else
				tin->totalfrags += tin->frags[j];
		}
	}

	WI_initStats();
}

void WI_SetState(stateenum_t st)
{
	if     (st == StatCount	 ) WI_initStats();
	else if(st == ShowNextLoc) WI_initShowNextLoc();
	else if(st == NoState	 ) WI_initNoState();
}
