//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dd_defns.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for use with R3D, includes;
//  R3D specific ded file loading
//  Sound loading routines
//  Link sound removed (only used for chaingun in Doom2) to
//  allow different sounds for pistol and chaingun
//  Further DEH compatibility
//  XG stuff removed
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  ded path Defs\Model\abbs\abbs-models.ded changed to
//           Defs\Model\Sitters\sitters.ded
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 11 (v220-11) 08/Aug/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  User Interface, ties to the console routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <direct.h>

#include "de_base.h"
#include "de_system.h"
#include "de_refresh.h"
#include "de_console.h"
#include "de_audio.h"
#include "de_misc.h"
#include "r3d_compat.h"
#include "r3d_mapinfo.h"

#include "AcFnLink.h"

#include "doom/info.h"
#include "doom/SndIdx.h"

#include <string.h>
#include <ctype.h>
#include <stdio.h>

// MACROS ------------------------------------------------------------------

#define LOOPi(n)	for(i=0; i<n; i++)
#define LOOPk(n)	for(k=0; k<n; k++)
#define MAX_READ	256	// max number of ded files that can be read

// The first five values must not be changed (see defdir)
#define DF_DEFBASE		0
#define DF_DEFMODEL		1
#define DF_FULLPATH		2
#define DF_ABBAPATH		3
#define DF_WSMPATH		4
#define DF_DEFPATH		5
#define DF_UDIRPATH		6
#define DF_THISPATH		7

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_SetStateTable(int num);

// v220-9
RDBOOL	hasDEHactors = false;
void	H_InitLODScript(void);

// v220-13
RDBOOL	R_CheckForPWADSprites(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void Def_ReadProcessDED(char *filename, int chdir);	// b211-0 *allowcg removed v220-3

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		detail_ded[MAXPATHNAMESIZE];
extern char		*R3Dinfine;
extern RDBOOL	dehdebug;
extern int		DEHchange;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

ded_t		defs;			// The main definitions database.
sprname_t	*sprnames;		// Sprite name list.
state_t		*states;		// State list.
mobjinfo_t	*mobjinfo;		// Map object info database.
sfxinfo_t	*sounds;		// Sound effect list.
detailtex_t	*dettex;		// Detail texture assignments.
mobjinfo_t	**stateowners;	// A pointer for each state.

int	customPunchRestState = -1;
int pistoldamage = 0;
int plasmadamage = 0;
int bfgdamage = 0;
int punchDamage = 0;
int bfgsound = 0;
int pistolsound = 0;
int plasmasound = 0;
int	punchsound1;
int	punchsound2;

ded_count_t	count_sprnames;
ded_count_t	count_states;
ded_count_t	count_mobjinfo;
ded_count_t	count_sounds;
ded_count_t	count_texts;
ded_count_t	count_details;
ded_count_t	count_stateowners;

// Pointer array to the def texts.
// Once loaded these are never changed.
// The value used is transferred to
// the ded_t ded_text_t text container
ddtext_t	*texts;

// Current mapinfo.
ded_mapinfo_t *mapinfo = NULL;

RDBOOL		first_ded;
RDBOOL		cache_ded_model;

// v220-13;
byte		*moTypeArray;

// v220-16
int			specialSprite = 0;

// v220-18
RDBOOL		EMAPINFO_script;

// v220-31	dynlights Revised v220-35

// MT_STROBE1	// type 5500 Flashing light slow
// MT_STROBE2	// type 5501 Flashing light fast
// MT_STROBE3	// type 5502 Strobe medium
// MT_STROBE4	// type 5503 Random flicker
// MT_STROBE5	// type 5504 Phased (color set using spawn angle)
// MT_STROBEX	// type 5505 used in a lighting special sector type 8 or 17

int fiparms[100][2];

// v220-35
int SittersLoaded = 0;
int stsndArray[MAXSTSOUNDS][2];

// Note: There is no ST_SND3 sound. Instead flag ST_DS3 references sound ST_DS2
char *nsnd[MAXSTSOUNDS] = { "ST_SND4" ,"ST_SND5" ,"ST_SND6","ST_SND7","ST_SND8","ST_SND9","ST_SND10",
							"ST_SND11","ST_SND12","ST_DS1" ,"ST_DS2" ,"ST_DS2" ,"ST_DS4" };
// v220-35 smoke stuff
char	*bfgdname = NULL;
char	*pistdname = NULL;
char	*plasdname = NULL;
char	*sawupname = NULL;
char	*punchname1 = NULL;
char	*punchname2 = NULL;

int pgsmokestatenum;
int pgsmokestatenumX;
int cgsmokestatenum;
int cgsmokestatenumX;
int sgsmokestatenum;
int sgsmokestatenumX;
int dbsmokestatenum;
int dbsmokestatenumX;
int gnsmokestatenum;
int gnsmokestatenumX;
int srsmokestatenum;
int srsmokestatenumX;
int igsmokestatenum;
int igsmokestatenumX;
int wnsmokestatenum;
int wnsmokestatenumX;
int aksmokestatenum;
int aksmokestatenumX;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// b211-0 Added
static char warn[] =
			{"WARNING   : %s def [%s] ignored as DD_DEFNS found in the pwad.\n"};

static RDBOOL		defs_inited = false;
static RDBOOL		has_dd_defn = false;
static int			read_ids[MAX_READ];
static int			read_count;
static int			*sprlist;

typedef struct {
	char	*dedname;
	int		val;
} r3ddefs_t;

r3ddefs_t r3ddefs[] = {
	{ "Audio.ded"		, 0x002AB6E0 },
	{ "CoreEffects.ded"	, 0x004F04F6 },
	{ "Maps.ded"		, 0x00111948 },
	{ "FinalesDoom.ded"	, 0x000B7EE0 },
	{ "FinalesPlut.ded"	, 0x000931B8 },
	{ "FinalesTNT.ded"	, 0x000945AD },
	{ "R3DFlags.ded"	, 0x000C2711 },
	{ "R3DSprites.ded"	, 0x0004C702 },
	{ "R3DStates.ded"	, 0x008D2520 },
	{ "R3DThings.ded"	, 0x0073B4D4 },
	{ "Text.ded"		, 0x001D60B0 },
	{ "Values.ded"		, 0x0001D6B3 },
	{ NULL, 0 }
};

char *defsnames[] = {
	"R3DFlags.ded",
	"R3DThings.ded",
	"R3DStates.ded",
	"R3DSprites.ded",
	"Maps.ded",
	"Audio.ded",
	"Text.ded",
	"Values.ded",
	"CoreEffects.ded",
	NULL
};

// The action routine link table
// v220-13 more rigorous state handling used

actionlink_t actionlinks[] =
{
// This must be first in the list
	{nmNullHandler,		"A_Null"			, NULL, NULL },  // R3D

// Player actions
	{nmLight0,          "A_Light0"			, NULL, NULL },
	{nmWeaponReady,     "A_WeaponReady"		, NULL, NULL },
	{nmLower,           "A_Lower"			, NULL, NULL },
	{nmRaise,           "A_Raise"			, NULL, NULL },
	{nmPunch,           "A_Punch"			, NULL, NULL },
	{nmReFire,          "A_ReFire"			, NULL, NULL },
	{nmFirePistol,      "A_FirePistol"		, NULL, NULL },
	{nmLight1,          "A_Light1"			, NULL, NULL },
	{nmFireShotgun,     "A_FireShotgun"		, NULL, NULL },
	{nmLight2,          "A_Light2"			, NULL, NULL },
	{nmFireShotgun2,    "A_FireShotgun2"	, NULL, NULL },
	{nmCheckReload,     "A_CheckReload"		, NULL, NULL },
	{nmOpenShotgun2,    "A_OpenShotgun2"	, NULL, NULL },
	{nmLoadShotgun2,    "A_LoadShotgun2"	, NULL, NULL },
	{nmCloseShotgun2,   "A_CloseShotgun2"	, NULL, NULL },
	{nmFireCGun,        "A_FireCGun"		, NULL, NULL },
	{nmGunFlash,        "A_GunFlash"		, NULL, NULL },
	{nmFireMissile,     "A_FireMissile"		, NULL, NULL },
	{nmSaw,             "A_Saw"				, NULL, NULL },
	{nmFirePlasma,      "A_FirePlasma"		, NULL, NULL },
	{nmBFGsound,        "A_BFGsound"		, NULL, NULL },
	{nmFireBFG,         "A_FireBFG"			, NULL, NULL },
	{nmWeapSound,		"A_WeapSound"		, NULL, NULL },

// Enemy actions
	{nmBFGSpray,        "A_BFGSpray"		, NULL, NULL },
	{nmExplode,         "A_Explode"			, NULL, NULL },
	{nmPain,            "A_Pain"			, NULL, NULL },
	{nmPlayerScream,    "A_PlayerScream"	, NULL, NULL },
	{nmFall,            "A_Fall"			, NULL, NULL },
	{nmXScream,         "A_XScream"			, NULL, NULL },
	{nmLook,            "A_Look"			, NULL, NULL },
	{nmChase,           "A_Chase"			, NULL, NULL },
	{nmFaceTarget,      "A_FaceTarget"		, NULL, NULL },
	{nmPosAttack,       "A_PosAttack"		, NULL, NULL },
	{nmScream,          "A_Scream"			, NULL, NULL },
	{nmSPosAttack,      "A_SPosAttack"		, NULL, NULL },
	{nmVileChase,       "A_VileChase"		, NULL, NULL },
	{nmVileStart,       "A_VileStart"		, NULL, NULL },
	{nmVileTarget,      "A_VileTarget"		, NULL, NULL },
	{nmVileAttack,      "A_VileAttack"		, NULL, NULL },
	{nmStartFire,       "A_StartFire"		, NULL, NULL },
	{nmFire,            "A_Fire"			, NULL, NULL },
	{nmFireCrackle,     "A_FireCrackle"		, NULL, NULL },
	{nmTracer,          "A_Tracer"			, NULL, NULL },
	{nmSkelWhoosh,      "A_SkelWhoosh"		, NULL, NULL },
	{nmSkelFist,        "A_SkelFist"		, NULL, NULL },
	{nmSkelMissile,     "A_SkelMissile"		, NULL, NULL },
	{nmFatRaise,        "A_FatRaise"		, NULL, NULL },
	{nmFatAttack1,      "A_FatAttack1"		, NULL, NULL },
	{nmFatAttack2,      "A_FatAttack2"		, NULL, NULL },
	{nmFatAttack3,      "A_FatAttack3"		, NULL, NULL },
	{nmBossDeath,       "A_BossDeath"		, NULL, NULL },
	{nmCPosAttack,      "A_CPosAttack"		, NULL, NULL },
	{nmCPosRefire,      "A_CPosRefire"		, NULL, NULL },
	{nmTroopAttack,     "A_TroopAttack"		, NULL, NULL },
	{nmSargAttack,      "A_SargAttack"		, NULL, NULL },
	{nmHeadAttack,      "A_HeadAttack"		, NULL, NULL },
	{nmBruisAttack,     "A_BruisAttack"		, NULL, NULL },
	{nmSkullAttack,     "A_SkullAttack"		, NULL, NULL },
	{nmMetal,           "A_Metal"			, NULL, NULL },
	{nmSpidRefire,      "A_SpidRefire"		, NULL, NULL },
	{nmBabyMetal,       "A_BabyMetal"		, NULL, NULL },
	{nmBspiAttack,      "A_BspiAttack"		, NULL, NULL },
	{nmHoof,            "A_Hoof"			, NULL, NULL },
	{nmCyberAttack,     "A_CyberAttack"		, NULL, NULL },
	{nmPainAttack,      "A_PainAttack"		, NULL, NULL },
	{nmPainDie,         "A_PainDie"			, NULL, NULL },
	{nmKeenDie,         "A_KeenDie"			, NULL, NULL },
	{nmBrainPain,       "A_BrainPain"		, NULL, NULL },
	{nmBrainScream,     "A_BrainScream"		, NULL, NULL },
	{nmBrainDie,        "A_BrainDie"		, NULL, NULL },
	{nmBrainAwake,      "A_BrainAwake"		, NULL, NULL },
	{nmBrainSpit,       "A_BrainSpit"		, NULL, NULL },
	{nmSpawnSound,      "A_SpawnSound"		, NULL, NULL },
	{nmSpawnFly,        "A_SpawnFly"		, NULL, NULL },
	{nmBrainExplode,    "A_BrainExplode"	, NULL, NULL },
	{nmDetonate,        "A_Detonate"		, NULL, NULL },	// killough 8/9/98
	{nmMushroom,        "A_Mushroom"		, NULL, NULL },	// killough 10/98
	{nmDie,             "A_Die"				, NULL, NULL },	// killough 11/98
	{nmSpawn,           "A_Spawn"			, NULL, NULL },	// killough 11/98
	{nmTurn,            "A_Turn"			, NULL, NULL },	// killough 11/98
	{nmFace,            "A_Face"			, NULL, NULL },	// killough 11/98
	{nmScratch,         "A_Scratch"			, NULL, NULL },	// killough 11/98
	{nmPlaySound,       "A_PlaySound"		, NULL, NULL },	// killough 11/98
	{nmRandomJump,      "A_RandomJump"		, NULL, NULL },	// killough 11/98
	{nmLineEffect,      "A_LineEffect"		, NULL, NULL },	// killough 11/98
	{nmdbAttack,		"A_dbAttack"		, NULL, NULL }, // R3D

	// This NULL entry must be the last in the list
	{nmEnd				, NULL, NULL, NULL }
};

// CODE --------------------------------------------------------------------

void loadActionlinks()
{
	int i;

	for (i = 0; i < nmEnd; i++)
	{
		switch(i)
		{
		case nmNullHandler	: actionlinks[i].func = A_NullHandler;	break;
// Player specific actions
		case nmLight0       : actionlinks[i].plfn = A_Light0;		break;
		case nmWeaponReady  : actionlinks[i].plfn = A_WeaponReady;	break;
		case nmLower        : actionlinks[i].plfn = A_Lower;		break;
		case nmRaise        : actionlinks[i].plfn = A_Raise;		break;
		case nmPunch        : actionlinks[i].plfn = A_Punch;		break;
		case nmReFire       : actionlinks[i].plfn = A_ReFire;		break;
		case nmFirePistol   : actionlinks[i].plfn = A_FirePistol;	break;
		case nmLight1       : actionlinks[i].plfn = A_Light1;		break;
		case nmFireShotgun  : actionlinks[i].plfn = A_FireShotgun;	break;
		case nmLight2       : actionlinks[i].plfn = A_Light2;		break;
		case nmFireShotgun2 : actionlinks[i].plfn = A_FireShotgun2;	break;
		case nmCheckReload  : actionlinks[i].plfn = A_CheckReload;	break;
		case nmOpenShotgun2	: actionlinks[i].plfn = A_OpenShotgun2;	break;
		case nmLoadShotgun2	: actionlinks[i].plfn = A_LoadShotgun2;	break;
		case nmCloseShotgun2: actionlinks[i].plfn = A_CloseShotgun2; break;
		case nmFireCGun     : actionlinks[i].plfn = A_FireCGun;		break;
		case nmGunFlash     : actionlinks[i].plfn = A_GunFlash;		break;
		case nmFireMissile  : actionlinks[i].plfn = A_FireMissile;	break;
		case nmSaw          : actionlinks[i].plfn = A_Saw;			break;
		case nmFirePlasma   : actionlinks[i].plfn = A_FirePlasma;	break;
		case nmBFGsound     : actionlinks[i].plfn = A_BFGsound;		break;
		case nmFireBFG      : actionlinks[i].plfn = A_FireBFG;		break;
		case nmWeapSound	: actionlinks[i].plfn = A_WeapSound;	break;

// Enemy actions
		case nmBFGSpray		: actionlinks[i].func = A_BFGSpray;		break;
		case nmExplode		: actionlinks[i].func = A_Explode;		break;
		case nmPain			: actionlinks[i].func = A_Pain;			break;
		case nmPlayerScream	: actionlinks[i].func = A_PlayerScream;	break;
		case nmFall			: actionlinks[i].func = A_Fall;			break;
		case nmXScream		: actionlinks[i].func = A_XScream;		break;
		case nmLook			: actionlinks[i].func = A_Look;			break;
		case nmChase		: actionlinks[i].func = A_Chase;		break;
		case nmFaceTarget	: actionlinks[i].func = A_FaceTarget;	break;
		case nmPosAttack	: actionlinks[i].func = A_PosAttack;	break;
		case nmScream		: actionlinks[i].func = A_Scream;		break;
		case nmSPosAttack	: actionlinks[i].func = A_SPosAttack;	break;
		case nmVileChase	: actionlinks[i].func = A_VileChase;	break;
		case nmVileStart	: actionlinks[i].func = A_VileStart;	break;
		case nmVileTarget	: actionlinks[i].func = A_VileTarget;	break;
		case nmVileAttack	: actionlinks[i].func = A_VileAttack;	break;
		case nmStartFire	: actionlinks[i].func = A_StartFire;	break;
		case nmFire			: actionlinks[i].func = A_Fire;			break;
		case nmFireCrackle	: actionlinks[i].func = A_FireCrackle;	break;
		case nmTracer		: actionlinks[i].func = A_Tracer;		break;
		case nmSkelWhoosh	: actionlinks[i].func = A_SkelWhoosh;	break;
		case nmSkelFist		: actionlinks[i].func = A_SkelFist;		break;
		case nmSkelMissile	: actionlinks[i].func = A_SkelMissile;	break;
		case nmFatRaise		: actionlinks[i].func = A_FatRaise;		break;
		case nmFatAttack1	: actionlinks[i].func = A_FatAttack1;	break;
		case nmFatAttack2	: actionlinks[i].func = A_FatAttack2;	break;
		case nmFatAttack3	: actionlinks[i].func = A_FatAttack3;	break;
		case nmBossDeath	: actionlinks[i].func = A_BossDeath;	break;
		case nmCPosAttack	: actionlinks[i].func = A_CPosAttack;	break;
		case nmCPosRefire	: actionlinks[i].func = A_CPosRefire;	break;
		case nmTroopAttack	: actionlinks[i].func = A_TroopAttack;	break;
		case nmSargAttack	: actionlinks[i].func = A_SargAttack;	break;
		case nmHeadAttack	: actionlinks[i].func = A_HeadAttack;	break;
		case nmBruisAttack	: actionlinks[i].func = A_BruisAttack;	break;
		case nmSkullAttack	: actionlinks[i].func = A_SkullAttack;	break;
		case nmMetal		: actionlinks[i].func = A_Metal;		break;
		case nmSpidRefire	: actionlinks[i].func = A_SpidRefire;	break;
		case nmBabyMetal	: actionlinks[i].func = A_BabyMetal;	break;
		case nmBspiAttack	: actionlinks[i].func = A_BspiAttack;	break;
		case nmHoof			: actionlinks[i].func = A_Hoof;			break;
		case nmCyberAttack	: actionlinks[i].func = A_CyberAttack;	break;
		case nmPainAttack	: actionlinks[i].func = A_PainAttack;	break;
		case nmPainDie		: actionlinks[i].func = A_PainDie;		break;
		case nmKeenDie		: actionlinks[i].func = A_KeenDie;		break;
		case nmBrainPain	: actionlinks[i].func = A_BrainPain;	break;
		case nmBrainScream	: actionlinks[i].func = A_BrainScream;	break;
		case nmBrainDie		: actionlinks[i].func = A_BrainDie;		break;
		case nmBrainAwake	: actionlinks[i].func = A_BrainAwake;	break;
		case nmBrainSpit	: actionlinks[i].func = A_BrainSpit;	break;
		case nmSpawnSound	: actionlinks[i].func = A_SpawnSound;	break;
		case nmSpawnFly		: actionlinks[i].func = A_SpawnFly;		break;
		case nmBrainExplode	: actionlinks[i].func = A_BrainExplode;	break;

// killough code pointers
		case nmDetonate		: actionlinks[i].func = A_Detonate;		break;
		case nmMushroom		: actionlinks[i].func = A_Mushroom;		break;
		case nmDie			: actionlinks[i].func = A_Die;			break;
		case nmSpawn		: actionlinks[i].func = A_Spawn;		break;
		case nmTurn			: actionlinks[i].func = A_Turn;			break;
		case nmFace			: actionlinks[i].func = A_Face;			break;
		case nmScratch		: actionlinks[i].func = A_Scratch;		break;
		case nmPlaySound    : actionlinks[i].func = A_PlaySound;	break;
		case nmRandomJump	: actionlinks[i].func = A_RandomJump;	break;
		case nmLineEffect	: actionlinks[i].func = A_LineEffect;	break;

// b211-0 dbshotgun R3D specific
		case nmdbAttack		: actionlinks[i].func = A_dbAttack;		break;
		default:	// for compiler purposes
			break;
		}
	}
}

char *DefCheckBase(int *val)
{
	DFILE	*file;
	char	fname[256];
	int		i, k, len, read, csm;
	int		allowf = -1;
	byte	*data;
	;
	pr_OFdirect("Def check : ");

	// b211-0 Load the base r3ddefs ptrs...
	for (i = 0; r3ddefs[i].dedname; i++)
	{
		sprintf(fname, "%s\\defs\\base\\%s", ddRuntimeDir.path, r3ddefs[i].dedname);

		if ( (file = F_OpenFile( fname, "rb" )) == NULL)
		{
			*val = -1; return r3ddefs[i].dedname;
		}

		F_Seek(file, 0, SEEK_END);

		len = F_Tell(file);
		F_Rewind(file);
		
		data = (byte*)M_Calloc(len + 1, 0);
		
		read = F_Read(data, len, file);
		F_Close(file);

		for (csm = k = 0; k < read; k++)
			csm += (byte)(data[k]);

		M_Free(data);

		*val = r3ddefs[i].val == csm;

#ifdef _DEBUG
		pr_OFinsert("\n%2i) %15s :  now 0x%.8X  was 0x%.8X %s", i, r3ddefs[i].dedname, csm, r3ddefs[i].val, *val ? "SAME" : "DIFFERENT");
		continue;
#else
		if (!*val)
		{
			if (ArgExists("-allowfile"))
			{
				pr_OFinsert("\n%2i) %15s :  now 0x%.8X  was 0x%.8X\n", i, r3ddefs[i].dedname, csm, r3ddefs[i].val);
				continue;
			}

			*val = 2; return r3ddefs[i].dedname;
		}
#endif
		pr_OFinsert(" %i", *val);
	}

	pr_OFdirect("\n");

	*val = 1;
	return NULL;
}

//===========================================================================
// Def_Init
//	Initializes the databases.
// b211-0 revised
//===========================================================================

void Def_Init()
{
	// more of a reminder than a necessity
	sprnames = 0;	// Sprite name list etc.
	mobjinfo = 0;
	states = 0;
	sounds = 0;
	texts = 0;
	dettex = 0;
	stateowners = 0;

	has_dd_defn = W_CheckNumForName("DD_DEFNS") >= 0;
	DED_ZCount(&count_sprnames);
	DED_ZCount(&count_mobjinfo);
	DED_ZCount(&count_states);
	DED_ZCount(&count_sounds);
	DED_ZCount(&count_texts); 
	DED_ZCount(&count_details);
	DED_ZCount(&count_stateowners);

	DED_Init(&defs);
}

//===========================================================================
// Def_Destroy
//	Destroy databases.
//   Clear global variables pointing to definitions.
//===========================================================================
void Def_Destroy(void)
{
	if (defs_inited)
	{
		// To make sure...
		DED_Destroy(&defs);
		DED_Init(&defs);

		// Destroy the databases.
		DED_DelArray((void**)&sprnames	 , &count_sprnames);
		DED_DelArray((void**)&states	 , &count_states);
		DED_DelArray((void**)&mobjinfo	 , &count_mobjinfo);
		DED_DelArray((void**)&sounds	 , &count_sounds);
		DED_DelArray((void**)&texts		 , &count_texts);
		DED_DelArray((void**)&dettex	 , &count_details);
		DED_DelArray((void**)&stateowners, &count_stateowners);

		defs_inited = false;
	}

	mapinfo = NULL;
}

static RDBOOL checkNULLstate(char *name)
{
	if (!_strnicmp("DOGS_" , name, 5)) return true;
	if (!_strnicmp("OLDBFG", name, 6)) return true;
	if (!_strnicmp("PLS1",   name, 4)) return true;
	if (!_strnicmp("PLS2",   name, 4)) return true;
	if (!_strnicmp("BON3",   name, 4)) return true;
	if (!_strnicmp("BON4",   name, 4)) return true;
	if (!_strnicmp("BSKUL_", name, 6)) return true;
	return false;
}

//===========================================================================
// Def_GetXnum
//	Returns the number of the given object, or -1 if it doesn't exist.
//===========================================================================
char *getSpriteIDfromStateID(char *statename)
{
	ded_state_t *dst = defs.states;
	int			i;

	for (i = 0; i <  count_states.num; i++, dst++)
	{
		if (!stricmp(dst->id, statename)) return (char*)&dst->sprite;
	}

	return NULL;
}

char *getSpriteIDfromDoomednum(int dmedn)
{
	ded_mobj_t	*dmo;

	for (int i = 0; i < count_mobjinfo.num; i++)
	{
		dmo = defs.mobjs + i;
		if (dmo->doomednum == dmedn) return (char*)&dmo->id;
	}

	return NULL;
}

int Def_GetSpriteNum(char *name)
{
	int i;

	if (!name || !name[0]) return -1;

	for(i = 0; i < count_sprnames.num; i++)
	{
		if (!_stricmp(sprnames[i].name, name))
			return i;
	}

	return -1;
}

int Def_GetMobjNum(char *id)
{
	int i;

	if (!id || !id[0]) return -1;

	for (i = 0; i <  defs.count.mobjs.num; i++)
	{
		if(!stricmp(defs.mobjs[i].id, id))
			return i;
	}

	return -1;

}

static int lodstatestart = 0;

int Def_GetStateLODNum(char *id)
{
	int i;

	if (!id || !id[0]) return -1;

	for (i = lodstatestart; i < defs.count.states.num; i++)
	{
		if(!strcmp(defs.states[i].id, id))
			return i;
	}

	return -1;
}

char *getStateID(int stnum)
{
	ded_state_t *dst = defs.states + stnum;
	return dst->id;
}

int Def_GetStateNum(char *id)
{
	int i;

	if (!id || !id[0]) return -1;

	if (DEHchange & DEHF_STATE)
	{
		for (i = 0; i < defs.count.states.num; i++)
		{
			if (!strcmp(defs.states[i].id, id))
				return i;
		}
	}
	else
	{
		for (i = defs.count.states.num - 1; i >= 0; --i)
		{
			if (i >= S_OLDBFG1 && i <= S_MUSHROOM) continue;

			if (!strcmp(defs.states[i].id, id))
				return i;
		}
	}

	return -1;
}

int Def_GetSoundNum(char *id)
{
	int i;

	if (!id || !id[0]) return sfx_None;

	for(i = 0; i < defs.count.sounds.num; i++)
	{
		if (!strcmp(defs.sounds[i].id, id))
			return i;
	}
	return sfx_None;
}

char *Defs_GetTextPtr(int texn)
{
	if ((uint)texn >= (uint)count_texts.num)
		Con_Error("Defs_GetTextPtr: Invalid texts index");

	return defs.text[texn].text;

}

// v220-33
char *Def_IDfromType(int type)
{
	ded_mobj_t	*dmo = defs.mobjs;

	if (type >= 0 && type < count_mobjinfo.num)
		return dmo[type].id;
	return "UNKNOWN";
}

// v220-33
int Def_GetThingHeight(char *id)
{
	int i;

	for(i = 0; i < defs.count.mobjs.num; i++)
	{
		if(!stricmp(defs.mobjs[i].id, id))
			return (int)(defs.mobjs[i].height);
	}

	return 0;
}

static int getSpriteNumFromDoomEdNum(int ednum)
{
	int		mtnum, stnum, sprnm;
	char	*stid, *sprid;

	if (ednum >= 0)
		mtnum = P_GetTypeFromDoomednum(ednum);

	if (ednum < 0 || mtnum > 3006)
		SR_SCPErr("invalid doomednum %i", ednum);

	// translate the spawnstate num to the state num
	stnum = mobjinfo[mtnum].spawnstate;

	if (stnum < 0 || stnum >= S_MUSHROOM)
		SR_SCPErr("invalid doomednum %i", ednum);

	// get stateID from the state num
	stid = getStateID(stnum);

	// Quite a palava: Get the sprite id from stid
	sprid = getSpriteIDfromStateID(stid);

	sprnm = Def_GetSpriteNum(sprid);
#ifdef _DEBUG	// ---- $remove
	pr_OFinsert("spawnstate ednum %i, mtnum %i stid %s sprid %s sprnm %i\n",
		ednum, mtnum, stid, sprid, sprnm);
#endif	// -----------
	if (sprnm < 0 || sprnm >= DH_NUMSPRITES)
		SR_SCPErr("invalid sprite with doomednum %i", ednum);

	return sprnm;
}

//===========================================================================
// Def_GetSoundNumForName
//	Looks up a sound using the Name key. If the name is not found, returns
//	the NULL sound index (zero).
//===========================================================================
int Def_GetSoundNumForName(char *name)
{
	int i;

	if (!name || !name[0]) return 0;

	for(i = 0; i < defs.count.sounds.num; i++)
	{
		if(!_stricmp(defs.sounds[i].name, name))
			return i;
	}

	return 0;
}

int Def_GetMusicNum(char *id)
{
	int i;

	if (!id || !id[0]) return -1;

	for(i = 0; i < defs.count.music.num; i++)
	{
		if(!strcmp(defs.music[i].id, id))
			return i;
	}

	return -1;
}

int Def_GetActionNum(char *name)
{
	int i;
	
	if (!name || !name[0]) return 0;

	for (i = 0; i < nmEnd; i++)
	{
		if (!stricmp(name, actionlinks[i].name))	// Use stricmp to accommodate typos
			return actionlinks[i].ref;
	}

	pr_OFinsert("GetActionPtr: action \"%s\" not found.\n", name);
	return 0;
}

ded_mapinfo_t *Def_GetMapInfo(char *map_id)
{
	int i, k = defs.count.mapinfo.num;

	for (i = 0; i < k; i++)
	{
		if(!_stricmp(defs.mapinfo[i].id, map_id))
			return defs.mapinfo + i;
	}

	return NULL;
}

int Def_GetFlagValue(char *flag)
{
	int i;

	if (!flag || !flag[0]) return 0;

	for(i = defs.count.flags.num-1; i >= 0; i--)
	{
		if(!strcmp(defs.flags[i].id, flag))
			return defs.flags[i].value;
	}

	pr_OFinsert("GetFlagValue: Undefined flag '%s' %i.\n", flag, defs.count.flags.num);
	return 0;
}

int Def_EvalFlags(char *ptr)
{
	int value = 0, len;
	char buf[64];
	
	while(*ptr)
	{
		ptr = M_SkipWhite(ptr);
		len = M_FindWhite(ptr) - ptr;
		strncpy(buf, ptr, len);
		buf[len] = 0;
		value |= Def_GetFlagValue(buf);
		ptr += len;
	}
	return value;
}

//===========================================================================
// Def_ModifyTextDef - R3D
//	Escape sequences are un-escaped (\n, \r, \t, \s, \_).
//  texn points to defs.text[texn].index not the def number
//===========================================================================
char *Def_ModifyTextDef(int texn, char *str)
{
	ddtext_t	*txt = texts + texn;

	if ((uint)texn >= (uint)count_texts.num)
		Con_Error("Def_ModifyTextDef: Invalid texts index");

	if (!str || !str[0]) // null pointer handling
	{
		txt->text = (char*)M_Realloc(txt->text, 1, 18);
		*txt->text = 0;
	}
	else
	{
		char		*out, *in;
		unsigned	len = strlen(str) + 1;

		// This is safe as the length of out will always
		// end up being <= to the length of  str
		txt->text = (char*)M_Realloc(txt->text, len, 19);
		memset(txt->text, 0, len);

		for (out = txt->text, in = str; *in; out++, in++)
		{
			if (*in == '\\') 
			{
				in++;
				if      (*in == 'n') *out = '\n';				// Newline.
				else if (*in == 'r') *out = '\r';				// Carriage return.
				else if (*in == 't') *out = '\t';				// Tab.
				else if (*in == '_' || *in == 's') *out = ' ';	// Space.
				else *out = *in;
				continue;
			}
			*out = *in;
		}

		// Adjust buffer to fit exactly 
		txt->text = (char*)M_Realloc(txt->text, strlen(txt->text) + 1, 20);
	}

	defs.text[texn].text = txt->text;	// Set the new defn ptr

	return (char*)txt->text;
}

//===========================================================================
// Def_ReadIncludedDEDs
//===========================================================================
void Def_ReadIncludedDEDs(directory_t *mydir)
{
	char fn[MAXPATH];

	// Also read the files that were listed as includes.
	while(defs.count.includes.num > 0)
	{
		strcpy(fn, defs.includes[0].path);

		// Remove the name from the list before reading it.
		// This'll prevent re-reading.
		DED_DelEntry(0, (void**)&defs.includes, &defs.count.includes, 
			sizeof(*defs.includes));
		Def_ReadProcessDED(fn, DF_THISPATH);
		// Back to our dir.
		Dir_ChDir(mydir);
	}
}

//==========================================================================
// Def_CheckFileID
//	Returns true if the given file can be read, or false if it has already
//	been read.
//==========================================================================
static RDBOOL Def_CheckFileID(char *path)
{
	int i, id;

	if(!F_Access(path))
	{
		if(verbose) pr_OFinsert("CheckFile: %s not found.\n", path);
		return false;
	}

	for (i = 0, id = Dir_FileID(path); i < read_count; i++)
		if(read_ids[i] == id) return false;

	read_ids[read_count++] = id;

	return true;
}

//===========================================================================
// Def_ReadDEDFile
//	Callback for DD_ReadProcessDED.
//===========================================================================
int Def_ReadDEDFile(char *fn, int parm)
{
	if (read_count == MAX_READ)
	{
		Con_Error("DEDFILE INDEX of %i exceeded - please report", MAX_READ);
	}

	if (Def_CheckFileID(fn) && !DED_Read(&defs, fn, read_count==1))
	{
		Con_Error("ReadDEDFile: Reading of %s failed.\n  %s\n", 
			fn, dedReadError);
	}

	return true;
}

//===========================================================================
// Def_ReadProcessDED
//  b211-0 Revised
//===========================================================================

static char *pext(int type)
{
	switch(type)
	{
	case DF_DEFBASE : return "Base\\";
	case DF_DEFMODEL: return "Model\\";
	case DF_ABBAPATH: return "Model\\Abbspack\\";
	case DF_WSMPATH : return "Model\\Sitters\\";	// b220-1
	default: return "";
	}
}

static void Def_ReadProcessDED(char *filename, int type)
{
	directory_t dir;
	char	fn[MAXFNAMEX], fname[MAXPATHNAMESIZE];
	uint	c = strlen(ddRuntimeDir.path);

	strcpy(fname, filename);

	switch(type)
	{
	case DF_UDIRPATH:
		if (has_dd_defn) { Con_Printf(warn, "UDIR", filename) ; return; }
		sprintf(fname, "%s%s", r3dUserDir.path, filename);
		break;

	case DF_FULLPATH:
		if (has_dd_defn)
		{
			char comp[256];
			uint n;

			// b220-0 Custom\*.deds are still allowed even if
			//        the file has a DD_DEFNS lump
			sprintf(comp, "%sDefs\\Custom", ddRuntimeDir.path);
			n = strlen(comp);

			if (n >= strlen(filename) || strnicmp(comp, filename, n))
			{
				Con_Printf(warn, "Command line", filename);
				return;
			}
		}
		break;

	default:
		break;
	}

	if (!Dir_IsAbsolute(fname))
	{
		sprintf(fname, "%sDefs\\%s", ddRuntimeDir.path, pext(type));
	}

	Dir_FileName(filename, fn);
	Dir_FileDir(fname, &dir);

	sprintf(fname, "%s%s", dir.path, fn);

	// Change to the directory of the file we're about to read.

	// b211-0 Check the definition file and ensure we don't allow
	//        non-existant filenames to pass through unnoticed.
	if (!Dir_ChDir(&dir) || !F_ForAll(fn, 0, Def_ReadDEDFile))
	{
		if (   (type == DF_DEFMODEL || type == DF_WSMPATH)	// v220-3
			&& !ge_inf.hasscp_defn
			&& !_stricmp(fn, "Monsters.ded"))
		{
			Con_Message("WARNING: r3d-models not installed");
			return;
		}

		Con_Error("%s not found%s", fname,
			(type == DF_DEFMODEL || type == DF_WSMPATH) ?	// v220-3
			"\nTO FIX: Install r3d-models" : "");
	}

	if (strnicmp(ddRuntimeDir.path, fname, c)) c = 0;

	pr_OFinsert(" %s\n", fname + c);

	if (ge_inf.useSitters || (ge_inf.hasdd_defns && ge_inf.hasscp_defn))
	{
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Decorations.ded"	)) SittersLoaded  = 1;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\FlameModels.ded"	)) SittersLoaded |= 2;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Items.ded"		)) SittersLoaded |= 4;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\HUDWeapons.ded"	)) SittersLoaded |= 8;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Monsters.ded"	)) SittersLoaded |= 16;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Players.ded"		)) SittersLoaded |= 32;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Splashes.ded"	)) SittersLoaded |= 64;
		if (!stricmp(fname + c, "Defs\\Model\\Sitters\\Weapons.ded"		)) SittersLoaded |= 128;
	}

	Def_ReadIncludedDEDs(&dir);
}

//===========================================================================
// Def_CountMsg
//	Prints a count with a 2-space indentation.
//===========================================================================
void Def_CountMsg(int count, const char *label)
{
	if(!verbose && !count) return; // Don't print zeros if not verbose.
	pr_OFinsert("%20s : %5i\n", label, count);
}

//===========================================================================
// DD_ReadLumpDefs
//	Reads all DD_DEFNS lumps found in the lumpinfo. 
//===========================================================================
void Def_ReadLumpDefs(void)
{
	int i, c;

	cache_ded_model = true;

	for(i = 0, c = 0; i < numlumps; i++)
	{
		if(!strnicmp(lumpinfo[i].name, "DD_DEFNS", 8))
		{
			if(!DED_ReadLump(&defs, i, false))
				Con_ErrorD("Syntax error in DD_DEFNS lump");
			
			c++;
		}
	}

	if (c)	// b211-0
		Con_Printf(" %i DD_DEFNS lump%s read\n", c, plural(c));

	// Read any included files.
	Def_ReadIncludedDEDs(&ddRuntimeDir);

	cache_ded_model = false;
}
int prc;
//===========================================================================
// Def_StateForMobj
//	Plug the inf ptr into the stateowners array - needed for client mobjs.
//===========================================================================
static int Def_StateForMobj(char *state_id, mobjinfo_t *inf)
{
	int num = Def_GetStateNum(state_id);
	int rev = inf->doomednum == 66;
	int stnum;

	// State zero is the NULL state.
	if (num > 0)
	{
		stateowners[num] = inf;

		stnum = states[num].nextstate;

		if (prc) pr_OFinsert("PRC %s\n", state_id); 

		while (stnum > 0 && !stateowners[stnum])
		{
			if (prc) pr_OFinsert("PRC %i\n", stnum); 
			stateowners[stnum] = inf;
			stnum = states[stnum].nextstate;
		}
	}
	else
		if (prc) pr_OFinsert("PRC end\n"); 

	return num > 0 ? num : 0;	// In case it was -1
}


// b211-0 iwad level names for plutonia/tnt names are read
//        in from the MNAMES lump in Risen3D.out
static void IWADMapName()
{
	char *iwadtype, *nextstr;
	int  i, lumpnum = W_CheckRisen3DWADNumForName("MNAMES", true);

	if (lumpnum < 0) goto error_out;

	if (gamemission == pack_plut)
		iwadtype = "PLUTONIA";
	else if (gamemission == pack_tnt)
		iwadtype = "TNT";
	else
		return;

	SR_LoadLump(lumpnum, "MNAMES");

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!strcmp(nextstr, iwadtype))
		{
			nextstr = SR_NextSTR(false);

			if (strcmp(nextstr, "{"))
				Con_ErrorD("Invalid MNAMES lump");

			while(1)
			{
				char id[8];

				if (!(nextstr = SR_NextSTR(false))) goto error_out;

				if (!strcmp(nextstr, "}")) break;

				if (!M_ValidateID(nextstr)) goto error_out;

				strcpy(id, nextstr);

				if (!(nextstr = SR_NextSTR(false))) goto error_out;

				// Search map ids
				for (i = 0; i < defs.count.mapinfo.num; i++)
				{
					if (!_stricmp(defs.mapinfo[i].id, id))
					{
						strcpy(defs.mapinfo[i].mapname, nextstr);
						break;
					}
				}
			}
			break;
		}
		else
		{
			while(nextstr = SR_NextSTR(false))
			{
				if(!strcmp(nextstr, "}"))
					break;
			}
		}
	}

	SR_EndScript();
	return;

error_out:;
	Con_ErrorD("Risen3D.wad - Wrong version");
}

// Set defaults
static void setDMTdefaults(ded_mapinfo_t *dmt, char *name)
{
	dmt->author[0] = 0;
	strcpy(dmt->sky_layers[0].flags, "slf_enable"); 
	strcpy(dmt->sky_layers[0].texture, "SKY1"); 
	dmt->sky_height = .49f;
	dmt->horizon_offset = -0.105f;
	dmt->gravity = 1.0f;
	dmt->mapinfomuslump = -1;
	dmt->muslump = -1;
	dmt->music.path[0] = 0;
	strcpy(dmt->id, name);
	dmt->R3Dflags |= R3Dhasmapinfo;
}

static RDBOOL SetUpMIParms()
{
	ded_mapinfo_t *dmt = NULL;
	RDBOOL	valid_next = false;
	RDBOOL	needs_flat = false;
	RDBOOL	braceset = false;
	RDBOOL	fiparmset = false;
	int		clusternum, musnum, picnum;
	char	*nextstr;

	if (!SR_LoadPWADLumpName("MAPINFO")) return false;

	while( (nextstr = SR_NextSTR(false)) )
	{
		// v220-33 IWAD check
		if (!strnicmp(nextstr, "IWAD", 4))
		{
			nextstr = SR_NextSTR(false);

			if (strlen(nextstr) > 6)
			{
				if (strstrCI(nextstr, "DOOM.WAD") && gamemission == doom) continue;
				if (strstrCI(nextstr, "DOOM1.WAD") && gamemission == doom) continue;
				if (strstrCI(nextstr, "DOOM2.WAD") && gamemission == doom2) continue;
				if (strstrCI(nextstr, "PLUTONIA.WAD") && gamemission == pack_plut) continue;
				if (strstrCI(nextstr, "TNT.WAD") && gamemission == pack_tnt) continue;
			}
			Con_Error("Wrong iwad selected\nRestart using %s", nextstr);
		}
		else
		if (!_stricmp(nextstr, "map"))
		{
			dmt = NULL;
			
			if (braceset) SR_SCPErr("Unmatched {");
			
			braceset = false;
			
			nextstr = SR_NextSTR(true);

			if (!M_ValidateID(nextstr) ) continue;

			if ( !(dmt = Def_GetMapInfo(nextstr)) )
			{
				dmt = (ded_mapinfo_t*)DED_NewEntry( (void**) &defs.mapinfo,
						&defs.count.mapinfo, sizeof(ded_mapinfo_t));
			}

			fiparmset = clusternum = musnum = picnum = 0;

			setDMTdefaults(dmt, nextstr);

			nextstr = SR_NextSTR(true);
			strncpy(dmt->mapname, nextstr, DED_STRINGID_LEN);
			dmt->mapname[DED_STRINGID_LEN] = 0;
			dmt->R3Dflags |= R3Dmapnameset;

			nextstr = SR_NextSTR(true);

			if (!_stricmp(nextstr, "{")) braceset = true;
		}

		// These are dealt with in r3d_mapinfo.c (MI_GetBriefings) - not here.
		if (!_stricmp(nextstr, "clusterdef"))
			dmt = NULL;

		if (!dmt) continue;

		if (!_stricmp(nextstr, "}"))
		{
			if (fiparmset && clusternum > 0 && clusternum < 100)
			{
				fiparms[clusternum][0] = musnum;
				fiparms[clusternum][1] = picnum;
			}

			fiparmset = braceset = false;
		}

		if (!_stricmp(nextstr, "map07special"))
			dmt->R3Dflags |= map07special;
		else if (!_stricmp(nextstr, "baronspecial"))
			dmt->R3Dflags |= baronspecial;
		else if (!_stricmp(nextstr, "cyberdemonspecial"))
			dmt->R3Dflags |= cyberdemonspecial;
		else if (!_stricmp(nextstr, "spidermastermindspecial"))
			dmt->R3Dflags |= spidermastermindspecial;
		else if (!_stricmp(nextstr, "specialaction_exitlevel"))
			dmt->R3Dflags |= specialaction_exitlevel;
		else if (!_stricmp(nextstr, "specialaction_blazedoor"))
			dmt->R3Dflags |= specialaction_blazedoor;
		else if (!_stricmp(nextstr, "specialaction_opendoor"))
			dmt->R3Dflags |= specialaction_opendoor;
		else if (!_stricmp(nextstr, "specialaction_lowerfloor"))
			dmt->R3Dflags |= specialaction_lowerfloor;
		else if (!_stricmp(nextstr, "specialaction_killmonsters"))
			dmt->R3Dflags |= specialaction_killmonsters;
		else if (!_stricmp(nextstr, "nointermission"))
			dmt->R3Dflags |= nointermission;
		else if (!_stricmp(nextstr, "resetinventory"))
			dmt->R3Dflags |= resetinventory;
		else if (!_stricmp(nextstr, "R3Dtype"))
		{
			nextstr = SR_NextSTR(true);

			if (!_stricmp(nextstr, "carrystatsover")) dmt->R3Dflags |= R3Dcarrystatsover;
			else
			if (!_stricmp(nextstr, "newmapfog")) dmt->R3Dflags |= R3Dnewmapfog;
		}
		else if (!_stricmp(nextstr, "sky1"))
		{
			ded_skylayer_t	*sl = dmt->sky_layers;
			// v220-9 modify to allow absence of rotate value
			RDBOOL			scperror = false;
			float			tmp;

			strcpy(sl->texture, SR_NextSTR(true));
			SR_NextSTR(true);
			tmp = SR_FloatForString(&scperror);
			if (scperror) SR_KeepString();

			sl->skyAngleIncrement = tmp/PI;
			if (sl->skyAngleIncrement >  1.0f) sl->skyAngleIncrement =  1.0f;
			if (sl->skyAngleIncrement < -1.0f) sl->skyAngleIncrement = -1.0f;

			strcpy(sl->flags, "slf_enable");
		}
		else if (!_stricmp(nextstr, "author"))
		{
			nextstr = SR_NextSTR(true);
			strncpy(dmt->author, nextstr, DED_STRINGID_LEN);
			dmt->author[DED_STRINGID_LEN] = 0;
		}
		else if (!_stricmp(nextstr, "next"))
		{
			nextstr = SR_NextSTR(true);

			if (!_stricmp(nextstr, "EndGame"))
			{
				nextstr = SR_NextSTR(true);

				if (strcmp(nextstr, "{")) SR_KeepString();
				else
				{
					RDBOOL endcb = false;
					
					fiparmset = true;

					while( (nextstr = SR_NextSTR(false)) )
					{
						if (!strcmp(nextstr, "}")) { endcb = true; break; }

						if (!_stricmp(nextstr, "pic"))
						{
							nextstr = SR_NextSTR(true);
							picnum = W_CheckNumForName(nextstr);
						}

						if (!_stricmp(nextstr, "music"))
						{
							nextstr = SR_NextSTR(true);
							musnum = W_CheckNumForName(nextstr);
						}
					}
					if (!endcb) SR_SCPErr("Missing }");
				}

				dmt->R3Dflags |= R3Dendgame;
			}
			else
			if (!_stricmp(nextstr, "EndGameC"))
			{
				dmt->R3Dflags |= R3DendgameC;
			}
			else if (M_ValidateID(nextstr))
			{
				strncpy(dmt->nextmap, nextstr, 7);
				dmt->nextmap[7] = 0;

				if (W_CheckPWADNumForName(nextstr) < 0)
					dmt->R3Dflags |= R3Dendmaps;
			}
		}
		else if (!_stricmp(nextstr, "secretnext"))
		{
			nextstr = SR_NextSTR(true);

			if (M_ValidateID(nextstr))
			{
				strncpy(dmt->secretnext, nextstr, 7);
				dmt->secretnext[strlen(nextstr)] = 0;
			}
		}
		else if (!_stricmp(nextstr, "fade")
			|| !_stricmp(nextstr, "outsidefog"))
		{
			RDBOOL	outer = false;
			int		num = 0x80;	// default

			// v220-3
			if (!_stricmp(nextstr, "outsidefog"))
			{
				strcpy(dmt->flags, "mif_fogoutside");
				outer = true;
			}
			else
			if (!_stricmp(nextstr, "underwaterfog"))
			{
				strcpy(dmt->flags, "mif_fogunderwater");
			}
			else
				strcpy(dmt->flags, "mif_fogexp");

			nextstr = SR_NextSTR(true);	// fog color (depth of grey)

			if (!strnicmp(nextstr, "grey", 4) || !strnicmp(nextstr, "gray", 4))
			{
				num = (int)(atoi(&nextstr[4]) * 255.0f/100.0f);

				if (num <   0) num =   0;
				if (num > 255) num = 255;
			}

			if (outer)
			{	// Linear fog
				dmt->fog_start =  200.0f;
				dmt->fog_end   = 1800.0f;
				dmt->fog_density = 0.0f;
			}
			else
			{	// exp fog
				dmt->fog_start = 0.0f;
				dmt->fog_end   = 0.0f;
				dmt->fog_density = 0.0005f;
			}

			dmt->fog_color[0] = (byte)num;
			dmt->fog_color[1] = (byte)num;
			dmt->fog_color[2] = (byte)num;
			dmt->fog_color[3] = 0xFF;
		}
		else if (!_stricmp(nextstr, "cluster"))
		{
			dmt->cluster = SR_NextNum();
			clusternum = dmt->cluster;
		}
		else if (!_stricmp(nextstr, "music"))
		{
			int n;

			nextstr = SR_NextSTR(true);

			// v220-24
			if (!strncmp(nextstr, "$MUSIC_E", 8) && strlen(nextstr) == 11)
			{
				char mstr[8];

				sprintf(mstr, "D_%s", nextstr + 7);
				mstr[6] = 0;

				if ( (n = W_CheckNumForName(mstr)) >= 0)
					dmt->mapinfomuslump = n;
			}
			else
			if ( (n = W_CheckNumForName(nextstr)) >= 0)
				dmt->mapinfomuslump = n;
		}
		else if (!_stricmp(nextstr, "par"))
		{
			nextstr = SR_NextSTR(true);
			dmt->partime = atoi(nextstr);
		}
		else if (!_stricmp(nextstr, "titlepatch"))
		{
			int pic = W_CheckNumForName(SR_NextSTR(true));
			if (pic >= 0) dmt->titlepatch = pic;
		}
		else if (!_stricmp(nextstr, "enterpic")
			|| !_stricmp(nextstr, "exitpic"))
		{
			nextstr = SR_NextSTR(true);

			if (!dmt->epitext[0] && nextstr[0] == '$')
				strncpy(dmt->epitext, &nextstr[1], DED_STRINGID_LEN);
		}
	}

	SR_EndScript();

	return true;
}

// v220-18 Only check for this if no MAPINFO lump exists
static void SetUpEtParms()
{
	ded_mapinfo_t	*dmt = NULL;
	char			scratch[128];
	char			*nextstr;
	int				len, n;

	if (!SR_LoadPWADLumpName("EMAPINFO")) return;
	
	EMAPINFO_script = true;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (   strlen(nextstr) == 6
			&& nextstr[0] == '['
			&& nextstr[1] == 'E'
			&& nextstr[3] == 'M'
			&& nextstr[5] == ']')
		{
			int ep = nextstr[2] - '0';
			int mp = nextstr[4] - '0';

			if (ep > 0 && ep < 10 && mp >= 0 && mp < 10)
			{
				char name[6];

				strncpy(name, nextstr + 1, 4);
				name[4] = 0;

				if ( !(dmt = Def_GetMapInfo(name)) )
				{
					dmt = (ded_mapinfo_t*)DED_NewEntry( (void**) &defs.mapinfo,
							&defs.count.mapinfo, sizeof(ded_mapinfo_t));
				}
				
				setDMTdefaults(dmt, name);
			}
		}
		else if (dmt)
		{
			if (!_stricmp(nextstr, "levelname"))
			{
				if ( (nextstr = SR_NextSTR(true)) )
				{
					strcpy(scratch, nextstr);
					len = strlen(scratch);
					if (len > DED_STRINGID_LEN) scratch[DED_STRINGID_LEN] = 0;
					strcpy(dmt->mapname, scratch);
					dmt->R3Dflags |= R3Dmapnameset;
				}
			}
			else
			if (   !_stricmp(nextstr, "music")
				&& (nextstr = SR_NextSTR(true)) )
			{
				if (stricmp(nextstr, "D_"))
				{
					strcpy(scratch, "D_");
					strcat(scratch, nextstr);
				}
				else
				{
					strcpy(scratch, nextstr);
				}

				if ( (n = W_CheckNumForName(scratch)) >= 0)
					dmt->mapinfomuslump = n;
			}
			else if (!_stricmp(nextstr, "levelpic")
				&& (nextstr = SR_NextSTR(true)) )
			{
				n = W_CheckNumForName(nextstr);
				if (n >= 0) dmt->titlepatch = n;
			}
			else if (!_stricmp(nextstr, "nextlevel")
				&& (nextstr = SR_NextSTR(true)) )
			{
				strcpy(scratch, nextstr);
				len = strlen(scratch);
				if (len > 8) scratch[8] = 0;
				strcpy(dmt->nextmap, scratch);
			}
			else if (!_stricmp(nextstr, "nextsecret")
				&& (nextstr = SR_NextSTR(true)) )
			{
				strcpy(scratch, nextstr);
				len = strlen(scratch);
				if (len > 8) scratch[8] = 0;
				strcpy(dmt->secretnext, scratch);
			}
			else if (!_stricmp(nextstr, "skyname")
				&& (nextstr = SR_NextSTR(true)) )
			{
				strcpy(scratch, nextstr);
				len = strlen(scratch);
				if (len > 8) scratch[8] = 0;
				strcpy(dmt->sky_layers[0].texture, scratch); 
			}
			else if (!_stricmp(nextstr, "boss-specials")
				&& (nextstr = SR_NextSTR(true)) )
			{
				if ( !_stricmp(nextstr, "E1M8") )
					dmt->R3Dflags |= (baronspecial|specialaction_lowerfloor);
				else if ( !_stricmp(nextstr, "E2M8") )
					dmt->R3Dflags |= (cyberdemonspecial|specialaction_exitlevel);
				else if ( !_stricmp(nextstr, "E3M8") )
					dmt->R3Dflags |= (spidermastermindspecial|specialaction_exitlevel);
				else if ( !_stricmp(nextstr, "E4M6") )
					dmt->R3Dflags |= (cyberdemonspecial|specialaction_blazedoor);
				else if ( !_stricmp(nextstr, "E4M8") )
					dmt->R3Dflags |= (spidermastermindspecial|specialaction_lowerfloor);
			}
		}
	}

	EMAPINFO_script = false;
	SR_EndScript();
}

static int TPnmapID(char *id)
{
	char	newid[8];
	int		lump;

	if (commercial_game)
	{
		sprintf(newid, "CWILV%.2i", atoi(&id[3]) - 1);
	}
	else
	{
		sprintf(newid, "WILV%c%c", id[1] - 1, id[3] - 1);
	}

	if ( (lump = W_CheckOriginalPatch(newid)) >= 0)
	{
		lumpinfo[lump].liflags |= LINF_MAPNAME;
		return lump;
	}

	return -1; 
}

//===========================================================================
//
//===========================================================================
static void Def_SetMapMusicNum(ded_mapinfo_t *mt)
{
	int		episode = commercial_game ? 0 : mt->id[1] - '0';
	int		map = commercial_game ? atoi(&mt->id[3]) : mt->id[3] - '0';
	int		mnum = -1;

	if (commercial_game)
	{
		mnum = mus_runnin + map - 1;
	}
	else
	{
		int spmus[]=
		{
			// Translate episode 4 to various from epis 1 to 3
			mus_e3m4,	// American		e4m1
			mus_e3m2,	// Romero		e4m2
			mus_e3m3,	// Shawn		e4m3
			mus_e1m5,	// American		e4m4
			mus_e2m7,	// Tim 			e4m5
			mus_e2m4,	// Romero		e4m6
			mus_e2m6,	// J.Anderson	e4m7 CHIRON.WAD
			mus_e2m5,	// Shawn		e4m8
			mus_e1m9	// Tim			e4m9
		};
		
		if (episode < 4)
			mnum = mus_e1m1 + (episode-1)*9 + map-1;
		else if (episode == 4)
			mnum = spmus[map-1];
	}
	
	if (mnum >= 0 && mnum < defs.count.music.num)
	{
		ded_music_t *def = defs.music + mnum;

		// Set the default
		mt->muslump = W_CheckNumForName(def->lumpname);

		if (def->path.path[0])
		{
			strcpy(mt->music.path, def->path.path);
		}
	}
	else if (mt->muslump < 0
		&& mt->mapinfomuslump >= 0)
	{
		mt->muslump = mt->mapinfomuslump;
	}

	// Transfer value (either a replacement lump or -1 if none defined)
	if (mt->R3Dflags & R3Dhasmapinfo)
	{
		// If a mapinfo has no music defined but a replacement
		// music lump was found in the pwad then this is allowed.
		if ( !(mt->muslump >= 0
			&& mt->mapinfomuslump == -1
			&& pwadHandle(lumpinfo[mt->muslump].handle)) )
		{
			mt->muslump = mt->mapinfomuslump;
		}
	}

	// Check any ded path set is valid
	if (   mt->music.path[0]
		&& !M_PrependBaseToFile(mt->music.path))
	{
		mt->music.path[0] = 0;
	}
	
	return;
}

// Order must be kept
static char *model_defs[] = {
	"Decorations.ded",
	"FlameModels.ded",
	"Items.ded",
	"HUDWeapons.ded",
	"Monsters.ded",
	"Players.ded",
	"Splashes.ded",
	"Weapons.ded",
	"Lod.ded",
	NULL
};

// v220-6
RDBOOL BaseModelDefsExist()
{
	char fname[512];
	int	 i, ret = 0;

	for (i = 0; model_defs[i]; i++)
	{
		if (ge_inf.useSitters)
			sprintf(fname, "%sDefs\\Model\\Sitters\\%s", ddRuntimeDir.path, model_defs[i]);
		else
			sprintf(fname, "%sDefs\\Model\\%s", ddRuntimeDir.path, model_defs[i]);

		if ( (_access(fname, 4)) != -1) ret |= (1 << i);
	}
			
	return ret;
}

// v220-13 Replacement sprite->model compatibilty (read in from R3D_WADmodify)
// v220-16 extended
static void subfn_Check_WADmodify()
{
	char	*vstr[] = { "sprites", "states" };
	char	fname[512], *nextstr;
	char	wname[128];
	RDBOOL	found = false;
	RDBOOL	firsttime = false;
	int		n, val, scperror;
	int		done[2] = { 0, 0 };
	int		action_num;
	uint	numchar;

	if (!CFX) return;

	sprintf(fname, "%sR3D_WADmodify", ddBinDir.path);

	if (!SR_LoadFile(fname, "R3D_WADmodify", NULL)) return;

	sprintf(wname, "[%s]", analysename);

	nextstr = SR_NextSTR(false);
	if (!nextstr) goto end_WADcheck;

	if (_stricmp("Version", nextstr) || SR_NextNum() != 3)
	{
		pr_OFdirect("ERROR: the R3D_WADmodify file is invalid\n");
		goto end_WADcheck;
	}
	
	while ( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, wname))
		{
			found = true; break;
		}

		if (   nextstr[0] == '['
			&& nextstr[1] == '\\'
			&&  (numchar = strlen(nextstr)) > 3
			&& nextstr[numchar - 1] == ']')
		{
			numchar -= 3;	// Eliminate end ] from count

			if (strlen(analysename) >= numchar && !_strnicmp(analysename, nextstr + 2, numchar))
			{
				found = true; break;
			}
		}
	}

	if (!found) goto end_WADcheck;

	pr_OFinsert("Wadmodify    : %s\n", wname);

find_next:;

	n = -1;

	while ( (nextstr = SR_NextSTR(false)) )
	{
		if (nextstr[0] == '[') goto end_WADcheck;

		if (!done [0] && !_stricmp(nextstr, vstr[0]))
		{
			done[0] = true; n = 0; break;
		}

		if (!done [1] && !_stricmp(nextstr, vstr[1]))
		{
			done[1] = true; n = 1; break;
		}
	}

	if (n < 0 || !nextstr)  goto end_WADcheck;

	while ( (nextstr = SR_NextSTR(false)) )
	{
		if (nextstr[0] == '[') goto end_WADcheck;

		if (!n && !stricmp(nextstr, "ALL"))
		{
			ge_inf.replaceall = ge_inf.amwds;
			pr_OFdirect("R3D_WADmodify: replace all sprites with models\n");
			goto find_next;
		}

		// v220-21 add keyword
		if (!n && !stricmp(nextstr, "ALLMODELSBAR"))
		{
			int i, term = 0, ednum, sprnm;

			pr_OFdirect("R3D_WADmodify: suppress models with list entries\n");

			for (i = 0; i < DH_NUMSPRITES; i++) DEHCheck[i] |= DEHF_USEMODEL;

			nextstr = SR_NextSTR(true);

			if (_stricmp(nextstr, "{")) SR_SCPErr("Missing {");

			while(1)
			{
				if ( !(nextstr = SR_NextSTR(false)) ) break;
				if (!_stricmp(nextstr, "}")) { term = true; break; }

				// find sprite from doomednum
				ednum = atoi(nextstr);
				sprnm = getSpriteNumFromDoomEdNum(ednum);
				DEHCheck[sprnm] &= ~DEHF_USEMODEL;
			}

			if (!term) SR_SCPErr("Missing }");

			goto find_next;
		}

		val = SR_NumForString(&scperror);

		if (scperror)
		{
			SR_KeepString();
			goto find_next;
		}
		else
		{
			if (!n)
			{
				if (val < 0 || val >= DH_NUMSPRITES)
					SR_SCPErr("Invalid sprite number %i", val);
			}

			if (!n)
			{
				DEHCheck[val] |= DEHF_USEMODEL;
			}
			else if (n == 1)
			{
				state_t	*st;

				if (val < 0 || val >= DH_NUMSTATES)
					SR_SCPErr("Invalid state number %i", val);

				st = &states[val];

				while (1)
				{
					if (!(nextstr = SR_NextSTR(false)) ) break;

					if (!stricmp(nextstr, "tics"))
					{
						if (!(nextstr = SR_NextSTR(false)) ) break;
						val = SR_NumForString(&scperror);
						if (scperror)
							SR_SCPErr("Invalid state tics \"%s\"", nextstr);
						st->tics = val;
					}
					else
					if (!stricmp(nextstr, "flags"))
					{
						if (!(nextstr = SR_NextSTR(false)) ) break;
						val = SR_NumForString(&scperror);
						if (scperror)
							SR_SCPErr("Invalid state flags \"%s\"", nextstr);
						st->stflags = val;
					}
					else
					if (!stricmp(nextstr, "frame"))
					{
						if (!(nextstr = SR_NextSTR(false)) ) break;
						val = SR_NumForString(&scperror);
						if (scperror)
							SR_SCPErr("Invalid state flags \"%s\"", nextstr);
						st->frame = val;
					}
					else
					if (!stricmp(nextstr, "action"))
					{
						if (!(nextstr = SR_NextSTR(false)) ) break;
						if (!(action_num = Def_GetActionNum(nextstr)))
							SR_SCPErr("Invalid action name \"%s\"", nextstr);
						st->action_num = action_num;
					}
					else
					if (!stricmp(nextstr, "killLight"))
					{
						st->light = NULL;
					}
					else
					{
						SR_KeepString();
						break;
					}
				}
			}
		}
	}

end_WADcheck:;

	SR_EndScript();
}

RDBOOL		XYonLine(fixed_t x, fixed_t y, line_t *l, fixed_t r);

typedef struct {
	fixed_t		x;
	fixed_t		y;
	fixed_t		radius;
	int			line;

} clinelist_t;

clinelist_t cll;

static RDBOOL PIT_CheckLineList(line_t* ld, void*)
{
	if (   ld - lines == cll.line
		&& XYonLine(cll.x, cll.y, ld, cll.radius))
	{
		return false;
	}

	return true;
}

static RDBOOL thingSpansLineCLL()
{
	int		bx, by, xl, xh, yl, yh;
	fixed_t	thbox[4];

	M_setBoxOffsets(thbox, cll.x, cll.y, cll.radius);
	M_getBlockOffsets(thbox, 0, &xl, &xh, &yl, &yh);

	validcount++;

	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			if (!P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_CheckLineList, NULL))
				return true;
		}
	}

	return false;
}

//
// P_GetTypeFromDoomednum
//
// Finds a mobj type from its doomednum
//
// killough 8/24/98: rewritten to use hashing
//
// b211-0 Ammended to use R3D win memory routine
//        and deal with variable numbers of mobjs

static struct dmedhash_s { int first, next; } *dmednumhash = NULL;

void Def_SetHashTable(int num)	// where num must be stats.num_mobjs
{
	int i;

	dmednumhash = (struct dmedhash_s*)W_Malloc(sizeof *dmednumhash * num, STATICID, 0);

	for (i = 0; i < num; i++) dmednumhash[i].first = num;
	for (i = 0; i < num; i++)
	{
		if (mobjinfo[i].doomednum != -1)
		{
			unsigned h = (unsigned) mobjinfo[i].doomednum % num;
			dmednumhash[i].next = dmednumhash[h].first;
			dmednumhash[h].first = i;
		}
	}
}

int P_GetTypeFromDoomednum(int type)
{
	int i;
	
	i = dmednumhash[type % stats.num_mobjs].first;

	while (i < stats.num_mobjs && mobjinfo[i].doomednum != type)
		i = dmednumhash[i].next;

	return i;
}

// v220-16
void R3D_SetParms(char *mapid)
{
	thinker_t	*thinker;
	int			mnum;
	int			line;
	int			radius;
	int			zpos;
	int			val;
	int			scperror;
	char		fname[512], *nextstr;
	char		wname[128];
	int			error = 0;

	//------------------------------

	if (!haspwads) return;

	sprintf(fname, "%sR3D_setparms", ddBinDir.path);

	if (!SR_LoadFile(fname, "R3D_setparms", NULL))
	{
		pr_OFinsert("ERROR: the file Risen3D\\bin\\R3D_setparms is missing\n");
		return;
	}

	sprintf(wname, "[%s]", analysename);

	while ( (nextstr = SR_NextSTR(false)) )
	{
		if (nextstr[0] != '[') continue;

check_wname:;
		if (stricmp(wname, nextstr)) continue;

		// found wad
		if ( !(nextstr = SR_NextSTR(false)) ) break;

		if (!stricmp("ALL", nextstr))	// special case for DEH
		{
			pr_OFinsert("Running R3D_setparm ALL %s\n", wname);

			if (!hasDEH)
			{
				SR_SCPWarn("Error: no DEHACKED loaded\nThe keyword ALL is dehacked specific");
					continue;
			}

			if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

			if (!stricmp("textbrighten", nextstr))
			{
				ge_inf.textbrighten = true;
			}
			else
			if (!stricmp("sprite", nextstr))
			{
				int		flag;

				if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

				if (   (val = Def_GetSpriteNum(nextstr)) < 0
					||  val >= DH_NUMSPRITES)
				{
					SR_SCPWarn("Error: Invalid sprite \"%s\"", nextstr);
					error = 1; break;
				}

				if (!(nextstr = SR_NextSTR(false)) ) break;

				if ( ((flag = SR_NumForString(&scperror)) && scperror)
					|| !flag)
				{
					SR_SCPWarn("Error: Invalid id \"%s\"", nextstr);
					error = 1; break;
				}

				specialSprite |= flag;
			}
			else
			{
				SR_SCPWarn("Error: sprite undefined", nextstr);
				error = 1; break;
			}
			continue;
		}
		
		if (stricmp(mapid, nextstr)) continue;

		pr_OFinsert("Running R3D_setparm %s %s\n", wname, mapid);
		// found mapid
check_next:;
		error = 0;

		if ( !(nextstr = SR_NextSTR(false)) ) break;
		
		if (nextstr[0] == '[') goto check_wname;

		if ( (val = SR_NumForString(&scperror)) && scperror)
		{
			SR_SCPWarn("Error: Invalid thing number \"%s\"", nextstr);
			break;
		}

		if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

		if ( (mnum = Def_GetMobjNum(nextstr)) < 0)
		{
			SR_SCPWarn("Error: Invalid mobjid \"%s\"", nextstr);
			error = 1; break;
		}

		if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

		line = radius = -1, zpos = DDMAXINT;

		if (!stricmp("radius", nextstr))
		{
			if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

			radius = SR_NumForString(&scperror);

			if (scperror || radius < 0 || radius > 100)
			{
				SR_SCPWarn("Error: Invalid radius \"%s\"", nextstr);
				error = 1; break;
			}
		}
		else if (!stricmp("zpos", nextstr))
		{
			if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

			zpos = SR_NumForString(&scperror);

			if (scperror || zpos < -32000 || zpos >= 32000)
			{
				SR_SCPWarn("Error: Invalid zpos \"%s\"", nextstr);
				error = 1; break;
			}
		}
		else if (!stricmp("line", nextstr))
		{
			if (!(nextstr = SR_NextSTR(false)) ) { error = 2; break; }

			line = SR_NumForString(&scperror);

			if (scperror || line < 0 || line >= numlines)
			{
				SR_SCPWarn("Error: Invalid line \"%s\"", nextstr);
				error = 1; break;
			}

			if (lines[line].frontsector != lines[line].backsector)
			{
				SR_SCPWarn("Error: Line \"%s\" must be same sector sided", nextstr);
				error = 1; break;
			}
		}
		else
		{
			SR_SCPWarn("Error: Unrecognised keyword \"%s\"", nextstr);
			error = 1; break;
		}

		for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
		{
			if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				mobj_t	*iter = (mobj_t*)thinker;

				if (MAPNUM(iter) == val && iter->type == mnum)
				{
					if (radius >= 0)
					{
						if (iter->flags & MF_SPECIAL)				// a pickup
							iter->pickupRadius = radius*FRACUNIT;
						else if (!(iter->flags & MF_SHOOTABLE) )	// a decoration
							iter->radius = radius*FRACUNIT;
						else
						{
							pr_OFinsert("T%i is neither non-shootable nor a pickup\n", MAPNUM(iter));
							break;
						}

						pr_OFinsert("R3D_setparms: %s %s T%i radius %i\n", analysename, mapid, val, radius);
					}
					else if (zpos < 32000)
					{
						iter->z = iter->floorz = iter->dropoffz = zpos*FRACUNIT;
						pr_OFinsert("R3D_setparms: %s %s T%i zpos %i\n", analysename, mapid, val, zpos);
					}
					else if (line >= 0)
					{
						cll.line = line, cll.x = iter->x, cll.y = iter->y, cll.radius = iter->radius;

						if (thingSpansLineCLL())
						{
							line_t	*ld = &lines[line];
							int		ht;

							if (iter->flags & MF_SPAWNCEILING)
							{
								ht = ld->frontsector->ceilingheight;
								if (ld->backsector) ht = MIN_OF(ht, ld->backsector->ceilingheight);
								iter->z = ht - iter->height;
							}
							else
							{
								ht = ld->frontsector->floorheight;
								if (ld->backsector) ht = MAX_OF(ht, ld->backsector->floorheight);
								iter->z = iter->floorz = iter->dropoffz = ht;
							}

							iter->scplightlevel = ld->frontsector->lightlevel;
#ifndef _DEBUG
						}
#else

							pr_OFinsert("R3D_setparms: %s %s T%i line %i\n", analysename, mapid, val, line);
						}
						else
						{
							pr_OFinsert("R3D_setparms: WARNING %s %s T%i does not span line %i\n", analysename, mapid, val, line);
						}
#endif
					}
					break;
				}
			}
		}

		goto check_next;
	}

	if (error == 1)
		pr_OFdirect("Errors found in R3D_Setparms - script abandoned\n");
	if (error == 2)
		SR_SCPWarn("Definition missing - script abandoned\n");

	SR_EndScript();
}

//===========================================================================
// Def_Read
//	Reads the specified definition files, and creates the sprite name, 
//	state, mobjinfo, sound, music, text and mapinfo databases accordingly.
// b211-0 Revised
//===========================================================================
void Def_Read(RDBOOL autoload)
{
	int		grenadebase = DF_DEFBASE;
	int		c, i, k;
	char	fname[256];
	char	*ptr;

	first_ded = true, read_count = 0;
	defs.model_path[0] = 0;	// Ensure model path is empty
	pr_OFdirect("Installing deds:\n");

	// b211-0 Load the base r3ddefs ptrs...
	for (i = 0; defsnames[i]; i++)
		Def_ReadProcessDED(defsnames[i], DF_DEFBASE);

	// ...and the required finales/mapnames
	if (gamemission == pack_plut)
	{
		ptr = "FinalesPLUT.ded";
		IWADMapName();
	}
	else if (gamemission == pack_tnt)
	{
		ptr = "FinalesTNT.ded";
		IWADMapName();
	}
	else
		ptr = "FinalesDOOM.ded";

	Def_ReadProcessDED(ptr, DF_DEFBASE);

	// Check that enough defs were found.
	if (   defs.count.states.num  < DH_NUMSTATES
		|| defs.count.mobjs.num   < DH_NUMOBJECTS
		|| defs.count.sprites.num < DH_NUMSPRITES)
	{
		Con_Error("DD_ReadDefs:\n Insufficent definitions where [num < : num];"
		"states [%i : %i]\nmobjs [%i : %i]\nsprites [%i : %i]",
		defs.count.states.num , DH_NUMSTATES,
		defs.count.mobjs.num  , DH_NUMOBJECTS,
		defs.count.sprites.num, DH_NUMSPRITES);
	}

	// R3D remove unwanted mapinfos
	// v220-33 do not delete, just zero the id string
	for (i = 0; i < defs.count.mapinfo.num; i++)
	{
		if (!M_ValidateID(defs.mapinfo[i].id)) *defs.mapinfo[i].id = 0;
	}

	if (!SetUpMIParms())	// Check for MAPINFO lump
		SetUpEtParms();		// Check for Eternity lump

	// v220-5 check there are mapinfos for all maps in the wad
	for (i = 0; i < 100; i++)
	{
		ded_mapinfo_t	*mi;
		ded_t			*ded = &defs;
		byte			*mapptr;
		char			id[8];
		int				k, idx;

		for (k = 0; k < 2; k++)
		{
			mapptr = k ? haslevelExMx : haslevelMapxx;

			if (mapptr[i])
			{
				if (k) sprintf(id, "E%iM%i", i/10, i%10); else sprintf(id, "MAP%02i", i);

				if (!Def_GetMapInfo(id))
				{
				//	Con_Message(">>>> CREATE MAPINFO for %s\n", id);
					// v220-5 None found so make a default one
					idx = DED_AddMapInfo(ded, id);
					mi = ded->mapinfo + idx;
					strcpy(mi->author, "unknown");
					strcpy(mi->sky_layers[0].flags, "slf_enable"); 
					strcpy(mi->sky_layers[0].texture, "SKY1"); 
				}
			}
		}
	}

	// Add unset mapinfo titlepatches
	for (i = 0; i < defs.count.mapinfo.num; i++)
	{
		ded_mapinfo_t	*mt = &defs.mapinfo[i];
		RDBOOL			mapinfset;

		if (!M_ValidateID(mt->id)) continue;

		if ( !(mapinfset = (mt->R3Dflags & R3Dhasmapinfo) != 0) )
			mt->mapinfomuslump = -1;

		if (commercial_game)
		{
			if (!mapinfset)
			{
				if ( !_stricmp(mt->id, "MAP07") )
					mt->R3Dflags |= map07special;
				if ( !_stricmp(mt->id, "MAP30") )
					mt->R3Dflags |= R3DendgameC;
			}
		}
		else
		{
			if ( mt->id[3] == '8')
				mt->R3Dflags |= R3Dendgame;

			if (!mapinfset)
			{
				if ( !_stricmp(mt->id, "E1M8") )
					mt->R3Dflags |= (baronspecial|specialaction_lowerfloor);
				else if ( !_stricmp(mt->id, "E2M8") )
					mt->R3Dflags |= (cyberdemonspecial|specialaction_exitlevel);
				else if ( !_stricmp(mt->id, "E3M8") )
					mt->R3Dflags |= (spidermastermindspecial|specialaction_exitlevel);
				else if ( !_stricmp(mt->id, "E4M6") )
					mt->R3Dflags |= (cyberdemonspecial|specialaction_blazedoor);
				else if ( !_stricmp(mt->id, "E4M8") )
					mt->R3Dflags |= (spidermastermindspecial|specialaction_lowerfloor);
			}
		}

		if (!mt->titlepatch)
			mt->titlepatch = TPnmapID(mt->id);

		if (mt->mapname[0] && mt->titlepatch > 0)
		{
			if (lumpinfo[mt->titlepatch].handle == IWADhandle)
			{
				mt->R3Dflags |= R3Dtpicfromiwad;
			}
			else
			{
				if (mt->R3Dflags & R3Dmapnameset)
					mt->R3Dflags &= ~R3Dmapnameset;
			}
		}
	}

	if (loadModels)	// Load the standard modeldefs
	{
		int flag = ge_inf.useSitters ? DF_WSMPATH : DF_DEFMODEL;	// v220-3

		for (i = 0; model_defs[i]; i++)
		{
			if (i != 8)	// Don't install the lod.ded yet
			{
				if (ge_inf.mdefflags & (1 << i))
				{
					// Skip Sitters weapons.ded
					if (i == 7 && flag == DF_WSMPATH && ge_inf.EXAPfound) continue;
					Def_ReadProcessDED(model_defs[i], flag);
				}
			}
		}
	}

	// End base deds

	if (detail_ded[0])
		Def_ReadProcessDED(detail_ded, DF_DEFPATH);

	// v220-24 check if custom deds have been called for:
	{
		char tmp[512];

		if (ArgExists("-cpz"))
		{
			sprintf(tmp, "%sDefs\\Custom\\Cheb's pistol packing zombies.ded", ddRuntimeDir.path);

			if ( (_access( tmp, 4 )) != -1) Def_ReadProcessDED(tmp, DF_FULLPATH);
		}

		if (ArgExists("-spz"))	// -spz also inhibits loading of the default sitters rifle bearing zombies
		{
			sprintf(tmp, "%sDefs\\Custom\\Sitters pistol packing zombies.ded", ddRuntimeDir.path);

			if ( (_access( tmp, 4 )) != -1) Def_ReadProcessDED(tmp, DF_FULLPATH);
		}

		if (ArgExists("-scg"))
		{
			sprintf(tmp, "%sDefs\\Custom\\Standard chaingun.ded", ddRuntimeDir.path);

			if ( (_access( tmp, 4 )) != -1) Def_ReadProcessDED(tmp, DF_FULLPATH);
		}

		if (ArgExists("-buded"))
		{
			sprintf(tmp, "%sDefs\\Custom\\Splash bubbles effect.ded", ddRuntimeDir.path);

			if ( (_access( tmp, 4 )) != -1) Def_ReadProcessDED(tmp, DF_FULLPATH);
		}
	}

	// Check for any .ded files specified on the command line.
	if (ArgCheckWith("-defs", 1))
	{
		while(1)
		{
			char *arg = ArgNext();
			int  argnum;

			if (!arg || arg[0] == '-') break;

			argnum = ArgGetLastMtch();

			// Add it to the list.
			if (strlen(arg) > 6 && arg[1] == ':' && !_stricmp(".ded", arg + strlen(arg) - 4))
				Def_ReadProcessDED(arg, DF_FULLPATH);
			else
				Con_Error("ERROR: %s\n-d or -defs parms must be .ded files with a complete path",arg);

			ArgSetLastMtch(argnum);
		}
	}

	if (autoload)
	{
		long	hFile;
		struct	_finddata_t d_file;

		if (Dir_ChDir(&r3dUserDir))
		{
			// Find first .ded file in current directory
			if( (hFile = _findfirst( "*.ded", &d_file )) != -1L )
			{
				do
				{
					if (Dir_ChExt(d_file.name, ".DED"))
						Def_ReadProcessDED(d_file.name, DF_UDIRPATH);
				}
				while(_findnext( hFile, &d_file ) == 0 );

				_findclose( hFile );
			}
			_chdir(ddRuntimeDir.path);
		}
	}

	if (ge_inf.switch_flags & SWF_FX)
	{
		if (loadModels)
		{
			sprintf(fname, "%sDefs\\Model\\ExtraFX.ded", ddRuntimeDir.path);

			if ( (_access(fname, 4)) != -1)
				Def_ReadProcessDED(fname, DF_THISPATH);
		}
		ge_inf.switch_flags &= ~SWF_FX;
	}

	if (loadModels)	
	{
		// Check for replacement model packs
		if (ge_inf.switch_flags & SWF_FP)
			Def_ReadProcessDED("fullpack.ded", DF_ABBAPATH);
		else if (ge_inf.switch_flags & SWF_AP)
		{
			if (ge_inf.holycrosswad)
			{
				Con_Error("PROBLEM: Alternative weapons cannot be\n"
					"used with %s.wad. Please deselect this\n"
					"option before restaring Risen3D", analysename); 
			}

			if (!ge_inf.EXAPfound)
				Def_ReadProcessDED("alternative.ded", DF_ABBAPATH);
		}

		if (ge_inf.grenade_id)				// Finally check grenade/pipebomb
		{
			if (ge_inf.switch_flags & (SWF_AP|SWF_FP))
			{
				grenadebase = DF_ABBAPATH;
				Def_ReadProcessDED("grenade2.ded", DF_ABBAPATH);
			}
			else
			{
				Def_ReadProcessDED("grenade.ded", DF_DEFMODEL);
			}
		}
	}

	if (ge_inf.switch_flags & SWF_GND)
		Def_ReadProcessDED("gndText.ded", grenadebase);
	else if (ge_inf.switch_flags & SWF_PBM)
		Def_ReadProcessDED("pbmText.ded", grenadebase);

	Dir_ChDir(&ddRuntimeDir);	// Back to the runtime directory
	Def_ReadLumpDefs();			// Check for DD_DEFNS lump(s)

	if (loadModels)				// Load the lod.ded last (if it exists)
	{
		int flag = ge_inf.useSitters ? DF_WSMPATH : DF_DEFMODEL;	// v220-3

		lodstatestart = defs.count.states.num;

		Def_ReadProcessDED(model_defs[8], flag);

		if (!(ge_inf.switch_flags & SWF_NC))
		{
			if (!ArgExists("-smn"))
				Def_ReadProcessDED("weapSmokeA.ded", DF_DEFMODEL);
			else	// enhanced weapn smoke
				Def_ReadProcessDED("weapSmoke.ded", DF_DEFMODEL);
		}

		if (ge_inf.mdefflags & (1 << 8))
		{
			if (ge_inf.switch_flags & SWF_FP)
				Def_ReadProcessDED("Lod-fullpack.ded", DF_ABBAPATH);
			else if (ge_inf.switch_flags & SWF_AP)
				Def_ReadProcessDED("Lod-alternative.ded", DF_ABBAPATH);
		}

		if ( ge_inf.switch_flags & (SWF_GND|SWF_PBM) )
			Def_ReadProcessDED("lod-grenade.ded", DF_DEFMODEL);

		Dir_ChDir(&ddRuntimeDir);	// Back to the runtime directory
	}

	H_InitLODScript();	// Can't be done prior to loading all Things deds

	// Sprite names.
	DED_NewEntries((void**)&sprnames, &count_sprnames, sizeof(*sprnames), defs.count.sprites.num);

	for (i = 0; i < count_sprnames.num; i++)
	{
		strncpy(sprnames[i].name, defs.sprites[i].id, 4);
		sprnames[i].name[4] = 0;
	}

	// v220-4 Texts are read before DEHACKED is run
	DED_NewEntries((void**)&texts, &count_texts, sizeof(*texts), defs.count.text.num);

	// Transfer the text strings to defs.text[].text
	// checking and modifying its format if necessary
	for(i = 0; i < count_texts.num; i++) Def_ModifyTextDef(i, defs.text[i].text);

	// At least NUMTEXT defs should have been loaded
	if (count_texts.num < NUMTEXT || strcmp(defs.text[TXT_VALIDITYCHECK].text, "valid"))
		Con_Error("%i %i Invalid or missing Risen3D\\Defs\\Base\\text.ded", count_texts.num, NUMTEXT);

	// Handle redefined strings. Modified v220-7
	// Note: new ids cannot be added. A replacement def with a Text definition
	//       must use an id that has already been defined in the base text.ded
	//       The string with id TXT_VALIDITYCHECK cannot be changed
	//       The base text.ded vals are stored to texts - these are never changed

	for (i = count_texts.num - 1; i > TXT_VALIDITYCHECK; --i)	// new entries
	{
		if (!defs.text[i].text) continue;		// should not happen

		for (k = TXT_VALIDITYCHECK - 1; k >= 0; --k)			// base entries
		{
			if (!defs.text[k].text) continue;	// should not happen

			if (!strcmp(defs.text[i].id, defs.text[k].id))
			{
				// Update the earlier string.
				defs.text[k].text = (char*)M_Realloc(defs.text[k].text, strlen(defs.text[i].text) + 1, 21);
				strcpy(defs.text[k].text, defs.text[i].text);

				// Free the later string, it isn't used (>NUMTEXT).
				M_Free(defs.text[i].text);
				defs.text[i].text = NULL;
			}
		}

		// Check the value has been transferred
		if (defs.text[i].text)
			Con_Message("TEXT has INVALID ID: %s \n", defs.text[i].id);
	}

	loadActionlinks();

	// Initialise state, mobj and stateowner arrays
	DED_NewEntries((void**)&states, &count_states, sizeof(*states), defs.count.states.num);
	DED_NewEntries((void**)&mobjinfo, &count_mobjinfo, sizeof(*mobjinfo), defs.count.mobjs.num);

	// Check that the correct base defs have been loaded
	if (count_states.num < S_DYNSPECIAL6 || count_mobjinfo.num < MT_STROBEX)
		Con_Error("Deds not present, incorrectly installed or are out of date %i %i",count_states.num,count_mobjinfo.num);

	P_SetStateTable(count_states.num);

	// Use Lee Killough's beta skulls
	if (ArgExists("-Bskull"))
	{
		mobjinfo[MT_SKULL].spawnstate   = S_BSKUL_STND;
		mobjinfo[MT_SKULL].seestate     = S_BSKUL_RUN1;
		mobjinfo[MT_SKULL].painstate    = S_BSKUL_PAIN1;
		mobjinfo[MT_SKULL].missilestate = S_BSKUL_ATK1;
		mobjinfo[MT_SKULL].deathstate   = S_BSKUL_DIE1;
		mobjinfo[MT_SKULL].damage       = 1;
	}

	// Check for and parse any DEHACKED lumps
	if (Defs_DEH(false))
	{
		// sprite name changes were detected
		ded_state_t *dst = defs.states;
		state_t		*st = states;

		sprlist = (int*)calloc(count_states.num, sizeof(int));

		// Remember original sprite pointers
		for (i = 0, dst; i < count_states.num; i++, dst++, st++)
		{
			sprlist[i] = Def_GetSpriteNum((char*)dst->sprite.id);;
		}
	}

	R3Dinfine = Defs_GetTextPtr(TXT_R3D_INFINE_TEXT);	// V220-13 init here

	if (loadModels && DEHchange & DEHF_STATE && ge_inf.switch_flags & (SWF_AP|SWF_FP))
	{
		Con_Error("PROBLEM: DEHACKED states detected\n\n"
		"Abbspack weapons cannot be used.\n"
		"Please deselect these before restarting");
	}

	if (DEHchange)
	{
		ded_state_t *dst = defs.states;
		state_t		*st = states;
		int			oldspr, spr;

		for (i = 0; i <  count_states.num; i++, dst++)
		{
			if ( (spr = Def_GetSpriteNum((char*)dst->sprite.id)) >= 0) goto sprcont;

			if (sprlist)
			{
				oldspr = spr = sprlist[i];

				// Check and change state id if valid
				if (i < DH_NUMSTATES || spr < 0)
				{
					ChangeDEHThingSPR(dst->sprite.id, i);

					spr = Def_GetSpriteNum((char*)dst->id);

					if (spr == oldspr) goto sprcont;	// no change
					if (spr >= 0 && spr < DH_NUMSPRITES) goto sprcont;

					if (checkNULLstate((char*)dst->id)) continue;

					if ((spr = sprlist[i]) >= 0) goto sprcont;
				}
				else
					goto sprcont;
			}
			else
			{
				if (checkNULLstate((char*)dst->id)) continue;
			}

			Con_Error("DEF Error: State \"%s\" has no sprite %s", dst->id, dst->sprite.id);
sprcont:;
			st = states + i;

			if (i < DH_NUMSTATES && DEHCheck[i] & DEHF_STATE)
			{
				// Ensure sprite has no model
				if (spr < DH_NUMSPRITES && !(DEHCheck[spr] & DEHF_SPRITE))
					DEHCheck[spr] |= DEHF_SPRITE;
			}

			st->sprite    = spr;
			st->stflags   = Def_EvalFlags(dst->flags) & ST_FLAGMASK;	// v220-11
			st->frame     = dst->frame;
			st->tics      = dst->tics;
			st->action_num= Def_GetActionNum(dst->action);
			st->nextstate = Def_GetStateNum(dst->nextstate);
			st->misct[0]  = dst->misct[0];
			st->misct[1]  = dst->misct[1];

			if (st->stflags & ST_DS1) stsndArray[ 9][0] = i;
			if (st->stflags & ST_DS2) stsndArray[10][0] = i;
			if (st->stflags & ST_DS3) stsndArray[11][0] = i;
			if (st->stflags & ST_DS4) stsndArray[12][0] = i;
		}

		if (sprlist) free(sprlist);

		// Update sprite list redirected names.
		for(i = 0; i < DH_NUMSPRITES; i++)
		{
			if (_stricmp(sprnames[i].name, defs.sprites[i].id))
			{
				if (dehdebug)
					pr_OFinsert("SPRITE LIST %s change to %s\n", sprnames[i].name, defs.sprites[i].id);
				strncpy(sprnames[i].name, defs.sprites[i].id, 4);
			}
		}
	}
	else
	{
		ded_state_t *dst = defs.states;
		state_t		*st = states;

		for (i = 0; i < count_states.num; i++, dst++, st++)
		{
			st->sprite    = Def_GetSpriteNum((char*)dst->sprite.id);
			st->stflags   = Def_EvalFlags(dst->flags) & ST_FLAGMASK;	// v220-11
			st->frame     = dst->frame;
			st->tics      = dst->tics;
			st->action_num= Def_GetActionNum(dst->action);
			st->nextstate = Def_GetStateNum(dst->nextstate);

			if (st->stflags & ST_DS1) stsndArray[ 9][0] = i;
			if (st->stflags & ST_DS2) stsndArray[10][0] = i;
			if (st->stflags & ST_DS3) stsndArray[11][0] = i;
			if (st->stflags & ST_DS4) stsndArray[12][0] = i;
		}

	/*	if (cfg.inFight)
			miscDEHval[Infight] = 0xDD;	// 0xDD == enable infighting */
	}

	ge_inf.states_count = count_states.num;

	if (ge_inf.hasscp_defn || ge_inf.switch_flags & (SWF_AP|SWF_FP))
	{
		state_t	*st = states + (int)S_BFG4;

		if (st->frame == 1) st->frame = 2;
	}

	if (count_states.num > S_DYNSPECIAL6)
	{
		int	offset = (int)S_DYNSPECIAL6 + 1;
		ded_state_t *dst = defs.states + offset;
		state_t		*st = states + offset;

		for (i = offset; i < count_states.num; i++, dst++, st++)
		{
			if (loadModels && i > S_DYNSPECIAL6)
			{	 
// pistol smoke
				if (!stricmp(dst->id, "pgsmoke" )) { pgsmokestatenum = i ; continue; }
				if (!stricmp(dst->id, "pgsmokeX")) { pgsmokestatenumX = i; continue; }
// chaingun smoke
				if (!stricmp(dst->id, "cgsmoke" )) { cgsmokestatenum = i ; continue; }
				if (!stricmp(dst->id, "cgsmokeX")) { cgsmokestatenumX = i; continue; }
// shotgun smoke
				if (!stricmp(dst->id, "sgsmoke" )) { sgsmokestatenum = i ; continue; }
				if (!stricmp(dst->id, "sgsmokeX")) { sgsmokestatenumX = i; continue; }
// db shotgun smoke
				if ( !(ge_inf.switch_flags & SWF_AP) )
				{
					if (!stricmp(dst->id, "dbsmoke" )) { dbsmokestatenum = i ; continue; }
					if (!stricmp(dst->id, "dbsmokeX")) { dbsmokestatenumX = i; continue; }
				}
// grenade launchersmoke
				if (!stricmp(dst->id, "gnsmoke" )) { gnsmokestatenum = i ; continue; }
				if (!stricmp(dst->id, "gnsmokeX")) { gnsmokestatenumX = i; continue; }

				if (ge_inf.switch_flags & SWF_AP)
				{
// SRM1216 smoke
					if (!stricmp(dst->id, "srsmoke" )) { srsmokestatenum = i ; continue; }
					if (!stricmp(dst->id, "srsmokeX")) { srsmokestatenumX = i; continue; }
// Ingram smoke
					if (!stricmp(dst->id, "igsmoke" )) { igsmokestatenum = i ; continue; }
					if (!stricmp(dst->id, "igsmokeX")) { igsmokestatenumX = i; continue; }
// winchester smoke
					if (!stricmp(dst->id, "wnsmoke" )) { wnsmokestatenum = i ; continue; }
					if (!stricmp(dst->id, "wnsmokeX")) { wnsmokestatenumX = i; continue; }
// ak47 smoke
					if (!stricmp(dst->id, "aksmoke" )) { aksmokestatenum = i ; continue; }
					if (!stricmp(dst->id, "aksmokeX")) { aksmokestatenumX = i; continue; }
				}
			}

			if (!_stricmp(dst->id, "CUSTOM_PUNCH"))
			{
				// COD support
				customPunchRestState = i;
			}
			else
			if (!_stricmp(dst->id, "PISTOL"))
			{
				if (st->stflags & ST_CHNGMASK)
				{
					pistoldamage = ((st->stflags & ST_GDMGMASK) >> ST_DMGSHIFT) * 20;
					if (st->stflags & ST_SND1) pistdname = "ST_SND1";
				}
			}
			else
			if (!_stricmp(dst->id, "PLASMA"))
			{
				if (st->stflags & ST_CHNGMASK)
				{
					plasmadamage = ((st->stflags & ST_GDMGMASK) >> ST_DMGSHIFT) * 20;
					if (st->stflags & ST_SND2) plasdname = "ST_SND2";
				}
			}
			else
			if (!_stricmp(dst->id, "BFG"))
			{
				if (st->stflags & ST_CHNGMASK)
				{
					bfgdamage = ((st->stflags & ST_GDMGMASK) >> ST_DMGSHIFT) * 20;
					if (st->stflags & ST_SND3) bfgdname = "ST_SND3";
				}
			}
			else
			if (!_stricmp(dst->id, "PUNCH"))
			{
				if (st->stflags & ST_SNDP1) punchname1 = "ST_SNDP1";
				if (st->stflags & ST_SNDP2) punchname2 = "ST_SNDP2";
			//	punchdamage = true;
			}
			else
			if (!_stricmp(dst->action, "A_WeapSound"))
			{
				int stnum = -1;

				if      (st->stflags & ST_SND4 ) stnum  = 0;
				else if (st->stflags & ST_SND5 ) stnum  = 1;
				else if (st->stflags & ST_SND6 ) stnum  = 2;
				else if (st->stflags & ST_SND7 ) stnum  = 3;
				else if (st->stflags & ST_SND8 ) stnum  = 4;
				else if (st->stflags & ST_SND9 ) stnum  = 5;
				else if (st->stflags & ST_SND10) stnum  = 6;
				else if (st->stflags & ST_SND11) stnum  = 7;
				else if (st->stflags & ST_SND12) stnum  = 8;

				if (stnum >= 0 && stnum < MAXSTSOUNDS)
				{
					stsndArray[stnum][0] = i;
				}
			}
		}
	}

	DED_NewEntries((void**) &stateowners, &count_stateowners, sizeof(mobjinfo_t*), defs.count.states.num);

	if (count_stateowners.num < S_DYNSPECIAL6)
		Con_Error("Deds not present, incorrectly installed or are out of date %i",count_stateowners.num);

	stats.num_mobjs = count_mobjinfo.num;

	moTypeArray = (byte*)W_Calloc(stats.num_mobjs, STATICID, 0);

	for (i = 0; i < count_mobjinfo.num; i++)
	{
		mobjinfo_t	*inf;
		ded_mobj_t	*dmo = defs.mobjs + i;
		int			mnum;

		if ( (mnum = Def_GetMobjNum(dmo->id)) < 0)
			Con_Error("Error: Invalid mobjid \"%s\"", dmo->id);

		if (hasDEHactors && mnum < DH_NUMOBJECTS && DEHCheck[mnum] & DEHF_THINGSET) continue;

		prc = 0;

		// Make sure duplicate defs overwrite the earliest.
		inf = mobjinfo + mnum;
		inf->doomednum	= dmo->doomednum;

		inf->spawnstate	= Def_StateForMobj(dmo->spawnstate	, inf);
		inf->seestate	= Def_StateForMobj(dmo->seestate	, inf);
		inf->meleestate	= Def_StateForMobj(dmo->meleestate	, inf);
		inf->missilestate = Def_StateForMobj(dmo->missilestate	, inf);
		inf->painstate	= Def_StateForMobj(dmo->painstate	, inf);
		inf->deathstate	= Def_StateForMobj(dmo->deathstate	, inf);
		inf->xdeathstate= Def_StateForMobj(dmo->xdeathstate	, inf);
		inf->raisestate	= Def_StateForMobj(dmo->raisestate	, inf);

		inf->spawnhealth= dmo->spawnhealth;
		inf->seesound	= Def_GetSoundNum(dmo->seesound);
		inf->reactiontime = dmo->reactiontime;
		inf->attacksound= Def_GetSoundNum(dmo->attacksound);
		inf->painchance	= dmo->painchance;
		inf->painsound	= Def_GetSoundNum(dmo->painsound);
		inf->deathsound	= Def_GetSoundNum(dmo->deathsound);
		inf->speed		= (int)(dmo->speed  * FRACUNIT);
		inf->radius     = (int)(dmo->radius * FRACUNIT);
		inf->height		= (int)(dmo->height * FRACUNIT);
		inf->mass		= dmo->mass;
		inf->damage		= dmo->damage;

		inf->weapon_offset = dmo->weapon_offset * FRACUNIT;	// v220-4
		inf->scplightlevel = dmo->scplightlevel;			// v220-13

		inf->activesound= Def_GetSoundNum(dmo->activesound);
		inf->flags2		= Def_EvalFlags(dmo->flags[1]) & MF2_LOADMASK;
		inf->flags3		= Def_EvalFlags(dmo->flags[2]) & MF3_LOADMASK;

		// Ensure a dehacked mobj does not get overwritten
		if (   hasDEHactors
			&& mnum < DH_NUMOBJECTS
			&& DEHCheck[mnum] & DEHF_THING)
		{
			DEHCheck[mnum] |= DEHF_THINGSET;

			if (loadModels == 3)
			{
				Con_Message("Models cannot be used with redirected dehacked\n"
							" things and have been partially disabled\n");

				loadModels = 1;
			}
		}

		if (   mnum < DH_NUMOBJECTS
			&& DEHCheck[mnum] & DEHF_THING
			&& flagsDEH[mnum] != 0xFFFFFFFF)
		{
			int dehf = flagsDEH[mnum] & (MF3_TOUCHY|MF3_BOUNCES|MF3_TRANSLUCENT|MF3_FRIEND);	// v220-35 tidy up
			inf->flags = (flagsDEH[mnum] & 0x0FFFFFFF);

			// v220-10 try to determine, from the myriad of possibilities, hacked
			//         parms that result in a Thing behaving as an ambient sound.
			if (inf->painsound)
			{
				if (   (inf->spawnstate == inf->deathstate && inf->flags & MF_NOCLIP)
					|| (inf->flags == MF_NOSECTOR && !inf->speed)
					|| (inf->flags == MF_NOSECTOR && !inf->attacksound && !inf->activesound && !inf->deathsound && !inf->deathstate))
				{
					inf->flags3 |= MF3_DEHAMBIENT;
					continue;
				}
			}

			if (dehf) inf->flags3 |= dehf;
		}
		else
		{
			inf->flags = Def_EvalFlags(dmo->flags[0]);
		}
	}

	Def_SetHashTable(stats.num_mobjs);

	// Dynamic lights. Update the sprite numbers.
	for(i = 0; i < defs.count.lights.num; i++)
	{
		k = Def_GetStateNum(defs.lights[i].state);
		if(k < 0)
		{
			Con_Error("ERROR: Defs(Lights): Undefined state '%s'", 
				defs.lights[i].state);
			continue;
		}
		states[k].light = defs.lights + i;
		defs.lights[i].flags = Def_EvalFlags(defs.lights[i].flags_string);
	}

	// b211-0 Revised ->
	if (ge_inf.grenade_id)
	{
		if (   flagsDEH[MT_ROCKET] != 0xFFFFFFFF
			|| DEHCheck[MT_ROCKET] & DEHF_SPRITE)
		{
			// do nothing if rocket or its sprite's been dehacked
		}
		else
		{
			strcpy(sprnames[SPR_BROK].name, "BGRE");
			strcpy(sprnames[SPR_LAUN].name, "GLNC");
			strcpy(sprnames[SPR_ROCK].name, "GHUD");

			if (ge_inf.grenade_id == MT_GRENADEX)
				states[mobjinfo[MT_GRENADEX].spawnstate].tics = ge_inf.grenade_tics;
		}
	}
	// b211-0 end <-

	DED_NewEntries((void**)&sounds, &count_sounds, sizeof(*sounds), defs.count.sounds.num);

	// Sound effects.
	// Modified to remove dependance on def_sound_t in sound code
	for (c = i = k = 0; i < count_sounds.num; i++)
	{
		sfxinfo_t	*si;
		ded_sound_t *snd = defs.sounds + i;
		int			sndnum = Def_GetSoundNum(snd->id);

		if (sndnum <= 0) continue;
		si = sounds + sndnum;

		strncpy(si->lumpname, snd->lumpname, 8);
		si->flags = Def_EvalFlags(snd->flags) & SFX_LOADFLAGS;
		si->lumpname[8]	= 0;
		si->sfxmaxchan = snd->dedmaxchan;

		// Ignore original ded link with chaingun to allow replacement sound.
		// This was the only linked sound and, as such, A_FireCGun must have
		// sfx_chgun changed to sfx_pistol if DSCHGUN does not exist - R3D

		if (!loadModels && !strcmp(si->lumpname, "DSGLAUNC"))
		{
			// v220-35 use P_WAD or IWAD default with sprites
			si->lumpnum = W_CheckPWADNumForName("DSGLAUNC");
			if (si->lumpnum < 0)
				si->lumpnum = W_CheckIWADNumForName("DSGLAUNC");
		}
		else if (!strcmp(si->lumpname, "DSCHGUN"))
		{
			si->lumpnum = W_CheckNumForName("DSCHGUN");

			if (si->lumpnum < 0)
				si->lumpnum = W_CheckNumForName("DSPISTOL");
		}
		else
		{
			int type;

			if (si->flags & SFX_R3DSOUND)
			{
				if (!strcmp(si->lumpname, "#STNMOV#"))
				{
					int n = W_CheckPWADNumForName("DSSTNMOV");

					// Use a pwad sound if replaced else the iwad sound (should be in all iwads)
					// for this reason a dummy "#STNMOV#" has been put in Risen3D.wad
					if (   n >= 0
						&& (type = DoomLumpIdentify(si->lumpnum, NULL, NULL, DLI_SOUND))
						&& (type == LMP_SND || type == LMP_WAV))
					{
						si->lumpnum = n;
					}
					else
					{
						si->lumpnum = W_CheckRisen3DWADNumForName(si->lumpname, true);
					}
				}
				else
				if (!strcmp(si->lumpname, "#JUMPDN#"))
				{
					// Use only the sound in Risen3D.wad
					si->lumpnum = W_CheckRisen3DWADNumForName(si->lumpname, true);
				}
				else
				{
					// Use only the sound in Risen3D.wad
					si->lumpnum = W_CheckRisen3DWADNumForName(si->lumpname, true);
				}

				c++;
			}
			else
			{
				si->lumpnum = W_CheckNumForName(si->lumpname);
			}

			type = DoomLumpIdentify(si->lumpnum, NULL, NULL, DLI_SOUND);

			if (si->lumpnum >= 0 && type != LMP_SND && type != LMP_WAV)
			{
				int iwadlump = W_CheckIWADNumForName(si->lumpname);

				if (iwadlump >= 0)
				{
					type = DoomLumpIdentify(iwadlump, NULL, NULL, DLI_SOUND);

					if (type != LMP_SND && type != LMP_RMI)
					{
						Con_Message("INVALID SOUND LUMP [%.8s] def [%s:%i] in PWAD has no data.\n",
						si->lumpname, snd->id, Def_GetSoundNum(snd->id));
						si->flags |= SFX_BADSOUND;
					}
					else
					{
						si->lumpnum = iwadlump;
						si->flags |= SFX_MENUONLY;
					}
				}
			}
		}

		if (si->lumpnum < 0) si->flags |= SFX_BADSOUND;

		// R3D - links and priority are not used
		//	si->link = NULL; si->link_pitch  = -1; si->link_volume = -1; si->priority = snd->priority;

		si->data = NULL;
		si->group = snd->group;
	}

	if (c != 18) Con_ErrorD(WADV);

	if (pistdname) pistolsound = R3D_AddNewSound(pistdname);
	if (plasdname) plasmasound = R3D_AddNewSound(plasdname);
	if (bfgdname) bfgsound = R3D_AddNewSound(bfgdname);
	if (punchname1)
	{
		punchsound1 = R3D_AddNewSound(punchname1);
		punchsound2 = R3D_AddNewSound(punchname2);
	}

	for (i = 0; i < MAXSTSOUNDS; i++)
	{
		if (stsndArray[i][0]) stsndArray[i][1] = R3D_AddNewSound(nsnd[i]);
	}

	// Music.
	for(i = 0; i < defs.count.music.num; i++)
	{
		ded_music_t	*mus = defs.music + i;
		// Make sure duplicate defs overwrite the earliest.
		ded_music_t	*earliest;
		int			ptr = Def_GetMusicNum(mus->id);
		
		if (ptr == i) continue;
		if (ptr >= defs.count.music.num) continue;

		earliest = defs.music + ptr;
		strcpy(earliest->lumpname, mus->lumpname);
		strcpy(earliest->path.path, mus->path.path);

	//	earliest->cdtrack = mus->cdtrack;				// b211-0 removed
	}

	// Set the default wad music lump
	for (i = 0; i < defs.count.mapinfo.num; i++)
		Def_SetMapMusicNum(&defs.mapinfo[i]);

	// Particle generators.
	for(i = 0; i < defs.count.ptcgens.num; i++)
	{
		int st = Def_GetStateNum(defs.ptcgens[i].state);

		if(defs.ptcgens[i].flat[0])
			defs.ptcgens[i].flat_num = W_CheckNumForName(defs.ptcgens[i].flat);
		else
			defs.ptcgens[i].flat_num = -1;

		defs.ptcgens[i].flags = Def_EvalFlags(defs.ptcgens[i].flags_string);
		defs.ptcgens[i].type_num = Def_GetMobjNum(defs.ptcgens[i].type);
		defs.ptcgens[i].type2_num = Def_GetMobjNum(defs.ptcgens[i].type2);
		defs.ptcgens[i].damage_num = Def_GetMobjNum(defs.ptcgens[i].damage);

		// If state triggered save def number to ptrigger
		if (st > 0) states[st].ptrigger = defs.ptcgens + i;
	}

	// v220-3 Check dehacked types that normally have blood that have had the MF_NOBLOOD flag set
	//        If so disable the relevant states' ptrigger
	if (hasDEH)
	{
		ded_mobj_t	*dmo = defs.mobjs;

		for (i = MT_POSSESSED; i < MT_BOSSBRAIN; i++, dmo++)
		{
			if (mobjinfo[i].flags & MF_NOBLOOD)
			{
				switch (i)
				{
				case MT_TROOP   : states[S_TROO_DIE1].ptrigger = states[S_TROO_DIE2].ptrigger = 0;
					break;
				case MT_SERGEANT: states[S_SARG_DIE1].ptrigger = states[S_SARG_DIE2].ptrigger = 0;
					break;
				case MT_HEAD    : states[S_HEAD_DIE3].ptrigger = states[S_HEAD_PAIN].ptrigger = 0;
					break;
				case MT_FATSO   : states[S_FATT_DIE5].ptrigger = 0;
					break;
				case MT_BABY    : states[S_SPID_DIE6].ptrigger = 0;
					break;
				case MT_KNIGHT:
				case MT_BRUISER : states[S_BOSS_DIE1].ptrigger = 0;
					break;
				case MT_SPIDER  : states[S_SPID_DIE6].ptrigger = states[S_CYBER_DIE3].ptrigger = 0;
					break;
				case MT_CYBORG  : states[S_CYBER_DIE2].ptrigger = states[S_CYBER_DIE3].ptrigger = 0;
					break;
				default:
					break;
				}
			}
		}
	}
	// Detail textures. Initialize later...

	// Model search path specified on the command line?
	if (ArgCheckWith("-modeldir", 1))
	{
		// Prepend to the search list; takes precedence.
		R_AddModelPath(ArgNext(), false);
	}

	numsprites = count_sprnames.num;

	// v220-13 pre-check for pwad sprites
	if (CFX && ArgExists("-mpd") && R_CheckForPWADSprites())
	{
		char buf[256];

		sprintf(buf,"The pwad %s has replacement sprites.\n\n"
					"The only way Risen3D can determine whether a model can\n"
					"be substituted is if all model groups are used.\n\n"
					"In the Launcher please ensure that no model group\n"
					"has been deselected.\n",
					analysename);

		Sys_MessageBox(buf, false);
		Sys_Quit(false);
	}

	subfn_Check_WADmodify();
}

void Def_PrintStats()
{
	pr_OFrepeat("\nDEF STATS ", "=", 32, 1);

	Def_CountMsg(count_sprnames.num    , "sprite names");
	Def_CountMsg(count_states.num      , "states");
	Def_CountMsg(count_mobjinfo.num    , "things");
	Def_CountMsg(ge_inf.models_count   , "models");
	Def_CountMsg(defs.count.lights.num , "lights");
	Def_CountMsg(count_sounds.num      , "sound effects");
	Def_CountMsg(defs.count.music.num  , "songs");
	Def_CountMsg(count_texts.num       , "text strings");
	Def_CountMsg(defs.count.ptcgens.num, "particle generators");
	Def_CountMsg(defs.count.details.num, "detail textures");
	// Other data:
	Def_CountMsg(defs.count.mapinfo.num, "map infos");
	Def_CountMsg(defs.count.finales.num, "finales");

	pr_OFrepeat(NULL, "=", 42, 2);

	defs_inited = true;
}

//===========================================================================
// Def_PostInit
//	Initialize detail textures - done after loading all textures.
//===========================================================================
void Def_PostInit(void)
{
	int i;

	DED_DelArray((void**)&dettex, &count_details);
	DED_NewEntries((void**)&dettex, &count_details, sizeof(*dettex), 
		defs.count.details.num);
	
	for(i = 0; i < defs.count.details.num; i++)
	{
		dettex[i].wall_texture = R_CheckTextureNumForName(defs.dettex[i].wall);
		dettex[i].flat_lump	= W_CheckFlatNumForName(defs.dettex[i].flat);
		dettex[i].detail_lump = W_CheckDetailNumForName(defs.dettex[i].det_lumpname);
	}
}

//==========================================================================
// Def_SameStateSequence
//	Can we reach 'snew' if we start searching from 'sold'?
//	Take a maximum of 16 steps.
//==========================================================================
int Def_SameStateSequence(state_t *snew, state_t *sold)
{
	int it, target = snew - states, start = sold - states;
	int count = 0;

	if(!snew || !sold) return false;
	if(snew == sold) return true; // Trivial.
	for(it = sold->nextstate; it >= 0 && it != start && count < 16; 
		it = states[it].nextstate, count++)
	{
		if(it == target) return true;
		if(it == states[it].nextstate) break;
	}
	return false;
}

//===========================================================================
// Def_Get
//	Returns true if the definition was found.
//===========================================================================
int Def_Get(int type, char *id, void *out)
{
	int				i;
	finalescript_t	*fin;

	switch(type)
	{
	case DD_DEF_MOBJ:
		return Def_GetMobjNum(id);

	case DD_DEF_STATE:
		return Def_GetStateNum(id);

	case DD_DEF_SOUND:
		return Def_GetSoundNum(id);

	case DD_DEF_MUSIC:
		return Def_GetMusicNum(id);

	case DD_DEF_TEXT:
		for(i = 0; i < defs.count.text.num; i++)
		{
			if(!_stricmp(defs.text[i].id, id))
			{
				if(out) *(char**)out = defs.text[i].text;
				return i;
			}
		}
		return -1;

	case DD_DEF_VALUE:
		// Read backwards to allow patching.
		for(i = defs.count.values.num - 1; i >= 0; i--)
		{
			if(!_stricmp(defs.values[i].id, id))
			{
				if(out) *(char**)out = defs.values[i].text;
				return true;
			}
		}
		return false;

	case DD_DEF_INFINE_BRIEF:
		fin = (finalescript_t*) out;
		i = defs.count.finales.num;
		while (--i >= 0)
		{
			if (!_stricmp(defs.finales[i].brief, id))
			{
				fin->debrief = defs.finales[i].debrief;
				fin->brief = defs.finales[i].brief;
				fin->script = defs.finales[i].script;
				return true;
			}
		}
		return false;

	case DD_DEF_INFINE_DEBRIEF:
		fin = (finalescript_t*) out;
		i = defs.count.finales.num;
		while (--i >= 0)
		{
			if (!_stricmp(defs.finales[i].debrief, id))
			{
				fin->debrief = defs.finales[i].debrief;
				fin->brief = defs.finales[i].brief;
				fin->script = defs.finales[i].script;
				return true;
			}
		}
		return false;

	default:
		return false;
	}
	return true;
}

// Routines to add a new sound - R3D
char *R3D_GetNameForSoundID(int sound_id)
{
	if (sound_id < 0 || sound_id >= count_sounds.num)
		Con_Error("Invalid sound num %i", sound_id);

	return sounds[sound_id].lumpname;
}

static int getSoundLump(char *lumpname)
{
	int i = numlumps;

	while (--i >= 0)
	{
		if (!_stricmp(lumpinfo[i].name, lumpname))
		{
			int type = DoomLumpIdentify(i, NULL, NULL, DLI_SOUND);

			if (type == LMP_WAV || type == LMP_SND)
				return i;
		}
	}

	return -1;
}

// New sounds: must not be added until after all the defs sounds
// have been loaded as this does not set a ded_sound_t entry.
int R3D_AddNewSound(char *lumpname)
{
	sfxinfo_t	*si = sounds;
	int			i, lumpnum = getSoundLump(lumpname);	// b220-1

	if (lumpnum < 0) return sfx_None;

	for (i = 0; i < count_sounds.num; i++)
	{
		if (!strnicmp(si[i].lumpname, lumpname, 8)) return i;
	}

	// used solely to assign memory which will be cleared when the deds are deleted
	si = (sfxinfo_t*)DED_NewEntry((void**)&sounds, &count_sounds, sizeof(*sounds));

	si->data		= NULL;
	si->lumpnum		= lumpnum;
	strncpy(si->lumpname, lumpname, 8);
	si->lumpname[8]	= 0;
	si->group		= 0;
	si->flags		= 0;
	si->sfxmaxchan	= -1;

/*	Not used - R3D
	si->usefulness	= 0;
	si->link		= NULL;
	si->link_pitch	= -1;
	si->link_volume	= -1;
	si->priority	= 64;
	si->external[0]	= 0; */

	// keep track of actual numbers of sounds.
	defs.count.sounds.num = count_sounds.num;
	return count_sounds.num - 1;
}

int R3D_ChangeSoundUsingAlias(int sound_id, char *lumpname)
{
	if (sound_id > 0 && sound_id < count_sounds.num)
	{
		sfxinfo_t *si = sounds + sound_id;

		if (strnicmp(si->lumpname, lumpname, 8))	// If strings are different
		{
			int lumpnum = getSoundLump(lumpname);	// b220-1

			if (lumpnum >= 0)
			{
				strcpy(si->lumpname, lumpname);
				si->lumpnum = lumpnum;
				return true;
			}
			return false;
		}

		return true;
	}

	return false;
}

// b220-0 Set up flag to indicate model needs cacheing
void Def_SetDMOnum(int num)
{
	ded_mobj_t	*dmo = defs.mobjs;

	for (int i = 0; i < count_mobjinfo.num; i++, dmo++)
	{
		int			mnum = Def_GetMobjNum(dmo->id);
		mobjinfo_t	*inf = mobjinfo + mnum;
		
		if (num == inf->doomednum)
		{
			dmo->cache = true;
			break;
		}
	}
}

// b220-0 Reload defs that may have been modified by a script
void Def_ResetMobjinfo()
{
	ded_mobj_t	*dmo = defs.mobjs;

	ge_inf.isfriendly = false;

	for (int i = 0; i < count_mobjinfo.num; i++, dmo++)
	{
		int			mnum = Def_GetMobjNum(dmo->id);
		mobjinfo_t	*inf = mobjinfo + mnum;
		ded_mobj_t	*dmo = defs.mobjs + i;
		
		// Currently [b220-0] only attacksound and painsound may have been changed
		inf->seesound	 = Def_GetSoundNum(dmo->seesound);
		inf->attacksound = Def_GetSoundNum(dmo->attacksound);
		inf->painsound	 = Def_GetSoundNum(dmo->painsound);
		inf->deathsound	 = Def_GetSoundNum(dmo->deathsound);
		inf->flags3	    &= ~(MF3_INTERREACT|MF3_FRIEND);
	}
}

// v220-31 reset Thing heights where older defs wiith incorrect scaling
//         may have been used with values stored in a savegame
void subfn_ResetHeight(mobj_t *iter)
{
	if (!R3D_ScriptedMap) return;
	if (iter->type > MIN_OF(count_mobjinfo.num, MT_MISC49) || iter->type < MT_VILE) return;

	if (iter->health > 0 && (iter->height>>16) != (iter->info->height>>16))
	{
		iter->height = iter->info->height;
	}
}

// v220-3 v220-4 Set the correct Thing heights values
//
// In Doom/Doom2 it must have been considered by iD during development that certain monsters
// looked too short. This meant the sprites were adjusted but the info height got overlooked
// probably  because none of the iwad maps, where these monsters were used, had celings low
// enough to cause a problem.
//
// This has resulted in many pwads with these monsters moving around with their heads in the
// ceiling. Unfortunately nothing can be done about offending pwads as their authors seemed to
// think this was perfectly OK. Thus to change these values with all pwads would affect the way
// they play. For R3D scripted maps, however, we can at least impose some common sense.

void Def_SetThingDimensions()
{
	int data[] =
	{
		MT_MISC49,		// torches
		MT_MISC46,
		MT_MISC45,
		MT_MISC44,
		MT_MISC43,
		MT_MISC42,
		MT_MISC41,
		MT_BARREL,		// 030
		MT_WOLFSS,		// 023
		MT_BABY,		// 020	- Arachnotron
		MT_KNIGHT,		// 017	- Hell Knight
		MT_BRUISER,		// 015	- Baron of Hell
		MT_HEAD,		// 014	- Cacodemon
		MT_SHADOWS,		// 013
		MT_SERGEANT,	// 012	- Demon
		MT_TROOP,		// 011	- Imp
		MT_FATSO,		// 008	- Mancubus
		MT_UNDEAD,		// 005	- Revenant
		MT_VILE,		// 003	- Arch Vile
		0
	};

	modeldef_t	*mf;
	mobjinfo_t	*inf;
	ded_mobj_t	*dmo = defs.mobjs;

	RDBOOL		prRes = ArgExists("-cdim");
	byte		*done;
	int			entries;
	int			i, k, mnum;
	int			mcount = hasDEH ? MT_PUSH : count_mobjinfo.num;
	int			limit = MIN_OF(mcount, MT_MISC49);

	//-------------------------------------------

	for (i = entries = 0; data[i]; i++)
		entries++;

	done = (byte*)M_Calloc(entries, 0);

	// Note: Def_GetMobjNum scans from last to first
	for (i = 0; i < mcount; i++)
	{
		dmo = defs.mobjs + i;

		mnum = Def_GetMobjNum(dmo->id);

		if (mnum > limit || mnum < MT_VILE) continue;

		for (k = 0; data[k]; k++)
		{
			if (mnum == data[k] && !done[k])
			{
				done[k] = true;

				inf = mobjinfo + mnum;
					
				if (R3D_ScriptedMap)	// Set to correct height/radius
				{
					mf = (modeldef_t*)states[inf->spawnstate].model;

					if (mf && mf->xyradius)
					{
						inf->height = mf->zmax - mf->zmin;
						inf->radius = mf->xyradius;

						if (prRes)
						{
							const char	spacestr[] = "          ";	// must be 10 spaces
							char		showstr[16];
							uint		len;

							if ( (len = strlen(dmo->id)) < 16)
							{
								strcpy(showstr, dmo->id);

								if (len < 10)
									strcat(showstr, &spacestr[len]);

								pr_OFinsert("[%s] r/h def %3i/%3i : model %3i/%3i doomednum %4i\n",
									showstr, (int)dmo->radius, (int)dmo->height,
									mf->xyradius>>16, (mf->zmax - mf->zmin)>>16, (int)dmo->doomednum);
							}
						}
					}
				}
				else					// Reset to height in the ded
				{
					inf->height = (int)dmo->height * FRACUNIT;
					inf->radius = (int)dmo->radius * FRACUNIT;
				}
				break;
			}
		}
	}

	M_Free(done);
}

// v220-13

int Def_SetScriptedNoInfight(char *id)
{
	ded_mobj_t	*dmo;
	int			i;

	for (i = 0, dmo = defs.mobjs; i < stats.num_mobjs; i++, dmo++)
	{
		if (!stricmp(dmo->id, id))
		{
			int mnum = Def_GetMobjNum(dmo->id);

			if (mnum > 0 && mnum < stats.num_mobjs)
			{
				moTypeArray[mnum] |= 1;
				return mnum;
			}
		}
	}

	return -1;
}

RDBOOL Def_GetScriptedNoInfight(int type)
{
	return type > 0
		&& type < stats.num_mobjs
		&& moTypeArray[type] & 1;
}
