//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen, Jack Vermeulen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dd_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Substantially modified/extended for Risen3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 14/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 32 (v220-32) 18/Mar/2017 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Main initialisation
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <direct.h>
#include <process.h>
#include <sys/timeb.h>
#include <time.h>
#include <errno.h>

#include "r3d_compat.h"

#include "doom/mbf_player.h"
#include "doom/g_game.h"
#include "doom/f_infine.h"
#include "doom/hu_stuff.h"
#include "doom/st_stuff.h"
#include "doom/m_menu.h"
#include "doom/p_setup.h"
#include "doom/p_saveg.h"
#include "doom/SndIdx.h"
#include "doom/d_englsh.h"
#include "sys_direc.h"
#include "con_action.h"
#include "con_bar.h"
#include "con_console.h"
#include "dd_wad.h"
#include "m_bams.h"
#include "m_misc.h"
#include "net_main.h"
#include "p_maputil.h"
#include "r_data.h"
#include "r_main.h"
#include "r_things.h"
#include "rend_model.h"
#include "rend_sky.h"
#include "rend_particle.h"
#include "s_main.h"
#include "sys_stwin.h"
#include "sys_console.h"
#include "sys_system.h"

#include "gl_main.h"
#include "gl_draw.h"
#include "gl_tex.h"
#include "gl_png.h"

#include "r3d_edit.h"
#include "r3d_mus2mid.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define	BGCOLOR			7
#define	FGCOLOR			8

// TYPES -------------------------------------------------------------------

const char *typedesc[] = {
	"UNDEF   ",
	"MARKER  ",
	"IGNORE  ",
	"TXT     ",
	"MAP     ",
	"PNG     ",
	"PCX     ",
	"JPG     ",
	"BMP     ",
	"FLAT    ",
	"FLAT320 ",
	"PATCH   ",
	"MIDI    ",
	"RMI     ",
	"MUS     ",
	"WAV     ",
	"SND     ",
	NULL
};

const char *grpdesc[] = {
	"LGT_NONE      ",
	"LGT_FLATS     ",
	"LGT_SPRITES   ",
	"LGT_PNAMES    ",
	"LGT_TX        ",
	"LGT_COLORMAPS ",
	"LGT_HI        ",
	"NUM_LGTAGS    ",	// Should never be printed out
	"LGT_DETAIL    ",
	"LGT_MARKER    ",
	NULL
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	EarlyQuit(void);
void	DD_GameLoop(void);
void	ErrorBoxAndQuit(char *err);
void	R_Init(void);
void	R_Init2(void);
void	D_DefaultBindings();
void	F_InitDFILES(void);
void	Def_PrintStats(void);
void	H_InitLODstates(void);		// v220-9
void	p2swd(char *msg, ...);		// v220-13
void	MC_SetMidiDevice(void);		// v220-13
void	R_InitIWADTextures(void);	// v220-16
void	MC_SetMidiDevs(void);		// v220-16
void	CloseSavefile(void);
// b211-0/v220-33
RDBOOL	SV_CheckLoadGame(RDBOOL checkswitch, RDBOOL versiononly, RDBOOL startup);
void	P_PresetIWADkeyLumps(void); // v220-35
void	LoadSmokeOffsets(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void PageDrawer(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	filloutlines;
extern RDBOOL	haslevels;
extern int		SittersLoaded;
extern char		OsVersionString[128];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// b211-0 r3dUDIRpath string replaced with r3dUserDir
directory_t		ddRuntimeDir, ddBinDir, r3dUserDir;
plhealth_t		plhealth;
maininfo_t		ma_inf;	// map specific
geninfo_t		ge_inf;	// global
skininfo_t		skin_inf;	// global

int		opennotepad = true;
byte	*palbuffer = NULL;
byte	*iwadpalbuffer = NULL;

// char	hiTexPath	[3][MAXPATHNAMESIZE];
// char	patchdirname  [10];
// char	patchpath     [MAXPATHNAMESIZE];

char	*analysename;
char	skyboxname[34];	// v220-3
char	dumppath[MAXPATH];
char	*headerDEHinsert;

RDBOOL	wasnomonsters;

byte	haslevelExMx[100];
byte	haslevelMapxx[100];
byte	hasIWADlevel[100];
byte	hasPWADlevel[100];
byte	palptr[256];

// Switches defined at load time
int		loadModels;
int		advancedemo;
int		autoload;
int		autostart;
int		nomonsters;		// checkparm of -nomonsters
int		paused;	
int		usergame;
int		verbose;
int		commercial_game;

// R3D switches
int		waterdynamics;
int		CFX;
int		BDMFX;

// Cvars
int		r_texglow;
int		r_bloodglow;
int		fastparm;		// checkparm of -fast
int		useDynLights;

char	*udirGame;
char	*r3ddump = "R3Ddump\\";

char	base_save_path[MAXPATHNAMESIZE];
char	detail_wad    [MAXFNAMEX];
char	detail_ded    [MAXPATHNAMESIZE];
char	*mainPwadname;
char	*mainIwadname;

GameMission_t	gamemission;
GameMode_t		gamemode  = undetermined;
gamestate_t		gamestate = GS_DEMOSCREEN; 
gameaction_t	gameaction; 

int		gameskill; 
int		startepisode;
int		startmap;

int		titlepic_num = -1;
int		interpic_num = -1;
int		pagepic_num = -1;

short	compat[NUM_COMPAT_OPTIONS];

int		haspwads;
int		pagetic;

// v220-13 set to 1 if music playing, 0 if music stopped, -1 if no music
int		music_status = -1;
// v220-13 set to 0 if WinMM midi selected, 1 if FMOD selected
int		midi_selected;

typedef struct {
	byte *hireslist;
	int num;
	int maxnum;	// b211-0
} hires_t;

hires_t wall_list;
hires_t flat_list;

int demo_lnum = -1;

char midiTempFname[256];


// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	mostnumlevels = 0, loadsaveset = false;

// Demo loop.
static int	demosequence;

static char	defdemoname [24];	// Only used for default lump names
static int	dumpflags;
static char	*dumpptr;


// CODE --------------------------------------------------------------------

// v220-7 added

long getFileTime(char *path)
{
	char	temp[MAXFNAMEX];
	char	szdrive[MAXDRIVE];
	char	szdir[MAXDIR];
	char	szfname[MAXFNAME];
	char	szext[MAXEXT];
	char	currentpath[MAXPATH];
	long	filetime = 0;

	_getcwd(currentpath, MAXPATH);

	_splitpath(path, szdrive, szdir, szfname, szext);

	sprintf(temp, "%s%s", szdrive, szdir);

	if ( !(_chdir(temp)) )
	{
		long	hFile;
		struct	_finddata_t g_file;

		sprintf(temp, "%s%s", szfname, szext);

		if ( (hFile = _findfirst(temp, &g_file )) != -1L )
		{
			filetime = (long)g_file.time_write;
			_findclose( hFile);
		}

		_chdir(currentpath);
	}

	return filetime;
}

// b211-0 Added
static int SetSwitchFlags()
{
	int switch_flags = 0;

// v220-35 revised

	if (ArgExists("-nc")  ) switch_flags |= SWF_NC;
	if (ArgExists("-nm")  ) switch_flags |= SWF_NM;
	if (ArgExists("-fp")  ) switch_flags |= SWF_FP;
	if (ArgExists("-ap")  ) switch_flags |= SWF_AP;
	if (ArgExists("-am")  ) switch_flags |= SWF_AM;
	if (ArgExists("-fx")  ) switch_flags |= SWF_FX;
	if (ArgExists("-xa")  ) switch_flags |= SWF_XA;
	if (ArgExists("-gnd") ) switch_flags |= SWF_GND;
	if (ArgExists("-pbm") ) switch_flags |= SWF_PBM;

	return switch_flags;
}

// =================================
// This cycles the start up and credit screens
// b211-0 demo support  removed (as of no use) and simplified
// =================================
void D_DoAdvanceDemo (void)
{
	int attempt = !compat[c_nomusic];
	int	id;

	id = commercial_game ? mus_dm2ttl : mus_intro;

    mbf_players[0].playerstate = PST_LIVE;  // not reborn
    advancedemo	= false;
    usergame	= false;               // no save / end game here
    paused		= false;
    gameaction	= ga_nothing;
	gamestate   = GS_DEMOSCREEN;

	if ( (demosequence = !demosequence) == 0)
		pagepic_num = titlepic_num;
	else
	{
		char *ptr = commercial_game || gamemode == retail ? "CREDIT" : "HELP2";
		pagepic_num = W_GetNumForName(ptr);
	}

	pagetic = TICRATE * 11;	//default

	global_looped = 0;

	if (attempt) attempt = S_StartMusicNum(id);

	if (attempt)
	{
		currentMapMusLumpnum = W_CheckNumForName(defs.music[id].lumpname);
		music_status = 1;
	}
	else
	{
		music_status = -1;
	}

	return;
}

void D_StartTitle (void)
{
    gameaction = ga_nothing;
    demosequence = -1;
    advancedemo = true;
}

RDBOOL LumpsFound(char **list)
{
	for(; *list; list++)
	{
		if(W_CheckIWADNumForName(*list) == -1) return false;	// was W_CheckNumForName v220-32		
	}

	return true;
}

static void Subfn_CheckGRENPBM()
{
	if (ge_inf.hasscp_defn && ge_inf.switch_flags & (SWF_GND|SWF_PBM))
	{
		Con_Error("SCRIPTED MAP: please DISABLE the %s launcher.\nThese are set automatically.",
			ge_inf.switch_flags & SWF_GND ? "grenade" : "pipebomb");
	}

	// b211-0 Set up pipebomb/grenade
	int type, val = W_CheckPwadSwitch(&type);

	if (type && type <= 3) pr_OFinsert("ADVISORY  : %s type set in pwad. Any switches ignored\n",
		type == 1 ? "ROCKET" : type == 2 ? "GRENADE" : "PIPEBOMB");

	if (CFX && !type && ge_inf.switch_flags & SWF_GND) type = 2;

	ge_inf.grenade_tics = 0;

	switch(type)
	{
	case 1:
		ge_inf.grenade_id = 0;
		ge_inf.switch_flags &= ~(SWF_GND|SWF_PBM);
		break;
	case 2:
		ge_inf.grenade_id = MT_GRENADEX;
		ge_inf.grenade_tics = val;

		if (!val)
		{
			if (ge_inf.switch_flags & SWF_GND)	// check cmd line settings
			{
				if (ArgCheckWith("-gnd", 1))
				{
					char *ptr = ArgNext();

					if (*ptr >= '0' && *ptr <= '9')
						ge_inf.grenade_tics = atoi(ptr) * 35;
				}
			}
		}
		pr_OFinsert("MT_GRENADEX %i\n", ge_inf.grenade_tics);
		ge_inf.switch_flags |=  SWF_GND;
		ge_inf.switch_flags &= ~SWF_PBM;
		break;
	case 3:
		ge_inf.grenade_id = MT_PIPEBOMB;
		ge_inf.switch_flags |=  SWF_PBM;
		ge_inf.switch_flags &= ~SWF_GND;
		break;
	default:
		if (ge_inf.switch_flags & SWF_GND)	// check cmd line settings
		{
			ge_inf.grenade_id = MT_GRENADEX;
			if (ArgCheckWith("-gnd", 1))
			{
				char *ptr = ArgNext();

				if (*ptr >= '0' && *ptr <= '9')
					ge_inf.grenade_tics = atoi(ptr);
			}
		}
		else if (ge_inf.switch_flags & SWF_PBM)
			ge_inf.grenade_id = MT_PIPEBOMB;
		break;
	}

	if (ge_inf.grenade_id == MT_GRENADEX)
	{
		if (ge_inf.grenade_tics <  60) ge_inf.grenade_tics =  60;
		if (ge_inf.grenade_tics > 255) ge_inf.grenade_tics = 255;
		pr_OFinsert("grenade_tics %i\n", ge_inf.grenade_tics);
	}

	if (!(ge_inf.switch_flags & SWF_NC) && ge_inf.switch_flags & (SWF_AP|SWF_FP))
	{
		char	temp[512];

		// e.g. scripted drocket.wad
		ge_inf.EXAPfound = ge_inf.hasscp_defn
			&& ge_inf.switch_flags & SWF_AP
			&& W_CheckNumForName("EXAP") >= 0;

		// b211-0 New path to search
		strcpy(temp, "md2\\abbspack\\abbspack.wad");

		if (!ge_inf.EXAPfound)
		{
			W_AddFile(ddRuntimeDir.path, temp, false);	// ABBS Alternative weapons
		}
	}

	pr_OFdirect("Files loaded:\n");
	ge_inf.MBFfriendallow = true;	// default

	// Note: analysename not set at this point
	for (int i = 0; i < numfilenameRecords; i++)
	{
		char	*fptr = filenameRecords[i].filename;
		// Use the earlier date as tiime_write will have been changed by Windows to the date when zipped/unzipped.
		const long	*tptr = (const long*)&MIN_OF(filenameRecords[i].time_write, filenameRecords[i].time_create);

		pr_OFinsert(" %s", fptr);
		pr_OFrepeat(NULL, " ", 64 - strlen(fptr), 0);
		pr_OFinsert("%s", ctime(tptr) );	// string has a line feed appended
	}

	pr_OFdirect("\n");
}

// =================================
// IdentifyVersion
// Checks availability of IWAD files by name,
// to determine whether registered/commercial features
// should be executed (notably when loading PWAD's).
// =================================
void subfn_WarnWadtype(char *type, char *found, int eps)
{
	char buff[256];

	sprintf(buff,
		"%s Doom requested but only %s doom found\n"
		"Episodes higher than %i will not be available.\n",
		type, found, eps);

	if (!Sys_CriticalMessage(buff, false))
		Sys_Quit(false);
}

int IdentifyVersion (int sw_auto, int *type)
{
	char *shareware_lumps[] = 
	{
		// List of lumps to detect shareware with.
		"e1m1", "e1m2", "e1m3", "e1m4", "e1m5", "e1m6", 
		"e1m7", "e1m8", "e1m9",
		"d_e1m1", "floor4_8", "floor7_2", NULL
	};
	char *registered_lumps[] = 
	{
		// List of lumps to detect registered with.
		"e2m1", "e2m2", "e2m3", "e2m4", "e2m5", "e2m6", 
		"e2m7", "e2m8", "e2m9",
		"e3m1", "e3m2", "e3m3", "e3m4", "e3m5", "e3m6", 
		"e3m7", "e3m8", "e3m9",
		"cybre1", "cybrd8", "floor7_2", NULL
	};
	char *retail_lumps[] =
	{
		// List of lumps to detect Ultimate Doom with.
		"e4m1", "e4m2", "e4m3", "e4m4", "e4m5", "e4m6", 
		"e4m7", "e4m8", "e4m9", 
		"m_epi4", NULL
	};
	char *commercial_lumps[] =
	{
		// List of lumps to detect Doom II with.
		"map01", "map02", "map03", "map04", "map10", "map20",
		"map25", "map30", "map31", "map32",
		"vilen1", "vileo1", "vileq1", "grnrock", NULL
	};
	char *plutonia_lumps[] =
	{
		"_deutex_", "mc5", "mc11", "mc16", "mc20", NULL
	};
	char *tnt_lumps[] =
	{
		"cavern5", "cavern7", "stonew1", NULL
	};

	if (LumpsFound(commercial_lumps))
	{
		gamemode = commercial;

		commercial_game = true;

		if (LumpsFound(plutonia_lumps))
		{
			gamemission = pack_plut;
			*type = 1;
			return 1;
		}

		if(LumpsFound(tnt_lumps))
		{
			gamemission = pack_tnt;
			*type = 2;
			return 1;
		}

		gamemission = doom2;
		*type = 3;
		return 1;
	}

	// Must be a doom wad of some description
	gamemission = doom;

	if (LumpsFound(registered_lumps))
	{
		char warn[] = {"WARNING   : "};
		char midm[] = {" doom loaded using "};

		if (LumpsFound(retail_lumps))
		{
			gamemode = retail;
			*type = 4;

			// Maybe too confusing to use case 5. Message printed if udoom
			// is loaded but the -doom switch is used. This is usually benign.
			if (sw_auto && sw_auto != 6 /* add */ && sw_auto != 5)
			{
				pr_OFinsert("%sUltimate%s%s switch\n",
				warn, midm, sw_auto == 5 ? "-doom" : "-doom1");
			}

			return 0;
		}
		if (sw_auto == 6)
			subfn_WarnWadtype("ULTIMATE", "registered", 3);
		if (sw_auto == 4)
			pr_OFinsert("%sRegistered%s-doom1 switch\n", warn, midm);

		gamemode = registered;
		*type = 5;
		return 0;
	}

	if (LumpsFound(shareware_lumps))
	{
		*type = 6;
		if (sw_auto > 4)
			subfn_WarnWadtype(
			sw_auto == 6 ? "ULTIMATE" : "REGISTERED", "shareware", 1);
		gamemode = shareware;
		return 0;
	}
	return -1;	// Nothing matches
}


//===========================================================================
// DetectIWADs
//	Looks for user defined iwad using the -iwad switch
//  Alternatively scans all compliant iwad locations if
//  a compliant iwad switch has been used
//===========================================================================

static RDBOOL subfn_CheckRedirected(char *str)
{
	return (strlen(str) > 1 &&
		str[0] == '\\' ||
		str[0] == '.'  ||
		str[1] == ':');
}

// Modified for R3D
static int DetectIWADs (char *iwadfile)
{
	typedef struct
	{
		char *file, *override;
	} fspec_t;
	
	fspec_t iwads[] = {
		{"TNT.wad"		, "-tnt"		},
		{"Plutonia.wad"	, "-plutonia"	},
		{"Doom2.wad"	, "-doom2"		},
		{"Doom1.wad"	, "-doom1"		},
		{"Doom.wad"		, "-doom"		},
		{"Doom.wad"		, "-ultimate"	},
		{NULL			, NULL			}
	};

	FILE	*check;
	char	fnf[512];
	int		i, iwadnum = -1, count = 0;

	fnf[0] = 0;

	// First check if an overriding command line option is being used.
	for (i = 0; iwads[i].file; i++)
	{
		if (ArgExists(iwads[i].override))
		{
			// Check for -tnt or -plutonia overriding -doom2
			if (i == 2 && (iwadnum >= 0 || iwadnum == 1))
				break;

			if (iwadnum >= 0)
			{
				Con_Error("Multiple switches of type %s used\n"
						  "Only one should be present", iwads[i].override);
			}

			iwadnum = i;
		}
	}

	// If no override found look for an -iwad definition
	if (iwadnum == -1)
	{
		if (ArgCheckWith("-iwad", 1))
			for(;;)
			{
				char *arg = ArgNext();

				if(!arg || arg[0] == '-') break;

				if (count)
					Con_ErrorD("-iwad takes only one argument");

				if (subfn_CheckRedirected(arg))
					strcpy(fnf, arg);
				else
					sprintf(fnf, "%s%s", ddRuntimeDir.path, arg);

				count++;
			}

		if (!count)
			Con_ErrorD("No IWAD specified with -iwad");
	}
	else
	{
		char	fn[512], parent_path[260], pathsbuff[1290];
		char	*libvar;
		RDBOOL	parent = false;
		uint	len;

		if (ArgExists("-iwad"))
			Con_Error("%s and -iwad cannot be used together",
						iwads[iwadnum].override);

		pathsbuff[0] = 0;
		strcpy(parent_path, ddRuntimeDir.path);

		if (!(libvar = getenv("$DOOMWADDIR")) ) libvar = getenv("DOOMWADDIR");

		if ((len = strlen(parent_path)) >= 4)
		{
			for (i = len - 2; i >= 2; --i)
			{
				if (parent_path[i] == '\\')
				{
					parent_path[i + 1] = 0;
					strcat(parent_path, "iwads\\");
					parent = true;
					break;
				}
			}
		}

		// Search all possible compliant directories.	
		for(i = 0; i < 6; i++)
		{
			strcpy(fn, ddRuntimeDir.path);

			// In order of preferential selection, 5) is highest
			switch(i)
			{
				//		..\iwads
			case 0: strcat(fn, "Data\\R3D\\")	; break;
			case 1: strcat(fn, "Data\\")		; break;
			case 2: /*  Games directory  */		  break;
			case 3: strcat(fn, "Iwads\\")		; break;
			case 4: fn[0] = 0; if (!parent)		  break;
				strcpy(fn, parent_path)			; break;
			case 5: fn[0] = 0;
					if (libvar) strcpy(fn, libvar);
				break;
			}

			if ( (len = strlen(fn)) > 0)
			{
				if (fn[len - 1] != '\\')
				{
					strcat(fn, "\\");
					len += 1;
				}
				strcat(fn, iwads[iwadnum].file);

				if ( (_access(fn, 4)) != -1)
				{
					strcpy(fnf, fn);
					fn[len] = '\n';
					fn[len + 1] = 0;
					strcat(pathsbuff, fn);
					count++;
				}
			}
		}

		if (count > 1)
		{
			pr_OFinsert(
				"\n%s found in multiple compliant locations;\n\n%s\n"
				"File selected is %s\n\n",
				 iwads[iwadnum].file, pathsbuff, fnf);
		}

		else if (!count)
		{
			Con_Error (
			"ERROR: The original id wad \"%s\" could not be found.\n"
			" Risen3D searched;\n\n"
			"1)  The game directory %s;\n"
			" the following locations in the game directory;\n"
			"2) \\data\\\n"
			"3) \\data\\R3D\\\n"
			"4) \\iwads\\\n"
			" and the special cases\n"
			"5) ..\\iwads\\ (one above the game directory)\n"
			"6) DOOMWADDIR or $DOOMWADDIR %s\n"
			"\nFor more information see the Risen3D\\Docs\\Risen3D_Readme.txt\n"
			"\nNOTE: This port requires that the iwad(s) from the original DOOM game are installed in one of the above.\n"
			"If you don't have this they are available for a nominal charge from;\n"DOOMWADSUPPLIER,
			iwads[iwadnum].file,ddRuntimeDir.path,
			libvar ? libvar : "(unset)");
		}
	}

	check = fopen(fnf, "rb");

	if (check != NULL)
	{
		char	header[5];
		int		hlen = fread(header, 1, 4, check);

		fclose(check);

		if (hlen == 4)
		{
			if (!strncmp(header, "IWAD", 4))
			{
				strcpy(iwadfile ,fnf);
				if (strstrCI(iwadfile, "doom.wad") || strstrCI(iwadfile, "doom1.wad"))
					ge_inf.doomwad = true;
				pr_OFinsert("Using IWAD: %s\n", iwadfile);
				return iwadnum + 1;
			}

			if (!strncmp(header, "PWAD", 4))
			{
				// v220-17 Extend message with surrogate wads
				if (strstr(fnf, "Doom2.wad") || strstr(fnf, "Doom.wad"))
				{
					Con_Error("PROBLEM: %s is configured as a PWAD\n"
						"This defeats all the logic used in Risen3D to manage pwads with iwads.\n\n"
						"To run Risen3D you will need to purchase the iwad(s) from either;\n\n"
						"www.idsoftware.com or www.steampowered.com\n\n"
						"If you do so then first delete this wad to prevent confusing it with\n"
						"any iwads purchased from the above. Do not associate Risen3D\n"
						"with this pwad in the iwad locations.\n\n"
						"If you do not own the original iwads and are unwilling to purchase them\n"
						"then you may find you can run it using another port such as ZDoom.\n", fnf);
				}
			}
		}
	}

	Con_Error("PROBLEM: %s is not a valid IWAD", fnf);
	return 0;	// happy compiler
}

static void InitCfg_and_PlrArray()
{
	int i;

	// Config defaults. The real settings are read from the
	// .cfg files but these will be used if one is not found.
	memset(&cfg, 0, sizeof(cfg));

	cfg.mouseSensiX = 8;
	cfg.mouseSensiY = 8;
	cfg.povLookAround = true;
	cfg.joyaxis[0] = JOYAXIS_TURN;
	cfg.joyaxis[1] = JOYAXIS_MOVE;
	cfg.joySensitivity = 5;
	cfg.joyDeadzone = 10;
	cfg.sbarscale = 1;	// fullscreen mode.
	cfg.echoMsg = true;
	cfg.lookSpeed = 3;
	cfg.menuScale = .9f;
	cfg.flashcolor[0] = .7f;
	cfg.flashcolor[1] = .9f;
	cfg.flashcolor[2] = 1;
	cfg.flashspeed = 4;
	cfg.turningSkull = true;
	cfg.hudShown[HUD_HEALTH] = true;
	cfg.hudShown[HUD_ARMOR] = true;
	cfg.hudShown[HUD_AMMO] = true;
	cfg.hudShown[HUD_KEYS] = true;
	cfg.hudShown[HUD_FRAGS] = true;
	cfg.hudScale = .6f;
	cfg.hudColor[0] = 1;
	cfg.hudColor[1] = cfg.hudColor[2] = 0;
	cfg.xhairSize = 1;
	for(i=0; i<3; i++) cfg.xhairColor[i] = 255;	
	cfg.jumpPower = 8.f;
	cfg.weaponAutoSwitch = true;
	cfg.secretMsg = true;
	cfg.levelTitle = true;
	cfg.menuColor[0] = 1;
	cfg.automapAlpha = .6f;
	cfg.linesz = 1.0f;
	cfg.msgCount = 4;
	cfg.msgScale = .8f;
	cfg.msgUptime = 5 * TICSPERSEC;
	cfg.msgBlink = true;
	cfg.bobWeapon = 1;
	cfg.bobView = 1;
	cfg.cameraNoClip = true;
	cfg.inFight = 1;
	cfg.ricochet = 4;
	cfg.underwaterDamage = false;
	cfg.footsteps = true;
	cfg.splashsprite = true;			// v220-35
	cfg.weaponsounds = true;			// v220-35
	cfg.weapsmoke = true;
	cfg.text_delay = 1.5f;
	cfg.usemlook = true;
	cfg.usemapmusic = true;
	cfg.flashdamage = 0.7f;
	cfg.flashpickup = 0.7f;
	cfg.use43 = false;
	cfg.doppler = 0.2f;
	cfg.pitch = 1;
	cfg.vid_cap = 10;
	cfg.useTorchLight = 0;
	cfg.sineCompress = 1;
	cfg.setVtxLight = 1;
	cfg.zbias = 0.001f;
	cfg.amDevboxWidth = 130;
	cfg.amEditboxWidth = 160;
	cfg.footstepvol = 1.f;
	cfg.splashvol = 1.f;

	if (ArgCheckWith("-ct", 1))
	{
		cfg.corpseTime = atoi(ArgNext());
		if (cfg.corpseTime < 0) cfg.corpseTime = 0;
		if (cfg.corpseTime > 9) cfg.corpseTime = 9;
	}
	else
		cfg.corpseTime = 0;

	for(i=0; i < MAXPLAYERS; i++)
	{
		mbf_players[i].plr = (ddplayer_t*) &players[i];
		mbf_players[i].plr->extradata = (void*) &mbf_players[i];
	}
}

static void warpError(int e, int m)
{
	if (gamemode == commercial)
		pr_OFinsert("WARP value %i out of range : defaulted to MAP01\n", m);
	else
		pr_OFinsert("WARP values %i %i out of range : defaulted to E1M1\n", e, m);
}

static void D_PostInit()
{
	char	name[8];
	int		mt = 1, et = 1;
	int		p;

    // get skill / episode / map from parms
	startepisode = startmap = 1;	// default
	autostart = false;

    fastparm = ArgExists ("-fast"); // w220-10 (console command game-fastmonsters removed)

    if((p = ArgCheck ("-skill")) && p < Argc()-1)
    {
		gameskill = Argv(p+1)[0] - '1';	// input values '1' to '5' so '1' is deducted
		if (gameskill < 0) gameskill = 0;
		if (gameskill > 4) gameskill = 4;
    }
	else
	{
		gameskill = sk_medium;	// default
	}

    if ( (p = ArgCheck("-warp")) && p < Argc()-1)
    {
		if (commercial_game)
		{
			mt = atoi(Argv(p+1));
		}
		else if (p + 1 < Argc()-1)
		{
			et = Argv(p+1)[0]-'0';
			mt = Argv(p+2)[0]-'0';
		}

		if (gamemode == commercial)
			sprintf(name, "map%.2i", mt);
		else
		{
			if (gamemode == shareware) et = 1;
			sprintf(name, "E%iM%i" , et, mt);
		}
		
		if (W_CheckNumForName(name) >= 0)
		{
			startepisode = et; startmap = mt;
		}
		else
		{
			warpError(et, mt);
		}

		autostart = true;
    }
	// episode ignored if commercial or warp used
	else if (   !commercial_game
		&& (p = ArgCheck ("-episode")) && p < Argc()-1)
    {
		et = Argv(p+1)[0] - '0';

		if (gamemode == shareware && et != 1)
			startepisode = 1;

		sprintf(name, "E%iM%i" , et, 1);
		
		if (W_CheckNumForName(name) < 0)
		{
			warpError(et, 1);
		}
		else
			startepisode = et;

		autostart = true;
    }

	if (autostart && ge_inf.devmap)
		brief_disabled = -1;
	
    // turbo option
	{
		extern int	forwardmove[2];
		extern int	sidemove[2];

		// turbo option
		if ( (p = ArgCheck("-turbo")) )
		{
			int			scale = 200;
		
			if (p < Argc()-1) scale = atoi(Argv(p+1));
			if (scale < 10  ) scale = 10;
			if (scale > 400 ) scale = 400;
			pr_OFinsert("turbo scale: %i%%\n", scale);
			forwardmove[0] = forwardmove[0]*scale/100;
			forwardmove[1] = forwardmove[1]*scale/100;
			sidemove[0] = sidemove[0]*scale/100;
			sidemove[1] = sidemove[1]*scale/100;
		}
	}
}

//==========================================================================
// HandlePWADS - R3D
//==========================================================================
static void subfn_AddPWAD(char* path, char* fname, RDBOOL load)
{
	char	completeFileName[MAXPATHNAMESIZE];
	char	szdrive[MAXDRIVE];
	char	szext[MAXEXT];
	uint	csm;
	int		type;
	
	cwinf.numlevels = 0;
	cwinf.pwadisdoom2 = false;	// Will be set to true if not doom

	// v220-5 This is the only pwad required to fix an iwad.
	if (   gamemission == pack_tnt
		&& !_stricmp(fname, "tnt31.wad"))
	{
		pr_OFdirect("The TNT map31 fix is not required. tnt31.wad has not been loaded\n");
		return; // map is fixed in P_SpawnMapThing
	}

	// v220-5 sort bug with completeFileName
	_splitpath(fname, szdrive, NULL, NULL, szext);

	// test added - maybe not needed
	if (_stricmp(szext, ".WAD")) return;	// We don't want non wads v220-6

	if (szdrive[0])
		sprintf(completeFileName, "%s", fname);
	else if (path[0])
		sprintf(completeFileName, "%s%s", path, fname);
	else
		sprintf(completeFileName, "%s%s", ddRuntimeDir.path, fname);

	// Check required lumps are present
	if (  !subfn_CheckWAD(completeFileName, &csm, &type)
		|| type != TYPE_PWAD)
	{
		Con_Error("PROBLEM: %s is not a valid PWAD", completeFileName);
	}
	
	if (cwinf.numlevels > mostnumlevels)
	{
		char szFname[MAXFNAME];
		char szExt[MAXEXT];

		_splitpath(fname, NULL, NULL, szFname, szExt);

		mainPwadname = (char*)W_Realloc(mainPwadname, strlen(szFname) + strlen(szExt) + 1, STATICID, 0);
		sprintf(mainPwadname, "%s%s", szFname, szExt);

		mostnumlevels = cwinf.numlevels;
	}

	if (cwinf.numlevels)
	{
		if (cwinf.iwadisdoom2 != cwinf.pwadisdoom2)
		{
			char buff[] = {"ERROR: Attempting to load a "};
			if (cwinf.iwadisdoom2)
				Con_Error("%s DOOM PWAD with a DOOM2 IWAD", buff);
			else
				Con_Error("%s DOOM2 PWAD with a DOOM IWAD", buff);
		}
		
		if (!analysename)
		{
			char	szFname[MAXFNAME];

			_splitpath(fname, NULL, NULL, szFname, NULL);
			analysename = (char*)W_Calloc(strlen(szFname) + 1, STATICID, 0);
			strcpy(analysename, szFname);
		}

		ge_inf.codlev = !stricmp(analysename, "CODLEV");
		ge_inf.holycrosswad = ge_inf.codlev || !stricmp(analysename, "CRUELD");
	}

	if (load)	// Deferred load for auto as order has to be determined
		W_AddFile("", completeFileName, cwinf.numlevels > 0);	// PWAD autoload
}

static RDBOOL subfn_CheckNotRedirected(char *fname)	// R3D
{
	unsigned i;
	if (subfn_CheckRedirected(fname)) return false;

	for (i = 0; i < strlen(fname); i++)
		if (fname[i] == '\\' ||
			fname[i] == '/')
			return false;
	return true;
}

// Read wads specified with the -file or -f switch
// v220-3 revised

typedef struct
{
	char	*levelname;
	long	time_write;
	int		levelcount;
} level_inf_t;

int __cdecl levelSort(const void *e1, const void *e2)
{
	int c1 = ((level_inf_t*) e1)->levelcount;
	int c2 = ((level_inf_t*) e2)->levelcount;
	int t1, t2;

	// Use number of levels. We assume that the pwad
	// with the lower number is the patchwad
	if (c1 != c2) return (c1 < c2) ? 1 : -1;

	// Failing this use the time written (could be tex patches)
	t1 = ((level_inf_t*) e1)->time_write;
	t2 = ((level_inf_t*) e2)->time_write;
	return (t1 > t2) ? 1 : (t1 < t2) ? -1 : 0;
}

static void HandlePWADS()	// R3D
{
	if (ArgExists("-file"))
	{
		int i = 0;

		while( i < Argc())
		{
			if (!_stricmp(Argv(i), "-file") || !_stricmp(Argv(i), "-f"))
			{
				while(++i < Argc() && !ArgIsOption(i))
				{
					if (r3dUserDir.path[0] && subfn_CheckNotRedirected(Argv(i)))	// b211-0
						subfn_AddPWAD(r3dUserDir.path, Argv(i), true);	// b211-0
					else
						subfn_AddPWAD("", Argv(i), true);

				}
			}
			else
				++i;
		}
	}
	else if (autoload)
	{
		level_inf_t *li, *lp;

		long	hFile;
		struct	_finddata_t g_file;
		int		c, i, levels;

		//-------------------------------

		if (!Dir_ChDir(&r3dUserDir))	// b211-0
		{
			Con_Error("Unable to locate the directory: %s", r3dUserDir.path);	// b211-0
        }

		li = lp = (level_inf_t*)W_Calloc(MAX_PWADS*sizeof(level_inf_t), TEMPID, 0);

		// Find first .wad file in current directory
		// v220-3 Revised

		levels = 0;

		if( (hFile = _findfirst( "*.wad", &g_file )) != -1L )
		{			
			do
			{
				if (Dir_ChExt(g_file.name, ".WAD"))
				{
					if ((lp - li) >= MAX_PWADS)
						Con_Error("Too many pwads for auto load. Max is %i", MAX_PWADS);				

					lp->levelname = (char*)W_Calloc(strlen(g_file.name) + 1, TEMPID, 0);
					strcpy(lp->levelname, g_file.name);

					// don't load - just set cwinf
					subfn_AddPWAD(r3dUserDir.path, g_file.name, false);

					lp->levelcount = cwinf.numlevels;
					lp->time_write = (long)g_file.time_write;
					levels += cwinf.numlevels;
					lp++;
				}
			}
			while( _findnext( hFile, &g_file ) == 0 );	// Find the rest of the .g files

			_findclose( hFile );
		}

		if (!levels)
			Con_Error("No pwads with levels for auto load found in dir:\n%s\n"
					  "Ensure also that the correct IWAD is selected", r3dUserDir.path);

		if ( (c = (lp - li)) > 1)
			qsort(li, c, sizeof(level_inf_t), levelSort);

		for (i = 0, lp = li; i < c; i++, lp++)
		{
			W_AddFile (r3dUserDir.path, lp->levelname, true);	// PWAD -file load
		}

		_chdir(ddRuntimeDir.path);

		W_Free(li, TEMPID, 0);
	}
}

// The -udir option sets a user game directory - R3D
static void R3D_UdirInit()
{
	char	temp[MAXFNAMEX];
	char	eror[] = {"ERROR UDIR: "};

	int s = ArgCheckWith("-savedir", 1);
	int u = ArgCheckWith("-udir", 1);

	if (s && u) Con_Error("%s-savedir invalid with -udir", eror);

	// Set initial savegame path
	if (s && s < Argc() - 1 )
	{
		M_PrependBasePath(Argv(s + 1), base_save_path);

		if ( (_access(base_save_path, 4)) == -1)
		{
			Con_Error("User savegame path \"%s\" does not exist" , base_save_path);
		}
	}

	temp[0] = 0, autoload = false;
	memset(&r3dUserDir, 0, sizeof(directory_t));	// b211-0

	if (u)
	{
		int i = 0, type = 0;
		
		while( i < Argc())
		{
			if(!_stricmp(Argv(i++), "-udir"))
			{
				while(i < Argc() && !ArgIsOption(i))
				{
					if (temp[0])
					{
						Con_Error("%sonly takes one argument\n"
							"If using a directory name with spaces then\n"
							"enclose it in double quotes.\n"
							"E.g.;  -udir \"my wads\"", eror);
					}
					strcpy(temp, Argv(i++));
				}
			}
		}

		if (!temp[0])
			Con_Error("%sunspecified", eror);

		if (strrchr(temp, '\\') || strrchr(temp, '/'))
			Con_Error("%sdirectory name only must be given", eror);

		udirGame = (char*)W_Calloc(strlen(temp) + 1, STATICID, 101);
		strcpy(udirGame, temp);

		// b211-0 ->
		memcpy(&r3dUserDir, &ddRuntimeDir, sizeof(directory_t));
		strcat(r3dUserDir.path, "R3DGames\\");
		strcat(r3dUserDir.path, temp);
		strcat(r3dUserDir.path, "\\");

		if(!Dir_ChDir(&r3dUserDir))
			Con_Error("%sudir %s not found", eror, r3dUserDir.path);

		sprintf(temp, "%s%s", r3dUserDir.path, "pwads\\");
		if ( (_access(temp, 0)) != -1) type |= 1;
		sprintf(temp, "%s%s", r3dUserDir.path, "gwads\\");
		if ( (_access(temp, 0)) != -1) type |= 2;
		// b211-0 <-

		switch (type)
		{
		case 0:
			Con_Error("%sudir must have a pwads or gwads directory", eror);
		case 1:
			strcat(r3dUserDir.path, "pwads\\");	// b211-0
			if (!ArgExists("-file") &&
				!ArgExists("-f") &&
				!ArgExists("-iwad"))
				autoload = true;
			break;
		case 2:
			strcat(r3dUserDir.path, "gwads\\");	// b211-0
			break;
		default:
			Con_Error("%sudir %s has a gwads and pwads directory\n"
					 "It can only have one or the other",
					  eror, r3dUserDir.path);	// b211-0
		}

		p2swd("UDIR      : \"%s\"\n", r3dUserDir.path);	// b211-0

		strcpy(base_save_path, r3dUserDir.path);	// b211-0
		// Get rid of \pwads or \gwads
		base_save_path[strlen(base_save_path) - 6] = 0;
	}
	else
	{
		p2swd("UDIR      : none set\n");
	}
}

static void Set_CompatOptions()	// R3D
{
	// Compatibility options
	if (R3D_edit)
	{
		compat[c_nuke24] = 0;
	}
	else
	{
		compat[c_nuke24] = (byte)(ArgExists("-c_nuke24_off"	) == 0);
	}

	compat[c_stairs     ] = (byte)(ArgExists("-c_stairs_off"	) == 0);
	compat[c_torches    ] = (byte)(ArgExists("-c_torches_off"	) == 0);
	compat[c_camera     ] = (byte)(ArgExists("-c_camera_off"	) == 0);
	compat[c_carry      ] = (byte)(ArgExists("-c_carryfix_off"	) == 0);	// b220-1
	compat[c_DEHamb     ] = (byte)(ArgExists("-c_DEHamb_off"    ) == 0);
	compat[c_model      ] = (byte)(ArgExists("-c_model_on"		) != 0);
	compat[c_dtex       ] = (byte)(ArgExists("-c_dtex_on"		) != 0);
	compat[c_skulllimit ] = (byte)(ArgExists("-c_skulllim_on"   ) != 0);	// v220-3
	compat[c_moddebug   ] = (byte)(ArgExists("-c_model_debug"	) != 0);
	compat[c_nocodep    ] = (byte)(ArgExists("-c_nocodep"	    ) != 0);	// v220-13
	compat[c_nextmap    ] = (byte)(ArgExists("-c_nextmap"	    ) != 0);	// v220-4
	compat[c_nosound    ] = (byte)(ArgExists("-nosound") || R3D_edit);
	compat[c_nomusic    ] = compat[c_nosound] || (byte)(ArgExists("-nomusic"));
}

// Modified for R3D
static void DumpLump(char *dumppath, char *arg)
{
	char	fname[10];

	FILE	*file;

	uint	psize;
	int		lump;
	int		type;
	int		i;

	// ----------------------------------

	for (lump = -1, i = numlumps - 1; i >= 0 ; --i)
	{
		if (!stricmp(lumpinfo[i].name, arg))
		{
			lump = i; break;
		}
	}

	if (lump < 0)
		Con_Error("Lump %8.8s not found", arg);

	type = DoomLumpIdentify(lump, NULL, NULL, DLI_ANY);

	pr_OFinsert("\n%8.8s %s %s\n",
		arg,
		typedesc[lumpinfo[lump].lumptype],
		grpdesc[lumpinfo[lump].groupMarker]);

	if ( !(psize = lumpinfo[lump].size) )
		Con_Error("Lump %8.8s has no data", arg);

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);


	sprintf(fname, "%.8s.", arg);

	if      (type == LMP_TXT ) strcat(fname, "txt");
	else if (type == LMP_PNG ) strcat(fname, "png");
	else if (type == LMP_JPG ) strcat(fname, "jpg");
	else if (type == LMP_BMP ) strcat(fname, "bmp");
	else if (type == LMP_PCX ) strcat(fname, "pcx");
	else if (type == LMP_MIDI) strcat(fname, "mid");
	else strcat(fname, "lmp");

	file = fopen(fname, "wb");

	if (file) 
	{
		char *ptr = (char*)M_CacheLumpNum(lump, &psize, false);
		fwrite(ptr, 1, psize, file);
		fclose(file);
	}
	else
	{
		Con_Error("Couldn't open %s for writing. %s", fname, strerror(errno));
	}

	Con_Error(">>> dumped to Risen3D\\R3Ddump\\%s", fname);
}

static void PrintLumpIfNameContains(int id, char *arg)
{
	int		i, n, w, h, type;
	char	*tnm[7] = { "PNG", "PCX", "JPG", "BMP", "FLAT", "FLAT320", "PATCH" };
	char	wbuff[12];

	// ----------------------------------

	pr_OFdirect("\n\n");

	for (i = n = 0; i < numlumps; i++)
	{
		if (strstrCI(lumpinfo[i].name, arg))
		{
			if ( (id == 0 && lumpinfo[i].handle == IWADhandle) || (id == 1 && pwadHandle(lumpinfo[i].handle)) )
			{
				type = DoomLumpIdentify(i, &w, &h, DLI_GRAPHIC);

				if (type >= LMP_PNG && type <= LMP_PATCH)
				{
					uint len = strlen(lumpinfo[i].name);	// 8 or less

					memset(&wbuff, ' ', 12);

					wbuff[12 - len] = 0;

					n++;
					pr_OFinsert("%3i) \"%s\"%s width %4i    height %4i    type %s\n", n, lumpinfo[i].name, wbuff, w, h, tnm[type - LMP_PNG]);
				}
			}
		}
	}

	Con_Error("%i names containing %s listed to Risen3D.out", n, arg);
}

// Modified for R3D
static void DumpPngs(char* dumppath)	// R3D
{
	char	fname[10];
	FILE	*file;
	int		i, count = 0;
	uint	psize;

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	for (i = 0; i < numlumps; i++)
	{
		if (   pwadHandle(lumpinfo[i].handle)
			&& DoomLumpIdentify(i, NULL, NULL, DLI_GRAPHIC) == LMP_PNG)
		{
			void *ptr = M_lumpCACHE(i, &psize, MLC_NOLOCK, MLC_NOERR);

			sprintf(fname, "%.8s.png", lumpinfo[i].name);

			if((file = fopen(fname, "wb")) != NULL) 
			{
				fwrite(ptr, 1, psize, file);
				pr_OFinsert("dumped %s\n", fname);
				fclose(file);
				count++;
			}
			else
			{
				pr_OFinsert("%s\n",fname);
			}
		}
	}

	Con_Error(">>> %i png's dumped to dir %s", count, r3ddump);
}

static void DumpJpgs(char* dumppath)	// R3D
{
	char	fname[10];
	FILE	*file;
	int		i, count = 0;
	uint	psize;

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	for (i = 0; i < numlumps; i++)
	{
		if (   pwadHandle(lumpinfo[i].handle)
			&& DoomLumpIdentify(i, NULL, NULL, DLI_GRAPHIC) == LMP_JPG)
		{
			void *ptr = M_lumpCACHE(i, &psize, MLC_NOLOCK, MLC_NOERR);

			sprintf(fname, "%.8s.jpg", lumpinfo[i].name);

			if((file = fopen(fname, "wb")) != NULL) 
			{
				fwrite(ptr, 1, psize, file);
				pr_OFinsert("dumped %s\n", fname);
				fclose(file);
				count++;
			}
			else
			{
				pr_OFinsert("%s\n",fname);
			}
		}
	}

	Con_Error(">>> %i jpg's dumped to dir %s", count, r3ddump);
}

static void DumpWaddir(char *dumppath)	// R3D
{
	FILE	*file;
	int		i;

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	if ((file = fopen("dumpwaddir.txt", "wt")) == NULL)
	{
		Con_ErrorD("Cannot open dumpwaddir.txt");
	}

	fprintf(file, "Lumps (%d total):\n\n", numlumps);

	for (i = 0; i < numlumps; i++)
	{
		fprintf(file, "%04d - %-8s (hndl: %d  size: %d)\n",
			i, lumpinfo[i].name, lumpinfo[i].handle, lumpinfo[i].size);
	}

	fclose(file);

	Con_Error(">>> dumpwad.txt stored to dir %s", r3ddump);
}

static void DumpLumpInf(char *dumppath, int ispwad)	// R3D
{
	FILE	*file;
	int		i, n;
	char	filetype[16];

	if (ispwad && !haspwads)
		Con_Error("No pwads are loaded");

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	if (ispwad)
		strcpy(filetype, "dumpPWADinf.txt");
	else
		strcpy(filetype, "dumpIWADinf.txt");

	if ((file = fopen(filetype, "wt")) == NULL)
	{
		Con_ErrorD("Cannot open dumpPWADinf.txt");
	}

	if (ispwad)
	{
		for (i = n = 0; i < numlumps; i++)
		{
			if (pwadHandle(lumpinfo[i].handle)) n++;
		}

		fprintf(file, "PWAD Lumps (%i of total %i):\n\n", n, numlumps);
	}
	else
	{
		for (i = n = 0; i < numlumps; i++)
		{
			if (lumpinfo[i].handle == IWADhandle) n++;
		}

		fprintf(file, "IWAD Lumps (%i of total number of lumps %i):\n\n", n, numlumps);
	}

	fprintf(file, "        Lumpname  lumptype     lumpgroup:\n"
				  "-----------------------------------------\n");

	for (i = n = 0; i < numlumps; i++)
	{
		RDBOOL show = false;

		if (ispwad)
			show = pwadHandle(lumpinfo[i].handle);
		else
			show = (lumpinfo[i].handle == IWADhandle);

		if (show)
		{
			DoomLumpIdentify(i, NULL, NULL, DLI_ANY);
			fprintf(file, "%5i - %-8s  %s %s\n",
				++n,
				lumpinfo[i].name,
				typedesc[lumpinfo[i].lumptype],
				grpdesc[lumpinfo[i].groupMarker]);
		}
	}

	fclose(file);

	Con_Error(">>> %s stored to dir %s", filetype, r3ddump);
}

static void DumpMidis(char *dumppath)	// R3D
{
	long	hFile;
	struct	_finddata_t g_file;
	char	fname[16];
	int		c = 0, i = numlumps, type;
	FILE	*f;

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	// b211-0 Revised
	if ((hFile = _findfirst( "*.mid", &g_file )) != -1L )
	{
		do
		{
			if (Dir_ChExt(g_file.name, ".MID"))
				remove(g_file.name);
		}
		while( _findnext( hFile, &g_file ) == 0 );

		_findclose( hFile );
	}

	// Scan backwards
	while (--i)
	{
		if (lumpinfo[i].handle == RISEN3DWADhandle) continue;

		sprintf(fname, "%.8s.mid", lumpinfo[i].name);
		
		if ( (_access(fname, 4)) == -1)
		{
			if ( (type = Mus_IsMUSorMIDILump(i)) == M2M_MUS)
			{
				ConvertMusToMidi(i, fname);
			}
			else if (type == M2M_MIDI && (f = fopen(fname, "wb")) != 0)
			{
				size_t psize;
				void *ptr = M_CacheLumpNum(i, &psize, false);
				fwrite(ptr, psize, 1, f);
				fclose(f);
				M_Free(ptr);
				pr_OFinsert("Dumped MIDI to %s\n", fname);
				c++;
			}
		}
	}

	Con_Error(">>> %i midis dumped to dir %s", c, r3ddump);
}

static void DumpWav(char *dumppath)	// R3D
{
	long	hFile;
	struct	_finddata_t g_file;
	char	fname[16];
	int		c = 0, i = numlumps;
	FILE	*f;

	if (!haspwads)
		Con_Error("Wav files are only dumped from pwads and no pwad is loaded");

	if (_chdir(dumppath))
		Con_Error("Can't access %s", dumppath);

	// b211-0 Revised
	if ((hFile = _findfirst( "*.wav", &g_file )) != -1L )
	{
		do
		{
			if (Dir_ChExt(g_file.name, ".WAV"))
				remove(g_file.name);
		}
		while( _findnext( hFile, &g_file ) == 0 );

		_findclose( hFile );
	}

	// Scan backwards
	while (--i)
	{
		if (   pwadHandle(lumpinfo[i].handle)
			&& DoomLumpIdentify(i, NULL, NULL, DLI_SOUND) == LMP_WAV)
		{
			sprintf(fname, "%.8s.wav", lumpinfo[i].name);
			
			if (   (_access(fname, 4)) == -1
				&& (f = fopen(fname, "wb")) != 0)
			{
				size_t psize;
				void *ptr = M_CacheLumpNum(i, &psize, false);
				fwrite(ptr, psize, 1, f);
				fclose(f);
				M_Free(ptr);
				pr_OFinsert("Dumped WAV to %s\n", fname);
				c++;
			}
		}
	}

	Con_Error(">>> %i wav files dumped to dir %s", c, r3ddump);
}

static void LoadIWADname(int type)
{
	char *str = NULL;

	switch(type)
	{
	case 1: str = "plutonia"; ge_inf.iwadid = 1; break;
	case 2: str = "tnt"     ; ge_inf.iwadid = 2; break;
	case 3: str = "doom2"   ; ge_inf.iwadid = 3; break;
	case 4:
	case 5:
	case 6: str = "doom"; ge_inf.iwadid = 4; break;

	default: Con_ErrorD("Cannot determine wad name");
	}

	mainIwadname = (char*)W_Calloc(strlen(str) + 1, STATICID, 0);
	strcpy(mainIwadname, str);
}


static void checkLoadGame()	// R3D
{
extern char	savefileDeletestr[1024];
    int p = ArgCheck ("-loadgame");

 	if (R3D_edit) return;

   if (p && p < Argc()-1 && Argv(++p)[0] != '-')
    {
		char	buff[512];
		int		i = -1, n = -1;

		strcpy(buff, Argv(p));

		if (!strchr(buff, (int)'/') &&
			!strchr(buff, (int)'\\') &&
			!strchr(buff, (int)':'))
		{
			char *ptr = strchr(buff, (int)'.');

			if (ptr) *ptr = 0;

			i = strnicmp(buff, "Risen3DSav", 10) ? 0 : 10;

			if ( (n = atoi(buff + i)) >= 0 && n < 10)	//NUMSAVESLOTS
			{
				SV_SaveGameFile(n);
				
				if (!SV_CheckLoadGame(true, false, true))	// b211-0 Check switch status
					EngineShutdown();

				CloseSavefile();
				loadsaveset = true;
			}
		}

		if (!loadsaveset)
		{
			sprintf(savefileDeletestr, "LOADGAME ERROR: %s\n\n",
				n < 0 || n > 9 ? "invalid index, must be in range 0 - 9" : i < 0 ? "Only send index e.g. -loadgame 2" : "");

			EngineShutdown();
		}
	}
}

static void checkDetailWad()	// R3D
{
	char	*arg;

	sprintf(detail_wad, "%sData\\Dtex\\", ddRuntimeDir.path);

	if (R3D_edit) return;

	if (   ArgCheckWith("-dtex", 1)
		&& (arg = ArgNext())
		&& arg[0] != '-')
	{
		strcat(detail_wad, arg);
		ge_inf.customdtex = true;
	}
	else
	{
		if (!ArgCheck("-detail")) return;
	}

	strcpy(detail_ded, detail_wad);
	strcat(detail_wad, "\\detail.wad");
	strcat(detail_ded, "\\detail.ded");

	if ( (_access(detail_wad, 4)) == -1)
	{
		Con_Error("DETAIL WAD LOAD ERROR: %s not found", detail_wad);
	}

	if ( (_access(detail_ded, 4)) == -1)
	{
		Con_Error("DETAIL DED LOAD ERROR: %s not found", detail_ded);
	}

	W_AddFile("", detail_wad, false);	// detail wad
}

// When loading a map the original -cmd values can get reset
static void Rerun_CMDArgs()
{
	int		p;

	for (p = 1; p < Argc() - 1; p++)
	{
		if (!_stricmp(Argv(p), "-command") || !_stricmp(Argv(p), "-cmd"))
		{
			// Get -cmd arguments
			// if cmd needs further parms then must be in ""  e.g. "give wa"
			// Each cmd will be processed so -cmd god "give wa" is fine,
			// Note that a space separator is needed; -cmd god"give wa" won't work.

			for (++p; p < Argc(); p++)
			{
				char *arg = Argv(p);

				// check if parm is a new switch, if so then decrement the ptr
				// (so it will be seen if another -cmd) and don't execute.

				if(arg[0] == '-') { p--; break; }

				Con_Execute(arg, false);
			}
		}
	}
}

// v220-13 rewritten to better reject invalid files. Previously this
//         could cause R3D to hang.
//         More than one instance of -data supported plus more than
//         one file following a single instance of -data supported.
static void dataSWITCHcheck(uint	*csm)
{
	FILE	*chkFile;
	RDBOOL	firsttime = true;
	RDBOOL	get_arg = true;
	char	fname[512];
	char	buf[256];
	char	*arg;
	int		type, cnt;

	if (!ArgCheckWith("-data", 1)) return;

	for ( ; ; )
	{
		while (!firsttime)	// Scan the rest of the arguments
		{
			if (!stricmp(arg, "-data")) break;
			if ( !(arg = ArgNext()) ) return;
			if (!get_arg && *arg != '-') break;

			get_arg = true;
		}

		firsttime = false;

		if (get_arg && !(arg = ArgNext()) ) return;

		if (*arg == '-')
		{
			get_arg = true;
			continue;
		}
	 
		get_arg = false;

		sprintf(fname, "%sData\\%s", ddRuntimeDir.path, arg);

		chkFile = NULL;

		if (   !Dir_ChExt(fname, ".WAD")
			|| !(chkFile = fopen(fname, "rb"))
			|| (cnt = fread(buf, 1, 128, chkFile)) < 128
			|| strnicmp(buf, "PWAD", 4))
		{
			if (chkFile) fclose(chkFile);

			Con_Error("Invalid file \"%s\" sent with -data", arg);
		}

		if (chkFile) fclose(chkFile);

		if (!subfn_CheckWAD(fname, csm, &type) || type != TYPE_PWAD)
		{
			Con_Error("File sent with -data is not a valid PWAD");
		}

		if (haslevels)
		{
			Con_Error("File sent with -data has levels. It should only have sounds or textures");
		}

		W_AddFile("", fname, false);	// Resource WADS (no levels)
	}
}

//===========================================================================
//
// DD_Main
//	Engine and game initialization. When complete, starts the game loop.
//
//
//===========================================================================

void palError(void)
{
	Con_Error("%s lump \"PLAYPAL\" is invalid", ge_inf.pwadplaypal ? "PWAD" : "IWAD");
}

void DD_Main(void)
{
	int		sw_auto, val;
	uint	csm;
	int		type;
	char	temp[1536];
	char	iwadfile[256];

	F_InitDFILES();

	memset(&ge_inf, 0, sizeof(geninfo_t));
	memset(&skin_inf, 0, sizeof(skininfo_t));

	// Set default config file.
	sprintf(configFileName, "%srisen3D.cfg", ddRuntimeDir.path);

	if(ArgExists("-dedicated") || ArgExists("-server"))
		ErrorBoxAndQuit("Dedicated or server mode not supported");

	if (ArgCheck("-def"))	// b211-0
		ErrorBoxAndQuit("-def is unsupported: use -d or -defs");

	CFX = !ArgExists("-nc") || ArgExists("-xa");
	BDMFX = CFX || ArgExists("-spf");

	ge_inf.switch_flags = SetSwitchFlags();		// b211-0

	if (ge_inf.devmap = ArgExists("-devmap"))
		ge_inf.noxhair = ArgExists("-noxhair");
	ge_inf.maxcheat = ArgExists("-mcht");
	ge_inf.scp_devpu = ArgExists("-devpu");
	ge_inf.usettfont = !ArgExists("-jfont");
	ge_inf.nomapinfo = ArgExists("-nmapinfo");			// v220-9
	ge_inf.usepwadmusic = ArgExists("-upm");			// v220-13
	ge_inf.usefmodmidi = ArgExists("-fmodmidi");		// v220-20
	ge_inf.nodup = ArgExists("-nodup");					// v220-21
	ge_inf.showconsistency = ArgExists("-showcon");		// v220-22
	ge_inf.showskytagerrors = ArgExists("-sste");		// v220-31
	ge_inf.showtptagerrors = ArgExists("-stte");		// v220-31
	ge_inf.l3dreport = ArgExists("-l3drep");			// v220-31
	ge_inf.l3dreportc = ArgExists("-l3drepc");			// v220-31
	if (ge_inf.l3dreportc) ge_inf.l3dreport = true;
	ge_inf.realtekCompat = ArgExists("-rtekcm");		// v220-33
	ge_inf.reportdamage = ArgExists("-reportdamage");	// v220-33

	if (ArgExists("-listMapTex" )) ge_inf.infflags |= 1;
	if (ArgExists("-noif1")   ) ge_inf.infightsametype |= 1;	// v220-13
	if (ArgExists("-noif2")   ) ge_inf.infightsametype |= 2;	// v220-13
	if (ArgExists("-useFMODs")) ge_inf.useFMODsoftware = true;	// v220-13

	// v220-4 defaults for custompunch
	ge_inf.customhits = 100;
	ge_inf.customdamage = 9999;

	// v220-4 allow corpse model load
	ge_inf.allowCorpse =
		    ArgExists("-nm")
		&& !ArgExists("-xm")
		&& !ArgExists("-nomonstermod")
		&& !ArgExists("-xa")
		&& !ArgExists("-nc");

	if (ArgCheckWith("-dumplump", 1 ))	{ dumpflags = 1; dumpptr = ArgNext(); }
	// R3D additional dump options
	if (ArgExists   ("-findnameI"	))  { dumpflags = 2; dumpptr = ArgNext(); };	// list lumps that contain a chsr string
	if (ArgExists   ("-findnameP"	))  { dumpflags = 3; dumpptr = ArgNext(); };	// list lumps that contain a chsr string
	if (ArgExists	("-dumppngs"	))	dumpflags =  4;
	if (ArgExists	("-dumpjpgs"	))	dumpflags =  5;
	if (ArgExists	("-dumpwaddir"	))	dumpflags =  6;
	if (ArgExists	("-dumpmidis"	))	dumpflags =  7;	// also changes mus to midi where necessary
	if (ArgExists   ("-dumpwavs"	))  dumpflags =  8;	// dumps all pwad wav files
	if (ArgExists   ("-dumpIWADinf" ))  dumpflags =  9;	// dump list of all iwad lumps
	if (ArgExists   ("-dumpPWADinf" ))  dumpflags = 10;	// dump list of all pwad lumps

	if (ArgCheck("-nojoy")) ge_inf.nojoystick = 1;
	if (ArgCheck("-nomouse")) ge_inf.nomouse = 1;

	nomonsters = ArgExists("-nomonsters");
	verbose    = ArgExists("-verbose");	//verbose reporting

	sprintf(midiTempFname, "%sr3dtemp", ddRuntimeDir.path);

	ge_inf.useSitters = ArgExists("-am");

	// v220-6 FlameModels.ded must at least be present
	if (CFX && !ArgExists("-xa") && (ge_inf.mdefflags = BaseModelDefsExist()) & 0x2)
		loadModels = 3;

	filloutlines = !ArgExists("-no_outlines");

	r_use_particles = CFX;
	wasnomonsters = nomonsters;	// remember inital loading state

	//----------------------------

	if (!Edit_Init()) EarlyQuit();
	
	if (R3D_edit)
	{
		opennotepad = false;
		ge_inf.noobjects  = ArgExists("-noobjects");

		if (ArgExists("-loadgame")) ErrorBoxAndQuit("Cannot edit a load game.");
		if (!ArgExists("-warp"))	ErrorBoxAndQuit("A map must be selected (or -warp must be used) with -texedit.");
	}
	else
	{
		if ( (ge_inf.switch_flags & (SWF_GND|SWF_PBM)) == (SWF_GND|SWF_PBM) )
			ErrorBoxAndQuit("Grenade and pipebomb both selected.\n"
							"Please select one or the other, not both");
	}
	
	appStatus = -1;
	Set_CompatOptions();

#ifdef _DEBUG

	if (   ArgExists("-dbline")
		|| ArgExists("-dbssec")
		|| ArgExists("-dbsec")
		|| ArgExists("-dbvtx"))
	{
		compat[c_nomusic] = true;
	}
#endif

	SW_Init();	// Create the startup messages window.

	// Following Con_Init messages can now be printed
	p2swd("Executable: "VERSIONTEXT" v"BUILDTEXT" "__DATE__"\n"); //R3DVER
	p2swd(OsVersionString);
	p2swd("Game path : %s\n",ddRuntimeDir.path);
	p2swd("EDIT MODE : %s\n", R3D_edit ? "ON" : "OFF");
	p2swd("Model defs: %.8x\n", ge_inf.mdefflags);

	{
		int		val = 0;
		char	*defname = DefCheckBase(&val);
	
		if (val < 0)
			Con_Error("ERROR: Missing base def %s\nFIX: Reinstall Risen3D", defname);
		if (val > 1)
		{
			Con_Error("ERROR: %s is out of date or has been modifed\n"
					  "Using this would cause a problem\n\n"
					  "FIX: Reinstall Risen3D", defname);
		}
	}

	if (verbose)
	{
		pr_OFdirect("===================================================\n");
		pr_OFdirect("NOTE: verbose messages are not echoed to the screen\n");
		pr_OFdirect("      they are only printed to Risen3D.out\n");
		pr_OFdirect("===================================================\n");
	}

	R3D_UdirInit();		// Also sets the savedir path

	sw_auto = DetectIWADs(iwadfile);

	// If there's a detail wad and this is wanted add this first
	checkDetailWad();

	// Set up IWAD
	if (!subfn_CheckWAD(iwadfile, &csm, &type) || type != TYPE_IWAD)
		Con_Error("File sent with -iwad is not a valid IWAD");

	W_AddFile("", iwadfile, false);	// IWAD
	IW_InitIWADTextures();	// v220-16

	cwinf.iwadisdoom2 = IdentifyVersion(sw_auto, &val);
	ge_inf.isDoom = !cwinf.iwadisdoom2;

	if (cwinf.iwadisdoom2 == -1)
		Con_ErrorD("IWAD type cannot be determined");

	sprintf(temp, "%sData\\Risen3D.wad", ddRuntimeDir.path);

	// v220-4 wad tampering check
	if (   !subfn_CheckWAD(temp, &csm, &type)
		|| type != TYPE_PWAD
		|| ((byte*)&csm)[0] != 0x7A
		|| ((byte*)&csm)[1] != 0x0F
		|| ((byte*)&csm)[2] != 0x15
		|| ((byte*)&csm)[3] != 0x00
		)
	{
		Con_Error("ERROR: The Risen3D.wad in\n  %sData\\\n"
			"sum check %.2X %.2X %.2X %.2X\n"
			"is the wrong version or has been altered",
			ddRuntimeDir.path, ((byte*)&csm)[0],((byte*)&csm)[1],((byte*)&csm)[2],((byte*)&csm)[3]);
	}

	W_AddFile(ddRuntimeDir.path, "Data\\Risen3D.wad", false); // FONTA and FONTB, M_THERM2 etc.

	if (ArgExists("-dogs"))
		W_AddFile(ddRuntimeDir.path, "Data\\MBFDOGS.wad", false); // MBF dogs + sounds

	dataSWITCHcheck(&csm);

	HandlePWADS(); //  detects PWADs defined with -file or with -udir.

#if 0	// external lump insertion. Ultimately removed as is potentially a recipe for disaster
	if (ArgExists("-lmp"))
	{
		int i = 0;

		while( i < Argc())
		{
			if (!_stricmp(Argv(i), "-lmp"))
			{
				while(++i < Argc() && !ArgIsOption(i))
				{
					if (r3dUserDir.path[0] && subfn_CheckNotRedirected(Argv(i)))	// b211-0
						W_AddFile(r3dUserDir.path, Argv(i), false);	// LMP files b211-0
					else
						W_AddFile("", Argv(i), false);				// LMP files
				}
			}
			else
				++i;
		}
	}
#endif

// v220-4	Autoload, for development purposes, scp_defn, scp_zoff
//			and sndinfo files stored in the same udir as the pwad.
//			Released files should have these as embedded lumps.
	if (ge_inf.devmap && autoload && r3dUserDir.path[0])
	{
		long	hFile;
		struct	_finddata_t d_file;

		if (Dir_ChDir(&r3dUserDir))
		{
			char	*fi[3] = { "SCP_DEFN.*", "SCP_ZOFF.*", "SNDINFO.*" };
			char	fname[64];
			int		i;

			for (i = 0; i < 3; i++)
			{
				fname[0] = 0;

				// Find first fi[i] file in current directory
				if ( (hFile = _findfirst( fi[i], &d_file )) != -1L )
				{
					// In case some idiot's used a ridiculous extension
					if (strlen(d_file.name) > 32)
					{
						_findclose( hFile );
						Con_Error("File %s has an excessively long extension", d_file.name);
					}

					strcpy(fname, d_file.name);

					_findclose( hFile );
				}

				// Call after _findclose in case Con_Error occurs
				if (fname[0]) W_AddFile(r3dUserDir.path, fname, false);	// external SCRIPTs (devmode only)
			}
		}
	}

	// Check that the dump folder and folder for analysis files exist
	sprintf(dumppath, "%s%s", ddRuntimeDir.path, r3ddump);
	if ( (_access(dumppath, 0)) == -1) mkdir(dumppath);
	sprintf(temp, "%sgwa\\", ddRuntimeDir.path);
	if ( (_access(temp, 0)) == -1) mkdir(temp);
	sprintf(temp, "%sR3Danfiles\\", ddRuntimeDir.path);
	if ( (_access(temp, 0)) == -1) mkdir(temp);

	if (ArgExists("-xstat")) ge_inf.xstat = true;	// v220-8

	LoadIWADname(val);
	if (!analysename ) analysename  = mainIwadname;
	if (!mainPwadname) mainPwadname = mainIwadname;

	// b220-0 Check if an addon wad is available
	// v220-35 extended to load more than one file. Also checks that a 
	// wad is not in the R3D_CheckWads list if models are called for.
	if (haspwads)
	{
		long    hFile;    
		struct  _finddata_t finddata;

		sprintf(temp, "%sData\\R3D_addon\\%s\\", ddRuntimeDir.path, analysename);

		if (!_chdir(temp))
		{
			if ((hFile = _findfirst( "*.*", &finddata )) != -1L )
			{
				do
				{
					uint n = strlen(finddata.name);

					if (n > 10)
					{
						if (!stricmp(finddata.name + n - 10, "_addon.wad"))
						{
							ge_inf.addon_handle = -1;
							sprintf(temp, "%sData\\R3D_addon\\%s\\%s", ddRuntimeDir.path, analysename, finddata.name);
							W_AddFile("", temp, false);		// add-on wads
						}
					}
				}
				while( _findnext( hFile, &finddata ) == 0 );

				_findclose( hFile );
			}
		}

		if (loadModels)
		{
			sprintf(temp, "%sR3D_CheckDEHWads", ddBinDir.path);

			if (SR_LoadFile(temp, "R3D_CheckDEHWads", NULL))
			{
				char *nextstr;

				while ( (nextstr = SR_NextSTR(false)) )
				{
					if (!stricmp(nextstr, analysename))
					{
						SR_EndScript();
						Con_Error(	"Models or core effects are not supported with\n"
									"heavily DEHACKED wads like %s as state\n"
									"incompatibility will lead to errant behaviour.\n\n"
									"Please restart using either Basic Doom Mode or FX\n",
									analysename);
					}
				}
			}
			else
			{
				Con_Error("R3D_CheckDEHWads not found");
			}

			SR_EndScript();
		}

		_chdir(ddRuntimeDir.path);
	}

	if (ArgCheckWith("-cd", 1)) pr_OFdirect("ERROR: -cd no longer supported");
	p2swd("Analyse base name: %s\n", analysename);
	p2swd("Wad base name    : %s\n", mainPwadname);

	if (haspwads)
	{
		if (gamemode == shareware)
			Con_Error(
				" You cannot use the shareware game to play pwads!\n"
				" To do so the full version is required which can\n"
				" be purchased for a nominal sum from;\n\n "
				DOOMWADSUPPLIER"\n\n");

		ge_inf.hasdd_defns = W_CheckPWADNumForName("DD_DEFNS");

		// Must be decided before calling Def_Read
		if (!R3D_edit && (ge_inf.hasscp_defn = W_CheckPWADNumForName("SCP_DEFN") >= 0) )
		{
			if (!CFX)
				Con_Error("R3D scripted map: In the launcher untick Basic Doom Mode");

			if ((ge_inf.mdefflags & 0xFF) != 0xFF)
				Con_ErrorD(	"R3D scripted map: All default models must be installed\n\n"
							"If models have been installed then in the Launcher, under\n"
							"the Models tab, ensure in the \'Disable Model Group\'\n"
							"that all boxes are unticked");
		}

		if (!compat[c_stairs]) ge_inf.flags |= GEF_COMPSTAIRS;
		if ( compat[c_model ]) ge_inf.flags |= GEF_COMPMODEL;
	}
	else
	{
		compat[c_model] = 1;
		ge_inf.flags |= GEF_COMPMODEL;
	}

	if (!ge_inf.use_sky_box)
	{
		if (ArgExists("-usb")) ge_inf.use_sky_box = GE_SKYBOXUSE;
		if (ArgExists("-fsb")) ge_inf.use_sky_box = GE_SKYBOXFORCE;
	}

	// v220-6 revised
	titlepic_num = W_CheckNumForName("TITLEPIC");

	if (titlepic_num < 0) Con_ErrorD("Error: IWAD is faulty");

	interpic_num = W_CheckNumForName("INTERPIC");

	if (interpic_num < 0)
	{
		interpic_num = titlepic_num;

		if (sw_auto > 3)
		{
			char msg[512];
			int  i;

			sprintf(msg,
				"\n\n You are using the shareware version of doom which limits your\n"
				" options. You may wish to consider purchasing the full version\n"
				" which is available from;\n\n "DOOMWADSUPPLIER"\n\n"
				);

			Sys_MessageBox(msg, false);

			pr_OFdirect("\n\n");
			for (i = 0; i < 64; i++) pr_OFdirect("!");
			pr_OFdirect(msg);
			for (i = 0; i < 64; i++) pr_OFdirect("!");
			pr_OFdirect("\n\n");
		}
		else
		{
			Con_Error(
				"Your doom wad is an early version that needs updating to version 1.9\n"
				"This can be done, free of charge, by using patches released by\n"
				"iD Software. These can be downloaded from %s", RISEN3DLINK
				);
		}
	}

	// v220-3 Development only - the TGASKYFOLDER script command is normally used
	if (ArgCheckWith("-skybox", 1))
	{
		char *ptr = ArgNext();

		if (strlen(ptr) && strlen(ptr) < 32)
		{
			strcpy(skyboxname, ptr);
			ge_inf.noskyreset = true;
		}
	}

	if (!R3D_edit && !dumpflags)
	{
		if (nomonsters && ge_inf.hasscp_defn)
		{
			char *msg = {	" R3D scripted map : nomonsters was set on entry\n"
							" but the script may require their being spawned.\n\n"
							" Please press No, uncheck the nomonsters button\n"
							" in the Risen3D launcher and restart.\n"};

			if (ge_inf.devmap)
			{
				pr_OFinsert("%s\nUser has invoked devmap mode\n", msg);
			}
			else if (!Sys_CriticalMessage(msg, true))
			{
				Sys_Quit(false);
			}

			if (compat[c_nextmap])
			{
				pr_OFdirect("ADVISORY  : -c_nextmap is ignored with scripted maps\n");
				compat[c_nextmap] = false;
			}
		}

		if (!base_save_path[0])
			sprintf(base_save_path, "%sUserSG\\%s\\", ddRuntimeDir.path, analysename);

		// Add trailing backslash to savepath if necessary
		if (base_save_path[strlen(base_save_path)-1] != '\\')
			strcat(base_save_path, "\\");
		if (!strstr(base_save_path, "savegame"))
			strcat(base_save_path, "savegame\\");

		M_CheckPath(base_save_path);

		pr_OFinsert("SAVE PATH : \"%s\"\n",base_save_path);
	}

	// Check that "PLAYPAL" exists and set the buffer
	// NOTE: a PWAD PLAYPAL lump can be oversized but remain viable
	{
		size_t		r = 768;	// 3 * 256
		int			iwadlump = W_CheckIWADNumForName("PLAYPAL");
		int			pwadlump = W_CheckPWADNumForName("PLAYPAL");
		long		pos;               // Position of file pointer
 		lumpinfo_t	*l;

		// Even if the PWAD has its own playpal the iwad playpal
		// will still be needed - load this first
		l = lumpinfo + iwadlump;

		if (iwadlump < 0 || iwadlump >= numlumps || r > l->size) palError();

		iwadpalbuffer = (byte*)W_Calloc(r, STATICID, 0);

		pos = lseek(l->handle, l->position, SEEK_SET);

		if (pos == -1L ||  (pos = read(l->handle, (void*)iwadpalbuffer, r)) == -1L) palError();

		if (pwadlump >= 0 && pwadlump < numlumps)
		{
			l = lumpinfo + pwadlump;

			if (r > l->size) goto use_iwadpal;

			ge_inf.pwadplaypal = true;
			palbuffer = (byte*)W_Calloc(r, STATICID, 0);

			pos = lseek(l->handle, l->position, SEEK_SET);

			if (pos == -1L ||  (pos = read(l->handle, (void*)palbuffer, r)) == -1L)
			{
				W_Free(palbuffer, STATICID, 0);
				goto use_iwadpal;
			}
			ge_inf.hadpwadplaypal = W_CheckPWADNumForName("STCFN033") >= 0;
		}
		else
		{
use_iwadpal:;
			if (ge_inf.pwadplaypal)
				pr_OFdirect("ERROR: Bad pwad PLAYPAL- reverting to IWAD PLAYPAL\n");
			palbuffer = iwadpalbuffer;
			iwadpalbuffer = NULL;
		}

		// v220-21 If a pwad palette exists check if it should
		//         inhibit any external hi-res texture commands.
		//			Note that a pwad palbuffer will already have
		//			been loaded at this point.
		if (palbuffer && iwadpalbuffer)
		{
			int	i, k, p, r = 768;
				
			// compare the two buffers to ensure the pwad palette
			// is not just a duplicate of the iwad palette

			for (i = p = 0; i < 256; i++)
			{
				for (k = 0; k < 3; k++)	// scan rgb values
				{	// If change is significant
					if (palbuffer[i*3 + k] != iwadpalbuffer[i*3 + k])
					{
						palptr[i] = true;
						p++;
					}
				}
			}
				
			if (p > 4)
			{
				pr_OFdirect("PWAD palette detected with significant changes\n"
							"All external hires substitution will be inhibited\n");
			}
			else	// stick with the iwad playpal
			{
				ge_inf.pwadplaypal = false;
				W_Free(palbuffer, STATICID, 0);
				palbuffer = iwadpalbuffer;
				iwadpalbuffer = NULL;
			}
		}

		pr_OFinsert("Using %s PLAYPAL\n", ge_inf.pwadplaypal ? "PWAD" : "IWAD");
	}

	// For development test/purposes (not the casual user)
#ifdef _DEBUG
	if (ArgExists("-zdnodg" )) ge_inf.modifier |= MODF_ZDNODG;
#endif

	FREEDOOMcheck();

	// R3D:	Function to aid finding lump types
	if (dumpflags)
	{
		if (dumpflags ==  1) DumpLump   (dumppath, dumpptr);
		if (dumpflags ==  2) PrintLumpIfNameContains(0, dumpptr);	// v220-35
		if (dumpflags ==  3) PrintLumpIfNameContains(1, dumpptr);	// v220-35
		if (dumpflags ==  4) DumpPngs   (dumppath);
		if (dumpflags ==  5) DumpJpgs   (dumppath);
		if (dumpflags ==  6) DumpWaddir (dumppath);
		if (dumpflags ==  7) DumpMidis  (dumppath);
		if (dumpflags ==  8) DumpWav(dumppath);
		if (dumpflags ==  9) DumpLumpInf(dumppath, 0);
		if (dumpflags == 10) DumpLumpInf(dumppath, 1);
	}

	Subfn_CheckGRENPBM();

	if (loadModels)
	{
		if (ArgExists("-noamwds"))
		{
			if (!ge_inf.hasscp_defn)	// don't care if scripted
				ge_inf.switch_flags |=  SWF_NOAMWDS;
		}
		else
		{
			ge_inf.amwds = true;
		}
	}

	p2swd("Switch flags     : %.8X\n", ge_inf.switch_flags);

	DD_InitInput();		// Initialize the key mappings.
	bamsInit();			// Binary angle calculations.
	Def_Init();			// Initialise the ded definitions.
	P_InitThinkers();	// Initialise the thinkers (see P_Ticker)
	P_InitParticleThinkers();

	Con_Init();	// Initialise console
	InitCfg_and_PlrArray();
	cfg.oldbarscale = -1;

	// Console background
	console_flat = W_CheckRisen3DWADNumForName("#CONBAK#", true);

	// Initialize subsystems
	Net_Init(); // Network before anything else.
	MC_SetMidiDevs();	// Must be run before loading config

	// Check for a custom config file.
	if (ArgCheckWith("-config", 1))
	{
		// This now overrides the default config file.
		strcpy(configFileName, ArgNext());
	}

	Con_ParseCommands(configFileName, true);

	// Parse additional files (that should be parsed BEFORE init).
	if (ArgCheckWith("-cparse", 1))
	{
		for(;;)
		{
			char *arg = ArgNext();
			if(!arg || arg[0] == '-') break;
			pr_OFinsert("Parsing: %s\n", arg);
			Con_ParseCommands(arg, false);						
		}
	}

//------ Order must not be changed

	F_InitDirec();

	// Now that we've read the WADs we can initialize definitions.
	Def_Read(autoload);

	if (!(SittersLoaded & 16) && ge_inf.hasscp_defn)
	{
		Con_ErrorD( "SCRIPTED MAP change with version v330-35\n\n"
			"Sitter's and Abbs model replacements must be selected\n"
			"in the launcher for use with R3D Scripted maps.\n\n"
			"If not installed then they can be obtained from\n"RISEN3DLINK);
	}

	// Check for DEH sprites/things
	// Must be done prior to calling checkLoadGame()
	if (hasDEH)
	{
		extern int	DEHchange;

		if (DEHchange & DEHF_STATE)
		{
			// v220-35 do not allow gnd/pbm if dehacked states have beem detected
			if (ge_inf.switch_flags & (SWF_GND|SWF_PBM))
			{
				Con_Error("DEHACKED states detected\n\n"
					"Please disable %s and restart", ge_inf.switch_flags & SWF_GND ? "grenade" : "pipebomb");
			}
			ge_inf.switch_flags |= SWF_DEH;	// for loadgame checks
		}

		if (!(ge_inf.switch_flags & SWF_DEH))
		{
			for (int i = 0; i < DH_CHECKNUM; i++)
			{
				if (DEHCheck[i] & (DEHF_SPRITE|DEHF_STATE))
				{
					ge_inf.switch_flags |= SWF_DEH;
					break;
				}
			}
		}
	}

	checkLoadGame();
//-------------------------------

	Sys_Init();
	S_Init();
	R_Init();
	// The message window can now be closed.
	// Only used by R_InitTextures()
	SW_Shutdown();

	if (hasDEH)
	{
		ge_inf.batman = !stricmp(analysename, "batman");	// v220-24
		ge_inf.arctlev = !stricmp(analysename, "arctlev");
	}

	IW_SetHiresTexPaths();
	R_InitSprites();
	H_InitLODstates();

	// Start the GL driver.
	GL_Init();
	GL_Info();
	GL_InitTextureManager();
	GL_InitRefresh();
	GL_SetHudsprites();

	pr_OFdirect("Con_Startup  : Init startup screen.\n");

	R_Init2();

	D_DefaultBindings();
	D_PostInit();

	P_Init() ;pr_OFdirect("P_Init : Playloop state initialised\n");
    HU_Init();pr_OFdirect("HU_Init: Heads up display set\n");
    ST_Init();pr_OFdirect("ST_Init: Status bar initialised\n");
    M_Init() ;pr_OFdirect("M_Init : Menu info initialised\n");
	P_PresetIWADkeyLumps();
	
	Def_PrintStats();

	if (ge_inf.models_count < 1)
		loadModels = CFX = 0;

	// Autoexec support removed - caused more trouble than it's worth - R3D
	//	Con_ParseCommands("autoexec.cfg", false);

	// Parse additional files.
	if(ArgCheckWith("-parse", 1))
	{
		for(;;)
		{
			char *arg = ArgNext();
			if(!arg || arg[0] == '-') break;
			pr_OFinsert("Parsing: %s\n", arg, false);
			Con_ParseCommands(arg, false);						
		}
	}

	appStatus = 0;	// R3D
	R_InitZDist();
	LoadSmokeOffsets();

	MC_SetMidiDevice();
	appStatus = 1;	// R3D

	// Save conditions on entry ------------------------------
	if (cfg.useTorchLight && cfg.setVtxLight)
		cfg.useTorchLight = 0;

	// set in console
	if (cfg.useTorchLight	) ge_inf.flags |= GEF_TORCHLIGHT;
	if (cfg.setVtxLight		) ge_inf.flags |= GEF_VTXLIGHT;
	if (cfg.sineCompress	) ge_inf.flags |= GEF_SINECOMPRESS;
	// set in game options
	if (cfg.fallingDamage	) ge_inf.flags |= GEF_FALLINGDAMAGE;
	if (cfg.underwaterDamage) ge_inf.flags |= GEF_UWDAMAGE;
	if (cfg.jumpEnabled		) ge_inf.flags |= GEF_JUMP;
	if (cfg.footsteps		) ge_inf.flags |= GEF_FOOTSTEPS;
	if (cfg.weaponsounds	) ge_inf.flags |= GEF_WEAPRSOUND;
	if (cfg.weapsmoke		) ge_inf.flags |= GEF_WEAPSMOKE;
	if (cfg.splashsprite	) ge_inf.flags |= GEF_SPLASHSPRITE;

	if (waterdynamics		) ge_inf.flags |= GEF_WATERDYNAMICS;
	if (useDynLights		) ge_inf.flags |= GEF_DYNLIGHTS;
	if (water_trans			) ge_inf.flags |= GEF_WATERTRANS;

	ge_inf.jumpPower = cfg.jumpPower;
	ge_inf.was_water_opacity = water_opacity;
	ge_inf.maxLightDistance = maxLightDistance;
	ge_inf.num_modelLight = modelLights;

	if (ArgCheckWith("-floff", 1))
		ge_inf.flashoffset = atoi(ArgNext());
	else
		ge_inf.flashoffset = 50;

	if (ArgCheckWith("-flint", 1))
		ge_inf.flashintensity = atoi(ArgNext());
	else
		ge_inf.flashintensity = 70;
	// --------------------------------------------------------

	appStatus = 2;
	{
		int tcd = consoleDump;

		consoleDump = false;
		Con_playFPrintf(true, CBLF_WHITE | CBLF_CENTER, ""VERSIONTEXT" v"BUILDTEXT" "__DATE__"\n"); //R3DVER
		Con_playFPrintf(true, CBLF_WHITE | CBLF_CENTER, OsVersionString);
		if (R3D_edit) Con_playFPrintf(true, CBLF_WHITE | CBLF_CENTER, "USING EDIT MODE\n");
		Con_Printf("\n\n\n");
		consoleDump = true;

		Con_playFPrintf(true, CBLF_RULER | CBLF_WHITE | CBLF_CENTER, 
			gamemode == retail? 
			"The Ultimate DOOM Startup\n"
			: gamemode == shareware?
			"DOOM Shareware Startup\n"
			: gamemode == registered?
			"DOOM Registered Startup\n"
			: gamemode == commercial?
			(gamemission == pack_plut?
				"Final DOOM: The Plutonia Experiment\n"
				: gamemission == pack_tnt? "Final DOOM: TNT: Evilution\n"
				: "DOOM 2: Hell on Earth\n")
			: "Public DOOM\n");
		Con_playFPrintf(false, CBLF_WHITE | CBLF_CENTER, 
			gamemission == pack_plut?
				"Dario Casali and Milo Casali\n"
				: gamemission == pack_tnt? "Team TNT\n"
				: "id Software\n");

		Con_playFPrintf(true, CBLF_GREEN | CBLF_CENTER, "Note: id Software cannot provide technical support for Risen3D\n");
		Con_playFPrintf(true, CBLF_GREEN | CBLF_CENTER, "Direct any issues to %s\n", RISEN3DLINK);
		Con_playFPrintf(true, CBLF_RULER, "\n");

		consoleDump = tcd;
	}

	if (headerDEHinsert)
	{
		Con_playFPrintf(true, CBLF_LIGHT | CBLF_CENTER, headerDEHinsert);
		M_Free(headerDEHinsert); headerDEHinsert = NULL;
	}

	ge_inf.oldgamenum = -1;

	if (loadsaveset)
	{
		brief_disabled = -1;
		G_DoLoadGame();
	}
	else
	{
		if (autostart)
		{
			G_InitNew(gameskill, startepisode, startmap);
			Rerun_CMDArgs();	// rerun any -cmd line args
		}
		else
			D_StartTitle();		// start up intro loop
	}

	Edit_SetVars();

	DD_GameLoop();				// Never returns...
}
