//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_MAINspec.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
//		LS_TraceLiquidNormal() removed
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: analyser entry point
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "sys_timer.h"
#include "doom\p_start.h"
#include "r_data.h"
#include "r_model.h"
#include "rend_list.h"
#include "con_bar.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "r3d_compat.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

ssline_t *sslines;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	TP_CheckTPSpawnedSectors(RDBOOL removeThing);
void	P_LoadThings (int lump);
void	LL_LineLoops(void);
void	WA_CheckAnalyseFile(char *name, int from);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void R3D_MAINsetarrays(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char	*tg_markline;
extern char	*tg_marklinemem;
extern char	*tg_specline;
extern char	*tg_linedone;
extern byte	*tg_ssecflags;
extern char	*show_wadname;
extern int	use_xl_lines;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

uint	maxsecsz;
uint	GLVertStart;

// timings
float	timecache[ITC_CACHENUM];

// arrays for R3D_MAINAnalyseMap
int		*countsec;
byte	*conarray;
byte	*hitline;
byte	*excludeline;
int		*hitssec;
int 	*lfsecarray;
int 	*lgrpnum;
int		*linearray;
int		*lnsec;
int		*sidelist;
short	*lineflags;
byte	*linetable;
byte	*lxarray;
byte	*mtarray;
char	*secdone;
char	*secmarker;
short	*secspflags;
byte	*ssecarray;
byte	*vtxcheck;
char	*vtxarray;
char	*vtxflags;
char	*vtxflags2;

int *ssinf_ownrcptr;
int *ssinf_otherlorcptr;
int *ssinf_otherlineptr;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static void markSwitchLines()
{
	line_t	*ld;
	int		i;

	// Mark switch lines for bullet aiming purposes
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int txn;

		if (ld->special)
		{
			if ( (txn = sides[ld->sidenum[0]].bottomtexture)
				&& texArray[txn]->txf_flags & TXF_SWITCH)
			{
				ld->lnflags2 |= LNF2_OWNS_SWITCH;
			}
			else
			if ( (txn = sides[ld->sidenum[0]].toptexture)
				&& texArray[txn]->txf_flags & TXF_SWITCH)
			{
				ld->lnflags2 |= LNF2_OWNS_SWITCH;
			}

			if (ld->sidenum[1] < 0) continue;

			if ( (txn = sides[ld->sidenum[1]].bottomtexture)
				&& texArray[txn]->txf_flags & TXF_SWITCH)
			{
				ld->lnflags2 |= LNF2_OWNS_SWITCH;
			}
			else
			if ( (txn = sides[ld->sidenum[1]].toptexture)
				&& texArray[txn]->txf_flags & TXF_SWITCH)
			{
				ld->lnflags2 |= LNF2_OWNS_SWITCH;
			}
		}
	}
}

static RDBOOL compareSectors(line_t *ld)
{
	sector_t *s1 = ld->frontsector;
	sector_t *s2 = ld->backsector;

	return	s2 &&
			s1 != s2 &&
			s1->floorheight   == s2->floorheight	&&
			s1->ceilingheight == s2->ceilingheight	&&
			s1->floorpic      == s2->floorpic		&&
			s1->ceilingpic    == s2->ceilingpic;
}

// Check for lines crossing R3D_04 lines
static void subfn_CheckLinesCrossingR3D_04()
{
	line_t	*ld = lines, *lc;
	int		i;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_3F &&
			lxarray[i] & 4 &&
			!ld->mtflags &&
			!ChkLptr(ld) &&
			!secdone[ld->frontsector - sectors] &&
			!secdone[ld->backsector - sectors] &&
			compareSectors(ld) && // Note: if front and back secs the same then returns false
			(lc = subfn_FindLineCrossing(ld, LC_NORMAL)) != NULL &&
			lc->specType == R3D_04)
		{
			Setspec(ld, ld->frontsector - sectors, 7001);
		}
	}
}

// Check for unset R3D_1D lines
static void subfn_CheckUnsetR3D_1D()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k;

	// Check for type such as in vile map24 sec 806
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& SecCtrl(i)
			&& sec->specFlags & SF_MOVSECTOR)
		{
			sector_t *s = NULL, *chk;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (ld->specType == R3D_1D
					// Eliminate map errors such as vile map19 sec 263/264
					&& !ChkLFGSecSet(ld)
					&& !ChkLptr(ld)
					&& ld->frontsector != ld->backsector)
				{
					if (   vtxarray[ld->v1 - vertexes] == 1
						|| vtxarray[ld->v2 - vertexes] == 1)
					{
						int chknum = LS_FindAdjacentSecnum(ld, 4);
						chk = chknum >= 0 ? &sectors[chknum] : NULL;
						if (chk == NULL)
						{
							s = NULL;
							break;
						}
					}
					else
					{
						chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;
					}

					if (s == NULL) s = chk;
					if (s != chk
						|| s->specFlags & SF_MOVSECTOR)
					{
						s = NULL;
						break;
					}
				}
				else
				{
					s = NULL;
					break;
				}
			}

			if (s != NULL)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (sec->lines[k]->specType == R3D_1D)
					{
						Setspec(sec->lines[k], s - sectors, 7003);
						SetLFGSecSet(sec->lines[k], 7004);
					}
				}

				Setsecdone3(sec, 7005);
			}
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_1D
			&& !ChkLFGSecSet(ld)
			&& !ChkLptr(ld)
			&& LSS(ld)
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR))
		{
			Setspec(ld, LS_GetAdjacentLineSecnum(ld, 0), 7006);
		}
	}
}

// Check unconnected trigger lines are in the correct sector;
static void subfn_CheckTriggerLinesInCorrectSec()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (LSS(ld) &&
			ld->frontsector->specFlags & SF_MOVSECTOR &&
			!ChkLptr(ld) &&
			ld->special &&
			ld->tag &&
			!origTEX[i] &&
			Vtxs1(ld) &&
			subfn_FindLineCrossing(ld, LC_NORMAL) == NULL)
		{
			// This will first check and then, if required, set ss->floorsec/ceilsec
			LS_SetOverrideDirect(ld);
		}
	}
}

static RDBOOL allLinesConnected(sector_t *sec, line_t *lc)
{
	int i, repeat = true;

	vtxcheck[lc->v1 - vertexes] = 1;
	vtxcheck[lc->v2 - vertexes] = 1;
	hitline[lc - lines] = 0;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (hitline[ld - lines])
			{
				if (vtxcheck[ld->v1 - vertexes]
					|| vtxcheck[ld->v2 - vertexes])
				{
					vtxcheck[ld->v1 - vertexes] = 1;
					vtxcheck[ld->v2 - vertexes] = 1;
					hitline[ld - lines] = 0;
					repeat = true;
				}
			}
		}
	}
	for (i = 0; i < sec->linecount; i++)
	{
		if (hitline[sec->lines[i] - lines])
			return false;
	}
	return true;
}

void LS_CheckConnected()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i]
			&& sectors[i].heightsec < 0
			&& SecCtrl(i)
			&& MSFC(i, 0x00210000))
		{
			sector_t *sec = &sectors[i];
			line_t	 *lc = NULL;
			int		 k, specSec = -1;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (   ld->frontsector != ld->backsector
					|| origTEX[ld - lines] & 0x22)
				{
					lc = NULL;
					break;
				}

				hitline[ld - lines] = 1;

				if (lnsec[ld - lines] >= 0)
				{
					if (specSec < 0) specSec = lnsec[ld - lines];
					if (specSec != lnsec[ld - lines])
					{
						lc = NULL;
						break;
					}
				}
				else
				{
					lc = ld;
				}
			}

			if (!lc || specSec < 0 || specSec == sec - sectors)
				continue;

			CLR_vtxcheck();

			if (allLinesConnected(sec, lc))
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (!(ld->spflags & SPF_CONTROLLINE))
					{
						Setspec(ld, specSec, 7007);
					}
				}
			}
		}
	}
}

//============================================================================

//
// R3D_MAINAnalyseMap
//
// This is the main special effects routine calling sub functions used
// to set sector pointer (specSec) in a linedef's line_t allowing for
// the correct setting of flats.  And if you think this was easy to work
// out then you haven't seen my grey hairs or haggard expression.
//

// Print stats
// b211-0 - map stats moved from p_setup.c and revised

extern float	stat_analysetime;
extern float	stat_nodestime;
extern int		stat_vispresent;
extern int		num_xlsame;
extern int		ambientfound;

void LS_PrintGeneralInfo(int preanalysed)
{
	pr_OFrepeat("\nGENERAL INFO ", "*", 32, 2);
	pr_OFinsert("  GL_VERT v%i.0 detected\n",  ma_inf.nodesFormat);
	pr_OFinsert("  LINEDEFS    %6i\n", numlines);
	pr_OFinsert("  SIDEDEFS    %6i", ma_inf.mapsides + 2 * num_xlsame);

	if (ma_inf.numsidecheck > ma_inf.mapsides)
		pr_OFinsert("  [unpacked] %i", ma_inf.numsidecheck);
	if (num_xlsame)
		pr_OFinsert("  [used] %i", ma_inf.mapsides);

	pr_OFdirect("\n");
	pr_OFinsert("  VERTEXES    %6i\n"
				"  SECTORS     %6i\n"
				"  GL_SEGS     %6i\n"
				"  GL_MINISEGS %6i\n"
				"  GL_SSECS    %6i\n"
				"  GL_NODES    %6i\n"
				"  GL_VERTS    %6i\n\n",
				GLVertStart,
				numsectors, numsegs, ma_inf.minisegs, numsubsectors,
				numnodes, (uint)numvertexes - GLVertStart);

#ifdef _DEBUG
	{
		int i;

		for (i = 0; i < NUM_MCLASS; i++)
			pr_OFinsert("  winmem ptrs set class %i : %10i\n", i, ma_inf.mempointers[i]);

		pr_OFinsert("  winmem bytes used       : %10u\n", ma_inf.memsize);
	}
#endif

	if (ge_inf.xstat)
	{
		if (use_xl_lines)
			pr_OFinsert("  num_xlsame              : %7i\n", num_xlsame);
		if (stat_vispresent)
			pr_OFinsert("  vispresent              : %7i\n", stat_vispresent);
		if (ma_inf.badsslines > 0)
			pr_OFinsert("  Bad 1's lines in map    : %7i\n", ma_inf.badsslines);
		if (ma_inf.degenerate > 0)
			pr_OFinsert("  Degenerate subsectors   : %7i\n", ma_inf.degenerate);
		// If not preanalysed then next two are recovered from analysis file
		if (ma_inf.changeconflicting)
			pr_OFinsert("  ssecs conflicting       : %7i\n", ma_inf.changeconflicting);
		if (ma_inf.sseclogical)
			pr_OFinsert("  SSECs [logical] changed : %7i [of %5i]\n", ma_inf.sseclogical, numsubsectors);
	}

	if (ge_inf.suppress_duplicate)
		pr_OFinsert("  unique pwad sprites     : %7i\n", ge_inf.suppress_duplicate);
	if (ma_inf.voodoodolls)
		pr_OFinsert("  Voodoo dolls spawned    : %7i\n", ma_inf.voodoodolls);
	if (ma_inf.friendsfound)
		pr_OFinsert("  MBF friends found       : %7i\n", ma_inf.friendsfound);

	if (ma_inf.scriptthings77 || ma_inf.scriptthings78 || ma_inf.scriptthings79)
	{
		pr_OFinsert("  Script things 77nn      : %7i\n", ma_inf.scriptthings77);
		pr_OFinsert("  Script things 78nn      : %7i\n", ma_inf.scriptthings78);
		pr_OFinsert("  Script things 79nn      : %7i\n", ma_inf.scriptthings79);
	}

	if (ma_inf.reverbthings)
		pr_OFinsert("  Script defined reverb   : %7i\n", ma_inf.reverbthings);
	if (ambientfound)
		pr_OFinsert("  Ambients found          : %7i\n", ambientfound);
	if (ma_inf.architmodels)
		pr_OFinsert("  Architectural models    : %7i [spawned]\n", ma_inf.architmodels);
	if (ma_inf.slopecount)
		pr_OFinsert("  Sloped sectors found    : %7i [of %5i]\n", ma_inf.slopecount, numsectors);
	if (ma_inf.num3Dlines)
		pr_OFinsert("  3D lines found          : %7i [of %5i]\n", ma_inf.num3Dlines, numlines);
	if (newtexcount)
		pr_OFinsert("  Textures added          : %7i\n", newtexcount);

	// Limits
	if (ge_inf.xstat)
	{
		pr_OFinsert("  Max polysides           : %7i\n", ma_inf.max_polynum);
		pr_OFinsert("  Max  sec/vertex found   : %7i [vertex %i]\n",											ma_inf.max_ownersec[0], ma_inf.max_ownersec[1]);
		pr_OFinsert("  Mobj thinkers spawned   : %7i [of %5i]\n", ma_inf.numthinkers - ma_inf.thinkersremoved, ma_inf.numthinkers);
	}

	// Timings
	if (timecache[ITC_UNPACK])
		pr_OFinsert("  Unpack sides            : %7.3f sec\n", timecache[ITC_UNPACK]);

	if (timecache[ITC_NODESBUILD])
		pr_OFinsert("  Time to build nodes     : %7.3f sec\n", timecache[ITC_NODESBUILD]);
	else
		pr_OFinsert("  Last node build time    : %7.3f sec\n", stat_nodestime);

	if (!preanalysed)
	{
		pr_OFinsert("  Trace analysis          : %7.3f sec\n", stat_analysetime);
		if (!R3D_edit)
			pr_OFinsert("  Calc reverb sectors     : %7.3f sec\n", timecache[ITC_SECREVERB]);
	}
	else
	{
		pr_OFinsert("  Last trace analysis     : %7.3f sec\n", stat_analysetime);
	}

	if (timecache[ITC_MODLOAD] > 0.1)
	{
		pr_OFinsert("  Load %4i models        : %7.3f sec\n", ge_inf.models_count, timecache[ITC_MODLOAD]);
		pr_OFinsert("  Model skin index size   : %7i\n", ge_inf.skincount);
		timecache[ITC_MODLOAD] = 0;	// This is only set once on startup
	}

	if (timecache[ITC_CACHE1] < 0) timecache[ITC_CACHE1] = 0;
	if (timecache[ITC_CACHE2] < 0) timecache[ITC_CACHE2] = 0;
	pr_OFinsert("  Cache textures/flats    : %7.3f sec\n", timecache[ITC_CACHE1]);
	pr_OFinsert("  Cache spr/model skins   : %7.3f sec\n", timecache[ITC_CACHE2]);
	{
		void Sfx_PreCache(void);
		Sfx_PreCache();
	}
	pr_OFinsert("\n");

	if (udirGame)
		Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  UDIR name   : %s\n", udirGame);
	if (ma_inf.doombuilder)
		Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  Map authored with Doombuilder\n");
	Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  Current wad : %s\n", show_wadname);
	Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  Current map : %s%s\n", R3D_ScriptedMap ? "R3D scripted " : "", mapinfo->id);
	Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  Map name    : %s\n", mapinfo->mapname);
	if (mapinfo->author[0])
		Con_playFPrintf(true, CBLF_LIGHT|CBLF_YELLOW, "  Map author  : %s\n", mapinfo->author);

	pr_OFrepeat(NULL, "*", 45, 2);
}

static mobj_t *checkFAKESS(int lnum)
{
	mobj_t	*it;
	line_t	*ld = &lines[lnum];
	int		s;

	for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
	{
		// Check for presence of *ONE* MT_FAKESOUNDSOURCE Thing
		for (it = sectors[s].thinglist; it; it = it->snext)
		{
			if (   it->thinker.function == (void (__cdecl *)(void))P_MobjThinker
				&& it->type == MT_FAKESOUNDSOURCE)	// 15551
			{
				ld->frontsector->specFlags2 |= SF2_FAKEMOVSECTOR;
				return it;
			}
		}
	}

	return NULL;
}

static void set_FAKESOUNDSOURCE()
{
	int i;

	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].special == 242
			&& lines[i].frontsector->specFlags & SF_MOVFLOOR
			&& sides[lines[i].sidenum[0]].lsiflags & LS_FAKESOUND)
		{
			checkFAKESS(i);
		}
		sides[lines[i].sidenum[0]].lsiflags &= ~LS_FAKESOUND;
	}
}

void redirect_fakeSND()
{
	mobj_t	*it;
	int		i, k;

	if (ma_inf.hasfakewater)
	{
		for (i = 0; i < numsectors; i++)
		{
			if (sectors[i].specFlags2 & SF2_FAKEMOVSECTOR)
			{
				for (k = 0; k < sectors[i].linecount; k++)
				{
					if (   sectors[i].lines[k]->special == 242
						&& sectors[i].lines[k]->frontsector - sectors == i
						&& (it = checkFAKESS(sectors[i].lines[k] - lines)) )
					{
						sectors[i].soundorg.x = it->x;
						sectors[i].soundorg.y = it->y;
						break;
					}
				}
			}
		}
	}
}

// Mark 3D lines
static void mark3DLines()
{
	if (ma_inf.num3Dlines)
	{
		int i;

		ma_inf.num3Dlines = 0;

		for (i = 0; i < numlines; i++)
		{
			if (lines[i].specType >= R3D_40)
			{
				ma_inf.num3Dlines++;
				lineflags[i] |= LFG_3DTYPE;
				Setsecdone(lines[i].frontsector, 7008);
				if (lines[i].backsector && lines[i].backsector != lines[i].frontsector)
					Setsecdone(lines[i].backsector , 7009);
			}
		}
	}

	if (R3D_slopetypes)
	{
		int i, k;

		for (i = 0; i < numsectors; i++)
		{
			if (sectors[i].stptr >= 0)
			{
				for (k = 0; k < sectors[i].linecount; k++)
				{
					lineflags[sectors[i].lines[k] - lines] |= LFG_3DTYPE;
				}
			}
		}
	}
}

// v220-7	subfn_chkOuterSecInside(), Look for missing lofloor loops
//          with the outer sector set as single sided lines inside.
static void subfn_chkOuterSecComplies()
{
	sector_t	*s;
	line_t		**ld; 
	int			c, i, k, m, lnum, box[4], rc;

	//---------------------------------

	ld = (line_t**)malloc(numsegs * sizeof(line_t*));

	for (i = 0; i < numsectors; i++)
	{
		RDBOOL	repeat = true;

		if (!countsec[i]) continue;

		s = &sectors[i];

		while (repeat)
		{
			int		ldcount = 0;

			repeat = false;

			lnum = -1; s = &sectors[i];

			for (c = k = 0; k < s->linecount; k++)
			{
				if (hitline[s->lines[k] - lines] == 1)
				{
					lnum = s->lines[k] - lines;
					linearray[c++] = lnum;
				}
			}

			if (c > 2)
			{
				if (   subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
					&& rc > 2
					&& LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE))
				{
					sector_t *os = NULL, *chk;

					for (m = 0; m < c; m++)
					{
						if (hitline[linearray[m]] == 4)
						{
							chk = lines[linearray[m]].frontsector == s ? lines[linearray[m]].backsector : lines[linearray[m]].frontsector;

							if (!os) os = chk;
							else if (os != chk) { os = NULL; break; }
						}
					}

					if (LS_HasOsecSSLinesInSSECarray(s, os, hitline, ld, &ldcount))
					{
						for (m = 0; m < c; m++)
						{
							if (hitline[linearray[m]] == 4)
								hitline[linearray[m]] = 8;
						}

						for (m = 0; m < hitssecptr; m++)
						{
							Setr3dssfs(&subsectors[hitssec[m]], os - sectors, 7010);
							Setr3dsscs(&subsectors[hitssec[m]], i, 29001);
						}

						for (m = 0; m < ldcount; m++)
						{
							SetLFG(ld[m], LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 7010);
						}

						repeat = true;
					}
				}
			}
			else
			{
				break;
			}
		}
	}

	free(ld);
}

static void subfn_chkOuterSecInside()
{
	int c, i;

	CLR_hitline();
	SET_countsec(0);

	for (c = i = 0; i < numlines; i++)
	{
		if (   lines[i].ml_flags & ML_TWOSIDED
			&& !lineflags[i]
			&& lines[i].lnflags2 == 0x0C000000
			&& !(origTEX[i] & 0x01)
			&& lines[i].frontsector->floorheight < lines[i].backsector->floorheight
			&& (lines[i].frontsector->specFlags & 0x1000C) == SF_NOLOTEX
			&& !(lines[i].backsector->specFlags & 0x1000C))
		{
			hitline[i] = 1; c++;
			countsec[lines[i].frontsector - sectors] = true;
		}
	}

	if (c > 3) subfn_chkOuterSecComplies();

	CLR_hitline();
	SET_countsec(0);

	for (c = i = 0; i < numlines; i++)
	{
		if (   lines[i].ml_flags & ML_TWOSIDED
			&& !lineflags[i]
			&& lines[i].lnflags2 == 0x0C000000
			&& !(origTEX[i] & 0x10)
			&& lines[i].frontsector->floorheight < lines[i].backsector->floorheight
			&& (lines[i].backsector->specFlags & 0x1000C) == SF_NOLOTEX
			&& !(lines[i].frontsector->specFlags & 0x1000C))
		{
			hitline[i] = 1; c++;
			countsec[lines[i].frontsector - sectors] = true;
		}
	}

	if (c > 3) subfn_chkOuterSecComplies();
}

static RDBOOL onlySSlines(subsector_t *ss, sector_t *ch)
{
	seg_t		*sg = &segs[ss->firstline];
	int			i, c;

	for (c = i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector == ch) continue;
		if (sg->frontsector != sg->backsector) return false;
		if (sg->linedef && origTEX[sg->linedef - lines] & 0x22) c++;
	}

	return c != ss->linecount;
}

// v220-16 check where a ssec references a textureless move sector
//         thus requiring the rendering sector to be calculated
//         including those that have no means of doing so directly
static void subfn_chkMovefloorRenderingSec()
{
	subsector_t	*ss;
	sector_t	*sc, *ch;
	seg_t		*sg;
	int			i, k, m, n, lnum;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (sc->ceilingpic != skyflatnum) continue;	// v220-20
		if ((sc->specFlags & 0x3040C) != 0x1000C) continue;

		for (k = 0; k < sc->linecount; k++)
		{
			lnum = sc->lines[k] - lines;

			if (lxarray[lnum] & 4)
			{
				ssline_t *sl = &sslines[lnum];

				ch = NULL;

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];

					if (ss->sector == sc && ss->floorsec < 0)
					{
						sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if (sg->frontsector == sc) continue;
							if (sg->frontsector == sg->backsector) continue;
							if (!ch) ch = sg->frontsector;
							else if (ch != sg->frontsector) { ch = NULL; break; }
						}

						if (ch)
						{
							SetOverride(ss, ch - sectors, 7012);
						}
					}
				}
			}
		}
	}
}

// v220-18 check for sectors with skyfloors/skyceilings and missingtex lines
static void subfn_checkSkyReplace()
{
	sector_t *fs, *bs;
	line_t	 *ld;
	int		c, i, k;

	SET_countsec(0);

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].specFlags & SF_MOVSECTOR) continue;

		if (sectors[i].floorpic == skyflatnum && sectors[i].ceilingpic != skyflatnum)
		{
			countsec[i] = 1;
		}
		else
		if (sectors[i].ceilingpic == skyflatnum && sectors[i].floorpic != skyflatnum)
		{
			countsec[i] = 2;
		}
	}

	for (i = 0; i < numsectors; i++)
	{
		if (!countsec[i]) continue;

		fs = &sectors[i];

		for (k = 0; k < fs->linecount; k++)
		{
			ld = fs->lines[k];

			if (ld->ml_flags & ML_TWOSIDED && !origTEX[ld - lines])
			{
				bs = ld->frontsector == fs ? ld->backsector : ld->frontsector;

				if (fs == bs) continue;
				if (fs->ceilingheight != bs->ceilingheight) continue;

				c = bs - sectors;

				if (   fs->floorpic != skyflatnum
					&& countsec[c] == 1
					&& countsec[i] == 2
					&& fs->floorheight != bs->floorheight)
				{
					fs->floorpic = skyflatnum;
				}
				else
				if (   fs->ceilingpic != skyflatnum
					&& countsec[c] == 2
					&& countsec[i] == 1
					&& fs->floorheight != bs->floorheight)
				{
					fs->ceilingpic = skyflatnum;
				}
			}
		}
	}
}

void R3D_MainSetupInit(int thinglumpnum, RDBOOL analyse)
{
	// NOTE : order of following routines must not be changed

	ma_inf.emptysec = LS_ClearEmptySectorSpecials();

	LS_Init_Neg_3D_tags();				// Must precede LS_InitTagLists
	LS_InitTagLists();

	if (analyse)
	{
		LS_ChkShortSegAssignment();
		LS_IdentifyMovsecList();		// Sets sector move type identifiers
		LS_InitVtxarray();
		LS_InitLXarray();
		LS_SetLinesclosedFlag();
	}

	LS_CheckDoorType();
	LS_SetLineTaggedSectors();			// Must precede LS_Process3DLines()
	LS_Process3DLines(analyse);			// Sets up 3D lines
	LS_SetTiltOffsets(analyse);			// Sets up slopes

	P_LoadThings(thinglumpnum);			// Must precede set_FAKESOUNDSOURCE()
	P_DealPlayerStarts();

	// v220-8
	if (ma_inf.unspawned)
	{
		char msg[512];

		sprintf(msg, "PROBLEM: %i unrecognised THINGS.\n", ma_inf.unspawned);

		if (ma_inf.modifier & MODF_ITE)
		{
			strcat(msg, "This wad was primarily authored for use with ZDoom\n");
			pr_OFinsert(msg);
		}
		else if (R3D_ScriptedMap)
		{
			pr_OFinsert(msg);
		}
		else
		{
#include "sys_system.h"
			if (R3D_maptype || R3D_lightextns)
				strcat(msg, "R3D specific wad: map authoring faults found.");
			else
				strcat(msg, "This wad may have been authored for ZDoom only.");

			if (ma_inf.unspawned > 4)	// $ZDdetect
			{
				Sys_MessageBox(msg, true);
				Sys_Quit(false);
			}

			if (!Sys_CriticalMessage(msg, true))
				Sys_Quit(false);
		}
	}

	if (ma_inf.dogsunspawned)
		pr_OFinsert("MBF dogs not supported: %i were found in the map\n", ma_inf.dogsunspawned);

	set_FAKESOUNDSOURCE();
}
void LS_SetLineTaggedSectors2();

void R3D_MAINAnalyseMap(int thinglumpnum)
{
	pr_OFdirect("=== Analysing level:\n");

	Con_InitProgress("ANALYSING MAP...", 110);	// $progress XH
	Con_Progress(5, 0);							// $progress XH 1

	markSwitchLines();

	R3D_MAINsetarrays();
	
	R3D_MainSetupInit(thinglumpnum, true);

	texwarn  = ArgCheck("-texwarn");
	texerror = verbose || R3D_edit || ArgCheck("-texerror");

	timecache[ITC_ANALYSE] = Sys_GetSeconds();

	LS_DeleteUnwantedHiTextures();		// v220-10
	LG_CheckMovTexUnset();				// v220-8
	LS_CheckUnmergedVtxs();				// b211-0 moved here
	LS_SetPASSUSEflags();
	LS_SetMultiLoopFlag();

	Con_Progress(5, 0);					// $progress XH 2

	LS_CheckFloorCeilHeightError();		// Jun 2004 rev06

	LS_InitVtflags();					// b211-0 moved here

	MX_RemoveUnwantedTexures();			// v220-10

	LS_SetLineMTFlags();
	LS_SetSectorsClosedFlag();

	ma_inf.taggederr = LS_CheckTaggedErrors();

	LS_FixSubsecErrorsWithShortSegs();
	TP_CheckTPSpawnedSectors(false);	// Results needed by TL_TraceBoundary()
	TL_TraceBoundary();

	// v220-35 deals speifically with line specials 271 and 272
	LS_SetLineTaggedSectors2();			// must not precede TL_TraceBoundary

	LS_SetDualMovSecFlag();
	LS_SetHiLoTexFlags();				// Results by LS_CheckBaseSSECsectors()
	subfn_checkSkyReplace();			// v220-18
	LS_CheckBaseSSECsectors();

	LS_SetLinesSpanning1sLines();		// v220-10
	LS_SetSameSecInGroupFlag();			// rev170f

	mark3DLines();

	LS_AddTexDefinitions();				// v220-8 e.g. 1's lines with no midtex
	LS_SetBOUNDS();						// b211-0 Bounds can now be set
	LS_CheckSameSecOwner();				// v220-7 Must be done first

	LS_CheckInvalidLoTexFinal();		// v220-18

	subfn_chkOuterSecInside();			// v220-7
	WA_CheckAnalyseFile("R3D_WADanalyse", 1);	// v220-22
	LS_chkOrphanedMov();				// v220-22
	TM_FindFlatRingSectors();			// v220-10
	LS_InheritanceMovFloorLoop();		// v220-10 look for bounded movfloor sector containing no lotex sectors
	MC_CheckFlatRings();				// v220-16
	MC_FalseFloorInBox();				// v220-16
	MX_CheckInvalidPlats();				// v220-22
	MX_CheckInvalidDoors();				// v220-16
	MC_CheckDoubleLowMove();			// v220-16 look for floor/ceil inheritance with no hi-tex double mov floor sector construct
	MX_MovInside();						// v220-10 look for bounded static sector containing a textureless mov sector
	MC_Nested();						// v220-16
	MC_DoubleLoop();					// v220-16
	LS_SameSidedLineSectors();			// v220-16
	MC_CheckIfHeightsecInvalid();		// v220-16
	MX_CheckNoHiAndSameHeightLine();	// v220-22

	LS_ChkBoundedSector();				// rev170g/b211-0 Must be done second
	LS_CheckForRingLinePairs();			// b211-0
	MC_FixedMLoContained();				// v220-18
	LS_CheckDonut();					// b211-0 Check donut types
	TM2_CheckMissLoNestedInMissHi();	// v220-22 must precede LL_LineLoops
	LL_LineLoops();						// v220-22
	MX_Prelim();						// v220-9
	MX_CheckNoTexDualRings();			// v220-16 
	MX_CheckContained();				// v220-18
	MX_ReversedMissingContained();		// v220-18
	MX_CheckNoTexRings();				// v220-16 
	MX_CheckLowMoveInFalseCeiling();	// v220-16 
	MX_MultiHiLineRing();				// v220-10
	LS_CheckLineLoops();				// v220-6

	XS_setLineTable();					// b211-0

	LS_XlineHacks();					// v220-7	Moved here v220-10
	TR_MissingBoundTo1sLine();			// v220-10
	MX_MovMixed();						// v220-10
	MX_MissingTexLoopSSLines();			// v220-9

	LS_CheckDualRingMissingSectors();	// v220-7 Must be done fourth or fifth
	LS_CheckDualRingSectors();			// b211-0 Must be done fourth or fifth
	MX_HeightsecMissing();				// v220-9 Check missing with heightsec
	TM2_CheckLoLoopIn0xCsectors();		// v220-22 Must preced MX_Main
	MX_Main();							// v220-9 Check basic missing tex structures 
	MX_Multi();							// v220-10 Check hi missing tex structures referencing several sectors 

	// Check for textureless line inheritance
	NF_CheckInheritance();				// b211-0
	NF_CheckDualSectorTypeLineRings();	// v220-10
	// Look for a particular kind of sector with midtex lines and
	// missing lower tex's that if drawn would cover the midtex.
	NF_SpecialFX_MidtexRing();
	NF_CheckForSubset();				// v220-16
	TM2_CheckMisHiAsymetricMovCeil();	// v220-22 must precede NF_CheckRingInheritance()
	NF_CheckRingInheritance();			// b211-0 revised v220-18
	// Note: SharedVertexRing must follow RingInheritance
	NF_CheckSharedVertexRing();			// v220-10 (must not precede NF_CheckRingInheritance)
	NF_ContainerRings();				// b211-0

	// Look for container rings wth ctrl lines
	NF_ContainRingsCTRL();				// b211-0
	NF_EnclosedFalseFloors();			// b211-0

	LS_SetMiscStaticType();				// b220-0 Contained static ring sectors vexpRC3
	LS_CheckRingSectors();				// b211-0 Contained ring sectors
	LS_SetMiscMoveType();				// b211-0 Contained mov ring sectors
	LS_SetMiscLineType();				// b211-0 Contained miscellaneous lines
	LS_PresetLCSEC();					// v220-4

	NF_CheckInheritance2();				// v220-4
	MX_MissingContained();				// v220-10
	MX_CheckLoMissingLoops();			// v220-12
	MX_CheckLoMissingFloorMovLoops();	// v220-13
	MX_CheckMovCeilInside();			// v220-13
	MX_CheckMovFloorInside();			// v220-13
	MX_CheckMissingHiLoSkySectors();	// v220-13
	MX_CheckNestedNotexSectors();		// v220-14
	MX_CheckMoveNoUpperTexSectors();	// v220-16
	MX_CheckUnclosedMissingContained();	// v220-18
	MX_CheckHighMissingContained();		// v220-18
	MX_CheckPartialHiMissing();			// v220-19
	MX_CheckAllConnectedLowMissing();	// v220-22
	TR_MovFloorInside();				// v220-19
	LG_AnalyseMovTexUnset();			// v220-22 (must precede TM2_TraceHiLoops)
	TM2_TraceHiLoops();					// v220-22 (must precede TM_TraceMissing)

	TM_TraceMissing();					// b211-0

	Con_Progress(5, 0);					// $progress XH 6

	// Look for container rings where all the bound sector's lines form one ring
	NF_ContainRingsFINAL();				// b211-0
	NF_UnclosedSecGroupsContained();	// b211-0
	NF_UnclosedSecGroupsContained2();	// v220-13

	TR_MovFloorContained_X();			// v220-18
	LS_TraceNormal();					// rev170f

	subfn_chkMovefloorRenderingSec();	// v220-16

	LS_MovSkyLoops();					// b211-0
// v220-18 moved here 
	LS_PresetContainedSSLines();		// b211-0
	LS_MovLogicalSecInRenderSec();		// b211-0
	LS_MarkLoopsWithContainedLoops();	// rev170s
	LG_RemoveAashitty();

	Con_HideProgress(); // $progress XH

	LS_NestedSectorRings();
	LS_Find4RingSectorsCrossing();
	LS_CheckSupportLineRings();	// rev50 iterate.c

	Con_InitProgress("General analysis...", 100); // $progress XF

	LS_SetHASTEXTUREflags();
	LS_TraceContainedLinesInRing();	// Added Apr 2004
	NS_CheckSSLineFlag();			// b211-0 Added

	LS_SetMissingTexLineType();
	MX_CheckMovFloor();				// v220-10
	MX_ContainedBySSLineRing();		// v220-9

	if (ma_inf.eliminate_sectors) MX_CheckCojoinedMissing();	// v220-16

	LS_CheckMapErrors();
	LS_CheckLinesInFakedSecs();
	LG_UnclosedMovSectors();		// Deal with unclosed mov secs

	LS_SetMissLowtexMovSectorsContained();
	LS_CheckFakeSectorInAnimated();
	LS_Check_MovSpecial_X();
	LS_Check_MovSpecial_Valid();
	LS_SetTexAndSamesecFlags();
	LS_SetTexInSector();
	LS_CheckOpenStaticContained();

	Con_Progress(20, 0);			// $progress XF 1

	LS_CheckMissingHiClosedLoops();
	LS_CheckMissingLoClosedLoops();
	LS_CheckValidLFSector();
	LS_CheckFlatsWithR3D_1ALoops();
	LS_FindLinkedAnimSectors();
	LS_CheckRaisedFloors();
	LS_CheckFlatsFromOrigTEX();

	Con_Progress(20, 0);			// $progress XF 2
	
	{
		sector_t *sec;
		int		 i;

		for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		{
			if (subfn_SecClosed(sec, R3D_03))  sec->specFlags |= SF_03_CLOSED;
			if (subfn_SecClosed(sec, R3D_04))  sec->specFlags |= SF_04_CLOSED;
		}
	}

	LS_CheckR3D_04Lines();	// checks for related variant type R3D_0E
	LS_SetChainedLines();	// rev170s

	// Must be done prior to LS_CheckLinesR3D_01()
	NF_Check1sLinesBoundedLiquidFloors();

	LS_CheckLinesR3D_01();	// Core checking routine

	// Further specific line type checking
	LS_CheckSpecialAnimSecs();	// Closed rings of lines in anim secs mainly used in vile.wad maps
	LS_CheckR3D_06Lines();	// Unconnected R3D_04 lines
	LS_CheckR3D_1ELines();	// Check if can be assigned to whole sector.

	// Tidy up
	LS_CheckUnsetR3D_1D();	// R3D_1D lines are checked at various points and needs tidying
	LS_CheckR3D_22Lines();	// Sets specSec for R3D_22 lines;
	LS_SetTypeR3D_2ECrossing(); // Search for crossing lines and set type accordingly

	Con_Progress(20, 0);	// $progress XF 3

	LS_CheckR3D_03Ladders();
	LS_DelSpecSecWithUnconnectedLinesInAnimSec();
	LS_CheckR3D_1AforR3D_3F();
	LS_MiscLineCrossAndLineCoincident();
	LS_CheckSameR3D_00TexLinesWithCL();
	LS_CheckR3D_06InheritanceFromR3D_03();
	LS_CheckSecInSecCeiling();

//Finals go here
	LS_CheckMissingTexInherited();
	LS_CheckR3D_2ASectors();
	LS_CheckForR3D_04Doors();
	LS_CheckForR3D_04InFakedSectors();
	LS_CheckForR3D_30Lines();
	LS_CheckR3D_2F();
	LS_CheckTex7WithR3D_03();
	LS_ResetClosedFlagForRenderer();
	LS_CheckZeroHeightSector();

	LS_CheckR3D_3FContained();
	LS_CheckLinkedAnimSecs();	// Catches any that were missed
	LS_CheckMovCeilInAnim();

	// GLNodes checking routines
	ma_inf.duallinesec  = LS_CheckDualLineSectorChoice();
	ma_inf.orphaned     = LS_CheckOrphanedSectors();

	LS_XcheckMissingClosedInFaked();
	LS_CheckR3D_1ALines();

	LS_XcheckR3D_04defs();
	LS_XcheckInheritanceGeneralLines();
	LS_XcheckR3D_1Bdefs();

	subfn_CheckLinesCrossingR3D_04();
	LS_CheckOpenUntexturedContained();

	Con_Progress(20, 0);		// $progress XF 4

	TR_TrickChainedFloorMissing();

	// Check for line on line E.g Eternall MAP28 line 953
	LS_CheckIfLineOnLineNeedsProjection();
	LS_CheckSpecialLinkedAnimSectors();
	LS_SectorArbitrator();

	subfn_CheckUnsetR3D_1D();
	subfn_CheckTriggerLinesInCorrectSec();

	LS_CheckMovSecWithSingleLine();
	NF_CheckLinesForInheritance();	// b211-0

	if (!R3D_maptype) LS_PreCheck_Types();
	LS_ChkSupportAssign();
	LS_ChkLinesInSkySec();
	LS_ChkContained808Csectors();
	LS_ChkNestedUnconnectedLinesInDifferentSectors();
	LS_CheckReversedLineMovCeilSecs();
	LS_CheckConnected();
	LS_CheckMovLiqFloorInMovFloor();
	
	Con_Progress(20, 0);			// $progress XF 5

	LS_chkInherit();
	LS_CheckUnclosedHighTexSectors();

	LG_CheckFixedInMovSector();
	LG_CheckUnsetStepLinePlats();

	LS_CheckLinkedFloors();		// rev170f
	LS_CheckMovAnimContained();	// rb211-0

	//========================================================================

	subfn_PRESET();	// Start process of pre-setting ssec flags etc.
	
	//========================================================================

	LS_CheckUnclosedMoveCrossingSecs();	// rev170s

	// Build list of sectors needing z-smoothing values
	LS_BuildMovsecList();

	// postcheck routines
	LS_CheckLinesInWrongSectors();
	LS_CheckMovSecWithAASHITTYTexLines();
	LS_CheckOVF_MISLOTEXLINEReset();

	LS_CheckAllCoincidentLines();
	LS_CheckBacksector();

	// Set line type defs for midtex lines
	LS_Setlinetypes();

	Con_HideProgress(); // $progress XF

	LS_Consolidate();
	LS_CheckR3D_29Loops();
	LS_CheckR3D_04SpecialCase();
	LS_SetFakeMoveFlag();
	LS_CheckSpecSecSettings();
	LS_RTP_Hacks();

	timecache[ITC_ANALYSE] = Sys_GetSeconds() - timecache[ITC_ANALYSE];
}

// Set the line subsec owner lists
void R3D_Set_sslinesArray()
{
	subsector_t *ss;
	line_t		*ld;
	seg_t		*sg;
	int			i, k;

	if (sslines) return;	// Already set up

	sslines = (ssline_t*)M_Calloc(numlines * sizeof(ssline_t), 1);

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) != NULL)
			{
				ssline_t	*sl = &sslines[ld - lines];
				int			m, found = false;

				for (m = 0; m < sl->num; m++)
				{
					if (sl->list[m] == i)
					{
						found = true;
						break;
					}
				}

				if (found) continue;

				if (sl->num >= sl->maxnum)
				{
					sl->maxnum += 4;
					sl->list = (int*)M_Realloc(sl->list, (sl->maxnum) * sizeof(int), 35);
				}
				sl->list[sl->num++] = i;
			}
		}
	}
}

void FreeSSLines()
{
	if (sslines)
	{
		int i;

		for (i = 0; i < numlines; i++) M_Free(sslines[i].list);

		M_Free(sslines);
		sslines = NULL;
	}
}

static void R3D_MAINsetarrays()
{
	uint nl = numlines      & 1 ? numlines      + 1 : numlines;
	uint ns = numsectors    & 1 ? numsectors    + 1 : numsectors;
	uint nv = numvertexes   & 1 ? numvertexes   + 1 : numvertexes;
	uint sb = numsubsectors & 1 ? numsubsectors + 1 : numsubsectors;
	uint si = (ma_inf.maxseclinecount + 2);

	maxsecsz = (ma_inf.maxseclinecount + 2);

	R3D_Set_sslinesArray();

	ssinf_ownrcptr		= (int  *)calloc(si, sizeof(int));
	ssinf_otherlorcptr	= (int  *)calloc(si, sizeof(int));
	ssinf_otherlineptr	= (int  *)calloc(nl, sizeof(int));
	countsec			= (int  *)calloc(ns, sizeof(int));
	conarray			= (byte *)calloc(nl, sizeof(byte));
	excludeline			= (byte *)calloc(nl, sizeof(byte));
	hitline				= (byte *)calloc(nl, sizeof(byte));
	hitssec				= (int  *)calloc(sb, sizeof(int));
	lfsecarray			= (int  *)malloc(ns*sizeof(int));	memset(lfsecarray, -1, ns*sizeof(int));
	lgrpnum				= (int  *)malloc(nl*sizeof(int));	memset(lgrpnum,    -1, nl*sizeof(int));
	linearray			= (int  *)calloc((numlines + 1),sizeof(int));
	lineflags			= (short*)calloc(nl, sizeof(short));
	linetable			= (byte *)calloc(nl, sizeof(byte));
	lnsec				= (int  *)malloc(nl*sizeof(int));	memset(lnsec,	   -1, nl*sizeof(int));
	lxarray				= (byte *)calloc(nl, sizeof(byte));
	mtarray				= (byte *)calloc(nl, sizeof(byte));
	secdone				= (char *)calloc(ns, sizeof(char));
	secmarker			= (char *)calloc(ns, sizeof(char));
	sidelist			= (int  *)calloc((numsides + 1),sizeof(int));
	ssecarray			= (byte *)calloc(sb, sizeof(byte));
	vtxarray			= (char *)calloc(nv, sizeof(char));
	vtxcheck			= (byte *)calloc(nv, sizeof(byte));
	vtxflags			= (char *)calloc(nv, sizeof(char));
	vtxflags2			= (char *)calloc(nv, sizeof(char));
	secspflags			= (short*)calloc(ns, sizeof(short));
	tg_markline			= (char *)calloc(nl, sizeof(char));
	tg_marklinemem		= (char *)calloc(nl, sizeof(char));
	tg_specline			= (char *)calloc(nl, sizeof(char));
	tg_linedone			= (char *)calloc(nl, sizeof(char));
	tg_ssecflags		= (byte *)calloc(sb, sizeof(byte));
	secbbox				= (bounds_t*)calloc(ns, sizeof(bounds_t));
}

// This is called from p_setup.c
void R3D_MAINfreeArrays()
{
	free(ssinf_ownrcptr);		ssinf_ownrcptr	= NULL;
	free(ssinf_otherlorcptr);	ssinf_otherlorcptr = NULL;
	free(ssinf_otherlineptr);	ssinf_otherlineptr = NULL;
	free(countsec);				countsec = NULL;
	free(conarray);				conarray = NULL;
	free(excludeline);			excludeline = NULL;
	free(hitline);				hitline = NULL;
	free(hitssec);				hitssec = NULL;
	free(lfsecarray);			lfsecarray = NULL;
	free(lgrpnum);				lgrpnum = NULL;
	free(linearray);			linearray = NULL;
	free(lineflags);			lineflags = NULL;
	free(linetable);			linetable = NULL;
	free(lnsec);				lnsec = NULL;
	free(lxarray);				lxarray = NULL;
	free(mtarray);				mtarray = NULL;
	free(secdone);				secdone = NULL;
	free(secmarker);			secmarker = NULL;
	free(sidelist);				sidelist = NULL;
	free(ssecarray);			ssecarray = NULL;
	free(vtxarray);				vtxarray = NULL;
	free(vtxcheck);				vtxcheck = NULL;
	free(vtxflags);				vtxflags = NULL;
	free(vtxflags2);			vtxflags2 = NULL;
	free(secspflags);			secspflags = NULL;
	free(tg_markline);			tg_markline = NULL;
	free(tg_marklinemem);		tg_marklinemem = NULL;
	free(tg_specline);			tg_specline = NULL;
	free(tg_linedone);			tg_linedone = NULL;
	free(tg_ssecflags);			tg_ssecflags = NULL;
	free(secbbox);				secbbox = NULL;

	// Free arrays set in P_SetSidePtrArray
	if (texname3D) { free(texname3D); texname3D = NULL; }
}
