//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_SinSceil.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: analyses ceiling inheritance (old code)
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"

//-----------------------------------------------------------------------------


//==========================================================
//
// LS_CheckSecInSecCeiling
//
// Routine to deal with sectors in sectors that have a higher ceiling height
// than the sector they are in, causing rendering problems with the containing
// sector's ceiling. EXAMPLE - PAR 1 6 - sector 818

static RDBOOL ssecExclude(sector_t *sec, byte *sslist)
{
	int		i, k, c1 = 0, bbox[4];
	int		*templist, *lnarray;
	int		tempptr;
	int		lnptr = 0;
	byte	*mrkline;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->validcount == validcount) continue;

			ld->validcount = validcount;

			if (hitline[ld - lines] == 4) continue;

			if (origTEX[ld - lines] & 0x55)
			{
				if (ld->frontsector == ld->backsector)
				{
					return false;
				}

				linearray[lnptr++] = ld - lines;
			}

			c1++;
		}
	}

	if (!c1   ) return false;	// Nothing to do
	if (!lnptr) return true;	// Nothing to exclude

	tempptr  = hitssecptr;
	templist = (int *)M_Malloc(tempptr * sizeof(int), 53);
	lnarray  = (int *)M_Malloc(lnptr   * sizeof(int), 54);
	mrkline  = (byte*)M_Calloc(lnptr, 267);

	memcpy(templist, hitssec, tempptr * sizeof(int));
	memcpy(lnarray, linearray, lnptr  * sizeof(int));

	c1 = lnptr;

	for (i = 0; i < lnptr; i++)
	{
		int lnum = lnarray[i];

		if (!mrkline[i])
		{
			int fh1 = lines[lnum].frontsector->floorheight;
			int fh2 = lines[lnum].backsector->floorheight;
			int ch1 = lines[lnum].frontsector->ceilingheight;
			int ch2 = lines[lnum].backsector->ceilingheight;
			int mark;

			sector_t *s = NULL;

			if (fh1 < fh2 && ch1 >= ch2)
			{
				s = lines[lnum].backsector;
				mark = 1;
			}
			if (fh1 > fh2 && ch1 <= ch2)
			{
				s = lines[lnum].frontsector;
				mark = 1;
			}
			if (fh1 <= fh2 && ch1 > ch2)
			{
				s = lines[lnum].backsector;
				mark |= 2;
			}
			if (fh1 >= fh2 && ch1 < ch2)
			{
				s = lines[lnum].frontsector;
				mark |= 2;
			}

			if (!s) break;

			if (subfn_CheckLoopClosed(i, lnptr, lnarray, bbox, mrkline))
			{
				int lnaptr = 0;

				for (k = 0; k < lnptr; k++)
				{
					if (mrkline[k] == 1)
					{
						linearray[lnaptr++] = lnarray[k];
						mrkline[k] = 2;
					}
				}

				if (LS_MarkSubsecsInAreaGeneral(s, NULL, lnaptr, NULL, bbox, -1, MSIA_NONE))
				{
					c1 -= lnaptr;

					for (k = 0; k < hitssecptr; k++)
					{
						sslist[hitssec[k]] = mark;
					}
				}
				else
					break;
			}
		}
	}

	memcpy(hitssec, templist, tempptr * sizeof(int));
	hitssecptr = tempptr;

	M_Free(mrkline);
	M_Free(lnarray);
	M_Free(templist);

	return !c1;	// All lines formed rings
}

static void subfn_CheckSecInSecCeiling(int secnum)
{
	sector_t	*sec = &sectors[secnum];
	side_t		*sd1, *sd2;
	line_t		*ld;
	int			i, k, ring_count, outer[4], count = 0, ht = sec->ceilingheight;

	CLR_vtxcheck();
	CLR_hitline();

	for ( i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->frontsector == ld->backsector)
			return;

		if (Vtxs2(ld)) continue;

		// check lines are compliant (we're looking for outer lines
		// that will be defining the sector as having a lowered ceiling)
		sd1 = &sides[ld->sidenum[0]];
		sd2 = &sides[ld->sidenum[1]];

		if ((!sd2->texFlags &&
			 (sd1->texFlags == T_LOW ||
			  sd1->texFlags == T_HI))
			||
			(!sd1->texFlags &&
			 (sd2->texFlags == T_LOW ||
			  sd2->texFlags == T_HI)))
		{
			hitline[ld - lines] = 1;
			count++;

			if (ld->frontsector == sec)
			{
				vtxcheck[ld->v1 - vertexes] |= 1;
				vtxcheck[ld->v2 - vertexes] |= 2;
			}
			else
			{
				vtxcheck[ld->v1 - vertexes] |= 2;
				vtxcheck[ld->v2 - vertexes] |= 1;
			}
		}
	}

	// If count not 4 then wouldn't work with an MBF visplane ??
//	if (count != 4) return;

	if (count < 2) return;

	for ( i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (hitline[ld - lines])
		{
			if (vtxcheck[ld->v1 - vertexes] != 3 ||
				vtxcheck[ld->v2 - vertexes] != 3)
			{
				return;
			}
		}
	}

	for ( i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 1)
		{
			if (subfn_CheckForRing(hitline, sec, lnum, &ring_count, outer, NULL, NULL)
				&& ring_count == count
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, outer, -1, MSIA_NONE))
			{
				byte *sslist = (byte*)M_Calloc(numsubsectors, 268);

				if (ssecExclude(sec, sslist))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						if (!sslist[hitssec[k]])
						{
							SetOverride(ss, sec - sectors, 15001);
						}
						else
						{
							if (!(sslist[hitssec[k]] & 2))
							{
								Setr3dsscs(ss, sec - sectors, 15001);
							}
							if (!(sslist[hitssec[k]] & 1))
							{
								Setr3dssfs(ss, sec - sectors, 15001);
							}
						}
						ss->ssflags |= SSF_NOCHANGE;
					}

					Setsecdone3(sec, 15001);
				}

				M_Free(sslist);
			}
			break;
		}
	}
}

// Look for inheritance in a mov floor sector
void LS_CheckSecInSecCeiling()
{
	sector_t *sec;
	int i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i]
			|| !(sec->specFlags & SF_MOVFLOOR)
			|| sec->linecount < 4
			|| sec->ceilingheight - sec->floorheight < 8*FRACUNIT
			|| !(sec->specFlags & SF_HITEXTRUE)
			|| !(sec->specFlags & SF_SECCLOSED)
			|| sec->heightsec >= 0)
		{
			continue;
		}

		subfn_CheckSecInSecCeiling(i);
	}
}
//==========================================================
