//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_ThingSpecials.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 23 (v220-23) 10/Jan/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: deals with linked Risen3D core effects and activates
//              Risen3D script commands in response to game actions.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "doom/d_englsh.h"
#include "doom/p_local.h"
#include "r3d_MAINspec.h"
#include "r_things.h"
#include "sys_system.h"
#include "r3d_edit.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "s_main.h"
#include "sound_var.h"
#include "r3d_mapinfo.h"
#include "rend_model.h"
#include "doom\g_game.h"

// v220-35
#include "gl_draw.h"
#include "gl_font.h"
#include "r3d_openGL.h"

// MACROS ------------------------------------------------------------------

#define		MAX_LINECOUNT	 8
#define		MINTEXTDELAY	30

// TYPES -------------------------------------------------------------------

static struct sfxnoway_t {
	mobj_t	*usething;
	int		tics;

} thinker_sfxnoway;

static struct mapmusic_t {
	mobj_t	*mapthing;
	mobj_t	*lastmapthing;
	int		tics;

} thinker_mapmusic;

static 	line_t	surrogate;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL	P_SoundSight(mobj_t *t1, mobj_t *t2);
mobj_t	*getArrayMobj(int id);
void	P_SetState(mobj_t *mo, int statenum);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void runScript(mobj_t *usething, int t79num, t77info_t *inf77, line_t *ld, int type);

static void SpawnOcclude(int t79num, RDBOOL setblocked);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern short	*listarray[NUMSPOTS];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

uwater_t	thinker_uwater;

th7400_t	th7400[NUM7400SPOTS];
th7700_t	th7700[NUMSPOTS];
th7800_t	th7800[NUMSPOTS];
th7900_t	th7900[NUMSPOTS];
int			T78_disable[NUMSPOTS][7];

int			linkedcard[NUMCARDS][2];
byte		t79list[NUMSPOTS];
int			maxtxtdelay = 90;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// P_MobjLinkFlame
// Transfers owner values to linked model
//===========================================================================
void P_MobjLinkFlame (mobj_t* mobj)
{
	mobj_t *mo;

	// Update link model x, y, z
	if ( (mo = mobj->corelink) != NULL)
	{
		mobj->x = mo->x;
		mobj->y = mo->y;
		mobj->z = mobj->floorz = mobj->dropoffz = mo->z;
		mobj->ceilingz  = mo->ceilingz;
		mobj->subsector = mo->subsector;
	}
}

//===========================================================================
// linkModel
//  Hooks to spawn flame models (where the flame is orientated to the viewer)
//===========================================================================
static void linkModel(mobj_t *iter, int linkedtype)
{
	mobj_t	*mobj;
	state_t	*st;

	// Check if it's been changed to an unsupported type
	if (iter->info->deathstate)
	{
		sprites[iter->sprite].sprflags |= SPR_NOMODEL;
		return;
	}

	// Create flame model.
	iter->corelink = mobj = (mobj_t*)W_Calloc(sizeof(mobj_t), THINKID, 0);

	mobj->corelink = iter;
	mobj->type = linkedtype;
	mobj->info = &mobjinfo[linkedtype];
	mobj->radius = mobj->info->radius;
	mobj->height = mobj->info->height;

	// Only flags allowed
	mobj->flags  = MF_NOSECTOR|MF_NOBLOCKMAP|MF_LOCAL|MF_NOGRAVITY;
	mobj->flags2 = MF2_NOTLIT|MF2_FLAGSSET;

	if (iter->flags2 & MF2_CANSLOPE) mobj->flags2 |= MF2_CANSLOPE;

	mobj->health = mobj->info->spawnhealth;
	mobj->reactiontime = mobj->info->reactiontime;

	// Values left at 0: iden_nums, flags3D, startangle

	mobj->thinker.function = (void (__cdecl *)(void))P_MobjLinkFlame;

	P_AddFNThinker (&mobj->thinker);

	st = states + mobj->info->spawnstate;

	mobj->state  = st;
	mobj->tics   = st->tics;
	mobj->sprite = (spritenum_t)st->sprite;
	mobj->frame  = st->frame;
	mobj->torchtic = (short)(rand() % 35);
	mobj->skintic  = (short)(rand() % 35);

	P_MobjLinkFlame(mobj);
}

// v220-5 mark flames that are angled into walls
typedef struct {
	int		fbox[4];
	int		flamebase;
	mobj_t	*it;
} cfl_t;

static RDBOOL PIT_CheckWallLine (line_t* ld, cfl_t *c)
{
	int		sd;

	// Check that the circumference crosses the line
	if (   c->fbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| c->fbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| c->fbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| c->fbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(c->fbox, ld) != -1)
	{
		return true;
	}

	if ( (sd = P_PointOnLineSide(c->it->x, c->it->y, ld))
		&& !(ld->ml_flags & ML_TWOSIDED) )
	{
		return true;
	}

	// Check if the top of the candelabra is in a fixed wall
	if (   ld->ml_flags & ML_TWOSIDED
		&& !(origTEX[ld - lines] & 0x22)
		&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
		&& !(ld->backsector->specFlags & SF_MOVSECTOR))
	{
		fixed_t fht = MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight);
		fixed_t cht = MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight);

		if (fht < c->it->z + 32 * FRACUNIT && c->it->z + c->it->height < cht)
		{
			return true;
		}
	}

	c->it->angle = R_PointToAngle2(ld->dx, ld->dy, 0, 0) - ANG90;
	c->it->visangle = c->it->angle >> 16;

	return false;
}

static RDBOOL PIT_CheckFlameLine (line_t* ld, cfl_t *c)
{
	// Check that the circumference crosses the line
	if (   c->fbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| c->fbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| c->fbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| c->fbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(c->fbox, ld) != -1)
	{
		return true;
	}

	if (ld->ml_flags & ML_TWOSIDED)
		return MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight) <= c->flamebase;

	return false;
}

void TS_CheckModelLink(mobj_t *iter, int sprnum)
{
	cfl_t		c;
	int			xl, xh;
	int			yl, yh;
	int			bx, by;

	if (   !iter->corelink
		&& sprnum > 0 && sprnum < numsprites
		&& !(sprites[sprnum].sprflags & SPR_NOMODEL))
	{
		switch (iter->type)
		{
			// evil eye
			case MT_MISC38:
				linkModel(iter, MT_EVILEYE2);
				break;
			// tall torches
			case MT_MISC41:
				linkModel(iter, MT_TALLBLUEFLAME);
				break;
			case MT_MISC42:
				linkModel(iter, MT_TALLGREENFLAME);
				break;
			case MT_MISC43:
				linkModel(iter, MT_TALLREDFLAME);
				break;
			// short torches
			case MT_MISC44:
				linkModel(iter, MT_SMALLBLUEFLAME);
				break;
			case MT_MISC45:
				linkModel(iter, MT_SMALLGREENFLAME);
				break;
			case MT_MISC46:
				linkModel(iter, MT_SMALLREDFLAME);
				break;
			// small candle
			case MT_MISC49:
				linkModel(iter, MT_CANDLEFLAMES);
				break;
			// skull pile
			case MT_MISC73:
				linkModel(iter, MT_SKULLFLAME);
				break;
			// firecan
			case MT_MISC77:
				linkModel(iter, MT_YELLOWFLAMES);
				break;
			case MT_MISC50:	// v220-5 align candelabras to walls
			case MT_MISC52:	// v220-5 align hanging arms out to walls
			case MT_MISC56:	// v220-5 align hanging arms out to walls
				if (!R3D_ScriptedMap)
				{
					M_setBoxOffsets(c.fbox, iter->x, iter->y, 16*FRACUNIT);
					M_getBlockOffsets(c.fbox, 0, &xl, &xh, &yl, &yh);

					validcount++;
					c.it = iter;

					for (bx = xl ; bx <= xh ; bx++)
					{
						for (by = yl ; by <= yh ; by++)
						{
							if (!P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_CheckWallLine, &c))
							{
								break;
							}
						}
					}
				}
				break;
			default:
				break;
		}

		if (iter->corelink)
		{
			// v220-4 Check the distance of model flames from any walls.

			// The flame is normally tilted about 10 degrees to allow viewing from
			// any position.  This is fine unless against a wall where it then gets
			// clipped. The following test checks for this, setting a flag if true,
			// to instruct the renderer to pitch the flame vertically.

			fixed_t size = 0;

			switch (iter->type)
			{
				case MT_MISC41:
				case MT_MISC42:
				case MT_MISC43:
				case MT_MISC44:
				case MT_MISC45:
				case MT_MISC46:
					size = 10*FRACUNIT;
					break;
				case MT_MISC77:
					size = 12*FRACUNIT;
					break;
				default:
					break;
			}

			if (size)
			{
				validcount++;
				c.flamebase = iter->z + iter->height;

				M_setBoxOffsets(c.fbox, iter->x, iter->y, size);
				M_getBlockOffsets(c.fbox, 0, &xl, &xh, &yl, &yh);

				for (bx = xl ; bx <= xh ; bx++)
				{
					for (by = yl ; by <= yh ; by++)
					{
						if (!P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_CheckFlameLine, &c))
						{
							iter->corelink->misc |= MJF_FLAMEWALL;
							return;
						}
					}
				}
			}
		}
	}
}

void TS_CheckLinkedModels()
{
	thinker_t	*thinker = thinkercap.next;
	thinker_t	*next;
	mobj_t		*iter;
	int			sprnum;

	// Reset the map music thinker
	thinker_mapmusic.lastmapthing = NULL;
	thinker_mapmusic.mapthing = NULL;
	thinker_mapmusic.tics = -1;

	if (!loadModels) return;

	while(thinker != &thinkercap)
	{
		next = thinker->next;

		if (thinker->function != (void (__cdecl *)(void))P_MobjThinker)
		{
			thinker = next; continue;
		}

		iter = (mobj_t *) thinker;
		sprnum = iter->sprite;

		TS_CheckModelLink(iter, iter->sprite);
		thinker = next;
	}
}

//  End Linked Flame Models =================================================

//	SCRIPTED models =========================================================

// Messages ----------------------------------------------

// v220-4 support linefeeds in "" (\n)
int TS_ConPrintMsg(char *ptr, RDBOOL prmsg)
{
	char	msg[300];
	int		len ;
	int		c, i;
	float	x = 0.0f;

	if (!ptr || !ptr[0]) return 0;

	if ( (len = strlen(ptr)) > 298) len = 298;

	for (c = i = 0; i < len; i++)
	{
		if (ptr[i] == '\\' && ptr[i + 1] == 'n')
		{
			if (c)
			{
				msg[c] = 0;
				if (prmsg) Con_Message("%s\n", msg);
				x += (float)strlen(msg);
			}
			c = 0; i++;
		}
		else
		{
			msg[c++] = ptr[i];
		}	
	}

	if (c)
	{
		msg[c] = 0;
		if (prmsg) Con_Message("%s\n", msg);
		x += (float)strlen(msg);
	}

	if (!x) return 0;

	x = MAX_OF(MINTEXTDELAY, x * 1.5f);

	int delay =  (int)(x * cfg.text_delay);
	return delay < maxtxtdelay ? delay: maxtxtdelay;
}

void TS_CheckMessageQueue()
{
	if (scpMessage.qmsgcount && scpMessage.delay_to_fade <= 0)
	{
		int   d, i;

		scpMessage.scriptMessage = scpMessage.queuemsg[0];
		d = TS_ConPrintMsg(scpMessage.scriptMessage, true);

		scpMessage.delay_to_next_text = 0;
		scpMessage.delay_to_fade = d;		// v220-4 revised delay
		scpMessage.alpha = 1.0f;

		// Shuffle the messages down
		for (i = 1; i < scpMessage.qmsgcount; i++)
		{
			scpMessage.queuemsg[i - 1] = scpMessage.queuemsg[i];
		}

		scpMessage.qmsgcount--;
	}
}

void TS_SetMessage(th7900_t *t79, char *msg, RDBOOL print)
{
	int d;

	if (    scpMessage.delay_to_fade > 0
		&& 	(!t79 || scpMessage.smflags & SM_ENDLEVEL))
	{
		// Don't bother to queue a non t79 message
		// Don't add messages if level is about to end
		return;
	}

	if (t79)
	{
		if (   !(msg = t79->textmsg)
			|| !(t79->t79flags & T79_MESSAGESET) )
		{
			return;
		}

		if (t79->t79flags & T79_MESSAGEMANY)
			print = false;
		else
			t79->t79flags &= ~T79_MESSAGESET;

		if (print && scpMessage.delay_to_fade > 0)
		{
			if (scpMessage.qmsgcount < MAXQUEUEMSG)
				scpMessage.queuemsg[scpMessage.qmsgcount++] = msg;
			return;
		}
	}

	d = TS_ConPrintMsg(msg, print);

	scpMessage.scriptMessage = msg;
	scpMessage.delay_to_next_text = 0;
	scpMessage.delay_to_fade = d;
	scpMessage.alpha = 1.0f;
}

static void TS_SetGeneralMessage(char *msg)
{
	int d;

	if (!msg || !msg[0]) return;

	if (scpMessage.delay_to_fade > 0)
	{
		scpMessage.delay_to_fade = 0;
		if (scpMessage.qmsgcount < MAXQUEUEMSG)
			scpMessage.queuemsg[scpMessage.qmsgcount++] = msg;
		return;
	}

	d = TS_ConPrintMsg(msg, false);
	scpMessage.scriptMessage = msg;
	scpMessage.delay_to_next_text = 0;
	scpMessage.delay_to_fade = d;
	scpMessage.alpha = 1.0f;
}

void ext_LockedMessage(line_t *ld, char *msg)
{
extern int lockedsound;

	if (ld->backsector && msg)
	{
		if (   ld->backsector->specFlags & SF_MOVCEIL
			&& ld->backsector->ceilingheight - ld->backsector->floorheight <= FRACUNIT)
		{
				S_StartSound (lockedsound, ppt.viewplrmo);
				TS_SetGeneralMessage(msg);
		}
		else
		if (   ld->frontsector->specFlags & SF_MOVCEIL
			&& ld->frontsector->ceilingheight - ld->frontsector->floorheight <= FRACUNIT)
		{
				S_StartSound (lockedsound, ppt.viewplrmo);
				TS_SetGeneralMessage(msg);
		}
	}
}

// ---------- UTILITIES

// Only single keys are matched (not multiples) 
int matchKey(int val)
{
	switch (val)
	{
	case T79_PK_BCARD : return  4;
	case T79_PK_RCARD : return  7;
	case T79_PK_YCARD : return 12;
	case T79_PK_BSKULL: return  5;
	case T79_PK_RSKULL: return  9;
	case T79_PK_YSKULL: return 13;
	default: return 0;
	}
}

// Only single keys are matched (not multiples)
static RDBOOL clearKey(int val)
{
	mbf_player_t *pl = ppt.viewplrmo->player;
	int	idx;

	if (!pl || !(idx = matchKey(val)) ) return false;

	GL_ClearIgnoreFlag(idx);

	switch (val)
	{
	case T79_PK_BCARD : pl->cards[it_bluecard   ] = 0; break;
	case T79_PK_RCARD : pl->cards[it_redcard    ] = 0; break;
	case T79_PK_YCARD : pl->cards[it_yellowcard ] = 0; break;
	case T79_PK_BSKULL: pl->cards[it_blueskull  ] = 0; break;
	case T79_PK_RSKULL: pl->cards[it_redskull   ] = 0; break;
	case T79_PK_YSKULL: pl->cards[it_yellowskull] = 0; break;
	default: return false;
	}

	return true;
}

static RDBOOL checkKeys(th7900_t *t79, mbf_player_t *player, int *singlekey, int linenum)
{
	ushort		 keys = t79->t79keys & T79_ALLKEYMASK;
	RDBOOL		 notallcardkeys, notallskullkeys;
	RDBOOL		 printmsg;
	RDBOOL		 ret = true;

	*singlekey = 0;

	if (player && ma_inf.keyoverride == linenum) return true;	// backward compatibilty for CruelD map02
	if (!keys) return true;

	// Enemies can open doors if the player has the key
	if (!player)
	{
		mbf_player_t	*pl;
		ushort			nmyk = t79->t79keys & T79_ALLNMYKEYMASK;

		if (!nmyk) return true;
		
		if (!(pl = ppt.viewplrmo->player)) return false;	// belt & braces

		if (nmyk & T79_MK_BCARD  && !pl->cards[it_bluecard   ]) return false;
		if (nmyk & T79_MK_RCARD  && !pl->cards[it_redcard    ]) return false;
		if (nmyk & T79_MK_YCARD  && !pl->cards[it_yellowcard ]) return false;
		if (nmyk & T79_MK_BSKULL && !pl->cards[it_blueskull  ]) return false;
		if (nmyk & T79_MK_RSKULL && !pl->cards[it_redskull   ]) return false;
		if (nmyk & T79_MK_YSKULL && !pl->cards[it_yellowskull]) return false;

		return true;
	}

	notallcardkeys  =  !player->cards[it_bluecard   ]
					|| !player->cards[it_redcard    ]
					|| !player->cards[it_yellowcard ];

	notallskullkeys =  !player->cards[it_blueskull  ]
					|| !player->cards[it_redskull   ]
					|| !player->cards[it_yellowskull];

	printmsg = player && !t79->lockmsg;

	if (keys == T79_ALLKEYMASK)
	{
		if (notallcardkeys || notallskullkeys)
		{
			if (printmsg) P_SetMessage(player, PD_SIXKEYS);
			return false;
		}
		return true;
	}

	if (keys == T79_CARDKEYMASK)
	{
		if (notallcardkeys)
		{
			if (printmsg) P_SetMessage(player, PD_THREECARDKEYS);
			return false;
		}
		return true;
	}

	if (keys == T79_SKULLKEYMASK)
	{
		if (notallskullkeys)
		{
			if (printmsg) P_SetMessage(player, PD_THREESKULLKEYS);
			return false;
		}
		return true;
	}

	if (keys & T79_PK_BCARD)
	{
		if (!player->cards[it_bluecard])
		{
			if (printmsg) P_SetMessage(player, PD_BLUEKC);
			ret = false;
		}
		*singlekey |= T79_PK_BCARD;
	}
	if (keys & T79_PK_RCARD)
	{
		if (!player->cards[it_redcard])
		{
			if (printmsg) P_SetMessage(player, PD_REDKC);
			ret = false;
		}
		*singlekey |= T79_PK_RCARD;
	}
	if (keys & T79_PK_YCARD)
	{
		if (!player->cards[it_yellowcard])
		{
			if (printmsg) P_SetMessage(player, PD_YELLOWKC);
			ret = false;
		}
		*singlekey |= T79_PK_YCARD;
	}
	if (keys & T79_PK_BSKULL)
	{
		if (!player->cards[it_blueskull])
		{
			if (printmsg) P_SetMessage(player, PD_BLUESK);
			ret = false;
		}
		*singlekey |= T79_PK_BSKULL;
	}
	if (keys & T79_PK_RSKULL)
	{
		if (!player->cards[it_redskull])
		{
			if (printmsg) P_SetMessage(player, PD_REDSK);
			ret = false;
		}
		*singlekey |= T79_PK_RSKULL;
	}
	if (keys & T79_PK_YSKULL)
	{
		if (!player->cards[it_yellowskull])
		{
			if (printmsg) P_SetMessage(player, PD_YELLOWSK);
			ret = false;
		}
		*singlekey |= T79_PK_YSKULL;
	}

	return ret;
}

static RDBOOL doorClosed(mobj_t *mo)
{
	return mo->tics == -1 && mo->health > 0;
}

static void modelRESET(int t79num, RDBOOL checkblocked)
{
	th7900_t	*t79  = &th7900[t79num];
	mobj_t		*mo   = t79->mo;
	int			conum = t79->cotrigger;
	mobj_t		*cmo  = conum > 0 ? th7900[conum - 1].mo : NULL;
	mobjinfo_t	*info = mo->info;
	int			snd;
	RDBOOL		occludeset = false;
	
	if (checkblocked)
	{
		th7700_t *t77 = &th7700[t79num];

		for (int i = 0; i < t77->spawn_count; i++)
		{
			if (t77->info[i].options & MTF_EASY)
			{
				if (cmo && !doorClosed(cmo))
				{
					SpawnOcclude(t79num, true);
					occludeset = true;
					break;
				}
				else
				{
					lines[t77->info[i].t77lnum].lnflags2 |= LNF2_BLOCKALL;
					t77->blocked = true;
				}
			}
		}
	}

	if (info->raisestate)
	{
		P_SetMobjState (mo, info->raisestate);

		if (   mo->flags2 & MF2_OCCLUDE
			&& !occludeset
			&& th7700[t79num].spawn_count)	// Not a cotriggered pair
		{
			SpawnOcclude(t79num, false);
		}
	}
	else
	{
		P_SetMobjState (mo, info->spawnstate);
	}

	mo->flags  = info->flags;
	mo->health = info->spawnhealth;
	mo->target = NULL;  // just in case..
	mo->corpsetics = 0;

	// Check for an ONRESET sound
	if ( (snd = mo->info->attacksound) > 0)
	{
		if (mo->modelboxnum)
		{	// sounds that can be suspended only apply to modelboxes
			S_StartSoundAtVolume(snd|HSF_CANSUSPEND, mo, 1.0f, DEF_ATTN);
		}
		else
		{
			t79->dmo1.x = mo->x;
			t79->dmo1.y = mo->y;
			t79->dmo1.z = mo->z + mo->height/2;

			S_StopAllSounds((mobj_t*)&t79->dmo1);
			S_StartSoundAtVolume(snd, (mobj_t*)&t79->dmo1, 1.0f,
						snd & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
		}
	}
}

static RDBOOL checkLineMove(mobj_t *mo, line_t *ld)
{
	if (PointIsOnLine(mo->x, mo->y, ld, mo->radius>>16))
		return true;

	if (!mo->momx && !mo->momy) return false;

	return PointIsOnLine(mo->x + mo->momx, mo->y + mo->momy, ld, mo->radius>>16);
}

static RDBOOL checkBlockedClear(th7700_t *t77)
{
	mobj_t	*mobj[64];		// should be enough
	int		mobj_count = 0;
	int		i, k;

	for (i = 0; i < t77->spawn_count; i++)
	{
		if (t77->info[i].options & MTF_EASY)
		{
			line_t		*ld = &lines[t77->info[i].t77lnum];
			sector_t	*sc = ld->frontsector;

			for (k = 0; k < 2; k++)
			{
				mobj_t *mo = sc->thinglist;

				for ( ; mo; mo = mo->snext)
				{
					int type = (mo->flags & (MF_SOLID|MF_SHOOTABLE)) == (MF_SOLID|MF_SHOOTABLE);

					if (   (type || mo->flags & MF_DROPPED)
						&& !(mo->flags3 & MF3_AMODEL)
						&& checkLineMove(mo, ld))
					{
						if (type)
						{
							return false;
						}
						else
						{
							if (mobj_count < 64) mobj[mobj_count++] = mo;
						}
					}
				}
				if (ld->ml_flags & ML_TWOSIDED && ld->backsector != ld->frontsector)
					sc = ld->backsector;
				else
					break;
			}
		}
	}

	for (i = 0; i < mobj_count; i++) P_RemoveMobj(mobj[i], 0);

	return true;
}

// ---------- THINKERS

// b211-0 sfx_noway and map music static thinkers
// v220-3 Player underwater bubble thinker
//        Note: not configured for multiplayer (if ever it's supported)
// v220-4 Inhibited bubbles and sound but not damage under anything other than water
void TS_ResetWater()
{
	memset(&thinker_uwater, 0, sizeof(uwater_t));
	thinker_uwater.tuwflags = TUWF_RESET;
}

void T_Underwater()
{
	RDBOOL plr_dead = (plhealth.health <= 0);
	RDBOOL notvalid = !underwater || !waterdynamics;

	if (!ppt.thinkers_have_run) return;

	// v220-4 eliminate falling in water when dead turning on air supply
	if (   thinker_uwater.tuwflags & TUWF_RESET
		&& (notvalid || plr_dead || R3D_edit))
	{
		return;
	}

	if (   notvalid
		|| ppt.viewer_fake_sector < 0
		|| !cfg.underwaterDamage
		|| plr_dead
		|| R3D_edit) 
	{
		if (!plr_dead)
		{
			if (thinker_uwater.breath_tics > 160)
				S_LocalSound(r3ds.breath_id, ppt.viewplrmo);

			ppt.viewplrmo->flags2 &= ~MF2_NOAIRSUPPLY;
		}

		TS_ResetWater();
		return;
	}

	if (thinker_uwater.tuwflags & TUWF_RESET)
	{
		thinker_uwater.tuwflags &= ~TUWF_RESET;
		thinker_uwater.tics = 30;

		if (   sectors[ppt.viewer_fake_sector].specFlags2 & SF2_HASWATERFLATS
			|| ppt.viewplrmo->subsector->sector->specFlags2 & SF2_ADJOINSWATER)
		{
			thinker_uwater.tuwflags |= TUWF_WATERTEX;
		}
	}

	if (thinker_uwater.tics <= 0)
		thinker_uwater.tics = 140;
	else
		thinker_uwater.tics--;

	thinker_uwater.breath_tics++;

	if (thinker_uwater.tics < 110)
	{
		thinker_uwater.tuwflags &= ~TUWF_HASBUBBLES;

		if (thinker_uwater.tuwflags & TUWF_WATERTEX)
			thinker_uwater.tuwflags |= TUWF_SOUNDPLAY;
	}
	else if (thinker_uwater.breath_tics < UWMAXTICS)
	{
		if (thinker_uwater.tuwflags & TUWF_WATERTEX)
			thinker_uwater.tuwflags |= TUWF_HASBUBBLES;

		if (thinker_uwater.tuwflags & TUWF_SOUNDPLAY)
		{	
			S_StartSound(r3ds.uwater_id, ppt.viewplrmo);
			thinker_uwater.tuwflags &= ~TUWF_SOUNDPLAY;
		}
	}
}

// Underwater damage with depleted air supply
void TS_PlayerInWaterSector(mobj_t *mo)
{
	if (thinker_uwater.breath_tics > UWMAXTICS)
	{
		if (   !(mbf_players[0].cheats & CF_GODMODE)
			&& !mbf_players[0].powers[pw_invulnerability])
		{
			mo->flags2 |= MF2_NOAIRSUPPLY;

			if (!(leveltime&0x3f))
			{
				if (thinker_uwater.tuwflags & TUWF_WATERTEX)
					S_StartSound(r3ds.drown_id|HSF_NO_ATTENUATION, NULL);

				if (mo->health > 20)
					P_DamageMobj(mo, NULL, NULL, 20, 0);
				else
					P_DamageMobj(mo, NULL, NULL, mo->health, 1);
			}
		}

		thinker_uwater.breath_tics = UWMAXTICS;
	}
}

//================================ end water

void T_SFXNoWay()
{
	if (thinker_sfxnoway.tics < 0)
		return;

	if (thinker_sfxnoway.tics > 0)
	{
		thinker_sfxnoway.tics--;
	}
	else if (!thinker_sfxnoway.tics)
	{
		S_StartSound (sfx_noway, thinker_sfxnoway.usething);
		thinker_sfxnoway.tics = -1;
	}
}

void SetSFXnoway(mobj_t *usething, int tics)
{
	if (tics < 0)
	{
		thinker_sfxnoway.tics = tics;
	}
	else if (usething && thinker_sfxnoway.tics < 0)
	{
		if (thinker_sfxnoway.tics < -1)
		{
			thinker_sfxnoway.tics++;
		}
		else
		{
			thinker_sfxnoway.usething = usething;
			thinker_sfxnoway.tics = tics;
		}
	}
}

void T_MAPMusic()
{
	if (   thinker_mapmusic.tics < 0
		|| !thinker_mapmusic.mapthing)
	{
		return;
	}

	if (thinker_mapmusic.tics > 0)
	{
		thinker_mapmusic.tics--;
	}
	else if (!thinker_mapmusic.tics
		&& thinker_mapmusic.lastmapthing != thinker_mapmusic.mapthing)
	{
		int arraypt = TIDNUM(thinker_mapmusic.mapthing);

		if (arraypt >= 0 && arraypt < MAX_MUS_ENTRIES)
		{
			int lumpnum = music_array[arraypt];

			if (lumpnum >= 0 && lumpnum < numlumps)
			{
				// global_looped may have been set to 0 by LINEMUSIC
				global_looped = !(mapinfo->R3Dflags & R3Dmusiconce);
				Music_PlayLump(lumpnum);
			}
		}

		thinker_mapmusic.tics = -1;
	}
}

void resetMapMusicThinkerIden()
{
	thinker_mapmusic.mapthing =NULL;
}

void SetMAPMusic(mobj_t *mapthing)
{
	if (   thinker_mapmusic.mapthing != mapthing
		&& ppt.view_subsector
		&& (!thinker_mapmusic.mapthing
		|| TIDNUM(thinker_mapmusic.mapthing) != TIDNUM(mapthing))
		&& mapthing->subsector->sector == ppt.view_subsector->sector)
	{
		thinker_mapmusic.lastmapthing = thinker_mapmusic.mapthing;
		thinker_mapmusic.mapthing = mapthing;
		thinker_mapmusic.tics = 30;
	}
}

//----------------- End fixed "thinkers"

// Thinker verification
static RDBOOL thingExists(th7900_t *t79)
{
	sector_t	*sec;
	mobj_t		*mo = t79->mo;
	mobj_t		*th;

	if (!mo || mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker) return false;

	sec = (R_PointInSubsector(t79->x<<FRACBITS,t79->y<<FRACBITS))->sector;

	for (th = sec->thinglist; th; th = th->snext)
		if (th == mo) return true;

	return false;
}

// Entry point for map thing trigger

void TS_CheckLinkSpecial(mobj_t *sp)
{
	int i;

	for (i = 0; i < NUMSPOTS; i++)
	{
		if (th7900[i].pickupmo == sp)
		{
			modelkill_t *mk = (modelkill_t*)W_Calloc(sizeof(modelkill_t), THINKID, 0);

			mk->tnum  = i;
			mk->delay = th7900[i].delay;

			mk->thinker.function = (void (__cdecl *)(void))T_ModelKill;

			P_AddFNThinker(&mk->thinker);
			break;
		}
	}
}

void TS_CheckLinkKey(card_t card)
{
	int t79num;

	if (  (t79num = linkedcard[card][0]) >= 0
		&& thingExists(&th7900[t79num]))
	{
		modelkill_t *mk = (modelkill_t*)W_Calloc(sizeof(modelkill_t), THINKID, 0);

		mk->tnum     = t79num;
		mk->delay    = linkedcard[card][1];

		mk->thinker.function = (void (__cdecl *)(void))T_ModelKill;

		P_AddFNThinker(&mk->thinker);
	}
}



// scripted dynamic thinkers

byte msgchangedarray[NUMSPOTS];

void T_ModelKill(modelkill_t *mk)
{
	if (mk->delay > 0) { mk->delay--; return; }

	if (thingExists(&th7900[mk->tnum]))
	{
		if (th7900[mk->tnum].changemsgnum > 0)
		{
			extern char* linetext[];
			th7900[th7900[mk->tnum].witht79num].textmsg = linetext[th7900[mk->tnum].changemsgnum - 1];
			msgchangedarray[mk->tnum] = true;
		}

		if (mk->occupied)
		{
			mobj_t *mo = th7900[mk->tnum].mo;

			int *bbox = (mo->flags3 & MF3_INTERREACT ?
				MB_GetBox(mo->modelboxnum - 1, NULL) : NULL);

			validcount++;

			mo->validcount = validcount;

			if (spotIsOccupied(mo->x, mo->y, mo->z, mo->radius, bbox, 0))
			{
				return;
			}
		}

		runScript(NULL, mk->tnum, NULL, NULL, 0);
	}

	P_RemoveThinker(&mk->thinker);
}

void T_ModelSpawn(modelspawn_t *ms)
{
	th7800_t *t78 = &th7800[ms->t78num];

	if (   ms->spawn_count >= t78->spawn_count
		|| (t78->setflags & T_UNSET && !(t78->setflags & T_ENABLE)) )
	{
		P_RemoveThinker(&ms->thinker);
	}
	else if (t78->setflags & T_UNSET)
	{
		return;
	}
	else if (ms->delay_tic > 0)
	{
		ms->delay_tic--;
	}
	else
	{
		int monum;

		if (!ms->spawn_sp.set)
		{
			int	c = ms->spawn_count;

			memcpy(&ms->spawn_sp, t78, sizeof(mapthing_t));

			ms->spawn_sp.type		= t78->spawn[c][0];
			ms->spawn_sp.z_off		= t78->spawn[c][1];
			ms->delay_tic			= t78->spawn[c][2];
			ms->spawn_sp.spawn_fog	= t78->spawn[c][3];
			ms->spawn_sp.z_nofall	= (t78->options & (MTF_EASY|MTF_NORMAL|MTF_HARD)) == MTF_NORMAL;	// v220-10

			if (t78->spawn[c][4])
				ms->spawn_sp.sspflags |= SSPF_NOFRIENDCOUNT;

			if (   nomonsters
				&& (ms->spawn_sp.type == MT_SKULL
				||  mobjinfo[ms->spawn_sp.type].flags & MF_COUNTKILL))
			{
				ms->spawn_sp.set = false;
				ms->spawn_count++;
			}
			else
				ms->spawn_sp.set = true;
		}
		else if (P_SpawnScriptThing(&ms->spawn_sp, &monum, get78N(ms->t78num), true, false))
		{
			mobj_t *mo = getArrayMobj(monum);

			ms->spawn_count++;
			ms->spawn_sp.set = false;

			if (mo)
			{
				if ( !(mo->flags3 & MF3_AMODEL)
					&& mo->flags & MF_SHOOTABLE
					&& mo->info->seestate
					&& ppt.viewplrmo)
				{
					// Anything spawned immediately starts looking for the console player 
					if (P_CheckSight(ppt.viewplrmo, mo))
						P_SetTarget(&mo->target, ppt.viewplrmo);
					else
						P_SetTarget(&mo->subsector->sector->soundtarget, ppt.viewplrmo);
				}
				
				// v220-4 Don't add carried over friends to total
				if (mo->flags3 & MF3_FRIEND && ms->spawn_sp.sspflags & SSPF_NOFRIENDCOUNT)
					G_DeductFriendStats();
			}
		}
		else
			ms->delay_tic = 12;	// Wait at least 12 tics before trying again
	}
}

void T_ModelReset(modelreset_t *mr)
{
	if (mr->delay-- <= 0)
	{
		modelRESET(mr->t79num, true);

		if (th7700[mr->t79num].info)	// v220-31 may not have been set
			lines[th7700[mr->t79num].info->t77lnum].special = th7700[mr->t79num].info->t77special;

		P_RemoveThinker(&mr->thinker);
	}
}

void T_SwitchReset(modelreset_t *mr)
{
	int activesec;

	if (mr->delay > 0)	// small delay to allow actions to get started
	{
		mr->delay--;
		return;
	}

	activesec = th7900[mr->t79num].activesec;

	if (activesec < 0 || (!sectors[activesec].floordata && !sectors[activesec].ceilingdata))
	{
		mobj_t *mo = th7900[mr->t79num].mo;

		if (mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
		{
			line_t	*li = &lines[mr->lnum];

			li->special = mr->special;
			li->tag     = 0;

			modelRESET(mr->t79num, false);
		}

		P_RemoveThinker(&mr->thinker);
	}
	else if (!mr->delay)
	{
		mr->delay--;

		if (lines[mr->lnum].linesound)
		{
			S_StartSoundAtVolume(lines[mr->lnum].linesound, (mobj_t*)&sectors[activesec].soundorg, 1.0f,
					lines[mr->lnum].linesound & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
		}
	}

}

// Note: t79->soundid replicated in ss->soundid
void T_SoundActivate(scriptedsound_t *ss)
{
	if (!ss->delay--)
	{
		if (ss->t79id > 0 && ss->soundid)	// purely defensive (race hazard maybe)
		{
			S_StartSoundAtVolume(ss->soundid, th7900[ss->t79id - 1].mo, 1.0f,
					ss->soundid & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
		}

		ss->delay = 0x7;	// reset the delay
	}
}

void T_DelayUnblock(delayline_t *dl)
{
	if (dl->tics-- < 0)
	{
		int i, num = dl->t77num;

		for (i = 0; i < th7700[num].spawn_count; i++)
		{
			// v220-3 modified to allow either ML_BLOCKING
			//        or LNF2_BLOCKALL to be set.
			line_t *ld = &lines[th7700[num].info[i].t77lnum];

			if (th7700[num].info[i].options & MTF_EASY)
			{
				if (th7700[num].info[i].options & MTF_ALIVE)
					ld->ml_flags &= ~ML_BLOCKING;
				else
					ld->lnflags2 &= ~LNF2_BLOCKALL;

				th7700[num].blocked = false;
			}
		}

		P_RemoveThinker(&dl->thinker);
	}
}

void T_DelayResurrect(modelreset_t *mr)
{
	th7700_t *t77;
	th7900_t *t79 = &th7900[mr->t79num];
	mobj_t	 *mo  = t79->mo;

	if (!mo || mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker)
	{
		P_RemoveThinker(&mr->thinker);
		return;
	}

	if (mr->delay2)
	{
		if (mo->tics != -1) return;

		t79->t79flags &= ~T79_DELAYACTIVE;
		P_RemoveThinker(&mr->thinker);
		return;
	}

	if (mr->delay > 0)
	{
		if (th7900[mr->t79num].mo->tics == -1)
			mr->delay--;
		
		return;
	}

	if (!(leveltime&7)) return;

	t77 = &th7700[mr->t79num];

	if (checkBlockedClear(t77))
	{
		if (mo && mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
		{
			int i, lnum = mr->lnum;

			modelRESET(mr->t79num, true);

			if (mr->lnum >= 0 && lines[mr->lnum].linesound)
			{
				S_StartSound (lines[mr->lnum].linesound, mo);
			}

			for (i = 0; i < t77->spawn_count; i++)
			{
				if (!(t77->info[i].options & MTF_NORMAL))
				{
					lines[t77->info[i].t77lnum].special = t77->info[i].t77special;	// Reset any line triggers
				}
			}
			mr->delay2 = true;
		}
	}
}

void T_StateMonitor(modelreset_t *mr)
{
	mobj_t	 *mo = th7900[mr->t79num].mo;

	if (!mo || mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker)
	{
		P_RemoveThinker(&mr->thinker);
		return;
	}

	if (mo->tics != -1) return;

	th7900[mr->t79num].t79flags &= ~T79_DELAYACTIVE;
	P_RemoveThinker(&mr->thinker);
}

void T_Endlevel(delayline_t *dl)
{
	if (scpMessage.smflags & SM_ENDSET)
	{
		if (scpMessage.delay_to_fade > 0 || scpMessage.qmsgcount)	// Wait for messages to clear
			return;

		scpMessage.smflags &= ~SM_ENDSET;
		scpMessage.scriptMessage = scpMessage.endlevelMessage;
		scpMessage.delay_to_next_text = scpMessage.endlevel_start_delay;
		scpMessage.delay_to_fade = scpMessage.endlevel_delay_to_fade;
		scpMessage.alpha = 1.0f;
		if ( !(scpMessage.smflags & SM_DONTENDLEVEL) )
			scpMessage.smflags |= SM_ENDLEVEL;
	}

	if (   scpMessage.delay_to_next_text <= 0
		&& !(scpMessage.smflags & SM_NOSPAWN)
		&& scpMessage.smflags & SM_MASSACRE)
	{
		int oldcount = mbf_players[0].killcount;

		scpMessage.smflags |= SM_NOSPAWN;
		P_Massacre(true);
		mbf_players[0].killcount = oldcount;
	}
	
	if (scpMessage.delay_to_fade <= 0)
	{
		P_RemoveThinker(&dl->thinker);

		if (scpMessage.smflags & SM_ENDLEVEL)
			G_ExitLevel();
	}
}

void T_PlayListsound(delayline_t *dl)
{
	if (!dl->tics--)
	{
		mobj_t	 *pl;
		th7900_t *t79 = &th7900[dl->t79num];
		int		 num  = t79->soundlist[0];

		if (!t79->mo || t79->mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker)
		{
			P_RemoveThinker(&dl->thinker);
			return;
		}

		if (t79->mo->health <= 0) 
		{
			if (dl->t77num < 0xFFFF)
				S_StopSound(dl->t77num, t79->mo);
			
			P_RemoveThinker(&dl->thinker);
			return;
		}

		if (   t79->mo->tics != t79->mo->state->tics
			|| t79->mo->state - states != t79->soundlist[5])
		{
			dl->tics++;
			return;
		}

		if (dl->t77num < 0xFFFF)
			S_StopSound(dl->t77num, t79->mo);
		
		dl->t77num = t79->soundlist[1];	// t77num here will actually be the sound_id
		
		if (num > 1) dl->t77num = t79->soundlist[ (1 + (rand() % num) ) ];

		dl->tics = t79->soundlist[6] + (rand() % t79->soundlist[7]);

		if ( (pl = ppt.viewplrmo) )
		{
			float a = FIX2FLT(t79->soundlist[9]);	// attenuation
			float d = P_MobjSoundDistancef(t79->mo, NULL, a);

			if (   d > SOUND_MAX_DISTANCE
				|| (t79->mo->flags3 & MF3_AMODEL && !P_SoundSight(t79->mo, pl)) )
			{
				return;
			}

			S_StartSoundAtVolume(dl->t77num|(a ? 0 : HSF_NO_ATTENUATION),
				t79->mo, FIX2FLT(t79->soundlist[8]), a);
		}
	}
}

void T_LineOcclude(lineocclude_t *lo)
{
	int		conum = th7900[lo->t79num].cotrigger;
	mobj_t	*mo = th7900[lo->t79num].mo;
	mobj_t	*cmo = conum > 0 ? th7900[conum - 1].mo : NULL;

	if (doorClosed(mo) && !(cmo && !doorClosed(cmo)))
	{
		if (mo->flags2 & MF2_OCCLUDE)
		{
			lines[mo->lineocclude].lnflags2 |= LNF2_OCCLUDE;
		}

		if (lo->block)
		{
			th7700_t *t77 = &th7700[lo->t79num];
			int		 i;

			for (i = 0; i < t77->spawn_count; i++)
			{
				if (t77->info[i].options & MTF_EASY)
				{
					if (t77->info[i].options & MTF_ALIVE)
						lines[t77->info[i].t77lnum].ml_flags |= ML_BLOCKING;
					else
						lines[t77->info[i].t77lnum].lnflags2 |= LNF2_BLOCKALL;

					t77->blocked = true;
				}
			}
		}

		P_RemoveThinker(&lo->thinker);
	}
}

// v220-4
void T_KillDelayed(delayline_t *dl)
{
	int c, d, i;

	if (--dl->tics >= 0) return;

	dl->tics = 3;

	for (c = i = 0; i < dl->t77num; i++)	// t77num was the original redir77[dl->t79num].mocount
	{
		thinker_t	*th;
		int			n = redir77[dl->t79num].monum[i];

		for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
		{
			mobj_t	*iter = (mobj_t*)th;

			if (   iter->thinker.function == (void (__cdecl *)(void))P_MobjThinker
				&& iter->health > 0
				&& iter->flags & MF_COUNTKILL
				&& MAPNUM(iter) == n)
			{
				d = rand() % 20;

				d *= iter->info->spawnhealth/100;

				if (d < 8) d = 8;

				P_DamageMobj(iter, NULL, NULL, d, 2);

				if (iter->health > 0) c++;
			}
		}
	}

	if (!c) P_RemoveThinker(&dl->thinker);	// Everything's been killed
}

void T_ActivateDelayed(modelreset_t *mr)
{
	line_t *ld = &lines[mr->lnum];

	if (ld->frontsector->ceilingdata || ld->frontsector->floordata) return;

	if (--mr->delay < 0)
	{
		// Always assume it's the player
		subfn_UseSpecialLine(ppt.viewplrmo, ld, 10);
		P_RemoveThinker(&mr->thinker);
	}
}

// Ceiling raise/lower functions =============================================

static RDBOOL ceilFunctionLower(adjustplat_t *ac, th7900_t *t79, sector_t *s)
{
	s->ceilingheight += ac->increment;

	if (s->ceilingheight <= t79->startheight)
	{
		s->ceilingheight = t79->startheight;
		t79->mo->z = t79->startheight;
		return true;
	}

	t79->mo->z = s->ceilingheight;

	if (P_ChangeSector(s, NULL, (t79->t79flags & T79_VDOORCRUSH) > 0))
	{
		if (t79->t79flags & T79_VDOORREVERSE)
			ac->increment = ac->inc_up;
		else if (t79->t79flags & T79_VDOORCRUSH)
			s->ceilingheight -= ac->increment;
		else	// in stasis
		{
			Sfx_SuspendChannel(t79->soundid, t79->mo, true);
			s->ceilingheight -= ac->increment;
			t79->mo->z = s->ceilingheight;
		}
	}

	return false;
}

static RDBOOL ceilFunctionRaise(adjustplat_t *ac, th7900_t *t79, sector_t *s)
{
	RDBOOL ret = false;

	s->ceilingheight += ac->increment;

	if (s->ceilingheight > t79->startheight + t79->offset*FRACUNIT)
	{
		s->ceilingheight = t79->startheight + t79->offset*FRACUNIT;
		ret = true;
	}

	t79->mo->z = s->ceilingheight;
	return ret;
}

static void ceilFunctionReset(adjustplat_t *ac, th7900_t *t79, sector_t *s)
{
	th7700_t	*t77 = &th7700[ac->t79num];
	mobj_t		*mo = t79->mo;
	mobjinfo_t	*info = mo->info;
	int			i;

	if (t79->wait)
	{
		for (i = 0; i < t77->spawn_count; i++)
			lines[t77->info[i].t77lnum].special = t77->info[i].t77special;

		t79->t79flags |= T79_VDOOR;

		P_SetMobjState (mo, info->spawnstate);
		mo->flags  = info->flags;
		mo->health = info->spawnhealth;
		mo->target = NULL;  // just in case..
		mo->corpsetics = 0;
	}
}

void T_AdjustCeil(adjustplat_t *ac)
{
	th7900_t	*t79 = &th7900[ac->t79num];
	sector_t	*s = &sectors[ac->sector];
	RDBOOL		removeThinker = false;

	if (ac->active)
	{
		if (!ac->increment)	// Shouldn't happen
			removeThinker = true;
		else if (ac->increment < 0)
		{
			if (ceilFunctionLower(ac, t79, s))
			{
				ceilFunctionReset(ac, t79, s);
				removeThinker = true;
			}
		}
		else if (ceilFunctionRaise(ac, t79, s))
		{
			if ( (ac->wait = t79->wait) )
				ac->active = false;
			else
				removeThinker = true;
		}
	}
	else if (ac->wait > 0)
	{
		if (ac->wait == 1)
		{
			int off = t79->offset*FRACUNIT;

			if (t79->soundid)
			{
				S_StartSoundAtVolume(t79->soundid, t79->mo, 1.0f,
					t79->soundid & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
			}

			// Ensure ceiling's at expected height
			s->ceilingheight = t79->startheight + t79->offset*FRACUNIT;

			ac->increment = ac->inc_down;
			ac->active = true;
		}

		ac->wait--;
	}
	else
		removeThinker = true;

	if (removeThinker) P_RemoveThinker(&ac->thinker);
}

int check_called[7];	// Debug check for multiple thinker calls (should always be 0)

// ---------- SPAWNERS

static RDBOOL checkUnspawned(int t79num, void* type)
{
#ifdef _DEBUG
	thinker_t *th;

	for (th = fnthinkcap.next; th != &fnthinkcap; th = th->next)
	{
		if ((void*)th->function == type)
		{
			if (th->function == (void (__cdecl *)(void))T_ModelReset)
			{
				if ( ((modelreset_t*)th)->t79num == t79num) { check_called[0]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_SwitchReset)
			{
				if ( ((modelreset_t*)th)->t79num == t79num) { check_called[1]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_DelayResurrect)
			{
				if ( ((modelreset_t*)th)->t79num == t79num) { check_called[2]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_StateMonitor)
			{
				if ( ((modelreset_t*)th)->t79num == t79num) { check_called[3]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_SoundActivate)
			{
				if ( ((scriptedsound_t*)th)->t79id - 1 == t79num) { check_called[4]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_DelayUnblock)
			{
				if ( ((delayline_t*)th)->t79num == t79num) { check_called[5]++; return false; }
			}
			else if (th->function == (void (__cdecl *)(void))T_LineOcclude)
			{
				if ( ((lineocclude_t*)th)->t79num == t79num) { check_called[6]++; return false; }
			}
		}
	}
#endif
	return true;
}

static RDBOOL checkUnspawned_t74(int lnum)
{
	thinker_t *th;

	for (th = fnthinkcap.next; th != &fnthinkcap; th = th->next)
	{
		if (   (void*)th->function == T_ActivateDelayed
			&& ((modelreset_t*)th)->lnum == lnum)
		{
			return false;
		}
	}
	return true;
}
// v220-31
static void spawnActive74Delay(int t74num)
{
	int lnum = th7400[t74num].refline - lines;

	if (checkUnspawned_t74(lnum))
	{
		modelreset_t *mr = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
		mr->delay = th7400[t74num].delay;
		mr->lnum = lnum;
		mr->thinker.function = (void (__cdecl *)(void))T_ActivateDelayed;

		P_AddFNThinker(&mr->thinker);
	}
}

static void SpawnOcclude(int t79num, RDBOOL setblocked)
{
	if (checkUnspawned(t79num, T_LineOcclude))
	{
		lineocclude_t *lo = (lineocclude_t*)W_Calloc(sizeof(lineocclude_t), THINKID, 0);

		lo->t79num = t79num;
		lo->block = setblocked;
		lo->thinker.function = (void (__cdecl *)(void))T_LineOcclude;
		P_AddFNThinker(&lo->thinker);
	}
}

static void SpawnModelKill(int num, th7900_t *t79)
{
	short tnum     = t79->killthing[num][0];
	short delay    = t79->killthing[num][1];
	short occupied = t79->killthing[num][2];

	if (delay || occupied)
	{
		modelkill_t *mk;

		mk = (modelkill_t*)W_Calloc(sizeof(modelkill_t), THINKID, 0);

		mk->tnum     = tnum;
		mk->delay    = delay;
		mk->occupied = occupied;
		mk->thinker.function = (void (__cdecl *)(void))T_ModelKill;

		P_AddFNThinker(&mk->thinker);
		return;
	}

	runScript(NULL, tnum, NULL, NULL, 0);
}

static void SpawnModelThinker(short t78num)
{
	if (!(th7800[t78num].setflags & T_UNSET))
	{
		modelspawn_t *ms = (modelspawn_t*)W_Calloc(sizeof(modelspawn_t), THINKID, 0);

		ms->t78num = t78num;
		ms->spawn_count = 0;
		ms->delay_tic = 0;
		ms->thinker.function = (void (__cdecl *)(void))T_ModelSpawn;

		P_AddFNThinker(&ms->thinker);
	}
}

static void SpawnResetThinker(int t79num)
{
	if (checkUnspawned(t79num, T_ModelReset))
	{
		modelreset_t *mr;

		mr = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);

		mr->t79num  = t79num;
		mr->delay   = th7900[t79num].resetdelay;
		mr->thinker.function = (void (__cdecl *)(void))T_ModelReset;

		P_AddFNThinker(&mr->thinker);
	}
}

static void SpawnSwitchResetThinker(int t79num, int lnum)
{
	if (checkUnspawned(t79num, T_SwitchReset))
	{
		modelreset_t *mr = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);

		mr->t79num  = t79num;
		mr->lnum    = lnum;
		mr->special = 8064;
		mr->delay   = 3;
		mr->thinker.function = (void (__cdecl *)(void))T_SwitchReset;

		P_AddFNThinker(&mr->thinker);
	}
}

static scriptedsound_t *SpawnSoundThinker(int t79num, int soundid)
{
	if (checkUnspawned(t79num, T_SoundActivate))
	{
		scriptedsound_t *ss = (scriptedsound_t*)W_Calloc(sizeof(scriptedsound_t), THINKID, 0);

		ss->soundid = soundid;
		ss->delay = 0x7;
		ss->t79id = t79num + 1;
		ss->thinker.function = (void (__cdecl *)(void))T_SoundActivate;

		P_AddFNThinker(&ss->thinker);

		return ss;
	}
	return th7900[t79num].ss;
}

static void SpawnCeilingThinker(int t79num)
{
	th7900_t	 *t79 = &th7900[t79num];
	adjustplat_t *ac  = (adjustplat_t*)W_Calloc(sizeof(adjustplat_t), THINKID, 0);

	ac->t79num    = t79num;
	ac->sector    = t79->mo->subsector->sector - sectors;
	ac->inc_up    = t79->offset*FRACUNIT;
	if (t79->doortics)
		ac->inc_up /= t79->doortics;
	ac->increment = ac->inc_up;
	ac->active    = true;

	ac->inc_down = -t79->offset*FRACUNIT;
	if (t79->falltics)
		ac->inc_down = ac->inc_down/t79->falltics;

	ac->thinker.function = (void (__cdecl *)(void))T_AdjustCeil;

	P_AddFNThinker(&ac->thinker);
}

static void SpawnUnblockThinker(th7700_t *t77, int t79num)
{
	if (th7900[t79num].delay <= 0)
	{
		int i;

		for (i = 0; i < t77->spawn_count; i++)
		{
			// v220-3 modified to allow either ML_BLOCKING
			//        or LNF2_BLOCKALL to be set.
			if (t77->info[i].options & MTF_EASY)
			{
				if (t77->info[i].options & MTF_ALIVE)
					lines[t77->info[i].t77lnum].ml_flags &= ~ML_BLOCKING;
				else
					lines[t77->info[i].t77lnum].lnflags2 &= ~LNF2_BLOCKALL;

				t77->blocked = false;
			}
		}
	}
	else
	{
		if (checkUnspawned(t79num, T_DelayUnblock))
		{
			delayline_t *dl= (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);

			dl->thinker.function = (void (__cdecl *)(void))T_DelayUnblock;
			dl->t79num = t79num;
			dl->tics   = th7900[t79num].delay;
			dl->t77num = t77 - th7700;
			P_AddFNThinker(&dl->thinker);
		}
	}
}

static void SpawnResurrectThinker(int t79num, int lnum)
{
	if (checkUnspawned(t79num, T_DelayResurrect))
	{
		modelreset_t *mr = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);

		th7900[t79num].t79flags |= T79_DELAYACTIVE;

		mr->t79num  = t79num;
		mr->lnum    = lnum;
		mr->delay   = th7900[t79num].resurrect;
		mr->delay2  = false;
		mr->thinker.function = (void (__cdecl *)(void))T_DelayResurrect;

		P_AddFNThinker(&mr->thinker);
	}
}

static void SpawnStateMonitor(int t79num)
{
	if (checkUnspawned(t79num, T_StateMonitor))
	{
		modelreset_t *mr = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);

		th7900[t79num].t79flags |= T79_DELAYACTIVE;

		mr->t79num  = t79num;
		mr->thinker.function = (void (__cdecl *)(void))T_StateMonitor;

		P_AddFNThinker(&mr->thinker);
	}
}

static void spawnKillThinker(int t79num)
{
	if (redir77[t79num].mocount)
	{
		delayline_t *dl = (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);

		dl->t79num = t79num;
		dl->t77num = redir77[t79num].mocount;
		dl->thinker.function = (void (__cdecl *)(void))T_KillDelayed;
		P_AddFNThinker(&dl->thinker);

		// Prevent a recall.  If a map is saved after the thinker's been
		// removed then it could be called again (if the circumstances
		// still allow) but this won't matter as the things will either
		// not be present or dead meaning the thinker will only run once
		// before removing itself (i.e. the redir77 array is not saved).
		redir77[t79num].mocount = 0;
	}
}

static void SpawnEndlevelThinker()
{
	delayline_t *dl= (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);

	dl->thinker.function = (void (__cdecl *)(void))T_Endlevel;

	P_AddFNThinker(&dl->thinker);
}

// ---------- soundlists

void TS_StartListSounds()
{
	th7900_t *t79 = th7900;
	int		 i, n;

	for (i = 0; i < 200; i++, t79++)
	{
		if ( (n = t79->soundlist[0]) )
		{
			delayline_t *dl= (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);

			dl->thinker.function = (void (__cdecl *)(void))T_PlayListsound;

			dl->t77num = 0xFFFF;	// Actually sound id. 0xFFFF means ignore
			dl->t79num = i;
			dl->tics = 20 + rand() % 70;

			P_AddFNThinker(&dl->thinker);
		}
	}
}

// ---------- player messages	// v220-35

#define SZSCPMSG	128
#define SZSCPNUM	16
#define PUTIMERTICS	105		// 3 seconds
int		pu_timer;
int		pu_counter;
char	pu_message[SZSCPNUM][SZSCPMSG];

//	Scripted pickups

// Add message to stack. Note single player only (pl not checked).
void TS_SCPmessage(mbf_player_t *pl, char *msg)
{
	if (pl - mbf_players == 0 && *msg)
	{
		static int	msglength[SZSCPMSG];

		pu_timer = PUTIMERTICS;	// reset, display

		int i, len = strlen(msg);

		if (len >= SZSCPMSG) len = SZSCPMSG - 1;

		if (pu_counter >= SZSCPNUM)
		{
			pu_counter = SZSCPNUM - 1;

			// lose the oldest message in the stack
			for (i = 0; i < pu_counter; i++)
			{
				strcpy(pu_message[i], pu_message[i + 1]);
				msglength[i] = msglength[i + 1];
			}
		}

		for (i = 0; i < pu_counter; i++)
		{
			if (   msglength[i] == len
				&& !strncmp(pu_message[i], msg, len))
			{
				// check if at end of list
				if (i == pu_counter - 1) return;

				// first remove from list to retain sequence of
				// adding as it could contain different items
				for (int k = i; k < pu_counter; k++)
				{
					strcpy(pu_message[k], pu_message[k + 1]);
					msglength[k] = msglength[k + 1];
				}

				--pu_counter;

				break;
			}
		}

		memcpy(pu_message[pu_counter], msg, len);
		msglength[pu_counter] = len;
		pu_message[pu_counter++][len] = 0;
	}
}

static int pu_yoff;

void TS_PUdisplay()
{
	extern int currentFont;

	if (pu_counter)
	{
		int i, y = 20;
		int idx = currentFont;

		FR_SetFont(1);
		int ch = FR_TextHeight("W");

		// Go into screen projection mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, current_w, current_h, 0, -1, 1);
		glColor4f(1.f, 1.f, .0f, .7f);

		for (i = 0; i < pu_counter; i++)
		{
			FR_TextOutOffset(pu_message[i], 20, float(y + pu_yoff));
			y += (ch + 2);
		}

		if (automap_active) FR_SetFont(0);

		// Restore original matrix.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		if (idx != currentFont) FR_SetFont(idx);
	}
}

// called each tic
void TS_PUticker()
{
	if (pu_counter)
	{
		if (pu_timer > 0)
		{
			if (--pu_timer < 5) pu_yoff -= 3;
		}
		else
		{
			// timed out, shuffle messages down the stack
			for (int i = 0; i < pu_counter; i++)
			{
				int len = strlen(pu_message[i]);
				strcpy(pu_message[i], pu_message[i + 1]);
			}
			
			pu_counter--;
			pu_timer = PUTIMERTICS;	// reset, display
			pu_yoff = 0;
		}
	}
}

// ---------- MAIN SCRIPT runtime code

static void disableMomonstersError(int progerror, int repspot, int reptype)
{
	if (nomonsters)
		Con_Error(	"SCRIPT FAILED\n"
					"CAUSE : nomonsters is set\n"
					"FIX : DISABLE nomonsters before relaunching");

	// If -nomonsters was not enabled then probably caused by a script error
	Con_Error("Program error [%i] spot %i type %i: Probable cause Script error", progerror, repspot, reptype);
}

static void checkOndeathSpawn(th7900_t *t79)
{
	if ( !(t79->t79flags & T79_SPAWNDONE) )
	{
		int i;

		for (i = 0; i < t79->spawndef_num; i++)
			SpawnModelThinker(t79->spawn78nn[i]);
	}

	if (t79->t79flags & T79_SPAWNONCE)
		t79->t79flags |= T79_SPAWNDONE;
}

static void runScript(mobj_t *usething, int t79num, t77info_t *inf77, line_t *ld, int type)
{
	th7700_t	*t77 = &th7700[t79num];
	th7900_t	*t79 = &th7900[t79num];
	th7900_t	*t79other;
	mobj_t		*use = usething ? usething : ppt.viewplrmo;
	mobj_t		*mo = t79->mo;
	int			i, n, health;

	SetSFXnoway(NULL, -2);	// Kill any pending sfx_noway sound

	if (!mo || mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker)
	{
		t79->mo = NULL;	// In case was not cleared

		if (ld)
		{
			checkOndeathSpawn(t79);
			ld->special = 0;
		}

		return;
	}

	if (!t79list[t79num])
	{
		t79list[t79num] = true;

		for (i = 0; i < t79->killcount; i++)
		{
			int t79nextnum = t79->killthing[i][0];

			if (thingExists(&th7900[t79nextnum]))
			{
				if (th7900[t79nextnum].resurrect)
				{
					runScript(usething, t79nextnum, NULL, NULL, 0);
				}
				else
				{
					SpawnModelKill(i, t79);
				}
			}
		}
	}

	if ( (n = t79->replacemocount) > 0)
	{
		for (i = 0; i < n; i++)
		{
			thinker_t	*th;
			int			oldtype  = t79->replacemo[i][0];
			int			newtype  = t79->replacemo[i][1];
			RDBOOL		isfriend = t79->replacemo[i][2];
			mobjinfo_t	*info = &mobjinfo[newtype];

			//----------------------------------

			for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
			{
				mobj_t *mo = (mobj_t*)th;

				if (   mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker
					&& mo->type == oldtype
					&& mo->health > 0)
				{
					RDBOOL wasfriend = (mo->flags3 & MF3_FRIEND ) != 0;
					RDBOOL wascnkill = (mo->flags & MF_COUNTKILL) != 0;

					// if different then swap types
					if (oldtype != newtype)
					{
						mo->type = newtype;
						mo->info = info;
						mo->radius = info->radius;
						mo->height = info->height;
						mo->flags  = info->flags;
						mo->flags2 = info->flags2 & MF2_LOADMASK;
						mo->flags3 = info->flags3 & MF3_LOADMASK;
						mo->health = info->spawnhealth;
						mo->reactiontime = info->reactiontime;
						mo->iden_nums = (mo->iden_nums & 0xFFFF0000)|mo->type;

						P_SetDDFlags(mo);
						P_SetState(mo, info->seestate);	// Do not use P_SetMobjState as action routines cannot be called yet
					}

					if (isfriend) mo->flags3 |= MF3_FRIEND;

					if (isfriend ^ wasfriend)
					{
						P_SetTarget(&mo->target,    NULL);
						P_SetTarget(&mo->tracer,    NULL);
						P_SetTarget(&mo->lastenemy, NULL);

						P_UpdateThinker(&mo->thinker);	// Transfer to correct list

						if (wascnkill)
						{
							if ( wasfriend) totalfriends--;
							if (!wasfriend) totalkills--;
						}

						if (mo->flags & MF_COUNTKILL)
						{
							if ( isfriend) totalfriends++;
							if (!isfriend) totalkills++;
						}
					}
				}
			}
		}
		t79->replacemocount = 0;
	}

	health = mo->health;

	if (t79->t79flags & T79_ACTIVATE)	// v220-4 activate line
	{
		if (t79->t79flags & T79_SWAPTRIGGER)	// v220-35 transfer line special/tag
		{
			int			n = t79->activate;
			sector_t	*sc = &sectors[th7400[n].msect];
			line_t		*ld = th7400[n].refline;

			if (ld)
			{
				if (sc->ceilingheight - sc->floorheight <= FRACUNIT)
					ld->special = th7400[n].triga;
				else
					ld->special = th7400[n].trigb;
			}
		}

		if (th7400[t79->activate].delay)
		{
			spawnActive74Delay(t79->activate);
		}
		else
		{
			subfn_UseSpecialLine(use, th7400[t79->activate].refline, 10);
		}
	}

	if (t79->t79flags & T79_CHECKACTIVATE)	// v220-4 activate line
	{
		if (t79->list79num >= 0)
		{
			th7900_t *tmp79;
			short	 *n = listarray[t79->list79num];
			int m = 0;

			while(n[m] >= 0)
			{
				if (th7900[n[m]].mo == t79->mo)
				{
					t79->t79flags |= T79_REMOVE;
				}
				m++;
			}

			m = 0;

			while(n[m] >= 0)
			{
				tmp79 = &th7900[n[m++]];

				if (   tmp79->mo
					&& tmp79->mo != t79->mo
					&& tmp79->mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)	// Exists
				{
					S_StartSound(t79->u79sndid, ppt.viewplrmo);
					TS_SetGeneralMessage(t79->u79msg);
					goto end_chkactive;
				}
			}
		}
		
		t79->t79flags &= ~T79_CHECKACTIVATE;
		subfn_UseSpecialLine(use, th7400[t79->activate].refline, 10);
	}

end_chkactive:;

	if (   inf77
		&& t79->t79flags & T79_MOVSEC
		&& !(inf77->options & MTF_HARD)
		&& (inf77->t77special < 0x1F80 || inf77->t77special >= 0x1FB0))	// 8064 8112
	{
		int msec = t79->activesec;

		if (sectors[msec].ceilingdata || sectors[msec].floordata)
		{
			return;
		}

		ld->special = inf77->t77special;
		ld->tag     = t79->tag;
		
		if (ld->special == 0)	// check for once only type
		{
			inf77->t77special = 0;
			t79->t79flags &= ~T79_MOVSEC;
		}
		else
		{
			subfn_UseSpecialLine(use, ld, 11);
			SpawnSwitchResetThinker(t79num, inf77->t77lnum);
		}
	}

	if (t79->killambient >= 0 && !(t79->t79flags & T79_NOAMBIENT))
	{
		S_KillAmbient(t79->killambient);
		t79->t79flags |= T79_NOAMBIENT;
	}

	if (health > 0 && (!inf77 || !(inf77->options & MTF_NOKILL)))
	{
		mo->flags |= MF_SHOOTABLE;	// else it won't work

		P_DamageMobj(mo, NULL, NULL, mo->info->spawnhealth, 3);

		mo->flags &= ~MF_SHOOTABLE;

		if (mo->flags2 & MF2_OCCLUDE)
			lines[mo->lineocclude].lnflags2 &= ~LNF2_OCCLUDE;

		if (mo->info->deathsound
			&& (mo->info->deathsound & ~HSF_FLAG_MASK)
			!= (t79->soundid & ~HSF_FLAG_MASK))
		{
			t79->dmo1.x = mo->x;
			t79->dmo1.y = mo->y;
			t79->dmo1.z = mo->z + mo->height/2;
			S_StopAllSounds((mobj_t*)&t79->dmo1);
			S_StartSoundAtVolume(mo->info->deathsound, (mobj_t*)&t79->dmo1, 1.0f, DEF_ATTN);
		}

		if (   t79->t79keys & T79_REMOVEHUDKEY
			&& GL_SetIgnoreFlag(t79->t79keys))
		{
			t79->t79flags |= T79_NOHUDKEY;	// savegame support
		}

		if (t79->soundid)
		{
			if (t79->soundid & HSF_REPEAT)
			{
				t79->ss = SpawnSoundThinker(t79num, t79->soundid);
			}
			else
			{
				if (mo->modelboxnum)
				{	// sounds that can be suspended only apply to modelboxes
					S_StartSoundAtVolume(t79->soundid|HSF_CANSUSPEND, mo, 1.0f, DEF_ATTN);
				}
				else
				{
					t79->dmo2.x = mo->x;
					t79->dmo2.y = mo->y;
					t79->dmo2.z = mo->z + mo->height/2;

					S_StopAllSounds((mobj_t*)&t79->dmo2);
					S_StartSoundAtVolume(t79->soundid, (mobj_t*)&t79->dmo2, t79->volume,
						t79->soundid & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
				}
			}
		}

		if (t79->t79flags & T79_MESSAGESET)
		{
			TS_SetMessage(t79, NULL, true);
		}

		if (t79->resurrect)
		{
			SpawnResurrectThinker(t79num, ld ? ld - lines : -1);
		}
	}

	if (inf77 && t79->t79flags & T79_TRIGGER)
	{
		for (i = 0; i < t77->spawn_count; i++)
		{
			line_t	*li;
			short	options = t77->info[i].options;

			if (options & MTF_HARD)
			{
				li = &lines[t77->info[i].t77lnum];

				if (li->special < 0x1F80 || li->special >= 0x1FB0)	// Prevent recursion with map error
				{
					// b211-0 New type added (note irrespective of the actual line special this is a once
					// only trigger type where a linked action with the same 77xx number kills an amodel.
					// This is so that, following its death, the next time runScript is called the
					// 1Fxx trigger lines are cleared from the inf77 array.
					if (   type == MS_WALKOVER
						&& options == (MTF_HARD|MTF_DEAD|MTF_ALIVE))	// usemask
					{
						subfn_UseSpecialLine(use, li, 12);
					}
					else
					{
						if (options & MTF_DEAD  && health <= 0) continue;
						if (options & MTF_ALIVE && health >  0) continue;

						P_CrossSpecialLine(li, usething);
					}
				}
			}
		}
	}

	if (t79->t79flags & T79_TRANSFER)	// v220-4 transfer line special/tag
	{
		int		k, n = t79->activate;
		line_t	*li;

		if (n >= 0 && (li = th7400[n].refline) )	// should be
		{
			th7400_t *t74 = &th7400[n];
			int rsec = li->frontsector - sectors;

			// Set all referenced line types
			for (k = 0; k < t74->ptrnum; k++)
			{
				line_t		*ls = t74->lptr[k];
				sector_t	*rs = ls->backsector;
				if (rs && rsec >= 0)
				{
					if (t74->flags & T74_CSECLIGHT)
						rs->ceilinglightsec = rsec;

					if (t74->flags & T74_FSECLIGHT)
						rs->floorlightsec = rsec;

					if (t74->flags & T74_CEILPIC)
						rs->ceilingpic = sectors[rsec].ceilingpic;

					if (t74->flags & T74_FLOORPIC)
						rs->floorpic = sectors[rsec].floorpic;

					if ((t74->flags & T74_FOG) >= 0)
					{
						int		x = rs->soundorg.x;
						int		y = rs->soundorg.y;

						P_SpawnMobj(x, y, ONFLOORZ, MT_TFOG, 0);
						
						if (   t74->flags & T74_FOGSOUND
							&& !S_IsPlaying(sfx_telept, (mobj_t*)&rs->soundorg))
						{
							S_StartSound(sfx_telept, (mobj_t*)&rs->soundorg);
						}
					}
				}

				ls->special = li->special; ls->tag = li->tag;

				if (t74->flags & T74_PLAYERONLY)
					ls->scpFlags |= SCPF_PLAYERONLY;
			}
		}
	}

	if (   t79->stopnum > 0
		&& (t79other = &th7900[t79->stopnum - 1])->soundid)
	{
		uint	soundid = t79other->soundid;
		mobj_t	*mo = soundid & HSF_NO_ATTENUATION ? NULL : t79other->mo;

		S_StopSound(soundid, mo);

		if (t79other->ss)
		{
			P_RemoveThinker(&t79other->ss->thinker);
			t79other->ss = NULL;
		}

		t79other->soundid = 0;
	}

	if (t79->replacetype[0])
	{
		spawnspot_t sp;
		mobj_t		*oldmo;

		t79other = &th7900[t79->replacespot];

		oldmo = t79other->mo;

		memcpy(&sp, t79other, sizeof(mapthing_t));

		sp.type		 = t79->replacetype[0];
		sp.z_off	 = 0;
		sp.z_nofall  = 0;
		sp.spawn_fog = 0;
		sp.mo		 = NULL;

		if (!P_SpawnScriptThing(&sp, NULL, get79N(t79num), false, false))
			disableMomonstersError(3, t79->replacespot, t79->replacetype[0]);	// Program error [3]

		if (oldmo && oldmo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
		{
			if (oldmo->flags2 & MF2_VDOOR)
			{
				sp.mo->flags2 |= MF2_VDOOR;
				sp.mo->z = oldmo->z;
			}

			// v220-35
			for (i = 0; i < vdoornum; i++)
			{
				if (vdoormo[i][0] == oldmo) vdoormo[i][0] = sp.mo;
				if (vdoormo[i][1] == oldmo) vdoormo[i][1] = sp.mo;
			}

			if (oldmo->flags2 & MF2_NOOCCLUDE) sp.mo->flags2 |= MF2_NOOCCLUDE;

			oldmo->flags &= ~MF_SPECIAL;	// just in case - not wanted in respawn queue
			P_RemoveMobj(oldmo, 1);
		}

		t79other->mo = sp.mo;
	}

	if (t79->t79flags & T79_ENDLEVELTYPE && scpMessage.smflags & SM_ENDSET)
	{
		RDBOOL	dospawn = true;
		
		t79->t79flags &= ~T79_ENDLEVELTYPE;	// One time only

		// Check all t79's with a delay have been cleared
		for (i = 0; i < NUMSPOTS; i++)
		{
			if (th7900[i].t79flags & T79_ENDLEVELTYPE)
			{
				dospawn = false; break;
			}
		}

		if (dospawn) SpawnEndlevelThinker();
	}

	if (t79->t79flags & T79_TRIGRESURRECT)
	{
		if (health <= 0)
		{
			modelRESET(t79num, false);

			if (t79->replacetype[1])
			{
				spawnspot_t sp;
				mobj_t		*oldmo;

				t79other = &th7900[t79->replacespot];

				oldmo = t79other->mo;

				memcpy(&sp, t79other, sizeof(mapthing_t));

				sp.type		 = t79->replacetype[1];
				sp.z_off	 = 0;
				sp.z_nofall  = 0;
				sp.spawn_fog = 0;
				sp.mo		 = NULL;

				if (!P_SpawnScriptThing(&sp, NULL, get79N(t79num), false, false))
					disableMomonstersError(4, t79->replacespot, t79->replacetype[1]);	// Program error [4]

				if (oldmo && oldmo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
				{
					if (oldmo->flags2 & MF2_NOOCCLUDE) sp.mo->flags2 |= MF2_NOOCCLUDE;
					oldmo->flags &= ~MF_SPECIAL;	// just in case - not wanted in respawn queue
					P_RemoveMobj(oldmo, 2);
				}

				t79other->mo = sp.mo;
			}

			if (t79->soundid)
			{
				S_StartSoundAtVolume(t79->soundid, t79->mo, t79->volume,
					t79->soundid & HSF_NO_ATTENUATION ? 0 : DEF_ATTN);
			}
		}
	}
	else if (ld)
	{
		ld->special = 0;
	}

	// Set any redirected line specials
	for (i = 0; i < redir77[t79num].count; i++)
	{
		if (!lines[redir77[t79num].lnum[i]].tag)
			lines[redir77[t79num].lnum[i]].special = redir77[t79num].spec[i];
	}
	
	// v220-4 Kill any mobjs set with KILLMOBJ
	spawnKillThinker(t79num);
	
	if (t79->t79flags & T79_VDOOR)
	{
		t79->t79flags &= ~T79_VDOOR;
		SpawnCeilingThinker(t79num);
	}

	if (t77->blocked)
	{
		SpawnUnblockThinker(t77, t79num);
	}

	if (health > 0 && t79->t79flags & (T79_RESET|T79_REMOVE))
	{
		if (t79->t79flags & T79_REMOVE)
		{
			t79->mo->flags &= ~MF_SPECIAL;	// just in case - not wanted in respawn queue
			P_RemoveMobj(t79->mo, 3);
			t79->mo = NULL;

			checkOndeathSpawn(t79);
		}
		else
			SpawnResetThinker(t79num);
	}
	else if (inf77 && inf77->options & MTF_NORMAL)
	{
		if (   health <= 0
			&& !t77->blocked
			&& checkBlockedClear(t77))
		{
			modelRESET(t79num, true);

			if (mo->info->raisestate) SpawnStateMonitor(t79num);

			if (ld->linesound)
			{
				S_StartSound (ld->linesound, usething);
			}

			for (i = 0; i < t77->spawn_count; i++)
			{
				if (!(t77->info[i].options & MTF_NORMAL))
				{
					lines[t77->info[i].t77lnum].special = t77->info[i].t77special;	// Reset any line triggers
				}
			}
		}

		ld->special = inf77->t77special;
	}
	else
		checkOndeathSpawn(t79);
}

// To save memory this is done using flags (Urgh!)
static void checkDisable(int num)
{
	int i, k, n, p;

	for (i = 0; i < 7; i++)
	{
		if ( (n = T78_disable[num][i]) )
		{
			for (k = 0, p = 1; k < 32; k++, p <<= 1)
			{
				if (n & p)
				{
					int dn = k + i*32;

					if ( !(th7800[dn].setflags & T_UNSET) )
						th7800[dn].setflags |= T_UNSET;
					else if (th7800[dn].setflags & T_ENABLE)
						th7800[dn].setflags &= ~T_UNSET;

					T78_disable[num][i] &= ~p;	// one time action
				}
			}
		}
	}
}

// v220-13
void TS_setHoldState(line_t *ld)
{
	mobj_t	*mo;

	for (mo = ld->backsector->thinglist; mo; mo = mo->snext)
	{
		if (  !(mo->ddflags & DDF_HOLDSTATE)
			&& mo->subsector->sector == ld->backsector	// not partially touching
			&& (mo->flags & (MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL)) == (MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL))
		{
			mo->ddflags |= DDF_HOLDSTATE;
		}
	}
}

// v220-13
static void clearHoldState(line_t *ld)
{
	mobj_t	*mo;

	for (mo = ld->backsector->thinglist; mo; mo = mo->snext)
	{
		if (mo->ddflags & DDF_HOLDSTATE)
			mo->ddflags &= ~DDF_HOLDSTATE;
	}
}

RDBOOL TS_Activate(mobj_t* thing, line_t *ld, int type)
{
	RDBOOL		isplayer;

	//-------------------------------------

	// Check for compliance
	if (!R3D_ScriptedMap) return false;

	// v220-13 add support for specials 8113/8114
	if (ld->special == 8113)	// Walkover release state trigger
	{
		if (type == MS_WALKOVER && thing->dplayer)
		{
			clearHoldState(ld);
			ld->special = 0;
		}
		return false;
	}

	if (ld->special == 8114)	// Gun release state trigger
	{
		if (type == MS_GUN && thing->dplayer)
		{
			clearHoldState(ld);
			ld->special = 0;
		}
		return false;
	}

	// v220-35 add support for specials 8142/8143
	if (ld->special == 8142)	// Use release state trigger
	{
		if (type == MS_USE && thing->dplayer)
			goto activateDirect;
		return false;
	}

	if (ld->special == 8143)	// Use trigger (but gun activated)
	{
		if (type == MS_USE && thing->dplayer)
			goto activateDirect;
		return false;
	}


	// lines with scripted specials must not have a tag
	// check special is in the allowed range
	if (ld->tag || ld->special < 0x1F80 || ld->special > 0x1FA2) return false;	// 8098

	isplayer = thing->dplayer != NULL;

	{
		int		range = ld->special & 0xF;

		switch(type)
		{
			// 8064 8065 8066 or 8080 8081 8082 or 8096 8097 8098
			case MS_USE:
				if (   (ld->special >= 8064 && ld->special <= 8066)		// 0x1F80 to 0x1F82
					|| (ld->special >= 8080 && ld->special <= 8082)		// 0x1F90 to 0x1F92
					|| (ld->special >= 8096 && ld->special <= 8098) )	// 0x1FA0 to 0x1FA2 resettable types
				{
					if (range == 0 &&  isplayer) break;	// player   triggered
					if (range == 1 && !isplayer) break;	// monster  triggered
					if (range == 2) break;				// anyTHING triggered
				}
				return false;

			// 8067 8068 8069 or 8083 8084 8085
			case MS_GUN:
				if (   (ld->special >= 8067 && ld->special <= 8069)		// 0x1F83 to 0x1F85
					|| (ld->special >= 8083 && ld->special <= 8085) )	// 0x1F93 to 0x1F95
				{
					if (range == 3 &&  isplayer) break;	// player   triggered
					if (range == 4 && !isplayer) break;	// monster  triggered
					if (range == 5) break;				// anyTHING triggered
				}
				return false;

			// 8070 8071 8072 8073
			case MS_WALKOVER:
				if (ld->special >= 8070 && ld->special <= 8073)			// 0x1F86 to 0x1F89
				{
					if (range == 6 &&  isplayer) break;	// player   triggered
					if (range == 7 && !isplayer) break;	// monster  triggered
					if (range == 8) break;				// anyTHING triggered
					// v220-4							// friend   triggered
					if (range == 9 && !isplayer && thing->flags3 & MF3_FRIEND)
						break;
				}
				return false;

			default:
				return false;
		}
	}

activateDirect:;
	int		i, k, m, n, linenum = ld - lines;

	linenum = ld - lines;

	for (i = 0; i < NUMSPOTS; i++)
	{
		th7700_t	*t77 = &th7700[i];
		th7900_t	*t79 = &th7900[i];
		RDBOOL		singlekey;

		if (t79->t79flags & T79_DELAYACTIVE) continue;

		for (k = 0; k < t77->spawn_count; k++)
		{
			if (t77->info[k].t77lnum == linenum)
			{
				checkDisable(i);

				// v220-35 check for surrogate trigger
				if (   (surrogate.special = t77->info[k].t77dummyspecial)
					&& (surrogate.tag = t77->info[k].t77dummytag))
				{
					subfn_UseSpecialLine(thing, &surrogate, 15);
				}

				if ( (n = t79->thingref78) >= 0)
				{
					if (   t79->t79flags & T79_NULLTRIGGER
						&& th7800[n].blocklines[0])
					{
						m = 0;

						while (th7800[n].blocklines[m])
						{
							th7800[n].blocklines[m++]->lnflags2 &= ~LNF2_BLOCKALL;
						}
					}
					else
					if (   t79->t79flags & T79_SETTRIGGER
						&& th7800[n].blocklines[0])
					{
						m = 0;

						while (th7800[n].blocklines[m])
						{
							th7800[n].blocklines[m++]->lnflags2 |= LNF2_BLOCKALL;
						}
					}
					else
					if (  t79->t79flags & (T79_SETTOGGLE|T79_NULLTRIGGER|T79_SETTRIGGER)
						&& th7800[n].refline78
						&& th7800[n].refmo
						&& th7800[n].refmo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
					{
						if (t79->t79flags & T79_SETTOGGLE)
						{
							if (lines[th7800[n].refline78].lnflags2 & LNF2_BLOCKALL)
							{
								lines[th7800[n].refline78].lnflags2 &= ~LNF2_BLOCKALL;
								th7800[n].refmo->flags2 |= MF2_NORENDER;
								th7800[n].refmo->subsector->sector->special = 0;
								if (!(th7800[n].setflags & T74_TRIGLINEONLY))
									TS_SetGeneralMessage(th7800[n].unblocktextmsg);
							}
							else
							{
								lines[th7800[n].refline78].lnflags2 |= LNF2_BLOCKALL;
								th7800[n].refmo->flags2 &= ~MF2_NORENDER;
								th7800[n].refmo->subsector->sector->special = th7800[n].secdamage;
								if (!(th7800[n].setflags & T74_TRIGLINEONLY))
									TS_SetGeneralMessage(th7800[n].blocktextmsg);
							}
						}
						else if (t79->t79flags & T79_NULLTRIGGER
							&& lines[th7800[n].refline78].lnflags2 & LNF2_BLOCKALL)
						{
							lines[th7800[n].refline78].lnflags2 &= ~LNF2_BLOCKALL;
							th7800[n].refmo->flags2 |= MF2_NORENDER;
							th7800[n].refmo->subsector->sector->special = 0;
							if (!(th7800[n].setflags & T74_TRIGLINEONLY))
								TS_SetGeneralMessage(th7800[n].unblocktextmsg);
						}
						else if (t79->t79flags & T79_SETTRIGGER
							&& !(lines[th7800[n].refline78].lnflags2 & LNF2_BLOCKALL))
						{
							lines[th7800[n].refline78].lnflags2 |= LNF2_BLOCKALL;
							th7800[n].refmo->flags2 &= ~MF2_NORENDER;
							th7800[n].refmo->subsector->sector->special = th7800[n].secdamage;
							if (!(th7800[n].setflags & T74_TRIGLINEONLY))
								TS_SetGeneralMessage(th7800[n].blocktextmsg);
						}
					}
				}

				if (!checkKeys(t79, thing->player, &singlekey, linenum) )
				{
					if (thing->player)
					{
						uint soundid = t79->locksndid ? t79->locksndid : sfx_oof;
						S_StartSound(soundid, ppt.viewplrmo);
						TS_SetGeneralMessage(t79->lockmsg);
					}
					return true;
				}

				// v220-23 support 7900 exists and in stasis list
				if (t79->list79num >= 0 && t79->t79flags & T79_CHECKLIST)
				{
					th7900_t *tmp79;
					short	 *n = listarray[t79->list79num];
					m = 0;

					while(n[m] >= 0)
					{
						tmp79 = &th7900[n[m++]];

						if (   tmp79->mo
							&& tmp79->mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker	// Exists
							&& (tmp79->mo->tics == -1))												// In stasis
						{
							S_StartSound(t79->u79sndid, ppt.viewplrmo);
							TS_SetGeneralMessage(t79->u79msg);
							return false;
						}
					}
				}

				// If any KILL things are active then abort
				for (m = 0; m < t79->killcount; m++)
				{
					if (th7900[t79->killthing[m][0]].t79flags & T79_DELAYACTIVE) return true;
				}

				for (m = 0; m < t77->direccount; m++)
				{
					lines[t77->info[t77->direcset[m][0]].t77lnum].special =
						t77->direcset[m][1];
				}

				t77->direccount = 0;

				if (t79->t79flags & T79_INITDEAD)
				{
					if (t79->mo)
					{
						modelRESET(i, false);
						t79->t79flags &= ~T79_INITDEAD;
					}

					// Not 0 but an unrecognised value (0x1FAA) to allow linesound to work
					ld->special = 8106;
					return true;
				}

				if (   thing
					&& type == MS_USE
					&& t79->t79flags & T79_VDOOR
					&& thing->z + thing->height < t79->startheight)
				{
					return false;
				}

				// Check if waiting on a vdoor to shut
				if (t79->t79flags & T79_VDOORWAIT
					&& sectors[t79->vdoorsec].ceilingheight > t79->vdoorsecheight)
				{
					TS_SetGeneralMessage(t79->vdoorwaitmsg);
					return false;
				}

				memset(t79list, 0, sizeof(t79list));
				runScript(thing, i, &t77->info[k], ld, type);

				if (t79->cotrigger > 0)
				{
					th7900_t *t = &th7900[t79->cotrigger - 1];

					if (t->mo && t->mo->tics < 0)
						runScript(thing, t79->cotrigger - 1, NULL, NULL, 0);
				}

				// v220-23
				if (isplayer && t79->t79keys & T79_REMOVEHUDKEY && clearKey(singlekey))
				{
					t79->t79flags &= ~T79_NOHUDKEY;
				}

				return true;
			}
		}
	}
	return true;
}

// External MS_USE Doom call
RDBOOL T_USEexternal(line_t *ld, mobj_t* thing)
{
	for (int i = 0; i < NUMSPOTS; i++)
	{
		if (   th7800[i].refline78 == ld - lines
			&& th7800[i].setflags & T74_TRIGLINEONLY)
		{
			if (th7800[i].blocktextmsg[0])
				TS_SetGeneralMessage(th7800[i].blocktextmsg);
			if (thing->player)
				S_StartSound(sfx_oof, thing->player->plr->mo);
			return true;
		}
	}

	return false;
}
