//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_anifile.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Note: Utilizes LZSS compression.
//
// DESCRIPTION: Analyser file save/load
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <direct.h>
#include <string.h>
#include <sys/timeb.h>

#include "lzss.h"

#include "r3d_MAINspec.h"
#include "r3d_nodefile.h"
#include "r3d_xline.h"
#include "dd_wad.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

#define	ANVERSION	0x02203104	// (v220-34) internal version 4
#define NBVERSION	0x02202510	// (v220-25) changed from 0x02206003

#define	CHECKVAL	0xB54EA3C7

// TYPES -------------------------------------------------------------------
	
// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void 	P_MobjThinker (mobj_t* mobj);
void	SV_Read(void *data, int len);
void	FW_SetFogLines(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	R3D_edit;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

aniseclight_t anisl;
float	stat_analysetime;
float	stat_nodestime;
int		stat_vispresent;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static LZFILE *anifile;
static LZFILE *nodfile;

// CODE --------------------------------------------------------------------

static void SV_ReadAN(void *data, int len)
{
	lzRead(data, len, anifile);
}

static void SV_WriteAN(void *data, int len)
{
	lzWrite(data, len, anifile);
}

void R3D_WriteAnalysedFile(char *filename)
{
	struct _timeb		timebuffer;

	subsector_t	*ss;
	sector_t	*sec;
	line_t		*ld;
	side_t		*sd;
	seg_t		*sg;
	int			i, m;

	// ----------------------------

	if ( !(anifile = lzOpen(filename, "wp")) )
	{
		pr_OFinsert("ERROR: Cannot create file %s", filename);
		if ((_access( filename, 4 )) != -1) remove(filename);
		return;
	}

	pr_OFinsert("WRITING: analysis file %s\n", filename);

	lzPutL(ANVERSION, anifile);

	_ftime( &timebuffer );
	lzPutL(timebuffer.time, anifile);

	// v220-9 Only check files having levels
	for (i = m = 0; i < numfilenameRecords; i++)
	{
		if (filenameRecords[i].numlevels)
		{
			lzPutL(RetrieveFileDateByNum(i), anifile); m++;
		}
	}

	lzPutC((byte)m, anifile);
	lzPutC((byte)ge_inf.iwadid, anifile);

	lzPutL(*(long*) &stat_analysetime, anifile);
	lzPutL(stat_vispresent, anifile);

	lzPutW(ma_inf.midtexerror, anifile);
	lzPutL(numlines, anifile);

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		lzPutC(ld->specType  , anifile);
		lzPutC(ld->mtflags   , anifile);
		lzPutC(ld->spflags   , anifile);
		lzPutC(ld->tyflags   , anifile);
		lzPutC(ld->ovflags   , anifile);
		lzPutC(ld->scpFlags  , anifile);
		lzPutL(ld->lnflags2  , anifile);
		lzPutW(ld->tag       , anifile);
		lzPutW(ld->ml_flags & ML_EXTRAFLAGS, anifile);
		lzPutL(ld->adjacentsec, anifile);
		lzPutW(ld->tag, anifile);
		lzPutW(linefog_group[i], anifile);
	}

	lzPutL(anisl.numentries, anifile);

	for (i = m = 0; i < anisl.numentries; i++)
	{
		lzPutL(anisl.entries[m], anifile); m++;
		lzPutL(anisl.entries[m], anifile); m++;
	}

	lzPutL(numsides, anifile);

	for (i = 0, sd = sides; i < numsides; i++, sd++)
	{
		lzPutW(sd->texFlags,      anifile);
		lzPutW(sd->bottomtexture, anifile);
		lzPutW(sd->midtexture,    anifile);
		lzPutW(sd->toptexture,    anifile);
		lzPutL(sd->rowoffadj,     anifile);
		lzPutL(sd->rowoffset,     anifile);
		lzPutL(sd->textureoffset, anifile);
	}

	lzPutL(numsegs, anifile);

	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		lzPutC(sg->sgflags>>8, anifile);
	}

	lzPutL(numsectors, anifile);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		SV_WriteAN(&sectors[i].sibounds, sizeof(sectors[i].sibounds));
		SV_WriteAN(&sec->reverb, sizeof(sec->reverb));

		lzPutL(sec->ceilingheight, anifile);
		lzPutL(sec->ceilinglightsec, anifile);
		lzPutL(sec->floorlightsec, anifile);
		lzPutL(sec->specFlags, anifile);
		lzPutL(sec->specFlags2, anifile);
		lzPutL(sec->R3Dfloorsec, anifile);
		lzPutL(sec->R3Dceilsec, anifile);
		lzPutL(sec->R3Dfloorheight, anifile);
		lzPutL(sec->R3Dceilheight, anifile);
		lzPutL(sec->R3Dlowsec, anifile);
		lzPutL(sec->floodline, anifile);
		lzPutL(sec->ceildiff, anifile);
		lzPutL(sec->sidesky, anifile);
		lzPutL(sec->skyfix, anifile);
		lzPutL(sec->specFlags, anifile);
		lzPutL(sec->oversec, anifile);
		lzPutL(sec->heightsec, anifile);
		lzPutW(sec->tag, anifile);
		lzPutL(sec->withheight, anifile);		// $v16added
		lzPutW(sec->texwithheight[0], anifile);	// $v16added
		lzPutW(sec->texwithheight[1], anifile);	// $v16added
		lzPutW(sec->ceilingpic, anifile);		// $v26added
		lzPutW(sec->floorpic, anifile);			// $v26added
	}

	lzPutL(numsubsectors, anifile);

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		lzPutL(ss->sector - sectors, anifile);
		lzPutL(ss->sslightsec, anifile);
		lzPutL(ss->floorsec, anifile);
		lzPutL(ss->ceilsec, anifile);
		lzPutL(ss->lcsec, anifile);
		lzPutL(ss->lfsec, anifile);
		lzPutL(ss->viewgroup, anifile);
		lzPutL(ss->ssflags , anifile);
	}

	lzPutL(ma_inf.sseclogical, anifile);
	lzPutL(ma_inf.changeconflicting, anifile);

	lzPutL(CHECKVAL, anifile);

	m = movsec_list[0];

	for (i = 0; i <= m; i++) lzPutL(movsec_list[i], anifile);

	lzPutL(CHECKVAL, anifile);

	lzClose(anifile);
}

int subfn_ReadAnalysedFile()
{
	subsector_t	*ss;
	sector_t	*sec;
	line_t		*ld;
	side_t		*sd;
	seg_t		*sg;
	int			i, val;

	if (lzGetL(anifile) != ANVERSION) return 1;

	lzGetL(anifile);

	for (i = val = 0; i < numfilenameRecords; i++)
	{
		if (filenameRecords[i].numlevels)
		{
			if (lzGetL(anifile) != RetrieveFileDateByNum(i)) return 3;
			val++;
		}
	}

	if (lzGetC(anifile) != (byte)val)  return 4;
	if (lzGetC(anifile) != (byte)ge_inf.iwadid)  return 5;

	val = lzGetL(anifile);
	stat_analysetime = *(float*) &val;
	stat_vispresent = lzGetL(anifile);

	ma_inf.midtexerror = (int)(lzGetW(anifile));

	val = lzGetL(anifile);
	if (val != numlines) return 6;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		ld->scpFlags &= SCPF_LOWMASK;

		ld->specType    = lzGetC(anifile);
		ld->mtflags	    = lzGetC(anifile);
		ld->spflags	    = lzGetC(anifile);
		ld->tyflags	    = lzGetC(anifile);
		ld->ovflags	    = lzGetC(anifile);
		ld->scpFlags   |= (lzGetC(anifile) & SCPF_HIMASK);	// v220-9 NOTE: ored
		ld->lnflags2    = lzGetL(anifile);
		ld->tag		    = lzGetW(anifile);
		ld->ml_flags   |= lzGetW(anifile);	// NOTE: ored
		ld->adjacentsec = lzGetL(anifile);
		ld->tag			= lzGetW(anifile);
		linefog_group[i]= lzGetW(anifile);
	}

	val = lzGetL(anifile);

	for (i = 0; i < val; i++)
	{
		int secidx = lzGetL(anifile);
		int llevel = lzGetL(anifile);

		sectors[secidx].lightlevel = llevel;
	}

	val = lzGetL(anifile);
	if (val != numsides) return 7;

	for (i = 0, sd = sides; i < numsides; i++, sd++)
	{
		sd->texFlags		= lzGetW(anifile);
		sd->bottomtexture	= lzGetW(anifile);
		sd->midtexture		= lzGetW(anifile);
		sd->toptexture		= lzGetW(anifile);
		sd->rowoffadj		= lzGetL(anifile);
		sd->rowoffset		= lzGetL(anifile);
		sd->textureoffset	= lzGetL(anifile);
	}

	val = lzGetL(anifile);
	if (val != numsegs) return 8;

	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		sg->sgflags = (lzGetC(anifile))<<8;
	}

	val = lzGetL(anifile);
	if (val != numsectors) return 9;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		SV_ReadAN(sectors[i].sibounds, sizeof(sectors[i].sibounds));
		SV_ReadAN(sec->reverb, sizeof(sec->reverb));

		sec->ceilingheight		= sec->smoothceil = lzGetL(anifile);
		sec->ceilinglightsec	= lzGetL(anifile);
		sec->floorlightsec		= lzGetL(anifile);
		sec->specFlags			= lzGetL(anifile);
		sec->specFlags2			= lzGetL(anifile);
		sec->R3Dfloorsec		= lzGetL(anifile);
		sec->R3Dceilsec			= lzGetL(anifile);
		sec->R3Dfloorheight		= lzGetL(anifile);
		sec->R3Dceilheight		= lzGetL(anifile);
		sec->R3Dlowsec			= lzGetL(anifile);
		sec->floodline			= lzGetL(anifile);
		sec->ceildiff			= lzGetL(anifile);
		sec->sidesky			= lzGetL(anifile);
		sec->skyfix				= lzGetL(anifile);
		sec->specFlags			= lzGetL(anifile);
		sec->oversec			= lzGetL(anifile);
		sec->heightsec			= lzGetL(anifile);
		sec->tag				= lzGetW(anifile);
		sec->withheight			= lzGetL(anifile);	// $v16added
		sec->texwithheight[0]	= lzGetW(anifile);	// $v16added
		sec->texwithheight[1]	= lzGetW(anifile);	// $v16added
		sec->ceilingpic			= lzGetW(anifile);	// $v27added
		sec->floorpic			= lzGetW(anifile);	// $v27added
	}

	val = lzGetL(anifile);
	if (val != numsubsectors) return 10;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		ss->sector		= &sectors[lzGetL(anifile)];
		ss->sslightsec	= lzGetL(anifile);
		ss->floorsec	= lzGetL(anifile);
		ss->ceilsec		= lzGetL(anifile);
		ss->lcsec		= lzGetL(anifile);
		ss->lfsec		= lzGetL(anifile);
		ss->viewgroup	= lzGetL(anifile);
		ss->ssflags		= lzGetL(anifile);
	}
	
	ma_inf.sseclogical = lzGetL(anifile);
	ma_inf.changeconflicting = lzGetL(anifile);

	val = lzGetL(anifile);
	if (val != CHECKVAL) return 11;

	val = lzGetL(anifile);
	if (val > numsectors) return 12;

	movsec_list[0] = val;

	for (i = 1; i <= val; i++) movsec_list[i] = lzGetL(anifile);

	val = lzGetL(anifile);

	return val != CHECKVAL ? 13 : 0;
}

int R3D_CheckAnalysedFile(char *filename, long nodetime)
{
	int i, m, w1, w2, error = 0;

	anifile = lzOpen(filename, "rp");
	if (!anifile) return -1;

	if (lzGetL(anifile) != ANVERSION)
	{
		error = 1;
	}
	else if (lzGetL(anifile) < nodetime)
	{
		error = 2;
	}
	else
	{
		for (i = m = 0; i < numfilenameRecords; i++)
		{
			if (filenameRecords[i].numlevels)
			{
				if (lzGetL(anifile) != RetrieveFileDateByNum(i))
				{
					error = 3; break;
				}
				m++;
			}
		}
	
		if (!error && lzGetC(anifile) != (byte)m)  error = 4;
	}

	w1 = ge_inf.iwadid;

	// v220-27 iwad check
	if (!error && (w2 = (int)lzGetC(anifile)) != w1)
	{
#include "sys_system.h"
		char *iwad[] = { "NONE", "Plutonia", "TnT", "Doom2", "Doom" };
		char tmp[1024];

		if (w2 > 0 && w2 < 5)	// insurance
		{
			sprintf(tmp,	"An analysis file has been detected which was\n"
							"built using the %s iwad\n\n"
							"To continue will cause the file to be rebuilt\n"
							"using the current %s iwad\n\n"
							"Not continuing will abort allowing the iwad\n"
							"to be reselected before restarting the game\n",
							iwad[w2], iwad[w1]);

			if (!Sys_CriticalMessage(tmp, true))
			{
				lzClose(anifile); Sys_Quit(false);
			}
		}

		error = 5;
	}

	lzClose(anifile);
	return error;
}

//==================================================================
// Read/Write nodes files v220-6
//==================================================================

int R3D_ReadAnalysedFile(char *filename)
{
	int error;

	anifile = lzOpen(filename, "rp");
	if (!anifile) return -1;

	FW_SetFogLines();

	pr_OFinsert("READING: analysis file %s\n", filename);

	error = subfn_ReadAnalysedFile();
	lzClose(anifile);
	return error;
}

int R3D_WriteNodesFile(char *fname)
{
	nodfile = lzOpen(fname, "wp");
	if (!nodfile) return -1;

	lzPutL(NBVERSION, nodfile);

	// For stats only
	lzPutL(*(long*) &stat_nodestime, nodfile);

	edn.xlsame = num_xlsame;

	if (!num_xlsame) edn.modifier &= ~MODF_MODNODES;

	lzPutL(edn.edit     , nodfile);
	lzPutL(edn.modifier , nodfile);
	lzPutL(edn.nodetime , nodfile);
	lzPutL(edn.filetime , nodfile);
	lzPutL(edn.length   , nodfile);
	lzPutL(edn.xlsame	, nodfile);

	lzWrite(edn.data, edn.length, nodfile);

	if (num_xlsame)
	{
		int i;

		for (i = 0; i < num_xlsame; i++)
			lzWrite(&xlinemem[i], NODSAVESIZE, nodfile);
	}

	lzClose(nodfile);

	return 0;
}

// modified v220-27
void R3D_ReadNodesFile(char *fname, int setxsame)
{
	memset(&edn, 0, sizeof(editnode_t));
	nodfile = lzOpen(fname, "rp");

	if (!nodfile) return;

	if (lzGetL(nodfile) != NBVERSION)
	{
		lzClose(nodfile);
		return;
	}

	// For stats only
	int val = lzGetL(nodfile);
	stat_nodestime = *(float*) &val;

	edn.edit     = lzGetL(nodfile);
	edn.modifier = lzGetL(nodfile);
	edn.nodetime = lzGetL(nodfile);
	edn.filetime = lzGetL(nodfile);
	edn.length   = lzGetL(nodfile);
	edn.xlsame   = lzGetL(nodfile);
	edn.data = (byte*)M_Malloc((uint)edn.length, 0);
	lzRead(edn.data, (long)edn.length, nodfile);

	if (num_xlsame) ma_inf.modifier |= MODF_MODNODES;

	// check the values match or are in range
	if (   edn.edit != R3D_edit
		|| edn.modifier != (ma_inf.modifier & MODF_NODEMASK)
		|| edn.length <= 0)
	{
		lzClose(nodfile);
		memset(&edn, 0, sizeof(editnode_t));
		return;
	}

	if (setxsame && (num_xlsame = edn.xlsame) > 0 )
	{
		int i;

		xlinemem = (xlinemem_t*)W_Calloc(num_xlsame * sizeof(xlinemem_t), LEVELID, 0);
		xlisrendered = (byte*)W_Malloc(num_xlsame, LEVELID, 11);

		for (i = 0; i < num_xlsame; i++)
			lzRead(&xlinemem[i], NODSAVESIZE, nodfile);
	}

	lzClose(nodfile);
}