//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_consolidate.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
//
// DESCRIPTION: aggregates results and sets subsectors' pointers
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_MAINspec.h"

// CODE --------------------------------------------------------------------

// Preset a few things - set SSF_SECTSET indicating to renderer
// that lines need to be inspected for references.
void subfn_PRESET()
{
	subsector_t *ss;
	int i;

	for (i = 0; i < numlines; i++)
	{
		if (lnsec[i] >= 0)
		{
			if (lxarray[i] & (1|32))
			{
				lnsec[i] = -1;
			}
		}
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		ss->dlceilheight  = ss->sector->ceilingheight;
		ss->dlfloorheight = ss->sector->floorheight;
		ss->ssfloorpic    = ss->sector->floorpic;
		ss->ssceilpic     = ss->sector->ceilingpic;
	}
}

static RDBOOL SecNoMoveLines(line_t *ld, int side)
{
	sector_t *sec = side ? ld->backsector : ld->frontsector;
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *lc = sec->lines[i];
		if (!(lc->ml_flags & ML_TWOSIDED) ||
			lc == ld)
			continue;
		if (lc->frontsector != sec &&
			lc->frontsector->specFlags & SF_MOVSECTOR)
			return false;
		if (lc->backsector != sec &&
			lc->backsector->specFlags & SF_MOVSECTOR)
			return false;
	}
	return true;
}

static int subfn_CheckR3D_1DSec(line_t *ld, int side)
{
	sector_t *sec = side ? ld->backsector : ld->frontsector;
	int i, newsec = -1, secnum = sec - sectors;

	if (sec->linecount != 4) return secnum;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *lc = sec->lines[i];

		if ( !(lc->ml_flags & ML_TWOSIDED) ) return -1;
		if (lc == ld) continue;
		if (!origTEX[lc - lines] &&
			lc->backsector->ceilingheight > lc->frontsector->ceilingheight &&
			lc->backsector->floorheight == lc->frontsector->floorheight)
		{
			if (newsec < 0)
				newsec = lc->frontsector - sectors;
			else
				return secnum;
		}
		else
			return secnum;
	}
	if (newsec < 0) return secnum;
	return newsec;
}

static RDBOOL noBackHeight(line_t *ld)
{
	if (   ld->frontsector == ld->backsector
		|| !(ld->ml_flags & ML_TWOSIDED)
		|| ld->frontsector->specFlags & SF_MOVSECTOR
		|| ld->backsector->specFlags & SF_MOVSECTOR
		|| ld->frontsector->floorheight == ld->backsector->floorheight
		|| ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
	{
		return false;
	}

	return (ld->frontsector->ceilingheight - ld->frontsector->floorheight < 4*FRACUNIT
		 || ld->backsector->ceilingheight  - ld->backsector->floorheight  < 4*FRACUNIT);
}

// Check for textureless mov ceiling with same sector sided line
static void checkR3D_03Type(line_t *ld)
{
	if (   !lineflags[ld - lines]
		&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
		&& (ld->frontsector->specFlags & 0x3000C) == 0x2000C)
	{
		vertexowner_t *vo = &vertexowners[ld->v1 - vertexes];
		int		i, secnum = -1, ldsec = ld->frontsector - sectors;

		for (i = 0; i < vo->num; i++)
		{
			if (vo->list[i] != ldsec)
			{
				if (secnum >= 0) return;
				secnum = vo->list[i];
			}
		}

		vo = &vertexowners[ld->v2 - vertexes];

		for (i = 0; i < vo->num; i++)
		{
			if (   vo->list[i] != ldsec
				&& vo->list[i] != secnum)
			{
				return;
			}
		}

		if (   secnum >= 0
			&& !(sectors[secnum].specFlags & (SF_MOVSECTOR|SF_NOHITEX|SF_NOLOTEX)) )
		{
			ssline_t *sl = &sslines[ld - lines];

			for (i = 0; i < sl->num; i++)
			{
				subsector_t *ss = &subsectors[sl->list[i]];

				if (ss->ceilsec < 0 && ss->floorsec < 0 && ss->sector - sectors != secnum)
					SetOverride(ss, secnum, 30009);

				SetLFG(ld, LFG_NEWSET, 30009);
				Setsecdone(ld->frontsector, 30009);
			}
		}
	}
}

// Consolidate lines and additionally set up values needed by renderer
// for ceiling flats and check for structural offset differences
void LS_Consolidate()
{
	line_t	*ld;
	int		c, i;

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		byte txn;
		RDBOOL movef, moveb;

		if (ld->specType >= R3D_40) continue;
		if (ld->lnflags2 & LNF2_SLOPESECLN) continue;		// v220-8
		if (ld->specType == R3D_22) ld->specType = R3D_00;
		if (lineflags[i] & LFG_NOCHANGE) continue;

		if (!(ld->ml_flags & ML_TWOSIDED))
		{
			if (ld->specType != R3D_00) Settype(ld, R3D_00, 30010);
			continue;
		}

		txn = origTEX[i];
		movef = (ld->frontsector->specFlags & SF_MOVSECTOR) > 0;
		moveb = (ld->backsector->specFlags & SF_MOVSECTOR) > 0;

		if ((txn == 0x40 && movef && !moveb &&
			 ld->mtflags == (R3D_MISHTEX1|R3D_INVHTEX2)) ||
			(txn == 0x04 && !movef && moveb &&
			 ld->mtflags == (R3D_MISHTEX2|R3D_INVHTEX1)))
		{
			if (Vtxs2(ld) && Vtfl3(ld))
			{
				Forcetype(ld, R3D_00, 30011);
				continue;
			}
		}
		
		// Consolidate line types for rendering
		switch(ld->specType)
		{
			case R3D_00:
				// Look for line with same sector each side but
				// with one hi-tex set.
				if (  LSS(ld) &&
					((sides[ld->sidenum[0]].texFlags == T_HI &&
					 !sides[ld->sidenum[1]].texFlags) ||
					( sides[ld->sidenum[1]].texFlags == T_HI &&
					 !sides[ld->sidenum[0]].texFlags)))
				{
					linearray[c++] = i;
					Settype(ld, R3D_22, 30012);	//hijack

					// Eliminate line such as 1117 in Doom2 map14
					if ( !(ld->lnflags2 & LNF2_OWNEQUALTOTWO) )	// $v73
					{
						// Mark this for test later on.
						// We are looking for lines such as 4522, 4549 in AV map19
						hitline[i] = 1;
					}
				}
				else if (txn == 0x22 && (movef ^ moveb))
				{
					Settype(ld, R3D_1D, 30013);
				}
				// v220-4 added to trap certain types
				else if (!movef && !moveb)
				{
					if (   (txn & 0x22) == 0x22
						&& (txn & 0x11) == 0x11
						&& !(lineflags[i] & (LFG_FIXED|LFG_3DTYPE))
						&& ld->frontsector->floorheight != ld->backsector->floorheight)
					{
						Settype(ld, R3D_03, 30013);
					}
					else if ((txn == 0x02 || txn == 0x20)
						&& lineflags[i] & LFG_FIXED
						&& ld->frontsector->floorheight == ld->backsector->floorheight
						&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
					{
						Forcetype(ld, R3D_03, 30011);
					}
				}
				break;
				
			case R3D_31:
				Forcetype(ld, R3D_00, 30014);
				break;

			case R3D_22:
			case R3D_23:
			case R3D_37:
			case R3D_3F:
				if (   lineflags[i] & LFG_FIXED		// No Forcetypes in this group
					|| !(ld->ml_flags & ML_TWOSIDED))	// should be
				{
					break;
				}

				// Test added rev69
				if (   origTEX[i]
					&& ld->frontsector->ceilingpic != skyflatnum
					&& ld->backsector->ceilingpic != skyflatnum)
				{
					Settype(ld, R3D_00, 30016);
				}
				else
				{
					int diff, mask = SF_SECCLOSED|SF_SSLINETRUE;

					if (   ld->frontsector->specFlags & SF_MOVSECTOR
						|| ld->backsector->specFlags & SF_MOVSECTOR)
					{
						Settype(ld, R3D_01, 30017);	// default
					}
					// v220-4 test added
					else if (ld->specType == R3D_3F
						&& ld->frontsector != ld->backsector
						&& ld->frontsector->stptr < 0
						&& ld->backsector->stptr < 0
						&& ld->mtflags & R3D_MISTEXANY)
					{
						RDBOOL	valid = true;
						
						diff = ld->frontsector->floorheight - ld->backsector->floorheight;

						if (   diff < 0
							&& (ld->backsector->specFlags & SF_SSLINETRUE
							|| !(ld->backsector->specFlags & SF_NOLOTEX)) )
						{
							valid = false;
						}

						if (   diff > 0
							&& (ld->frontsector->specFlags & SF_SSLINETRUE
							|| !(ld->frontsector->specFlags & SF_NOLOTEX)) )
						{
							valid = false;
						}

						if (valid)
						{
							diff = ld->frontsector->ceilingheight - ld->backsector->ceilingheight;

							if (   diff > 0
								&& (ld->backsector->specFlags & SF_SSLINETRUE
								|| !(ld->backsector->specFlags & SF_NOHITEX)) )
							{
								valid = false;
							}

							if (   diff < 0
								&& (ld->frontsector->specFlags & SF_SSLINETRUE
								|| !(ld->frontsector->specFlags & SF_NOHITEX)) )
							{
								valid = false;
							}
						}

						if (!valid)
						{
							Settype(ld, R3D_00, 30017);	// default v220-4, was R3D_16
						}
						// add test v220-9
						else if (ld->frontsector->R3Dfloorheight != DDMAXINT
							|| ld->backsector->R3Dfloorheight  != DDMAXINT)
						{
							// do nothing
						}
						else
							Settype(ld, R3D_16, 30017);
					}
					else
						Settype(ld, R3D_3F, 30017);	// default

					if (   ld->frontsector != ld->backsector
						&& IsLiquidFloorFlat(ld->frontsector)
						&& IsLiquidFloorFlat(ld->backsector)
						&& ((diff = ld->backsector->floorheight - ld->frontsector->floorheight)
						|| ld->backsector->ceilingheight != ld->frontsector->ceilingheight))
					{
						if (diff)
						{
							if (   (diff < 0 && !(ld->backsector->specFlags  & mask) == mask)
								|| (diff > 0 && !(ld->frontsector->specFlags & mask) == mask) )
							{
								Settype(ld, R3D_16, 30018);
							}
							else
							if (   (diff < 0
								&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
								&& ld->backsector->specFlags  == 0x1C)
								|| (diff > 0
								&& !(ld->backsector->specFlags & SF_MOVSECTOR)
								&& ld->frontsector->specFlags == 0x1C) )
							{
								Settype(ld, R3D_16, 30018);
							}
							else
								Settype(ld, R3D_00, 30019);
						}
						else
							Settype(ld, R3D_16, 30020);
					}

					// added rev91
					if (!lineflags[i]
						&& ld->frontsector->heightsec < 0
						&& ld->backsector->heightsec < 0
						// Next test added v2.1.00 build_53
						&& ld->frontsector->ceilingheight - ld->frontsector->floorheight > 3*FRACUNIT
						&& ld->backsector->ceilingheight - ld->backsector->floorheight > 3*FRACUNIT)
					{
						if (ld->frontsector->specFlags & SF_NOLOTEX &&
							ld->frontsector->specFlags & SF_SECCLOSED &&
							ld->frontsector->specFlags & SF_MOVFLOOR &&
							!(ld->frontsector->specFlags & SF_MOVCEIL) &&
							!(ld->backsector->specFlags & SF_MOVFLOOR))
						{
							Settype(ld, R3D_16, 30023);
						}
						else if (ld->backsector->specFlags & SF_NOLOTEX &&
							ld->backsector->specFlags & SF_SECCLOSED &&
							ld->backsector->specFlags & SF_MOVFLOOR &&
							!(ld->backsector->specFlags & SF_MOVCEIL) &&
							!(ld->frontsector->specFlags & SF_MOVFLOOR))
						{
							Settype(ld, R3D_16, 30024);
						}
						else if (vtxflags[ld->v1 - vertexes]
							&& vtxflags[ld->v2 - vertexes]
							&& vtxflags[ld->v1 - vertexes] != vtxflags[ld->v2 - vertexes])
						{
							Settype(ld, R3D_00, 30025);
						}
					}
				}
				break;

			case R3D_3E:
				if (!(txn & 0x22))
				{
					if (!txn && movef &&
						ld->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2))
					{
						break;
					}

					if (!txn && Vtfl3(ld) && LS_TestLineTex(ld) & LT_MISHI)
						Settype(ld, R3D_1A, 30026);
					else
						Settype(ld, R3D_00, 30027);
					break;
				}
			case R3D_04: 
			case R3D_05:
			case R3D_06:
			case R3D_0E:
			case R3D_3D:
				// v220-16 deal with asymetric heightsec lines
				if (   txn & 0x22
					&& ((ld->frontsector->heightsec >= 0 && ld->backsector->heightsec  < 0)
					||  (ld->frontsector->heightsec  < 0 && ld->backsector->heightsec >= 0)) )
				{
					Forcetype(ld, R3D_03, 30028);
					break;
				}
				if (txn == 0x02 && moveb)
				{
					Settype(ld, R3D_1D, 30028);
					break;
				}
				if (txn == 0x20 && movef)
				{
					Settype(ld, R3D_1D, 30029);
					break;
				}
				if (txn & 0x22)
				{
					if (   movef
						&& ld->frontsector == ld->backsector
						&& ld->frontsector->heightsec >= 0
						&& ld->frontsector->ceilingheight
						== ld->frontsector->floorheight
						&& ld->frontsector->specFlags)
					{
						Settype(ld, R3D_11, 30030);
					}
					else
					{
						if (   txn == 0x22
							&& noBackHeight(ld))
						{
							Settype(ld, R3D_1D, 30031);
						}
						// v220-3 code to deal with lines backing a movceil sector with ill-constructed maps
						else if (!(txn & 0x55)
							&& (ld->frontsector->specFlags & SF_MOVCEIL)^(ld->backsector->specFlags & SF_MOVCEIL)
							&& ld->frontsector->floorheight   != ld->backsector->floorheight
							&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight
							&& Spans1sLines(ld))
						{
							if ((ld->frontsector->specFlags & SF_MOVSECTOR) == SF_MOVCEIL)
							{
								Settype(ld, R3D_00, 30031);
								sides[ld->sidenum[1]].bottomtexture = sides[ld->sidenum[1]].midtexture;
								ld->ml_flags |= ML_DONTPEGBOTTOM;
							}
							else
							{
								sides[ld->sidenum[0]].bottomtexture = sides[ld->sidenum[0]].midtexture;
								Settype(ld, R3D_00, 30031);
								ld->ml_flags |= ML_DONTPEGBOTTOM;
							}
						}
						// v220-9 test added
						else if ((ld->mtflags == 0x10 && (ld->backsector->specFlags  & 0xFFFFC) == 0x0009809C)
							||   (ld->mtflags == 0x40 && (ld->frontsector->specFlags & 0xFFFFC) == 0x0009809C))
						{
							Settype(ld, R3D_1D, 30032);
						}
						// v220-4 determine whether type should be R3D_03 or R3D_1D
						else
						{
							int		diff;
							byte	type = R3D_03;
							int		n, h1, h2;

							checkR3D_03Type(ld);

							// v220-4 take fake sector heights into account
							h1 = (n = ld->frontsector->heightsec) >= 0 ? sectors[n].floorheight : ld->frontsector->floorheight;
							h2 = (n = ld->backsector->heightsec ) >= 0 ? sectors[n].floorheight : ld->backsector->floorheight;

							diff = h1 - h2;

							if (diff)
							{
								if (diff < 0)
								{
									if (ld->frontsector->specFlags & SF_NOLOTEX)
									{
										type = ld->frontsector->heightsec >= 0 && ld->ml_flags & ML_DONTPEGBOTTOM ? R3D_03 : R3D_1D;	// v220-24
									}
								}
								else
								{
									if (ld->backsector->specFlags & SF_NOLOTEX) type = R3D_1D;
								}

								if (   type != R3D_1D && !movef && !moveb
									&& ((origTEX[i] == 0x02 && diff < 0)
									||  (origTEX[i] == 0x20 && diff > 0)) )
								{
									type = R3D_1D;
								}
							}

							Settype(ld, type, 30032);	// R3D_03 or R3D_1D

							if (type == R3D_1D)
							{
								// v220-7 if setting to R3D_1D ensure it is not changed back
								SetLFG(ld, LFG_FIXED, 30032);
							}
						}
					}
					break;
				}
				if (!txn)
				{
					if (moveb ^ movef
						&& Vtxs2(ld))
						Settype(ld, R3D_14, 30032);
					else
						Settype(ld, R3D_01, 30033);
					break;
				}
				break;

			case R3D_1A:
				if (!txn &&
					ld->mtflags & R3D_MISHITEX &&
					ld->mtflags & R3D_MISLOTEX)
				{
					Settype(ld, R3D_00, 30034);
					break;
				}
				if (ld->mtflags & R3D_MISHITEX &&
					!movef && !moveb &&
					!Vtfl3(ld) &&
					abs(ld->frontsector->ceilingheight -
						ld->backsector->ceilingheight) <= 8*FRACUNIT)
				{
					Settype(ld, R3D_00, 30035);
					break;
				}
				// Fall through
			case R3D_30:
			case R3D_1E:
				// Ammended Jun 2004 - now tests movsec one side only.
				if (!(txn == 0x20 && !movef && moveb && Vtfl3(ld) &&
					  ld->mtflags & (R3D_MISHTEX1|R3D_MISLTEX1)))
				{
					if (!(!movef && !moveb &&
						ld->backsector->ceilingheight >
						ld->frontsector->ceilingheight &&
						ld->backsector->floorheight <
						ld->frontsector->floorheight))
					{
						if (origTEX[i])
							Settype(ld, R3D_1A, 30037);
					}
					else
						Settype(ld, R3D_00, 30038);
					break;
				}
				if (!(txn == 0x02 && movef && !moveb && Vtfl3(ld) &&
					  ld->mtflags & (R3D_MISHTEX2|R3D_MISLTEX2)))
				{
					if (!(!movef && !moveb &&
						ld->backsector->ceilingheight <
						ld->backsector->ceilingheight &&
						ld->backsector->floorheight >
						ld->frontsector->floorheight))
						Settype(ld, R3D_1A, 30039);
					else
						Settype(ld, R3D_00, 30040);
					break;
				}
				Settype(ld, R3D_00, 30041);
				break;

			case R3D_2B:
				Settype(ld, R3D_2A, 30042);
				break;

			default:
				break;
		}

		// 1) Set up ceildiff values for static sectors with missing tex's
		// 2) Set lines requiring a skyflat to be drawn
		//    Test near end will reset line to R3D_00 if conditions are not met
		// Fake sectors must be screened out to prevent occlusion problems
		if (ld->mtflags & R3D_MISHITEX

			// $change 1 low tex test added - rev170f
			&& !(origTEX[i] & 0x33)
			// next test belt and braces only (tex 0x33 test should be enough)
			&&  !((ld->frontsector->ceilingpic == skyflatnum)
			^     (ld->backsector->ceilingpic == skyflatnum))
			&& !movef && !moveb
			&& ld->frontsector->heightsec < 0
			&& ld->backsector->heightsec < 0
			&& !(lineflags[i] & (LFG_NHCONNOBSLN|LFG_3DTYPE))
			&& !Spans1sLines(ld)
			&& vtxflags[ld->v1 - vertexes] != 3
			&& vtxflags[ld->v2 - vertexes] != 3
			&& vtxarray[ld->v1 - vertexes] > 2
			&& vtxarray[ld->v2 - vertexes] > 2)
		{
			if (ld->mtflags & R3D_MISHTEX1
				&& ld->frontsector->floorheight <= ld->backsector->floorheight)
			{
				ld->frontsector->ceildiff =
				ld->frontsector->ceilingheight -
				ld->backsector->ceilingheight;
				linearray[c++] = i;
				Settype(ld, R3D_22, 30043);
				hitline[i] = 2;
			}
			else if (ld->mtflags & R3D_MISHTEX2
				&& ld->frontsector->floorheight >= ld->backsector->floorheight)
			{
				ld->backsector->ceildiff =
				ld->backsector->ceilingheight -
				ld->frontsector->ceilingheight;
				linearray[c++] = i;
				Settype(ld, R3D_22, 30044);
				hitline[i] = 2;
			}
		}

		if (ld->specType == R3D_1A)
		{
			if (txn & 0x04 && txn & 0x40)
			{
				Settype(ld, R3D_00, 30045);
			}
			else if (txn & 0x01 && txn & 0x10)
			{
				Settype(ld, R3D_00, 30046);
			}
			else if (ld->mtflags & R3D_MISLOTEX &&
				ld->mtflags & R3D_MISHITEX)
			{
				Settype(ld, R3D_00, 30047);
			}
			continue;
		}

		// DOORS
		if (ld->mtflags == R3D_MISHTEX2 &&
			txn == 0x01 && Vtfl3(ld) &&
			!(ld->frontsector->specFlags & (SF_MOVSECTOR|SF_LINESCLOSED)) &&
			!(ld->backsector->specFlags & (SF_MOVSECTOR|SF_LINESCLOSED)))
		{
			Settype(ld, R3D_29, 30048);
		}
		else if (ld->frontsector != ld->backsector &&
			(txn == 0x22 || txn == 0x02 || txn == 0x20) &&

			// v220-5 replaced movef && moveb with ceil tests
			ld->frontsector->specFlags & SF_MOVCEIL &&
			ld->backsector->specFlags & SF_MOVCEIL &&

			ld->frontsector->R3Dfloorheight == DDMAXINT &&
			ld->backsector->R3Dfloorheight == DDMAXINT &&
			vtxarray[ld->v1 - vertexes] == 3 &&
			vtxarray[ld->v2 - vertexes] == 3 &&
			ld->frontsector->ceilingheight ==
			ld->backsector->ceilingheight &&
			ld->frontsector->floorheight ==
			ld->backsector->floorheight &&
			SecNoMoveLines(ld, 0) &&
			SecNoMoveLines(ld, 1))
		{
			// v220-5 added test
			if (   ld->frontsector->floorheight < ld->frontsector->ceilingheight
				&& ld->backsector->floorheight < ld->backsector->ceilingheight)
			{
				Setr3dfh(ld->frontsector, ld->frontsector, 30049);
				Setr3dfh(ld->backsector, ld->backsector, 30050);
				Setr3dch(ld->frontsector, ld->frontsector, 30051);
				Setr3dch(ld->backsector, ld->backsector, 30052);
			}

			Settype(ld, R3D_03, 30053);
			Setspec(ld, -1, 30054);
		}
		else if (ld->mtflags & R3D_MISHITEX &&
			movef ^ moveb &&
			ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0 &&
			!ChkLptr(ld) &&
			(txn == 0x02 ||
			 txn == 0x20 ||
			 txn == 0x03 ||
			 txn == 0x23 ||
			 txn == 0x33))
		{
			int txmid = sides[ld->sidenum[0]].midtexture;

			if (movef && !moveb)
			{
				if (ld->frontsector->floorheight ==
					ld->frontsector->ceilingheight &&
					ld->frontsector->floorheight & FRACUNIT)
				{
					Settype(ld, R3D_1D, 30055);
					Setspec(ld, ld->backsector - sectors, 30056);
				}
				else if (ld->frontsector->ceilingheight -
					ld->frontsector->floorheight <= FRACUNIT)
				{
					int secnum = LS_FindAdjacentSecnum(ld, 1);
					if (secnum < 0) secnum = ld->backsector - sectors;
					if (sectors[secnum].ceilingheight <=
						ld->frontsector->ceilingheight)
					{
						Settype(ld, R3D_1D, 30057);
						Setspec(ld, secnum, 30058);
					}
				}
			}
			else if (!movef && moveb && ld->specType != R3D_2D)
			{
				if (ld->backsector->floorheight ==
					ld->backsector->ceilingheight &&
					ld->backsector->floorheight & FRACUNIT)
				{
					int secnum = subfn_CheckR3D_1DSec(ld, 0);
					Settype(ld, R3D_1D, 30059);
					Setspec(ld, secnum, 30060);
					SetNOocclude(ld->frontsector, 30059);
					SetNOocclude(ld->backsector , 30060);
				}
				else if (ld->backsector->ceilingheight -
					ld->backsector->floorheight <= FRACUNIT)
				{
					int secnum = LS_FindAdjacentSecnum(ld, 1);
					if (secnum < 0) secnum = ld->frontsector - sectors;
					if (sectors[secnum].ceilingheight <=
						ld->backsector->ceilingheight)
					{
						Settype(ld, R3D_1D, 30061);
						Setspec(ld, secnum, 30062);
						SetNOocclude(ld->frontsector, 30061);
						SetNOocclude(ld->backsector , 30062);
					}
				}
			}
		}
	}

	for (i = 0; i < c; i++)
	{
		ld = &lines[linearray[i]];

		// If ceildiff values were found then a line's
		// TYF_SKYWALL flag may need to be set.
		if (!origTEX[linearray[i]])
		{
			sector_t *f = ld->frontsector;
			sector_t *b = ld->backsector;
			RDBOOL fs = f->ceilingpic == skyflatnum;
			RDBOOL bs = b->ceilingpic == skyflatnum;
			RDBOOL fh = f->ceilingheight - f->floorheight;
			RDBOOL bh = b->ceilingheight - b->floorheight;

			if (   (fs && !bs && fh <= 0 && bh > 0)
				|| (bs && !fs && bh <= 0 && fh > 0))
			{
				SetSkywall(ld, 30063);
				Settype(ld, R3D_01, 30063);
				hitline[linearray[i]] |= 4;
				continue;
			}
		}

		// Should really check for a closed line loop
		// but this will have to suffice for the moment.
		if (hitline[linearray[i]] == 1)	// $v73 - hitline test introduced (was count)
		{
			subsector_t	*ss;
			sector_t	*sec = ld->frontsector, *osec = NULL;
			seg_t		*seg;
			int			k, m;

			// First find the other sector
			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->ml_flags & ML_TWOSIDED &&
					sec->lines[k]->mtflags & R3D_MISHITEX)
				{
					osec = sec->lines[k]->frontsector == ld->frontsector ?
						sec->lines[k]->backsector : sec->lines[k]->frontsector;

					break;
				}
			}

			if (  !osec
				|| osec->specFlags & SF_MOVSECTOR
				|| osec->ceilingpic == skyflatnum
				|| osec->ceilingheight >= sec->ceilingheight)
			{
				continue;
			}

			for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
			{
				line_t *lc;

				seg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, seg++)
				{
					if ((lc = seg->linedef) != NULL &&
						lc == ld &&
						!seg->sidedef->toptexture)
					{
						Setr3dsscs(ss, osec - sectors, 30064);
						hitline[linearray[i]] |= 4;
						break;
					}
				}
			}
		}
	}

	// v220-10 reset lines that don't meet the required conditions
	for (i = 0; i < c; i++)
	{
		if (!(hitline[linearray[i]] & 4))
		{
			Settype(&lines[linearray[i]], R3D_00, 30065);
		}
	}
}