//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_ctrlines.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
//
// DESCRIPTION: attempts to find the lines in a sectors that are used
//              to pass control values (e.g. height to raise plat to)
//				or for 'tricks' in general that it was discovered that
//				the doom engine can be fooled into doing (e.g. floor
//				over floor).
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

#define		MRK_ONESIDED	1
#define		MRK_SAMESEC		2
#define		MRK_IGNORE		3

static int		*tempsseclist;
static byte		*excludessecs;
static int		*ringtmp;
static int		*linelist;

//==========================================================

static int checkForRingInList(byte *lnarray, int ringcount, int *ringlist, int lnum, int *bbox)
{
	line_t		*ld = &lines[lnum];
	vertex_t	*v1 = ld->v1;
	RDBOOL		isclosed = false, repeat = true;
	int			c, i;

	lnarray[lnum] = 2;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < ringcount; i++)
		{
			int k = ringlist[i];

			if (lnarray[k] == 1)
			{
				line_t	*li = &lines[k];

				if (li->v1 == v1)
				{
					v1 = li->v2; lnarray[k] = 2; repeat = true;
				}
				else if (li->v2 == v1)
				{
					v1 = li->v1; lnarray[k] = 2; repeat = true;
				}
			}
		}
	}

	isclosed = (v1 == ld->v2);

	if (isclosed)
	{
		M_ClearBox(bbox);
	}
	
	for (c = i = 0; i < ringcount; i++)
	{
		if (lnarray[ringlist[i]] == 2)
		{
			if (isclosed)
			{
				lnarray[ringlist[i]] = 4; c++;
				M_AddLineToBox(bbox, &lines[ringlist[i]]);
			}
			else
			{
				lnarray[ringlist[i]] = 1;
			}
		}
	}

	return c;
}

static void checkSUBSSECS(sector_t *sec, int ringtmpcnt)
{
	int		i, k, tempptr, bbox[4];

	if (ringtmpcnt < 3) return;

	memcpy(tempsseclist, hitssec, hitssecptr * sizeof(int));
	tempptr = hitssecptr;

	CLR_hitline();

	for (i = 0; i < ringtmpcnt; i++) hitline[ringtmp[i]] = 1;

	for (i = 0; i < ringtmpcnt; i++)
	{
		if (   hitline[ringtmp[i]] == 1
			&& checkForRingInList(hitline, ringtmpcnt, ringtmp, ringtmp[i], bbox))
		{
			int		flag = MSIA_NONE;
			int		count;

			for (count = k = 0; k < ringtmpcnt; k++)
			{
				if (hitline[ringtmp[k]] == 4) linelist[count++] = ringtmp[k];
			}

			if (!LS_MarkSubsecsInLoop(sec, count, linelist, bbox, MSIA_TESTAREA, NULL))
				flag = MSIA_REVERSED;

			// This should not fail
			if (LS_MarkSubsecsInLoop(sec, count, linelist, bbox, flag, NULL))
			{
				for (k = 0; k < hitssecptr; k++)
					excludessecs[hitssec[k]] = true;				
			}

			for (k = 0; k < ringtmpcnt; k++)
			{
				if (hitline[ringtmp[k]] == 4)
				{
					hitline[ringtmp[k]] = 8;
				}
			}
		}
	}

	hitssecptr = tempptr;
	memcpy(hitssec, tempsseclist, hitssecptr * sizeof(int));
}

static RDBOOL chkSSEC(subsector_t *ss, sector_t *s)
{
	sector_t	*sec = NULL;
	seg_t		*sg;
	int			c, i;

	if (ss->linecount < 3 || !(ss->sector->specFlags & SF_MOVFLOOR)) return false;

	for (c = i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (   sg->linedef
			&& sg->linedef->frontsector == sg->linedef->backsector)
		{
			if (sg->linedef->frontsector == ss->sector)
			{
				if (sec || !(sg->linedef->lnflags2 & LNF2_OWNBOTHGTRTWO))
				{
					return false;
				}

				sec = ss->sector;
				continue;
			}

			if (sg->linedef->frontsector->specFlags & SF_MOVFLOOR)
				return false;
		}

		if (sg->frontsector == s) c++; else return false;
	}

	return (c == ss->linecount - 1);
}

void LS_PresetContainedSSLines()
{
	sector_t	*s, *fs, *bs;
	line_t		*ld;
	byte		*mrkline, *lnarray, *mrksect;

	int			c, i, k, m, rc, bbox[4];

	//--------------------------------------------------

	if (numsectors < 3) return;

	mrkline = (byte*)W_Calloc(numlines  , TEMPID, 0);
	lnarray = (byte*)W_Calloc(numlines  , TEMPID, 0);
	mrksect = (byte*)W_Calloc(numsectors, TEMPID, 0);

	validcount++;
	ss_validcount++;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   lineflags[i] & (LFG_NEWSET|LFG_3DTYPE)
			|| lxarray[i]
			|| ld->spflags & SPF_CONTROLLINE
			|| ld->lnflags2 & LNF2_SLOPESECLN)
		{
			mrkline[i] = (byte)MRK_IGNORE;
			continue;
		}

		if ( !(ld->ml_flags & ML_TWOSIDED) )
		{
			mrkline[i] = (byte)MRK_ONESIDED;
		}
		else if (ld->frontsector == ld->backsector)
		{
			mrkline[i] = (byte)MRK_SAMESEC;
		}
		else if (Vtfl(ld)
			|| !(ld->lnflags2 & LNF2_OWNBOTHGTRONE))
		{
			mrkline[i] = (byte)MRK_IGNORE;
		}
	}

	for (c = i = 0; i < numlines; i++)
	{
		if (mrkline[i] == MRK_IGNORE) continue;

		if (mrkline[i] == MRK_ONESIDED)
		{
			lnarray[i] = 1; continue;
		}

		if (!mrkline[i])
		{
			lnarray[i] = 1; c++; continue;
		}

		if (lineflags[i]) continue;
		if (mrkline[i] != MRK_SAMESEC) continue;
		if (!(lines[i].frontsector->specFlags & SF_MOVFLOOR)) continue;

		if (origTEX[i] == 0x22)
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					seg_t *sg = &segs[ss->firstline];

					ss->ssvalidcount = ss_validcount;

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						ld = sg->linedef;

						if (ld && ld->validcount != validcount)
						{
							ld->validcount = validcount;

							if (!mrkline[ld - lines])
							{
								mrksect[ld->frontsector - sectors] = true;
								mrksect[ld->backsector  - sectors] = true;
							}
						}
					}
				}
			}
		}
	}

	if (c < 3)
	{
		W_FreeALL(TEMPID); return;
	}

	linelist     = (int *)W_Malloc(maxsecsz*sizeof(int), TEMPID, 12);
	ringtmp      = (int *)W_Malloc(maxsecsz*sizeof(int), TEMPID, 13);
	excludessecs = (byte*)W_Malloc(numsubsectors, TEMPID, 14);
	tempsseclist = (int *)W_Malloc(numsubsectors*sizeof(int), TEMPID, 15);

	for (i = 0; i < numlines; i++)
	{
		if (lnarray[i] == 1)
		{
			fs = lines[i].frontsector, bs = lines[i].backsector;

			if (!bs)
			{
				s = fs;
			}
			else
			if (fs->floorheight != bs->floorheight)
			{
				s = fs->floorheight < bs->floorheight ? fs : bs;
			}
			else
			if (fs->ceilingheight != bs->ceilingheight)
			{
				s = fs->ceilingheight > bs->ceilingheight ? fs : bs;
			}
			else
				continue;

			if (   mrksect[s - sectors]
				&& subfn_CheckForRing(lnarray, s, i, &rc, bbox, NULL, NULL))
			{
				for (c = k = 0; k < s->linecount; k++)
				{
					int lnum = s->lines[k] - lines;

					if (lnarray[lnum] == 4)
					{
						int n = lines[lnum].frontsector == s ? 0 : 1;

						linearray[c] = lnum;
						sidelist[c++] = lines[lnum].sidenum[n];
					}
				}

				if (LS_MarkSubsecsInMultiSecLoop(0, NULL, c, bbox, NULL, true, MSIA_NONE))
				{
					int		ringtmpcnt = 0;
					RDBOOL	noskyfix = false;
					RDBOOL	islx = false;

					memset(excludessecs, 0, numsubsectors);

					validcount++;

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						if (   !noskyfix
							&& ss->sector->ceilingpic == skyflatnum
							&& ss->sector->specFlags & SF_MOVCEIL)
						{
							noskyfix = true;
						}

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ( (ld = sg->linedef) && ld->validcount != validcount)
							{
								ld->validcount = validcount;

								fs = ld->frontsector, bs = ld->backsector;

								if (lnarray[ld - lines] == 4 || fs == bs) continue;

								if (fs == s || bs == s)
									ringtmp[ringtmpcnt++] = ld - lines;
							}
						}
					}

					if (ringtmpcnt > 2)
						checkSUBSSECS(s, ringtmpcnt);

					// v220-21 check for special case where a same sided line not belonging to s is connected to s
					if (s->specFlags & SF_SSLINETRUE)
					{
						int	sn = s - sectors;

						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];

							if (!excludessecs[hitssec[k]])
							{
								seg_t *sg = &segs[ss->firstline];

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (  (ld = sg->linedef)
										&& ld->frontsector == ld->backsector
										&& ld->frontsector != s)
									{
										vertexowner_t	*vo;
										int				vc;

										for (vo = &vertexowners[ld->v1 - vertexes], vc = 0; vc < vo->num; vc++)
										{
											if (vo->list[vc] == sn) {islx = true; goto found_islx; }
										}
										for (vo = &vertexowners[ld->v2 - vertexes], vc = 0; vc < vo->num; vc++)
										{
											if (vo->list[vc] == sn) {islx = true;  goto found_islx; }
										}
									}
								}
							}
						}
found_islx:;
					}

					validcount++;

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						if (!excludessecs[hitssec[k]])
						{
							seg_t *sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( (ld = sg->linedef) )
								{
									if (mrkline[ld - lines] == MRK_SAMESEC)
									{
										if (ss->sector != s)
										{
											// v220-3
											if (chkSSEC(ss, s))
											{
												ChangeSSECsec(ss, s, 3001);
											}
											else if (ss->sector != s)
											{
												// v220-20 add height tests for ceilsec and floorsec
												if (   ss->ceilsec < 0
													&& (islx 
													|| ss->sector->ceilingheight <= s->ceilingheight
													|| (s->specFlags & (SF_MOVCEIL|SF_NOHITEX)) == SF_MOVCEIL) )
												{
													Setr3dsscs(ss, s - sectors, 3001);
													if (noskyfix) ss->ssflags |= SSF_NOSKYFIX;
												}

												if (   ss->floorsec < 0
													&& (islx
													|| ss->sector->floorheight >= s->floorheight
													|| (s->specFlags & (SF_MOVFLOOR|SF_NOLOTEX)) == SF_MOVFLOOR) )
												{
													Setr3dssfs(ss, s - sectors, 3001);
												}
											}
										}

										if (ld->validcount != validcount)
										{
											ld->validcount = validcount;

											if (origTEX[ld - lines])
												Settype(ld, R3D_03, 3002);
											else
												Settype(ld, R3D_3F, 3003);

											SetLFG(ld, LFG_FIXED|LFG_NEWSET, 3004);
										}
									}
								}
							}
						}
					}
				}

				for (k = 0; k < s->linecount; k++)
				{
					int lnum = s->lines[k] - lines;

					if (lnarray[lnum] == 4) lnarray[lnum] = 8;
				}
			}
		}
	}

	W_FreeALL(TEMPID);
}
