//-----------------------------------------------------------------------------
//
// Copyright (C) 2016 Graham Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_dynlights.cpp
//
// First release 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
//
// DESCRIPTION:
//  Implements active dynlights and extends types (from the initial
//  type 5000 introduced by JK in his Doomsday port).
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "con_console.h"
#include "m_misc.h"
#include "p_data.h"
#include "p_maputil.h"
#include "doom/p_spec.h"
#include "r3d_utils.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "sys_system.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

dynphase_t	dynphase;
trig5505_t	trig5505;
secmo_t		*secmo;

int		 maxcount5505;
int		 maxmobjcount5505;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static thinker_t	**newthinkers;
static sector_t		**seclist5505;
static mobj_t		**mobjlist5505;
static int			newmaxcount = 0;
static int			countmobj5505;
static int			count5505;

typedef struct {
	mobj_t	*mo;
	int		lnum;
	short	onfloor;
	short	onceil;
} dynzlist_t;

static dynzlist_t	*dynzlist;
static int			dzcount;
static int			maxdzcount;

// CODE --------------------------------------------------------------------

// THINKERS ----------------------------------------------------------------

// Analogue for T_Flicker
void T_DynFlicker(dynstrobe_t* flash)
{
    if (--flash->count) return;
	
    int	amount = (M_RandomD() & 3) * 16;
    
    if (flash->brighttime - amount < flash->darktime)
		flash->mo->flags2 &= ~MF2_NORENDER;
    else
		flash->mo->flags2 |= MF2_NORENDER;
	
    flash->count = 4;
}

// Analogue for T_LightFlash
void T_DynLightFlash(dynstrobe_t* flash)
{
    if (--flash->count) return;
	
    if (flash->mo->flags2 & MF2_NORENDER)
    {
		flash->mo->flags2 &= ~MF2_NORENDER;
		flash->count = (M_RandomD() & flash->brighttime) + 1;
    }
    else
    {
		flash->mo->flags2 |= MF2_NORENDER;
		flash->count = (M_RandomD() & flash->darktime) + 1;
    }
	
}

// Analogue for T_StrobeFlash
void T_DynStrobeFlash(dynstrobe_t* flash)
{
    if (--flash->count) return;
	
    if (flash->mo->flags2 & MF2_NORENDER)
    {
		flash->mo->flags2 &= ~MF2_NORENDER;
		flash->count = flash->brighttime;
    }
    else
    {
		flash->mo->flags2 |= MF2_NORENDER;
		flash->count = flash->darktime;
    }
}

// Replacement for T_Glow
void T_DynGlow(dynglow_t* dg)
{
	switch(dg->direction)
	{
	case -1:
		// DOWN
		dg->lightlevel -= GLOWSPEED;

		if (dg->lightlevel <= dg->minlight)
		{
			dg->lightlevel += GLOWSPEED;
			dg->direction = 1;
		}
		break;
		
	case 1:
		// UP
		dg->lightlevel += GLOWSPEED;

		if (dg->lightlevel >= dg->maxlight)
		{
			dg->lightlevel -= GLOWSPEED;
			dg->direction = -1;
		}
		break;
	}

	// copy the new lightlevel to each 5505 mobj in the sector
	for (int i = 0; i < dg->listcount; i++)
	{
		dg->list[i]->torchtic = dg->lightlevel; // torchtic is otherwise unused
	}
}

// Replacement for T_FireFlicker
void T_DynFireFlicker(dynfireflicker_t* dflick)
{
    if (--dflick->count) return;
	
    int amount = (M_RandomD() & 3) * 16;
    
    if (dflick->lightlevel - amount < dflick->minlight)
		dflick->lightlevel = dflick->minlight;
    else
		dflick->lightlevel = dflick->maxlight - amount;

	// copy the new lightlevel to each 5505 mobj in the sector
	for (int i = 0; i < dflick->listcount; i++)
	{
		dflick->list[i]->torchtic = dflick->lightlevel; // torchtic is otherwise unused
	}
	
    dflick->count = 4;
}

//------------------------------------------------------------------------------

// Called as part of the load game process.
// Mobj information mangled when the Thinkers are respawned is regenerated.
// The phase is not preserved but this hardly matters for these types.
static void P_RegenerateDynLists()
{
	thinker_t	*fnthinker = fnthinkcap.next;
	int			c, i;

	while (fnthinker != &fnthinkcap)
	{
		thinker_t *next = fnthinker->next;

		if (fnthinker->function == (void (__cdecl *)(void))T_DynGlow)
		{
			dynglow_t *dg = (dynglow_t*)fnthinker;

			// mark all 5505 types in this particular sector
			for (c = i = 0; i < countmobj5505; i++)
			{
				if (mobjlist5505[i]->subsector->sector - sectors == dg->secn)
				{
					if (c < MAXMOBJLIST)
						dg->list[c++] = mobjlist5505[i];
					else
						goto consistency_error;	// should never happen
				}
			}

			if (dg->listcount != c) goto consistency_error;
		}
		else
		if (fnthinker->function == (void (__cdecl *)(void))T_DynFireFlicker)
		{
			dynfireflicker_t *dflick = (dynfireflicker_t*)fnthinker;

			// mark all 5505 types in this particular sector
			for (c = i = 0; i < countmobj5505; i++)
			{
				if (mobjlist5505[i]->subsector->sector - sectors == dflick->secn)
				{
					if (c < MAXMOBJLIST)
						dflick->list[c++] = mobjlist5505[i];
					else
						goto consistency_error;	// should never happen
				}
			}

			if (dflick->listcount != c) goto consistency_error;
		}

		fnthinker = next;
	}
	return;

consistency_error:;
	Con_Error("Stored savegame mobj count does not match regenerated count\n");
}

static void P_SpawnDynFlash(mobj_t *mo)
{
	dynstrobe_t *flash;

	flash = (dynstrobe_t*)W_Calloc(sizeof(*flash), THINKID, 0);
	flash->mo = mo;
	
	if (mo->type == MT_STROBE1)
	{
		flash->thinker.function = (void (__cdecl *)(void))T_DynLightFlash;
		flash->darktime = 64, flash->brighttime = 7,

		flash->count = 1;
	}
	else
	if (mo->type == MT_STROBE2)
	{
		int inSync = ( mo->angle == STROBE_N
					|| mo->angle == STROBE_E
					|| mo->angle == STROBE_S
					|| mo->angle == STROBE_W);

		flash->thinker.function = (void (__cdecl *)(void))T_DynStrobeFlash;
		flash->darktime = 15, flash->brighttime = 5,

		flash->count = inSync ? 1 : (M_RandomD() & 7) + 1;
	}
	else
	if (mo->type == MT_STROBE3)
	{
		int inSync = ( mo->angle == STROBE_N
					|| mo->angle == STROBE_E
					|| mo->angle == STROBE_S
					|| mo->angle == STROBE_W);

		flash->thinker.function = (void (__cdecl *)(void))T_DynStrobeFlash;
		flash->darktime = 35, flash->brighttime = 5,

		flash->count = inSync ? 1 : (M_RandomD() & 7) + 1;
	}
	else
	if (mo->type == MT_STROBE4)
	{
		flash->thinker.function = (void (__cdecl *)(void))T_DynFlicker;
		flash->darktime = mo->subsector->sector->lightlevel + 16;;
		// since getting a max sector lighting is not possible add 32 to the darktime
		flash->brighttime = flash->darktime + 32;
		flash->count = 4;
	}
	else
	{
		Con_Error("Program error: P_SpawnDynFlash");
	}
	
	P_AddFNThinker (&flash->thinker);
}

// Collates any 5505 Things into their own structure
void P_CreateDynLightStructures(RDBOOL regenerate)
{
	thinker_t	*it;
	int			i;

	memset(&dynphase, 0, sizeof(dynphase_t));
	count5505 = countmobj5505 = 0;
	for (i = 0; i < trig5505.count; i++) secmo[i].count = 0;
	trig5505.count = 0;

	if (!R3D_lightextns) return;

	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*iter = (mobj_t*)it;

			// Note: STROBE5 has no dedicated Thinker
			if (   iter->type == MT_STROBE1
				|| iter->type == MT_STROBE2
				|| iter->type == MT_STROBE3
				|| iter->type == MT_STROBE4)
			{
				P_SpawnDynFlash(iter);
			}
			else if (iter->type == MT_STROBEX)
			{
				sector_t *s = iter->subsector->sector;

				if (s->special == 6)	// this special is not cleared
				{
					int len = -1;
					int k;

					for (i = 0; i < trig5505.count; i++)
					{
						if (trig5505.seclist[i] == s) goto add_mobj;
					}

					// set up arrays for new sector entry
					if (trig5505.maxcount <= trig5505.count)
					{
						k = (trig5505.maxcount += 16);
						trig5505.seclist = (sector_t**)W_Realloc(trig5505.seclist, k*sizeof(sector_t*), STATICID, 0);
						secmo = (secmo_t*)W_ReallocCLR(secmo, k*sizeof(secmo_t), STATICID, 0);
					}

					// check for a line special type 81
					for (k = 0; k < numlines; k++)
						if (lines[k].special == 81 && lines[k].tag == s->tag) break;

					if (k == numlines)
					{
						iter->type = MT_LIGHTSOURCE;
						iter->torchtic = 255;	// Set permanently on
						pr_OFinsert("WARNING: No trigger line special 81 found for T%i in S%i tag %i\n",
							MAPNUM(iter), s - sectors, s->tag);
					}
					else
					{
						trig5505.seclist[trig5505.count++] = s;
						len = (int)M_AccurateLength(&lines[k]);
					}

add_mobj:;
					iter->flags2 &= ~(MF2_NORENDER|MF2_CORPSEFADE|MF2_NOTLIT);

					// set up mobj list size
					if (secmo[i].maxcount <= secmo[i].count)
					{
						k = (secmo[i].maxcount += 16);
						secmo[i].list = (mobj_t**)W_Realloc(secmo[i].list, k*sizeof(mobj_t*), STATICID, 0);
					}

					secmo[i].list[secmo[i].count++] = iter;

					if (regenerate)
						secmo[i].timer = secdmo[i].timer;
					else
					{
						if (len >= 0) { secmo[i].period = len >> 1; secmo[i].timer = 0; }
						iter->torchtic = 0;	// start dark
					}
				}
				else
				{
					// Construct a sector list
					for (i = 0; i < count5505; i++)
					{
						if (seclist5505[i] == s) goto next_mobj;
					}

					if (count5505 >= maxcount5505)
					{
						maxcount5505 += 32;
						seclist5505 = (sector_t**)W_Realloc(seclist5505, sizeof(sector_t*)*maxcount5505, STATICID, 0);
					}

					seclist5505[count5505++] = s;

next_mobj:;
					// Construct a mobj list
					if (countmobj5505 >= maxmobjcount5505)
					{
						maxmobjcount5505 += 32;
						mobjlist5505 = (mobj_t**)W_Realloc(mobjlist5505, sizeof(mobj_t*)*maxmobjcount5505, STATICID, 0);
					}

					mobjlist5505[countmobj5505++] = iter;
				}
			}
		}
	}

	if (regenerate && count5505) P_RegenerateDynLists();
}

static void MapError(int secn, int type)
{
	Con_Error("Too many 5505 Things in S%i type %i (max supported is %i)\n", secn, type, MAXMOBJLIST);
	// never returns
}

// Called after standard Boom T_Glow and T_FireFlicker functions have been spawned
// If any 5505 types are found in the thinkers sectors then the thinker is replaced
// with a new type.(either T_DynGlow or T_DynFireFlicker)
void P_CheckInheritance5505()
{
	if (!countmobj5505) return;

	thinker_t	*fnthinker = fnthinkcap.next;
	int			newcount = 0;
	byte		*mark = (byte*)calloc(countmobj5505, 1);
	RDBOOL		update = false;
	int			errors = 0;
	int			c, i;

	while (fnthinker != &fnthinkcap)
	{
		thinker_t *next = fnthinker->next;

		if (fnthinker->function == (void (__cdecl *)(void))T_Glow)
		{
			glow_t *glow = (glow_t*)fnthinker;

			for (c = i = 0; i < count5505; i++)
			{
				if (glow->sector == seclist5505[i])
				{
					c++; break;
				}
			}

			if (c)	// true if a 5505 type resides in a sector with special 8;
			{
				dynglow_t *dg = (dynglow_t*)W_Calloc(sizeof(*dg), THINKID, 0);

				dg->thinker.function = (void (__cdecl *)(void))T_DynGlow;
				dg->direction = glow->direction;
				dg->minlight = glow->minlight;
				dg->maxlight = glow->maxlight;
				dg->secn = glow->sector - sectors;	// needed for savegame regeneration

				// set the sector to minlight as all lighting is now dynamic
				glow->sector->lightlevel = glow->minlight;

				// mark all 5505 types in this particular sector
				for (c = i = 0; i < countmobj5505; i++)
				{
					if (mobjlist5505[i]->subsector->sector == glow->sector)
					{
						mark[i] = true;

						if (c < MAXMOBJLIST)
							dg->list[c++] = mobjlist5505[i];
						else
						{
							free(mark);
							MapError(glow->sector - sectors, 8);
						}
					}
				}

				dg->listcount = c;

				fnthinker->function = (think_t) -1;			// mark this thinker as defunct

				if (newcount >= newmaxcount)
				{
					newmaxcount += 32;
					newthinkers = (thinker_t**)W_Realloc(newthinkers, sizeof(thinker_t*)*newmaxcount, STATICID, 301);
				}

				newthinkers[newcount++] = &dg->thinker;	// store this one to be linked in

				update = true;
			}
		}
		else
		if (fnthinker->function == (void (__cdecl *)(void))T_FireFlicker)
		{
			fireflicker_t *flick = (fireflicker_t*)fnthinker;

			for (c = i = 0; i < count5505; i++)
			{
				if (flick->sector == seclist5505[i])
				{
					c++; break;
				}
			}

			if (c)	// true if a 5505 type resides in a sector with special 17;
			{
				dynfireflicker_t *dflick = (dynfireflicker_t*)W_Calloc(sizeof(*dflick), THINKID, 0);

				dflick->thinker.function = (void (__cdecl *)(void))T_DynFireFlicker;
				dflick->minlight = flick->minlight;
				dflick->maxlight = flick->maxlight;
				dflick->count = flick->count;
				dflick->secn = flick->sector - sectors;	// needed for savegame regeneration

				// set the sector to minlight as all lighting is now dynamic
				flick->sector->lightlevel = flick->minlight - 16;

				// mark all 5505 types in this particular sector
				for (c = i = 0; i < countmobj5505; i++)
				{
					if (mobjlist5505[i]->subsector->sector == flick->sector)
					{
						mark[i] = true;

						if (c < MAXMOBJLIST)
						{
							dflick->list[c++] = mobjlist5505[i];
						}
						else
						{
							free(mark);
							MapError(flick->sector - sectors, 17);
						}
					}
				}

				dflick->listcount = c;

				fnthinker->function = (think_t) -1;			// mark this thinker as defunct

				if (newcount >= newmaxcount)
				{
					newmaxcount += 32;
					newthinkers = (thinker_t**)W_Realloc(newthinkers, sizeof(thinker_t*)*newmaxcount, STATICID, 301);
				}

				newthinkers[newcount++] = &dflick->thinker;	// store this one to be linked in

				update = true;
			}
		}

		fnthinker = next;
	}

	if (update)
	{
		fnthinker = fnthinkcap.next;
		
		while (fnthinker != &fnthinkcap)
		{
			thinker_t *next = fnthinker->next;

			if (fnthinker->function == (think_t)-1)
			{
				// Time to remove it
				fnthinker->next->prev = fnthinker->prev;
				fnthinker->prev->next = next;

				W_Free(fnthinker, THINKID, 509);
			}

			fnthinker = next;
		}

		for (i = 0; i < newcount; i++)
			P_AddFNThinker(newthinkers[i]);
	}

	for (i = 0; i < countmobj5505; i++)
	{
		if (!mark[i])
		{
			PrintProgInfo("MAP ERROR: T%i is a 5505 type that is not in a valid sector\n", MAPNUM(mobjlist5505[i]));
			errors++;
		}
	}

	free(mark);

	if (errors)
	{
		Sys_MessageBox("MAP ERROR: 5505 sector errors were found\n"
				 "See R3Ddebug.out for lines having faults", false);
		Sys_Quit(false);
	}
}

static int subfn_checkThingLine(mobj_t *mo)
{
	sector_t	*sc = mo->subsector->sector;
	line_t		*ld;
	int			i;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (  (ld->frontsector == ld->backsector)
			&& XYonLine(mo->x, mo->y, ld, 1))
		{
			return ld - lines;
		}
	}

	return -1;
}

// Dyn lights that potentially need z heights to be changed.
//
// These are not dealt with by the usual functions and are hooked, immutably,
// to their x, y values  As such it is just their z heights that may need to
// be altered. To save trawling the whole of the mobj thinkers each game tic
// a list of any that *may* need to be checked is created here.

void P_SetupDynlightList()
{
	thinker_t	*it;
	mobj_t		*mo;

	dzcount = 0;

	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (   it->function == (void (__cdecl *)(void))P_MobjThinker
			&& (mo = (mobj_t*)it)->flags3 & MF3_DYNLIGHT)
		{
			sector_t	*sc = mo->subsector->sector;
			int			lnum = -1;
			RDBOOL		onfloor = false, onceil = false;

			if (R3D_lightextns)	// For backward compatibility with Things 5000
			{
				if (mo->angle == STROBE_N || mo->angle == STROBE_NE)
				{
					mo->floorz = SL_PointHeight(sc, mo->x, mo->y, LOSLOPE);
					mo->ceilingz = DDMININT;
					onfloor = true;
				}
				else
				if (mo->angle == STROBE_E || mo->angle == STROBE_SE)
				{
					mo->floorz = DDMAXINT;
					mo->ceilingz = SL_PointHeight(sc, mo->x, mo->y, HISLOPE);
					onceil = true;
				}
				else
				{
					mo->floorz = SL_PointHeight(sc, mo->x, mo->y, LOSLOPE);
					mo->ceilingz = SL_PointHeight(sc, mo->x, mo->y, HISLOPE);
					onceil = onfloor = true;
				}
			}
			else
			if (mo->type == MT_LIGHTSOURCE)
			{
				if (mo->angle)	// = any except EAST
				{
					mo->floorz = SL_PointHeight(sc, mo->x, mo->y, LOSLOPE);
					mo->ceilingz = DDMININT;
				}
				else
				{
					mo->floorz = DDMAXINT;
					mo->ceilingz = SL_PointHeight(sc, mo->x, mo->y, HISLOPE);
				}
			}

			// check if Thing is on a same sector sided line
			// First check the mo's sseclines. If that fails check the  sector ssec lines
			if (onfloor && ((lnum = FindIfThingIsOnLine(mo, true)) >= 0 || (lnum = subfn_checkThingLine(mo)) >= 0))
			{
				line_t	*ld = &lines[lnum];

				sc = ld->frontsector;

				if (ld->specType >= R3D_40)
				{
					int lift = ld->frontsector->liftsec3D;
					int lf = ld->backsector->floorheight;
					int height = lift >= 0 ? sectors[lift].floorheight : lf;
					height += ld->d3d->offset3D + ld->d3d->totalheight;

					mo->floorz = height + ld->d3d->lineheightlow;
				}
				else
				{
					mo->floorz = ld->frontsector->floorheight;
				}
			}

			onfloor = onfloor ? (sc->specFlags & SF_MOVFLOOR || sc->liftsec3D >= 0) : false;	// $liftsec3D
			onceil  = onceil  ? (sc->specFlags & SF_MOVCEIL ) > 0 : false;

			if (onfloor || onceil)
			{
				if (dzcount >= maxdzcount)
				{
					maxdzcount += 128;	// we could expect tons of these
					dynzlist = (dynzlist_t*)W_Realloc(dynzlist, maxdzcount*sizeof(dynzlist_t), STATICID, 0);
				}

				dynzlist[dzcount  ].onfloor = onfloor;
				dynzlist[dzcount  ].onceil  = onceil;
				dynzlist[dzcount  ].lnum = lnum;
				dynzlist[dzcount++].mo = mo;
			}
		}
	}
}

// called on each level tic
void P_UpdatePhasedLights()
{
	// values for phased lighting
	dynphase.val += dynphase.dir ? -8 : 8;
	if (dynphase.val > 255) { dynphase.val = 255; dynphase.dir = 1; }
	if (dynphase.val <   0) { dynphase.val =   0; dynphase.dir = 0; }

	// 5505 types in sector 6 special sectors
	for (int i = 0; i < trig5505.count; i++)
	{
		if (secmo[i].timer-- < 0)
		{
			secmo[i].timer = 0;	// suspended until next trigger event

			for (int k = 0; k < secmo[i].count; k++)
			{
				secmo[i].list[k]->torchtic = 0;
			}
		}
	}
}

// Called on each game tic following all scsm adjustments
void P_UpdateDynLights()
{
	for (int i = 0; i < dzcount; i++)
	{
		mobj_t	*mo = dynzlist[i].mo;
		int		lnum =  dynzlist[i].lnum;

		if (dynzlist[i].onceil) mo->ceilingz = mo->subsector->sector->smoothceil;

		if (lnum >= 0)
		{
			line_t *ld = &lines[lnum];

			if (ld->specType >= R3D_40)
			{
				int lift = ld->frontsector->liftsec3D;
				int lf = ld->frontsector->floorheight;
				int height = lift >= 0 ? sectors[lift].smoothfloor : lf;

				if (ld->ovflags & OVF_3DLADDERLINE)
				{
					mo->floorz = height;
				}
				else
				{
					mo->floorz = height + ld->d3d->lineheightlow;
				}
			}
			else if (dynzlist[i].onfloor)
			{
				mo->floorz = ld->frontsector->floorheight;
			}
		}
		else
		{
			if (dynzlist[i].onfloor) mo->floorz = mo->subsector->sector->smoothfloor;
		}
	}
}
