//-----------------------------------------------------------------------------
//
// Copyright (C) 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_fogwalls.c
// New 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//
// DESCRIPTION: Identify lines that will need a fogwall texture
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "dd_types.h"
#include "dd_share.h"

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "r3d_edit.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

fogvalues_t fogval;

ushort	*linefog_group;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static ushort	alpha_groups = 0;

// CODE --------------------------------------------------------------------


// Set sky lines spanning 1's lines. Must be done last as hijacks a
// flag. This must also be done by the analysis code. As such it is
// always done although only needed if outside fog is set.
// Fortunately it does not take much time.
static RDBOOL FW_MiscFogWalls()
{
	line_t	*ld;
	int		c, i, k, m;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		ld->lnflags2 &= ~LNF2_SKY2FOGWALL;	// Clear LNF2_NOCHANGE flags

		if ( !(ld->ml_flags & ML_TWOSIDED) ) continue;

		if (   origTEX[i] == 0x22
			&& ld->frontsector != ld->backsector
			&& !(ld->frontsector->specFlags & SF_MOVCEIL)
			&& !(ld->backsector->specFlags  & SF_MOVCEIL)
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight/*
			&& Spans1sLines(ld)*/)
		{
			ld->lnflags2 |= LNF2_DRAWSKY;
		}
		else
		if (  !(origTEX[i] & 0x66)
			&& ld->frontsector != ld->backsector
			&& ld->frontsector->specFlags2 & SF2_SETFOG
			&& ld->backsector->specFlags2 & SF2_SETFOG
			&& Spans1sLines(ld))
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				seg_t		*sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					line_t *li = sg->linedef;

					if (   li
						&& li->backsector
						&& origTEX[li - lines] & 0x22
						&& (!(li->frontsector->specFlags2 & SF2_SETFOG)
						||  !(li->backsector->specFlags2  & SF2_SETFOG))
						&& Spans1sLines(li))
					{
						ld->lnflags2 |= LNF2_SKY2FOGWALL; c++;
						goto end_search;
					}
				}
			}
		}
end_search:;
	}

	return c > 0;
}

static void iterateFogLines(line_t *ld)
{
	RDBOOL	found = true;
	line_t	*lc;

	hitline[ld - lines] = false;
	linefog_group[ld - lines] = alpha_groups;

	while (found)
	{
		found = false;

		if ( (lc = subfn_FindVtxConnected(ld, 0)) )
		{
			iterateFogLines(lc);
			found = true;
		}
	}
}

void FW_SetFogLines()
{
	static int	linefog_groupmax = 0;

	if (numlines > linefog_groupmax)
	{
		linefog_group = (ushort*)W_Realloc(linefog_group , numlines*sizeof(ushort), STATICID, 0);
		linefog_groupmax = numlines;
	}
	memset(linefog_group, ~0, numlines*sizeof(ushort));
}

void FW_SetFogwallGroups()
{
	sector_t	*fs, *bs, *sec;
	line_t		*ld;
	int			c, d, e, i, k, m, n;
	RDBOOL		special_walls;
	RDBOOL		valid;

	//------------------------------------

	if (ma_inf.fogwalls_set) return;

	alpha_groups = 0;

	FW_SetFogLines();

	ma_inf.fogwalls_set = true;

	SET_countsec(0);
	CLR_hitline();
	
	// 1) Set compliant fog sectors
	for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->ceilingpic == skyflatnum)
		{
			valid = true;

			// check for replacement ceiling
			if (sec->specFlags & SF_NOHITEX)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					int		 lnum = sec->lines[k] - lines;
					ssline_t *sl = &sslines[lnum];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (   ss->sector == sec
							&& ss->ceilsec >= 0
							&& sectors[ss->ceilsec].ceilingpic != skyflatnum)
						{
							valid = false; break;
						}
					}
					if (!valid) break;
				}
			}

			if (valid)
			{
				sec->specFlags2 |= SF2_SETFOG; c++;
			}
		}
	}

	if (!c) return;

	special_walls = FW_MiscFogWalls();

	// set lines that need to act as fog walls
	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->lnflags2 & LNF2_SKY2FOGWALL)
		{
			hitline[i] = true; c++;
		}
		else
		if (   ld->ml_flags & ML_TWOSIDED
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& ((fs->specFlags2 & SF2_SETFOG) ^ (bs->specFlags2 & SF2_SETFOG)) )
		{
			hitline[i] = true; c++;

			if (fs->specFlags2 & SF2_SETFOG)
				countsec[fs - sectors]++;
			else
				countsec[bs - sectors]++;
		}
	}

	if (!c && !special_walls) return;

	if (c)
	{
		int		*wallmem = (int*)M_Malloc(numlines*sizeof(int), 0);
		int		*wallnum = (int*)M_Malloc(numlines*sizeof(int), 0);
		int		wallnumcount;

		// eliminate closed sky sectors
		for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		{
			if (countsec[i] < 2) continue;

			CLR_conarray();

			valid = false;

			for (d = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] && !(ld->lnflags2 & LNF2_SKY2FOGWALL))
				{
					conarray[ld - lines] = 1; d++;
				}
			}

			if (d > 2)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					int lnum = sec->lines[k] - lines;
					int bbox[4];

					if (   conarray[lnum] == 1
						&& subfn_CheckForRing(conarray, sec, lnum, NULL, bbox, NULL, NULL))
					{
						if (!LS_MarkSubsecsInAreaGeneral(sec, conarray, 0, NULL, bbox, -1, MSIA_TESTAREA))
						{
							valid = false; break;
						}

						for (m = 0; m < sec->linecount; m++)
						{
							lnum = sec->lines[m] - lines;

							if (conarray[lnum] == 4)
							{
								conarray[lnum] = 8;
								hitline[lnum] = false;
								valid = true;
							}
						}
					}
				}
			}

			if (valid && d == sec->linecount)
				sec->specFlags2 &= ~SF2_SETFOG;
		}

		memcpy(wallmem, hitline, numlines);

		// set the fog groups
		for (i = 0; i < numlines; i++)
		{
			if (hitline[i])
			{
				iterateFogLines(lines + i);
				alpha_groups++;
			}
		}

		memcpy(hitline, wallmem, numlines);
		M_Free(wallmem);

		// Force non-sky sectors to display fog
		// All fog walls must be in the same group
		for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		{
			ushort grp = 0xFFFF;

			if (sec->ceilingpic == skyflatnum || countsec[i]) continue;

			for (d = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] && !(ld->lnflags2 & LNF2_SKY2FOGWALL))
				{
					if (grp == 0xFFFF)
						grp = linefog_group[ld - lines];
					else if (grp != linefog_group[ld - lines])
					{
						d = 0; break;
					}
					d++;
				}
			}

			if (d > 1 && grp != 0xFFFF)
			{
				for (wallnumcount = e = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] || ld->lnflags2 & LNF2_SKY2FOGWALL)
						continue;

					if (!(ld->ml_flags & ML_TWOSIDED))
					{
						e++;
					}
					else
					{
						ssline_t *sl = &sslines[ld - lines];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector == sec)
							{
								seg_t *sg = &segs[ss->firstline];

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if (sg->linedef == ld)
									{
										if (   sg->sidedef->bottomtexture
											&& sg->sidedef->toptexture)
										{
											wallnum[wallnumcount++] = ld - lines;
										}
										goto endssectest;
									}
								}
							}
						}
endssectest:;
					}
				}

				if ((e || wallnumcount) && d + e + wallnumcount == sec->linecount)
				{
					sec->specFlags2 |= SF2_SETFOG;

					for (k = 0; k < sec->linecount; k++)
					{
						int lnum = sec->lines[k] - lines;
						hitline[lnum] = false;
						linefog_group[lnum] = 0xFFFF;
					}

					// set fog walls
					for (k = 0; k < wallnumcount; k++)
					{
						hitline[wallnum[k]] = true;
						linefog_group[wallnum[k]] = grp;
					}
				}
			}
		}

		M_Free(wallnum);
	}
}
