//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_hack.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
//
// DESCRIPTION: deals with map constructs that defy all reason
//
//-----------------------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

// CODE -----------------------------------------------------------------------

void LS_ChkNestedUnconnectedLinesInDifferentSectors()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& MSFC(i, 0x0080000C))
		{
			int k, secnum = -1;

			for (k = 0; k < sec->linecount; k++)
			{
				if (Vtxs1(sec->lines[k]))
				{
					secnum = LS_FindAdjacentSecnum(sec->lines[k], 9);
					break;
				}
			}
			if (secnum >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (Vtxs1(sec->lines[k]))
						Setspec(sec->lines[k], secnum, 32002);
				}
				Setsecdone3(sec, 32003);
			}
		}
	}
}

// Ritenour's Phobos ===========================================

static void setSSECS(int lnum, int ovrnum)
{
	ssline_t *sl = &sslines[lnum];
	int		 i;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		// v220-5 test added
		if (ss->floorsec < 0 && ss->ceilsec < 0)
			SetOverride(ss, ovrnum, 32001);
	}
}

static RDBOOL subfn_checkRTPowner(sector_t *sec)
{
	sector_t *chk, *osec = NULL;
	vertex_t *v1 = NULL, *v2 = NULL;
	int		 i;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED)
			|| ld->frontsector == ld->backsector
			|| origTEX[ld - lines]
			|| (ld->frontsector->floorheight
			== ld->backsector->floorheight
			&& ld->frontsector->ceilingheight
			== ld->backsector->ceilingheight))
		{
			return false;
		}

		chk = ld->frontsector == sec ?
			ld->backsector : ld->frontsector;

		if (!osec) osec = chk;
		else if (osec != chk) return false;

		vtxcheck[ld->v1 - vertexes]++;
		vtxcheck[ld->v2 - vertexes]++;
	}

	if (!osec
		|| osec == sec
		|| osec->specFlags & SF_NOLOTEX
		|| osec->specFlags & SF_MOVSECTOR)
	{
		return false;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];

		int		vn1 = vtxcheck[ld->v1 - vertexes];
		int		vn2 = vtxcheck[ld->v2 - vertexes];

		if (vn1 > 2 || vn2 > 2) return false;

		if		(vn1 == 1 && !v1) v1 = ld->v1;
		else if (vn1 == 1 && !v2) v2 = ld->v1;
		else if	(vn2 == 1 && !v1) v1 = ld->v2;
		else if (vn2 == 1 && !v2) v2 = ld->v2;
		else if (vn1 == 1 || vn2 == 1) return false;
	}

	if (!v1 || !v2)
	{
		return false;
	}

	for (i = 0; i < osec->linecount; i++)
	{
		line_t	*ld = osec->lines[i];

		if (ld->frontsector == ld->backsector
			&& ((ld->v1 == v1 && ld->v2 == v2)
			||  (ld->v1 == v2 && ld->v2 == v1)))
		{
			Settype(ld, R3D_01, 32001);
			setSSECS(ld - lines, osec - sectors);

			for (i = 0; i < sec->linecount; i++)
			{
				Settype(sec->lines[i], R3D_01, 32002);
				setSSECS(sec->lines[i] - lines, osec - sectors);
			}
			Setsecdone3(osec, 32002);
			Setsecdone3(sec, 32003);
			return true;
		}
	}

	return false;
}

// Check for unclosed rings requiring inheritance
void LS_RTP_Hacks()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& !SecCtrl(i)
			&& !SecONE(i)
			&& sec->linecount > 1
			&& (sec->specFlags & 0x00FFFFFC) == 0xC)
		{
			subfn_checkRTPowner(sec);
		}
	}
}

// v220-10 Look for hacks using crossing lines
// v220-7 Some horrendous things can be got away with when using a software renderer.
//        When an author compounds things further then it's only possible to resolve
//        the issue with a specific function especially when it does not even work
//        properly with a software renderer.
void LS_XlineHacks()
{
	sector_t	*fs, *bs;
	line_t		*ld, *lx, *li;
	int			i, box[4];

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !lineflags[i]
			&& ld->ml_flags & ML_TWOSIDED
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& !lxarray[i]
			&& origTEX[i] & 0x22
			&& !(origTEX[i] & 0x55)
			&& Vtxs1(ld)
			&& fs->floorheight != bs->floorheight
			&& fs->ceilingheight != bs->ceilingheight
			&& (lx = subfn_FindLineCrossing(ld, LC_NOVTXONLINEGEN|LC_DIFFSECLINE|LC_MIDTEXONLY))
			&& origTEX[lx - lines] & 0x22
			&& !(origTEX[lx - lines] & 0x55)
			&& ((lx->frontsector == fs && lx->backsector  == bs)
			||  (lx->backsector  == fs && lx->frontsector == bs))
			&& !(lx->frontsector->specFlags & SF_MOVSECTOR
			&&   lx->backsector->specFlags  & SF_MOVSECTOR) )
		{
			// check all textured lines in fs and bs form a loop
			sector_t *use, *chk = NULL;
			int		 c, k;

			use = lx->frontsector->specFlags & SF_MOVSECTOR ?
				lx->frontsector : lx->backsector;

			if (!(use->specFlags & SF_MOVSECTOR)) continue;

			CLR_hitline();

			validcount++;

			for (c = k = 0; k < use->linecount; k++)
			{
				li = use->lines[k];

				li->validcount = validcount;

				if (li->spflags & SPF_CONTROLLINE) continue;

				if (li->ml_flags & ML_TWOSIDED)
				{
					if (li->frontsector != fs && li->frontsector != bs && li->frontsector != chk)
					{
						if (chk) break;
						if (!(li->scpFlags & SCPF_DRAWTOLIMITS)) chk = li->frontsector;
					}

					if (li->backsector != fs && li->backsector != bs && li->backsector != chk)
					{
						if (chk) break;
						if (!(li->scpFlags & SCPF_DRAWTOLIMITS)) chk = li->backsector;
					}
				}

				if (!(li->ml_flags & ML_TWOSIDED) || origTEX[li - lines] & 0x44)
				{
					linearray[c++] = li - lines;
				}
			}

			if (!chk || chk->specFlags & SF_MOVSECTOR)
			{
				goto end_loop;
			}

			for (k = 0; k < chk->linecount; k++)
			{
				li = chk->lines[k];

				if (li->validcount == validcount) continue;

				if (!(li->ml_flags & ML_TWOSIDED) || origTEX[li - lines] & 0x44)
				{
					linearray[c++] = li - lines;
				}
			}

			if (c > 3 && c == nf_chkLineListRing(c, box, 0, hitline))
			{
				int m;

				validcount++;
				ss_validcount++;

				for (c = k = 0; k < fs->linecount; k++)
				{
					ssline_t *sl = &sslines[fs->lines[k] - lines];

					fs->lines[k]->validcount = validcount;

					linearray[c++] = fs->lines[k] - lines;

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						if (ss->sector != chk)
						{
							SetOverride(ss, chk - sectors, 32004);
						}
					}
				}

				for (k = 0; k < bs->linecount; k++)
				{
					ssline_t *sl = &sslines[bs->lines[k] - lines];

					if (bs->lines[k]->validcount != validcount)
					{
						linearray[c++] = bs->lines[k] - lines;
					}

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						if (ss->sector != chk)
						{
							SetOverride(ss, chk - sectors, 32005);
						}
					}
				}

				for (k = 0; k < c; k++)
				{
					li = &lines[linearray[k]];

					if ( !(li->ml_flags & ML_TWOSIDED)
						|| li->scpFlags & SCPF_DRAWTOLIMITS)
						Settype(li, R3D_36, 32006);
					else if (!(origTEX[li - lines] & 0x22))
						Settype(li, R3D_3F, 32006);
					else
						Settype(li, R3D_1D, 32006);

					SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_LINESSECSET|LFG_NOCHANGE, 32007);
				}
			}
		}

end_loop:;
	}
}

typedef struct {
	int		*sslist;
	int		sscount;
	int		box[4];
	int		secn;
	byte	dir;
} nloop_t;

// v220-10
// Look for nested textureless missing tex rings and, if valid, set a false floor inside
void LS_FalseFloorRingSpan()
{
	sector_t	*s;
	nloop_t		*loop;
	int			looplimit;
	int			c, count, i;
	int			bn = 0;

	CLR_hitline();
	SET_countsec(0);

	for (count = i = 0; i < numlines; i++)
	{
		if (   lines[i].ml_flags & ML_TWOSIDED
			&& lines[i].mtflags == 0x10
			&& lines[i].specType == R3D_00
			&& (lineflags[i] == 0x160 || lineflags[i] == 0x170)
			&& !(origTEX[i] & 0x11)
			&& Vtxs2(&lines[i])
			&& abs(lines[i].frontsector->floorheight - lines[i].backsector->floorheight) > 8*FRACUNIT
			&& !(lines[i].frontsector->specFlags & SF_MOVSECTOR)
			&& !(lines[i].backsector->specFlags  & SF_MOVSECTOR))
		{
			hitline[i] = 1;
			linearray[count++] = i;
			countsec[lines[i].frontsector - sectors]++;
		}
	}

	if (count < 6) return;

	for (c = i = 0; i < numsectors; i++)
		if (countsec[i] > 2) c++;

	if (!c) return;

	// Make sure enough loop storage is available
	// This cannot be reallocated on the fly because of sslist
	looplimit = count/3;
	loop = (nloop_t*)W_Malloc(looplimit * sizeof(nloop_t), TEMPID, 16);

	while(1)
	{
		int			n, rc;
		int			lnum = -1;

		for (i = 0; i < count; i++)
		{
			n = linearray[i];

			if (hitline[n] == 1) { lnum = n; break; }
		}

		if (lnum < 0) break;

		s = lines[lnum].frontsector->floorheight < lines[lnum].backsector->floorheight ?
			lines[lnum].backsector : lines[lnum].frontsector;

		if (bn >= looplimit) bn = looplimit - 1;	// insurance

		if (subfn_CheckForRing(hitline, s, lnum, &rc, loop[bn].box, NULL, NULL) && rc > 2)
		{			
			loop[bn].dir = 1;
			loop[bn].secn = s - sectors;

			if (!LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, loop[bn].box, -1, MSIA_NONE))
			{
				loop[bn].dir = 0;
				LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, loop[bn].box, -1, MSIA_REVERSED);
			}

			loop[bn].sscount = hitssecptr;

			if (hitssecptr)
			{
				loop[bn].sslist = (int*)W_Malloc(hitssecptr * sizeof(int), TEMPID, 17);
				memcpy(loop[bn].sslist, hitssec, hitssecptr * sizeof(int));
			}

			for (i = 0; i < s->linecount; i++)
			{
				n = s->lines[i] - lines;

				if (hitline[n] == 4) hitline[n] = 0;
			}

			bn++;
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	if (bn > 2)
	{
		while(1)
		{
			nloop_t	*lpi = NULL, *lpo = NULL;
			fixed_t	fht;
			int		k, ssn;

			// Look for ring with a direction of 0
			for (i = 0; i < bn; i++)
			{
				if (!loop[i].dir) { lpo = &loop[i]; break; }
			}

			if (!lpo) break;
			
			// Mark loop off
			lpo->dir = 2;
			fht = sectors[lpo->secn].floorheight;

			// Look for smallest ring with a direction of 1 that fits inside loop found
			for (i = 0; i < bn; i++)
			{
				if (   loop[i].dir == 1
					&& sectors[loop[i].secn].floorheight >= fht)
				{
					if (M_CompareBox(loop[i].box, lpo->box))
					{
						if (!lpi) lpi = &loop[i];
						else if (M_CompareBox(loop[i].box, lpi->box)) lpi = &loop[i];
					}
				}
			}

			if (!lpi) continue;
			
			// form a list of ssectors between the two rings
			hitssecptr = 0;

			for (i = 0; i < lpo->sscount; i++)
			{
				ssn = lpo->sslist[i];

				for (k = 0; k < lpi->sscount; k++)
				{
					if (lpi->sslist[k] == ssn) goto end_num;
				}

				hitssec[hitssecptr++] = ssn;
end_num:;
			}

			if (hitssecptr)
			{
				subsector_t *ss;
				RDBOOL		valid = true;

				// Since this can only be done following all previous
				// analysis then check that subsecs have not already
				// been modified
				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];

					if (   ss->lfsec >= 0
						|| ss->floorsec >= 0
						|| ss->sector->R3Dfloorheight < DDMAXINT
						|| ss->sector->R3Dfloorsec >= 0
						|| ss->sector->R3Dlowsec >= 0)
					{
						valid = false; break;
					}

					// eliminate any subsectors where its
					// sector's floorheight is > fht unless
					// it's a floor move sector;
					if (   ss->sector->floorheight > fht
						&& !(ss->sector->specFlags & SF_MOVFLOOR) )
					{
						hitssec[i] = -1;
					}
				}

				if (valid)
				{
					// Set all subsecs between loops to have a false floor
					for (i = 0; i < hitssecptr; i++)
					{
						if (hitssec[i] >= 0)
							subsectors[hitssec[i]].lfsec = lpo->secn;
					}
				}
			}
		}
	}

	W_FreeALL(TEMPID);
}

static void subfn_FalseCeilRingSpan(sector_t	*sc)
{
	sector_t	*chk, *nsec[2] = { NULL, NULL };
	line_t		*ld;
	int			c, i, k, lnum, rc, box[4];

	for (i = 0; i < sc->linecount; i++)
	{
		ld   = sc->lines[i];
		lnum = ld - lines;

		if (   lineflags[lnum]
			|| !(ld->mtflags & R3D_MISHITEX)
			|| (ld->specType != R3D_00 && ld->specType != R3D_01)
			|| vtxflags[ld->v1 - vertexes] != 3
			|| vtxflags[ld->v2 - vertexes] != 3)
		{
			return;
		}
	}

	CLR_hitline();

	for (i = 0; i < sc->linecount; i++)
	{
		ld   = sc->lines[i];

		// Find non-sky sector
		chk = ld->ml_flags & R3D_MISHTEX1 ? ld->backsector : ld->frontsector;

		if (chk->ceilingpic == skyflatnum) continue;

		if (!nsec[0]) nsec[0] = chk;
		else if (chk->ceilingheight < nsec[0]->ceilingheight) nsec[0] = chk;

		if (!nsec[1]) nsec[1] = chk;
		else if (chk->ceilingheight > nsec[1]->ceilingheight) nsec[1] = chk;
	}

	if (!nsec[0] || nsec[0]->ceilingheight > sc->ceilingheight) return;
	if (!nsec[1] || nsec[1]->ceilingheight > sc->ceilingheight) return;
	if (nsec[0] == nsec[1] || nsec[0]->ceilingheight == nsec[1]->ceilingheight) return;

	for (c = i = 0; i < 2; i++)
	{
		sector_t *use = nsec[i];

		CLR_hitline();

		for (k = 0, lnum = -1; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

			if (chk->ceilingheight == use->ceilingheight)
			{
				lnum = ld - lines;
				hitline[lnum] = 1;
			}
		}

		if (   lnum >= 0
			&& subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
			&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE) )
		{
			for (k = 0; k < hitssecptr; k++)
			{
				Setr3dsscs(&subsectors[hitssec[k]], use - sectors, 32008);
			}

			c++;
		}
	}

	if (c == 2)
	{
		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (ld->specType == R3D_01)
				Settype(ld, R3D_3F, 32009);

			if (ld->mtflags & R3D_MISLOTEX)
			{
				ssline_t *sl = &sslines[ld - lines];
				
				pr_OFinsert("R3D_MISLOTEX L%i\n", ld - lines);

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->sector != sc && ss->lfsec != sc - sectors)
						Setr3dsslfsec(ss, sc - sectors, 32009);
				}
			}
		}
	}
}

// v220-13
// Look for missing tex ceiling spanning two different ceiling sectors in one floor sector
// NOTE: may not work in edit mode
void LS_FalseCeilRingSpan()
{
	sector_t	*sc;
	int			i;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (  !secdone[i]
			&& sc->ceilingpic == skyflatnum
			&& (sc->specFlags & 0xFFFFFF) == 0x018188)
		{
			subfn_FalseCeilRingSpan(sc);
		}
	}
}

// =====================================================

static void subfn_secContaining(sector_t *sc)
{
	subsector_t	*ss;
	ssline_t	*sl;
	sector_t	*os;
	seg_t		*sg;
	line_t		*ld;
	int			lnum, rc, box[4];
	int			c, i, k, m;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		if (lineflags[ld - lines] & LFG_3DTYPE) return;
		if (ld->frontsector != ld->backsector) return;
	}

	validcount++;
	ss_validcount++;

	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (!sg->linedef || sg->linedef->validcount == validcount) continue;
				sg->linedef->validcount = validcount;
				if (lineflags[sg->linedef - lines] & LFG_3DTYPE) return;
				if (sg->frontsector == sc) continue;

				if (!os)
				{
					if (sg->frontsector->floorheight >= sc->floorheight) return;
					if (sg->frontsector->specFlags & SF_MOVSECTOR) return;
					if (!(sg->frontsector->specFlags & SF_SSLINETRUE)) return;
					os = sg->frontsector;
				}
				else if (os != sg->frontsector) return;

				if (sg->frontsector == sg->backsector) c++;
			}
		}
	}

	if (!os || c > 1) return;

	// check the sector
	CLR_hitline();

	for (c = i = 0, lnum = -1; i < os->linecount; i++)
	{
		ld = os->lines[i];

		if (ld->frontsector != ld->backsector)
		{
			if (   ld->backsector
				&& !origTEX[ld - lines]
				&& ld->backsector->floorheight != ld->frontsector->floorheight)
			{
				return;
			}

			lnum = ld - lines;
			hitline[lnum] = 1; c++;
		}
	}

	if (   c > 2
		&& subfn_CheckForRing(hitline, os, lnum, &rc, box, NULL, NULL)
		&& rc == c
		&& LS_MarkSubsecsInAreaGeneral(os, hitline, 0, NULL, box, -1, MSIA_NONE))
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];

			if (ss->sector != sc) ChangeSSECsec(ss, sc, 32010);
			if (ss->floorsec != os - sectors) Setr3dssfs(ss, os - sectors, 32010);
			if (ss->ceilsec != os - sectors) Setr3dsscs(ss, os - sectors, 32010);
		}
	}
}

static void oneSecReferenced(int lidx, sector_t *ref)
{
	subsector_t	*ss;
	ssline_t	*sl = &sslines[lidx];
	sector_t	*chk = NULL;
	int			i;

	hitssecptr = 0;

	for (i = 0; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->sector == ref) continue;
		if (ss->floorsec >= 0) return;
		if (!chk)
		{
			chk = ss->sector;
			if ((chk->specFlags & 0xC) != 0xC) return;
		}
		else
		if (ss->sector != chk) return;

		hitssec[hitssecptr++] = sl->list[i];
	}

	if (!chk) return;

	for (i = 0; i < hitssecptr; i++)
		SetOverride(&subsectors[hitssec[i]], ref - sectors, 32011);

	SetLFG(&lines[lidx], LFG_NEWSET, 32011);
}

void LS_SupportSecInside()
{
	sector_t	*sc;
	line_t		*ld;
	int			i;

	// v220-17 0x1C support sectors to be inherited by containing sector with 1's lines
	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& sc->linecount >= 3	// this type of construction always seem to use 3 lines
			&& sc->linecount <= 4	// but also allow 4 lines
			&& sc->stptr < 0
			&& sc->heightsec < 0
			&& (sc->specFlags & 0xFFFFC) == 0x1C)
		{
			subfn_secContaining(sc);
		}
	}

	// v220-17 Check disconnected same sec lines' ssecs are referencing the correct sector
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& !lineflags[i]
			&& ld->frontsector->specFlags & 0x400
			&& Vtxs1(ld)
			&& !lxarray[i])
		{
			oneSecReferenced(i, ld->frontsector);
		}
	}
}