//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_intersect.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: looks for lines that intersect
//
//-----------------------------------------------------------------------------

#include <math.h>
#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "p_maputil.h"

line_t	*multicrossline;
int		backlinesecnum;
int		vtxisonline;
int		xcrosspoint;
int		ycrosspoint;
int		vtxc_x;
int		vtxc_y;

//==============================================================================


static line_t *LS_BlockLinesIterator(int x, int y, line_t *ldc, int condition, int type, RDBOOL(*func)(line_t*, line_t*, int))
{
	int			offset;
	const long	*list;
	
	if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
		return NULL;

	offset = y*bmapwidth + x;

	offset = *(blockmap + offset);
	list = blockmaplump + offset;

	// skip starting delimiter
	while( *(++list) >= 0)
	{
		line_t *ld = &lines[*list];

		if (ld->validcount == validcount)
			continue;		// line has already been checked

		ld->validcount = validcount;

		if (type != 3 && ld->specType >= R3D_40) continue;
		if (type == 3 && (ld->specType <  R3D_40 || ld->specType > R3D_43) ) continue;
		
		if(func(ldc, ld, condition))
		{
			switch(type)
			{
			case 0:
			case 3:
				return ld;
				break;
			case 1:
				if (ld->ml_flags & ML_TWOSIDED &&
					ld->frontsector != ld->backsector &&
					ld->frontsector != ldc->frontsector &&
					ld->backsector  != ldc->frontsector)
					return ld;
				break;
			case 2:
				if (LSS(ld) &&
					ld->tyflags & TYF_MIDTEXLINE &&
					texArray[sides[ld->sidenum[0]].midtexture]->height <= 16)
					return ld;
				break;
			default:
				Con_Error("LS_BlockLinesIterator: ERROR [1]: type %i", type);
				break;
			}
		}
	}
	
	return NULL;	// everything was checked
}

static RDBOOL subfn_GetLineFacingBacksector(line_t *ld, line_t *li)
{
	// Check lines don't form a single straight line
	if (ld->specType < R3D_40 &&
		li->specType < R3D_40 &&
		!subfn_CheckConnectedLineSlope(ld, li))
	{
		float v1x  = FIX2FLT(li->v1->x);
		float v1y  = FIX2FLT(li->v1->y);
		float nx   = v1y - FIX2FLT(li->v2->y);
		float ny   = FIX2FLT(li->v2->x) - v1x;
		float vvx1 = v1x-FIX2FLT(ld->v1->x), vvy1 = v1y-FIX2FLT(ld->v1->y);
		float vvx2 = v1x-FIX2FLT(ld->v2->x), vvy2 = v1y-FIX2FLT(ld->v2->y);

		// The dot product - test each end of the reference line to
		// ensure, absolutely, the side of the line being faced.
		if (nx*vvx1 + ny*vvy1 <= 0 && nx*vvx2 + ny*vvy2 <= 0)	// [Not <=]  Set 2nd term set to < in the past as found problem with <= which is what it should be and has now been set back to. Can't recall why now. May2003
		{
			backlinesecnum = li->backsector - sectors;
			return true;
		}
	}
	return false;
}

RDBOOL checkSuperimposed(line_t *cmp, line_t *ref)
{
	sector_t *s1, *s2, *s3;

	return (s1 = cmp->frontsector) != (s2 = cmp->backsector)
		&& (s3 = ref->frontsector) != s1
		&& s2 != s3
		&& (s1->floorheight != s3->floorheight
		||  s2->floorheight != s3->floorheight);
}

RDBOOL subfn_CheckLineIntersect(line_t *ref, line_t *cmp, int condition)
{
	vertex_t	*s, *e, *d;
	double		ax, ay, bx, by;
	double		cx, cy, dx, dy;
	double		dv, r;
	fixed_t		x1, y1;

	xcrosspoint = ycrosspoint = DDMAXINT;

	if (cmp->ml_flags & ML_TWOSIDED)
	{
		if (condition & LC_SINGLESIDED)
			return false;
	}
	else
	{
		if (!(condition & LC_SINGLESIDED))
			return false;

		// Ensure this and only this condition is set
		condition = LC_TESTLINE;
	}

	if (condition & LC_NOTEX &&
		origTEX[cmp - lines])
	{
		return false;
	}

	if (condition & LC_MIDTEXONLY &&
		(!(origTEX[cmp - lines] & 0x22) || origTEX[cmp - lines] & 0x55))
	{
		return false;
	}

	// Check if bounding boxes overlap
	// Lines can't intersect if they don't
	if (ref->bbox[BOXLEFT ] > cmp->bbox[BOXRIGHT] ||
		ref->bbox[BOXRIGHT] < cmp->bbox[BOXLEFT ] ||
		ref->bbox[BOXBASE ] > cmp->bbox[BOXTOP  ] ||
		ref->bbox[BOXTOP  ] < cmp->bbox[BOXBASE ])
	{
		return false;
	}

	// Quick vertical/horizontal slope reject
	if ((ref->slopetype == ST_HORIZONTAL && cmp->slopetype == ST_HORIZONTAL) ||
		(ref->slopetype == ST_VERTICAL   && cmp->slopetype == ST_VERTICAL  ) ||
		(ref->dx == cmp->dx && ref->dy == cmp->dy))
	{
		if (   condition & LC_SUPERIMPOSED
			&& checkSuperimposed(cmp, ref))
		{
			int chk0 = ref->v1->x == cmp->v1->x && ref->v1->y == cmp->v1->y;
			int chk1 = ref->v2->x == cmp->v2->x && ref->v2->y == cmp->v2->y;

			// If both vertexes are coincident then return false since
			// one line will not have been split by the other's vertex
			if (chk0 && chk1) return false;

			if (chk0 || chk1)
				return VertexIsOnLine(chk0 ? cmp->v2 : cmp->v1, ref);

			chk0 = ref->v2->x == cmp->v1->x && ref->v2->y == cmp->v1->y;
			chk1 = ref->v1->x == cmp->v2->x && ref->v1->y == cmp->v2->y;

			if (chk0 && chk1) return false;

			if (chk0 || chk1)
				return VertexIsOnLine(chk0 ? cmp->v2 : cmp->v1, ref);

			// Check if both vertexes lie on ref's line
			return VertexIsOnLine(cmp->v1, ref) && VertexIsOnLine(cmp->v2, ref);
		}
		 
		return false;
	}
	
	if (condition & LC_LI_CLOSED)
	{
		if (cmp->frontsector == cmp->backsector ||
			!(cmp->frontsector->specFlags & SF_LINESCLOSED ||
			  cmp->backsector->specFlags & SF_LINESCLOSED) ||
			LineHasMovSectors(cmp) ||
			cmp->frontsector->floorheight != cmp->backsector->floorheight ||
			cmp->frontsector->floorheight >= ref->frontsector->floorheight)
		{
			return false;
		}
	}

	// Tests only if crossing a two sided line with different floor heights.
	if (condition & LC_DIFF_FLR_HT)
	{
		if (cmp->frontsector->floorheight == cmp->backsector->floorheight)
		{
			return false;
		}
	}

	// Tests only if crossing a two sided with different sectors - both static.
	if (condition & LC_DIFFSECSTATIC)
	{
		if (   cmp->frontsector == ref->frontsector
			|| cmp->frontsector == ref->backsector
			|| cmp->backsector == ref->frontsector
			|| cmp->backsector == ref->backsector
			|| cmp->frontsector->specFlags & SF_MOVSECTOR
			|| cmp->backsector->specFlags & SF_MOVSECTOR)
		{
			return false;
		}
	}

	// Textureless connected same sided line (sent) crossing
	// a same sided line in a different sector with both
	// midtex's set.  Used with LS_CheckMoveSupportLineSSECsec
	if (condition & LC_SAMEWITHMID)
	{
		if (cmp->frontsector != cmp->backsector ||
			!(origTEX[cmp - lines] & 0x22))
		{
			return false;
		}
		if (condition & LC_DIFFSECMOV)
		{
			if ( !(cmp->frontsector->specFlags & SF_MOVFLOOR)
				|| cmp->frontsector == ref->frontsector
				|| cmp->frontsector->floorpic == ref->frontsector->floorpic)
			{
				return false;
			}
		}
	}

	if (condition & LC_DIFFSECLINE)
	{
		if (cmp->frontsector == cmp->backsector)
		{
			return false;
		}
	}

	if (condition & (LC_SAMESECLINE|LC_SAMESECFMOV))
	{
		if (cmp->frontsector != cmp->backsector)
		{
			return false;
		}

		if (   condition & LC_SAMESECFMOV
			&& !(cmp->frontsector->specFlags & SF_MOVFLOOR))
		{
			return false;
		}
	}

	// b211-0
	if (condition & LC_OWNSECLINE)
	{
		if (   cmp->frontsector == cmp->backsector
			&& cmp->frontsector == ref->frontsector )
		{
			return false;
		}
	}

	// Special test to eliminate test for line in sector with all lines
	// same crossing a line in another sector with the same criteria.
	if (condition & LC_NOT_SHARED)
	{
		if (SecONE(cmp->frontsector - sectors) &&
			SecMIDC(cmp->frontsector - sectors))
			return false;

		if (SecONE(cmp->backsector - sectors) &&
			SecMIDC(cmp->backsector - sectors))
			return false;

		// Eliminate shared sectors
		if ((ref->frontsector == cmp->frontsector &&
			 ref->frontsector == cmp->backsector) ||
			(ref->backsector  == cmp->frontsector &&
			 ref->backsector  == cmp->backsector))
			return false;
	}

	// Check ref's vertexes not on the test line
	if (condition & LC_NOVTXONLINE)
	{
		if (VertexIsOnLine(ref->v1, cmp))
		{
			return false;
		}
		if (VertexIsOnLine(ref->v2, cmp))
		{
			return false;
		}
	}

	if (condition & LC_NOVTXONLINEGEN)
	{
		if (PointIsOnLine(ref->v1->x, ref->v1->y, cmp, 1))
		{
			return false;
		}
		if (PointIsOnLine(ref->v2->x, ref->v2->y, cmp, 1))
		{
			return false;
		}
	}

	if (condition & LC_NOTEXMOV)	// v220-6
	{
		if (   cmp->frontsector != cmp->backsector
			|| (cmp->frontsector->specFlags & 0x1000C) != 0x1000C)
		{
			return false;
		}
	}

	// Check ref's vertexes not on the test line
	if (condition & (LC_VTXISONLINE|LC_SUPERIMPOSED))
	{
		if (    condition & LC_VTXNOTSHARED
			|| (condition & LC_SUPERIMPOSED
			&& checkSuperimposed(cmp, ref)) )
		{
			if (ref->v1->x == cmp->v1->x && ref->v1->y == cmp->v1->y) return false;
			if (ref->v1->x == cmp->v2->x && ref->v1->y == cmp->v2->y) return false;
			if (ref->v2->x == cmp->v1->x && ref->v2->y == cmp->v1->y) return false;
			if (ref->v2->x == cmp->v2->x && ref->v2->y == cmp->v2->y) return false;

			if (VertexIsOnLine(cmp->v1, ref))
			{
				return true;
			}
			if (VertexIsOnLine(cmp->v2, ref))
			{
				return true;
			}
			return false;
		}

		if (condition & LC_SUPERIMPOSED) return false;

		if (VertexIsOnLine(ref->v1, cmp))
		{
			return true;
		}

		if (VertexIsOnLine(ref->v2, cmp))
		{
			return true;
		}
	}

	if (condition & LC_IFVTXONLINE &&
		cmp->frontsector == cmp->backsector &&
		ref->frontsector == ref->backsector &&
		ref->frontsector != cmp->frontsector)
	{
		if (VertexIsOnLine(ref->v1, cmp))
		{
			vtxisonline = cmp->frontsector - sectors;
			return false;
		}
		if (VertexIsOnLine(ref->v2, cmp))
		{
			vtxisonline = cmp->frontsector - sectors;
			return false;
		}
	}

	s = ref->v1, e = ref->v2, d = cmp->v1;
	ax = (double)s->x, ay = (double)s->y, bx = (double)e->x, by = (double)e->y;
	cx = (double)d->x, cy = (double)d->y, dx = cx+(double)cmp->dx, dy = cy+(double)cmp->dy;

	// If dv zero then lines have same slope (and therefore cannot intersect)
	if (!(dv = (bx - ax)*(dy - cy) - (by - ay)*(dx - cx))) return false;

	r = ((ay - cy)*(dx - cx) - (ax - cx)*(dy - cy)) / dv;
	x1 = (fixed_t)(ax + r*(bx - ax));
	y1 = (fixed_t)(ay + r*(by - ay));

	xcrosspoint = x1;
	ycrosspoint = y1;

	// Check intersect lies in both bounding boxes
	if (x1 < ref->bbox[BOXLEFT] || x1 > ref->bbox[BOXRIGHT] ||
		x1 < cmp->bbox[BOXLEFT] || x1 > cmp->bbox[BOXRIGHT])
	{
		return false;
	}

	if (y1 < ref->bbox[BOXBASE] || y1 > ref->bbox[BOXTOP] ||
		y1 < cmp->bbox[BOXBASE] || y1 > cmp->bbox[BOXTOP])
	{
		return false;
	}

	// Check not a vertex misplacement (map) error
	if (!(condition & LC_TESTLINE) && Vtxs1(ref))	// Only test for unconnected lines
	{
		if (abs(ref->v1->x - x1) <= FRACUNIT &&
			abs(ref->v1->y - y1) <= FRACUNIT)
			return false;

		if (abs(ref->v2->x - x1) <= FRACUNIT &&
			abs(ref->v2->y - y1) <= FRACUNIT)
			return false;
	}

	// Check intersection point not at line end
	if (condition)	// i.e any other than NONE
	{
		if ((ref->v1->x == x1 && ref->v1->y == y1) ||
			(ref->v2->x == x1 && ref->v2->y == y1))
			return false;
		if ((cmp->v1->x == x1 && cmp->v1->y == y1) ||
			(cmp->v2->x == x1 && cmp->v2->y == y1))
			return false;
	}

	// Some tests need the multicross test line.
	if (!multicrossline && condition & (LC_NOVTXONLINE|LC_DIFFSECLINE))
		multicrossline = cmp;

	return true;
}

static RDBOOL PIT_IntersectLine(line_t* cl, line_t* ld, int condition)
{
	return subfn_CheckLineIntersect(cl, ld, condition);
}

// b211-0 Revised
// conditions;
//	0 - tests both vertexes of line sent - owners are not checked
//	1 - tests vtx1
//	2 - tests vtx1 and also number of owners of unmerged vtx
//	3 - tests vtx2
//	4 - tests vtx2 and also number of owners of unmerged vtx

static RDBOOL PIT_UnmergedVTXLine(line_t* ls, line_t* ld, int condition)
{
	switch(condition)
	{
	case 0:
	case 1:
	case 3:
		if (  ls->v1 != ld->v1 && ls->v1 != ld->v2 &&
			((ls->v1->x == ld->v1->x && ls->v1->y == ld->v1->y) ||
			 (ls->v1->x == ld->v2->x && ls->v1->y == ld->v2->y)) )
		{
			if (   condition == 0
				|| condition == 1
				|| vertexowners[ld->v1 - vertexes].num > 1)
			{
				return true;	// Unmerged vtx1
			}
		}
		if (condition) break;
	case 2:
	case 4:
		if (  ls->v2 != ld->v1 && ls->v2 != ld->v2 &&
			((ls->v2->x == ld->v1->x && ls->v2->y == ld->v1->y) ||
			 (ls->v2->x == ld->v2->x && ls->v2->y == ld->v2->y)) )
		{			
			if (   condition == 0
				|| condition == 2
				|| vertexowners[ld->v2 - vertexes].num > 1)
			{
				return true;	// Unmerged vtx2
			}
		}
		break;
	}

	return false;
}

static RDBOOL PIT_ConnectedLine(line_t* ls, line_t* ld, int condition)
{
	backlinesecnum = -1;	// Used with condition 8

	if (condition == 200)	// v220-10
	{
		if (ld->v1->x == ls->v1->x && ld->v1->y == ls->v1->y) return true;
		if (ld->v2->x == ls->v1->x && ld->v2->y == ls->v1->y) return true;
		return false;
	}

	if (condition == 201)	// v220-10
	{
		if (ld->v1->x == ls->v2->x && ld->v2->y == ls->v2->y) return true;
		if (ld->v2->x == ls->v2->x && ld->v2->y == ls->v2->y) return true;
		return false;
	}

	if (condition == 100 || condition == 101)	//v220-7
	{
		vertex_t *v = (condition == 100 ? ls->v1 : ls->v2);

		return (v == ld->v1 || v == ld->v2)
			&& ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector
			&&  origTEX[ld - lines] & 0x44;
	}

	if (ls->v1 == ld->v1 ||
		ls->v1 == ld->v2 ||
		ls->v2 == ld->v1 ||
		ls->v2 == ld->v2)
	{
		switch(condition)
		{
		case 0:
			if (!(ld->ml_flags & ML_TWOSIDED))
				return true;
			break;
		case 1:
			if (sides[ld->sidenum[0]].texFlags)
				return true;

			if (ld->ml_flags & ML_TWOSIDED &&
				sides[ld->sidenum[1]].texFlags == T_LOW)
				return true;
			break;
		case 2:
			if (ld->tyflags & TYF_MIDTEXLINE &&
				ld->frontsector != ls->frontsector &&
				ld->frontsector == ld->backsector)
			{
				return true;
			}
			break;
		case 3:
			if (LSS(ld) &&
				!countsec[ld->frontsector - sectors] &&
				CmpLptr(ld, -1) &&
				ld->frontsector->floorpic !=
				ls->frontsector->floorpic &&
				ld->frontsector->ceilingpic ==
				ls->frontsector->ceilingpic &&
				ld->frontsector->floorheight ==
				ls->frontsector->floorheight &&
				ld->frontsector->ceilingheight >=
				ls->frontsector->ceilingheight)
				return true;
			break;
		case 4:
			if (ld->tyflags & TYF_MIDTEXLINE &&
				!hitline[ld - lines] &&
				ld->frontsector == ls->frontsector)
			{
				return true;
			}
			break;
		case 5:
			if (ld->ml_flags & ML_TWOSIDED &&
				origTEX[ld - lines] &&
				Vtfl3(ld) &&
				(vtxarray[ld->v1 - vertexes] == 2 ||
				 vtxarray[ld->v2 - vertexes] == 2) &&
				ld->frontsector == ls->frontsector &&
				ld->backsector  == ls->backsector)
			{
				return true;
			}
			break;
		case 6:
			if (ld->ml_flags & ML_TWOSIDED &&
				ld->specType == R3D_00 &&
				!ld->mtflags &&
				!ld->spflags &&
				ld->frontsector == ls->frontsector &&
				ld->backsector == ls->backsector)
			{
				return true;
			}
			break;
		case 7:
			if (!(ld->ml_flags & ML_TWOSIDED) &&
				ld->frontsector->ceilingpic == skyflatnum)
				return true;
			break;
		case 8:
			if (ld->ml_flags & ML_TWOSIDED &&
				ld->frontsector != ld->backsector &&
				origTEX[ld - lines] == 0x01 &&
				subfn_GetLineFacingBacksector(ls, ld))
			{
				return true;
			}
			break;
		case 9:
			if (ls->frontsector == ld->frontsector &&
				ls->backsector != ld->backsector)
				return true;
			if (ls->frontsector != ld->frontsector &&
				ls->backsector == ld->backsector)
				return true;
			break;
		case 10:
			if (LSS(ld)) return true;
			break;
		case 11:
			if (ld->ml_flags & ML_TWOSIDED &&
				ld->frontsector != ld->backsector &&
				!origTEX[ld - lines] &&
				ld->frontsector->ceilingheight !=
				ld->backsector->ceilingheight)
				return true;
			break;
		case 12:
			if (ld->ml_flags & ML_TWOSIDED &&
				lnsec[ld - lines] < 0 &&
				ld->frontsector == ls->frontsector &&
				ld->backsector == ls->backsector)
			{
				return true;
			}
			break;
		case 13:
			if (LSS(ld)
				&& ld->frontsector != ls->frontsector
				&& ld->frontsector != ls->backsector
				&& origTEX[ld - lines] & 0x22
				&& !hitline[ld - lines])
			{
				return true;
			}
			break;
		case 14:
			if (ld->special && ld->special != 91)
				return true;
			break;
		case 15: //rev170f
			if (ld->ml_flags & ML_TWOSIDED
				&& !(origTEX[ld - lines] & 0x44)
				&& ld->frontsector->ceilingheight
				!= ld->backsector->ceilingheight)
			{
				return true;
			}
			break;
		case 16: //rev170f
			if (origTEX[ld - lines] & 0x44)
			{
				return true;
			}
			break;
		case 17: //rev170f
			if (LSS(ld)
				&& lnsec[ld - lines] >= 0
				&& ld->frontsector == ls->frontsector)
			{
				vertexowner_t *own;

				own = vertexowners + (ld->v1 - vertexes);
				if (own->num != 1) break;
				own = vertexowners + (ld->v2 - vertexes);
				if (own->num != 1) break;
				own = vertexowners + (ls->v1 - vertexes);
				if (own->num != 1) break;
				own = vertexowners + (ls->v2 - vertexes);
				if (own->num != 1) break;
				return true;
			}
			break;
		case 18:	// rev170j
			if (ld->backsector
				&& ld->backsector == ls->backsector
				&& ld->frontsector->floorheight == ls->frontsector->floorheight
				&& origTEX[ld - lines] & 0x01)
			{
				return true;
			}
			break;
		case 19:	// rev170j
			if (ld->backsector
				&& ld->frontsector == ls->frontsector
				&& ld->backsector->floorheight == ls->backsector->floorheight
				&& origTEX[ld - lines] & 0x10)
			{
				return true;
			}
			break;
		case 20:
			if (ld->frontsector != ld->backsector)
				return true;
			break;
		case 21:	// rev170f
			if (ld->ml_flags & ML_TWOSIDED
				&& origTEX[ld - lines] == 0x05
				&& ((ld->frontsector == ls->frontsector && ls->frontsector->ceilingheight == ld->frontsector->ceilingheight)
				||  (ld->backsector == ls->backsector && ls->backsector->ceilingheight == ld->backsector->ceilingheight)))
			{
				return true;
			}
			break;
		case 22:	// Rev 0ct/05 - R3D_3B tighten up condition
			if (!(ld->ml_flags & ML_TWOSIDED))
			{
				if (ld->frontsector == ls->frontsector
					&& !(origTEX[ls - lines] & 0x07))
					return true;
				if (ld->frontsector == ls->backsector
					&& !(origTEX[ls - lines] & 0x70))
					return true;
			}
			break;
		case 23:	// Added v2.1.00 build_32 - missing tex substitution
			if (sides[ld->sidenum[0]].bottomtexture)
				return true;
			break;
		case 24:	// Added v2.1.00 build_32
			if (ld->ml_flags & ML_TWOSIDED
				&& sides[ld->sidenum[1]].bottomtexture)
				return true;
			break;
		case 25:	// Added b211-0
			if (ld->frontsector == ld->backsector
				&& ld->frontsector != ls->frontsector
				&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
				&& ld->frontsector->floorpic
				== ls->frontsector->floorpic)
				return true;
			break;
		case 26:	// Added b211-0
			return vtxarray[ld->v1 - vertexes] == 1
				|| vtxarray[ld->v2 - vertexes] == 1;
		case 27:	// Added v220-4
			{
				if (   ld->frontsector == ld->backsector
					&& ld->frontsector != ls->frontsector
					&& origTEX[ld - lines] & 0x22
					&& !(origTEX[ld - lines] & 0x55))
				{
					return true;
				}
			}
			break;
		case 28:	// Added v220-4 Used following test 27 above
			if (   ld->frontsector == ld->backsector
				&& ld->frontsector != ls->frontsector
				&& hitline[ld - lines] & 1)
			{
				return true;
			}
			break;
		case 29:	// Added v220-4 Used with test 27 above
			if (origTEX[ld - lines] & 0x55)
			{
				return true;
			}
			break;
		case 30:	// v220-6
			if (!(ld->ml_flags & ML_TWOSIDED) && ld->ovflags & OVF_NOSKYWALL)
				return true;
			break;
		case 31:	// v220-6
			if (   ld->frontsector != ld->backsector
				&& ld->ml_flags & ML_TWOSIDED)
				return true;
			break;
		case 32:	// v220-6
			if (   ld->frontsector == ld->backsector
				&& ld->frontsector != ls->frontsector
				&& origTEX[ld - lines] == 0x22)
				return true;
			break;
		case 33:	// v220-7
			if ( !(ld->ml_flags & ML_TWOSIDED)
				&& ld->frontsector->floorheight + 4*FRACUNIT < ld->frontsector->ceilingheight
				&& ld->frontsector != ls->frontsector
				&& ld->frontsector != ls->backsector)
				return true;
			break;
		case 34:	// v220-8
			return !ld->backsector && origTEX[ld - lines] & 0x02;
		case 35:	// v220-16
			if (ld->mtflags & R3D_MISLOTEX)
			{
				int fh = ls->mtflags & R3D_MISLTEX1 ? ls->backsector->floorheight : ls->frontsector->floorheight;

				if (   ld->mtflags & R3D_MISLTEX1
					&& ld->backsector->floorheight != fh)
				{
					return true;
				}

				if (   ld->mtflags & R3D_MISLTEX2
					&& ld->frontsector->floorheight != fh)
				{
					return true;
				}
			}
			break;
		case 36:	// v220-33
			if (ld->mtflags & R3D_MISHITEX)
			{
				int ch = ls->mtflags & R3D_MISHTEX1 ? ls->backsector->ceilingheight : ls->frontsector->ceilingheight;

				if (   ld->mtflags & R3D_MISHTEX1
					&& ld->backsector->ceilingheight != ch)
				{
					return true;
				}

				if (   ld->mtflags & R3D_MISHTEX2
					&& ld->frontsector->ceilingheight != ch)
				{
					return true;
				}
			}
			break;
		default:
			Con_Error("PIT_ConnectedLine: unknown case %i",condition);
			break;
		}
	}
	return false;
}

static RDBOOL PIT_LineOnLine(line_t *ls, line_t *ld, int condition)
{
	// currently unused
	if (condition & 1)
	{
		if (vtxarray[ld->v1 - vertexes] < 2 ||
			vtxarray[ld->v2 - vertexes] < 2)
			return false;
	}

	if (condition & 2)
	{
		if ((ls->v1->x == ld->v1->x && ls->v1->y == ld->v1->y) ||
			(ls->v2->x == ld->v2->x && ls->v2->y == ld->v2->y) ||
			(ls->v2->x == ld->v1->x && ls->v2->y == ld->v1->y) ||
			(ls->v1->x == ld->v2->x && ls->v1->y == ld->v2->y))
			return false;
	}

	// currently unused
	if (condition & 4)
	{
		if (ld->frontsector == ld->backsector
			|| ld->dx && !ls->dx
			|| ld->dy && !ls->dy)
			return false;

		if (ls->v1 == ld->v1)
		{
			return  ls->v2 == ld->v2 ? false : VertexIsOnLine(ls->v2, ld);
		}
		if (ls->v1 == ld->v2)
		{
			return  ls->v2 == ld->v1 ? false : VertexIsOnLine(ls->v2, ld);
		}
		if (ls->v2 == ld->v1)
		{
			return  ls->v1 == ld->v2 ? false : VertexIsOnLine(ls->v1, ld);
		}
		if (ls->v2 == ld->v2)
		{
			return  ls->v1 == ld->v1 ? false : VertexIsOnLine(ls->v1, ld);
		}
		return false;
	}

	if (condition & (8|16))
	{
		if (condition & 8)
		{
			// v220-27
			if (   ls->frontsector == ld->frontsector
				&& ls->backsector == ld->backsector
				&& ls->v1->x == ld->v2->x && ls->v1->y == ld->v2->y
				&& ls->v2->x == ld->v1->x && ls->v2->y == ld->v1->y)
				return true;

			if (   ls->frontsector == ld->backsector
				&& ls->backsector == ld->frontsector
				&& ls->v1->x == ld->v1->x && ls->v1->y == ld->v1->y
				&& ls->v2->x == ld->v2->x && ls->v2->y == ld->v2->y)
				return true;

			return false;
		}

		if (ld->frontsector == ld->backsector)
		{
			if (  (ls->v1->x == ld->v1->x && ls->v1->y == ld->v1->y
				&& ls->v2->x == ld->v2->x && ls->v2->y == ld->v2->y))
				return true;

			if (  (ls->v1->x == ld->v2->x && ls->v1->y == ld->v2->y
				&& ls->v2->x == ld->v1->x && ls->v2->y == ld->v1->y))
				return true;
		}
		
		return false;
	}

	return  VertexIsOnLine(ls->v1, ld) &&
			VertexIsOnLine(ls->v2, ld);
}

static RDBOOL PIT_CoincidentLine(line_t* ls, line_t* ld, int condition)
{
	if (condition & 1
		&& ld->frontsector == ld->backsector)
	{
		return false;
	}

	if (condition & 2
		&& (ld->frontsector != ld->backsector
		|| ld->frontsector != ls->frontsector))
	{
		return false;
	}

	if (condition & 4)
	{
		if (ld->ml_flags & ML_BADLINE) return false;

		if (   ld->backsector
			|| vtxarray[ld->v1 - vertexes] < 2
			|| vtxarray[ld->v2 - vertexes] < 2)
		{
			return false;
		}

		return  VertexIsOnLine(ls->v1, ld) &&
				VertexIsOnLine(ls->v2, ld);
	}

	if (condition & 8)	// v220-22
	{
		if (ld->frontsector != ld->backsector) return false;

		if (   vtxarray[ld->v1 - vertexes] < 2
			|| vtxarray[ld->v2 - vertexes] < 2)
		{
			return false;
		}

		return  (VertexIsOnLine(ls->v1, ld) && VertexIsOnLine(ls->v2, ld))
			||  (VertexIsOnLine(ld->v1, ls) && VertexIsOnLine(ld->v2, ls));
	}

	if (condition & 16)	// v220-31
	{
		if (ld->frontsector == ld->backsector) return false;

		if (   ld->frontsector == ls->frontsector
			&& ld->backsector  == ls->backsector)
		{
			return false;
		}
	}

	return (ls->v1->x == ld->v1->x &&
			ls->v1->y == ld->v1->y &&
			ls->v2->x == ld->v2->x &&
			ls->v2->y == ld->v2->y)
			||
		   (ls->v2->x == ld->v1->x &&
			ls->v2->y == ld->v1->y &&
			ls->v1->x == ld->v2->x &&
			ls->v1->y == ld->v2->y);
}

static RDBOOL PIT_ConnectedVtx(line_t* ls, line_t* ld, int condition)
{
	if (condition == 0)
	{
		if (hitline[ld - lines])
		{
			return  (ls->v1->x == ld->v1->x && ls->v1->y == ld->v1->y)
				 || (ls->v1->x == ld->v2->x && ls->v1->y == ld->v2->y)
				 || (ls->v2->x == ld->v1->x && ls->v2->y == ld->v1->y)
				 || (ls->v2->x == ld->v2->x && ls->v2->y == ld->v2->y);
		}
	}
	else if (condition == 1 || condition == 2)	// v220-17
	{
		// NOTE: ls is a reference line and only the frontsector is set
		if (condition == 1)
		{
			if ( !(ld->mtflags & R3D_MISLOTEX) ) return false;
		}
		else
		{
			if (ld->mtflags & R3D_MISLOTEX) return false;
		}

		if (ld->frontsector == ld->backsector) return false;
		if (ld->frontsector == ls->frontsector) return false;
		if (ld->backsector  == ls->frontsector) return false;
		if (   ld->backsector
			&& ld->frontsector->floorheight == ld->backsector->floorheight)
		{
			return false;
		}

		if (P_PointOnLineSide(vtxc_x, vtxc_y, ld))
		{
			return ld->backsector
				&& ld->backsector->floorheight == ls->frontsector->floorheight
				&& ld->backsector->floorpic == ls->frontsector->floorpic;
		}

		return ld->frontsector->floorheight == ls->frontsector->floorheight
			&& ld->frontsector->floorpic == ls->frontsector->floorpic;
	}
	else if (condition == 3)	// v220-19
	{
		byte txn = origTEX[ld - lines];

		if (!hitline[ld - lines] && (txn == 0x20 || txn == 0x02))
		{
			if (  !ld->mtflags
				|| ld->mtflags == R3D_MISHTEX1
				|| ld->mtflags == R3D_MISHTEX2)
			{
				return  (ls->v1 == ld->v1 && ls->v2 == ld->v2)
					 || (ls->v1 == ld->v2 && ls->v2 == ld->v1);
			}
		}
	}

	return false;
}

// NOTE: ls a reference line and only the frontsector is set
static RDBOOL PIT_ConnectedBothVtx(line_t* ls, line_t* ld, int condition)
{
	if (ld->frontsector == ld->backsector) return false;
	if ( !(ld->mtflags & R3D_MISLOTEX) ) return false;
	if (ld->frontsector == ls->frontsector) return false;
	if (ld->backsector  == ls->frontsector) return false;

	if (    ld->frontsector->floorheight != ls->frontsector->floorheight
		&& !(ld->backsector && ld->backsector->floorheight == ls->frontsector->floorheight))
	{
		return false;
	}

	// NOTE: if in edit mode and vertexes are unmerged this will fail
	return  (ls->v1 == ld->v1 && ls->v2 == ld->v2)
		 || (ls->v1 == ld->v2 && ls->v2 == ld->v1);
}

static line_t *subfn_Type(line_t *ldc, int condition, int fnnum)
{
	line_t	*ldf;
    int		xl, xh, yl, yh, bx, by;

	if (fnnum != 9 && ldc->specType >= R3D_40) return NULL;

	validcount++;
	ldc->validcount = validcount;	//Don't test self

	M_getBlockOffsets(ldc->bbox, 0, &xl, &xh, &yl, &yh);

	for (bx=xl ; bx<=xh ; bx++)
	{
		for (by=yl ; by<=yh ; by++)
		{
			switch(fnnum)
			{
			case 0:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_CoincidentLine)) != NULL)
					return ldf;
				break;
			case 1:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_ConnectedLine)) != NULL)
					return ldf;
				break;
			case 2:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,1,1,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_IntersectLine)) != NULL)
					return ldf;
				break;
			case 3:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition, 0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_IntersectLine)) != NULL)
					return ldf;
				break;
			case 4:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition, 2,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_IntersectLine)) != NULL)
					return ldf;
				break;
			case 5:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_LineOnLine)) != NULL)
					return ldf;
				break;
			case 6:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_UnmergedVTXLine)) != NULL)
					return ldf;
				break;
			case 7:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_ConnectedVtx)) != NULL)
					return ldf;
				break;
			case 8:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_ConnectedBothVtx)) != NULL)
					return ldf;
				break;
			case 9:
				if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,3,(int (__cdecl *)(struct line_s *,struct line_s *, int))PIT_CoincidentLine)) != NULL)
					return ldf;
				break;
			default:
				Con_Error("ERROR [2]: fnnum %i", fnnum);
				break;
			}
		}
	}

	return NULL;
}

line_t *subfn_FindCoincidentLine(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 0);
}

line_t *subfn_FindLineConnected(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 1);
}

line_t *subfn_FindR3D_0BLineCrossing(line_t *ldc)
{
	return subfn_Type(ldc, LC_NOCONDITION, 2);
}

line_t *subfn_FindLineCrossing(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 3);
}

line_t *subfn_FindLineCrossingMidTex(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 4);
}

line_t *subfn_FindLineOnLine(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 5);
}

line_t *subfn_UnmergedVTXline(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 6);
}

// v220-3
line_t *subfn_FindVtxConnected(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 7);
}

// v220-18 (not currently used)
line_t *subfn_FindVtxBothConnected(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 8);
}

// v220-31
line_t *subfn_FindCoincident3DLine(line_t *ldc, int condition)
{
	return subfn_Type(ldc, condition, 9);
}

int subfn_FindAllLinesCrossing(line_t *ldc)
{
	int		xl, xh, yl, yh, bx, by;
	int		lnptr = 0;

	if (ldc->specType >= R3D_40) return 0;

	validcount++;
	ldc->validcount = validcount;	//Don't test self

	M_getBlockOffsets(ldc->bbox, 0, &xl, &xh, &yl, &yh);

	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			int			offset;
			const long	*list;
			
			if (bx < 0 || by < 0 || bx >= bmapwidth || by >= bmapheight)
				continue;

			offset = by*bmapwidth + bx;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				line_t *ld = &lines[*list];

				if (ld->validcount == validcount)
					continue;		// line has already been checked

				ld->validcount = validcount;

				if (LS_CheckLineIntersect(ldc, ld, LC_NORMAL|LC_NOVTXONLINE))
				{
					linearray[lnptr++] = *list;
				}
			}
		}
	}
	return lnptr;
}

RDBOOL LS_CheckLineIntersect(line_t *ref, line_t *cmp, int condition)
{
	if (ref == cmp ||
		ref->specType >= R3D_40 || cmp->specType >= R3D_40)
		return false;

	sampleline = NULL;

	return subfn_CheckLineIntersect(ref, cmp, condition);
}

int subfn_FindMultiLineCrossing(line_t *ldc, int condition)
{
	line_t *ldf;
	int		xl, xh, yl, yh, bx, by;
	int		count = 0;

	multicrossline = NULL;

	if (ldc->specType >= R3D_40) return 0;

	validcount++;
	ldc->validcount = validcount;	//Don't test self

	SET_countsec(0);

	// Extend search area by a mapblock in all directions - found to be necessary
	// as in some cases a line partially in a mapblock may not have been included
	// in that block when built by P_CreateBlockmap().

	M_getBlockOffsets(ldc->bbox, MAPBLOCKSIZE, &xl, &xh, &yl, &yh);

	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			if ((ldf = LS_BlockLinesIterator (bx,by,ldc,condition,0,PIT_IntersectLine)) != NULL)
			{
				if (!countsec[ldf->frontsector - sectors]
					&& ldf->frontsector != ldc->frontsector)
				{
					countsec[ldf->frontsector - sectors] = 1;
					count++;
				}
				if (ldf->ml_flags & ML_TWOSIDED
					&& !countsec[ldf->backsector - sectors]
					&& ldf->backsector != ldc->frontsector)
				{
					countsec[ldf->backsector - sectors] = 1;
					count++;
				}
			}
		}
	}
	return count;
}

// Deal with unclosed, unconnected support loops (rare)
void LS_CheckUnclosedMoveCrossingSecs()
{
	line_t	*ld, *lc;
	int		i, ssecptr = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& ld->frontsector->specFlags == 0x002D0000
			&& SecCtrl(ld->frontsector - sectors)
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& !IsLiquidFloorFlat(ld->frontsector)
			&& ld->frontsector->heightsec < 0
			&& (lc = subfn_FindLineCrossing(ld, LC_NORMAL)) != NULL)
		{
			sector_t *osec = NULL;
			int		 k, m, lnptr = 0, valid = true;

			if (   lnsec[ld - lines] < 0
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& lc->frontsector != ld->frontsector
				&& lc->frontsector == lc->backsector
				&& SecCtrl(ld->frontsector - sectors)
				&& !(lc->frontsector->specFlags & SF_MOVSECTOR))
			{
				lnsec[ld - lines] = lc->frontsector - sectors;
			}

			if (lnsec[ld - lines] < 0) continue;

			if ( (lnptr = subfn_FindAllLinesCrossing(ld)) < 2)
				continue;

			for (k = 0; k < lnptr; k++)
			{
				line_t	 *lc = &lines[linearray[k]];
				sector_t *chk = lc->frontsector;

				if (origTEX[linearray[k]] & 0x55
					|| lc->frontsector != lc->backsector
					|| chk == ld->frontsector)
				{
					valid = false;
					break;
				}

				if (!osec) osec = chk;
				if (osec != chk) { valid = false; break; }
			}

			if (!valid) continue;

			linearray[lnptr++] = ld - lines;

			for (k = 0; k < lnptr; k++)
			{
				ssline_t *sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					if (   ss->floorsec == ss->ceilsec
						&& ss->floorsec != lnsec[ld - lines]
						&& !(ss->ssflags & SSF_NOCHANGE))
					{
						Clear_SSEC(ss);
							
						if (ss->sector - sectors != lnsec[ld - lines])
							SetOverride(ss, lnsec[ld - lines], 14001);

						if (ss->sector != ld->frontsector)
							ChangeSSECsec(ss, ld->frontsector, 14002);
					}
				}
			}
		}
	}
}
