//
// R3D_LINEDEBUG.C - b220-0 Copyright (C) 2006 - 2008 G Jackson
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty
// of FITNESS FOR A PARTICULAR PURPOSE.
// See the DOOM Source Code License for more details.

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <process.h>

#include "m_bams.h"
#include "r_data.h"
#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "sys_system.h"

// public
FILE	*dbfile = NULL;
RDBOOL	dbfile_active;

// private
static char *header = NULL;

// Debug routines
////////////////////////////////////////////////////////////////////////////
//
// Checking aid sub functions for routines called in R3D map analyser files
// These can be used to trap specific calls using the num parm
//

RDBOOL openDBFILE()
{
	if (!dbfile)
	{
extern char		DEBUGfile[MAXPATH];		// Path\R3Ddebug.out

		if ( (dbfile = fopen(DEBUGfile, "w")) == NULL)
			return false;
		fprintf(dbfile, "======= DEBUG INFO file %s ========\n\n", DEBUGfile);
	}

	return true;
}

void SetProgHeader(char *message)
{
	header = message;
}

void ResetProgHeader()
{
	header = NULL;
}

void PrintProgInfo(char *message, ...)
{
	openDBFILE();

	if (dbfile && header)
	{
		fprintf(dbfile, header);
		header = NULL;
	}

	if (dbfile)
	{
		char buffer[1024];
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(buffer, 1022, message, argptr);
		va_end(argptr);
		
		buffer[1023] = 0;
		fprintf(dbfile, buffer);
		dbfile_active = true;
	}
}

void PrintProgError(char *message, ...)
{
	openDBFILE();

	if (dbfile && header)
	{
		fprintf(dbfile, header);
		header = NULL;
	}

	if (dbfile)
	{
		char buffer[1024];
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(buffer, 1022, message, argptr);
		va_end(argptr);
		
		buffer[1023] = 0;
		fprintf(dbfile, buffer);
		dbfile_active = true;
	}

	Sys_Quit(false);
}

void SetControlline(line_t *ld, int num)
{
	ld->spflags |= SPF_CONTROLLINE;
}

void SetNOocclude(sector_t *sec, int num)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld;
	int			i, k;

	sec->specFlags |= SF_NOOCCLUDE;

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if ( !(ld->spflags & SPF_CONTROLLINE) )
		{
			sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];
				ss->ssflags |= SSF_NOOCCLUDE;
			}
		}
	}
}

void Setspec(line_t *ld, int val, int num)
{
	int specSec = (ld->spflags & SPF_CONTROLLINE ? -1 : val);

	if (ld->specType >= R3D_40) return;

	lnsec[ld - lines] = specSec;
}

void SetLFG(line_t *ld, short lfgflag, int num)
{
	lineflags[ld - lines] |= lfgflag;
}

void SetLFGSecSet(line_t *ld, int num)
{
	SetLFG(ld, LFG_LINESSECSET, num);
}

void Setloop(line_t *ld, int val, int num)
{
	ld->tyflags |= val;
}

// Settype PreAnalysed
// If using an analysis file then lineflags will be NULL but R3D_4x
// types are always set up.  So this is used by r3d_sl3D.c
void SettypePA(line_t *ld, byte val, int num)
{
	if (!(ld->ml_flags & ML_TWOSIDED))
		Con_Error("SettypePA: 3D line %i from %i\n", ld - lines, num);
	else
		ld->specType = val;
}

void Settype(line_t *ld, byte val, int num)
{
	byte	oldval  = ld->specType;
	int		linenum = ld - lines;

	if (   lineflags[linenum] & LFG_FIXED
		|| oldval == R3D_0F
		|| oldval == R3D_31
		|| oldval == R3D_32
		|| oldval == R3D_36
		|| oldval >= R3D_40
		|| (val == R3D_1B
		&& ((ld->mtflags == R3D_MISLTEX1 && ld->frontsector->specFlags & SF_SSLINETRUE)
		||  (ld->mtflags == R3D_MISLTEX2 && ld->backsector->specFlags  & SF_SSLINETRUE))) )
	{
		return;
	}

	if (!(ld->ml_flags & ML_TWOSIDED))
	{
		ld->specType = R3D_00;
	}
	else
	{
		ld->specType = val;
	}
}

void Forcetype(line_t *ld, byte val, int num)
{
	if (lineflags[ld - lines] & (LFG_FIXABS|LFG_3DTYPE))
		return;

	if (!(ld->ml_flags & ML_TWOSIDED))
		ld->specType = R3D_00;
	else
		ld->specType = val;
}


void Absolutetype(line_t *ld, byte val, int num)
{
	if (lineflags[ld - lines] & LFG_3DTYPE)	// always excluded
		return;

	if (!(ld->ml_flags & ML_TWOSIDED))
		ld->specType = R3D_00;
	else
		ld->specType = val;
}

void Setlfsec(sector_t *sec, int secnum, int num)
{
	lfsecarray[sec - sectors] = secnum;
}

void SetNoskywall(int lnum, int num)
{
	lines[lnum].ovflags |= OVF_NOSKYWALL;
}

void SetSkywall(line_t *ld, int num)
{
	ld->tyflags |= TYF_SKYWALL;
}

void Setsecdone(sector_t *sec, int num)
{
	secdone[sec - sectors] = 1;
}

void Setsecdone2(sector_t *sec, int num)
{
	secdone[sec - sectors] = 2;
}

void Setsecdone3(sector_t *sec, int num)
{
	secdone[sec - sectors] = 3;
}

void SetSecidxDone(int secidx, int num)
{
	secdone[secidx] = 1;
}

void SetMissinglo(line_t *ld, int num)
{
	ld->ml_flags |= ML_IGNOREMISSINGLO;
}

void ResetMissinglo(line_t *ld, int num)
{
	if (lineflags[ld - lines] & LFG_OVERRIDE) return;
	ld->ml_flags &= ~ML_IGNOREMISSINGLO;
}

void OrLN2flag(line_t *ld, int flag, int num)
{
	ld->lnflags2 |= flag;
}

void Clr_allmtflags(line_t *ld, int num)
{
	ld->mtflags = 0;
}

void Clr_mthiflags(line_t *ld, int num)
{
	ld->mtflags &= ~R3D_MISHITEX;
}

void Clr_mtloflags(line_t *ld, int num)
{
	ld->mtflags &= ~R3D_MISLOTEX;
}

void inhibitSSEC(subsector_t *ss, int num)
{
	ss->ssflags |= SSF_01INHIBIT;
}

void setSSECflag(subsector_t *ss, int flag, int num)
{
	ss->ssflags |= flag;
}

//=========================================================================================

void Setr3dcs(sector_t *s, int newsec, int num)
{
	s->R3Dceilsec = newsec;
}

void Setr3dch(sector_t *sec, sector_t *s, int num)
{
	sec->R3Dceilheight = s->ceilingheight;
	if (sec->ceilinglightsec < 0)
		sec->ceilinglightsec = s - sectors;
}

void Setr3dch_byH(sector_t *sec, int height, int num)
{
	sec->R3Dceilheight = height;
}

void Setr3dchangech(sector_t *sec, int ht, int num)
{
	pr_OFinsert("  MAP ERROR FIXED: sector %i incorrect ceiling height changed from %i to %i\n",
		sec - sectors, sec->ceilingheight>>16, ht>>16);

	ma_inf.maperror++;
	sec->ceilingheight = ht;
}

//=========================================================================================

void Setr3dfs(sector_t *s, int newsec, int num)
{
	s->R3Dfloorsec = newsec;
}

void Setr3dfh(sector_t *sec, sector_t *s, int num)
{
	if (s->heightsec < 0 && !(s->floorheight & FRACUNIT))
	{
		sec->R3Dfloorheight = s->floorheight;
		sec->floorlightsec = s - sectors;	// $light needs looking at
	}

	Setsecdone(sec, num);
}

void Setr3dfh_byH(sector_t *sec, int height, RDBOOL test, int num)
{
	if (!test || !(height & FRACUNIT))
	{
		sec->R3Dfloorheight = height;
		Setsecdone(sec, num);
	}
}

//=========================================================================================

void Setr3dssfs(subsector_t *ss, int newssfs, int num)
{
	ss->floorsec = newssfs;
	if (newssfs >= 0) ss->ssflags |= SSF_INHSECLIGHT;
}

void Setr3dsscs(subsector_t *ss, int newsscs, int num)
{
	if (ss->ssflags & SSF_CSECFIXED) return;

	ss->ceilsec = newsscs;
}

void Sethir3dssfs(subsector_t *ss, int newssfs, int num)
{
	if (ss->floorsec >= 0 && newssfs >= 0
		&& sectors[ss->floorsec].floorheight > sectors[newssfs].floorheight)
	{
		return;
	}
	if (ss->sector->floorheight >= sectors[newssfs].floorheight)
	{
		return;
	}

	ss->floorsec = newssfs;
	ss->ssflags |= SSF_INHSECLIGHT;
}

void Setlor3dsscs(subsector_t *ss, int newsscs, int num)
{
	if (ss->ceilsec >= 0 && newsscs >= 0
		&& sectors[ss->ceilsec].ceilingheight < sectors[newsscs].ceilingheight)
	{
		return;
	}

	ss->ceilsec = newsscs;
}

void Setr3dsslcsec(subsector_t *ss, int newsscs, int num)
{
	if (newsscs >= 0 && sectors[newsscs].ceilingpic == skyflatnum) return;

	ss->lcsec = newsscs;
}

void Setr3dsslfsec(subsector_t *ss, int newsslfs, int num)
{
	ss->lfsec = newsslfs;
}

//=========================================================================================

void SetOverride(subsector_t *ss, int ovrnum, int num)
{
	if (ss->sector->stptr < 0)
		ss->floorsec = ss->ceilsec = ovrnum;
}

void SetFloorlightsec(subsector_t *ss, int newsec, int num)
{
	ss->sslightsec = newsec;
	ss->ssflags &= ~SSF_INHSECLIGHT;
}

void ChangeSSECsec(subsector_t *ss, sector_t *s, int num)
{
	if (ss->ssflags & SSF_BASESSECFIX) return;

	if (ss->sector != s) ss->sector = s;
}

void SetSSF_BASESSECFIX(subsector_t *ss, int num)
{
	ss->ssflags |= SSF_BASESSECFIX;
}
