//-----------------------------------------------------------------------------
//
// Copyright (C) 2010 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_lodmodels.c
//
// First written (v220-9) 18/Dec/2010 gmj
//
//
// DESCRIPTION:   State management for lod models.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>

#include "con_console.h"
#include "dd_main.h"
#include "dedfile.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct {
	int type;
	int lodstate;
	char lodstatename[24];
} lodlink_t;

// state names for lod replacements
lodlink_t *lodlinks;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int	dropped_lodlink[3];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int maxlodnum;

// CODE --------------------------------------------------------------------

void H_InitLODScript()
{
	int		c, i, dummy_err = false;
	char	fname[512];
	char	*nextstr;

	if (!(loadModels & 2)) return;

	sprintf(fname, "%sDefs\\Base\\R3D_LodDefs.ded", ddRuntimeDir.path);
	
	if (!SR_LoadFile(fname, "LOD_FILE", &dummy_err)) return;

	if (dummy_err) 	Con_Error("ERROR: Bad file %s\nReinstall Risen3D", fname);

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "LODdef"))
		{
			nextstr = SR_NextSTR(true);
			if (_stricmp(nextstr, "{")) break;

			nextstr = SR_NextSTR(true);
			nextstr = SR_NextSTR(true);
			nextstr = SR_NextSTR(true);
			if (!_stricmp(nextstr, "}")) maxlodnum++;
		}
	}

	if (!maxlodnum) { SR_EndScript(); return; }

	lodlinks = (lodlink_t*)W_Calloc(maxlodnum * sizeof(lodlink_t), STATICID, 400);

	SR_RewindScript();

	c = 0;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "LODdef"))
		{
			RDBOOL	found = false;

			nextstr = SR_NextSTR(true);
			if (_stricmp(nextstr, "{")) break;

			nextstr = SR_NextSTR(true);

			// Note: MT_STROBE1 is the minimum number that defs.count.mobjs.num will have
			//       LOD's are not supported for thing numbers greater than this
			for (i = 0; i < MT_STROBE1; i++)
			{
				ded_mobj_t	*dmo = defs.mobjs + i;

				if (!_stricmp(dmo->id, nextstr))
				{
					nextstr = SR_NextSTR(true);

					if (strlen(nextstr) > 23)
					{
						SR_EndScript();
						Con_Error("LODdef state name \"%s\" length exceeds max allowed of 23\n", nextstr);
					}

					strcpy(lodlinks[c].lodstatename, nextstr);

					lodlinks[c].type = i;
					found = true;
					break;
				}
			}

			if (!found)
			{
				SR_EndScript();
				Con_Error("LODdef ID \"%s\" not recognised\n", nextstr);
			}

			nextstr = SR_NextSTR(true);
			if (_stricmp(nextstr, "}")) break;
				
			c++;				
		}
	}
	
	SR_EndScript();

	if (c != maxlodnum)
		Con_Error("PROGRAM ERROR: file %s\n", fname);
}

void H_InitLODstates()
{
int Def_GetStateLODNum(char *id);

	// Get the lodstates (if they exist)
	for (int i = 0; i < maxlodnum; i++)
	{
		int stnum = Def_GetStateLODNum(lodlinks[i].lodstatename);

		if (stnum > 0) lodlinks[i].lodstate = stnum;
	}
}

static int getLodState(int mtnum)
{
	int		i;

	for (i = 0; i < maxlodnum; i++)
	{
		if (lodlinks[i].type == mtnum)
			return lodlinks[i].lodstate;
	}

	return 0;
}

void H_SetTHINGsLODstate()
{
	thinker_t	*th;

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		mobj_t *iter = (mobj_t*)th;

		if (   iter->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& iter->type < MT_STROBE1)
		{
			iter->lodstate = getLodState(iter->type);
		}
	}

	// set pickup lodstates
	dropped_lodlink[0] = getLodState(MT_CLIP);
	dropped_lodlink[1] = getLodState(MT_SHOTGUN);
	dropped_lodlink[2] = getLodState(MT_CHAINGUN);
}
