//-----------------------------------------------------------------------------
//
// Copyright (C) 2013 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_LoopLines.cpp	First release December 2013 (v220-22) gmj
//
// DESCRIPTION: Determine, if possible, subsectors belonging inside a
//				closed line loop containing same sided lines owned by
//				another sector. This relies on the map being properly
//              constructed (most are). Where not then later routines
//              are used to mop up bad map design.
// Revision 2.2.0 build 23 (v220-23) 10/Jan/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "sys_system.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------


// ===============================================================

static RDBOOL addSecLocal(int secidx, int *array, int *count)
{
	int i;

	for (i = 0; i < *count; i++)
	{
		if (array[i] == secidx) return false;
	}

	i = (*count)++;
	array[i] = secidx;

	return true;
}

static int countLoops(sector_t *sc)
{
	int i, n, cnt = 0;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			n = sc->lines[i] - lines;
			if (conarray[n] == 1) { lnum = n; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(conarray, sc, lnum, NULL, NULL, NULL, NULL))
			cnt++;
		else
			conarray[lnum] = 0;
	}

	return cnt;
}

static void checkConflicting()
{
	subsector_t *ss;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k;

	SET_countsec(0);

	for (c = i = 0; i < numsectors; i++)
	{
		// Note sector will be set closed if, amongst other things, all lines are same sided
		RDBOOL	valid = (sectors[i].specFlags & 0x3041C) == 0x1C;

		if (valid)
		{
			int		lc = sectors[i].linecount;

			for (k = 0; k < lc; k++)
			{
				ld = sectors[i].lines[k];

				if (ld->frontsector != ld->backsector)
				{
					valid = false; break;
				}

				// quick check that lines are not looped/disconnected
				if (   vtxarray[ld->v1 - vertexes] <= 2
					|| vtxarray[ld->v2 - vertexes] <= 2)
				{
					valid = false; break;
				}
			}
		}

		if (valid)
		{
			countsec[i] = true; c++;
		}
	}

	if (c < 2) return;

	for (c = i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (countsec[ss->sector - sectors])
		{
			// There are two possibilities;
			// 1) the subsec only references one valid sector
			//    in which case an alternative must be found
			// 2) the subsec references two valid sectors
			//    in which case that with the lowest floor height
			//    must be used.

			RDBOOL		setflag = false;

			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (   sg->frontsector == sg->backsector
					&& sg->frontsector != ss->sector
					&& countsec[sg->frontsector - sectors])
				{
					setflag = true; break;
				}
			}

			if (setflag)
			{
				for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
				{
					if (   sg->frontsector == sg->backsector
						&& sg->frontsector != ss->sector
						&& countsec[sg->frontsector - sectors])
					{
						if (sg->frontsector->floorheight < ss->sector->floorheight)
						{
							ChangeSSECsec(ss, sg->frontsector, 51001);
							ma_inf.changeconflicting++;
							break;
						}
					}
				}
			}

			if (setflag) hitssec[c++] = i;
		}
	}

	// check lines belonging to the ssecs

	if (c)
	{
		int m, lac = 0;

		validcount++;;
		ss_validcount++;;

		for (i = 0; i < c; i++)
		{
			ss = &subsectors[hitssec[i]];
			ss->ssvalidcount = ss_validcount;

			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (   sg->frontsector == sg->backsector
					&& sg->frontsector != ss->sector
					&& countsec[sg->frontsector - sectors])
				{
					if (sg->linedef && sg->linedef->validcount != validcount)
					{
						sg->linedef->validcount = validcount;
						linearray[lac++] = sg->linedef - lines;
					}
				}
			}
		}

		for (i = 0; i < lac; i++)
		{
			ssline_t *sl = &sslines[linearray[i]];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;
				ss->ssvalidcount = ss_validcount;
				if (!countsec[ss->sector - sectors]) continue;

				for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
				{
					if (sg->frontsector != sg->backsector
						&& (!(sg->frontsector->specFlags & 0xC) || sg->frontsector->specFlags & 0x400))
					{
						ss->ssflags &= ~SSF_BASESSECFIX;
						ChangeSSECsec(ss, sg->frontsector, 51002);
						ss->ssflags |= SSF_BASESSECFIX;
						ma_inf.changeconflicting++;
						break;
					}
				}
			}
		}
	}
}

// If sec is true then test for one's lines can be made
static RDBOOL subfn_checkSSECarray(sector_t *sc, int count, int *fhc, int *flc)
{
	sector_t	*s2 = NULL, *ch;
	line_t		*ld;
	int			*lmem = (int*)malloc(sc->linecount * sizeof(int));
	int			c, d, e, i, k;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (conarray[ld - lines] == 1)
		{
			if (origTEX[ld - lines] & 0x55) { free(lmem); return false; }
			lmem[c++] = ld - lines;
		}
	}

	if (c < 6 || countLoops(sc) != 2)
	{
		goto end_check;
	}

	for (d = e = i = 0; i < c; i++)
	{
		ld = &lines[lmem[i]];

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (!ch) { e++; continue; }
		if (!s2) s2 = ch; else if (s2 != ch) { s2 = NULL; goto end_check; }

		if (ch->floorheight > sc->floorheight) d++ ; else { s2 = NULL; goto end_check; }
	}

	if (s2 && d + e == c) *flc = s2 - sectors; else s2 = NULL;

end_check:;

	if (!s2)
	{
		for (i = 0; i < c; i++) conarray[lmem[i]] = 1;
		free(lmem);
		return false;
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& hitline[sg->linedef - lines] != 4
				&& hitline[sg->linedef - lines] != 64)
			{
				if (!sg->linedef->backsector)
				{
					if (sg->linedef->frontsector != sc)
					{
						for (i = 0; i < c; i++) conarray[lmem[i]] = 1;
						free(lmem);
						return false;
					}
				}
				else if (sg->linedef->frontsector != sg->linedef->backsector)
				{
					if (sg->linedef->frontsector == sc && sg->linedef->backsector == s2) continue;
					if (sg->linedef->frontsector == s2 && sg->linedef->backsector == sc) continue;

					for (i = 0; i < c; i++) conarray[lmem[i]] = 1;
					free(lmem);
					return false;
				}
			}
		}
	}

	for (i = 0; i < c; i++) hitline[lmem[i]] = 8;
	free(lmem);
	*flc = s2 - sectors;

	return true;
}

static int mtxline(line_t *ld)
{
	if (ld->backsector && ld->frontsector != ld->backsector)
	{
		if (   !(origTEX[ld - lines] & 0x11)
			&& ld->backsector->floorheight != ld->frontsector->floorheight
			&& !(ld->backsector->floorpic == skyflatnum
			&&   ld->frontsector->floorpic == skyflatnum))
		{
			if (   vtxflags2[ld->v1 - vertexes] == 2
				&& vtxflags2[ld->v2 - vertexes] == 2
				&& Vtxs2(ld))
			{
				return -1;
			}

			return 1;
		}

		if (   !(origTEX[ld - lines] & 0x44)
			&& ld->backsector->ceilingheight != ld->frontsector->ceilingheight
			&& !(ld->backsector->ceilingpic == skyflatnum
			&&   ld->frontsector->ceilingpic == skyflatnum))
		{
			if (   vtxflags2[ld->v1 - vertexes] == 2
				&& vtxflags2[ld->v2 - vertexes] == 2
				&& Vtxs2(ld))
			{
				return -2;
			}

			return 1;
		}
	}

	return 0;
}

// box must be set
static int subfn_FindClosedRing(byte *lnarray, sector_t *sec, int *lnum, int *box, RDBOOL uselnum)
{
	line_t		*ld, *li;
	vertex_t	*v;
	short		*n1 = NULL, *n2 = NULL;
	int			*c1 = NULL;
	int			c, i, n;
	int			count = 0;
	int			*lnmem;
	RDBOOL		isclosed = false, repeat = true;

	if (!uselnum) *lnum = -1;
	lnmem = (int*)malloc(sec->linecount * sizeof(int));

	for (c = i = 0; i < sec->linecount; i++)
	{
		li = sec->lines[i];

		if (lnarray[li - lines] == 1)
		{
			lnmem[c++] = li - lines;
		}
	}

	if (c < 3) { count = -1; goto tidyup; }

	c1 = (int*)malloc(c * sizeof(int));
	n1 = (short*)calloc(c, sizeof(short));
	n2 = (short*)calloc(c, sizeof(short));

	if (!uselnum)
	{
		for (i = 0; i < c; i++)
		{
			n1[i] = vtxcheck[lines[lnmem[i]].v1 - vertexes];	
			n2[i] = vtxcheck[lines[lnmem[i]].v2 - vertexes];
		}

		for (i = 0; i < c; i++)
		{
			if      (n1[i] < 2 || n2[i] < 2)
				lnarray[lnmem[i]] = 0;
			else if (n1[i] % 2 || n2[i] % 2)
				lnarray[lnmem[i]] = 0;
			else if (n1[i] > 2 || n2[i] > 2)
			{
				*lnum = lnmem[i];
			}
		}

		if (*lnum < 0) *lnum = lnmem[0];
	}

	lnarray[*lnum] = 2;
	ld = &lines[*lnum];
	v = ld->v1;
	if (vtxcheck[v - vertexes] > 2) v = ld->v2;

	while (repeat)
	{
		for (i = 0, repeat = false; i < c; i++)
		{
			n = lnmem[i];
			li = &lines[n];

			if (lnarray[n] == 1)
			{
				if (li->v1 == v)
				{
					v = li->v2; lnarray[n] = 2; repeat = true;
				}
				else if (li->v2 == v)
				{
					v = li->v1; lnarray[n] = 2; repeat = true;
				}
			}
		}
	}

	isclosed = v == ld->v2;

	if (box) M_ClearBox(box);

	for (i = 0; i < c; i++)
	{
		if (lnarray[lnmem[i]] == 2)
		{
			if (isclosed)
			{
				lnarray[lnmem[i]] = 4; count++;
				if (box) M_AddLineToBox(box, &lines[lnmem[i]]);
			}
			else
			{
				lnarray[lnmem[i]] = 1;
			}
		}
	}

tidyup:;

	free(c1);
	free(n1);
	free(n2);
	free(lnmem);

	return count < 0 ? -1 : count > 2 ? count : 0;
}

static RDBOOL checkOrphanSectors(sector_t *sc, byte *bSSF, int cnt)
{
	subsector_t	*ss;
	sector_t	*ch;
	line_t		*ld;
	seg_t		*sg;
	byte		*liarray = (byte*)calloc(numlines, sizeof(byte));
	RDBOOL		valid = false;
	int			*llmem = (int*)calloc(cnt, sizeof(int));
	int			*ssecmem;
	int			hsp, box[4];
	int			i, k;

	memcpy(llmem, linearray, cnt * sizeof(int));

	for (i = 0; i < cnt; i++) liarray[llmem[i]] = 1;

	hsp = hitssecptr;
	ssecmem = (int*)malloc(hsp * sizeof(int));
	memcpy(ssecmem, hitssec, hsp * sizeof(int));

	while(1)
	{
		int		rc, lnum = -1;

		for (i = 0; i < cnt; i++) { if (liarray[llmem[i]] == 1) { lnum = llmem[i]; break; } }

		if (lnum < 0) break;

		ch = lines[lnum].frontsector;

		if ( (rc = subfn_FindClosedRing(liarray, ch, &lnum, box, true)) < 0) break;

		if (!rc)
		{
			ch = lines[lnum].backsector;

			if (subfn_FindClosedRing(liarray, ch, &lnum, box, true) <= 0) { valid = false; goto tidy_up; }
		}

		if (   LS_MarkSubsecsInAreaGeneral(ch, liarray, 0, NULL, box, -1, MSIA_NONE)
			|| LS_MarkSubsecsInAreaGeneral(ch, liarray, 0, NULL, box, -1, MSIA_REVERSED))
		{
			byte flags = 0;

			valid = true;
			validcount++;

			for (i = 0; i < hitssecptr; i++)
			{
				ss = &subsectors[hitssec[i]];
				sg = &segs[ss->firstline];

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if ( (ld = sg->linedef) && ld->validcount != validcount)
					{
						ld->validcount = validcount;

						if (origTEX[ld - lines] & 0x11) flags |= 1;
						if (origTEX[ld - lines] & 0x44) flags |= 2;
					}
				}

				if (flags == 3) break;
			}

			for (i = 0; i < hitssecptr; i++) bSSF[hitssec[i]] |= flags;
		}

		for (i = 0; i < cnt; i++) { if (liarray[llmem[i]] == 4) liarray[llmem[i]] = 0; }
	}

	memcpy(hitssec, ssecmem, hsp * sizeof(int));
	hitssecptr = hsp;

tidy_up:;

	free(ssecmem);
	free(liarray);
	free(llmem);
	return valid;
}

static int getOtherSec(int *lmem, int lmemc, sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*ch;
	sector_t	*os = NULL;
	ssline_t	*sl;
	line_t		*ld, *li;
	seg_t		*sg;
	int			i, k, m;

	ss_validcount++;

	for (i = 0; i < lmemc; i++)
	{
		sl = &sslines[lmem[i]];
		li = &lines[lmem[i]];
		ch = li->frontsector == sc ? li->backsector : li->frontsector;

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if ( !(ld = sg->linedef) ) continue;
				if (ld == li) continue;
				if (ld->frontsector == ld->backsector) goto next_ssec;
				if (sg->frontsector == sc || sg->frontsector == ch) continue;
				if (!os)
				{
					os = sg->frontsector;
					if ((os->specFlags & 0x40C) == 0xC) return -1;
				}
				else if (os != sg->frontsector)
				{
					return -1;
				}
			}
		}
next_ssec:;
	}

	return os ? os - sectors : -1;
}

// If the sector has no hilo textures or 1's lines then it may not qualify as
// a viable sector. This can happen if a line loop has different front and
// back sectors that do not qualify contained in another sector that does but
// has not been added to the list. Such are the vagaries of map design that
// have to be dealt with. In all cases that have been encountered it has been
// caused by overly and unnecessarily complex map design.
static void verifySector(int secidx, int *cntsecnum)
{
	sector_t *s = &sectors[secidx];

	if ((s->specFlags & 0x40C) == 0xC)
	{
		sector_t	*ch;
		line_t		*ld;
		int			c, i;

		CLR_hitline();
		CLR_vtxcheck();

		for (c = i = 0; i < s->linecount; i++)
		{
			ld = s->lines[i];

			if (   ld->frontsector != ld->backsector
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& vtxarray[ld->v1 - vertexes] > 1
				&& vtxarray[ld->v2 - vertexes] > 1)
			{
				linearray[c++] = ld - lines;
				hitline[ld - lines] = 1;
				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}
		}

		if (c < 3) return;	// nothing to check

		while (1)
		{
			sector_t	*os = NULL;
			RDBOOL		valid;
			int			*lmem;
			int			lmemc;
			int			addsec = -1;
			int			lnum = -1;
			int			rc;

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 1)
				{
					lnum = linearray[i]; break;
				}
			}

			if ( (rc = subfn_FindClosedRing(hitline, s, &lnum, NULL, false)) <= 0)
				return;

			lmem = (int*)malloc(rc * sizeof(int));
			lmemc = 0;
			
			for (i = 0, valid = true; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 8;
					ld = &lines[linearray[i]];
					ch = ld->frontsector == s ? ld->backsector : ld->frontsector;
					if (!ch) continue;
					if (   s->floorheight == ch->floorheight
						&& s->ceilingheight == ch->ceilingheight)
					{
						lmem[lmemc++] = linearray[i];
					}

					if (!os)
					{
						os = ch;
						if ((ch->specFlags & 0x40C) != 0xC)
						{
							addsec = ch - sectors;
						}
					}
					else if (os != ch)
					{
						valid = false;
					}
				}
			}
			
			if (valid && addsec >= 0)
			{
				int repsec;

				// check for superimposed sector lines (very rare but has to be done)
				if (lmemc == rc && (repsec = getOtherSec(lmem, lmemc, os)) >= 0)
				{
					addSec(repsec, cntsecnum);
				}
				else
				{
					addSec(addsec, cntsecnum);
				}

			}
			else if (!valid)
				addSec(secidx, cntsecnum);

			free(lmem);
		}
	}
	else
	{
		addSec(secidx, cntsecnum);
	}
}

static void freeSortbox(sortbox_t *sbx, int bc)
{
	int		i;

	for (i = 0; i < bc; i++) free(sbx[i].llist);
}

static int boxSort(sector_t *s, sortbox_t *sbx, int *lnlist)
{
	sortbox_t	*sbp;
	line_t		*ld;
	int			bc = 0;
	int			lcnt, lnum;
	int			c, i, rc;

	CLR_hitline();
	CLR_vtxcheck();

	for (lcnt = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (   ld->frontsector != ld->backsector
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& vtxarray[ld->v1 - vertexes] > 1
			&& vtxarray[ld->v2 - vertexes] > 1)
		{
			lnlist[lcnt++] = ld - lines;
			hitline[ld - lines] = 1;
			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;
		}
	}

	if (lcnt < 3) return 0;	// nothing to check

	while (1)
	{
		if (bc == MAXSORTBOXES)	Con_Error("program error: boxSort max number of %i exceeded", MAXSORTBOXES);

		sbp = &sbx[bc];

		if ( (rc = subfn_FindClosedRing(hitline, s, &lnum, sbp->box, false)) < 0) break;
		
		if (rc > 0 )
		{
			bc++;
			
			sbp->area = ((sbp->box[BOXRIGHT] - sbp->box[BOXLEFT]) >> FRACBITS)
					  * ((sbp->box[BOXTOP  ] - sbp->box[BOXBASE]) >> FRACBITS);
			sbp->llist = (int*)malloc(rc * sizeof(int));
			sbp->listcnt = rc;

			if (sbp->area <= 0) Con_Error("program error: boxSort area");

			for (c = i = 0; i < lcnt; i++)
			{
				if (hitline[lnlist[i]] == 4)
				{
					if (c == rc) Con_Error("program error: boxSort c %i == rc %i", c, rc);
					sbp->llist[c++] = lnlist[i];
				}
			}

			if (c != rc) Con_Error("program error: boxSort c %i != rc %i", c, rc);

			for (i = 0; i < lcnt; i++)
			{
				if (hitline[lnlist[i]] == 4) hitline[lnlist[i]] = 0;
			}
		}
		else	// loop is not closed
		{
			freeSortbox(sbx, bc);
			return 0;
		}
	}

	return bc;
}

static void subfn_marklines(sector_t *s, byte *linemark)
{
	subsector_t	*ss;
	line_t		*lc, *ld;
	seg_t		*sg;
	int			i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) )
			{
				linemark[ (ld - lines) ] = true;

				// Check logical sector reference
				if (   ss->sector == s
					&& ld->frontsector == ld->backsector
					&& lxarray[ld - lines] & 64
					&& (lc = subfn_FindLineCrossing(ld, LC_SAMESECFMOV)) != NULL)
				{
					int flags = ss->ssflags;

					ss->ssflags &= ~SSF_BASESSECFIX;
					ChangeSSECsec(ss, lc->frontsector, 51003);
					SetSSF_BASESSECFIX(ss, 51003);
				}
			}
		}
	}
}

static RDBOOL subfn_chkLinesInSSECarray(sector_t *sec, byte *mrklinearray, int *same)
{
	sector_t	*fs, *bs;
	int			i, k;

	for (i = 0, *same = false; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& mrklinearray[sg->linedef - lines] != 4)
			{
				fs = sg->linedef->frontsector;
				bs = sg->linedef->backsector;

				if (!bs)
				{
					if (fs != sec) return false;
				}
				else if (fs != bs)
				{
					if (   fs->floorheight != bs->floorheight
						|| fs->ceilingheight != bs->ceilingheight)
					{
						*same = false;
						return false;
					}

					*same = true;
				}
			}
		}
	}

	return true;
}

static void chkMissing(sector_t *s, sortbox_t *sbp, byte *linemark)
{
	sector_t	*ch, *os = NULL;
	line_t		*ld;
	RDBOOL		flv = true, fhv = true;
	int			c, i, same;

	// check for textureless sector
	for (c = i = 0; i < sbp->listcnt; i++)
	{
		ld = &lines[sbp->llist[i]];

		ch = ld->frontsector == s ? ld->backsector : ld->frontsector;

		if (!ch) return;
		if (!os) os = ch; else if (os != ch) return;
		if (origTEX[ld - lines] & 0x11) flv = false;
		if (origTEX[ld - lines] & 0x44) fhv = false;
		if (ld->spflags & SPF_NOTEX || ld->mtflags & R3D_MISTEXANY) c++;
	}

	if (c < 2)	// v220-31 At least two lines should have missing textures
	{
		return;
	}

	if (flv && ((os->specFlags  & 0x3041C) != 0x1C || s->floorheight   >= os->floorheight  )) flv = false;
	if (fhv && ((os->specFlags  & 0x3040C) != 0x0C || s->ceilingheight >= os->ceilingheight)) fhv = false;

	CLR_conarray();

	for (i = 0; i < sbp->listcnt; i++)
	{
		conarray[sbp->llist[i]] = 4;
	}

	if (!subfn_chkLinesInSSECarray(s, conarray, &same)) return;
	if (!flv && !fhv && !same) return;

	if (same && !flv && !fhv)
	{
		// V220-33 ammended (sometimes no changes are needed at all depending on how the nodes have been built)
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *hssec = &subsectors[hitssec[i]];

			if (hssec->sector == s)	continue;

			if (hssec->sector->specFlags & 0x4 && s->floorpic != skyflatnum)
			{
				if (hssec->sector->floorheight < s->floorheight || hssec->sector->specFlags & SF_MOVFLOOR)
					Setr3dssfs(hssec, s - sectors, 51004);
			}
			if (hssec->sector->specFlags & 0x8 && s->ceilingpic != skyflatnum)
			{
				if (hssec->sector->ceilingheight > s->ceilingheight || hssec->sector->specFlags & SF_MOVCEIL)
					Setr3dsscs(hssec, s - sectors, 51004);
			}
		}
	}
	else
	{
		for (i = 0; i < hitssecptr; i++)
		{
			if (flv)
			{
				Setr3dssfs(&subsectors[hitssec[i]], os - sectors, 51005);
				Setr3dsslfsec(&subsectors[hitssec[i]], s - sectors, 51005);
			}
		}
	}

	subfn_marklines(NULL, linemark);
}

static int eliminateSsecsInside(sector_t *s, sortbox_t *sbp, byte *ssecptr, byte *linemark)
{
	RDBOOL	valid = false;
	RDBOOL	remove = false;
	int		c, i;

	for (c = i = 0; i < sbp->listcnt; i++)
	{
		linearray[i] = sbp->llist[i];
		if (!lines[sbp->llist[i]].backsector) c++;
	}

	if (c == sbp->listcnt)
	{
		hitssecptr = 0;
		valid = true;
	}
	else
	if (   LS_MarkSubsecsInAreaGeneral(s, NULL, sbp->listcnt, NULL, sbp->box, -1, MSIA_REVERSED)
		|| LS_MarkSubsecsInAreaGeneral(s, NULL, sbp->listcnt, NULL, sbp->box, -1, MSIA_NONE))
	{
		chkMissing(s, sbp, linemark);
		valid = true;
	}

	for (i = 0; i < hitssecptr; i++)
	{
		ssecptr[hitssec[i]] = true;	// true means reject
	}

	for (i = 0; i < sbp->listcnt; i++)
	{
		hitline[sbp->llist[i]] = 8;
	}

	return valid;
}

static RDBOOL subfn_loopLines(sector_t *s, byte *linemark, int *lnlist)
{
	sortbox_t	sbx[MAXSORTBOXES];
	sortbox_t	*sbp;
	sector_t	*os;
	line_t		*ld;
	byte		*bSSF;
	__int64		area;
	int			lfsec = -1;
	int			c, d, i, k;
	int			ssecptr;
	int			numloops;
	int			rc;
	int			diff;
	byte		txn;

	// the following must be initialised
	RDBOOL		lfs;
	RDBOOL		found;

	int			mainindex;
	int			fhv , flv;

	// ------------------------------

	memset(sbx, 0, sizeof(sortbox_t) * MAXSORTBOXES);

	if ( !(numloops = boxSort(s, sbx, lnlist)) ) return false;

	bSSF = (byte*)malloc(numsubsectors * sizeof(byte));

restart:;

	lfs = false;
	found = false;
	fhv = -1, flv = -1;
	mainindex = -1;

	for (area = 0, i = 0; i < numloops; i++)
	{
		if (!sbx[i].done && sbx[i].area > area) { area = sbx[i].area; mainindex = i; }
	}

	if (mainindex < 0) goto abort;

	sbp = &sbx[mainindex];
	sbp->done = true;
	rc = sbp->listcnt;

	memcpy(lnlist, sbp->llist, rc * sizeof(int));

	// check the loop for missing textures
	for (c = d = i = 0; i < rc; i++)
	{
		ld = &lines[lnlist[i]];
		if (!ld->backsector) continue;

		os = ld->frontsector == s ? ld->backsector : ld->frontsector;
		txn = origTEX[ld - lines];

		diff = os->floorheight - s->floorheight;

		if ((!diff || !(txn & 0x11)) && os->floorpic != skyflatnum)
		{
			if (!diff || diff > 8*FRACUNIT)
			{
				c++;

				if (diff && flv != -2)	// bug fix v220-33 flv test added
				{
					if (flv == -1) flv = os - sectors;
					else if (os->floorheight != sectors[flv].floorheight) flv = -2;
				}
			}
		}

		if (!(txn & 0x44) && os->ceilingheight < s->ceilingheight)
		{
			d++;

			if (fhv != -2 && os->ceilingpic != skyflatnum)
			{
				if (fhv == -1) fhv = os - sectors;
				else if (os->ceilingheight != sectors[fhv].ceilingheight) fhv = -2;
			}
		}
		else
		if (!(txn & 0x44) && os->ceilingheight > s->ceilingheight)
		{
			d++;

			if (fhv != -2 && s->ceilingpic != skyflatnum)
			{
				if (fhv == -1) fhv = s - sectors;
				else if (s->ceilingheight != sectors[fhv].ceilingheight) fhv = -2;
			}
		}
	}

	if (c != rc) flv = -1;
	if (d != rc) { if (fhv != -1) goto abort; else fhv = -1; }

	for (i = 0; i < rc; i++) hitline[lnlist[i]] = 4;

	if (!LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, sbp->box, -1, MSIA_NONE))
		goto abort;

	{
		// Checks need to be made
		// 1) are all lines inside same sided?
		subsector_t	*ss;
		seg_t		*sg;
		RDBOOL		valid = false;
		RDBOOL		inhibitskyfix;
		int			lncount = 0;
		int			*sstype = NULL;
		int			sscnt = 0;
		int			licnt = 0;
		int			skycount = 0;

		memset(bSSF, 0, numsubsectors);

		CLR_hitline();

		for (i = 0; i < sbp->listcnt; i++)
		{
			hitline[sbp->llist[i]] = 8;
		}

		if (numloops > 1)
		{
			byte	*ssecptr = (byte*)calloc(numsubsectors, 1);
			int		*ssecmem = (int*)malloc(hitssecptr * sizeof(int));
			int		ptrmem = hitssecptr;

			memcpy(ssecmem, hitssec, ptrmem * sizeof(int));

			for (i = 0; i < numloops; i++)
			{
				if (!sbx[i].done && M_CompareBox(sbx[i].box, sbx[mainindex].box))
				{
					sbx[i].done = true;

					if (!eliminateSsecsInside(s, &sbx[i], ssecptr, linemark))
					{
						char msg[128];

						sprintf(msg,
							"LOOP INDEX ERROR S%i L%i : Base sec line %i\n"
							"Please report this by sending us the Risen3D.out file",
							s - sectors, sbx[i].llist[0], sbx[mainindex].llist[0]);

						Sys_MessageBox(msg, true);
					}
				}
			}

			for (hitssecptr = i = 0; i < ptrmem; i++)
			{
				if (!ssecptr[ssecmem[i]]) hitssec[hitssecptr++] = ssecmem[i];
			}

			free(ssecptr);
			free(ssecmem);
		}

		if (!hitssecptr) goto restart;

		validcount++;
		sstype = (int*)malloc(hitssecptr * sizeof(int));

		for (licnt = i = 0; i < hitssecptr; i++)
		{
			RDBOOL	uselfsec = false;

			ss = &subsectors[hitssec[i]];
			sg = &segs[ss->firstline];

			if (ss->sector->ceilingpic == skyflatnum
				&& (ss->sector == s
				|| (ss->sector->specFlags & 0x20008) == 0x20008) )
			{
				skycount++;
			}

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (ld->frontsector == ld->backsector)
					{
						valid = true;
						continue;
					}

					if (hitline[ld - lines]) continue;

					if (origTEX[ld - lines] & 0x55)
					{
						licnt++;
					}
					// Check if lfsec should be used for this ssec
					else if (!uselfsec
						&& ld->backsector
						&& origTEX[ld - lines] & 0x22
						&& ld->frontsector->floorheight > s->floorheight
						// check not mov sectors
						&& !(ld->backsector->specFlags & SF_MOVFLOOR)
						&& !(ld->frontsector->specFlags & SF_MOVFLOOR)
						&& ld->frontsector->floorheight == ld->backsector->floorheight
						&& (ld->frontsector->ceilingheight == ld->backsector->ceilingheight
						|| (ld->frontsector->ceilingpic == skyflatnum
						&&  ld->backsector->ceilingpic == skyflatnum)) )
					{
						uselfsec = true;	// indicate this ssec needs lfsec set
					}
				}
			}

			if (uselfsec) sstype[sscnt++] = hitssec[i];
		}

		if (valid && licnt && licnt < 3) valid = false;
		if (valid && lncount && lncount < 3) valid = false;

		if (valid && lncount)
		{
			// do nothing
		}
		// check lines forming primary loop
		else if (valid && !licnt && !lncount && fhv == -1 && flv == -1)
		{
			sector_t	*os;
			byte	 txn;

			// Find if a different ceiling/floor needs to be adopted
			for (i = 0; i < s->linecount; i++)
			{
				ld = s->lines[i];
				if (hitline[ld - lines] != 4) continue;
				if (!ld->backsector) continue;

				os = ld->frontsector == s ? ld->backsector : ld->frontsector;
				txn = origTEX[ld - lines];

				if (!(txn & 0x11) && os->floorheight < s->floorheight)
				{
					if (flv != -2)
					{
						if (flv == -1) flv = s - sectors;
						else if (s->floorheight != sectors[flv].floorheight) flv = -2;
					}
				}

				if (!(txn & 0x44) && os->ceilingheight < s->ceilingheight)
				{
					if (fhv != -2)
					{
						if (fhv == -1) fhv = os - sectors;
						else if (os->ceilingheight != sectors[fhv].ceilingheight) fhv = -2;
					}
				}
			}
		}

		if (valid)
		{
			int	secidx = s - sectors;

			found = true;					

			// Look for lines belonging to other sectors in the array
			// Rare but an instance can be seen in PAR E1M6 (S804 in S818).
			if (licnt)
			{
				int	cnt;

				validcount++;

				for (cnt = i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					sg = &segs[ss->firstline];

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						if ( (ld = sg->linedef) && ld->validcount != validcount)
						{
							ld->validcount = validcount;

							if (ld->frontsector == ld->backsector) continue;
							if (ld->frontsector == s) continue;
							if (!ld->backsector) continue;
							if (ld->backsector == s) continue;

							if (origTEX[ld - lines] & 0x55) linearray[cnt++] = ld - lines;
						}
					}
				}

				if (cnt > 2) checkOrphanSectors(s, bSSF, cnt);
			}

			subfn_marklines(s, linemark);
			inhibitskyfix = skycount == hitssecptr;

			for (i = 0; i < hitssecptr; i++)
			{
				ssecptr = hitssec[i];
				ss = &subsectors[ssecptr];

				if (inhibitskyfix) ss->ssflags |= SSF_NOSKYFIX2;

				if (fhv >= 0 || flv >= 0)
				{
					for (valid = c = 0; c < sscnt; c++)
					{
						if (sstype[c] == hitssec[i]) { valid = true; break; }
					}

					if ( !(bSSF[ssecptr] & 2) )
					{
						if (fhv >= 0)
						{
							if (sectors[fhv].ceilingheight >= s->ceilingheight)
							{
								Setr3dsscs(ss, s - sectors, 51006);
							}
							else if (!lncount)
							{
								Setr3dsslcsec(ss, fhv, 51006);
							}
							else if (ss->sector - sectors != fhv
								&& sectors[fhv].ceilingheight < ss->sector->ceilingheight)
							{
								Setr3dsscs(ss, fhv, 51007);
							}
						}
						else if (ss->sector != s)
						{
							Setr3dsscs(ss, s - sectors, 51008);
						}
					}

					if ( !(bSSF[ssecptr] & 1) )
					{
						if (flv >= 0)
						{
							if (ss->sector - sectors != flv)
							{
								if (lfs)
								{
									Setr3dsslfsec(ss, flv, 51009);
									Setr3dssfs(ss, s - sectors, 51009);
								}
								else
									Setr3dssfs(ss, flv, 51010);
							}
							else if (s->floorheight < sectors[flv].floorheight)
							{
								Setr3dsslfsec(ss, s - sectors, 51010);
							}

							ss->ssflags |= SSF_NOCHANGE;
						}
						else if (ss->sector != s
							&& !(ss->sector->heightsec >= 0
							&&   sectors[ss->sector->heightsec].floorheight > s->floorheight))
						{
							Setr3dssfs(ss, s - sectors, 51011);
						}
					}

				}
				else if (ss->sector - sectors != secidx)
				{
					for (valid = c = 0; c < sscnt; c++)
					{
						if (sstype[c] == hitssec[i]) { valid = true; break; }
					}

					if (ss->sector != s && !(bSSF[ssecptr] & 2))
					{
						// v220-33 add test
						if (   s->ceilingheight == ss->sector->ceilingheight
							&& s->ceilingpic == ss->sector->ceilingpic
							&& !((s->heightsec >= 0)^(ss->sector->heightsec >= 0)) )
						{
							// do nothing
						}
						else
						{
							Setr3dsscs(ss, s - sectors, 51011);
						}
						setSSECflag(ss, SSF_NOCHANGE, 51012);
					}

					if ( !(bSSF[ssecptr] & 1) )
					{
						if (valid && lfsec < 0)
						{
							Setr3dsslfsec(ss, secidx, 51013);
						}
						// test for skyfloor
						else if (ss->sector != s
							&& !(ss->ssflags & SSF_CONTAINER)
							&& !(ss->sector->floorpic == skyflatnum
							&& ss->sector->floorheight > sectors[secidx].floorheight) )
						{
							if (   (s->specFlags & 0xFFFFC) == 0x18
								&& (ss->sector->specFlags & 0xFFFFC) == 0x81C
								&& ss->sector->floorheight == s->floorheight
								&& ss->sector->ceilingheight == s->ceilingheight)
							{
								ss->ssflags &= ~SSF_BASESSECFIX;
								ChangeSSECsec(ss, &sectors[secidx], 51013);
							}
							else
							{
								// v220-33 set ceilsec if containing sector has no textures
								if ((ss->sector->specFlags & 0xC) == 0xC)
									SetOverride(ss, secidx, 51014);
								else
									Setr3dssfs(ss, secidx, 51014);
							}
						}

						ss->ssflags |= SSF_NOCHANGE;
					}
				}

				if (lfsec >= 0 && !(bSSF[ssecptr] & 1))
				{
					Setr3dsslfsec(ss, lfsec, 51015);
				}
			}

			Setsecdone(s, 51016);
		}

		free(sstype);
	}

	for (i = 0; i < numloops; i++)
	{
		if (!sbx[i].done) goto restart;
	}

abort:;

	free(bSSF);
	freeSortbox(sbx, numloops);

	return found;
}

static RDBOOL isValidSsec(subsector_t *ss)
{
	line_t	*ld;
	seg_t	*sg = &segs[ss->firstline];
	int		c, i;

	validcount++;

	for (c = i = 0; i < ss->linecount; i++, sg++)
	{
		if (  (ld = sg->linedef)
			&& ld->validcount != validcount)
		{
			if (ld->specType >= R3D_40) return false;
			
			ld->validcount = validcount;
			
			if (   ld->frontsector == ld->backsector
				&& ld->frontsector == ss->sector
				&& Vtxs2(ld))
			{
				c++;
			}
			else
			{
				return true;
			}
		}
	}

	return c < 3;
}

static void checkLogicalContaining(sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*ch = NULL;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			lnum, box[4];
	int			c, i, k, m, n;

	for (i = 0, lnum = -1; i < sc->linecount; i++)
	{
		n = sc->lines[i] - lines;
		sl = &sslines[n];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->frontsector != sc)
				{
					if (!ch) { ch = sg->frontsector; if (ch->specFlags & SF_MOVFLOOR) return; }
					else if (ch != sg->frontsector) return;

					if (  (ld = sg->linedef)
						&& ld->validcount != validcount
						&& ld->frontsector != ld->backsector)
					{
						ld->validcount = validcount;
						lnum = ld - lines;
					}
				}
			}
		}
	}

	if (!ch || lnum < 0) return;

	CLR_hitline();

	for (c = i = 0; i < ch->linecount; i++)
	{
		ld = ch->lines[i];

		if (ld->frontsector != ld->backsector) { hitline[ld - lines] = 1; c++; }
	}

	if (c < 3) return;

	if (   subfn_FindClosedRing(hitline, ch, &lnum, box, true) > 0
		&& LS_MarkSubsecsInAreaGeneral(ch, hitline, 0, NULL, box, -1, MSIA_NONE)
		&& LS_OnlySSLinesInSSECarray(ch, hitline))
	{
		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			if (!(ss->ssflags & SSF_CONTAINER) && isValidSsec(ss))
			{
				ss->ssflags &= ~SSF_BASESSECFIX;
				// v220-33 test added
				ChangeSSECsec(ss, sc, 51024);
				ss->ssflags |= SSF_BASESSECFIX;
			}
		}

		for (i = 0; i < ch->linecount; i++)
		{
			ld = ch->lines[i];

			if (   hitline[ld - lines] == 4
				&& !origTEX[ld - lines]
				&& ld->backsector
				&& ld->frontsector->floorheight != ld->backsector->floorheight)
			{
				sl = &sslines[ld - lines];

				for (k = 0; k < sl->num; k++)
				{
					ss = &subsectors[sl->list[k]];

					if ( !(ss->ssflags & SSF_BASESSECFIX) )
					{
						sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (!sg->sidedef) continue;

							if (!sg->backsector || sg->sidedef->bottomtexture)
							{
								ss->floorsec = sg->frontsector - sectors;
								if (ss->sector != sc) ChangeSSECsec(ss, sc, 51025);
								ss->ssflags |= SSF_BASESSECFIX;
								break;
							}
						}
					}
				}
			}
		}
	}
}

// Used to check sectors with same sector sided lines partially enlosed in loops
// that would have been set in LL_LineLoops. Any remaining, outside of any
// coherent loop, are checked to see if a common lowfloor sector can be adopted.
// Only closed static sectors with with lotex are checked.
static void subfn_checkSSlineIneritance(byte *linemark)
{
	sector_t	*ch;
	line_t		*ld;
	int			c, i, k;

	SET_countsec(0);

	for (c = i = 0; i < numlines; i++)
	{
		if (   linemark[i]
			&& lxarray[i] & 0x4
			&& lines[i].frontsector == lines[i].backsector
			&& (lines[i].frontsector->specFlags & 0x30414) == 0x10)
		{
			subsector_t	*ss;
			ssline_t	*sl = &sslines[i];
			RDBOOL		valid = false;

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->floorsec >= 0 || ss->lfsec >= 0)
				{
					valid = true; break;
				}
			}

			if (valid)
			{
				for (k = 0; k < sl->num; k++)
				{
					ss = &subsectors[sl->list[k]];

					if (ss->floorsec < 0 && ss->lfsec < 0)
					{
						countsec[lines[i].frontsector - sectors]++; c++;
						break;
					}
				}
			}
		}
	}

	if (!c)	return;

	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i])
		{
			ch = &sectors[i];

			for (c = k = 0; k < ch->linecount; k++)
			{
				ld = ch->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   !linemark[ld - lines]
					&& !lineflags[ld - lines]
					&& ld->frontsector == ld->backsector)
				{
					linearray[c++] = ld - lines;
				}
			}

			if (c)
			{
				subsector_t	*ss;
				ssline_t	*sl;
				sector_t	*os;
				seg_t		*sg;
				int			m;

				os = NULL;
				hitssecptr = 0;
				ss_validcount++;

				for (k = 0; k < c; k++)
				{
					sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						hitssec[hitssecptr++] = sl->list[m];

						if (ss->sector != ch && ss->floorsec < 0)
						{
							if (!os) os = ss->sector;
							else if (os != ss->sector
								&& (os->floorheight != ss->sector->floorheight
								||  os->floorpic != ss->sector->floorpic
								||  os->ceilingheight != ss->sector->ceilingheight
								||  os->ceilingpic != ss->sector->ceilingpic))
							{
								goto next_sec;
							}
						}
					}
				}

				if (!os) continue;

				for (k = 0; k < hitssecptr; k++)
				{
					ss = &subsectors[hitssec[k]];
					sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef) )
							linemark[ld - lines] = 1;
					}

					SetOverride(ss, os - sectors, 51031);
				}

				Setsecdone(ch, 51032);
				Setsecdone(os, 51033);
			}
		}
next_sec:;
	}
}

void LL_LineLoops()
{
	subsector_t	*ss;
	ssline_t	*sl;
	sector_t	*ch;
	line_t		*ld, *li;
	seg_t		*sg;
	int			i, k, m;
	byte		*linemark = (byte*)calloc(numlines, sizeof(byte));
	byte		*tmp = (byte*)calloc(numlines, sizeof(byte));
	int			*lnlist = (int*)malloc(numlines * sizeof(int));
	int			*cntsecloc = (int*)malloc(numsectors * sizeof(int));
	int			cntseclocnum;
	int			cntsecnum = 0;
	RDBOOL		found = false;

	// First fix assignments for maps like hr2final map24 where a nodes builder
	// is unable to determine whether a ssec's logical sector should, in this
	// case, be S56 or S99. The fact that maps like this worked in the first
	// place was down to 50:50 chance.

	checkConflicting();

	// Second check for construction where a textureless same sector sided line
	// loop is contained within another unbroken loop of diff sec lines. In
	// this instance the original doom nodes builder usually sets the contained
	// loop as the logical sector.

	SET_countsec(0);

	validcount++;

	for (i = 0; i < numlines; i++)
	{
		if (   !lineflags[i]
			&& !origTEX[i]
			&& lines[i].lnflags2 == LNF2_OWNEQUALTOONE
			&& (ch = lines[i].frontsector) == lines[i].backsector	// should be
			&& (ch->specFlags & 0x3001C) == 0x1C
			&& Vtxs2(&lines[i])
			&& !lxarray[i])
		{
			countsec[ch - sectors]++;
			lines[i].validcount = validcount;	// needed for checkLogicalContaining
		}
	}

	// keep this simple. No shared sectors, no linecounts above 4. This
	// prevents having to establish that the lines form a loop.
	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i] > 2 && countsec[i] < 5 && countsec[i] == sectors[i].linecount)
		{
			checkLogicalContaining(&sectors[i]);
		}
	}

	// Third attempt to process all sectors containing same sided lines in a different
	// sector. The following routine depends on a map being properly designed (which
	// most are). Those that are not are thrown on the mercy of the old analysis code.

	for (cntseclocnum = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !lineflags[i]
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& ld->frontsector == ld->backsector
			&& (origTEX[i] & 0x22
			|| ld->frontsector->specFlags & SF_MOVFLOOR
			|| (vertexowners[ld->v1 - vertexes].num == 1
			&&  vertexowners[ld->v2 - vertexes].num == 1)) )
		{
			sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];
				sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (   sg->frontsector != sg->backsector
						&& sg->frontsector != ld->frontsector)
					{
						if ( !(sg->frontsector->floorpic == skyflatnum
							&& sg->frontsector->ceilingpic == skyflatnum))
						{
							addSecLocal(sg->frontsector - sectors, cntsecloc, &cntseclocnum);
							// don't bail out, more than one sec may be involved
						}
					}
				}
			}
		}
	}

	for (cntsecnum = i = 0; i < cntseclocnum; i++)
	{
		// adds sectors to countsec
		verifySector(cntsecloc[i], &cntsecnum);
	}

	for (i = 0; i < cntsecnum; i++)
	{
		if (   sectors[countsec[i]].specFlags & 0x400
			&& subfn_loopLines(&sectors[countsec[i]], linemark, lnlist))
		{
			found = true;
		}
	}

	for (i = 0; i < cntsecnum; i++)
	{
		if ( !(sectors[countsec[i]].specFlags & 0x400)
			&& subfn_loopLines(&sectors[countsec[i]], linemark, lnlist))
		{
			found = true;
		}
	}

	for (i = 0; i < cntsecnum; i++)
	{
		// check for instance of failing to deal with a sector where
		// a missing lotex line borders a sector with 1's lines.
		if (!secdone[countsec[i]] && sectors[countsec[i]].specFlags & 0x400)
		{
			sector_t	*sc = &sectors[countsec[i]];
			RDBOOL		hasmissing = false;
			int			ln;

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				ln = ld - lines;

				if (   !lineflags[ln]
					&& !(ld->spflags & SPF_CONTROLLINE)
					&& ld->backsector
					&& ld->frontsector != ld->backsector
					&& !(origTEX[ld - lines] & 0x11)
					&& ld->frontsector->floorheight != ld->backsector->floorheight)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (ch->floorheight < sc->floorheight) { hasmissing = true; break; }
				}
			}

			if (hasmissing)
			{
				int rc, box[4];

				CLR_hitline();
				memset(tmp, 0, numlines);

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];
					if (ld->spflags & SPF_CONTROLLINE) continue;
					if (ld->frontsector != ld->backsector) hitline[ld - lines] = 1;
				}

				while (1)
				{
					int lnum = -1;

					for (k = 0; k < sc->linecount; k++)
					{
						ld = sc->lines[k];
						if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
					}

					if (lnum < 0) break;

					if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
					{
						if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
						{
							RDBOOL valid = true, high = true;

							// check lines inside
							for (cntseclocnum = k = 0; k < hitssecptr; k++)
							{
								ss = &subsectors[hitssec[k]];
								sg = &segs[ss->firstline];

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (!(ld = sg->linedef)) continue;
									if (hitline[ld - lines] == 4)
									{
										// a missing tex line may still need to be processed elsewhere
										if (!mtxline(ld)) tmp[ld - lines] = 1;
										continue;
									}
									tmp[ld - lines] = 2;
									if (!ld->backsector) continue;
									if (ld->frontsector == ld->backsector) continue;
									if (ld->frontsector == sc && !(ld->backsector->specFlags & 0x4)) { valid = false; break; }
									if (ld->backsector == sc && !(ld->frontsector->specFlags & 0x4)) { valid = false; break; }
									// v220-33 next 2 tests added
									if (ld->frontsector == sc && ld->backsector->ceilingheight > sc->ceilingheight) { valid = false; break; }
									if (ld->backsector == sc && ld->frontsector->ceilingheight > sc->ceilingheight) { valid = false; break; }

									// Look for a missing hi tex
									if (high && ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
									{
										if (ld->backsector == sc && !(ld->frontsector->specFlags & 0x8))
										{
											addSecLocal(ld->frontsector - sectors, cntsecloc, &cntseclocnum);
										}
										else
										if (ld->frontsector == sc && !(ld->backsector->specFlags & 0x8))
										{
											addSecLocal(ld->backsector - sectors, cntsecloc, &cntseclocnum);
										}
									}
								}
							}

							if (valid)
							{
								for (k = 0; k < cntseclocnum; k++)
								{
									if (high)
									{
										high = false; ch = &sectors[cntsecloc[k]];

										for (m = 0; m < ch->linecount; m++)
										{
											li = ch->lines[m];
											if (tmp[ld - lines] != 2) continue;
											if (!(origTEX[li - lines] & 0x44)) { high = true; break; }	// One line's enough
										}
									}
								}

								for (k = 0; k < hitssecptr; k++)
								{
									ss = &subsectors[hitssec[k]];
									if (ss->sector == sc) continue;
									if (high) SetOverride(ss, sc - sectors, 0); else Setr3dssfs(ss, sc - sectors, 0);
								}

								for (k = 0; k < numlines; k++)
								{
									if (tmp[k] == 2 && !mtxline(&lines[k])) tmp[k] = 1;
								}

								found = true;
								for (k = 0; k < cntseclocnum; k++) Setsecdone(&sectors[cntsecloc[k]], 51017);
								Setsecdone(sc, 51018);
							}
						}
						else
						{
							break;
						}

						for (k = 0; k < sc->linecount; k++)
						{
							ld = sc->lines[k];

							if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
						}
					}
					else
					{
						break;
					}
				}
			}
		}
	}

	if (found)
	{
		byte	txn;
		int		c;

		subfn_checkSSlineIneritance(linemark);

		for (i = 0; i < numlines; i++)
		{
			if (linemark[i] || tmp[i] == 1)
			{
				ld = &lines[i];

				if (!tmp[i])
				{
					if ( (c = mtxline(ld)) == 1) continue;
					if (c)
					{
						if (   !secdone[ld->frontsector - sectors]
							|| !secdone[ld->backsector - sectors])
							continue;
					}
				}

				txn = origTEX[ld - lines];

				if (!txn)
					Settype(ld, R3D_3F, 51019);
				else if (ld->frontsector == ld->backsector)
					Settype(ld, R3D_03, 51020);
				else if (ld->backsector
					&& txn & 0x22
					&& !(txn & 0x55)
					&& ld->frontsector->floorheight == ld->backsector->floorheight)
					Settype(ld, R3D_03, 51021);
				else if (mtxline(ld))
					Settype(ld, R3D_00, 51022);
				else
					Settype(ld, R3D_36, 51023);

				SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 51024);
			}
		}
	}

	free(cntsecloc);
	free(lnlist);
	free(tmp);
	free(linemark);
}

// --------------------------------------------------------------------------

static int getDonutSectors(line_t *ld, byte *scmark)
{
	sector_t	*s1 = NULL, *s2 = NULL, *s3 = NULL;
	int			i, f;
	int			count = 0, secnum = -1;

	while ((secnum = P_FindSectorFromTag(ld->tag, secnum)) >= 0)
	{
		if (scmark[secnum]) return 0;

		s1 = &sectors[secnum];

		for (f = i = 0; i < count; i++)
		{
			if (countsec[i] == secnum) {f = 1; break; }
		}

		if (f) continue;

		countsec[count++] = secnum;
		
		if (!(s2 = getNextSector(s1->lines[0], s1))) continue;
		
		if (scmark[s2 - sectors]) return 0;

		for (f = i = 0; i < count; i++)
		{
			if (countsec[i] == s2 - sectors) {f = 1; break; }
		}

		if (!f) countsec[count++] = s2 - sectors;

		for (i = 0; i < s2->linecount; i++)
		{
			if ( !(s2->lines[i]->ml_flags & ML_TWOSIDED)
				|| s2->lines[i]->backsector == s1)
			{
				continue;
			}
			
			s3 = s2->lines[i]->backsector;      // s3 is model sector for changes
			if (scmark[s3 - sectors]) return 0;

			for (f = i = 0; i < count; i++)
			{
				if (countsec[i] == s3 - sectors) {f = 1; break; }
			}

			if (!f) countsec[count++] = s3 - sectors;

			break;
		}
	}

	if (count > 1)
	{
		int *temp = (int*)M_Malloc(count*sizeof(int), 0);
		int k, c = count;

		memcpy(temp, countsec, count*sizeof(int));

		for (count = i = 0; i < c; i++)
		{
			sector_t *s = &sectors[temp[i]];
			
			for (k = 0; k < s->linecount; k++)
			{
				if (!(s->lines[k]->spflags & SPF_CONTROLLINE))
				{
					countsec[count++] = temp[i];
					break;
				}
			}
		}
		M_Free(temp);
	}
	return count;
}

static void subfn_donutloopLines(sector_t *s, int count, int pos, byte *lnmark, byte *scmark, int *lnlist, int *ret, int *exitloop)
{
	subsector_t	*ss;
	sortbox_t	sbx[MAXSORTBOXES];
	sortbox_t	*sbp;
	__int64		area;
	int			mainindex;
	int			numloops;
	int			bc, i;
	int			rc, fht, lfsec;

	// ------------------------------

	memset(sbx, 0, sizeof(sortbox_t) * MAXSORTBOXES);

	if ( !(numloops = boxSort(s, sbx, lnlist)) ) return;

	for (bc = 0; bc < numloops; bc++)
	{
		mainindex = -1;

		for (area = 0, i = 0; i < numloops; i++)
		{
			if (!sbx[i].done && sbx[i].area > area) { area = sbx[i].area; mainindex = i; }
		}

		if (mainindex < 0) goto tidy_up;

		sbp = &sbx[mainindex];
		sbp->done = true;
		rc = sbp->listcnt;

		memcpy(lnlist, sbp->llist, rc * sizeof(int));

		CLR_hitline();

		for (i = 0; i < rc; i++) hitline[lnlist[i]] = 4;

		if (!LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, sbp->box, -1, MSIA_NONE))
		{
			continue;
		}

		if (!bc && !pos) *exitloop = true;

		// scan the ssecs for the highest floor
		fht = s->floorheight; lfsec = s - sectors;

		if (*exitloop == false)
		{
			for (i = 0; i < hitssecptr; i++)
			{
				ss = &subsectors[hitssec[i]];

				if (findSec(ss->sector - sectors, count))
				{
					if (ss->sector->floorheight > fht)
					{
						lfsec = ss->sector - sectors;
						fht = ss->sector->floorheight;
					}
				}
			}
		}

		if (lfsec >= 0)
		{
			for (i = 0; i < hitssecptr; i++)
			{
				ss = &subsectors[hitssec[i]];
				scmark[ss->sector - sectors] = true;

				if (ss->lfsec < 0 || sectors[ss->lfsec].floorheight < fht)
				{
					Setr3dsslfsec(ss, lfsec, 51025);

					if ( (ss->sector->specFlags & 0x1000C) == 0xC)
					{
						SetOverride(ss, s - sectors, 51025);
					}
				}
			}
			
			subfn_marklines(NULL, lnmark);

			*ret = true;
		}

		scmark[s - sectors] = true;

		if (*exitloop) break;
	}

tidy_up:;
	freeSortbox(sbx, numloops);
}

// v220-22  Trick types often have same sector sided lines and therefore share
// much of the same code as LL_LineLoops. They must, however, be checked first.
void LS_CheckDonut()
{
	line_t	*ld;
	int		i, k, count;
	int		*lnlist = NULL;
	byte	*lnmark = NULL;
	byte	*scmark = NULL;
	byte	txn;


	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].special
			&& lines[i].tag
			&& (lines[i].special ==   9
			||  lines[i].special == 146
			||  lines[i].special == 155
			||  lines[i].special == 191))
		{
			// test added v220-23
			int	c = 0, d = 0, sn = -1;

			ld = &lines[i];

			while ((sn = P_FindSectorFromTag(ld->tag, sn)) >= 0)
			{
				c++;
				if (SecCtrl(sn)) d++;
			}

			if ( c < 2 || c != d) continue;

			if (!scmark)
			{
				scmark = (byte*)calloc(numsectors, sizeof(byte));

				for (k = 0; k < numsectors; k++)
				{
					if (secdone[k]) scmark[k] = true;
				}
			}

			if ( (count = getDonutSectors(&lines[i], scmark)) )
			{
				int		ret = false;
				int		exitloop = false;

				for (k = 0; k < count; k++)
				{
					sector_t *s = &sectors[countsec[k]];

					if ((s->specFlags & 0x810014) == 0x10010)
					{
						if (!lnlist) lnlist = (int*)malloc(numlines * sizeof(int));
						if (!lnmark) lnmark = (byte*)calloc(numlines, sizeof(byte));

						subfn_donutloopLines(s, count, k, lnmark, scmark, lnlist, &ret, &exitloop);
					}
				}

				if (!ret) goto tidy_up;

				if (lnmark)
				{
					for (k = 0; k < numlines; k++)
					{
						if (lnmark[k])
						{
							ld = &lines[k];
							txn = origTEX[ld - lines];

							if (!txn)
								Settype(ld, R3D_3F, 51026);
							else if (ld->frontsector == ld->backsector)
								Settype(ld, R3D_03, 51027);
							else if (ld->backsector
								&& txn & 0x22
								&& !(txn & 0x55)
								&& ld->frontsector->floorheight == ld->backsector->floorheight)
								Settype(ld, R3D_03, 51028);
							else if (mtxline(ld))
								Settype(ld, R3D_00, 51029);
							else
								Settype(ld, R3D_36, 51030);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 51031);
						}
					}

					if (scmark)
					{
						for (k = 0; k < numsectors; k++)
						{
							if (scmark[k] && !secdone[k]) Setsecdone(&sectors[k], 51032);
						}
					}
				}
			}
tidy_up:;
			if (lnmark) { free(lnmark); lnmark = NULL; }
			if (scmark) { free(scmark); scmark = NULL; }
		}
	}

	if (lnmark) free(lnmark);
	if (scmark) free(scmark);
	if (lnlist) free(lnlist);
}
