//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_nodesfixspec.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: looks for subsector assignment errors from nodes
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>
#include "r3d_MAINspec.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static void subfn_checkShortSegAssignment(subsector_t *ss, seg_t *ssg)
{
	sector_t	*s = NULL;
	seg_t		*sg = &segs[ss->firstline];
	int			i, count = 0;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->linedef != NULL
			&& sg->frontsector != sg->backsector)
		{
			if (!s) s = sg->frontsector;
			if (s != sg->frontsector
				|| s == ss->sector)
			{
				return;
			}
		}
	}

	if (!s) return;

	ChangeSSECsec(ss, s, 0);
}

void LS_ChkShortSegAssignment()	// rev99n
{
	subsector_t *ss;
	int			i, k;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef
				&& sg->length < 1
				&& sg->frontsector == sg->backsector
				&& ss->sector == sg->frontsector
				&& M_AccurateLength(sg->linedef) > 4)
			{
				subfn_checkShortSegAssignment(ss, sg);
				break;
			}
		}
	}
}

//=================================================================

static void subfn_setSSEC(sector_t *sec)
{
	subsector_t *ss;
	int			i, k, outer[4];

	CLR_hitline();

	M_ClearBox(outer);

	for (i = 0; i < sec->linecount; i++)
	{
		M_AddLineToBox (outer, sec->lines[i]);
	}

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].bbox[BOXLEFT ] > outer[BOXLEFT ] &&
			lines[i].bbox[BOXRIGHT] < outer[BOXRIGHT] &&
			lines[i].bbox[BOXBASE ] > outer[BOXBASE ] &&
			lines[i].bbox[BOXTOP  ] < outer[BOXTOP  ] &&
			lines[i].specType < R3D_40)
		{
			hitline[i] = 1;
		}
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->sector != sec &&
			ChkSSECunset(ss))
		{
			seg_t *sg = &segs[ss->firstline];
			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef != NULL && hitline[sg->linedef - lines])
				{
					ChangeSSECsec(ss, sec, 37001);
					break;
				}
			}
		}
	}
}

// Attempt to fix subsecs in sec 91 Vile map31
void LS_SNF1()
{
	subsector_t *ss;
	int			i;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!(ss->sector->specFlags & SF_MOVSECTOR) &&
			ChkSSECunset(ss))
		{
			seg_t	 *sg = &segs[ss->firstline];
			sector_t *s1 = NULL, *s2 = NULL, *s3 = NULL, *s4 = NULL;
			int		 k, count = 0;

			CLR_hitline();

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef)
				{
					if (hitline[sg->linedef - lines] ||
						sg->linedef->specType >= R3D_40)
						continue;

					hitline[sg->linedef - lines] = 1;

					if (++count > 2) { s1 = NULL; break; }

					if (!sg->backsector ||
						sg->frontsector == sg->backsector)
					{ s1 = NULL; break; }

					if (sg->frontsector->specFlags & SF_MOVFLOOR)
					{
						s1 = sg->frontsector;
						s2 = sg->backsector;
					}
					else
					{
						s3 = sg->frontsector;
						s4 = sg->backsector;
					}
				}
			}

			if (!s1 || !s2 || !s3 || !s4 ||
				s4 == s1 || s4 == s2 || s4 == s3)
				continue;

			if (MSFC(s1 - sectors, 0x00018390) &&
				s1->linecount == 4 &&
				s2 == ss->sector &&
				s2 == s3)
			{
				subfn_setSSEC(s1);
			}
		}
	}
}

//=====================================================================

static int CheckIsRing(int lnum, int *outer)
{
	line_t	 *ld = &lines[lnum];
	sector_t *s = ld->frontsector;
	vertex_t *v = ld->v1;
	int		 i, isclosed = false, repeat = true;

	M_ClearBox(outer);
	M_AddLineToBox (outer, ld);

	hitline[lnum] = 2;

	while (repeat)
	{
		repeat = false;
		for (i = 0; i < s->linecount; i++)
		{
			int l = s->lines[i] - lines;

			if (!Vtxs2(s->lines[i]) ||
				s->lines[i]->frontsector !=
				s->lines[i]->backsector)
				continue;

			if (hitline[l] == 1)
			{
				if (s->lines[i]->v1 == v)
				{
					v = s->lines[i]->v2;
					hitline[l] = 2;
					M_AddLineToBox (outer, s->lines[i]);
					repeat = true;
				}
				else if (s->lines[i]->v2 == v)
				{
					v = s->lines[i]->v1;
					hitline[l] = 2;
					M_AddLineToBox (outer, s->lines[i]);
					repeat = true;
				}
			}
			else if (l == lnum && v == ld->v2)
			{
				isclosed = true;
				break;
			}
		}
	}
	if (!isclosed)
	{
		for (i = 0; i < s->linecount; i++)
			if (hitline[s->lines[i] - lines] == 2)
				hitline[s->lines[i] - lines] = 1;
	}
	return isclosed;
}

static int ConferToInnerLines(int specSec, int *outer)
{
	line_t	*ld;
	int		i, count = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (LSS(ld) &&
			IsLiquidFloorFlat(ld->frontsector) &&
			(ld->frontsector->specFlags & 0x000F000C) == 0x0000000C &&
			!origTEX[i] &&
			ld->specType < R3D_40 &&
			Vtxs2(ld) &&
			ld->bbox[BOXLEFT ] > outer[BOXLEFT ] &&
			ld->bbox[BOXRIGHT] < outer[BOXRIGHT] &&
			ld->bbox[BOXBASE ] > outer[BOXBASE ] &&
			ld->bbox[BOXTOP  ] < outer[BOXTOP  ])
		{
			Setspec(ld, specSec, 37001);
			Settype(ld, R3D_3D, 37001);
			count++;
		}
	}
	return count;
}
// This does not use the line owned subsecs due to vagaries of the nodes builder.
void LS_NestedSectorRings()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k, count = 0;
	int		 outer[4];

	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if ((sec->specFlags & 0x0000000C) != 0x0000000C &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			!IsLiquidFloorFlat(sec) &&
			sec->heightsec < 0)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (LSS(ld) &&
					Vtxs2(ld) &&
					ld->specType < R3D_40 &&
					!origTEX[ld - lines] &&
					subfn_FindLineCrossing(ld, LC_NORMAL) == NULL)
				{
					hitline[ld - lines] = 1;
					count++;
				}
			}
		}
	}

	if (!count) return;

	for (i = 0; i < numlines; i++)
	{
		if (hitline[i] = 1 && CheckIsRing(i, outer))
		{
			ConferToInnerLines(lines[i].frontsector - sectors, outer);

			sec = lines[i].frontsector;

			for (k = 0; k < sec->linecount; k++)
				if (hitline[sec->lines[k] - lines] == 2)
					hitline[sec->lines[k] - lines] = 0;
		}
	}
}

void LS_Find4RingSectorsCrossing()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->linecount == 4 &&
			MSFC(i, 0x0000001C))
		{
			sector_t *xsec = NULL;
			int		 k, count = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (LSS(ld)
					&& Vtxs2(ld)
					&& !origTEX[ld - lines]
					&& !ChkLFGSecSet(ld)
					&& ld->specType < R3D_40)
				{
					line_t *lc = subfn_FindLineCrossing(ld, LC_NORMAL);

					if (lc != NULL
						&& LSS(lc)
						&& lc->frontsector != sec
						&& lc->frontsector->floorheight == sec->floorheight
						&& lc->frontsector->floorpic == sec->floorpic
						&& lc->frontsector->ceilingheight > sec->ceilingheight)
					{
						if (!xsec)
							xsec = lc->frontsector;
						else if (xsec != lc->frontsector)
						{
							xsec = NULL;
							break;
						}
					}
					hitline[ld - lines] = 1;
					count++;
				}
			}
			if (!xsec || count != 4) continue;

			for (k = 0; k < sec->linecount; k++)
				Setspec(sec->lines[k], xsec - sectors, 37002);
		}
	}
}

// Check unassigned fixed support lines
void LS_ChkSupportAssign()
{
	subsector_t *ss;
	line_t		*ld;
	int			i;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!(ss->sector->specFlags & SF_MOVSECTOR) &&
			IsLiquidFloorFlat(ss->sector) &&
			ChkSSECunset(ss))
		{
			sector_t	*chksec = NULL;
			seg_t		*sg = &segs[ss->firstline];
			int			k, valid = false;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ((ld = sg->linedef) != NULL)
				{
					if (ld->frontsector != ld->backsector ||
						ld->frontsector->specFlags & SF_MOVSECTOR ||
						!IsLiquidFloorFlat(ld->frontsector) ||
						ld->specType >= R3D_40)
					{
						valid = false;
						break;
					}

					if (!origTEX[ld - lines] &&
						ld->frontsector != ss->sector &&
						Vtxs2(ld))
					{
						valid = true;
						if (!chksec)
							chksec = ld->frontsector;
						
						if (chksec != ld->frontsector ||
							lnsec[ld - lines] >= 0 ||
							chksec->floorheight <
							ss->sector->floorheight)
						{
							valid = false;
							break;
						}
					}
					else if (origTEX[ld - lines] == 0x22 &&
						ld->frontsector == ss->sector)
					{
						// do nothing
					}
					else
					{
						valid = false;
						break;
					}
				}
			}

			if (!valid) continue;

			sg = &segs[ss->firstline];
			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ((ld = sg->linedef) != NULL &&
					ld->frontsector != ss->sector)
				{
					Setspec(ld, ss->sector - sectors, 37003);
				}
			}
		}
	}
}

// More problems (sec 2,3 Eternal map13)
void LS_ChkLinesInSkySec()
{
	subsector_t *ss;
	int			i, k;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (MSFC(ss->sector - sectors, 0x0000808C)
			&& SecCtrl(ss->sector - sectors)
			&& ss->ceilsec != skyflatnum
			&& ChkSSECunset(ss))
		{
			sector_t	*newsec = NULL, *fs, *bs;
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef != NULL &&
					sg->frontsector->ceilingpic == skyflatnum)
				{
					newsec = sg->frontsector;
					break;
				}
			}

			if (!newsec
				|| SecCtrl(newsec - sectors))
			{
				continue;
			}

			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef
					&& (bs = sg->linedef->backsector) != NULL
					&& (fs = sg->linedef->frontsector) != NULL
					&& fs != bs)
				{
					if (MSFC(fs - sectors, 0x0000808C)
						&& MSFC(bs - sectors, 0x0000808C)
						&& SecCtrl(fs - sectors)
						&& SecCtrl(bs - sectors)
						&& fs->floorheight == bs->floorheight
						&& fs->ceilingheight == bs->ceilingheight)
					{
						Setspec(sg->linedef, newsec - sectors, 37004);
					}
				}
			}
		}
	}
}
