//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_postcheck.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: Newer routines being developed to replace old routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "dd_types.h"

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_compat.h"
#include "r3d_edit.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static void subfn_PrintStats()
{
#ifdef _DEBUG
	int i, count;
#endif

	int val;

	if (ma_inf.sslinecheck && verbose)
		pr_OFinsert("  WARNING : Map may have single sided line error(s).\n"
					"           Line total %i lines tested %i\n", numlines, ma_inf.sslinecheck);

#ifdef _DEBUG
	if (ma_inf.nogroup ) pr_OFinsert("%5i subsector%s ungrouped.\n", ma_inf.nogroup, ma_inf.nogroup == 1 ? " is" : "s are");
	if (ma_inf.ovrcount) pr_OFinsert("%5i spurious OVR settings cleared.\n", ma_inf.ovrcount);

	count = sizeof(ma_inf.dbcount)/sizeof(int);

	for (i = 0; i < count; i++)
		if (ma_inf.dbcount[i]) pr_OFinsert("%5i subsec ptrs corrected  [type %i].\n", ma_inf.dbcount[i], i);

	if (ma_inf.sscount2) pr_OFinsert("%5i subsector ptrs set [pass 2].\n", ma_inf.sscount2);
	if (ma_inf.sscount3) pr_OFinsert("%5i subsector ptrs set [pass 3].\n", ma_inf.sscount3);
#endif

	if ((val = ma_inf.called_arbitrator) > 0 && !verbose)
		pr_OFinsert("  Times sector arbitrator called %i\n", val);

	if ((val = ma_inf.unrequiredtex) > 0)
		pr_OFinsert("%5i unused side tex def%s found.\n", val, plural(val));

	if (   ma_inf.seclinefixed
		|| ma_inf.misstexfixed
		|| ma_inf.yofferror
		|| ma_inf.emptysec
		|| ma_inf.taggederr
		|| ma_inf.invalidtex)
	{
		pr_OFinsert("Map errors fixed:\n");

		if ((val = ma_inf.seclinefixed) > 0)
			pr_OFinsert("%5i line%s in wrong sector.\n", val, plural(val)), ma_inf.maperror += val; 
		if ((val = ma_inf.yofferror) > 0)
			pr_OFinsert("%5i incorrect rowoffset%s.\n", val, plural(val)), ma_inf.maperror += val;
		if ((val = ma_inf.misstexfixed) > 0)
			pr_OFinsert("%5i missing texture%s.\n", val, plural(val)), ma_inf.maperror += val;
		if ((val = ma_inf.emptysec) > 0)
			pr_OFinsert("%5i empty sector%s noted.\n", val, plural(val)), ma_inf.maperror += val;
		if ((val = ma_inf.tagerrors) > 0)
			pr_OFinsert("%5i missing tag error%s.\n", val, plural(val)), ma_inf.maperror += val; 
		if ((val = ma_inf.taggederr) > 0)
			pr_OFinsert("%5i line%s found with unset tags.\n", val, plural(val)), ma_inf.maperror += val;
		if ((val = ma_inf.invalidtex) > 0)
			pr_OFinsert("%5i bad side tex def%s removed.\n", val, plural(val)), ma_inf.maperror += val;
	}

	if ((val = ma_inf.thingheight) > 0)
	{
		pr_OFinsert("Spawn errors fixed;\n");
		pr_OFinsert("%5i mobj's spawned height adjusted.\n", val);
	}
}

void LS_CheckLinesInWrongSectors()
{
	sector_t *sec;
	int i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i]) continue;

		// Check for invalid R3D_1B types
		if (   sec->linecount == 4
			&& (sec->specFlags & 0x00FFFFFF) == 0x00808198
			&& LS_ChkAllVtx2(sec, false, false))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!origTEX[ld - lines]
					&& ld->specType == R3D_1B)
				{
					Settype(sec->lines[k], R3D_00, 29001);
				}
			}
		}
		// Check for unclosed movef secs with no lower tex
		else if (sec->specFlags & SF_MOVFLOOR
			&& !SecCtrl(i)
			&& !sec->special
			&& sec->ceilingpic != skyflatnum
			&& sec->tag
			&& sec->specFlags & SF_NOLOTEX
			&& sec->specFlags & SF_SIDETEXTRUE
			&& !(sec->specFlags & (SF_MOVCEIL|SF_SECCLOSED|SF_NOHITEX)))
		{
			line_t *li = NULL;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED
					&& ld->frontsector != ld->backsector)
				{
					sector_t *s = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (s->floorheight == s->ceilingheight)
					{
						li = ld;
						break;
					}
				}
			}

			if (!li) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->special && ld->tag == sec->tag)
				{
					sector_t *s = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (s && s != sec
						&& !(s->specFlags & SF_MOVSECTOR)
						&& s->floorheight < sec->floorheight
						&& s->ceilingpic == sec->ceilingpic
						&& s->floorpic == sec->floorpic)
					{
						ssline_t *sl = &sslines[ld - lines];
						int		 m;

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector == sec)
								SetOverride(ss, s - sectors, 29002);
						}

						sl = &sslines[li - lines];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->floorsec < 0 || ss->ceilsec < 0)
								SetOverride(ss, s - sectors, 29003);
						}

						Setsecdone3(sec, 29004);
						break;
					}
				}
			}
		}
	}


	// Deal with floor offset defined by lines inside a line ring with
	// wrong sectors facing
	for(i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			IsLiquidFloorFlat(sec) &&
			sec->specFlags & SF_SECCLOSED &&
			sec->specFlags & SF_NOHITEX)
		{
			int c1 = 0, valid = false;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (!(ld->ml_flags & ML_TWOSIDED) ||
					ld->mtflags & ~R3D_MISLTEX2 ||
					ld->frontsector == ld->backsector ||
					vtxarray[ld->v1 - vertexes] != 2 ||
					vtxarray[ld->v2 - vertexes] != 2 ||
					ld->frontsector->ceilingheight !=
					ld->backsector->ceilingheight)
				{
					valid = false;
					break;
				}
				if (ld->spflags & SPF_NOTEX &&
					ld->mtflags & R3D_MISLTEX2 &&	// height check really
					IsLiquidFloorFlat(ld->frontsector) &&
					IsLiquidFloorFlat(ld->backsector))
				{
					valid = true;
					c1++;
				}
				else if (sides[ld->sidenum[0]].texFlags == T_LOW &&
					!sides[ld->sidenum[1]].texFlags)
					c1--;
				else
				{
					valid = false;
					break;
				}
			}
			if (!valid || c1) continue;

			for (k = 0; k < sec->linecount; k++)
				if (!sec->lines[k]->mtflags)
				{
					Settype(sec->lines[k], R3D_0D, 29005);
				}

		}
		// Check moving secs with missing tex lines
		else if (!secdone[i] &&
			sec->specFlags & SF_MOVSECTOR)
		{
			for (k = 0; k < sec->linecount; k++)
				if (sec->lines[k]->specType == R3D_1A &&
					sec->lines[k]->spflags & SPF_NOTEX)
					Settype(sec->lines[k], R3D_00, 29006);
		}
	}
}

// Deal with moving secs with lines having AASHITTY textures
void LS_CheckMovSecWithAASHITTYTexLines()
{
	subsector_t *ss;
	sector_t *sec;
	line_t *ld;
	int i, k, cs;

	for(i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags & SF_MOVSECTOR &&	//No secdone test - may or may not have been done
			!IsLiquidFloorFlat(sec) &&
			sec->specFlags & SF_SECCLOSED)
		{
			int c1 = 0, c2 = 0;
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ld->ml_flags & ML_TWOSIDED)
				{
					c1++;
					if (ld->specType == R3D_2D &&
						origTEX[ld - lines] == 0x70 &&
						ld->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2) &&
						!sides[ld->sidenum[0]].texFlags)
					{
						c2++;
					}
					else if (ld->specType == R3D_00 &&
						origTEX[ld - lines] == 0x50 &&
						ld->spflags & SPF_NOTEX)
					{
						continue;
					}
					else
					{
						c1 = 0;
						break;
					}
				}
				else
				{
					c1 = 0;
					break;
				}
			}

			if (c1 != 4 || c2 < 2) continue;

			if ( (cs = XS_GetBoundarySecnum(sec)) < 0) continue;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];
				if (!(ld->spflags & SPF_CONTROLLINE))
				{
					hitline[ld - lines] = 1;
					if (ld->specType == R3D_2D)
						Settype(ld, R3D_1D, 29011);
				}
			}
			for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
			{
				seg_t *seg = &segs[ss->firstline];
				int m;
				for (m = 0; m < ss->linecount; m++, seg++)
					if ((ld = seg->linedef) != NULL &&
						hitline[ld - lines])
					{
						SetOverride(ss, cs, 29012);
						break;
					}
			}
			Setsecdone(sec, 29013);
		}
	}
}

void LS_CheckOVF_MISLOTEXLINEReset()
{
	sector_t *sec;
	int i, k, bbox[4];
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			sec->specFlags & SF_MOVSECTOR &&
			!IsLiquidFloorFlat(sec))
		{
			int count1 = 0, count2 = 0, snum1 = -1, snum2 = -1;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->spflags & SPF_CONTROLLINE)
					continue;
				if (ld->backsector &&
					ld->frontsector != sec &&
					origTEX[ld - lines] == 0x01 &&
					ld->backsector->floorheight ==
					ld->frontsector->floorheight &&
					Vtxs2(ld))
				{
					count2++;
					if (snum2 == -1) snum2 = ld->frontsector - sectors;
					if (snum2 != ld->frontsector - sectors)
					{
						count1 = 0;
						break;
					}
				}
				else
				{
					count1 = 0;
					break;
				}
			}
			if (!count1 ||
				!count2 ||
				snum1 == i ||
				sectors[snum1].floorheight <= sec->floorheight) continue;

			M_ClearBox(bbox);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				M_AddLineToBox(bbox, ld);
			}

			for (k = 0; k < numsubsectors; k++)
			{
				int midx = (int)(subsectors[k].midpoint.x * FRACUNIT);
				int midy = (int)(subsectors[k].midpoint.y * FRACUNIT);

				if (midx > bbox[BOXLEFT ] &&	// Quick bounds test for rejection
					midx < bbox[BOXRIGHT] &&
					midy > bbox[BOXBASE ] &&
					midy < bbox[BOXTOP  ])
					Setr3dsslfsec(&subsectors[k], snum1, 29014);
			}
			Setsecdone(sec, 29015);
		}
	}
}

// Checks for structure that cannot be analysed in the normal way
// since it does not even work properly with a software renderer.
// Note: This is totally dependent on *hasDualRings ref: 46108 having been run.
static RDBOOL subfn_checkMaskedNested(sector_t *sc)
{
	sector_t	*ch, *os = NULL, *is = NULL;
	line_t		*ld;
	RDBOOL		valid = false;
	int			*ssecmem = NULL;
	int			c, d, i, n, rc;
	int			lnum1, lnum2, box1[4], box2[4];
	int			hsp;

	//------------------------

	CLR_hitline();

	for (c = d = i = 0; i < sc->linecount; i++)
	{
		if ((ld = sc->lines[i])->spflags & SPF_CONTROLLINE) continue;
		if (ld->frontsector == ld->backsector) continue;
		if (!ld->backsector) return false;

		n = ld - lines;
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (lineflags[n])
		{
			// Note: LFG_DUAL is set for the outer lines in *hasDualRings ref: 46108
			// This tells us that the sec has dual rings
			if (lineflags[n] & LFG_DUAL)
			{
				if (origTEX[n] != 0x73) return false;

				hitline[n] = 1;
				lnum1 = n;
				linearray[c++] = n;
				if (!os) os = ch; else if (os != ch) return false;
			}
			else return false;
		}
		else
		{
			if (origTEX[n] != 0x51) return false;
			hitline[n] = 1;
			lnum2 = n; d++;
			if (!is) is = ch; else if (is != ch) return false;
		}
	}

	// looking for double loop with static inner and outer sectors
	if (c < 3  || d < 3 || !is || !os) return false;
	if (is->specFlags & SF_MOVSECTOR || os->specFlags & SF_MOVSECTOR) return false;
	if (is->ceilingheight < sc->ceilingheight) return false;
	if (!subfn_CheckForRing(hitline, sc, lnum1, &rc, box1, NULL, NULL) || rc != c) return false;
	if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box1, -1, MSIA_NONE)) return false;
	if ( !(hsp = hitssecptr) ) return false;

	ssecmem = (int*)malloc(hsp * sizeof(int));
	memcpy(ssecmem, hitssec, hsp * sizeof(int));

	for (i = 0; i < c; i++) { if (hitline[linearray[i]] == 4) hitline[linearray[i]] = 64; }

	if (!subfn_CheckForRing(hitline, sc, lnum2, &rc, box2, NULL, NULL) || rc != c) goto tidy_up;
	if (!M_CompareBox(box2, box1)) goto tidy_up;
	// test whole area
	if (!LS_OnlySSLinesInSSECarray(sc, hitline)) goto tidy_up;

	if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box2, -1, MSIA_REVERSED))
	{

		for (i = 0; i < hsp; i++)
		{
			if (subsectors[ssecmem[i]].sector == sc)
				Setr3dsscs(&subsectors[ssecmem[i]], os - sectors, 999);
		}

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			n = ld - lines;

			if (!hitline[n]) continue;

			sides[ld->sidenum[0]].bottomtexture = 0;
			sides[ld->sidenum[1]].bottomtexture = 0;
			sides[ld->sidenum[0]].toptexture = 0;
			sides[ld->sidenum[1]].toptexture = 0;
			origTEX[n] &= 0x22; 

			if (!origTEX[n])
				Forcetype(ld, R3D_1D, 999);
			else
				Forcetype(ld, R3D_03, 999);

			SetLFG(ld, LFG_NEWSET, 999);
		}

		Setsecdone(sc, 999);
		Setsecdone(is, 999);
		valid = true;
	}
tidy_up:;
	free(ssecmem);
	return valid;
}

static void LS_SetPostcheckR3D_16Type()
{
	sector_t *sec;
	line_t	*ld;
	int		c, d, i, k, txn;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		// rev91
		if (sec->ceilingheight == sec->floorheight)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->specType == R3D_16
					|| ld->frontsector->heightsec >= 0
					|| (ld->backsector && ld->backsector->heightsec >= 0))
				{
					SetNOocclude(sec, 29014);
					break;
				}
			}

			// v220-22 added
			if (  !secdone[i]
				&& (sec->specFlags & 0x303859C) == 0x03028190
				&& !R3D_edit)
			{
				if (subfn_checkMaskedNested(sec)) continue;
			}

			// v220-5 Door sector.
			//        specific type - 2 lines with upper masked textures, 2 1's lines
			if (  !(sec->specFlags & SF_NOOCCLUDE)
				&& sec->heightsec < 0
				&& (sec->specFlags & 0xFFFFC) == 0x00028794)
			{
				sector_t *s1 = NULL;

				for (c = d = k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (ld->spflags & SPF_CONTROLLINE) continue;

					if (ld->ml_flags & ML_TWOSIDED)
					{
						sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

						if (!s1) s1 = s;

						if (s->specFlags & SF_MOVSECTOR
							||  (s1 != s	// only heights are used for an adjacent sec
							&& (s1->floorheight != s->floorheight
							||  s1->ceilingheight != s->ceilingheight)) )
						{
							c = 0; break;
						}

						txn = origTEX[ld - lines];

						if (   (txn == 0x04 && texArray[sides[ld->sidenum[0]].toptexture]->txf_flags & TXF_MASKED)
							|| (txn == 0x40 && texArray[sides[ld->sidenum[1]].toptexture]->txf_flags & TXF_MASKED) )
						{
							c++;
						}
					}
					else
					{	d++;
					}
				}

				if (c == 2 && d == 2)
				{
					SetNOocclude(sec, 29015);
					
					for (c = k = 0; k < sec->linecount; k++)
					{
						line_t *ld = sec->lines[k];

						if ( !(ld->ml_flags & ML_TWOSIDED) )
						{
							ld->adjacentsec = s1 - sectors;
						}
					}
				}
			}
		}
		// b211-0 Added - these types must not occlude when closed
		else if (sec->heightsec >= 0)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->frontsector->specFlags & SF_MOVSECTOR
					|| (ld->backsector && ld->backsector->specFlags & SF_MOVSECTOR))
				{
					SetNOocclude(sec, 29015);
					break;
				}
			}
		}

		if (!secdone[i] &&
			SecCtrl(i) &&
			sec->ceilingpic != skyflatnum &&
			sec->specFlags & SF_MOVFLOOR)
		{
			int valid = true;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (!(ld->ml_flags & ML_TWOSIDED) ||
					!(ld->mtflags & R3D_MISLOTEX) ||
					origTEX[ld - lines] ||
					!Vtxs2(ld) ||
					!Vtfl3(ld))
				{
					valid = false;
					break;
				}
			}

			if (!valid) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE || ld->specType == R3D_01)
					continue;

				if (ld->mtflags == R3D_MISLTEX1 &&
					ld->backsector->specFlags & SF_MOVFLOOR &&
					!(ld->frontsector->specFlags & SF_MOVFLOOR))
					Settype(ld, R3D_16, 29016);
				if (ld->mtflags == R3D_MISLTEX2 &&
					ld->frontsector->specFlags & SF_MOVFLOOR &&
					!(ld->backsector->specFlags & SF_MOVFLOOR))
					Settype(ld, R3D_16, 29017);
			}
		}
	}
}

static RDBOOL SechasMidtexOneSide(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (LSS(sec->lines[i]))
		{
			byte txn = origTEX[sec->lines[i] - lines];
			byte mtx = (txn & 0x22);

			if (mtx && mtx != 0x22
				&& !(txn & 0x55)
				&& !(sec->lines[i]->spflags & SPF_CONTROLLINE))
			{
				return true;
			}
		}
	}
	return false;
}

static void subfn_PostSetupRoutines()
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, val = 0;

	//====== Set R3D_14 lines

	// check move lines with no tex and missing tex's
	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->R3Dfloorheight < DDMAXINT ||
			sec->R3Dfloorsec >= 0 ||
			sec->heightsec >= 0)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->mtflags & R3D_MISLOTEX)
					hitline[sec->lines[k] - lines] = 1;
			}
		}
	}

	// Specific type identifying heightsec in heightsec
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->specFlags == 0x0080001C
			&& sec->heightsec >= 0
			&& !SecCtrl(i)
			&& sec->stptr < 0
			&& sec->linecount > 2)
		{
			sector_t *s;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->specType != R3D_00
					|| lineflags[ld - lines]
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| !Vtxs2(ld))
				{
					break;
				}

				s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (   s
					&& s != sec
					&& s->heightsec == sec->heightsec
					&& s->stptr < 0
					&& !(s->specFlags & SF_MOVSECTOR)
					&& s->floorheight > sec->floorheight
					&& s->ceilingheight == sec->ceilingheight)
				{
					c++;
				}
				else
					break;
			}

			if (c == sec->linecount)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];
					Settype(ld, R3D_14, 29018);
					SetLFG(ld, LFG_NEWSET|LFG_NOCHANGE, 29018);
				}
				Setsecdone(sec, 29018);
			}
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !hitline[i]
			&& ld->frontsector != ld->backsector)	// v220-9 test added
		{
			sector_t *fs = ld->frontsector;
			sector_t *bs = ld->backsector;

			if (!origTEX[i]
				&& !(ld->mtflags & R3D_MISTEXANY
					&& fs->specFlags & SF_MOVSECTOR
					&& bs->specFlags & SF_MOVSECTOR)
				&& ((fs->specFlags & SF_MOVSECTOR
					&& SechasMidtexOneSide(fs))
					||
					(bs->specFlags & SF_MOVSECTOR)
					&& SechasMidtexOneSide(bs))
				)
			{
				 Settype(ld, R3D_14, 29019);
			}
		}
	}

	// Mark all subsectors with a spawn glow sector to prevent any dynlight
	SET_countsec(0);

	// Check for mbf spawnglow - we don't want to add CFX glow
	// as it can swamp the mbf effect
	for (c = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if ((sec->special & 31) == 8) { c++; countsec[i] = true; }
	}

	if (c)
	{
		for (i = 0; i < numsubsectors; i++)
		{
			if (countsec[subsectors[i].sector - sectors])
				subsectors[i].ssflags |= SSF_NOGLOW;
		}
	}


	// Look for ceiling inheritance
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(secspflags[i] & SECSPF_CTRLLINES) &&
			MSFC(i, 0x00818198) &&
			sec->ceilingpic == skyflatnum)
		{
			sector_t *s = NULL, *cs;
			int		k, ht = sec->ceilingheight;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->ml_flags & ML_TWOSIDED) ||
					ld->frontsector == ld->backsector ||
					!Vtxs2(ld))
				{
					s = NULL;
					break;
				}
				cs = ld->frontsector == sec ?
					ld->backsector : ld->frontsector;
				if (cs->ceilingheight != ht ||
					cs->ceilingpic == skyflatnum)
				{
					s = NULL;
					break;
				}
				if (!s)
					s = cs;
				else if (s != cs)
				{
					s = NULL;
					break;
				}
			}
			if (s)
			{
				Setr3dcs(sec, s - sectors, 29023);
			}
		}
	}
}

static void subfn_ChkLiquidFloorRaisedSectors()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k, count = 0;

	byte *markline = (byte*)M_Calloc(numlines, 285);

	SET_countsec(0);

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int side = 0;

		if (ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector
			&& origTEX[i] == 0x11
			&& Vtxs2(ld)
			&& (
				(ld->mtflags == 0x04
				 && !IsLiquidFloorFlat(ld->frontsector)
				 && IsLiquidFloorFlat(ld->backsector)
				 && IsAnimTex(sides[ld->sidenum[0]].bottomtexture))
				||
				(side = 1,
				 ld->mtflags == 0x40
				 && IsLiquidFloorFlat(ld->frontsector)
				 && !IsLiquidFloorFlat(ld->backsector)
				 && IsAnimTex(sides[ld->sidenum[1]].bottomtexture))
				)
			)
		{
			int secnum = side ? ld->backsector - sectors :
								ld->frontsector - sectors;

			countsec[secnum] = 1;
			markline[i] = 1;
			count++;
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& !IsLiquidFloorFlat(sec))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];
				if (markline[ld - lines] == 1)
				{
					sector_t *s = ld->mtflags == 0x04 ?
						ld->frontsector : ld->backsector;

					if (s == sec
						&& LS_CheckRingInSec(markline, ld, s))
					{
						int fillsec = ld->frontsector == sec ?
							ld->backsector - sectors :
							ld->frontsector - sectors;

						int m;

						CLR_conarray();

						for (m = 0; m < sec->linecount; m++)
						{
							int lnum = sec->lines[m] - lines;


							if (markline[lnum] == 2)
							{
								markline[lnum] = 0; // It's job is done
								conarray[lnum] = 1;
							}
						}
					
						LS_FillGenArea(sec, fillsec, FGA_LFSEC, 29001); 
					}
				}
			}
		}
	}
	M_Free(markline);
}

// Line type R3D_3D - (has nothing to do with 3D lines!)
static void subfn_CheckR3D_3Dsec()
{
	sector_t *sec;
	int		 mask = SF_MOVLOWER|SF_MOVFLOOR|SF_HILOTEXTRUE|SF_SECCLOSED;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& SecCtrl(i)
			&& (sec->specFlags & mask) == mask
			&& sec->ceilingheight == sec->floorheight)
		{
			int valid = false;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ld->specType == R3D_3D
					&& !ld->mtflags
					&& Vtxs2(ld))
				{
					valid = true;
				}
				else
				{
					valid = false;
					break;
				}
			}
			if (valid)
			{
				Setr3dch(sec, sec, 29027);
				Setsecdone3(sec, 29028);
			}
		}
	}
}

// E.g Eternal map04 sec 772
static void subfn_Chk_MovFloorCeilSecs()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& sec->specFlags & SF_MOVCEIL
			&& sec->specFlags & SF_SECCLOSED
			&& sec->specFlags & SF_NOHITEX
			&& sec->specFlags & SF_NOLOTEX
			&& sec->specFlags & SF_SIDETEXTRUE
			&& !(sec->specFlags & SF_MOVFLOOR)
			&& !SecCtrl(i))
		{
			sector_t *s = NULL;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->specType == R3D_1D)
				{
					sector_t *chk = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;
					if (!s) s = chk;
					if (s != chk
						|| s == sec
						|| s->specFlags & SF_MOVSECTOR)
					{
						s = NULL;
						break;
					}
				}
				else
				{
					s = NULL;
					break;
				}
			}
			if (s)
			{
				Setr3dfs(sec, s - sectors, 29029);
			}
		}
	}
}

// rev60 - deals with upper tex lines wih crossing lines
static void subfn_ChkR3D_33lines()
{
	int			i, k;
	byte		*mark = (byte*)M_Calloc(numsubsectors, 286);

	hitssecptr = 0;

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].specType == R3D_33)
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				int n = sl->list[k];

				if (!mark[n])
				{
					mark[n] = true;
					hitssec[hitssecptr++] = n;
				}
			}
		}
	}

	M_Free(mark);

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (   ld
				&& ld->specType == R3D_33
				&& sg->frontsector == ld->frontsector)
			{
				SetOverride(ss, sg->frontsector - sectors, 29030);
				Setr3dsscs(ss, -1, 29031);
				Setr3dssfs(ss, -1, 29032);
				ss->ssflags |= SSF_NOCHANGE;	// v220-3
				break;
			}
		}
	}
}

static void subfn_Setsubsecs(sector_t *sec)
{
	subsector_t	*ss;
	int			i, k;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t *sg = &segs[ss->firstline];
			
		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && hitline[sg->linedef - lines])
			{
				sector_t *s = sg->linedef->frontsector == sec ?
					sg->linedef->backsector : sg->linedef->frontsector;
				SetOverride(ss, s - sectors, 29033);
				break;
			}
		}
	}
}

static RDBOOL chkValidR3D_1A(line_t *ld)
{
	vertexowner_t *own = &vertexowners[ld->v1 - vertexes];
	int i, found = false;

	for (i = 0; i < own->num; i++)
	{
		sector_t *s = &sectors[own->list[i]];
		int spf = s->specFlags;

		if (IsLiquidFloorFlat(s)) return false;

		if ((spf & 0xC) == 0xC)
		{
			found = true;
		}
		if (spf & SF_MOVSECTOR)
		{
			return false;
		}
	}
	if (!found) return false;
	
	own = &vertexowners[ld->v2 - vertexes];
	found = false;

	for (i = 0; i < own->num; i++)
	{
		sector_t *s = &sectors[own->list[i]];
		int spf = s->specFlags;

		if (IsLiquidFloorFlat(s)) return false;

		if ((spf & 0xC) == 0xC)
		{
			found = true;
		}
		if (spf & SF_MOVSECTOR)
		{
			return false;
		}
	}
	return found;
}

static int findSecFromLine(line_t *ld)
{
	ssline_t *sl = &sslines[ld - lines];
	line_t	 *li;
	int		 i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (li = sg->linedef)
				&& li->ml_flags & ML_TWOSIDED
				&& !origTEX[li - lines])
			{
				if (   li->backsector->ceilingheight
					<  li->frontsector->ceilingheight
					&& li->backsector != ld->frontsector
					&& li->backsector != ld->backsector)
				{
					return li->backsector - sectors;
				}
				if (   li->backsector->ceilingheight
					>  li->frontsector->ceilingheight
					&& li->frontsector != ld->frontsector
					&& li->frontsector != ld->backsector)
				{
					return li->frontsector - sectors;
				}
			}
		}
	}
	return -1;
}

static RDBOOL verifySSECS(sector_t *s)
{
	int i, k, m;

	for (i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		if (hitline[ld - lines])
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->sector == s)
				{
					seg_t *sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (   sg->backsector
							&& sg->frontsector == s
							&& sg->frontsector != sg->backsector
							&& sg->frontsector->floorheight <= sg->backsector->floorheight)
						{
							return false;
						}
					}
				}
			}
		}
	}

	return true;
}

static void LS_Finalise_Types()
{
	subsector_t *ss;
	seg_t		*sg;
	int			i, k;

	CLR_hitline();

	// The following checks are only performed where subsectors are unset
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		line_t   *l1 = NULL;
		line_t   *l2 = NULL;
		sector_t *s1 = NULL;
		sector_t *s2 = NULL;

		int count = 0, lineptr = 0, valid = true;

		// ========================================
		
		if (!ChkSSECunset(ss)) continue;

		validcount++;

		// Check for node assignment error with crossing lines
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef != NULL)
			{
				int lnum = sg->linedef - lines, found = false;

				if ( !(sg->linedef->ml_flags & ML_TWOSIDED) ) continue;

				if (sg->linedef->validcount == validcount) continue;

				if (sg->frontsector != sg->backsector)
				{
					if (!l1) l1 = sg->linedef;
					if (!s1) s1 = sg->frontsector;

					if (   l1 != sg->linedef
						|| s1 != sg->frontsector
						|| s1 == ss->sector)
					{
						valid = false; break;
					}

					if (++count > 1) break;
				}
				else
				{
					if (!l2) l2 = sg->linedef;
					if (!s2) s2 = sg->frontsector;

					if (   l2 != sg->linedef
						|| s2 != sg->frontsector
						|| s2 != ss->sector)
					{
						valid = false; break;
					}
				}
				
				if (++lineptr > 2) break;

				sg->linedef->validcount = validcount;
			}
		}

		if (   valid
			&& count == 1
			&& lineptr == 2
			&& subfn_FindLineCrossing(l2, LC_NORMAL) != NULL)
		{
			SetOverride(ss, s1 - sectors, 29036);

			// Mark lines as set
			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef != NULL) hitline[sg->linedef - lines] = 1;
			}
		}
	}
}

static RDBOOL ChkSubsecFloorUnset(subsector_t *ss)
{
	return (ss->floorsec < 0 || ss->floorsec == ss->sector - sectors)
		&& ss->sector->R3Dfloorsec < 0
		&& ss->sector->R3Dfloorheight == DDMAXINT
		&& ss->sector->floodline < 0;
}

static void subfn_checkInheritance()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		line_t *ld;

		if (  !secdone[i]
			&& SecCtrl(i)
			&& (sec->specFlags & 0xC) == 0xC)
		{
			int c = 0, d = 0, m;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
				{
					c++;
				}
				else if (!lineflags[ld - lines]
					&& ld->lnflags2 & LNF2_OWNEQUALTOONE
					&& ld->frontsector == ld->backsector
					&& Vtxs1(ld)
					&& !lxarray[ld - lines]
					&& !subfn_UnmergedVTXline(ld, 0))
				{
					linearray[d++] = ld - lines;
				}
				else
					break;
			}

			if (c && d == 1 && c + d == sec->linecount)
			{
				subsector_t *ss;
				int			losec = -1, hisec = -1;

				ss_validcount++;

				for (c = k = 0; k < d; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;

							if ((ss->sector->specFlags & 0xC) != 0xC)
							{
								losec = hisec = ss->sector - sectors;
							}
							else
							{
								if (hisec < 0)
								{
									if (ss->ceilsec  >= 0) hisec = ss->ceilsec;
								}
								if (losec < 0)
								{
									if (ss->floorsec >= 0) losec = ss->floorsec;
								}
							}

							hitssec[c++] = sl->list[m];
						}
					}
					if (hisec >= 0 || losec >= 0) break;
				}

				if (!c || d != 1) continue;

				ld = &lines[linearray[0]];

				if (hisec < 0 && losec < 0)
				{
					XS_GetBoundsFromLine(ld, &losec, &hisec, NULL);
				}

				if (hisec >= 0 || losec >= 0)
				{
					RDBOOL	isset = false;
					int		n;

					for (k = 0; k < c; k++)
					{
						ss = &subsectors[hitssec[k]];

						n = ss->sector - sectors;

						if (hisec >= 0 && hisec != n) Setr3dsscs(ss, hisec, 29001), isset = true;
						if (losec >= 0 && hisec != n) Setr3dssfs(ss, losec, 29001), isset = true;
					}

					SetLFG(ld, LFG_NEWSET, 29001);

					Setsecdone(sec, 29001);
				}
			}
		}

		if (sec->linecount == 4
			&& !secdone[i]
			&& sec->specFlags & SF_MOVCEIL
			&& !(sec->specFlags & SF_MOVFLOOR))
		{
			sector_t *osec = NULL;
			int count = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->specType == R3D_1D
					&& origTEX[ld - lines] == 0x02
					&& texArray[sides[ld->sidenum[0]].midtexture]->txf_flags & TXF_MASKED)
				{
					if (!osec)
					{
						osec = ld->frontsector;
						if (osec->specFlags & SF_MOVSECTOR)
						{
							osec = NULL;
							break;
						}
					}

					count++;
				}
			}
			if (count == 4 && osec)
			{
				CLR_conarray();
				for (k = 0; k < sec->linecount; k++)
					conarray[sec->lines[k] - lines] = 1;
				LS_FillGenArea(sec, osec - sectors, FGA_LFSEC, 29002);
			}
		}

		// rev93 Check for inheritance where AASHITTY tex is used
		if (!secdone[i]
			&& sec->floorheight == sec->ceilingheight
			&& sec->specFlags & SF_MOVSECTOR
			&& sec->specFlags & SF_NOHITEX
			&& sec->specFlags & SF_NOLOTEX
			&& !(sec->specFlags & SF_SECCLOSED)
			&& SecCtrl(i))
		{
			int found = false;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->backsector)
				{
					if (!origTEX[ld - lines]
						&&
						(ld->mtflags == 0x30 ||
						 ld->mtflags == 0x03))
					{
						Settype(ld, R3D_01, 29049);

						if (ld->frontsector->floorheight == ld->frontsector->floorheight
							|| ld->backsector->floorheight == ld->backsector->ceilingheight)

						{
							SetNOocclude(ld->frontsector, 29049);
							SetNOocclude(ld->backsector, 29049);
						}

						hitline[ld - lines] = 1;
						found = true;
					}
				}
			}
			if (found)
				subfn_Setsubsecs(sec);
		}
	}
}

// Used with LS_CheckNoDraw()
// Look for nodraw loops with at least one 1's line
static void checkInvisible()
{
	int i, k, m;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
			{
				if ( !(sg->linedef->ml_flags & ML_DONTDRAW))
				{
					return;
				}
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		ss->ssflags |= SSF_NOAMFLOOR;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
			{
				if (hitline[sg->linedef - lines] == 4)
				{
					if (   vtxcheck[sg->linedef->v1 - vertexes] > 2
						&& vtxcheck[sg->linedef->v2 - vertexes] > 2)
					{
						ssline_t *sl = &sslines[sg->linedef - lines];

						for (m = 0; m < sl->num; m++)
						{
							subsectors[sl->list[m]].ssflags |= SSF_NOAMFLOOR;
						}
					}

					if (!sg->linedef->backsector)
						hitline[sg->linedef - lines] = 0;
				}
			}
		}
	}
}

// v220-15 check ML_DONTDRAW flags for use with floor drawing in the automap
static void LS_CheckNoDraw()
{
	line_t		*ld = lines;
	int			lnum;
	int			a, c, i;

	CLR_vtxcheck();

	for (a = c = i = 0; i < numlines; i++, ld++)
	{
		if (ld->specType >= R3D_40) continue;

		if (ld->ml_flags & ML_DONTDRAW)
		{
			lnum = ld - lines;

			if (!ld->backsector) a++;

			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;
			linearray[c++] = ld - lines;
		}
	}

	if (!a || c < 3) return;
	
	// check no lines are 'hanging'
	for (i = 0; i < c; i++)
	{
		if (   vtxcheck[lines[linearray[i]].v1 - vertexes] > 1
			&& vtxcheck[lines[linearray[i]].v2 - vertexes] > 1)
		{
			hitline[linearray[i]] = 1;
		}
		else if (!lines[linearray[i]].backsector)
		{
			if (--a < 1) return;
		}
	}

	while(1)
	{
		sector_t	*sc;
		int			box[4], rc;

		lnum = -1;

		for (i = 0; i < c; i++)
		{
			ld = &lines[linearray[i]];

			if (!ld->backsector && hitline[linearray[i]])
			{
				lnum = linearray[i]; break;
			}
		}

		if (lnum < 0) break;

		sc = lines[lnum].frontsector;

		if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
			&& rc > 2
			&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
		{
			checkInvisible();
		}

		hitline[lnum] = 0;
	}
}

static void subfn_checkMovesecLadder(line_t *ld)
{
	subsector_t *ss;
	sector_t	*os = NULL;
	seg_t		*sg;
	ssline_t	*sl = &sslines[ld - lines];
	int		i, k;

	hitssecptr = 0;
	ss_validcount++;

	for (i = 0; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount == ss_validcount) continue;

		ss->ssvalidcount = ss_validcount;

		hitssec[hitssecptr++] = ss - subsectors;

		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (sg->frontsector != ld->frontsector)
			{
				if (os && os != sg->frontsector)
				{
					return;
				}
				else
				{
					os = sg->frontsector;
				}
			}
		}
	}

	if (!os || os->specFlags & SF_MOVFLOOR) return;

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];

		if (ss->sector != os)
		{
			ss->ssflags &= ~SSF_BASESSECFIX;
			ChangeSSECsec(ss, os, 2900);
			ss->ssflags |= SSF_BASESSECFIX;
		}
	}
}

// Done following graphics being cached as some routines need TXF_MASKED set
// Note: calling order for functions must not be changed
void LS_PostSetupRoutines()
{
	LS_CheckLineType();
	LS_CheckMaskedDoors();
	subfn_CheckR3D_3Dsec();

	subfn_PostSetupRoutines();
	LS_CheckSkyfix();
	LS_SetPostcheckR3D_16Type();

	LS_SNF1();
	subfn_ChkLiquidFloorRaisedSectors();	// rev44
	subfn_Chk_MovFloorCeilSecs();			// rev50
	subfn_ChkR3D_33lines();					// rev60
	LS_Set3DSpecialConditions();
	LS_SetSkyLoops();						// b211-0
	LS_CheckAdjacentSkyFromSubsector();		// v220-10
	LS_ResetUnwantedSkyfix();				// b211-0
	subfn_checkInheritance();

	MX_CheckNestedTexInMovRings();		// v220-19
	MC_CheckSSLinesContained();			// v220-27
	MX_CheckNoTexMovRings();			// v220-9
	// Nodes subsector issue and R3D_16, R3D_20 assignment
	LS_Finalise_Types();	// rev99L

	PS_UnwantedTextures();
	PS_SetTexInhibitFsecEqualBsec();
	PS_CheckUnsetLoTexSectors();
	PS_CheckSameSecMovSectors();		// v220-10
	MX_CheckLineRings();				// v220-9
	PS_CheckUnsetSupportLines();
	PS_FunctionsVarious();
	PS_CheckUnsetOwnerSectors();
	PS_CheckLiquidRings();
	PS_CheckLowerTexRings();
	PS_BoundedLiquidSectors();
	PS_MiscellaneousRoutines();
	PS_CheckFakeSecStructures();
	PS_CheckUnsetR3D_16Lines();
	PS_MidtexDoorsSpanning1SLines();
	PS_UnsetSkysecFromMidtexLine();
	PS_SectorHeightChecks();
	PS_SetTranslucentWaterFlags();		// v220-3/220-4
	PS_CheckUnclosedNoHILOtexSectors();
	PS_CheckCTRLSpecial();				// b211-0 $v71
	PS_CheckLiqLines();					// b211-0 $v71
	PS_CheckGenInheritance();			// b211-0 $v67
	PS_CheckNestedFakeMovSectors();		// b211-0 $v90
	PS_CheckUnsetCTRLsectors();
	PS_FinaliseSSECsettings();			// Final checks on ssec settings
	PS_FinaliseAASHITTYLines();			// b211-0 $v100
	PS_FinaliseMissingTexLoops();		// v220-6
	PS_FinaliseCrossingLines();			// v220-5
	PS_CheckEndLineTypes();				// v220-6
	PS_CheckEndSectorTypes();			// v220-6
	PS_CheckR3D_16Type();				// v220-6
	PS_CheckInvalidHiWalls();			// v220-6
	PS_CheckSSLinesAccessingMidTEX();	// v220-6
	PS_CheckMultiSecMissing();			// v220-7
	PS_ReplaceMissingHi();				// v220-7
	PS_CheckNoHiMovSectors();			// v220-7
	PS_CheckSameSidedLineRings();		// v220-7
	PS_Check3LineMissingLoTexSecs();	// v220-9
	PS_CheckSingleLineSsecs();			// v220-9
	MX_CheckMultiFloorMove();			// v220-10
	MX_R3D_16Nested();					// v220-10
	MX_UnsetLines();					// v220-9
	MX_Contained();						// v220-9
	MX_Unset();							// v220-9
	MX_InnerMovLoopContained();			// v220-10
	MX_Postlim();						// v220-9
	MX_ConnectedLo();					// v220-10
	MX_PartialMissing();				// v220-10
	LS_SameSectorContained();			// v220-10

	// v220-10 Find ssec that needs a different rendering sector but
	//         should not really need one, It is, strictly speaking,
	//         a map design fault but one that the rather vague
	//         software renderer will accept (it is also very rare).
	{
		subsector_t *ss;
		seg_t		*sg;
		int			c, i, k;

		for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
		{
			sector_t *s = ss->sector;

			if (s - sectors != ssinitial[i]) ma_inf.sseclogical++;

			if (ss->floorsec >= 0 || s->specFlags & 0x3000C) continue;

			validcount++;

			for (sg = &segs[ss->firstline], c = k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef && sg->linedef->validcount != validcount)
				{
					if (   sg->frontsector != sg->backsector
						|| sg->frontsector != s
						|| lineflags[sg->linedef - lines])
					{
						goto end_ssec;
					}

					linearray[c++] = sg->linedef - lines;
				}
			}

			if (c < 3) continue;

			// Check this subsector's lines form a loop

			if (nf_chkLineListRing(c, NULL, 0, NULL) == c)
			{
				sector_t *os = NULL, *chk;

				hitssecptr = 0;
				ss_validcount++;

				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];
					int		 m;

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						chk = ss->sector;

						if (chk == s)
						{
							hitssec[hitssecptr++] = sl->list[m];
						}
						else
						{
							if (!os) os = chk; else if (os != chk) goto end_ssec;
						}
					}
				}

				if (os && os->floorheight < s->floorheight)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						Setr3dssfs(&subsectors[hitssec[k]], os - sectors, 29051);

						if (s->ceilingheight > os->ceilingheight)
							Setr3dsscs(&subsectors[hitssec[k]], os - sectors, 29051);
					}
				}
			}
end_ssec:;
		}
	}

	// v220-10
	{
		line_t	*ld;
		int		i, k, m;
		
	//	1)	check ladder lines in 3D lift	v220-31
	//	2)	check R3D_03 types	v220-12
	//	3)  clear mtflags for missing hitex lines spanning 1's lines
		for (i = 0; i < numlines; i++)
		{
			if (lines[i].ovflags & OVF_3DLADDERLINE)
			{
				ld = &lines[i];

				if (   ld->frontsector == ld->backsector
					&& ld->frontsector->liftsec3D >= 0)
				{
					subfn_checkMovesecLadder(ld);
				}
				continue;
			}

			if (   lines[i].specType != R3D_00
				|| !(lines[i].ml_flags & ML_TWOSIDED))
			{
				continue;
			}

			if (lines[i].frontsector == lines[i].backsector)
			{
				if (origTEX[i] & 0x22) Forcetype(&lines[i], R3D_03, 29052);
				continue;
			}

			if (lineflags[i]) continue;

			if (   lines[i].mtflags & R3D_MISTEXANY
				&& lines[i].lnflags2 == 0x0C000000
				&& lines[i].spflags == (SPF_NOULTEX|SPF_NOTEX)
				&& lines[i].tyflags == TYF_NOMIDTEX
				&& !(lines[i].frontsector->specFlags & SF_MOVSECTOR)
				&& !(lines[i].backsector->specFlags & SF_MOVSECTOR)
				&& lines[i].ovflags == 0
				&& Vtfl1(&lines[i])
				&& Spans1sLines(&lines[i]))
			{
				Settype(&lines[i], R3D_36, 29052);
				Clr_allmtflags(&lines[i], 29052);
				continue;
			}

			// v220-10 project line coincident with 1's line forward.
			//         Done here so as not to affect preceding analysis
			if ( !(origTEX[i] & 0x55)
				&& origTEX[i] & 0x22
				&& !(lines[i].spflags & SPF_LOL_MASK))
			{
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t	*ss = &subsectors[sl->list[k]];
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef)
							&& !(ld->ml_flags & ML_TWOSIDED)
							&& VertexIsOnLine(ld->v1, &lines[i])
							&& VertexIsOnLine(ld->v2, &lines[i]))
						{
							lines[i].spflags |= SPF_LINEONLINE;
							goto end_linecheck;
						}
					}
				}
			}
end_linecheck:;
		}

	//	2)	Set OVF_MISHITEXLINE to indicate that a missing tex wall must be drawn
	//      Done after specType has been finalised
		for (i = 0; i < numsubsectors; i++)
		{
			if (subsectors[i].ceilsec >= 0)
			{
				seg_t *sg = &segs[subsectors[i].firstline];

				for (k = 0; k < subsectors[i].linecount; k++, sg++)
				{
					if (   sg->backsector
						&& (ld = sg->linedef)
						&& ld->specType == R3D_00
						&& !sg->sidedef->toptexture
						&& sg->backsector->ceilingheight < sectors[subsectors[i].ceilsec].ceilingheight)
					{
						ssline_t *sl = &sslines[ld - lines];

						for (m = 0; m < sl->num; m++)
						{
							if (   subsectors[sl->list[m]].sector != subsectors[i].sector
								&& subsectors[sl->list[m]].ceilsec < 0
								&& subsectors[sl->list[m]].sector->ceilingpic != skyflatnum)
							{
								ld->ovflags |= OVF_MISHITEXLINE;
								break;
							}
						}
					}
				}
			}
		}
	}

	// v220-10 Reset sectors that may have been affected by same sector sided lines
	{
		int i, k;

		for (i = 0; i < numsectors; i++)
		{
			if ((sectors[i].specFlags & 0xFFFFC) == 0x8198)
			{
				sector_t *s = &sectors[i];
				RDBOOL	 valid = true;

				for (k = 0; k < s->linecount; k++)
				{
					line_t *ld = s->lines[k];

					if ((origTEX[ld - lines] & 0x11) != 0x11) { valid = false; break; }
				}

				if (!valid) continue;

				CLR_hitline();

				for (k = 0; k < s->linecount; k++)
					hitline[s->lines[k] - lines] = 1;

				while (1)
				{
					int rc, box[4];
					int lnum = -1;

					for (k = 0; k < s->linecount; k++)
					{
						int n = s->lines[k] - lines;
						
						if (hitline[n] == 1) {lnum = n; break; }
					}

					if (lnum < 0) break;

					if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
					{
						if (   LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE)
							&& LS_OnlySSLinesInSSECarray(NULL, hitline))
						{
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];

								if (ss->ceilsec >= 0 && sectors[ss->ceilsec].ceilingheight >= s->ceilingheight)
									Setr3dsscs(ss, -1, 29053);

								if (ss->floorsec >= 0 && sectors[ss->floorsec].floorheight <= s->floorheight)
									Setr3dssfs(ss, -1, 29053);

								if (ss->sector != s)
								{
									Setr3dssfs(ss, s - sectors, 29054);

									// v220-13 Test added
									if ((ss->sector->specFlags & 0xC) == 0xC)
										Setr3dsscs(ss, s - sectors, 29054);
								}
							}
						}

						for (k = 0; k < s->linecount; k++)
						{
							int n = s->lines[k] - lines;
							
							if (hitline[n] == 4) hitline[n] = 0;
						}
					}
					else
					{
						hitline[lnum] = 0;
					}
				}
			}
		}
	}

	LS_FalseFloorRingSpan();	// v220-10 done at end
	LS_FalseCeilRingSpan();		// v220-13
	PS_CheckLotexPreference();	// v220-13
	PS_CheckMissingHeightsec();	// v220-16
	PS_CheckPartialMissingHi();	// v220-13
	PS_CheckPartialMissingLo();	// v220-13
	MC_ContiguousMissingHi();	// v220-16
	MC_MissingLoMove();			// v220-16
	MX_UnsetR3D_16();			// v220-16

	LS_SupportSecInside();		// v220-17
	LS_SetVisSplit();
	LS_CheckSlopeAnalysis();
	LS_CheckNoDraw();

	PS_SelfreferencingMovCeil();		// v220-18
	PS_CheckSubsecOcclude();			// v220-16
	PS_UnsetPartialMissingTEXLoops();	// v220-22
	TM2_CheckMissHiInDoubleMovCeil();	// v220-22
	TM2_PartialMissingLoMov();			// v220-22
	TM2_SingleMissingLowLoop();			// v220-33

	// Added v220-22
	{
		sector_t	*fs, *bs, *os;
		ssline_t	*sl;
		int			i, k, m, v;
		byte		*lmem = (byte*)calloc(numlines, 1);

		SET_countsec(0);
		CLR_vtxcheck();

		for (i = 0; i < numlines; i++)
		{
			if (   lines[i].specType == R3D_00
				&& !lineflags[i]
				&& (bs = lines[i].backsector)
				&& (fs = lines[i].frontsector) != bs
				&& (v = fs->floorheight - bs->floorheight))
			{
				if (   origTEX[i] == 0x22
					&& lines[i].ml_flags & ML_DONTPEGBOTTOM
					&& fs->ceilingheight == bs->ceilingheight)
				{
					// This may also need to have floorsec set
					sector_t	*s1 = NULL, *s2 = NULL;
					fixed_t		fh1 = DDMININT;
					fixed_t		fh2 = DDMAXINT;

					for (sl = &sslines[i], k = 0; k < sl->num; k++)
					{
						if ( (os = subsectors[sl->list[k]].sector)->floorheight > fh1) { fh1 = os->floorheight; s1 = os; }
						if ( (os = subsectors[sl->list[k]].sector)->floorheight < fh2) { fh2 = os->floorheight; s2 = os; }
					}

					if (fh1 > fh2 && s2 && s1 && (s1->specFlags & 0xC) == 0xC)
					{
						for (k = 0; k < sl->num; k++)
						{
							if (subsectors[sl->list[k]].sector->floorheight == fh1)
								Setr3dssfs(&subsectors[sl->list[k]], s2 - sectors, 29055);
						}
					}

					Settype(&lines[i], R3D_03, 29055);
				}
				// v220-24 Look for any unset missing lo loops inside sectors with
				//         other lines set. Usually confined to humunguous sectors
				//         caused by competitions leading to wads like 10sector.wad.
				else if ( !(origTEX[i] & 0x11)
					&& v > 8*FRACUNIT
					&& lines[i].mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)
					&& !(lines[i].mtflags & R3D_INVLOTEX)
					&& !(fs->specFlags & 0x3000C)
					&& !(bs->specFlags & 0x3000C))
				{
					if (lines[i].mtflags & R3D_MISLTEX1) os = bs; else os = fs;
					if (os->floorpic == skyflatnum) goto end_check;
					if (lines[i].mtflags & R3D_MISLTEX1) os = fs; else os = bs;

					if (   os->floodline >= 0
						|| os->R3Dfloorheight != DDMAXINT
						|| os->R3Dfloorsec >= 0
						|| os->R3Dlowsec >= 0)
					{
						goto end_check;
					}

					subsector_t	*ss;
					seg_t		*sg;

					for (sl = &sslines[i], k = 0; k < sl->num; k++)
					{
						if ( (ss = &subsectors[sl->list[k]])->sector == os)
						{
							if (ss->floorsec >= 0 || ss->lfsec >= 0) goto end_check;

							for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
							{
								if (sg->frontsector == sg->backsector) continue;
								if (sg->linedef && !(sg->linedef->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)) )
									goto end_check;
							}
						}
					}

					countsec[os - sectors]++;

					vtxcheck[lines[i].v1 - vertexes]++;
					vtxcheck[lines[i].v2 - vertexes]++;
					lmem[i] = true;
				}
end_check:;
			}
		}

		// v220-24
		for (i = 0; i < numsectors; i++)
		{
			line_t	*ld;
			int		c, box[4], rc;

			if (countsec[i] > 2)
			{
				fs = &sectors[i];
				CLR_hitline();

				for (c = k = 0; k < fs->linecount; k++)
				{
					ld = fs->lines[k];

					if (   lmem[ld - lines]
						&& vtxcheck[ld->v1 - vertexes] == 2
						&& vtxcheck[ld->v2 - vertexes] == 2)
					{
						linearray[c++] = ld - lines;
						hitline[ld - lines] = 1;
					}
				}

				if (c > 2)
				{
					while(1)
					{
						int lnum = -1;

						for (k = 0; k < c; k++)
						{
							if (hitline[linearray[k]] == 1) { lnum = linearray[k]; break; }
						}

						if (lnum < 0) break;

						if (subfn_CheckForRing(hitline, fs, lnum, &rc, box, NULL, NULL) && rc > 2)
						{
							for (os = NULL, k = 0; k < c; k++)
							{
								if (hitline[linearray[k]] == 4)
								{
									ld = &lines[linearray[k]];
									bs = ld->frontsector == fs ? ld->backsector : ld->frontsector;

									if (!os) os = bs; else if (os != bs) { os = NULL; break; }
								}
							}

							if (os && LS_MarkSubsecsInAreaGeneral(fs, hitline, 0, NULL, box, -1, MSIA_NONE))
							{
								for (k = 0; k < hitssecptr; k++)
									Setr3dssfs(&subsectors[hitssec[k]], os - sectors, 29056);

								// check for missinghi
								if (fs->ceilingpic != skyflatnum)
								{
									for (os = NULL, k = 0; k < c; k++)
									{
										if (hitline[linearray[k]] == 4)
										{
											ld = &lines[linearray[k]];

											if ( !(origTEX[ld - lines] & 0x44) )
											{
												bs = ld->frontsector == fs ? ld->backsector : ld->frontsector;

												if (bs->ceilingheight - fs->ceilingheight > 8*FRACUNIT)
												{
													sl = &sslines[ld - lines];

													for (m = 0; m < sl->num; m++)
													{
														subsector_t *ss = &subsectors[sl->list[m]];

														if (   ss->sector == bs
															&& ss->ceilsec == -1
															&& ss->lcsec == -1)
														{
															Setr3dsscs(ss, fs - sectors, 29057);
														}
													}
												}
											}
										}
									}
								}
							}

							for (k = 0; k < c; k++)
							{
								if (hitline[linearray[k]] == 4) hitline[linearray[k]] = 0;
							}
						}
						else
						{
							hitline[lnum] = 0;
						}
					}
				}
			}

			// v220-24 set SF2_NONEMISSMOVLO
			if (sectors[i].specFlags & SF_MOVFLOOR)
			{
				RDBOOL valid = true;

				for (k = 0; k < sectors[i].linecount; k++)
				{
					ld = sectors[i].lines[k];
					if (ld->spflags & SPF_CONTROLLINE) continue;

					v = ld->frontsector - sectors == i ? 1 : 0;

					if (ld->sidenum[v] >= 0)
					{
						if (sides[ld->sidenum[v]].bottomtexture == 0)
						{
							valid = true;
							break;
						}
					}
				}

				if (valid) sectors[i].specFlags2 |= SF2_NONEMISSMOVLO;
			}
		}

		free(lmem);
	}

	if (R3D_maptype) LS_3DLiftSectorSSecCheck();	// v220-31

	subfn_PrintStats();
}
