//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_precheck.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
//
// DESCRIPTION: checks made before first pass consolidation
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

//------------------------------------------------------------------------------
// sub functions

void subfn_CheckMissingTexInherited(sector_t *sec, int type)
{
	sector_t *s = NULL;
	int i, h, count = 0;
	int outer[4];

	switch(type)
	{
	case 0:
		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];
			if (ld->specType == R3D_2A &&
				ld->mtflags & R3D_MISLOTEX &&
				Vtxs2(ld) && Vtfl3(ld))
				count++;
			else
				return;
		}

		if (count < 3) return;

		s = sec->lines[0]->frontsector != sec ?
			sec->lines[0]->frontsector : sec->lines[0]->backsector;

		h = s->R3Dceilheight;

		if (h > DDMININT)
		{
			Setr3dch_byH(sec, h, 4001); 
			sec->ceilinglightsec = s->ceilinglightsec;
		}
		break;
	case 1:
		M_ClearBox(outer);
		count = 0;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];
			if (ld->specType == R3D_2A &&
				ld->mtflags & R3D_MISHITEX &&
				Vtxs2(ld) && Vtfl3(ld))
			{
				M_AddLineToBox(outer, ld);
				count++;
				if (!s)
					s = ld->frontsector != sec ? ld->frontsector : ld->backsector;
			}
			else if (ld->specType != R3D_00 ||
				ld->mtflags ||
				(ld->ml_flags & ML_TWOSIDED &&
				 (sides[ld->sidenum[0]].texFlags == T_ALL ||
				  sides[ld->sidenum[0]].texFlags == T_ALL)))
				return;
		}

		if (count < 3) return;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];
			if (ld->specType == R3D_00 &&
				ld->bbox[BOXLEFT ] >= outer[BOXLEFT ]   &&
				ld->bbox[BOXRIGHT] <= outer[BOXRIGHT]  &&
				ld->bbox[BOXBASE ] >= outer[BOXBASE ] &&
				ld->bbox[BOXTOP  ] <= outer[BOXTOP  ])
			{
				Setspec(ld, s - sectors, 4001);
			}
		}
		break;
	}

	Setsecdone(sec, 4001);
}

int subfn_GetAdjacentSecnum(sector_t *sec)
{
	int i, secnum;
	for ( i = 0; i < sec->linecount; i++)
		if (!(sec->lines[i]->spflags & SPF_CONTROLLINE) &&
			sec->lines[i]->specType == R3D_01 &&
			(secnum = LS_GetAdjacentLineSecnum(sec->lines[i], 9))>=0)
		{
			return secnum;
		}
	return -1;
}

RDBOOL subfn_Check0190Closed(sector_t *sec)
{
	line_t *ld;
	int i;

	CLR_vtxcheck();

	// Check marked lines are connected (i.e. closed)
	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE) &&
			(vtxarray[ld->v1 - vertexes] == 1 ||
			 vtxarray[ld->v2 - vertexes] == 1))
		{
			 return false;
		}
	}

	return true;
}

//------------------------------------------------------------------------------

void LS_CheckMissingTexInherited()
{
	sector_t *sec;
	int i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!SecCtrl(i) &&
			sec->heightsec < 0)
		{
			if (SecACTN(i) &&
				!(sec->specFlags & SF_SIDETEXTRUE))
				subfn_CheckMissingTexInherited(sec, 0);
			else if (!secspflags[i] &&
				sec->specFlags & SF_SIDETEXTRUE)
				subfn_CheckMissingTexInherited(sec, 1);
		}
	}
}

// Check closed/unclosed R3D_2A sectors
void LS_CheckR3D_2ASectors()
{
	sector_t *sec;
	int i, secnum;
	for (sec = sectors, i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(sec->specFlags & SF_SECCLOSED) &&
			!secspflags[i] &&
			sec->linecount > 2 &&
			sec->heightsec < 0)
		{
			int k, valid = true;
			for (k = 0; k < sec->linecount; k++)
				if (sec->lines[k]->specType != R3D_2A)
				{
					valid = false;
					break;
				}
			if (!valid) continue;

			secnum = sec->lines[0]->frontsector != sec ?
				sec->lines[0]->frontsector - sectors : sec->lines[0]->backsector - sectors;

			for (k = 0; k < sec->linecount; k++)
			{
				Setspec(sec->lines[k], secnum, 4002);
				Settype(sec->lines[k], R3D_3F, 4001);
			}
		}
	}
	for (sec = sectors, i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			sec->specFlags & SF_SECCLOSED &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			SecCtrl(i))
		{
			line_t *ldf = NULL;
			int k, valid = true, count = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ld->specType == R3D_2A &&
					Vtxs2(ld) && Vtfl3(ld))
				{
					ldf = ld;
					count++;
				}
				else
				{
					valid = false;
					break;
				}
			}

			if (valid && count > 2 && ldf)
			{
				secnum = ldf->frontsector != sec ?
					ldf->frontsector - sectors : ldf->backsector - sectors;

				// Check for height inheritance
				if (!(sec->specFlags & SF_MOVSECTOR))
					for (k = 0; k < sectors[secnum].linecount; k++)
						if (sectors[secnum].lines[k]->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2))
						{
							secnum = sectors[secnum].lines[k]->frontsector - sectors;
							break;
						}

				for (k = 0; k < sec->linecount; k++)
					if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
					{
						Setspec(sec->lines[k], secnum, 4003);
						Settype(sec->lines[k], R3D_3F, 4002);
					}
				Setsecdone(sec, 4002);
			}
		}
	}
}

// Check R3D_04 doors
void LS_CheckForR3D_04Doors()
{
	sector_t *sec = sectors;
	int i;
	for ( i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			sec->specFlags & SF_SECCLOSED &&
			(secspflags[i] & SECSPF_MASK1) == (SECSPF_HASACTION|SECSPF_CTRLLINES))
		{
			line_t *typebd_04 = NULL;
			int k, valid = true;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ld->specType == R3D_04 && LSS(ld))
				{
					if (!typebd_04 && ld->specType == R3D_04)
						typebd_04 = ld;
					else
					{
						valid = false; break;
					}
				}
				else
				{
					valid = false; break;
				}
			}
			if (!valid || !typebd_04) continue;

			Setr3dfh(sec, sec, 4002);
		}
	}
}

// Set R3D_04 lines if heightsec >= 0
void LS_CheckForR3D_04InFakedSectors()
{
	sector_t *sec = sectors;
	int i = 0, k;
	for (; i < numsectors; i++, sec++)
		if (sec->heightsec >= 0)
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->ml_flags & ML_TWOSIDED &&
					ld->specType == R3D_00 &&
					!ld->mtflags)
				{
					side_t *sd1 = &sides[ld->sidenum[0]];
					side_t *sd2 = &sides[ld->sidenum[1]];

					if (sd1->texFlags & T_MID &&
						sd1->midtexture ==
						sd2->midtexture)
					{
						// E.g tntrecl map03 and 05
						if ((sd1->texFlags & T_LOW && ld->backsector->heightsec >= 0) ||
							(sd2->texFlags & T_LOW && ld->frontsector->heightsec >= 0))
						{
							Settype(ld, R3D_04, 4003);
						}
					}
				}
			}
}

// Mainly batman stuff
void LS_CheckForR3D_30Lines()
{
	sector_t *sec = sectors;
	int i = 0, k;
	for (; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			secspflags[i] & SECSPF_FLOORTEXHT && // Height check set in missingtex
			sec->linecount > 3 &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			!IsLiquidFloorFlat(sec) &&
			sec->heightsec < 0)
		{
			int count = 0;
			side_t *sd1, *sd2;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->specType == R3D_3F ||
					(ld->specType == R3D_00 &&
					 ld->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)))
					continue;

				if (ld->specType != R3D_30)
				{
					count = 0;
					break;
				}
				count++;
			}
			if (!count) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE ||
					ld->specType == R3D_3F)
					continue;

				sd1 = &sides[ld->sidenum[0]];
				sd2 = &sides[ld->sidenum[1]];

				if (!SecACTN(ld->frontsector - sectors) &&
					!SecACTN(ld->backsector - sectors) &&
					((ld->mtflags & (R3D_MISLTEX1|R3D_MISHTEX1) &&
					  sd2->bottomtexture == sd2->toptexture) ||
					 (ld->mtflags & (R3D_MISLTEX2|R3D_MISHTEX2) &&
					  sd1->bottomtexture == sd1->toptexture)))
				{
					if (ld->backsector->R3Dfloorheight == DDMAXINT)
					{
						if (ld->mtflags & R3D_MISLTEX1)
							Setr3dfh(ld->backsector, ld->frontsector, 4003);
						else if (ld->mtflags & R3D_MISLTEX2)
							Setr3dfh(ld->frontsector, ld->backsector, 4004);
					}
					Settype(ld, R3D_3F, 4004);
				}
				else
					Settype(ld, R3D_1A, 4005);
			}
			Setsecdone(sec, 4003);
		}
	}
}

// R3D_2F handler
void LS_CheckR3D_2F()
{
	sector_t *sec = sectors;
	int i = 0, k, count1, count2, secnum = -1;

	for (; i < numsectors; i++, sec++)
	{
		if (secdone[i] ||
			secspflags[i]) continue;

		count1 = count2 = 0;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (   ld->specType == R3D_2A
				&& origTEX[ld - lines] & 0x11
				&& (ld->frontsector->specFlags & SF_MOVFLOOR)
				^  (ld->backsector->specFlags & SF_MOVFLOOR))
			{
				int tempsec = ld->frontsector != sec ?
					ld->frontsector - sectors : ld->backsector - sectors;

				if (secnum < 0)
					secnum = tempsec;
				else if (secnum != tempsec)
				{
					count1 = count2 = 0;
					break;
				}
				count1++;
			}
			else if (ld->specType == R3D_2F)
			{
				count2++;
			}
			else
			{
				count1 = count2 = 0;
				break;
			}
		}

		if (count1 < 3 || count2 < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (ld->specType == R3D_2A)
				Settype(ld, R3D_01, 11070);
		}

		Setlfsec(sec, secnum, 4001);
		Setsecdone(sec, 4006);
	}
}

// Check R3D_03 lines with sd2->texFlags == T_ALL for crossing
void LS_CheckTex7WithR3D_03()
{
	line_t *ld = lines, *ldc;
	int i = 0;
	for (; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_03 &&
			ChkLptr(ld) &&
			sides[ld->sidenum[0]].texFlags == T_MID &&
			sides[ld->sidenum[1]].texFlags == T_ALL &&
			(ldc = subfn_FindR3D_0BLineCrossing(ld)) != NULL &&
			ldc->frontsector != ldc->backsector &&
			ldc->frontsector != ld->frontsector &&
			ldc->backsector != ld->frontsector &&
			!sides[ldc->sidenum[0]].texFlags &&
			sides[ldc->sidenum[1]].texFlags == T_HI)
		{
			Setspec(ld, ldc->backsector - sectors, 4006);
		}
	}
}

// Mark sector flag if lines all in one sector
// Needed in rend_main.c for lift sectors
void LS_ResetClosedFlagForRenderer()
{
	sector_t *sec = sectors;
	int i = 0;

	for (; i < numsectors; i++, sec++)
	{
		// For renderer needs to be more specific
		sec->specFlags &= ~SF_03_CLOSED;

		if ((secspflags[i] & SECSPF_MASK2) == (SECSPF_LINONESEC|SECSPF_CTRLLINES|SECSPF_HASACTION)
			&& subfn_Check0190Closed(sec))
		{
			sec->specFlags |= SF_03_CLOSED;
		}
	}
}

void LS_CheckZeroHeightSector()
{
	sector_t *sec;
	int i, k;
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			IsLiquidFloorFlat(sec) &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->floorheight == sec->ceilingheight)
		{
			sector_t *s = NULL;
			int valid = true, mtflags = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->ml_flags & ML_TWOSIDED &&
					ld->specType == R3D_00 &&
					!origTEX[ld - lines] &&
					ld->mtflags & R3D_MISTEXANY)
				{
					if (!mtflags) mtflags = ld->mtflags;
					if (mtflags != ld->mtflags)
					{
						valid = false;
						break;
					}

					s = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (s->floorpic != sec->floorpic ||
						s->ceilingpic != sec->ceilingpic ||
						s->specFlags & SF_MOVSECTOR)
					{
						valid = false;
						break;
					}
				}
				else
				{
					valid = false;
					break;
				}
			}

			if (!valid || !s) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				Settype(sec->lines[k], R3D_01, 4007);
				Setspec(sec->lines[k], -1, 4007);
				Clr_allmtflags(sec->lines[k], 4001);
			}
			Setr3dfh(sec, s, 4008);
			Setr3dch(sec, s, 4009);
			SetNOocclude(sec, 4008);
			Setsecdone(sec, 4008);
		}
	}
}

void subfn_CheckR3D_3FContained(sector_t *sec)
{
	sector_t *csec;
	line_t *ld;
	int c, i, k = sec->linecount;

	if (sec->ceilingpic == skyflatnum ||
		sec->floorpic == skyflatnum)
		return;

	if (k < 3 ||
		sec->R3Dfloorsec >= 0 ||
		sec->R3Dceilsec >= 0)
		return;

	for (i = 0; i < k; i++)
	{
		ld = sec->lines[i];
		if (ld->specType != R3D_3F
			|| ChkLFGSecSet(ld)
			|| !(ld->spflags & SPF_NOTEX)
			|| !Vtxs2(ld)
			||
			(ld->backsector &&
			 ld->frontsector->ceilingheight == ld->backsector->ceilingheight &&
			 ld->frontsector->floorheight == ld->backsector->floorheight))
			return;

	}

	if ( (c = XS_GetBoundarySecnum(sec)) < 0) return;

	csec = &sectors[c];
	if (csec->ceilingpic != sec->ceilingpic)
		return;

	if ((sec->heightsec >= 0 && csec->heightsec >= 0) ||
		(csec->ceilingheight == sec->ceilingheight &&
		 csec->floorheight   == sec->floorheight))
		return;

	for (i = 0; i < k; i++)
		Setspec(sec->lines[i], c, 4008); // only set where heights don't match
}

void LS_CheckR3D_3FContained()
{
	int i;

	for (i = 0; i < numsectors; i++)
		if (secdone[i] != 2 && !(secdone[i] && IsLiquidFloorFlat(&sectors[i]))) // secdone can't be relied on
			subfn_CheckR3D_3FContained(&sectors[i]);
}

static void subfn_CheckUnsetR3D_1D(sector_t *sec)
{
	int i, count = 0, secnum;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (ld->frontsector != ld->backsector
			|| ChkLFGSecSet(ld)
			|| ChkLptr(ld))
			return;

		if (origTEX[ld - lines] &&
			!(ld->specType == R3D_1D ||
			  ld->specType == R3D_2D))
			 return;

		count++;
		vtxcheck[ld->v1 - vertexes] |= 1;
		vtxcheck[ld->v2 - vertexes] |= 2;
	}
	if (count != 4) return;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (vtxcheck[ld->v1 - vertexes] != 3 ||
			vtxcheck[ld->v2 - vertexes] != 3)
		{
			return;
		}
	}

	// Use even if -1
	secnum = XS_GetBoundarySecnum(sec);

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE))
			Setspec(ld, secnum, 4009);
	}
}
	
void LS_CheckUnsetR3D_1D()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			sec->specFlags & SF_MOVSECTOR)
		{
			subfn_CheckUnsetR3D_1D(sec);
		}
	}
}

// Check if R3D_29 types form a loop
void LS_CheckR3D_29Loops()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k;

	CLR_hitline();

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (!hitline[i] && ld->specType == R3D_29)
		{
			int valid = true;
			sec = ld->frontsector;
			CLR_vtxcheck();
			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->specType == R3D_29)
				{
					hitline[sec->lines[k] - lines] = 1;
					vtxcheck[sec->lines[k]->v1 - vertexes] |= 1;
					vtxcheck[sec->lines[k]->v2 - vertexes] |= 2;
				}
			}
			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->specType == R3D_29 &&
					(vtxcheck[sec->lines[k]->v1 - vertexes] != 3 ||
					 vtxcheck[sec->lines[k]->v2 - vertexes] != 3))
				{
					valid = false;
					break;
				}
			}
			if (!valid)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (sec->lines[k]->specType == R3D_29)
						Settype(sec->lines[k], R3D_1A, 4008);
				}
			}
		}
	}
}

// 1)	Check for unset R3D_04 special case
//		In the renderer this will be treated as an R3D_00 line other
//		than where type R3D_00 is checked for.  It's status as R3D_04
//		will mean this type will be ignored for specific R3D_00 checks.
void LS_CheckR3D_04SpecialCase()
{
	line_t *ld;
	int i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_00 &&
			ld->ml_flags & ML_TWOSIDED &&
			origTEX[i] == 0x27 &&
			ld->frontsector->heightsec >= 0 &&
			ld->backsector->heightsec < 0)
			Settype(ld, R3D_04, 4009);
	}
}

static RDBOOL subfn_CheckMovSecValidType1(sector_t *sec, int *fh)
{
	int i, count1 = 0, count2 = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];
		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			if (ld->mtflags & R3D_MISTEXANY) return false;

			if (Vtxs1(sec->lines[i]))
				count1++;
			else
			{
				if (!(ld->frontsector->specFlags & SF_MOVSECTOR))
					*fh = ld->frontsector->floorheight;
				else if (ld->backsector &&
					!(ld->backsector->specFlags & SF_MOVSECTOR))
					*fh = ld->backsector->floorheight;
				count2++;
			}
		}
	}
	return (count1 && count2 == count1 * 4);
}

// If only has unconnected move lines then nodes builder may have been confused
static RDBOOL subfn_CheckMovSecValidType2(sector_t *sec)
{
	int i, count = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		if (!(sec->lines[i]->spflags & SPF_CONTROLLINE) && !Vtxs1(sec->lines[i]))
			return false;
	}
	return true;
}

// Look for a moving sector with an enclosed unconnected line.
void LS_CheckMovSecWithSingleLine()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   LSS(ld)
			&& lnsec[ld - lines] < 0
			&& SecCtrl(ld->frontsector - sectors)
			&& !secdone[ld->frontsector - sectors])
		{
			if (Vtxs1(ld) && ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				int fh = ld->frontsector->floorheight;

				if (origTEX[i] & 0x22 &&
					subfn_CheckMovSecValidType1(ld->frontsector, &fh))
				{
					Setr3dfh_byH(ld->frontsector, fh, false, 4010);
					Setsecdone(ld->frontsector, 4009);
				}
				else if (!origTEX[i] && ld->frontsector->floorheight & FRACUNIT &&
					subfn_CheckMovSecValidType2(ld->frontsector))
				{
					int secnum = LS_GetAdjacentLineSecnum(ld, 10);
					Setspec(ld, secnum, 4010);
				}
			}
			else
			// v220-9 check inheritance of movceil lines
			if (  (ld->frontsector->specFlags & SF_MOVSECTOR) == SF_MOVCEIL
				&& vtxarray[ld->v1 - vertexes] > 2
				&& vtxarray[ld->v2 - vertexes] > 2
				&& vertexowners[ld->v1 - vertexes].num == 2
				&& vertexowners[ld->v2 - vertexes].num == 2)
			{
				// check subsectors belonging to this
				ssline_t *sl = &sslines[i];
				sector_t *chk = NULL;
				int		 k;

				hitssecptr = 0;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->sector == ld->frontsector)
						hitssec[hitssecptr++] = sl->list[k];
					else
					{
						if (!chk) chk = ss->sector;
						else if (chk != ss->sector) { chk = NULL; break; }
					}
				}

				if (   hitssecptr
					&& chk
					&& chk->ceilingheight > ld->frontsector->ceilingheight
					&& !(chk->specFlags & SF_MOVSECTOR))
				{
					int		secn = chk - sectors;
					int		valid = 0;

					for (k = 0; k < vertexowners[ld->v1 - vertexes].num; k++)
					{
						if (vertexowners[ld->v1 - vertexes].list[k] == secn) valid++;
						if (vertexowners[ld->v2 - vertexes].list[k] == secn) valid++;
					}

					if (valid == 2)
					{
						for (k = 0; k < hitssecptr; k++)
							Setr3dsscs(&subsectors[hitssec[k]], secn, 4010);
					}
				}
			}
		}
	}
}

static int subfn_findSec()
{
	int			i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t	*ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!sg->backsector && sg->linedef)
				return sg->linedef->frontsector - sectors;
		}
	}

	return -1;
}

// Check line rings with a same sided line in a mov sector
// that inherits a mov sector
// modified v220-19
void LS_MovLogicalSecInRenderSec()
{
	sector_t	*s, *sec;
	line_t		*ld;
	int			c, i, k, m, n;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!(sec->specFlags & SF_MOVFLOOR)) continue;

		for (s = NULL, k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (lxarray[ld - lines] & 16)
			{
				line_t *lc = subfn_FindLineCrossing(ld, LC_NORMAL);

				if (lc && (s = lc->frontsector)->specFlags & SF_MOVFLOOR)
					break;
				s = NULL;
			}
		}

		if (!s) continue;

		CLR_hitline();

		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (   !(origTEX[ld - lines] & 0x22)
				&& Vtxs2(ld)
				&& ld->lnflags2 & LNF2_OWNEQUALTOONE)
			{
				hitline[ld - lines] = 1; c++;
			}
		}

		if (c < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (   hitline[ld - lines] == 1
				&& lxarray[ld - lines] & 16
				&& subfn_CheckForRing(hitline, sec, ld - lines, NULL, NULL, NULL, NULL))
			{
				int v;

				ss_validcount++;
				hitssecptr = 0;

				for (c = m = 0; m < sec->linecount; m++)
				{
					v = sec->lines[m] - lines;

					if (hitline[v] == 4)
					{
						ssline_t *sl = &sslines[v];

						for (n = 0; n < sl->num; n++)
						{
							subsector_t	*ss = &subsectors[sl->list[n]];

							if (ss->ssvalidcount != ss_validcount)
							{
								if (ss->floorsec < 0)
									hitssec[hitssecptr++] = sl->list[n];
							}
						}

						linearray[c++] = v;
					}
				}

				v = -1;

				if (hitssecptr)
				{
					// Check for a 1's line
					if ( (v = subfn_findSec()) < 0)
						v = s - sectors;
				}

				for (k = 0; k < c; k++)
				{
					Setspec(&lines[linearray[k]], v, 4012);
					hitline[linearray[k]] = 8;
				}
			}
		}
	}
}

void LS_PreCheck_Types()
{
	sector_t	*sec;
	line_t		*ld;

	int	i, k, specSec;
	int	snum;
	int	mask1, mask2;

	// Look for unset unconnected textureless support lines
	// with step lines.
	// 1) Check lines
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (LSS(ld) &&
			ld->frontsector->specFlags & SF_MOVFLOOR &&
			lnsec[ld - lines] < 0 &&
			Vtxs1(ld) &&
			(specSec = LS_FindAdjacentSecnum(ld, 8)) >= 0)
		{
			Setspec(ld, specSec, 4013);
		}
	}

	// 2) Check unclosed sectors.
	//	  All other types must be eliminated
	//	  as lines may be contained by different sectors.  SF_MOVTYPE_X test
	//	  also incorporated as the AV23 map hack will probably only ever
	//	  be seen with such a move type.

	mask1 = SF_HILOTEXTRUE|SF_SSLINETRUE|SF_HASMIDTEXL|SF_SECCLOSED;
	mask2 = SF_MOVFLOOR|SF_MOVLOWER|SF_MOVTYPE_X;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& SecCtrl(i)
			&& !(sec->specFlags & mask1)
			&& (sec->specFlags & mask2) == mask2
			&& !IsLiquidFloorFlat(sec))
		{
			int specSec = -1;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ld->frontsector != ld->backsector ||
					origTEX[ld - lines] ||
					Vtxs1(ld) ||
					ld->mtflags ||
					subfn_FindLineCrossing(ld, LC_NORMAL) != NULL)
				{
					specSec = -1;
					break;
				}
				if (lnsec[ld - lines] >= 0)
				{
					if (specSec < 0)
						specSec = lnsec[ld - lines];
					else if (specSec != lnsec[ld - lines])
					{
						specSec = -1;
						break;
					}
				}
				else
				{
					int chksec = LS_FindAdjacentSecnum(ld, 8);
					if (chksec >= 0)
					{
						if (specSec < 0)
							specSec = chksec;
						else if (specSec != chksec)
						{
							specSec = -1;
							break;
						}
					}
				}
			}
			if (specSec >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
					{
						Setspec(sec->lines[k], specSec, 4014);
						Settype(sec->lines[k], R3D_01, 4014);
					}
				}
			}
		}
	}

	// 3)	Looks for same sided unconnected lines in a moving
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
	
		if (LSS(ld)
			&& !ld->special
			&& !ld->tag
			&& lnsec[ld - lines] < 0
			&& !origTEX[i] && Vtxs1(ld)
			&& ld->frontsector->specFlags & SF_MOVSECTOR
			&& ld->frontsector->ceilingpic != skyflatnum
			&& !IsLiquidFloorFlat(ld->frontsector)
			&& subfn_FindCoincidentLine(ld, 0) == NULL
			&& subfn_FindLineCrossing(ld, LC_NORMAL) == NULL
			&& (snum = LS_FindAdjacentSecnum(ld, 3)) >= 0)
		{
			Setspec(ld, snum, 4015);
		}
	}
}

