//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_secarbiter.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: determines which sector is needed from a mix
//              Not often needed but if called is printed out to the stats
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>
#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	master;

// CODE --------------------------------------------------------------------

void subfn_ChangePtrsForSingleLine(line_t *ld, line_t *lc)
{
	subsector_t *ss;
	int i;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t *sg = &segs[ss->firstline];
		int k, val1 = 0, val2 = 0;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->frontsector == ld->frontsector)
			{
				val1 = true;
			}
			if (sg->frontsector == lc->backsector)
			{
				val2 = true;
			}
		}

		if (val1 && val2)
			SetOverride(ss, lc->backsector - sectors, 24001);
	}
}

void subfn_SecArea(int secnum)
{
	sector_t *sec = &sectors[secnum];
	int i, count = IsLiquidFloorFlat(sec);

	CLR_conarray();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->frontsector != ld->backsector)
		{
			if (!(ld->spflags & SPF_CONTROLLINE))
				conarray[ld - lines] = 1;
		}
	}

	if (!LS_SetSectorArea(sec, secnum, 24100))
	{
		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i], *lc;

			if (LSS(ld) &&
				origTEX[ld - lines] == 0x22 &&
				vtxarray[ld->v1 - vertexes] > 1 &&
				vtxarray[ld->v2 - vertexes] > 1)
			{
				Setspec(ld, master, 24001);

				if (   lxarray[ld - lines] & 4
					&& (lc = subfn_FindLineCrossing(ld, LC_NORMAL|LC_DIFFSECLINE)) != NULL)
				{
					Settype(ld, R3D_04, 24001);
					subfn_ChangePtrsForSingleLine(ld, lc);
				}
			}
			else if (lnsec[sec->lines[i] - lines] != master)
				Setspec(sec->lines[i], -1, 24002);
		}
	}
	else
	{
		for (i = 0; i < sec->linecount; i++)
			Setspec(sec->lines[i], -1, 24003);
	}
}

void subfn_SectorArbiter(int secnum, int *sectormarker)
{
	int i;

	// Look for elevated lines
	for (i = 0; i < numsectors; i++)
	{
		if (sectormarker[i] == secnum && sectors[i].linecount > 4)
			subfn_SecArea(i);	
	}
}

static void subfn_CheckLineCrossing(line_t *ld, int *f, int *b)
{
	line_t *lc = subfn_FindLineCrossing(ld, LC_DIFF_FLR_HT);

	if (lc != NULL
		&& lc->ml_flags & ML_TWOSIDED
		&& lc->frontsector != ld->frontsector
		&& lc->frontsector != lc->backsector)
	{
		if (IsLiquidFloorFlat(lc->frontsector))
			*f = lc->frontsector - sectors;
		if (IsLiquidFloorFlat(lc->backsector))
			*b = lc->backsector - sectors;
	}
}

static RDBOOL imbalanced(line_t *ld)
{
	if (   lxarray[ld - lines] & 4
		&& ld->lnflags2 & LNF2_OWNEQUALTOONE)
	{
		return true;
	}

	if (   lxarray[ld - lines]
		|| origTEX[ld - lines] != 0x22)
	{
		return false;
	}

	if (   vertexowners[ld->v1 - vertexes].num == 1
		&& vertexowners[ld->v2 - vertexes].num >  1)
	{
		return true;
	}

	if (   vertexowners[ld->v2 - vertexes].num == 1
		&& vertexowners[ld->v1 - vertexes].num >  1)
	{
		return true;
	}

	return false;
}

// What we're looking for is a sector with floor animation which
// has same sector sided lines crossing it that are not referenced
// by any of the lines in the sector and which reference another
// sector without floor animation but having the same ceilingpic
// and identical floor and ceiling heights.
void LS_SectorArbitrator()
{
	subsector_t	*ss;
	sector_t	*sec;
	line_t		*ld;
	int			i, k;
	
	int		*sectormarker = (int*)M_Calloc(numsectors*sizeof(int), 288);
	byte	*sectordone = (byte*)M_Calloc(numsectors, 289);

	//----------------------------------

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   lineflags[i] & (LFG_NOARBITER|LFG_3DTYPE)
			|| !lxarray[i]
			|| Spans1sLines(ld))
		{
			continue;
		}

		if (linetable[i] & 32)
		{
			XS_SetLineSegs(ld);
		}
		else if (ld->frontsector == ld->backsector
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
			&& (lxarray[i] & 3 || imbalanced(ld))
			&& !(vtoflags[ld->v1 - vertexes]
			&&   vtoflags[ld->v2 - vertexes]) )
		{
			XS_SetLineSegs(ld);
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !sectordone[i]
			&& SecCtrl(i)
			&& (sec->specFlags & 0x3040C) == 0xC
			&& !IsAnimFlat(sec->floorpic))
		{
			int valid = true, nscount = 0, lflr = -1;
			int closed = sec->specFlags & SF_SECCLOSED;

			memset(sectormarker, 0, numsectors);
			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->spflags & SPF_CONTROLLINE
					|| lineflags[ld - lines] & LFG_3DTYPE
					|| (closed && lineflags[ld - lines] & LFG_NEWSET)
					|| origTEX[ld - lines] != 0x22)
				{
					continue;
				}

				if (ld->frontsector != ld->backsector)
				{
					valid = false;
					break;
				}

				hitline[ld - lines] = 1;

				if (lflr < 0)
				{
					if (ChkLptr(ld))
					{
						sector_t *s = &sectors[GetLptr(ld)];
						if (s != sec &&
							IsAnimFlat(s->floorpic) &&
							(s->specFlags & 0x400 || !(s->specFlags & 0x4)) &&
							s->floorheight ==
							sec->floorheight &&
							s->ceilingheight ==
							sec->ceilingheight)
						{
							lflr = GetLptr(ld);
						}
					}
					if (lflr < 0)
					{
						int b = -1;
						// Note: lflr can be front or back, (ld, &b, &lflr)
						//		 would be equally as valid.  The puzzle is to
						//		 identify the sector requiring arbitration,
						//		 lflr is purely the key.
						subfn_CheckLineCrossing(ld, &lflr, &b);
						if (lflr < 0 && b >= 0) lflr = b;
					}
				}
			}

			if (!valid || lflr < 0) continue;

			if (verbose)
				pr_OFinsert("  Arbitrator called, main sector %.5i : seed sector %.5i\n", i, lflr);

			ma_inf.called_arbitrator++;

			master = lflr;

			for (k = 0; k < sec->linecount; k++)
			{
				Setspec(sec->lines[k], -1, 24004);
			}

			for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
			{
				seg_t *sg;
				int m, valid = false;

				sg = &segs[ss->firstline];
				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->linedef != NULL && hitline[sg->linedef - lines])
					{
						SetOverride(ss, lflr, 24002);
						valid = true;
						break;
					}
				}
				if (!valid) continue;

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->linedef != NULL)
					{
						sector_t *s1 = sg->linedef->frontsector;
						sector_t *s2 = sg->linedef->backsector;

						if (s1 != sec)
						{
							sectormarker[s1 - sectors] = i;
						}
						if (s2 && s2 != sec)
						{
							sectormarker[s2 - sectors] = i;
						}
					}
				}
			}

			for (k = 0; k < numsectors; k++)
			{
				if (sectormarker[i]) sectordone[i] = 1;
			}

			sectormarker[lflr] = -1;
			sectormarker[i] = -1;

			subfn_SectorArbiter(i, sectormarker);
		}
	}

	M_Free(sectordone);
	M_Free(sectormarker);
}

