//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_secbbox.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: very old code - needs better algos. Tries to find
//              a containing sector.  Mainly succeeds - remarkably.
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

// Public
bounds_t	*secbbox;
int			csbox[4];


// Sets sector sibounds box
void LS_SetBOUNDS()
{
	sector_t *sec;
	int		 i, k;

	// Setup the sectors bounds and secbbox bounds
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		float	 *fbox = sectors[i].sibounds;
		int		 *box  = secbbox[i].bounds;

		M_ClearBox(box);

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (!(ld->spflags & SPF_CONTROLLINE))
				M_AddLineToBox(box, ld);
		}

		// Set the soundorg.z
		sec->soundorg.z = sec->floorheight;

		for (k = 0; k < 4; k++) fbox[k] = FIX2FLT(box[k]);
	}
}

//===========================================================================
//
// The following routines analyse lines in contained sectors to create
// required bounding boxes for lines visible on both sides and closed.
// The containing sector may not always be closed so this isn't checked.
//
//===========================================================================

// Check for invalid texture used to raise floor height
static void subfn_CheckRaisedFloors(sector_t *sec)
{
	sector_t		*invsec = NULL, *othersec = NULL;
	int				i;
	RDBOOL			inverror = false;

	for(i = 0; i < sec->linecount; i++)
	{
		line_t *li = sec->lines[i];

		if (!(li->ml_flags & ML_TWOSIDED))
			return;

		if (inverror) break;

		if (li->mtflags & R3D_MISTEXANY)
		{
			if (li->frontsector == sec &&
				li->mtflags & R3D_MISLTEX1)
			{
				if (!invsec)
					invsec = li->backsector;
				else
					if (invsec != li->backsector)
					inverror = true;
			}
			if (li->backsector == sec &&
				li->mtflags & R3D_MISLTEX2)
			{
				if (!invsec)
					invsec = li->frontsector;
				else
					if (invsec != li->frontsector)
					inverror = true;
			}
		}
		if (li->backsector != li->frontsector &&
			li->mtflags == (R3D_INVLTEX1|R3D_INVLTEX2))
		{
			if (li->frontsector == sec)
			{
				if (!othersec)
					othersec = li->backsector;
				else
					if (othersec != li->backsector)
					inverror = true;
			}
			if (li->backsector == sec)
			{
				if (!othersec)
					othersec = li->frontsector;
				else
					if (othersec != li->frontsector)
					inverror = true;
			}
		}
	}

	if (invsec && othersec && invsec != othersec)
	{
		Setr3dfs(sec, invsec - sectors, 13001);
		Setr3dfs(othersec, invsec - sectors, 13002);
		Setsecdone(sec, 13001);
		Setsecdone(invsec, 13001);
		Setsecdone(othersec, 13001);
	}
}

void LS_CheckRaisedFloors()
{
	sector_t *sec;
	int i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		if (sec->ceilingheight - sec->floorheight > MAX_SECHEIGHT &&
			sec->ceilingpic != skyflatnum &&
			sec->floorpic != skyflatnum)
			subfn_CheckRaisedFloors(sec);
}

//===========================================================================
// LS_GetContainingSectorOf
//===========================================================================

static RDBOOL subfn_CheckContingentLines(sector_t *csec, sector_t *sec)
{
	line_t	*ld;
	int		i, ret = true;
	byte	*marksec = (byte*)M_Calloc(numsectors, 290);

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		marksec[ld->frontsector - sectors] = 1;
		if (ld->ml_flags & ML_TWOSIDED)
			marksec[ld->backsector - sectors] = 1;
	}

	for (i = 0; i < csec->linecount; i++)
	{
		ld = csec->lines[i];

		if (marksec[ld->frontsector - sectors])
		{
			ret = false;
			break;
		}

		if (ld->ml_flags & ML_TWOSIDED &&
			marksec[ld->backsector - sectors])
		{
			ret = false;
			break;
		}
	}

	M_Free(marksec);

	return ret;
}

//=============================================================================
//
//	subfn_CheckContainingSectorLine
//
// To be inside a box then each line of the inner box must be facing a
// line in the outer box whose facing sector is the containing sector
// NOTE: 1) Sectors sharing lines eliminated (sec for this case can be
//          derived directly from a lines other sector).
//		2) Only double sided lines returned.
//

static line_t *subfn_NearestLine(line_t *ld, sector_t *csec)
{
	line_t *ln = NULL;
	int i, diff = DDMAXINT;

	// Imaginary line
	line_t lx;
	vertex_t v1, v2;
	// Get ld's centre points - these will be used for v1 of lx
	int lx1 = (ld->bbox[BOXLEFT]>>1) + (ld->bbox[BOXRIGHT]>>1);
	int ly1 = (ld->bbox[BOXBASE]>>1) + (ld->bbox[BOXTOP  ]>>1);

	lx.v1 = &v1;
	lx.v2 = &v2;
	lx.v1->x = lx1;
	lx.v1->y = ly1;

	for (i = 0; i < csec->linecount; i++)
	{
		line_t *ldc = csec->lines[i];

		if (!(ldc->spflags & SPF_CONTROLLINE)
			// Horizontal and vertical lines at 90deg eliminated as;
			//	a) must be at least one line in each sector not meeting this condition
			//	b) could cause line crossing check to fail.
			&& !((!ldc->dx && !ld->dy) || (!ldc->dy && !ld->dx)))
		{
			int lx2 = (ldc->bbox[BOXLEFT]>>1) + (ldc->bbox[BOXRIGHT]>>1);
			int ly2 = (ldc->bbox[BOXBASE]>>1) + (ldc->bbox[BOXTOP  ]>>1);
			int ndiff = abs(lx2 - lx1) + abs(ly2 - ly1);

			if (ndiff < diff)
			{

				ln = ldc;	// default
				diff = ndiff;

				// Set up line
				lx.v2->x = lx2;
				lx.v2->y = ly2;
				lx.dx = lx1 - lx2;
				lx.dy = ly1 - ly2;
				
				if (!lx.dx)
					lx.slopetype = ST_VERTICAL;
				else if (!lx.dy)
					lx.slopetype = ST_HORIZONTAL;
				else
				{
					if (FixedDiv (lx.dy , lx.dx) > 0)
						lx.slopetype = ST_POSITIVE;
					else
						lx.slopetype = ST_NEGATIVE;
				}

				lx.bbox[BOXLEFT ] = MIN_OF(v1.x, v2.x);
				lx.bbox[BOXRIGHT] = MAX_OF(v1.x, v2.x);
				lx.bbox[BOXBASE ] = MIN_OF(v1.y, v2.y);
				lx.bbox[BOXTOP  ] = MAX_OF(v1.y, v2.y);
				
				lx.frontsector = ld->frontsector;
				lx.backsector  = ld->backsector;

				// Don't use backsector in case a special fx line
				lx.ml_flags = (ld->ml_flags & ML_TWOSIDED) ? ML_TWOSIDED : 0;
		
				// If the imaginary line lx crosses a line in csec then, if it
				// has the shortest distance measured from the centre of ld
				// to the crossover point, ldc will be taken as being closer.
				if (LS_CheckLineIntersect(&lx, ldc , LC_NORMAL|LC_TESTLINE))
				{
					extern int xcrosspoint, ycrosspoint;

					ln = ldc;
					diff = abs(xcrosspoint - lx1) + abs(ycrosspoint - ly1);
				}
			}
		}
	}

	return (ln && ln->ml_flags & ML_TWOSIDED) ? ln : NULL;
}

RDBOOL subfn_CheckContainingSectorLine(line_t *ld, sector_t *csec)
{
	line_t *li = subfn_NearestLine(ld, csec);

	if (!li) return false;

	if (li->frontsector == csec &&
		li->backsector  == csec)
		return true;

	if (li->frontsector != csec &&
		li->backsector  != csec)
		return false;

	{
		float v1x  = FIX2FLT(li->v1->x);
		float v1y  = FIX2FLT(li->v1->y);
		float nx   = v1y - FIX2FLT(li->v2->y);
		float ny   = FIX2FLT(li->v2->x) - v1x;
		float vvx1 = v1x-FIX2FLT(ld->v1->x), vvy1 = v1y-FIX2FLT(ld->v1->y);
		float vvx2 = v1x-FIX2FLT(ld->v2->x), vvy2 = v1y-FIX2FLT(ld->v2->y);

		// The dot product - test each end of the reference line to
		// ensure, absolutely, the side of the line being faced.
		if (li->frontsector == csec &&
			nx*vvx1 + ny*vvy1 > 0 && nx*vvx2 + ny*vvy2 > 0)
				return true;

		if (li->backsector == csec &&
			nx*vvx1 + ny*vvy1 <= 0 && nx*vvx2 + ny*vvy2 <= 0)
				return true;
	}
	return false;
}

//=============================================================================
// LS_GetContainingSecnum
//  b211-0 rewritten
//=============================================================================
int LS_GetContainingSecnum(sector_t *sec, RDBOOL type)
{
	float	*inner, *outer, maxarea = 32768.0f * 32768.0f, area;
	int		secmem[256];
	int		i, k, m, n, secnum, seccnt;
	int		csky = sec->ceilingpic == skyflatnum;

	ss_validcount++;
	validcount++;

	// Get a list of the sectors sec is related to
	for (seccnt = i = 0; i < sec->linecount; i++)
	{
		subsector_t *ss;
		sector_t	*fs, *bs;
		seg_t		*sg;
		ssline_t	*sl;
		line_t		*ld = sec->lines[i];
		RDBOOL		found;

		if (ld->spflags & SPF_CONTROLLINE) continue;

		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if ( !(ld = sg->linedef) || ld->validcount == validcount) continue;

				if (seccnt >= 256) break;	// Unlikely

				ld->validcount = validcount;

				fs = ld->frontsector, bs = ld->backsector;

				if (   fs != sec
					&& !(csky && fs->ceilingpic != skyflatnum)
					&& !(type && SecONE(fs - sectors)))
				{
					found = false;

					for (n = 0; n < seccnt; n++)
					{
						if (secmem[n] == fs - sectors) { found = true; break; }
					}

					if (!found) secmem[seccnt++] = fs - sectors;
				}

				if (  !bs
					|| fs == bs
					|| bs == sec
					|| (csky && bs->ceilingpic != skyflatnum)
					|| (type && SecONE(bs - sectors)))
				{
					continue;
				}

				found = false;

				for (n = 0; n < seccnt; n++)
				{
					if (secmem[n] == bs - sectors) { found = true; break; }
				}

				if (!found) secmem[seccnt++] = bs - sectors;
			}
		}
	}

	inner = sec->sibounds;

	for (secnum = -1, i = 0; i < seccnt; i++)
	{
		n = secmem[i];

		outer = sectors[n].sibounds;

		if (   outer[BOXLEFT] == outer[BOXRIGHT]
			|| outer[BOXTOP ] == outer[BOXBASE ])
		{
			continue;
		}

		if (M_FitFPBox(inner, outer))
		{
			if ( (area = M_AreaFPBox(outer)) < maxarea)
			{
				secnum = n; maxarea = area;
			}
		}
	}

	if (secnum >= 0 && (sectors[secnum].specFlags & 0xC) == 0xC)
	{
		sec = &sectors[secnum];

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			if (s && !SecCtrl(s - sectors) && (s->specFlags & 0xC) != 0xC)
			{
				secnum = s - sectors; break;
			}
		}
	}

	return secnum;
}


sector_t *LS_GetContainingSectorOf(sector_t *sec, RDBOOL type)
{
	int secnum = LS_GetContainingSecnum(sec, type);

	sampleline = NULL;

	return secnum >= 0 ? &sectors[secnum] : NULL;
}

// End of containing sec
//==============================================================================

static sector_t	*otxsec;
static int		otxcount, otxmask;

void subfn_TraceOTEX(line_t *ld)
{
	vertex_t *v1 = ld->v1, *v2 = ld->v2;
	int i;

	hitline[ld - lines] = 1;
	otxcount++;

	if (vtxcheck[v1 - vertexes] == 3 &&
		vtxcheck[v2 - vertexes] == 3)
		return;

	vtxcheck[v1 - vertexes] |= 1;
	vtxcheck[v2 - vertexes] |= 2;

	for (i = 0; i < otxsec->linecount; i++)
	{
		ld = otxsec->lines[i];

		if (!hitline[ld - lines] &&
			ld->ml_flags & ML_TWOSIDED &&
			ld->spflags & SPF_NOTEX &&
			ld->backsector == otxsec &&
			origTEX[ld - lines] == otxmask &&
			(ld->v1 == v2 || ld->v2 == v1))
			subfn_TraceOTEX(ld);
	}
}

void subfn_CheckFlatsFromOrigTEX(sector_t *sec, int ceil)
{
	int i, k, valid, secnum, repeat = true, repeat2;

	otxmask = ceil ? 0x04 : 0x01;

	SET_countsec(0);
	CLR_hitline();
	countsec[sec - sectors] = 1;
	otxcount = 0;
	otxsec = sec;

	while (repeat)
	{
		line_t *ld, *lx;
		repeat = false;

		// Mark lines processed as done
		for (i = 0; i < sec->linecount; i++)
		{
			ld = sec->lines[i];
			if (hitline[ld - lines] == 1)
				hitline[ld - lines] = 2;
		}

		for (i = 0; i < sec->linecount; i++)
		{
			ld = sec->lines[i];
			otxcount = 0;

			// First check for lines setting the ceiling sector
			if (!hitline[ld - lines] &&
				ld->ml_flags & ML_TWOSIDED &&
				ld->spflags & SPF_NOTEX &&
				ld->backsector == sec &&
				origTEX[ld - lines] == otxmask)
			{
				CLR_vtxcheck();
				secnum = ld->frontsector - sectors;
				subfn_TraceOTEX(ld);
			}

			// If found then check they are closed
			if (otxcount)
			{
				valid = true;
				for (k = 0; k < sec->linecount; k++)
				{
					lx = sec->lines[k];
					if (hitline[lx - lines] == 1)
					{
						if (vtxcheck[lx->v1 - vertexes] != 3 ||
							vtxcheck[lx->v2 - vertexes] != 3)
						{
							valid = false;
							break;
						}
					}
				}

				// may need to look for more
				repeat = repeat2 = true;

				if (!valid) break;

				while (repeat2)
				{
					sector_t *nsec = &sectors[secnum];

					repeat2 = false;

					Setsecdone(nsec, 13002);

					if (sec->ceilingheight <= sec->floorheight)
					{
						SetNOocclude(sec, 13002);

						for (k = 0; k < sec->linecount; k++)
						{
							Setspec(sec->lines[k], secnum, 13001);
							Settype(sec->lines[k], R3D_3F, 13001);
						}
						return;
					}
					
					if (ceil)
						Setr3dcs(nsec, sec - sectors, 13003);
					else
						Setr3dfs(nsec, sec - sectors, 13004);

					for (k = 0; k < nsec->linecount; k++)
					{
						lx = nsec->lines[k];
						if (hitline[lx - lines] == 1)
						{
							Setspec(lx, -1, 13002);
							Settype(lx, R3D_3F, 13002);
						}

						if (lx->ml_flags & ML_TWOSIDED &&
							!origTEX[lx - lines])
						{
							hitline[lx - lines] = 2;
							Setspec(lx, -1, 13003);
							if (!countsec[lx->frontsector - sectors])
							{
								secnum = lx->frontsector - sectors;
								repeat2 = true;
								countsec[secnum] = 1;
							}
						}
					}
				}
			}
		}
	}
}

void LS_CheckFlatsFromOrigTEX()
{
	sector_t *sec;
	int i;
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->specFlags & (SF_HASORIGHITEX))
		{
			subfn_CheckFlatsFromOrigTEX(sec, 1);
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->specFlags & (SF_HASORIGLOTEX))
		{
			subfn_CheckFlatsFromOrigTEX(sec, 0);
		}
	}
}
