//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_ssline.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: some routines for finding a line's containing sector
//
//-----------------------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

line_t  *sampleline;

static byte *linesinsec;
static int  S5bbox[4];

int subfn_GetLineFacingLine(line_t *ld, line_t *li)
{
	// Check lines don't form a single straight line
	if (!subfn_CheckConnectedLineSlope(ld, li))
	{
		float v1x  = FIX2FLT(li->v1->x);
		float v1y  = FIX2FLT(li->v1->y);
		float nx   = v1y - FIX2FLT(li->v2->y);
		float ny   = FIX2FLT(li->v2->x) - v1x;
		float vvx1 = v1x-FIX2FLT(ld->v1->x), vvy1 = v1y-FIX2FLT(ld->v1->y);
		float vvx2 = v1x-FIX2FLT(ld->v2->x), vvy2 = v1y-FIX2FLT(ld->v2->y);

		if (ld->specType >= R3D_40 || li->specType >= R3D_40)
			return -1;

		// The dot product - test each end of the reference line to
		// ensure, absolutely, the side of the line being faced.
		if (nx*vvx1 + ny*vvy1 > 0 && nx*vvx2 + ny*vvy2 > 0)
		{
			if ((li->mtflags == (R3D_MISLTEX1|R3D_MISHTEX1)) ||
				((li->mtflags & R3D_MISTEXANY) == R3D_MISLTEX1 &&
				 Vtfl3(ld) &&
				 sides[li->sidenum[0]].texFlags == T_HI &&
				 sides[li->sidenum[1]].texFlags == T_HI))
			{
				sampleline = li;
				return li->backsector - sectors;
			}
			if (li->frontsector != ld->frontsector)
			{
				sampleline = li;
				return li->frontsector - sectors;
			}
		}
		else if (nx*vvx1 + ny*vvy1 <= 0 && nx*vvx2 + ny*vvy2 <= 0)
		{
			if ((li->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2)) ||
				(li->mtflags == R3D_MISLTEX2 &&
				 Vtfl3(ld) &&
				 sides[li->sidenum[0]].texFlags == T_HI &&
				 sides[li->sidenum[1]].texFlags == T_HI))
			{
				sampleline = li;
				return li->frontsector - sectors;
			}
			if (li->backsector != ld->frontsector)
			{
				sampleline = li;
				return li->backsector - sectors;
			}
		}
	}
	return -1;
}

int subfn_GetAdjacentAnimLineSecnum2(subsector_t *ss, line_t *ld)
{
	seg_t *seg;
	line_t *li;
	int i;

	seg = &segs[ss->firstline];
	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if ((li = seg->linedef) == NULL ||
			li->validcount == validcount)
			continue;

		li->validcount = validcount;

		if (li->backsector &&
			li->frontsector != ld->frontsector &&
			li->backsector  != ld->frontsector)
		{
			int secnum = subfn_GetLineFacingLine(ld, li);
			if (secnum >= 0 && IsLiquidFloorFlat(&sectors[secnum]))
				return secnum;
		}
	}
	return -1;
}

int subfn_CheckAdjacentHeightsec(subsector_t *ss, line_t *ld)
{
	seg_t	*seg;
	line_t	*li;
	int		i, secnum;

	seg = &segs[ss->firstline];
	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if ((li = seg->linedef) == NULL ||
			li->validcount == validcount)
			continue;

		li->validcount = validcount;

		if (li->ml_flags & ML_TWOSIDED &&
			(li->frontsector->heightsec >= 0 ||
			 li->backsector->heightsec >= 0) &&
			(secnum = subfn_GetLineFacingLine(ld, li))>=0)
			return secnum;
	}

	return -1;
}

static int subfn_CheckLSV(vertex_t *v1, vertex_t  *v2, vertex_t  *v3)
{
	__int64 x  = v1->x, y = v1->y;
	__int64 dx = (v2->x - v3->x);	// Line vector
	__int64 dy = (v2->y - v3->y);
	__int64 a  = (v2->x) - (x);		// point-->v1 vector
	__int64 b  = (v2->y) - (y);
	
	// Check if common vertex lies on line drawn from other two vertices.
	
	a *= dy;
	b *= dx;
	a -= b;
	b = dx + dy;
	if (((a-b)^(a+b)) < 0) return true;
	dy -= dx;
	b = a+dy;
	a -= dy;
	return (a^b) < 0;
}

static RDBOOL LinesAt180deg(line_t *li, line_t *ld)
{
	vertex_t v1, v2, v3;

	if (li->dy != ld->dy && li->dx != ld->dx) return false;
	if (li->v1->x != ld->v1->x && li->v1->y != ld->v1->y) return false;

	v1.x = MIN_OF(MAX_OF(li->v1->x, ld->v1->x),MAX_OF(li->v2->x, ld->v2->x));
	v1.y = MIN_OF(MAX_OF(li->v1->y, ld->v1->y),MAX_OF(li->v2->y, ld->v2->y));
	v2.x = MIN_OF(MIN_OF(li->v1->x, ld->v1->x),MIN_OF(li->v2->x, ld->v2->x));
	v3.x = MAX_OF(MAX_OF(li->v1->x, ld->v1->x),MAX_OF(li->v2->x, ld->v2->x));
	v2.y = MIN_OF(MIN_OF(li->v1->y, ld->v1->y),MIN_OF(li->v2->y, ld->v2->y));
	v3.y = MAX_OF(MAX_OF(li->v1->y, ld->v1->y),MAX_OF(li->v2->y, ld->v2->y));

	if (subfn_CheckLSV(&v1, &v2, &v3))
	{
		return true;
	}
	return false;
}

// Ammended rev170f
int subfn_GetLineSideFacingSecnum(line_t *li, line_t *ld, RDBOOL misstest)
{
	// If line has same sector each side then there's no point testing for the side facing
	if (li->frontsector == li->backsector) return li->frontsector - sectors;

	if (!LinesAt180deg(ld, li))
	{
		float v1x  = FIX2FLT(li->v1->x);
		float v1y  = FIX2FLT(li->v1->y);
		float nx   = v1y - FIX2FLT(li->v2->y);
		float ny   = FIX2FLT(li->v2->x) - v1x;
		float vvx1 = v1x-FIX2FLT(ld->v1->x), vvy1 = v1y-FIX2FLT(ld->v1->y);
		float vvx2 = v1x-FIX2FLT(ld->v2->x), vvy2 = v1y-FIX2FLT(ld->v2->y);

		// The dot product - test each end of the reference line to
		// ensure, absolutely, the side of the line being faced.
		if (nx*vvx1 + ny*vvy1 > 0 && nx*vvx2 + ny*vvy2 > 0)
		{
			if (misstest)
			{
				if (li->mtflags == (R3D_MISLTEX1|R3D_MISHTEX1))
					return li->backsector - sectors;
				if (li->frontsector != ld->frontsector)
					return li->frontsector - sectors;
			}
			else
				return li->frontsector - sectors;
		}
		else if (nx*vvx1 + ny*vvy1 <= 0 && nx*vvx2 + ny*vvy2 <= 0)
		{
			if (misstest)
			{
				if (li->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2))
					return li->frontsector - sectors;
				if (li->backsector != ld->frontsector)
					return li->backsector - sectors;
			}
			else
				return li->backsector - sectors;
		}
	}
	return -1;
}

int subfn_GetLineSideFacingSecnumAny(line_t *li, line_t *ld)
{
	float v1x  = FIX2FLT(li->v1->x);
	float v1y  = FIX2FLT(li->v1->y);
	float nx   = v1y - FIX2FLT(li->v2->y);
	float ny   = FIX2FLT(li->v2->x) - v1x;
	float vvx1 = v1x-FIX2FLT(ld->v1->x), vvy1 = v1y-FIX2FLT(ld->v1->y);
	float vvx2 = v1x-FIX2FLT(ld->v2->x), vvy2 = v1y-FIX2FLT(ld->v2->y);

	// The dot product - test each end of the reference line to
	// ensure, absolutely, the side of the line being faced.
	if (nx*vvx1 + ny*vvy1 > 0 && nx*vvx2 + ny*vvy2 > 0)
	{
		if (li->mtflags == R3D_MISHTEX1 &&
			Vtxs2(li))
		{
			sampleline = li;
			return li->backsector - sectors;
		}

		return li->frontsector - sectors;
	}

	if (li->mtflags == R3D_MISHTEX2 && Vtxs2(li))
	{
		sampleline = li;
		return li->frontsector - sectors;
	}

	return li->backsector - sectors;
}

static int subfn_FindAdjacentSecnum(subsector_t *ss, line_t *ld, int type)
{
	line_t	*li;
	seg_t	*seg = &segs[ss->firstline];
	int		i, secptr = -1;
	//----------------------

	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if ((li = seg->linedef) == NULL ||
			li->validcount == validcount)
			continue;

		li->validcount = validcount;

		switch(type)
		{
		case 1:
			if ( !(li->ml_flags & ML_TWOSIDED)
				&& li->frontsector != ld->frontsector)
			{
				return li->frontsector - sectors;
			}
			break;
		case 2:
			if (li->frontsector != ld->frontsector)
			{
				if ( !(li->ml_flags & ML_TWOSIDED) )
					return li->frontsector - sectors;

				if (li->frontsector != li->backsector)
				{
					if (li->backsector->floorheight > li->frontsector->floorheight)
						return li->frontsector - sectors;
					else
						return li->backsector - sectors;
				}
			}
			break;
		case 3:
			if (li->ml_flags & ML_TWOSIDED)
			{
				if (   li->frontsector != li->backsector
					&& (secptr = subfn_GetLineFacingLine(ld, li)) >= 0)
				{
					return secptr;
				}
			}
			else if (li->frontsector != ld->frontsector)
			{
				sampleline = NULL;
				return li->frontsector - sectors;
			}
			break;
		case 4:
			if ( !(li->ml_flags & ML_TWOSIDED) )
				return li->frontsector - sectors;

			secptr = subfn_GetLineFacingLine(ld, li);

			if (   secptr >= 0
				&& secptr != ld->frontsector - sectors
				&& secptr != ld->backsector - sectors)
			{
				return secptr;
			}
			break;
		case 5:
			if (   li->specType == R3D_04
				&& LSS(li)
				&& ChkLptr(li))
			{
				return GetLptr(li);
			}
			break;
		case 6:	// Added for vile_geo
			if (   seg->frontsector != ld->frontsector
				&& li->ml_flags & ML_TWOSIDED
				&& li->frontsector != li->backsector
				&& (ld->frontsector == li->frontsector
				||  ld->frontsector == li->backsector)
				&& origTEX[li - lines] & 0x11
				&& !(origTEX[li - lines] & 0x44)
				&& subfn_SegFacingMidpoint(seg->fv1, seg->fv2, ss->midpoint.x, ss->midpoint.y))
			{
				return ld->frontsector - sectors;
			}
			break;
		case 7:
			if ( !(li->ml_flags & ML_TWOSIDED) )
			{
				if (li->frontsector != ld->frontsector)
					return li->frontsector - sectors;
				break;
			}
			if ((secptr = subfn_GetLineSideFacingSecnum(li, ld, true)) != ld->frontsector - sectors)
				return secptr;
			break;
		case 8:
			if (   LSS(li)
				&& li->frontsector != ld->backsector
				&& origTEX[li - lines] == 0x22
				&& !(IsLiquidFloorFlat(li->frontsector))
				&& !(li->frontsector->specFlags & SF_MOVSECTOR))
			{
				return li->frontsector - sectors;
			}
			break;
		case 9:
			if (   Vtxs1(li)
				&& li->frontsector == li->backsector
				&& !(li->frontsector->specFlags & 0x0083000C))
			{
				return li->frontsector - sectors;
			}
			break;
		// Case 10 is for paired moving sectors.
		case 10:
			if (   li->ml_flags & ML_TWOSIDED
				&& !(li->spflags & SPF_CONTROLLINE)
				&& origTEX[li - lines] & 0x11
				&& li->frontsector->floorheight != li->backsector->floorheight
				&& !LineHasMovSectors(li))
			{
				sampleline = li;

				if (li->frontsector->floorheight == ld->frontsector->floorheight)
				{
					return li->frontsector - sectors;
				}
				if (li->frontsector->floorheight == ld->backsector->floorheight)
				{
					return li->frontsector - sectors;
				}
				if (li->backsector->floorheight == ld->frontsector->floorheight)
				{
					return li->backsector - sectors;
				}
				if (li->backsector->floorheight == ld->backsector->floorheight)
				{
					return li->backsector - sectors;
				}
			}
			break;
		case 11:
			if (   LSS(li)
				&& li->lnflags2 & LNF2_OWNEQUALTOONE
				&& li->frontsector != ld->frontsector
				&& lnsec[li - lines] < 0
				&& vtxarray[li->v1 - vertexes] > 1
				&& vtxarray[li->v2 - vertexes] > 1)
			{
				Setspec(li, lnsec[ld - lines], 18001);
			}
			break;
		default:
			Con_Error("FindAdjacentSector case %i",type);
			break;
		}
	}
	return secptr;
}

static void subfn_SetOverrideDirect(subsector_t *ss, line_t *li)
{
	sector_t *sec = ss->sector, *ns = NULL;
	seg_t *seg = &segs[ss->firstline];
	int i, count = 0;

	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if (seg->linedef != NULL && seg->linedef->special)
		{
			if (seg->frontsector == sec)
				count++;
			else
			{
				if (seg->frontsector == seg->backsector ||
					seg->frontsector->specFlags & SF_MOVSECTOR)
					return;

				if (!ns) ns = seg->frontsector;
				else if (ns != seg->frontsector)
					return;
			}
		}
	}
	if (ns && count == 1)
	{
		SetOverride(ss, ns - sectors, 18001);
		ma_inf.seclinefixed++;
		pr_OFinsert("  map error fixed : line %i should be in sector %i\n",
			li - lines, ns - sectors);
	}
}

int CallFNNUM(subsector_t *ss, line_t *ld, int fnnum, int type, int height)
{
	int secptr;

	switch (fnnum)
	{
	case 1:
		secptr = subfn_GetAdjacentAnimLineSecnum2(ss, ld);
		if (secptr >= 0) return secptr;
		break;
	case 2:
		if (type == 26 &&
			( (ss->floorsec >= 0 && ss->floorsec == ss->ceilsec)
			|| ss->sector != ld->frontsector))
		{
			return -1;
		}

		secptr = subfn_FindAdjacentSecnum(ss, ld, type);
		if (secptr >= 0) return secptr;
		break;
	case 3:
		secptr = subfn_CheckAdjacentHeightsec(ss, ld);
		if (secptr >= 0) return secptr;
		break;
	case 4:
		if (!(ss->floorsec >= 0 && ss->ceilsec == ss->floorsec)
			&& ld->frontsector == ss->sector)
		{
			subfn_SetOverrideDirect(ss, ld);
		}
		break;
	default:
		Con_Error("CallFNNUM: unknown index %i", fnnum);
		break;
	}

	return -1;
}


int subfn_GetSec(line_t *ld, int fnnum, int type, int height)
{
	int secptr;
	int i;
	ssline_t *sl = &sslines[ld - lines];

	sampleline = NULL;

	validcount++;

	ld->validcount = validcount;	//Don't check self

	for (i = 0; i < sl->num; i++)
	{
		secptr = CallFNNUM(&subsectors[sl->list[i]], ld, fnnum, type, height);
		if (secptr >= 0) return secptr;
	}

	return -1;
}

static int getAdjacentLineSecnum(line_t *ld)
{
	ssline_t *sl = &sslines[ld - lines];
	sector_t *cmp = ld->frontsector;
	sector_t *chk, *osec = NULL;
	line_t	 *li;
	int		 i, k;

	if (lxarray[ld - lines] & 4) return -1;

	validcount++;

	ld->validcount = validcount;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (li = sg->linedef) && li->validcount != validcount)
			{
				li->validcount = validcount;

				if (   (chk = sg->frontsector) != cmp
					&& !Vtxs1(li)
					&& !(!(origTEX[li - lines] & 0x66) // $v60 added
						&& sg->frontsector->specFlags & SF_MOVSECTOR)
					&& (sg->frontsector != sg->backsector
						|| (lxarray[li - lines] && !origTEX[li - lines])) )
				{
					if (!osec) osec = chk;
					else
					if (chk != osec) return -1;
				}
			}
		}
	}
	return osec ? osec - sectors : -1;
}

int LS_GetAdjacentLineSecnum(line_t *ld, int num)
{
	if (ld->frontsector != ld->backsector)
		return -1;

	return getAdjacentLineSecnum(ld);
}

int LS_GetAdjacentAnimLineSecnum(line_t *ld)
{
	return subfn_GetSec(ld, 1, 0, 0);
}

int LS_FindAdjacentSecnum(line_t *ld, int type)
{
	return subfn_GetSec(ld, 2, type, 0);
}

int LS_CheckAdjacentHeightsec(line_t *ld)
{
	return subfn_GetSec(ld, 3, 0, 0);
}

void LS_SetOverrideDirect(line_t *ld)
{
	subfn_GetSec(ld, 4, 0, 0);
}
