//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceInherited.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: loop trace function
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static void subfn_MarkSSECLines()
{
	int	i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t *sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;
			if (ld != NULL)
			{
				Settype(ld, R3D_3D, 36001);
				Setspec(ld, -1, 36001);
				Clr_allmtflags(ld, 36001);
				if (!secdone[ld->frontsector - sectors])
				{
					Setsecdone(ld->frontsector, 36001);
				}
				if (ld->backsector &&
					!secdone[ld->backsector - sectors])
				{
					Setsecdone(ld->backsector, 36002);
				}
			}
		}
	}
}

void LS_TraceContainedLinesInRing()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& (sec->specFlags & 0x30001C) == 0x1C
			&& sec->linecount > 2
			&& sec->ceilingpic == skyflatnum
			&& !IsLiquidFloorFlat(sec)
			&& sec->ceilingheight - sec->floorheight >= 8*FRACUNIT
			&& sec->stptr < 0
			&& sec->heightsec < 0)
		{
			sector_t *s = NULL, *t;
			int count = 0;

			CLR_conarray();

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (  !(ld->ml_flags & ML_TWOSIDED)
					|| lineflags[ld - lines]
					|| ld->frontsector == ld->backsector
					|| origTEX[ld - lines]
					|| ld->specType >= R3D_40
					|| !Vtxs2(ld)
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| lxarray[ld - lines]
					|| ld->frontsector->ceilingpic
					!= ld->backsector->ceilingpic
					|| ld->frontsector->floorpic
					!= ld->backsector->floorpic
					|| LineHasMovSectors(ld))
				{
					count = 0;
					break;
				}

				t = ld->frontsector == sec ?
					ld->backsector : ld->frontsector;

				if (!s) s = t;
				if (s != t) { count = 0; break; }

				if (ld->mtflags & R3D_MISTEXANY)
				{
					if (ld->frontsector->floorheight !=
						ld->backsector->floorheight)
					{
						count = 0;
						break;
					}

					if (  s->ceilingpic != skyflatnum
						&& ld->frontsector->ceilingheight
						!= ld->backsector->ceilingheight)
					{
						count = 0;
						break;
					}
				}

				conarray[ld - lines] = 1;
				count++;
			}

			if (count == sec->linecount)
			{
				int			sn = LS_GetContainingSecnum(sec, 1);
				RDBOOL		valid = true;

				if (   sn == -1
					|| sn == s - sectors
					|| sectors[sn].floorheight >= sec->floorheight)
				{
					sn = s - sectors, valid = false;
				}

				if (valid)
				{
					// Floodfill assumed
					LS_FillGenArea(sec, sn, FGA_SSOVRSEC, 36001);
					subfn_MarkSSECLines();

					for (k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k];

						SetLFG(ld, LFG_NEWSET, 36001);

						if (valid)
						{
							ssline_t *sl = &sslines[ld - lines];

							for (m = 0; m < sl->num; m++)
							{
								subsector_t *ss = &subsectors[sl->list[m]];

								if (ss->floorsec < 0)
									SetOverride(ss, sn, 36001);
							}
						}
					}
				}

				Setsecdone(sec, 36001);
				Setsecdone(&sectors[sn], 36001);
			}
		}
	}
}
