//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceNormal.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//		LS_TraceLiquidNormal() removed
//
// DESCRIPTION: finds all subsectors in a line loop
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "con_bar.h"

// Main functions - listed here for convenience when searching
// LS_TraceNormal()
// LS_CheckLinkedFloors()

//CODE ---------------------------------------------------------------------

static RDBOOL validLine(line_t *ld)
{
	if (!(ld->ml_flags & ML_TWOSIDED))
	{
		return true;
	}

	if (   ld->frontsector == ld->backsector
		|| chkSkyFloor(ld))
	{
		return false;
	}

	if (origTEX[ld - lines]
		|| (!Vtfl3(ld)
		&& (ld->frontsector->floorheight   != ld->backsector->floorheight
		||  ld->frontsector->ceilingheight != ld->backsector->ceilingheight)))
	{
		return true;
	}

	return !idenLine(ld);
}

static void setRefsecs(int secnum, int *fsec, int *csec)
{
	sector_t *sec = &sectors[secnum];
	int		 i;

	// All lines must have missing tex on boundary
	// to invoke a different sector reference
	for (i = 0; i < sec->linecount; i++)
	{
		line_t *li = sec->lines[i];

		if (hitline[li - lines] == 4)
		{
			if (!Vtfl3(li)) return;
		}
	}

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *li = sec->lines[i];
		int diff, num = li - lines;

		if (hitline[num] != 4
			|| !(li->ml_flags & ML_TWOSIDED))	// insurance
		{
			continue;
		}

		if (*fsec == secnum
			&& !(origTEX[num] & 0x11)
			&& (diff = li->frontsector->floorheight - li->backsector->floorheight) != 0)
		{
			*fsec = (diff > 0) ?
				li->frontsector - sectors : li->backsector - sectors;
		}
		if (*csec == secnum
			&& !(origTEX[num] & 0x44)
			&& !(li->frontsector->ceilingpic == skyflatnum
			&&   li->backsector->ceilingpic == skyflatnum)
			&& (diff = li->frontsector->ceilingheight - li->backsector->ceilingheight) != 0)
		{
			*csec = (diff > 0) ?
				li->backsector - sectors : li->frontsector - sectors;
		}
	}
}

static RDBOOL setConcentric(sector_t *sec, sector_t *osec, int dsptr, int *dslnmem, byte *sseclist)
{
	byte *templnlst1 = (byte*)M_Calloc(numlines, 301);
	byte *templnlst2 = (byte*)M_Malloc(numlines, 79);
	int  i, ret = false;

	int  tempptr = hitssecptr;
	int  *tempsslist = (int*)M_Malloc(tempptr * sizeof(int), 80);

	memcpy(tempsslist, hitssec, tempptr * sizeof(int));

	// First mark all subsectors with osec
	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		sseclist[hitssec[i]] = true;

		Setr3dsscs(ss, osec - sectors, 25001);
		Setr3dssfs(ss,  sec - sectors, 25002);
	}
	
	for (i = 0; i < dsptr; i++) templnlst1[dslnmem[i]] = 1;

	for (i = 0; i < dsptr; i++)
	{
		if (templnlst1[dslnmem[i]] == 1)
		{
			sector_t *s = lines[dslnmem[i]].frontsector;
			int		 k, ring_count, seccheck, bbox[4];

			if (subfn_CheckForRing(templnlst1, s, dslnmem[i], &ring_count, bbox, NULL, &seccheck))
			{
				if (seccheck >= 0)
				{
					ret = false;
					break;
				}

				// Mark all lines with this sector as we want the subsecs between them
				memset(templnlst2, 0, numlines);

				for (k = 0; k < dsptr; k++)
				{
					if (lines[dslnmem[i]].frontsector == s
						|| lines[dslnmem[i]].backsector == s)
					{
						templnlst2[dslnmem[i]] = 4;
					}
				}

				if (LS_MarkSubsecsInAreaGeneral(s, templnlst2, 0, NULL, bbox, -1, MSIA_NONE))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						sseclist[hitssec[i]] = true;

						Setr3dsscs(ss, s - sectors, 25003);
					}
					ret = true;
				}
			}
		}
	}

	memcpy(hitssec, tempsslist, tempptr * sizeof(int));
	hitssecptr = tempptr;

	if (ret)
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];

			sseclist[hitssec[i]] = true;
			ss->ssflags |= SSF_NOCHANGE;
		}
	}

	M_Free(tempsslist);
	M_Free(templnlst2);
	M_Free(templnlst1);

	return ret;
}

static void excludeSSECinternal(sector_t *sec, sector_t *osec, byte *sseclist)
{
	int i, k, txn;

	int  lnptr = 0, limit = 64;
	int *lnarray = (int*)M_Malloc(limit * sizeof(int), 81);


	//---------------------------------------------------

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;
			
			txn = origTEX[ld - lines];

			if (conarray[ld - lines] != 4
				&& Vtxs2(ld)
				&& ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector != ld->backsector)
			{
				if ((txn & 0x44
					&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
					|| (txn & 0x11
					&& ld->frontsector->floorheight != ld->backsector->floorheight))
				{
					if (lnptr >= limit)
					{
						limit += 64;
						lnarray = (int*)M_Realloc(lnarray, limit * sizeof(int), 40);
					}

					lnarray[lnptr++] = ld - lines;
				}
			}
		}
	}
			
	if (lnptr > 2)
	{
		int	 lo = false, hi = false, found = false;
		int	 ring_count, bbox[4];
		byte *lnmarker = (byte*)M_Calloc(numlines, 302);

		int  *templist = (int*)M_Malloc(hitssecptr * sizeof(int), 82);
		int  tempptr = hitssecptr;

		memcpy(templist, hitssec, hitssecptr * sizeof(int));

		for (i = 0; i < lnptr; i++)
		{
			lnmarker[lnarray[i]] = 1;
		}
		for (i = 0; i < lnptr; i++)
		{
			int lnum = lnarray[i];

			if (lnmarker[lnum] == 1)
			{
				sector_t *s = lines[lnum].frontsector;

				if (subfn_CheckForRing(lnmarker, s, lnum, &ring_count, bbox, NULL, NULL))
				{
					if (LS_MarkSubsecsInAreaGeneral(s, lnmarker, 0, NULL, bbox, -1, MSIA_NONE)
						||
						(s = lines[lnum].backsector,
						LS_MarkSubsecsInAreaGeneral(s, lnmarker, 0, NULL, bbox, -1, MSIA_NONE)))
					{
						for (k = 0; k < hitssecptr; k++)
						{
							sseclist[hitssec[k]] |= 4;
						}
						for (k = 0; k < lnptr; k++)
						{
							if (lnmarker[lnarray[k]] == 4)
							{
								lnmarker[lnarray[k]] = 8;

								if (origTEX[lnarray[k]] & 0x11) lo = true;
								if (origTEX[lnarray[k]] & 0x44) hi = true;
							}
						}
						found = true;
					}
				}
			}
		}

		hitssecptr = tempptr;
		memcpy(hitssec, templist, hitssecptr * sizeof(int));

		M_Free(templist);

		if (found && (!lo || !hi))
		{
			RDBOOL	skyflatlo = false;
			RDBOOL	skyflathi = false;

			// Test added v2.1.00 build_31
			// If the ring has unequal ceiling heights with a skyflat
			// each side then it does not qualify as a missing hi ring
			for (i = 0; i < sec->linecount; i++)
			{
				if (conarray[sec->lines[i] - lines] == 4)
				{
					if (sec->lines[i]->ml_flags & ML_TWOSIDED)
					{
						skyflathi = sec->lines[i]->frontsector->ceilingpic == skyflatnum
							&& sec->lines[i]->backsector->ceilingpic == skyflatnum;
						skyflatlo = sec->lines[i]->frontsector->floorpic == skyflatnum
							&& sec->lines[i]->backsector->floorpic == skyflatnum;
						break;
					}
				}
			}

			for (i = 0; i < hitssecptr; i++)
			{
				if (sseclist[hitssec[i]] & 4)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (!lo && !skyflatlo) Sethir3dssfs(ss, osec - sectors, 25004);
					if (!hi && !skyflathi) Setlor3dsscs(ss,  sec - sectors, 25005);
				}
			}
		}
	}

	M_Free(lnarray);
}

// Required test for type R3D_1D as this precedes r3d_missingtex.c
// and SetLFGSecSet() has been called prior to this test which inhibits
// any missingtex analysis.
static void subfn_checkR3D_1D(line_t *ld)
{
	int txn;

	if (   ld->ml_flags & ML_TWOSIDED
		&& ld->specType == R3D_00
		&& ((txn = origTEX[ld - lines]) == 0x02 || txn == 0x20)
		&& Vtxs2(ld)
		&& ld->frontsector != ld->backsector
		&& ld->frontsector->heightsec < 0
		&& ld->backsector->heightsec < 0
		&& (ld->frontsector->specFlags & SF_MOVSECTOR
		||  ld->backsector->specFlags & SF_MOVSECTOR))
	{
		fixed_t	floorh_f = ld->frontsector->floorheight;
		fixed_t	ceilh_f  = ld->frontsector->ceilingheight;
		fixed_t	floorh_b = ld->backsector->floorheight;
		fixed_t ceilh_b  = ld->backsector->ceilingheight;

		if (txn & 0x02 && !(ld->frontsector->specFlags & SF_MOVSECTOR))
		{
			if (ceilh_f > ceilh_b &&
				ld->backsector->specFlags & SF_MOVCEIL)
			{
				Settype(ld, R3D_1D, 25006);
			}
			else if (floorh_f < floorh_b &&
				ld->backsector->specFlags & SF_MOVFLOOR)
			{
				Settype(ld, R3D_1D, 25007);
			}
		}

		if (txn & 0x20 && !(ld->backsector->specFlags & SF_MOVSECTOR))
		{
			if (ceilh_f < ceilh_b &&
				ld->frontsector->specFlags & SF_MOVCEIL)
			{
				Settype(ld, R3D_1D, 25008);
			}
			else if (floorh_f > floorh_b &&
				ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				Settype(ld, R3D_1D, 25009);
			}
		}
	}
}

static void LS_MarkLineflagsInSSECarray(byte *mrklinearray, byte *sseclist, RDBOOL r3d_01type, int from)
{
	int i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		if (!sseclist || !sseclist[hitssec[i]])
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (   sg->linedef
					&& sg->linedef->ml_flags & ML_TWOSIDED
					&& mrklinearray[sg->linedef - lines] != 4)
				{
					SetLFGSecSet(sg->linedef, from);

					if (r3d_01type)
					{
						if (!origTEX[sg->linedef - lines])
						{
							// v220-6 if a movceil type then textures may need to be drawn
							if (   sg->linedef->frontsector->specFlags & SF_MOVCEIL
								|| sg->linedef->backsector->specFlags  & SF_MOVCEIL)
							{
								Settype(sg->linedef, R3D_37, from);
							}
							else
								Settype(sg->linedef, R3D_01, from);

							SetLFG(sg->linedef, LFG_FIXED, from);
						}
						else // Added v2.1.0 build_62
						{
							subfn_checkR3D_1D(sg->linedef);
						}
					}
				}
			}
		}
	}
}

//START-------------------------

static RDBOOL onlyOwnSecInside(sector_t *sec)
{
	int i, k, lnum, txn, count = 0;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->validcount == validcount) continue;

			ld->validcount = validcount;

			if (ld->frontsector == ld->backsector) continue;

			lnum = ld - lines;
			txn  = origTEX[lnum];

			if (conarray[lnum] == 4)
			{
				if (ld->ml_flags & ML_TWOSIDED)
				{
					if (!(txn & 0x11)
						&& ld->frontsector->floorheight
						!= ld->backsector->floorheight)
					{
						return 0;
					}
					if (!(txn & 0x44)
						&& ld->frontsector->ceilingheight
						!= ld->backsector->ceilingheight)
					{
						return 0;
					}
				}
				continue;
			}

			if (ld->ml_flags & ML_TWOSIDED
				&& !(origTEX[lnum] & 0x55)
				&& (ld->frontsector == sec
				||  ld->backsector == sec))
			{
				count++;
			}
			else
				return false;
		}
	}
	return count > 2;
}

//END---------------------------

//===============================================================================

static void tn_excludeSSEC(sector_t *sec, byte *sseclist, int *ptr, int count, int flag, RDBOOL *recheck)
{
	int		i, k, lnum, bbox[4];
	byte	*temphitline;
	int		*tempsseclist;
	int		tempptr = hitssecptr;

	//----------------------------

	if (count < 3) return;

	temphitline  = (byte*)M_Calloc(numlines, 303);
	tempsseclist = (int *)M_Malloc(hitssecptr * sizeof(int), 85);
	memcpy(tempsseclist, hitssec, hitssecptr*sizeof(int));

	for (i = 0; i < count; i++) temphitline[ptr[i]] = 1, hitline[ptr[i]] = 4;

	for (i = 0; i < sec->linecount; i++)
	{
		lnum = sec->lines[i] - lines;

		if (   temphitline[lnum] == 1
			&& subfn_CheckForRing(temphitline, sec, lnum, NULL, bbox, NULL, NULL))
		{

			if (LS_MarkSubsecsInAreaGeneral(sec, temphitline, 0, NULL, bbox, -1, flag))
			{
				for (k = 0; k < hitssecptr; k++)
					sseclist[hitssec[k]] = 3;

				*recheck = true;
			}

			for (k = 0; k < sec->linecount; k++)
			{
				lnum = sec->lines[k] - lines;

				if (temphitline[lnum] == 4)
				{
					temphitline[lnum] = 8;
					excludeline[lnum] = true;
				}
			}
		}
	}

	hitssecptr = tempptr;
	memcpy(hitssec, tempsseclist, hitssecptr*sizeof(int));

	M_Free(tempsseclist);
	M_Free(temphitline);
}

typedef struct {
	int secn;
	int box[4];
	int boxarea;
	int hitssecptr;
	int *sseclist;
} cringi_t;

static cringi_t *cri;

static RDBOOL checkRingsInside()
{
	sector_t	*s;
	line_t		*ld;
	RDBOOL		ret = false;
	int			c, i, k, n;
	int			*chkline = (int *)W_Calloc(numlines * sizeof(int), TEMPID, 0);
	byte		*setline = (byte*)W_Calloc(numlines, TEMPID, 0);
	int			maxcricount = 16;
	int			excludesseccount = 0;
	byte		*excludessecs;

	validcount++;

	cri = (cringi_t*)W_Calloc(sizeof(cringi_t)*maxcricount, TEMPID, 0);

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (hitline[ld - lines] == 4) continue;

				if (   !(ld->ml_flags & ML_TWOSIDED)
					|| (ld->frontsector != ld->backsector
					&& !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)) )
					goto end;

				if (   ld->frontsector == ld->backsector
					|| !Vtxs2(ld)
					|| ld->frontsector->specFlags & SF_MOVFLOOR
					|| ld->backsector->specFlags  & SF_MOVFLOOR
					|| ld->frontsector->floorheight != ld->backsector->floorheight
					|| ld->frontsector->floorpic != ld->backsector->floorpic
					// v220-5 eliminate if skysectors and same ceilingheights
					|| (ld->frontsector->ceilingpic == skyflatnum
					&&  ld->backsector->ceilingpic == skyflatnum
					&&  ld->frontsector->ceilingheight == ld->backsector->ceilingheight) )
				{
					continue;
				}

				if (  ld->frontsector->ceilingpic != skyflatnum
					|| ld->backsector->ceilingpic != skyflatnum)
				{
					goto end;
				}

				if (ld->frontsector->ceilingpic == ld->backsector->ceilingpic
					&& origTEX[ld - lines] & 0x44)
					continue;

				chkline[c++] = ld - lines;
				setline[ld - lines] = 1;
			}
		}
	}

	if (c < 3) goto end;

	for (n = i = 0; i < c; i++)
	{
		int rc;

		if (setline[chkline[i]] != 1) continue;

		ld = &lines[chkline[i]];
		s = ld->frontsector;

		if (subfn_CheckForRing(setline, s, chkline[i], &rc, cri[n].box, NULL, NULL))
		{
			int flag = MSIA_NONE;

			if (n >= maxcricount)
			{
				maxcricount += 16;
				cri = (cringi_t*)W_ReallocCLR(cri, sizeof(cringi_t)*maxcricount, TEMPID, 0);
			}

			if (!LS_MarkSubsecsInAreaGeneral(s, setline, 0, NULL, cri[n].box, -1, flag))
				flag = MSIA_REVERSED;
			if (!LS_MarkSubsecsInAreaGeneral(s, setline, 0, NULL, cri[n].box, -1, flag))
			{
				goto end;
			}

			if (flag == MSIA_REVERSED) s = ld->backsector;

			for (k = 0; k < n; k++) 
			{
				if (cri[k].secn == s - sectors) goto end;
			}

			cri[n].sseclist = (int*)W_Malloc(hitssecptr*sizeof(int), TEMPID, 37);
			memcpy(cri[n].sseclist, hitssec, hitssecptr*sizeof(int));
			cri[n].hitssecptr = hitssecptr;

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (setline[ld - lines] == 4) setline[ld - lines] = 8;
			}

			cri[n].boxarea =
				((cri[n].box[BOXRIGHT] - cri[n].box[BOXLEFT])>>FRACBITS) *
				((cri[n].box[BOXTOP  ] - cri[n].box[BOXBASE])>>FRACBITS);

			cri[n++].secn = s - sectors;
		}
	}

	if (!n) goto end;

	excludessecs = (byte*)W_Calloc(numsubsectors, TEMPID, 0);

	for (i = 0; i < n; i++)
	{
		int v = i;

		for (k = 0; k < n; k++)
		{
			if (cri[k].boxarea < cri[v].boxarea) v = k;
		}

		for (k = 0; k < cri[v].hitssecptr; k++)
		{
			int	ssecidx = cri[v].sseclist[k];

			if (!excludessecs[ssecidx])
			{
				subsector_t *ss = &subsectors[ssecidx];

				if ( !(ss->ssflags & SSF_NOCHANGE)
					&& ss->floorsec < 0	// v220-9 added
					&& ss->sector - sectors != cri[v].secn
					&& ss->sector->specFlags & SF_NOLOTEX
					&& sectors[cri[v].secn].floorheight > ss->sector->floorheight)
				{
					SetOverride(ss, cri[v].secn, 25038);
				}
				excludessecs[ssecidx] = true;
			}
		}

		Setsecdone(&sectors[cri[v].secn], 25039);

		cri[v].boxarea = DDMAXINT;
	}

	validcount++;

	for (i = 0; i < numsubsectors; i++)
	{
		if (excludessecs[i])
		{
			subsector_t *ss = &subsectors[i];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && ld->validcount != validcount)
				{
					if (!secdone[ld->frontsector - sectors])
						Setsecdone(ld->frontsector, 25040);

					if (ld->backsector
						&& !secdone[ld->backsector - sectors])
						Setsecdone(ld->backsector, 25041);

					SetLFG(ld, LFG_NEWSET, 25042);
					ld->validcount = validcount;
				}
			}
		}
	}

	ret = true;

end:;

	W_FreeALL(TEMPID);
	return ret;
}

static RDBOOL sameSecSsec(subsector_t *ss)
{
	seg_t	*sg;
	int		i;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (   sg->frontsector == sg->backsector
			&& ss->sector == sg->frontsector
			&& sg->frontsector->specFlags & SF_MOVFLOOR)
		{
			return true;
		}
	}

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector != sg->backsector)
		{
			if (   sg->linedef
				&& origTEX[sg->linedef - lines] & 0x11)
			{
				return false;
			}
		}
	}

	return true;
}

// v220-22 some code removed
static void subfn_TraceNormal(int secnum, int *missline, byte *sseclist)
{
	sector_t *sec = &sectors[secnum];

	int		i, count = 0, mcount = 0;
	int		secmisscount = 0;
	int		fht = sec->floorheight, cht = sec->ceilingheight;
	RDBOOL skyc = (sec->ceilingpic == skyflatnum);
	RDBOOL secmovf = (sec->specFlags & SF_MOVFLOOR) > 0;

	//---------------------------------------------------

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->frontsector == ld->backsector || lineflags[ld - lines] & LFG_3DTYPE)
			continue;

		if (   ld->ml_flags & ML_TWOSIDED
			&& !origTEX[ld - lines]
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
			&& !(ld->backsector->specFlags & SF_MOVSECTOR)
			&& chkSkyFloor(ld))
		{
			lineflags[ld - lines] |= LFG_NEWSET;
			Settype(ld, R3D_01, 11111);
			continue;
		}

		if (lineflags[ld - lines] & (LFG_NEWSET|LFG_FIXABS))
			continue;

		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector)
		{
			sector_t *s = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			RDBOOL skyflag = skyc && s->ceilingpic == skyflatnum;

			int mtx = s == ld->frontsector ? 0x20 : 0x02;

			if (skyflag
				&& s->floorpic == skyflatnum)
			{
				if (!origTEX[ld - lines]
					&& Vtxs2(ld)
					&& Vtfl3(ld)
					&& abs(s->floorheight - fht) > MAX_SECHEIGHT)
				{
					missline[mcount++] = ld - lines;
					count++;
					continue;
				}
			}

//START-------------------------
			// Detect textureless in non sky sectors with missing hi and low texs
			if (!secmovf
				&& !skyflag
				&& Vtxs2(ld)
				&& Vtfl3(ld)
				&& !(origTEX[ld - lines] & 0x55)
				&& !IsLiquidFloorFlat(s)
				&& !chkSkyFloor(ld)
				&& ((s->floorheight >  fht && s->ceilingheight <= cht)
				||  (s->floorheight >= fht && s->ceilingheight < cht)))
			{
				secmisscount++;
				continue;
			}
//END---------------------------

			// Bypass textureless lines with missing hi and low texs
			if (!(origTEX[ld - lines] & 0x55)
				&& !skyflag
				&& !(s->specFlags & SF_MOVSECTOR))
			{
				if (   (s->floorheight > fht && s->ceilingheight > cht)
					|| (s->floorheight < fht && s->ceilingheight < cht))
				{
					continue;
				}
			}

			if ((!(origTEX[ld - lines] & (0x11|mtx)) && fht > s->floorheight)
				||
				(!(origTEX[ld - lines] & 0x44) && cht < s->ceilingheight))
			{
				if (secmovf) continue;
				missline[mcount++] = ld - lines;
			}

			if (secmovf)
			{
				if (origTEX[ld - lines] & 0x11)
				{
					hitline[ld - lines] = 1;
					count++;
				}
			}
		}

		if (!secmovf && validLine(ld))
		{
			hitline[ld - lines] = 1;
			count++;
		}
	}
}


void LS_TraceNormal()
{
	sector_t *sec;
	int		 i, barpos = 5;

	int		 *missline = NULL;
	byte	 *sseclist = (byte*)M_Malloc(numsubsectors, 86);

	//-------------------------------------------------------

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if ((int)(60.0f * ((float)i/(float)numsectors)) > barpos)
		{
			barpos += 5;
			Con_Progress(5, 0); // $progress XH 4
		}

		if (   !secdone[i]
			&& sec->linecount > 2
			&& !(sec->specFlags & SF_MOVCEIL)
			&& !IsLiquidFloorFlat(sec)
			&& sec->ceilingheight - sec->floorheight > FRACUNIT
			&& sec->floorpic != skyflatnum)
		{
			missline = (int*)M_Realloc(missline, sec->linecount * sizeof(int), 41);

			subfn_TraceNormal(i, missline, sseclist);
		}
	}

	M_Free(missline);
	M_Free(sseclist);
}

// --------------------------------------------------------

static int LS_MarkLinesInLoopGeneral(sector_t *sec, byte *mrkarray, int *bbox, int *setarray, int limit)
{
	int i, k, lnptr = 0;

	if (!LS_MarkSubsecsInAreaGeneral(sec, mrkarray, 0, NULL, bbox, -1, MSIA_NONE))
		return 0;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!sg->linedef || sg->linedef->validcount == validcount)
				continue;

			sg->linedef->validcount = validcount;


			if (mrkarray[sg->linedef - lines] != 4)
			{
				if (lnptr >= limit) return 0;
				setarray[lnptr++] = sg->linedef - lines;
			}
		}
	}
	return lnptr;
}

static void subfn_CheckLinkedFloors(sector_t *sec)
{
	sector_t *s = NULL;
	int		 i, k, c1 = 0, c2 = 0;
	int		 ring_count, lnptr, bbox[4];
	int		 *setarray;

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t  *ld = sec->lines[i];
		int		txn;

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| !Vtxs2(ld)
			|| ChkLFGSecSet(ld)
			|| ld->frontsector == ld->backsector
			|| !((txn = origTEX[ld - lines]) & 0x55)
			|| txn & 0x22)
		{
			return;
		}

		if (txn & 0x11)
		{
			hitline[ld - lines] = 1;
			c1++;
		}
		else
		{
			sector_t *chk = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			if      (!s      ) s  = chk;
			else if (s != chk) return;

			hitline[ld - lines] = 16;
			c2++;
		}
	}

	if (c1 < 3 || c2 < 3 || !s || !(s->specFlags & SF_MOVFLOOR)) return;

	setarray = (int*)M_Malloc(c2 * sizeof(int), 87);

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 1
			&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL)
			&& (lnptr = LS_MarkLinesInLoopGeneral(sec, hitline, bbox, setarray, c2)) > 0)
		{
			for (k = 0; k < lnptr; k++)
			{
				if (hitline[setarray[k]] != 16)
				{
					M_Free(setarray);
					return;
				}
				
				hitline[setarray[k]] = 32;
			}

			for (k = 0; k < sec->linecount; k++)
			{
				if (hitline[sec->lines[k] - lines] == 4)
					hitline[sec->lines[k] - lines] = 8;
			}

			for (k = 0; k < hitssecptr; k++)
			{
				subsector_t *ss = &subsectors[hitssec[k]];

				if (ss->sector == s)
				{
					Setr3dssfs(ss, sec - sectors, 25052);
					SetNOocclude(ss->sector, 25053);
				}
			}

			// v220-22
			if ((s->specFlags & 0x4003040C) == 0x40010004)
			{
				line_t	 *ld;
				int		 side;

				for (k = 0; k < s->linecount; k++)
				{
					ld = s->lines[k];

					if (!ld->backsector) continue;

					side = ld->frontsector == s ? 1 : 0;

					if (ld->sidenum[side] < 0) continue;
					if (!sides[ld->sidenum[side]].toptexture) continue;

					Settype(ld, R3D_36, 25053);
				}
			}
		}
	}
	M_Free(setarray);
}

void LS_CheckLinkedFloors()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& (sec->specFlags  & 0x4003839C) == 0x40018390
			&& !SecCtrl(i)
			&& !IsLiquidFloorFlat(sec))
		{
			subfn_CheckLinkedFloors(sec);
		}
	}
}

//===============================================================================
