//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_xcheck.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: extra checks
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int backlinesecnum;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------


static void subfn_CheckR3D_1AAcessingFaked(int secnum)
{
	sector_t *sec = &sectors[secnum];
	int i, hs = -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];
		if (ld->backsector == NULL ||
			!(ld->tyflags & TYF_NOMIDTEX))
			return;

		if (ld->frontsector == sec &&
			ld->mtflags == R3D_MISHTEX1)
		{
			if ((hs = ld->backsector->heightsec) >= 0)
				break;
			return;
		}
		if (ld->backsector == sec &&
			ld->mtflags == R3D_MISHTEX2)
		{
			if ((hs = ld->frontsector->heightsec) >= 0)
				break;
			return;
		}
	}
	if (hs >= 0)
	{
		Setr3dcs(sec, hs, 20001);
		sec->ceilinglightsec = secnum;
	}
}

// Check for closed sectors with missing tex lines that
// could be referencing fake sectors;
void LS_XcheckMissingClosedInFaked()
{
	int i;
	for ( i = 0; i < numsectors; i++)
		if (sectors[i].specFlags & SF_SECCLOSED &&
			!(sectors[i].specFlags & SF_MOVSECTOR) &&
			secdone[i] &&
			!SecCtrl(i))
			subfn_CheckR3D_1AAcessingFaked(i);
}

// b211-0 revised
void LS_XcheckR3D_04defs()
{
	sector_t	*fs, *bs;
	line_t		*ld = lines;
	int			i, mt;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_00 &&
			(bs = ld->backsector) )
		{
			fs = ld->frontsector;

			if (ld->tyflags & TYF_MIDTEXNEGY &&
				fs->heightsec < 0 &&
				bs->heightsec < 0)
			{
				Settype(ld, R3D_04, 20001);
			}
			else if (ld->mtflags == R3D_MISLTEX1 &&
				sides[ld->sidenum[0]].texFlags == T_MID &&
				!sides[ld->sidenum[1]].texFlags)
			{
				lineflags[i] &= ~LFG_FIXED;
				Settype(ld, R3D_04, 20002);
			}
			else if (LSS(ld))
			{
				if (ld->tyflags & TYF_MIDTEXLINE)
					Settype(ld, R3D_04, 20003);
				else if (origTEX[i] == 0x22)
				{
					if (Vtxs1(ld))
						Settype(ld, R3D_3D, 20004);
					else
						Settype(ld, R3D_04, 20005);
				}
			}
			// b211-0 moving ceil line in faked sector
			else if (origTEX[i] == 0x22
				&& ((fs->heightsec >= 0 && (fs->specFlags & 0x3041C) == 0x2001C)
				||  (bs->heightsec >= 0 && (bs->specFlags & 0x3041C) == 0x2001C)) )
			{
				Settype(ld, R3D_04, 20006);
				SetLFG(ld, LFG_FIXED, 20006);
			}
			else
			// Type with low tex but missing upper on same side
			{
				if (   (origTEX[i] & 0x22) == 0x22
					&& origTEX[i] & 0x11
					&& ( ((mt = ld->mtflags & R3D_MISHITEX) == R3D_MISHTEX1
					&& (origTEX[i] & 0x11) == 0x01)
					|| (mt == R3D_MISHTEX2 && (origTEX[i] & 0x11) == 0x10) )
					&& !(ld->mtflags & R3D_INVTEXANY)
					&& !(fs->specFlags & SF_MOVSECTOR)
					&& !(bs->specFlags & SF_MOVSECTOR))
				{
					Settype(ld, R3D_04, 20006);
				}
			}
		}
	}
}

//==========================================================================


// Inheritance checks for support lines in anim sectors
static void subfn_XSetSpecForType(int secnum, int specSec, byte type)
{
	sector_t *sec = &sectors[secnum];
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (sec->lines[i]->specType == type)
			Setspec(sec->lines[i], specSec, 20006);
	}

	Setsecdone(sec, 20002);
}

static void subfn_FindXCheckSector(int secnum, int specSec, byte type)
{
	int i;
	// Find containing sector
	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		line_t *ld = sectors[secnum].lines[i];
		if (ld->specType == type)
		{
			int specSec2 = LS_FindAdjacentSecnum(ld, 5);
			if (specSec2 >= 0 &&
				specSec2 != specSec &&
				IsLiquidFloorFlat(&sectors[specSec2]))
			{
				subfn_XSetSpecForType(secnum, specSec2, type);
				return;
			}
		}
	}
}

static void subfn_XCheckR3D_3F(int secnum)
{
	line_t *ld;
	int i;

	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		ld = sectors[secnum].lines[i];
		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->specType != R3D_3F ||
			ld->frontsector != ld->backsector ||
			ChkLptr(ld) ||
			ld->mtflags ||
			vtxarray[ld->v1 - vertexes] != 2 ||
			vtxarray[ld->v2 - vertexes] != 2)
			return;
	}
		
	subfn_FindXCheckSector(secnum, -1, R3D_3F);
}

//===================================================================

// See if inheritance should be updated 
void LS_XcheckInheritanceGeneralLines(void)
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !IsLiquidFloorFlat(&sectors[i])
			&& !(sectors[i].specFlags & SF_MOVSECTOR))
		{
			subfn_XCheckR3D_3F(i);
		}
	}
}

// Check for unset R3D_1B lines in R3D_3F sector
void LS_XcheckR3D_1Bdefs()
{
	sector_t *sec = sectors;
	int i;

	for (i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->linecount > 7)
		{
			int k, count1 = 0, count2 = 0, specSec = -1, other = -1;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				int vtest = Vtxs2(ld);

				if (ld->specType == R3D_3F)
				{
					if (specSec == -1) specSec = GetLptr(ld);
					if (specSec == -1 ||
						!CmpLptr(ld, specSec) ||
						!vtest)
					{
						count1 = 0;
						break;
					}
					count1++;
				}
				else if (ld->specType == R3D_1B)
				{
					int os = ld->frontsector != sec ?
						ld->frontsector - sectors : ld->backsector - sectors;
					if (other == -1) other = os;
					if (ChkLptr(ld) ||
						other != os ||
						other == -1 ||
						!vtest)
					{
						count1 = 0;
						break;
					}
					count2++;
				}
				else if (vtxarray[ld->v1 - vertexes] != 1 ||
					vtxarray[ld->v2 - vertexes] != 1)
				{
					count1 = 0;
					break;
				}
			}

			if (other < 0 || count1 != 4 || count2 != 4) continue;

			subfn_resetSector(other, R3D_3F);

			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->specType != R3D_3F)
				{
					Setspec(sec->lines[k], specSec, 20010);
					if (sec->lines[k]->specType == R3D_1B)
						Settype(sec->lines[k], R3D_3F, 20012);
				}
			}
		}
	}
}
