//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Rewritten for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Rend_AddZ removed - smoothing done in r_main
//  New sector values used - smoothfloor, smoothceil
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 2 17/Jun/2008 gmj (b220-2)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//  The max amount of memory is now dynamically allocated.
//  As such sspoly must not be cleared using memset as this would cause
//  the address of rendpoly_vertex_t *vertices in rendpoly_t to be lost.
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
//  Added sector.withheight support
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 30 (v220-30) 03/Dec/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 14/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:   Derive parameters for rendering lists
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "de_render.h"
#include "de_refresh.h"
#include "de_play.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "r3d_xline.h"

#include "rend_clip.h"
#include "rend_dyn.h"
#include "rend_sky.h"

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_items.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

enum // For Wall Height Division.
{
	WHD_TOP,
	WHD_MIDDLE,
	WHD_BOTTOM
};

static struct polylist_t
{
	int			count;
	rendpolyq_t	list[MAXPOLYSIDES];
	seg_t		*segptr[MAXPOLYSIDES];

} plist;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	R_SetMissingParms(sector_t *f, sector_t *c);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		R3D_edit;
extern int		blacktexnum;
extern int		hasmask, transfloor;
extern int		thing_validcount;		// b211-0
extern RDBOOL	R3D_maptype;
extern RDBOOL	nearestClip;
extern float	occAngle;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

rendwallvars_t rv;
rendpolyq_t	*quad;


float		truevang, vang, dplookdir;
float		fvx, fvy, fvz, vpitch;
float		viewsidex, viewsidey;	
float		rend_lightcompress;		// R3D

int			r_ambient;
int			current_skyvarnum;		// v220-5

RDBOOL		whitefog;				// All fog types
RDBOOL		setNoFog;				// v220-3 for outside fog
RDBOOL		faked;					// R3D
RDBOOL		underwater;				// R3D
RDBOOL		underfakefloor;			// v220-3
RDBOOL		abovefakeceil;			// v220-3

RDBOOL		vssecunderfake;
RDBOOL		vssecoverfake;
rendpoly_t	sspoly;					// v220-5 use also for shadows

xlinemem_t		*xlinemem;
unsigned char	*xlisrendered;
int				use_xl_lines;
int				num_xlsame;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// R3D

static side_t	**lsiarray;
static int		lsiarray_count;
static RDBOOL	firstsubsector;

static int		rend_count;
static float	rend_flats_list[(MAX_FLATS3D + 4) * 2];

// v220-3
static byte		fogcol[4] = { 255, 255, 255, 0};

// v220-9
double rad_pitch;

// v220-18
static int currentnumsides;


// CODE --------------------------------------------------------------------

//===========================================================================
// Rend_LsiArrayInit v220-3
//  Called after map has been loaded but before rendering begins
//===========================================================================
void Rend_LsiArrayInit()
{
	int i;

	if (rv.maxsegcount < ma_inf.max_polynum)
	{
		rv.maxsegcount = ma_inf.max_polynum;
		rv.facingsegptr = (seg_t**)W_Realloc(rv.facingsegptr, rv.maxsegcount * sizeof(seg_t*), STATICID, 0);
	}

	for (i = 0; i < numsides; i++)
		sides[i].lsiflags &= ~LS_TEXPRESENT;

	if (numsides > currentnumsides)
	{
		currentnumsides = numsides;
		lsiarray = (side_t**)W_Realloc(lsiarray, numsides * sizeof(side_t*), STATICID, 42);
	}

	lsiarray_count = 0;
}

//===========================================================================
// Rend_PointDist2D
//	How far the point is from the viewside plane?
//===========================================================================
float Rend_PointDist2D(float *c)
{
/*          (YA-YC)(XB-XA)-(XA-XC)(YB-YA)
        s = -----------------------------
                        L**2
	Luckily, L**2 is one. dist = s*L. Even more luckily, L is also one.
*/
	float dist = (fvz - c[VY])*viewsidex - (fvx - c[VX])*viewsidey;

	return (dist < 0) ? -dist : dist;	// Always return positive.
}

#define LIGHTFACTOR 40

// v220-18 Use a sine curve adjustment (compress low values and expand higher values)
short SectorDoomLight(short ll)
{
	if (ma_inf.addlolight)
	{
		double val = (3.1415926535/2) * (double)(ll)/256;
		double sine = sin(val);
		ll = (int)(sine * 256);
	}
	else
	{
		if (ll > LIGHTFACTOR)	// $r_ambient usually 128 is best but criticality needs 40
		{
			int v = 3 * 2048;			// point to 270deg in finesine table

			v = v + ((ll - LIGHTFACTOR) * 2048) / (256 - LIGHTFACTOR);	// offset light level between 0 and 2048
			ll = LIGHTFACTOR + (finesine[v] + 65536) / (256 - LIGHTFACTOR);
		}
	}

	if (ll <   0) ll = 0;
	if (ll > 255) ll = 255;

	return ll;
}

//===========================================================================
//	R_SkyFromLinedef - R3D
//	 Routine to find sky from sidedef[0]
//   v220-5 extended to support 272
//===========================================================================

static RDBOOL sky_not_set;
int GLOBAL_SKY_SCROLL_LINE = -1;

static int R_SkyFromLinedef(sector_t *sect, int idx)
{
	if (r_skybox) return idx;	// v220-3

	int snum = sect->heightsec >= 0 ? sect->heightsec : sect - sectors;

	if (idx == skyflatnum || sectors[snum].ceilingpic == skyflatnum)
	{
		// v220-35 tag zero test added
		if (sectors[snum].sidesky & PL_SKYFLAT && sectors[snum].tag)
		{
			if (sky_not_set || ppt.view_subsector->sector == sect)
			{
				skyLayers[1].texture = sectors[snum].sidesky & ~PL_SKYFLAT;
				skyLayers[1].flags |= SLF_ENABLED;

				if (!(sectors[snum].specFlags2 & SF2_HASTYPE272))
					skyLayers[1].flags |= SLF_SKYFLIPY;
				else
					skyLayers[1].flags &= ~SLF_SKYFLIPY;

				skyLayers[1].sky_offy = sectors[snum].sky_offy;

				skyLayers[0].flags &= ~SLF_ENABLED;
				current_skyvarnum = sectors[snum].skyvar;

				// tagged sectors take precedence
				sky_not_set = !sectors[snum].tag;
			}
		}
		else if (sky_not_set)
			current_skyvarnum = sectors[snum].skyvar;
	}

	return idx;
}

//===========================================================================
// Rend_VtxFacingDir - R3D
//===========================================================================
static RDBOOL Rend_VtxFacingDir(float *v1, float *v2)
{
	return ((v1[VY] - v2[VY])*(v1[VX] - fvx)
			+ (v2[VX] - v1[VX])*(v1[VY] - fvz)) > 0;
}

//===========================================================================
// Rend_SegFacingDir
//===========================================================================
static RDBOOL Rend_SegFacingDir(seg_t *seg)
{
	return ((seg->fv1[VY] - seg->fv2[VY])*(seg->fv1[VX] - fvx)
		  + (seg->fv2[VX] - seg->fv1[VX])*(seg->fv1[VY] - fvz)) > 0;
}

static int __cdecl DivSortAscend(const void *e1, const void *e2)
{
	float f1 = *(float*) e1, f2 = *(float*) e2;

	if(f1 > f2) return 1;
	if(f2 > f1) return -1;
	return 0;
}

static int __cdecl DivSortDescend(const void *e1, const void *e2)
{
	float f1 = *(float*) e1, f2 = *(float*) e2;

	if(f1 > f2) return -1;
	if(f2 > f1) return 1;
	return 0;
}
//===========================================================================
// Rend_CheckDiv
//	Returns true if the quad has a division at the specified height.
//===========================================================================
static int Rend_CheckDiv(int side, fixed_t height, float *val)
{
	int		i;

	*val = FIX2FLT(height);

	for(i = 0; i < quad->divs[side].num; i++)
		if(quad->divs[side].pos[i] == *val) return true;
	return false;
}

//===========================================================================
// Rend_WallHeightDivision
//	Division will only happen if it must be done.
//	Converts quads to divquads.
//===========================================================================
static void R_WallHeightDivision(seg_t *seg, int mode)
{
	int				i, k, vtx[2];		// Vertex indices.
	vertexowner_t	*own;
	sector_t		*sec;
	int				hi = rv.fceil, lo = rv.ffloor;
	int				thi, tlo;
	int				lceil, lfloor;
	float			v;

	if (hi <= lo || quad->top <= quad->bottom) return;

	switch(mode)
	{
	case WHD_MIDDLE:
		break;

	case WHD_TOP:
		tlo = (int)(quad->top*FRACUNIT);
		if(lo < tlo) lo = tlo;
		break;

	case WHD_BOTTOM:
		thi = (int)(quad->bottom*FRACUNIT);
		if(hi > thi) hi = thi;
		break;

	default:
		return;
	}

	vtx[0] = seg->v1 - vertexes;
	vtx[1] = seg->v2 - vertexes;
	quad->divs[0].num = 0;
	quad->divs[1].num = 0;

	// Check both ends.
	for (i = 0; i < 2; i++)
	{
		own = vertexowners + vtx[i];

		if(own->num > 1)
		{
			// Check for sectors other than the segs front and back sector.
			for (k = 0; k < own->num; k++)
			{
				sec = &sectors[own->list[k]];

				if (sec == seg->frontsector ||
					sec == seg->backsector) continue;

				lceil  = sec->smoothceil;
				lfloor = sec->smoothfloor;

				if(lceil > lo && lceil < hi)
				{
					if (!Rend_CheckDiv(i, lceil, &v))
						quad->divs[i].pos[quad->divs[i].num++] = v;
					// Do we need to break?
					if (quad->divs[i].num == MAX_DIVS) break;
				}

				if(lfloor > lo && lfloor < hi)
				{
					if (!Rend_CheckDiv(i, lfloor, &v))
						quad->divs[i].pos[quad->divs[i].num++] = v;
					// Do we need to break?
					if (quad->divs[i].num == MAX_DIVS) break;
				}					
			}
			if (quad->divs[i].num)
			{
				quad->type = rp_divquad;

				if(quad->divs[i].num > 1)
				{
					// Sorting is required. This shouldn't take too long...
					// There are seldom more than one or two divisions.
					qsort(quad->divs[i].pos, quad->divs[i].num, sizeof(float), 
						i? DivSortDescend : DivSortAscend);
				}
			}
		}
	}
}

//===========================================================================
// R_SectorLight
//===========================================================================
static short R_SectorLight(short lightlevel)
{
	int ll;

	if (LevelFullBright) return 255;

	if (rend_lightcompress && lightlevel)	// $r_ambient
	{
		float fll = (float)lightlevel;
		lightlevel = (short)(fll + (255 - fll) * rend_lightcompress);
	}

	if (lightlevel < (short)r_ambient)
		ll = (short)r_ambient;
	else
		ll = lightlevel;

	if (ll <   0) ll = 0;
	if (ll > 255) ll = 255;

	if (cfg.sineCompress)
		return SectorDoomLight(ll);
	else
		return (short)(ll);
}

//===========================================================================
// rend_UpdateSide
//===========================================================================
void rend_UpdateSide(side_t *ls, short pos, int flatn)
{
	fixed_t	hi, lo;
	short	pflag = pos & LS_TEXPRESENT;
	RDBOOL	setlsiarray, masked = false;

	quad->flags |= RPF_QUADSET;	// This must be set

	if (quad->top <= quad->bottom)
	{
		ls->lsiflags &= ~pos;
		return;
	}

	if (rendtexflags & TXF_MASKED)
	{
		masked = true;

		if (whitefog || rendtexflags & TXF_NOALPHACUT)	// (TXF_ANIMPIC|TXF_NOALPHACUT)
		{
			quad->flags |= RPF_NOALPHACUT;
		}
	}

	{
		short	llev, usellev;
		int		secnum = rv.sect->heightsec;

		if (   secnum >= 0
			&& (( vssecunderfake && quad->bottom <= FIX2FLT(sectors[secnum].smoothfloor))
			||  (!vssecunderfake && quad->top    <= FIX2FLT(sectors[secnum].smoothfloor))) )
		{
			llev = sectors[secnum].lightlevel;
		}
		else
		{
		//	secnum = ls->sector - sectors;
			llev   = ls->sector->lightlevel;
		}

		usellev = R_SectorLight(llev);

		RL_VertexColors(quad->vertices, quad->numvertices, usellev, rendtexflags & TXF_MASKED);
	}


	if ((ls->lsiflags & pos) == pflag) return;	// Already set
	
	setlsiarray = !(ls->lsiflags & LS_TEXPRESENT);

	lo = (fixed_t)(quad->bottom * FRACUNIT);
	hi = (fixed_t)(quad->top * FRACUNIT);

	switch(pflag)
	{
	case LS_LOWTEX:
		ls->ll = lo;
		ls->lh = hi;
		ls->flatl = flatn;
		ls->lsiflags |= pos;
		if (masked) ls->lsiflags |= LS_LOWALPHA;
		break;

	case LS_MIDTEX:
		ls->ml = lo;
		ls->mh = hi;
		ls->lsiflags |= pos;
		if (masked)
		{
			ls->lsiflags |= LS_MIDALPHA;
			if (R3D_edit)
			{
				ls->editml = ctex->loalpha;
				ls->editmh = ctex->hialpha;
			}
		}
		break;

	case LS_MIDTEXH:
		ls->sl0 = lo;
		ls->sh0 = hi;
		ls->lsiflags |= pos;
		if (masked) ls->lsiflags |= LS_MIDALPHA;
		break;

	case LS_MIDTEXL:
		ls->sl1 = lo;
		ls->sh1 = hi;
		ls->lsiflags |= pos;
		if (masked) ls->lsiflags |= LS_MIDALPHA;
		break;

	case LS_HITEX:
		ls->hl = lo;
		ls->hh = hi;
		ls->lsiflags |= pos;
		ls->flath = flatn;
		if (masked) ls->lsiflags |= LS_HIALPHA;
		break;
	}

	if (setlsiarray && ls->lsiflags & LS_TEXPRESENT)
	{
		lsiarray[lsiarray_count++] = ls;
	}
}

static void ProjectVertices(seg_t *sg, fixed_t dis, int type)
{
	binangle_t angle = bamsAtan2(
		(sg->v2->y - sg->v1->y) >> FRACBITS, 
		(sg->v2->x - sg->v1->x) >> FRACBITS) >> (ANGLETOFINESHIFT - FRACBITS);
	float dx = FIX2FLT(FixedMul(dis,finesine  [angle]));
	float dy = FIX2FLT(FixedMul(dis,finecosine[angle]));

	switch(type)
	{
	case 0: break;									// towards the viewer
	case 1: dx = -dx, dy = -dy; break;				// away from the viewer
	case 2: dx =  FABS(dx), dy =  FABS(dy); break;	// fixed up 
	case 3: dx = -FABS(dx), dy = -FABS(dy); break;	// fixed down
	case 4: dx =  FABS(dx); break;					// fixed left 
	case 5: dx = -FABS(dx); break;					// fixed right
	case 6: dy =  FABS(dy); break;					// fixed up 
	case 7: dy = -FABS(dy); break;					// fixed down
	}
	// Apply the offsets
	quad->vertices[0].pos[VX] += dx;
	quad->vertices[0].pos[VY] -= dy;
	quad->vertices[1].pos[VX] += dx;
	quad->vertices[1].pos[VY] -= dy;
}

rendpolyq_t *Get_Quad(seg_t *sg)
{
	rendpolyq_t *ptr;

	if (plist.count >= MAXPOLYSIDES)
		Con_Error("[1] plist.count %i is >= to MAXPOLYSIDES %i", plist.count, MAXPOLYSIDES);

	plist.segptr[plist.count] = sg;

	ptr = &plist.list[plist.count++];

	ptr->type = rp_quad;
	ptr->flags = 0;
	ptr->numvertices = 2;
	ptr->length = sg->length;

	ptr->vertices[0].pos[VX] = sg->fv1[VX];
	ptr->vertices[0].pos[VY] = sg->fv1[VY];
	ptr->vertices[1].pos[VX] = sg->fv2[VX];
	ptr->vertices[1].pos[VY] = sg->fv2[VY];

	// rv.distance1/2 have to be set prior to calling.
	ptr->vertices[0].vpdist = rv.distance1;
	ptr->vertices[1].vpdist = rv.distance2;

	ptr->light = NULL;
	ptr->detail = NULL;
	ptr->divs[0].num = 0;
	ptr->divs[1].num = 0;
	ptr->tex = 0;
	ptr->texoffx = 0;
	ptr->texoffy = 0;

	return ptr;
}

static void setVISquad(line_t *ld, int texn, RDBOOL dir, seg_t *sg)
{
	if (plist.count >= MAXPOLYSIDES)
		Con_Error("[2] plist.count %i is >= to MAXPOLYSIDES %i", plist.count, MAXPOLYSIDES);

	plist.segptr[plist.count] = sg;

	quad = &plist.list[plist.count++];

	quad->type   = rp_quad;
	quad->flags  = 0;
	quad->length = M_AccurateLength(ld);

	quad->numvertices = 2;

	quad->light  = NULL;

	quad->divs[0].num = 0;
	quad->divs[1].num = 0;

	quad->vertices[!dir].pos[VX] = FIX2FLT(ld->v1->x);
	quad->vertices[!dir].pos[VY] = FIX2FLT(ld->v1->y);
	quad->vertices[ dir].pos[VX] = FIX2FLT(ld->v2->x);
	quad->vertices[ dir].pos[VY] = FIX2FLT(ld->v2->y);

	quad->vertices[0].vpdist = Rend_PointDist2D(quad->vertices[0].pos);
	quad->vertices[1].vpdist = Rend_PointDist2D(quad->vertices[1].pos);

	// set the texture parms
	quad->tex  = GL_PrepareTexture(texn);
	quad->texh = ctex->rendh;
	quad->texw = ctex->rendw;
	quad->detail = NULL;
}

//===========================================================================
// Rend_RenderWallSeg
//===========================================================================
static void Rend_RenderWallSeg(seg_t *seg)
{
	line_t			*ldef = seg->linedef;
	sector_t		*frontsec = seg->frontsector;
	sector_t		*backsec = (ldef->scpFlags & SCPF_IGNOREBS) ? NULL : seg->backsector;
	side_t			*sid = seg->sidedef;
	RDBOOL			topvis, midvis, botvis, skyflag;
	RDBOOL			fmovsec = (frontsec->specFlags & SF_MOVSECTOR) != 0;
	RDBOOL			bmovsec = backsec && (backsec->specFlags & SF_MOVSECTOR) != 0;
	RDBOOL			midtexmask = false;
	RDBOOL			midtexquadset = false;
	RDBOOL			transvalid;
	RDBOOL			isSloped = (ldef->lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL)) != 0;
	float			xoff, roff, segoff;
	float			topskywall = 32767.0f;
	float			basetoptex = 32767.0f;
	short			mtflags = BYTE(ldef->mtflags);
	short			midtex = sid->midtexture, targettex;
	byte			stype = ldef->specType;
	int				txn = origTEX[ldef - lines];
	int				flatnum;
	int				floortest = DDMININT;
	int				bsh;

	//---------------------------------------

	transvalid = false;
	quad = NULL;

	// Set heights
	if (frontsec->heightsec >= 0)
	{
		sector_t *tfsec = frontsec, tempfsec;

		if (water_trans)
		{
			if (rv.istrans)
			{
				if (frontsec->smoothfloor <= sectors[frontsec->heightsec].smoothfloor)
					transvalid = true;
			}
			else if (frontsec->specFlags2 & SF2_ADJOINSWATER)
				transvalid = true;
		}

		// v220-6/v220-31 code simplified
		if (!backsec)
		{
			rv.ffloor = MIN_OF(sectors[frontsec->heightsec].smoothfloor, frontsec->smoothfloor);
			rv.fceil = MAX_OF(sectors[frontsec->heightsec].smoothceil, frontsec->smoothceil);
		}
		else
		if (!vssecoverfake)
		{
			rv.fceil = sectors[frontsec->heightsec].smoothceil;
			rv.ffloor = MIN_OF(sectors[frontsec->heightsec].smoothfloor, frontsec->smoothfloor);
		}
		else
		{
			tfsec = R_FakeFlat(frontsec, &tempfsec, NULL, NULL);
			rv.ffloor = tfsec->smoothfloor;
			rv.fceil  = tfsec->smoothceil;
		}
	}
	else
	{
		rv.ffloor = frontsec->smoothfloor;
		if (water_trans) floortest = rv.ffloor;
		rv.fceil = frontsec->ceiloffset < DDMAXINT ?
				frontsec->smoothfloor + frontsec->ceiloffset : frontsec->smoothceil;
	}

	if (backsec)
	{
		if (backsec->heightsec >= 0 && !(sectors[backsec->heightsec].specFlags2 & SF2_PSEUDOFAKE))
		{
			if (   frontsec->heightsec >= 0
				&& sectors[backsec->heightsec].smoothfloor != sectors[frontsec->heightsec].smoothfloor)
			{
				sector_t tempbsec, *tbsec = R_FakeFlat(backsec, &tempbsec, NULL, NULL);

				if (vssecunderfake)
				{
					rv.bfloor = tbsec->smoothfloor;
				}
				else if (sectors[backsec->heightsec].smoothfloor > sectors[frontsec->heightsec].smoothfloor)
				{
					rv.bfloor = sectors[backsec->heightsec].smoothfloor;
				}
				else
				{
					rv.ffloor = sectors[frontsec->heightsec].smoothfloor;
					rv.bfloor = frontsec->smoothfloor;
				}
			}
			else
			{
				if (frontsec->heightsec >= 0)
					rv.bfloor = backsec->smoothfloor;
				else
					rv.bfloor = sectors[backsec->heightsec].smoothfloor;

				rv.bceil = sectors[backsec->heightsec].smoothceil;

				if (rv.bceil == rv.bfloor)
				{
					rv.ffloor = frontsec->smoothfloor;
					rv.bfloor = backsec->smoothfloor;
				}
			}

			if (backsec->specFlags2 & SF2_FAKEDHIGHMISS)	// v220-31 special case
				rv.bceil = sectors[backsec->heightsec].smoothceil;
			else if (vssecoverfake)
				rv.bceil = MAX_OF(backsec->smoothceil, sectors[backsec->heightsec].smoothceil);
			else
				rv.bceil = MIN_OF(backsec->smoothceil, sectors[backsec->heightsec].smoothceil);

			if (backsec->specFlags2 & SF2_FAKEDLOWMISS)	// v220-31 special case
			{
				rv.bfloor = sectors[backsec->heightsec].smoothfloor;
			}
			else
			// v220-4 the back floor may need resetting
			if (   frontsec->heightsec >= 0
				&& frontsec->smoothfloor == backsec->smoothfloor
				&& sectors[frontsec->heightsec].smoothfloor
				== sectors[backsec->heightsec].smoothfloor)
			{
				rv.ffloor = rv.bfloor = frontsec->smoothfloor;
			}
		}
		else
		{
			rv.bfloor = backsec->smoothfloor;
			rv.bceil = backsec->ceiloffset < DDMAXINT ?
				backsec->smoothfloor + backsec->ceiloffset : backsec->smoothceil;
		}

		if (ldef->ml_flags & (ML_IGNOREMISSINGHI|ML_IGNOREMISSINGLO))
		{
			if (ldef->ml_flags & ML_IGNOREMISSINGHI)
			{
				rv.csec  = rv.bceil  > rv.fceil  ? backsec : frontsec;
				rv.sceil = rv.csec->smoothceil;
			}
			if (   ldef->ml_flags & ML_IGNOREMISSINGLO
				&& !(sid->bottomtexture && rv.bfloor > rv.ffloor))
			{
				rv.fsec  = rv.bfloor < rv.ffloor ? backsec : frontsec;
				rv.sfloor= rv.fsec->smoothfloor;
			}
		}
		else if (stype == R3D_28)
		{
			if (fmovsec && rv.ffloor > rv.bfloor)
			{
				rv.fsec = frontsec;
				rv.sfloor = rv.ffloor;
				return;
			}
			if (bmovsec && rv.bfloor > rv.ffloor)
			{
				rv.fsec = backsec;
				rv.sfloor = rv.bfloor;
				return;
			}
		}

		// NOTE: R3D_14 only affects floors - use the other sector's floorlight level
		//       R3D_16 uses the existing sector's lightlevel
		//       R3D_20 uses R3D_16 logic but the other sector's floorlight level
		else if (stype == R3D_14
			&& frontsec->heightsec == backsec->heightsec	// i.e both unset or both set the same
			&& rv.ffloor != rv.bfloor
			&& !(sid->bottomtexture && rv.ffloor < rv.bfloor))
		{
			if (rv.ssec->lfsec >= 0)
			{
				rv.fsec = &sectors[rv.ssec->lfsec];
				rv.sfloor = rv.fsec->smoothfloor;
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
			}
			else if (!fmovsec && !bmovsec && frontsec->heightsec >= 0)
			{
				rv.sfloor = MAX_OF(frontsec->smoothfloor, backsec->smoothfloor);
				return;
			}
			else if (ldef->frontsector->floorpic	//rev1.7.00
				== ldef->backsector->floorpic
				&& IsLiquidFloorFlat(ldef->frontsector))
			{
				// v220-7 add more precise sector tex tests
				if (  !(frontsec->specFlags & SF_NOLOTEX)
					&& backsec->specFlags & SF_NOLOTEX)
				{
					rv.sfloor = rv.ffloor;
				}
				else
				if (  !(backsec->specFlags & SF_NOLOTEX)
					&& frontsec->specFlags & SF_NOLOTEX)
				{
					rv.sfloor = rv.bfloor;
				}
				// fall back to missing tex test
				else
				if (   rv.ffloor > rv.bfloor
					&& !fmovsec && bmovsec)
				{
					rv.sfloor = rv.ffloor;
				}
				else
				if (   rv.ffloor < rv.bfloor
					&& fmovsec && !bmovsec)
				{
					rv.sfloor = rv.bfloor;
				}
				// Check for changing sectors
				// E.g cchest2 map29 sec 272
				return;
			}
			else if (rv.bfloor > rv.ffloor)
			{
				if (viewz > rv.bfloor)
				{
					rv.fsec = backsec;
					rv.sfloor = rv.ffloor = rv.bfloor;
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
					rv.ceilinglightlevel = rv.floorlightlevel;
				}
			}
			else
			{
				if (viewz < rv.ffloor)
				{
					rv.fsec = backsec;
					rv.sfloor = rv.ffloor = rv.bfloor;
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				}
			}

			// v220-8 if ceilings match adopt the floor light level
			//        for the ceiling
			if (rv.fceil == rv.bceil)
			{
				if (frontsec->ceilingpic == backsec->ceilingpic)
						rv.ceilinglightlevel = rv.floorlightlevel;
				return;
			}
		}

		if (   stype != R3D_00
			&& ldef->frontsector != ldef->backsector
			&& subfn_MissingFirstPass(frontsec, backsec, sid, ldef, stype, mtflags))
		{
			return;
		}

		// v220-10
		if (   stype == R3D_1D
			&& !midtex
			&& (ldef->mtflags == 0x4B || ldef->mtflags == 0x1E)
			&& rv.fceil == rv.bceil
			&& !faked)
		{
			R_SetMissingParms(backsec, NULL);
			return;
		}

		// b211-0 rehacked
		// Check for missing wall being drawn/not drawn depending
		// on the position of a moving floor
		if (   (stype == R3D_1D || stype == R3D_16)
			&& (txn & 0x55) == 0x10
			&& sid->bottomtexture
			&& frontsec->specFlags & SF_MOVFLOOR
			&& !bmovsec)
		{
			if (rv.ffloor > rv.bfloor)
			{
				if (rv.lfsec < 0)
					rv.lfsec = backsec - sectors;
			}
			else if (stype == R3D_16)
				stype = R3D_00;
		}

		// Don't overlap midtex top with a raised flat
		if (   stype == R3D_16
			&& midtex
			&& rv.ffloor < rv.bfloor
			&& sid->rowoffset == rv.ffloor - rv.bfloor)
		{
			stype = R3D_03;
		}

		if (   !faked
			&& (stype == R3D_16 || stype == R3D_1D || stype == R3D_20)
			&& !(midtex && txn != 0x22 && (fmovsec || bmovsec)) )
		{
			if (   (stype == R3D_16
				&& ((txn & 0x22) != 0x22 || backsec->smoothfloor != frontsec->smoothfloor))
				|| (stype == R3D_1D
				&& ldef->lnflags2 & LNF2_OWNBOTHGTRTWO
				&& (txn == 0x22 || fmovsec || bmovsec))
				|| stype == R3D_20)
			{
				RDBOOL validhi = true, validlo = true;
				RDBOOL notexf = false, notexb = false;
				RDBOOL above_floor = false;

				if (   !txn
					&& rv.ffloor == rv.bfloor
					&& rv.fceil != rv.bceil
					&& (frontsec->floorpic == skyflatnum)^(backsec->floorpic == skyflatnum))
				{
					if (rv.fceil < rv.bceil && viewz < rv.fceil)
					{
						rv.csec = rv.usesec = frontsec;
						rv.sceil = rv.fceil;
						return;
					}
					if (rv.fceil > rv.bceil && viewz < rv.bceil)
					{
						rv.csec = rv.usesec = backsec;
						rv.sceil = rv.bceil;
						return;
					}
				}

				if (stype == R3D_16
					&& ((rv.fceil - rv.ffloor < 4*FRACUNIT)
					^   (rv.bceil - rv.bfloor < 4*FRACUNIT)) )
				{
					notexf = (frontsec->specFlags & 0x8C) == 0xC;
					notexb = (backsec->specFlags  & 0x8C) == 0xC;
				}

				if (notexf^notexb)
				{
					if (notexf)
					{
						rv.fsec = rv.csec = backsec;
					}
					else
					{
						rv.fsec = rv.csec = frontsec;
					}
					rv.sfloor = rv.fsec->smoothfloor;
					rv.sceil  = rv.fsec->smoothceil;
					rv.ceilinglightlevel = subfn_LLevel(rv.fsec, 0);
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				}
				else
				{
					if (   rv.ssec->ceilsec < 0
						// If floor height set in a movceil structure then draw the missing textures
						&& !(fmovsec
						&& frontsec->specFlags & SF_MOVCEIL
						&& rv.sect->R3Dfloorheight != DDMAXINT)
						&& rv.fceil != rv.bceil)
					{
						int ht = DDMININT;

						validhi = sid->toptexture && rv.fceil > rv.bceil;

						if (   frontsec->floorpic == skyflatnum
							&& backsec->floorpic  != skyflatnum)
						{
							ht = rv.bceil;
							rv.csec = backsec;
							rv.ceilinglightlevel = subfn_LLevel(backsec, 0);
						}
						else
						if (   frontsec->floorpic != skyflatnum
							&& backsec->floorpic  == skyflatnum)
						{
							ht = rv.fceil;
							rv.csec = frontsec;
							rv.ceilinglightlevel = subfn_LLevel(frontsec, 0);
						}
						else if (!validhi)
						{
							ht = MIN_OF(rv.fceil, rv.bceil);

							if (viewz > ht)
								ht = MAX_OF(rv.fceil, rv.bceil);

							if (frontsec->smoothceil == ht)
							{
								rv.csec = frontsec;
								rv.ceilinglightlevel = subfn_LLevel(frontsec, 1);
							}
							// Check if mid would extend above the backsec ceiling
							else if ( !(midtex && !fmovsec && !bmovsec
								&& rv.bceil - rv.bfloor <= sid->rowoffset) )
							{
								rv.csec = backsec;
								rv.ceilinglightlevel = subfn_LLevel(backsec, 1);
							}
						}

						if (ht != DDMININT)
						{
							if (midtex)
								rv.sceil = ht;
							else
								rv.sceil = rv.fceil = ht;
						}
					}

					if ( !(ldef->lnflags2 & LNF2_NORENDLOW)
						&& rv.ssec->floorsec < 0
						&& rv.bfloor != rv.ffloor)
					{
						int ht = DDMAXINT;

						validlo = sid->bottomtexture && rv.ffloor <= rv.bfloor;

						if (   frontsec->floorpic == skyflatnum
							&& backsec->floorpic  != skyflatnum)
						{
							ht = rv.bfloor;
							rv.fsec = backsec;
							rv.floorlightlevel = subfn_LLevel(backsec, 0);
						}
						else
						if (   frontsec->floorpic != skyflatnum
							&& backsec->floorpic  == skyflatnum)
						{
							ht = rv.ffloor;
							rv.fsec = frontsec;
							rv.floorlightlevel = subfn_LLevel(frontsec, 0);
						}
						else if (!validlo)
						{
							ht = MAX_OF(rv.ffloor, rv.bfloor);

							above_floor = viewz > ht;

							if (viewz < ht
								|| frontsec->floorpic == skyflatnum
								|| backsec->floorpic == skyflatnum)
							{
								ht = MIN_OF(rv.ffloor, rv.bfloor);
							}

							if (frontsec->smoothfloor == ht)
							{
								rv.fsec = frontsec;
								rv.floorlightlevel = subfn_LLevel(frontsec, 0);
							}
							else
							{
								rv.fsec = backsec;
								rv.floorlightlevel = subfn_LLevel(backsec, 0);

							}

							if (!subfn_LLevel(rv.fsec, 0))
							{
								rv.floorlightlevel = subfn_LLevel(rv.fsec == backsec ? frontsec : backsec, 0);
							}
						}

						if (ht != DDMAXINT)
						{
							if (midtex)
								rv.sfloor = ht;
							else
								rv.sfloor = rv.ffloor = ht;
						}
					}
				}

				// If viewheight elevated draw floor flat
				if (above_floor && rv.lfsec < 0)
				{
					if (rv.sfloor > frontsec->smoothfloor)
					{
						rv.lfsec = frontsec - sectors;
					}

					if (rv.lfsec < 0)
					{
						if (rv.ffloor > rv.bfloor)
						{
							if (rv.bfloor < rv.sfloor) rv.lfsec = backsec - sectors;
						}
						else
						{
							if (rv.ffloor < rv.sfloor) rv.lfsec = frontsec - sectors;
						}
					}
				}

				if (!midtex && !validhi && !validlo)
				{
					if (   viewz < rv.bfloor
						&& seg->sgflags & DDSEGF_FACING
						&& !fmovsec && !bmovsec
						&& ((flatnum = frontsec->floorpic) != skyflatnum))
					{
						quad = Get_Quad(seg);

						quad->tex = GL_PrepareFlat(flatnum);
						quad->texw = ctex->rendw;
						quad->texh = ctex->rendh;
						quad->detail = rendtexdetail;

						quad->top = FIX2FLT(backsec->smoothfloor);
						quad->bottom = FIX2FLT(frontsec->smoothfloor);
						quad->texoffx = 0.0f;
						quad->texoffy = 0.0f;
						R_WallHeightDivision(seg, WHD_BOTTOM);
						rend_UpdateSide(sid, LS_LOWTEX, flatnum);
					}

					return;
				}
			}
		}

		if (   ldef->ovflags & OVF_MISHITEXLINE
			&& stype == R3D_00
			&& rv.ssec->ceilsec < 0
			&& !faked)
		{
			if (rv.bceil < rv.fceil)
				R_SetMissingParms(NULL, backsec);
			else if (rv.bceil > rv.fceil)
				R_SetMissingParms(NULL, frontsec);
		}
	}

	// end actions prior to seg facing test ==================================

	if (!(seg->sgflags & DDSEGF_FACING)) return;

	// v220-16 modified
	if (ldef->lnflags2 & LNF2_SKYOCCLUDE)
	{
		// Add test v220-33
		if (   stype == R3D_00
			&& rv.ffloor == rv.bfloor
			&& !mtflags
			&& rv.ffloor < rv.bceil
			&& !midtex && !sid->bottomtexture
			&& !((frontsec->specFlags & SF_MOVFLOOR) ^ (backsec->specFlags & SF_MOVFLOOR)) )
		{
			return;
		}

		quad = Get_Quad(seg);

		if (rv.fceil > rv.ffloor)
		{
			quad->top = FIX2FLT(rv.fceil);
			quad->bottom = FIX2FLT(rv.ffloor);
		}
		else
		{
			quad->top = quad->bottom = 0.f;	// occlusion only
		}

		quad->flags = RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE;
		return;
	}

	// v220-9 add skymask/occlusion test for a movfloor sec
	if (   bmovsec
		&& stype == R3D_00
		&& backsec->specFlags & SF_MOVFLOOR	// added v220-35
		&& frontsec->ceilingpic == skyflatnum
		&& backsec->ceilingheight == backsec->floorheight
		&& !midtex
		&& !sid->toptexture
		&& !sid->bottomtexture
		&& (frontsec->ceilingheight - backsec->ceilingheight < MAX_SECHEIGHT)	// added v220-13
		&& frontsec->ceilingheight > backsec->ceilingheight)
	{
		ldef->ovflags &= ~OVF_NOSKYWALL;	// in case it was set
		quad = Get_Quad(seg);
		quad->flags = RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE;
		quad->top = FIX2FLT(frontsec->ceilingheight);
		quad->bottom = FIX2FLT(frontsec->floorheight);
		return;
	}

	// v220-5 test added
	if (!txn && stype == R3D_16 && (seg->frontsector->specFlags & 0xC) == 0xC)
	{
		return;
	}

	// Note: none of the next four values are used by a skymask quad
	// rv.distance1 and rv.distance2 are assigned to a quad on calling Get_Quad().
	rv.distance1 = Rend_PointDist2D(seg->fv1);
	rv.distance2 = Rend_PointDist2D(seg->fv2);
	segoff = FIX2FLT(seg->offset);
	xoff   = FIX2FLT(sid->textureoffset);

	if (stype == R3D_21)
		roff = FIX2FLT(sid->rowoffset);
	else
		roff = FIX2FLT(sid->rowoffset + sid->rowoffadj);

	//	$slopes - note not pre-calculated in case sectors are moving types
	if (frontsec->stptr >= 0 && backsec && !isSloped)
	{
		if (backsec->stptr >= 0 && txn == 0x02)
		{
			rendSlopesInfill(seg, sid, midtex, xoff, roff, segoff);					
			return;	// nothing else to render on this line
		}

		if (backsec->stptr < 0 && sid->lsiflags & (LS_LOSLOPEB|LS_HISLOPEB))
		{
			rendSlopesInfill2(seg, sid, midtex, xoff, roff, segoff);
			return;	// nothing else to render on this line
		}
	}

	// v220-10 more weird doom behaviour hacks
	if (stype == R3D_02)
	{
		if (midtex)
		{
			quad = Get_Quad(seg);
			quad->tex = GL_PrepareTexture(midtex);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			if (ldef->ml_flags & ML_DONTPEGBOTTOM)
				quad->top = FIX2FLT(backsec->floorheight);
			else
				quad->top = FIX2FLT(frontsec->ceilingheight);

			quad->bottom = quad->top - ctex->rendh;
			quad->texoffx = xoff;
			quad->texoffy = roff;

			if (rendtexmask) quad->flags |= RPF_MASKED;
			rend_UpdateSide(sid, LS_MIDTEX, midtex);
		}
		return;
	}

	// v220-10 deal with horrible map hack.
	if (ldef->scpFlags & SCPF_DRAWTOLIMITS)
	{
		fixed_t a, b;

		if (  (a = rv.ssec->ceilsec  >= 0 ? sectors[rv.ssec->ceilsec ].ceilingheight : frontsec->ceilingheight)
			> (b = rv.ssec->floorsec >= 0 ? sectors[rv.ssec->floorsec].floorheight   : backsec->floorheight))
		{
			quad = Get_Quad(seg);
			quad->top = FIX2FLT(a);
			quad->bottom = FIX2FLT(b);

			if (midtex)
			{
				quad->tex = GL_PrepareTexture(midtex);
			}
			else
			{
				if (rv.sect->floorpic == skyflatnum)
				{
					quad->flags = RPF_SKY_MASK|RPF_QUADSET;
					return;
				}

				// v220-13 If a toptexture exists then use it
				if (sid->toptexture)
					quad->tex = GL_PrepareTexture(sid->toptexture);
				else
					quad->tex = GL_PrepareFlat(rv.sect->floorpic);
			}

			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->texoffx = xoff;
			quad->texoffy = roff;
			rend_UpdateSide(sid, LS_MIDTEX, midtex);
			return;
		}
	}

	if (   frontsec->floorpic == skyflatnum
		&& rv.setfloorheight < DDMAXINT)
	{
		int ht = rv.sect->heightsec >= 0 ? sectors[rv.sect->heightsec].smoothfloor : rv.sect->smoothfloor;

		if (ht > rv.setfloorheight)
		{
			quad = Get_Quad(seg);
			quad->flags = RPF_SKY_MASKLO|RPF_QUADSET;
			quad->top = FIX2FLT(ht);
			quad->bottom = FIX2FLT(rv.setfloorheight);
		}
	}

	// b211-0 Added
	if (stype == R3D_09)
	{
		flatnum = frontsec->floorpic;

		if (rv.bfloor > rv.ffloor && flatnum != skyflatnum)
		{
			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->top = FIX2FLT(rv.bfloor);
			quad->bottom = FIX2FLT(rv.ffloor);
			quad->texoffx = segoff + xoff;
			quad->texoffy = roff;

			R_WallHeightDivision(seg, WHD_BOTTOM);
			rend_UpdateSide(sid, LS_LOWTEX, flatnum);
		}
		return;
	}

	// b211-0 Added
	if (stype == R3D_07 && rv.sect->R3Dceilheight != DDMININT)
	{
		quad = Get_Quad(seg);

		rv.sceil  = rv.sect->R3Dceilheight;
		rv.sfloor = MIN_OF(rv.ffloor, rv.bfloor);

		quad->tex = GL_PrepareTexture(sid->toptexture);
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;

		quad->top = FIX2FLT(rv.sceil);
		quad->bottom = FIX2FLT(rv.sfloor);
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;
		rend_UpdateSide(sid, LS_HITEX, 0);
		return;
	}

	// Missing texture lines 2
	if (stype == R3D_2C)
	{
		// Type bypassing skyfix - plus never occlude
		if ((frontsec == ldef->frontsector &&
			 mtflags == R3D_MISHTEX1 &&
			 ldef->backsector->ceilingpic == skyflatnum) ||
			 (frontsec == ldef->backsector &&
			 mtflags == R3D_MISHTEX2 &&
			 ldef->frontsector->ceilingpic == skyflatnum))
		{
			quad = Get_Quad(seg);
			quad->flags = (RPF_SKY_MASK|RPF_QUADSET);
			quad->top = FIX2FLT(rv.fceil);
			quad->bottom = FIX2FLT(MAX_OF(rv.bceil, rv.ffloor));
		}
		return;
	}

	// floor height change types
	if (stype == R3D_2A) //plus case R3D_2B:$consolidated
	{
		if (faked)
		{
			if (frontsec == ldef->frontsector &&
				mtflags & (R3D_MISHTEX2|R3D_MISLTEX2) &&
				ldef->backsector->heightsec >= 0)
			{
				rv.sfloor = sectors[ldef->backsector->heightsec].smoothfloor;
				rv.sceil  = sectors[ldef->backsector->heightsec].smoothceil;
				return;
			}
			if (frontsec == ldef->backsector &&
				mtflags & (R3D_MISHTEX1|R3D_MISLTEX1) &&
				ldef->frontsector->heightsec >= 0)
			{
				rv.sfloor = sectors[ldef->frontsector->heightsec].smoothfloor;
				rv.sceil  = sectors[ldef->frontsector->heightsec].smoothceil;
				return;
			}
		}
		return;
	}

	if (stype == R3D_19)	// indicates roff may need changing 
	{
		if (!faked && frontsec == ldef->frontsector)
		{
			int t;
			t = rv.bfloor, rv.bfloor = rv.ffloor, rv.ffloor = t;
			t = rv.bceil , rv.bceil  = rv.fceil , rv.fceil  = t;
			midtex = sides[ldef->sidenum[1]].midtexture;
			roff = FIX2FLT(sides[ldef->sidenum[1]].rowoffset);
		}
	}

	// v220-7 skywall
	if (  !txn
		&& backsec
		&& rv.bfloor == rv.bceil
		&& rv.bfloor <= rv.ffloor
		&& frontsec->ceilingpic != skyflatnum
		&& backsec->ceilingpic == skyflatnum)
	{
		quad = Get_Quad(seg);
		quad->bottom = FIX2FLT(rv.ffloor);
		rv.ceil2 = rv.fceil + FRACUNIT;	// set ceil2 else the skymask won't work
		quad->top = FIX2FLT(rv.ceil2);
		quad->flags |= (RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE);
		rv.inhibit = false;
		return;
	}

	if (!midtex && (rv.ffloor >= rv.fceil || !ldef->backsector))
	{
		if (!backsec)
		{
			if (   frontsec->skyfix > 0
				&& !(frontsec->specFlags & SF_NOOCCLUDE)
				&& !(ldef->ovflags & OVF_NOSKYWALL) )
			{
				quad = Get_Quad(seg);
				quad->bottom = FIX2FLT(rv.ffloor);
				quad->top    = quad->bottom + FIX2FLT(frontsec->skyfix);
				quad->flags |= (RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE);
				return;
			}
		}
	}

	// The middle texture, single sided.========================================
	if (!backsec)
	{
		fixed_t		fh, ch, ht;
		fixed_t		sf;

		// v220-7 check no midtex lines. These have usually been
		//        occluded. If so they will never be processed
		if (!midtex)
		{
			if (frontsec->skyfix)
			{
				quad = Get_Quad(seg);
				quad->bottom = FIX2FLT(rv.ffloor);
				quad->top = FIX2FLT(rv.fceil) + FIX2FLT(frontsec->skyfix);
				if (ldef->backsector == frontsec)	// v220-35 special case
					quad->flags |= (RPF_SKY_MASK|RPF_QUADSET);
				else
					quad->flags |= (RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE);
			}

			return;
		}
		
		if (   rv.sect->ceilingpic == skyflatnum
			&& rv.ssec->ceilsec >= 0)
		{
			sf = sectors[rv.ssec->ceilsec].smoothceil - rv.sect->smoothceil;
		}
		else
		// Added b220-1
		if (   fmovsec
			&& frontsec->specFlags & SF_MOVCEIL
			&& rv.sect->ceilingpic == skyflatnum
			&& rv.ssec->ceilsec >= 0
			&& (ch = sectors[rv.ssec->ceilsec].smoothceil)
			>  rv.sect->smoothceil)
		{
			sf = rv.skyfix + ch - rv.sect->smoothceil;
		}
		else
			sf = rv.skyfix;
		
		if (rv.skyfix > sf) sf = rv.skyfix;

		quad = Get_Quad(seg);
		quad->tex = GL_PrepareTexture(midtex);	// Note: must be set if no alpha type required

		if (rendtexmask)
			quad->tex = GL_PrepareTextureNoAlpha(midtex);

		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = xoff*ctex->texscalex + segoff;
		quad->texoffy = roff*ctex->texscaley;
		ht = frontsec->smoothceil - frontsec->smoothfloor;	// ignore fake

		// v220-33 revised slopes handling
		if (sid->lsiflags & (LS_LOSLOPEB|LS_HISLOPEB))
		{	// ignore faked
			if (sid->lsiflags & LS_HISLOPEB)
			{
				ch = SL_LowCeilHeight(frontsec);
			}
			else
			{
				ch = frontsec->smoothceil;
			}

			if (sid->lsiflags & LS_LOSLOPEB)
			{
				fh = SL_HighFloorHeight(frontsec);
			}
			else
			{
				fh = frontsec->smoothfloor;
			}

			if (ldef->ml_flags & ML_DONTPEGBOTTOM)
				quad->texoffy += (ctex->rendh - FIX2FLT(ht));
		}
		else
		{
			// v220-33 rendering fix for map bodges
			if (rv.ssec->ceilsec >= 0 && rv.ssec->ceilsec == rv.ssec->floorsec)
			{
				fh = sectors[rv.ssec->floorsec].smoothfloor;
				ch = sectors[rv.ssec->floorsec].smoothceil;

				if (ht != (ch - fh)) quad->texoffy += FIX2FLT(ht - (ch - fh));
			}
			else	// v220-31 added
			if (ldef->adjacentsec >= 0 && rv.sect->heightsec < 0)
			{
				fh = sectors[ldef->adjacentsec].smoothfloor;
				ch = sectors[ldef->adjacentsec].smoothceil;

				if (   rv.ssec->lcsec < 0	
					&& sectors[ldef->adjacentsec].ceilingpic != skyflatnum)
				{
					rv.ssec->lcsec = ldef->adjacentsec;
				}
			}
			else
			{
				fh = rv.ffloor, ch = rv.fceil;

				if (ldef->ml_flags & ML_DONTPEGBOTTOM)
				{
					quad->texoffy += (ctex->rendh - FIX2FLT(rv.fceil - rv.ffloor));
					// 1) Check for a singles side wall in a mov sec
					// 2) If a high-sky sector then skywall test must be done
					if (fh == ch && rv.common >= 0)
					{
						fh = sectors[rv.common].smoothfloor;
						ch = sectors[rv.common].smoothceil;
					}

					if (frontsec->heightsec < 0)	// v220-35 stop scrolling in faked sector
						quad->texoffy += FIX2FLT(ht - (ch - fh));
				}	
			}
		}

		quad->top    = FIX2FLT(ch);
		quad->bottom = FIX2FLT(fh);

		// Check for neighbourhood division.
		R_WallHeightDivision(seg, WHD_MIDDLE);
		rend_UpdateSide(sid, LS_MIDTEX, 0);


		if ( !(ldef->ovflags & OVF_NOSKYWALL) )
		{
			if (sf > 0 && !(frontsec->specFlags & SF_NOOCCLUDE))
			{
				quad = Get_Quad(seg);
				quad->bottom = FIX2FLT(ch);
				quad->top    = quad->bottom + FIX2FLT(sf);
				quad->flags  = (RPF_SKY_MASK|RPF_QUADSET);
			}

			quad->flags |= RPF_OCCLUDE;
		}

		return;
	}

	skyflag = frontsec->ceilingpic == skyflatnum && backsec->ceilingpic == skyflatnum;

	// v220-6 Is this just a skywall?
	if (   skyflag
		&& !midtex
		&& rv.bfloor == rv.bceil
		&& rv.ffloor >= rv.bfloor
		&& backsec->skyfix)
	{
		quad = Get_Quad(seg);
		quad->flags |= RPF_SKY_MASK|RPF_OCCLUDE|RPF_QUADSET;
		quad->bottom = FIX2FLT(rv.ffloor);
		quad->top = FIX2FLT(rv.fceil + backsec->skyfix);
		return;
	}

	if (ldef->tyflags & TYF_SKYWALL || (!R3D_ScriptedMap && !txn && skyflag))	// v220-35 special cases
	{
		if (rv.fceil > rv.bceil)
		{
			quad = Get_Quad(seg);
			quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
			quad->bottom = FIX2FLT(rv.bceil);
			quad->top = FIX2FLT(rv.fceil + rv.sect->skyfix);
		}

		if (!sid->bottomtexture) return;
	}
	else if (!(ldef->ovflags & OVF_NOSKYWALL))
	{
		// v220-6 draw skywall and a midtexture (should be else sector has no point)
		//  with a heightless backsector e.g. Eternal III map29 sec 164
		if (   backsec
			&& stype == R3D_00
			&& rv.bfloor >= rv.bceil
			&& backsec->skyfix > 0
			&& !sid->toptexture
			&& rv.fceil > rv.bceil
			&& rv.bfloor < rv.fceil)
		{
			float	ffloor = FIX2FLT(rv.ffloor), offset = ffloor;
			float	skyf = FIX2FLT(backsec->skyfix);
			float	texh = 0.0f;

			if (sid->bottomtexture && rv.bfloor > rv.ffloor)
			{
				quad = Get_Quad(seg);

				quad->tex = GL_PrepareTexture(sid->bottomtexture);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;

				quad->bottom= offset;
				quad->top = quad->bottom + FIX2FLT(rv.bfloor - rv.ffloor);

				offset = quad->top;
				texh = ctex->rendh;

				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy = roff*ctex->texscaley;

				R_WallHeightDivision(seg, WHD_BOTTOM);
				rend_UpdateSide(sid, LS_LOWTEX, 0);
			}

			if (midtex)
			{
				// v220-33 Prevent drawing midtex if no backspace
				if (rv.ffloor < rv.bfloor)
				{
					quad = Get_Quad(seg);
					quad->tex = GL_PrepareTexture(midtex);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;

					quad->bottom = offset;
					quad->top = quad->bottom + ctex->rendh;

					offset = quad->top;
					texh = ctex->rendh;

					quad->texoffx = segoff + xoff*ctex->texscalex;
					quad->texoffy = roff*ctex->texscaley;

					R_WallHeightDivision(seg, WHD_MIDDLE);
					rend_UpdateSide(sid, LS_MIDTEX, 0);
				}
			}

			if (offset - ffloor < skyf)
			{
				quad = Get_Quad(seg);
				quad->bottom = offset;
				quad->top    = quad->bottom + skyf - (offset - ffloor);
				//	Note: occlusion's not possible if textures are drawn.
				//  This is why some  ports, even if using software rendering, do
				//  not display the mid texture through wrongly applying occlusion.
				quad->flags |= (RPF_SKY_MASK|RPF_QUADSET);
			}
			
			return;
		}

		if (   seg->sgflags & R3DSEGF_CHKSKYFIX	// Added rev170f
			&& !midtex
			&& backsec							// Insurance (should have)
			&& rv.bceil <= rv.fceil)
		{
			int skytop = rv.fceil + rv.skyfix;
			// v220-6 cover case where floor is higher than ceiling
			int skybottom = MAX_OF(rv.bceil + backsec->skyfix, rv.bfloor);

			if (skytop > skybottom)
			{
				quad = Get_Quad(seg);
				quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
				if (rv.bceil <= rv.ffloor) quad->flags |= RPF_OCCLUDE;
				quad->bottom = FIX2FLT(skybottom);
				quad->top = FIX2FLT(skytop);
			}
		}

		if (fmovsec || !(frontsec->specFlags & SF_NOOCCLUDE) )
		{
			int skytop = 0, skybottom = 0, flags = RPF_SKY_MASK|RPF_QUADSET;

			if (   rv.skyfix
				&& (!backsec
				|| (backsec->ceilingpic != skyflatnum && rv.bceil <= rv.fceil)))
			{
				skybottom = rv.fceil, skytop = skybottom + rv.skyfix;

				if (!backsec && !midtex && !(frontsec->specFlags & SF_NOOCCLUDE))
				{
					flags |= RPF_OCCLUDE;
				}
			}
			else if (!faked && backsec)
			{
				int		bsf = backsec->skyfix;
				RDBOOL	skywc = false;

				// b220-1 skyfix with heightless 'trick' sectors
				if (   rv.bfloor == rv.bceil
					&& backsec->skyfix
					&& !midtex
					&& backsec->ceilingpic == frontsec->ceilingpic
					&& backsec != frontsec
					&& sid->bottomtexture
					&& !fmovsec
					&& !bmovsec)
				{
					bsf = 0;
					skywc = true;
				}

				if (rv.skyfix)
				{
					if (   backsec != frontsec
						&& !(backsec->specFlags & SF_NOOCCLUDE)
						&& (rv.bceil + bsf <  rv.fceil + frontsec->skyfix))
					{
						skybottom = rv.fceil, skytop = skybottom + rv.skyfix;
					}
				}
				else if (ldef->tyflags & TYF_SKYWALLCANDIDATE
					&& frontsec == ldef->frontsector
					&& rv.bceil + backsec->skyfix < rv.fceil)
				{
					skybottom = rv.bceil + rv.skyfix, skytop = rv.fceil;

					if (!(frontsec->specFlags & SF_NOOCCLUDE))
						flags |= RPF_OCCLUDE;
				}
				else if (skywc && rv.bceil + bsf < rv.fceil)
				{
					skybottom = rv.bceil, skytop = rv.fceil;

					if (!(frontsec->specFlags & SF_NOOCCLUDE))
						flags |= RPF_OCCLUDE;
				}
			}

			if (skytop > skybottom)
			{
				quad = Get_Quad(seg);
				quad->flags = flags;
				quad->top = FIX2FLT(skytop);
				quad->bottom = FIX2FLT(skybottom);
				if (midtex) topskywall = quad->bottom;
			}
		}
	}

	// Determine which parts of the segment are visible.
	bsh = rv.bceil - rv.bfloor;
	midvis = midtex && bsh > 0;
	topvis = rv.bceil  < rv.fceil;
	botvis = rv.bfloor > rv.ffloor;

	// Check for occlusion
	// Missing top or bottom textures don't occlude visibility.
	if (   quad
		&& quad->flags & RPF_QUADSET
		&& !(quad->flags & RPF_OCCLUDE) 
		&& !faked
		&& !(midtex && topvis && !sid->toptexture)
		&& !(midtex && botvis && !sid->bottomtexture)
		&& !(rv.ssec->ssflags & SSF_NOOCCLUDE)
		&& !(ldef->mtflags & R3D_MISTEXANY)
		&& !(frontsec->specFlags & (SF_MASKEDDOOR|SF_NOOCCLUDE))
		&& !(backsec->specFlags  & (SF_MASKEDDOOR|SF_NOOCCLUDE|SF_LINESCLOSED)) )
	{
		// Add skyfix b220-1
		fixed_t fc = frontsec->smoothceil + frontsec->skyfix;
		fixed_t bc = backsec->smoothceil  + backsec->skyfix;
		fixed_t ff = frontsec->smoothfloor;
		fixed_t bf = backsec->smoothfloor;

		if (   (bc != fc || ff != bf)
			&& (bsh <= 0 || bc <= ff || bf >= fc))
		{
			// The backsector is occluded.
			quad->flags |= RPF_OCCLUDE;
		}
	}

	// With a back sector check for upper and lower walls.
	{
		float	midtextop = -32767.0f;
		float	midtexbottom = 32767.0f;

		RDBOOL	mid_covers_top = false;
		RDBOOL	btextest, transwall = false;
		RDBOOL	mid_faked = frontsec->heightsec >= 0 || backsec->heightsec >= 0;
		RDBOOL	mid_direct = false;

		// v220-3 Set fog wall
		if (   outsidefog
			&& !setNoFog
			&& (!midtex || isSloped)
			&& linefog_group[ldef - lines] != 0xFFFF)
		{
			float	oT = FIX2FLT(MIN_OF(rv.bceil , rv.fceil ));
			float	oB = FIX2FLT(MAX_OF(rv.bfloor, rv.ffloor));

			if (oB < oT)
			{
				quad = Get_Quad(seg);
				quad->top    = oT;
				quad->bottom = oB;
				quad->flags  = (RPF_QUADSET|RPF_FOGWALL);
			}
		}

		//	$slopes
		if (isSloped && midtex)
		{
			// ignore faked
			fixed_t	ffloor = seg->frontsector->smoothfloor;
			fixed_t bfloor = seg->backsector->smoothfloor;
			fixed_t fceil  = seg->frontsector->smoothceil;
			fixed_t bceil  = seg->backsector->smoothceil;
			RDBOOL	himissing  = false;
			RDBOOL	lomissing  = false;

			if (ldef->lnflags2 & LNF2_HITRIWALL && seg->backsector->stptr >= 0)
			{
				float	zh0 = ssecvtxz[seg->zptr1].z[1];
				float	zh1 = ssecvtxz[seg->zptr2].z[1];

				if (zh0 != zh1)
				{
					quad = Get_Quad(seg);
					quad->tex = GL_PrepareTexture(midtex);
					quad->texw   = ctex->rendw;
					quad->texh   = ctex->rendh;
					quad->detail = rendtexdetail;
					quad->texoffx = segoff + xoff*ctex->texscalex;
					quad->texoffy = roff*ctex->texscaley + MIN_OF(zh0, zh1);

					// Adjust if upper unpegged flag is unset
					if (!(ldef->ml_flags & ML_DONTPEGTOP))
						quad->texoffy += FIX2FLT(fceil - bceil) - MAX_OF(zh0, zh1);

					quad->vertices[0].pos[VZ] = zh0;
					quad->vertices[1].pos[VZ] = zh1;
					quad->top    = FIX2FLT(bceil);
					// This sets the triangle's height (note: the z values are negative)
					quad->bottom = quad->top + MIN_OF(zh0, zh1);
					quad->flags = RPF_TRIHI;

					rend_UpdateSide(sid, (byte)(LS_MIDTEXH), 0);
					himissing = !sid->toptexture;
				}
			}

			if (ldef->lnflags2 & LNF2_LOWTRIWALL && seg->backsector->stptr >= 0)
			{
				float	zl0 = ssecvtxz[seg->zptr1].z[0];
				float	zl1 = ssecvtxz[seg->zptr2].z[0];

				if (zl0 != zl1)
				{
					quad = Get_Quad(seg);
					quad->tex = GL_PrepareTexture(midtex);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;
					quad->texoffx = segoff + xoff*ctex->texscalex;
					quad->texoffy = roff*ctex->texscaley - MAX_OF(zl0, zl1);

					// Adjust if lower unpegged flag is set
					if (ldef->ml_flags & ML_DONTPEGBOTTOM)
						quad->texoffy += FIX2FLT(fceil - bfloor);

					quad->vertices[0].pos[VZ] = zl0;
					quad->vertices[1].pos[VZ] = zl1;
					quad->bottom = FIX2FLT(bfloor);
					// This sets the triangle's height
					quad->top = quad->bottom + MAX_OF(zl0, zl1);
					quad->flags = RPF_TRILOW;

					rend_UpdateSide(sid, (byte)(LS_MIDTEXL), 0);
					lomissing = !sid->bottomtexture;
				}
			}

			if ((sid->toptexture || himissing) && fceil > bceil)
			{
				quad = Get_Quad(seg);
				quad->tex = GL_PrepareTexture(himissing ? sid->midtexture : sid->toptexture);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;
				quad->bottom = FIX2FLT(bceil);
				quad->top = FIX2FLT(fceil);
				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy = roff*ctex->texscaley;
				if (ldef->ml_flags & ML_DONTPEGTOP)
					quad->texoffy += (ctex->rendh - FIX2FLT(fceil - bceil));	// Normal alignment to bottom

				R_WallHeightDivision(seg, WHD_TOP);
				rend_UpdateSide(sid, LS_HITEX, 0);
			}

			if ((sid->bottomtexture || lomissing) && rv.bfloor > rv.ffloor)
			{
				quad = Get_Quad(seg);
				quad->tex = GL_PrepareTexture(lomissing ? sid->midtexture : sid->bottomtexture);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;
				quad->bottom = FIX2FLT(ffloor);
				quad->top = FIX2FLT(bfloor);
				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy = roff*ctex->texscaley;
				if (ldef->ml_flags & ML_DONTPEGBOTTOM)
					quad->texoffy += FIX2FLT(fceil - bfloor);

				R_WallHeightDivision(seg, WHD_BOTTOM);
				rend_UpdateSide(sid, LS_LOWTEX, 0);
			}
			return;
		}

		if (stype == R3D_3E)
		{
			flatnum = rv.ffloor > rv.bfloor ?
				frontsec->floorpic : backsec->floorpic;

			if (flatnum == skyflatnum) return;

			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->top = FIX2FLT(MAX_OF(rv.ffloor, rv.bfloor));
			quad->bottom = FIX2FLT(MIN_OF(rv.ffloor, rv.bfloor));

			if (rendtexmask)
			{
				quad->texoffx = segoff;
				quad->texoffy = 0.0f;
			}
			else
			{
				quad->texoffx = segoff + xoff;
				quad->texoffy = roff;
			}

			rend_UpdateSide(sid, LS_HITEX, flatnum);
			return;
		}

		btextest =	rv.bfloor > rv.ffloor;

		if (midtex)
		{
			float	mceil, mfloor, lowtexceil;
			float	openingTop    = FIX2FLT(MIN_OF(rv.bceil, rv.fceil));
			float	openingBottom = FIX2FLT(MAX_OF(rv.bfloor, rv.ffloor));
			float	mroff;
			float	d;
			DGLuint	tn = 0;

			// v220-35 test added
			if (skyflag && !faked && openingTop <= openingBottom) goto upperwall;
			
			// v220-33 test added
			if (   !rv.pseudo
				&& sid->bottomtexture
				&& ldef->ml_flags & ML_DONTPEGBOTTOM
				&& frontsec->heightsec >= 0
				&& backsec->heightsec >= 0
				&& sectors[backsec->heightsec].floorheight < sectors[backsec->heightsec].ceilingheight)
			{
				rv.bfloor  = sectors[backsec->heightsec].floorheight;
				rv.bceil = sectors[backsec->heightsec].ceilingheight;
				openingTop = FIX2FLT(MIN_OF(rv.bceil, rv.fceil));
				openingBottom = FIX2FLT(MAX_OF(rv.bfloor, rv.ffloor));
			}

			tn = GL_PrepareTexture(midtex);

			if (ctex->high < 0)	// v220-33
				mroff = -ctex->high + ctex->offset + roff;
			else if (ctex->high2 && !(ldef->ml_flags & ML_DONTPEGBOTTOM))	// v220-33
				mroff = ctex->high2 + ctex->offset + roff;
			else if (ctex->high > 0 && !(ldef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP)))	// v220-33
				mroff = ctex->high + ctex->offset + roff;
			else
				mroff = roff + ctex->offset;

			// v220-16 Prevent drawing midtex if no backspace
			if ( !(ldef->ml_flags & ML_DONTPEGBOTTOM)
				&& rv.bceil <= rv.bfloor
				&& !(stype == R3D_10 && rv.ssec->floorsec >= 0)
				&& -mroff > ctex->rendh)
			{
				goto upperwall;
			}
			
			if (   openingTop - openingBottom <= 0
				&& frontsec != backsec	// v220-31
				&& !rendtexmask
				&& !mid_faked
				&& !(ldef->spflags & SPF_MIDTRANSLUCENT)
				&& stype != R3D_2D) // rev170s
			{
				goto upperwall;
			}

			// v220-35 midtex below heightsec floor
			if (ldef->ml_flags & ML_DONTPEGBOTTOM && ctex->rendh <= -roff)	// could be below the floor
			{
				if (fmovsec || bmovsec)	// assume sector moves else what's the point?
				{
					// One side at least must have a heightsec def
					int hsec = frontsec->heightsec >=0 ? frontsec->heightsec : backsec->heightsec >= 0 ? backsec->heightsec : -1;

					if (hsec >= 0)
					{
						// One side at least must have a sector that can change height
						sector_t *nsec = fmovsec ? frontsec : backsec;

						// If the moving sec's floorheight is <= to the heightsec's floorheight
						// then the midtex will be below the floor to be rendered
						if (nsec->floorheight <= sectors[hsec].floorheight)
							goto upperwall;
					}
				}
			}

			quad = Get_Quad(seg);
			quad->tex = tn;
			quad->texw    = ctex->rendw;
			quad->texh    = ctex->rendh;
			quad->detail  = rendtexdetail;
			midtexquadset = true;
			midtexmask = rendtexmask;

			if (sid->lsiflags & LS_ORIGTEXOFF)
				quad->texoffx = ((float)sid->orig_textureoffset)*ctex->texscalex;
			else
				quad->texoffx = xoff*ctex->texscalex;

			quad->texoffy = 0.0f;
			
			lowtexceil = FIX2FLT(rv.bfloor > rv.fceil ? rv.fceil : rv.bfloor);

			switch(stype)
			{
			// R3D_04, R3D_05, R3D_06, R3D_0E, R3D_3D, R3D_3E $consolidated
			case R3D_03:
				if (sid->bottomtexture) goto setas_default;	// v220-33
				if (rv.ssec->ssflags & SSF_LOWFLOORSEC)
				{
					if (ldef->ml_flags & ML_DONTPEGBOTTOM)
					{
						mfloor = FIX2FLT(backsec->floorheight) + mroff;
						mceil = mfloor + ctex->rendh;
					}
					else
					if (mroff <= 0.f && rv.sect->R3Dlowsec >= 0)
					{
						mfloor = FIX2FLT(sectors[rv.sect->R3Dlowsec].floorheight) - mroff;
						mceil  = mfloor + ctex->rendh;
					}
					else
					{
						mceil = FIX2FLT(backsec->ceilingheight) + mroff;
						mfloor = mceil - ctex->rendh;
					}

					if (sid->toptexture && mceil > openingTop)
					{
						d = mceil - openingTop;

						if (d < ctex->rendh)
						{
							mceil -= d;
							quad->texoffy += d;
						}
					}

					mid_direct = true;
					break;
				}
				// fall through
			case R3D_16:
				if (ldef->ml_flags & ML_DONTPEGBOTTOM)
				{
					mfloor = mid_faked ?
						FIX2FLT(MAX_OF(frontsec->smoothfloor,backsec->smoothfloor)) : openingBottom;

					if (rv.bfloor != rv.ffloor && !rendtexmask)
					{
						float h = ctex->rendh;

						if (-mroff > h)
						{
							// do nothing;
						}
						else if (mroff < 0 && -mroff < h)	// v220-6 test added
						{
							stype = R3D_00;
							goto setas_default;
						}
						else
							mfloor += mroff;
					}
					else
						mfloor += mroff;

					mceil = mfloor + ctex->rendh;
				}
				else
				{
					// ignore the fake sector's heights
					mceil = mid_faked ?
						FIX2FLT(MIN_OF(frontsec->smoothceil, backsec->smoothceil)) : openingTop;

					mceil += mroff;

					// Check for lifts/elevators
					if (!mid_faked && bmovsec && mceil < openingBottom && !(backsec->specFlags & SF_MOVINVALID))
						mceil = openingBottom;
				
					mfloor = mceil - ctex->rendh;
				}
				break;

			case R3D_21:
				mfloor = FIX2FLT(backsec->floorheight) + mroff;
				mceil = mfloor + ctex->rendh + FIX2FLT(sid->rowoffadj);
				break;

			case R3D_10:
				if (openingTop == openingBottom && rv.ssec->floorsec >= 0)
				{
					mceil = FIX2FLT(sectors[rv.ssec->floorsec].ceilingheight);
					mfloor = FIX2FLT(sectors[rv.ssec->floorsec].floorheight);
				}
				else
				{
					goto setas_default;	// v220-33
				//	mceil = openingTop;
				//	mfloor = openingBottom;
				}
				break;

			case R3D_11:
				// v220-4 test added
				if (frontsec == backsec && mid_faked && frontsec->specFlags & SF_MOVCEIL)
				{
					mfloor = FIX2FLT(rv.sect->smoothceil);
					mceil = mfloor + ctex->rendh;
				}
				else if (backsec->specFlags & SF_MOVFKR3D_11)
				{
					// Special type in fake sector
					int		bf = backsec->smoothfloor;
					int		bc = backsec->smoothceil;
					float	ht = FIX2FLT(rv.bceil - rv.bfloor);

					mceil = FIX2FLT(rv.fceil + (bc - bf));
					mfloor = mceil - MIN_OF(ht, ctex->rendh);	// v220-8 tex must not tile
				}
				else
				{
					int		ff = frontsec->smoothfloor;
					int		fc = frontsec->smoothceil;
					float	ht = FIX2FLT(rv.fceil - rv.ffloor);

					mceil = FIX2FLT(rv.fceil + (fc - ff));

					if (sid->toptexture)
						mceil -= ctex->rendh;
					mfloor = mceil - MIN_OF(ht, ctex->rendh);	// v220-8 tex must not tile
				}
				break;

			case R3D_12:
				if (rendtexmask)
				{
					// This cannot be tiled
					mfloor = FIX2FLT(rv.ffloor);
					mceil = mfloor + ctex->rendh;
				}
				else
				{
					mfloor = FIX2FLT(rv.ffloor);
					mceil = mfloor + mroff;
				}
				break;

			case R3D_13:
				{
					// v220-4 add lo alpha rows
					mfloor = FIX2FLT(sides[ldef->sidenum[1]].sector->smoothceil);
					mceil = mfloor + ctex->rendh;
				}
				break;

			case R3D_1D:
			case R3D_2D:
				mfloor = openingBottom;
				topvis = topvis && sid->toptexture;
				botvis = false;

				if(ldef->ml_flags & ML_DONTPEGBOTTOM)
				{
					mfloor += mroff;
					mceil = mfloor + ctex->rendh;
				}
				else
				{
					if (backsec->heightsec >= 0)
					{
						mceil = FIX2FLT(MIN_OF(frontsec->smoothceil, backsec->smoothceil));
					}
					else
						mceil = openingTop;

					mceil += mroff;
					mfloor = mceil - ctex->rendh;
				}
				break;

			default:
setas_default:;
				if (!mid_faked)
				{
					mceil = openingTop;
					mfloor = openingBottom;
				}
				else
				{
					mceil  = FIX2FLT(MIN_OF(frontsec->smoothceil , backsec->smoothceil ));
					mfloor = FIX2FLT(MAX_OF(frontsec->smoothfloor, backsec->smoothfloor));
				}

				if(ldef->ml_flags & ML_DONTPEGBOTTOM)
				{
					mfloor += mroff;
					mceil = mfloor + ctex->rendh;
				}
				else
				{
					mceil += mroff;
					mfloor = mceil - ctex->rendh;
				}

				// If necessary clip it.
				if (mceil > openingBottom && mfloor < openingBottom)
					mfloor = openingBottom;

				// Easiest way to align abutting, identical mid and bottom
				// textures is to draw one quad rather than two.
				// v220-5, v220-16 revised
				if (   ldef->ml_flags & ML_DONTPEGBOTTOM
					&& midtex == sid->bottomtexture		// if same tex
					&& !sid->toptexture					// and no toptextrue
					&& rv.ffloor < rv.bfloor			// and room for bottom texture
					&& mfloor == openingBottom			// and abutt
					&& mfloor < mceil					// and midtex values are valid
					&& !rendtexmask							// and not masked
					&& !(ldef->spflags & (SPF_MIDTRANSLUCENT|SPF_LOWTRANSLUCENT)) ) // and not trans
				{
					quad->top     = mceil;
					quad->bottom  = FIX2FLT(rv.ffloor);
					quad->texoffy = mroff *ctex->texscaley;

					rend_UpdateSide(sid, LS_MIDTEX, 0);
					return;
				}

				break;
			}

			quad->texoffx += segoff;

			if (!mid_direct)
			{
				// added v220-35 ---->
				RDBOOL hsbackonly = backsec->heightsec >= 0 && frontsec->heightsec < 0;
				RDBOOL hsbtex = false;
				float  temp;

				if (hsbackonly && (temp = FIX2FLT(MIN_OF(frontsec->smoothceil, sectors[backsec->heightsec].ceilingheight))) < mceil)
				{
					mceil  = temp;
					hsbtex = true;
				}
				
				if (hsbackonly && (temp = FIX2FLT(MAX_OF(frontsec->smoothfloor, sectors[backsec->heightsec].floorheight))) > mfloor)
				{
					mfloor = temp;
					hsbtex = true;
				}

				if (hsbtex) goto endhsbtex;
				// <-----------------

				// Prevent midtex bleeding into uppertex
				if (topskywall < 32767.0f && mceil > topskywall)
				{
					quad->texoffy += mceil - topskywall;
					mceil = topskywall;
				}
				else
				if ( (d = mceil - FIX2FLT(rv.bceil)) > 0)
				{
					if (sid->toptexture)
					{
						if (d < ctex->rendh)
						{
							mceil -= d;
							quad->texoffy += d;
						}
					}
					else if (!mid_faked)
						mid_covers_top = true;
				}

				// Prevent midtex bleeding into lowertex
				if (btextest && sid->bottomtexture && mfloor < lowtexceil)
				{
					mfloor = lowtexceil;
				}
			
				// v220-31 reduce clip if passing through a midtex line
				if (   rv.ssec == ppt.view_subsector
					&& !nearestClip
					&& !(ldef->ml_flags & ML_BLOCKING)
					&& fvy > mfloor
					&& fvy < mceil
					&& PointIsOnLine(ppt.viewplrmo->x, ppt.viewplrmo->y, ldef, 8))
				{
					nearestClip = true;
				}

				// Prevent midtex bleeding into uppertex
				if (   sid->toptexture
				//	&& !mid_faked	// Added v220-4 removed v220-35
					&& rv.bceil < rv.fceil
					&& mceil > (d = FIX2FLT(rv.bceil)) )
				{
					quad->texoffy -= (d - mceil);
					mceil = d;
				}

				if (mceil <= mfloor) goto upperwall;
			}
endhsbtex:;
			// Fill in the remainder of the data.
			quad->top    = mceil;
			quad->bottom = mfloor;
			
			midtextop = mceil;
			midtexbottom = mfloor + quad->texoffy;

			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;

			// Check for translucent walls R3D
			if (ldef->spflags & SPF_MIDTRANSLUCENT)
			{
				// Translucent polys (walls) are dealt with separately.
				quad->flags |= RPF_TRANS;
				rend_UpdateSide(sid, LS_MIDTEX|LS_MIDTRANS, 0);
			}
			else
			{
				if (rendtexmask)
				{
					quad->flags |= RPF_MASKED;
				}
				else if (ldef->ml_flags & ML_BLOCKING	//else will get problems when walking through
					&& quad->top >= openingTop
					&& quad->bottom <= openingBottom
					&& !(frontsec->skyfix && backsec->skyfix)
					&& rv.common < 0
					&& stype != R3D_16
					&& frontsec != backsec
					&& !(mtflags & R3D_MISTEXANY))
				{
					quad->flags |= RPF_OCCLUDE;
				}

				if (ldef->spflags & SPF_LINEONLINE)
				{
					ProjectVertices(seg, 0x3334, 0);	// .FRACUNIT/5
					rv.ssec->ssflags |= SSF_NOGLOW;
				}
				else if (ldef->spflags & SPF_LINEONLINE2)
				{
					ProjectVertices(seg, 0x199A, 0);	// .FRACUNIT/10
					rv.ssec->ssflags |= SSF_NOGLOW;
				}
				else if (ldef->xtraflags & XT_FRACPROJECT)
				{
					ProjectVertices(seg, 0x20000, 2);	// .2 * FRACUNIT
					rv.ssec->ssflags |= SSF_NOGLOW;
				}
				else if (ldef->xtraflags & XT_FRACPROJECTBACK)
				{
					ProjectVertices(seg, -0x20000, 3);	// -2 * FRACUNIT
					rv.ssec->ssflags |= SSF_NOGLOW;
				}

				rend_UpdateSide(sid, LS_MIDTEX, 0);

				// Added Dec2003
				if (  !mid_faked
					&& frontsec == rv.sect
					&& frontsec->ceilingpic == skyflatnum
					&& sid->lsiflags & LS_MIDTEX
					&& sid->mh > frontsec->smoothceil)
				{
					rv.sceil = sid->mh;
				}
			}
			
			// v220-3 Set fog wall
			if (   outsidefog
				&& !setNoFog
				&& linefog_group[ldef - lines] != 0xFFFF
				&& openingTop > openingBottom)
			{
				float	top = quad->top;
				float	bas = quad->bottom;

				if (quad->flags & (RPF_TRANS|RPF_MASKED))
				{
					if (quad->flags & RPF_QUADSET)	// should be
					{
						// Move the midtex quad's vertexes back
						// 1/2 FRACUNIT to prevent texture fighting.
						ProjectVertices(seg, 1<<15, 1);
						quad = Get_Quad(seg);
					}
					quad->top    = openingTop;
					quad->bottom = openingBottom;
					quad->flags  = (RPF_QUADSET|RPF_FOGWALL);
				}
				else
				{
					// Two fog walls might be required if the midtex
					// is between the top and bottom limits.
					if (top < openingTop)
					{
						if (quad->flags & RPF_QUADSET)
							quad = Get_Quad(seg);
						quad->top    = openingTop;
						quad->bottom = top;
						quad->flags  = (RPF_QUADSET|RPF_FOGWALL);
					}
					if (bas > openingBottom)
					{
						if (quad->flags & RPF_QUADSET)
							quad = Get_Quad(seg);
						quad->top    = bas;
						quad->bottom = openingBottom;
						quad->flags  = (RPF_QUADSET|RPF_FOGWALL);
					}
				}
			}
		}

upperwall:;

		if (rv.faked && rv.fceil == rv.bceil) goto lowerwall;

		if (   !fmovsec
			&& ldef->spflags & SPF_DRAWTOP
			&& backsec->ceilingpic != skyflatnum)
		{			
			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(backsec->ceilingpic);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->top = FIX2FLT(rv.fceil);
			quad->bottom = quad->top - 1.f;
			quad->texoffx = 0.0f;
			quad->texoffy = 0.0f;
			rend_UpdateSide(sid, LS_HITEX, backsec->ceilingpic);
			basetoptex = quad->bottom;
			goto lowerwall;
		}

		if (stype == R3D_0A && rv.ssec->floorsec >= 0)
			rv.ffloor = sectors[rv.ssec->floorsec].floorheight;

		// Added v220-4 Don't draw missing textures with mid_faked
		//				unless texture from other side can be used
		if (mid_faked && !(txn & 0x44))
		{
			// fakesec anomaly
			if (   frontsec->heightsec >= 0
				&& backsec->heightsec < 0
				&& backsec->smoothceil <= frontsec->smoothceil
				&& frontsec->smoothceil < sectors[frontsec->heightsec].smoothceil)
			{
				rv.sceil = frontsec->smoothceil;
				goto lowerwall;	// v220-33 moved here
			}
		}

		if (ldef->ml_flags & ML_IGNOREMISSINGHI)
		{
			goto lowerwall;
		}

		if (!midtex && rv.fceil == rv.bceil && !fmovsec && !bmovsec)
		{
			if (stype == R3D_22 && ldef->frontsector == ldef->backsector)
			{
				if (sid->toptexture && frontsec->ceildiff)
					goto type22;
				return;
			}
			goto lowerwall;
		}

		if (  !(targettex = sid->toptexture)
			&& ((stype == R3D_16 && !fmovsec && !bmovsec)
			|| (frontsec->ceilingpic == skyflatnum && viewz > MIN_OF(rv.fceil, rv.bceil))) )
		{
			goto lowerwall;
		}

		// v220-10
		if (   !targettex
			&& !(ldef->ovflags & OVF_MISHITEXLINE)
			&& rv.fceil - rv.bceil > 1024*FRACUNIT)
		{
			goto lowerwall;
		}

		// v220-8 modified
		if (  !targettex
			&& stype == R3D_00
			&& rv.ssec->ceilsec < 0)
		{
			if (   rv.bceil < rv.bfloor
				&& fmovsec && !bmovsec
				&& (frontsec->specFlags & 0x30018) == 0x30018)
			{
				rv.csec = backsec;
				rv.sceil = rv.bceil;
				goto lowerwall;
			}

			if (   !fmovsec && bmovsec
				&& (backsec->specFlags & 0x30018) == 0x20018)
			{
				goto lowerwall;
			}
		}

		// v220-31	dodgy toptexture map design with fake sector
		if (   targettex
			&& ldef->mtflags == 0x41
			&& ldef->frontsector->heightsec >= 0
			&& ldef->backsector->heightsec < 0
			&& rv.fceil > rv.bfloor)
		{
			quad = Get_Quad(seg);

			quad->tex = GL_PrepareTexture(targettex);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->top = FIX2FLT(rv.fceil);
			quad->bottom = FIX2FLT(rv.bfloor);
			roff = FIX2FLT(sid->rowoffset);
			quad->texoffy = roff*ctex->texscaley;
			quad->texoffx = segoff + xoff*ctex->texscalex;
			rend_UpdateSide(sid, LS_HITEX, 0);
			basetoptex = quad->bottom;
			goto lowerwall;
		}

		// v220-3 Check for invalid missing top textures in fake sectors
		if (   stype == R3D_03
			&& txn == 0x22
			&& frontsec->heightsec >= 0
			&& backsec->heightsec >= 0
			&& frontsec->heightsec != backsec->heightsec
			&& rv.fceil > rv.bceil
			&& !targettex
			&& frontsec->smoothceil == backsec->smoothceil)
		{
			goto lowerwall;
		}

		if (   ldef->lnflags2 & LNF2_NORENDHI
			|| (sid->lsiflags & LS_HISLOPEB
			&& rv.fceil + ssecvtxz[seg->zptr1].z[1] * FRACUNIT == rv.bceil))
		{
			goto lowerwall;
		}

		// Following two routines handle movceilings in a faked sector
		// This depends on a lower texture that is never drawn
		if (   !targettex						// 1st test as targettex must not be set
			&& frontsec->heightsec >= 0			// check if frontsector's faked
			&& fmovsec							// if front sector can change height
			&& ldef->backsector == frontsec		// Must be the line's backsector
			&& txn == 0x14						// Side 2 must have a lower texture
			&& !bmovsec							// backsec must be static
			&& (frontsec->specFlags & 0xFFFFFFC) == 0x03028590)
		{
			rv.sfloor = frontsec->smoothfloor;
			return;
		}

		// special case
		if (   targettex						// 1st test as targettex must be set
			&& bmovsec							// if back sector can change height
			&& backsec->heightsec >= 0			// check if backsector's faked
			&& !fmovsec							// frontsec must be static
			&& ldef->backsector == backsec		// Must be the line's backsector
			&& txn == 0x14						// Side 2 must have a lower texture
			&& (backsec->specFlags & 0xFFFFFFC) == 0x03028794)	// can have SF_NOLOTEX
		{
			int textop  = sid->sector->smoothceil;
			int texbase = MAX_OF(backsec->smoothceil, frontsec->smoothfloor);

			if (textop > texbase)
			{
				quad = Get_Quad(seg);
				quad->tex = GL_PrepareTexture(sid->toptexture);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;

				quad->top    = FIX2FLT(textop);
				quad->bottom = FIX2FLT(texbase);

				roff = FIX2FLT(sid->rowoffset + sid->rowoffadj);
				quad->texoffy = roff*ctex->texscaley;
				quad->texoffx = segoff + xoff*ctex->texscalex;
				basetoptex = quad->bottom;

				if ( !(ldef->ml_flags & ML_DONTPEGTOP) )
					quad->texoffy += (ctex->rendh - FIX2FLT(textop - backsec->smoothceil));

				if (rendtexmask)
					quad->flags |= RPF_MASKED;
			
				rend_UpdateSide(sid, LS_HITEX, 0);
			}
			goto lowerwall;
		}

		if (   !targettex
			&& ldef->ovflags & OVF_MISHITEXLINE	// only set with R3D_00 lines
			&& rv.ssec->ceilsec >= 0
			&& !faked)
		{
			if (rv.bceil >= sectors[rv.ssec->ceilsec].smoothceil)
				goto lowerwall;

			// b211-0 - Draw wall to required height using value in ss->ceilsec
			flatnum = sectors[rv.ssec->ceilsec].ceilingpic;

			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->bottom  = FIX2FLT(rv.bceil);
			quad->top     = FIX2FLT(sectors[rv.ssec->ceilsec].smoothceil);

			if (rendtexmask)
			{
				quad->texoffx = segoff;
				quad->texoffy = 0.0f;
			}
			else
			{
				quad->texoffx = segoff + xoff;
				quad->texoffy = roff;
			}

			R_WallHeightDivision(seg, WHD_TOP);
			rend_UpdateSide(sid, LS_HITEX, flatnum);
			basetoptex = quad->bottom;
			goto lowerwall;
		}

		if (!sid->texFlags
			&& backsec->floorpic == skyflatnum
			&& backsec->ceilingpic == skyflatnum 
			&& frontsec->ceilingpic == skyflatnum 
			&& rv.ffloor <= rv.bfloor)
		{
			if (bsh && rv.ffloor < rv.bfloor)
			{
				rv.fsec = backsec;
				rv.sfloor = rv.bfloor;
			}
			return;
		}

		if (ldef->spflags & SPF_NOTEX && rv.common >= 0)
		{
			return;
		}

		if (stype == R3D_38)
			goto lowerwall;

		if (stype == R3D_11 && frontsec->specFlags & SF_MOVFKR3D_11)
		{
			return;
		}

type22:;
		if (stype == R3D_1D || stype == R3D_22)
		{
			if (sid->toptexture)
			{
				float t = 0.0f, b = 0.0f;

				if (stype == R3D_1D && rv.fceil > rv.bceil)
				{
					t = FIX2FLT(MAX_OF(rv.fceil, rv.bceil));
					b = FIX2FLT(MIN_OF(rv.fceil, rv.bceil));
				}
				else if (frontsec->ceildiff > 0)
				{
					t = FIX2FLT(rv.fceil);	// front used with R3D_22
					b = t - FIX2FLT(frontsec->ceildiff);	// v220-8 bug fix: t substituted for quad->top
				}

				if (t > b)
				{
					quad = Get_Quad(seg);

					quad->tex = GL_PrepareTexture(sid->toptexture);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;

					quad->top = t;
					basetoptex = quad->bottom = b;
					quad->texoffx = segoff + xoff*ctex->texscalex;
					// v220-9 don't apply y offset if masked
					quad->texoffy = (rendtexmask || ldef->ml_flags & ML_DONTPEGTOP) ? 0 : roff*ctex->texscaley;

					R_WallHeightDivision(seg, WHD_TOP);
					rend_UpdateSide(sid, LS_HITEX, 0);
				}

				if (sid->bottomtexture) goto lowerwall;

				return;
			}

			if (   !sid->bottomtexture
				&& (rv.bceil == rv.fceil
				|| (quad && quad->flags & RPF_QUADSET && quad->bottom <= FIX2FLT(rv.ffloor))))
			{
				return;	// Don't draw lower missing textures
			}

			if (fmovsec ^ bmovsec)
			{
				if ((txn & 0x11) == 0x11 ||
					ldef->mtflags & R3D_MISHITEX)
					goto lowerwall;

				if (midtex || sid->toptexture)
				{
					if (fmovsec)
						R_SetMissingParms(backsec, backsec);
					else
						R_SetMissingParms(frontsec, frontsec);
				}
			}
			goto lowerwall;
		}

		if (stype == R3D_39 && (flatnum = backsec->ceilingpic) != skyflatnum)
		{
			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->bottom  = midtextop;
			quad->top     = FIX2FLT(seg->frontsector->smoothceil);
			if (rendtexmask)
			{
				quad->texoffx = segoff;
				quad->texoffy = 0.0f;
			}
			else
			{
				quad->texoffx = segoff + xoff;
				quad->texoffy = roff;
			}

			R_WallHeightDivision(seg, WHD_TOP);
			rend_UpdateSide(sid, LS_HITEX, flatnum);
			basetoptex = quad->bottom;
			goto lowerwall;
		}

		if (!targettex && !fmovsec && !bmovsec)
		{
			side_t *sc = (sid == &sides[ldef->sidenum[0]]) ?
				&sides[ldef->sidenum[1]] : &sides[ldef->sidenum[0]];
			targettex = sc->toptexture;
		}

		// Upper wall
		if (topvis && !skyflag && !mid_covers_top)
		{
			if(targettex)	// A texture present?
			{
				quad = Get_Quad(seg);
				quad->tex = GL_PrepareTexture(targettex);

				if (rendtexmask)
				{
					if (ldef->lnflags2 & LNF2_NOALPHA)
					{
						quad->tex = GL_PrepareTextureNoAlpha(targettex);
					}
					else if (!(backsec->specFlags & SF_NOOCCLUDE)
						&& !(!midtex && backsec->specFlags & SF_MASKEDDOOR)
						&& backsec->heightsec < 0)
					{
						// speed test on future pass
						ldef->lnflags2 |= LNF2_NOALPHA;
						quad->tex = GL_PrepareTextureNoAlpha(targettex);
					}
					else
					{
						quad->flags |= RPF_MASKED;
					}
				}

				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;

				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy = roff*ctex->texscaley;

				if (  !fmovsec
					&& !(ldef->ml_flags & ML_DONTPEGTOP)
					&& bmovsec
					&& backsec->heightsec >= 0
					&& backsec->specFlags & SF_MOVCEIL)
				{
					quad->texoffy += FIX2FLT(backsec->smoothceil - rv.bceil);
				}
			}
			else if (rv.csec->R3Dceilsec >= 0 || stype == R3D_3D)
			{
				goto lowerwall;
			}
			// A special effect?
			else if (midtexquadset
				&& bmovsec
				&& !(sid->lsiflags & LS_MIDTEX)
				&& !sid->bottomtexture)
			{
				// Overwrite the midtex quad
				quad->tex  = GL_PrepareTexture(midtex);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;

				if (ldef->ml_flags & ML_DONTPEGTOP)
				{
					quad->top = FIX2FLT(rv.bceil);
					quad->bottom = quad->top - quad->texh;
				}
				else
				{
					quad->bottom = FIX2FLT(rv.bceil);
					quad->top = quad->bottom + quad->texh;
					quad->texoffy = ctex->rendh - roff;
				}

				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy *= ctex->texscaley;

				rend_UpdateSide(sid, LS_MIDTEX, 0);

				if (ldef->lnflags2 & LNF2_SPAN1DOORS)
					goto lowerwall;
				return;
			}
			else if (!faked
				&& ldef->specType == R3D_00
				&& ldef->ovflags & OVF_MISHITEXLINE)
			{
				goto lowerwall;
			}
			else if (stype == R3D_12 ||
				(!midtex && backsec->ceilingpic == skyflatnum))
			{
				goto lowerwall;
			}
			else
			{
				if (   mtflags & R3D_MISHITEX
					&& !bmovsec && !faked)
				{
					// E.g. batman map 11 sec 88
					if (   stype == R3D_00
						&& rv.common < 0
						&& !fmovsec
						&& txn & 0x44
						&& !(mtflags & R3D_MISLOTEX))
					{
					//	do nothing
					}
					else if (mtflags & R3D_MISHTEX1)
					{
						if (!txn || frontsec->ceilingpic == skyflatnum)
						{
							if (viewz > ldef->backsector->smoothceil
								|| backsec->R3Dceilheight != DDMININT)
							{
								goto lowerwall;
							}
						}
					}
					else
					{
						if (!txn || backsec->ceilingpic == skyflatnum)
						{
							if (viewz > ldef->frontsector->smoothceil)
								goto lowerwall;
						}
					}
				}

				// Is this a special effect?
				if (!midtex && bmovsec && txn == 0x02)
				{
					targettex = sides[ldef->sidenum[0]].midtexture;
				}
				else // Still no texture? Use backsec unless skyflatnum.
				{
					if (ldef->lnflags2 & LNF2_DRAWSKY)
					{
						if (   rv.bceil > rv.fceil
							&& backsec->ceilingpic == skyflatnum)
						{
							quad = Get_Quad(seg);
							quad->flags |= RPF_SKY_MASK|RPF_OCCLUDE|RPF_QUADSET;
							quad->top = FIX2FLT(rv.bceil + backsec->skyfix);
							quad->bottom = FIX2FLT(rv.fceil);
						}

						goto lowerwall;
					}

					// v220-30 if a bottom texture exists and the following
					// conditions are met then use it in place of a flat (rare).
					if (   sid->bottomtexture
						&& ldef->specType == R3D_00
						&& rv.fceil == rv.bfloor
						&& rv.bceil == rv.ffloor)
					{
						goto lowerwall;
					}

					flatnum = backsec->ceilingpic;

					if (flatnum == skyflatnum)
						flatnum = frontsec->ceilingpic;
					
					quad = Get_Quad(seg);

					quad->tex = GL_PrepareFlat(flatnum);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;

					quad->top = FIX2FLT(rv.fceil);
					quad->bottom = FIX2FLT(MAX_OF(rv.bceil, rv.ffloor));

					if (rendtexmask)
					{
						quad->texoffx = segoff;
						quad->texoffy = 0.0f;
					}
					else
					{
						quad->texoffx = segoff + xoff;
						quad->texoffy = roff;
					}

					R_WallHeightDivision(seg, WHD_TOP);
					rend_UpdateSide(sid, LS_HITEX, flatnum);
					basetoptex = quad->bottom;
					goto lowerwall;
				}
			}

			if (!targettex)
				goto lowerwall;

			if (!quad || quad->flags & RPF_QUADSET)
				quad = Get_Quad(seg);
			else
			if (!(ldef->ml_flags & ML_DONTPEGTOP))
				quad->texoffy += (ctex->rendh - FIX2FLT(rv.fceil - rv.bceil));	// Normal alignment to bottom

			if (!quad->tex) quad->tex = GL_PrepareTexture(targettex);

			if (!compat[c_fake] && backsec->heightsec >= 0 && sectors[backsec->heightsec].ceilingheight > rv.bceil)
				rv.bceil = sectors[backsec->heightsec].ceilingheight;	// fake fix added v220-35

			quad->top = FIX2FLT(rv.fceil);

			if (transvalid)
				quad->bottom = FIX2FLT(rv.bceil);
			else
				quad->bottom = FIX2FLT(MAX_OF(rv.bceil, rv.ffloor));

			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			if (rv.ffloor >= rv.bceil)
			{
				if ( !(quad->flags & RPF_MASKED) ) quad->flags |= RPF_OCCLUDE;

				// v220-4 check for hitex covering midtex
				if (mid_faked && midtex && quad->bottom < midtextop )
				{
					quad->bottom = midtextop;

					if (sid->lsiflags & LS_MIDALPHA)
						quad->flags &= ~RPF_OCCLUDE;
				}
			}

			R_WallHeightDivision(seg, WHD_TOP);
			rend_UpdateSide(sid, LS_HITEX, 0);
			basetoptex = quad->bottom;
		}

lowerwall:

		if (mid_direct) return;

		// v220-33 test added to eliminate unwanted low tex added to a side or ignore a missing lotex
		if (!rv.pseudo && backsec->heightsec >= 0 && !(sectors[backsec->heightsec].specFlags2 & SF2_PSEUDOFAKE))
		{
			if (!(sid->bottomtexture && ldef->mtflags & (R3D_INVLTEX1|R3D_INVLTEX2) && frontsec->heightsec < 0) )
			{
				if (frontsec->heightsec >= 0)
				{
					if (!sid->bottomtexture)
					{
						if (frontsec->floorheight == frontsec->ceilingheight) return;	// v220-35
						if (frontsec->specFlags2 & SF2_NONEMISSMOVLO) return;
					}
					if (frontsec->floorheight >= backsec->floorheight && rv.bfloor == frontsec->floorheight) return;
				}
				else
				if (   frontsec->heightsec < 0
					&& rv.ffloor < rv.bfloor
					&& rv.ffloor > backsec->floorheight
					&& frontsec->sibounds[BOXLEFT ] > backsec->sibounds[BOXLEFT ]
					&& frontsec->sibounds[BOXRIGHT] < backsec->sibounds[BOXRIGHT]
					&& frontsec->sibounds[BOXBASE ] > backsec->sibounds[BOXBASE ]
					&& frontsec->sibounds[BOXTOP  ] < backsec->sibounds[BOXTOP  ])
				{
					return;
				}
				if (bmovsec && !(ldef->ml_flags & ML_DONTPEGBOTTOM))
					rv.bfloor = MIN_OF(backsec->smoothfloor, sectors[backsec->heightsec].smoothfloor);	// clamp tex to moving sec
			}
		}

		if (transvalid && !vssecunderfake)
		{
			if (!sid->bottomtexture && frontsec->floorheight == backsec->floorheight)
				return;	// v220-31 no missingtex allowed with this type

			if (backsec->smoothfloor < rv.ffloor)
			{
				if (   sid->bottomtexture
					&& frontsec->heightsec >= 0	// bug fix v220-33 add >= 0
					&& backsec->heightsec >= 0	// bug fix v220-33 add >= 0
					&& sectors[frontsec->heightsec].floorheight < sectors[backsec->heightsec].floorheight
					&& frontsec->floorheight > backsec->floorheight)
				{
					rv.ffloor = sectors[frontsec->heightsec].floorheight;
					rv.bfloor = sectors[backsec->heightsec].floorheight;
				}
				else
				{
					rv.ffloor = MIN_OF(rv.sfloor, frontsec->smoothfloor);
					rv.bfloor = backsec->smoothfloor;
					btextest = rv.bfloor > rv.ffloor;
				}
			}
		}

		// v220-6 Added
		if (!txn && stype == R3D_16 && fmovsec && bmovsec) return;

		// Added v220-4 Don't draw missing textures with mid_faked
		//				unless texture from other side can be used
		// removed v220-33
	/*	if (!(txn & 0x11) && stype != R3D_00 && mid_faked)
		{
			return;
		}*/

		if (   ldef->ml_flags & ML_IGNOREMISSINGLO
			&& !(sid->bottomtexture && rv.bfloor > rv.ffloor))
		{
			return;
		}

		if (  !(targettex = sid->bottomtexture)
			&& rv.sect->oversec >= 0)
		{
			return;
		}

		if (!targettex && rv.ffloor - rv.bfloor > 8*FRACUNIT && frontsec->floorpic == skyflatnum) return;

		// if (stype == R3D_16 && (midtexmask || (!targettex && !fmovsec && !bmovsec)))
		// v220-8 change to:
		if (!targettex && stype == R3D_16 && (midtexmask || (!fmovsec && !bmovsec)))
		{
			return;
		}

		if (!targettex)
		{
			if (midtex)
			{
				int top = rv.bfloor > rv.fceil ? rv.fceil : rv.bfloor;
				if (FIX2FLT(top) > midtexbottom) return;
			}
			else		// v220-8
			if (   stype == R3D_00
				&& fmovsec && !bmovsec
				&& rv.ssec->floorsec < 0
				&& rv.bfloor > rv.ffloor
				&& !(!frontsec->floordata && abs(rv.bfloor - rv.ffloor) <= 2*FRACUNIT)
				&& (frontsec->specFlags & 0x10014) == 0x10014)
			{
				rv.fsec = backsec;
				rv.sfloor = rv.bfloor;
				return;
			}
		}

		if (  !targettex
			&& stype == R3D_00
			&& rv.bfloor != rv.ffloor
			&& (flatnum = backsec->floorpic) != skyflatnum
			&& !fmovsec && !bmovsec
			&& frontsec->heightsec >= 0
			&& backsec->heightsec >= 0)
		{
			quad = Get_Quad(seg);

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			quad->top = FIX2FLT(MAX_OF(rv.bfloor, rv.ffloor));
			quad->bottom = FIX2FLT(MIN_OF(rv.bfloor, rv.ffloor));

			if (rendtexmask)
			{
				quad->texoffx = segoff;
				quad->texoffy = 0.0f;
			}
			else
			{
				quad->texoffx = segoff + xoff;
				quad->texoffy = roff;
			}

			if (basetoptex < quad->top)
			{
				if (quad->bottom >= basetoptex) return;

				quad->texoffy += quad->top - basetoptex;
				quad->top = basetoptex;
			}

			R_WallHeightDivision(seg, WHD_BOTTOM);
			rend_UpdateSide(sid, LS_LOWTEX, flatnum);
			return;
		}

		if (   ldef->lnflags2 & LNF2_NORENDLOW
			|| (sid->lsiflags & LS_LOSLOPEB &&
			frontsec->smoothfloor + ssecvtxz[seg->zptr1].z[0] * FRACUNIT == backsec->smoothfloor))
		{
			return;
		}

		if (faked && sid->lsiflags & LS_LOSLOPEB)
		{
			float z = ssecvtxz[seg->zptr1].z[0];

			if (z && z == ssecvtxz[seg->zptr2].z[0])
			{
				int zht = (int)(z * FRACUNIT) - (rv.ffloor - rv.sect->smoothfloor);

				if (zht > 0) rv.ffloor += zht;
			}
		}

		if (!txn && backsec->R3Dfloorheight < DDMAXINT)
		{
			if (backsec->floorpic != skyflatnum)
				rv.sfloor = backsec->R3Dfloorheight;
			return;
		}

		// b211-0 - Added
		if (stype==R3D_3A && rv.ffloor > rv.bfloor)
		{
			int tmp = rv.ffloor;
			rv.ffloor = rv.bfloor, rv.bfloor = tmp;
			targettex = 0;
			goto set_lower;
		}

		// Lower wall.
		if(!btextest)
		{
			if (!targettex
				&& viewz < rv.ffloor
				&& ldef->lnflags2 & LNF2_RENDVIEWH)
			{
				rv.fsec = backsec;
				rv.sfloor = rv.bfloor;
			}
			return;
		}

		if (!targettex)	// No texture tests
		{
			if (   stype == R3D_14
				|| stype == R3D_12
				|| stype == R3D_0F)
			{
				return;
			}

			if (viewz < rv.bfloor
				&& ldef->lnflags2 & LNF2_RENDVIEWH)
			{
				return;
			}

			if (!fmovsec && !bmovsec)
			{
				// test added v220-8
				// e.g. Scythe2 map30 S572
				if (abs(rv.ffloor - rv.bfloor) > MAX_SECHEIGHT)
				{
					if (  (frontsec->specFlags & 0xFFFFFC) == 0x80001C
						&& rv.bfloor > rv.ffloor)
					{
						rv.fsec = backsec;
						rv.sfloor = rv.fsec->floorheight;
					}
					return;
				}
				else
				{
					side_t *sc = (sid == &sides[ldef->sidenum[0]]) ?
						&sides[ldef->sidenum[1]] : &sides[ldef->sidenum[0]];
					targettex = sc->bottomtexture;
				}
			}
		}

		if (!targettex)	// Still no texture?
		{
			if (rv.ffloor >= rv.bceil) return;

			if (midtex && FIX2FLT(rv.bfloor) > basetoptex) return;

			if (midtex && bmovsec &&
				!(sid->lsiflags & LS_MIDTEX) &&
				!sid->toptexture)
			{
				quad = Get_Quad(seg);

				quad->tex  = GL_PrepareTexture(midtex);
				quad->texw = ctex->rendw;
				quad->texh = ctex->rendh;
				quad->detail = rendtexdetail;

				if (ldef->ml_flags & ML_DONTPEGBOTTOM)
				{
					quad->top = FIX2FLT(rv.bfloor);
					quad->bottom = quad->top - ctex->rendh;
				}
				else
				{
					quad->bottom = FIX2FLT(rv.bfloor);
					quad->top = quad->top + ctex->rendh;
					quad->texoffy = ctex->rendh - roff;
				}

				quad->texoffx = segoff + xoff*ctex->texscalex;
				quad->texoffy *= ctex->texscaley;

				if (basetoptex < quad->top)
				{
					if (quad->bottom >= basetoptex) return;

					quad->texoffy += quad->top - basetoptex;
					quad->top = basetoptex;
				}

				rend_UpdateSide(sid, LS_HITEX, 0);
				return;
			}
		}
set_lower:;

		if (ldef->lnflags2 & LNF2_NORENDLOW) return;

		if (!targettex)
		{
			quad = Get_Quad(seg);

			if ( (flatnum = backsec->floorpic ) == skyflatnum)
			{
				if (rv.bfloor > rv.ffloor)
				{
					quad->top = FIX2FLT(rv.bfloor);
					quad->bottom = FIX2FLT(rv.ffloor);
					quad->flags |= (RPF_SKY_MASK|RPF_QUADSET);
				}
				return;
			}

			quad->tex = GL_PrepareFlat(flatnum);
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;

			// v220-26 add test for type R3D_26
			quad->top = FIX2FLT(stype == R3D_26 ? rv.bfloor : (rv.bfloor > rv.fceil ? rv.fceil : rv.bfloor) );
			quad->bottom = FIX2FLT(rv.ffloor);

			if (rendtexmask)
			{
				quad->texoffx = segoff;
				quad->texoffy = 0.0f;
			}
			else
			{
				quad->texoffx = segoff + xoff;
				quad->texoffy = roff;
			}

			if (basetoptex < quad->top)
			{
				if (quad->bottom >= basetoptex) return;

				quad->texoffy += quad->top - basetoptex;
				quad->top = basetoptex;
			}

			R_WallHeightDivision(seg, WHD_BOTTOM);
			rend_UpdateSide(sid, LS_LOWTEX, flatnum);
			return;
		}

		quad = Get_Quad(seg);

		quad->tex = GL_PrepareTexture(targettex);

		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;

		// v220-25 offset code revised --------------
		if (rendtexmask)
		{
			if (backsec->heightsec < 0)
				quad->tex = GL_PrepareTextureNoAlpha(targettex);
			else
				quad->flags |= RPF_MASKED;

			if (   roff
				&& !(ldef->ml_flags & (ML_DONTPEGTOP|ML_DONTPEGBOTTOM)))
			{
				quad->texoffy = roff;
				if (roff < 0) quad->texoffy += ctex->rendh;
			}
		}
		
		if (ldef->ml_flags & ML_DONTPEGBOTTOM)
		{
			if (   rv.fceil != rv.bceil
				&& frontsec->ceilingpic == skyflatnum
				&& backsec->ceilingpic == skyflatnum)
			{
				if (1 || ma_inf.r3dskyoffsets)	// v220-35 always true
				{
					// v220-27 use the max of the front and back ceiling heights
					//         if both line sectors reference a sky texture
					quad->texoffy += FIX2FLT(MAX_OF(rv.fceil, rv.bceil) - rv.bfloor);
				}
				else
				{
					// v220-31 use Doom method
					quad->texoffy += FIX2FLT(rv.bceil - rv.fceil - rv.bfloor);
				}
			}
			else
			{
				quad->texoffy += FIX2FLT(rv.fceil - rv.bfloor);
			}
		}
		// -----------------------

		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;

		quad->bottom = FIX2FLT(rv.ffloor);

		if (rv.bfloor > rv.fceil)
		{
			quad->texoffy += FIX2FLT(rv.bfloor - rv.fceil);
			quad->top = FIX2FLT(rv.fceil);
		}
		else
		{
			quad->top = FIX2FLT(rv.bfloor);
		}

		if (basetoptex < quad->top)
		{
			if (quad->bottom >= basetoptex) return;

			quad->texoffy += quad->top - basetoptex;
			quad->top = basetoptex;
		}

		if (rv.sect->specFlags2 & SF2_NEGOFFSET) quad->texoffy -= 1.f;

		R_WallHeightDivision(seg, WHD_BOTTOM);
		rend_UpdateSide(sid, (byte)(transwall ? LS_LOWTEX|LS_LOWTRANS : LS_LOWTEX), 0);

		if (   !transwall
			&& !(quad->flags & (RPF_MASKED|RPF_OCCLUDE)) )
		{
			// v220-16 occlusion added
			if (   quad->top >= FIX2FLT(frontsec->ceilingheight)
				&& quad->bottom <= FIX2FLT(frontsec->floorheight))
			{
				quad->flags |= RPF_OCCLUDE;
			}
			else if (seg->length >= 4.f)	// v220-25
			{
				// v220-9
				float diff = quad->top - fvy;

				if (diff > 0.f)
				{
					float ds = MAX_OF(quad->vertices[0].vpdist, quad->vertices[1].vpdist);
					double an = atan2((double)(diff*1.2f), (double)ds);

					if (an - rad_pitch > occAngle) quad->flags |= RPF_OCCLUDE;
				}
			}
		}
	}
}

// R3D 3D lines
// Only send if seg facing
// v220-13 $rotate removed
// v220-31 revised
static void Rend_RenderWallSeg3D(seg_t *seg)
{
	side_t		*sid = seg->sidedef;
	sector_t	*frontsec = seg->frontsector;

	if (!sid->midtexture && !sid->bottomtexture) return;

	line_t		*ldef = seg->linedef;
	float		roff, xoff;

	xoff = FIX2FLT(seg->offset);
	roff = FIX2FLT(sid->rowoffset);

	// Normal upper/lower tex line in ssec with flats3DArray set. Initially 
	// identified by a mid tex carrying the light levels. When loaded the light
	// levels are put in lo/hilightlevel and the line type is set to R3D_4B
	if (ldef->specType == R3D_4B)
	{
		int count, *flatslist = rv.ssec->flats3DArray;

		if (flatslist && (count = *(int*)flatslist++) == 1)
		{
			if (sid->bottomtexture)
			{
				int	*loflatptr = flatslist + (count - 1)*FLIST_ENTRIES;	// End of list

				quad = Get_Quad(seg);
				quad->bottom = FIX2FLT(frontsec->smoothfloor);
				quad->top = quad->bottom + FIX2FLT(loflatptr[2]);	// + offset

				if (quad->top > quad->bottom)
				{
					quad->tex = GL_PrepareTexture(sid->bottomtexture);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;

					quad->texoffx = xoff + (FIX2FLT(sid->textureoffset) * ctex->texscalex);
					quad->texoffy = roff * ctex->texscaley;

					rend_UpdateSide(sid, LS_LOWTEX, 0);
				}
			}
			if (sid->toptexture)
			{
				int hiflat = sectors[flatslist[1]].smoothfloor + flatslist[2] + flatslist[3];	// offset + depth;
				
				quad = Get_Quad(seg);
				quad->bottom = FIX2FLT(hiflat);
				quad->top = FIX2FLT(seg->backsector ? seg->backsector->smoothfloor : frontsec->smoothceil);

				if (quad->top > quad->bottom)
				{
					quad->tex = GL_PrepareTexture(sid->toptexture);
					quad->texw = ctex->rendw;
					quad->texh = ctex->rendh;
					quad->detail = rendtexdetail;

					quad->texoffx = xoff + (FIX2FLT(sid->textureoffset) * ctex->texscalex);
					quad->texoffy = roff * ctex->texscaley;

					rend_UpdateSide(sid, LS_HITEX, 0);
				}
			}
		}
		return;
	}

	// Type scrolling translucent is not scaled
	if (ldef->specType == R3D_4A)
	{
		int usellev;

		quad = Get_Quad(seg);
		quad->bottom  = FIX2FLT(frontsec->smoothfloor);
		quad->top     = quad->bottom + Q_FIX2FLT(ldef->d3d->lineheightlow);
		quad->texoffx = xoff + FIX2FLT(sid->textureoffset);
		quad->texoffy = roff;

		quad->tex = GL_PrepareTexture(sid->bottomtexture);
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;

		if (rendtexmask) quad->flags |= RPF_MASKED;
		quad->flags = (RPF_TRANS|RPF_QUADSET);

		// With trans textures clamp light to 96 else texture can disappear
		if ( (usellev = R_SectorLight(rv.floorlightlevel)) < 96) usellev = 96;

		RL_VertexColors(quad->vertices, quad->numvertices, usellev, quad->flags & RPF_MASKED);
		return;
	}

	int		liftsec = ldef->frontsector->liftsec3D;	// Lift sector
	int		height  = liftsec >= 0 ?
				sectors[liftsec].smoothfloor : ldef->backsector->smoothfloor;

	if (sid->bottomtexture && ldef->d3d->lineheightlow >= FRACUNIT)
	{
		quad = Get_Quad(seg);
		quad->bottom = FIX2FLT(ldef->d3d->offset3D + height);
		quad->top = quad->bottom + FIX2FLT(ldef->d3d->lineheightlow);

		quad->tex = GL_PrepareTexture(sid->bottomtexture);
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;

		if (rendtexmask) quad->flags |= RPF_MASKED;

		quad->texoffx = xoff + (FIX2FLT(sid->textureoffset) * ctex->texscalex);

		// Translucent poly (walls) get a special treatment.
		if (ldef->spflags & SPF_LOWTRANSLUCENT)
		{
			quad->texoffy = 0.0f;
			quad->flags |= RPF_TRANS;
		}
		else if (ldef->ml_flags & ML_DONTPEGBOTTOM)
			quad->texoffy = roff*ctex->texscaley + quad->texh - FIX2FLT(ldef->d3d->lineheightlow);
		else
			quad->texoffy = roff*ctex->texscaley;

		rend_UpdateSide(sid, quad->flags & RPF_TRANS ? LS_LOWTEX|LS_LOWTRANS : LS_LOWTEX, 0);
	}

	if (sid->midtexture && ldef->d3d->lineheightmid >= FRACUNIT)
	{
		quad = Get_Quad(seg);
		quad->bottom = FIX2FLT(ldef->d3d->offset3D + ldef->d3d->lineheightlow + height);
		quad->top = quad->bottom + FIX2FLT(ldef->d3d->lineheightmid);

		quad->tex = GL_PrepareTexture(sid->midtexture);
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;

		if (rendtexmask) quad->flags |= RPF_MASKED;

		// Check for translucent walls
		if (ldef->spflags & SPF_MIDTRANSLUCENT)
		{
			quad->texoffy = 0.0f;
			// Translucent polys (walls) get a special treatment.
			quad->flags |= RPF_TRANS;
		}
		else
		{
			quad->texoffy = roff*ctex->texscaley;
			quad->texoffx = xoff + (FIX2FLT(sid->textureoffset) * ctex->texscalex);
		}

		rend_UpdateSide(sid, quad->flags & RPF_TRANS ? LS_MIDTEX|LS_MIDTRANS : LS_MIDTEX, 0);
	}
}


//===========================================================================
// Rend_RenderSubsector
//===========================================================================

static RDBOOL CheckRendered(float height)
{
	int count = rend_count;

	while(--count >= 0)
	{
		if (rend_flats_list[count] == height)
			return true;
	}

	rend_flats_list[rend_count++] = height;
	return false;
}

static void getHeights(sector_t *sec, int *fh, int *ch)
{
	if (sec->heightsec >= 0)
	{
		*fh = sectors[sec->heightsec].smoothfloor;
		*ch = sectors[sec->heightsec].smoothceil;
	}
	else
	{
		*fh = sec->smoothfloor;
		*ch = sec->smoothceil;
	}
}

// Note: seg->backsector must be set
static void subfn_DoSkyWall(seg_t *seg)
{
	int ffh, fch, bfh, bch;

	if (   seg->backsector->floorpic == skyflatnum
		&& (bfh = seg->backsector->smoothfloor) > (ffh = seg->frontsector->smoothfloor))
	{
		quad = Get_Quad(seg);
		quad->flags = RPF_SKY_MASK|RPF_QUADSET;
		quad->top    = FIX2FLT(bfh);
		quad->bottom = FIX2FLT(ffh);
	}
	else
	if (   seg->frontsector->ceilingpic == skyflatnum
		&& seg->backsector->ceilingpic == skyflatnum
		&& (seg->frontsector->specFlags & SF_MOVCEIL)^(seg->backsector->specFlags & SF_MOVCEIL)
		&& (fch = seg->frontsector->ceilingheight) > (bch = seg->backsector->ceilingheight)
		&& seg->frontsector->floorheight == bch)
	{
		quad = Get_Quad(seg);
		quad->flags = RPF_SKY_MASK|RPF_QUADSET|RPF_OCCLUDE;
		quad->top    = FIX2FLT(fch);
		quad->bottom = FIX2FLT(bch);
		rv.inhibit = false;
	}

	if (seg->linedef->tyflags & TYF_SKYWALL)
	{
		getHeights(seg->frontsector, &ffh, &fch);
		getHeights(seg->backsector , &bfh, &bch);

		if (bch < fch)
		{
			if (bch > bfh)
			{
				quad = Get_Quad(seg);
				quad->flags = RPF_SKY_MASK|RPF_QUADSET;
				quad->top    = FIX2FLT(fch + rv.skyfix);
				quad->bottom = FIX2FLT(bch);
			}
			else if (ffh < fch)	// && bch <= bfh
			{
				quad = Get_Quad(seg);
				quad->flags = RPF_SKY_MASK|RPF_QUADSET;
				quad->bottom = FIX2FLT(ffh);
				quad->top = FIX2FLT(fch + seg->frontsector->skyfix);

				if (   seg->frontsector->ceilingpic != skyflatnum
					&& seg->backsector->ceilingpic  == skyflatnum)
				{
					quad->flags |= RPF_OCCLUDE;

					// If the ceilingpic is not a skyflat then the masked
					// wall won't render. Use rvceil2 to fix this. Setting
					// it higher than the true ceiling means the wanted
					// ceiling flat will continue to be seen.
					if (rv.csec->ceilingpic != skyflatnum)
					{
						rv.ceil2 = fch + FRACUNIT;
						quad->top = FIX2FLT(rv.ceil2);
					}
				}
			}
		}
	}
}

// NOTE: don't send minisegs
static void subfn_DoWallSeg(seg_t *seg)
{
	sector_t	*bs = seg->backsector;
	sector_t	*fs = seg->frontsector;

	if (   bs
		&& fs->specFlags2 & SF2_MOVSKYCEIL
		&& bs->ceilingpic == skyflatnum
		&& fs->ceilingpic == skyflatnum
		&& bs->smoothceil > rv.skyheight
		&& bs->smoothceil > fs->smoothceil)
	{
		rv.skyheight = bs->smoothceil;
	}

	if (seg->linedef->specType >= R3D_40)
	{
		if (seg->sgflags & DDSEGF_FACING)
		{
			// rv.distance1 and rv.distance2 are assigned to a quad on calling Get_Quad().
			rv.distance1 = Rend_PointDist2D(seg->fv1);
			rv.distance2 = Rend_PointDist2D(seg->fv2);
			Rend_RenderWallSeg3D(seg);
		}
	}
	else
	{
		if (seg->linedef->specType == R3D_01 && bs)	// should be
		{
			if (!rv.fixedsec && (fs == rv.sect || bs == rv.sect))
			{
				int ff = fs->specFlags;
				int fb = bs->specFlags;

				if (!(rv.ssec->ssflags & SSF_01INHIBIT))
				{
					if (fs->R3Dfloorsec < 0 && bs->R3Dfloorsec < 0)
					{
						if (ff & SF_MOVFLOOR && !(fb & SF_MOVFLOOR))
						{
							rv.fixedsec = bs;
						}
						else if (!(ff & SF_MOVFLOOR) && fb & SF_MOVFLOOR)
						{
							rv.fixedsec = fs;
						}
						else if (ff & SF_MOVFLOOR && fb & SF_MOVFLOOR)
						{	// v220-4 modified
							if (ff & SF_NOLOTEX && !(fb & SF_NOLOTEX))
								rv.fixedsec = bs;
							else
							if (!(ff & SF_NOLOTEX) && fb & SF_NOLOTEX)
								rv.fixedsec = fs;
						}
					}

				}
				else if (rv.lfsec < 0
					&& rv.ssec->ceilsec < 0
					&& rv.ssec->floorsec < 0
					&& rv.ssec->lfsec < 0)
				{
					if (   bs->smoothfloor > fs->smoothfloor
						&& ff & SF_MOVFLOOR
						&& (fb & 0xFFFFC) == 0x1C)
					{
						rv.lfsec = bs - sectors;
					}
					else if (fs->smoothfloor > bs->smoothfloor
						&& fb & SF_MOVFLOOR
						&& (ff & 0xFFFFC) == 0x1C)
					{
						rv.lfsec = fs - sectors;
					}
				}

				// v220-8 added
				if (  !rv.fixedsec
					&& rv.sect == fs
					&& (fs->specFlags & 0xFFFFFFFC) == 0x0080001C)
				{
					if (   fs->ceilingheight > bs->ceilingheight
						&& bs->specFlags & SF_MOVCEIL)
					{
						rv.fixedsec = bs;
					}
					else if (fs->floorheight < bs->floorheight)
					{
						rv.fixedsec = bs;
					}
				}
			}

			if (seg->sgflags & DDSEGF_FACING && seg->backsector)
				subfn_DoSkyWall(seg);	// Possible sky fix or skywall
		}
		// v220-6 Test indicates unreferenced 2nd sector in subsec (rare)
		else if ( !(seg->sgflags & R3DSEGF_IGNORE) )
		{
			Rend_RenderWallSeg(seg);
		}
	}
}

extern RDBOOL textrans;

static void subfn_rendHiFlat(rendpoly_t *sspoly, short planepic, float cfp, RDBOOL trans, int flags)
{
	// Check for sky.
	if (planepic == skyflatnum) 
	{
		skyhemispheres |= SKYHEMI_UPPER;
		if (ma_inf.modifier & MODF_NOSKYMASK) return;

		sspoly->flags = RPF_SKY_MASK;

		if (!(rv.ssec->ssflags & SSF_NOSKYFIX))
		{
			// Set cfp to sect + skyfix height
			if (rv.skyheight > rv.sect->smoothceil)
			{
				rv.ssec->dlceilheight = rv.skyheight + rv.skyfix;
				cfp = FIX2FLT(rv.ssec->dlceilheight);
			}
			else
			{
				rv.ssec->dlceilheight = (int)(cfp*FRACUNIT);
			}
		}
	}
	else
	{
		sspoly->tex = GL_PrepareFlat(planepic);
		sspoly->detail = rendtexdetail;
		sspoly->texoffx = FIX2FLT(rv.csec->ceiloffx);
		sspoly->texoffy = FIX2FLT(rv.csec->ceiloffy);
		sspoly->texw = ctex->rendw;
		sspoly->texh = ctex->rendh;
		sspoly->vertices[0].rgba[CR] = (byte)R_SectorLight(rv.ceilinglightlevel);
		sspoly->flags = 0;

	/*	if (rendtexmask)	Masked doom flats are not supported
			sspoly->flags = RPF_MASKED;
		// $translucent_water
		else*/ if (textrans && trans)
		{
			if (fvy < cfp && ppt.view_subsector == rv.ssec) underwater = true;
			rv.fpwaterplane = cfp;
			rv.waterplane = (int)(cfp * FRACUNIT);
			if (!(rv.ssec->sector->specFlags2 & SF2_FAKETPSEC))
				sspoly->flags = RPF_TRANS;
			rv.ssec->ssflags |= SSF_FAKETRANS;
		}
		else
		{
			if (R_CheckFlatGlowFlag(planepic))
				sspoly->flags = RPF_PLANE_GLOW;

			if (rv.ssec->ssflags & DDSUBF_DLIGHT_CEILING && !(flags & RLPF_NOTLIT))
				sspoly->flags |= RPF_DLIT;
		}
	}

	sspoly->top = cfp;
	RL_PrepareFlat(sspoly, 0, 0, flags, rv.ssec);
	RL_AddPolyFlat(sspoly);
}

static void subfn_rendLoFlat(rendpoly_t *sspoly, short planepic, float ffp, RDBOOL trans, int flags)
{
	if (planepic == skyflatnum)	// Check for sky floor
	{
		skyhemispheres |= SKYHEMI_LOWER;
		if (ma_inf.modifier & MODF_NOSKYMASK) return;
		if (rv.sect->heightsec >= 0)
		{
			if (rv.sect->floorpic == sectors[rv.sect->heightsec].floorpic)
			{
				float fmin = FIX2FLT(MIN_OF(rv.sect->floorheight, sectors[rv.sect->heightsec].floorheight));
				if (fmin != ffp) return;
			}
		}
		sspoly->flags = RPF_SKY_MASK;
	}
	else
	{
		sspoly->tex = GL_PrepareFlat(planepic);

		if (rv.ssec == ppt.view_subsector) sspoly->flags |= RPF_PLRINSSEC;

		if (   rv.sect->heightsec >= 0
			&& rv.sect->siflags & SIF_FLOOR
			&& sectors[rv.sect->heightsec].floorheight > rv.sect->floorheight
			&& ffp == FIX2FLT(rv.sect->smoothfloor))
		{
			// Don't scroll floor underneath the fake floor
			sspoly->texoffx = sspoly->texoffy = 0.0f;
		}
		else
		{
			sspoly->texoffx = FIX2FLT(rv.fsec->flooroffx);
			sspoly->texoffy = FIX2FLT(rv.fsec->flooroffy);
		}

		sspoly->detail = rendtexdetail;
		sspoly->texw = ctex->rendw;
		sspoly->texh = ctex->rendh;
		sspoly->vertices[0].rgba[CR] = (byte)R_SectorLight(rv.floorlightlevel);

	/*	if (rendtexmask)	Masked doom flats are not supported
		{
			sspoly->flags = RPF_MASKED;
		}
		// $translucent_water
		else*/ if (textrans && trans)
		{
			rv.fpwaterplane = ffp;
			rv.waterplane = (int)(ffp * FRACUNIT);
			sspoly->flags = RPF_TRANS;
			rv.ssec->ssflags |= SSF_FAKETRANS;
		}
		else
		{
			sspoly->flags = 0;

			if (R_CheckFlatGlowFlag(planepic))
				sspoly->flags = RPF_PLANE_GLOW;

			if (rv.ssec->ssflags & DDSUBF_DLIGHT_FLOOR)
				sspoly->flags |= RPF_DLIT;
		}
	}

	sspoly->top = ffp;
	RL_PrepareFlat(sspoly, 0, 0, flags, rv.ssec);
	RL_AddPolyFlat(sspoly);
}

// R3D 3D flats
static void subfn_Rend3DFlats()
{
	RDBOOL		fakefloor = rv.fakedsec && rv.fakedsec->smoothfloor > rv.sect->smoothfloor;
	RDBOOL		trans = rv.istrans && fakefloor;
	RDBOOL		floorlit = (rv.ssec->ssflags & DDSUBF_DLIGHT_FLOOR) != 0;
	RDBOOL		ceillit = (rv.ssec->ssflags & DDSUBF_DLIGHT_CEILING) != 0;
	float		smoothfloor, smoothceil;
	float		fakeflrht = fakefloor ? FIX2FLT(rv.fakedsec->smoothfloor) : -32767.f;
	float		cfp = fakefloor ? FIX2FLT(rv.sect->smoothceil ) : FIX2FLT(rv.sceil);
	float		ffp = fakefloor ? FIX2FLT(rv.sect->smoothfloor) : FIX2FLT(rv.sfloor);
	short		planepic = R_SkyFromLinedef(rv.sect, rv.sect->ceilingpic);
	int			*flatslist = rv.ssec->flats3DArray;
	int			flats3D_count = *(int*)flatslist++;
	int			i;

	sector_t	*lsec;
	sector_t	*rsec;
	RDBOOL		alpha;
	int			ll;
	int			offset3D;
	int			fdiff;

	// ------------------------------------------

	CheckRendered(cfp);	// add to check list
	CheckRendered(ffp);	// add to check list

	sspoly.flags = 0;

	ll = rv.ceilinglightlevel;

	if (rv.sect->ceilinglightsec >= 0)
		rv.ceilinglightlevel = sectors[rv.sect->ceilinglightsec].lightlevel;
	else
		rv.ceilinglightlevel = rv.sect->lightlevel;

	subfn_rendHiFlat(&sspoly, planepic, cfp, false, RLPF_REVERSE);

	rv.ceilinglightlevel = ll;

	// Preset the floor's planepic
	planepic = R_SkyFromLinedef(rv.sect, rv.fsec->floorpic);

	// v220-16 since masked texture are now rendered using alphacut this has been simplified
	for (i = 0; i < flats3D_count; i++)
	{
		lsec = &sectors[flatslist[0]];	// Sector with flats/lightlevel definitions
		rsec = &sectors[flatslist[1]];	// Rendering sector
		offset3D = flatslist[2];
		fdiff	 = flatslist[3];

		if (!fdiff && !lsec->ceillight3D && !lsec->floorlight3D)
		{
			flatslist += FLIST_ENTRIES;
			continue;
		}

		smoothceil = FIX2FLT(rsec->smoothfloor + offset3D);
		smoothfloor = smoothceil + FIX2FLT(fdiff);

		alpha = GL_Prepare3DFlat(&sspoly, lsec->ceilingpic, flatslist[0]);

		if ((alpha || fvy < smoothceil) && !CheckRendered(smoothceil))
		{
			if (lsec->ceilingpic == skyflatnum) 
			{
				sspoly.flags |= RPF_SKY_MASK;
				skyhemispheres |= SKYHEMI_UPPER;
			}
			else if (!alpha && R_CheckFlatGlowFlag(lsec->ceilingpic))
			{
				sspoly.flags |= RPF_PLANE_GLOW;
			}

			sspoly.texoffx = FIX2FLT(lsec->ceiloffx);
			sspoly.texoffy = FIX2FLT(lsec->ceiloffy);
			sspoly.top = smoothceil;
			sspoly.vertices[0].rgba[CR] = (byte)R_SectorLight(lsec->floorlight3D);
			sspoly.flags |= fvy < smoothceil ? RLPF_REVERSE : RLPF_NORMAL;
			if (alpha) sspoly.flags |= RPF_MASKED;
			RL_PrepareFlat(&sspoly, 0, 0, sspoly.flags, rv.ssec);
			if (!alpha && ceillit) sspoly.flags |= RPF_DLIT;
			RL_AddPolyFlat(&sspoly);
		}

		alpha = GL_Prepare3DFlat(&sspoly, lsec->floorpic, flatslist[0]);

		if ((alpha || fvy > smoothfloor) && !CheckRendered(smoothfloor))
		{
			if (lsec->floorpic == skyflatnum) 
			{
				sspoly.flags |= RPF_SKY_MASK;
				skyhemispheres |= SKYHEMI_LOWER;
			}
			else if (!alpha && R_CheckFlatGlowFlag(lsec->floorpic))
			{
				sspoly.flags |= RPF_PLANE_GLOW;
			}

			sspoly.texoffx = FIX2FLT(lsec->flooroffx);
			sspoly.texoffy = FIX2FLT(lsec->flooroffy);
			sspoly.top = smoothfloor;
			sspoly.vertices[0].rgba[CR] = (byte)R_SectorLight(lsec->ceillight3D);
			sspoly.flags |= fvy > smoothfloor ? RLPF_NORMAL : RLPF_REVERSE;
			if (alpha) sspoly.flags |= RPF_MASKED;
			RL_PrepareFlat(&sspoly, 0, 0, sspoly.flags, rv.ssec);
			if (!alpha && floorlit) sspoly.flags |= RPF_DLIT;
			RL_AddPolyFlat(&sspoly);
		}

		flatslist += FLIST_ENTRIES;
	}

	ll = rv.floorlightlevel;

	if (trans || fakefloor)
	{
		if (fvy <= fakeflrht)
		{
			subfn_rendHiFlat(&sspoly, rv.fakedsec->ceilingpic, fakeflrht, rv.istrans, RLPF_REVERSE);
		}
		else
		{
			subfn_rendLoFlat(&sspoly, planepic, fakeflrht, rv.istrans, RLPF_NORMAL);
		}

		if (rv.pseudo)
		{
			subfn_rendLoFlat(&sspoly, planepic, ffp, false, RLPF_NORMAL);
		}
		else
		if (trans || fvy <= fakeflrht)
		{
			rv.floorlightlevel = rv.fakedsec->lightlevel;
			subfn_rendLoFlat(&sspoly, rv.fakedsec->floorpic, ffp, trans, RLPF_NORMAL);
		}
	}
	else if (viewz >= rv.sfloor)
	{
		if (rv.sect->floorlightsec >= 0)
			rv.floorlightlevel = sectors[rv.sect->floorlightsec].lightlevel;

		subfn_rendLoFlat(&sspoly, planepic, FIX2FLT(rv.sfloor), false, RLPF_NORMAL);

	}

	rv.floorlightlevel = ll;
}

static void subfn_CheckCeilFloor()
{
	if (rv.fixedsec)
	{
		rv.csec = rv.fixedsec;

		if (   rv.fixedsec->smoothfloor != rv.fsec->smoothfloor
			|| rv.fsec->specFlags & SF_NOLOTEX)
		{
			rv.fsec = rv.fixedsec;
		}

		if (rv.fsec->heightsec >= 0)
			rv.fsec = &sectors[rv.fsec->heightsec];

		if (rv.csec->specFlags & SF_MOVCEIL)
			rv.sceil = rv.fsec->smoothceil;

		rv.sfloor = rv.fsec->smoothfloor;
		rv.floorlightlevel = rv.ceilinglightlevel = subfn_LLevel(rv.fsec, 0);

		return;
	}

	if (rv.ssec->ceilsec >= 0)
	{
		rv.csec = R_FakeFlat(&sectors[rv.ssec->ceilsec], &rv.cfakesec,
			NULL, &rv.ceilinglightlevel);
		rv.sceil = rv.csec->smoothceil;
	}
	else if (!faked)
	{
		if (rv.csec->R3Dceilsec >= 0)
		{
			rv.csec = &sectors[rv.csec->R3Dceilsec];
			rv.sceil = rv.csec->smoothceil;
			rv.ceilinglightlevel = subfn_LLevel(rv.csec, 1);
		}
		else if (rv.csec->R3Dceilheight > DDMININT)
		{
			rv.sceil = rv.csec->R3Dceilheight;
		}
	}

	if (rv.ssec->floorsec >= 0)
	{
		rv.fsec = R_FakeFlat(&sectors[rv.ssec->floorsec], &rv.ffakesec,
				&rv.floorlightlevel, NULL);
		rv.sfloor = rv.fsec->smoothfloor;

		if ( !(rv.ssec->ssflags & SSF_USEFSEC)
			&& rv.sfloor == rv.ssec->sector->smoothfloor
			&& rv.ssec->sector->specFlags & SF_MOVFLOOR
			&& !rv.ssec->sector->floordata)	// Floor must be inactive
		{
			rv.fsec = rv.ssec->sector;
			rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
		}
	}
	else if (rv.fsec->floodline >= 0)
	{
		sector_t *s = lines[rv.fsec->floodline].frontsector;
		int fh = s->R3Dfloorheight < DDMAXINT ? s->R3Dfloorheight : s->smoothfloor;

		if (rv.sfloor < fh)
		{
			rv.fsec   = s;
			rv.sfloor = fh;
		}
	}
	else if (!faked)
	{
		// Note: R3Dfloorheight will override R3Dfloorsec
		if (rv.fsec->R3Dfloorheight < DDMAXINT)
		{
			rv.sfloor = rv.fsec->R3Dfloorheight;

			if (   rv.sect->specFlags & SF_MOVFLOOR
				&& rv.sfloor == rv.sect->smoothfloor)
			{
				rv.floorlightlevel = rv.sect->lightlevel;
			}
		}
		else if (rv.fsec->R3Dfloorsec >= 0)
		{
			sector_t *temp = &sectors[rv.fsec->R3Dfloorsec];
			int nf = temp->smoothfloor;

			// Detect 'trick' sector
			if (temp->specFlags & SF_MOVSECTOR)
			{
				if (nf > rv.sfloor && viewz > nf)
				{
					rv.fsec = temp;
					rv.sfloor = nf;
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				}
			}
			else	// standard setting
			{
				rv.fsec = temp;
				rv.sfloor = nf;
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
			}
		}
	}
}

// Discharge wall list
static void R_TransferList()
{
	rendpolyq_t	*ptr;
	seg_t		*seg;
	int			i;

	for (i = 0; i < plist.count; i++)
	{
		ptr = &plist.list[i];
		seg = plist.segptr[i];

		if (   (seg->linedef && (!seg->linedef->backsector || seg->linedef->xtraflags & XT_OCCLUDE))
			|| (ptr->flags & RPF_OCCLUDE && !rv.inhibit))
		{
			C_AddViewRelSeg(seg);
		}

		// If the quad was not flagged with RPF_QUADSET continue
		if (!(ptr->flags & RPF_QUADSET)) continue;

		// Not strictly necessary as these flags
		// are not used by the rend_list code.
		ptr->flags &= ~(RPF_QUADSET|RPF_OCCLUDE);

		if (ptr->flags & RPF_SKY_MASK && seg->linedef->ovflags & OVF_NOSKYWALL)
			continue;

		if (ptr->top > ptr->bottom)
		{
			if ( !(ptr->flags & (RPF_SKY_MASK|RPF_MASKED|RPF_TRANS|RPF_FOGWALL)) )
			{
				if (seg->sgflags & DDSEGF_DLIGHT)
				{
					ptr->flags |= RPF_DLIT;
				}
			}

			RL_AddPolyWall(ptr, seg);
			th_items[TH_WALLTEX]++;
		}
	}
}

static void Rend_RenderSubsector(subsector_t *ssec)
{
	// sfloorfp2 is for 2D rendering - 3D flats are on their own list
	seg_t			*seg;
	float			sfloorfp, sceilfp, lcsecfp;
	float			roff;
	RDBOOL			trans = false;
	RDBOOL			dir;
	int				i, lffht, fakesec;
	int				sfl;	//slopeflags
	short			planepic;
	short			cllight;

	//--------------------------------

	// Reset dynamic lighting/heightsec visibility flags.
	ssec->ssflags &= ~DDSUBF_CLEAR_MASK;

	if (GLOBAL_SKY_SCROLL_LINE < 0 && skl[ssec->sector - sectors].sky_line >= 0)
		GLOBAL_SKY_SCROLL_LINE = skl[ssec->sector - sectors].sky_line;

	if (!firstsubsector)	// The first subsec is not clipped
	{
		if (!C_CheckBspBox(ssec->ssbox)) return;
	}
	else
	{
		firstsubsector = false;
	}

	// Check the player's group belongs
	if (!M_GroupVis(ssec))
	{
		th_items[TH_SUBSECBYPASSED]++;
		return;
	}

	rv.ssec = ssec;
	rv.inhibit = (rv.ssec->ssflags & SSF_NOOCCLUDE);
	rv.sect = rv.usesec = rv.ssec->sector;

	vssecoverfake = vssecunderfake = false;

	if (!rv.inhibit && rv.sect->smoothfloor >= rv.sect->smoothceil)
	{
		seg = &segs[rv.ssec->firstline];

		// intially only need to check the first seg
		if ( !(seg->frontsector->specFlags & SF_NOOCCLUDE) )
		{
			for (i = 0; i < rv.ssec->linecount; i++, seg++)
			{
				seg->sgflags &= ~DDSEGF_CLEARMASK;

				if (  (seg->frontsector == rv.sect
					|| seg->backsector && seg->backsector == rv.sect)
					&& Rend_SegFacingDir(seg))
				{
					C_AddViewRelSeg(seg);
				}
			}
			return;
		}
	}

	//------------------------ sector visible

	sfl = rv.sect->specFlags2 & (SF2_LOSLOPE|SF2_HISLOPE);
	rv.facingsegcount = 0;
	th_items[TH_SUBSEC]++;

	// Visibility now established
	// Note: SIF_VISIBLE is cleared before each frame. SIF_WASSEEN is not.
	rv.sect->siflags |= (SIF_VISIBLE|SIF_WASSEEN);

	if (outsidefog)
		setNoFog = !(rv.sect->specFlags2 & SF2_SETFOG);
	else
		setNoFog = true;

	// Things can be added following subsector scanning
	if (outsidefog && setNoFog)
		ssec->ssflags |= SSF_FOGUNSET;
	else
		ssec->ssflags &= ~SSF_FOGUNSET;

	faked = false;
	rend_count  = 0;
	plist.count = 0;
	rv.pseudo = false;

	if (rv.ssec->floorsec >= 0 && rv.ssec->floorsec == rv.ssec->ceilsec)
	{
		rv.common = rv.ssec->floorsec;
		rv.sect = rv.fsec = rv.csec = &sectors[rv.ssec->floorsec];
	}
	else
	{
		rv.common = -1;
		rv.sect = rv.ssec->sector;
		rv.fsec = R_FakeFlat(rv.sect, &rv.ffakesec, &rv.floorlightlevel, NULL  );
		rv.csec = R_FakeFlat(rv.sect, &rv.cfakesec, NULL, &rv.ceilinglightlevel);
	}

	rv.faked = faked;
	rv.sceil  = rv.csec->smoothceil;
	rv.sfloor = rv.fsec->smoothfloor;
	rv.ceil2    = DDMININT;
	rv.skyfix   = rv.csec->skyfix;
	rv.lfsec    = rv.ssec->lfsec;
	rv.fixedsec = NULL;
	rv.fakedsec = (fakesec = rv.sect->heightsec) >= 0 ? &sectors[fakesec] : NULL;
	rv.waterplane = DDMININT;
	rv.skyheight = DDMININT;

	// Check if a flooroffset required
	if (   rv.sect->R3Dfloorheight < DDMAXINT
		&& rv.sect->floorpic == skyflatnum)
	{
		rv.setfloorheight = rv.sect->R3Dfloorheight;
	}
	else
		rv.setfloorheight = DDMAXINT;

	// v220-4 may need resetting
	if (fakesec >= 0  || (rv.ssec->floorsec >= 0 && (fakesec = sectors[rv.ssec->floorsec].heightsec) >= 0) )
	{
		rv.ssec->s_waterplane = sectors[fakesec].smoothfloor;
	}
	else
		rv.ssec->s_waterplane = DDMININT;

	if (R3D_ScriptedMap)
	{
		if (rv.fakedsec && rv.fakedsec->wvar)
			water_opacity = rv.fakedsec->wvar->opacity;
		else
			water_opacity = 0xFF;

		rv.istrans = water_opacity != 0xFF;
	}
	else
	{
		rv.istrans = water_trans && water_opacity != 0xFF;
	}

	// Prepare the walls ----------------------------------

	// Set/preset seg flags, check for changed settings and set rv.skyfix
	for (i = 0, seg = &segs[rv.ssec->firstline]; i < rv.ssec->linecount; i++, seg++)
	{
		seg->sgflags &= ~DDSEGF_CLEARMASK;

		if (Rend_SegFacingDir(seg) && C_CheckViewSeg(seg))
		{
			seg->sgflags |= DDSEGF_FACING;

			if (seg->linedef)
			{
				// flag used in automap
				seg->linedef->ml_flags |= ML_MAPPED;

				if (   numLuminous
					&& !(seg->linedef->ml_flags & (ML_VISPRESENT|ML_BADLINE))
					&& seg->length >= 1.0f)
				{
					rv.facingsegptr[rv.facingsegcount++] = seg;
				}
			}
		}
	}

	for (i = 0, seg = &segs[rv.ssec->firstline]; i < rv.ssec->linecount; i++, seg++)
	{
		if (!seg->linedef || seg->linedef->specType == R3D_3F) continue;

		if (seg->linedef->ml_flags & ML_VISPRESENT)
		{
			if (   !(seg->sgflags & DDSEGF_FACING)
				|| !seg->sidedef->midtexture
				|| seg->sidedef->lsiflags & LS_MIDTEX)	// i.e. if not already drawn
			{
				continue;	// nothing to render
			}

			dir = seg->sidedef - sides == seg->linedef->sidenum[0];
			setVISquad(seg->linedef, seg->sidedef->midtexture, dir, seg);
			quad->texoffy = 0.0f;

			if (seg->sidedef->lsiflags & LS_ORIGTEXOFF)
				quad->texoffx = ((float)seg->sidedef->orig_textureoffset)*ctex->texscalex;
			else
				quad->texoffx = FIX2FLT(seg->sidedef->textureoffset)*ctex->texscalex;

			if (ctex->high < 0)	// v220-33
				roff = (FIX2FLT(seg->sidedef->rowoffset) - ctex->high + ctex->offset)*ctex->texscaley;
			else if (ctex->high2 && !(seg->linedef->ml_flags & ML_DONTPEGBOTTOM))	// v220-33
				roff = (FIX2FLT(seg->sidedef->rowoffset) + ctex->high2 + ctex->offset)*ctex->texscaley;
			else if (ctex->high > 0 && !(seg->linedef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP)))	// v220-33
				roff = (FIX2FLT(seg->sidedef->rowoffset) + ctex->high + ctex->high2 + ctex->offset)*ctex->texscaley;
			else
				roff = (FIX2FLT(seg->sidedef->rowoffset) + ctex->offset)*ctex->texscaley;

			if (seg->linedef->ml_flags & ML_DONTPEGBOTTOM)
			{
				quad->bottom = FIX2FLT(seg->sidedef->sector->smoothfloor) + roff;
				quad->top = quad->bottom + ctex->rendh;
			}
			else
			{
				quad->top = FIX2FLT(seg->sidedef->sector->smoothceil) + roff;
				quad->bottom = quad->top - ctex->rendh;
			}

			quad->flags |= (RPF_MASKED|RPF_NOSPLIT);

			rend_UpdateSide(seg->sidedef, LS_MIDTEX, 0);
		}
		else
		// filter out inaccurate vertex placement if same sector sided
		if (seg->length > 0.1 || seg->frontsector != seg->backsector)
		{
			if (seg->linedef->lnflags2 & LNF2_OCCLUDE)
				C_AddViewRelSeg(seg);	// has no other function
			else
				subfn_DoWallSeg(seg);
		}
	}

	if (   rv.ssec->ssflags & SSF_CHKFLOORDIFF
		&& rv.lfsec >= 0
		&& sectors[rv.lfsec].smoothfloor - rv.sect->smoothfloor < 64*FRACUNIT)
	{
		rv.lfsec = -1;
	}

	if (rv.sect->R3Dlowsec >= 0)
	{
		if (rv.lfsec < 0 || viewz < sectors[rv.lfsec].floorheight)
			rv.lfsec = rv.sect->R3Dlowsec;
	}

	subfn_CheckCeilFloor();

	// Check if a flooroffset required
	if (rv.setfloorheight < DDMAXINT)
	{
		rv.sfloor = rv.setfloorheight;
	}

	// Check if a ceiloffset required
	if (rv.csec->ceiloffset < DDMAXINT && !faked)
	{
		rv.sceil = rv.csec->smoothfloor + rv.csec->ceiloffset;
	}

	sfloorfp = FIX2FLT(rv.sfloor);
	sceilfp = FIX2FLT(rv.sceil);

	// Must be set prior to glow calcs
	if (rv.lfsec >= 0 && viewz <= (lffht = sectors[rv.lfsec].smoothfloor))	// added v2.0.01 $build 30
		rv.lfsec = -1;

	rv.ssec->dlfloorheight = rv.sfloor;
	rv.ssec->dlceilheight  = vssecunderfake && rv.fakedsec ? rv.fakedsec->floorheight : rv.sceil;
	rv.ssec->ssfloorpic	   = rv.fsec->floorpic;
	rv.ssec->ssceilpic	   = vssecunderfake && rv.fakedsec ? rv.fakedsec->ceilingpic : rv.csec->ceilingpic;
	rv.ssec->dynlight	   = R_SectorLight(rv.floorlightlevel);
	rv.overfake            = rv.sect->oversec >= 0 ? &sectors[sectors[rv.sect->oversec].heightsec] : NULL;

	if (rv.overfake)
	{
		if (rv.sect->smoothfloor < rv.overfake->smoothfloor)
		{
			if (   water_trans
				&& !rv.istrans
				&& rv.overfake->specFlags2 & SF2_HASWATERFLATS)
			{
				if (R3D_ScriptedMap && rv.overfake->wvar)
				{
					water_opacity = rv.overfake->wvar->opacity;
					rv.istrans = water_opacity != 0xFF;
				}
				else
					rv.istrans = true;
			}
		}
		else
			rv.overfake = NULL;
	}

	if (   numLuminous	// v220-35
		&& !(rv.ssec->ssflags & SSF_NOGLOW)
	//	&& rv.sect->ceiloffset == DDMAXINT	// no glow with SPAWNVDOOR - $remove v220-35
		&& rv.ssec->dlceilheight > rv.ssec->dlfloorheight)
	{
		// Must be done before sending floor/wall polys
		DL_ProcessSubsector(&sspoly, rv.ssec, rv.facingsegcount, rv.facingsegptr);
	}

	// Prepare the floor and ceiling flats ----------------------------------

	// v220-3 Set up conditions for water translucency
	// $translucent_water

	float inhibitoverflat = 32767.f;

	if (rv.ssec->flats3DArray)
	{
		// 3D ceilings and floors
		subfn_Rend3DFlats();
	}
	else
	{
		sspoly.flags = 0;

		if (rv.lfsec >= 0 && rv.sect->withheight < DDMAXINT && viewz > lffht)
		{
			sfloorfp = FIX2FLT(lffht);

			if (!CheckRendered(sfloorfp))
			{
				sector_t *tmp = &sectors[rv.lfsec];

				sspoly.tex = GL_PrepareFlat(tmp->floorpic);
				sspoly.texoffx = FIX2FLT(tmp->flooroffx);
				sspoly.texoffy = FIX2FLT(tmp->flooroffy);
				sspoly.detail = rendtexdetail;
				sspoly.texw = ctex->rendw;
				sspoly.texh = ctex->rendh;
				sspoly.vertices[0].rgba[CR] = (byte)R_SectorLight(rv.floorlightlevel);
				sspoly.top = sfloorfp;
				RL_PrepareFlat(&sspoly, 0, 0, RLPF_NORMAL, rv.ssec);
				RL_AddPolyFlat(&sspoly);
				rv.lfsec = -1;
			}
		}

		if ( !(rv.ssec->ssflags & SSF_INHSECLIGHT)
			&& rv.ssec->sslightsec >= 0)
		{
			if (rv.ssec->lcsec >= 0)	// b211-0 If set use sector's lightlevel
			{
				rv.floorlightlevel = sectors[rv.ssec->sslightsec].lightlevel;

				// v220-10 $light !rv.faked added
				if (!rv.faked && rv.ssec->lfsec >= 0)	// v220-6 both lcsec and lfsec must be set
					rv.ceilinglightlevel = sectors[rv.ssec->sslightsec].lightlevel;
			}
		}

		// Check for lfsec overriding sect.
		if (   rv.lfsec >= 0
			&& !faked
			&& rv.sect->withheight == DDMAXINT
			&& rv.sect->specFlags & SF_MOVFLOOR
			&& !(sectors[rv.lfsec].specFlags & SF_MOVFLOOR)
			&& (lffht > rv.sect->smoothfloor || (rv.ssec->ssflags & SSF_RENDLFSEC)) )
		{
			sfloorfp = FIX2FLT(lffht);

			if (fvy > sfloorfp && !CheckRendered(sfloorfp)) 
			{
				sspoly.flags = 0;
				planepic = sectors[rv.lfsec].floorpic;
				rv.floorlightlevel = sectors[rv.lfsec].lightlevel;

				if ( !(planepic == skyflatnum && rv.fsec->specFlags & SF_MOVFLOOR) )
				{
					subfn_rendLoFlat(&sspoly, planepic, sfloorfp, false, RLPF_NORMAL);
				}
			}
			rv.lfsec = -1;	// Don't draw secondary floor
		}
		else if (fvy > sfloorfp || rv.sect->stptr >= 0) 
		{
			int flags = rv.sect->specFlags2 & SF2_LOSLOPE ? RLPF_NORMAL|RLPF_TILT : RLPF_NORMAL;		// $slopes

			// R3D moving water floor
			if (rv.fakedsec && rv.fakedsec->specFlags2 & SF2_FAKEMOVSECTOR)
			{
				rv.ssec->ssfloorpic = rv.fakedsec->floorpic;

				// Add the sector's movesec floor
				if (rv.fakedsec->smoothfloor > rv.sect->smoothfloor)
				{
					if (   viewz > rv.fakedsec->smoothfloor
						&& !CheckRendered( (sfloorfp = FIX2FLT(rv.fakedsec->smoothfloor)) ))
					{
						subfn_rendLoFlat(&sspoly, rv.sect->floorpic, sfloorfp, rv.istrans, RLPF_NORMAL);
						rv.ssec->ssfloorpic = rv.sect->floorpic;
						rv.ssec->dlfloorheight = rv.fakedsec->smoothfloor;
					}
				}

				// draw the faked sector floor
				sfloorfp = FIX2FLT(rv.sect->smoothfloor);

				if (fvy > sfloorfp && !CheckRendered(sfloorfp))
					subfn_rendLoFlat(&sspoly, rv.fakedsec->floorpic, sfloorfp, false, flags);

				trans = false;
			}
			else
			{
				// v220-3 Check for floor going under a fake liquid floor
				// This is visually expected despite the move sector having no fake sector.
				if (rv.overfake)
				{
					int		ht = rv.overfake->smoothfloor;

					if (ht > rv.sfloor)
					{
						float	sfloorfp2 = FIX2FLT(ht);
						int		faketrans = water_trans;
						int		opacity = water_opacity;	// Remember current value

						CheckRendered(sfloorfp2);	// Add to list
						planepic = sectors[rv.sect->oversec].floorpic;

						if (R3D_ScriptedMap)
						{
							if (rv.overfake->wvar)
							{
								water_opacity = rv.overfake->wvar->opacity;
								faketrans = water_opacity != 0xFF;
							}
						}
						else if (faketrans)	// v220-29
						{
							faketrans = rv.overfake->specFlags2 & SF2_HASWATERFLATS;
						}

						if (viewz > ht)
						{
							subfn_rendLoFlat(&sspoly, planepic, sfloorfp2, faketrans, RLPF_NORMAL);
							rv.ssec->dlfloorheight = ht;
							rv.ssec->ssfloorpic = planepic;
						}
						else
							subfn_rendHiFlat(&sspoly, planepic, sfloorfp2, faketrans, RLPF_REVERSE);

						// v220-4 used with footsteps/particles
						rv.ssec->s_waterplane = ht;
						water_opacity = opacity;
					}
				}

				if (flags & RLPF_TILT && rv.fakedsec && !underwater)
				{
					float sfloorfp2 = FIX2FLT(rv.sect->smoothfloor);

					if (rv.fakedsec->specFlags2 & SF2_PSEUDOFAKE)
						subfn_rendLoFlat(&sspoly, rv.sect->floorpic, sfloorfp2, false, flags);
					else
					if (fvy >= sfloorfp2)
						subfn_rendLoFlat(&sspoly, rv.fakedsec->floorpic, sfloorfp2, false, flags);

					CheckRendered(sfloorfp2);	// Add to list
					flags = RLPF_NORMAL;	// Don't tilt the following fake floor
				}

				// $v16added
				if (viewz > rv.sect->withheight)	// Note: DDMAXINT if unset
				{
					sfloorfp = FIX2FLT(rv.sect->withheight);
					planepic = rv.sect->texwithheight[0];
				}
				else
				{
					planepic = R_SkyFromLinedef(rv.sect, rv.fsec->floorpic);
				}

				if (   !(planepic == skyflatnum && rv.lfsec >= 0 && rv.fsec->specFlags & SF_MOVFLOOR)
					&& !CheckRendered(sfloorfp))
				{
					int fls = -1;

					if (rv.fakedsec && vssecoverfake)
					{
						subfn_rendLoFlat(&sspoly, rv.fakedsec->floorpic, FIX2FLT(rv.fakedsec->ceilingheight), false, RLPF_NORMAL);
					}
					else if (fvy > sfloorfp || sfl & SF2_LOSLOPE)
					{
						// $translucent_water
						trans = rv.istrans
							&& rv.fakedsec != NULL
							&& !underwater
							&& sfloorfp > FIX2FLT(rv.sect->floorheight)
							&& !(rv.fakedsec->specFlags2 & SF2_FAKEDLOWMISS);

						subfn_rendLoFlat(&sspoly, planepic, sfloorfp, trans, flags);
						inhibitoverflat = sfloorfp;
					}
				}
			}
		}

		// A possible scripted dummy floor
		if (rv.sect->specFlags2 & SF2_DUMMYFLOOR)
		{
			int fht;

			rv.fsec = &sectors[rv.sect->R3Dfloorsec];
			fht = rv.fsec->smoothfloor;
			sfloorfp = FIX2FLT(fht);

			if (fvy > sfloorfp && !CheckRendered(sfloorfp)) 
			{
				if (rv.fsec->heightsec >= 0)
					planepic = sectors[rv.fsec->heightsec].floorpic;
				else
					planepic = rv.fsec->floorpic;

				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				rv.ssec->ssfloorpic	= planepic;
				rv.ssec->dlfloorheight = fht;

				sspoly.flags = 0;

				subfn_rendLoFlat(&sspoly, planepic, sfloorfp, false, RLPF_NORMAL);
			}
		}
		
		// A possible secondary floor
		// The second test will normally pass as withheight, if unset, will be
		// DDMAXINT. If set then the player's height must be < withheight as
		// this type relies on the the sector being in it's lower position.
		if (rv.lfsec >= 0 && rv.sect->withheight == DDMAXINT)
		{
			int fht;

			rv.fsec = R_FakeFlat(&sectors[rv.lfsec], &rv.ffakesec, NULL, NULL);

			fht = rv.fsec->smoothfloor;

			sfloorfp = FIX2FLT(fht);

			if (fvy > sfloorfp && !CheckRendered(sfloorfp)) 
			{
				int flags = RLPF_NORMAL;

				// v220-7 always use the lfsec's light level
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);

				if (viewz < rv.sfloor)
				{
					rv.ssec->ssfloorpic	= rv.fsec->floorpic;
					rv.ssec->dlfloorheight = fht;
				}
				else
				// v220-18 store the highest visible flat
				if (fht >= rv.ssec->dlfloorheight && viewz > fht)
				{
					rv.ssec->ssfloorpic	= rv.fsec->floorpic;
					rv.ssec->dlfloorheight = fht;
				}

				sspoly.flags = 0;
				planepic = rv.fsec->floorpic;

				if ( !(planepic == skyflatnum && rv.fsec->specFlags & SF_MOVFLOOR) )
				{
					trans = sectors[rv.lfsec].heightsec >= 0
						&& fht == sectors[sectors[rv.lfsec].heightsec].floorheight;	// $demonizd

					subfn_rendLoFlat(&sspoly, planepic, sfloorfp, trans && water_trans, flags);

					if (trans) rv.ssec->s_waterplane = fht;
				}
			}
		}

		// Draw floors below a faked floor if water translucency is set
		// v220-31 3rd test removed
		if (   rv.fakedsec
			&& !(rv.sect->specFlags2 & SF2_LOSLOPE)/*
			&& (trans || (water_trans && rv.fakedsec->smoothfloor > rv.sect->smoothfloor))*/ )
		{
			sfloorfp = FIX2FLT(rv.sect->smoothfloor);

			if (sfloorfp < inhibitoverflat && fvy > sfloorfp && !CheckRendered(sfloorfp))
			{
				rv.fsec = rv.fakedsec;
				rv.floorlightlevel = subfn_LLevel(rv.fakedsec, 0);
				planepic = rv.pseudo ? rv.sect->floorpic : rv.fakedsec->floorpic;
				subfn_rendLoFlat(&sspoly, planepic, sfloorfp, false, RLPF_NORMAL);
			}
		}

		if (rv.ssec->ssfloorpic < 0)
			rv.ssec->ssfloorpic = rv.fsec->floorpic;

// The sector's ceiling

		RDBOOL inhibitceiling = false;

		if (   rv.ssec->ssflags & SSF_FORCESKY
			&& rv.ssec->ceilsec >= 0
			&& sectors[rv.ssec->ceilsec].ceilingheight >= rv.sect->ceilingheight)
		{
			// The sector's ceiling
			planepic = skyflatnum;
			sceilfp = FIX2FLT(rv.sect->smoothceil + rv.sect->skyfix);

			if (fvy < sceilfp && !CheckRendered(sceilfp) )
			{
				subfn_rendHiFlat(&sspoly, planepic, sceilfp, false, RLPF_REVERSE);
			}
		}
		else
		{
			planepic = R_SkyFromLinedef(rv.usesec, rv.csec->ceilingpic);
			// v220-22 lcsecfp added to allow testing that a sector with a sky ceiling
			//         of the same height, where lcsec is set, does not override it.
			lcsecfp = rv.ssec->lcsec >= 0 ? FIX2FLT(sectors[rv.ssec->lcsec].smoothceil) : 32768.f;

			int flags = rv.csec->specFlags2 & SF2_HISLOPE		// $slopes
					&& planepic != skyflatnum ?
					RLPF_REVERSE|RLPF_TILT : RLPF_REVERSE;

			if (rv.fakedsec)
			{
				float	sceilfp2 = FIX2FLT(rv.fakedsec->smoothfloor);
				sceilfp = FIX2FLT(rv.sect->smoothceil + rv.sect->skyfix);
		
				if (fvy < sceilfp2)
				{
					int fpic;

					if (   vssecunderfake
						&& sceilfp >= sceilfp2
						&& (fpic = rv.pseudo ? rv.fakedsec->floorpic : rv.fakedsec->ceilingpic) != skyflatnum
						&& !CheckRendered(sceilfp2))
					{
						subfn_rendHiFlat(&sspoly, fpic, sceilfp2, sceilfp > sceilfp2 && water_trans && rv.istrans, RLPF_REVERSE);
					}
					else if (rv.fakedsec->specFlags2 & SF2_FAKEMOVSECTOR)
					{
						if (   underwater
							&& rv.fakedsec->floorheight > rv.sect->floorheight
							&& !CheckRendered(sceilfp2))
						{
							subfn_rendHiFlat(&sspoly, rv.sect->floorpic, sceilfp2, rv.istrans, RLPF_REVERSE);
						}
					}
					else
					// v220-33 added
					if (   vssecunderfake
						&& !CheckRendered(sceilfp2))
					{
						subfn_rendHiFlat(&sspoly, rv.fakedsec->floorpic, sceilfp2, water_trans && rv.istrans, RLPF_REVERSE);
					}
				}

				sceilfp2 = FIX2FLT(rv.fakedsec->smoothceil);

				RDBOOL rend1 = CheckRendered(sceilfp);
				RDBOOL rend2 = sceilfp != sceilfp2 ? CheckRendered(sceilfp2) : rend1;

				if (!rend1 || !rend2)	// Add to list
				{
					if (!rend2)
					{
						if (rv.sect->specFlags2 & SF2_FAKEDHIGHMISS)
						{
							if (rv.sect->ceilingpic == skyflatnum)	// v220-35 real sec sky takes precedence
							{
								planepic = rv.sect->ceilingpic;
							}
							else if (  rv.fakedsec->ceilingheight > rv.sect->ceilingheight
								&& viewz < rv.sect->ceilingheight)
							{
								planepic = rv.sect->ceilingpic;
							}
							else
							{
								planepic = rv.fakedsec->ceilingpic;
							}

							subfn_rendHiFlat(&sspoly, planepic, sceilfp2, false, flags);
							rend2 = true;
							if (sceilfp == sceilfp2) goto endrend12;
						}
						// v220-35 test added for case where fake ceiling is higher than real ceiling
						else if (!rend2
							&& !compat[c_fake]
							&& sceilfp2 > sceilfp
							&& rv.ssec->ceilsec < 0
							&& viewz < rv.fakedsec->ceilingheight)
						{
							// Use the real sector's ceilingpic
							subfn_rendHiFlat(&sspoly, rv.sect->ceilingpic, sceilfp2, false, flags);
							rend1 = rend2 = true;	// Don't draw lower ceiling
							goto endrend12;
						}
					}

					if (vssecoverfake)
					{
						if (!rend1)
						{
							subfn_rendHiFlat(&sspoly, rv.fakedsec->ceilingpic, sceilfp, false, flags);
							rend1 = true;
							if (sceilfp == sceilfp2) goto endrend12;
						}
					}
					else
					{
						RDBOOL isrendered = rend1;

						if (!rend2 && !rv.pseudo && fvy < sceilfp2)
						{
							if (rv.fakedsec->ceilingheight < rv.sect->ceilingheight)
							{
								// under fake ceiling so use control sector's ceiling pic unless its sky
								if (!rend2)
								{
									if (rv.sect->ceilingpic != skyflatnum)
										subfn_rendHiFlat(&sspoly, rv.sect->ceilingpic, sceilfp2, false, flags);
								}

								isrendered = true;
							}
							else if (!R3D_ScriptedMap && sceilfp2 > sceilfp)
							{
								if (fvy < sceilfp)
								{
									if (   rv.fakedsec->ceilingheight > rv.sect->ceilingheight
										&& viewz < rv.sect->ceilingheight)
									{
										planepic = rv.sect->ceilingpic;
										sceilfp = FIX2FLT(rv.fakedsec->smoothceil); // v220-35 added $hschange
									}
									else
									{
										planepic = rv.fakedsec->ceilingpic;
									}
									subfn_rendHiFlat(&sspoly, planepic, sceilfp, false, flags);
								}
								inhibitceiling = isrendered = true;
							}
						}

						if (!isrendered)
						{
							if (rv.sect->specFlags2 & SF2_FAKEDHIGHMISS)
							{
								 if (!rend2 && fvy < sceilfp2)
									 subfn_rendHiFlat(&sspoly, rv.sect->ceilingpic, sceilfp2, false, flags);
							}
							else if (fvy < sceilfp || sfl & SF2_HISLOPE)
							{
								subfn_rendHiFlat(&sspoly, rv.sect->ceilingpic, sceilfp, false, flags);
							}
						}

						if (  !vssecunderfake
							&& rv.fakedsec && viewz < rv.ssec->s_waterplane
							&& rv.fakedsec->floorheight <= rv.sect->floorheight) // v220-35 added $hschange
						{
							sceilfp = FIX2FLT(rv.ssec->s_waterplane);
							if (fvy < sceilfp && !CheckRendered(sceilfp))
								subfn_rendHiFlat(&sspoly, rv.fakedsec->ceilingpic, sceilfp, water_trans && rv.istrans, RLPF_REVERSE);
						}
					}
				}
endrend12:;

				flags = RLPF_REVERSE;	// Don't tilt any following ceilings

				if (rv.ssec->ceilsec >= 0)
				{
					sector_t *cs = &sectors[rv.ssec->ceilsec];

					sceilfp2 = FIX2FLT(cs->smoothceil);

					if (fvy < sceilfp2 && !CheckRendered(sceilfp2))
					{
						subfn_rendHiFlat(&sspoly, cs->ceilingpic, sceilfp2, false, flags);
					}
				}
			}
			else if (rv.ssec->ceilsec >= 0 && sectors[rv.ssec->ceilsec].heightsec >= 0)	// $demonizd
			{
				float sceilfp2 = FIX2FLT(sectors[sectors[rv.ssec->ceilsec].heightsec].smoothfloor);
				if (fvy < sceilfp2 && !CheckRendered(sceilfp2))
				{
					subfn_rendHiFlat(&sspoly, sectors[sectors[rv.ssec->ceilsec].heightsec].ceilingpic, sceilfp2, water_trans && rv.istrans, RLPF_REVERSE);
				}
				planepic = rv.sect->ceilingpic;
				rv.ssec->s_waterplane = sectors[sectors[rv.ssec->ceilsec].heightsec].smoothfloor;
				trans = false;
			}
			else
			// $v16added
			if (rv.sect->texwithheight[1] && viewz < rv.sect->withheight)
			{
				sceilfp = FIX2FLT(rv.sect->withheight);
				planepic = rv.sect->texwithheight[1];
			}

			if (planepic != skyflatnum || rv.ssec->lcsec < 0)
			{
				if (planepic == skyflatnum)
				{
					if (!(rv.ssec->ssflags & SSF_NOSKYFIX2))
					{
						int a, h = rv.usesec->smoothceil;

						if (   rv.ssec->ceilsec >= 0
							&& sectors[rv.ssec->ceilsec].ceilingpic == skyflatnum
							&& (a = sectors[rv.ssec->ceilsec].smoothceil) > h)
						{
							// v220-33 ammend
							h = a + sectors[rv.ssec->ceilsec].skyfix;
						}
						else
							h += rv.sect->skyfix;

						sceilfp = FIX2FLT(h);
					}

					trans = false;
				}

				if (!CheckRendered(sceilfp) && (fvy < sceilfp || sfl & SF2_HISLOPE))
					subfn_rendHiFlat(&sspoly, planepic, sceilfp, trans && water_trans, flags);
			}
		}

		// A possible secondary ceiling
		if (rv.ssec->lcsec >= 0)
		{
			sector_t *s = &sectors[rv.ssec->lcsec];

			if (fvy < (lcsecfp = FIX2FLT(s->smoothceil)) )
			{
				planepic = s->ceilingpic;

				// v220-33 Ceiling may be changed by map
				if (planepic == skyflatnum)
					lcsecfp += FIX2FLT(rv.sect->skyfix);

				if (!CheckRendered(lcsecfp))
				{
					if (s->smoothceil < rv.ssec->dlceilheight)
						rv.ssec->dlceilheight = s->smoothceil;

					rv.ceilinglightlevel = subfn_LLevel(s, 1);	// v220-6
					subfn_rendHiFlat(&sspoly, planepic, lcsecfp, false, RLPF_NOTLIT|RLPF_REVERSE);
				}
			}
		}
		else
		// Draw ceilings above a faked ceiling
		if (   rv.fakedsec
			&& !inhibitceiling
			&& !(rv.sect->specFlags2 & SF2_HISLOPE)
			&& (!vssecunderfake || trans || (water_trans && rv.fakedsec->smoothceil <= rv.sect->smoothceil)) )
		{
			planepic = R_SkyFromLinedef(rv.sect, rv.sect->ceilingpic);

			// The sector's ceiling
			if (planepic == skyflatnum)
				sceilfp = FIX2FLT(rv.sect->smoothceil + rv.sect->skyfix);
			else
				sceilfp = FIX2FLT(rv.sect->smoothceil);

			if (fvy < sceilfp && !CheckRendered(sceilfp) )
			{
				rv.ceilinglightlevel = rv.sect->lightlevel;	// v220-16 set correct light level
				subfn_rendHiFlat(&sspoly, planepic, sceilfp, false, RLPF_REVERSE);
			}
		}

		// A skyflat hack ceiling to get a 1's skywall to be drawn in certain rare circumstances
		if (rv.ceil2 > DDMININT)
		{
			subfn_rendHiFlat(&sspoly, (short)skyflatnum, FIX2FLT(rv.ceil2), false, RLPF_NOTLIT|RLPF_REVERSE);
		}

		if (underwater && viewz < rv.ssec->s_waterplane)
		{
			rv.ssec->dlceilheight = rv.ssec->s_waterplane;
		}
	}

	{
	sector_t	*ls;
//	RDBOOL		cls = true;

//	if (rv.ssec->flats3DArray) cls = false;	// FIXME 3Dflat player/sprite light should not be affected by this

	// v220-21 player light setting
	if (rv.fakedsec && viewz < rv.fakedsec->floorheight)
	{
		ls = rv.fakedsec; // cls = false;
	}
	else if (rv.ssec->floorsec >= 0)
	{
		ls = &sectors[rv.ssec->floorsec];
	}
	else if (rv.ssec->lfsec >= 0 && sectors[rv.ssec->lfsec].floorheight < rv.sect->floorheight)
	{
		ls = &sectors[rv.ssec->lfsec];
	}
	else
	{
		ls = rv.sect;
	}

//	if (cls && ls->floorlightsec >= 0 && !ma_inf.useSectorLight)
//		ls = &sectors[ls->floorlightsec];

	cllight = R3D_ScriptedMap ? ls->lightlevel : SectorDoomLight(ls->lightlevel);
	}

	if (rv.ssec == ppt.view_subsector)
	{
		ppt.viewer_lightlevel = cllight;			// used with psprite not model
	}

	fixed_t fakefloor = rv.fakedsec && !(rv.ssec->ssflags & SSF_FAKETRANS) ? rv.fakedsec->floorheight : DDMININT;

	R_AddSprites(rv.ssec, fakefloor, cllight);

	// R3D - Moved here as translucent flats now come in to play
	if ( !(rv.sect->siflags & SIF_RENDERED) )
	{
		rv.sect->siflags |= SIF_RENDERED;

		// Mark the particle generators in the sector visible.
		PG_SectorIsVisible(rv.sect - sectors);
	}

	R_TransferList();	// Send the list of (wall) quads

	// v220-25 Transfer any visible *xlinemem lines
	if (ssec->xlines)
	{
		xlinemem_t	*lm;
		sector_t	*sc;
		int			dir;
		int			n;

		quad = &plist.list[0];
		quad->numvertices = 2;
		quad->type   = rp_quad;
		quad->light  = NULL;
		quad->detail = NULL;
		quad->texoffy = 0.0f;
		quad->divs[0].num = 0;
		quad->divs[1].num = 0;

		i = 0;

		while ( (n = ssec->xlines[i++]) >= 0)
		{
			if (xlisrendered[n]) continue;
			
			xlisrendered[n] = true;

			lm = &xlinemem[n];

			dir = P_PointOnViewSide(lm->startx, lm->starty, lm->dx, lm->dy);

			quad->length = lm->length;

			quad->vertices[ dir].pos[VX] = lm->startx;
			quad->vertices[ dir].pos[VY] = lm->starty;
			quad->vertices[!dir].pos[VX] = lm->endx;
			quad->vertices[!dir].pos[VY] = lm->endy;

			quad->vertices[0].vpdist = Rend_PointDist2D(quad->vertices[0].pos);
			quad->vertices[1].vpdist = Rend_PointDist2D(quad->vertices[1].pos);

			// set the texture parms
			quad->tex  = GL_PrepareTexture(lm->texn);
			quad->texh = ctex->rendh;
			quad->texw = ctex->rendw;
			quad->texoffx = lm->x_off[dir] * ctex->texscalex;

			if (lm->flags & ML_DONTPEGBOTTOM)
			{
				quad->bottom = lm->floorheight + lm->y_off * ctex->texscaley;
				quad->top = quad->bottom + ctex->rendh;
			}
			else
			{
				quad->top = lm->ceilheight + lm->y_off * ctex->texscaley;
				quad->bottom = quad->top - ctex->rendh;
			}

			sc = &sectors[lm->secn];
			RL_VertexColors(quad->vertices, quad->numvertices, R_SectorLight(sc->lightlevel), rendtexmask);

			quad->flags = RPF_NOSPLIT;
			if (rendtexmask) quad->flags |= RPF_MASKED;

			RL_AddPolyWall(quad, NULL);
		}
	}
}

//===========================================================================
// Rend_RenderNode
//  v220-6 modified to use C_CheckBspBox
//===========================================================================

void Rend_RenderNode(int bspnum)
{
	while ( !(bspnum & NFx_SUBSECTOR) )
	{
		node_t	*bsp = &nodes[bspnum];
		int		side = P_PointOnNodeSide(viewx, viewy, bsp);

		// Recursively divide front space
		Rend_RenderNode(bsp->children[side]);

		// v220-6 If the node's totally clipped (away from
		//        the viewer) then it's time to quit.
		if (!C_CheckBspBox(bsp->bbox[!side])) return;

		// Recursively divide back space
		bspnum = bsp->children[!side];
	}

	Rend_RenderSubsector(&subsectors[bspnum & ~NFx_SUBSECTOR]);
}

int rendStartList;	// Used for debugging

//===========================================================================
// Rend_PrepareList
//==========================================================================
void Rend_PrepareList(void)
{
	truevang = viewangle / (float)ANGMAX * 360.0f;
	if ((vang = truevang - 90.0f) < 0) vang += 360.0f;
	vpitch = viewpitch * RATIO;
	rad_pitch = vpitch / 180 * PI;

	if (!paused && !menu_active && !con_freezegame)	// v220-28 add con_freezegame
	{
		float	halfyfov = yfov * 0.5f;
		int		i;

		nearestClip = false;

		if (ge_inf.devmap) memset(th_items, 0, sizeof(th_items));

		M_SetGlobalYaw();

		// Prepare for rendering.
		thing_validcount++;		// b211-0
		ppt.viewplrmo->mjvalidcount = thing_validcount;	// Don't render self
		visspritecnt = 0;		// init the vissprite counter
		hasmask = transfloor = false;

		skyhemispheres = 0;
		sky_not_set = true;

		skyLayers[0].flags |=  SLF_ENABLED;
		skyLayers[1].flags &= ~SLF_ENABLED;

		// clear the siflags
		for (i = 0; i < numsectors; i++)
			sectors[i].siflags &= ~SIF_CLEARMASK;

		// v220-3 Clear lsiflags (for sides set in last pass)
		for (i = 0; i < lsiarray_count; i++)
			lsiarray[i]->lsiflags &= ~LS_TEXPRESENT;

		lsiarray_count = 0;

		RL_ClearLists();		// Clear the lists for new quads.
		C_ClearRanges();		// Clear the clipper.

		// Add the backside clipping range (if vpitch allows).
		if (vpitch <= halfyfov && vpitch >= -halfyfov)
		{
			float vpu = vpitch >= 0 ? vpitch : -vpitch;

			angle_t startAngle = (angle_t)((ANG45 * fieldOfView / 90.0f) * (1.0f + vpu / halfyfov));

			C_SafeAddRangeRealAngles(viewangle + startAngle, viewangle - startAngle);
		}

		// The viewside line for the depth cue.
		viewsidex = -FIX2FLT(viewsin);
		viewsidey =  FIX2FLT(viewcos);
		firstsubsector = true;

		if (num_xlsame) memset(xlisrendered, 0, num_xlsame);

		rendStartList = 0;
		GLOBAL_SKY_SCROLL_LINE = -1;

		if (!numnodes)
			Rend_RenderSubsector(subsectors);
		else
			Rend_RenderNode(numnodes-1);

		if (!LevelFullBright && !R3D_slopetypes)
			Rend_RenderShadows();

		PG_PreparePG();			// Prepare the particle lists
		R_SortVisSprites();		// Sort visprites based on distance
		ppt.listset = true;
	}
	else if (ppt.listset)	// v220-4 continue to render sprites/particles whilst paused
	{
		uint i;

		for (i = 0; i < visspritecnt; i++)
			vispointer[i]->visflags &= ~VIS_RENDERED;

		PG_PtcClearRender();	// Clear particles rendered flag
	}
}
