//-----------------------------------------------------------------------------
//
// Copyright (C) 2005
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_environ.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//	Support for pseudofake added
//
// Based on: BOOM v2.01
//
// DESCRIPTION:  Fake sector, missing texture handling
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>

#include "de_base.h"
#include "de_render.h"
#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// Data definitions used in conjunction with GL_Rend.c
extern RDBOOL	underfakefloor, faked;
extern RDBOOL	vssecunderfake;
extern RDBOOL	vssecoverfake;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------


// killough 3/7/98: Hack floor/ceiling heights for deep water etc.
//
// If player's view height is underneath fake floor replace
// the drawn floor and ceiling textures, and light level, with
// the control sector's. Similar for ceiling, only reflected.
//
// Modified as needed for use with Risen3D
//
sector_t *R_FakeFlat(sector_t *sec, sector_t *tempsec, short *flight, short *clight)
{
	sector_t *s = sec->heightsec >= 0 ? &sectors[sec->heightsec] : NULL;

	if (flight) *flight = (short)(sec->floorlightsec == -1 ? sec->lightlevel :
					sectors[sec->floorlightsec].lightlevel); // killough 3/16/98

	if (clight) *clight = (short)(sec->ceilinglightsec == -1 ? sec->lightlevel :
					sectors[sec->ceilinglightsec].lightlevel); // killough 4/11/98

	if (s && s->specFlags2 & SF2_PSEUDOFAKE)
	{
		*tempsec = *sec;
		rv.pseudo = true;

		if (underfakefloor)
		{	
			tempsec->smoothceil = s->smoothfloor;
			tempsec->ceilingpic = s->ceilingpic;
			tempsec->ceiloffx   = s->ceiloffx;
			tempsec->ceiloffy   = s->ceiloffy;

			tempsec->lightlevel  = s->lightlevel;
			vssecunderfake = true;

			if (flight) *flight = (short)(s->floorlightsec == -1 ? s->lightlevel :
							sectors[s->floorlightsec].lightlevel); // killough 3/16/98
		
			if (clight) *clight = (short)(s->ceilinglightsec == -1 ? s->lightlevel :
							sectors[s->ceilinglightsec].lightlevel); // killough 4/11/98
		}
		else if (sec->smoothfloor < s->smoothfloor)
		{
			tempsec->smoothfloor = s->smoothfloor;
			tempsec->floorpic = s->floorpic;
		}

		sec = tempsec;
		faked = true;

		return sec;
	}

	if (s && !(s->specFlags2 & SF2_FAKEMOVSECTOR) )
	{
		// Replace sector being drawn, with a copy to be hacked
		*tempsec = *sec;

		// Replace floor and ceiling height with other sector's heights.
		tempsec->floorheight   = s->floorheight;
		tempsec->ceilingheight = s->ceilingheight;
		
		// killough 11/98: prevent sudden light changes from non-water sectors:
		if (underfakefloor && (tempsec->smoothfloor = sec->smoothfloor,
                          tempsec->ceilingheight = s->smoothfloor-1))
		{	// head below floor
			tempsec->floorpic    = s->floorpic;
			tempsec->flooroffx   = s->flooroffx;
			tempsec->flooroffy   = s->flooroffy;
			
			if (s->ceilingpic == skyflatnum)
			{
				tempsec->smoothfloor = tempsec->ceilingheight+1;
				tempsec->ceilingpic = tempsec->floorpic;
				tempsec->ceiloffx   = tempsec->flooroffx;
				tempsec->ceiloffy   = tempsec->flooroffy;
			}
			else
			{
				tempsec->ceilingpic = s->ceilingpic;
				tempsec->ceiloffx   = s->ceiloffx;
				tempsec->ceiloffy   = s->ceiloffy;
			}

			tempsec->lightlevel = s->lightlevel;
			vssecunderfake = true;

			if (flight) *flight = (short)(s->floorlightsec == -1 ? s->lightlevel :
							sectors[s->floorlightsec].lightlevel); // killough 3/16/98
		
			if (clight) *clight = (short)(s->ceilinglightsec == -1 ? s->lightlevel :
							sectors[s->ceilinglightsec].lightlevel); // killough 4/11/98
		}
		else if (abovefakeceil && rv.sect->ceilingheight > s->ceilingheight)
		{	// head above ceiling
			tempsec->ceilingheight = s->ceilingheight;
			tempsec->floorheight   = s->ceilingheight + 1;
			
			tempsec->floorpic  = tempsec->ceilingpic = s->ceilingpic;
			tempsec->flooroffx = tempsec->ceiloffx   = s->ceiloffx;
			tempsec->flooroffy = tempsec->ceiloffy   = s->ceiloffy;
			
			if (s->floorpic != skyflatnum)
			{
				tempsec->ceilingheight = sec->ceilingheight;
				tempsec->floorpic   = s->floorpic;
				tempsec->flooroffx  = s->flooroffx;
				tempsec->flooroffy  = s->flooroffy;
			}
			else
				tempsec->smoothceil = s->smoothceil;
			
			tempsec->lightlevel = s->lightlevel;
			vssecoverfake = true;

			if (flight) *flight = (short)(s->floorlightsec == -1 ? s->lightlevel :
							sectors[s->floorlightsec].lightlevel); // killough 3/16/98
		
			if (clight) *clight = (short)(s->ceilinglightsec == -1 ? s->lightlevel :
							sectors[s->ceilinglightsec].lightlevel); // killough 4/11/98
		}
		else
		{
			tempsec->smoothfloor = s->smoothfloor;
			tempsec->smoothceil  = s->smoothceil;
		}
			
		sec = tempsec;
		faked = true;
	}

	return sec;
}

short subfn_LLevel(sector_t *s, RDBOOL type)
{
	if (type)
	{
		if (!rv.faked && s->ceilinglightsec >= 0)
			return sectors[s->ceilinglightsec].lightlevel;

		return s->lightlevel;
	}

	if (!rv.faked && s->floorlightsec >= 0)
		return sectors[s->floorlightsec].lightlevel;

	return s->lightlevel;
}

void R_SetMissingParms(sector_t *f, sector_t *c)
{
	if (f && f->floorpic != skyflatnum)
	{
		rv.fsec = (sector_t*)f;
		rv.sfloor = f->smoothfloor;
		rv.ffloor = rv.bfloor = rv.sfloor;
		rv.floorlightlevel = subfn_LLevel(f, 0);
	}
	if (c && c->ceilingpic != skyflatnum)
	{
		rv.csec = (sector_t*)c;
		rv.sceil = c->smoothceil;
		rv.fceil = rv.bceil = rv.sceil;
		rv.ceilinglightlevel = subfn_LLevel(c, 1);
	}
}

// This mainly deals with missing tex definitions
RDBOOL subfn_MissingFirstPass(sector_t *frontsec, sector_t *backsec, side_t *sid,
							   line_t *ldef, int stype, short mtflags)
{
	switch(stype)
	{
	default:
		break;
	case R3D_0D:
		if (frontsec == ldef->backsector)
		{
			rv.fsec = ldef->backsector;
			rv.sfloor = rv.fsec->smoothfloor;
		}
		break;
	// Missing tex lines - used to set heights of flats.
	// As such, must be called before SegFacing test.
	case R3D_3C:
		if (frontsec == ldef->backsector &&
			mtflags == R3D_MISHTEX1)
			R_SetMissingParms(NULL, ldef->frontsector);
		else if (frontsec == ldef->frontsector &&
			mtflags == R3D_MISHTEX2)
			R_SetMissingParms(NULL, ldef->backsector);
		if (frontsec == ldef->backsector &&
			mtflags == R3D_MISLTEX1)
			R_SetMissingParms(ldef->frontsector, NULL);
		else if (frontsec == ldef->frontsector &&
			mtflags == R3D_MISLTEX2)
			R_SetMissingParms(ldef->backsector, NULL);
		return true;
	case R3D_29:
		R_SetMissingParms(NULL, ldef->frontsector);
		break;
	case R3D_1A:
		if (sid->midtexture)
			break;

		if (frontsec->ceilingpic == skyflatnum)
		{
			if (mtflags & R3D_MISLOTEX && rv.lfsec < 0)
			{
				RDBOOL sa = frontsec->specFlags & SF_NOLOTEX;
				RDBOOL sb = backsec->specFlags  & SF_NOLOTEX;

				if (sa && !sb)
				{
					R_SetMissingParms(backsec, NULL);
				}

				return (sa^sb && !(sid->texFlags & T_ALL));
			}
			break;
		}
	//case R3D_1E:$consolidated
	//case R3D_30:$consolidated
		if (ldef->frontsector->heightsec < 0 &&
			ldef->backsector->heightsec < 0)
		{
			short tx1 = sides[ldef->sidenum[0]].texFlags;
			short tx2 = sides[ldef->sidenum[1]].texFlags;

			if (mtflags & R3D_MISHITEX)
			{
				if (!(mtflags & R3D_MISHTEX1 && (tx2 & 1 && !(mtflags & R3D_INVLTEX2))) && 
					!(mtflags & R3D_MISHTEX2 && (tx1 & 1 && !(mtflags & R3D_INVLTEX1))) && 
					rv.fceil > rv.bceil &&
					rv.csec->R3Dceilsec < 0 &&
					rv.fceil - rv.bceil <= MAX_SECHEIGHT)
				{
					R_SetMissingParms(NULL, rv.bceil < rv.fceil ? backsec : frontsec);
				}
				else if (IsLiquidCeilFlat(frontsec) &&
					IsLiquidCeilFlat(backsec))
				{
					R_SetMissingParms(NULL, rv.bceil < rv.fceil ? backsec : frontsec);
				}
				else if (mtflags & R3D_MISHTEX1 && IsLiquidFloorFlat(ldef->frontsector) &&
					!IsLiquidFloorFlat(ldef->backsector))
				{
					R_SetMissingParms(NULL, rv.bceil < rv.fceil ? backsec : frontsec);
				}
				else if (mtflags & R3D_MISHTEX2 && IsLiquidFloorFlat(ldef->backsector) &&
					!IsLiquidFloorFlat(ldef->frontsector))
				{
					R_SetMissingParms(NULL, rv.bceil < rv.fceil ? backsec : frontsec);
				}
			}
			else if (rv.fsec->R3Dfloorsec < 0)
			{
				if (origTEX[ldef - lines] & 0x11)
				{
					rv.sfloor = MIN_OF(rv.ffloor, rv.bfloor);
					rv.bfloor = rv.ffloor = rv.sfloor;
				}
				else if (rv.bfloor > rv.ffloor
					&& rv.bfloor - rv.ffloor <= MAX_SECHEIGHT)
				{
					R_SetMissingParms(backsec, NULL);
				}
			}
		}
		break;
	case R3D_32:
		rv.ssec->lcsec = -1;	// b211-0
		return true;
	case R3D_2A:
	//case R3D_2B: $consolidated
		if (faked)
			break;
		if (frontsec == ldef->frontsector)
		{
			if (mtflags & R3D_MISHTEX1)
				R_SetMissingParms(NULL, ldef->backsector);
			if ((rv.lfsec < 0 || stype == R3D_32)
				&& mtflags & R3D_MISLTEX1
				&& viewz > ldef->backsector->smoothfloor
				&& rv.ffloor < ldef->backsector->smoothfloor)
				R_SetMissingParms(ldef->backsector, NULL);
		}
		else if (frontsec == ldef->backsector)
		{
			if (mtflags & R3D_MISHTEX2)
				R_SetMissingParms(NULL, ldef->frontsector);
			if ((rv.lfsec < 0 || stype == R3D_32)
				&& mtflags & R3D_MISLTEX2
				&& viewz > ldef->frontsector->smoothfloor
				&& rv.ffloor < ldef->frontsector->smoothfloor)
				R_SetMissingParms(ldef->frontsector, NULL);
		}
		return true;
	case R3D_0B:
		// Added Apr 2004
		if (mtflags == (R3D_MISHTEX1|R3D_MISLTEX2|R3D_INVLTEX1))
		{
			rv.csec = ldef->frontsector;
			rv.sceil = rv.csec->smoothceil;
			rv.bceil = rv.fceil = rv.sceil;
			if (frontsec == ldef->backsector)
			{
				rv.fsec  = ldef->backsector;
				rv.sfloor = rv.fsec->smoothfloor;
			}
		}
		else
		if (rv.csec == backsec)
		{
			rv.csec  = frontsec;
			rv.sceil = frontsec->smoothceil;
		}
		break;
	case R3D_1B:
		if (mtflags & R3D_MISHTEX2 &&
			ldef->backsector->smoothceil >= ldef->frontsector->smoothceil)
		{
			rv.csec  = ldef->frontsector;
			rv.sceil = rv.csec->smoothceil;
			if (!(frontsec->specFlags & SF_MOVSECTOR) &&
				rv.fsec->lightlevel < 255)
			{
				rv.ceilinglightlevel = subfn_LLevel(rv.csec, 1);
			}
		}
		else if (mtflags & R3D_MISHTEX1 &&
			ldef->frontsector->smoothceil >= ldef->backsector->smoothceil)
		{
			rv.csec  = ldef->backsector;
			rv.sceil = rv.csec->smoothceil;
			if (!(backsec->specFlags & SF_MOVSECTOR) &&
				rv.fsec->lightlevel < 255)
			{
				rv.ceilinglightlevel = subfn_LLevel(rv.csec, 1);
			}
		}
		if (mtflags & R3D_MISLTEX1 &&
			ldef->frontsector->smoothfloor <= ldef->backsector->smoothfloor)
		{
			if (rv.ssec->lfsec >= 0 &&	// rev62
				ldef->backsector->smoothfloor <= sectors[rv.ssec->lfsec].smoothfloor)
			{
				rv.fsec = &sectors[rv.ssec->lfsec];
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				break;
			}
			if (rv.ffloor < ldef->backsector->smoothfloor)
			{
				rv.fsec = ldef->backsector;
				rv.sfloor = rv.fsec->smoothfloor;
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
			}
		}
		else if (mtflags & R3D_MISLTEX2 &&
			ldef->backsector->smoothfloor <= ldef->frontsector->smoothfloor)
		{
			if (rv.ssec->lfsec >= 0 &&	// rev62
				ldef->frontsector->smoothfloor <= sectors[rv.ssec->lfsec].smoothfloor)
			{
				rv.fsec = &sectors[rv.ssec->lfsec];
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				break;
			}
			if (rv.ffloor < ldef->frontsector->smoothfloor)
			{
				rv.fsec = ldef->frontsector;
				rv.sfloor = rv.fsec->smoothfloor;
				rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
			}
		}
		if (ldef->spflags & SPF_NOTEX) return true;
		break;
	case R3D_1C:
		if (!faked)
		{
			RDBOOL ft, bt;

			if (!IsLiquidFloorFlat(frontsec) &&
				(frontsec->specFlags & SF_MOVSECTOR ||
				 backsec->specFlags & SF_MOVSECTOR))
			{
				// The floor
				if (frontsec == ldef->frontsector &&
					frontsec->specFlags & SF_MOVSECTOR)
				{
					rv.fsec = ldef->backsector;
					rv.sfloor = rv.fsec->smoothfloor;
					rv.ffloor = rv.bfloor = rv.sfloor;
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				}
				else
				if (frontsec == ldef->backsector &&
					frontsec->specFlags & SF_MOVSECTOR)
				{
					rv.fsec = ldef->frontsector;
					rv.sfloor = rv.fsec->smoothfloor;
					rv.ffloor = rv.bfloor = rv.sfloor;
					rv.floorlightlevel = subfn_LLevel(rv.fsec, 0);
				}
			}
			else
			{
				R_SetMissingParms(rv.bfloor > rv.ffloor ? backsec : frontsec, NULL);
			}

			if (rv.fceil == rv.bceil) return true;

			// The ceiling
			if ( ((ft = (frontsec->specFlags & 0xC)) == 0xC) ^ ((bt = (backsec->specFlags  & 0xC)) == 0xC) )
			{
				if (ft && rv.bceil < rv.fceil) { R_SetMissingParms(NULL, backsec ); return true; }
				if (bt && rv.bceil > rv.fceil) { R_SetMissingParms(NULL, frontsec); return true; }
				return false;	// hi walls will need to be drawn
			}
			else
			{
				R_SetMissingParms(NULL, rv.bceil < rv.fceil ? backsec : frontsec);
				return true;
			}
		}
		break; //Invalid type R3D_1C may require walls to be drawn
	case R3D_17:
	case R3D_1F:
		if (mtflags & R3D_MISLTEX1)
			R_SetMissingParms(ldef->frontsector, NULL);
		if (mtflags & R3D_MISLTEX2)
			R_SetMissingParms(ldef->backsector, NULL);
		if (stype == R3D_17)
			return true;
		break;
	case R3D_18:
		if (mtflags & R3D_MISLTEX1 &&
			frontsec->smoothfloor <= backsec->smoothfloor)
			R_SetMissingParms(ldef->frontsector, NULL);
		if (mtflags & R3D_MISLTEX2 &&
			frontsec->smoothfloor >= backsec->smoothfloor)
			R_SetMissingParms(ldef->backsector, NULL);
		return true;
	}
	
	return false;
}
