//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_stwin.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified as needed for R3D
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Startup message and progress bar window
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <stdarg.h>

#include "de_base.h"
#include "de_console.h"
#include "sys_stwin.h"
#include "resource.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define CREF_BACKGROUND		0
#define CREF_PROGRESS		0xc08080
#define CREF_TEXT			0xffc0c0

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern uint swp_flags;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static HWND			msgWnd = NULL;
static int			barPos, barMax;

// CODE --------------------------------------------------------------------

//===========================================================================
// SW_DialogProc
//===========================================================================
static BOOL CALLBACK SW_DialogProc
	(HWND dlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static HBRUSH	bgBrush = NULL;
	static HBRUSH	progressBrush = NULL;

	char buf[300];
	HWND ed;

	switch(uMsg)
	{
	case WM_CTLCOLORSTATIC:
		// Set the background and text color of the messages edit box.
		ed = GetDlgItem(dlg, IDC_MESSAGES);
		if( (HWND) lParam != ed) return FALSE;

		SetBkColor  ( (HDC) wParam, CREF_BACKGROUND);
		SetTextColor( (HDC) wParam, CREF_TEXT);
       
   		return (BOOL) bgBrush;// jav this is supposed to return the brush handle

	case WM_INITDIALOG:
		// Get the title from the main window.
     	bgBrush       = CreateSolidBrush(CREF_BACKGROUND);
     	progressBrush = CreateSolidBrush(CREF_PROGRESS);

		GetWindowText(hWndMain, buf, 300);
		SetWindowText(dlg, buf);
		return TRUE;
 
    case WM_DESTROY:   // jav let it handle it's own destruction
     	if (bgBrush      ) DeleteObject(progressBrush);
    	if (progressBrush) DeleteObject(bgBrush);
     	bgBrush       = NULL;
        progressBrush = NULL;
        return FALSE;

	}
	// The message was not processed.
	return FALSE;
}

//===========================================================================
//  This is only called during startup
//===========================================================================
static void sw_Printf(char *buf)
{
	if(msgWnd)
	{
		char rep[256];
		uint len = strlen(buf);
		uint  pos = 1;

		if (len > 253) return;	// should never happen

		strcpy(rep + pos, buf);

		// Only '\n' at the start and ......
		if (rep[1] == '\n')
		{
			rep[pos] = '\n';
			pos = 0;
			rep[pos] = '\r';
		}

		// ...... at the end are dealt with
		if (rep[len - 1 + pos] == '\n')
		{
			rep[len - 1 + pos] = '\r';
			rep[len + pos] = '\n';
			rep[len + 1 + pos] = 0;
		}

		SendDlgItemMessage(msgWnd, IDC_MESSAGES, EM_REPLACESEL, 0, (LPARAM) &rep[pos]);
	}
}

// Used to print dotted line as textures are loaded
void p2swd(char *msg, ...)
{
	char	buff[256];
	va_list	args;
	
	va_start(args, msg);
	_vsnprintf(buff, 254, msg, args);
	va_end(args);

	buff[255] = 0;
	
	sw_Printf(buff);
	pr_OFdirect(buff);
}

//===========================================================================
// SW_Init
//===========================================================================
void SW_Init(void)
{
	if (!msgWnd) //  Check it's not already initialized.
	{
		msgWnd = CreateDialog(hInstApp, MAKEINTRESOURCE(IDD_STARTUP_WINDOW),
			hWndMain, SW_DialogProc);

		// v220-3 inherit msgWnd size (just a nicety when first starting up)
		if (msgWnd)
		{
			RECT rect;

			GetWindowRect(msgWnd, &rect);
			SetWindowLong(hWndMain, GWL_STYLE, WS_THICKFRAME);
			SetWindowPos(hWndMain, 0, rect.left, rect.top,
				rect.right - rect.left, rect.bottom - rect.top, swp_flags);
		}
	}
}

//===========================================================================
// SW_Shutdown
//===========================================================================
void SW_Shutdown(void)
{
	if(msgWnd) // Check it's been successfully initialized.
	{
		DestroyWindow(msgWnd);
		msgWnd = NULL;
	}
}
