//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_timer.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Time functions
//
//-----------------------------------------------------------------------------

//**************************************************************************
//**
//** SYS_TIMER.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <mmsystem.h>

#include "de_base.h"
#include "de_console.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

//float ticsPerSecond = TICSPERSEC;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static uint	timeresolution = 0;

// CODE --------------------------------------------------------------------

//==========================================================================
// Sys_InitTimer
//==========================================================================
void Sys_InitTimer(void)
{
	TIMECAPS tc;
	MMRESULT rs = timeGetDevCaps (&tc, sizeof(tc));

	// Set the timer to be as accurate as possible
	if (rs == TIMERR_NOERROR)
		timeresolution = tc.wPeriodMin;
	else
		timeresolution = 1;	// Try 1 ms

	Con_Message("Timer period : %imS [rs %i]\n", timeresolution, rs);

	if (   timeresolution != 1
		|| timeBeginPeriod(timeresolution) != TIMERR_NOERROR)
	{
		timeresolution = 0;
		Con_Error("Failed to set timer resolution to 1mS");
	}
}

//==========================================================================
// Sys_ShutdownTimer
//==========================================================================
void Sys_ShutdownTimer(void)
{
	if (timeresolution)
	{
		timeEndPeriod(timeresolution);
		timeresolution = 0;
	}
}

//===========================================================================
// Sys_GetRealTime
//	Returns the time in milliseconds.
//===========================================================================
uint Sys_GetRealTime (void)
{
	static RDBOOL first = true;
	static DWORD start;
	DWORD now = timeGetTime();

	if (first)
	{
		first = false;
		start = now;
		return 0;
	}
	
	// Wrapped around? (Every 50 days...)
	if (now < start) return 0xffffffff - start + now + 1;

	return now - start;
}

//===========================================================================
// Sys_GetSeconds
//	Returns the timer value in seconds.
//===========================================================================
float Sys_GetSeconds (void)
{
	uint t = Sys_GetRealTime();

	return  (float)t * 0.001f; //  / 1000.0;
}

//===========================================================================
// Sys_GetTimef
//	Returns time in 35 Hz floating point tics.
//===========================================================================
float Sys_GetTimef(void)
{
	return Sys_GetRealTime() * 0.035f; // TICSPERSEC / 1000.0;
}

//==========================================================================
// Sys_GetTime
//	Returns time in 35 Hz tics.
//==========================================================================
int Sys_GetTime (void)
{
	return (int) Sys_GetTimef();
}
