//------------------------------------------------------------------------
// BLOCKMAP : Generate the blockmap
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#include "rd_system.h"

#include <ctype.h>
#include <math.h>
#include <limits.h>

#include "rd_blockmap.h"
#include "rd_util.h"

int block_x, block_y;
int block_w, block_h;

//
// InitBlockmap
//
void InitBlockmap(void)
{
	int i;

	// find limits of linedefs, and store as map limits
	int minx = SHRT_MAX, miny = SHRT_MAX;
	int maxx = SHRT_MIN, maxy = SHRT_MIN;

	for (i = 0; i < num_linedefs; i++)
	{
		ldef_t *L = LookupLinedef(i);
		
		if (!L->zero_len)
		{
			float x1 = L->start->x;
			float y1 = L->start->y;
			float x2 = L->end->x;
			float y2 = L->end->y;
			
			int lx = (int)floor(MIN(x1, x2));
			int ly = (int)floor(MIN(y1, y2));
			int hx = (int)ceil(MAX(x1, x2));
			int hy = (int)ceil(MAX(y1, y2));
			
			if (lx < minx) minx = lx;
			if (ly < miny) miny = ly;
			if (hx > maxx) maxx = hx;
			if (hy > maxy) maxy = hy;
		}
	}

	block_x = minx - (minx & 0x7);
	block_y = miny - (miny & 0x7);

	block_w = ((maxx - block_x) / 128) + 1;
	block_h = ((maxy - block_y) / 128) + 1;
}
