//------------------------------------------------------------------------
// MAIN : Command-line version main program
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  $Log: general, r3dglbsp v5.0 - nodes generator for Risen3D v2.1.0
//
//	Modified to extend map limits and to create subsectors with the
//  the NULL backsector of one sided lines used for special effects.
//  Resultant files have the extension .R3Dgwa to differentiate them
//  from normal .gwa files.  This prevents confusion and allows them
//  to be stored in the same directory as normal .gwa files where
//  generated for use with other OGL ports.
//  Extended type generated is gNd4 only.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#include "rd_glbsp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <io.h>

void pr_OFinsert(char *message, ...);

static nodebuildinfo_t info;
static nodebuildcomms_t comms;

/* ----- entry point ----------------------------- */

// Error handling modified $R3Dgwa
int bsp_BuildNodes(int type, int factor, int noprune, char *infile, char *outfile)
{	
	int error = 0;

	info  = default_buildinfo;
	comms = default_buildcomms;
	
	GlbspCheckInfo(&info, &comms, type, factor, noprune, infile, outfile);
	
	if ( (error = GlbspBuildNodes(&info, &comms)) == GLBSP_E_Unknown)
	{
		pr_OFinsert("Error: %s\n", comms.message ? comms.message : 
				"No levels found");
	}
	else if (error == GLBSP_E_Cancelled)
	{
		pr_OFinsert("Error: %s\n", comms.message ? comms.message : "Thread was cancelled");
	}
	
	return error;
}

