//------------------------------------------------------------------------
// SYSTEM : System specific code
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#include "rd_system.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <assert.h>

#include "con_console.h"

void pr_OFinsert(char *message, ...);

int total_big_warn;
int total_small_warn;

static char message_buf[1024];

//
// FatalError
//
void FatalError(const char *str, ...)
{
	va_list args;
	
	va_start(args, str);
	vsprintf(message_buf, str, args);
	va_end(args);
	
	Con_Error("\nNODES BUILD FATAL ERROR: *** %s ***\n\n", message_buf);
}

//
// InternalError
//
void InternalError(const char *str, ...)
{
	va_list args;
	
	va_start(args, str);
	vsprintf(message_buf, str, args);
	va_end(args);
	
	Con_Error("\nNODES BUILD INTERNAL ERROR: *** %s ***\n\n", message_buf);
}

//
// PrintWarn
//
void PrintWarn(const char *str, ...)
{
	va_list args;
	
	va_start(args, str);
	vsprintf(message_buf, str, args);
	va_end(args);
	
	pr_OFinsert("WARN major: %s", message_buf);
	
	total_big_warn++;
}

//
// PrintMiniWarn
//
void PrintMiniWarn(const char *str, ...)
{
	va_list args;
	
	if (cur_info->mini_warnings)
	{
		va_start(args, str);
		vsprintf(message_buf, str, args);
		va_end(args);

		pr_OFinsert("WARN minor: %s", message_buf);
	}
	
	total_small_warn++;
}
