//-----------------------------------------------------------------------------
//
// Copyright (C) 2010 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_Mtex.c v220-9
//
// First released (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 11 (v220-11) 08/Aug/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: analyse missing textures.
//
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "r3d_MAINspec.h"
#include "r3d_edit.h"
#include "p_maputil.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int setSsecList(sector_t *sc, int sconly);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

osx_t osx;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

typedef struct {
	sector_t	*use;
	int			box[4];
	int			lnum;
	int			done;	// or flag
	float		sz;

	// used only in BiggestPIAever
	int			*lring;
	int			count;
} lring_t;

// CODE --------------------------------------------------------------------

static int setCountsec(sector_t *sc, int c)
{
	line_t	*ld;
	int		i, k, m, p;
	int		n, ln, vn;
	RDBOOL	valid = true;

	for (i = 0, n = c ; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		ln = ld - lines;
		vn = lines[ln].v1 - vertexes;

		for (k = 0; k < 2; k++)
		{
			if (k) vn = lines[ln].v2 - vertexes;

			for (m = 0; m < vertexowners[vn].num; m++)
			{
				RDBOOL	found = false;
				int		s = vertexowners[vn].list[m];

				if (s == sc - sectors) continue;
				if (sectors[s].heightsec >= 0) return 0;
				if (sectors[s].specFlags & SF_MOVSECTOR) return 0;

				// special case
				if ((sectors[s].specFlags & 0xC) != 0xC)
				{
					if (   sectors[s].tag == 999
						&& sc->ceilingpic != skyflatnum
						&& SecCtrl(s))
					{
						sector_t *ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

						if (   ch
							&& ch->ceilingpic != skyflatnum
							&& ch->ceilingheight - sectors[s].ceilingheight <= 4*FRACUNIT)
						{
							countsec[1] = s; return -1;
						}
					}
					valid = false;
				}

				if (!valid) continue;	// purely searching for special case

				for (p = 0; p < n; p++)
				{
					if (countsec[p] == s) { found = true; break; }
				}

				if (!found)
				{
					countsec[n++] = s;
				}
			}
		}
	}

	return valid ? n : 0;
}

static RDBOOL secInCountsec(sector_t *s, int c)
{
	int i;

	for (i = 0; i < c; i++)
	{
		if (countsec[i] == s - sectors) return true;
	}

	return false;
}

static	sector_t *ownerSec(int c)
{
	sector_t *sc, *ch, *os = NULL;
	line_t	 *ld;
	int		 i, k, n;

	for (i = 1; i < c; i++)
	{
		sc = &sectors[countsec[i]];

		for (k = 0; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			for (n = 0; n < 2; n++)
			{
				ch = n ? ld->backsector : ld->frontsector;

				if (ch && !(ch->specFlags & SF_MOVSECTOR))
				{
					if (ch->specFlags & 0xC)
					{
						// check if this sector is in the countsec array
						if (secInCountsec(ch, c)) continue;
					}

					if (!os) os = ch;
					else if (os
						&& os != ch
						&& (os->floorheight != ch->floorheight
						||  os->ceilingheight != ch->ceilingheight))
					{
						return NULL;
					}
				}
			}
		}
	}

	return os;
}

static void setCtrlOwner()
{
	sector_t *s = &sectors[countsec[0]];
	int		 i, k;

	for (i = 0; i < s->linecount; i++)
	{
		if ( !(s->lines[i]->spflags & SPF_CONTROLLINE) )
		{
			int		 lnum = s->lines[i] - lines;
			ssline_t *sl = &sslines[lnum];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->sector == s && ss->ceilsec < 0)
				{
					Setr3dsscs(ss, countsec[1], 48001);
					Setr3dsslcsec(ss, countsec[0], 48002);
				}
			}

			if (origTEX[lnum] & 0x22)
				s->lines[i]->spflags |= (SPF_DRAWTOP|SPF_LINEONLINE);
			Settype(s->lines[i], R3D_36, 48003);
			SetLFG(s->lines[i], LFG_NEWSET, 48004);
		}
	}
}

static void setOwner(sector_t *os, int c, RDBOOL adopt, int num)
{
	sector_t *sc;
	ssline_t *sl;
	line_t	 *ld;
	int		 i, k, m, p;

	validcount++;
	ss_validcount++;

	for (i = 0; i < c; i++)
	{
		sc = &sectors[countsec[i]];
		Setsecdone(sc, num);

		for (k = 0; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			sl = &sslines[ld - lines];

			for (m = 0; m < sl->num; m++)
			{
				subsector_t *ss = &subsectors[sl->list[m]];

				if (ss->ssvalidcount != ss_validcount)
				{
					int		n = ss->sector - sectors;
					RDBOOL	found = false;

					ss->ssvalidcount = ss_validcount;

					for (p = 0; p < c; p++)
					{
						if (countsec[p] == n)
						{
							found = true;
							SetOverride(ss, os - sectors, num);
						}
					}

					if (found)
					{
						seg_t *sg = &segs[ss->firstline];

						for (p = 0; p < ss->linecount; p++, sg++)
						{
							if (  (ld = sg->linedef)
								&& ld->validcount != validcount)
							{
							/*	if (   adopt
									&& !ld->backsector
									&& ld->frontsector->specFlags & 0xC
									&& !(ld->frontsector->specFlags & SF_MOVCEIL))
								{
									ld->adjacentsec = os - sectors;
									sides[*ld->sidenum].rowoffadj = ld->frontsector->ceilingheight - os->ceilingheight;
								}*/

								ld->validcount = validcount;
								SetLFG(ld, LFG_NEWSET, num);
							}
						}
					}
				}
			}
		}
	}
}


// v220-16 Look for structures that are invalid (i.e. do not
//         even render correctly using a software renderer).
void MX_CheckInvalidDoors()
{
	sector_t	*sc, *os;
	int			i;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		RDBOOL	valid = true;

		// Find textureless door sectors only accessing
		// other notex sectors and find the adoptee sector
		if (   SecCtrl(i)
			&& (sc->specFlags & 0x3000C) == 0x2000C
			&& !secdone[i]
			&& sc->stptr < 0)
		{
			int		c = 0;

			countsec[c++] = i;

			// type 0
			if ( (c = setCountsec(sc, c)) < 0)
			{
				setCtrlOwner();
				continue;
			}
			else
			// type 1 - multiple sectors involved
			if (c > 2 && (os = ownerSec(c)))
			{
				setOwner(os, c, false, 48005);
			}
			
			// type 2 - single sector involved
			if (!secdone[i])
			{
				vertexowner_t	*vt[2], *vtm = NULL;	// v220-24 bug fix
				sector_t		*ch;
				line_t			*ld;
				int				k, m, n;

				for (k = 0, os = NULL; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (   !(ld->spflags & SPF_CONTROLLINE)
						&& Vtfl1(ld)
						&& Spans1sLines(ld)
						&& (vt[0] = &vertexowners[ld->v1 - vertexes])->num == 3
						&& (vt[1] = &vertexowners[ld->v2 - vertexes])->num == 3)
					{
						for (m = 0; m < 2; m++)
						{
							for (n = 0; n < 3; n++)
							{
								ch = &sectors[vt[m]->list[n]];

								if (!(ch->specFlags & 0x3000C))
								{
									if (!os) os = ch;
									else if (os != ch
										&& (ch->floorheight != os->floorheight
										||  ch->ceilingheight != os->ceilingheight))
									{
										goto end_check;
									}

									vtm = &vertexowners[ld->v1 - vertexes];
								}
							}
						}
					}
				}

				if (vtm)
				{
					for (m = n = 0; n < 3; n++)
					{
						ch = &sectors[vtm->list[n]];

						if (ch->specFlags & 0x3000C)
							countsec[m++] = ch - sectors;
					}

					setOwner(os, m, true, 48006);
				}
			}
		}
end_check:;
	}

	// similar for floors
	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		RDBOOL	valid = true;

		// Find textureless door sectors only accessing
		// other notex sectors and find the adoptee sector
		if (   SecCtrl(i)
			&& (sc->specFlags & 0x13001C) == 0x11001C
			&& !secdone[i]
			&& sc->stptr < 0)
		{
			sector_t *ch;
			ssline_t *sl;
			line_t	 *ld;
			int		 c, k, m;

			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				if (ld->spflags & SPF_CONTROLLINE) continue;
				if (origTEX[ld - lines] & 0x55) { c = 0; break; }
				if (!ld->backsector) { c = 0; break; }
				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (  ch->specFlags & 0x30000) continue;
				if (!(ch->specFlags & 0x400) ) continue;
				if (!os) os = ch;
				else if (os != ch
					&& (os->ceilingheight != ch->ceilingheight
					||  os->floorheight != ch->floorheight))
				{
					c = 0; break;
				}
				linearray[c++] = ld - lines;
			}

			if (os && c > 2)
			{
				for (k = 0; k < c; k++)
				{
					ld = &lines[linearray[k]];
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sc && ss->ceilsec < 0)
							SetOverride(ss, os - sectors, 48007);
					}
				}
			}
		}
	}
}

// v220-22 Look for naive structures.
void MX_CheckInvalidPlats()
{
	sector_t	*ch, *sc, *os;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, k;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		RDBOOL	valid = false;
		int		*lmem;

		if (   SecCtrl(i)
			&& (sc->specFlags & 0x1001C) == 0x1001C
			&& !secdone[i]
			&& sc->stptr < 0)
		{
			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   lineflags[ld - lines]
					|| !ld->backsector
					|| ld->frontsector->floorheight == ld->backsector->floorheight)
				{
					goto next_sec;
				}

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!os)
				{
					if (ch->specFlags & SF_MOVFLOOR) goto next_sec;
					os = ch;
				}
				else if (os != ch)
				{
					if (   ch->floorheight != os->floorheight
						|| ch->floorpic != os->floorpic
						|| ch->ceilingpic != os->ceilingpic
						|| (ch->ceilingpic != skyflatnum && ch->ceilingheight != os->ceilingheight))
					{
						goto next_sec;
					}
				}

				linearray[c++] = ld - lines;;
			}

			if (c < 3) goto next_sec;

			CLR_hitline();

			lmem = (int*)malloc(c * sizeof(int));
			memcpy(lmem, linearray, c * sizeof(int));

			for (k = 0; k < c; k++) hitline[lmem[k]] = 1;

			while (1)
			{
				int lnum = -1;

				for (k = 0; k < c; k++) { if (hitline[lmem[k]] == 1) lnum = lmem[k]; break; }

				if (lnum < 0) break;

				if (   subfn_CheckForRing(hitline, sc, linearray[0], &rc, box, NULL, NULL) && rc > 2
					&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
					&& LS_OnlySSLinesInSSECarray(sc, hitline))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						// v220-33 add test for where the sector will move up
						if (subsectors[hitssec[k]].sector->specFlags & SF_MOVINVALID)
							Setr3dssfs(&subsectors[hitssec[k]], os - sectors, 48008);
						else
							Setr3dsslfsec(&subsectors[hitssec[k]], os - sectors, 48008);
					}

					for (k = 0; k < c; k++)
					{
						if (hitline[lmem[k]] == 4)
						{
							hitline[lmem[k]] = 0;

							if (origTEX[lmem[k]] & 0x22)
								Settype(&lines[lmem[k]], R3D_2D, 48009);
							else
								Settype(&lines[lmem[k]], R3D_3F, 48011);

							SetLFG(&lines[lmem[k]], LFG_NEWSET, 48012);
						}
					}
					
					valid = true;

				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
next_sec:;

		 if (valid) SetSecidxDone(i, 48013);
	}
}

// v220-10 check for line rings with just one line with no missing tex
void MX_RemoveUnwantedTexures()
{
	int diff, i;

	// 1) remove unwanted bottom textures that could affect missing tex functione
	for (i = 0; i < numlines; i++)
	{
		if (  !lineflags[i]
			&& origTEX[i] & 0x11
			&& (origTEX[i] & 0x11) != 0x11
			&& Vtfl1(&lines[i])
			&& lines[i].ml_flags & ML_TWOSIDED
			&& (diff = lines[i].backsector->floorheight - lines[i].frontsector->floorheight)
			&& abs(diff) > 8*FRACUNIT)
		{
			sector_t *s = diff > 0 ? lines[i].frontsector : lines[i].backsector;
			sector_t *chk = lines[i].frontsector == s ? lines[i].backsector : lines[i].frontsector;
			int		 k, rc, fht = chk->floorheight;

			if (   s->linecount < 3
				|| s->floorheight >= fht
				|| chk->specFlags & SF_MOVFLOOR)
			{
				continue;
			}

			CLR_hitline();

			for (k = 0; k < s->linecount; k++)
			{
				hitline[s->lines[k] - lines] = 1;
			}

			if (subfn_CheckForRing(hitline, s, i, &rc, NULL, NULL, NULL) && rc > 2)
			{
				int c, d, tx, tn;

				for (c = d = k = 0; k < s->linecount; k++)
				{
					int	lnum = s->lines[k] - lines;

					if (hitline[lnum] != 4) continue;

					if (!Vtfl(&lines[lnum])) { c = 0; break; }

					chk = s->lines[k]->frontsector == s ? s->lines[k]->backsector : s->lines[k]->frontsector;

					if (  !chk
						|| chk->specFlags & SF_MOVFLOOR
						|| chk->specFlags != fht)
					{ 
						c = 0; break;
					}

					tx = s->lines[k]->frontsector == s ? 0x01 : 0x10;

					if ( (tn = origTEX[lnum] & 0x11) )
					{
						if (tn != tx) { c = 0; break; }

						linearray[d++] = lnum;
					}

					c++;
				}

				if (c == rc)
				{
					for (k = 0; k < d; k++)
					{
						line_t  *ld = &lines[linearray[k]];
						int		side = tx = 0x01 ? 0 : 1;

						origTEX[linearray[k]] &= ~0x11;
						sides[ld->sidenum[side]].bottomtexture = 0;
						if (R3D_edit)
							pr_OFinsert("MAP ISSUE: L%i has an unwanted bottom texture side %i\n", linearray[k], side++);
					}
				}
			}
		}
	}
	// 2) remove unwanted top textures that could affect missing tex functione
	for (i = 0; i < numlines; i++)
	{
		if (  !lineflags[i]
			&& origTEX[i] & 0x44
			&& (origTEX[i] & 0x44) != 0x44
			&& Vtfl1(&lines[i])
			&& lines[i].ml_flags & ML_TWOSIDED
			&& (diff = lines[i].frontsector->ceilingheight - lines[i].backsector->ceilingheight)
			&& abs(diff) > 8*FRACUNIT)
		{
			sector_t *s = diff > 0 ? lines[i].frontsector : lines[i].backsector;
			sector_t *chk = lines[i].frontsector == s ? lines[i].backsector : lines[i].frontsector;
			int		 k, rc, cht = chk->ceilingheight;

			if (   s->linecount < 3
				|| cht >= s->ceilingheight
				|| chk->specFlags & SF_MOVCEIL)
			{
				continue;
			}

			CLR_hitline();

			for (k = 0; k < s->linecount; k++)
			{
				hitline[s->lines[k] - lines] = 1;
			}

			if (subfn_CheckForRing(hitline, s, i, &rc, NULL, NULL, NULL) && rc > 2)
			{
				int c, d, tx, tn;

				for (c = d = k = 0; k < s->linecount; k++)
				{
					int	lnum = s->lines[k] - lines;

					if (hitline[lnum] != 4) continue;

					if (!Vtfl(&lines[lnum])) { c = 0; break; }

					chk = s->lines[k]->frontsector == s ? s->lines[k]->backsector : s->lines[k]->frontsector;

					if (  !chk
						|| chk->specFlags & SF_MOVCEIL
						|| chk->ceilingheight != cht)
					{ 
						c = 0; break;
					}

					tx = s->lines[k]->frontsector == s ? 0x04 : 0x40;

					if ( (tn = origTEX[lnum] & 0x44) )
					{
						if (tn != tx) { c = 0; break; }

						linearray[d++] = lnum;
					}

					c++;
				}

				if (c == rc)
				{
					for (k = 0; k < d; k++)
					{
						line_t  *ld = &lines[linearray[k]];
						int		side = tx = 0x04 ? 0 : 1;

						origTEX[linearray[k]] &= ~0x44;
						sides[ld->sidenum[side]].toptexture = 0;
						if (R3D_edit)
							pr_OFinsert("MAP ISSUE: L%i has an unwanted top texture side %i\n", linearray[k], side++);
					}
				}
			}
		}
	}
}

// v220-16 allow skyflatnum if lower sector
static RDBOOL loValid(sector_t *fs, sector_t *bs, int txn)
{
	int d = (fs->floorheight == fs->ceilingheight || bs->floorheight == bs->ceilingheight) ? 0 : 8*FRACUNIT;

	return !(txn & 0x11)
		&& !(fs->floorpic == skyflatnum && fs->floorheight > bs->floorheight)
		&& !(bs->floorpic == skyflatnum && bs->floorheight > fs->floorheight)
		&& abs(fs->floorheight - bs->floorheight) > d;
}

// v220-16 allow skyflatnum if higher sector
static RDBOOL hiValid(sector_t *fs, sector_t *bs, int txn)
{
	return !(txn & 0x44)
		&& !(fs->ceilingpic == skyflatnum && fs->ceilingheight < bs->ceilingheight)
		&& !(bs->ceilingpic == skyflatnum && bs->ceilingheight < fs->ceilingheight)
		&& abs(fs->ceilingheight - bs->ceilingheight) > 8*FRACUNIT;
}

static RDBOOL osInvalid(sector_t *s)
{
	return !s || s->stptr >= 0 || s->heightsec >= 0;
}

static int __cdecl boxSize(const void *r1, const void *r2)
{
	float diff = ((lring_t*)r1)->sz - ((lring_t*)r2)->sz;

	return diff > 0 ? -1 : diff < 0 ? 1 : 0;
}

static RDBOOL setLinesContained(sector_t *sec, int secn, int from)
{
	subsector_t *ss;
	seg_t		*sg;
	line_t		*ld;
	RDBOOL		trickfloor;
	int			c, i, k;

	for (c = i = 0; i < hitssecptr; i++)
	{
		if (subsectors[hitssec[i]].sector - sectors != secn)
		{
			c = 1; break;
		}
	}

	if (!c) return false;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (   hitline[ld - lines] != 4
					&& ld->ml_flags & ML_TWOSIDED
					&& origTEX[ld - lines] & 0x55)
				{
					return false;
				}

				linearray[c++] = ld - lines;
			}
		}
	}

	// v220-16 Check for single sector in loop referenced by a movfloor sector
	trickfloor = false;

	if ((sectors[secn].specFlags & 0x30404) == 0x10400)
	{
		for (k = 0; k < hitssecptr; k++)
		{
			ss = &subsectors[hitssec[k]];

			trickfloor = ss->sector - sectors != secn
				&& ss->sector->ceilingheight == sectors[secn].ceilingheight
				&& (ss->sector->specFlags & 0x3040C) == 0x40C;

			if (!trickfloor) break;
		}
	}

	if (!trickfloor)
	{
		for (k = 0; k < hitssecptr; k++)
		{
			ss = &subsectors[hitssec[k]];

			ss->lcsec = -1;
			ss->lfsec = -1;

			if (trickfloor) continue;

			if (ss->sector - sectors != secn)
			{
				Setr3dsscs(ss, secn, 48014);
				Setr3dssfs(ss, secn, 48015);
			}
		}
	}

	for (k = 0; k < c; k++)
	{
		ld = &lines[linearray[k]];

		if (   ld->ml_flags & ML_TWOSIDED
			&& !origTEX[ld - lines])
		{
			Settype(ld, trickfloor ? R3D_16 : R3D_3F, 48016);
		}
		else if (ld->frontsector == ld->backsector)
		{
			Settype(ld, R3D_03, 48017);
		}
		else
		{
			Settype(ld, R3D_36, 48018);
		}

		SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48019);
	}

	return true;
}

static int checkLines(sector_t *ref, int *count, int *type, byte *array)
{
	line_t	*ld;
	int		c, i, k;
	int		ret = 0;

	//-------------------

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (array[ld - lines] == 4)
				{
					sector_t *chk = ld->frontsector == ref ? ld->backsector : ld->frontsector;

					if ((chk->specFlags & 0xF000C) != 0xC) return -1;
				}
				else
				{
					if ( !(ld->ml_flags & ML_TWOSIDED) ) return -1;

					if (   ld->frontsector == ld->backsector
						&& ld->frontsector != ref)
					{
						// Could have nested rings which cause light level problems;
						// this is dealt with elsewhere.
						ret |= 1;
					}

					if (   (ld->frontsector->specFlags & 0xF000C) != 0xC
						|| (ld->backsector->specFlags  & 0xF000C) != 0xC)
					{
						ret |= 2; break;
					}
				}

				linearray[c++] = ld - lines;

				if (!(*type) && origTEX[ld - lines]) *type = 1;
			}
		}
	}

	*count = c;
	return ret;
}

static RDBOOL fillLines(sector_t *use, sector_t *in, sector_t *out, byte *array, int type)
{
	int		c, i, k;

	for (i = 0; i < use->linecount; i++)
	{
		line_t *ld = use->lines[i];

		if (array[ld - lines] == 4)
		{
			if (   !(ld->frontsector ==  in && ld->backsector == out)
				&& !(ld->frontsector == out && ld->backsector == in))
			{
				return false;
			}
		}
	}

	c = use == out && (in->specFlags & 0xC) == 0xC && in->floorheight > out->floorheight;

	if (!c && !type) return true;

	for (i = 0; i < use->linecount; i++)
	{
		line_t *ld = use->lines[i];

		if (array[ld - lines] == 4)
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->sector == out && ss->lfsec < 0)
				{
					if (type)
					{
						Setr3dsscs(ss, in - sectors, 48020);
						Setr3dsslcsec(ss, out - sectors, 48021);
					}
					else if (!c)
					{
						Setr3dssfs(ss, in - sectors, 48022);
						Setr3dsslfsec(ss, out - sectors, 48023);
					}
				}
			}
		}
	}

	return true;
}

static void subfn_Prelim(sector_t *s)
{
	line_t	*ld;
	int		c, i, k;

	CLR_hitline();

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (!lineflags[i] && Vtfl1(ld)) hitline[ld - lines] = 1, c++;
	}

	if (c < 3) return;

	while(1)
	{
		int rc, box[4], lnum = -1;

		for (i = 0; i < s->linecount; i++)
		{
			ld = s->lines[i];

			if (hitline[ld - lines] == 1 && !(ld->ml_flags & ML_TWOSIDED))
			{
				lnum = ld - lines; break;
			}
		}

		if (lnum < 0) return;

		if (   subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
			&& rc > 2
			&& LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE))
		{
			sector_t *os = NULL;

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (hitline[ld - lines] == 4 && ld->ml_flags & ML_TWOSIDED)
				{
					sector_t *chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

					if (!os)
						os = chk;
					else if (os != chk
						&& (os->floorheight != chk->floorheight
						|| os->ceilingheight != chk->ceilingheight))
					{
						os  = NULL; break;
					}
				}
			}

			if (os) setLinesContained(s, os - sectors, 1);

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

static sector_t *subfn_CheckAdjacentSector(sector_t *ref, sector_t *s)
{
	sector_t	*nsec = NULL;
	line_t		*ld;
	int			i, k, m, lim;
	int			addssec[4];

	if ( !(s->specFlags & 0x4) || !(s->specFlags & 0x400) ) return s;

	for (i = lim = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (hitline[ld - lines] == 4)
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					ss->ssvalidcount = ss_validcount;

					if (ss->sector == s)
					{
						seg_t *sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (   sg->backsector
								&& sg->backsector != ref
								&& sg->linedef
								&& !(sg->backsector->specFlags & SF_MOVFLOOR)
								&& (sg->backsector->floorheight - sg->frontsector->floorheight) > 8*FRACUNIT
								&& !(origTEX[sg->linedef - lines] & 0x44) )
							{
								if (!nsec)
								{
									nsec = sg->backsector;
								}
								else if (sg->backsector->floorheight != nsec->floorheight
									|| sg->backsector->ceilingheight != nsec->ceilingheight)
								{
									return s;
								}

								if (lim < 4) addssec[lim++] = sl->list[k];
							}
						}
					}
				}
			}
		}
	}

	if (nsec)
	{
		for (i = 0; i < lim; i++) hitssec[hitssecptr++] = addssec[i];

		return nsec;
	}

	return s;
}

static int subfn_CheckVirtualSectors(sector_t *sc, int lnum)
{
	sector_t	*osec = NULL;
	line_t		*ld;
	ssline_t	*sl;
	int			k, m, n;
	int			refsec = -1;

	hitssecptr = 0;
	ss_validcount++;

	for (k = 0; k < sc->linecount; k++)
	{
		ld = sc->lines[k];

		if (hitline[ld - lines] != 4) continue;

		sl = &sslines[ld - lines];

		for (m = 0; m < sl->num; m++)
		{
			subsector_t *ss = &subsectors[sl->list[m]];

			if (   ss->sector == sc
				&& ss->ssvalidcount != ss_validcount)
			{
				seg_t *sg = &segs[ss->firstline];

				ss->ssvalidcount = ss_validcount;
				if (ss->sector->heightsec >= 0) continue;	// v220-35 add heightsec test
				hitssec[hitssecptr++] = sl->list[m];

				for (n = 0; n < ss->linecount; n++, sg++)
				{
					if (sg->frontsector != sc)
					{
						if (!osec)
						{
							osec = sg->frontsector;
						}
						else if (osec != sg->frontsector)
						{
							if (   osec->floorheight != sg->frontsector->floorheight
								|| osec->ceilingheight != sg->frontsector->ceilingheight)
							{
								return false;
							}

							if (refsec < 0) refsec = sg->frontsector - sectors;
						}
					}

					if (sg->backsector && sg->backsector != sc)
					{
						if (!osec)
						{
							osec = sg->backsector;
						}
						else if (osec != sg->backsector)
						{
							if (   osec->floorheight != sg->backsector->floorheight
								|| osec->ceilingheight != sg->backsector->ceilingheight)
							{
								return false;
							}

							if (refsec < 0) refsec = sg->backsector - sectors;
						}
					}
				}
			}
		}
	}

	if (   osec
		&& !(osec->specFlags & 0x30000)
		&& (osec->specFlags & SF_SSLINETRUE || (osec->specFlags & 0xC) != 0xC))
	{
		osec = subfn_CheckAdjacentSector(sc, osec);

		if (refsec >= 0) osec = subfn_CheckAdjacentSector(sc, &sectors[refsec]);

		for (k = 0; k < hitssecptr; k++)
		{
			SetOverride(&subsectors[hitssec[k]], osec - sectors, 48024);
		}

		for (k = 0; k < sc->linecount; k++)
		{
			line_t *ld = sc->lines[k];

			if (hitline[ld - lines] == 4)
			{
				if (ld->ml_flags & ML_TWOSIDED)
				{
					if (!origTEX[ld - lines])
						Forcetype(ld, R3D_3F, 48025);
				}
				else
					ld->adjacentsec = osec - sectors;
			}
		}

		Setsecdone(sc, 48026);
		return true;
	}
	return false;
}

int subfn_CVS(sector_t *sc)
{
	int c, k, rc;

	CLR_hitline();

	for (c = k = 0; k < sc->linecount; k++)
	{
		line_t *ld = sc->lines[k];

		if (lineflags[ld - lines]) { c = 0; break; }

		if (ld->lnflags2 & LNF2_OWNEQUALTOTWO)
		{
			hitline[ld - lines] = 1; c++;
		}
	}

	if (c < 3) return false;

	while(1)
	{
		int lnum = -1;

		for (k = 0; k < sc->linecount; k++)
		{
			line_t *ld = sc->lines[k];
			
			if (hitline[ld - lines] == 1)
			{
				lnum = ld - lines;
				break;
			}
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			int a, b;

			for (a = b = k = 0; k < sc->linecount; k++)
			{
				line_t *ld = sc->lines[k];
				
				if (hitline[ld - lines] == 4)
				{
					if (ld->ml_flags & ML_TWOSIDED) a++; else b++;
				}
			}

			if (a > 1 && b > 1)
				subfn_CheckVirtualSectors(sc, lnum);

			for (k = 0; k < sc->linecount; k++)
			{
				line_t *ld = sc->lines[k];
				
				if (hitline[ld - lines] == 4)
					hitline[ld - lines] = 0;
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	return secdone[sc - sectors];
}

// Look for missing tex loops with 1's lines included
void MX_Prelim()
{
	subsector_t	*ss;
	sector_t	*os, *s = sectors;
	seg_t		*sg;
	line_t		*ld;
	int			i, k, m, n, ht;

	for (i = 0; i < numsectors; i++, s++)
	{
		// v220-16 check for mov ceiling with no upper tex
		if (   !secdone[i]
			&& s->heightsec < 0	// v220-31 added
			&& (s->specFlags & 0x20008) == 0x20008
			&& s->ceilingheight <= s->floorheight)
		{
			ss_validcount++;

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (ld->backsector)
				{
					ssline_t *sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;

							if (ss->sector == s)
							{
							//	ss->ssflags |= SSF_NOOCCLUDE;
								sg = &segs[ss->firstline];

								for (os = NULL, n = 0; n < ss->linecount; n++, sg++)
								{
									if (sg->backsector)
									{
										if (!os) os = sg->backsector;
										else if (os->ceilingheight != sg->backsector->ceilingheight)
										{
											os = NULL; break;
										}
									}
								}

								if (os && !(os->specFlags & SF_MOVCEIL))
									Setr3dsscs(ss, os - sectors, 48027);
							}
						}
					}
				}
			}
		}

		if (  (s->specFlags & 0x30410) == 0x410
			&& s->ceilingpic != skyflatnum
			&& s->floorpic != skyflatnum
			&& (ht = s->ceilingheight - s->floorheight) < 4*FRACUNIT
			&& ht >= 0)
		{
			if (subfn_CVS(s)) continue;
		}

		if (SecCtrl(i) || s->stptr >= 0 || s->linecount < 3)
		{
			continue;
		}

		if ((s->specFlags & 0x3041C) == 0x41C)
		{
			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (   !lineflags[i]
					&& !(ld->ml_flags & ML_TWOSIDED)
					&& Vtfl1(ld))
				{
					subfn_Prelim(s);
					break;
				}
			}
		}
	}
}

// v220-10 Check for subsectors having a complete line loop. This only
//         works for convex ssecs and where a line has not been split
//         by a gl node. This is invariably the case when needed, however.
static RDBOOL ssLoop(subsector_t *ss, sector_t *chk)
{
	line_t	*ld;
	seg_t	*sg = &segs[ss->firstline];
	int		c, i, rc, lnum = -1;

	CLR_conarray();

	for (c = i = 0; i < ss->linecount; i++, sg++)
	{
		if ( (ld = sg->linedef))
		{
			if (   sg->backsector
				&& sg->backsector->floorheight < chk->floorheight)
			{
				return false;
			}

			lnum = ld - lines;

			if (!conarray[lnum])
			{
				conarray[lnum] = 1; c++;
			}
		}
	}

	return c > 2
		&& subfn_CheckForRing(conarray, ss->sector, lnum, &rc, NULL, NULL, NULL)
		&& rc == c;
}

static RDBOOL diffSectors(int idx)
{
	return hitline[idx] != 4
		&& lines[idx].frontsector != lines[idx].backsector;
}

static int getSSECowner(subsector_t *ss)
{
	line_t	*ld;
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	if (ss->linecount == 1)
	{
		if (!(ld = sg->linedef)) return -1;

		if (ld->frontsector == ld->backsector)
		{
			goto check_subsec;
		}
		else
		{
			fixed_t x = (fixed_t)(ss->midpoint.x * FRACUNIT);
			fixed_t y = (fixed_t)(ss->midpoint.y * FRACUNIT);

			return P_PointOnLineSide(x, y, ld) ? ld->backsector - sectors : ld->frontsector - sectors;
		}
	}
	else
	{
		for (i = 0; i < ss->linecount; i++, sg++)
		{
			if (  (ld = sg->linedef)
				&& ld->frontsector != ld->backsector
				&& sg->frontsector)	// should be set
			{
				return sg->frontsector - sectors;
			}
		}
	}

check_subsec:;
	
	sg = &segs[ss->firstline];

	// if we get here then the subsec is likely to be inside a loop with no other lines
	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if ((ld = sg->linedef) && ld->frontsector == ld->backsector)
		{
			ssline_t *sl = &sslines[ld - lines];
			int		 k, m;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ref = &subsectors[sl->list[k]];
				seg_t		*osg = &segs[ref->firstline];

				for (m = 0; m < ref->linecount; m++, osg++)
				{
					if (  (ld = osg->linedef)
						&& ld->frontsector != ld->backsector
						&& osg->frontsector)	// should be set
					{
						return osg->frontsector - sectors;
					}
				}
			}
		}
	}

	return -1;
}

static void subfn_CheckLinesContained(int ln0, int ln1, sector_t *ref, int* chkline)
{
	sector_t	*chk, *use1 = NULL, *use2 = NULL;
	line_t		*ld;
	byte		*mrkssec = NULL;
	int			c, d, i, k, box[4], rc;
	int			error = false;

	if (!subfn_CheckForRing(hitline, ref, ln1, &rc, box, NULL, NULL) || rc < 3) return;

	for (c = i = 0; i < ref->linecount; i++)
	{
		ld = ref->lines[i];

		if (hitline[ld - lines] == 4)
		{
			chk = ld->frontsector == ref ? ld->backsector : ld->frontsector;

			if (chk && !use1)
			{
				if (chk->specFlags & 0x4 && !(chk->specFlags & 0x400)) return;
				use1 = chk;
			}
			else
				if (!chk || use1 != chk) return;

			chkline[c++] = ld - lines;
		}
	}

	if (   !use1
		|| !LS_MarkSubsecsInAreaGeneral(use1, hitline, 0, NULL, box, -1, MSIA_NONE)
		|| !hitssecptr)
	{
		return;
	}

	mrkssec = (byte*)calloc(numsubsectors, 1);

	for (i = 0; i < hitssecptr; i++) mrkssec[hitssec[i]] = true;
	for (i = 0; i < c; i++) hitline[linearray[i]] = 0;

	if (!subfn_CheckForRing(hitline, ref, ln0, &rc, box, NULL, NULL) || rc < 3)
	{
		M_Free(mrkssec); return;
	}

	for (i = 0; i < ref->linecount; i++)
	{
		ld = ref->lines[i];

		if (hitline[ld - lines] == 4)
		{
			chkline[c++] = ld - lines;

			if (!error)
			{
				chk = ld->frontsector == ref ? ld->backsector : ld->frontsector;

				if (chk && !use2)
				{
					error = chk->specFlags & 0x4 && !(chk->specFlags & 0x400);
					use2 = chk;
				}
				else
					error = (!chk || use2 != chk);
			}
		}
	}

	if (   error
		|| !use2
		|| (use1 != use2 && (use1->floorheight != use2->floorheight || use1->ceilingheight != use2->ceilingheight))
		|| !LS_MarkSubsecsInAreaGeneral(ref, hitline, 0, NULL, box, -1, MSIA_NONE))
	{
		M_Free(mrkssec); return;
	}

	validcount++;

	for (d = i = 0; i < c; i++)
	{
		lines[chkline[i]].validcount = validcount;
		hitline[chkline[i]] = 0;
		linearray[d++] = chkline[i];
	}

	for (i = 0; i < hitssecptr; i++)
	{
		if (!mrkssec[hitssec[i]])
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef && sg->linedef->validcount != validcount)
				{
					sg->linedef->validcount = validcount;
					linearray[d++] = sg->linedef - lines;

					if (sg->linedef->frontsector != sg->linedef->backsector)
					{
						error = true;
						break;
					}
				}
			}
		}

		if (error) break;
	}

	if (!error)
	{
		for (i = 0; i < hitssecptr; i++)
		{
			if (!mrkssec[hitssec[i]])
			{
				subsector_t *ss = &subsectors[hitssec[i]];
			
				SetOverride(ss, use2 - sectors, 48028);
			}
		}

		for (i = 0; i < d; i++)
		{
			ld = &lines[linearray[i]];

			if (origTEX[linearray[i]] & 0x22)
				Settype(ld, R3D_03, 48029);
			else
				Settype(ld, R3D_3F, 48030);

			SetLFG(ld, LFG_NEWSET, 48031);
		}
	}

	M_Free(mrkssec);
}

// v220-11 Function added. The code is similar to the first part of MX_Main
//         but it was felt easier to duplicate it for this function. It
//         looks for single loops of missing hi-textures where the reference
//         sectors are different.
// 
void MX_MultiSecHi(
	lring_t		*lr,
	sector_t	**secidx,
	int			*lstart,
	byte		*tmpline,
	byte		*mrkline
)
{
	sector_t	*fs, *bs;
	int			lrcnt;
	int			lscnt;
	int			seccnt;
	int			mt, f1, f2;
	int			i, k, rc;
	//-----------------------

	CLR_hitline();
	SET_countsec(0);

	for (seccnt = i = 0; i < numlines; i++)
	{
		if (lineflags[i]) continue;

		fs = lines[i].frontsector;
		bs = lines[i].backsector;

		if (  !bs
			|| fs == bs
			|| !((mt = lines[i].mtflags) & R3D_MISHITEX)
			|| fs->heightsec >= 0
			|| bs->heightsec >= 0
			|| fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR
			|| fs->ceilingheight - fs->floorheight <= 8*FRACUNIT
			|| bs->ceilingheight - bs->floorheight <= 8*FRACUNIT)
		{
			continue;
		}

		f1 = fs->specFlags & SF_SSLINETRUE ? 0 : fs->specFlags & 0x8;
		f2 = bs->specFlags & SF_SSLINETRUE ? 0 : bs->specFlags & 0x8;

		if ( (!f1 && !f2) || ( f1 &&  f2) )
		{
			continue;
		}

		if (hiValid(fs, bs, origTEX[i]))
		{
			hitline[i] = 1;

			if (f1 && !countsec[fs - sectors])
			{
				countsec[fs - sectors] = true;
				secidx[seccnt++] = fs;
			}
			else
			if (f2 && !countsec[bs - sectors])
			{
				countsec[bs - sectors] = true;
				secidx[seccnt++] = bs;
			}
		}
	}

	memcpy(tmpline, hitline, numlines);

	for (lrcnt = i = 0; i < seccnt; i++)
	{
		RDBOOL	repeat = true;
		int		lnum;

		fs = secidx[i];

		while (repeat)
		{
			repeat = false;

			for (k = 0; k < fs->linecount; k++)
			{
				lnum = fs->lines[k] - lines;

				if (   hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, fs, lnum, &rc, lr[lrcnt].box, NULL, NULL)
					&& rc > 2)
				{
					repeat = true;

					lr[lrcnt].use = fs;
					lr[lrcnt].lnum = lnum;
					lr[lrcnt].sz = M_FPAreaBox(lr[lrcnt].box);

					lrcnt++;
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}

	// Sort the boxes into descending size
	qsort(lr, lrcnt, sizeof(lring_t), boxSize);

	// Scan the boxes
	for (i = 0; i < lrcnt; i++)
	{
		lscnt = 0;

		if (lr[i].done) continue;

		lr[i].done = true;

		lstart[lscnt++] = i;

		// check if any of the other boxes fit inside
		for (k = i + 1; k < lrcnt; k++)
		{
			if (!lr[k].done && M_FitBox(lr[k].box, lr[i].box))
			{
				lr[k].done = true;
				lstart[lscnt++] = k;
			}
		}

		memcpy(hitline, tmpline, numlines);

		fs = lines[lr[lstart[0]].lnum].frontsector;
		bs = lines[lr[lstart[0]].lnum].backsector;

		f1 = fs->specFlags & SF_SSLINETRUE ? 0 : fs->specFlags & 0xC;
		f2 = bs->specFlags & SF_SSLINETRUE ? 0 : bs->specFlags & 0xC;

		if (lscnt == 1)
		{
			sector_t	*use = lr[lstart[0]].use;
			int			num0 = lr[lstart[0]].lnum;
			int			box[4];
			line_t		*ld = &lines[num0];

			if (   hitline[num0] == 1
				&& subfn_CheckForRing(hitline, use, num0, &rc, box, NULL, NULL)
				&& rc > 2)
			{
				sector_t	*chk, *ref = f1 == 0xC ? bs : fs;
				RDBOOL		diffsec = false;
				int			cht = fs == use ? bs->ceilingheight : fs->ceilingheight;
				int			sky = fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum;

				//---------------------------------

				// Check for hitex missing rings referencing more than one sector
				for (k = 0; k < use->linecount; k++)
				{
					ld = use->lines[k];

					if (hitline[ld - lines] == 4)
					{
						chk = ld->backsector == use ? ld->frontsector : ld->backsector;

						if (chk != ref)
						{
							if (chk->ceilingheight != ref->ceilingheight)
							{
								diffsec = false; break;
							}

							diffsec = true;
						}
					}
				}

				if (diffsec)
				{
					if (   LS_MarkSubsecsInAreaGeneral(use, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& LS_OnlySSLinesInSSECarray(use, hitline))
					{
						for (k = 0; k < hitssecptr; k++)
							Setr3dsscs(&subsectors[hitssec[k]], ref - sectors, 48032);
					}
					else
					{
						diffsec = false;
					}
				}

				for (k = 0; k < use->linecount; k++)
				{
					ld = use->lines[k];

					if (hitline[ld - lines] == 4)
					{
						hitline[ld - lines] = 0;

						// don't mark off lines with missing lo tex
						if (diffsec && !(ld->mtflags & R3D_MISLOTEX))
							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48033);
					}
				}
			}
		}
	}
}

static RDBOOL rejectVTX(vertexowner_t *vtx)
{
	int a, i, f;

	if (vtx->num < 4) return false;

	for (a = i = 0; i < vtx->num; i++)
	{
		f = (sectors[vtx->list[i]].specFlags & 0xC);

		if (!f) a |= 1;
		else
		if (f == 0x4) a |= 2;
		else
		if (f == 0x8) a |= 4;
		else
		if (f == 0xC) a |= 8;
	}

	return a == 0xF;
}

void subfn_MXHIcontained(int secn, int *chkline)
{
	sector_t *sc = &sectors[secn];
	line_t	*ld;
	int		i, k;

	if (	sc->linecount < 6	// min would be triangle inside a triangle
		|| (sc->specFlags & (SF_SSLINETRUE|SF_SECCLOSED|SF_NOHITEX|SF_MOVSECTOR)) != (SF_SECCLOSED|SF_NOHITEX)
		|| sc->ceilingpic == skyflatnum)
	{
		// v220-33 check 3 line door sectors
		if (   sc->linecount == 3
			&& (sc->specFlags & (SF_MOVFLOOR|SF_MOVLOWER|SF_NOHITEX|SF_NOLOTEX))
			== (SF_MOVFLOOR|SF_MOVLOWER|SF_NOHITEX|SF_NOLOTEX))
		{
			sector_t *os = NULL;
			int		 mtn;

			for (i = 0; i < 3; i++)
			{
				if ( !(ld = sc->lines[i])->backsector
					|| ld->frontsector == ld->backsector
					|| ld->specType >= R3D_40
					|| ld->spflags & SPF_CONTROLLINE )
				{
					return;
				}

				if (!origTEX[ld - lines])
				{
					mtn = ld->frontsector == sc ? 0x40 : 0x10;
					if (mtn == 0x10) os = ld->frontsector;
					if (mtn == 0x40) os = ld->backsector;
				}
			}

			if (os)
			{
				subsector_t *ss;

				setSsecList(sc, true);

				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					Setr3dssfs(ss, os - sectors, 48300);
					Setr3dsscs(ss, os - sectors, 48300);
				}
				Setsecdone(sc, 48300);
			}
		}
		return;
	}

	int		lcount = 0, scount = 0;

	// check all lines in the sector
	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (!ld->backsector || ld->mtflags & R3D_MISLOTEX) return;
		if (ld->specType >= R3D_40 || ld->spflags & SPF_CONTROLLINE) continue;

		if (   vtxflags[ld->v1 - vertexes] == 3
			&& vtxflags[ld->v2 - vertexes] == 3
			&& vertexowners[ld->v1 - vertexes].num == 2
			&& vertexowners[ld->v2 - vertexes].num == 2)
		{
			int temp = -1;

			if (vertexowners[ld->v1 - vertexes].list[0] != secn)
				temp = vertexowners[ld->v1 - vertexes].list[0];
			else
			if (vertexowners[ld->v1 - vertexes].list[1] != secn)
				temp = vertexowners[ld->v1 - vertexes].list[1];

			if (temp >= 0)
			{
				if (sectors[temp].specFlags & SF_MOVSECTOR) return;

				for (k = 0; k < scount; k++)
				{
					if (countsec[k] == temp) goto inlist;
				}
				countsec[scount++] = temp;
inlist:;
			}

			chkline[lcount++] = ld - lines;
		}
		else
		{
			scount = 0; break;
		}
	}

	if (scount != 2) return;

	int usesec = -1;

	for (i = 0; i < scount; i++)
	{
		int flags = sectors[countsec[i]].specFlags;

		if (flags & SF_SSLINETRUE || !(flags & SF_NOHITEX) )
		{
			if (usesec >= 0) return;
			usesec = countsec[i];
		}
	}
	

	if (usesec < 0 || sectors[usesec].ceilingheight > sc->ceilingheight) return;

	// get the outer box
	fixed_t bbox[4];

	for (i = k = 0; i < lcount; i++)
	{
		ld = &lines[chkline[i]];
		if (ld->frontsector - sectors == usesec || ld->backsector - sectors == usesec)
		{
			M_AddLineToBox(bbox, ld); k++;
		}
	}

	if (k < 3) return;

	// check all other lines fit inside the box
	for (i = 0; i < lcount; i++)
	{
		ld = &lines[chkline[i]];
		if (ld->frontsector - sectors != usesec && ld->backsector - sectors != usesec)
		{
			if (ld->v1->x >= bbox[BOXRIGHT] || ld->v1->x <= bbox[BOXLEFT]) return;
			if (ld->v2->x >= bbox[BOXRIGHT] || ld->v2->x <= bbox[BOXLEFT]) return;
			if (ld->v1->y >= bbox[BOXTOP] || ld->v1->y <= bbox[BOXBASE]) return;
			if (ld->v2->y >= bbox[BOXTOP] || ld->v2->y <= bbox[BOXBASE]) return;
		}
	}

	// If we get here then now find all subsectors in sc
	setSsecList(sc, true);

	for (i = 0; i < hitssecptr; i++)	// should have been named hitsseccount but hey ho..
		Setr3dsscs(&subsectors[hitssec[i]], usesec, 48000);

	for (i = 0; i < lcount; i++)
	{
		ld = &lines[chkline[i]];
		if (!origTEX[chkline[i]])
			Settype(ld, R3D_3F, 48000);
		else
			Settype(ld, R3D_36, 48000);
		SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48000);
		Clr_allmtflags(ld, 48000);
	}
}

// Check nested textureless (hi or lo) rings in static sectors
void MX_Main()
{
	sector_t	*fs, *bs;
	sector_t	**secidx;
	lring_t		*lr;
	byte		*tmpline;
	byte		*mrkline;
	int			*lstart;
	int			*chkline;
	int			i, k, m, rc;
	int			lrcnt;
	int			lscnt;
	int			seccnt;
	int			mt, txn;
	int			f1, f2;

	//-----------------------

	if (numsectors < 2 || numlines < 6) return;

	lr      = (lring_t*  )calloc(numlines, sizeof(lring_t));
	secidx  = (sector_t**)malloc(numsectors * sizeof(sector_t*));
	mrkline = (byte*     )malloc(numlines);
	tmpline = (byte*     )malloc(numlines);
	lstart  = (int*      )malloc(sizeof(int) * numlines);
	chkline = (int*      )malloc(ma_inf.maxseclinecount * sizeof(int));

	// v220-33 check for simple loop in loop cases first
	for (i = 0; i < numsectors; i++) subfn_MXHIcontained(i, chkline);

	MX_MultiSecHi(lr, secidx, lstart, tmpline, mrkline);

	memset(lr, 0, numlines * sizeof(lring_t));
	CLR_hitline();
	SET_countsec(0);

	for (seccnt = i = 0; i < numlines; i++)
	{
		if (lineflags[i]) continue;

		fs = lines[i].frontsector;
		bs = lines[i].backsector;

		if (   !bs
			|| fs == bs
			|| fs->heightsec >= 0
			|| bs->heightsec >= 0
			|| fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR
			|| fs->ceilingheight - fs->floorheight <= 8*FRACUNIT
			|| bs->ceilingheight - bs->floorheight <= 8*FRACUNIT
			|| Spans1sLines(&lines[i]))
		{
			continue;
		}

		if (   fs->floorheight == bs->floorheight
			&& fs->ceilingheight == bs->ceilingheight)
		{
			continue;
		}

		if (fs->ceilingpic != skyflatnum)
		{
			if (   rejectVTX(&vertexowners[lines[i].v1 - vertexes])
				|| rejectVTX(&vertexowners[lines[i].v2 - vertexes]) )
			{
				continue;
			}
		}

		f1 = fs->specFlags & SF_SSLINETRUE ? 0 : fs->specFlags & 0xC;
		f2 = bs->specFlags & SF_SSLINETRUE ? 0 : bs->specFlags & 0xC;

		if (f1 != 0xC && f2 != 0xC) continue;

		mt = lines[i].mtflags;
		txn = origTEX[i];

		if (mt & R3D_INVLTEX1 && mt & R3D_MISLTEX2) txn &= ~0x01;
		if (mt & R3D_INVLTEX2 && mt & R3D_MISLTEX1) txn &= ~0x10;

		if (loValid(fs, bs, txn) || hiValid(fs, bs, txn))
		{
			hitline[i] = 1;

			if (!countsec[fs - sectors])
			{
				countsec[fs - sectors] = true;
				secidx[seccnt++] = fs;
			}

			if (!countsec[bs - sectors])
			{
				countsec[bs - sectors] = true;
				secidx[seccnt++] = bs;
			}
		}
	}

	memcpy(tmpline, hitline, numlines);

	for (lrcnt = i = 0; i < seccnt; i++)
	{
		RDBOOL	repeat = true;
		int		lnum;

		fs = secidx[i];

		while (repeat)
		{
			repeat = false;

			for (k = 0; k < fs->linecount; k++)
			{
				lnum = fs->lines[k] - lines;

				if (   hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, fs, lnum, &rc, lr[lrcnt].box, NULL, NULL)
					&& rc > 2)
				{
					repeat = true;

					lr[lrcnt].use = fs;
					lr[lrcnt].lnum = lnum;
					lr[lrcnt].sz = M_FPAreaBox(lr[lrcnt].box);

					lrcnt++;
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}

	// Sort the boxes into descending size
	qsort(lr, lrcnt, sizeof(lring_t), boxSize);

	// Scan the boxes
	for (i = 0; i < lrcnt; i++)
	{
		lscnt = 0;

		if (lr[i].done) continue;

		lr[i].done = true;

		lstart[lscnt++] = i;

		// check if any of the other boxes fit inside
		for (k = i + 1; k < lrcnt; k++)
		{
			if (!lr[k].done && M_FitBox(lr[k].box, lr[i].box))
			{
				lr[k].done = true;
				lstart[lscnt++] = k;
			}
		}

		memcpy(hitline, tmpline, numlines);

		fs = lines[lr[lstart[0]].lnum].frontsector;
		bs = lines[lr[lstart[0]].lnum].backsector;

		f1 = fs->specFlags & SF_SSLINETRUE ? 0 : fs->specFlags & 0xC;
		f2 = bs->specFlags & SF_SSLINETRUE ? 0 : bs->specFlags & 0xC;

		if (lscnt == 1)
		{
			sector_t	*use = lr[lstart[0]].use;
			int			num0 = lr[lstart[0]].lnum;
			int			box[4];
			int			error = false;
			line_t		*ld = &lines[num0];
			int			chklinecnt;

			if (   hitline[num0] == 1
				&& subfn_CheckForRing(hitline, use, num0, &rc, box, NULL, NULL)
				&& rc > 2)
			{
				sector_t	*ref = f1 == 0xC ? bs : fs;
				RDBOOL		validhi = true, validlo = true;
				int			cht = fs == use ? bs->ceilingheight : fs->ceilingheight;
				int			fht = fs == use ? bs->floorheight   : fs->floorheight  ;
				int			sky = fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum;
				int			c;

				sector_t	*inv;
				int			invalidhicheck, boxinvalid[4];
				int			invlnum = -1;
				//---------------------------------

				M_ClearBox(boxinvalid);

				// Check for a descending ceiling
				for (invalidhicheck = k = 0, inv = NULL; k < use->linecount; k++)
				{
					ld = use->lines[k];

					if (hitline[ld - lines] == 4)
					{
						if (   !(origTEX[ld - lines] & 0x44)
							&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
						{
							if (invlnum < 0) invlnum = ld - lines;

							if (ld->frontsector->ceilingheight > ld->backsector->ceilingheight)
							{
								if (!inv) inv = ld->backsector; else if (inv != ld->backsector) break;
								invalidhicheck++;
								M_AddLineToBox(boxinvalid, ld);
							}
							else
							{
								if (!inv) inv = ld->frontsector; else if (inv != ld->frontsector) break;
								invalidhicheck++;
								M_AddLineToBox(boxinvalid, ld);
							}
						}
						else
						{
							break;
						}
					}
				}

				if (   inv
					&& invalidhicheck == rc
					&& LS_MarkSubsecsInAreaGeneral(inv, hitline, 0, NULL, boxinvalid, -1, MSIA_NONE))	// v220-22 $bugfix (6th parm was 0)
				{
					validhi = false;
				}

				for (c = chklinecnt = k = 0; k < use->linecount; k++)
				{
					ld = use->lines[k];

					if (hitline[ld - lines] == 4)
					{
						sector_t *chk = ld->frontsector == use ? ld->backsector  : ld->frontsector;
						sector_t *rf2 = ld->frontsector == ref ? ld->frontsector : ld->backsector;

						if (ref != rf2) error = true;

						if (validhi && chk->ceilingheight != cht)
						{
							validhi = false;
						}

						if (validlo && chk->floorheight != fht)
						{
							validlo = false;
						}

						linearray[c++] = ld - lines;

						if (!error && !origTEX[ld - lines])
						{
							if (   (vtxarray[ld->v1 - vertexes] == 2 && vtxarray[ld->v2 - vertexes] > 2)
								|| (vtxarray[ld->v2 - vertexes] == 2 && vtxarray[ld->v1 - vertexes] > 2))
							{
								chkline[chklinecnt++] = ld - lines;
							}
						}
					}
				}

				if (error)
				{
					for (k = 0; k < c; k++) hitline[linearray[k]] = 8;
					continue;
				}

				memcpy(mrkline, hitline, numlines);

				if (!validhi || !validlo)
				{
					for (k = 0; k < use->linecount; k++)
					{
						ld = use->lines[k];

						if (hitline[ld - lines] == 4)
						{
							Settype(ld, R3D_36, 48034);
							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48035);
							Clr_allmtflags(ld, 48036);
						}
					}
				}
				else if (c > 2)
				{
					int		flag = MSIA_NONE;
					int		ok = LS_MarkSubsecsInLoop(use, c, NULL, box, MSIA_TESTAREA, NULL);
					int		ret;

					if (!ok) flag = MSIA_REVERSED;

					if (LS_MarkSubsecsInLoop(use, c, NULL, box, flag, NULL))
					{
						sector_t	*chk, *other = fs == ref ? bs : fs;
						RDBOOL		lofill = false;
						RDBOOL		hifill = false;
						int			type = 0;

						//--------------------------------

						if ( (ret = checkLines(ref, &c, &type, mrkline)) < 0) continue;

						if (ret & 2) continue;

						if (ret == 1)
						{
							if (validhi && validlo)
							{
								int h1 = lines[num0].frontsector->ceilingheight;
								int h2 = lines[num0].backsector->ceilingheight;
								int d, h3;
								int *container;

								// ----------------------------

								if (h1 != h2) continue;
								
								h2 = lines[num0].frontsector->floorheight;
								h3 = lines[num0].backsector->floorheight;

								// check all internal lines
								for (d = k = 0; k < c; k++)
								{
									ld = &lines[linearray[k]];

									if (!Vtxs2(ld)) { ret |= 2; break; }

									if (mrkline[linearray[k]] != 4)
									{
										if (   ld->frontsector != ld->backsector
											&& (lxarray[ld - lines]
											|| ld->frontsector->ceilingheight != h1
											|| ld->backsector->ceilingheight != h1
											|| ld->frontsector->floorheight != ld->backsector->floorheight
											|| (ld->frontsector->floorheight != h2 && ld->frontsector->floorheight != h3)) )
										{
											d = 0; break;
										}
										d++;
									}
								}

								if (d < 3) continue;

								container = (int*)malloc(hitssecptr * sizeof(int));

								for (k = 0; k < hitssecptr; k++)
								{
									subsector_t *ss = &subsectors[hitssec[k]];

									container[k] = getSSECowner(ss);

									if (container[k] < 0) { d = 0; break; }
								}

								if (d)
								{
									for (k = 0; k < hitssecptr; k++)
									{
										subsector_t *ss = &subsectors[hitssec[k]];

										Setr3dsscs(ss, container[k], 48037);
										Setr3dssfs(ss, ref - sectors, 48038);

										if (!secdone[container[k]])
											Setsecdone(sectors + container[k], 48039);
									}

									Setsecdone(ref, 48040);
								}

								M_Free(container);

								if (!d) continue;

								validlo = validhi = false;
							}
							else
								continue;
						}

						if (validlo && ref->floorheight < other->floorheight)
						{
							lofill = fillLines(use, other, ref, mrkline, 0);
						}
						
						if (validhi && ref->ceilingheight > other->ceilingheight)
						{
							hifill = fillLines(use, other, ref, mrkline, 1);
						}
						
						if (sky && validhi && validlo && !hifill && !lofill) continue;

						for (k = 0; k < c; k++)
						{
							ld = &lines[linearray[k]];

							if (origTEX[ld - lines] & 0x22)
							{
								// v220-28 test added
								if (   ld->backsector
									&& abs(ld->frontsector->floorheight - ld->backsector->floorheight) > 1024*FRACUNIT)
								{
									Settype(ld, R3D_03, 48041);
								}
								else
								{
									Settype(ld, R3D_1D, 48041);
								}
							}
							else
							{
								Settype(ld, R3D_3F, 48042);
							}

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_FIXABS, 48043);
							Clr_allmtflags(ld, 48044);
						}

						if (hitssecptr && (validhi|validlo))
						{
							RDBOOL sameceil  = !(ref->specFlags & SF_MOVCEIL ) && (ref->specFlags & 0x400 || !(ref->specFlags & 0x8));
							RDBOOL samefloor = !(ref->specFlags & SF_MOVFLOOR) && (ref->specFlags & 0x400 || !(ref->specFlags & 0x4));

							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];

								if (validhi && ss->sector != ref)
								{
									if (   sameceil
										&& !(ss->sector->specFlags & SF_MOVCEIL)
										&& ss->sector->ceilingheight == ref->ceilingheight)
									{
										// do nothing
									}
									else if (hifill)
										Setr3dsslcsec(ss, ref - sectors, 48045);
									else if (ss->sector->ceilingheight != ref->ceilingheight)
										Setr3dsscs(ss, ref - sectors, 48046);
								}
								if (validlo)
								{
									if (   samefloor
										&& !(ss->sector->specFlags & SF_MOVFLOOR)
										&& ss->sector->floorheight == ref->floorheight)
									{
										// do nothing
									}
									else if (lofill)
									{
										if (ss->sector != ref)
											Setr3dsslfsec(ss, ref - sectors, 48047);
										else if (flag == MSIA_REVERSED)
											Setr3dsslfsec(ss, other - sectors, 48048);
									}
									else if (ss->sector->floorheight != ref->floorheight)
									{
										Setr3dssfs(ss, ref - sectors, 48049);
									}
								}
							}

							// Check for adjoining subsec inheritance. If facing a visible missing tex line
							// on the missing side software renderers just extend the flat (this is why
							// walls needing missing textures to be drawn look odd with software renderers).
							// Rather than mapping a texture to a wall, as is normally the case in R3D, in
							// this instance the flat, instead, is extended as long as a compliant line is
							// found in the subsec.	v220-10
							if (lofill && chklinecnt)
							{
								for (k = 0; k < chklinecnt; k++)
								{
									ssline_t *sl = &sslines[chkline[k]];

									ld = &lines[chkline[k]];

									if (ld->frontsector->specFlags & 0x4) chk = ld->frontsector;
									else if (ld->backsector->specFlags & 0x4) chk = ld->backsector;
									else
										chk = NULL;

									if (chk)
									{
										if (chk->floorheight > ref->floorheight)
										{
											for (m = 0; m < sl->num; m++)
											{
												subsector_t *ss = &subsectors[sl->list[m]];

												if (   ss->sector == ref
													&& ss->floorsec < 0
													&& ss->lfsec < 0
													&& ssLoop(ss, chk))	// if this fails not a lot more can be done
												{
													Setr3dsslfsec(ss, chk - sectors, 48050);
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		else if (lscnt == 2)
		{
			int		num0 = lr[lstart[0]].lnum;
			int		num1 = lr[lstart[1]].lnum;
			line_t	*ld = &lines[num0];

			if (   !(f1 & 0xC)
				|| !(f2 & 0xC)
				|| (bs && fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum && ( !(f1 & 0x4) || !(f2 & 0x4))) )
			{
				sector_t	*use;
				line_t		*li = &lines[num1];
				RDBOOL		contained = false;
				int			box[4];
				int			cht = fs->specFlags & 0xC ? fs->ceilingheight : bs->ceilingheight;
				int			fht = fs->specFlags & 0xC ? fs->floorheight   : bs->floorheight;
				int			ok = 0;

				//---------------------------------

				// check that the inner ring has no tex sectors. These types are dealt with elsewhere
				if ((li->frontsector->specFlags & 0xC) != 0xC || (li->backsector->specFlags & 0xC) != 0xC)
				{
					if (   li->frontsector->floorheight == li->backsector->floorheight
						|| origTEX[li - lines] & 0x11)
					{
						continue;
					}

					if (   li->frontsector->floorheight > li->backsector->floorheight
						&& (li->frontsector->specFlags & 0xC) != 0xC)
					{
						if (  (lr[lstart[1]].use->specFlags & 0xC) == 0xC
							&& lr[lstart[0]].use == lr[lstart[1]].use)
						{
							subfn_CheckLinesContained(num0, num1, lr[lstart[1]].use, chkline);
						}

						continue;
					}

					if (   li->frontsector->floorheight < li->backsector->floorheight
						&& (li->backsector->specFlags & 0xC) != 0xC)
					{
						if (  (lr[lstart[1]].use->specFlags & 0xC) == 0xC
							&& lr[lstart[0]].use == lr[lstart[1]].use)
						{
							subfn_CheckLinesContained(num0, num1, lr[lstart[1]].use, chkline);
						}

						continue;
					}

					contained = true;
				}

				if (hitline[num1] == 1)	// should be
				{
					use = lr[lstart[1]].use;
					ok = subfn_CheckForRing(hitline, use, num1, &rc, box, NULL, NULL) && rc > 2;
				}

				if (ok)
				{
					line_t *li = &lines[num1];
					RDBOOL validhi = li->frontsector->ceilingheight >= cht && li->backsector->ceilingheight >= cht;
					RDBOOL validlo = li->frontsector->floorheight   <= fht && li->backsector->floorheight   <= fht;

					// Check for unusual case where a textureless sector is used for an intermediate floor
					if (contained)
					{
						int	chk, losec = -1;

						for (k = 0; k < use->linecount; k++)
						{
							ld = use->lines[k];

							if (hitline[ld - lines] == 4)
							{
								chk = ld->frontsector == use ? ld->backsector - sectors : ld->frontsector - sectors;

								if (losec >= -1)
								{
									if (losec == -1) losec = chk;
									else if (losec != chk) losec = -2;
								}

								hitline[ld - lines] = 64;
							}
						}

						if (losec >= 0)
						{
							ld = &lines[num0];

							if (   hitline[ld - lines] == 1
								&& validhi
								&& validlo
								&& lr[lstart[0]].use == use
								&& (use->specFlags & 0xC) == 0xC
								&& subfn_CheckForRing(hitline, use, num0, &rc, box, NULL, NULL)
								&& rc > 2)
							{
								int c;

								for (c = k = 0; k < use->linecount; k++)
								{
									ld = use->lines[k];

									if (hitline[ld - lines] == 4)
									{
										hitline[ld - lines] = 64;
										linearray[c++] = ld - lines;
									}
								}
								
								if (   LS_MarkSubsecsInMultiSecLoop(0, NULL, c, box, NULL, true, MSIA_NONE)
									&& LS_OnlySSLinesInSSECarray(NULL, hitline))
								{
									for (k = 0; k < hitssecptr; k++)
									{
										subsector_t *ss = &subsectors[hitssec[k]];
										seg_t		*sg = &segs[ss->firstline];

										for (m = 0; m < ss->linecount; m++, sg++)
										{
											if ( (ld = sg->linedef) && ld->validcount != validcount)
											{
												ld->validcount = validcount;

												if (origTEX[ld - lines] & 0x22)
													Settype(ld, R3D_03, 48051);
												else
													Settype(ld, R3D_3F, 48052);

												SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48053);
												Clr_allmtflags(ld, 48054);
											}
										}

										Setr3dsslfsec(ss, losec, 48055);
										Setr3dssfs(ss, use - sectors, 48056);
									}

									Setsecdone(use, 48057);
									Setsecdone(sectors + losec, 48058);
								}
							}
						}

						for (k = 0; k < use->linecount; k++)
						{
							ld = use->lines[k];
							if (hitline[ld - lines] == 64) hitline[ld - lines] = 0;
						}

						continue;
					}

					if (validhi || validlo)
					{
						for (k = 0; k < use->linecount; k++)
						{
							ld = use->lines[k];

							if (hitline[ld - lines] == 4)
							{
								if (validhi && ld->frontsector->ceilingheight != cht && ld->backsector->ceilingheight != cht)
								{
									validhi = false;
								}
								if (validlo && ld->frontsector->floorheight != fht && ld->backsector->floorheight != fht)
								{
									validlo = false;
								}
							}

							if (!validhi && !validlo) break;
						}
					}

					ok = 0;

					if (!validhi || !validlo || (fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum))
					{
						int flag = MSIA_NONE;
						int c;

						ld = &lines[num0];
						use = lr[lstart[0]].use;

						if (   hitline[ld - lines] == 1
							&& subfn_CheckForRing(hitline, use, num0, &rc, box, NULL, NULL)
							&& rc > 2)
						{
							for (c = k = 0; k < use->linecount; k++)
							{
								ld = use->lines[k];

								if (hitline[ld - lines] == 4)
								{
									linearray[c++] = ld - lines;
								}
							}
						
							ok = LS_MarkSubsecsInLoop(use, c, NULL, box, MSIA_TESTAREA, NULL);

							if (!ok) flag = MSIA_REVERSED;

							ok = LS_MarkSubsecsInLoop(use, c, NULL, box, flag, NULL);
						}
					}

					if (ok)
					{
						RDBOOL ceilfill  = li->frontsector->ceilingpic == skyflatnum
										&& li->backsector->ceilingpic  == skyflatnum;
						RDBOOL floodfill = false;

						if (!ceilfill)
						{
							floodfill = li->frontsector->ceilingheight - li->frontsector->floorheight <= FRACUNIT
									 || li->backsector->ceilingheight  - li->backsector->floorheight  <= FRACUNIT;
						}

						validcount++;
						use = NULL;

						if (validhi || validlo) use = fs->specFlags & 0xC ? bs : fs;

						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];
							seg_t		*sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( (ld = sg->linedef) && ld->validcount != validcount)
								{
									ld->validcount = validcount;

									if (origTEX[ld - lines] == 0x22)
									{
										Settype(ld, R3D_03, 48059);
									}
									else if (ceilfill)
									{
										Settype(ld, R3D_14, 48060);
									}
									else
									if (floodfill && !(origTEX[ld - lines] & 0x22))
									{
										Settype(ld, R3D_3F, 48061);
									}
									else
									{
										Settype(ld, R3D_36, 48062);
									}

									SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_NOCHANGE, 48063);
									Clr_allmtflags(ld, 48064);
								}
							}

							if (!ceilfill)
							{
								if ((floodfill || validhi) && ss->sector->ceilingheight != use->ceilingheight)
									Setr3dsscs(ss, use - sectors, 48065);
								if ((floodfill || validlo) && ss->sector->floorheight != use->floorheight)
									Setr3dsslfsec(ss, use - sectors, 48066);
							}
						}
					}
				}
			}
		}
	}

	M_Free(chkline);
	M_Free(lstart);
	M_Free(tmpline);
	M_Free(mrkline);
	M_Free(secidx);
	M_Free(lr);
}

//--------------------------------------------------------------------------------

void MX_HeightsecMissing()
{
	sector_t	*s, *fs, *bs, *chk;
	line_t		*ld;
	int			c, i, k, n, fl, lnum, txn;

	for (i = 0; i < numsectors; i++)
	{
		if (   sectors[i].heightsec >= 0
			&& (fl = sectors[sectors[i].heightsec].specFlags) & 0xC
			&& !(fl & SF_MOVSECTOR))
		{
			CLR_hitline();

			s = &sectors[i];

			for (c = k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];
				lnum = ld - lines;

				if (   !lineflags[lnum]
					&& ld->ml_flags & ML_TWOSIDED
					&& (fs = ld->frontsector) != (bs = ld->backsector))
				{
					chk = fs == s ? bs : fs;
					fl = chk->specFlags;

					if (chk->heightsec > 0 || !(fl & 0x400 || !(fl & 0xC)) )
						continue;

					txn = origTEX[lnum];

					if (   (!(txn & 0x44) && abs(s->ceilingheight - chk->ceilingheight) > 8*FRACUNIT)
						|| (!(txn & 0x11) && abs(s->floorheight   - chk->floorheight  ) > 8*FRACUNIT) )
					{
						hitline[lnum] = 1; c++;
					}
				}
			}

			if (c < 3) continue;

			validcount++;

			while(1)
			{
				int		rc, box[4];
				RDBOOL	valid = false;
				//------------------

				lnum = -1;

				for (k = 0; k < s->linecount; k++)
				{
					n = s->lines[k] - lines;
					if (hitline[n] == 1) { lnum = n; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
				{
					sector_t	*osec = NULL;
					int			validhi = 0, validlo = 0;
					int			hi, lo;
					//------------------

					for (c = k = 0; k < s->linecount; k++)
					{
						ld = s->lines[k];
						n = ld - lines;

						if (hitline[n] == 4)
						{
							chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

							if (!osec) osec = chk;
							else if (osec != chk) { osec = NULL; break; }

							if (origTEX[ld - lines] & 0x44) validhi++;
							if (origTEX[ld - lines] & 0x11) validlo++;
						}
					}

					if (osec)
					{
						chk = &sectors[s->heightsec];
						hi = osec->ceilingheight < chk->ceilingheight - 8*FRACUNIT;
						lo = osec->floorheight > chk->floorheight + 8*FRACUNIT;

						valid = (validhi < 2 && hi) || (validlo < 2 && lo);
					}

					if (   valid
						&& LS_MarkSubsecsInAreaGeneral(osec, hitline, 0, NULL, box, -1, MSIA_REVERSED)
						&& LS_OnlySSLinesInSSECarray(s, hitline))
					{
						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];
							seg_t		*sg = &segs[ss->firstline];

							if (validhi < 2 && hi) Setr3dsscs(ss, osec - sectors, 48067);
							if (validlo < 2 && lo) Setr3dssfs(ss, osec - sectors, 48068);

							for (n = 0; n < ss->linecount; n++)
							{
								if (  (ld = sg->linedef)
									&& hitline[ld - lines] == 4
									&& ld->validcount != validcount)
								{
									ld->validcount = validcount;
									linearray[c++] = ld - lines;
								}
							}
						}
					}

					if (valid)
					{
						for (k = 0; k < c; k++)
						{
							Settype(&lines[linearray[k]], R3D_36, 48069);
							SetLFG(&lines[linearray[k]], LFG_NEWSET|LFG_FIXED, 48070);
						}
					}

					for (k = 0; k < s->linecount; k++)
					{
						n = s->lines[k] - lines;

						if (hitline[n] == 4) hitline[n] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}
}

// v220-33
static RDBOOL secCeilSame(sector_t *s1, sector_t *s2)
{
	return s1->ceilingheight == s2->ceilingheight
		&& s1->ceilingpic == s2->ceilingpic
		&& s1->lightlevel == s2->lightlevel
		&& s1->ceilinglightsec == s2->ceilinglightsec
		&& !(s1->specFlags & SF_MOVCEIL)
		&& !(s2->specFlags & SF_MOVCEIL)
		&& s1->heightsec == s2->heightsec;
}

// v220-33
static RDBOOL secFloorSame(sector_t *s1, sector_t *s2)
{
	return s1->floorheight == s2->floorheight
		&& s1->floorpic == s2->floorpic
		&& s1->lightlevel == s2->lightlevel
		&& s1->floorlightsec == s2->floorlightsec
		&& !(s1->specFlags & SF_MOVFLOOR)
		&& !(s2->specFlags & SF_MOVFLOOR)
		&& s1->heightsec == s2->heightsec;
}

//--------------------------------------------------------------------------------

static RDBOOL noLoTex(sector_t *s, byte *array, int flag)
{
	int c, d, i;
	int tx1 = 0x01, tx2 = 0x10;

	if (flag & MSIA_REVERSED) tx1 = 0x10, tx2 = 0x01;

	for (c = d = i = 0; i < s->linecount; i++)
	{
		int n = s->lines[i] - lines;

		if (array[n] == 4)
		{
			d++;

			if (    s->lines[i]->frontsector->specFlags & SF_MOVSECTOR
				&&	s->lines[i]->backsector->specFlags  & SF_MOVSECTOR)
			{
				continue;
			}

			if (lines[n].frontsector == s && !(origTEX[n] & tx1)) c++;
			if (lines[n].backsector  == s && !(origTEX[n] & tx2)) c++;
		}
	}

	return d - c < 2 || (d > 4 && d - c < 3);
}

static RDBOOL noHiTex(sector_t *s, byte *array, int flag)
{
	int c, d, i;
	int tx1 = 0x04, tx2 = 0x40;

	if (flag & MSIA_REVERSED) tx1 = 0x40, tx2 = 0x04;

	for (c = d = i = 0; i < s->linecount; i++)
	{
		int n = s->lines[i] - lines;

		if (array[n] == 4)
		{
			d++;

			if (    s->lines[i]->frontsector->specFlags & SF_MOVSECTOR
				&&	s->lines[i]->backsector->specFlags  & SF_MOVSECTOR)
			{
				continue;
			}

			if (lines[n].frontsector == s && !(origTEX[n] & tx1)) c++;
			if (lines[n].backsector  == s && !(origTEX[n] & tx2)) c++;
		}
	}

	return d - c < 2;
}

static void checkInfill0(int idx)
{
	sector_t *chk, *sc = &sectors[idx];
	line_t	 *ld;
	byte	 *mrkline = (byte*)malloc(numlines);
	int		 c, i, n, box[4];

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (   ld->ml_flags & ML_TWOSIDED
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& ld->frontsector != ld->backsector)
		{
			n = ld - lines;

			if (   !(origTEX[n] & 0x22)
				&& !lxarray[n]
				&& !(lineflags[n] & ~LFG_NHCONNOBSLN))
			{
				chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if ( !(chk->specFlags & SF_MOVSECTOR) ) c++;

				hitline[n] = 1;
			}
		}
	}

	if (c > 2)
	{
		RDBOOL repeat = true;

		while (repeat)
		{
			int		rc = 0;
			int		lnum = -1;

			RDBOOL	setarray = false;
			RDBOOL	validhi = false;
			RDBOOL	validlo = false;
			RDBOOL	valid_ring = false;

			//------------------------------------------

			repeat = false;

			for (i = 0; i < sc->linecount; i++)
			{
				int n = sc->lines[i] - lines;

				// eliminate movsec lines
				if (   hitline[n] == 1
					&& !(sc->lines[i]->frontsector->specFlags & SF_MOVSECTOR
					&&   sc->lines[i]->backsector->specFlags  & SF_MOVSECTOR) )
				{
					lnum = n; repeat = true; break;
				}
			}

			if (!repeat) break;

			if (!subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
			{
				hitline[lnum] = 0; continue;
			}
			
			if (rc > 2)
			{
				sector_t	*ns = NULL;
				sector_t	*os[2] = { NULL, NULL };
				int			osc[2] = { 0, 0 };
				int			f = 0;

				// get the adopting sector. We have to fart about because one wrong line sector
				// should not prevent the required adoption

				for (i = 0; i < sc->linecount; i++)
				{
					int n = sc->lines[i] - lines;

					if (hitline[n] == 4)
					{
						if (    sc->lines[i]->frontsector->specFlags & SF_MOVSECTOR
							&&	sc->lines[i]->backsector->specFlags  & SF_MOVSECTOR)
						{
							continue;
						}

						chk = sc->lines[i]->frontsector == sc ? sc->lines[i]->backsector : sc->lines[i]->frontsector;

						if (!os[0])
						{
							os[0] = chk; osc[0]++;
						}
						else if (os[0] == chk)
						{
							osc[0]++;
						}
						else if (os[0]->floorheight == chk->floorheight
							&& os[0]->ceilingheight == chk->ceilingheight
							&& os[0]->floorpic == chk->floorpic
							&& os[0]->ceilingpic == chk->ceilingpic)
						{
							osc[0]++;
						}
						else
						{
							if (!os[1])
							{
								os[1] = chk; osc[1]++;
							}
							else if (os[1] == chk)
							{
								osc[1]++;
							}
							else
							{
								osc[0] = osc[1] = 0; break;
							}
						}
					}
				}

				if (os[0] && osc[0] > osc[1] && osc[1] < 2) ns = os[0];
				else
				if (os[1] && osc[1] > osc[0] && osc[0] < 2) ns = os[1];

				if (ns && !(ns->specFlags & SF_MOVFLOOR))
				{
					int		flag = MSIA_NONE;
					int		osec;

					memcpy(mrkline, hitline, numlines); setarray = true;

					for (c = i = 0; i < sc->linecount; i++)
					{
						int n = sc->lines[i] - lines;

						if (hitline[n] == 4) linearray[c++] = n;
					}

					if (LS_MarkSubsecsInLoop(sc, rc, NULL, box, MSIA_TESTAREA, NULL))
					{
						osec = ns - sectors;
					}
					else
					{
						flag = MSIA_REVERSED;
						osec = sc - sectors;
					}

					if (LS_MarkSubsecsInLoop(sc, c, NULL, box, flag, NULL))
					{
						if (noHiTex(sc, mrkline, flag))
						{
							validhi = true; f |= 1;
						}

						if (noLoTex(sc, mrkline, flag))
						{
							validlo = true; f |= 2;
						}
					}
				
					if ( (validlo || validhi) && LS_OnlySSLinesInSSECarray(NULL, mrkline))
					{
						for (c = i = 0; i < hitssecptr; i++)
						{
							if (validlo && subsectors[hitssec[i]].sector->floorpic != skyflatnum)
							{
								if (!secFloorSame(subsectors[hitssec[i]].sector, &sectors[osec]))	// v220-33 test added
									Setr3dsslfsec(&subsectors[hitssec[i]], osec, 48071);
							}
							if (validhi && subsectors[hitssec[i]].sector->ceilingpic != skyflatnum)
							{
								if (!secCeilSame(subsectors[hitssec[i]].sector, &sectors[osec]))	// v220-33 test added
									Setr3dsslcsec(&subsectors[hitssec[i]], osec, 48072);
							}
						}
					}
				}
			}

			if (setarray) memcpy(hitline, mrkline, numlines);

			for (i = 0; i < sc->linecount; i++)
			{
				n = sc->lines[i] - lines;

				if (hitline[n] == 4)
				{
					hitline[n] = 8;

					if (   (validhi || validlo)
						&& !(sc->lines[i]->frontsector->specFlags & SF_MOVSECTOR
						&&   sc->lines[i]->backsector->specFlags  & SF_MOVSECTOR) )
					{
						SetLFG(sc->lines[i], LFG_MARKED, 48073);

						if (   !origTEX[sc->lines[i] - lines]
							&& (sc->lines[i]->mtflags == R3D_MISLTEX1 || sc->lines[i]->mtflags == R3D_MISLTEX2))
						{
							chk = sc->lines[i]->mtflags == R3D_MISLTEX1 ?
								sc->lines[i]->backsector : sc->lines[i]->frontsector;

							if (chk && (chk->specFlags & 0x3000C) == 0x1000C)
								Settype(sc->lines[i], R3D_14, 48074);
						}
					}
				}
			}
		}
	}

	M_Free(mrkline);
}

// fixed sector enclosed by moving floor sector where the fixed sector
// starts higher than the surrounding floor sector and has low textures
static RDBOOL checkInfill1(int idx)
{
	sector_t *ch, *os = NULL, *sc = &sectors[idx];
	line_t	 *ld;
	int		 a, i, tx;

	for (a = -1, i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		if (lineflags[ld - lines]) return false;
		if (ld->specType != R3D_00) return false;
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if (!ch) return false;
		tx = origTEX[ld - lines];
		if (a < 0)
		{
			if (ch == ld->frontsector && tx == 0x01) a = 0;
			else
			if (ch == ld->backsector  && tx == 0x10) a = 0;
			else
			if (ch == ld->frontsector && tx == 0x10) a = 1;
			else
			if (ch == ld->backsector  && tx == 0x01) a = 1;

			if (a < 0) return false;
		}
		else
		{
			if (a == 0)
			{
				if (ch == ld->frontsector && tx != 0x01) return false;
				if (ch == ld->backsector  && tx != 0x10) return false;
			}
			else
			{
				if (ch == ld->frontsector && tx != 0x10) return false;
				if (ch == ld->backsector  && tx != 0x01) return false;
			}
		}

		if (!(ch->specFlags & SF_MOVFLOOR)) return false;
		if (!os) os = ch; else if (os != ch) return false;
	}

	if (!os || a < 0) return false;
	if (!M_CompareFPBox(sc->sibounds, os->sibounds)) return false;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		Settype(ld, R3D_14, 48075);
		SetLFG(ld, LFG_NEWSET, 48076); 
	}

	Setsecdone(sc, 48077);
	Setsecdone(os, 48078);

	return true;
}

// check static sector containing a moving floor sector
static RDBOOL checkInfill2(int idx)
{
	sector_t *ch, *os = NULL, *sc = &sectors[idx];
	line_t	 *ld;
	int		 c, i, tx;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		if (lineflags[ld - lines]) return false;
		if (ld->specType != R3D_00) return false;
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if (!ch) continue;

		tx = origTEX[ld - lines] & 0x11;

		if (ch->specFlags & SF_MOVFLOOR)
		{
			if (sc == ld->frontsector && tx != 0x01) return false;
			if (sc == ld->backsector  && tx != 0x10) return false;
			if (!os) os = ch; else if (os != ch) return false;
			linearray[c++] = ld - lines;
		}
		else
		{
			if (sc == ld->frontsector && tx != 0x10) return false;
			if (sc == ld->backsector  && tx != 0x01) return false;
		}
	}

	if (!os || c < 3) return false;
	if (!M_CompareFPBox(os->sibounds, sc->sibounds)) return false;
	if (!LS_SsecAllSegsFrontSame(sc)) return false;

	for (i = 0; i < os->linecount; i++)
	{
		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			ld = os->lines[i];
			Settype(ld, R3D_14, 48079);
			SetLFG(ld, LFG_NEWSET, 48080);
		}
	}

	Setsecdone(sc, 48081);
	Setsecdone(os, 48082);

	return true;
}

static void subfn_checkInfill3(sector_t *s1, sector_t *s2)
{
	line_t	*ld;
	int		c, i;

	CLR_hitline();

	for (c = i = 0; i < s1->linecount; i++)
	{
		ld = s1->lines[i];

		if (   ld->spflags & SPF_CONTROLLINE
			|| ld->specType != R3D_00
			|| origTEX[ld - lines] & 0x33)
		{
			continue;
		}

		if (   (ld->frontsector == s1 && ld->backsector == s2)
			|| (ld->frontsector == s2 && ld->backsector == s1) )
		{
			hitline[ld - lines] = 1; c++;
		}
	}

	if (c > 2)
	{
		RDBOOL repeat = true;

		while (repeat)
		{
			RDBOOL	valid;
			int		lnum, rc, box[4];

			for (i = 0, repeat = false; i < s1->linecount; i++)
			{
				int n = s1->lines[i] - lines;

				if (hitline[n] == 1) { lnum = n; repeat = true; break; }
			}

			if (!repeat) break;

			if (!subfn_CheckForRing(hitline, s1, lnum, &rc, box, NULL, NULL))
			{
				hitline[lnum] = 0; continue;
			}

			valid = rc > 2
				&& LS_MarkSubsecsInAreaGeneral(s1, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(NULL, hitline);

			for (i = 0; i < s1->linecount; i++)
			{
				int n = (ld = s1->lines[i]) - lines;

				if (hitline[n] == 4)
				{
					hitline[n] = 8;

					if (valid)
					{
						Settype(ld, R3D_14, 48083);
						SetLFG(ld, LFG_MARKED, 48084);
					}
				}
			}
		}
	}
}

static void checkInfill3()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	byte		*marksec = (byte*)calloc(numsectors, sizeof(byte));
	int			*sectidx = (int*) malloc(numsectors*sizeof(int));
	int			c, i;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !lineflags[i]
			&& ld->ml_flags & ML_TWOSIDED
			&& ld->specType == R3D_00
			&& !(origTEX[i] & 0x33)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->specFlags & SF_MOVFLOOR
			&& bs->specFlags & SF_MOVFLOOR)
		{
			int s1 = fs - sectors;
			int s2 = bs - sectors;
			int nf = fs->specFlags & SF_NOLOTEX;
			int nb = bs->specFlags & SF_NOLOTEX;
			int store = !secdone[s1] && !secdone[s2];

			if (!marksec[s1] && !marksec[s2])
			{
				marksec[s1] = true;
				marksec[s2] = true;

				if (store && (nf^nb))
				{
					if (nf)
						sectidx[c++] = s1, sectidx[c++] = s2;
					else
						sectidx[c++] = s2, sectidx[c++] = s1;
				}
			}
		}
	}

	for (i = 0; i < c;)
	{
		// set sectors outside of parameter list else
		// correct compilation cannot be guaranteed
		fs = &sectors[sectidx[i++]];
		bs = &sectors[sectidx[i++]];

		subfn_checkInfill3(fs, bs);
	}

	M_Free(marksec);
	M_Free(sectidx);
}

// This has been made quite specific for constructions where the lines
// for a sector with an invalid move bound other same tagged sectors.
static void checkNestedInfill(sector_t *sec)
{
	sector_t	*chk;
	line_t		*ld;
	int			c, i, n, tx;

	CLR_hitline();

	for (c = i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->spflags & SPF_CONTROLLINE) continue;

		chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;
		if (!chk) continue;

		n = ld - lines;

		if (lineflags[n]) continue;
		if ( !((tx = origTEX[n]) & 0x11) ) continue;
		if (chk->specFlags & SF_MOVFLOOR) continue;
		if (chk == ld->frontsector && tx & 0x10) continue;
		if (chk == ld->backsector  && tx & 0x01) continue;

		hitline[n] = 1; c++;
	}

	if (c < 3) return;

	while(1)
	{
		sector_t	*osec = NULL;
		int			k, lnum = -1, rc, box[4];
		//-------------------

		for (i = 0; i < sec->linecount; i++)
		{
			ld = sec->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			int	valid = 1;

			for (i = 0; i < sec->linecount; i++)
			{
				ld = sec->lines[i];
				chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (hitline[ld - lines] == 4)
				{
					if (!osec) osec = chk; else if (osec != chk) valid = 0;
				}
			}

			if (   valid
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE) )
			{
				for (c = k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];

					if (  (ss->sector->specFlags & (SF_MOVFLOOR|SF_MOVINVALID)) == (SF_MOVFLOOR|SF_MOVINVALID)
						&& ss->sector->floorheight >= sec->floorheight
						&& ss->sector->tag == sec->tag)
					{
						c++;
					}
					else
						break;
				}

				if (c == hitssecptr)	// all subsecs must comply
				{
					valid = 2;

					for (c = k = 0; k < hitssecptr; k++)
						Setr3dsslfsec(&subsectors[hitssec[k]], osec - sectors, 48085);
				}
			}

			for (i = 0; i < sec->linecount; i++)
			{
				ld = sec->lines[i];

				if (hitline[ld - lines] == 4)
				{
					if (valid == 2) SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48086);
					hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

//=======================================================

// v220-13 sub function of MX_CheckNoTexMovRings()
static RDBOOL checkInfillx(int idx)
{
	sector_t	*chk, *sc = &sectors[idx];
	line_t		*ld;
	RDBOOL		repeat = true;
	RDBOOL		valid;
	int			c, d, i, n, box[4];

	CLR_hitline();

	for (c = d = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (   ld->ml_flags & ML_TWOSIDED
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& ld->frontsector != ld->backsector)
		{
			n = ld - lines;

			if (   !(origTEX[n] & 0x22)
				&& !lxarray[n]
				&& !(lineflags[n] & ~LFG_NHCONNOBSLN))
			{
				chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (chk->specFlags & SF_MOVSECTOR) continue;
				if (chk->ceilingpic == skyflatnum) continue;

				if (sc->ceilingheight - chk->ceilingheight > 8*FRACUNIT)
				{
					hitline[n] = 1; c++;
				}
				else
					d++;
			}
		}
	}

	if (c < 3 || d < 3) return false;

	while (repeat)
	{
		sector_t	*osec;
		int			lc, rc = 0;
		int			lnum = -1;

		//------------------------------------------

		repeat = false;
		valid = true;

		for (i = 0; i < sc->linecount; i++)
		{
			n = sc->lines[i] - lines;

			if (hitline[n] == 1)
			{
				lnum = n; repeat = true; break;
			}
		}

		if (!repeat) break;

		if (!subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
		{
			hitline[lnum] = 0; continue;
		}

		for (lc = i = 0, osec = NULL; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			n = ld - lines;

			if (hitline[n] == 4)
			{
				chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!osec) osec = chk;
				else if (osec->floorheight != chk->floorheight
					|| osec->ceilingheight != chk->ceilingheight)
				{
					valid = false; break;
				}

				linearray[lc++] = n;
			}
		}

		if (valid)
		{
			if ( (valid = LS_MarkSubsecsInLoop(sc, lc, NULL, box, MSIA_NONE, NULL)) )
			{
				valid = LS_NoHiTexInSSECarray();
			}

			if (valid)
			{
				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];
					seg_t		*sg = &segs[ss->firstline];

					Setr3dsslcsec(ss, osec - sectors, 48087);

					for (n = 0; n < ss->linecount; n++, sg++)
					{
						if (sg->linedef && !(lineflags[sg->linedef - lines] & LFG_NEWSET) )
							SetLFG(sg->linedef, LFG_NEWSET|LFG_FIXED, 48088);
					}
				}

				Setsecdone(sc, 48089);
			}
		}

		for (i = 0; i < sc->linecount; i++)
		{
			int n = sc->lines[i] - lines;

			if (hitline[n] == 4) hitline[n] = 0;
		}
	}

	return secdone[idx] & 1;
}

// v220-18
static RDBOOL secMovesHigher(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*ld;
	RDBOOL		valid = false;
	int			testht = sc->ceilingheight + 8*FRACUNIT;
	int			lnum, rc;
	int			c, i;

	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (ld->spflags & SPF_CONTROLLINE)
		{
			if (ch)
			{
				if (ch->ceilingheight <= testht) return false;
				valid = true;
			}
			continue;
		}

		if (!ch) return false;
		if (ch == sc) continue;
		if (!(ld->lnflags2 & LNF2_OWNEQUALTOTWO)) return false;
		if (!(origTEX[ld - lines] & 0x11) && ch->floorheight != sc->floorheight) return false;
		if (!os)
		{
			if (ch->ceilingheight < sc->ceilingheight) return false;
			if (ch->specFlags & 0x3000C && !(ch->specFlags & 0x400)) return false;
			os = ch;
		}
		else if (os != ch)
		{
			if (ch->specFlags & 0x3000C && !(ch->specFlags & 0x400)) return false;
			if (   ch->floorheight != os->floorheight
				|| ch->ceilingheight != os->ceilingheight)
			{
				return false;
			}
		}

		linearray[c++] = ld - lines;
	}

	if (!os || c < 3) return false;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while (1)
	{
		lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1) { lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (!subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) || rc < 3) return false;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 4) hitline[linearray[i]] = 0;
		}
	}

	ss_validcount++;
	hitssecptr = 0;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	// check the ssec references
	for (i = 0; i < c; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];
		int		 k, m;

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg;

			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;
			if (ss->sector != sc) continue;
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->frontsector == sg->backsector) continue;
				if (!(ld = sg->linedef)) continue;
				if (sg->frontsector != sc) return false;
				if (sg->backsector && sg->backsector != os) return false;
			}

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	for (i = 0; i < hitssecptr; i++) Setr3dsslcsec(&subsectors[hitssec[i]], os - sectors, 48090);

	for (i = 0; i < c; i++)
	{
		if (!origTEX[linearray[i]]) Settype(&lines[linearray[i]], R3D_3F, 48091);
		SetLFG(&lines[linearray[i]], LFG_FIXABS|LFG_NEWSET|LFG_NOCHANGE, 48092);
	}

	Setsecdone(sc, 48093);
	Setsecdone(os, 48094);

	return true;
}

// v220-9 check for continuous line rings that need an infill flat
void MX_CheckNoTexMovRings()
{
	int		i;

	if (R3D_ScriptedMap) return;	// Scripted maps use different methods

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].stptr >= 0) continue;

		if (sectors[i].specFlags & SF_MOVSECTOR)
		{
			if (  (sectors[i].specFlags & 0xFBFFC) == 0x98198
				&& sectors[i].ceilingpic != skyflatnum
				&& checkInfillx(i))
			{
				continue;
			}

			if (sectors[i].tag && !(sectors[i].specFlags & SF_MOVINVALID))
				checkInfill0(i);
		}

		// Test added v220-18
		if (sectors[i].tag && SecCtrl(i))
		{
			if ( (sectors[i].specFlags & (SF_MOVFLOOR|SF_MOVINVALID)) == (SF_MOVFLOOR|SF_MOVINVALID))
			{
				checkNestedInfill(&sectors[i]);
			}
			else
			if (  (sectors[i].specFlags & 0x30018) == 0x20018
				&& secMovesHigher(&sectors[i]))
			{
				// do nothing
			}
		}
	}

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].stptr >= 0) continue;

		if (  !(sectors[i].specFlags & SF_MOVSECTOR)
			&& !SecCtrl(i)
			&& sectors[i].specFlags & SF_SECCLOSED)
		{
			if (!checkInfill1(i) && sectors[i].linecount > 5)
				checkInfill2(i);
		}
	}

	// check for double movfloor sec line rings one having
	// low tex the other without;
	checkInfill3();
}

//===================================================================

static sector_t *getInternalLines(sector_t *s, int *cl, int *cr, int *lnum)
{
	sector_t	*osec = NULL;
	line_t		*ld;
	int			c, d, i, m;

	validcount++;

	for (c = d = i = 0; i < hitssecptr; i++)
	{
		subsector_t	*ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (m = 0; m < ss->linecount; m++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;
				linearray[c++] = ld - lines;

				if (   ld->frontsector == ld->backsector
					&& ld->frontsector != s)
				{
					if (!osec) osec = ld->frontsector;
					else if (osec != ld->frontsector) return NULL;

					hitline[ld - lines] = 1;
					*lnum = ld - lines; d++;
				}
			}
		}
	}

	*cl = c;
	*cr = d;
	return osec;
}

// Look for loops of lines with no upper/lower textures with a static sector
// which enclose a ring of same sided lines with a mov sector.

void MX_MissingTexLoopSSLines()
{
	line_t	*ld;
	byte	*mrkline;
	int		c, i, k;
	int		cl, cr, rc, box[4];
	//----------------------

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !lineflags[i]
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& ld->frontsector != ld->backsector
			&& !(origTEX[i] & 0x55))
		{
			int f1 = ld->frontsector->specFlags & SF_MOVSECTOR;
			int f2 = ld->backsector->specFlags  & SF_MOVSECTOR;

			if ((!f1 && f2) || (f1 && !f2))
			{
				hitline[i] = 1; c++;
			}
		}
	}

	if (c < 3) return;

	mrkline  = (byte*)malloc(numlines);

	while(1)
	{
		sector_t	*s, *osec;
		int			lnum = -1;
		//-------------------

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (hitline[i] == 1)
			{
				s = lines[i].backsector && lines[i].frontsector->specFlags & SF_MOVSECTOR ?
													lines[i].backsector : lines[i].frontsector;
				lnum = i;
				break;
			}
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			memcpy(mrkline, hitline, numlines);

			if (LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_REVERSED))
			{
				CLR_hitline();

				if (  (osec = getInternalLines(s, &cl, &cr, &lnum)) )
				{
					if (   cr > 2
						&& subfn_CheckForRing(hitline, osec, lnum, &rc, box, NULL, NULL)
						&& rc == cr)
					{
						int		hitex = 0, lotex = 0;

						// Hi is only true if 2 or more lines are counted
						// and if no lines share a vertex
						for (k = 0; k < cl; k++)
						{
							ld = &lines[linearray[k]];

							if (!Vtxs2(ld)) { hitex = 0; break; }
								
							if (origTEX[linearray[k]] & 0x44) hitex++;
							if (origTEX[linearray[k]] & 0x11) lotex++;
						}

						for (k = 0; k < hitssecptr; k++)
						{
							if (hitex < 2)
								Setr3dsscs(&subsectors[hitssec[k]], s - sectors, 48095);
							if (lotex < 2)
								Setr3dssfs(&subsectors[hitssec[k]], s - sectors, 48096);
						}

						for (k = 0; k < cl; k++)
						{
							ld = &lines[linearray[k]];

							if (!ld->backsector || origTEX[linearray[k]] & 0x55)
								Settype(ld, R3D_00, 48097);
							else if (origTEX[linearray[k]] & 0x22)
								Settype(ld, R3D_03, 48098);
							else if (hitex < 2 && lotex < 2)
								Settype(ld, R3D_3F, 48099);
							else if (hitex < 2)
								Settype(ld, R3D_16, 48100);
							else
								Settype(ld, R3D_14, 48101);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48102);
						}
					}
				}
			}

			memcpy(hitline, mrkline, numlines);

			for (k = 0; k < s->linecount; k++)
			{
				lnum = s->lines[k] - lines;

				if (hitline[lnum] == 4) hitline[lnum] = 0;
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	M_Free(mrkline);
}

// Look for R3D_1A lines that form loops in sectors with sky ceilings (where they are usually found)
// that access different infill sectors. These should all have the same characteristics.
void MX_UnsetLines()
{
	line_t	*ld;
	int		rc, box[4];
	int		c, i, k, m, n;

	for (i = 0; i < numsectors; i++)
	{
		if (   sectors[i].ceilingpic == skyflatnum
			&& sectors[i].floorpic != skyflatnum
			&& sectors[i].specFlags & SF_NOLOTEX
			&& sectors[i].stptr < 0)
		{
			sector_t	*sec = &sectors[i], *chk;
			RDBOOL		revmove = (sectors[i].specFlags & 0x13FFFC) == 0x11889C;

			CLR_hitline();

			for (c = k = 0; k < sectors[i].linecount; k++)
			{
				ld = sec->lines[k];
				n = ld - lines;

				if (   !lineflags[n]
					&& !origTEX[n]
					&& !lxarray[i]
					&& !(ld->spflags & SPF_CONTROLLINE)
					&& ld->backsector)
				{
					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (   !(chk->specFlags & SF_MOVFLOOR)
						&& (abs(chk->floorheight - sec->floorheight) > 8*FRACUNIT
						|| (revmove && chk->floorheight == sec->floorheight)) )
					{
						hitline[n] = 1; c++;
					}
				}
			}

			if (c < 3) continue;

			while(1)
			{
				int lnum = -1;

				for (k = 0; k < sectors[i].linecount; k++)
				{
					ld = sectors[i].lines[k];

					if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
				{
					// check the adopting sector
					sector_t	*osec = NULL;

					for (k = 0; k < sectors[i].linecount; k++)
					{
						ld = sectors[i].lines[k];

						if (hitline[ld - lines] == 4)
						{
							chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

							if (!osec)
								osec = chk;
							else if (osec != chk)
							{
								if (  osec->floorheight != chk->floorheight
									|| osec->floorpic   != chk->floorpic
									|| osec->lightlevel != chk->lightlevel)
								{
									osec = NULL; break;
								}
							}
						}
					}

					if (   osec
						&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& LS_OnlySSLinesInSSECarray(sec, hitline))
					{
						validcount++;

						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];
							seg_t		*sg = &segs[ss->firstline];

							Reset_SSEC(hitssec[k], 48103);
							Setr3dsslfsec(ss, osec - sectors, 48104);

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( (ld = sg->linedef) && ld->validcount != validcount)
								{
									ld->validcount = validcount;
									SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48105);
									Clr_allmtflags(ld, 48106);

									if (origTEX[ld - lines] & 0x22)
										Forcetype(ld, R3D_03, 48107);
									else
										Forcetype(ld, R3D_3F, 48108);
								}
							}
						}
					}

					for (k = 0; k < sectors[i].linecount; k++)
					{
						n = sectors[i].lines[k] - lines;

						if (hitline[n] == 4) hitline[n] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}
}

// Find bounded areas including 1's lines in mov sectors and then check if
// any lines inside the area, with different sectors, have hi or lo textures.
// If not then set all subsectors in the area to the containing ring's sector.
// This is done at the end of all other analysis to prevent it being called
// unnecessarily as it is a fairly expensive function and is rarely needed.
void MX_Contained()
{
	sector_t	*sec, *chk;
	line_t		*ld;
	int			c, i, k, tx;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->stptr < 0
			&& sec->specFlags & SF_MOVSECTOR
			&& sec->specFlags & SF_SSLINETRUE)
		{
			CLR_hitline();

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (  !lineflags[i]
					&& ld->frontsector != ld->backsector
					&& !(ld->spflags & SPF_CONTROLLINE))
				{
					if ( !(ld->ml_flags & ML_TWOSIDED) )
					{
						hitline[ld - lines] = 1; c++;
					}
					else if ( (tx = origTEX[ld - lines]) & 0x55)
					{
						if (ld->frontsector == sec)
						{
							chk = ld->backsector;

							if (chk->floorheight > sec->floorheight && !(tx & 0x01)) continue;
							if (chk->floorheight < sec->floorheight && !(tx & 0x10)) continue;

							if (chk->ceilingheight > sec->ceilingheight && !(tx & 0x40)) continue;
							if (chk->ceilingheight < sec->ceilingheight && !(tx & 0x04)) continue;
						}
						else
						{
							chk = ld->frontsector;

							if (chk->floorheight < sec->floorheight && !(tx & 0x01)) continue;
							if (chk->floorheight > sec->floorheight && !(tx & 0x10)) continue;

							if (chk->ceilingheight < sec->ceilingheight && !(tx & 0x40)) continue;
							if (chk->ceilingheight > sec->ceilingheight && !(tx & 0x04)) continue;
						}
						
						hitline[ld - lines] = 1, c++;
					}
				}
			}
		}

		if (c < 3) continue;

		while (1)
		{
			int rc, box[4], lnum = -1;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if ( !(ld->ml_flags & ML_TWOSIDED)
					&& hitline[ld - lines] == 1)
				{
					lnum = ld - lines; break;
				}
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
			{
				int d, chksec[1];

				chksec[0] = i;

				for (c = d = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						if (ld->ml_flags & ML_TWOSIDED) d++;
						linearray[c++] = ld - lines;
					}
				}
			
				if (   d	// eliminate loop of all 1's lines
					&& LS_MarkSubsecsInMultiSecLoop(1, chksec, rc, box, NULL, false, MSIA_NONE))
				{
					setLinesContained(sec, i, 2);
				}

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						hitline[ld - lines] = 0;
					}
				}
			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}
}

static void subfn_containedBySSLineRing(line_t *li)
{
	ssline_t	*sl = &sslines[li - lines];
	sector_t	*fs = li->frontsector;
	sector_t	*bs = li->backsector;
	sector_t	*os = NULL;
	line_t		*ld;
	int			c, f = 0, i, k, rc, box[4];

	if (!bs || fs == bs) return;

	validcount++;

	for (c = i = 0; i < sl->num; i++)
	{
		subsector_t	*ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) )
			{
				if (ld == li)
				{
					// Ensure lines form a ring with li inside
					if (sg->frontsector == li->frontsector && sg->backsector == li->backsector ) f |= 1;
					if (sg->frontsector == li->backsector  && sg->backsector == li->frontsector) f |= 2;
				}
				else if (ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (ld->frontsector == ld->backsector)
					{
						if (!os) os = ld->frontsector; else if (os != ld->frontsector) return;

						linearray[c++] = ld - lines;
					}
					else
						return;
				}
			}
		}
	}

	if (c < 3 || f != 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++)
		hitline[linearray[i]] = 1;

	if (subfn_CheckForRing(hitline, os, linearray[0], &rc, box, NULL, NULL) && rc == c)
	{
		Setspec(li, os - sectors, 48109);
	}
}

// --------------------------------------------------

static RDBOOL setSubsecs(sector_t *s, int count)
{
	int i, k, m;

	hitssecptr = 0;
	ss_validcount++;

	for (i = 0; i < count; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			if (ss->sector == s)
			{
				seg_t *sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->frontsector && sg->frontsector != s) return false;
				}
			}

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	return hitssecptr > 0;
}

// v220-10 check unset mov floor sectors
void MX_CheckMovFloor()
{
	line_t	*ld;
	int		c, i, k, rc;

	CLR_conarray();

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& SecCtrl(i)
			&& (sectors[i].specFlags & 0x11B79C) == 0x11801C)
		{
			sector_t	*chk, *s = &sectors[i];

			for (c = k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   ld->lnflags2 == 0x54000000
					&& ld->ml_flags & ML_TWOSIDED	// insurance
					&& ld->frontsector != ld->backsector
					&& !lineflags[ld - lines])
				{
					// test added v220-13
					chk = ld->frontsector != s ? ld->frontsector : ld->backsector;
					if (chk->specFlags & SF_MOVFLOOR) continue;

					linearray[c++] = ld - lines;
				}
				else
				{
					c = 0; break;
				}
			}

			if (c < 3) continue;

			CLR_conarray();
			CLR_hitline();

			for (k = 0; k < c; k++)
			{
				ld = &lines[linearray[k]];
				chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

				// check for possibly shared line
				if ( (chk->specFlags & 0x40C) == 0xC) conarray[linearray[k]] = 1;

				hitline[linearray[k]] = 1;
			}

			while(1)
			{
				int n, lnum = -1;
				
				for (k = 0; k < s->linecount; k++)
				{
					n = s->lines[k] - lines;

					if (hitline[n] == 1 && !conarray[n]) { lnum = n; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, s, lnum, &rc, NULL, NULL, NULL))
				{
					sector_t	*os = NULL;

					for (k = 0; k < s->linecount; k++)
					{
						n = s->lines[k] - lines;

						if (hitline[n] == 4 && !conarray[n])
						{
							if (!os) os = chk;
							else if (os != chk)
							{
								if (   os->floorheight != chk->floorheight
									|| os->ceilingheight != chk->ceilingheight)
								{
									os = NULL; break;
								}
							}
						}
					}

					if (os && setSubsecs(s, c))
					{
						for (k = 0; k < hitssecptr; k++)
							SetOverride(&subsectors[hitssec[k]], os - sectors, 48066);

						Setsecdone(s, 48110);
					}
					else
					{
						os = NULL;
					}

					for (k = 0; k < s->linecount; k++)
					{
						n = s->lines[k] - lines;

						if (hitline[n] == 4)
						{
							if (conarray[n])
								hitline[n] = 1;	// reset (possibly) shared line
							else
							{
								hitline[n] = 0;

								if (os)
								{
									if (!origTEX[n]) Settype(s->lines[k], R3D_3F, 48111);
									SetLFG(s->lines[k], LFG_NEWSET|LFG_FIXED, 48112);
								}
							}
						}
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}
}

// Search for lines with missing hi and lo textures, disconnected
// from other missing texture lines, and check if any found are
// enclosed in a same sector sided line ring.
void MX_ContainedBySSLineRing()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector
			&& !origTEX[i]
			&& !(lineflags[i] & ~LFG_FIXED)
			&& !lxarray[i]
			&& (ld->mtflags == 0x30 || ld->mtflags == 0xC0)
			&& vtxflags[ld->v1 - vertexes] == 1
			&& vtxflags[ld->v2 - vertexes] == 1
			&& vertexowners[ld->v1 - vertexes].num > 2
			&& vertexowners[ld->v2 - vertexes].num > 2)
		{
			subfn_containedBySSLineRing(ld);
		}
	}
}

// Check for rings of same sided lines inside a ring of missing tex lines
void MX_CheckLineRings()
{
	sector_t	*sec;
	line_t		*ld;
	byte		*la;
	int			c, i;

	SET_countsec(0);
	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& ld->frontsector->specFlags & SF_MOVSECTOR
			&& !lineflags[i]
			&& !lxarray[i]
			&& !origTEX[i])
		{
			hitline[i] = 1;
			countsec[ld->frontsector - sectors]++;
			if (countsec[ld->frontsector - sectors] > 2) c++;
		}
	}

	if (!c) return;

	la = (byte*)malloc(numlines);

	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i] < 3) continue;

		sec = &sectors[i];

		while(1)
		{
			int lnum = -1, k, m, n, rc, box[4], labox[4];

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] == 1)
				{
					lnum = ld - lines; break;
				}
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
			{
				sector_t *chk, *os = NULL;
				RDBOOL	 valid = false;
				int		 lanum = -1;

				for (c = k = 0; k < sec->linecount; k++)
				{
					int num = sec->lines[k] - lines;

					if (hitline[num] == 4)
						linearray[c++] = num;
				}

				memset(la, 0, numlines);
				lanum = -1;
				ss_validcount++;
				validcount++;

				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount != ss_validcount)
						{
							seg_t *sg = &segs[ss->firstline];

							ss->ssvalidcount = ss_validcount;

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (  (ld = sg->linedef)
									&& ld->validcount != validcount)
								{
									ld->validcount = validcount;

									if (hitline[ld - lines] == 4) continue;

									if (   ld->frontsector == ld->backsector
										|| !(ld->ml_flags & ML_TWOSIDED)
										|| sg->frontsector == sec
										|| sg->backsector == sec
										|| !sg->backsector
										|| sg->sidedef->bottomtexture
										|| sg->sidedef->midtexture
										|| sg->sidedef->toptexture)
									{
										os = NULL; goto end_check;
									}

									chk = NULL;

									if (   ld->frontsector->floorheight   < ld->backsector->floorheight
										&& ld->frontsector->ceilingheight > ld->backsector->ceilingheight)
									{
										chk = ld->backsector;
									}
									else
									if (   ld->frontsector->floorheight   > ld->backsector->floorheight
										&& ld->frontsector->ceilingheight < ld->backsector->ceilingheight)
									{
										chk = ld->frontsector;
									}

									if (!chk) { os = NULL; goto end_check; }

									lanum = ld - lines;
									la[lanum] = 1;

									if (!os)
										os = chk;
									else if (os != chk)
									{
										os = NULL; goto end_check;
									}
								}
							}
						}
					}
				}

				if (  !osInvalid(os)
					&& subfn_CheckForRing(la, os, lanum, &rc, labox, NULL, NULL)
					&& rc > 2
					&& M_FitBox(box, labox))
				{
					valid = true;

					ss_validcount++;

					for (k = 0; k < c; k++)
					{
						ssline_t *sl = &sslines[linearray[k]];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->ssvalidcount != ss_validcount)
							{
								ss->ssvalidcount = ss_validcount;
								Setr3dssfs(ss, os - sectors, 48113);
								Setr3dsscs(ss, os - sectors, 48114);
							}
						}
					}
				}

end_check:;
				for (k = 0; k < sec->linecount; k++)
				{
					lnum = sec->lines[k] - lines;

					if (hitline[lnum] == 4)
					{
						hitline[lnum] = 0;
						if (valid) SetLFG(sec->lines[k], LFG_NEWSET, 48115);
					}
				}
			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}

	M_Free(la);
}

// Check lines set with ML_IGNOREMISSINGLO to see if they are in
// a missing tex line ring with 2 mov ceil sectors (rare).
void MX_Unset()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			c, i, k, rc, box[4];

	for (i = 0; i < numlines; i++)
	{
		if (   !lineflags[i]
			&& !lxarray[i]
			&& lines[i].ml_flags & ML_IGNOREMISSINGLO
			&& ((fs = lines[i].frontsector)->specFlags & SF_MOVSECTOR) == SF_MOVCEIL
			&& (bs = lines[i].backsector)
			&& (bs->specFlags  & SF_MOVSECTOR) == SF_MOVCEIL)
		{
			sector_t *sec = fs->floorheight > bs->floorheight ? bs : fs;
			sector_t *chk = fs == sec ? bs : fs;

			CLR_hitline();

			hitline[i] = 1;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   !lineflags[ld - lines]
					&& !origTEX[ld - lines]
					&& (ld->frontsector == chk || ld->backsector == chk))
				{
					hitline[ld - lines] = 1; c++;
				}
			}

			if (   c > 1
				&& subfn_CheckForRing(hitline, sec, i, &rc, box, NULL, NULL)
				&& rc > 2
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(sec, hitline))
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					Setr3dsslfsec(ss, chk - sectors, 48116);
					Setr3dsslcsec(ss, chk - sectors, 48117);
				}

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						if (!(lineflags[ld - lines] & LFG_OVERRIDE))
							ld->ml_flags &= ~ML_IGNOREMISSINGLO;
						SetLFG(ld, LFG_NEWSET, 48118);
					}
				}
			}
		}
	}
}

static void subfn_Postlim(sector_t *sec)
{
	sector_t *os = NULL;
	sector_t *chk;
	line_t	 *ld;
	RDBOOL	 set_hi, set_lo;
	int		 diff, lnum, rc, box[4];
	int		 i, n;

	//----------------------

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->frontsector == ld->backsector) continue;

		lnum = ld - lines;

		if (   lineflags[lnum]
			|| origTEX[lnum]
			|| lxarray[lnum]
			|| !Vtfl(ld))
		{
			return;
		}

		chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (  !chk
			|| chk->specFlags & SF_MOVSECTOR
			|| ((chk->specFlags & 0xC) == 0xC && !(chk->specFlags & 0x400)) )
		{
			return;
		}

		if (!os)
			os = chk;
		else if (os != chk
			&& (os->floorheight != chk->floorheight
			|| os->ceilingheight != chk->ceilingheight))
		{
			return;
		}
	}

	if (osInvalid(os)) return;

	if (   sec->ceilingheight == os->ceilingheight
		&& sec->floorheight == os->floorheight)
	{
		return;
	}

	set_hi = (diff = sec->ceilingheight - os->ceilingheight) && !(diff > 0 && diff <= 8*FRACUNIT);
	set_lo = (diff = os->floorheight    - sec->floorheight ) && !(diff > 0 && diff <= 8*FRACUNIT);

	if (!set_hi && !set_lo) return;

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->frontsector != ld->backsector)  hitline[ld - lines] = 1;
	}

	while(1)
	{
		lnum = -1;

		for (i = 0; i < sec->linecount; i++)
		{
			int n = sec->lines[i] - lines;

			if (hitline[n] == 1) { lnum = n; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			RDBOOL	valid = false;

			if (   LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(sec, hitline))
			{
				valid = true;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (set_hi) Setr3dsscs(ss, os - sectors, 48119);
					if (set_lo) Setr3dssfs(ss, os - sectors, 48120);
				}
			}

			for (i = 0; i < sec->linecount; i++)
			{
				ld = sec->lines[i];
				n = sec->lines[i] - lines;

				if (hitline[n] == 4)
				{
					hitline[n] = 0;

					if (valid)
					{
						byte type = R3D_36;

						if (ld->ml_flags & ML_TWOSIDED)
						{
							if (set_hi && set_lo)
							{
								type = R3D_3F;
							}
							else if (!set_hi && set_lo)
							{
								if (ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
									type = R3D_3F;
								else
									ld->ml_flags |= ML_IGNOREMISSINGLO;
							}
							else if (set_hi && !set_lo)
							{
								if (ld->frontsector->floorheight == ld->backsector->floorheight)
									type = R3D_3F;
								else
									ld->ml_flags |= ML_IGNOREMISSINGHI;
							}
						}

						Settype(ld, type, 48121);
						SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48122);
					}
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

static void addTo_smk(int idx, int *smk, int *cs)
{
	int i;

	for (i = 0; i < *cs; i++)
	{
		if (smk[i] == idx) return;
	}

	smk[*cs] = idx; (*cs)++;
}

static void setSSECval(int *smk, byte *pref, int cs, int type)
{
	sector_t	*sec;
	line_t		*ld;
	int			rc, box[4];
	int			c, d, i, k, n;
	//-----------------------

	for (d = i = 0; i < cs; i++)
	{
		sec = &sectors[smk[i]];

		CLR_hitline();

		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k]; n = ld - lines;

			if (pref[n])
			{
				hitline[n] = 1; d++;
			}
			else if (ld->backsector
				&& !lineflags[n]
				&& Vtfl(ld))
			{
				hitline[n] = 1; c++;
			}
		}

		if (c < 2 && c + d > 2)
		{
			while (1)
			{
				int	lnum = -1;

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k]; n = ld - lines;

					if (pref[n] && hitline[n] == 1) { lnum = n; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
				{
					sector_t *chk, *osec = NULL;

					for (c = k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k]; n = ld - lines;
						chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

						if (hitline[n] == 4)
						{
							if (pref[n]) c++;

							if (!osec)
								osec = chk;
							else if (osec != chk
								&& (chk->floorheight != osec->floorheight
								||  chk->ceilingheight != osec->ceilingheight))
							{
								osec = NULL; break;
							}
						}
					}
						

					if (   osec
						&& c > 1
						&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& LS_OnlySSLinesInSSECarray(sec, hitline))
					{
						for (n = 0; n < hitssecptr; n++)
						{
							if (type)
								Setr3dsslcsec(&subsectors[hitssec[n]], osec - sectors, 48123);
							else
								Setr3dsslfsec(&subsectors[hitssec[n]], osec - sectors, 48124);
						}
					}

					for (k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k]; n = ld - lines;
						if (hitline[n] == 4) hitline[n] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}
}

// Look for unset basic missing tex loops
void MX_Postlim()
{
	byte		*plo = (byte*)calloc(numlines, 1);
	byte		*phi = (byte*)calloc(numlines, 1);
	int			*smk = (int *)malloc(numsectors * sizeof(int));
	int			lc, hc, cs;
	int			f, i;

	//--------------------------------

	for (i = 0; i < numsectors; i++)
	{
		if ( (sectors[i].specFlags & 0x3041C) == 0x1C
			&& sectors[i].linecount > 2
			&& sectors[i].heightsec < 0
			&& !SecCtrl(i)
			&& sectors[i].stptr < 0)
		{
			subfn_Postlim(&sectors[i]);
		}
	}

	// Mop up unset missing tex loops
	for (lc = hc = cs = i = 0; i < numlines; i++)
	{
		if (  !lineflags[i]
			&& lines[i].ml_flags & ML_TWOSIDED
			&& lines[i].specType == R3D_00
			&& (f = lines[i].mtflags) & R3D_MISTEXANY)
		{
			sector_t *fs = lines[i].frontsector;
			sector_t *bs = lines[i].backsector;
			int		 fdiff = 0, cdiff = 0;

			if (  !bs
				|| fs == bs
				|| (fs->specFlags & 0x40C) == 0xC
				|| (bs->specFlags & 0x40C) == 0xC
				|| fs->ceilingheight - fs->floorheight <= FRACUNIT
				|| bs->ceilingheight - bs->floorheight <= FRACUNIT)
			{
				continue;
			}

			if (f & R3D_MISLTEX1 && bs->floorpic != skyflatnum && (fdiff = bs->floorheight - fs->floorheight) > 8*FRACUNIT)
			{
				plo[i] = 1; lc++;
				addTo_smk(fs - sectors, smk, &cs);
			}
			else
			if (f & R3D_MISLTEX2 && fs->floorpic != skyflatnum && (fdiff = fs->floorheight - bs->floorheight) > 8*FRACUNIT)
			{
				plo[i] = 1; lc++;
				addTo_smk(bs - sectors, smk, &cs);
			}

			if (f & R3D_MISHTEX1 && bs->ceilingpic != skyflatnum && (cdiff = fs->ceilingheight - bs->ceilingheight) > 8*FRACUNIT)
			{
				phi[i] = 1; hc++;
				addTo_smk(fs - sectors, smk, &cs);
			}
			else
			if (f & R3D_MISHTEX2 && fs->ceilingpic != skyflatnum && (cdiff = bs->ceilingheight - fs->ceilingheight) > 8*FRACUNIT)
			{
				phi[i] = 1; hc++;
				addTo_smk(bs - sectors, smk, &cs);
			}
		}
	}

	if (lc > 1) setSSECval(smk, plo, cs, 0);
	if (hc > 2) setSSECval(smk, phi, cs, 1);

	M_Free(plo);
	M_Free(phi);
	M_Free(smk);
}

//======================================================================

static RDBOOL noLowerTex(subsector_t *ss)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->sidedef && sg->sidedef->bottomtexture) return false;
	}

	return true;
}

static void subfn_MovMixed(sector_t *s, sector_t *smov, int lnum)
{
	sector_t	*fs, *bs, *chk, *sfix = NULL;
	sector_t	*os = NULL, *movsec = NULL;
	line_t		*ld;
	RDBOOL		setceil = true;
	int			rc, box[4];
	int			c, i, k;

	CLR_hitline();
	hitline[lnum] = 1;	// seed line

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (   lineflags[ld - lines]
			|| ld - lines == lnum
			|| ld->spflags & SPF_CONTROLLINE)
		{
			continue;
		}

		fs = ld->frontsector, bs = ld->backsector;

		if (!bs || fs == bs) continue;

		if (origTEX[ld - lines] & 0x22)
		{
			if (   ld->lnflags2 == 0x54000000
				&& (bs = ld->backsector)
				&& (fs = ld->frontsector) != bs
				&& (fs->specFlags & SF_MOVFLOOR)^(bs->specFlags & SF_MOVFLOOR))
			{
				hitline[ld - lines] = 1; c++;
			}
			else
			if (   fs->floorheight != bs->floorheight
				&& !(fs->specFlags & SF_MOVFLOOR)
				&& !(bs->specFlags & SF_MOVFLOOR))
			{
				chk = fs == s ? bs : fs;

				if (!sfix) sfix = chk; else if (sfix != chk)
				{
					continue;
				}

				hitline[ld - lines] = 1; c++;
			}
		}
	}

	if (c < 2 || !sfix || (sfix->specFlags & 0x1000C) != 0xC) return;

	if (   !subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
		|| !LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_REVERSED))
	{
		return;
	}

	// check the lines in the subsectors
	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				linearray[c++] = ld - lines;
				fs = ld->frontsector, bs = ld->backsector;

				if (!bs || fs == bs) continue;
				if (hitline[ld - lines] == 4) continue;

				if (   (origTEX[ld - lines] & 0x01 && fs->floorheight < bs->floorheight)
					|| (origTEX[ld - lines] & 0x10 && fs->floorheight > bs->floorheight))
				{
					return;
				}

				if (   (origTEX[ld - lines] & 0x04 && fs->ceilingheight > bs->ceilingheight)
					|| (origTEX[ld - lines] & 0x40 && fs->ceilingheight < bs->ceilingheight))
				{
					setceil = false;
				}
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		if (setceil) Setr3dsscs(ss, s - sectors, 48125);

		if (noLowerTex(ss))
		{
			Setr3dssfs(ss, sfix - sectors, 48126);
			Setr3dsslfsec(ss, s - sectors, 48127);
		}
	}

	for (i = 0; i < c; i++)
	{
		ld = &lines[linearray[i]];

		if ( !(origTEX[ld - lines] & 0x55) )
		{
			if (origTEX[ld - lines] & 0x22)
				Settype(ld, R3D_16, 48128);
			else
				Settype(ld, R3D_3F, 48129);
		}
		else
		{
			if (origTEX[ld - lines] & 0x22)
				Settype(ld, R3D_16, 48130);
			else
				Settype(ld, R3D_14, 48131);
		}

		SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48132);
	}
}

// v220-10
// Where analysis has trouble is with a moving sector inside
// and forming part of a static group where the moving sector
// only has the unshared part retaining lower textures.
void MX_MovMixed()
{
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			i;

	// Find any lower missing lines in a moving sector that are
	// possibly sharing with a static loop with no lower textures
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !lineflags[i]
			&& ld->lnflags2 == 0x54000000
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& (bs = ld->backsector)
			&& (fs = ld->frontsector) != bs
			&& origTEX[i] & 0x22
			&& (fs->specFlags & SF_MOVFLOOR)^(bs->specFlags & SF_MOVFLOOR))
		{
			chk = fs->specFlags & SF_MOVFLOOR ? fs : bs;

			if (   (chk == fs && !(origTEX[ld - lines] & 0x10))
				|| (chk == bs && !(origTEX[ld - lines] & 0x01)) )
			{
				subfn_MovMixed(fs == chk ? bs : fs, chk, i);
			}
		}
	}
}

//=============================================================

static void subfn_MX_Multi(sector_t *sec)
{
	sector_t	*fs, *bs, *chk, *osec;
	line_t		*ld;
	fixed_t		ch = sec->ceilingheight;
	short		cp = sec->ceilingpic;
	int			i, rc, tx, box[4];
	RDBOOL		multisec = false;

	for (i = 0, osec = NULL; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (   lineflags[ld - lines]
			|| !(ld->ml_flags & ML_TWOSIDED)
			|| (fs = ld->frontsector) == (bs = ld->backsector))
		{
			return;
		}

		chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (!osec)
		{
			if ((ch = chk->ceilingheight) <= sec->ceilingheight) return;
			cp = chk->ceilingpic;
			osec = chk;
		}

		if (   chk->ceilingheight != ch
			|| chk->ceilingpic != cp
			|| chk->specFlags & SF_MOVCEIL
			|| (chk->specFlags & 0xC && !(chk->specFlags & SF_SSLINETRUE)) )
		{
			return;
		}

		if (osec != chk) multisec = true;
	}

	if (!osec || !multisec) return;	// standard contained missing tex loops are dealt with elsewhere

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
		hitline[sec->lines[i] - lines] = 1;

	if (   subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &rc, box, NULL, NULL)
		&& rc == sec->linecount
		&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
		&& LS_OnlySSLinesInSSECarray(sec, hitline))
	{
		for (i = 0; i < hitssecptr; i++)
		{
			Setr3dsscs(&subsectors[hitssec[i]], osec - sectors, 48133);
			Setr3dsslcsec(&subsectors[hitssec[i]], sec - sectors, 48134);
		}

		for (i = 0; i < sec->linecount; i++)
		{
			ld = sec->lines[i];
			tx = ld->frontsector == sec ? 0x20 : 0x02;
			if (origTEX[ld - lines] & tx)
				Settype(ld, R3D_03, 48135);
			else if (!origTEX[ld - lines])
				Settype(ld, R3D_3F, 48136);
			else
				Settype(ld, R3D_36, 48137);

			SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48138);
		}
	}

}

// v220-10 look for missing hi in multi sec referenced loops
void MX_Multi()
{
	sector_t	*sec;
	int			i;

	// Find any higher missing lines in a static sector
	// that reference more than one other sector
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  (sec->specFlags & 0x30001C) == 0x18
			&& !secdone[i]
			&& sec->stptr < 0
			&& sec->ceilingpic != skyflatnum)
		{
			subfn_MX_Multi(sec);
		}
	}
}

// ===============================================

static RDBOOL line_error;

static void subfn_addLine(sector_t *sc, line_t *ld, int *c, int *lnum)
{
	ssline_t *sl = &sslines[ld - lines];
	int		 i, k;

	if (line_error) return;
	
	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			ss->ssvalidcount = ss_validcount;

			if (ss->sector == sc)
			{
				seg_t *sg = &segs[ss->firstline];

				hitssec[hitssecptr++] = sl->list[i];

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if (   sg->linedef
						&& sg->linedef->validcount != validcount)
					{
						int v = sg->linedef - lines;

						sg->linedef->validcount = validcount;

						if (hitline[v] & 4) continue;

						if (!hitline[v])
						{
							if (sg->linedef->frontsector == sg->linedef->backsector)
								subfn_addLine(sc, sg->linedef, c, lnum);

							line_error = true;
						}

						conarray[v] = 1; *lnum = v; (*c)++;
					}
				}
			}
		}
	}
}

static int subfn_ChkRing2(sector_t *sc, int *rc, int *box)
{
	int c, i;
	int lnum = -1;

	ss_validcount++;
	validcount++;

	CLR_conarray();

	hitssecptr = 0;

	for (c = i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (hitline[ld - lines] == 4) subfn_addLine(sc, ld, &c, &lnum);
	}

	return lnum >= 0
		&& !line_error
		&& c > 2
		&& subfn_CheckForRing(conarray, sc, lnum, rc, box, NULL, NULL)
		&& *rc == c;
}

static int findFloorsec(sector_t *s)
{
	int i, k;

	for (i = 0; i < s->linecount; i++)
	{
		if (s->lines[i]->spflags & SPF_CONTROLLINE) continue;

		ssline_t *sl = &sslines[s->lines[i] - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->sector == s && ss->floorsec >= 0)
				return -1;	// already set
			if (ss->sector != s && ss->floorsec >= 0)
				return ss->floorsec;
		}
	}
	return -1;
}

//============================================================

static int subfn_missingClosedContained(int idx, int *osec_idx)
{
	sector_t	*osec = NULL;
	sector_t	*chk;
	sector_t	*sec = &sectors[idx];
	int			chk_ht = sec->floorheight;
	int			c, i, k, m;

	for (c = i = 0, *osec_idx = -1; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (lineflags[ld - lines] || Vtxs2(ld)) return 1;

		chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (!chk || chk == sec) return 2;

		if (!osec)
		{
			if (chk->specFlags & (SF_MOVSECTOR|SF_NOLOTEX)) return 3;
			if (chk_ht - chk->floorheight < 8*FRACUNIT) return 4;
			osec = chk;
		}
		else if (osec != chk)
		{
			return 5;
		}

		if (!origTEX[ld - lines])
		{
			vertexowner_t	*vo = &vertexowners[ld->v1 - vertexes];
			int				ht = DDMININT;

			for (k = 0; k < 2; k++, vo = &vertexowners[ld->v2 - vertexes])
			{
				if (vo->num > 2)
				{
					for (m = 0; m < vo->num; m++)
					{
						if (vo->list[m] == idx || vo->list[m] == osec - sectors) continue;

						if ( (ht = sectors[vo->list[m]].floorheight) < chk_ht)
						{
							return 6;
						}
					}
				}
			}

			if (ht > DDMININT) c++;
		}
	}

	if (c < 4) return 7;
	if (!osec) return 8;

	*osec_idx = osec - sectors;

	return 0;
}

// Look for a static missing tex loop with midtex lines
// contained inside a standard static sector
void MX_MissingContained()
{
	sector_t	*s;
	line_t		*ld;
	int			osec_idx, rc, box[4];
	int			i, k, m;

	for (i = 0; i < numsectors; i++)
	{
		if (   secdone[i]
			|| SecCtrl(i)
			|| (sectors[i].specFlags & 0xF001C) != 0x1C
			|| sectors[i].heightsec >= 0)
		{
			continue;
		}

		if ( (k = subfn_missingClosedContained(i, &osec_idx)) )
		{
			continue;
		}

		s = &sectors[i];

		// Check that the lines form a single loop and that chk is contained in osec_idx
		CLR_hitline();

		for (k = 0; k < s->linecount; k++) hitline[s->lines[k] - lines] = true;

		if (   subfn_CheckForRing(hitline, s, s->lines[0] - lines, &rc, box, NULL, NULL)
			&& rc == s->linecount
			&& LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE)
			&& LS_OnlySSLinesInSSECarray(s, hitline))
		{
			// First set the contained subsecs
			for (k = 0; k < hitssecptr; k++)
			{
				Setr3dsslfsec(&subsectors[hitssec[k]], osec_idx, 48139);
				Setr3dssfs(&subsectors[hitssec[k]], i, 48140);
			}

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (origTEX[ld - lines])	// midtex line
					Settype(ld, R3D_03, 48141);
				else
				{
					// Fill in any gaps
					ssline_t *sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->lfsec < 0 && ss->sector - sectors == osec_idx)
							Setr3dsslfsec(ss, s - sectors, 48085);
					}
					
					Settype(ld, R3D_3F, 48142);
				}
			}

			subfn_setSSEC_LFG(48143, true);
			Setsecdone(&sectors[osec_idx], 48144);
			Setsecdone(s, 48145);
		}
	}
}

static RDBOOL nohitexInSsecs(int height)
{
	int			i;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		if (   ss->sector->ceilingheight <= height
			|| !(ss->sector->specFlags & SF_NOHITEX)
			|| ss->sector->specFlags & (SF_SSLINETRUE|SF_MOVCEIL))
		{
			return false;
		}
	}

	return true;
}

// v220-10
void MX_R3D_16Nested()
{
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			ht, rc, rc2, box[4], box2[4];
	int			c, i, k, n;

//	1) Deal with nohi/lotex sectors that may or may not have R3D_16 set
	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && (sectors[i].specFlags & 0xFFC) == 0x01C)
		{
			int lnum = -1, fsec;

			CLR_hitline();
			chk = &sectors[i];

			for (c = k = 0; k < chk->linecount; k++)
			{
				ld = chk->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   ld->lnflags2 != 0x0C000000
					|| !Vtxs2(ld)
					|| lineflags[ld - lines] & (LFG_3DTYPE|LFG_NOARBITER))
				{
					c = 0; break;
				}

				lnum = ld - lines; hitline[lnum] = 1; c++;
			}

			if (   c > 2
				&& subfn_CheckForRing(hitline, chk, lnum, &rc, box, NULL, NULL)
				&& rc == c
				&& LS_MarkSubsecsInAreaGeneral(chk, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(chk, hitline)
				&& (fsec = findFloorsec(chk)) >= 0)
			{
				for (k = 0; k < hitssecptr; k++)
				{
					Setr3dsslfsec(&subsectors[hitssec[k]], -1, 48146);
					Setr3dssfs(&subsectors[hitssec[k]], fsec, 48147);
					if (chk->lightlevel != sectors[fsec].lightlevel)
						chk->lightlevel = sectors[fsec].lightlevel;
				}

				for (k = 0; k < chk->linecount; k++)
				{
					ld = chk->lines[k];

					if ( !(ld->spflags & SPF_CONTROLLINE) )
						Forcetype(ld, R3D_3F, 48148);
				}

				subfn_setSSEC_LFG(48149, false);
			}
		}
	}

//	2)	Deal with closed R3D_00 loops in static textureless sectors
//      where the lines only own other static textureless sectors.
//      These would normally remain unset.
	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || (sectors[i].specFlags & 0x3001C) != 0x1C) continue;

		for (c = k = 0; k < sectors[i].linecount; k++)
		{
			ld = sectors[i].lines[k];
			n = ld - lines;

			if (   ld->ml_flags & ML_TWOSIDED
				&& !origTEX[n]
				&& !lineflags[n]
				&& ld->specType == R3D_00
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& (ld->frontsector->specFlags & 0x3000C) == 0xC
				&& (ld->backsector->specFlags & 0x3000C) == 0xC
				&& (ld->lnflags2 == 0x34000000 || ld->lnflags2 == 0x54000000)
				&& ( (ld->mtflags & (R3D_MISLTEX1|R3D_INVLTEX1)) == R3D_MISLTEX1
				||   (ld->mtflags & (R3D_MISLTEX2|R3D_INVLTEX2)) == R3D_MISLTEX2) )
			{
				fs = ld->mtflags == R3D_MISLTEX1 ? ld->backsector : ld->frontsector;
				bs = fs == ld->frontsector ? ld->backsector : ld->frontsector;
				ht = fs->floorheight - bs->floorheight;

				if (ht < 8*FRACUNIT) continue;

				linearray[c++] = n;
			}
		}

		// v220-11 Moved out of previous loop
		if (c > 2)
		{
			CLR_hitline();

			for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

			while(1)
			{
				int lnum = -1;

				for (k = 0; k < c; k++)
				{
					if (hitline[linearray[k]] == 1) { lnum = linearray[k]; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, fs, lnum, &rc, box, NULL, NULL))
				{
					if (   rc > 2
						&& LS_MarkSubsecsInAreaGeneral(fs, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& LS_OnlySSLinesInSSECarray(fs, hitline))
					{
						for (k = 0; k < hitssecptr; k++)
						{
							SetOverride(&subsectors[hitssec[k]], bs - sectors, 48150);
						}

						for (k = 0; k < c; k++)
						{
							n = linearray[k];

							if (hitline[n] == 4)
							{
								Settype(&lines[n], R3D_3F, 48151);
								SetLFG(&lines[n], LFG_NEWSET, 48152);
							}
						}
					}

					for (k = 0; k < c; k++)
					{
						n = linearray[k];

						if (hitline[n] == 4) hitline[n] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}

//	3)	Deal with closed R3D_16/R3D_00 loops in static or floor move sectors
//      where floorsec inheritance may be needed

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !origTEX[i]
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& (ld->specType == R3D_16 || (ld->specType == R3D_00 && ld->lnflags2 == 0x0C000000 && !lineflags[i]))
			&& ( (ld->mtflags & (R3D_MISLTEX1|R3D_INVLTEX1)) == R3D_MISLTEX1
			||   (ld->mtflags & (R3D_MISLTEX2|R3D_INVLTEX2)) == R3D_MISLTEX2) )
		{
			fs = ld->mtflags == R3D_MISLTEX1 ? ld->backsector : ld->frontsector;
			bs = fs == ld->frontsector ? ld->backsector : ld->frontsector;

			if (  !(bs->specFlags & SF_MOVFLOOR)
				&& ((fs->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR || !(fs->specFlags & SF_MOVSECTOR)) )
			{
				hitline[i] = 1;

				for (k = 0; k < c; k++)
				{
					if (countsec[k] == fs - sectors) goto end_loop;
				}

				countsec[c++] = fs - sectors;
			}
		}
end_loop:;
	}

	for (i = 0; i < c; i++)
	{
		fs = &sectors[countsec[i]];

		while(1)
		{
			int		lnum = -1;

			for (k = 0; k < fs->linecount; k++)
			{
				ld = fs->lines[k];

				if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, fs, lnum, &rc, box, NULL, NULL))
			{
				int		linetagchk;
				int		mask = 0;
				int		typeR3D_16 = 0;

				if (lines[lnum].mtflags == (R3D_MISLTEX1|R3D_MISHTEX2))
					mask = R3D_MISLTEX1|R3D_MISHTEX2;
				else
				if (lines[lnum].mtflags == (R3D_MISLTEX2|R3D_MISHTEX1))
					mask = R3D_MISLTEX2|R3D_MISHTEX1;

				// v220-10 if mask is same on all lines and chkRing fails
				//         then type will default to R3D_16
				if (mask)
				{
					for (k = 0; k < fs->linecount; k++)
					{
						ld = fs->lines[k];

						if (hitline[ld - lines] == 4 && ld->mtflags != mask)
						{
							mask = 0;
							break;
						}
					}
				}

				if (!mask)
				{
					for (k = 0; k < fs->linecount; k++)
					{
						ld = fs->lines[k];

						if (hitline[ld - lines] == 4 && ld->specType == R3D_16)
						{	typeR3D_16++;
						}
					}
				}

				typeR3D_16 = (typeR3D_16 == rc);

				line_error  = false;
				bs = NULL;

				if (   subfn_ChkRing2(fs, &rc2, box2)
					&& (M_CompareBox(box, box2) || M_CompareBox(box2, box)) )
				{
					for (k = 0; k < fs->linecount; k++)
					{
						ld = fs->lines[k];

						if (hitline[ld - lines] == 4 || conarray[ld - lines] == 4)
						{
							chk = ld->frontsector == fs ? ld->backsector : ld->frontsector;

							hitline[ld - lines] = 4;

							if (!line_error)
							{
								if (!bs)
									bs = chk;
								else if (chk->floorheight != bs->floorheight
									|| chk->ceilingheight != bs->ceilingheight)
								{
									line_error = true; bs = NULL;
								}
							}
						}
					}
				}

				// This may now be redundant: see MX_CheckNoTexDualRings()
				// If double ring structure is tagged and trigger lines are inside the
				// structure then assumes structure should be drawn. Normally it is not.
				// This is only done if the lines have already been preset to R3D_16
				linetagchk  = (bs && typeR3D_16 && fs->tag && fs->specFlags & SF_MOVFLOOR) ? fs->tag : 0;

				if (linetagchk)
				{
					for (k = 0; k < numlines; k++)
					{
						if (   lines[k].tag == linetagchk
							&& M_FitBox(lines[k].bbox, box))
						{
							bs = NULL; break;
						}
					}
				}

				if (bs)
				{
					for (k = 0; k < hitssecptr; k++)
						SetOverride(&subsectors[hitssec[k]], bs - sectors, 48153);
				}
				// v220-20 check if this is a containing hitex ring
				else if (mask && !linetagchk)
				{
					sector_t *os = NULL;

					for (k = 0; k < fs->linecount; k++)
					{
						ld = fs->lines[k];

						if (hitline[ld - lines] == 4)
						{
							chk = ld->frontsector == fs ? ld->backsector : ld->frontsector;

							if (   chk
								&& (!(chk->specFlags & SF_NOHITEX) || chk->specFlags & SF_SSLINETRUE)
								&& chk->ceilingheight < fs->ceilingheight
								&& chk->ceilingpic != skyflatnum)
							{
								if (!os) os = chk; else if (os != chk) { os = NULL; break; }
							}
							else
							{ os = NULL; break;
							}
						}
					}

					if (   os
						&& LS_MarkSubsecsInAreaGeneral(fs, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& nohitexInSsecs(os->ceilingheight))
					{					
						for (k = 0; k < hitssecptr; k++)
						{
							Setr3dsscs(&subsectors[hitssec[k]], os - sectors, 48154);
						}
					}
				}

				for (k = 0; k < fs->linecount; k++)
				{
					ld = fs->lines[k];

					if (hitline[ld - lines] == 4)
					{
						hitline[ld - lines] = 0;
						if (bs) Forcetype(ld, R3D_3F, 48155); else if (mask) Forcetype(ld, R3D_16, 48156);
					}
				}

			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}
}

// ===========================================

void subfn_MX_PartialMissing(sector_t *sc)
{
	sector_t	*chk;
	sector_t	*osl = NULL, *osh = NULL;
	sector_t	*fs, *bs;
	line_t		*ld;
	byte		tx;
	int			rc, box[4];
	int			i, lnum;
	int			c = 0, hi = 0, lo = 0;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		lnum = ld - lines;

		if (   !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[lnum] & LFG_3DTYPE
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR)
		{
			return;
		}

		if (lineflags[lnum]) c++;

		tx = origTEX[lnum];

		chk = fs == sc ? bs : fs;

		if (chk->ceilingpic != skyflatnum && !(tx & 0x44) && abs(fs->ceilingheight - bs->ceilingheight) > 8*FRACUNIT)
		{
			if (!osh) (osh = chk); else if (osh->ceilingheight != chk->ceilingheight) return;
			hi++;
		}

		if (chk->floorpic != skyflatnum && !(tx & 0x11) && abs(fs->floorheight - bs->floorheight) > 8*FRACUNIT)
		{
			if (!osl) (osl = chk); else if (osl->floorheight != chk->floorheight) return;
			lo++;
		}
	}

	if (c == sc->linecount) return;

	if (hi) hi = (hi >= sc->linecount - 1);
	if (lo) lo = (lo >= sc->linecount - 1);
	if (hi && osh && osh->specFlags & SF_NOHITEX && !(osh->specFlags & SF_SSLINETRUE)) hi = 0;
	if (lo && osl && osl->specFlags & SF_NOLOTEX && !(osl->specFlags & SF_SSLINETRUE)) lo = 0;

	if (!hi && !lo) return;

	CLR_hitline();

	for (i = 0; i < sc->linecount; i++) hitline[sc->lines[i] - lines] = 1;

	if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
		&& rc == sc->linecount
		&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
		&& LS_OnlySSLinesInSSECarray(NULL, hitline))
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			if (hi && osh) Setr3dsscs(ss, osh - sectors, 48157);
			if (lo && osl) Setr3dsslfsec(ss, osl - sectors, 48158);
		}

		subfn_setSSEC_LFG(48159, false);
	}
}


// v220-10
void MX_PartialMissing()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !SecCtrl(i)
			&& sectors[i].linecount > 3
			&& sectors[i].specFlags & 0xC
			&& (sectors[i].specFlags & 0xC) != 0xC
			&& sectors[i].specFlags & 0x10
			&& !(sectors[i].specFlags & 0x30400)
			&& sectors[i].ceilingpic != skyflatnum
			&& sectors[i].floorpic != skyflatnum)
		{
			subfn_MX_PartialMissing(&sectors[i]);
		}
	}
}

// ===========================================

static void subfn_checkMultiFloorMove(sector_t *sc, int *mrkline)
{
	int c, i, rc, box[4];

	for (c = i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (origTEX[ld - lines]) return;

		if (   ld->specType == R3D_00
			|| ld->specType == R3D_01
			|| ld->specType == R3D_3E)
		{
			if (lineflags[ld - lines] & 0xFFFB)
			{
				return;
			}
		}
		else
			return;

		mrkline[c++] = ld - lines;
	}

	if (c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[mrkline[i]] = 1;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[mrkline[i]] == 1) { lnum = mrkline[i]; break; }
		}

		if (lnum < 0) break;

		if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
			&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
			&& LS_OnlySSLinesInSSECarray(NULL, hitline))
		{
			sector_t *os = NULL, *chk;

			for (i = 0; i < c; i++)
			{
				if (hitline[mrkline[i]] == 4)
				{
					line_t *ld = &lines[mrkline[i]];
					chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;
					if (!os)
						os = chk;
					else if (chk->floorheight != os->floorheight
						|| chk->ceilingheight != os->ceilingheight)
					{
						os = NULL; break;
					}
				}
			}

			if (os)
			{
				for (i = 0; i < hitssecptr; i++)
				{
					SetOverride(&subsectors[hitssec[i]], os - sectors, 48160);
				}
			}

			for (i = 0; i < c; i++)
			{
				if (hitline[mrkline[i]] == 4)
				{
					hitline[mrkline[i]] = 0;

					if (os) Forcetype(&lines[mrkline[i]], R3D_3F, 48161);
				}
			}

			if (os) subfn_setSSEC_LFG(48162, false);
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-10
void MX_CheckMultiFloorMove()
{
	int	i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& (sectors[i].specFlags & 0x13FF1C) == 0x0011001C)
		{
			int	*mrkline = (int*)malloc(sectors[i].linecount * sizeof(int));
			subfn_checkMultiFloorMove(&sectors[i], mrkline);
			M_Free(mrkline);
		}
	}
}

// ===================================================

static int count_sec;
static byte *mrksec;

static void subfn_iterateSec(sector_t *sc)
{
	int i;

	if (mrksec[sc - sectors]) return;

	mrksec[sc - sectors] = 1;
	countsec[count_sec++] = sc - sectors;

	for (i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   ld->ml_flags & ML_TWOSIDED)
		{
			if (   ((ld->frontsector->specFlags & 0xFFBFC) == 0x8394
				||  (ld->frontsector->specFlags & 0xFBFFC) == 0x809C
				||  (ld->frontsector->specFlags & 0xFFFFC) == 0x9001C)
				&& ld->frontsector->heightsec < 0
				&& ld->frontsector->stptr < 0)
			{
				subfn_iterateSec(ld->frontsector);
			}

			if (   ((ld->backsector->specFlags & 0xFFBFC) == 0x8394
				||  (ld->backsector->specFlags & 0xFBFFC) == 0x809C
				||  (ld->backsector->specFlags & 0xFFFFC) == 0x9001C)
				&& ld->backsector->heightsec < 0
				&& ld->backsector->stptr < 0)
			{
				subfn_iterateSec(ld->backsector);
			}
		}
	}
}

RDBOOL seedSector(sector_t *sc)
{
	sector_t *fs, *bs;
	int		 a, b, i;
	byte	 tx;

	for (a = b = i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| ld->spflags & SPF_CONTROLLINE
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| (tx = origTEX[ld - lines]) & 0x11
			|| (tx & 0x44) == 0x44
			|| (tx & 0x22) == 0x22
			|| (tx & 0x40 && fs->ceilingheight >= bs->ceilingheight)
			|| (tx & 0x04 && fs->ceilingheight <= bs->ceilingheight))
		{
			return false;
		}

		if (!tx && fs->floorheight == bs->floorheight) a++;
		if (tx & 0x22 && !(tx & 0x11) && fs->floorheight != bs->floorheight) b++;
	}

	return a && b;
}

// v220-10 Deals with rare exploitation of the more esoteric vagaries of the software renderer.
void MX_ConnectedLo()
{
	sector_t *os;
	RDBOOL	 setssec = false;
	int		 lnc, i, k, m;
	
	mrksec = (byte*)calloc(numsectors, 1);

	validcount++;

	for (lnc = i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0xFBFFC) == 0x8394
			&& sectors->heightsec < 0
			&& sectors->stptr < 0
			&& seedSector(&sectors[i]))
		{
			os = &sectors[i];
			count_sec = 0;
			subfn_iterateSec(os);

			if (count_sec > 1)
			{
				for (k = 0; k < count_sec; k++)
				{
					sector_t *s = &sectors[countsec[k]];
					int		 sm = s->specFlags & SF_MOVSECTOR;

					for (m = 0; m < s->linecount; m++)
					{
						line_t *ld = s->lines[m];
						byte	tx = origTEX[ld - lines];

						if (ld->validcount == validcount) continue;
						ld->validcount = validcount;
						if (ld->spflags & SPF_CONTROLLINE) continue;

						if (!sm && ld->ml_flags & ML_TWOSIDED)
						{
							if (tx & 0x22)
							{
								// v220-20 add test
								if (   (tx == 0x06 || tx == 0x02 || tx == 0x60 || tx == 0x20)
									&& (ld->frontsector->floorheight   != ld->backsector->floorheight
									||  ld->frontsector->ceilingheight != ld->backsector->ceilingheight))
								{
									if (tx & 0x44)
										Forcetype(ld, R3D_2D, 48163);
									else
										Forcetype(ld, R3D_1D, 48164);
								}
								else
								if (   (tx == 0x06 || tx == 0x60)
									&& ld->frontsector->ceilingpic == ld->backsector->ceilingpic
									&& ld->frontsector->floorpic   == ld->backsector->floorpic
									&& ld->frontsector->floorheight != ld->backsector->floorheight)
								{
									linearray[lnc++] = ld - lines;
								}
								else
									Forcetype(ld, R3D_03, 48165);
							}
							else if ((!tx
								|| (!(tx & 0x33) && tx & 0x44 && Vtfl3(ld)))
								&& ld->frontsector->floorheight != ld->backsector->floorheight)
							{
								Forcetype(ld, R3D_14, 48166);
							}
							else
							{
								Forcetype(ld, R3D_36, 48167);
							}

							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48168);
						}

						if (ld->frontsector->R3Dlowsec < 0) ld->frontsector->R3Dlowsec = os - sectors;
						if (ld->backsector && ld->backsector->R3Dlowsec < 0) ld->backsector->R3Dlowsec = os - sectors;
					}

					Setsecdone(&sectors[countsec[k]], 48169);
				}

				setssec = true;
			}
		}
	}

	ss_validcount++;

	// Gets even more complicated as we have to check subsectors
	// belonging to these lines for any unset R3D_14 lines
	for (i = 0; i < lnc; i++)
	{
		int			d, lnum = linearray[i];
		subsector_t	*refss = NULL;
		ssline_t	*sl = &sslines[lnum];
		sector_t	*s = origTEX[lnum] == 0x06 ? lines[lnum].frontsector : lines[lnum].backsector;

		for (d = k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				if (ss->sector == s && !(ss->ssflags & DDSUBF_MIDPOINT))
				{
					if (d) { refss = NULL; break; }	// only one allowed
					refss = ss; d++;
				}
			}
		}

		if (refss)
		{
			sector_t	*fs = NULL;
			seg_t		*sg = &segs[refss->firstline];

			validcount++;

			for (d = m = 0; m < refss->linecount; m++, sg++)
			{
				if (sg->linedef && sg->linedef->validcount != validcount)
				{
					sg->linedef->validcount = validcount;

					if (!fs) fs = sg->frontsector;
					else if (fs != sg->frontsector) { d = 0; break; }

					d++;
				}
			}

			if (d > 1)	// eliminate single line ssecs
			{
				os = s == lines[lnum].frontsector ? lines[lnum].backsector : lines[lnum].frontsector;
				if (refss->floorsec < 0) refss->floorsec = os - sectors;
				lines[lnum].specType = R3D_00;
			}
		}
	}

	M_Free(mrksec);

	if (setssec)
	{
		for (i = 0; i < numsubsectors; i++)
		{
			if (  (os = subsectors[i].sector)->R3Dlowsec >= 0)
			{
				subsectors[i].ssflags |= SSF_NOOCCLUDE;

				if (   (os->specFlags & 0xFFBFC) == 0x809C
					|| ((os->specFlags & 0xFFBFC) == 0x8394
					&& sectors[os->R3Dlowsec].floorheight == os->floorheight) )
				{
					subsectors[i].ssflags |= SSF_LOWFLOORSEC;
				}
			}
		}
	}
}

// ===================================================

static RDBOOL checkLoop(int lnum, int *mrkline, int loop_count, int *ring_count, int *bbox, int *seccount, int *os)
{
	line_t		*ld = &lines[lnum];
	vertex_t	*v = ld->v1;
	RDBOOL		isclosed = false, repeat = true;
	byte		*mrksec = NULL;
	int			chkht;	// used to get the reference sector by finding the lowest ceiling
	int			i;

	//--------------------------------

	*seccount = 0;
	*os = -1;
	hitline[lnum] = 2;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < loop_count; i++)
		{
			int k = mrkline[i];

			if (hitline[k] == 1)
			{
				line_t	*li = &lines[k];

				if (li->v1 == v)
				{
					v = li->v2;
					hitline[k] = 2;
					repeat = true;
				}
				else if (li->v2 == v)
				{
					v = li->v1;
					hitline[k] = 2;
					repeat = true;
				}
			}
		}
	}
	
	isclosed = (v == ld->v2);

	*ring_count = 0;

	if (isclosed)
	{
		mrksec = (byte*)calloc(numsectors, sizeof(byte));
		M_ClearBox(bbox);
	}

	for (i = 0; i < loop_count; i++)
	{
		int n = mrkline[i];

		if (hitline[n] == 2)
		{
			if (isclosed)
			{
				sector_t *ref;
				line_t	*li = &lines[n];
				int		side = li->frontsector->ceilingheight > li->backsector->ceilingheight ? 0 : 1;
				int		snum = side ? li->backsector - sectors : li->frontsector - sectors;
				int		snum2;

				ref = side ? li->frontsector : li->backsector;
				snum2 = ref - sectors;

				if (!(mrksec[snum] & 1)) { (*seccount)++; mrksec[snum] |= 1; }

				if (!(mrksec[snum2] & 2))
				{
					if (*os < 0)
					{
						*os = ref - sectors;
						chkht = ref->ceilingheight;
					}
					else if (ref->ceilingheight < chkht)
					{
						*os = ref - sectors;
						chkht = ref->ceilingheight;
					}

					mrksec[snum2] |= 2;
				}

				M_AddLineToBox(bbox, li);
				hitline[n] = 4;

				// Set the linearray and sidelist lists
				linearray[*ring_count] = n;
				sidelist[*ring_count] = li->sidenum[side];

				(*ring_count)++;
			}
			else
			{
				hitline[n] = 1;
			}
		}
	}

	M_Free(mrksec);

	return isclosed;
}

// v220-10 look specifically for missing hi loops traced with multiple sectors
//         with surrogate missing lines that can cause a loop failure.
void MX_MultiHiLineRing()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			rc, box[4];
	int			a, c, i, k, mlc;
	int			*mrkline = (int*)malloc(numlines * sizeof(int));
	int			*ml;
	byte		*vf;

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !lineflags[i]
			&& ld->ml_flags & ML_TWOSIDED
			&& !(origTEX[i] & 0x44)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->ceilingpic != skyflatnum
			&& bs->ceilingpic != skyflatnum
			&& !(fs->specFlags & SF_MOVCEIL)
			&& !(bs->specFlags & SF_MOVCEIL)
			&& abs(fs->ceilingheight - bs->ceilingheight) > 8*FRACUNIT)
		{
			hitline[i] = 1;
			mrkline[c++] = i;
		}
	}

	if (c < 4)
	{
		M_Free(mrkline);
		return;
	}

	// eliminate standard rings
	memcpy(conarray, hitline, numlines);

	a = 0;

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (conarray[mrkline[i]] == 1) { lnum = mrkline[i]; break; }
		}

		if (lnum < 0) break;

		fs = lines[lnum].frontsector->ceilingheight > lines[lnum].backsector->ceilingheight ?
			 lines[lnum].frontsector : lines[lnum].backsector;

		if (subfn_CheckForRing(conarray, fs, lnum, NULL, NULL, NULL, NULL))
		{
			for (i = 0; i < c; i++)
			{
				if (conarray[mrkline[i]] == 4)
				{
					conarray[mrkline[i]] = 0;
					hitline[mrkline[i]] = 0; a++;
				}
			}
		}
		else
		{
			conarray[lnum] = 0;
		}
	}

	// Add missing line count to their vertex numbers
	ml = (int* ) calloc(c, sizeof(int));
	vf = (byte*) calloc(numvertexes, sizeof(byte));

	for (i = 0; i < c; i++)
	{
		int n = mrkline[i];

		if (hitline[n])
		{
			vf[lines[n].v1 - vertexes]++;
			vf[lines[n].v2 - vertexes]++;
		}
	}

	// Find vertexes connected to more than one missing line
	// and, if found, mark the owner lines
	for (mlc = i = 0; i < c; i++)
	{
		int n = mrkline[i];

		if (hitline[n])
		{
			if (   vf[lines[n].v1 - vertexes] < 2
				|| vf[lines[n].v2 - vertexes] < 2)
			{
				mrkline[i] = 0;
			}
			else
			if (   vf[lines[n].v1 - vertexes] > 2
				|| vf[lines[n].v2 - vertexes] > 2)
			{
				ml[mlc++] = n;
			}
		}
	}

	if (mlc > 5)
	{
		line_t *li;

		for (i = 0; i < mlc; i++)
		{
			if (!hitline[ml[i]])  continue;

			ld = &lines[ml[i]];

			fs = ld->frontsector->ceilingheight < ld->backsector->ceilingheight ?
				ld->frontsector : ld->backsector;

			for (k = 0; k < mlc; k++)
			{
				if (i == k) continue;

				if (!hitline[ml[k]])  continue;

				li = &lines[ml[k]];

				if (li->v1 == ld->v1 || li->v1 == ld->v2 || li->v2 == ld->v1 || li->v2 == ld->v2)
				{
					bs = li->frontsector->ceilingheight < li->backsector->ceilingheight ?
						li->frontsector : li->backsector;

					if (fs != bs)
					{
						fs = ld->frontsector == fs ? ld->backsector : ld->frontsector;
						bs = li->frontsector == bs ? li->backsector : li->frontsector;

						if (fs == bs)
						{
							// remove line from loop
							hitline[ml[k]] = 0;
							break;
						}
					}
				}
			}
		}

		while (1)
		{
			int			seccount, secnum;
			int			d, lnum = -1;

			for (i = 0; i < c; i++)
			{
				if (   hitline[mrkline[i]] == 1
					&& lines[mrkline[i]].lnflags2 == 0x34000000)
				{
					lnum = mrkline[i]; break;
				}
			}

			if (lnum < 0) break;

			if (checkLoop(lnum, mrkline, c, &rc, box, &seccount, &secnum) && seccount > 1)
			{
				for (d = i = 0; i < c; i++)
				{
					if (hitline[mrkline[i]] == 4)
					{
						hitline[mrkline[i]] = 0;
					}
				}

				if (LS_MarkSsecsUsingSidelist(rc, box, MSIA_NONE))
				{
					for (i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];

						Setr3dsscs(ss, secnum, 50001);
					}
				}
			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}

	M_Free(vf);
	M_Free(ml);
	M_Free(mrkline);
}

// ==================================================================

static RDBOOL checkLinesInBox(int *n)
{
	line_t	*ld;
	int		c, i, k;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (!hitline[ld - lines] && ld->frontsector != ld->backsector) return false;

				linearray[c++] = ld - lines; c++;
			}
		}
	}

	*n = c;
	return true;
}

// v220-18 new test added (moving floor within containing floor)
// v220-15 new test added (moving floor within containing floor - different type)
// v220-10 look for bounded static sector containing a textureless mov sector and a reference sector
void MX_MovInside()
{
	int		c, i, k;

	if (R3D_ScriptedMap) return;

	for (i = 0; i < numsectors; i++)
	{
		if (SecCtrl(i) || sectors[i].stptr >= 0 || sectors[i].heightsec >= 0) continue;

		// v220-18 look for bounded ring having an inner textureless floor mov sector
		if (   sectors[i].linecount > 5
			&& (sectors[i].specFlags & 0x3041C) == 0x1C
			&& sectors[i].floorpic != skyflatnum)
		{
			sector_t *ch, *os, *sc = &sectors[i];
			line_t	 *ld;
			int		 a, box[2][4];

			M_ClearBox(box[0]);
			M_ClearBox(box[1]);

			for (a = c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX1)
				{
					if (  (ld->lnflags2 == 0x0C000000 || ld->lnflags2 == 0x0C000008)
						&& Vtxs2(ld))
					{
						ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

						if (!ch) { c = 0; break; }
						if (ch->specFlags & SF_MOVSECTOR) { a++; M_AddLineToBox(box[1], ld); continue; }
						if ((ch->specFlags & 0x400) || !(ch->specFlags & 0xC))
						{
							if (!os) os = ch; else if (os != ch) { c = 0; break; }
							linearray[c++] = ld - lines;
							M_AddLineToBox(box[0], ld);
						}
					}
				}
				else
				{
					c = 0; break;
				}
			}

			if (c > 2  && a && M_CompareBox(box[1], box[0]))
			{
				int m, rc;

				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

				if (   subfn_CheckForRing(hitline, sc, linearray[0], &rc, NULL, NULL, NULL)
					&& rc == c
					&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box[0], -1, MSIA_NONE))
				{
					subsector_t	*ss;
					seg_t		*sg;
					byte		txn;

					for (k = 0; k < hitssecptr; k++)
					{
						ss = &subsectors[hitssec[k]];
						if ((ss->sector->specFlags & 0xC) != 0xC) { hitssecptr = 0; break; }
					}

					if (hitssecptr)
					{
						int		count = 0;

						validcount++;

						for (c = k = 0; k < hitssecptr; k++)
						{
							ss = &subsectors[hitssec[k]];
							sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ((ld = sg->linedef) && ld->validcount != validcount)
								{
									ld->validcount = validcount;
									linearray[c++] = ld - lines;
								}
							}

							Setr3dssfs(ss, os - sectors, 48170);
						}

						addSec(sc - sectors, &count);

						for (k = 0; k < c; k++)
						{
							ld = &lines[linearray[k]];
							txn = origTEX[ld - lines];

							if (!txn)
							{
								Settype(ld, R3D_3F, 48171);
							}
							else if (txn == 0x22)
							{
								Settype(ld, R3D_03, 48172);
							}
							else
							{
								Settype(ld, R3D_36, 48173);
							}

							if (ld->frontsector == os && ld->backsector == os) continue;

							if (ld->frontsector != os) addSec(ld->frontsector - sectors, &count);
							if (ld->backsector && ld->backsector != os) addSec(ld->backsector - sectors, &count);

							SetLFG(ld, LFG_FIXABS|LFG_NEWSET|LFG_NOCHANGE, 48174);
						}

						for (k = 0; k < count; k++) Setsecdone(&sectors[countsec[k]], 48175);
					}
				}
			}
		}
	}

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || SecCtrl(i) || sectors[i].stptr >= 0 || sectors[i].heightsec >= 0) continue;

		if (   sectors[i].linecount > 5
			&& (sectors[i].specFlags & 0x3041C) == 0x1C
			&& sectors[i].floorpic != skyflatnum)
		{
			sector_t *s = &sectors[i], *chk;
			sector_t *ref[3] = { NULL, NULL, NULL};
			line_t	 *ld;
			int		 lnum, rc, box[4];
			short	 cpic = s->ceilingpic;

			CLR_hitline();

			// look for inner ring having a floor mov sector
			for (c = k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (   ld->lnflags2 == 0x0C000000
					&& Vtxs2(ld))
				{
					chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

					if (!chk) continue;

					if ((chk->specFlags & 0x3041C) == 0x10410)
					{
						byte txn = ld->frontsector == s ? 0x01 : 0x10;
						if ( !(origTEX[ld - lines] & txn) ) { c = 0; break; }
						if (!ref[0]) ref[0] = chk; else if (ref[0] != chk) { c = 0; break; }
					}

					if ((chk->specFlags & 0x3041C) == 0x410)
					{
						if (origTEX[ld - lines]) { c = 0; break; }
						if (!ref[1]) ref[1] = chk; else if (ref[1] != chk) { c = 0; break; }

						hitline[ld - lines] = 1; c++;
					}
				}
			}

			if (c > 2 && ref[0] && ref[1])
			{
				while(1)
				{
					for (lnum = -1, k = 0; k < s->linecount; k++)
					{
						ld = s->lines[k];
						if (hitline[ld - lines] == 1) { lnum = ld - lines; break; };
					}

					if (lnum < 0) break;

					if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
						&& rc > 2)
					{
						RDBOOL valid = false;

						if (LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE))
						{
							validcount++;

							for (c = k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];
								seg_t		*sg = &segs[ss->firstline];
								int			m;

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if ( (ld = sg->linedef) && ld->validcount != validcount)
									{
										ld->validcount = validcount;

										if (  !hitline[ld - lines]
											&& ld->backsector
											&& ld->frontsector != ld->backsector)
										{
											linearray[c++] = ld - lines;
										}
									}
								}
							}

							if (c > 2 && (rc = nf_chkLineListRing(c, NULL, 0, hitline)) == c)
							{
								valid = true;

								for (k = 0; k < hitssecptr; k++)
								{
									subsector_t *ss = &subsectors[hitssec[k]];

									Setr3dsslfsec(ss, ref[1] - sectors, 48176);
								}
							}
						}

						for (k = 0; k < s->linecount; k++)
						{
							ld = s->lines[k];

							if (hitline[ld - lines] == 4)
							{
								hitline[ld - lines] = 8;

								if (valid)
								{
								/*	if (!origTEX[ld - lines])
										Settype(ld, R3D_3F, 48177);
									else*/
										Settype(ld, R3D_36, 48178);

									SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_NOCHANGE, 48179);
								}
							}
						}
					}
					else
					{
						hitline[lnum] = 8;
					}
				}
			}
		}
	}


	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& sectors[i].linecount > 7
			&& !SecCtrl(i)
			&&  sectors[i].stptr < 0
			&& sectors[i].heightsec < 0
			&& (sectors[i].specFlags & 0x3041C) == 0x410
			&& sectors[i].floorpic != skyflatnum)
		{
			sector_t *s = &sectors[i], *chk;
			sector_t *ref[3] = { NULL, NULL, NULL};
			line_t	 *ld;
			int		 lnum, rc, box[4];
			short	 cpic = s->ceilingpic;

			// look for inner ring having a floor mov sector
			for (c = k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (lineflags[ld - lines]) continue;

				if (   ld->lnflags2 == 0x0C000000
					&& Vtxs2(ld))
				{
					chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

					if (!chk) continue;

					if ((chk->specFlags & 0x3041C) == 0x1001C)
					{
						if (origTEX[ld - lines]) { c = 0; break; }
						if (!ref[0]) ref[0] = chk; else if (ref[0] != chk) { c = 0; break; }

						linearray[c++] = ld - lines;
					}
				}
			}

			if (c < 3 || s->linecount - c < 3 || ref[0]->linecount - c < 3) continue;

			// It now gets complicated. This is a wonderful example of how abandoning
			// the visplane method (as, although just about good enough  for software
			// rendering, it's not up to the job for accurate OGL rendering) leads to
			// extraordinary complication.
			CLR_hitline();

			// 1) check the lines obtained form a ring
			for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

			if (!subfn_CheckForRing(hitline, s, linearray[0], &rc, NULL, NULL, NULL) || rc != c) continue;

			// 2) check the validity of sector ref
			for (k = 0; k < c; k++) hitline[linearray[k]] = 8;	// currently set to 4 by the ring check

			for (c = k = 0, lnum = -1; k < ref[0]->linecount; k++)
			{
				ld = ref[0]->lines[k];

				if (   !(ld->spflags & SPF_CONTROLLINE)
					&& !hitline[ld - lines])
				{
					if (origTEX[ld - lines] || ld->frontsector == ld->backsector || !ld->backsector) { c = 0; break; }

					chk = ld->frontsector == ref[0] ? ld->backsector : ld->frontsector;

					if (!ref[1])
					{
						if (   chk->specFlags & SF_MOVFLOOR
							|| !(chk->specFlags & SF_NOLOTEX)
							|| chk->floorheight != s->floorheight
							|| (chk->ceilingheight != s->ceilingheight
							&& !(cpic == skyflatnum && chk->ceilingpic == cpic)) )
						{
							c = 0; break;
						}

						ref[1] = chk;
					}
					else if (ref[1] != chk) { c = 0; break; }

					lnum = ld - lines; hitline[lnum] = 1; c++; 
				}
			}

			if (   lnum < 0
				|| c < 3
				|| !subfn_CheckForRing(hitline, ref[0], lnum, &rc, NULL, NULL, NULL)
				|| rc != c)
			{
				continue;
			}

			// 3) We now need to find an internal sector with a lower ceiling height

			for (c = k = 0, lnum = -1; k < ref[1]->linecount; k++)
			{
				ld = ref[1]->lines[k];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 8;
				}
				else
				{
					if (origTEX[ld - lines] || ld->frontsector == ld->backsector || !ld->backsector) { c = 0; break; }

					chk = ld->frontsector == ref[1] ? ld->backsector : ld->frontsector;

					if (!ref[2])
					{
						if (   chk->specFlags & SF_MOVFLOOR
							|| !(chk->specFlags & SF_NOLOTEX)
							|| chk->ceilingpic == skyflatnum
							|| chk->floorheight != s->floorheight
							|| s->ceilingheight - chk->ceilingheight < 8*FRACUNIT)
						{
							c = 0; break;
						}

						ref[2] = chk;
					}
					else if (ref[2] != chk) { c = 0; break; }

					lnum = ld - lines; hitline[lnum] = 1; c++; 
				}
			}

			if (!c) continue;

			// If we get here then we're on to something.
			// Now's the time to check the containing ring and get all the subsecs inside.
			// This is left to last as the preceding routimes don't take too long and will
			// early out if non-compliant.

			for (c = k = 0, lnum = -1; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if (!hitline[ld - lines])
				{
					lnum = ld - lines; hitline[lnum] = 1; c++;
				}
			}

			if (   c < 3
				|| !subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
				|| rc < c
				|| !LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE)
				|| !checkLinesInBox(&c))
			{
				continue;	// shame..
			}

			for (k = 0; k < hitssecptr; k++)
			{
				subsector_t *ss = &subsectors[hitssec[k]];

				if (ss->sector != ref[0])
				{
					Setr3dssfs(ss, ref[0] - sectors, 48180);
				}

				Setr3dsslfsec(ss, s - sectors, 48181);

				if (ss->sector != ref[2])
				{
					Setr3dsslcsec(ss, ref[2] - sectors, 48182);
				}
			}

			for (k = 0; k < c; k++)
			{
				ld = &lines[linearray[k]];

				if (!origTEX[linearray[k]]) Settype(ld, R3D_3F, 48183);
				SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48184);
			}

			Setsecdone(s, 48185);

			for (k = 0; k < 3; k++) Setsecdone(ref[k], 48186);
		}
	}
}

// v220-10
void MX_InnerMovLoopContained()
{
	int i, c;

	for (c = i = 0; i < numlines; i++)
	{
		if (  !lineflags[i]
			&& lines[i].specType == R3D_00
			&& lines[i].ml_flags & ML_TWOSIDED
			&& lines[i].frontsector != lines[i].backsector
			&& (lines[i].mtflags == R3D_MISLTEX1 || lines[i].mtflags == R3D_MISLTEX2))
		{
			if (   !(lines[i].frontsector->specFlags & SF_NOLOTEX)
				&& !(lines[i].backsector->specFlags  & SF_NOLOTEX))
			{
				linearray[c++] = i;
			}
		}
	}

	if (c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	for (i = 0; i < c; i++)
	{
		if (hitline[linearray[i]] == 1)
		{
			line_t	 *ld = &lines[linearray[i]];
			sector_t *sc = ld->mtflags == 0x10 ? ld->backsector  : ld->frontsector;
			sector_t *os = ld->mtflags == 0x10 ? ld->frontsector : ld->backsector;
			int		 k, m, n, rc;

			if (subfn_CheckForRing(hitline, sc, linearray[i], &rc, NULL, NULL, NULL))
			{
				RDBOOL valid = true;

				ss_validcount++;
				hitssecptr = 0;

				for (k = 0; k < c; k++)
				{
					if (hitline[linearray[k]] == 4)
					{
						ssline_t *sl = &sslines[linearray[k]];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->ssvalidcount != ss_validcount)
							{
								seg_t *sg = &segs[ss->firstline];

								ss->ssvalidcount = ss_validcount;

								if (ss->sector != sc) continue;

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if (sg->frontsector && sg->frontsector != sc) { valid = false; goto end_check; }
								}

								hitssec[hitssecptr++] = sl->list[m];
							}
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
					Setr3dsslfsec(&subsectors[hitssec[k]], os - sectors, 48187);
end_check:;
				for (k = 0; k < c; k++)
				{
					if (hitline[linearray[k]] == 4)
					{
						hitline[linearray[k]] = 0;

						if (valid) SetLFG(&lines[linearray[k]], LFG_NEWSET|LFG_FIXED, 48188);
					}
				}
			}
			else
			{
				hitline[linearray[i]] = 0;
			}
		}
	}
}

// v220-12
void MX_CheckLoMissingLoops()
{
	sector_t	*fs, *bs, *ref;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, k, fht;

//	1) Deal with nohi/lotex sectors that may or may not have R3D_16 set
	for (i = 0; i < numsectors; i++)
	{
		if (   secdone[i]
			|| (sectors[i].specFlags & 0x1000C) != 0x4)
		{
			continue;
		}

		CLR_hitline();
		ref = &sectors[i];
		fht = ref->floorheight + 8*FRACUNIT;

		for (c = k = 0; k < ref->linecount; k++)
		{
			ld = ref->lines[k];

			if (   !lineflags[ld - lines]
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2)
				&& Vtfl3(ld)
				&& (fs = ld->frontsector)->heightsec < 0
				&& (bs = ld->backsector)->heightsec  < 0)
			{
				hitline[ld - lines] = 1; c++;
			}
		}

		if (c < 3) continue;

		while(1)
		{
			int lnum = -1;

			for (k = 0; k < ref->linecount; k++)
			{
				ld = ref->lines[k];

				if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, ref, lnum, &rc, box, NULL, NULL))
			{
				sector_t *chk = NULL;	// verify type
			
				for (c = k = 0; k < ref->linecount; k++)
				{
					ld = ref->lines[k];
					
					if (hitline[ld - lines] == 4)
					{
						fs = ld->frontsector == ref ? ld->backsector : ld->frontsector;

						if (!chk)
						{
							if (  !fs
								|| fs->floorheight <= fht
								|| fs->specFlags & SF_MOVFLOOR
								|| (fs->specFlags & SF_NOLOTEX && !(fs->specFlags & SF_SSLINETRUE)) )
							{
								c = 1; break;
							}

							chk = fs;
						}
						else if (fs != chk)
						{
							c = 1; break;
						}
					}
				}

				if (  !c
					&& LS_MarkSubsecsInAreaGeneral(ref, hitline, 0, NULL, box, -1, MSIA_NONE)
					&& LS_OnlySSLinesInSSECarray(ref, hitline))
				{
					// v220-21 Add ceiling test
					RDBOOL	 needsceil = (ref->specFlags & SF_MOVCEIL) > 0;

					for (k = 0; k < ref->linecount; k++)
					{
						ld = ref->lines[k];

						if (hitline[ld - lines] == 4)
						{
							SetLFG(ld, LFG_NEWSET, 48189);

							if (needsceil)
							{
								fs = ld->frontsector == ref ? ld->backsector : ld->frontsector;

								if (   !fs
									|| fs->ceilingpic == skyflatnum
									|| fs->specFlags & SF_MOVCEIL
									|| fs->ceilingheight - ref->ceilingheight <= 8*FRACUNIT)
								{
									needsceil = false;
								}
								else
								{
									if (   (fs == ld->frontsector && origTEX[ld - lines] & 0x40)
										|| (fs == ld->backsector  && origTEX[ld - lines] & 0x04))
									{
										needsceil = false;
									}
								}
							}
						}
					}

					for (k = 0; k < hitssecptr; k++)
					{
						Setr3dssfs(&subsectors[hitssec[k]], chk - sectors, 48190);

						if (needsceil)
							Setr3dsslcsec(&subsectors[hitssec[k]], chk - sectors, 48191);
					}
				}

				for (k = 0; k < ref->linecount; k++)
				{
					ld = ref->lines[k];
					
					if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
				}
			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}
}

// =======================================================

// v220-13 Check connected no lotex floor mov sectors
static int testLines_lfg(sector_t *s, int *lcnt, RDBOOL first)
{
	line_t	*ld;
	int		c, i;

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (lineflags[ld - lines])
		{
			if (first) return 0;
			continue;
		}

		if (ld->mtflags & R3D_MISLOTEX && Vtfl3(ld))
		{
			if (lcnt) linearray[(*lcnt)++] = ld - lines;
			hitline[ld - lines] = 1;
			c++;
		}
	}

	return c;
}
static int setSref(sector_t *s, int *mem)
{
	line_t		*ld;
	int			c, i, k, m, n;

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->ml_flags & ML_TWOSIDED)
		{
			for (k = 0; k < 2; k++)
			{
				n = k ? ld->backsector - sectors : ld->frontsector - sectors;

				for (m = 0; m < c; m++)
				{
					if (mem[m] == n) goto next;
				}

				mem[c++] = n;
next:;
			}
		}
	}

	return c;
}

// v220-21 rewritten
void MX_CheckLoMissingFloorMovLoops()
{
	sector_t	*sc;
	int			c, i;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (  !secdone[i]
			&& (sc->specFlags & 0x3001F) == 0x14
			&& !SecCtrl(i)
			&& sc->linecount > 5
			&& (c = testLines_lfg(sc, NULL, true)) > 1)
		{
			vertexowner_t	*vo;
			sector_t		*ch, *os;
			line_t			*ld;
			int				k, m, n, v;

			sector_t	*bs[16];	// bounding sectors
			int			bcnt = 0;

			int			box[4];
			int			lcnt_mem;
			int			lcnt = 0;
			int			seccnt = 0;

			// get all other sectors referenced
			int *sref = (int*)malloc((sc->linecount + 1) * sizeof(int));	// max possible
			int scnt = setSref(sc, sref);

			// ---------------------------------

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& !(origTEX[ld - lines] & 0x11)
					&& abs(ld->frontsector->floorheight - ld->backsector->floorheight) > 8*FRACUNIT)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (ch == sc) continue;
					if (!(ch->specFlags & SF_NOLOTEX)) continue;

					linearray[lcnt++] = ld - lines;

					// Store the sector's index
					for (m = 0; m < seccnt; m++)
					{
						if (countsec[m] == ch - sectors) goto end_seccnt;
					}

					countsec[seccnt++] = ch - sectors;
end_seccnt:;

					// Find vertex owner sector that is not referenced by the sector's lines
					for (v = 0; v < 2; v++)
					{
						vo = v ? &vertexowners[ld->v2 - vertexes] : &vertexowners[ld->v1 - vertexes];

						for (m = 0; m < vo->num; m++)
						{
							for (n = 0; n < scnt; n++)
							{
								if (sref[n] == vo->list[m]) goto next;
							}

							os = &sectors[vo->list[m]];

							if (os->floorpic != sc->floorpic) continue;	// low floor sec to be set
							if (os->floorheight != sc->floorheight) continue;
							// Note: bounding sector expected to have a wall
							if (os->specFlags & (SF_NOLOTEX|SF_MOVFLOOR)) continue;

							for (n = 0; n < bcnt; n++)
							{
								if (bs[n] == os) goto next;
							}

							if (bcnt < 16) bs[bcnt++] = os;
next:;
						}
					}
				}
			}

			free(sref);

			if (!bcnt) continue;

			for (k = 0; k < bcnt; k++)
			{
				countsec[seccnt++] = bs[k] - sectors;

				for (lcnt_mem = lcnt, m = 0; m < bs[k]->linecount; m++)
				{
					ld = bs[k]->lines[m];

					if (   origTEX[ld - lines] & 0x11
						&& (ch = ld->frontsector == bs[k] ? ld->backsector : ld->frontsector)
						&& ch->floorheight > bs[k]->floorheight)
					{
						linearray[lcnt++] = ld - lines;
					}
				}
			}

			if (   lcnt > lcnt_mem
				&& (lcnt_mem = nf_chkLineListRing(lcnt, box, 0, NULL)) > 3
				&& LS_MarkSubsecsInMultiSecLoop(seccnt, countsec, lcnt_mem, box, NULL, false, MSIA_NONE))
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef) && !lineflags[ld - lines])
						{
							if (ld->mtflags & R3D_MISLOTEX)
							{
								Clr_mtloflags(ld, 48196);
								Settype(ld, R3D_16, 48197);
							}

							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48198);
						}
					}

					Setr3dsslfsec(ss, i, 48199);
				}

				for (k = 0; k < seccnt; k++)
				{
					Setsecdone(&sectors[countsec[k]], 48200);
				}
			}
		}
	}
}

// =======================================================

// v220-13
static void subfn_CheckMovCeilInside(sector_t *sc)
{
	subsector_t	*ss;
	seg_t		*sg;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, k, m;
	int 		*mrkline;
	RDBOOL		no_outer;

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (!(ld->ml_flags & ML_TWOSIDED) ) return;
		if ((ld->lnflags2 & 0x0C000000) != 0x0C000000) return;
		if (lineflags[ld - lines]) return;

		hitline[ld - lines] = 1; linearray[c++] = ld - lines;
	}

	if (c < 3) return;

	mrkline = (int*)M_Calloc(maxsecsz * sizeof(int), 0);

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{ lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
		{
			sector_t	*osec, *outer, *outerc, *chk;
			int			mline = 0;

			for (osec = NULL, i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					ld = &lines[linearray[i]];

					chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!chk) continue;

					if (!osec)
					{
						osec = chk;

						if (  (chk->specFlags & 0xF0004) != 0x4)
						{ osec = NULL; goto end_check; }
					}
					else if (osec != chk)
					{ osec = NULL; goto end_check; }
				}
			}

			if (!osec) goto end_check;

			validcount++;
			ss_validcount++;
			hitssecptr = 0;

			for (outer = outerc = NULL, i = no_outer = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					ssline_t *sl = &sslines[linearray[i]];

					for (k = 0; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						hitssec[hitssecptr++] = sl->list[k];

						if (ss->sector != sc)
						{
							if (osec != ss->sector)
							{ osec = NULL; goto end_check; }

							sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if (sg->frontsector != osec)
								{ osec = NULL; goto end_check; }

								if (   !no_outer
									&& sg->backsector
									&& sg->backsector != sc
									&& sg->backsector != osec
									&& sg->backsector != outerc)
								{
									if (!outerc) outerc = sg->backsector;
									else if (outerc != sg->backsector)
									{
										outerc = NULL; no_outer = true;
									}
								}

								if (  (ld = sg->linedef)
									&& ld->validcount != validcount
									&& !hitline[ld - lines])
								{
									ld->validcount = validcount;
									chk = ld->frontsector == osec ? ld->backsector : ld->frontsector;

									if ( !(ld->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2))
										|| ld->mtflags & (R3D_MISHTEX1|R3D_MISHTEX2))
									{ osec = NULL; goto end_check; }

									if (chk)
									{
										if (!outer)
										{
											outer = chk;

											if (  (chk->specFlags & 0xF0004 && !(chk->specFlags & SF_SSLINETRUE))
												|| chk->floorheight <= osec->floorheight)
											{ osec = NULL; goto end_check; }
										}
										else if (outer != chk)
										{ osec = NULL; goto end_check; }
									}

									mrkline[mline++] = ld - lines;
								}
							}
						}
					}
				}
			}

			if (osec && mline > 2)
			{
				if (outerc)
				{
					for (i = 0; i < hitssecptr; i++)
					{
						Setr3dsslfsec(&subsectors[hitssec[i]], outerc - sectors, 48201);
					}
				}

				for (i = 0; i < mline; i++)
				{
					ld = &lines[mrkline[i]];
					if (!outerc) Settype(ld, R3D_16, 48202);
					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48203);
				}

				Setsecdone(sc  , 48204);
				Setsecdone(osec, 48205);
			}

end_check:;
			for (i = 0; i < mline; i++) mrkline[i] = 0;

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4) hitline[linearray[i]] = 8;
			}
		}
		else
		{
			hitline[lnum] = 8;
		}
	}

	M_Free(mrkline);
}

// v220-13
void MX_CheckMovCeilInside()
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0xF000C) == 0x20004)
		{
			subfn_CheckMovCeilInside(&sectors[i]);
		}
	}
}

//=======================================================

// v220-13
static void subfn_CheckMovFloorInside(sector_t *sc)
{
	subsector_t	*ss;
	seg_t		*sg;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, k, m;
	int 		*mrkline;
	RDBOOL		no_outer;

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (!(ld->mtflags & R3D_MISHITEX)) continue;
		if (!(origTEX[ld - lines] & 0x11)) continue;

		if (!(ld->ml_flags & ML_TWOSIDED) ) return;
		if ((ld->lnflags2 & 0x0C000000) != 0x0C000000) return;
		if (lineflags[ld - lines]) return;

		hitline[ld - lines] = 1; linearray[c++] = ld - lines;
	}

	if (c < 3) return;

	mrkline = (int*)M_Calloc(maxsecsz * sizeof(int), 0);

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{ lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
		{
			sector_t	*osec, *outer, *outerc, *chk;
			int			mline = 0;

			for (osec = NULL, i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					ld = &lines[linearray[i]];

					chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!chk) continue;

					if (!osec)
					{
						osec = chk;

						if (  (chk->specFlags & 0xF0008) != 0x8)
						{ osec = NULL; goto end_check; }
					}
					else if (osec != chk)
					{ osec = NULL; goto end_check; }
				}
			}

			if (!osec) goto end_check;

			validcount++;
			ss_validcount++;
			hitssecptr = 0;

			for (outer = outerc = NULL, i = no_outer = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					ssline_t *sl = &sslines[linearray[i]];

					for (k = 0; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						hitssec[hitssecptr++] = sl->list[k];

						if (ss->sector != sc)
						{
							if (osec != ss->sector)
							{ osec = NULL; goto end_check; }

							sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if (sg->frontsector != osec)
								{ osec = NULL; goto end_check; }

								if (   !no_outer
									&& sg->backsector
									&& sg->backsector != sc
									&& sg->backsector != osec
									&& sg->backsector != outerc)
								{
									if (!outerc) outerc = sg->backsector;
									else if (outerc != sg->backsector)
									{
										outerc = NULL; no_outer = true;
									}
								}

								if (  (ld = sg->linedef)
									&& ld->validcount != validcount
									&& !hitline[ld - lines])
								{
									ld->validcount = validcount;
									chk = ld->frontsector == osec ? ld->backsector : ld->frontsector;

									if ( !(ld->mtflags & (R3D_MISHTEX1|R3D_MISHTEX2)) )
									{ osec = NULL; goto end_check; }

									if (chk)
									{
										if (!outer)
										{
											outer = chk;

											if (chk->specFlags & 0xF0004 && !(chk->specFlags & SF_SSLINETRUE) )
											{ osec = NULL; goto end_check; }
										}
										else if (outer != chk)
										{ osec = NULL; goto end_check; }
									}

									mrkline[mline++] = ld - lines;
								}
							}
						}
					}
				}
			}

			if (outer && osec && outer->floorheight - osec->floorheight >= 8*FRACUNIT && mline > 2)
			{
				if (outerc)
				{
					for (i = 0; i < hitssecptr; i++)
					{
						Setr3dsslfsec(&subsectors[hitssec[i]], outerc - sectors, 48206);
					}
				}

				for (i = 0; i < mline; i++)
				{
					ld = &lines[mrkline[i]];
					if (!outerc) Settype(ld, R3D_16, 48207);
					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48208);
				}

				Setsecdone(sc  , 48209);
				Setsecdone(osec, 48210);
			}

end_check:;
			for (i = 0; i < mline; i++) mrkline[i] = 0;

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4) hitline[linearray[i]] = 8;
			}
		}
		else
		{
			hitline[lnum] = 8;
		}
	}

	M_Free(mrkline);
}

void MX_CheckMovFloorInside()	// v220-13 v220-19
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i]) continue;

		if ( (sectors[i].specFlags & 0xF000C) == 0x10008)
		{
			subfn_CheckMovFloorInside(&sectors[i]);
		}
		else
		if ( sectors[i].ceilingpic == skyflatnum && (sectors[i].specFlags & 0x3001C) == 0x1C)
		{
			// check for lotex ring
			sector_t *ch, *os, *sc = &sectors[i];
			line_t	 *ld;
			int		 mask, side;
			int		 box1[4], box2[4];
			int		 c, d, k, m;

			M_ClearBox(box2);

			for (c = d = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (!(ld->ml_flags & ML_TWOSIDED)) { c = 0; break; }
				if (!(ld->mtflags)) { c = 0; break; }
				if (!(ld->lnflags2 & LNF2_OWNEQUALTOTWO)) { c = 0; break; }
				if (!Vtxs2(ld)) { c = 0; break; }
				if (ld->frontsector == ld->backsector) { c = 0; break; }
				if (lxarray[ld - lines]) { c = 0; break; }

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (ch->specFlags & SF_MOVCEIL) { c = 0; break; }

				mask = sc == ld->frontsector ? R3D_MISLTEX1 : R3D_MISLTEX2;
				side = sc == ld->frontsector ? 0 : 1;

				if (ld->mtflags == mask)
				{
					if (!os)
					{
						if (    ch->specFlags & 0x30000
							|| (ch->specFlags & 0x4 && !(ch->specFlags & 0x400)) )
						{
							c = 0; break;
						}

						os = ch;
					}
					else if (os != ch) { c = 0; break; }
					if (ch->specFlags & SF_MOVFLOOR) { c = 0; break; }
					linearray[c++] = ld - lines;
				}
				else
				{
					if (!sides[ld->sidenum[side]].bottomtexture) { c = 0; break; }
					if (!(ch->specFlags & SF_MOVFLOOR)) { c = 0; break; }
					M_AddLineToBox(box2, ld);
					d++;
				}
			}

			if (c < 3 || !d) continue;

			if (   subfn_CheckRingFromList(linearray, c, box1)
				&& M_CompareBox(box2, box1)
				&& LS_MarkSubsecsInLoop(sc, c, NULL, box1, MSIA_NONE, NULL))
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (  (ld = sg->linedef)
							&& ld->frontsector != sc
							&& ld->backsector != sc)
						{
							goto next_check;
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];

					Setr3dsslfsec(ss, os - sectors, 48211);
				}

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (ld->spflags & SPF_CONTROLLINE) continue;

					if (origTEX[ld - lines])
						Settype(ld, R3D_36, 48212);
					else
						Settype(ld, R3D_3F, 48213);

					SetLFG(ld, LFG_NEWSET|LFG_FIXABS, 48214);
				}

				Setsecdone(sc, 48215);
			}
		}
next_check:;
	}
}

static RDBOOL subfn_onlySCinside(sector_t *sc, int lidx)
{
	subsector_t *ss;
	seg_t		*sg;
	ssline_t	*sl = &sslines[lidx];
	int			i, k;

	ss_validcount++;

	for (i = 0; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount == ss_validcount) continue;

		ss->ssvalidcount = ss_validcount;

		if (ss->sector == sc)
		{
			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->frontsector != sc) return false;
			}
		}
	}
	return true;
}

static void subfn_CheckMissingLoSkyFloors(sector_t *sc)
{
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			a, c, i;

	for (a = c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if ((fs = ld->frontsector) == (bs = ld->backsector))
			return;

		if (lineflags[ld - lines]) return;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			chk = fs == sc ? bs : fs;

			if (chk->floorpic == skyflatnum) return;

			if (chk->floorheight - sc->floorheight > 8*FRACUNIT)
			{
				a++;
			}
			else
				return;
		}

		linearray[c++] = ld - lines;
	}

	if (a < 2 || c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		int rc, lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{ lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			RDBOOL	valid = true;

			for (i = 0; i < c; i++)
			{
				if (   hitline[linearray[i]] == 4
					&& !subfn_onlySCinside(sc, linearray[i]))
				{
					valid = false; break;
				}
			}

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 0;

					ld = &lines[linearray[i]];

					if (valid)
					{
						// v220-18 Test added for R3D_1B type
						if ( (valid = (ld->mtflags & R3D_MISTEXANY) != 0) )
						{
							fs = ld->frontsector, bs = ld->backsector;

							valid = !(fs->specFlags & SF_MOVSECTOR)
								&& !(bs->specFlags & SF_MOVSECTOR)
								&& abs(fs->ceilingheight - bs->ceilingheight) > 8*FRACUNIT;

							if (valid)
							{
								short cpic = -1;

								if (   fs->ceilingheight > bs->ceilingheight
									&& fs->specFlags & SF_NOHITEX
									&& !(fs->specFlags & SF_SSLINETRUE)
									&& (!(bs->specFlags & SF_NOHITEX) || (bs->specFlags & SF_SSLINETRUE)) )
								{
									cpic = bs->ceilingpic;
								}
								else
								if (   bs->ceilingheight > fs->ceilingheight
									&& bs->specFlags & SF_NOHITEX
									&& !(bs->specFlags & SF_SSLINETRUE)
									&& (!(fs->specFlags & SF_NOHITEX) || (fs->specFlags & SF_SSLINETRUE)) )
								{
									cpic = fs->ceilingpic;
								}

								if ( (valid = cpic >= 0 && cpic != skyflatnum) )
								{
									Settype(ld, R3D_1B, 48216);
								}
							}
						}
						
						if (!valid)
						{
							Settype(ld, R3D_14, 48217);
							valid = true;
						}

						SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48218);
					}
				}
			}

			if (valid)
			{
				Setsecdone(sc, 48219);
			}
		}
		else
			hitline[lnum] = 0;

	}
}


static void subfn_CheckMissingHiSkyCeilings(sector_t *sc)
{
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			a, c, i;

	for (a = c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if ((fs = ld->frontsector) == (bs = ld->backsector))
			return;

		if (lineflags[ld - lines]) return;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			chk = fs == sc ? bs : fs;

			if (chk->ceilingpic == skyflatnum) return;

			if (sc->ceilingheight - chk->ceilingheight > 8*FRACUNIT)
			{
				a++;
			}
			else
				return;
		}

		linearray[c++] = ld - lines;
	}

	if (a < 2 || c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		int rc, lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{ lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			RDBOOL	valid = true;

			for (i = 0; i < c; i++)
			{
				if (   hitline[linearray[i]] == 4
					&& !subfn_onlySCinside(sc, linearray[i]))
				{
					valid = false; break;
				}
			}

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 0;

					ld = &lines[linearray[i]];

					if (valid)
					{
						Settype(ld, R3D_16, 48220);
						SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48221);
					}
				}
			}

			if (valid)
			{
				Setsecdone(sc, 48222);
			}
		}
		else
			hitline[lnum] = 0;
	}
}

// v220-13
// Look for;
//	1) missing lo with a sky floor
//	2) missing hi with a sky ceiling
void MX_CheckMissingHiLoSkySectors()
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i])
		{
			if (  (sectors[i].specFlags & 0xF000C) == 0x4
				&& sectors[i].floorpic == skyflatnum)
			{
				subfn_CheckMissingLoSkyFloors(&sectors[i]);
			}
			else
			if (  (sectors[i].specFlags & 0xF000C) == 0x8
				&& sectors[i].ceilingpic == skyflatnum)
			{
				subfn_CheckMissingHiSkyCeilings(&sectors[i]);
			}
		}
	}
}

static void Subfn_NestedNotexSectors(sector_t *sc)
{
	sector_t *os = NULL, *osc = NULL;
	sector_t *cs;
	int		 i;

	for (i = 0; i < sc->linecount; i++)
	{
		line_t	*ld = sc->lines[i];

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| ld->frontsector == ld->backsector
			|| lineflags[ld - lines])
		{
			return;
		}

		cs = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (   !cs
			|| (cs->specFlags & 0xF001C) != 0xC
			|| cs->ceilingpic != skyflatnum)
		{
			return;
		}

		if (!os) os = cs;
	}

	if (os) sc->ceilingpic = skyflatnum;
}

// v220-14 Specifically look for a non-skyflat ceiling sector in a
//         skyflat ceiling sector with neither having line textures.
void MX_CheckNestedNotexSectors()
{
	int		 i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& sectors[i].heightsec < 0
			&& (sectors[i].specFlags & 0xF001C) == 0x1C
			&& sectors[i].ceilingpic != skyflatnum
			&& sectors[i].ceilingheight - sectors[i].floorheight == FRACUNIT)
		{
			Subfn_NestedNotexSectors(&sectors[i]);
		}
	}
}


// =============================================================

static RDBOOL checkEnclosedBy1s(sector_t *s, RDBOOL testouter)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k, m;
	int			rc;

	ss_validcount++;
	validcount++;

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];
		if (hitline[ld - lines] != 4) continue;
		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			if (ss->sector != s) continue;
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if ( (ld = sg->linedef) )
				{
					if (ld->backsector && hitline[ld - lines] != 4) return testouter ? false : true;

					if (!ld->backsector && ld->validcount != validcount)
					{
						ld->validcount = validcount;
						linearray[c++] = ld - lines;
					}
				}
			}
		}
	}

	if (c < 3) return false;
	
	CLR_conarray();
	for (i = 0; i < c; i++) conarray[linearray[i]] = 1;
	return subfn_CheckForRing(conarray, s, linearray[0], &rc, NULL, NULL, NULL) && rc == c;
}

void subfn_CheckMoveNoUpperTexSectors(sector_t *sc)
{
	line_t		*ld;
	RDBOOL		sshas1s;
	int			i, k, lnum, rc;

	while (1)
	{
		lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) && rc > 2)
		{
			sector_t	*chk, *os = NULL;
			int			a, b;

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (hitline[ld - lines] == 4)
				{
					if (!subfn_onlySCinside(sc, ld - lines))
					{
						os = NULL; break;
					}

					chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!os) os = chk;
					else if (os != chk)
					{
						if (   os->tag != chk->tag
							|| os->floorpic != chk->floorpic
							|| os->ceilingpic != chk->ceilingpic
							|| os->floorheight != chk->floorheight
							|| os->ceilingheight != chk->ceilingheight
							|| (os->specFlags & SF_MOVSECTOR) != (chk->specFlags & SF_MOVSECTOR))
						{
							os = NULL; break;
						}

						if ( (os->specFlags & 0xC) == 0xC) os = chk;
					}
				}
			}

			if (os && (os->specFlags & 0xC) == 0xC) os = NULL;

			if (os)
			{
				// v220-19 revised
				if ( !(sshas1s = checkEnclosedBy1s(os, true)) )
				{
					// If not surrounded by 1's lines check that all lines inside belong
					// to the ring. Double ring types are dealt with elsewhere.
					if (checkEnclosedBy1s(sc, false)) os = NULL;
				}
			}

			for (a = b = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 0; a++;

					if (os)
					{
						if (sshas1s)
						{
							Settype(ld, R3D_36, 48223);
						}
						else
						{
							Settype(ld, R3D_16, 48224);

							if (origTEX[ld - lines] & 0x11)
								linearray[b++] = ld - lines;
						}

						SetLFG(ld, LFG_NEWSET, 48225);
					}
				}
			}

			// v220-20 add test
			if (   a
				&& a == b
				&& (sc->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR
				&& !(os->specFlags & SF_MOVSECTOR)
				&& sc->ceilingheight - os->ceilingheight > 8*FRACUNIT
				&& os->ceilingpic != skyflatnum)
			{
				subsector_t	*ss;
				ssline_t	*sl;
				seg_t		*sg;
				int			m, n;

				hitssecptr = 0;
				ss_validcount++;

				for (k = 0; k < b; k++)
				{
					sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						ss = &subsectors[sl->list[m]];

						if (ss->sector == os) continue;
						if (ss->sector != sc) { hitssecptr = 0; break; }
						if (ss->ssvalidcount == ss_validcount) continue;
						ss->ssvalidcount = ss_validcount;
						hitssec[hitssecptr++] = sl->list[m];

						sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if (sg->frontsector != sc) { hitssecptr = 0; break; }
						}

						if (!hitssecptr) break;
					}
				}

				if (hitssecptr)
				{
					for (k = 0; k < b; k++)
						Settype(&lines[linearray[k]], R3D_36, 48226);

					for (k = 0; k < hitssecptr; k++)
						Setr3dsslcsec(&subsectors[hitssec[k]], os - sectors, 48227);
				}
			}
		}
		else
			hitline[lnum] = 0;
	}
}

// v220-16 Deal with mov sectors with missing upper textures
void MX_CheckMoveNoUpperTexSectors()
{
	sector_t	*sc, *os;
	line_t		*ld = lines;
	int			c, i;

	CLR_hitline();
	SET_countsec(0);

	for (c = i = 0; i < numlines; i++, ld++)
	{
		if (  !ld->backsector
			|| ld->spflags & SPF_CONTROLLINE
			|| lineflags[i]
			|| origTEX[i] & 0x44
			|| (sc = ld->frontsector) == (os = ld->backsector)
			|| abs(os->ceilingheight - sc->ceilingheight) <= 8*FRACUNIT
			|| !((os->specFlags & SF_MOVSECTOR) || (sc->specFlags & SF_MOVSECTOR))
			||  ((os->specFlags & SF_MOVSECTOR) && (sc->specFlags & SF_MOVSECTOR))
			|| os->heightsec >= 0
			|| sc->heightsec >= 0)
		{
			continue;
		}

		if (os->ceilingheight > sc->ceilingheight)
		{
			if (sc->ceilingpic == skyflatnum) continue;
			countsec[os - sectors]++; c++;
		}
		else
		{
			if (os->ceilingpic == skyflatnum) continue;
			countsec[sc - sectors]++; c++;
		}

		hitline[i] = 1;
	}

	if (c > 2)
	{
		for (i = 0; i < numsectors; i++)
		{
			if (countsec[i] > 2) subfn_CheckMoveNoUpperTexSectors(&sectors[i]);
		}
	}
}

// =======================================================

static RDBOOL subfn_connected(sector_t *sc, line_t *lt)
{
	vertex_t	*v1 = lt->v1;
	vertex_t	*v2 = lt->v2;
	line_t		*ld;
	int			c, i;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld == lt || !hitline[ld - lines]) continue;

		if (ld->v1 == v1 || ld->v2 == v1) c++;
		else
		if (ld->v1 == v2 || ld->v2 == v2) c++;

		if (c == 2) return true;
	}

	return false;
}

static int subfn_chkR3D_16loop(sector_t *sc, line_t *li)
{
	ssline_t	*sl = &sslines[li - lines];
	line_t		*ld;
	int			i, k;
	RDBOOL		found = false;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->sector == sc)
		{
			seg_t *sg = &segs[ss->firstline];

			if (sg->frontsector != sc) return false;
			if (sg->frontsector == sg->backsector) continue;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef)
					&& ld != li
					&& hitline[ld - lines])
				{
					found = true;
				}
			}
		}
	}

	return found;
}

// v220-16
// Set lines with missing upper/lower texture to R3D_16 where necessary.
void MX_UnsetR3D_16()
{
	sector_t	*sc, *os, *rf;
	line_t		*ld;
	int			c, i, k;

	// missing lo
	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& !(sc->specFlags & 0x3000C)
			&& sc->linecount > 2
			&& sc->specFlags & SF_SECCLOSED)
		{
			MC_CheckNonContiguous(sc);
		}
	}

	SET_countsec(0);
	CLR_hitline();

	// missing hi
	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  (sc = ld->frontsector) == (os = ld->backsector)
			|| !os
			|| lineflags[i] & LFG_3DTYPE
			|| ld->spflags & SPF_CONTROLLINE
			|| origTEX[i] & 0x44
			|| os->specFlags & SF_MOVSECTOR
			|| sc->specFlags & SF_MOVSECTOR
			|| os->heightsec >= 0
			|| sc->heightsec >= 0)
		{
			continue;
		}
		
		// Added v220-16
		if (   !origTEX[i]
			&& !lineflags[i]
			&& ld->mtflags & R3D_MISLOTEX
			&& !(ld->mtflags & R3D_MISHITEX)
			&& Vtfl3(ld)
			&& ld->specType == R3D_00
			&& ld->lnflags2 == 0x54000000)
		{
			if (subfn_FindLineConnected(ld, 35)) continue;


			Settype(ld, R3D_16, 48228);
			continue;
		}
		
		// Added v220-33
		if (   ld->specType == R3D_00
			&& !(origTEX[i] & 0x44)
			&& !lineflags[i]
			&& ld->mtflags & R3D_MISHITEX
			&& !(ld->mtflags & R3D_MISLOTEX)
			&& Vtfl3(ld)
			&& ld->lnflags2 == 0x0C000000)
		{
			if (subfn_FindLineConnected(ld, 36)) 	continue;

			Settype(ld, R3D_16, 48228);
			continue;
		}

		if (abs(os->ceilingheight - sc->ceilingheight) <= 8*FRACUNIT) continue;

		if (sc->ceilingheight > os->ceilingheight) rf = os, os = sc; else rf = sc;

		if (ld->specType != R3D_00)
		{
			if (   rf->ceilingpic != skyflatnum
				&& !(rf->specFlags & 0xC) )
			{
				hitline[i] = 2;	// line connected reference only
			}
			continue;
		}

		if (   lineflags[i]
			|| secdone[rf - sectors]
			|| secdone[os - sectors])
		{
			continue;
		}

		if (   vtxflags[ld->v1 - vertexes] < 2
			|| vtxflags[ld->v2 - vertexes] < 2)
		{
			if (   (vtxflags[ld->v1 - vertexes] == 1 && vtoflags[ld->v1 - vertexes])
				&& (vtxflags[ld->v2 - vertexes] == 1 && vtoflags[ld->v2 - vertexes]))
			{
				if (os->specFlags & SF_NOHITEX)
				{
					if (   !origTEX[i]
						&& os->ceilingheight > rf->ceilingheight
						&& os->floorheight   < rf->floorheight
						&& os->ceilingpic == skyflatnum
						&& os->floorpic == skyflatnum)
					{
						Settype(ld, R3D_3F, 48229);
						SetSkywall(ld, 48230);
						SetLFG(ld, LFG_NEWSET, 48231);
					}
				}
			}
			continue;
		}

		if (   rf->ceilingpic != skyflatnum
			&& !(rf->specFlags & 0xC) )
		{
			hitline[i] = 1; c++;
			countsec[os - sectors]++;
		}
	}

	if (!c) return;

	for (i = 0; i < numsectors; i++)
	{
		if (!countsec[i]) continue;

		sc = &sectors[i];

		for (c = k = 0; k < sc->linecount; k++)
		{
			ld = sectors[i].lines[k];

			if (hitline[ld - lines] == 1)
			{
				if (!subfn_connected(sc, ld))
				{
					hitline[ld - lines] = 0;
					continue;
				}

				linearray[c++] = ld - lines;
			}
		}

		for (k = 0; k < c; k++)
		{
			ld = &lines[linearray[k]];

			if (subfn_chkR3D_16loop(sc, ld))
			{
				Settype(ld, R3D_16, 48234);
				SetLFG(ld, LFG_NEWSET, 48235);
			}
		}
	}
}

// =======================================================

// returns mov ceil sector if conditiona are met
static sector_t *linetypeMOVceil(line_t *ld)
{
	if ( !(origTEX[ld - lines] & 0x66) )
	{
		sector_t *fs, *bs;
		int fh = (fs = ld->frontsector)->ceilingheight;
		int bh = (bs = ld->backsector )->ceilingheight;

		if (abs(fh - bh) <= 8*FRACUNIT) return NULL;

		if (fh > bh)
		{
			if (    (bs->specFlags & SF_MOVSECTOR) == SF_MOVCEIL
				&& !(fs->specFlags & SF_MOVSECTOR) )
				return bs;

			return NULL;
		}

		if (    (fs->specFlags & SF_MOVSECTOR) == SF_MOVCEIL
			&& !(bs->specFlags & SF_MOVSECTOR) )
			return fs;
	}

	return NULL;
}

// v220-16 Deal with straightforward ring loops first
void MX_CheckNoTexRings()
{
	subsector_t	*ss;
	sector_t	*sec, *fs, *bs, *chk, *s;
	line_t		*ld;
	RDBOOL		secset;
	int			c, i, k;
	//---------------------

	SET_countsec(0);
	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   lineflags[i]
			|| !ld->backsector
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| fs->heightsec >= 0
			|| bs->heightsec >= 0
			|| fs->ceilingpic == skyflatnum
			|| bs->ceilingpic == skyflatnum
			|| !Vtxs2(ld)
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO))
		{
			continue;
		}

		if (origTEX[i])
		{
			if ( !(s = linetypeMOVceil(ld)) ) continue;
		}
		else
		{
			int afdf = abs(fs->floorheight - bs->floorheight);

			if (afdf <= 8*FRACUNIT && abs(fs->ceilingheight - bs->ceilingheight) <= 8*FRACUNIT)
			{
				continue;
			}

			if (fs->ceilingheight == bs->ceilingheight && afdf < 24*FRACUNIT)
			{
				if (   (fs->specFlags & 0xFFFFC) == 0x9001C
					|| (bs->specFlags & 0xFFFFC) == 0x9001C)
				{
					continue;
				}
			}

			if ( (fs->specFlags & 0x1C) == 0x1C)
			{
				s = fs; chk = bs;
			}
			else if ( (bs->specFlags & 0x1C) == 0x1C)
			{
				s = bs; chk = fs;
			}
			else
			{
				continue;
			}

			if ( (chk->specFlags & 0xC) == 0xC) continue;
			if (s->ceilingheight > chk->ceilingheight) continue;
		}


		hitline[ld - lines] = 1; c++;
		countsec[s - sectors]++;
	}

	if (c < 3) return;

	for (i = 0; i < numsectors; i++)
	{
		int setR3D_3F = 0;

		if (countsec[i] < 2) continue;

		sec = &sectors[i];
		secset = false;

		while(1)
		{
			RDBOOL	movceil = false;
			int		lnum = -1;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, sec, lnum, NULL, NULL, NULL, NULL))
			{
				for (k = 0, s = chk = NULL; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

						if (!s)
						{
							if (   chk->floorheight == sec->floorheight
								&& chk->ceilingheight == sec->ceilingheight)
							{
								break;
							}
							s = chk;
						}
						else if (chk != s) { s = NULL; break; }
					}
				}

				if (s && !SetAndCheckSSECSInLoop(sec, NULL)) s = NULL;

				if (s)
				{
					secset = true;
					movceil = linetypeMOVceil(&lines[lnum]) != NULL;

					for (k = 0; k < hitssecptr; k++)
					{
						ss = &subsectors[hitssec[k]];

						ss->ssflags |= SSF_NOOCCLUDE;

						if (ss->sector->ceilingheight > s->ceilingheight)
						{
							Setr3dsscs(ss, s - sectors, 48236);
						}
						// v220-20 add test
						else if (ss->sector->ceilingheight < s->ceilingheight
							&& (ss->sector->specFlags & 0x30418) == 0x20018
							&& !(s->specFlags & 0x20000)
							&& (s->specFlags & 0x400 || !(s->specFlags & 0x8)) )
						{
							Setr3dsscs(ss, s - sectors, 48237);
						}

						if (!movceil)
						{
							int a;

							if (  (ss->sector->specFlags & 0x3001C) == 0x1C
								&& ((ss->sector->ceilingheight >= s->ceilingheight
								&& ss->sector->floorheight <= s->floorheight)
								|| ss->sector->ceilingheight - ss->sector->floorheight <= 4*FRACUNIT))
							{
								setR3D_3F++;
							}

							if (!setR3D_3F && s->specFlags & SF_MOVFLOOR)
							{
								Setr3dsslfsec(ss, s - sectors, 48238);
							}
							// v220-33 if both the sector has missing textures and a lighting special the same
							//         as the containing sector then adopt the containing sector (looks better)
							else if (setR3D_3F
								||  (ss->sector->lightlevel == s->lightlevel
								&& (a = (ss->sector->special & 0x3001f)) == (s->special & 0x3001f)
								&& (a == 2 || a == 3 || a == 8 || a == 12 || a == 13 || a == 17)) )
							{
								SetOverride(ss, s - sectors, 48239);
							}
							else
								Setr3dssfs(ss, s - sectors, 48240);
						}
					}

					if (setR3D_3F != hitssecptr) setR3D_3F = 0;
				}

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						if (s)
						{
							sector_t *os = ld->frontsector == s ? ld->backsector : ld->frontsector;

							if (   !setR3D_3F
								&& (movceil
								|| origTEX[ld - lines]
								// v220-22 add test
								|| (os && os->specFlags & 0x8 && os->ceilingheight != s->ceilingheight)) )
							{
								Settype(ld, R3D_00, 48241);
							}
							else
							{
								Settype(ld, R3D_3F, 48241);
							}

							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48242);
						}

						hitline[ld - lines] = 0;
					}
				}
			}
			else
				hitline[lnum] = 0;
		}

		if (secset) Setsecdone(sec, 48243);
	}
}

// ===========================================================

static void testTexDualRings(sector_t *sc)
{
	sector_t	*fs, *bs, *chk, *os;
	line_t		*ld;
	int			box[16][4];
	int			c, d, e, i, k, m;

	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if ((fs = ld->frontsector) == (bs = ld->backsector)) continue;
		if (lineflags[ld - lines] || origTEX[ld - lines]) return;

		if (   !bs
			|| (fs->ceilingheight != bs->ceilingheight && !(origTEX[ld - lines] & 0x44))
			|| !Vtxs2(ld)
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO))
		{
			return;
		}

		chk = fs == sc ? bs : fs;

		if (!os)
		{
			if (chk->specFlags & 0x3000C) return;
			if (chk->floorheight - sc->floorheight > 8*FRACUNIT) return;
			if (chk->ceilingpic == skyflatnum) return;
			os = chk;
		}
		else if (os != chk)
		{
			return;
		}
		
		linearray[c++] = ld - lines;
	}

	if (c < 6 || !os) return;

	CLR_hitline();

	for (d = i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		int	rc;
		int lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1) { lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box[d], NULL, NULL))
		{
			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4) hitline[linearray[i]] = 8;
			}
			if (++d > 15) return;
		}
		else
		{
			hitline[lnum] = 16;
		}
	}

	if (d < 2) return;

	for (e = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (hitline[ld - lines] == 8) linearray[e++] = ld - lines;
	}

	// check any one box fits in another
	for (i = 0; i < d - 1; i++)
	{
		for (k = 1; k < d; k++)
		{
			if (M_CompareBox(box[i], box[k]) || M_CompareBox(box[k], box[i]))
			{
				for (m = 0; m < e; m++)
				{
					ld = &lines[linearray[m]];

					if ( M_FitBox(ld->bbox, box[i]) || M_FitBox(ld->bbox, box[k]) )
					{
						hitline[linearray[m]] = 4;
					}
				}

				if (SetAndCheckSSECSInLoop(sc, os))
				{
					int n;

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];
						seg_t		*sg = &segs[ss->firstline];

						Setr3dsslfsec(ss, sc - sectors, 48244);

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if ( (ld = sg->linedef) && !lineflags[ld - lines])
							{
								if (!origTEX[ld - lines]) Settype(ld, R3D_3F, 48245);
								SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48246);
							}
						}
					}

					Setsecdone(sc, 48247);
					Setsecdone(os, 48248);
				}
			}
		}
	}
}

// v220-16 Deal with dual ring loops before calling MX_CheckNoTexRings()
void MX_CheckNoTexDualRings()
{
	int			i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x20014) == 0x14)
		{
			testTexDualRings(&sectors[i]);
		}
	}
}

// ====================================================

static RDBOOL checkSec(int secnum)
{
	int i;

	for (i = 0; i < osx.scount; i++)
	{
		if (countsec[i] == secnum) return true;
	}

	return false;
}

static int errnumber(sector_t *s, int num)
{
	return false;
}

static RDBOOL secReferenced(sector_t *s)
{
	sector_t	*ch;
	line_t		*ld;
	int			box[2][4];
	int			i, secnum;
	int			invsec = -1;
	int			sec1 = -1;
	int			osec;
	byte		txn;

	osx.scount = osx.lcount = 0;
	M_ClearBox(box[0]);
	M_ClearBox(box[1]);

	for (i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (!ld->backsector) return errnumber(s, 0);
		if (ld->frontsector == ld->backsector) continue;

		if (   lineflags[ld - lines]
			|| !Vtxs2(ld)
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO))
		{
			return errnumber(s, 1);
		}

		ch = ld->frontsector == s ? ld->backsector : ld->frontsector;

		if (   ch->specFlags & SF_MOVCEIL
			|| ch->heightsec >= 0
			|| ((txn = origTEX[ld - lines]) & 0x44) == 0x44
			|| !(txn & 0x11)
			|| (txn & 0x04 && ld->backsector->ceilingheight > ld->frontsector->ceilingheight)
			|| (txn & 0x40 && ld->backsector->ceilingheight < ld->frontsector->ceilingheight))
		{
			return errnumber(s, 2);
		}

		secnum = ch - sectors;

		if (txn & 0x44)
		{
			// compensate for invalid textures
			if (invsec < 0) invsec = secnum;
			else if (invsec != secnum) return errnumber(s, 3);
		}

		if (!checkSec(secnum)) countsec[osx.scount++] = secnum;

		if (sec1 < 0) sec1 = secnum;

		M_AddLineToBox(box[ (sec1 == secnum) ], ld);
		linearray[osx.lcount++] = ld - lines;
	}
	
	if (osx.scount != 2 || osx.lcount < 6) return errnumber(s, 4);

	// Get the outer sector num
	osec = M_CompareBox(box[0], box[1]) ? sec1 : countsec[0] == sec1 ? countsec[1] : countsec[0];
	
	if (sectors[osec].ceilingpic == skyflatnum) return errnumber(s, 5);
	if ((sectors[osec].specFlags & 0x408) == 0x8) return errnumber(s, 6);

	if (invsec >= 0)
	{
		if (invsec == osec) return errnumber(s, 7);
	}
	else
	{
		if ( !(s->specFlags && 0x8) ) return errnumber(s, 8);
	}

	return U_checkDoubleLoop(s, 1);
}

void MX_CheckLowMoveInFalseCeiling()
{
	int	i, k;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& sectors[i].linecount > 5
			&& (sectors[i].specFlags & 0x30414) == 0x10010)
		{
			if (!secReferenced(&sectors[i])) continue;

			for (k = 0; k < hitssecptr; k++)
			{
				Setr3dsscs(&subsectors[hitssec[k]], osx.outer - sectors, 48249);
			}

			for (k = 0; k < osx.lcount; k++)
			{
				SetLFG(&lines[linearray[k]], LFG_NEWSET, 48250);
			}
			Setsecdone(&sectors[i], 48251);
		}
	}
}

// =====================================================

// v220-16
void MX_CheckCojoinedMissing()
{
	vertexowner_t	*v1, *v2;
	sector_t		*fs, *bs;
	int				c, i, k, m, n;

	for (i = 0; i < numlines; i++)
	{
		if (   (bs = lines[i].backsector)
			&& (fs = lines[i].frontsector) != bs
			&& origTEX[i] & 0x22
			&& !(origTEX[i] & 0x55)
			&& (v1 = &vertexowners[lines[i].v1 - vertexes])->num > 3 
			&& (v2 = &vertexowners[lines[i].v2 - vertexes])->num == v1->num
			&& fs->ceilingpic != skyflatnum
			&& bs->ceilingpic != skyflatnum
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& ((fs->ceilingheight <= fs->floorheight)
			||  (bs->ceilingheight <= bs->floorheight)) )
		{
			line_t	*ld;
			int		rc, box[4];

			for (c = k = 0, bs = NULL; k < v1->num; k++)
			{
				fs = &sectors[v1->list[k]];

				if ((fs->specFlags & 0x40C) == 0xC) continue;

				if (!bs)
				{
					bs = fs;
				}
				else if (bs != fs)
				{
					if (bs->floorheight != fs->floorheight) { bs = NULL; break; }
					if (bs->ceilingheight != fs->ceilingheight) { bs = NULL; break; }
				}

				if (!bs) break;

				for (m = 0; m < fs->linecount; m++)
				{
					ld = fs->lines[m];

					if (   ld->ml_flags & ML_TWOSIDED
						&& !origTEX[ld - lines])
					{
						linearray[c++] = ld - lines;
					}
				}
			}

			if (!bs || c < 3) continue;

			CLR_hitline();

			if ( (rc = nf_chkLineListRing(c, box, 0, hitline)) > 2)
			{
				ssline_t	*sl;

				validcount++;

				for (k = 0; k < v1->num; k++)
				{
					fs = &sectors[v1->list[k]];

					if ( (fs->specFlags & 0x40C) != 0xC ) continue;

					for (m = 0; m < fs->linecount; m++)
					{
						if ( (ld = fs->lines[m]) && ld->validcount == validcount) continue;

						ld->validcount = validcount;

						if (!M_FitBox(ld->bbox, box)) continue;

						if (origTEX[ld - lines] & 0x22)
							Settype(ld, R3D_10, 48252);
						else
							Settype(ld, R3D_3F, 48253);

						SetLFG(ld, LFG_NEWSET|LFG_FIXABS|LFG_FIXED, 48254);

						sl = &sslines[ld - lines];

						for (n = 0; n < sl->num; n++)
						{
							subsector_t *ss = &subsectors[sl->list[n]];

							if ((ss->sector->specFlags & 0x40C) != 0xC ) continue;
							if (ss->floorsec >= 0) continue;
							SetOverride(ss, bs - sectors, 48255);
						}
					}
				}
			}
		}
	}
}

// ======================================================

static RDBOOL outerRingValid(sector_t *sc)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k, m, rc;
	int			lnum = -1;

	validcount++;
	ss_validcount++;

	for (i = 0; i < hitssecptr; i++)
		subsectors[hitssec[i]].ssvalidcount = ss_validcount;

	CLR_excludeline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (hitline[ld - lines] == 4)
		{
			sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				if ( (ss = &subsectors[sl->list[k]])->ssvalidcount == ss_validcount)
					continue;

				ss->ssvalidcount = ss_validcount;
				hitssec[hitssecptr++] = sl->list[k];
				sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if ( (ld = sg->linedef) && ld->validcount != validcount)
					{
						ld->validcount = validcount;

						if (   !hitline[ld - lines]
							&& (ld->frontsector == sc || ld->backsector == sc))
						{
							lnum = ld - lines;
							excludeline[lnum] = 1;
							c++;
						}
					}
				}
			}
		}
	}

	return c > 2 && subfn_CheckForRing(excludeline, sc, lnum, &rc, NULL, NULL, NULL) && rc == c;
}

static void containedLoops(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*ld;
	byte		txn;
	int			rc, box[4];
	int			c, i;

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (hitline[ld - lines] == 1)
			{
				lnum = ld - lines; break;
			}
		}

		if (lnum < 0) return;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			for (i = 0, os = NULL; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!os) os = ch; else if (os != ch) { os = NULL; break; }
				}
			}

			if (os && !SecCtrl(sc - sectors) && !SecCtrl(os - sectors)) return;	// v220-28 test added

			if (   os
				&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_REVERSED)	// v220-22 $bugfix (6th parm was NULL)
				&& outerRingValid(sc) /* excludeline set in this function*/)
			{
				subsector_t	*ss;
				seg_t		*sg;
				int			k;

				validcount++;

				for (c = i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					sg = &segs[ss->firstline];

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						if ( (ld = sg->linedef) && ld->validcount != validcount)
						{
							ld->validcount = validcount;

							if (   ld->frontsector != ld->backsector
								&& excludeline[ld - lines] != 4
								&& !((ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2)))
							{
								return;
							}

							linearray[c++] = ld - lines;
						}
					}
				}

				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					if (ss->sector->floorheight < sc->floorheight)
						Setr3dssfs(ss, os - sectors, 48256);
					Setr3dsslfsec(ss, sc - sectors, 48257);
				}

				for (i = 0; i < c; i++)
				{
					int diff;

					ld = &lines[linearray[i]];
					txn = origTEX[ld - lines];

					if ( !(txn && 0x44)
						&& (diff = ld->backsector->ceilingheight - ld->frontsector->ceilingheight)
						&& abs(diff) == FRACUNIT
						&& !((diff > 0 && ld->backsector->ceilingpic == skyflatnum)
						||   (diff < 0 && ld->frontsector->ceilingpic == skyflatnum)) )
					{
						ld->spflags |= SPF_DRAWTOP;
					}

					if (!txn)
					{
						if (ld->spflags & SPF_DRAWTOP)
						{
							Settype(ld, R3D_36, 48258);
							SetMissinglo(ld, 48259);
						}
						else if (ss->sector == sc && os->floorheight > sc->floorheight)
						{
							Settype(ld, R3D_16, 48260);
							SetMissinglo(ld, 48261);
						}
						else
							Settype(ld, R3D_3F, 48262);
					}
					else
					if (   txn & 0x22
						&& ((ld->mtflags & R3D_MISLTEX1 && txn & 0x02 && txn & 0x04)
						 || (ld->mtflags & R3D_MISLTEX2 && txn & 0x20 && txn & 0x40)) )
					{
						Settype(ld, R3D_2D, 48263);
					}
					else
					{
						if (txn == 0x22)
							Settype(ld, R3D_03, 48264);
						else
							Settype(ld, R3D_36, 48265);
					}

					SetLFG(ld, LFG_FIXABS|LFG_NEWSET|LFG_NOCHANGE, 48266);
				}

				Setsecdone(sc, 48267);
				Setsecdone(os, 48268);
			}

			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

static void subfn_CheckContained(sector_t *sc)
{
	line_t		*ld;
	int			c, i;

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[ld - lines]) return;
		if (ld->specType != R3D_00 && ld->specType != R3D_2D) return;
		if (  (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2)
			&& Vtfl3(ld))
		{
			hitline[ld - lines] = 1;
			c++;
		}
	}

	if (c > 2)
	{
		containedLoops(sc);
	}
}

// v220-18
void MX_CheckContained()
{
	int			i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
		//	&& SecCtrl(i)	// test added v220-21 moved from here to containedLoops v220-28
			&& (sectors[i].specFlags & 0xFFFEC) == 0x4
			&& sectors[i].heightsec < 0)
		{
			subfn_CheckContained(&sectors[i]);
		}
	}
}

// ========================================================

static RDBOOL chkInternal(sector_t *sc, int *c)
{
	subsector_t	*ss;
	seg_t		*sg;
	line_t		*ld;
	int			i, k;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( !(ld = sg->linedef) || ld->validcount == validcount) continue;
			linearray[(*c)++] = ld - lines;
			ld->validcount = validcount;
			if (hitline[ld - lines] == 4) continue;
			if (origTEX[ld - lines]) return false;
			if (ld->frontsector != sc && ld->backsector != sc) return false;
		}
	}

	return true;
}

static void subfn_unclosedMissingContained(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*ld;
	int			box[4], rc;
	int			c, i;


	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (!(ld->ml_flags & ML_TWOSIDED)) continue;
		if (lineflags[ld - lines]) return;
		if (origTEX[ld - lines]) return;
		if (!Vtfl3(ld)) return;
		if (ld->mtflags != R3D_MISLTEX1 && ld->mtflags != R3D_MISLTEX2) return;

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;	
		if (!os) os = ch; else if (os != ch) { os = NULL; break; }
		linearray[c++] = ld - lines;
	}

	// The containing sector is not expected to have lower (valid) textures or too be closed
	if (c < 3 || !os || (os->specFlags & 0x30001C) != 0x4)
	{
		return;
	}

	CLR_hitline();

	for (i = 0; i < os->linecount; i++)
	{
		ld = os->lines[i];
		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (origTEX[ld - lines]) hitline[ld - lines] = 1;
	}

	while(1)
	{
		int lnum = -1, valid;

		for (i = 0; i < os->linecount; i++)
		{
			ld = os->lines[i];

			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		c = 0;

		if (subfn_CheckForRing(hitline, os, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			if (  (valid = LS_MarkSubsecsInAreaGeneral(os, hitline, 0, NULL, box, -1, MSIA_NONE))
				&& chkInternal(sc, &c) )
			{
				byte	txn;

				for (i = 0; i < c; i++)
				{
					ld = &lines[linearray[i]];
					txn = origTEX[ld - lines];

					if (!txn)
					{
						Settype(ld, R3D_3F, 48269);
					}
					else
					if (   txn & 0x22
						&& ((ld->mtflags & R3D_MISLTEX1 && txn & 0x02 && txn & 0x04)
						 || (ld->mtflags & R3D_MISLTEX2 && txn & 0x20 && txn & 0x40)) )
					{
						Settype(ld, R3D_2D, 48270);
					}
					else
					{
						Settype(ld, R3D_36, 48271);
					}

					SetLFG(ld, LFG_FIXABS|LFG_NEWSET, 48272);
				}

				Setsecdone(sc, 48273);
				Setsecdone(os, 48274);
			}

			for (i = 0; i < os->linecount; i++)
			{
				ld = os->lines[i];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			return;
		}
	}
}

// v220-18
void MX_CheckUnclosedMissingContained()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& SecCtrl(i)
			&& (sectors[i].specFlags & 0x3001C) == 0xC)
		{
			subfn_unclosedMissingContained(&sectors[i]);
		}
	}
}

// ===========================================

static RDBOOL reversedContained(sector_t *os, int c, int *linemem, int *box1, int *linec)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			i, k, m, ec;
	int			lnum, rc, box2[4];

	CLR_excludeline();
	ss_validcount++;
	validcount++;
	hitssecptr = 0;
	ec = 0;

	for (i = 0; i < c; i++)
	{
		if (hitline[linemem[i]] != 4) continue;

		sl = &sslines[linemem[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;
			hitssec[hitssecptr++] = sl->list[k];

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if ( !(ld = sg->linedef) || ld->validcount == validcount) continue;

				ld->validcount = validcount;
				linearray[(*linec)++] = ld - lines;

				if (hitline[ld - lines] != 4)
				{
					lnum = ld - lines;
					excludeline[lnum] = 1; ec++;
				}
			}
		}
	}

	return (subfn_CheckForRing(excludeline, os, lnum, &rc, box2, NULL, NULL) && rc == ec && M_FitBox(box1, box2));
}

// v220-18
void MX_ReversedMissingContained()
{
	sector_t	*ch, *os;
	line_t		*ld;
	int			c, i, k;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& SecCtrl(i)
			&& (sectors[i].specFlags & 0x3001C) == 0x1C)
		{
			sector_t	*sc = &sectors[i];
			int			*linemem = (int*)malloc(sc->linecount * sizeof(int));

			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;
				if (origTEX[ld - lines]) { c = 0; break; }
				if (ld->mtflags != R3D_MISLTEX1 && ld->mtflags != R3D_MISLTEX2) { c = 0; break; }

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!ch) { c = 0; break; }

				if (!os)
				{
					if (ch->specFlags & 0x3000C) break;
					os = ch;
				}
				else if (os != ch)
				{
					c = 0; break;
				}

				linemem[c++] = ld - lines;
			}

			if (c > 2 && os->floorheight < sc->floorheight)
			{
				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linemem[k]] = 1;

				while(1)
				{
					int lnum = -1, rc, box1[4];

					for (k = 0; k < c; k++)
					{
						if (hitline[linemem[k]] == 1)
						{
							lnum = linemem[k]; break;
						}
					}

					if (lnum < 0) break;

					if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box1, NULL, NULL)
						&& rc > 2)
					{
						int		linec = 0, valid;

						valid = reversedContained(os, c, linemem, box1, &linec);


						for (k = 0; k < c; k++)
						{
							if (hitline[linemem[k]] == 4) hitline[linemem[k]] = 0;
						}

						if (valid)
						{
							byte	txn;

							for (k = 0; k < hitssecptr; k++)
								Setr3dsslfsec(&subsectors[hitssec[k]], os - sectors, 48275);

							for (k = 0; k < linec; k++)
							{
								ld = &lines[linearray[k]];
								txn = origTEX[ld - lines];

								if (!txn)
								{
									Settype(ld, R3D_16, 48276);
									SetMissinglo(ld, 48277);
								}
								else
								if (   txn & 0x22
									&& ((ld->mtflags & R3D_MISLTEX1 && txn & 0x02 && txn & 0x04)
									 || (ld->mtflags & R3D_MISLTEX2 && txn & 0x20 && txn & 0x40)) )
								{
									Settype(ld, R3D_2D, 48278);
								}
								else
								{
									Settype(ld, R3D_36, 48279);
								}

								SetLFG(ld, LFG_FIXABS|LFG_NEWSET|LFG_NOCHANGE, 48280);
							}

							Setsecdone(sc, 48281);
							Setsecdone(os, 48282);
						}
					}
					else
						hitline[lnum] = 0;
				}
			}

			free(linemem);
		}
	}
}

// ===============================================================

static void subfn_CheckHighMissingContained(sector_t *sc, int c)
{
	line_t	*ld;
	int		i, rc;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) && rc > 2)
		{
			subsector_t	*ss;
			ssline_t	*sl;
			seg_t		*sg;
			sector_t	*os = NULL;
			int			k, m;

			ss_validcount++;
			validcount++;

			for (c = i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					for (k = 0, sl = &sslines[ld - lines]; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (ss->ssvalidcount == ss_validcount) continue;
						ss->ssvalidcount = ss_validcount;
						sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ((ld = sg->linedef) && ld->validcount != validcount)
							{
								ld->validcount = validcount;
								linearray[c++] = ld - lines;

								if (hitline[ld - lines] == 4) continue;
								if (!sg->backsector) return;
								if (!os) os = sg->backsector;
								else if (os != sg->backsector) return;
							}
						}
					}
				}
			}

			if (os && os->ceilingheight < sc->ceilingheight && !(os->specFlags & 0xC && !(os->specFlags & 0x400)) )
			{
				Setsecdone(sc, 48283);

				ss_validcount++;
				hitssecptr = 0;

				for (i = 0; i < c; i++)
				{
					ld = &lines[linearray[i]];

					for (k = 0, sl = &sslines[ld - lines]; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (ss->sector != os && ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;
							hitssec[hitssecptr++] = sl->list[k];
						}
					}

					if (!origTEX[ld - lines])
						Settype(ld, R3D_3F, 48284);
					else
					{
						Settype(ld, R3D_36, 48285);
						ld->ml_flags |= ML_IGNOREMISSINGHI;
					}

					SetLFG(ld, LFG_FIXABS|LFG_NEWSET|LFG_NOCHANGE, 48286);

					// Don't set secdone with os as this may contain other constructs
					if (!secdone[ld->frontsector - sectors] && ld->frontsector != os)
					{
						Setsecdone(ld->frontsector, 48287);
					}

					if (ld->backsector && !secdone[ld->backsector - sectors] && ld->backsector != os)
					{
						Setsecdone(ld->backsector, 48288);
					}
				}

				for (i = 0; i < hitssecptr; i++) Setr3dsscs(&subsectors[hitssec[i]], os - sectors, 48289);
			}

			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 0;
				}
			}
		}
		else
			hitline[lnum] = 0;
	}
}


void MX_CheckHighMissingContained()
{
	int i, k;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x3001C) == 0x1C
			&& sectors[i].stptr < 0
			&& sectors[i].heightsec < 0
			&& !SecCtrl(i))
		{
			sector_t	*ch, *sc = &sectors[i];
			line_t		*ld;
			int			c, lnum;

			for (c = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				lnum = ld - lines;

				if (   lineflags[lnum]
					|| origTEX[lnum] & 0x44
					|| !Vtfl3(ld))
				{
					c = 0; break;
				}

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (!ch) { c = 0; break; }
				if (ch == sc) continue;
				if ((ch->specFlags & 0x3001C) != 0x18) { c = 0; break; }
				if (ch->ceilingheight != sc->ceilingheight || ch->floorheight != sc->floorheight) { c = 0; break; }
				linearray[c++] = ld - lines;
			}

			if (c > 2) subfn_CheckHighMissingContained(sc, c);
		}
	}
}

// ===============================================================

void subfn_CheckPartialHiMissing(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*lb, *ld, ref;
	int			a, b, c, i, rc, box[4];

	// ------------------------------

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		// All lines in sector must form a closed loop
		if (!subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			hitline[lnum] = 0; continue;
		}
		
		if (rc < 4) goto tidyup;	// Min is re-entrant V

		// Put the missing tex lines into linearray
		// and check that the inheritable sector is valid
		M_ClearBox(box);

		for (c = i = 0, os = NULL; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (   hitline[ld - lines] == 4
				&& (ld->mtflags == R3D_MISHTEX1 || ld->mtflags == R3D_MISHTEX2))
			{
				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!os)
				{
					if (  !ch
						|| ch->specFlags & SF_MOVSECTOR
						|| ch->ceilingheight >= sc->ceilingheight
						|| ch->specFlags & 0x8 && !(ch->specFlags & 0x400))
					{
						goto tidyup;
					}

					os = ch;
				}
				else if (os != ch)
				{
					goto tidyup;
				}

				if (ld->mtflags == R3D_MISHTEX1)
					sidelist[c] = ld->sidenum[0];
				else
					sidelist[c] = ld->sidenum[1];

				linearray[c++] = ld - lines;

				M_AddLineToBox(box, ld);
			}
		}

		if (c < 2 || c > rc - 2) goto tidyup;

		// Check for re-entrancy
		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (!ld->mtflags && hitline[ld - lines] == 4)
			{
				if (   !M_ComparexyBox(ld->v1->x, ld->v1->y, box)
					&& !M_ComparexyBox(ld->v2->x, ld->v2->y, box))
				{
					goto tidyup;
				}
			}
		}

		// Start the process of finding a line bridging
		// the end vertexes of the missinghi line string
		CLR_vtxcheck();

		for (i = 0; i < c; i++)
		{
			ld = &lines[linearray[i]];

			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;
		}

		for (a = b = -1, i = 0; i < c; i++)
		{
			ld = &lines[linearray[i]];

			if (vtxcheck[ld->v1 - vertexes] == 1)
			{
				if (a < 0) a = ld->v1 - vertexes; else goto tidyup;
			}
			if (vtxcheck[lines[linearray[i]].v2 - vertexes] == 1)
			{
				if (b < 0) b = ld->v2 - vertexes; else goto tidyup;
			}
		}

		if (a < 0 || b < 0) goto tidyup;

		ref.specType = 0;
		ref.v1 = &vertexes[a];
		ref.v2 = &vertexes[b];
		ref.bbox[BOXLEFT ] = MIN_OF(ref.v1->x, ref.v2->x);
		ref.bbox[BOXRIGHT] = MAX_OF(ref.v1->x, ref.v2->x);
		ref.bbox[BOXBASE ] = MIN_OF(ref.v1->y, ref.v2->y);
		ref.bbox[BOXTOP  ] = MAX_OF(ref.v1->y, ref.v2->y);

		// Find bridging line
		if ( !(lb = subfn_FindVtxConnected(&ref, 3)) ) goto tidyup;

		// Note the opposite sidenum is used to sc's mishi lines
		// This is what makes this particular structure so awkward
		if (origTEX[lb - lines] == 0x02)
			sidelist[c] = lb->sidenum[1];
		else
			sidelist[c] = lb->sidenum[0];

		linearray[c++] = lb - lines;

		M_AddLineToBox(box, lb);

		// Get all subsectors in the loop
		if (LS_MarkSsecsUsingSidelist(c, box, MSIA_NONE))
		{
			subsector_t	*ss;
			seg_t		*sg;
			byte		txn;
			int			d, k;

			validcount++;

			for (c = d = i = 0; i < hitssecptr; i++)
			{
				ss = &subsectors[hitssec[i]];

				if (ss->sector->ceilingheight < os->ceilingheight)
				{
					goto tidyup;
				}

				addSec(ss->sector - sectors, &d);

				for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
				{
					if ( (ld = sg->linedef) && ld->validcount != validcount)
					{
						ld->validcount = validcount;
						linearray[c++] = ld - lines;
					}
				}
			}

			for (i = 0; i < hitssecptr; i++)
			{
				ss = &subsectors[hitssec[i]];

				if (ss->sector->ceilingheight > os->ceilingheight)
					Setr3dsscs(ss, os - sectors, 48290);
			}

			for (i = 0; i < c; i++)
			{
				ld = &lines[linearray[i]];
				txn = origTEX[ld - lines];

				if (!txn)
				{
					Settype(ld, R3D_3F, 48291);
				}
				else
				if (   txn & 0x22
					&& ((ld->mtflags & R3D_MISHTEX1 && txn & 0x02)
					 || (ld->mtflags & R3D_MISHTEX2 && txn & 0x20)) )
				{
					Settype(ld, R3D_2D, 48292);
				}
				else
				{
					Settype(ld, R3D_36, 48293);
				}

				SetLFG(ld, LFG_FIXABS|LFG_NEWSET, 48294);
			}

			for (i = 0; i < d; i++) Setsecdone(&sectors[countsec[i]], 48295);
		}
tidyup:;
		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
		}
	}
}

// Look for stuff that the visplane will accept where a sector has a mix
// of lines with high textures together with others having missing hitex.
void MX_CheckPartialHiMissing()
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& !SecCtrl(i)
			&& (sectors[i].specFlags & 0x3041C) == 0x14)
		{
			sector_t *sc = &sectors[i];
			line_t	 *ld;
			int		 c, d, k;

			CLR_hitline();

			for (c = d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				hitline[ld - lines] = 1;

				if (ld->mtflags)
				{
					if (ld->mtflags == R3D_MISHTEX1 || ld->mtflags == R3D_MISHTEX2)
					{
						c++;
					}
					else
					{
						c = 0; break;
					}
				}
				else if (origTEX[ld - lines] & 0x44)
				{
					d++;
				}
			}

			if (d && c > 1) subfn_CheckPartialHiMissing(sc);
		}
	}
}

// ====================================================

// v220-19
void MX_CheckNestedTexInMovRings()
{
	int		c, d, e, i, k;

	if (R3D_ScriptedMap) return;	// Scripted maps use different methods

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& SecCtrl(i)
			&& !(sectors[i].specFlags & 0x30404))
		{
			sector_t *ch, *os, *sc = &sectors[i];
			line_t	 *ld;

			for (c = d = e = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (!(ld->ml_flags & ML_TWOSIDED)) { c = 0; break; }
				if (ld->frontsector == ld->backsector) { c = 0; break; }
				if (lxarray[ld - lines]) { c = 0; break; }
				if (!(ld->lnflags2 & LNF2_OWNEQUALTOTWO)) { c = 0; break; }
				if (!Vtxs2(ld)) { c = 0; break; }

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (ch->specFlags & SF_MOVCEIL) { c = 0; break; }

				if (origTEX[ld - lines] & 0x55)
				{
					if (ch->specFlags & SF_MOVFLOOR) { c = 0; break; }
					linearray[c++] = ld - lines;
					if (!os) os = ch; else if (os != ch) { c = 0; break; }
				}
				else
				{
					if (!(ch->specFlags & SF_MOVFLOOR)) { c = 0; break; }
					d++;
				}

				e++;
			}

			if (os && c > 2 && d > 2 && c + d == e)
			{
				int rc;

				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

				if (subfn_CheckForRing(hitline, sc, linearray[0], &rc, NULL, NULL, NULL) && rc == c)
				{
					subsector_t	*ss;
					ssline_t	*sl;
					seg_t		*sg;
					int			m, n;
					int			*lnmem = (int*)malloc(c * sizeof(int));

					memcpy(lnmem, linearray, c * sizeof(int));

					ss_validcount++; validcount++;

					for (e = k = 0; k < c; k++)
					{
						sl = &sslines[lnmem[k]];

						for (m = 0; m < sl->num; m++)
						{
							ss = &subsectors[sl->list[m]];

							if (ss->ssvalidcount == ss_validcount) continue;

							ss->ssvalidcount = ss_validcount;

							sg = &segs[ss->firstline];

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (!sg->backsector) { e = -1; break; }

								if ((ld = sg->linedef) && ld->validcount != validcount)
								{
									ld->validcount = validcount;

									if (hitline[ld - lines] == 4)
									{
										if (sg->frontsector != sc && sg->frontsector != os) { e = -1; break; }
									}
									else if (origTEX[ld - lines] & 0x55)
									{
										goto tidy_up;
									}
									else
									{
										if ((sg->backsector->specFlags & SF_MOVSECTOR) != SF_MOVFLOOR) { e = -1; break; }
									}

									linearray[e++] = ld - lines;
								}
							}
						}
					}

					if (e > 0)
					{
						for (k = 0; k < e; k++)
						{
							ld = &lines[linearray[k]];

							if (origTEX[ld - lines])
								Settype(ld, R3D_36, 48128);
							else
								Settype(ld, R3D_16, 48128);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS, 48128);
						}

						Setsecdone(sc, 48129);
						Setsecdone(os, 48129);
					}
tidy_up:;
					free(lnmem);
				}
			}
		}
	}
}

// --------------------------------------------

static RDBOOL subfn_CheckNoHiAndSameHeightLine(sector_t *sc, sector_t *ns, sector_t *os, int *lnmem)
{
	subsector_t	*ss;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k;
	int			lc, rc, box[2][4];

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box[0], NULL, NULL) && rc > 2)
		{
			if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box[0], -1, MSIA_NONE))
			{
				for (i = 0; i < sc->linecount; i++)
				{
					if (hitline[ ((ld = sc->lines[i]) - lines) ] == 4)
					{
						hitline[ld - lines] = 8;
					}
				}

				for (c = i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					sg = &segs[ss->firstline];

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						if ( (ld = sg->linedef) && hitline[ld - lines] == 16)
						{
							c++; lnum = ld - lines; hitline[lnum] = 1;
						}

						if (   (ld->frontsector == sc && ld->backsector == os)
							|| (ld->frontsector == os && ld->backsector == sc))
						{
							hitline[ld - lines] = 8;
						}
					}
				}

				if (c < 3) return false;

				if (   subfn_CheckForRing(hitline, os, lnum, &rc, box[1], NULL, NULL)
					&& rc > 2
					&& M_CompareBox(box[1], box[0]))
				{
					int valid;
					int *ssecmem = (int*)malloc(hitssecptr * sizeof(int));
					int	ptrmem = hitssecptr;
					byte txn;

					memcpy(ssecmem, hitssec, hitssecptr * sizeof(int));

					valid = LS_MarkSubsecsInAreaGeneral(os, hitline, 0, NULL, box[1], -1, MSIA_REVERSED);
					
					if (!valid) goto tidy_up;

					for (i = 0; i < ptrmem; i++)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							if (ssecmem[i] == hitssec[k]) ssecmem[i] = -1;
						}
					}

					for (hitssecptr = i = 0; i < ptrmem; i++)
						if (ssecmem[i] >= 0) hitssec[hitssecptr++] = ssecmem[i];

					for (i = 0; i < hitssecptr; i++)
					{
						ss = &subsectors[hitssec[i]];

						Setr3dsscs(ss, ns - sectors, 48301);
					}

					for (lc = i = 0; i < sc->linecount; i++)
					{
						ld = sc->lines[i]; lnum = ld - lines;

						if (hitline[lnum] == 4 || hitline[lnum] == 8)
						{
							lnmem[lc++] = lnum;
							hitline[lnum] = 0;
						}
					}

					for (i = 0; i < os->linecount; i++)
					{
						ld = sc->lines[i]; lnum = ld - lines;

						if (hitline[lnum] == 4 || hitline[lnum] == 8)
						{
							lnmem[lc++] = lnum;
							hitline[lnum] = 0;
						}
					}

					for (i = 0; i < lc; i++)
					{
						ld = &lines[lnmem[i]];
						txn = origTEX[ld - lines];

						if (!txn)
							Settype(ld, R3D_3F, 48302);
						else if (ld->frontsector == ld->backsector)
							Settype(ld, R3D_03, 48303);
						else if (ld->backsector
							&& txn & 0x22
							&& !(txn & 0x55)
							&& ld->frontsector->floorheight == ld->backsector->floorheight)
							Settype(ld, R3D_03, 48304);
						else
							Settype(ld, R3D_36, 48305);

						SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 48306);
					}

					Setsecdone(sc, 48307);
					Setsecdone(ns, 48308);

tidy_up:;
					free(ssecmem);

					if (!valid) return false;
				}
				else
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}

	return true;
}

// v220-22
void MX_CheckNoHiAndSameHeightLine()
{
	sector_t	*sc, *ch, *ns, *os;
	line_t		*ld;
	int			*lnmem;
	int			sd, sn;
	int			c, d, i, k;

	for (i = 0; i < numsectors; i++)
	{
		if ( (sectors[i].specFlags & 0x3041C) != 0x18) continue;

		sc = &sectors[i];
		CLR_hitline();

		for (c = d = k = 0, ns = NULL, os = NULL; k < sc->linecount; k++)
		{
			ld = sc->lines[k];
			ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
			if (ch == sc) continue;
			if (!ch) { os = NULL; break; }
			sd = ld->frontsector == sc ? 0 : 1;

			if ( (sn = ld->sidenum[sd]) < 0) { os = NULL; break; }

			if (  !(origTEX[ld - lines] & 0x44)
				&& sides[sn].bottomtexture
				&& sc->ceilingheight > ch->ceilingheight
				&& sc->floorheight < ch->floorheight)
			{
				hitline[ld - lines] = 1; c++;
				if (!ns) ns = ch; else if (ns != ch) { ns = NULL; break; }
			}
			else
			if (  !(origTEX[ld - lines] & 0x55)
				&& sc->ceilingheight == ch->ceilingheight)
			{
				d++;
				if (!os) os = ch; else if (os != ch) { ns = NULL; break; }
			}
		}

		if (!ns || !os || ns->ceilingpic == skyflatnum) continue;
		if (c < 3 || d < 3) continue;
		if (ns->specFlags & 0x8 && !(ns->specFlags & 0x400)) continue;

		for (c = k = 0; k < os->linecount; k++)
		{
			ld = os->lines[k];
			if (ld->frontsector == ld->backsector) continue;

			sd = ld->frontsector == os ? 0 : 1;

			if (ld->sidenum[sd] < 0 || sides[ld->sidenum[sd]].toptexture)
			{
				hitline[ld - lines] = 16; c++;
			}
		}

		if (c < 3) continue;

		lnmem = (int*)malloc((sc->linecount + os->linecount) * sizeof(int));

		subfn_CheckNoHiAndSameHeightLine(sc, ns, os, lnmem);

		free(lnmem);
	}
}

// ======================================================

static void subfn_CheckAllConnectedLowMissing(sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*ch, *os = NULL;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k, m;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (lineflags[ld - lines]) return;
		if (!ld->backsector) return;
		if (origTEX[ld - lines]) return;
		if (!Vtfl3(ld)) return;

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (ch->floorheight >= sc->floorheight || ch->ceilingheight != sc->ceilingheight) return;

		if (!os)
		{
			os = ch;
			addSec(ch - sectors, &c);
		}
		else if (os != ch)
		{
			if (   ch->floorheight != os->floorheight
				|| ch->ceilingheight != os->ceilingheight
				|| ch->floorpic != os->floorpic)
			{
				return;
			}
			addSec(ch - sectors, &c);
		}
	}

	ss_validcount++;
	hitssecptr = 0;

	for (i = 0; i < sc->linecount; i++)
	{
		sl = &sslines[ (sc->lines[i] - lines) ];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;

			if (ss->sector == sc)
			{
				sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->frontsector != sc) return;
				}
			}
			else
			{
				if (ss->sector != os)
				{
					if (   ss->sector->floorheight != os->floorheight
						|| ss->sector->ceilingheight != os->ceilingheight
						|| ss->sector->floorpic != os->floorpic)
					{
						return;
					}

					sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (sg->frontsector != ss->sector) return;
					}
				}
			}

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];

		Setr3dsslfsec(ss, os - sectors, 48309);
		if (ss->sector != sc) SetOverride(ss, sc - sectors, 48310);
	}

	addSec(sc - sectors, &c);

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		Settype(ld, R3D_3F, 48311);
		SetLFG(ld, LFG_NEWSET|LFG_NOCHANGE, 48312);
	}

	for (i = 0; i < c; i++) SetSecidxDone(countsec[i], 48313);
}

// v220-22
// Find static sector with no hilo texs where all lines connect one to another
// This type defeats the loop code as it is not known which lines to go to next
// without further analysis 
void MX_CheckAllConnectedLowMissing()
{
	int			i;

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && (sectors[i].specFlags & 0xFFFFC) == 0xC && !SecCtrl(i))
			subfn_CheckAllConnectedLowMissing(&sectors[i]);
	}
}