//-----------------------------------------------------------------------------
//
// Copyright (C) 2008 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_PSroutines.c
//
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
//
// DESCRIPTION: Routines relocated from r3d_postcheck.c plus 
//              newer replacement routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "dd_types.h"
#include "dd_wad.h"
#include "m_misc.h"

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	pwad_nodesNotRebuilt;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int ss_validcount;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// Get rid of textures that will never be rendered, check for R3D_03
// Check for unset R3D_0C types.
// v220-7 add substitution for missing hi textures
void PS_UnwantedTextures()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (lineflags[i] & (LFG_FIXED|LFG_3DTYPE))
		{
			continue;
		}

		if (ld->frontsector == ld->backsector)
		{
			if (   ld->specType != R3D_22
				&& !(ld->frontsector->specFlags & SF_MOVSECTOR))
			{
				if (origTEX[i] & 0x55)
				{
					checkULTEX(ld, 3);
					sides[ld->sidenum[0]].bottomtexture = 0;
					sides[ld->sidenum[1]].bottomtexture = 0;
					sides[ld->sidenum[0]].toptexture = 0;
					sides[ld->sidenum[1]].toptexture = 0;
					origTEX[i] &= ~0x55;
					sides[ld->sidenum[0]].texFlags &= ~(T_LOW|T_HI);
					sides[ld->sidenum[1]].texFlags &= ~(T_LOW|T_HI);
				}
				else
				if (origTEX[i] == 0x22
					&& ld->specType != R3D_03
					&& sides[ld->sidenum[0]].rowoffset < 0
					&& sides[ld->sidenum[0]].rowoffset
					== sides[ld->sidenum[1]].rowoffset)
				{
					Settype(ld, R3D_03, 43001);
				}
			}
		}

		if (ld->specType == R3D_0C)
		{
			byte settype = origTEX[i] ? R3D_00 : R3D_01;

			Settype(ld, settype, 43002);
		}
	}
}

// Set tex inhibit flags where rendering frontsec == backsec
void PS_SetTexInhibitFsecEqualBsec()
{
	line_t	*ld;
	seg_t	*sg;
	int		a, b, i, k, m;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[i] & (LFG_FIXABS|LFG_3DTYPE)
			|| ld->frontsector == ld->backsector)
		{
			continue;
		}

		a = !(origTEX[i] & 0x11) && ld->frontsector->floorheight   != ld->backsector->floorheight;
		b = !(origTEX[i] & 0x44) && ld->frontsector->ceilingheight != ld->backsector->ceilingheight;

		if (a || b)
		{
			int fsecf = -1;
			int fsecb = -1;
			int csecf = -1;
			int csecb = -1;

			ssline_t *sl = &sslines[i];

			ss_validcount++;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				ss->ssvalidcount = ss_validcount;

				if (ss->floorsec >= 0 || ss->ceilsec >= 0)
				{
					for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
					{
						if (sg->linedef == ld && sg->backsector)
						{
							if (a && ss->floorsec >= 0)
							{
								if (fsecf < 0 && sg->frontsector == ld->frontsector)
									fsecf = ss->floorsec;
								if (fsecb < 0 && sg->frontsector == ld->backsector)
									fsecb = ss->floorsec;
							}

							if (b && ss->ceilsec >= 0)
							{
								if (csecf < 0 && sg->frontsector == ld->frontsector)
									csecf = ss->ceilsec;
								if (csecb < 0 && sg->frontsector == ld->backsector)
									csecb = ss->ceilsec;
							}
						}
					}
				}
			}

			if (fsecf >= 0 && fsecb >= 0 && sectors[fsecf].floorheight   == sectors[fsecb].floorheight  ) ld->lnflags2 |= LNF2_NORENDLOW;
			if (csecf >= 0 && csecb >= 0 && sectors[csecf].ceilingheight == sectors[csecb].ceilingheight) ld->lnflags2 |= LNF2_NORENDHI;
		}
	}
}

// b211-0 Check for any unset sectors with no lotex
void PS_CheckUnsetLoTexSectors()
{
	sector_t *sec;
	int		 i, k, bbox[4];

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->specFlags == 0x00008394
			&& sec->heightsec < 0
			&& sec->stptr < 0
			&& nf_lineList(sec->lines[0], sec, bbox, linearray, 0) == sec->linecount
			&& nf_boxAreaSameSec(sec, NULL, bbox, sec->linecount))
		{
			RDBOOL	valid = true;
			int		*list = (int*)M_Malloc(sec->linecount * sizeof(int), 51);
			int		csec = -1, lnumcnt = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if ( !(ld->ml_flags & ML_TWOSIDED)
					|| lineflags[ld - lines]
					|| origTEX[ld - lines] & 0x22)
				{
					valid = false; break;
				}
				if (ld->frontsector == ld->backsector)
				{
					if (   (vertexowners[ld->v1 - vertexes].num == 1 && vtxarray[ld->v1 - vertexes] == 2)
						|| (vertexowners[ld->v2 - vertexes].num == 1 && vtxarray[ld->v2 - vertexes] == 2))
					{
						int chksec = LS_FindAdjacentSecnum(ld, 3);

						if (chksec < 0) { valid = false; break; }
						if (csec < 0) csec = chksec;
						else
						if (chksec != csec) { valid = false; break; }
						list[lnumcnt++] = ld - lines;
					}
				}
			}

			if (valid && lnumcnt)
			{
				for (k = 0; k < lnumcnt; k++)
				{
					SetLFG(&lines[list[k]], LFG_NEWSET, 43003);
				}
				// Only mark the same sector sided lines
				for (k = 0; k < hitssecptr; k++)
				{
					Setr3dssfs(&subsectors[hitssec[k]], csec, 43004);
				}
			}

			M_Free(list);
		}
	}
}

void PS_CheckUnsetSupportLines()
{
	subsector_t *ss;
	sector_t	*s, *chk;
	ssline_t	*sl;
	line_t		*ld;
	int			i, k, m;

	// v220-10
	// Check for any partially set mov sectors
	for (i = 0, s = sectors; i < numsectors; i++, s++)
	{
		if (secdone[i] || s->stptr >= 0 || s->heightsec >= 0 || !ssUnset1(s)) continue;

		// v220-10 Check for any unset subsectors in a floor move
		//         sector only referencing itself where it is partially set
		if ((s->specFlags  & 0xFFFEC) == 0x1000C && SecCtrl(i) && SecONE(i))
		{
			// v220-16 do nothing
		}
		else
		// v220-8 Deal with horrible bodge in Requiem map21 (sectors 90/91 and 96/97) where the
		//      inheritable sectors (S87, S841) are muddled together with differing sky heights.
		if ( (s->specFlags & 0xFFFFC) == 0xC && !SecCtrl(i))
		{
			int cs, ff, os = -1;

			hitssecptr = 0;

			ss_validcount++;

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];

				if ( !(ld->ml_flags & ML_TWOSIDED) || lineflags[ld - lines])
				{ os = -1; break;
				}

				chk = ld->frontsector == s ? ld->backsector : ld->frontsector;
				ff  = chk->specFlags & 0xFFFFC;

				if ((ff & 0xF000C) != 0xC) { os = -1; break; }

				sl = &sslines[ld - lines];

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];

					if (ss->ssvalidcount != ss_validcount)
					{
						ss->ssvalidcount = ss_validcount;

						if ( (cs = ss->floorsec) >= 0)
						{
							if (cs != i && cs != chk - sectors) break;
							cs = -1;
						}

						hitssec[hitssecptr++] = sl->list[m];
					}
				}

				if (cs >= 0) { os = -1; break; }

				if (   ff == 0x889C
					&& !origTEX[ld - lines]
					&& !lineflags[ld - lines])
				{
					if (linetable[ld - lines] & 2)
					{
						int lo, hi;

						XS_GetBoundsFromLine(ld, &lo, &hi, NULL);

						if ( lo != hi || lo == i || lo == chk - sectors) { os = -1; break; }

						if (os < 0)
							os = lo;
						else if (os != lo)	// and herein lies the rub
						{
							if (sectors[lo].floorheight != sectors[os].floorheight) { os = -1; break; }
							if (sectors[lo].ceilingheight != sectors[os].ceilingheight
								&& sectors[lo].ceilingpic != skyflatnum
								&& sectors[lo].ceilingpic != skyflatnum)
							{ os = -1; break;
							}

							if (sectors[lo].ceilingheight > sectors[os].ceilingheight)
								os = lo;
						}
					}
				}
			}

			if (os >= 0)
			{
				for (k = 0; k < hitssecptr; k++)
					SetOverride(&subsectors[hitssec[k]], os, 989);

				for (k = 0; k < s->linecount; k++)
				{
					ld = s->lines[k];
					SetLFG(ld, LFG_NEWSET, 989);
				}

				Setsecdone(s, 989);
			}
		}
	}

	// PS_CheckUnsetSupportLines() line test
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& !origTEX[i]
			&& !lxarray[i]
			&& !(lineflags[i] & (LFG_NEWSET|LFG_LINESSECSET|LFG_3DTYPE))
			&& ld->lnflags2 & LNF2_OWNEQUALTOONE)
		{
			ssline_t	*sl = &sslines[i];
			int			count = 0;

			ss_validcount++;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t	*ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				ss->ssvalidcount = ss_validcount;

				if (ss->ssflags & SSF_NOCHANGE)
				{
					count = 0; break;
				}

				if (   ss->sector == ld->frontsector
					&& (ss->floorsec >= 0 || ss->ceilsec >= 0))
				{
					continue;
				}

				hitssec[count++] = sl->list[k];
			}

			if (count)
			{
				int losec, hisec;
				
				XS_GetBoundsFromLine(ld, &losec, &hisec, NULL);

				if (losec >= 0 && losec == hisec && losec != ld->frontsector - sectors)
				{
					for (k = 0; k < count; k++)
					{
						SetOverride(&subsectors[hitssec[k]], losec, 43005);
					}

					SetLFG(ld, LFG_NEWSET, 43006);
				}
			}
		}
	}

	// v220-3 added
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		int c, d;

		if (   ss->linecount > 2
			&& ss->floorsec < 0
			&& ss->ceilsec < 0
			&& ss->sector->heightsec < 0
			&& !(ss->sector->specFlags & 0xC))
		{
			sector_t *os = NULL;
			seg_t	 *sg = &segs[ss->firstline];

			for (c = d = k = 0; k < ss->linecount; k++, sg++)
			{
				ld = sg->linedef;

				if (!ld->backsector)
				{
					os = NULL; break;
				}

				if (ld->frontsector != ld->backsector)
				{
					if (c || sg->frontsector != ss->sector)
					{
						os = NULL; break;
					}
					c++;
				}
				else if (ld->frontsector == ss->sector)
				{
					d++;
				}
				else if ( (ld->frontsector->specFlags & 0xF84FC) == 0xC
					&& lxarray[ld - lines])
				{
					if (   ld->frontsector != ss->sector
						&& ld->frontsector->floorheight > ss->sector->floorheight)
					{
						if (os && os != ld->frontsector)
						{
							os = NULL; break;
						}
						os = ld->frontsector;
					}
				}
				else
				{
					os = NULL; break;
				}
			}

			if (os && d)
			{
				SetOverride(ss, ss->sector - sectors, 43007);	// The rendering sector
				ChangeSSECsec(ss, os, 43008);
			}
		}
	}
}

static RDBOOL chk1Cvtx(sector_t *sec)
{
	int		i, k;
	RDBOOL	vn = false;

	for (i = 0; i < sec->linecount; i++)
	{
		vertexowner_t *v = &vertexowners[sec->lines[i]->v1 - vertexes];

		if (v->num > 1) vn = true;

		for (k = 0; k < v->num; k++)
			if (sectors[v->list[k]].specFlags != 0x1C) return false;
		
		v = &vertexowners[sec->lines[i]->v2 - vertexes];

		if (v->num > 1) vn = true;

		for (k = 0; k < v->num; k++)
			if (sectors[v->list[k]].specFlags != 0x1C) return false;
	}
	return vn;
}

// b211-0 various (3 functions)
void PS_FunctionsVarious()
{
	sector_t *sec, *s;
	int		 i, k, m, ref, secnum;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i]) continue;

		if (sec->specFlags == 0x1C)
		{
			if (!SecCtrl(i) && !chk1Cvtx(sec))
				continue;

			s = XS_GetBoundary(sec, &ref);
			secnum = s ? s - sectors : -1;

			if (secnum >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
					{
						ssline_t *sl = &sslines[sec->lines[k] - lines];

						for (m = 0; m < sl->num; m++)
							SetOverride(&subsectors[sl->list[m]], secnum, 43009);
					}
				}
			}
		}

		// Check for unclosed sectors contained in a sky sector
		else if (   sec->ceilingpic != skyflatnum
			&& (sec->specFlags & 0x09B3B39C) == 0x0000808C)	// include SF_HEIGHSKY
		{
			sector_t *s = XS_GetBoundary(sec, &ref);

			secnum = s ? s - sectors : -1;

			if (secnum >= 0
				&& sectors[secnum].specFlags & SF_SECCLOSED
				&& sectors[secnum].ceilingpic == skyflatnum)
			{
				RDBOOL	 valid = true;
				ssline_t *sl;

				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (!(ld->spflags & SPF_CONTROLLINE))
					{
						sector_t *fs = ld->frontsector;
						sector_t *bs = ld->backsector;

						if (   !(ld->ml_flags & ML_TWOSIDED)
							|| fs == bs
							|| fs->floorpic != bs->floorpic
							|| fs->ceilingpic != bs->ceilingpic
							|| fs->floorheight != bs->floorheight
							|| fs->ceilingheight != bs->ceilingheight)
						{
							valid = false; break;
						}

						sl = &sslines[ld - lines];

						for (m = 0; m < sl->num; m++)
						{
							if (!SSECunset(&subsectors[sl->list[m]]))
							{
								valid = false; break;
							}
						}
						if (!valid) break;
					}
				}

				if (valid)
				{
					for (k = 0; k < sec->linecount; k++)
					{
						if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
						{
							ssline_t *sl = &sslines[sec->lines[k] - lines];

							for (m = 0; m < sl->num; m++)
								SetOverride(&subsectors[sl->list[m]], secnum, 43010);

							SetLFG(sec->lines[k], LFG_NEWSET, 43011);
						}
					}
				}
			}
		}
		else if ((sec->specFlags & 0x0903B39C) == 0x0001818C	// include SF_HEIGHSKY
			&& IsLiquidFloorFlat(sec))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t	*li = sec->lines[k];
				int		seclo, sechi;

				XS_GetBoundsFromLine(li, &seclo, &sechi, NULL);

				if (seclo >= 0 && sechi == seclo)
				{
					ssline_t	*sl = &sslines[li - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->floorsec < 0 && ss->ceilsec < 0)
						{
							SetOverride(ss, seclo, 43012);
						}
					}
				}
			}
			Setsecdone(sec, 43013);
		}
	}
}

// b211-0 catch any unset owner sectors for same-sector lines
void PS_CheckUnsetOwnerSectors()
{
	sector_t	*s;
	line_t		*ld, *li;
	int			i, k, seclo, sechi;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  (s = ld->frontsector) == ld->backsector
			&& !lineflags[i]
			&& !secdone[s - sectors]
			&& (s->specFlags & 0x3000C) == 0xC
			&& !lxarray[i]
			&& ld->specType < R3D_40)
		{
			if ( (li = subfn_FindLineOnLine(ld, 0)) )	// v220-9
			{
				if (ge_inf.devmap && (li->frontsector != s || li->backsector != s))
					pr_OFinsert("DEVMAP: line %i lying on line %i has been ignored\n", i, li - lines);
				// ignore
				SetLFG(li, LFG_NEWSET, 43015);
			}
			else
			{
				XS_GetBoundsFromLine(ld, &seclo, &sechi, NULL);

				if (seclo >= 0 && sechi == seclo && seclo != s - sectors)
				{
					ssline_t *sl = &sslines[i];
					RDBOOL	 lineset = false;

					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (   ss->sector - sectors != seclo
							&& ss->floorsec < 0
							&& ss->ceilsec < 0)
						{
							// v220-10 add ceiling test
							if (   ss->sector->specFlags & SF_NOHITEX
								|| ss->sector->ceilingheight > sectors[seclo].ceilingheight)
							{
								Setr3dsscs(ss, seclo, 43014);
							}

							Setr3dssfs(ss, seclo, 43014);
							lineset = true;
						}
					}

					if (lineset)
						SetLFG(ld, LFG_NEWSET, 43015);
				}
			}
		}
	}
}

// b211-0 Check for water rings
void PS_CheckLiquidRings()
{
	sector_t	*sec;
	int			c, i, k, m;
	int			ring_count, bbox[4];

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& IsLiquidFloorFlat(sec)
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& sec->linecount > 2
			&& (sectors[i].specFlags & 0xF0014) == 0x10)
		{
			RDBOOL valid = true;

			CLR_hitline();

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED)
				{
					if (ld->frontsector == ld->backsector)
					{
						valid = false; break;
					}
					if (   !(ld->frontsector->specFlags & SF_NOLOTEX)
						&& !(ld->backsector->specFlags  & SF_NOLOTEX))
					{
						hitline[sec->lines[k] - lines] = 1;
						c++;
					}
				}
				else
				{
					hitline[sec->lines[k] - lines] = 1;
					c++;
				}
			}

			if (!valid || c < 3) continue;

			valid = false;

			for (k = 0; k < sec->linecount; k++)
			{
				int lnum = sec->lines[k] - lines;

				if (   hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL)
					&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE))
				{
					RDBOOL justlow = false;
					RDBOOL nothi   = sec->ceilingpic != skyflatnum;
					RDBOOL movsec  = false;

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];

						if (ss->sector != sec)
						{
							if (ss->sector->specFlags & SF_MOVSECTOR)
							{
								movsec = true; break;
							}

							if (!(ss->sector->specFlags & SF_NOLOTEX))
							{
								valid = false; break;
							}

							if (   nothi
								&& ss->ceilsec >= 0
								&& ss->ceilsec != i)
							{
								nothi = false;
							}

							if (!(ss->sector->specFlags & SF_NOHITEX))
							{
								nothi = false;
								justlow = true;
							}

							if (sec->floorheight > ss->sector->floorheight)
								valid = true;
						}
					}

					if (!movsec)
					{
						if (valid)
						{
							for (m = 0; m < hitssecptr; m++)
							{
								subsector_t *ss = &subsectors[hitssec[m]];

								Clear_SSEC(ss);

								if (justlow)
									Setr3dssfs(ss, i, 43016);
								else
									SetOverride(ss, i, 43017);
							}
							Setsecdone(sec, 43018);
						}
						else if (nothi)
						{	// This checks for a missing hi-tex ring that has not been
							// set due to a problem with sector assignments
							// E.g.where incorrectly set from 47031 in r3d_NewFunc
							// This resolves here through checking if ceilsec is set to sec
							sector_t *osec = NULL;

							for (c = m = 0; m < sec->linecount; m++)
							{
								lnum = sec->lines[m] - lines;

								if (hitline[lnum] == 4)
								{
									line_t	*ld = sec->lines[m];

									if (   !lineflags[lnum]
										&& Vtxs2(ld)
										&& Vtfl3(ld)
										&& !(origTEX[lnum] & 0x44))
									{
										sector_t *chk = ld->frontsector == sec ?
											ld->backsector : ld->frontsector;

										if (!osec) osec = chk;
										else
										if (osec != chk || osec->ceilingheight >= sec->ceilingheight)
										{
											c = 0; break;
										}
										c++;
									}
								}
							}
							if (c == ring_count)
							{
								for (m = 0; m < hitssecptr; m++)
									Setr3dsscs(&subsectors[hitssec[m]], osec - sectors, 43019);
							}
						}
					}
				}

				for (m = 0; m < sec->linecount; m++)
				{
					lnum = sec->lines[m] - lines;

					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}
			}
		}
	}
}

// Check line rings with lower tex in a different containing sector
void PS_CheckLowerTexRings()
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, m, n, p;
	int			lnum, ring_count, bbox[4];

	for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->specFlags & SF_NOHITEX
			&& sec->specFlags & SF_SECCLOSED
			&& !(sec->specFlags & (SF_NOLOTEX|SF_MOVSECTOR)))
		{
			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->specType == R3D_00
					&& ld->backsector == sec
					&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
					&& Vtxs2(ld)
					&& origTEX[ld - lines] & 0x01
					&& !(origTEX[ld - lines] & 0x66)
					&& ld->backsector->floorheight   >  ld->frontsector->floorheight
					&& ld->backsector->ceilingheight == ld->frontsector->ceilingheight)
				{
					hitline[ld - lines] = 1;
					c++;
				}
			}
		}
		if (c < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			lnum = sec->lines[k] - lines;

			if (   hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL)
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE))
			{
				if (!LS_LinesInSSECarray(hitline, LSS_NONE|LSS_SSECSET, NULL))
				{
					sector_t *fs = lines[lnum].frontsector;
					sector_t *bs = lines[lnum].backsector;
					sector_t *nsec = NULL;

					for (m = 0; m < sec->linecount; m++)
					{
						if (hitline[sec->lines[m] - lines] == 4)
						{
							ssline_t *sl = &sslines[sec->lines[m] - lines];

							for (n = 0; n < sl->num; n++)
							{
								subsector_t	*ss = &subsectors[sl->list[n]];
								seg_t		*sg = &segs[ss->firstline];
								sector_t	*chk = ss->sector;

								if (chk != fs && chk != bs && SSECunset(ss))
								{
									// Subsec must have a textured line
									for (p = 0; p < ss->linecount; p++, sg++)
									{
										// b211-0 Revised ->
										if (   sg->linedef
											&& sg->frontsector == chk
											&& 
											( (sg->backsector && sg->sidedef->toptexture
											&& sg->frontsector->ceilingheight > sg->backsector->ceilingheight)
											||
											(!sg->backsector && sg->sidedef->midtexture)) )
										{
											if (sg->frontsector->specFlags & SF_SECCLOSED
												&& Vtxs2(sg->linedef))
											{
												nsec = chk; break;
											}
											else
											{
												nsec = sec; break;
											}
										}
										// b211-0 end revised <-
									}
								}
								if (nsec) break;
							}
						}
						if (nsec) break;
					}

					if (nsec && nsec != sec)	// b211-0 changed
					{
						// All line adjacent subsecs must be set
						for (m = 0; m < sec->linecount; m++)
						{
							if (hitline[sec->lines[m] - lines] == 4)
							{
								ssline_t *sl = &sslines[sec->lines[m] - lines];

								for (n = 0; n < sl->num; n++)
								{
									subsector_t *ss = &subsectors[sl->list[n]];
									if (ss->sector != nsec && SSECunset(ss))
									{
										Setr3dsscs(ss, nsec - sectors, 43020);
									}
								}
							}
						}
					}
				}
				for (m = 0; m < sec->linecount; m++)
				{
					int lnum = sec->lines[m] - lines;
					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}
			}
		}
	}
}

// b211-0 Add - Look for sectors bounded by liquid floor flats
void PS_BoundedLiquidSectors()
{
	sector_t	*s, *chk, *sec;
	int			c, i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->specFlags & SF_MOVSECTOR
			|| IsLiquidFloorFlat(sec) )
		{
			continue;
		}

		CLR_hitline();
		s = NULL;

		for (c = k = 0; k < sec->linecount; k++)
		{
			line_t	*ld = sec->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			if (   !origTEX[ld - lines]
				&& !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
				&& ld->specType == R3D_00
				&& ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector->floorheight != ld->backsector->floorheight)
			{
				chk = ld->frontsector == sec ?
					ld->backsector : ld->frontsector;

				if (   !s
					&& (!IsLiquidFloorFlat( (s = chk) )
					|| sec->floorheight - s->floorheight < 64*FRACUNIT))
				{
					c = 0; break;
				}

				if (   s == chk
					&& !subfn_FindLineCrossing(ld, LC_NORMAL))
				{
					hitline[ld - lines] = 1; c++;
				}
			}
		}

		if (c < 3)	continue;

		for (k = 0; k < sec->linecount; k++)
		{
			int rcount, lnum = sec->lines[k] - lines;

			if (   hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, &rcount, NULL, NULL, NULL))
			{
				for (m = 0; m < sec->linecount; m++)
				{
					lnum = sec->lines[m] - lines;

					if (hitline[lnum] == 4)
					{
						Settype(&lines[lnum], R3D_16, 43021);
						hitline[lnum] = 8;
					}
				}
			}
		}
	}
}

// b211-0 Added routines
void PS_MiscellaneousRoutines()
{
	typedef struct { sector_t *s; int lnum; int maxlnum; int *ln; } lmem_t;
	lmem_t		*lmptr = NULL;
	int			lmnum = 0;
	int			lmmaxnum = 0;
	line_t		*ld;
	int			i, k;
	int			mask = SF_MOVCEIL|SF_NOHITEX;

	//----------------------------------

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType >= R3D_40) continue;

		if (ld->frontsector == ld->backsector)
		{
			if (origTEX[i] & 0x22)
				SetNoskywall(i, 43022);

			if (ld->specType == R3D_00)
			{
				if (origTEX[i] & 0x22)
				{
					Settype(ld, R3D_03, 43023);
					continue;
				}
			}
		}
		else if (ld->backsector
			&& (ld->frontsector->ceilingpic == skyflatnum
			||  ld->backsector->ceilingpic  == skyflatnum) )
		{
			if (   ld->frontsector->ceilingpic == skyflatnum
				&& ld->backsector->ceilingpic  == skyflatnum)
			{
				if (   (ld->backsector->specFlags & mask) == mask
					&& ld->backsector->floorheight != ld->backsector->ceilingheight)
				{
					SetNoskywall(i, 43024);
				}
				else if ((ld->frontsector->specFlags & mask) == mask
					&& ld->frontsector->floorheight != ld->frontsector->ceilingheight)
				{
					SetNoskywall(i, 43024);
				}
			}
			else if (ld->lnflags2 & LNF2_OWNEQUALTOTWO
				&& Vtxs2(ld))
			{
				int sidex = ld->frontsector->ceilingpic == skyflatnum ?
					ld->sidenum[0] : ld->sidenum[1];
				int sidey = ld->sidenum[0] == sidex ?
					ld->sidenum[1] : ld->sidenum[0];

				if (   sides[sidex].toptexture
					&& sides[sidex].sector->ceilingheight
					>  sides[sidey].sector->ceilingheight)
				{
					sector_t *s = ld->frontsector->ceilingpic == skyflatnum ?
						ld->backsector : ld->frontsector;
					lmem_t *lm = NULL;

					if (lmnum >= lmmaxnum)
					{
						lmmaxnum += 16;
						lmptr = (lmem_t*)W_ReallocCLR(lmptr, lmmaxnum * sizeof(lmem_t), TEMPID, 0);
					}

					for (k = 0; k < lmnum; k++)
					{
						if (lmptr[k].s == s) { lm = &lmptr[k]; break; }
					}

					if (!lm)
					{
						lm = &lmptr[lmnum++];
						lm->s = s;
					}

					if (lm->lnum >= lm->maxlnum)
					{
						lm->maxlnum += 16;
						lm->ln = (int*)W_ReallocCLR(lm->ln, lm->maxlnum * sizeof(int), TEMPID, 0);
					}

					lm->ln[lm->lnum++] = i;
				}
			}
		}

		if (   ld->specType == R3D_00
			&& ld->ml_flags & ML_TWOSIDED
			&& origTEX[i] == 0x50
			&& ld->backsector->ceilingheight - ld->backsector->floorheight == FRACUNIT
			&& ld->frontsector->ceilingheight - ld->frontsector->floorheight > 3*FRACUNIT
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
			&& !(ld->backsector->specFlags & SF_MOVSECTOR))
		{
			Settype(ld, R3D_01, 43025);
			{
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->sector == ld->backsector && SSECunset(ss))
					{
						SetOverride(ss, ld->frontsector - sectors, 43026);
					}
				}
			}
		}

		// Reset certain textureless line types in sky and fake sectors
		if (  !origTEX[i]
			&& ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector)
		{
			sector_t *fs = ld->frontsector;
			sector_t *bs = ld->backsector;

			if (fs->heightsec >= 0 && bs->heightsec < 0 && fs->ceilingpic == skyflatnum)
			{
				sector_t *s = &sectors[fs->heightsec];

				if (   s->ceilingheight != bs->ceilingheight
					|| s->floorheight   != bs->floorheight)
				{
					Forcetype(ld, R3D_00, 43027);
				}
			}
			if (bs->heightsec >= 0 && fs->heightsec < 0 && bs->ceilingpic == skyflatnum)
			{
				sector_t *s = &sectors[bs->heightsec];

				if (   s->ceilingheight != fs->ceilingheight
					|| s->floorheight   != fs->floorheight)
				{
					Forcetype(ld, R3D_00, 43028);
					Clr_allmtflags(ld, 43028);
					ld->tyflags = 0;
				}
			}
		}
	}

	if (lmptr) W_FreeALL(TEMPID);

	// v220-4 Code to suppress skywalls with line loops having a hi tex removed	
}

// b211-0 Added routine - check structures in fake sectors
void PS_CheckFakeSecStructures()
{
	sector_t	*sec, *osec;
	line_t		*ld;
	int			i, k, m, n, lnum;
	int			bbox[4], seccount = 0;

	CLR_hitline();
	SET_countsec(-1);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		int ffht, found = false;

		if (sec->heightsec < 0) continue;

		ffht = sectors[sec->heightsec].floorheight;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->ml_flags & ML_TWOSIDED
				&& origTEX[ld - lines] & 0x11
				&& !(origTEX[ld - lines] & 0x66)
				&& Vtxs2(ld))
			{
				sector_t *s = ld->frontsector == sec ?
					ld->backsector : ld->frontsector;

				if (s->heightsec < 0
					&& !(s->specFlags & SF_MOVSECTOR)
					&& s->ceilingheight == sec->ceilingheight
					&& s->floorheight > sec->floorheight
					&& s->floorheight < ffht
					&& s->ceilingpic == sec->ceilingpic)
				{
					hitline[ld - lines] = true;
					found = true;
				}
			}
		}

		if (found) countsec[seccount++] = i;
	}

	ss_validcount++;	// b211-0

	for (i = 0; i < seccount; i++)
	{
		sec = &sectors[countsec[i]];

		for (k = 0; k < sec->linecount; k++)
		{
			osec = sec->lines[k]->frontsector == sec ?
				sec->lines[k]->backsector : sec->lines[k]->frontsector;
			
			lnum = sec->lines[k] - lines;

			if (   hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, NULL, bbox, NULL, NULL))
			{
				int valid = LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_REVERSED);

				for (m = 0; m < sec->linecount; m++)
				{
					lnum = sec->lines[m] - lines;

					if (hitline[lnum] == 4)
					{
						hitline[lnum] = 8;

						if (valid)
						{
							ssline_t *sl = &sslines[lnum];

							for (n = 0; n < sl->num; n++)
							{
								subsector_t *ss = &subsectors[sl->list[n]];

								if (ss->ssvalidcount == ss_validcount) continue;

								ss->ssvalidcount = ss_validcount;

								if (ss->sector == osec)
								{
									Setr3dsscs(ss, sec - sectors, 43029);
									Setr3dsslfsec(ss, sec - sectors, 43030);
								}
							}
						}
					}
				}
			}
		}
	}
}

// B211-0 Check unset R3D_16 lines.
void PS_CheckUnsetR3D_16Lines()
{
	sector_t *sec;
	line_t	 *li;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags == 0x40858394
			&& NF_CheckLoop(sec, 0))	// Freedoom map30 sec 388
		{
			SetNOocclude(sec, 43031);
		}

		if (   !secdone[i]
			&& sec->specFlags & SF_MOVFLOOR
			&& sec->specFlags & SF_MOVINVALID
			&& sec->ceilingpic == skyflatnum)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				li = sec->lines[k];

				if (li->spflags & SPF_CONTROLLINE || li->specType == R3D_16) continue;

				if (   li->ml_flags & ML_TWOSIDED
					&& li->frontsector != li->backsector
					&& !(li->lnflags2 & LFG_NEWSET))
				{
					sector_t *s = li->frontsector == sec ?
						li->backsector : li->frontsector;

					if (   s->ceilingpic == skyflatnum
						&& !(s->specFlags & SF_MOVINVALID))
					{
						int txn = origTEX[li - lines];

						if (   (txn == 0x01 && li->frontsector == sec)
							|| (txn == 0x10 && li->backsector  == sec))
						{
							Forcetype(li, R3D_16, 43032);
						}
					}
				}
			}
		}
	}
}

// Check for 1's lines in mov sectors with masked doors
static void chkMaskedDoorsAdjoining1sLines()
{
	line_t	*ld;
	int		txn;
	int		c, i, k;
	int		m = movsec_list[0];
	int		*sn;

	if (!m) return;

	sn = (int*)M_Malloc(m * sizeof(int), 52);
	memset(sn, -1, m * sizeof(int));
	CLR_vtxcheck();

	for(c = 0, i = 1; i <= m; i++)
	{
		sector_t *s = &sectors[movsec_list[i]];
		RDBOOL	 found = false;

		if ( !(s->specFlags & SF_SSLINETRUE) ) continue;
		if ( !(s->specFlags & SF_MOVCEIL)    ) continue;
		if (s->floorheight + FRACUNIT < s->ceilingheight) continue;

		for (k = 0; k < s->linecount; k++)
		{
			ld = s->lines[k];

			// Some already set
			if (ld->adjacentsec >= 0) { found = false; break; }

			if (   ld->ml_flags & ML_TWOSIDED
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& ld->frontsector != ld->backsector
				&& origTEX[ld - lines] & 0x22
				&& !(origTEX[ld - lines] & 0x55))	// v220-4 further test added
			{
				int	v1 = vtoflags[ld->v1 - vertexes] & 1;
				int	v2 = vtoflags[ld->v2 - vertexes] & 1;

				if (v1 || v2)
				{
					if (   ((txn = sides[ld->sidenum[0]].midtexture)
						&& texArray[txn]->txf_flags & TXF_MASKED)
						|| ((txn = sides[ld->sidenum[1]].midtexture)
						&& texArray[txn]->txf_flags & TXF_MASKED) )
					{
						if (v1) vtxcheck[ld->v1 - vertexes] = true;
						if (v2) vtxcheck[ld->v2 - vertexes] = true;

						found = true;

						if (sn[c] < 0)
						{
							if (!(ld->frontsector->specFlags & SF_MOVSECTOR))
							{
								sn[c] = ld->frontsector - sectors;
							}
							else
							if (!(ld->backsector->specFlags & SF_MOVSECTOR))
							{
								sn[c] = ld->backsector - sectors;
							}
						}
					}
				}
			}
		}
		if (found) countsec[c++] = movsec_list[i];
	}

	for(i = 0; i < c; i++)
	{
		sector_t *s = &sectors[countsec[i]];

		for (k = 0; k < s->linecount; k++)
		{
			ld = s->lines[k];

			if (   !ld->backsector
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& vtxcheck[ld->v1 - vertexes]
				&& vtxcheck[ld->v2 - vertexes])
			{
				if (sectors[sn[i]].ceilingpic != skyflatnum)
					ld->adjacentsec = sn[i];
			}
		}
	}

	M_Free(sn);
}

// v220-9
static RDBOOL hasSingleLineSSEC(int lnum)
{
	ssline_t *sl = &sslines[lnum];

	if (sl->num == 2)
	{
		int	i;

		for (i = 0; i < sl->num; i++)
			if (subsectors[sl->list[i]].linecount == 1) return true;
	}

	return false;
}

// detect midtex doors spanning 1 sided lines or sectors marked SECSPF_DOORSECTOR
void PS_MidtexDoorsSpanning1SLines()
{
	sector_t	*sec;
	side_t		*sd1, *sd2;
	line_t		*ld;
	int			c, i, k;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		// Lines with mid textures only
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& origTEX[i] & 0x22
			&& !(origTEX[i] & 0x55)
			&& ld->mtflags & R3D_MISLOTEX)
		{
			if (Spans1sLines(ld))
			{
				if (   (ld->frontsector->specFlags & SF_MOVCEIL)
					^  (ld->backsector->specFlags  & SF_MOVCEIL))
				{
					int tx = origTEX[i] & 0x02 ? sides[ld->sidenum[0]].midtexture : sides[ld->sidenum[1]].midtexture;

					if (texArray[tx]->txf_flags & TXF_MASKED)
						ld->lnflags2 |= LNF2_SPAN1DOORS;
					// v220-9 detect map error
					else if (ld->frontsector->ceilingheight < ld->backsector->ceilingheight
						&& origTEX[i] == 0x02
						&& hasSingleLineSSEC(i))
					{
						pr_OFinsert("MAP ERROR: L%i should have no backsector\n", i);
						ma_inf.maperror++;
						ld->scpFlags |= SCPF_IGNOREBS;
						ld->ml_flags |= ML_BLOCKING;
					}
				}
				// Eliminate lines with a midtex one side only
				else if (origTEX[i] == 0x22)
				{
					if ( !(ld->mtflags & R3D_MISHITEX) )
					{
						sector_t *s = ld->frontsector->floorheight > ld->backsector->floorheight ?
							ld->frontsector : ld->backsector;

						if ( !(s->specFlags & SF_MOVSECTOR)
							&& s->specFlags & SF_SSLINETRUE
							&& ssecAllSameSec(s, ld - lines))
						{
							// do nothing
						}
						else
							SetMissinglo(ld, 43033);
					}
					else
						SetMissinglo(ld, 43034);
					
					Settype(ld, R3D_03, 43035);
				}
			}
			else
			if (  (secspflags[ld->frontsector - sectors] & SECSPF_DOORSECTOR)
				^ (secspflags[ld->backsector  - sectors] & SECSPF_DOORSECTOR))
			{
				ssline_t *sl = &sslines[i];
				sector_t *sc = secspflags[ld->frontsector - sectors] & SECSPF_DOORSECTOR ?
															ld->backsector : ld->frontsector;
				RDBOOL	 sctype = sc->specFlags & SF_MOVSECTOR || sc->ceilingpic == skyflatnum;
				int		 k;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (  (ss->sector->specFlags & 0xC) == 0xC
						&& ss->sector->ceilingpic != skyflatnum
						&& !sctype)
						SetOverride(ss, sc - sectors, 43036);
					else
						Setr3dssfs(ss, sc - sectors, 43037);
				}
			}
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		// Check rising sectors for masked hi tex's
		if (   sec->specFlags & SF_MASKEDDOOR
			&& sec->specFlags & SF_MOVCEIL
			&& sec->specFlags & SF_SSLINETRUE
			&& sec->floorheight >= sec->ceilingheight)
		{
			RDBOOL valid = true;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& !(ld->spflags & SPF_CONTROLLINE)
					&& !(origTEX[ld - lines] & 0x22)
					&& origTEX[ld - lines] & 0x44)
				{
					if (   !(vtoflags[ld->v1 - vertexes] & 1)
						&& !(vtoflags[ld->v2 - vertexes] & 1))
					{
						c = 0; break;
					}

					sd1 = &sides[ld->sidenum[0]];
					sd2 = &sides[ld->sidenum[1]];

					if ((sd1->toptexture > 0)^(sd2->toptexture > 0))
					{
						int txn = sd1->toptexture ? sd1->toptexture : sd2->toptexture;

						if (texArray[txn]->txf_flags & TXF_MASKED)
						{
							linearray[c++] = ld - lines;
						}
						else
						{
							valid = false;
						}
					}
				}
			}

			if (!c || valid) continue;

			for (k = 0; k < c; k++)
			{
				lines[linearray[k]].lnflags2 |= LNF2_NOALPHA;
			}
		}
	}

	chkMaskedDoorsAdjoining1sLines();	// Check in movsectors
}

//////////////////////////////////////////////////////////////
// b211-0 PS_UnsetSkysecFromMidtexLine
//
static void subfn_USFmidtexline(sector_t *s1, sector_t *s2, int lnum)
{
	sector_t	*s = NULL;
	ssline_t	*sl;
	vertex_t	*v1, *v2;
	line_t		*li;
	int			i, k, m, count = 0;
	int			secn = -1;

	CLR_hitline();
	hitline[lnum] = true;
	linearray[count++] = lnum;

	if (   secdone[s1 - sectors]
		|| s1->ceilingpic == skyflatnum
		|| (s1->specFlags & 0xFFFFF) != 0x808C)
	{
		return;	// Uncompliant
	}

	for (i = 0, sl = &sslines[lnum]; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];
		int			ch1, ch2;

		if (ss->sector == s2) continue;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( !(li = sg->linedef)
				|| hitline[li - lines]) continue;

			if ( !(li->ml_flags & ML_TWOSIDED)
				|| li->frontsector == li->backsector
				|| li->frontsector == s1 || li->backsector == s1
				|| li->frontsector->floorheight != li->backsector->floorheight
				|| (ch1 = li->frontsector->ceilingheight) == (ch2 = li->backsector->ceilingheight)
				|| origTEX[li - lines] & 0x33)
			{
				return;	// Invalid line type
			}

			hitline[li - lines] = 1;

			if (origTEX[li - lines])	// Has an upper texture
			{
				if (MIN_OF(ch1, ch2) < s1->ceilingheight)
				{
					return;	// Invalid line type
				}
				linearray[count++] = li - lines;
			}
			else if (li->frontsector->ceilingpic == li->backsector->ceilingpic
				  && li->frontsector->ceilingpic == skyflatnum)
			{
				sector_t *chk = ch1 < ch2 ? li->frontsector : li->backsector;

				if (!s) s = chk;
				else if (s != chk) return;

				if ((s->specFlags & 0xFFFFF) != 0x1C)
				{
					return;	// Invalid line type
				}

				linearray[count++] = li - lines;
				hitline[li - lines] = 2;
			}
			else
			{
				return;	// Invalid line type
			}
		}
	}

	if (!s) return;

	// v220-6 added. Check if contained by several sectors. If so the correct
	//               one must be derived as light levels could be different.
	{
		int ht, m, n, sn, fs, bs;

		// Find a valid sector from the vertex owners

		m = lines[lnum].v1 - vertexes;
		n = vertexowners[m].num;

		if (n <= 2)
		{
			m = lines[lnum].v2 - vertexes;
			n = vertexowners[m].num;
		}

		if (n <= 2) return;	// Can't derive anything from this line

		// Set the sector idx's of the test line
		fs = lines[lnum].frontsector - sectors;
		bs = lines[lnum].backsector  - sectors;

		// Get the lowest height - the sector we want must have this height
		ht = MIN_OF(sectors[fs].ceilingheight, sectors[bs].ceilingheight);

		for (i = 0; i < n; i++)
		{
			sn = vertexowners[m].list[i];

			if (sn == fs || sn == bs) continue;			// eliminate the test line's sectors

			if (sectors[sn].ceilingheight != ht			// eliminate a sector by ceiling height
			||  sectors[sn].specFlags & SF_NOHITEX)		// This is no use
			{
				continue;
			}

			if (secn < 0) secn = sn;
			else
			if (secn >= 0 && secn != sn) {secn = -1; break; }	// Can't make sense of this
		}

		if (secn < 0)
			secn = s - sectors;	// Assume it's the one we started with
	}

	for (i = 0; i < count; i++)
	{
		sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->sector != s2 && ss->ceilsec < 0)
				Setr3dsscs(ss, secn, 43038);
		}
		
		if (!origTEX[linearray[i]])
			Settype(&lines[linearray[i]], R3D_3F, 43039);
	}

	for (i = 0; i < s->linecount; i++)
	{
		li = s->lines[i];

		v1 = li->v1, v2 = li->v2;

		for (k = 0; k < count; k++)
		{
			int n = linearray[k];

			if (   hitline[n] == 2
				&& !hitline[li - lines]
				&& n != li - lines
				&& (lines[n].v1 == v1
				||  lines[n].v1 == v2
				||  lines[n].v2 == v1
				||  lines[n].v2 == v2) )
			{
				hitline[li - lines] = true;
				sl = &sslines[li - lines];

				Settype(li, R3D_3F, 43040);

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					if (ss->sector != s2 && ss->ceilsec < 0)
						Setr3dsscs(ss, secn, 43041);
				}
			}
		}
	}
}

void PS_UnsetSkysecFromMidtexLine()
{
	line_t	*ld;
	int		i, diff;

	// Look for the "seed" line
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->specType == R3D_03	// so should have midtextures
			&& !lineflags[i]
			&& Vtfl3(ld)
			&& ld->lnflags2 & LNF2_OWNGTRTHANTWO
			&& ld->frontsector != ld->backsector
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& (diff = ld->frontsector->ceilingheight - ld->backsector->ceilingheight))
		{
			sector_t *s1, *s2;

			if ((diff < 0 && (s1 = ld->frontsector)->ceilingpic != skyflatnum
				 && (s2 = ld->backsector)->ceilingpic == skyflatnum)
				 ||
				 (diff > 0 && (s1 = ld->backsector)->ceilingpic != skyflatnum
				 && (s2 = ld->frontsector)->ceilingpic == skyflatnum))
			{
				subfn_USFmidtexline(s1, s2, i);
			}
		}
	}
}

static void iterateWaterSecs(sector_t *sec)
{
	int i, fh, bh;

	if (sec->heightsec < 0)
		Con_Error("Program error: sec %i heightsec iterate", sec-sectors);

	sec->scvalidcount = -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED))
		{
			if (  !ld->frontsector->scvalidcount
				&& (fh = ld->frontsector->heightsec) >= 0
				&& sectors[fh].floorheight
				>= sectors[fh].ceilingheight)
			{
				ld->frontsector->specFlags2 |= SF2_ADJOINSWATER;
				iterateWaterSecs(ld->frontsector);
			}
		}
		else if ( (fh = ld->frontsector->heightsec) >= 0
			   && (bh = ld->backsector->heightsec ) >= 0)
		{
			if (  !ld->frontsector->scvalidcount
				&& sectors[fh].floorheight
				>= sectors[fh].ceilingheight)
			{
				ld->frontsector->specFlags2 |= SF2_ADJOINSWATER;
				iterateWaterSecs(ld->frontsector);
			}

			if (  !ld->backsector->scvalidcount
				&& sectors[bh].floorheight
				>= sectors[bh].ceilingheight)
			{
				ld->backsector->specFlags2 |= SF2_ADJOINSWATER;
				iterateWaterSecs(ld->backsector);
			}
		}
	}
}

// v220-3 Used to identify translucent water sectors
void PS_SetTranslucentWaterFlags()
{
	sector_t *sec;
	int		 i, c;

	for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		sec->scvalidcount = 0;

		if (sec->heightsec >= 0)
		{
			if (  R_CheckFlatFlag(sectors[sec->heightsec].ceilingpic, FLT_WATER)
				|| (R_CheckFlatFlag(sectors[sec->heightsec].floorpic, FLT_WATER)
				&& sectors[sec->heightsec].floorpic == sec->floorpic))
			{
				sectors[sec->heightsec].specFlags2 |= SF2_HASWATERFLATS;
				countsec[c++] = i;
				sec->scvalidcount = -1;
			}
			else if (sectors[sec->heightsec].ceilingpic == skyflatnum
				&& R_CheckFlatFlag(sectors[sec->heightsec].floorpic, FLT_WATER))
			{
				sectors[sec->heightsec].specFlags2 |= SF2_HASWATERFLATS;
				sec->scvalidcount = -1;
			}
		}
	}

	// v220-3 Look for sectors that do not have waterflats
	//        but which are connected to those that have.
	for (i = 0; i < c; i++)
	{
		iterateWaterSecs(&sectors[countsec[i]]);
	}
}

//////////////////////////////////////////////////////////////
// b211-0
// 1) Deals with moving doors in fake sectors
// 2) Deals with near heightless static sectors used for blocking
void PS_SectorHeightChecks()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		// Check for sector that could move below a fake sector floor
		if (   sec->specFlags & SF_MOVFLOOR
			&& sec->heightsec < 0)
		{
			sector_t	*s, *os = NULL;
			int			box[4], c_reject = 0;
			//--------------------------------

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->spflags & SPF_CONTROLLINE))
				{
					s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (s == sec) continue;

					if (s)
					{
						if (s->heightsec >= 0)
						{
							if (!os && IsLiquidFloorFlat(s)) os = s;
							if (os == s) M_AddLineToBox(box, ld);
						}
						else	// v220-4 additional check added
						{
							linearray[c_reject++] = ld - lines;
						}
					}
				}
			}

			if (os)
			{
				if (c_reject)	// v220-4 additional check added
				{
					for (k = 0; k < c_reject; k++)
					{
						if ( !M_FitBox(lines[linearray[k]].bbox, box) )
						{
							os = NULL; break;
						}
					}
				}

				if (os)
				{
					sec->oversec = os - sectors;
				}
			}
		}

		// Inhibit fake sector in special circumstances E.g. Beluga sec 767
		// This is used where a sector with a moving ceiling is faked but the textures
		// are masked hi tex's in a non-faked non-moving sector spanning 1's lines.
		//  Note: This type should also, a priori, had the adjacentsec setting invoked.
		if (   sec->heightsec >= 0
			&& (sec->specFlags & (SF_MOVCEIL|SF_MOVFLOOR)) == SF_MOVCEIL
			&& !IsLiquidFloorFlat(sec)
			&& !IsLiquidCeilFlat(&sectors[sec->heightsec]))
		{
			int	txn;
			int	valid = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (ld->adjacentsec >= 0)
				{
					valid |= 1;
				}
				
				if (!(ld->ml_flags & ML_TWOSIDED)) continue;

				if (origTEX[ld - lines] & 0x22)
				{
					valid = 0; break;
				}						

				if (   origTEX[ld - lines] & 0x44
					&& (ld->frontsector->heightsec >= 0)^(ld->backsector->heightsec >= 0)
					&& (ld->frontsector->specFlags & SF_MOVCEIL)^(ld->backsector->specFlags & SF_MOVCEIL)
					&& Spans1sLines(ld))
				{
					side_t *sd;

					if (   (sd = &sides[ld->sidenum[0]])->sector != sec
						&& (txn = sd->toptexture) )
					{
						if (texArray[txn]->txf_flags & TXF_MASKED)
						{
							valid |= 2;
						}
						else
						{
							valid = 0; break;
						}
					}
					if (   (sd = &sides[ld->sidenum[1]])->sector != sec
						&& (txn = sd->toptexture) )
					{
						if (texArray[txn]->txf_flags & TXF_MASKED)
						{
							valid |= 2;
						}
						else
						{
							valid = 0; break;
						}
					}
				}
			}

			if (valid == 3)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					int		 m, lnum = sec->lines[k] - lines;
					ssline_t *sl = &sslines[lnum];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec)
						{
							ss->ssflags |= SSF_IGNOREFAKE;
						}
					}
				}
			}
		}

		// Check for near heightless sector with hi and lo tex
		if (  !secdone[i]
			&& (sec->specFlags & 0xFBFFF) == 0x8390
			&& sec->ceilingheight - sec->floorheight < 4*FRACUNIT)
		{
			int		ht = DDMAXINT;
			RDBOOL	valid = true;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
				{
					valid = false; break;
				}

				if (ld->ml_flags & ML_TWOSIDED)	// should be as the SF_SSLINETRUE is unset
				{
					sector_t *s = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (s->ceilingheight < ht
						&& !(s->specFlags & SF_MOVSECTOR))
					{
						ht = s->ceilingheight;
					}
				}
			}

			if (   valid
				&& ht != DDMAXINT
				&& ht > sec->ceilingheight
				&& SecUnset(sec))
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (   ld->ml_flags & ML_TWOSIDED	// insurance
						&& origTEX[ld - lines] != 0x44)
					{
						if ( (ld->frontsector == sec && origTEX[ld - lines] & 0x04)
						||   (ld->backsector  == sec && origTEX[ld - lines] & 0x40) )
						{
							Settype(ld, R3D_07, 43042);
						}
						else if (!(origTEX[ld - lines] & 0x44))
							ld->ml_flags |= ML_IGNOREMISSINGHI;
					}
				}

				Setr3dch_byH(sec, ht, 43043);
				Setsecdone(sec, 43044);
			}
		}
	}
}

//////////////////////////////////////////////////////////////
// b211-0
// Checks for unset, unclosed sectors with no hi or lo textures
//
void PS_CheckUnclosedNoHILOtexSectors()
{
	sector_t *sec, *chk;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& (sec->specFlags & 0xFFFFF) == 0x818C)
		{
			sector_t	*os = NULL;
			int			box[4];
			RDBOOL		valid = true;

			M_ClearBox(box);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t	*ld = sec->lines[k];
				int		n = ld - lines;
				int		t = origTEX[n];

				if (   lineflags[n]
					|| !(ld->ml_flags & ML_TWOSIDED)
					|| ld->frontsector == ld->backsector
					|| (t != 0x20 && t != 0x02))
				{
					valid = false; break;
				}

				chk = t == 0x02 ? ld->frontsector : ld->backsector;

				if (!os)
				{
					os = chk;

					if ( !(os->floorheight < sec->floorheight)
						|| os->ceilingheight != sec->ceilingheight)
					{
						valid = false; break;
					}
				}
				else if (os != chk)
				{
					valid = false; break;
				}

				M_AddLineToBox(box, ld);
			}

			if (os && valid)
			{
				int		c = 0;

				LS_SectorsFromLinesInBox(box, countsec, &c);

				for (k = 0; k < c; k++)
				{
					if (   countsec[k] == i									// same sector inside is invalid
						|| (sectors[countsec[k]].specFlags & 0xC) != 0xC)	// or sectors with hi/lo textures
					{
						valid = false; break;
					}
				}

				if (valid)
				{
					subsector_t	*ss;
					float		fpbox[4];
					
					M_SetFPBoxFromIntBox(fpbox, box);

					for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
					{
						if (M_CompareFPxyBox(ss->midpoint.x, ss->midpoint.y, fpbox))
						{
							Clear_SSEC(ss);
							SetOverride(ss, os - sectors, 43045);
						}
					}

					for (k = 0; k < sec->linecount; k++)
					{
						line_t	*ld = sec->lines[k];
						
						Settype(ld, R3D_1D, 43046);
						SetLFG(ld, LFG_NEWSET, 43047);

						// These might have had a low tex set
						checkULTEX(ld, 1);
						sides[ld->sidenum[0]].bottomtexture = 0;
						sides[ld->sidenum[1]].bottomtexture = 0;
					}

					Setsecdone(sec, 43048);
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////
// b211-0
//  Checks for Legacy type unset,paired sectors with no hi or
//  lo textures that use a light level difference of 1 to
//  determine their status;

typedef struct {
	int			snum;
	int			scnt;
	int			sptr[4];
	RDBOOL		invalid;
} svar_t;

static svar_t	 *svt;

static void chknonCtrl(sector_t *sec, sector_t *chk, int *lo, int *hi)
{
	line_t	*t;
	int		i;

	*hi = -1, *lo = -1;

	for (i = 0, t = NULL; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[ld - lines]
			|| origTEX[ld - lines]
			|| ld->specType >= R3D_40
			|| (ld->frontsector != chk && ld->backsector != chk) )
		{
			return;
		}
		if (!t) t = ld;
	}

	if (t) XS_GetBoundsFromLine(t, lo, hi, NULL);
}

static void setSVAR(int fs, int bs, int snum, int *count)
{
	int		i;
	svar_t	*sc;

	if (!svt)
	{
		svt = (svar_t*)W_Calloc(sizeof(svar_t), STATICID, 0);
		sc = svt; (*count)++;
		sc->snum = snum;
	}
	else
	{
		RDBOOL found = false;

		for (i = 0; i < *count; i++)
		{ 
			sc = &svt[i];

			if (sc->snum == snum)
			{
				found = true; break;
			}
		}
		if (!found)
		{
			svt = (svar_t*)W_ReallocCLR(svt, sizeof(svar_t) * (*count + 1), STATICID, 0);
			sc = &svt[(*count)++];
			sc->snum = snum;
		}
	}

	if (sc->invalid) return;

	if (sc->scnt > 2) { sc->invalid = true; return; }

	if (sc->scnt)
	{
		// Split line?
		if (   (sc->sptr[0] == fs && sc->sptr[1] == bs)
			|| (sc->sptr[1] == fs && sc->sptr[0] == bs) )
		{
			return;
		}

		if (   sc->sptr[0] == fs || sc->sptr[1] == fs
			|| sc->sptr[0] == bs || sc->sptr[1] == bs )
		{
			sc->invalid = true;
			return;
		}
	}

	sc->sptr[sc->scnt++] = fs;
	sc->sptr[sc->scnt++] = bs;
}

// Strange construction used in AV map31 secs 506, 507, 508. 509
// It seems to rely on two defining control loops with a light diff of one
void PS_CheckCTRLSpecial()
{
	sector_t *fs, *bs;
	line_t	 *ld;
	int		 i, k, m, n, scount = 0;

	svt = NULL;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->spflags & SPF_CONTROLLINE
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->stptr < 0
			&& bs->stptr < 0
			&& abs(fs->lightlevel - bs->lightlevel) == 1
			&& !secdone[fs - sectors]
			&& !secdone[bs - sectors]
			&& (fs->specFlags & 0xFFBEC) == 0xC		// v220-9 changed from 0xFFFEF
			&& (bs->specFlags & 0xFFBEC) == 0xC)	// v220-9 changed from 0xFFFEF
		{
			int hi1, lo1, hi2, lo2;

			chknonCtrl(fs, bs, &hi1, &lo1);

			if (hi1 > 0 && hi1 == lo1)
			{
				chknonCtrl(bs, fs, &hi2, &lo2);

				if (hi2 == hi1 && lo2 == lo1)
				{
					setSVAR(fs - sectors, bs - sectors, hi1, &scount);
				}
			}
		}
	}

	if (scount)
	{
		svar_t	*sc;

		ss_validcount++;

		for (i = 0, sc = svt; i < scount; i++, sc++)
		{
			if (!sc->invalid && sc->scnt == 4)
			{
				// Get the hifloor sector
				int highest = sc->sptr[0];

				for (k = 1; k < sc->scnt; k++)
				{
					if (sectors[sc->sptr[k]].floorheight > sectors[highest].floorheight)
						highest = sc->sptr[k];
				}

				for (k = 0; k < sc->scnt; k++)
				{
					sector_t *s = &sectors[sc->sptr[k]];

					Setsecdone(s, 43049);

					for (m = 0; m < s->linecount; m++)
					{
						ld = s->lines[m];

						if (!(ld->spflags & SPF_CONTROLLINE))
						{
							ssline_t *sl = &sslines[ld - lines];

							SetLFG(ld, LFG_NEWSET, 43050);

							for (n = 0; n < sl->num; n++)
							{
								subsector_t *ss = &subsectors[sl->list[n]];

								if (ss->ssvalidcount == ss_validcount) continue;

								ss->ssvalidcount = ss_validcount;

								Setr3dsscs(ss, sc->snum, 43051);
								Setr3dssfs(ss, highest, 43052);
								Setr3dsslfsec(ss, sc->snum, 43053);
							}
						}
					}
				}
			}
		}
		W_Free(svt, STATICID, 0);
	}
}

// b211-0 modified
void PS_CheckLiqLines()
{
	sector_t *fs, *bs;
	line_t	 *ld;
	int		 i, k;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (!(ld->ml_flags & ML_TWOSIDED)) continue;

		if (   !lineflags[i]
			&& ld->specType == R3D_00
			&& ld->lnflags2 == 0x0C000000
			&& !(origTEX[i] & 0x33)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& (fs->specFlags & SF_MOVFLOOR) ^ (bs->specFlags & SF_MOVFLOOR)
			&& fs->heightsec < 0
			&& bs->heightsec < 0
			&& (vtoflags[ld->v1 - vertexes] & 1
			||  vtoflags[ld->v2 - vertexes] & 1) )
		{
			RDBOOL	l1 = IsLiquidFloorFlat(ld->frontsector);
			RDBOOL	l2 = IsLiquidFloorFlat(ld->backsector);

			if (l1 ^ l2)
			{
				sector_t *s1 = l1 ? fs : bs;
				sector_t *s2 = l1 ? bs : fs;

				if (s1->floorheight > s2->floorheight)
				{
					ssline_t *sl = &sslines[i];

					Settype(ld, R3D_08, 43054);		// Purely for debug purposes
					SetLFG(ld, LFG_NEWSET, 43055);
					
					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->sector != s1
							&& (ss->floorsec < 0 || ss->floorsec == ss->sector - sectors)
							&& ss->lfsec < 0)
						{
							Setr3dsslfsec(ss, s1 - sectors, 43056);
						}
					}
				}
			}
		}
	}
}

static RDBOOL getVTXowner(int s, vertexowner_t *vo, int *secn)
{
	int		i;

	if (vo->num != 2) return vo->num == 1;

	for (i = 0; i < 2; i++)
	{
		if (vo->list[i] != s)
		{
			if (*secn == -1)
			{
				*secn = vo->list[i];

				if (sectors[*secn].specFlags & SF_MOVSECTOR) return false;
			}
			else
				return vo->list[i] == *secn;
		}
	}

	return true;
}

static int chkSecLfsec(int secidx)
{
	sector_t *s = &sectors[secidx];
	int		 i, k;

	for (i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->sector == s && ss->lfsec >= 0)
					return ss->lfsec;
			}
		}
		break;
	}
	return -1;
}

static RDBOOL all2SandCTRL(sector_t *s)
{
	ssline_t *sl;
	int		 rsec, secn = -1;
	int		 hisec, losec, rf;
	int		 c, i, k, n;

	for (c = i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			if (ld->frontsector != ld->backsector)
				return -1;
			linearray[c++] = ld - lines;
		}
	}

	if (!c) return false;

	ss_validcount++;
	hitssecptr = 0;

	// Check if any have been set
	for (i = 0, rsec = -1; i < c; i++)
	{
		sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				if (ss->ssflags & SSF_RENDLFSEC) return false; // v220-4

				if ( (n = ss->floorsec) == ss->ceilsec && n >= 0)
				{
					if (rsec < 0)
						rsec = n;
					else if (rsec != n)
						return false;

					linearray[i] = -1;	// ignore from this point
				}
				hitssec[hitssecptr++] = sl->list[k];
			}
		}
	}

	if (rsec < 0) return -1;

	for (i = 0; i < c; i++)
	{
		if ( (n = linearray[i]) < 0) continue;

		XS_GetBoundsFromLine(&lines[n], &losec, &hisec, &rf);

		if (hisec != losec || hisec != rsec) return false;
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		if (ss->floorsec < 0)
			SetOverride(ss, rsec, 43162);
		if (ss->lfsec >= 0)
			Setr3dsslfsec(ss, -1, 43162);
	}

	return true;
}

static int check9809Clines(sector_t *sec, int *ring1 , int *ring2,
							  int *count1, int *count2, int *box1, int *box2)
{
	sector_t *s, *chk = NULL, *isec = NULL;
	int		 c1 = 0, c2 = 0;
	int		 i, k;

	M_ClearBox(box1); M_ClearBox(box2);

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		int		lnum = ld - lines;

		if (   !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
			|| !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[lnum]
			|| !Vtxs2(ld)
			|| origTEX[lnum] & 0x55
			|| lxarray[lnum])
		{
			return -1;
		}

		s = ld->frontsector == sec ? ld->backsector : ld->frontsector;


		if ((s->specFlags & 0x3001C) == 0x1C)
		{
			if (!isec)
			{
				if (SecCtrl(s - sectors)) return -1;
				isec = s;
			}
			else if (isec != s)
			{
				return -1;
			}

			ring1[c1++] = lnum;
			M_AddLineToBox(box1, ld);
		}
		else if ((s->specFlags & 0xFFFFFC) == 0x8198)
		{
			if (!chk)
			{
				if (SecCtrl(s - sectors)) return -1;
				chk = s;
			}
			else if (chk != s)
			{
				return -1;
			}

			ring2[c2++] = lnum;
			M_AddLineToBox(box2, ld);
		}
		else
			return -1;
	}

	*count1 = c1, *count2 = c2;

	if (   c1 > 2
		&& c2 > 2
		&& chk->floorheight == isec->floorheight
		&& sec->linecount == c1 + c2
		&& M_CompareBox(box1, box2))
	{
		hitssecptr = 0;
		ss_validcount++;

		for (i = 0; i < sec->linecount; i++)
		{
			ssline_t *sl = &sslines[sec->lines[i] - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					ss->ssvalidcount = ss_validcount;

					if (ss->sector == sec)
					{
						if (ss->floorsec >= 0) return -1;
						hitssec[hitssecptr++] = sl->list[k];
					}
				}
			}
		}
		return chk - sectors;
	}

	return -1;
}

static int SectorContainingUnclosed(sector_t *s)
{
	ssline_t *sl;
	int		 i, k, n = LS_GetContainingSecnum(s, 1);

	if (   n < 0
		|| sectors[n].specFlags & SF_MOVSECTOR
		|| sectors[n].stptr >= 0
		|| n == s - sectors
		|| n != XS_GetBoundarySecnum(s))	// Double agreement
	{
		return -1;
	}

	hitssecptr = 0;
	validcount++;
	ss_validcount++;

	for (i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		if (   ld->frontsector != ld->backsector
			|| lineflags[ld - lines]
			|| ld->specType >= R3D_40)
		{
			return -1;
		}

		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				if (ss->floorsec >= 0 || ss->ceilsec >= 0)
					return -1;

				if (ss->sector - sectors != n)
					hitssec[hitssecptr++] = sl->list[k];
			}
		}
	}
	return n;
}

// v220-3
static void nestedCeilSecs(sector_t *sec)
{
	vertexowner_t	*vx[3];
	sector_t		*newsec = NULL;
	float			*fbox;
	int				cv, i, k, m;

	for (cv = i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];
		vertexowner_t *v1 = &vertexowners[ld->v1 - vertexes];
		vertexowner_t *v2 = &vertexowners[ld->v2 - vertexes];

		if ( !(ld->ml_flags & ML_TWOSIDED) ) return;

		if (ld->frontsector->ceilingheight == ld->backsector->ceilingheight) continue;

		newsec = ld->frontsector->ceilingheight < ld->backsector->ceilingheight ?
			ld->frontsector : ld->backsector;

		if (v1->num == 2 && v2->num > 2)
		{
			if (cv > 1) { cv = 0; break; }
			vx[cv++] = v2; continue;
		}

		if (v1->num > 2 && v2->num == 2)
		{
			if (cv > 1) { cv = 0; break; }
			vx[cv++] = v1; continue;
		}
	}

	// All else being equal check that sec fits in newsec
	// From then on check all other selected sectors fit in sec.
	if (   cv == 2
		&& newsec
		&& vx[0]->num == vx[1]->num
		&& M_FitFPBox( (fbox = sec->sibounds), newsec->sibounds))
	{
		int		*secn = (int*)M_Malloc(vx[0]->num * sizeof(int), 0);

		for (cv = i = 0; i < vx[0]->num; i++)
		{
			if (vx[0]->list[i] != vx[1]->list[i]) { cv = 0; break; }

			if (  (sectors[vx[0]->list[i]].specFlags & 0xFFFFC) == 0x8198
				&& sectors[vx[0]->list[i]].ceilingheight > newsec->ceilingheight
				&& M_FitFPBox(sectors[vx[0]->list[i]].sibounds, fbox))
			{
				secn[cv++] = vx[0]->list[i];
			}
		}

		if (cv)
		{
			for (i = 0; i < cv; i++)
			{
				sector_t *s = &sectors[secn[i]];

				for (k = 0; k < s->linecount; k++)
				{
					line_t *ld = s->lines[k];
					ssline_t *sl = &sslines[ld - lines];

					SetLFG(ld, LFG_NEWSET, 43160);

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ceilsec < 0 && ss->sector == s)
							Setr3dsscs(ss, newsec - sectors, 43160);
					}
				}

				Setsecdone(s, 43160);
			}
		}

		M_Free(secn);
	}
}

static RDBOOL chkUnsetSector(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[ld - lines]) return false;
	}

	return true;
}

void PS_CheckGenInheritance()
{
	sector_t *s, *sec;
	line_t	 *ld;
	int		 c, i, k, m, n;
	RDBOOL	 valid;
	int		 *hitssecmem;
	int		 hitssecmemptr;
	int		 secn;

	//-------------------

	hitssecmem = (int*)W_Malloc(numsubsectors * sizeof(int), STATICID, 7);

// Check line types in mov ceil sectors $v91
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		int mask;

		if (secdone[i] || sec->stptr >= 0)
		{
			continue;
		}

		mask = sec->specFlags & 0xFFFFC;

		if (mask == 0x9888C)
		{
			if (SecCtrl(i) && all2SandCTRL(sec))
				Setsecdone(sec, 43163);

			continue;
		}
		
		if (!chkUnsetSector(sec)) continue;

		// v220-9 movfloor loop with missing tex (1 not missing allowed)
		if ((sec->specFlags & 0xFFFF4) == 0x18390 && IsLiquidFloorFlat(sec))
		{
			sector_t *os, *is;
			fixed_t  fh = sec->floorheight;

			for (c = k = n = 0, os = NULL; k < sec->linecount; k++)
			{
				if ((ld = sec->lines[k])->spflags & SPF_CONTROLLINE) continue;

				if (lineflags[ld - lines]) { os = NULL; break; }

				is = NULL; linearray[n++] = ld - lines;

				if (ld->mtflags == 0x10 || ld->mtflags == 0x01)
				{
					is = ld->frontsector == sec ? ld->backsector : ld->frontsector;
				}

				if (is && is->floorheight > fh)
				{
					if (!os)
					{
						if (is->specFlags & SF_MOVFLOOR) break;
						os = is;
					}
					else if (is != os)
					{
						os = NULL; break;
					}
					c++;
				}
			}

			if (   os
				&& c > 2
				&& (c == n || n - c == 1)
				&& (subfn_CheckLoopClosed(0, n, linearray, NULL, NULL) == n)
				&& !LS_LinesInsideLoop(sec, n))
			{
				for (k = 0; k < hitssecptr; k++)
				{
					Setr3dsslfsec(&subsectors[hitssec[k]], os - sectors, 43161);
				}

				for (k = 0; k < sec->linecount; k++)
					SetLFG(sec->lines[k], LFG_NEWSET, 43161);

				Setsecdone(sec, 43161);
			}
		}

		// v220-3 check for a floor and ceil mov sector containing a static
		//        textureless sector which has not had inheritance set (rare)
		if (mask == 0x3001C && SecCtrl(i))
		{
			sector_t *os = NULL, *is = NULL;
			int		 inherit = -1;

			for (c = k = 0; k < sec->linecount; k++)
			{
				if ((ld = sec->lines[k])->spflags & SPF_CONTROLLINE) continue;

				if (lineflags[ld - lines])
				{
					os = NULL; break;
				}

				s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (!s || s == sec) continue;

				if ((s->specFlags & 0xFFFFC) == 0x1C)
				{
					if (!os) os = s;
					else if (os != s) { os = NULL; break; }
					linearray[c++] = ld - lines;
				}
				else
				{
					if (!is) is = s;
					else if (is != s) { is = NULL; break; }
				}
			}

			if (os && is && c == os->linecount)
			{
				int		 inherit = -1;

				hitssecptr = 0;
				ss_validcount++;

				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						if (ss->sector == os)
						{
							// Check all references are unset
							if (ss->ceilsec >= 0 || ss->floorsec >= 0)
							{
								hitssecptr = 0; break;
							}

							hitssec[hitssecptr++] = sl->list[m];
						}
						else 
						{
							// This should already have been set via specset in
							// missingtex.c (11018 - transferred at point 23036)
							if (ss->ceilsec < 0 || ss->floorsec < 0 || ss->ceilsec != ss->floorsec)
							{
								hitssecptr = 0; break;
							}
							else
							{
								if (inherit < 0) inherit = ss->ceilsec;
								else if (inherit != ss->ceilsec)
								{
									hitssecptr = 0; break;
								}
							}
						}
					}

					if (!hitssecptr) break;
				}

				if (hitssecptr)
				{
					for (k = 0; k < hitssecptr; k++)
						SetOverride(&subsectors[hitssec[k]], inherit, 43164);

					Setsecdone(os, 43164); Setsecdone(is, 43164);

					for (k = 0; k < c; k++)
						SetLFG(&lines[linearray[k]], LFG_NEWSET, 43164);
					
					continue;
				}
			}
		}

		// v220-3 check for unset moving support sectors
		if (mask == 0x1000C && SecCtrl(i))
		{
			for (c = k = 0; k < sec->linecount; k++)
			{
				if ((ld = sec->lines[k])->spflags & SPF_CONTROLLINE) continue;

				if (   lineflags[ld - lines]
					|| ld->frontsector != ld->backsector)
				{
					c = 0; break;
				}
				c++;
			}

			if (c)
			{
				int c1 = XS_GetBoundarySecnum(sec);

				if (   c1 >= 0
					&& c1 != i
					&& !(sectors[c1].specFlags & (SF_MOVSECTOR|SF_NOLOTEX))
					&& c1 == LS_GetContainingSecnum(sec, 1))	// Double check
				{
					for (c = k = 0; k < sec->linecount; k++)
					{
						ssline_t *sl = &sslines[(ld = sec->lines[k]) - lines];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->floorsec < 0 && ss->ceilsec < 0)
							{
								SetOverride(ss, c1, 43165);
							}
						}

						lineflags[ld - lines] |= LFG_NEWSET;
					}
				}

				Setsecdone(sec, 43165); continue;
			}
		}

		if (mask == 0x9809C)
		{
			int count1 ,  count2;
			int box1[4], box2[4];
			int *ring1 ,  *ring2;

			if (SecCtrl(i)) continue;

			ring1 = (int*)M_Malloc(sizeof(int)*sec->linecount, 0);
			ring2 = (int*)M_Malloc(sizeof(int)*sec->linecount, 0);

			secn = check9809Clines(sec, ring1, ring2, &count1, &count2, box1, box2);

			if (secn >= 0)
			{
				for (k = 0; k < hitssecptr; k++)
				{
					Setr3dssfs(&subsectors[hitssec[k]], secn, 43166);
				}

				Setsecdone(sec, 43166);
				Setsecdone(&sectors[secn], 43166);
			}

			M_Free(ring1); M_Free(ring2);

			continue;
		}

		if (mask == 0x181BC)
		{
			int	a, b = -1, d;

			for (c = d = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) { d++; continue; }

				if (   ld->frontsector != ld->backsector
					|| (m = vertexowners[ld->v1 - vertexes].num) > 2
					|| (n = vertexowners[ld->v2 - vertexes].num) > 2)
				{
					c = 0; break;
				}

				if (m > 1)
				{
					a = vertexowners[ld->v1 - vertexes].list[0];
					if (a == i) a = vertexowners[ld->v1 - vertexes].list[1];

					if (b < 0) b = a;
					else if (b != a) { c = 0; break; }
				}

				if (n > 1)
				{
					a = vertexowners[ld->v2 - vertexes].list[0];
					if (a == i) a = vertexowners[ld->v2 - vertexes].list[1];

					if (b < 0) b = a;
					else if (b != a) { c = 0; break; }
				}

				linearray[c++] = ld - lines;
			}

			if (   b >= 0
				&& sectors[b].specFlags & SF_MOVFLOOR
				&& sec->ceilingheight < sectors[b].ceilingheight
				&& c > 2
				&& c + d == sec->linecount)
			{
				int			rc, box[4];

				CLR_conarray();

				for (k = 0; k < c; k++) conarray[linearray[k]] = 1;

				// check for a contiguous area
				if (subfn_CheckForRing(conarray, sec, linearray[0], &rc, box, NULL, NULL)
					&& rc == c)
				{
					for (k = 0; k < c; k++)
					{
						ssline_t *sl = &sslines[linearray[k]];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector - sectors == i && ss->ceilsec < 0)
								Setr3dsscs(ss, b, 43167);
						}
					}

					Setsecdone(sec, 43167);
					Setsecdone(&sectors[b], 43167);
				}
			}
		}

		if (   mask == 0xC
			&& !SecCtrl(i)
			&& sec->stptr < 0
			&& (secn = SectorContainingUnclosed(sec)) >= 0
			&& (sectors[secn].specFlags & 0xFFF8C) == 0x8780)
		{
			// Continue even if hitssecptr == 0 in order to set the lineflags/secdone
			for (k = 0; k < hitssecptr; k++)
				SetOverride(&subsectors[hitssec[k]], secn, 43168);

			for (k = 0; k < sec->linecount; k++)
				SetLFG(sec->lines[k], LFG_NEWSET, 43168);

			Setsecdone(sec, 43168);
			continue;
		}

		if (   (sec->specFlags & 0x13FFFC) == 0x11C88C
			&& SecCtrl(i)
			&& (secn = XS_GetBoundarySecnum(sec)) >= 0
			&& sectors[secn].heightsec >= 0)
		{
			ssline_t *sl;

			for (k = 0; k < sec->linecount; k++)
			{
				if ( ((ld = sec->lines[k])->frontsector != ld->backsector || lineflags[ld - lines])
					&& !(ld->spflags & SPF_CONTROLLINE))
				{
					goto end_loop;
				}
			}

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				SetLFG(ld, LFG_NEWSET, 43059);
				
				sl = &sslines[ld - lines];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					SetOverride(ss, secn, 43060);
				}
			}
			Setsecdone(sec, 43061);
			continue;
		}

		if ((mask & 0x3001C) == 0x1C)
		{
			ssline_t *sl;

			for (k = 0; k < sec->linecount; k++)
			{
				if ( (ld = sec->lines[k])->frontsector != ld->backsector
					|| lxarray[ld - lines])
					goto bailout;
			}

			if(    (secn = LS_GetContainingSecnum(sec, 1)) >= 0
				&& (secn = chkSecLfsec(secn)) >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					SetLFG(ld, LFG_NEWSET, 43062);
					
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						Setr3dsslfsec(ss, secn, 43063);
					}
				}
				Setsecdone(sec, 43064);
				continue;
			}
		}

bailout:;
		// v220-3 Rising liquid floor sector
		if (   (sec->specFlags & 0xFFFFFFFC) == 0x00918198
			&& IsLiquidFloorFlat(sec)
			&& !SecCtrl(i)
			&& sec->heightsec < 0)
		{
			int secn = LS_GetContainingSecnum(sec, 1);

			if (   secn >= 0
				&& sectors[secn].floorheight > sec->floorheight
				&& sectors[secn].floorpic == sec->floorpic)
			{
				ssline_t	*sl;

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];
					sl = &sslines[ld - lines];

					Settype(ld, R3D_3F, 988);

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec && ss->floorsec < 0)
						{
							 Setr3dssfs(ss, secn, 988);
						}
					}
				}
			}
		}
		else
		// Floor move type inheritance check
		if (   (sec->specFlags & 0xFFFFFFFC) == 0x0181809C	// too specific?
			&& !SecCtrl(i)
			&& sec->heightsec < 0)
		{
			subsector_t	*ss;
			ssline_t	*sl;
			int			chk = -1;

			ss_validcount++;
			valid = true;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];
				sl = &sslines[ld - lines];

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];
					if (ss->ssvalidcount != ss_validcount)
					{
						ss->ssvalidcount = ss_validcount;

						if (ss->sector == sec)
						{
							if (ss->floorsec >= 0 || ss->ceilsec >= 0)
							{
								valid = false; break;
							}
							hitssec[c++] = sl->list[m];
						}
						else if (ss->floorsec >= 0)
						{
							if (chk < 0) chk = ss->floorsec;
							if (chk != ss->floorsec) { valid = false; break; }
						}
					}
				}
				if (!valid) break;
			}

			if (valid && c && chk >= 0)
			{
				for (k = 0; k < c; k++)
					Setr3dsslfsec(&subsectors[hitssec[k]], chk, 988);
			}
		}

		if (  !SecCtrl(i)
			&& sec->heightsec < 0
			&& sec->ceilingpic == skyflatnum
			&& (sec->specFlags & 0x00FFFFFC) == 0x00918198)
		{
			sector_t *ck, *os = NULL;

			for (valid = true, c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   lineflags[ld - lines]
					|| !(ld->ml_flags & ML_TWOSIDED)
					|| origTEX[ld - lines] & 0x67
					|| ld->mtflags & R3D_MISTEXANY
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| !Vtxs2(ld))
				{
					valid = false; break;
				}

				ck = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (!os) os = ck;

				if (   os != ck
					|| os == sec
					|| os->ceilingpic == skyflatnum
					|| os->specFlags & (SF_MOVSECTOR|SF_NOHITEX)
					|| os->ceilingheight > sec->ceilingheight)
				{
					valid = false; break;
				}

				linearray[c++] = ld - lines;
			}

			if (valid && os && c == sec->linecount)
			{
				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					ld = &lines[linearray[k]];
					
					Settype(ld, R3D_00, 43065);
					SetLFG(ld, LFG_NEWSET, 43066);

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec)
							Setr3dsscs(ss, os - sectors, 43067);
					}
				}
			}
		}
		else if (mask == 0x809C)
		{
			RDBOOL	valid = !SecCtrl(i)
				&& sec->ceilingheight - sec->floorheight == FRACUNIT;

			if (valid)
			{
				for (c = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					n = origTEX[ld - lines];

					if (   lineflags[ld - lines]
						|| n & 0x55
						|| ld->mtflags & R3D_MISHITEX)
					{
						valid = false; break;
					}

					if (   ld->ml_flags & ML_TWOSIDED
						&& ld->specType == R3D_00
						&& ld->mtflags & R3D_MISLOTEX
						&& n & 0x22)
					{
						s = ld->frontsector == sec ?
							ld->backsector : ld->frontsector;

						if (    s->specFlags & 0x3000C
							|| (s == ld->frontsector && n & 0x20)
							|| (s == ld->backsector  && n & 0x02))
						{
							valid = false; break;
						}

						linearray[c++] = ld - lines;
					}
				}
			}

			if (valid && c)
			{
				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					ld = &lines[linearray[k]];
					
					SetLFG(ld, LFG_NEWSET, 43068);

					Settype(ld, R3D_12, 43069);

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						SetOverride(ss, s - sectors, 43070);
					}
				}
			}
		}
		else if (mask == 0x8198)
		{
			if (   sec->ceilingpic != skyflatnum
				&& sec->R3Dceilheight == DDMININT
				&& !SecCtrl(i)
				&& sec->R3Dceilsec < 0
				&& sec->heightsec < 0)
			{
				sector_t	*chk;
				RDBOOL		valid2 = false;
				int			diff, box[4];

				// v220-3
				nestedCeilSecs(sec); if (secdone[i]) continue;

				valid = true; s = sec;

				// Look for line loops
				CLR_hitline();

				for (c = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (ld->specType != R3D_00 && origTEX[ld - lines])
					{
						valid = false; break;
					}

					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (  !chk
						|| chk == sec
						|| chk->specFlags & SF_MOVSECTOR
						|| chk->heightsec >= 0
						|| chk->stptr >= 0)
					{
						valid = false; break;
					}

					diff = abs(chk->ceilingheight - sec->ceilingheight);
					
					if (diff && diff <= 8*FRACUNIT)
					{
						valid = false; break;
					}

					if (   diff
						&& (asyflags[ld->sidenum[0]] & ASY_H
						||  asyflags[ld->sidenum[1]] & ASY_H))
					{
						hitline[ld - lines] = 1; c++;
						valid2 = true;
					}
					else
					{
						hitline[ld - lines] = 1; c++;
					}

					if (chk->ceilingheight < s->ceilingheight)
					{
						s = chk;
					}
				}

				if (!valid || !valid2 || s == sec)
				{
					// v220-9 there's another possibility here
					if (valid && sec->ceilingheight - sec->floorheight == FRACUNIT)
					{
						for (s = NULL, c = k = 0; k < sec->linecount; k++)
						{
							ld = sec->lines[k];

							if ((ld->mtflags == 0x10 || ld->mtflags == 0x40) && !origTEX[ld - lines])
							{
								chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

								if (!s)
								{
									if (!chk || chk->specFlags & 0x3000C) break;
									s = chk;
								}
								else if (chk != s
									&& (!chk
									|| s->specFlags & 0x3000C
									|| chk->specFlags & 0x3000C
									|| s->ceilingheight != chk->ceilingheight
									|| s->floorheight != chk->floorheight))
								{
									s = NULL; break;
								}

								linearray[c++] = ld - lines;
							}
						}

						if (s && c == 2)
						{
							Setr3dcs(sec, s - sectors, 43070);
							Setr3dfs(sec, s - sectors, 43070);

							for (k = 0; k < c; k++)
							{
								ld = &lines[linearray[k]];
								Settype(ld, R3D_3F, 43070);
								SetLFG(ld, LFG_NEWSET|LFG_FIXED, 43070);
							}
						}
					}
					continue;
				}
				
				hitssecmemptr = 0;

				for (c = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 1
						&& subfn_CheckForRing(hitline, sec, ld - lines, NULL, box, NULL, NULL))
					{
						if (LS_MarkSubsecsInAreaGeneral(sec, hitline, ld - lines, NULL, box, -1, MSIA_NONE))
						{
							memcpy(hitssecmem + hitssecmemptr, hitssec, hitssecptr * sizeof(int));
							hitssecmemptr += hitssecptr;
							c++;
						}

						for (m = 0; m < sec->linecount; m++)
						{
							ld = sec->lines[m];

							if (hitline[ld - lines] == 4) hitline[ld - lines] = 8;
						}
					}
				}

				for (k = 0; k < hitssecmemptr; k++)
				{
					if (subsectors[hitssecmem[k]].sector->ceilingheight < s->ceilingheight)
					{
						valid = false; break ;
					}
				}

				if (valid && hitssecmemptr)
				{
					for (k = 0; k < hitssecmemptr; k++)
						Setr3dsscs(&subsectors[hitssecmem[k]], s - sectors, 43071);

					for (k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k];

						if (hitline[ld - lines] == 8
							&& !(origTEX[ld - lines] & 0x33))
						{
							Settype(ld, R3D_3F, 43071);
							SetLFG(ld, LFG_NEWSET, 43072);
						}
					}

					Setsecdone(sec, 43073);
				}
			}
		}
		else if (mask == 0x80DC)		// E.g Eternal map12 sectors 661, 663
		{
			int		 h, fht = sec->floorheight;
			sector_t *fsec = NULL;

			if (sec->ceilingpic != skyflatnum) continue;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   lineflags[ld - lines]
					|| !(ld->ml_flags & ML_TWOSIDED)
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| !Vtxs2(ld))
				{
					c = 0; break;
				}

				linearray[c++] = ld - lines;

				if ( (h = ld->frontsector->floorheight) > sec->floorheight
					&& h > fht)
				{
					fht = h; fsec = ld->frontsector;
				}
				if ( (h =ld->backsector->floorheight) > sec->floorheight
					&& h > fht)
				{
					fht = h; fsec = ld->backsector;
				}
			}

			if (c == sec->linecount && fsec)
			{
				ss_validcount++;

				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec && ss->ssvalidcount != ss_validcount)
						{
							seg_t *sg = &segs[ss->firstline];

							ss->ssvalidcount = ss_validcount;

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (sg->frontsector && sg->frontsector != sec)
								{
									goto endMask1;
								}
							}
						}
					}
				}

				for (k = 0; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec && ss->floorsec < 0)
							Setr3dssfs(ss, fsec - sectors, 43074);
					}
				}
				Setsecdone(sec, 43075);
			}
endMask1:;
		}
		else if (mask == 0x28198)
		{
			int lnum, box[4];

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   lineflags[ld - lines]
					|| !(ld->ml_flags & ML_TWOSIDED)
					|| ld->specType != R3D_00
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| !Vtxs2(ld))
				{
					c = 0; break;
				}
				linearray[c++] = ld - lines;
			}

			if (c < 3) continue;

			CLR_hitline();

			for (k = 0; k < c; k++)
				hitline[linearray[k]] = 1;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k]; lnum = ld - lines;

				if (   hitline[lnum]
					&& subfn_CheckForRing(hitline, sec, lnum, NULL, box, NULL,NULL)
					&& LS_MarkSubsecsInAreaGeneral(sec, hitline, lnum, NULL, box, -1, MSIA_NONE))
				{
					for (m = 0, s = NULL; m < sec->linecount; m++)
					{
						ld = sec->lines[m]; lnum = ld - lines;

						if (hitline[lnum] == 4)
						{
							if (!s)
							{
								s = ld->frontsector == sec ?
									ld->backsector : ld->frontsector;
							}
							
							Settype(ld, R3D_3D, 43076);
							SetLFG(ld, LFG_NEWSET, 43077);
							hitline[lnum] = 8;
						}
					}

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];
					//	ss->ssflags |= SSF_NOOCCLUDE;
						Setr3dsscs(ss, s - sectors, 43078);
					}
				}
			}
		}
// Check unset SF_MOVFLOOR floor types $v93
		else if ((mask == 0x98F94 || mask == 0x18D88) && SecCtrl(i))
		{
			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (lineflags[ld - lines]) { c = 0; break; }

				s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (s && s != sec && s->specFlags & SF_MOVFLOOR) { c = 0; break; }

				linearray[c++] = ld - lines;
			}

			if (c)
			{
				sector_t *chk;

				ss_validcount++;

				hitssecptr = 0;

				for (k = 0, chk = NULL; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						if (ss->sector == sec)
							hitssec[hitssecptr++] = sl->list[m];
						else
							continue;

						if (!chk)
						{
							seg_t *sg = &segs[ss->firstline];

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (sg->frontsector == sec
									&& sg->backsector
									&& sg->backsector != sec
									&& sg->backsector->floorheight == sec->floorheight
									&& sg->backsector->floorpic == sec->floorpic
									&& !(sg->backsector->specFlags & SF_MOVFLOOR)
									&& (!(sg->backsector->specFlags & SF_NOLOTEX)
									||  sg->backsector->specFlags & SF_SSLINETRUE))
								{
									chk = sg->backsector; break;
								}
							}
						}
					}
				}

				if (chk)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						Setr3dssfs(&subsectors[hitssec[k]], chk - sectors, 43079);
					}
				}
			}
		}
		else if (mask == 0x1808C && SecCtrl(i))
		{
			sector_t *chk, *s = NULL;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if ( !(ld->ml_flags & ML_TWOSIDED) ) { s = NULL; break; }

				chk = sec == ld->frontsector ? ld->backsector : ld->frontsector;
				
				if (chk == sec) continue;

				if (!s)
				{
					s = chk;
					if (   s->specFlags & SF_MOVSECTOR
						|| (s->specFlags & 0xC) == 0xC)
					{
						s = NULL; break;
					}
				}
				else if (s != chk)
				{
					if (   chk->specFlags & SF_MOVSECTOR
						|| (chk->specFlags & 0xC) == 0xC
						|| chk->floorheight != s->floorheight
						|| chk->ceilingheight != s->ceilingheight)
					{
						s = NULL; break;
					}
				}

				if (   lineflags[ld - lines]
					|| ld->specType != R3D_14
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| origTEX[ld - lines])
				{
					s = NULL; break;
				}
				linearray[c++] = ld - lines;
			}

			if (s && c > 2)
			{
				int rc, box[4];

				CLR_hitline();

				for (k = 0; k < c; k++)
				{
					hitline[linearray[k]] = 1;
				}

				if (subfn_CheckForRing(hitline, sec, linearray[0], &rc, box, NULL, NULL)
					&& rc == c
					&& LS_MarkSubsecsInAreaGeneral(sec, NULL, c, NULL, box, -1, MSIA_NONE))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						SetOverride(&subsectors[hitssec[k]], s - sectors, 43080);
					}

					for (k = 0; k < c; k++)
					{
						Settype(&lines[linearray[k]], R3D_3F, 43081);
						SetLFG(&lines[linearray[k]], LFG_NEWSET, 43082);
					}
					Setsecdone(sec, 43083);
				}
			}
		}
		// v220-5 one that got missed
		else if (mask == 0x1888C
			&& !(sec->specFlags & 0xE0000000) )
		{
			ssline_t	*sl;
			int			cs;

			if (sec->heightsec >= 0) continue;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   ld->frontsector != ld->backsector
					|| ld->specType >= R3D_40
					|| lxarray[ld - lines]
					|| origTEX[ld - lines] & 0x55
					|| lineflags[ld - lines] & ~LFG_FIXED)
				{
					c = 0; break;
				}

				if (!(lineflags[ld - lines] & LFG_FIXED))
					c++;
			}

			if (   !c
				|| (cs = LS_GetContainingSecnum(sec, 1)) < 0
				|| cs == i)
			{
				continue;
			}

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				sl = &sslines[ld - lines];
				SetLFG(ld, LFG_NEWSET, 43084);

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					SetOverride(ss, cs, 43085);
				}
			}
			Setsecdone(sec, 43086);
		}
		else if (mask == 0x2808C
			&& !(sec->specFlags & SF_MASKEDDOOR)
			&& sec->ceilingheight == sec->floorheight)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				int txn;

				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   ld->ml_flags & ML_TWOSIDED
					&& ld->specType == R3D_00
					&& origTEX[ld - lines] & 0x02
					&& (s = ld->frontsector) != sec
					&& s->ceilingheight > s->floorheight
					&& (txn = sides[*ld->sidenum].midtexture)
					&& texArray[txn]->txf_flags & TXF_MASKED)
				{
					sec->specFlags |= SF_MASKEDDOOR;
					break;
				}
			}
		}
		// Check sector for movfloor sec
		else if (SecCtrl(i)
			&& mask == 0x1C8AC
			&& sec->specFlags & SF_MOVINVALID)
		{
			int		n, p;
			int		secn, diff;

			valid = true;

			for (c = k = 0, secn = -1; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   lineflags[ld - lines]
					|| ld->frontsector != ld->backsector)
				{
					valid = false; break;
				}

				if (   !getVTXowner(i, &vertexowners[ld->v1 - vertexes], &secn)
					|| !getVTXowner(i, &vertexowners[ld->v2 - vertexes], &secn) )
				{
					valid = false; break;
				}

				linearray[c++] = ld - lines;
			}

			if (c && valid && secn >= 0)
			{
				hitssecptr = 0;
				ss_validcount++;

				for (m = 0; m < c; m++)
				{
					ssline_t *sl = &sslines[linearray[m]];

					for (n = 0; n < sl->num; n++)
					{
						subsector_t *ss = &subsectors[sl->list[n]];
						seg_t		*sg = &segs[ss->firstline];

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						for (p = 0; p < ss->linecount; p++, sg++)
						{
							if (  (ld = sg->linedef)
								&& ld->ml_flags & ML_TWOSIDED
								&& (ld->frontsector != sec || ld->backsector != sec))
							{
								diff = ld->frontsector->floorheight - ld->backsector->floorheight;
								
								if (   (diff > 0 && !(origTEX[ld - lines] & 0x10))
									|| (diff < 0 && !(origTEX[ld - lines] & 0x01)) )
								{
									valid = false; break;
								}

								diff = ld->frontsector->ceilingheight - ld->backsector->ceilingheight;
								
								if (   (diff > 0 && !(origTEX[ld - lines] & 0x04))
									|| (diff < 0 && !(origTEX[ld - lines] & 0x40)) )
								{
									valid = false; break;
								}
							}

							hitssec[hitssecptr++] = ss - subsectors;
						}

						if (!valid) break;
					}
				}

				if (valid)
				{
					for (m = 0; m < hitssecptr; m++)
					{
						SetOverride(&subsectors[hitssec[m]], secn, 43087);
					}
					Setsecdone(sec, 43088);
				}
			}
		}
		else if ((sec->specFlags & 0xFFFFFFFC) == 0x00118188
			&& SecCtrl(i)
			&& sec->heightsec < 0
			&& sec->stptr < 0
			&& !IsLiquidFloorFlat(sec))
		{
			for (c = k = 0; k < sec->linecount; k++)
			{
				sector_t	*fs, *bs;
				line_t		*ld = sec->lines[k];
				int			lnum = ld - lines;

				if (   ld->spflags & SPF_CONTROLLINE
					|| ld->frontsector == ld->backsector)
				{
					continue;
				}
				
				if (   ld->ml_flags & ML_TWOSIDED
					&& !lineflags[lnum]
					&& Vtxs2(ld)
					&& ld->specType == R3D_00
					&& ld->lnflags2 == 0x0C000000
					&& (fs = ld->frontsector) != (bs = ld->backsector))
				{
					sector_t *s = (fs == sec ? bs : fs);

					if (   IsLiquidFloorFlat(s)
						|| s->specFlags & SF_MOVFLOOR
						|| s->floorheight >= sec->floorheight
						|| s->heightsec >= 0
						|| s->stptr >= 0)
					{
						c = 0; break;
					}

					if (!origTEX[lnum])
					{
						linearray[c++] = lnum;
					}
				}
				else
				{
					c = 0; break;
				}
			}

			if (c > 2)
			{
				for (k = 0; k < c; k++)
				{
					line_t	 *ld = &lines[linearray[k]];
					sector_t *s  = (ld->frontsector == sec ? ld->backsector : ld->frontsector);
					ssline_t *sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec && ss->floorsec < 0 && ss->lfsec < 0)
						{
							Setr3dsslfsec(ss, s - sectors, 43169);
						}
					}
				}
			}
		}
		else if (mask == 0x888C
			&& !SecCtrl(i)
			&& (secn = LS_GetContainingSecnum(sec, 1)) >= 0
			&& !(sectors[secn].specFlags & 0xF000C)
			&& sectors[secn].stptr < 0
			&& XS_GetBoundarySecnum(sec) == secn)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				if (   lineflags[sec->lines[k] - lines]
					|| sec->lines[k]->specType >= R3D_40)
				{
					secn = -1; break;
				}
			}

			if (secn < 0) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				ssline_t *sl = &sslines[sec->lines[k] - lines];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					if (ss->ceilsec < 0 && ss->floorsec < 0)
					{
						SetOverride(ss, secn, 877);
					}
				}

				SetLFG(sec->lines[k], LFG_NEWSET, 877);
			}

			Setsecdone(s, 877);
		}
end_loop:;
	}

	W_Free(hitssecmem, STATICID, 0);
}

// Check for stuff like Herian2 map29 sec 447/449
void PS_CheckNestedFakeMovSectors()
{
	sector_t	*sec, *osec;
	line_t		*ld;
	int			c, i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->heightsec >= 0
			&& (sec->specFlags & 0x1879C) == 0x18790)
		{
			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (lineflags[ld - lines]) { c = 0; break; }

				if (   ld->ml_flags & ML_TWOSIDED
					&& Vtxs2(ld)
					&& ld->lnflags2 & LNF2_OWNEQUALTOTWO)
				{
					osec = sec == ld->frontsector ? ld->backsector : ld->frontsector;

					if (  (osec->specFlags & 0x3001C) == 0x1001C
						&& osec->heightsec >= 0)
					{
						linearray[c++] = ld - lines;
					}
				}
			}

			if (c > 2)
			{
				int m, n, num;

				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = true;

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (hitline[ld - lines] == 1
						&& subfn_CheckForRing(hitline, sec, ld - lines, &num, NULL, NULL, NULL) )
					{
						osec = sec == ld->frontsector ? ld->backsector : ld->frontsector;

						for (m = 0; m < sec->linecount; m++)
						{
							num = sec->lines[m] - lines;

							if (hitline[num] == 4)
							{
								ssline_t *sl = &sslines[num];

								for (n = 0; n < sl->num; n++)
								{
									subsector_t *ss = &subsectors[sl->list[n]];

									if (ss->sector == osec && ss->floorsec < 0)
										SetOverride(ss, i, 43089);
								}

								hitline[num] = 8;
								SetLFG(sec->lines[m], LFG_NEWSET, 43090);
							}
						}
						Setsecdone(osec, 43091);
					}
				}
			}
		}
	}
}

// PS_FinaliseSSECsettings -------------------------------------------------

static RDBOOL validMovSec(sector_t *sf, sector_t *sb)
{
	if (  !sb
		|| sf == sb
		|| !((sf->specFlags & SF_MOVFLOOR)^(sb->specFlags & SF_MOVFLOOR)) )
	{
		return false;
	}

	if (sf->specFlags & SF_MOVFLOOR)
	{
		if (sf->ceilingpic == skyflatnum && sb->ceilingpic != skyflatnum) return false;
		return !(sf->specFlags & SF_NOLOTEX);
	}

	if (sb->ceilingpic == skyflatnum && sf->ceilingpic != skyflatnum) return false;
	return !(sb->specFlags & SF_NOLOTEX);
}

// b211-0 Check if line should have a lower or upper wall drawn
// $v66
// v220-22 some code removed
static RDBOOL checkLFSEC(line_t *ld)
{
	sector_t *fs = ld->frontsector;
	sector_t *bs = ld->backsector;
	int		 lfsec = -1;

	if (  !origTEX[ld - lines]
		&& bs
		&& ((fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum)
		||   fs->ceilingheight == bs->ceilingheight)
		&& IsLiquidFloorFlat(fs)
		|| IsLiquidFloorFlat(bs))
	{
		ssline_t *sl = &sslines[ld - lines];
		int		 i;

		for (i = 0; i < sl->num; i++)
		{
			if ( subsectors[sl->list[i]].lfsec >= 0)
			{
				lfsec = subsectors[sl->list[i]].sector - sectors;
				break;
			}
		}

		if (   lfsec >= 0
			&& IsLiquidFloorFlat(&sectors[lfsec])
			&& !(origTEX[ld - lines] & 0x11)
			&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
			&& Vtxs2(ld)
			&& Vtfl3(ld))
		{
			int ht = sectors[lfsec].floorheight;
			int pic = sectors[lfsec].floorpic;

			for (i = 0; i < sl->num; i++)
			{
				if (   subsectors[sl->list[i]].lfsec < 0
					&& subsectors[sl->list[i]].sector->floorpic == pic
					&& subsectors[sl->list[i]].floorsec < 0
					&& subsectors[sl->list[i]].sector->floorheight > ht)
					subsectors[sl->list[i]].lfsec = lfsec;
			}
		}
	}
	return lfsec >= 0;
}

// v220-3 extended 
void PS_FinaliseSSECsettings()
{
	subsector_t *ss;
	sector_t	*fs, *bs;
	int			a, c, i, k;

	for (a = c = i = 0; i < numlines; i++)
	{
		if (   !(lines[i].lnflags2 & LNF2_OWNEQUALTOONE)
			&& !(lines[i].ml_flags & ML_TWOSIDED)
			&& !(lineflags[i] & LFG_BADLINE)
			&& vtxarray[lines[i].v1 - vertexes]  > 1
			&& vtocount[lines[i].v1 - vertexes] == 1
			&& vtxarray[lines[i].v2 - vertexes]  > 1
			&& vtocount[lines[i].v2 - vertexes] == 1)
		{
			if (lines[i].backsector)	// if backsector set then a spfx line
			{
				a++;
			}
			else
			{
				linearray[c++] = i;
			}
		}
	}

	if (a) pr_OFinsert("\n%i special effects lines detected\n", a);

	if (c)	// v220-25 ammended to only show lines in edit mode
	{
		pr_OFinsert("\n%i MAP ERROR%s: line%s missing side 2\n"
					"  This fault cannot be fixed internally.\n"
					"  The map should be fixed using an editor.\n",
					c,
					c > 1 ? "S" : "",
					c > 1 ? "s" : "");

		if (R3D_edit)
		{
			pr_OFinsert("  line%s requiring a backsector:\n", c > 1 ? "s" : "");

			for (i = 0; i < c; i++)
				pr_OFinsert("  %i\n", linearray[i]);
		}
		else
		{
			pr_OFinsert("  To get the line number%s run in edit mode.\n",
						c > 1 ? "s" : "");
		}

		pr_OFdirect("\n");
		ma_inf.maperror += c;
	}

	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].tyflags & TYF_SKYWALL
			&& lines[i].ovflags & OVF_NOSKYWALL)
		{
			lines[i].tyflags &= ~TYF_SKYWALL;
		}
		else if (lines[i].specType == R3D_01)
		{
			sector_t *fs = lines[i].frontsector;
			sector_t *bs = lines[i].backsector;

			if (   bs
				&& fs != bs
				&& fs->ceilingpic == bs->ceilingpic
				&& (fs->specFlags & SF_MOVCEIL)^(bs->specFlags & SF_MOVCEIL))
			{
				sector_t *s1 = fs->specFlags & SF_MOVCEIL ? bs : fs;
				sector_t *s2 = fs->specFlags & SF_MOVCEIL ? fs : bs;
				RDBOOL   type = (s2->specFlags & 0xFFFFC) == 0x2001C
					&& s2->floorheight == s2->ceilingheight
					&& s1->floorheight < s2->floorheight;	// v220-3

				if (s1->ceilingheight > s2->ceilingheight)
				{
					ssline_t *sl = &sslines[i];

					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						// v220-3 check for no floor tex added
						if (ss->sector == s2 && ss->ceilsec < 0)
						{
							if (type && ss->floorsec < 0)
								SetOverride(ss, s1 - sectors, 43170);
							else
								Setr3dsscs(ss, s1 - sectors, 43170);
						}
					}
				}
			}
		}
		else if (lines[i].ml_flags & ML_TWOSIDED
			&& (fs = lines[i].frontsector) != (bs = lines[i].backsector))
		{
			if (   lines[i].specType == R3D_00
				&& !origTEX[i]
				&& SecCtrl(fs - sectors)
				&& SecCtrl(bs - sectors)
				&& (fs->specFlags & SF_MOVFLOOR)^(bs->specFlags & SF_MOVFLOOR)
				&& fs->heightsec < 0
				&& bs->heightsec < 0
				&& fs->stptr < 0
				&& bs->stptr < 0)
			{
				if (fs->specFlags & SF_MOVFLOOR && IsLiquidFloorFlat(fs))
				{
					if (fs->floorheight - bs->floorheight > 8*FRACUNIT)
						Settype(&lines[i], R3D_28, 43092);
				}
				else
				if (bs->specFlags & SF_MOVFLOOR && IsLiquidFloorFlat(bs))
				{
					if (bs->floorheight - fs->floorheight > 8*FRACUNIT)
						Settype(&lines[i], R3D_28, 43093);
				}
			}

			// Added b220-1
			if (   lines[i].specType == R3D_00
				&& (fs->specFlags & SF_MOVCEIL)^(bs->specFlags & SF_MOVCEIL)
				&& fs->ceilingpic == bs->ceilingpic
				&& fs->ceilingpic == skyflatnum)
			{
				sector_t *s1 = fs->specFlags & SF_MOVCEIL ? bs : fs;
				sector_t *s2 = fs->specFlags & SF_MOVCEIL ? fs : bs;
				int      n;

				if (s1->ceilingheight + s1->skyfix >= s2->ceilingheight + s2->skyfix)
				{
					RDBOOL valid = true;

					for (k = 0; k < s2->linecount; k++)
					{
						line_t *ld = s2->lines[k];

						if (   ld->ml_flags & ML_TWOSIDED
							&& !(ld->spflags & SPF_CONTROLLINE))
						{
							sector_t *s = ld->frontsector == s2 ? ld->backsector : ld->frontsector;

							if (s != s1 && s != s2)
							{
								valid = false; break;
							}
						}
					}

					if (valid)
					{
						for (k = 0; k < s2->linecount; k++)
						{
							ssline_t *sl = &sslines[s2->lines[k] - lines];
							
							for (n = 0; n < sl->num; n++)
							{
								subsector_t *ss = &subsectors[sl->list[n]];

								if (   ss->sector == s2
									&& ss->ceilsec < 0)
								{
									Setr3dsscs(ss, s1 - sectors, 43171);
								}
							}
						}
					}
				}
			}
		}
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		int			bs = ss->sector - sectors;
		int			count = 0;

		if (  !R3D_maptype
			&& ss->sector->heightsec < 0
			&& ss->floorsec < 0
			&& ss->lfsec < 0
			&& ss->linecount >= 4
			&& ss->linecount <= 6)
		{
			sector_t	*s = NULL;
			seg_t		*sg = &segs[ss->firstline];

			validcount++;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (  !sg->linedef
					|| sg->linedef->specType != R3D_00
					|| sg->frontsector == sg->backsector
					|| sg->linedef->spflags & SPF_CONTROLLINE
					|| origTEX[sg->linedef - lines] & 0x11
					|| lineflags[sg->linedef - lines]
					|| lxarray[sg->linedef - lines])
				{
					count = 0; break;
				}

				if ( !(sg->linedef->ml_flags & ML_TWOSIDED) )
				{
					if (sg->linedef->validcount != validcount)
					{
						if (!s) s = sg->linedef->frontsector;
						else if (s != sg->linedef->frontsector) { count = 0; break; }
						count++;
						sg->linedef->validcount = validcount;
					}
				}
			}

			if (count == 2 && ss->sector != s && s->floorheight >= ss->sector->floorheight)
			{
				ss->lfsec = s - sectors;
			}
		}

		if (secdone[ss->sector - sectors] == 4) continue;

		if (ss->floorsec >= 0 && ss->lfsec == ss->floorsec)
			ss->lfsec = -1;

		if (ss->ceilsec  == bs) ss->ceilsec      = -1;
		if (ss->floorsec == bs) ss->floorsec     = -1;
		if (ss->lfsec == ss->floorsec) ss->lfsec = -1;

		if (ss->sector->specFlags2 & (SF2_LOSLOPE|SF2_HISLOPE))
		{
			if (ss->sector->specFlags2 & SF2_HISLOPE)
			{
				ss->lcsec    = -1;
				ss->ceilsec  = -1;
				ss->sector->R3Dceilheight = DDMININT;
				ss->sector->R3Dceilsec = -1;
			}

			if (ss->sector->specFlags2 & SF2_LOSLOPE)
			{
				ss->lfsec    = -1;
				ss->floorsec = -1;
				ss->sector->R3Dfloorheight = DDMAXINT;
				ss->sector->R3Dfloorsec = -1;
			}

			ss->ssflags &= ~SSF_INHSECLIGHT;
			ss->sslightsec  = -1;
			continue;
		}

		// v220-5 Catch any subsectors that use the wrong line sector and have evaded analysis
		if (   ss->sector->ceilingpic == skyflatnum
			&& !(ss->ssflags & SSF_BASESSECFIX)
			&& ss->linecount > 3
			&& ss->linecount < 6
			&& ss->sector->stptr < 0
			&& ChkSSECunset(ss)
			&& SecUnset(ss->sector))
		{
			sector_t	*chk = NULL;
			seg_t		*sg;
			RDBOOL		valid = false;

			// v220-9 check that ssec has no crossing or same sector lines and
			//        that any lines with a midtex are not in the same sector
			//        as the subsectors sector
			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef)
				{
					if (   sg->linedef->frontsector == sg->linedef->backsector
						|| lxarray[sg->linedef - lines])
					{
						valid = false; break;
					}

					if (origTEX[sg->linedef - lines] & 0x22)
					{
						if (sg->frontsector == ss->sector) { valid = false; break; }

						if (!chk) chk = sg->frontsector;
						else if (chk != sg->frontsector) { valid = false; break; }

						valid = true;
					}
				}
			}

			if (valid)
			{
				for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
				{
					if (   sg->linedef
						&& sg->sidedef
						&& (fs = sg->sidedef->sector) != ss->sector
						&& fs->ceilingheight < ss->sector->ceilingheight
						&& fs->ceilingpic != skyflatnum
						&& fs->floorheight == ss->sector->floorheight
						&& fs->floorpic == ss->sector->floorpic)
					{
						// changing just the ceilsec may not work if there
						// are multiple instances in the same sector
						ChangeSSECsec(ss, fs, 43094);
						break;
					}
				}
			}
		}

		// v211-0 1) final check on ssec->sector (where nodes have not been rebuilt)
		// v211-0 2) set R3D_3F, check/set SSF_01INHIBIT flag
		// v2.1.00 build_62 Transfer lightsecs
		{
			sector_t	*rs;
			line_t		*ld;
			seg_t		*sg, *firstsg = &segs[ss->firstline];

			// Check that subsector with a one's line is set to the line's frontsec
			if ((ss->sector->specFlags & 0xFFFFC) == 0x8188)
			{
				if (   pwad_nodesNotRebuilt
					&& !ss->ssflags
					&& !secdone[ss->sector - sectors]
					&& SSECunset(ss))
				{
					rs = NULL;

					for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
					{
						if ( (ld = sg->linedef) )
						{
							if ( !ld->backsector
								&& ss->sector != ld->frontsector)
							{
								if (!rs)
								{
									rs = ld->frontsector;
									if (secdone[rs - sectors])
									{
										rs = NULL; break;
									}
								}
								else if (rs != ld->frontsector)
								{
									rs = NULL; break;
								}
							}
							if (lineflags[ld - lines])
							{
								rs = NULL; break;
							}
						}
					}

					if (rs)
						ChangeSSECsec(ss, rs, 43095);
				}
			}
			else if (   ss->sector->specFlags == 0x0200001C
				&& ss->sector->floorheight == ss->sector->ceilingheight
				&& ss->sector->ceilingpic == skyflatnum)
			{
				rs = NULL;

				for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
				{
					if ( (ld = sg->linedef) && ld->specType == R3D_01)
					{
						Forcetype(ld, R3D_3F, 43096);

						if (!rs)
						{
							if (ld->frontsector != ss->sector)
								rs = ld->frontsector;
							else if (ld->backsector != ss->sector)
								rs = ld->frontsector;

							if ((rs->specFlags & 0xC) == 0xC)
								rs = NULL;
						}
					}
				}

				if (   rs
					&& ss->floorsec < 0
					&& ss->ceilsec < 0)
				{
					SetOverride(ss, rs - sectors, 43097);
				}
			}
			else if (ss->ssflags & SSF_01INHIBIT)
			{
				// v220-7 add mov check
				for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
				{
					if (   sg->linedef
						&& sg->linedef->specType == R3D_01
						&& sg->backsector
						&& !(sg->frontsector->specFlags & SF_MOVSECTOR)
						&& !(sg->backsector->specFlags & SF_MOVSECTOR))
					{
						Forcetype(sg->linedef, R3D_3F, 43098);
					}
				}
			}
			else
			{
				if (!(ss->sector->specFlags & SF_NOLOTEX))
				{
					inhibitSSEC(ss, 43098);
				}
				else
				{
					RDBOOL noinhibit = false;
					RDBOOL test = ss->floorsec < 0 && ss->ceilsec < 0;

					if (test && ss->lfsec < 0)
					{
						for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
						{
							if (sg->frontsector && sg->backsector)
							{
								if (   IsLiquidFloorFlat(sg->frontsector)
									|| IsLiquidFloorFlat(sg->backsector ))
								{
									noinhibit = false; break;
								}
								else if (sg->frontsector != sg->backsector
									&& sg->linedef	// should be
									&& sg->linedef->specType == R3D_01)
								{
									if (   sg->frontsector->specFlags & SF_MOVSECTOR
										|| sg->backsector->specFlags  & SF_MOVSECTOR)
									{
										noinhibit = true;
									}
								}
							}
						}
					}
					else if (test && ss->lfsec >= 0)	// v220-6 second test added
					{
						for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
						{
							if (sg->frontsector && sg->backsector)
							{
								if (   IsLiquidFloorFlat(sg->frontsector)
									|| IsLiquidFloorFlat(sg->backsector ))
								{
									noinhibit = false; break;
								}
								else if (sg->frontsector != sg->backsector
									&& sg->linedef	// should be
									&& sg->linedef->specType == R3D_01)
								{
									if (   sg->frontsector->specFlags & SF_MOVFLOOR
										&& sg->backsector->specFlags  & SF_MOVFLOOR
										&& (sg->frontsector->specFlags & SF_NOLOTEX)
										^  (sg->backsector->specFlags  & SF_NOLOTEX))
									{
										noinhibit = true;
									}
									else
									{
										noinhibit = false; break;
									}
								}
							}
						}
					}

					if (!noinhibit)
					{
						// v220-7 add test for line order nodes build inconsistency
						if (ss->floorsec < 0 && ss->lfsec < 0)
						{
							if (   firstsg->linedef
								&& firstsg->backsector
								&& firstsg->linedef->specType == R3D_1D
								&& !(firstsg->frontsector->specFlags & SF_MOVSECTOR)
								&& !(firstsg->backsector->specFlags & SF_MOVSECTOR)
								&& firstsg->frontsector == ss->sector
								&& firstsg->frontsector->specFlags & SF_NOLOTEX)
							{
								sector_t *os = NULL;

								sg = firstsg;

								for (k = 1, ++sg; k < ss->linecount; k++, sg++)
								{

									if (sg->frontsector != firstsg->frontsector)
									{
										if (   !sg->backsector
											|| sg->frontsector->floorheight != sg->backsector->floorheight)
										{	os = NULL; break;
										}

										if (!os) os = sg->frontsector;
										else if (os != sg->frontsector) { os = NULL; break; }
									}
								}

								if (   os
									&& !(os->specFlags & SF_MOVSECTOR)
									&& os->floorheight == firstsg->backsector->floorheight
									&& os->floorpic == firstsg->backsector->floorpic)
								{
									Setr3dssfs(ss, os - sectors, 43099);
								}
							}
						}

						inhibitSSEC(ss, 43099);
					}
				}
			}

			for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
			{
				if (   sg->frontsector
					&& sg->frontsector->floorlightsec >= 0)
				{
					SetFloorlightsec(ss, sg->frontsector->floorlightsec, 43040);
					break;
				}
			}

			// ChangeLineTypes
			if (ss->floorsec >= 0 || ss->lfsec >= 0)
			{
				for (k = 0, sg = firstsg; k < ss->linecount; k++, sg++)
				{
					if (   (ld = sg->linedef)
						&& !(lineflags[ld - lines] & (LFG_NOCHANGE|LFG_3DTYPE))
						&& !origTEX[ld - lines])
					{
						RDBOOL validheight = true;

						if (sg->backsector)
						{
							if (sg->frontsector->ceilingheight - sg->frontsector->floorheight
								< 4*FRACUNIT
								|| sg->backsector->ceilingheight - sg->backsector->floorheight
								< 4*FRACUNIT)
							{
								validheight = false;
							}
						}

						if (validMovSec(sg->frontsector, sg->backsector)
							|| (validheight
							&& vtoflags[ld->v1 - vertexes]
							&& vtoflags[ld->v2 - vertexes]))
						{
							// v220-4 test added
							if (!(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)))
								Forcetype(ld, R3D_00, 43099);
						}
						else if (ld->specType != R3D_3F)
						{
							if (   ld->ml_flags & ML_TWOSIDED
								&& ld->mtflags & R3D_MISLOTEX
								&& !(lineflags[ld - lines] & (LFG_FIXED|LFG_3DTYPE)
								&& ld->specType == R3D_01)
								&& sg->backsector
								&& sg->frontsector->ceilingpic == skyflatnum
								&& sg->backsector->ceilingpic == skyflatnum)
							{
								int sa = ld->frontsector->specFlags & 0x4;
								int sb = ld->backsector->specFlags  & 0x4;

								if ((sa && !sb) || (sb && !sa))
								{
									if (!checkLFSEC(ld))
									{
										Forcetype(ld, R3D_1A, 43102);
										continue;
									}
								}
							}

							// b211-0 R3D_16 test added for textureless missing tex rings
							if (ss->lfsec >= 0)
							{
								if ( !(ld->frontsector->ceilingpic == skyflatnum
									&& ld->backsector->ceilingpic  == skyflatnum) )
								{
									if (   ld->frontsector->floorheight < ld->backsector->floorheight
										&& ld->frontsector - sectors == ss->lfsec
									//	&& ld->frontsector->specFlags & SF_SSLINETRUE
										&& (ld->backsector->specFlags & 0x3FF1C) == 0x1C)
									{
										Forcetype(ld, R3D_3F, 43102);
										continue;
									}

									if (   ld->backsector->floorheight < ld->frontsector->floorheight
										&& ld->backsector - sectors == ss->lfsec
									//	&& ld->backsector->specFlags & SF_SSLINETRUE
										&& (ld->frontsector->specFlags & 0x3FF1C) == 0x1C)
									{
										Forcetype(ld, R3D_3F, 43102);
										continue;
									}
								}
							}

							// 2) Check for R3D_16 test
							if (   Vtfl3(ld)
								&& ld->mtflags & R3D_MISLOTEX
								&& !origTEX[ld - lines]
								&& ld->frontsector->floorpic != skyflatnum
								&& ld->backsector->floorpic != skyflatnum
								&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
								&& !(ld->backsector->specFlags  & SF_MOVSECTOR)
								&& abs(ld->frontsector->floorheight - ld->backsector->floorheight) > 8*FRACUNIT
								&& (ld->frontsector->ceilingheight == ld->backsector->ceilingheight
								|| (ld->frontsector->ceilingpic == skyflatnum && ld->backsector->ceilingpic  == skyflatnum))
								&& ld->frontsector->heightsec < 0
								&& ld->backsector->heightsec < 0 )
							{
								Forcetype(ld, R3D_16, 43103);
							}
							else
							{
								if (ld->tyflags & TYF_SKYWALL)
								{
									if (ld->specType != R3D_01)
										Forcetype(ld, R3D_01, 43104);
								}
							}
						}
					}
				}
			}
		}
	}
}

void PS_FinaliseAASHITTYLines()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			d, i, k;

	CLR_hitline();

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !origTEX[i]
			&& ld->ml_flags & ML_TWOSIDED
			&& !(lineflags[i] & LFG_3DTYPE)
			&& ld->lnflags2 & LNF2_OWNTWOANDPLUS
			&& ld->specType != R3D_16
			&& Vtfl(ld)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& (d = fs->floorheight - bs->floorheight)
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& fs->heightsec < 0
			&& bs->heightsec < 0
			&& fs->floorpic != skyflatnum
			&& bs->floorpic != skyflatnum
			&& ((d > 0 && asyflags[ld->sidenum[1]] & ASY_L)
			||  (d < 0 && asyflags[ld->sidenum[0]] & ASY_L)) )
		{
			if (lineflags[i])
			{
				sector_t *st = d > 0 ? fs : bs;
				sector_t *sb = d < 0 ? fs : bs;
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					Setr3dssfs(ss, st - sectors, 43101);
					Setr3dsslfsec(ss, sb - sectors, 43101);
				}
			}
			else
				Forcetype(ld, R3D_16, 43101);
		}
	}
}

//==============================================================================
//
// Control group check

static struct ctrlgroup_t {
	sector_t *groupcontain;
	sector_t *groupcontainlfsec;
	RDBOOL	valid;
	int		*groupsec;
	int		*chkgrpsec;
	int		gscount;
	int		grpcount;
} cg;

static RDBOOL notCTRL(sector_t *s)
{
	int	snum = s - sectors;
	int i, k;

	for (i = 0; i < cg.grpcount; i++)
	{
		if (countsec[i] == snum)
		{
			RDBOOL found = false;

			for (k = 0; k < cg.gscount; k++)
			{
				if (countsec[cg.groupsec[k]] == snum) { found = true; break; }
			}

			if (!found)
			{
				cg.chkgrpsec[i] = true;
				cg.groupsec[cg.gscount++] = i;
			}

			return false;
		}
	}
	return true;
}

static void addRelatedCtrlSec(sector_t *sec)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld, *li;
	seg_t		*sg;
	int			i, k, m;

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			sg = &segs[ss->firstline];

			if (ss->ssflags & SSF_NOCHANGE)
			{
				cg.valid = false;
				return;
			}

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (  (li = sg->linedef)
					&& li->ml_flags & ML_TWOSIDED)
				{
					// At this point check for the containing sector
					if (notCTRL(sg->frontsector))
					{
						if (!cg.groupcontain)
							cg.groupcontain = sg->frontsector;
						else if (cg.groupcontain != sg->frontsector)
							cg.valid = false;
					}
				}
			}
		}
	}
}

static RDBOOL chkCtrlsec(sector_t *sec)
{
	sector_t	*s;
	int			i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| notCTRL(s = ld->frontsector == sec ? ld->backsector : ld->frontsector))
		{
			return false;
		}
	}

	addRelatedCtrlSec(sec);
	return true;
}

static void setGroupSsecs(int secn)
{
	subsector_t	*ss;
	sector_t	*sec = &sectors[secn];
	ssline_t	*sl;
	line_t		*ld;
	int			i, k;
	int			ovrnum = cg.groupcontain - sectors;

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->validcount != validcount)
		{
			ld->validcount = validcount;

			if (ld->spflags & SPF_CONTROLLINE) continue;

			sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					ss->ssvalidcount = ss_validcount;

					SetOverride(ss, ovrnum, 43103);

					if (cg.groupcontainlfsec)
						Setr3dsslfsec(ss, cg.groupcontainlfsec - sectors, 43104);
				}
			}
		}
	}
}

//  Check unreferenced sectors with control lines
void PS_CheckUnsetCTRLsectors()
{
	sector_t	*sec;
	line_t		*ld;
	int			i, k;
	RDBOOL		valid;

	for (cg.grpcount = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& (sec->specFlags & 0x3000C) == 0xC	// v220-3 static sector only
			&& SecCtrl(i))
		{
			for (k = 0, valid = true; k < sec->linecount; k++)
			{
				if (  !( (ld = sec->lines[k])->spflags & SPF_CONTROLLINE)
					&& lineflags[ld - lines]
					&& !(lineflags[ld - lines] & LFG_3DTYPE))
				{
					valid = false; break;
				}

				// Eliminate line like 5579 in Eternal map28
				if (Vtxs1(ld) && ld->lnflags2 & LNF2_OWNEQUALTOTWO)
				{
					valid = false; break;
				}
			}
			if (valid) countsec[cg.grpcount++] = i;
		}
	}

	if (cg.grpcount)
	{
		cg.groupsec  = (int*)M_Malloc(cg.grpcount * sizeof(int), 0);
		cg.chkgrpsec = (int*)M_Calloc(cg.grpcount * sizeof(int), 0);

		// Group any referencing another
		for (i = 0; i < cg.grpcount; i++)
		{
			int		ref = i;
			RDBOOL	repeat = true;

			if (cg.chkgrpsec[i]) continue;
			
			cg.groupcontain = NULL;
			cg.gscount = 0;
			cg.valid = true;

			valid = true;

			while (repeat)
			{
				int oldcount = cg.gscount;

				repeat = false;				
				cg.chkgrpsec[ref] = true;
				sec = &sectors[countsec[ref]];
				valid = chkCtrlsec(sec);

				if (!valid || cg.gscount <= oldcount) break;

				for (k = 0; k < cg.gscount; k++)
				{
					if (!cg.chkgrpsec[cg.groupsec[k]])
					{
						ref = cg.groupsec[k];
						repeat = true;
					}
				}
			}

			if (valid && cg.valid && cg.groupcontain)
			{

				validcount++;
				ss_validcount++;

				cg.groupcontainlfsec = NULL;

				for (k = 0; k < cg.gscount; k++)
				{
					sec = &sectors[countsec[cg.groupsec[k]]];

					if (   sec->specFlags & SF_MOVFLOOR
						&& sec->floorpic   == cg.groupcontain->floorpic
						&& sec->ceilingpic == cg.groupcontain->ceilingpic)
					{
						cg.groupcontainlfsec = cg.groupcontain;
						cg.groupcontain = sec;
					}
				}

				for (k = 0; k < cg.gscount; k++)
					setGroupSsecs(countsec[cg.groupsec[k]]);
			}
		}

		M_Free(cg.chkgrpsec);
		M_Free(cg.groupsec);
	}
}

static RDBOOL verifySSEC(ssline_t *sl, sector_t *sec, sector_t *rsec)
{
	int c, i, k;

	// ensure the subsector is only referencing the marked lines
	for (c = i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->sector == sec)
		{
			seg_t *sg = &segs[ss->firstline];

			c++;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (   sg->linedef
					&& hitline[(sg->linedef - lines)] != 4)
				{
					return false;
				}
			}
		}
		else if (rsec)
		{
			seg_t *sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (   sg->linedef
					&& sg->linedef->frontsector == sg->linedef->backsector)
				{
					return false;
				}
			}
		}
	}

	return c > 0;	// It's possible that no subsecs access sec
}

// v220-6 Look for movfloor sectors that need a false floor
//        Types not yesy processed will be those with no low
//        textures set (e.g Eternal map15 sec 156)
void PS_FinaliseMissingTexLoops()
{
	int i, c;

	for (i = 0; i < numsectors; i++)
	{
		if ((sectors[i].specFlags & 0xFFFFF) == 0x18390)
		{
			sector_t *sec = &sectors[i];
			int		 k, m, lnum, hitcount = 0;
			int		 *ptr = NULL, ptrcnt;

			CLR_hitline();

repeat_check:;	// could be multiple loops in any one sector

			c = LS_ChkLineLoops(sec, false, true, false);

			if (c)
			{
				if (!ptr) ptr = (int*)malloc(sec->linecount * sizeof(int));

				ptrcnt = 0;

				for (k = 0; k < sec->linecount; k++)
				{
					if (hitline[ (lnum = sec->lines[k] - lines) ] == 4)
					{
						ptr[ptrcnt++] = lnum;
					}
				}

				if (ptrcnt)
				{
					sector_t *s = lines[ptr[0]].frontsector == sec ?
						lines[ptr[0]].backsector: lines[ptr[0]].frontsector;

					if (s)
					{
						RDBOOL	valid = true;

						for (k = 0; k < ptrcnt; k++)
						{
							line_t		*ld = &lines[ptr[k]];
							ssline_t	*sl = &sslines[ptr[k]];

							if (   !(ld->frontsector == sec && ld->backsector == s  )
								&& !(ld->frontsector == s   && ld->backsector == sec))
							{
								valid = false;	break;
							}
							
							if (!verifySSEC(sl, sec, NULL)) { valid = false; break; }
						}

						if (valid)
						{
							for (k = 0; k < ptrcnt; k++)
							{
								ssline_t *sl = &sslines[ptr[k]];

								lineflags[ptr[k]] |= LFG_NEWSET;

								// Set lfsec as this won't be used if the mov sector's floor
								// should rise higher than the containing sector's floor
								for (m = 0; m < sl->num; m++)
								{
									subsector_t *ss = &subsectors[sl->list[m]];
									if (ss->sector == sec) Setr3dsslfsec(ss, s - sectors, 43106);
								}
							}
						}
					}

					// Mark off the lines in this loop
					for (k = 0; k < ptrcnt; k++)
					{
						if (hitline[ptr[k]] == 4) hitline[ptr[k]] = 8;
					}

					goto repeat_check;
				}
			}

			M_Free(ptr);
		}
	}
}

// 1) check crossing lines 2) v220-5 check if R3D_00 should be R3D_01
void PS_FinaliseCrossingLines()
{
	line_t	*ld;
	int		i, k;
	int		mask = SF_NOLOTEX|SF_NOHITEX;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->frontsector == ld->backsector)
		{
			if (ld->specType >= R3D_40) continue;

			// v220-6 lighting inheritance from 0xC secs with lightlevel == 0
			if (   ld->frontsector->lightlevel == 0
				&& ld->lnflags2 & LNF2_OWNEQUALTOONE
				&& Vtxs1(ld))
			{
				ssline_t	*sl = &sslines[i];
				subsector_t *ss = &subsectors[sl->list[0]];
				seg_t		*sg = &segs[ss->firstline];
				RDBOOL		found = false;
				int			m, n, secidx, llevel;

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if (sg->frontsector->lightlevel)
					{
						for (m = n = 0; m < anisl.numentries; m++)
						{
							secidx = anisl.entries[n]; n++;
							llevel = anisl.entries[n]; n++;

							if (   secidx == ld->frontsector - sectors
								&& llevel == sg->frontsector->lightlevel)
							{
								found = true; break;
							}
						}

						if (!found)
						{
							if (anisl.numentries >= anisl.maxentries)
							{
								anisl.maxentries += 16;
								anisl.entries = (int*)W_Realloc(anisl.entries, anisl.maxentries*2*sizeof(int), STATICID, 0);
							}
							ld->frontsector->lightlevel = sg->frontsector->lightlevel;
							anisl.entries[anisl.numentries*2] = ld->frontsector - sectors;
							anisl.entries[anisl.numentries*2 + 1] = sg->frontsector->lightlevel;
							anisl.numentries++;
							break;
						}
					}
				}
			}
			else	// v220-6 special movfloor case
			if (   ld->lnflags2 & LNF2_OWNBOTHGTRONE
				&& origTEX[i] & 0x22
				&& lxarray[i] == 4
				&& ld->frontsector->specFlags & SF_MOVFLOOR
				&& !(ld->frontsector->specFlags & SF_NOLOTEX))
			{
				ssline_t *sl = &sslines[i];
				int newsec = -1;


				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (   ss->floorsec >= 0
						&& ss->sector->specFlags & SF_NOLOTEX
						&& ss->floorsec != ss->sector - sectors
						&& ss->floorsec == ld->frontsector - sectors)
					{
						newsec = ld->frontsector - sectors; break;
					}
				}

				if (newsec >= 0)
				{
					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->floorsec != newsec)
							Setr3dssfs(ss, newsec, 43107);
					}

					continue;
				}
			}
			else
			// v220-6 special movfloor case. Needs to differentiate between
			//        two possible sectors both of which have lotex set.
			if (   ld->lnflags2 & LNF2_OWNBOTHGTRONE
				&& origTEX[i] & 0x22
				&& !lxarray[i]
				&& ld->frontsector->specFlags & SF_MOVFLOOR
				&& !(ld->frontsector->specFlags & SF_NOLOTEX))
			{
				ssline_t *sl = &sslines[i];
				int newsec = -1;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (   ss->floorsec >= 0
						&& ss->floorsec != ss->sector - sectors
						&& ss->floorsec == ld->frontsector - sectors)
					{
						newsec = ld->frontsector - sectors; break;
					}
				}

				if (newsec >= 0)
				{
					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->floorsec != newsec)
							Setr3dssfs(ss, newsec, 43108);
					}

					continue;
				}
			}

			// v220-5 check rendering sector inheritance with single side crossing lines
			// v220-6 modified
			if (   lxarray[i] & 4
				&& !lineflags[i]
				&& ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				if ((ld->frontsector->specFlags & 0xC) == 0xC)
				{
					ssline_t *sl = &sslines[i];
					int s1 = -1, s2 = false;

					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->ssflags & (SSF_NOCHANGE|SSF_FIXED)) { s1 = -1; break; }
						
						if (ss->floorsec < 0)
						{
							s2 = true;
						}
						else if ((sectors[ss->floorsec].specFlags & 0xC) == 0xC)
						{
							s2 = true;
						}
						else
						{
							if (s1 < 0) s1 = ss->floorsec;

							if (s1 != ss->floorsec || s1 != ss->ceilsec)
							{
								s1 = -1;
								break;
							}
						}
					}

					if (s1 > 0 && s2)
					{
						for (k = 0; k < sl->num; k++)
						{
							subsector_t *ss = &subsectors[sl->list[k]];
							
							SetOverride(ss, s1, 43109);
						}
					}
				}
			}

			// v220-5 check rendering sector inheritance with crossing lines
			if (   lxarray[i] & 4
				&& ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				ssline_t *sl = &sslines[i];
				int		 s1 = -1, s2 = false;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->ssflags & (SSF_NOCHANGE|SSF_FIXED)) { s1 = -1; break; }
					
					if (ss->floorsec < 0)
					{
						s2 = true;
					}
					else
					{
						if (s1 < 0) s1 = ss->floorsec;

						if (s1 != ss->floorsec || s1 != ss->ceilsec)
						{
							s1 = -1;
							break;
						}
					}
				}

				if (s1 > 0 && s2)
				{
					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];
						
						if (   ss->floorsec < 0
							&& (ss->sector->specFlags & mask) == mask
							&& SSECunset(ss))
						{
							SetOverride(ss, s1, 43110);
						}
					}
				}
			}
			else	// first added v220-6, modified v220-10
			if (!lxarray[i])
			{
				ssline_t *sl = &sslines[i];
				int		 sf1 = -1, sf2 = 0, sc1 = -1, sc2 = 0;
				int		 m;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];
					seg_t		*sg = &segs[ss->firstline];

					if (ss->ssflags & (SSF_NOCHANGE|SSF_FIXED)) { sf1 = -1; sf2 = -1; break; }
					
					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (sg->backsector != sg->frontsector)
						{
							if (sg->frontsector != ld->frontsector) { sf1 = -1; sf2 = -1; break; }
						}

						if (sg->sidedef)
						{
							if (sg->sidedef->bottomtexture) sf1 = -2;
							if (sg->sidedef->toptexture   ) sc1 = -2;
						}
					}

					if (sf1 > -2)
					{
						if (ss->floorsec < 0)
							sf2 = 1;							// floorsec needs setting
						else
						{
							if (sf1 < 0) sf1 = ss->floorsec;	// valid floorsec
							else
							if (sf1 != ss->floorsec) sf1 = -2;	// different floorsecs
						}
					}
					
					if (sc1 > -2)
					{
						if (ss->ceilsec < 0)
							sc2 = 1;							// ceilsec needs setting
						else
						{
							if (sc1 < 0) sc1 = ss->ceilsec;		// valid ceilsec
							else
							if (sc1 != ss->ceilsec) sc1 = -2;	// different ceilsecs
						}
					}

					if (sf1 < -1 && sc1 < -1) break;
				}

				if (!sf2) sf1 = -1;	// Use sf1 to signal no floors to set
				if (!sc2) sc1 = -1;	// Use sc1 to signal no ceilings to set

				if (sf1 >= 0  || sc1 >= 0)
				{
					int gen = sf1 == sc1 ? sf1 : sf1 >= 0 && sc1 < 0 ? sf1 : sc1 >= 0 && sf1 < 0 ? sc1 : -1;
					int lt = !(ld->frontsector->specFlags & SF_NOLOTEX);
					int ht = !(ld->frontsector->specFlags & SF_NOHITEX);

					// Establish whether lt and/or ht remain true if
					// just looking at seg facing textures.
					if (gen >= 0 && (lt || ht))
					{
						for (m = 0; m < ld->frontsector->linecount; m++)
						{
							line_t *li = ld->frontsector->lines[m];

							if (li->spflags & SPF_CONTROLLINE) continue;

							if (li->backsector && li->backsector != li->frontsector)
							{
								int side = ld->frontsector == li->frontsector ? 0 : 1;

								if (lt && sf1 >= 0 && sides[li->sidenum[side]].bottomtexture) sf1 = -1;
								if (ht && sc1 >= 0 && sides[li->sidenum[side]].toptexture   ) sc1 = -1;
							}
						}
					}

					if (sf1 >= 0  || sc1 >= 0)
					{
						for (k = 0; k < sl->num; k++)
						{
							subsector_t *ss = &subsectors[sl->list[k]];

							if (SSECunset(ss))
							{
								// v220-12 modified
								if (gen >= 0)
								{
									if (ss->sector - sectors != gen)
									{
										// v220-10 add ceiling test
										if (   ss->sector->specFlags & SF_NOHITEX
											|| ss->sector->ceilingheight > sectors[gen].ceilingheight)
										{
											Setr3dsscs(ss, gen, 43111);
										}

									//	if ( (sectors[gen].specFlags & (SF_NOLOTEX|SF_SSLINETRUE)) != SF_NOLOTEX)
										if (   ss->sector->specFlags & SF_NOLOTEX
											&& !(ss->sector->specFlags & SF_SSLINETRUE))
										{
											Setr3dssfs(ss, gen, 43111);
										}
									}
								}
								else
								{
									if (sf1 >= 0) Setr3dssfs(ss, sf1, 43112);
									if (sc1 >= 0) Setr3dsscs(ss, sc1, 43112);
								}
							}
						}
					}
				}
			}
		}
		else
		{
			if (!ld->backsector || origTEX[i] || !(ld->mtflags & R3D_MISLOTEX)) continue;
 
			// v220-5 check if R3D_16 should have ML_IGNOREMISSINGLO set
			if (   ld->specType == R3D_16
				&& !(ld->ml_flags & ML_IGNOREMISSINGLO)
				&& ((ld->backsector->specFlags  & 0x3FFFC) == 0x1001C
				||  (ld->frontsector->specFlags & 0x3FFFC) == 0x1001C) )
			{
				// v220-6 test added
				if ( ((ld->backsector->specFlags & 0x3FFFC) == 0x1001C
					&& !(ld->frontsector->specFlags & 0x10004)
					&& ld->backsector->floorheight > ld->frontsector->floorheight)
					||
					( (ld->frontsector->specFlags & 0x3FFFC) == 0x1001C
					&& !(ld->backsector->specFlags & 0x10004)
					&& ld->frontsector->floorheight > ld->backsector->floorheight) )
				{
					SetMissinglo(ld, 43113);
				}
			}
			else
			// v220-5 check if R3D_00 should be R3D_01
			if (   ld->specType == R3D_00
				&& ld->frontsector->tag == ld->backsector->tag
				&& (ld->frontsector->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR
				&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
			{
				// if the line has the same floorsec each side then it will
				// have no rendered height even if the logical heights differ
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (   ss->floorsec >= 0
						&& sectors[ss->floorsec].tag == ld->frontsector->tag)
					{
						Settype(ld, R3D_01, 43114);
						break;
					}
				}
			}
		}
	}
}

// v220-17
static RDBOOL subfn_checkSsecs(sector_t *sc, int ovsec)
{
	subsector_t	*ss;
	ssline_t	*sl;
	seg_t		*sg;
	int			i, k, m, ln;

	ss_validcount++;
	hitssecptr = 0;

	for (i = 0; i < sc->linecount; i++)
	{
		ln = sc->lines[i] - lines;
		if (excludeline[ln]) continue;
		sl = &sslines[ln];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if ((ss->sector->specFlags & 0x3040C) != 0xC) continue;
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;
			hitssec[hitssecptr++] = sl->list[k];
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->backsector && excludeline[sg->linedef - lines]) continue;
				if (sg->frontsector != sc) return false;
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
		SetOverride(&subsectors[hitssec[i]], ovsec, 43173);

	return hitssecptr > 0;
}

static void subfn_notexType(sector_t *sc, int *lnacount)
{
	sector_t	*os, *ch;
	line_t		*ld;
	int			c, i;

	CLR_hitline();

	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (lineflags[ld - lines]) return;
		if (ld->mtflags & R3D_MISHITEX) return;	// v220-20 add test

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (!ch) return;
		if (ch->specFlags & SF_MOVSECTOR) return;
		if (ch->floorheight > sc->floorheight) return;
		if (ch->floorheight == sc->floorheight)
		{
			if ((ch->specFlags & 0x3FF8C) != 0xC) return;
			excludeline[ld - lines] = 1;
			hitline[ld - lines] = 1; c++;
			continue;
		}
		if (!os)
		{
			os = ch;
		}
		else if (os != ch)
		{
			if (os->floorheight != ch->floorheight) return;
			if (os->ceilingheight != ch->ceilingheight) return;
		}

		hitline[ld - lines] = 1; c++;
	}

	if (c < 3) return;

	int k, m, rc, box[4];

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1 && !excludeline[ld - lines])
			{
				lnum = ld - lines; break;
			}
		}

		if (lnum < 0) return;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			RDBOOL valid = subfn_checkSsecs(sc, os - sectors);

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (hitline[ld - lines] == 4)
				{
					if (valid)
					{
						for (m = 0; m < *lnacount; m++)
						{
							if (linearray[m] == ld - lines) goto skip_count;
						}

						linearray[ (*lnacount)++] = ld - lines;
skip_count:;
					}

					if (!excludeline[ld - lines]) hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-6 mop up unset lines where the floorsec needs to be set
//        These are lines with a missing lower texture that work alone
//        For this reason (i.e. don't fit any particular category)
//        they are processed last.
static void subfn_checkUnsetLowMissingLines()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			i, k;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   lineflags[i]
			|| lxarray[i]
			|| !(ld->specType == R3D_00 || ld->specType == R3D_03)
			|| origTEX[i] & 0x55
			|| !(bs = ld->backsector)
			|| (fs = ld->frontsector) == bs
			|| !(k = fs->floorheight - bs->floorheight)
			|| fs->ceilingheight == bs->ceilingheight
			|| fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR
			|| fs->floorpic == skyflatnum
			|| bs->floorpic == skyflatnum
		// v220-16 add 2 heightsec tests
			|| (fs->heightsec >= 0 && bs->heightsec <  0)
			|| (fs->heightsec <  0 && bs->heightsec >= 0)
			|| fs->stptr >= 0
			|| bs->stptr >= 0
			|| Spans1sLines(ld))
		{
			continue;
		}

		if (   (k > 0 && (fs->specFlags & 0xC) == 0xC && !(bs->specFlags & 0xC))
			|| (k < 0 && (bs->specFlags & 0xC) == 0xC && !(fs->specFlags & 0xC)) )
		{
			subsector_t	*ss;
			sector_t	*s1 = k < 0 ? bs : fs;
			sector_t	*s2 = k > 0 ? bs : fs;
			ssline_t	*sl = &sslines[i];
			RDBOOL		valid = true;
			byte		type = ld->specType;

			// v220-17 test added
			if (abs(k) <= 8*FRACUNIT)
			{
				Settype(ld, R3D_36, 43125);
				SetLFG(ld, LFG_NEWSET, 43125);
				continue;
			}

			ss_validcount++;
			hitssecptr = 0;

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					ss->ssvalidcount = ss_validcount;

					if (  !SSECunset(ss)
						&& ss->floorsec != s2 - sectors)
					{
						valid = false; break;
					}

					hitssec[hitssecptr++] = sl->list[k];
				}
			}

			if (valid)
			{
				sector_t	*chk, *other, *test = NULL;

				// v220-16 add test
				if (ld->mtflags & R3D_MISLOTEX && !(ld->mtflags & ~R3D_MISLOTEX))
				{
					test  = (fs->specFlags & 0xC) == 0xC ? fs : bs;
					other = test == fs ? bs : fs;
					chk = ld->mtflags & R3D_MISLTEX1 ? bs : fs;

					if (   chk != test
						|| !SecCtrl(test - sectors)
						|| test->specFlags & 0x10)
					{
						test = NULL;
					}
				}

				for (k = 0; k < hitssecptr; k++)
				{
					ss = &subsectors[hitssec[k]];

					if (test) Setr3dsslfsec(ss, test - sectors, 43125);
					// v220-17 floorheight test test added
					if (!test && ss->sector->floorheight >= s2->floorheight) continue;
					if (ss->sector == s1) Setr3dssfs(&subsectors[sl->list[k]], s2 - sectors, 43125);
				}

				if (!test) type = R3D_2D; else if (test && !origTEX[ld - lines]) type = R3D_3F;

				Settype(ld, type, 43126);
			}
		}
	}
}

void subfn_SetSSECFloorsec(line_t *ld, sector_t *chk, sector_t *set)
{
int setSsecList(sector_t *sc, int sconly);
	setSsecList(chk, true);

	for (int i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		Setr3dssfs(ss, set - sectors, 43186);
	}
	
	Settype(ld, R3D_3F, 43186);
}

// v220-6 check types that can only be done at end
void PS_CheckEndLineTypes()
{
	sector_t *sec;
	line_t	 *ld;
	int		 lnacount = 0;
	int		 d, i, k, n, n1, n2;

	//-----------------------------

	// v220-18 add routine ->
	CLR_excludeline();	// Used with subfn_notexType()

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && (sectors[i].specFlags & 0x3FFFC) == 0x1C)
		{
			subfn_notexType(&sectors[i], &lnacount);
		}
	}

	// Set any lines pointed to by subfn_notexType()
	for (i = 0; i < lnacount; i++)
	{
		ld = &lines[linearray[i]];
		Settype(ld, R3D_3F, 43174);
		SetLFG(ld, LFG_NEWSET, 43174);
	}

	// <---------------------------

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		n1 = 0, n2 = 0;

		// v220-33 check for lines with missing tex that should not be rendered
		if (   ld->specType == R3D_00
			&& lineflags[i] == LFG_MARKED
			&& !origTEX[i]
			&& ld->backsector
			&& ld->frontsector != ld->backsector
			&& vtxflags[ld->v1 - vertexes] == 3
			&& vtxflags[ld->v2 - vertexes] == 3
			&& vertexowners[ld->v1 - vertexes].num == 2
			&& vertexowners[ld->v2 - vertexes].num == 2)
		{
			if (   ld->mtflags == 0x80
				&& ld->frontsector->specFlags == 0x0103809C
				&& ld->backsector->specFlags == 0x00008394)
			{
				subfn_SetSSECFloorsec(ld, ld->frontsector, ld->backsector);
			}
			else
			if (   ld->mtflags == 0x08
				&& ld->backsector->specFlags == 0x0103809C
				&& ld->frontsector->specFlags == 0x00008394)
			{
				subfn_SetSSECFloorsec(ld, ld->backsector, ld->frontsector);
			}
		}

		// v220-6 map authors sometimes put the wrong sector in trigger lines
		if (   ld->special
			&& ld->frontsector == ld->backsector
			&& lxarray[i] == 0x28	// touching a vertex but not crossing a line
			&& ld->lnflags2 == LNF2_OWNEQUALTOONE
			&& (ld->frontsector->specFlags & 0x3000C) == 0x1000C)
		{
			ssline_t *sl = &sslines[i];
			sector_t *chk = NULL;

			for (d = 0; d < sl->num; d++)
			{
				subsector_t *ss = &subsectors[sl->list[d]];
				seg_t		*sg = &segs[ss->firstline];

				for (n = 0; n < ss->linecount; n++, sg++)
				{
					if (sg->frontsector != ld->frontsector)
					{
						if (!chk) chk = sg->frontsector;
						else
						if (chk != sg->frontsector) { chk = NULL; break; }
					}
				}
			}

			if (chk && !(chk->specFlags & SF_MOVSECTOR))
			{
				for (d = 0; d < sl->num; d++)
					ChangeSSECsec(&subsectors[sl->list[d]], chk, 43115);
			}
		}

		if (  !ld->backsector
			|| ld->frontsector == ld->backsector
			|| ld->specType >= R3D_40
			|| ld->frontsector->stptr >= 0
			|| ld->backsector->stptr >= 0
			|| ld->frontsector->heightsec >= 0
			|| ld->backsector->heightsec >= 0
			|| Spans1sLines(ld))
		{
			continue;
		}

		if (   (origTEX[i] & 0x33) == 0x22
			&& (ld->specType == R3D_00 || ld->specType == R3D_1D)
			&& (d = ld->backsector->floorheight - ld->frontsector->floorheight)
			&& abs(d) > FRACUNIT
			&& !(ld->frontsector->specFlags & SF_MOVFLOOR)
			&& !(ld->backsector->specFlags & SF_MOVFLOOR)
			&& !(lineflags[i] & 0xFFEF)
			&& ((n1 = vertexowners[ld->v1 - vertexes].num) > 1
			||  (n2 = vertexowners[ld->v2 - vertexes].num) > 1))
		{
			sector_t *s = d > 0 ? ld->frontsector : ld->backsector;
			vertex_t *v = ld->v1;
			int		 secn = -1;

			if (s->specFlags & SF_SSLINETRUE) continue;

			if (n2 > n1) n1 = n2, v = ld->v2;

			for (n = k = 0; k < n1; k++)
			{
				if (sectors[vertexowners[v - vertexes].list[k]].specFlags & SF_NOLOTEX)
					n++;
				else
				{
					if (secn == -1)
						secn = vertexowners[v - vertexes].list[k];
					else if (secn != vertexowners[v - vertexes].list[k])
					{
						secn = -2;
						break;
					}
				}
			}

			if (secn == -1 && ld->specType == R3D_00)
			{
				goto set_linetype;
			}

			if (   secn >= 0
				&& sectors[secn].floorheight == s->floorheight)
			{
				{
					ssline_t *sl = &sslines[i];

					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->sector != s)
							Setr3dssfs(ss, s - sectors, 43116);
					}
				}

set_linetype:;
				if (ld->specType == R3D_00)
				{
					if (ld->frontsector->floorheight != ld->backsector->floorheight)
						Settype(ld, R3D_1D, 43117);
					else
						Settype(ld, R3D_03, 43118);
				}

				SetMissinglo(ld, 43119);
			}
		}
		// v220-6 look for inheritance of lcsec and lfsec. This is used
		// to close gaps with certain special effects. With a software
		// renderer the gaps are not visible to the uncritical eye.
		else if (!origTEX[i]
			&& !lxarray[i]
			&& ld->frontsector->specFlags == 0xC
			&& ld->backsector->specFlags  == 0xC
			&& SecCtrl(ld->frontsector - sectors)
			&& SecCtrl(ld->backsector  - sectors)
			&& ld->frontsector->floorheight   == ld->backsector->floorheight
			&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
		{
			ssline_t	*sl = &sslines[i];
			subsector_t	*ref = NULL;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (   ss->lcsec >= 0
					&& ss->lfsec >= 0)
				{
					ref = ss;
					break;
				}
			}

			if (!ref) continue;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->lcsec == -1) Setr3dsslcsec(ss, ref->lcsec, 43120);
				if (ss->lfsec == -1) Setr3dsslfsec(ss, ref->lfsec, 43121);
			}
		}
	}


	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->stptr >= 0 || sec->heightsec >= 0) continue;

		if ((sec->specFlags & 0x008DFF0C) == 0x00808304)
		{
			sector_t *osec = sec->lines[0]->frontsector == sec ? sec->lines[0]->backsector : sec->lines[0]->frontsector;

			if (   osec
				&& osec->stptr < 0
				&& osec->heightsec < 0
				&& (osec->specFlags & 0xFF0FC) == 0x00098090)
			{
				RDBOOL valid = true;

				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (   ld->specType != R3D_00
						|| lineflags[ld - lines])
					{
						valid = false; break;
					}
				}

				if (valid)
				{
					for (k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k];
						Settype(ld, R3D_14, 43122);
						SetLFG(ld, LFG_NEWSET, 43123);
					}
				}
			}
		}
		else
		if (   (sec->specFlags & 0x3000C) == 0x10008
			&& !sec->special
			&& !sec->tag
			&& sec->ceilingheight - sec->floorheight > MAX_SECHEIGHT)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				int m, fs;

				ld = sec->lines[k];

				if (   ld->specType == R3D_00
					&& ld->lnflags2 == 0x0C000000
					&& ld->mtflags & R3D_MISLOTEX
					&& !origTEX[ld - lines]
					&& Vtxs2(ld))
				{
					ssline_t *sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						if ( (fs = subsectors[sl->list[m]].floorsec) >= 0)
							break;
					}

					if (fs == -1)
					{
						Settype(ld, R3D_14, 43124);
					}
				}
			}
		}
	}

	// v220-21 Run just prior to routine that immediately
	//         follows this one that mops up unset lines.
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& !SecCtrl(i)
			&& sec->heightsec < 0
			&& (sec->specFlags & 0xFFFFF) == 0x809C)
		{
			sector_t *ch, *os = NULL;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (lineflags[ld - lines]) { os = NULL; break; }
				if (!ld->backsector) { os = NULL; break; }
				if (ld->frontsector == ld->backsector) continue;
				if (abs(ld->frontsector->floorheight - ld->backsector->floorheight) < 8*FRACUNIT) { os = NULL; break; }

				ch = ld->frontsector->floorheight > ld->backsector->floorheight ? ld->backsector : ld->frontsector;

				if (ch->specFlags & SF_MOVSECTOR) { os = NULL; break; }
				if (ch->heightsec >= 0) { os = NULL; break; }
				if (ch->specFlags & 0xC && !(ch->specFlags & 0x400)) { os = NULL; break; }

				if (!os) os = ch;
				else
				if (   os != ch
					&& (os->floorpic != ch->floorpic
					|| os->ceilingpic != ch->ceilingpic
					|| os->floorheight != ch->floorheight
					|| os->ceilingheight != ch->ceilingheight))
				{
					os = NULL; break;
				}
			}

			if (os)
			{
				subsector_t *ss;
				seg_t		*sg;
				ssline_t	*sl;
				int			c, m, n;

				validcount++;
				ss_validcount++;
				hitssecptr = 0;

				for (c = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						ss = &subsectors[sl->list[m]];

						if (ss->sector != sec) continue;
						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						if (!SSECunset(ss)) { os = NULL; break; }

						for (sg = &segs[ss->firstline], n = 0; n < ss->linecount; n++, sg++)
						{
							if (ld = sg->linedef)
							{
								if (sg->frontsector != sec) { os = NULL; break; }

								if (ld->validcount != validcount)
								{
									ld->validcount = validcount;
									linearray[c++] = ld - lines;
								}
							}
						}

						if (!os) break;

						hitssec[hitssecptr++] = sl->list[m];
					}

					if (!os) break;
				}

				if (os)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						ss = &subsectors[hitssec[k]];
						Setr3dsslfsec(ss, os - sectors, 43191);
					}

					for (k = 0; k < c; k++)
					{
						ld = &lines[linearray[k]];

						if (origTEX[ld - lines] & 0x22)
							Settype(ld, R3D_03, 43191);
						else
							Settype(ld, R3D_3F, 43191);

						SetLFG(ld, LFG_NEWSET, 43191);
					}
				}
			}
		}
	}

	subfn_checkUnsetLowMissingLines();
}

// v220-16
static fixed_t getLowestSkyFloor(sector_t *s, fixed_t fht)
{
	line_t	*ld;
	int		i, ht;

	for (i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (   ld->frontsector->floorpic == skyflatnum
			&& ld->frontsector->floorheight < fht)
		{
			ht = ld->frontsector->floorheight;
			fht = getLowestSkyFloor(ld->frontsector, ht);
		}

		if (   ld->backsector
			&& ld->backsector->floorpic == skyflatnum
			&& ld->backsector->floorheight < fht)
		{
			ht = ld->backsector->floorheight;
			fht = getLowestSkyFloor(ld->backsector, ht);
		}
	}

	return fht;
}

// v220-6 check;
//	1) low sky walls
//	2) verify ML_IGNOREMISSINGLO/ML_IGNOREMISSINGHI
//	3) R3D_14/R3D_16/R3D_1D loops
//  4) suspect map hacks

void PS_CheckEndSectorTypes()
{
	sector_t	*osec, *sec;
	sector_t	*fs, *bs;
	ssline_t	*sl;
	line_t		*ld;
	byte		*valline;
	int			*tmparray = (int *)malloc(numsectors * sizeof(int));
	int			c, i, k;
	int			lfc = 0;

	// 1) Sky floors (modified v220-16)
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->floorpic == skyflatnum
			&& sec->floorheight < sec->ceilingheight
			&& sec->R3Dfloorsec == -1
			&& sec->R3Dfloorheight == DDMAXINT)
		{
			RDBOOL	found = false;
			int		ht = sec->floorheight;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED)
				{
					byte sidetx = ld->frontsector == sec ? 0x01 : 0x10;
					int  bht = ld->frontsector == sec ? ld->backsector->floorheight : ld->frontsector->floorheight; 

					if (origTEX[ld - lines] & sidetx && bht > sec->floorheight) { found = true; break; }
				}
			}

			// iterate to see if there's a lower value
			if (found && (ht = getLowestSkyFloor(sec, ht)) < sec->floorheight)
			{
				Setr3dfh_byH(sec, ht, false, 43127);
			}
		}
	}

	// Line loops
	// 2) verify ML_IGNOREMISSINGLO/ML_IGNOREMISSINGHI
	// (identify lines for 3)
	valline = (byte*)calloc(numlines, 1);
	SET_countsec(0);

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->ml_flags & ML_IGNOREMISSINGLO)
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				if (subsectors[sl->list[k]].floorsec >= 0)
				{
					ResetMissinglo(ld, 43128);
					break;
				}
			}
		}

		if (ld->ml_flags & ML_IGNOREMISSINGHI && !(lineflags[ld - lines] & LFG_OVERRIDE))
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				if (subsectors[sl->list[k]].ceilsec >= 0)
				{
					ld->ml_flags &= ~ML_IGNOREMISSINGHI;
					break;
				}
			}
		}

		if (   ld->ml_flags & ML_TWOSIDED
			&& (fs = ld->frontsector) != (bs = ld->backsector))
		{
			if (   ld->ovflags & OVF_NOSKYWALL
				&& vertexowners[ld->v1 - vertexes].num > 2
				&& vertexowners[ld->v2 - vertexes].num > 2
				&& ((origTEX[i] & 0x04 &&  fs->skyfix && !bs->skyfix)
				||  (origTEX[i] & 0x40 && !fs->skyfix &&  bs->skyfix))
				&& Spans1sLines(ld))
			{
				ld->ovflags &= ~OVF_NOSKYWALL;
			}

			if (  !(origTEX[i] & 0x55)
				&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
				&& (ld->specType == R3D_14
				||  ld->specType == R3D_16
				||  ld->specType == R3D_1D)
				&& !(ld->spflags & SPF_CONTROLLINE)
				&& fs->ceilingheight == bs->ceilingheight
				&& fs->floorheight != bs->floorheight
				&& fs->ceilingheight - fs->floorheight >= 4*FRACUNIT
				&& bs->ceilingheight - bs->floorheight >= 4*FRACUNIT
				&& !(fs->specFlags & SF_MOVFLOOR)
				&& !(bs->specFlags & SF_MOVFLOOR)
				&& (fs->specFlags & SF_NOHITEX || bs->specFlags & SF_NOHITEX)
				&& fs->stptr < 0
				&& bs->stptr < 0)
			{
				valline[i] = true;
				if ((fs->specFlags & 0x41C) == 0x1C) countsec[fs - sectors]++;
				if ((bs->specFlags & 0x41C) == 0x1C) countsec[bs - sectors]++;
			}
		}
	}

	// 3) check loops
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  (sectors[i].specFlags & 0x3001C) == 0x20010
			&& sectors[i].ceilingpic != skyflatnum
			&& !SecCtrl(i))
		{
			tmparray[lfc++] = i;
		}

		if (countsec[i] > 2)
		{
			sector_t *chk, *s = NULL;
			RDBOOL	 reversed = false;
			int		 m, n, ring_count;
			int		 lnum;

repeat_check:;	// could be multiple loops in any one sector

			CLR_hitline();

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];
				n = ld - lines;

				if (valline[n])
				{
					lnum = n; c++;
					hitline[n] = true;

					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!s) s = chk;
					else if (s != chk) { c = 0; break; }
				}
			}
			
			if (c > 2)
			{
				if (!(s->specFlags & SF_SSLINETRUE) && s->specFlags & SF_NOLOTEX)
				{
					c = 0;
				}
			}

			if (c > 2 && subfn_CheckForRing(hitline, s, lnum, &ring_count, NULL, NULL, NULL))
			{
				RDBOOL	valid = true;
				int		R3D_1Dtype = 0;	// v220-12

				reversed = sec->floorheight > s->floorheight;
				valline[lnum] = false;	// insurance
				ss_validcount++;

				for (c = k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];
					lnum = ld - lines;

					if (hitline[lnum] == 4)
					{
						sl = &sslines[lnum];
						valline[lnum] = false;
						linearray[c++] = lnum;

						if (valid)
							valid = verifySSEC(sl, sec, reversed ? s : NULL);

						if (reversed && ld->specType == R3D_1D) R3D_1Dtype++;
					}
				}

				if (valid)
				{
					R3D_1Dtype = R3D_1Dtype == c;

					for (k = 0; k < c; k++)
					{
						lnum = linearray[k];
						sl = &sslines[lnum];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->ssvalidcount == ss_validcount) continue;

							ss->ssvalidcount = ss_validcount;
							Reset_SSEC(sl->list[m], 50004);

							// v220-7 don't set the floorsec if lfsec is already set the same
							if (ss->sector == sec && ss->lfsec != s - sectors)
								Setr3dssfs(ss, s - sectors, 43129);
						}

						if (origTEX[lnum] & 0x22)
						{
							if (!R3D_1Dtype)
								Forcetype(&lines[lnum], R3D_03, 43130);
						}
						else
							Forcetype(&lines[lnum], R3D_3F, 43131);

						SetLFG(ld, LFG_NEWSET, 43132);
					}
				}
				
				goto repeat_check;
			}
		}
	}

	// 4) suspect map hacks
	for (i = 0; i < lfc; i++)
	{
		sector_t *secn[2] = { NULL, NULL };
		sector_t *secv[2] = { NULL, NULL };
		int m, n;

		sec = &sectors[tmparray[i]];

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (   !lineflags[ld - lines]
				&& (osec = ld->frontsector == sec ? ld->backsector : ld->frontsector)
				&& osec->floorheight == sec->ceilingheight
				&& ld->lnflags2 & LNF2_OWNGTRTHANTWO)
			{
				for (m = 0; m < 2; m++)
				{
					int vn = m ? ld->v2 - vertexes : ld->v1 - vertexes;
					vertexowner_t *vo = &vertexowners[vn];

					if (vo->num > 3)
					{
						for (n = 0; n < vo->num; n++)
						{
							if (  (sectors[vo->list[n]].specFlags & 0xF001C) == 0x1C)
							{
								secn[m] = &sectors[vo->list[n]];
								secv[m] = m ? osec : sec;

								// v220-9 add single sided line test
								if (!(sectors[vo->list[n]].specFlags & SF_SSLINETRUE))
									Settype(ld, R3D_16, 43133);
							}
						}
					}
				}
			}
		}

		// This is the penalty for trying to render suspect map hacks
		// that relied on the vagaries of software rendering (rare)
		if (   secn[0]
			&& secn[0] == secn[1]
			&& secv[0] != secv[1]
			&& secv[0] != secn[0]
			&& secv[1] != secv[0]
			&& secv[0]->ceilingpic != skyflatnum
			&& secv[1]->ceilingpic != skyflatnum)
		{
			fs = NULL, bs = NULL;

			for (k = 0; k < secn[0]->linecount; k++)
			{
				ld = secn[0]->lines[k];

				if (ld->specType == R3D_3F)
				{
					bs = ld->frontsector == secn[0] ? ld->backsector : ld->frontsector;

					if (bs->ceilingheight < secn[0]->ceilingheight)
					{
						if (!fs) fs = bs; else if (fs != bs) goto end_bs;
					}
				}
			}

			if (   fs
				&& fs->ceilingpic != skyflatnum
				&& fs->ceilingheight > secv[0]->ceilingheight
				&& fs->ceilingheight > secv[0]->ceilingheight)
			{
				if (secv[0]->ceilingheight < secv[1]->ceilingheight)
				{
					sector_t *t = secv[0]; secv[0] = secv[1]; secv[1] = t;
				}

				for (k = 0; k < fs->linecount; k++)
				{
					ld = fs->lines[k];
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == fs)
						{
							Setr3dsscs(ss, secv[0] - sectors, 43134);
							Setr3dsslcsec(ss, secv[1] - sectors, 43134);
						}
					}
				}

				for (k = 0; k < secn[0]->linecount; k++)
				{
					ld = secn[0]->lines[k];
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == secn[0])
						{
							Setr3dsscs(ss, secv[0] - sectors, 43135);
							Setr3dsslcsec(ss, secv[1] - sectors, 43135);
						}
					}
				}
			}
		}
end_bs:;
	}

	M_Free(tmparray);
	M_Free(valline);
}

static RDBOOL ssecHasSameSided(subsector_t	*ss)
{
	line_t	*ld;
	seg_t	*sg;
	int		i;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (  (ld = sg->linedef)
			&& ld->frontsector == ld->backsector
			&& ld->frontsector == ss->sector)
		{
			return true;
		}
	}
	return false;
}

// v220-6 1) mark segs that must not have missing textures drawn
//        2) check for double referenced textureless ceiling sectors (rare)
void PS_CheckR3D_16Type()
{
	subsector_t	*ss;
	sector_t	*fs, *bs;
	seg_t		*sg;
	line_t		*ld;
	int			i, k;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		// v220-6 add check to prevent the sector's floor overriding
		// a floorsec with a moving floor (see rend_main.c)
		if (ss->floorsec >= 0)
		{
			if (   SecCtrl(ss->sector - sectors)
				|| !(ss->sector->specFlags & SF_MOVFLOOR)
				|| sectors[ss->floorsec].specFlags & SF_MOVFLOOR)
			{
				ss->ssflags |= SSF_USEFSEC;
			}
			// v220-16 function added
			else if (ss->ceilsec == ss->floorsec
				&& ss->sector->specFlags & SF_MOVFLOOR
				&& ssecHasSameSided(ss))
			{
				ss->ssflags |= SSF_USEFSEC;
			}
		}

		if (ss->ceilsec >= 0 || ss->sector->specFlags & SF_MOVSECTOR) continue;

		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (  (ld = sg->linedef)
				&& ld->specType == R3D_16
				&& !origTEX[ld - lines]
				&& (fs = sg->frontsector)
				&& fs->ceilingpic == skyflatnum
				&& (bs = sg->backsector)
				&& bs->ceilingpic != skyflatnum
				&& fs->floorheight == bs->floorheight)
			{
				if (   fs != ss->sector
					&& !(ss->sector->specFlags & SF_NOLOTEX)
					&& (fs->specFlags & 0x3001C) == 0x1C)
				{
					sg->sgflags |= R3DSEGF_IGNORE;
					continue;
				}

				if (   ss->ceilsec != bs - sectors
					&& fs->specFlags & SF_NOHITEX
					&& (bs->specFlags & 0x1C) != 0x1C)
				{
				//	RDBOOL	found = false;
				//	int		lnum = ld - lines;

					Setr3dsscs(ss, bs - sectors, 43136);
				}
			}
		}
	}
}

// v220-6 Find lines that have the same ceiling height both sides after taking
//        a ceilsec revision into account and mark it not to draw a high wall.
void PS_CheckInvalidHiWalls()
{
	line_t	*ld;
	int		i, k;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_0A) ld->specType = R3D_00;

		if (   !(ld->ml_flags & ML_TWOSIDED)
			|| ld->frontsector == ld->backsector)
		{
			continue;
		}

		if (!(ld->ml_flags & ML_IGNOREMISSINGHI) && !(origTEX[i] & 0x44))
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ceilsec >= 0)
				{
					sector_t *s = ld->frontsector == ss->sector ? ld->backsector : ld->frontsector;
					int ht = sectors[ss->ceilsec].floorheight;

					if (ht == s->ceilingheight)
					{
						ld->lnflags2 |= LNF2_NORENDHI;
					}
				}
			}
		}

		if (!(ld->ml_flags & ML_IGNOREMISSINGLO) && !(origTEX[i] & 0x11))
		{
			ssline_t *sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->floorsec >= 0)
				{
					sector_t *s = ld->frontsector == ss->sector ? ld->backsector : ld->frontsector;
					int ht = sectors[ss->floorsec].floorheight;

					if (ht == s->floorheight)
					{
						// v220-10 test added
						if (   ld->specType == R3D_00
							&& sectors[ss->floorsec].specFlags & SF_MOVFLOOR
							&& ss->sector->specFlags & SF_MOVFLOOR)
						{
							ld->spflags &= ~SPF_NOTEX;
						}
						else
						{
							ld->lnflags2 |= LNF2_NORENDLOW;
						}
					}
				}
			}
		}
	}
}

// v220-10 some tests removed
// Note: R3D_0A type reset prior to calling (see PS_CheckInvalidHiWalls)
void PS_CheckSSLinesAccessingMidTEX()
{
	sector_t	*sec;
	line_t		*ld;
	int			i, k, m, n;

	if (R3D_maptype) return;

	// 1) check disconnected lines for owming sector
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->heightsec >= 0
			|| sec->R3Dceilheight > DDMININT
			|| sec->R3Dceilsec >= 0
			|| sec->R3Dfloorheight < DDMAXINT
			|| sec->R3Dfloorsec >= 0)
		{
			continue;
		}

		for (k = 0; k < sec->linecount; k++)
		{
			if (  (ld = sec->lines[k])
				&& ld->frontsector == ld->backsector
				&& !lineflags[ld - lines]
				&& !origTEX[ld - lines]
				&& !lxarray[ld - lines])
			{
				int vnum1 = ld->v1 - vertexes;
				int vnum2 = ld->v2 - vertexes;

				if (   vtxarray[vnum1] > 1
					&& vtxarray[vnum2] > 1
					&& (vertexowners[vnum1].num > 1
					||  vertexowners[vnum2].num > 1))
				{
					line_t *li = NULL;

					if ( (li = subfn_FindLineConnected(ld, 31)) )
					{
						// The subsectors will have to be checked
						ssline_t *sl = &sslines[ld - lines];
						RDBOOL	 found = false;

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector == ld->frontsector)
							{
								seg_t *sg = &segs[ss->firstline];

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if (  (li = sg->linedef)
										&& li->frontsector == li->backsector
										&& origTEX[li - lines] == 0x22)
									{
										found = true; break;
									}
								}
							}
							if (found) break;
						}

						if (!found || !li || li->frontsector == sec) continue;
					}

					if (li || (li = subfn_FindLineConnected(ld, 32)) )
					{
						if ( !(li->frontsector->specFlags & 0x1000C)
							&& li->frontsector->floorheight   < sec->floorheight
							&& li->frontsector->ceilingheight > sec->ceilingheight)
						{
							ssline_t *sl = &sslines[ld - lines];

							ss_validcount++;

							for (m = 0; m < sl->num; m++)
							{
								subsector_t *ss = &subsectors[sl->list[m]];

								if (   ss->ssvalidcount != ss_validcount
									&& ss->floorsec < 0)
								{
									ss->ssvalidcount = ss_validcount;

									if (ss->sector != li->frontsector)
									{
										SetOverride(ss, li->frontsector - sectors, 43137);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	// v220-7 Check textureless R3D_14, R3D_16 lines in a floor mov sector
	for (ld = lines, i = 0; i < numlines; i++, ld++)
	{
		if (   !origTEX[i]
			&& (ld->specType == R3D_14 || ld->specType == R3D_16)
			&& ld->ml_flags & ML_TWOSIDED)	// insurance
		{
			subsector_t *chk = NULL;
			ssline_t	*sl = &sslines[i];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->lfsec >= 0 || ss->floorsec >= 0)
				{
					chk = ss; break;
				}
			}

			if (chk && (chk->sector == ld->frontsector || chk->sector == ld->backsector))
			{
				sector_t *s = chk->sector == ld->frontsector ? ld->backsector : ld->frontsector;

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->sector == s)
					{
						if (   (chk->floorsec >= 0 && ss->lfsec == chk->floorsec)
							|| (chk->lfsec >= 0 && ss->lfsec == chk->lfsec))
						{
							Forcetype(ld, R3D_3F, 43138);
						}
					}
				}
			}
		}
	}
}

// v220-7 check for any multisec lo missing loops that are unset (rare)
//        This is limited to missing tex lines associated with a same sector line
void PS_CheckMultiSecMissing()
{
	sector_t	*sec, *chk;
	int			chkseccnt;
	int			chksec[MAX_SEC_PER_VTX];
	int			startline;
	int			d, i, k, m;
	int			lsc, box[4], lnum, rc;

	//-----------------------------------

	for (sec = sectors, i = 0; i < numsectors; i++, sec++)
	{
		if (   secdone[i]
			|| sec->specFlags & SF_MOVSECTOR
			|| sec->stptr >= 0)
		{
			continue;
		}

		CLR_hitline();
		chkseccnt = 0;

		for (lsc = d = k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			lnum = ld - lines;

			if (lineflags[lnum] || ld->specType > R3D_03) goto end_check;
			if (!(ld->ml_flags & ML_TWOSIDED)) continue;

			if (ld->frontsector == ld->backsector)
			{
				linearray[lsc++] = lnum; continue;
			}

			if (ld->mtflags)
			{
				if (   origTEX[ld - lines]
					|| ld->mtflags & (R3D_MISHITEX|R3D_INVTEXANY))
				{
					goto end_check;
				}

				if (chkseccnt < MAX_SEC_PER_VTX)
				{
					int secn = ld->frontsector == sec ? ld->backsector - sectors : ld->frontsector - sectors;

					startline = lnum;
					hitline[lnum] = 1; d++;

					for (m = 0; m < chkseccnt; m++)
					{
						if (chksec[m] == secn) goto chk_found;
					}

					chksec[chkseccnt++] = secn;
				}
			}
chk_found:;
		}

		// There must be more than one referenced sector
		// and these sectors must have the same values.
		if (   !lsc
			|| chkseccnt < 2
			|| !subfn_CheckForRing(hitline, sec, startline, &rc, box, NULL, NULL)
			|| rc != d)	// all lines identified must form the ring
		{
			continue;
		}

		chk = &sectors[chksec[0]];

		// The ceiling values must match the sector being tested
		// The referenced sectors' floorheights must be less than the test sector's
		if (   chk->floorheight >= sec->floorheight
			|| chk->ceilingheight != sec->ceilingheight
			|| chk->ceilingpic != sec->ceilingpic)
		{
			chkseccnt = 0;
		}

		for (k = 1; k < chkseccnt; k++)
		{
			if (   chk->floorheight   != sectors[chksec[k]].floorheight
				|| chk->ceilingheight != sectors[chksec[k]].ceilingheight
				|| chk->floorpic      != sectors[chksec[k]].floorpic
				|| chk->ceilingpic    != sectors[chksec[k]].ceilingpic)
			{
				chkseccnt = 0; break;
			}
		}

		// All the same sided lines must be referenced by a missing tex line
		if (chkseccnt)
		{
			byte	*lscheck = (byte*)calloc(lsc, 1);
			int		n, p;

			ss_validcount++;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t	*ld = sec->lines[k];

				lnum = ld - lines;

				if (hitline[lnum] == 4)
				{
					ssline_t *sl = &sslines[lnum];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec && ss->ssvalidcount != ss_validcount)
						{
							seg_t *sg = &segs[ss->firstline];

							ss->ssvalidcount = ss_validcount;

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (  (ld = sg->linedef)
									&& ld->frontsector == ld->backsector)
								{
									for (p = 0; p < lsc; p++)
									{
										if (linearray[p] == ld - lines)
										{
											lscheck[p] = true;
										}
									}
								}
							}
						}
					}
				}
			}

			for (p = 0; p < lsc; p++)
			{
				if (!lscheck[p])
				{
					chkseccnt = 0; break;
				}
			}

			M_Free(lscheck);
		}

		if (chkseccnt < 2) continue;

		for (d = k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			lnum = ld - lines;

			if (hitline[lnum] == 4)
			{
				linearray[d++] = lnum;
			}
		}

		// These lines must not enclose an area 
		//(this is what makes this type of construction very rare).
		if (   LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, d, box, NULL, false, MSIA_NONE)
			|| LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, d, box, NULL, false, MSIA_REVERSED))
		{
			continue;
		}

		ss_validcount++;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			lnum = ld - lines;

			if (hitline[lnum] == 4)
			{
				ld->specType = R3D_3F;
			}
			else if (ld->frontsector == ld->backsector)
			{
				ssline_t *sl = &sslines[lnum];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];
					seg_t		*sg = &segs[ss->firstline];
					RDBOOL		valid = true;
					int			n;

					if (ss->ssvalidcount != ss_validcount)
					{
						// The subsector must contain only missing tex and same sided lines
						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if (  (ld = sg->linedef)
								&& hitline[ld - lines] != 4
								&& ld->frontsector != ld->backsector)
							{
								valid = false; break;
							}
						}

						ss->ssvalidcount = ss_validcount;

						if (valid)
						{
							Setr3dssfs(ss, i, 43175);
							Setr3dsslfsec(ss, chksec[0], 43175);
						}
					}
				}
			}
		}

end_check:;
	}
}

// v220-7 if a textureless line with a missing hi-tex is found then the adjacent
//        lines are checked to see if a matching type with a hi-texture exists.
//        If one is found then it is used to replace the missing texture.
//        This is limited to dealing with notex lines where one ceiling flat is
//        a sky sector and where the ceilings have a fixed height.
void PS_ReplaceMissingHi()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->specType == R3D_00
			&& !(lineflags[i] & (LFG_FIXED|LFG_NEWSET|LFG_3DTYPE))
			&& ld->ml_flags & ML_TWOSIDED
			&& ld->spflags & SPF_NOTEX
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& ld->frontsector != ld->backsector
			&& ld->mtflags & R3D_MISHITEX
			&& Vtfl1(ld)
			&& vtxarray[ld->v1 - vertexes] >= 2
			&& vtxarray[ld->v2 - vertexes] >= 2
			&& ((ld->frontsector->ceilingpic == skyflatnum)
			^   (ld->backsector->ceilingpic == skyflatnum))
			&& !(ld->frontsector->specFlags & SF_MOVCEIL)
			&& !(ld->backsector->specFlags & SF_MOVCEIL))
		{
			int		 side = ld->mtflags & R3D_MISHTEX1 ? 0 : 1;
			line_t	 *li;
			side_t	 *sd = NULL;
			int		 num[] = { 100, 101, 0 };	// v220-8 changed from 34, 35
			int		 i;
			short	 txn = 0;

			for (i = 0; num[i]; i++)
			{
				if ( (li = subfn_FindLineConnected(ld, num[i])) )
				{
					if (   origTEX[li - lines] & 0x04
						&& sides[ld->sidenum[side]].sector->floorheight == li->frontsector->floorheight
						&& sides[ld->sidenum[side]].sector->ceilingheight == li->frontsector->ceilingheight)
					{
						sd = &sides[li->sidenum[0]];
					}
					else
					if (   origTEX[li - lines] & 0x40
						&& sides[ld->sidenum[side]].sector->floorheight == li->backsector->floorheight
						&& sides[ld->sidenum[side]].sector->ceilingheight == li->backsector->ceilingheight)
					{
						sd = &sides[li->sidenum[1]];
					}

					if (sd)
					{
						if (R3D_edit)
							pr_OFinsert("ERROR: L%i has unset hi-tex; tex from L%i substituted\n", ld-lines, li-lines);

						sides[ld->sidenum[side]].toptexture = sd->toptexture;
						sides[ld->sidenum[side]].textureoffset = sd->textureoffset;
						sides[ld->sidenum[side]].rowoffset = sd->rowoffset;
						ld->mtflags &= ~R3D_MISHITEX;
						ld->spflags &= ~SPF_NOTEX;
						origTEX[ld - lines] |= (byte)(side ? 0x40 : 0x04);
						break;
					}
				}
			}
		}
	}
}

static int getBoundarySecnum(sector_t *s, int *sslist, int *sslistcnt)
{
	sector_t	*osec;
	line_t		*ld;
	int			i, k, m;
	
	*sslistcnt = 0;
	ss_validcount++;

	for (i = 0, osec = NULL; i < s->linecount; i++)
	{
		if (hitline[s->lines[i] - lines] == 4)
		{
			ssline_t *sl = &sslines[linearray[i]];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				
				if (ss->ssvalidcount != ss_validcount)
				{
					seg_t *sg = &segs[ss->firstline];

					ss->ssvalidcount = ss_validcount;
					sslist[*sslistcnt] = sl->list[k];

					if (*sslistcnt < 256) *sslistcnt += 1;

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (  (ld = sg->linedef)
							&& !hitline[ld - lines]
							&& !(sg->frontsector->specFlags & SF_NOHITEX))
						{
							if (!osec) osec = sg->frontsector;
							else if (osec != sg->frontsector)
							{
								return -1;
							}
						}
					}
				}
			}
		}
	}

	return osec ? osec - sectors : -1;
}

void PS_CheckNoHiMovSectors()
{
	sector_t	*sec, *osec;
	int			box[4];
	int			sslist[256], sslistcnt;
	int			rc, secn, seccnt = 0;
	int			c, i, k;
	int			*mrksects = (int*)calloc(numsectors, sizeof(int));

	// v220-7 First check for rings of R3D_1D lines
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->linecount > 2
			&& (sec->specFlags & 0xFFFFC) == 0x849C
			&& !SecCtrl(i)
			&& sec->ceilingpic != skyflatnum)
		{
			osec = NULL;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->specType == R3D_1D)
				{
					sector_t *chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!chk || !(chk->specFlags & SF_MOVCEIL)) { osec = NULL; break; }

					if (!osec) osec = chk;
					else if (osec != chk) { osec = NULL; break; }

					linearray[c++] = ld - lines;
				}
			}

			if (osec && c > 2)
			{
				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

				if (   subfn_CheckForRing(hitline, sec, linearray[0], &rc, box, NULL, NULL)
					&& rc == c
					&& LS_MarkSubsecsInAreaGeneral(osec, hitline, 0, NULL, box, -1, MSIA_NONE))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						if (ss->lcsec < 0)
						{
							ChangeSSECsec(ss, osec, 43139);
							Setr3dsslcsec(ss, sec - sectors, 43140);
						}
					}
				}
			}
		}
	}

	// v220-7 check validity of R3D_16 line rings
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->linecount > 2
			&& (sec->specFlags & 0xFFFFC) == 0x819C
			&& !SecCtrl(i)
			&& sec->ceilingpic != skyflatnum)
		{
			osec = NULL;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->specType == R3D_16)
				{
					sector_t *chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!chk) { osec = NULL; break; }

					if (!osec) osec = chk;
					else if (osec != chk) { osec = NULL; break; }

					linearray[c++] = ld - lines;
				}
			}

			if (osec && c > 2 && osec->ceilingheight >= sec->ceilingheight)
			{
				RDBOOL	repeat = true;

				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

				while (repeat)
				{
					int lnum = -1;

					repeat = false;

					for (k = 0; k < c; k++)
					{
						if (hitline[linearray[k]] == 1)
						{
							lnum = linearray[k]; break;
						}
					}

					if (   lnum >= 0
						&& subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL))
					{
						if (LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_TESTAREA))
						{
							byte type = osec->floorheight < sec->floorheight ? R3D_00 : R3D_14;

							for (k = 0; k < c; k++)
							{
								if (hitline[linearray[k]] == 4)
									Forcetype(&lines[linearray[k]], type, 43141);
							}
						}

						for (k = 0; k < c; k++)
						{
							if (hitline[linearray[k]] == 4)
								hitline[linearray[k]] = 8;
						}

						repeat = true;
					}
				}
			}
		}
	}

	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->linecount > 2
			&& (sec->specFlags & 0xFFFFC) == 0x98198
			&& !SecCtrl(i)
			&& sec->ceilingpic != skyflatnum)
		{
			for (c = k = 0, osec = NULL; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& !(lineflags[ld - lines])
					&& ld->frontsector != ld->backsector
					&& ld->lnflags2 == 0x0C000000
					&& !(ld->mtflags)
					&& origTEX[ld - lines] & 0x11
					&& !(origTEX[ld - lines] & 0x66))
				{
					sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!osec) osec = s; else if (osec != s) { c = 0; break; }

					linearray[c++] = ld - lines;
				}
				else
				{
					c = 0; break;
				}
			}

			if (c == sec->linecount)
			{
				mrksects[seccnt] = osec - sectors;
				countsec[seccnt++] = i;

				for (k = 0; k < c; k++)
					hitline[linearray[k]] = 1;
			}
		}
	}

	for (i = 0; i < seccnt; i++)
	{
		osec = &sectors[mrksects[i]];
		sec = &sectors[countsec[i]];

		if (   subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &rc, box, NULL, NULL)
			&& rc == sec->linecount
			&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
			&& !LS_LinesInside(hitline, sec)
			&& (secn = getBoundarySecnum(sec, sslist, &sslistcnt)) >= 0
			&& sectors[secn].ceilingheight != sec->ceilingheight
			&& !(sectors[secn].specFlags & (SF_MOVCEIL|SF_NOHITEX)))
		{
			int ss_set = 0;

			for (k = 0; k < sslistcnt; k++)
			{
				if (   subsectors[sslist[k]].sector - sectors != secn
					&& subsectors[sslist[k]].ceilsec != secn)
				{
					Setr3dsscs(&subsectors[sslist[k]], secn, 43142); ss_set++;
				}
			}
		}

		for (k = 0; k < sec->linecount; k++)
		{
			hitline[sec->lines[k] - lines] = 8;
		}
	}

	M_Free(mrksects);
}

// v220-9
static RDBOOL floorCHK(sector_t *sec)
{
	return (sec->specFlags & 0x3001C) == 0x10014;
}

// v220-7/v220-8/v220-9
void PS_CheckSameSidedLineRings()
{
	sector_t	*sec, *chk, *isec, *osec;
	ssline_t	*sl;
	line_t		*ld;
	int			c, i, k, m, n;
	int			lnc, lnum, b1[4], b2[4], rc;

	// v220-8 Look for move floor inside missing lotex nested rings of lines.
	//        These don't have false floors (exception to usual behaviour)
	for (sec = sectors, i = 0; i < numsectors; i++, sec++)
	{
		if (!floorCHK(sec)) continue;

		for (lnc = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if ( !(ld->ml_flags & ML_TWOSIDED)
				|| ld->spflags & SPF_CONTROLLINE)
			{
				continue;
			}

			if (   ld->mtflags & 0x10
				&& ld->frontsector == sec
				&& ld->backsector->floorheight > sec->floorheight)
			{
				linearray[lnc++] = ld - lines;
			}
			else
			if (   ld->mtflags & 0x40
				&& ld->backsector == sec
				&& ld->frontsector->floorheight > sec->floorheight)
			{
				linearray[lnc++] = ld - lines;
			}
			else
			if (   sec->tag
				&& ld->frontsector->tag == ld->backsector->tag
				&& floorCHK(ld->frontsector)
				&& floorCHK(ld->backsector ))
			{
				linearray[lnc++] = ld - lines;
			}
		}

		isec = osec = NULL;
		
		if (lnc > 5)
		{
			// first check the bounding sectors - they must all be the same height
			for (k = 0; k < lnc; k++)
			{
				ld = &lines[linearray[k]];

				chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (chk->specFlags & SF_MOVFLOOR)
				{
					// v220-9 check for inner floors moving in sync
					if (   chk->tag
						&& chk->tag == sec->tag
						&& floorCHK(chk))
					{
						isec = chk;
					}
					else
					{
						osec = NULL;
						break;
					}
				}

				if (chk != isec)
				{
					if (!osec)
						osec = chk;
					else if (osec != chk && osec->floorheight != chk->floorheight)
					{
						osec = NULL; break;
					}
				}
			}
		}

		if (osec)	// if set then height floor will be raised to is same for all missing tex rings
		{
			int res, *ring = (int*)malloc(sec->linecount * sizeof(int));

			if ( (res = largestRingNestedSector(sec, lnc, ring, &c, b1)) > 1 && c > 2)
			{
				CLR_hitline();

				for (k = 0; k < c; k++) hitline[ring[k]] = 4;

				ld = &lines[ring[0]];

				if (LS_MarkSubsecsInAreaGeneral(sec, hitline, ring[0], NULL, b1, -1, MSIA_NONE))
				{
					// Use the floor of the outer ring's containing sector
					osec = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					validcount++;

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t	*ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						if (ss->sector->specFlags & SF_NOLOTEX)
						{
							seg_t *sg = &segs[ss->firstline];

							Clear_SSEC(ss);
							ss->lfsec = osec - sectors;

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( (ld = sg->linedef) && sg->linedef->validcount != validcount)
								{
									ld->validcount = validcount;

									if (ld->ml_flags & ML_TWOSIDED)
									{
										if (ld->mtflags & 0x50)
										{
											ld->specType = R3D_31;
											ld->lnflags2 &= ~LNF2_NORENDLOW;
										}
									}
								}
							}
						}
					}
				}
			}

			M_Free(ring);
		}
	}

	// v220-7
	for (sec = sectors, i = 0; i < numsectors; i++, sec++)
	{
		RDBOOL	repeat = true;
		int		check16 = 0;

		if (sec->ceilingpic == skyflatnum) continue;	// v220-8

		while (repeat)
		{
			repeat = false;

			for (lnc = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->specType == R3D_16) check16++;

				if (   ld->frontsector == ld->backsector
					&& Vtxs2(ld)
					&& !lineflags[ld - lines])
				{
					linearray[lnc++] = ld - lines;
				}
			}

			if (lnc < 2) break;

			CLR_hitline();

			for (k = 0; k < lnc; k++)
				hitline[linearray[k]] = 1;

			lnum = linearray[0];

			if (subfn_CheckForRing(hitline, sec, lnum, &rc, b1, NULL, NULL))
			{
				int cs;

				for (lnc = k = 0; k < sec->linecount; k++)
				{
					lnum = sec->lines[k] - lines;

					if (hitline[lnum] == 4)
					{
						linearray[lnc++] = lnum;
						SetLFG(&lines[lnum], LFG_NEWSET, 43143);
					}
				}

				// v220-8 modified to explicitly look for an outer ring
				if (   nf_boxAreaSameSec(sec, &cs, b1, rc)
					&& cs >= 0
					&& !((sectors[cs].specFlags & 0xC) == 0xC // v220-9 added
					&&   sectors[cs].specFlags & SF_MOVSECTOR))
				{
					// Check the outer subsectors for a ring of
					// same sector lines in a different sector

					validcount++;

					for (c = k = 0; k < sec->linecount; k++)
					{
						lnum = sec->lines[k] - lines;

						if (hitline[lnum] == 4)
						{
							sl = &sslines[lnum];

							for (m = 0; m < sl->num; m++)
							{
								subsector_t *ss = &subsectors[sl->list[m]];
								seg_t		*sg = &segs[ss->firstline];
								int			n;
								RDBOOL		found = false;

								for (n = 0; n < hitssecptr; n++)
								{
									if (hitssec[n] == ss - subsectors) { found = true; break; }
								}

								if (found) continue;

								hitssec[hitssecptr++] = ss - subsectors;

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if (sg->linedef
										&& sg->linedef->validcount != validcount)
									{
										sg->linedef->validcount = validcount;

										if (   sg->linedef->frontsector - sectors == cs
											&& sg->linedef->backsector == sg->linedef->frontsector
											&& Vtxs2(sg->linedef))
										{
											linearray[c++] = sg->linedef - lines;
										}
									}
								}
							}
						}
					}

					if (c > 2)
					{
						CLR_hitline();

						for (k = 0; k < c; k++)
						{
							lnum = linearray[k];
							hitline[lnum] = 1;
						}

						if (   subfn_CheckForRing(hitline, &sectors[cs], lnum, &rc, b2, NULL, NULL)
							&& M_CompareBox(b1, b2))
						{
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];

								if (ss->floorsec < 0) Setr3dssfs(ss, cs, 43144);
								if (ss->ceilsec  < 0) Setr3dsscs(ss, cs, 43144);
							}
						}
					}
				}

				repeat = true;
			}
		}

		if (check16 < 3 || (sec->specFlags & 0x3004C) != 0xC) continue;

		CLR_hitline();

		// v220-7 recheck R3D_16 types as degenerate ssecs can cause problems
		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			if (!chk || chk->specFlags & SF_MOVSECTOR) continue;

			if (   (ld->specType == R3D_16 && sec->floorheight < chk->floorheight)
				|| (sec->floorheight == chk->floorheight && sec->ceilingheight == chk->ceilingheight))
			{
				lnum = ld - lines; hitline[lnum] = true; c++;
			}
		}

		if (   c == sec->linecount
			&& subfn_CheckForRing(hitline, sec, lnum, &rc, b1, NULL, NULL)
			&& rc == c
			&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, b1, -1, MSIA_NONE))
		{
			int		fs = -1, cs = -1;
			RDBOOL	valid = true;

			for (k = 0; k < hitssecptr; k++)
			{
				subsector_t *ss = &subsectors[hitssec[k]];

				if ((ss->sector->specFlags & 0x3004C) != 0xC) { valid = false; break; }

				if (ss->floorsec >= 0) if (fs < 0) fs = ss->floorsec; else if (fs != ss->floorsec) { valid = false; break; }
				if (ss->ceilsec  >= 0) if (cs < 0) cs = ss->ceilsec ; else if (cs != ss->ceilsec ) { valid = false; break; }
			}

			if (valid)
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					if (fs >= 0 && ss->floorsec < 0) Setr3dssfs(ss, fs, 43146);
					if (cs >= 0 && ss->ceilsec  < 0) Setr3dsscs(ss, cs, 43146);
				}

				if (fs >= 0 && cs >= 0)
				{
					for (k = 0; k < sec->linecount; k++)
					{
						line_t *ld = sec->lines[k];

						if (ld->specType == R3D_16 && !origTEX[ld - lines])
							Settype(ld, R3D_3F, 43147);
					}
				}
			}
		}
	}

	// v220-8 certain ssec types need unsetting
	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].stptr >= 0) continue;

		if (   (sectors[i].specFlags & 0xFFFFFC) == 0x0080001C
			&& (ld = sectors[i].lines[0])->specType == R3D_01
			&& (osec = (ld->frontsector - sectors == i ? ld->backsector : ld->frontsector)) )
		{
			if (   osec->ceilingheight < sectors[i].ceilingheight
				&& osec->specFlags & SF_MOVCEIL)
			{
				sectors[i].R3Dceilheight  = DDMININT;
				sectors[i].R3Dfloorheight = DDMAXINT;
				sectors[i].R3Dceilsec = -1;
				sectors[i].R3Dfloorsec = -1;
				
				osec->R3Dceilheight  = DDMININT;
				osec->R3Dfloorheight = DDMAXINT;
				osec->R3Dceilsec = -1;
				osec->R3Dfloorsec = -1;
			
				for (k = 0; k < sectors[i].linecount; k++)
				{
					sl = &sslines[sectors[i].lines[k] - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						ss->sector->R3Dceilheight  = DDMININT;
						ss->sector->R3Dfloorheight = DDMAXINT;
						ss->sector->R3Dceilsec = -1;
						ss->sector->R3Dfloorsec = -1;
						Setr3dsscs(ss, -1, 43148);
						Setr3dssfs(ss, -1, 43148);
					}
				}
			}
			// check for lines in wrong sector
			else if (SecUnset(&sectors[i]))
			{
				for (chk = NULL, k = 0; k < sectors[i].linecount; k++)
				{
					sl = &sslines[sectors[i].lines[k] - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector - sectors != i)
						{
							if (ss->sector == osec) { chk = NULL; break; }

							if (!chk) chk = ss->sector;
							else if (chk != ss->sector) { chk = NULL; break; };
						}
					}
				}

				if (chk && !(chk->specFlags & SF_MOVSECTOR))
				{
					pr_OFinsert("MAP ERROR FIXED: S%i lines other sec use S%i but should use S%i\n", i, osec - sectors, chk - sectors);
					ma_inf.maperror++;

					for (k = 0; k < sectors[i].linecount; k++)
					{
						sl = &sslines[sectors[i].lines[k] - lines];

						Settype(sectors[i].lines[k], R3D_3F, 43150);

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector != chk && ss->floorsec < 0)
								SetOverride(ss, chk - sectors, 43151);
						}
					}
				}
			}
		}

		// v220-8 check inheritance from internal line loop
		if (   sectors[i].linecount == 3
			&& (sectors[i].specFlags & 0xF000C) == 0xC)
		{
			CLR_hitline();

			for (k = 0; k < sectors[i].linecount; k++)
			{
				ld = sectors[i].lines[k];

				if (   ld->frontsector != ld->backsector
					|| lineflags[ld - lines] & LFG_3DTYPE)
				{
					goto end_loop;
				}

				lnum = sectors[i].lines[k] - lines;
				hitline[lnum] = 1;
			}
			
			if (  !subfn_CheckForRing(hitline, &sectors[i], lnum, &rc, b1, NULL, NULL)
				|| rc != 3)
			{
				continue;
			}

			validcount++;
			ss_validcount++;
			hitssecptr = 0;
			osec = NULL;

			for (c = k = 0; k < sectors[i].linecount; k++)
			{
				sl = &sslines[sectors[i].lines[k] - lines];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					if (ss->ssvalidcount != ss_validcount)
					{
						seg_t *sg = &segs[ss->firstline];

						ss->ssvalidcount = ss_validcount;

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if ( (ld = sg->linedef) && ld->validcount != validcount)
							{
								ld->validcount = validcount;

								if (lineflags[ld - lines] & LFG_3DTYPE)
									goto end_loop;

								if (   ld->frontsector == ld->backsector
									|| origTEX[ld - lines])
								{
									if (ld->frontsector - sectors != i)
									{
										if (c > 1) goto end_loop;	// Just two uncompliant lines are allowed
										c++;
									}
								}
								else if ( !(ld->ml_flags & ML_TWOSIDED) )
								{
									goto end_loop;
								}
								else
								{
									if (!osec)
									{
										osec = sg->frontsector;
										lnum = ld - lines;
									}
									else if (osec != sg->frontsector)
										goto end_loop;
								}
							}
						}
					}
				}
			}

			// I'm guessing that this is what the visplane would require
			// with the software renderer for compliance (but who knows?)
			// In those maps known where this is needed it is the case.
			if (  !osec
				|| osec->linecount > 6	// arbitrary, usually 4
				|| osec->floorheight >= sectors[i].floorheight
				|| osec->ceilingheight != sectors[i].ceilingheight
				|| osec->floorpic != sectors[i].floorpic
				|| osec->ceilingpic != sectors[i].ceilingpic)
			{
				goto end_loop;
			}

			CLR_hitline();

			for (k = 0; k < osec->linecount; k++)
			{
				ld = osec->lines[k];

				if (   ld->frontsector == ld->backsector
					|| !(ld->ml_flags & ML_TWOSIDED) )
				{
					goto end_loop;
				}

				hitline[ld - lines] = 1;
			}

			if (   subfn_CheckForRing(hitline, osec, lnum, &rc, b2, NULL, NULL)
				&& ((c < 3 && rc > 3) || (c < 2 && rc > 2))
				&& M_CompareBox(b1, b2)
				&& LS_MarkSubsecsInAreaGeneral(osec, hitline, 0, NULL, b2, -1, MSIA_NONE))
			{
				validcount++;

				// Double check we've got compliant subsectors. Unfortunately we cannot
				// determine this a priori because of the way subsectors are built. The 3
				// inner lines may not reference all of the outer subsectors required.
				for (c = k = 0; k <hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t       *sg = &segs[ss->firstline];
					
					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef) && ld->validcount != validcount)
						{
							ld->validcount = validcount;

							if (ld->frontsector == ld->backsector)
							{
								if (ld->frontsector - sectors != i)
								{
									if (c) goto end_loop;
									c++;
								}
							}
							else
							{
								if (!osec)
								{
									osec = sg->frontsector;
									lnum = ld - lines;
								}
								else if (osec != sg->frontsector)
									goto end_loop;
							}
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];

					ChangeSSECsec(ss, &sectors[i], 43150);
					// If the floorsec is already set then don't change it.
					// If set then this will be the correct value. Using
					// the outer sector cannot always be relied on.
					if (ss->floorsec < 0) SetOverride(ss, osec - sectors, 43152);
				}
			}

			goto end_loop;
		}

		// v220-8 look for R3D_16 loops inside another R3D_16 loop. The inner
		//        loop will cause unwanted textures to be drawn unless dealt with
		if ( (sectors[i].specFlags & 0xF000C) == 0x90004)
		{
			CLR_hitline();

			osec = NULL;

			for (c = k = 0; k < sectors[i].linecount; k++)
			{
				ld = sectors[i].lines[k];

				if (ld->specType == R3D_16 && ld->backsector)
				{
					chk = (ld->frontsector - sectors == i) ? ld->backsector : ld->frontsector;

					if (!osec)
						osec = chk;
					else if (osec != chk)
						goto end_loop;

					lnum = ld - lines;
					hitline[lnum] = 1;
					c++;
				}
				else
					goto end_loop;
			}

			if (   sectors[i].linecount == c
				&& (osec->specFlags & 0xF000C) == SF_NOLOTEX
				&& subfn_CheckForRing(hitline, &sectors[i], lnum, &rc, b1, NULL, NULL)
				&& rc == c)
			{
				sector_t *refsec = NULL;

				M_ClearBox(b2);

				for (k = 0; k < osec->linecount; k++)
				{
					ld = osec->lines[k];

					if (ld->specType != R3D_16)
						goto end_loop;

					chk = NULL;

					if (!(ld->frontsector->specFlags & SF_NOLOTEX) )
						chk = ld->frontsector;
					else if (ld->backsector && !(ld->backsector->specFlags & SF_NOLOTEX) )
						chk = ld->backsector;

					if (chk)
					{
						if (!refsec)
							refsec = chk;
						else if (refsec->floorheight != chk->floorheight
							|| ((refsec->specFlags & 0xF000C) != (chk->specFlags & 0xF000C)) )
							goto end_loop;
					}

					M_AddLineToBox(b2, ld);
				}

				if (refsec && M_FitBox(b1, b2))
				{
					for (k = 0; k < osec->linecount; k++)
					{
						ld = osec->lines[k];
						ld->ml_flags |= ML_IGNOREMISSINGLO;
						sl = &sslines[ld - lines];

						for (m = 0; m < sl->num; m++)
						{
							if (subsectors[sl->list[m]].sector != refsec)
								Setr3dssfs(&subsectors[sl->list[m]], refsec - sectors, 43153);
						}
					}
				}
			}
		}
end_loop:;
	}
}

// v220-9/v220-10
void PS_Check3LineMissingLoTexSecs()
{
	sector_t	*sec, *osec;
	int			c, d, i, k, rc, box[4];
	int			mask;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   secdone[i]
			|| sec->stptr >= 0
			|| sec->heightsec >= 0)
		{
			continue;
		}

		mask = sec->specFlags & 0xFFFFC;

		// v220-10 check for possible missing lo tex when sector's
		//         floor changes height.
		if (!SecCtrl(i) && (mask == 0x98790 || mask == 0x98390))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->specType == R3D_03
					&& !lineflags[ld - lines]
					&& ld->frontsector != ld->backsector
					&& ld->ml_flags & ML_TWOSIDED)
				{
					byte txn = origTEX[ld - lines] & 0x11;

					if (   ((txn == 0x01 && ld->backsector->specFlags  & SF_MOVFLOOR)
						||  (txn == 0x10 && ld->frontsector->specFlags & SF_MOVFLOOR)) )
					{
						Settype(ld, R3D_36, 43154);
					}
				}
			}
		}
		else if (mask == 0x809C)
		{
			CLR_hitline();

			for (c = d = k = 0, osec = NULL; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   !lineflags[ld - lines]
					&& ld->specType == R3D_00
					&& !(origTEX[ld - lines] && 0x11)
					&& (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2))
				{
					sector_t *chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!osec) osec = chk; else if (osec != chk) break;
					c++;
				}

				hitline[ld - lines] = 1;
			}

			if (   osec
				&& c == sec->linecount - 1
				&& osec->floorheight > sec->floorheight
				&& !(osec->specFlags & SF_MOVSECTOR)
				&& subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &rc, box, NULL, NULL)
				&& rc == sec->linecount
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(NULL, hitline))
			{
				for (k = 0; k < hitssecptr; k++)
					Setr3dssfs(&subsectors[hitssec[i]], osec - sectors, 43155);

				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];
					SetLFG(ld, LFG_NEWSET, 43156);
				}

				Setsecdone(sec, 43156);
			}
		}
		// v220-10 check unclosed sectors
		else if (!SecCtrl(i) && (mask & 0xFFFEC) == 0x8384)
		{
			sector_t	*chk;
			int			fht = DDMAXINT;

			CLR_hitline();

			for (c = d = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   lineflags[ld - lines]
					|| ld->spflags & SPF_CONTROLLINE
					|| origTEX[ld - lines] & 0x22
					|| !(ld->ml_flags & ML_TWOSIDED))
				{
					goto end_loop;
				}

				if (ld->specType == R3D_00)
				{
					if ((ld->lnflags2 & 0xFFFFFFF7) != 0x54000000) goto end_loop;

					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;
					if (chk->floorheight < fht) fht = chk->floorheight;
					c++;
				}
				else if (ld->specType == R3D_01)
				{
					if (ld->frontsector != ld->backsector) goto end_loop;
					d++;
				}
				else
				{
					goto end_loop;
				}

				hitline[ld - lines] = 1;
			}

			if (   c
				&& d
				&& subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &rc, box, NULL, NULL)
				&& rc == sec->linecount)
			{
				ss_validcount++;

				for (d = k = 0, osec = NULL; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (ld->specType == R3D_01)
					{
						ssline_t *sl = &sslines[ld - lines];

						for (c = 0; c < sl->num; c++)
						{
							subsector_t *ss = &subsectors[sl->list[c]];

							if (ss->ssvalidcount != ss_validcount)
							{
								ss->ssvalidcount = ss_validcount;

								if (ss->sector == sec)
								{
									hitssec[d++] = sl->list[c];
								}
								else if (!osec)
								{
									osec = ss->sector;
								}
								else if (osec != ss->sector)
								{
									goto end_loop;
								}
							}
						}
					}
				}

				if (   d
					&& osec
					&& !(osec->specFlags & SF_MOVSECTOR)
					&& fht == sec->floorheight
					&& osec->floorheight > sec->floorheight)
				{
					for (k = 0; k < d; k++)
						Setr3dssfs(&subsectors[hitssec[k]], osec - sectors, 43157);

					for (k = 0; k < sec->linecount; k++)
					{
						line_t *ld = sec->lines[k];

						if (ld->frontsector == ld->backsector)
						{
							Settype(ld, R3D_3F, 43157);
							SetLFG(ld, LFG_NEWSET, 43157);
						}
					}

					Setsecdone(sec, 43157);
				}
			}
		}
end_loop:;
	}
}

// v220-9, v220-19 revised
void PS_CheckSingleLineSsecs()
{
	subsector_t	*ssec, *ss;
	sector_t	*fs, *bs;
	line_t		*ld;
	int			flf, flb;
	int			i, k, os;

	for (i = 0, ssec = subsectors; i < numsubsectors; i++, ssec++)
	{
		if (   ssec->linecount != 1
			|| ssec->floorsec >= 0
			|| !(ld = segs[ssec->firstline].linedef))
		{
			continue;
		}

		if ((ssec->sector->specFlags & 0xC) == 0xC)
		{
			if (ld->frontsector == ld->backsector && lxarray[ld - lines] == 0x8)
			{
				if (SSECunset(ssec))
				{
 					ssline_t	*sl = &sslines[ld - lines];

					for (k = 0, os = -1; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (ss->floorsec >= 0 && ss->floorsec == ss->ceilsec)
						{
							if (os < 0) os = ss->floorsec; else if (os != ss->floorsec) { os = -1; break; }
						}
					}

					if (os >= 0) SetOverride(ssec, os, 43176);
				}
			}
			else if (ld->specType == R3D_00
				&& (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2)
				&& !lxarray[ld - lines]
				&& !origTEX[ld - lines]
				&& (fs = ld->frontsector) && (bs = ld->backsector)
				&& (flf = fs->specFlags) && (flb = bs->specFlags)
				&& !(flf & SF_MOVSECTOR) && !(flb & SF_MOVSECTOR)
				&& SSECunset(ssec))
			{
				// v220-31 removed
			/*	if (fs == ssec->sector) fs = bs;

				if (fs->specFlags & 0x400 || !(fs->specFlags & 0xC))
				{
					ssline_t *sl = &sslines[ld - lines];

					for (k = 0; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];
						if (ss->floorsec < 0) Setr3dssfs(ss, fs - sectors, 43177);
					}

					Forcetype(ld, R3D_3F, 43177);
				}*/
			}
		}
		// added v220-17
		else if ((ssec->sector->specFlags & 0x3001C) == 0x14)
		{
			if (   ld->specType == R3D_16
				&& (ld->mtflags == R3D_MISLTEX1 || ld->mtflags == R3D_MISLTEX2)
				&& !lxarray[ld - lines]
				&& !origTEX[ld - lines]
				&& (fs = ld->frontsector) && (bs = ld->backsector)
				&& (flf = fs->specFlags) && (flb = bs->specFlags)
				&& !(flf & SF_MOVSECTOR) && !(flb & SF_MOVSECTOR)
				&& (flf = (flf & 0x1C)) && (flb = (flb & 0x1C))
				&& ((flf == 0x1C && flb == 0x14) || (flb == 0x1C && flf == 0x14))
				&& SSECunset(ssec))
			{
				ssline_t *sl = &sslines[ld - lines];

				for (k = 0; k < sl->num; k++)
				{
					ss = &subsectors[sl->list[k]];

					// Look to inherit only
					if (ss->sector == ssec->sector && ss->floorsec >= 0 && ss->lfsec < 0)
					{
						Setr3dssfs(ssec, ss->floorsec, 43177);
						Forcetype(ld, R3D_3F, 43177);
						break;
					}
				}
			}
		}
	}
}

// =====================================================

static void checkLinesUnset(sector_t *s)
{
	sector_t	*ch, *fs, *bs;
	int			c, i, lnum, rc, box[4];

	CLR_hitline();

	for (c = i = 0; i < s->linecount; i++)
	{
		lnum = s->lines[i] - lines;
		if (lineflags[lnum]) return;
		if (origTEX[lnum]) return;
		if (!(s->lines[i]->ml_flags & ML_TWOSIDED)) return;
		if (s->lines[i]->spflags & SPF_CONTROLLINE) return;
		if ((fs = s->lines[i]->frontsector) == (bs = s->lines[i]->backsector)) continue;
		ch = fs == s ? bs : fs;
		if (ch->tag != s->tag) return;
		if (ch->floorheight != s->floorheight) return;
		if (ch->specFlags & 0xC && !(ch->specFlags & 0x400)) return;
		hitline[lnum] = 1;
		linearray[c++] = lnum;
	}

	while(1)
	{
		lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1) { lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (   subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
			&& rc > 3
			&& LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE))
		{
			RDBOOL valid = true;

			for (i = 0; i < hitssecptr; i++)
			{
				if (subsectors[hitssec[i]].sector != s) { valid = false; break; }
			}

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 0;

					if (valid)
					{
						Settype(&lines[linearray[i]], R3D_16, 43190);
						SetLFG(&lines[linearray[i]], LFG_NEWSET|LFG_FIXED, 43190);
					}
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-16 1) Look for textureless floormov sector bounded by another floormov sec with same tag
// v220-13 2) Try an establish if there is a preferred floor texture
void PS_CheckLotexPreference()
{
	sector_t	*sc, *ch, *os;
	line_t		*ld;
	int			c, i, k, lnum, rc;

	//----------------------------------

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& (sc->specFlags & 0xFFFFC) == 0x9001C
			&& !SecCtrl(i)
			&& sc->heightsec < 0)
		{
			checkLinesUnset(sc);
		}
		else
		if ( (sc->specFlags & 0xFFFFFC) == 0x008394)
		{
			CLR_hitline();

			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (   (ld->specType == R3D_14       || ld->specType == R3D_16)
					&& (ld->mtflags  == R3D_MISLTEX1 || ld->mtflags  == R3D_MISLTEX2)
					&& (ch = ld->frontsector == sc ? ld->backsector : ld->frontsector) )
				{
					if (!os) os = ch;

					if (os->floorheight == ch->floorheight
						&& !(os->specFlags & SF_MOVFLOOR))
					{
						lnum = sc->lines[k] - lines;
						hitline[lnum] = 1;
						c++;
					}
				}
			}

			if (c < 3) continue;

			while(1)
			{
				lnum = -1;

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (hitline[ld - lines] == 1)
					{
						lnum = ld - lines;
						break;
					}
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
				{
					line_t	*ref[2];

					// Only interested in 3 line rings
					if (   rc == 3
						&& (lines[lnum].mtflags == R3D_MISLTEX1 || lines[lnum].mtflags == R3D_MISLTEX2))
					{
						int		ct[2] = {  0,  0};
						int		ft[2] = { -1, -1};
						int		pic;

						for (k = 0; k < sc->linecount; k++)
						{
							line_t *ld = sc->lines[k];

							if (hitline[ld - lines] == 4)
							{
								pic   = ld->mtflags == R3D_MISLTEX1 ? ld->backsector->floorpic :
										ld->mtflags == R3D_MISLTEX2 ? ld->frontsector->floorpic : -1;

								if (pic < 0)
								{
									ct[0] = 0; 	break;		// give up - line not valid
								}

								if (ft[0] < 0)
								{
									ft[0] = pic; ct[0]++; ref[0] = ld;

								}
								else if (ft[0] == pic)
								{
									ct[0]++;
								}
								else if (ft[1] < 0)
								{
									ft[1] = pic; ct[1]++; ref[1] = ld;
								}
								else if (ft[1] == pic)
								{
									ct[1]++;
								}
								else
								{
									ct[0] = 0; break;		// give up - multiple floor textures
								}
							}
						}

						if (   ct[0] && ct[1]				// both set so must be two textures
							&& (ct[0] == 1 || ct[1] == 1)	// only one line can be reset
							&& ct[0] + ct[1] == rc)			// all lines must be an R3D_14 or R3D_16 type
						{
							ld = ct[0] == 1 ? ref[0] : ref[1];

							Forcetype(ld, R3D_00, 43158);
						}
					}

					for (k = 0; k < sc->linecount; k++)
					{
						line_t *ld = sc->lines[k];

						if (hitline[ld - lines] == 4)
							hitline[ld - lines] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}
		}
	}
}

// ===================================================

static void subfn_CheckSameSecMovSectors(sector_t *s)
{
	subsector_t *ss;
	sector_t	*os;
	sector_t	*use;
	line_t		*ld;
	seg_t		*sg;
	int			c, d, i, k, m;

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   ld->frontsector != ld->backsector
			|| lineflags[ld - lines])
		{
			return;
		}

		linearray[c++] = ld - lines;
	}

	if (c < 4) return;

	for (i = 0; i < c; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];

		for (k = 0, os = NULL; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if ( (ld = sg->linedef)
					&& sg->frontsector != s)
				{
					if (!os)
					{
						os = sg->frontsector;
						if (os->specFlags & SF_MOVSECTOR != SF_MOVFLOOR) return;
					}
					else if (os != sg->frontsector) return;
				}
			}
		}
	}

	if (!os || os->floorheight >= s->floorheight) return;


	// If we got here then a moving floor is contained in a secondary moving floor
	// Check for adoptive sector and that lines have no appropriate lower textures.
	for (d = i = 0, use = NULL; i < os->linecount; i++)
	{
		ld = os->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			sector_t *chk = ld->frontsector == os ? ld->backsector : ld->frontsector;
			byte	 txn = ld->frontsector == os ? 0x10 : 0x01;

			if (chk == os) continue;

			if (os->floorheight != chk->floorheight) return;

			if (chk->specFlags & SF_MOVFLOOR) return;

			linearray[c++] = ld - lines;

			if (!origTEX[ld - lines] & txn) d++;

			if (!use) use = chk;
		}
	}

	if (!use || !d) return;

	ss_validcount++;

	for (i = 0; i < c; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				if (ss->sector == s || ss->sector == os)
				{
					Setr3dssfs(ss, use - sectors, 43158);
				}
			}
		}
	}

	for (i = 0; i < c; i++)
		SetLFG(&lines[linearray[i]], LFG_NEWSET, 43158);
}

// b211-0 Check for any unset sectors with no lotex
void PS_CheckSameSecMovSectors()
{
	int			i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& (sectors[i].specFlags & 0xFFFFC) == 0x9003C
			&& sectors[i].stptr < 0)
		{
			subfn_CheckSameSecMovSectors(&sectors[i]);
		}
	}
}

// =====================================================================

static void subfn_CheckL1secMissingHi(sector_t *sc, int *lna)
{
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			a, b, c, i;
	int			lnum;
	int			side;

	// ----------------------

	for (a = b = c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i]; lnum = ld - lines;

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   lineflags[lnum]
			|| (ld->specType != R3D_00 && ld->specType != R3D_03)
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| !(ld->lnflags2 & LNF2_OWNBOTHGTRONE) )
		{
			return;
		}

		if ( !(ld->ml_flags & ML_TWOSIDED) )
		{
			if (ld->lnflags2 & LNF2_OWNEQUALTOTWO)
			{
				lna[c++] = ld - lines; a++;
			}
			continue;
		}

		chk = fs == sc ? bs : fs;

		if (chk->specFlags & SF_NOHITEX) return;	// v220-16 test added

		side = ld->frontsector == sc ? 0 : 1;

		if (   (!side && !(origTEX[lnum] & 0x04) && fs->ceilingheight - bs->ceilingheight > 8*FRACUNIT)
			|| ( side && !(origTEX[lnum] & 0x40) && bs->ceilingheight - fs->ceilingheight > 8*FRACUNIT) )
		{
			b++;
		}
		else if (origTEX[lnum] & 0x44)
		{
			 continue;
		}
		
		lna[c++] = ld - lines;
	}

	if (!a || !b || c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[lna[i]] = 1;

	while(1)
	{
		RDBOOL	valid = false;
		int		k, m, rc;

		for (i = 0, lnum = -1; i < c; i++)
		{
			if (hitline[lna[i]] == 1) { lnum = lna[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			sector_t *os = NULL;
			int		 lc;

			for (lc = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (hitline[ld - lines] == 4)
				{
					linearray[lc++] = ld - lines;

					if (os != sc && ld->ml_flags & ML_TWOSIDED)
					{
						for (m = 0; m < c; m++)
						{
							if (lna[m] == ld - lines)
							{
								chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

								if (chk->ceilingheight == sc->ceilingheight) break;	// shared line

								if (!os) os = chk;
								else if (os->ceilingheight != chk->ceilingheight)
								{
									os = sc; break;
								}
							}
						}
					}
				}
			}

			if (os && os != sc)
			{
				ssline_t *sl;

				valid = true;

				ss_validcount++;
				hitssecptr = 0;

				for (k = 0; k < lc; k++)
				{
					lnum = linearray[k];

					sl = &sslines[lnum];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector != sc) continue;

						if (ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;

							hitssec[hitssecptr++] = sl->list[m];
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
					Setr3dsscs(&subsectors[hitssec[k]], os - sectors, 43159);

			}

			for (k = 0; k < lc; k++)
			{
				lnum = linearray[k];
				hitline[lnum] = 0;

				if (valid)
				{
					ld = &lines[lnum];

					if ( !(ld->ml_flags & ML_TWOSIDED)
						|| ld->frontsector->ceilingheight != ld->backsector->ceilingheight)	// don't mark shared lines
					{
						SetLFG(ld, LFG_NEWSET|LFG_FIXED, 43159);
					}
				}
			}

			if (valid && !secdone[sc - sectors]) Setsecdone(sc, 43159);
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

static void subfn_CheckPartialMissingHi(sector_t *sc, int *lna)
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			a, b, c, i;
	int			lnum;
	int			side;

	// ----------------------

	for (a = b = c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i]; lnum = ld - lines;

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[lnum]
			|| ld->specType != R3D_00
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| !(ld->lnflags2 & LNF2_OWNBOTHGTRTWO) )
		{
			return;
		}

		a++;

		side = ld->frontsector == sc ? 0 : 1;

		if (   (!side && !(origTEX[lnum] & 0x04) && fs->ceilingheight - bs->ceilingheight > 8*FRACUNIT)
			|| ( side && !(origTEX[lnum] & 0x40) && bs->ceilingheight - fs->ceilingheight > 8*FRACUNIT) )
		{
			lna[c++] = ld - lines;
		}
		else if (origTEX[lnum] & 0x44)
		{
			b++;
		}
	}

	if (!c || !b || a < 3) return;

	CLR_hitline();

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		hitline[ld - lines] = 1;
	}

	while(1)
	{
		RDBOOL	valid = false;
		int		k, m, rc;

		for (i = 0, lnum = -1; i < c; i++)
		{
			if (hitline[lna[i]] == 1) { lnum = lna[i]; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL))
		{
			sector_t *os = NULL;
			int		 lc;

			for (lc = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (hitline[ld - lines] == 4)
				{
					linearray[lc++] = ld - lines;

					if (os != sc)
					{
						for (m = 0; m < c; m++)
						{
							if (lna[m] == ld - lines)
							{
								sector_t *chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

								if (chk->heightsec >= 0) { os = NULL; break; }	// v220-33 add test

								if (!os) os = chk;
								else if (os->ceilingheight != chk->ceilingheight)
								{
									os = NULL; break;
								}
							}
						}
					}
				}
			}

			if (os && os != sc)
			{
				ssline_t *sl;

				valid = true;

				ss_validcount++;
				hitssecptr = 0;

				for (k = 0; k < lc; k++)
				{
					lnum = linearray[k];

					sl = &sslines[lnum];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector != sc) continue;

						if (ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;

							hitssec[hitssecptr++] = sl->list[m];
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
					Setr3dsscs(&subsectors[hitssec[k]], os - sectors, 43185);
			}

			for (k = 0; k < lc; k++)
			{
				lnum = linearray[k];
				hitline[lnum] = 0;

				if (valid) SetLFG(&lines[lnum], LFG_NEWSET|LFG_FIXED, 43185);
			}

			if (valid && !secdone[sc - sectors]) Setsecdone(sc, 43185);
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-13 check if a closed sector has missing hi-tex lines
// together with lines with hi-textures.
void PS_CheckPartialMissingHi()
{
	int *lna = (int*)W_Calloc(numlines*sizeof(int), STATICID, 0);
	int c, i;

	SET_countsec(0);

	// 1) check for missing hi textures in sectors with a 1's line
	for (c = i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& (sectors[i].specFlags & 0xF0FFC) == 0x598
			&& sectors[i].heightsec < 0
			&& sectors[i].ceilingpic != skyflatnum
			&& sectors[i].stptr < 0)
		{
			countsec[c++] = i;
		}
	}

	if (c)
	{
		for (i = 0; i < c; i++)
		{
			subfn_CheckL1secMissingHi(&sectors[countsec[i]], lna);
		}
	}

	// 2) check for missing hi textures in sectors with no 1's line
	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& (sectors[i].specFlags & 0xFBFFC) == 0x8390
			&& sectors[i].heightsec < 0
			&& sectors[i].ceilingpic != skyflatnum
			&& sectors[i].stptr < 0)
		{
			subfn_CheckPartialMissingHi(&sectors[i], lna);
		}
	}

	W_Free(lna, STATICID, 0);
}

void subfn_CheckPartialMissingLo(sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*fs, *bs, *chk;
	line_t		*ld;
	int			c, i, box[4];;

	//-------------------------

	M_ClearBox(box);

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (!(ld->ml_flags & ML_TWOSIDED)) return;
		if (!(ld->lnflags2 & LNF2_OWNBOTHGTRONE)) return;
		if (lineflags[ld - lines]) return;
		if ((fs = ld->frontsector) == (bs = ld->backsector)) return;

		chk = fs == sc ? bs : fs;

		if (chk->ceilingpic != skyflatnum) return;
		if (chk->heightsec >= 0) return;
		if ((chk->specFlags & 0xF000C) != 0x8 ) return;
		if (chk->ceilingheight != sc->ceilingheight) return;
		if (chk->floorheight >= sc->floorheight) return;
		if (!(chk->specFlags & SF_NOHITEX)) return;
		if (chk->heightsec >= 0) return;

		if (ld->mtflags)
		{
			if      (ld->mtflags == R3D_MISLTEX1) c++;
			else if (ld->mtflags == R3D_MISLTEX2) c++;
			else return;
		}

		M_AddLineToBox(box, ld);
	}

	if (!c) return;

	// check all the subsectors in the box
	hitssecptr = 0;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (M_FitBox(ss->ssbox, box))
		{
			if ( (chk = ss->sector) == sc) continue;

			if (!ssUnset1(ss->sector)) return;
			if ((chk->specFlags & 0xC) != 0xC && !ssUnset2(ss)) return;
			if ((chk->specFlags & 0x80000C) == 0x80000C) continue;
			if (!(chk->specFlags & 0xC)) return;
			if (chk->floorheight >= sc->floorheight) return;

			hitssec[hitssecptr++] = i;
		}
	}

	if (!hitssecptr) return;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		seg_t *sg;
		int   k;

		ss = &subsectors[hitssec[i]];

		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef)
				&& ld->validcount != validcount)
			{
				if (!(ld->ml_flags & ML_TWOSIDED)) return;
				if (ld->spflags & SPF_CONTROLLINE) return;

				ld->validcount = validcount;

				if (ld->specType != R3D_00 && ld->specType != R3D_03) return;

				if (lineflags[ld - lines])
				{
					if (   ld->frontsector != ld->backsector
						&& (ld->frontsector->floorheight   != ld->backsector->floorheight
						||  ld->frontsector->ceilingheight != ld->backsector->ceilingheight) )
					{
						return;
					}
				}

				linearray[c++] = ld - lines;
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];

		if (ss->sector->floorheight < sc->floorheight)
		{
			Setr3dsslfsec(ss, ss->sector - sectors, 43180);
			Setr3dssfs(ss, sc - sectors, 43180);
		}
		else
			Setr3dsslfsec(ss, sc - sectors, 43180);
	}

	Setsecdone(sc, 43180);
}

// v220-13
// Check if sector with no-tex in a sky sector has partially missing lo-tex
// lines and encloses textureless sectors needing a high floor to be drawn
void PS_CheckPartialMissingLo()
{
	int i;

	if (R3D_maptype) return;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& sectors[i].ceilingpic == skyflatnum
			&& (sectors[i].specFlags & 0xFBFFC) == 0x8198
			&& ssUnset1(&sectors[i])
			&& sectors[i].heightsec < 0)
		{
			subfn_CheckPartialMissingLo(&sectors[i]);
		}
	}
}

// ===============================================================

static void subfn_CheckMissingHeightsec(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*ld;
	int			i, k, m, rc;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (lineflags[ld - lines]) return;
		if (!(ld->mtflags & R3D_MISTEXANY)) return;

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (!ch || ch->heightsec < 0) return;
	}

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		hitline[ld - lines] = 1;
	}

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) && rc > 2)
		{
			hitssecptr = 0;

			for (i = 0, os = NULL; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
					if (!os) os = ch;
					else if (ch != os
						&& (ch->floorheight != sc->floorheight
						|| ch->ceilingheight != sc->ceilingheight))
					{
						os = NULL; break;
					}
				}
			}

			if (os)
			{
				ssline_t	*sl;

				ss_validcount++;

				for (i = 0; i < sc->linecount; i++)
				{
					ld = sc->lines[i];

					if (hitline[ld - lines] == 4)
					{
						sl = &sslines[ld - lines];

						for (k = 0; k < sl->num; k++)
						{
							subsector_t *ss = &subsectors[sl->list[k]];

							if (ss->ssvalidcount == ss_validcount) continue;

							ss->ssvalidcount = ss_validcount;
							
							if (ss->sector == sc)
							{
								seg_t *sg = &segs[ss->firstline];

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (sg->linedef && hitline[sg->linedef - lines] != 4)
									{
										os = NULL; break;
									}
								}

								if (!os) break;

								hitssec[hitssecptr++] = sl->list[k];
							}
						}
					}
				}
			}

			if (os)
			{
				for (i = 0; i < hitssecptr; i++)
					SetOverride(&subsectors[hitssec[i]], os - sectors, 43178);
			}

			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					if (os) Settype(ld, R3D_3F, 43178);
					hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// added v220-16
void PS_CheckMissingHeightsec()
{
	sector_t	*sc;
	int			i;

	//-------------------------------

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (  !secdone[i]
			&& sc->heightsec < 0
			&& (sc->specFlags & 0x3001C) == 0x1C)
		{
			subfn_CheckMissingHeightsec(sc);
		}
	}
}

// ===============================================

static void subfn_SelfreferencingMovCeil(sector_t *sc)
{
	line_t	*ld;
	int		c, i;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (   ld->frontsector != ld->backsector
			|| lineflags[ld - lines])
		{
			return;
		}
		linearray[c++] = ld - lines;
	}
	
	subsector_t	*ss;
	sector_t	*os;
	ssline_t	*sl;
	seg_t		*sg;
	int			k, m;

	hitssecptr = 0;
	ss_validcount++;

	for (i = 0, os = NULL; i < c; i++)
	{
		sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;
			
			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (  (ld = sg->linedef)
					&& sg->frontsector != sc)
				{
					if (!os) os = sg->frontsector;
					else if (os != sg->frontsector) return;
				}
			}

			if (ss->sector == sc)
				hitssec[hitssecptr++] = sl->list[k];
		}
	}

	if (!os) return;

	for (i = 0; i < hitssecptr; i++)
		SetOverride(&subsectors[hitssec[i]], os - sectors, 43179);
}

// v220-18
void PS_SelfreferencingMovCeil()
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x3040C) == 0x2000C)
		{
			subfn_SelfreferencingMovCeil(&sectors[i]);
		}
	}
}

// ===============================================

static void subfn_UnsetPartialMissingTEXLoops(sector_t *sc)
{
	sector_t	*ch, *os = NULL;
	line_t		*ld;
	int			d = sc->linecount;
	int			lnum;
	int			b, c, i;
	
	for (b = c = i = 0; i < d; i++)
	{
		ld = sc->lines[i];
		lnum = ld - lines;

		if (lineflags[lnum])
		{
			if (ld->specType == R3D_16 && !origTEX[lnum])
			{
				if ( !(ld->mtflags == 0x40 || ld->mtflags == 0x10) ) return;
				b++;
			}
			else
			{
				return;
			}
		}
		else if (ld->specType == R3D_16 && !origTEX[lnum])	// test added v220-31
		{
			return;
		}

		ch = NULL;

		if (ld->mtflags)
		{
			if (origTEX[lnum]) return;
			if (ld->mtflags == 0x40) ch = ld->backsector;
			else
			if (ld->mtflags == 0x10) ch = ld->frontsector;
			else if (ld->mtflags) return;

			if (ch)
			{
				if (!os) { if (ch->floorpic == skyflatnum) return; else os = ch; }
				else if (os != ch
					&& (os->floorheight != ch->floorheight
					|| os->ceilingheight != ch->ceilingheight) )
				{
					return;
				}
			}

			c++;
		}
	}

	if (!os || !c || os == sc || b >= d - 1) return;

	int rc, box[4], *lmem;

	lmem = (int*)malloc(sc->linecount * sizeof(int));
	CLR_hitline();

	for (c = i = 0; i < d; i++)
	{
		ld = sc->lines[i];
		hitline[ld - lines] = 1;
		lmem[c++] = ld - lines;
	}

	while(1)
	{
		lnum = -1;

		for (i = 0; i < d; i++) { if (hitline[lmem[i]] == 1 && lines[lmem[i]].mtflags) { lnum = lmem[i]; break; } }

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			RDBOOL	valid = false;

			if (   LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(sc, hitline))
			{
				valid = true;

				for (i = 0; i < hitssecptr; i++)
				{
					Setr3dsslfsec(&subsectors[hitssec[i]], sc - sectors, 43184);
					Setr3dssfs(&subsectors[hitssec[i]], os - sectors, 43184);
				}
			}

			for (i = 0; i < d; i++)
			{
				if (hitline[lmem[i]] == 4)
				{
					byte txn = origTEX[lmem[i]];

					hitline[lmem[i]] = 0;

					if (valid)
					{
						ld = &lines[lmem[i]];

						if (!txn)
							Settype(ld, R3D_3F, 43184);
						else if (txn & 0x55)
							Settype(ld, R3D_36, 43184);
						else
							Settype(ld, R3D_03, 43184);

						SetLFG(ld, LFG_NEWSET|LFG_FIXABS, 43184);
					}
				}
			}

			if (valid)
			{
				Setsecdone(os, 43184);
				Setsecdone(sc, 43184);
			}

			if (rc == c) break;
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	free(lmem);
}

void PS_UnsetPartialMissingTEXLoops()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && !SecCtrl(i))
		{
			int f = sectors[i].specFlags & 0x3041C;
			
			if (f == 0x18 || f == 0x1C)
			{
				subfn_UnsetPartialMissingTEXLoops(&sectors[i]);
			}
		}
	}
}
